/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DirectApplication.h"
#include <common/InputConsole.h>
#include <common/output/OutputStream.h>
#include <common/base/Date.h>
#include <common/base/ServerProperties.h>
#include <dataprovider/DirectDataProvider.h>
#include <command/application/DirectCommandProcessor.h>
#include <core/DirectCore.h>
#include <iostream>
#include <fstream>
#include <signal.h>
#include <cstdlib>
#include <ics/IcsServerManager.h>
#include "Utilities/Simulated.h"
#include <cameo/api/cameo.h>

using namespace std;
using namespace common;
using namespace dataprovider;
using namespace command;
using namespace core;

const std::string DirectApplication::SIMULATION = "sim";
const std::string DirectApplication::VISA = "visa";
const std::string DirectApplication::SIMULATION3D = "sim3d";
const std::string DirectApplication::SIMULATION_NEUTRON = "simneutron";
const std::string DirectApplication::TEST_EXECUTION = "-test";
const std::string DirectApplication::TEST_ESTIMATED_TIME = "-estimatedtime";
const std::string DirectApplication::REPLACE_PROPERTY_NAMES = "-replace-property-names-with-alias";
const std::string DirectApplication::STATEFILENAME = ".nomadstate";
const std::string DirectApplication::INIT_STATE = "init";
const std::string DirectApplication::EXEC_STATE = "exec";
const std::string DirectApplication::EXIT_STATE = "exit";

DirectApplication::DirectApplication() :
	_console(0),
	_testExecution(false),
	_testEstimatedTimeExecution(false),
	_replacePropertyNamesWithAlias(false),
	_state(STARTING),
	_commandProcessorStarted(2),
	_terminated(2),
	_returnCode(EXIT_SUCCESS) {

	_dataprovider = new DirectDataProvider();
	_commandprocessor = new DirectCommandProcessor(this);
	_core = new DirectCore();
}

DirectApplication::~DirectApplication() {

	delete _core;
	delete _commandprocessor;
	delete _dataprovider;
	delete _console;
}

void DirectApplication::init(int argc, char* argv[]) {

	string configPath;
	bool remote = false;

	for (int i = 0; i < argc; i++) {
		string oldSimulation = string("-") + SIMULATION;

		if (string(argv[i]) == CONFIG_PATH && i+1 < argc) {
			i++;
			configPath = argv[i];
		} else if (string(argv[i]) == REMOTE) {
			remote = true;
		} else if (string(argv[i]) == SIMULATION || string(argv[i]) == oldSimulation) {
			Simulated::activated = true;
		} else if (string(argv[i]) == SIMULATION3D) {
			Simulated::for3d = true;
		} else if (string(argv[i]) == SIMULATION_NEUTRON) {
			Simulated::forneutron = true;
		} else if (string(argv[i]) == VISA) {
			Simulated::visa = true;
		} else if (string(argv[i]) == TEST_EXECUTION) {
			_testExecution = true;
		} else if (string(argv[i]) == TEST_ESTIMATED_TIME) {
			_testEstimatedTimeExecution = true;
		} else if (string(argv[i]) == REPLACE_PROPERTY_NAMES) {
			_replacePropertyNamesWithAlias = true;
		}
	}

	try {
		ServerProperties::initInstance(configPath);

	} catch (const ServerProperties::NoSuchDirectoryException& e) {
		err << e.printStack() << eol;
	}

	setFileState(INIT_STATE);

	if (!remote && !_testExecution) {
		_console = new InputConsole(this);
	}

	_dataprovider->init(argc, argv);
	_commandprocessor->init(argc, argv);
	_core->init(argc, argv);
}

void DirectApplication::setFileState(std::string const & state) {

	// writing state file
	fstream stateFile;

	string stateFileName = ServerProperties::getInstance()->getHomePath() + "/" + STATEFILENAME;

	stateFile.open(stateFileName.c_str(), ios::in | ios::out | ios::trunc);
	if (stateFile.is_open()) {
		stateFile << state;
		stateFile.close();
	} else {
		err << "problem writing state file" << eol;
	}
}

void DirectApplication::run() {

	Date d1;

	// running the 3 parts of the server
	_dataprovider->run();
	_commandprocessor->run();

	// rendezvous
	_commandProcessorStarted.wait();

	//sleep(1);

	//cout << "command processor started" << endl;
	// possible synchro error : initialisation of command processor sends events asynchronously,
	// before core is properly started for which subscriber can be not completely started!

	_core->run();

	_state = STARTED;

	setFileState(EXEC_STATE);

	cout << "Server started in " << (Date() - d1).getSeconds() << "s" << endl;

	cameo::This::setRunning();

	// if server started in test mode, quit immediately
	if (_testExecution) {
		// Only test and exit
		test();
		return;
	}
	// if server started in test mode, quit immediately
	if (_testEstimatedTimeExecution) {
		// Only test and exit
		testEstimatedTime();
		return;
	}

	// If the server is started to replace property names.
	if (_replacePropertyNamesWithAlias) {
		replacePropertyNamesWithAlias();
		return;
	}

	if (_console != 0) {
		_console->run();
	}

	// wait for the end of the application
	_terminated.wait();
}

void DirectApplication::reset() {

	if (_console != 0) {
		_console->reset();
	}

	_core->reset();
	_commandprocessor->reset();
	_dataprovider->reset();
}

void DirectApplication::execute(const std::string& commandName, const std::vector<std::string>& args) {

	if (args.empty()) {
		return;
	}

	vector<string> newArgs;
	vector<string>::const_iterator a = args.begin();
	a++;
	while (a != args.end()) {
		newArgs.push_back(*a);
		a++;
	}

	string newCommandName = args.front();

	if (commandName == "dp") {
		_dataprovider->execute(newCommandName, newArgs);
	} else if (commandName == "cp") {
		_commandprocessor->execute(newCommandName, newArgs);
	} else if (commandName == "cr") {
		_core->execute(newCommandName, newArgs);
	}
}

void DirectApplication::quit(const std::vector<std::string>& args, bool isRemote) {

	// be able to stop the server when quitting
	_state = STOPPING;
	setFileState(EXIT_STATE);

	_console->quit();

	_core->quit(args, isRemote);
	_commandprocessor->quit(args, isRemote);
	_dataprovider->quit(args, isRemote);

	_terminated.wait();
}

void DirectApplication::setCommandProcessorStarted() {
	_commandProcessorStarted.wait();
}

bool DirectApplication::isStarted() const {
	return _state == STARTED;
}

bool DirectApplication::isStopping() const {
	return _state == STOPPING;
}

void DirectApplication::forceQuit() {

	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::getInstance()->forceReset();
	}
}

void DirectApplication::test() {

	if (ics::ServerManager::getInstance() != 0) {
		bool result = ics::ServerManager::getInstance()->test();
		if (!result) {
			_returnCode = EXIT_FAILURE;
		}
	}

	// quit
	_core->quit(std::vector<std::string>(), false);
	_commandprocessor->quit(std::vector<std::string>(), false);
	_dataprovider->quit(std::vector<std::string>(), false);
}

void DirectApplication::testEstimatedTime() {

	if (ics::ServerManager::getInstance() != 0) {
		bool result = ics::ServerManager::getInstance()->testEstimatedTime();
		if (!result) {
			_returnCode = EXIT_FAILURE;
		}
	}

	// quit
	_core->quit(std::vector<std::string>(), false);
	_commandprocessor->quit(std::vector<std::string>(), false);
	_dataprovider->quit(std::vector<std::string>(), false);
}

void DirectApplication::replacePropertyNamesWithAlias() {

	_core->replacePropertyNamesWithAlias();

	// quit
	_core->quit(std::vector<std::string>(), false);
	_commandprocessor->quit(std::vector<std::string>(), false);
	_dataprovider->quit(std::vector<std::string>(), false);
}

int DirectApplication::getReturnCode() const {
	return _returnCode;
}
