/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DIRECTAPPLICATION_H_
#define DIRECTAPPLICATION_H_

#include <common/Application.h>
#include <string>
#include <boost/thread/thread.hpp>
#include <boost/thread/condition.hpp>
#include <boost/thread/barrier.hpp>

namespace common {

	class Console;
}

namespace dataprovider {

class DirectDataProvider;

}

namespace command {

class DirectCommandProcessor;

}

namespace core {

class DirectCore;

}

/**
 * Direct application.
 */
class DirectApplication : public common::Application {

public :
	/**
	 * Constructor.
	 */
	DirectApplication();

	/**
	 * Virtual destructor.
	 */
	virtual ~DirectApplication();

	/**
	 * Initialises the application.
	 */
	virtual void init(int argc, char* argv[]);

	/**
	 * Runs the application.
	 */
	virtual void run();

	/**
	 * Resets the application.
	 */
	virtual void reset();

	/**
	 * Executes the command called commandName.
	 */
	virtual void execute(const std::string& commandName, const std::vector<std::string>& args);

	/**
 	 * Quits the application.
 	 */
	virtual void quit(const std::vector<std::string>& args, bool isRemote);

	/**
	 * Sets the comman processor started.
	 */
	void setCommandProcessorStarted();

	/**
	 * Returns true if the application is started.
	 */
	bool isStarted() const;

	/**
	 * Returns true if the application is stopping.
	 */
	bool isStopping() const;

	/**
	 * Forces to quit the application.
	 */
	void forceQuit();

	/**
	 * Returns the return code.
	 */
	int getReturnCode() const;

private:
	void setFileState(std::string const & state);
	void test();
	void testEstimatedTime();
	void replacePropertyNamesWithAlias();

	common::Console* _console;

	dataprovider::DirectDataProvider * _dataprovider;
	command::DirectCommandProcessor * _commandprocessor;
	core::DirectCore * _core;

	bool _testExecution;
	bool _testEstimatedTimeExecution;
	bool _replacePropertyNamesWithAlias;

	enum State {STARTING, STARTED, STOPPING};
	State _state;

	boost::barrier _commandProcessorStarted;
	boost::barrier _terminated;

	int _returnCode;

	static const std::string SIMULATION;
	static const std::string VISA;
	static const std::string SIMULATION3D;
	static const std::string SIMULATION_NEUTRON;
	static const std::string TEST_EXECUTION;
	static const std::string TEST_ESTIMATED_TIME;
	static const std::string REPLACE_PROPERTY_NAMES;
	static const std::string STATEFILENAME;
	static const std::string INIT_STATE;
	static const std::string EXEC_STATE;
	static const std::string EXIT_STATE;
};

#endif /*DIRECTAPPLICATION_H_*/
