#include "ics/Module/ModuleManager.h"
#include "common/base/ServerProperties.h"
#include <iostream>

#include <sysexits.h>

using namespace std;
using namespace common;

/**
 * The module load sequencer has 3 tasks:
 * - generate the load sequence
 *       The application is started at each iteration to avoid the persistent state of dlopen when an attempt to load a module is unsuccessful.
 *       The LoadSequence.xml file (in mod) grows at each iteration (until it cannot, i.e. the end of the algorithm).
 *       At each iteration the set L = A - S (A = all modules, S = sequence, L = modules to load) of modules is iterated. A module that is loaded successfully
 *       is added to S.
 *
 * - generate the minimum load sequence (for the InstrumentConfig.xml and HardwareConfig.xml files)
 *       Two types of dependencies are defined : direct (the module has a type that is present in the config files), indirect (the module is necessary for another module that is direct)
 *       First the algorithm computes the list of direct modules.
 *       Then the application (started at each iteration) tries to not load each module by reverse order of the load sequence.
 *       If the new sequence (1 module is skipped) is loaded successfully, then the skipped module is unnecessary.
 *       Otherwise it means that the module is indirect (indirectly necessary).
 */

int generateSequence(const std::string& directoryPath) {

	try {
		bool finished = ModuleManager::getInstance()->loadModulesForOrdering(directoryPath);

		if (finished) {
			return EXIT_SUCCESS;
		} else {
			// special value to continue to iterate
			return 11;
		}

	} catch (const Exception & e) {
		cout << "Unable to find modules path" << endl;
	}

	return EXIT_SUCCESS;
}

int generateDirectDependencies(const std::string& directoryPath, const std::string& configPath) {
	ModuleManager::getInstance()->setDirectDependenciesInSequence(directoryPath, configPath);

	return EXIT_SUCCESS;
}

int generateMinimumSequence(const std::string& directoryPath, const std::string& moduleConfigFilename) {

	try {
		bool finished = ModuleManager::getInstance()->loadModulesForMinimizing(directoryPath, moduleConfigFilename);

		if (finished) {
			return EXIT_SUCCESS;
		} else {
			// special value to continue to iterate
			return 11;
		}

	} catch (const Exception & e) {
		cout << "Unable to find modules path" << endl;
	}

	return EXIT_SUCCESS;
}

int main(int argc, char* argv[]) {

	string task = "sequence";

	if (argc > 1) {
		task = argv[1];
	}

	const char* home = getenv("NOMAD_HOME");
	string nomadHomePath = string(home);
	string nomadModPath = nomadHomePath + "/mod";

	home = getenv("HOME");
	string nomadConfigPath = string(home) + "/.nomad";

	ModuleManager::initInstance();

	// dispatch task
	if (task == "sequence") {
		return generateSequence(nomadModPath);
	} else if (task == "config") {
		return generateDirectDependencies(nomadModPath, nomadConfigPath);
	} else if (task == "minimize") {

		string moduleConfigFilename;
		if (argc > 2) {
			moduleConfigFilename = argv[2];
		} else {
			cout << "Error, a file must be passed in argument" << endl;
			return EXIT_FAILURE;
		}

		return generateMinimumSequence(nomadModPath, moduleConfigFilename);
	}

}
