/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DirectCommandProcessor.h"
#include <command/processor/ProcessorManager.h>
#include <command/base/ServantManager.h>
#include <command/transportlayer/MultiThreadProcessorManagerCommunication.h>
#include <command/transportlayer/MultiThreadProcessorCommunication.h>
#include <command/log/LogHandler.h>
#include <ics/IcsServerManager.h>
#include <common/base/ServerProperties.h>
#include <common/base/Convert.h>
#include <common/base/Real.h>
#include <common/output/OutputStream.h>
#include <common/output/Debug.h>
#include <DirectApplication.h>
#include <cameo/api/cameo.h>
#include <boost/lexical_cast.hpp>
#include <iostream>
#include <cstdlib>

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;

namespace command {

const std::string DirectCommandProcessor::ICS = "-ics";
const std::string DirectCommandProcessor::SIMULATION = "sim";
const std::string DirectCommandProcessor::SIMULATION3D = "sim3d";
const std::string DirectCommandProcessor::DEBUG_ON = "debug-on";

DirectCommandProcessor::DirectCommandProcessor(DirectApplication * application) :
	Application(),
	_application(application),
	_onError(false),
	_thread(nullptr) {
}

DirectCommandProcessor::~DirectCommandProcessor() {

	_thread->join();
	delete _thread;
}

void DirectCommandProcessor::init(int argc, char* argv[]) {

	string serverId;

	for (int i = 0; i < argc; i++) {

		string oldSimulation = string("-") + SIMULATION;
		if ((string(argv[i]) == SIMULATION || string(argv[i]) == SIMULATION3D || string(argv[i]) == oldSimulation)) {
			serverId = lexical_cast<string>(This::getId());
		} else if (string(argv[i]) == ICS) {
			ics::ServerManager::initInstance();
		} else if (string(argv[i]) == DEBUG_ON) {
			Debug::enable();
		}
	}

	out.setStdOut(true);
	err.setStdOut(true);
	warn.setStdOut(true);
	title.setStdOut(true);
	trace.setStdOut(true);

	// Processor Manager
	MultiThreadProcessorManagerCommunication::initInstance(argc, argv, false, serverId);
	ProcessorManager::initInstance(MultiThreadProcessorManagerCommunication::getInstance()->getProcessorManagerMessenger(), this);

	// Processor
	MultiThreadProcessorCommunication::initInstance(argc, argv, serverId);

	Real::init();

	LogHandler::initInstance(MultiThreadProcessorCommunication::getInstance()->getLogAccessor());
	StringHelper::init();
}

void DirectCommandProcessor::loop() {

	ProcessorManager::getInstance()->run();
}

void DirectCommandProcessor::run() {

	// Initialising the servants here, because the notifications are asynchronous and need the Core to be
	// initialised.
	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::getInstance()->init();
	}

	ServantManager::initInstance(MultiThreadProcessorCommunication::getInstance()->getServantAccessor(), MultiThreadProcessorCommunication::getInstance()->getCommandAccessor());

//	if (ics::ServerManager::getInstance() != 0) {
//		ics::ServerManager::getInstance()->cloneControllers();
//	}

	MultiThreadProcessorManagerCommunication::getInstance()->run();

	_thread = new std::thread(&DirectCommandProcessor::loop);
}

void DirectCommandProcessor::shutdown(bool isRemote) {

	MultiThreadProcessorManagerCommunication::getInstance()->stop();

	ProcessorManager::getInstance()->shutdown();
}

void DirectCommandProcessor::reset() {

	out << common::shutdown << "Command processor shutdown" << eol;

	_thread->join();

	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::resetInstance();
	}

	// Processor
	LogHandler::resetInstance();

	ServantManager::resetInstance();

	MultiThreadProcessorCommunication::resetInstance();

	// Processor Manager
	ProcessorManager::resetInstance();

	MultiThreadProcessorManagerCommunication::resetInstance();
}

void DirectCommandProcessor::setStarted() {
	_application->setCommandProcessorStarted();
}

void DirectCommandProcessor::setOnError() {
	_onError = true;
}

bool DirectCommandProcessor::onError() const {
	return _onError;
}

void DirectCommandProcessor::show() {

	vector<ProcessorItem*> processors = ProcessorManager::getInstance()->getProcessors();

	out << nolog << "processors (" << processors.size() << ") :" << eol;

	for (vector<ProcessorItem*>::const_iterator p = processors.begin(); p != processors.end(); p++) {

		ProcessorItem* processorItem = *p;

		string ready;
		bool isResponding = false;

		if (processorItem->isReady()) {
			ready = "ready    ";

			isResponding = ProcessorManager::getInstance()->isProcessorResponding(processorItem->getID());

		} else {
			ready = "not ready";
		}

		string active;

		if (processorItem->isActive()) {
			active = "active  ";
		} else {
			active = "inactive";
		}

		string responding;

		if (isResponding) {
			responding = "ok";
		} else {
			responding = "not responding";
		}

		out << nolog << "\tprocessor (pid = " << processorItem->getPID() << ", id = " << processorItem->getID() << ") : " << ready << ", " << active << ", " << responding << eol;
	}
}

void DirectCommandProcessor::where(unsigned long processorID) {

	vector<ProcessorItem*> processors = ProcessorManager::getInstance()->getProcessors();

	out << nolog << "processors (" << processors.size() << ") :" << eol;
	for (vector<ProcessorItem*>::const_iterator p = processors.begin(); p != processors.end(); p++) {

		ProcessorItem* processorItem = *p;

		if (processorItem->getID() == processorID) {

			string ready;
			bool isResponding = false;

			if (processorItem->isReady()) {
				ready = "ready    ";
				isResponding = ProcessorManager::getInstance()->isProcessorResponding(processorItem->getID());
			} else {
				ready = "not ready";
			}

			string active;

			if (processorItem->isActive()) {
				active = "active  ";
			} else {
				active = "inactive";
			}

			string responding;

			if (isResponding) {
				responding = "ok";
			} else {
				responding = "not responding";
			}

			string currentStack;
			if (processorItem->isActive() && isResponding) {
				currentStack = ProcessorManager::getInstance()->getProcessorCallStack(processorID);
			}

			out << nolog << "\tprocessor (pid = " << processorItem->getPID() << ", id = " << processorItem->getID() << ") : " << ready << ", " << active << ", " << responding << eol;
			out << nolog << "\t\tstack : " << currentStack << eol;
		}
	}
}

void DirectCommandProcessor::unloadUnusedModules() {
	ics::ServerManager::getInstance()->unloadUnusedModules();
}

void DirectCommandProcessor::execute(const std::string& commandName, const std::vector<std::string>& args) {

	// All commands that arrive are ONLY for the Processor Manager.
	if (commandName == "show") {
		show();

	} else if (commandName == "where") {

		if (args.size() > 0) {
			unsigned long processorID = to<int>(args[0]);
			where(processorID);
		}

	} else if (commandName == "unmod") {
		unloadUnusedModules();

	} else if (commandName == "debug-on") {
		Debug::enable();

	} else if (commandName == "debug-off") {
		Debug::disable();

	} else if (commandName == "help") {

		out << nolog << "commands :" << eol;
		out << nolog << "\tshow : shows the state of the processes" << eol;
		out << nolog << "\twhere <ID> : shows the current line of the processor <ID>" << eol;
		out << nolog << "\tunmod : unloads the modules that are not used" << eol;
		out << nolog << "\tquit : quits the application" << eol;

	} else {

		out << nolog << "command '" << commandName << "' unknown" << eol;
	}

}

void DirectCommandProcessor::quit(const std::vector<std::string>& args, bool isRemote) {

	shutdown(isRemote);
}

}
