/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ServantManager.h"
#include <command/transportlayer/dataprovider/ServantAccessor.h>
#include <command/transportlayer/dataprovider/PropertyAccessor.h>
#include <command/transportlayer/dataprovider/CommandAccessor.h>
#include <command/base/controller/Base.h>
#include <command/base/driver/Base.h>
#include <common/base/Convert.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <ics/IcsDriver.h>
#include <ics/IcsController.h>
#include "InstrumentManager/HardwareManager.h"
#include "InstrumentManager/InstrumentManager.h"
#include <vector>
#include <list>
#include <iostream>

using namespace std;
using namespace common;

namespace command {

ServantManager* ServantManager::_instance = 0;

ServantManager::NoSuchServantException::NoSuchServantException() :
	TraceableException("command::ServantManager::NoSuchServantException", string("unable to find servant")) {
}

ServantManager::NoSuchServantException::NoSuchServantException(unsigned long ID) :
	TraceableException("command::ServantManager::NoSuchServantException", string("unable to find servant with ID <") + to<string>(ID) + ">") {
}

ServantManager::ServantManager() :
	_onError(false) {
}

ServantManager::~ServantManager() {
}

ServantManager* ServantManager::getInstance() {
	return _instance;
}

void ServantManager::initInstance(ServantAccessor* servantAccessor, CommandAccessor* commandAccessor) {

	if (_instance == 0) {

		_instance = new ServantManager();

		_instance->init(servantAccessor, commandAccessor);
	}
}

void ServantManager::resetInstance() {

	if (_instance != 0) {
		_instance->reset();
		delete _instance;
		_instance = 0;
	}
}

void ServantManager::init(ServantAccessor* servantAccessor, CommandAccessor* commandAccessor) {

	_servantAccessor = servantAccessor;
	_commandAccessor = commandAccessor;


	// get list of installed controllers and drivers
	vector<ServantDescriptor> controllers = _servantAccessor->getControllers();
	vector<ServantDescriptor> drivers = _servantAccessor->getDrivers();

	bool problem = false;

	out << "Synchronising controllers" << eol;

	// create controllers
	vector<ServantDescriptor>::iterator s;
	for (s = controllers.begin(); s != controllers.end(); s++) {
		ServantDescriptor descriptor = *s;

		controller::Base* controller = new controller::IcsController(descriptor.getID(), InstrumentManager::getInstance()->getExperimentController(descriptor.getName()));

		if (controller != 0) {

			try {
				// synchronise controllers with data provider
				controller->_synchronise(_commandAccessor);

				// add controller to map
				_controllers.insert(descriptor.getID(), controller);

			} catch (Exception& e) {
				err << "Cannot synchronise controller \"" << descriptor.getName() << "\" (" << descriptor.getType() << ")" << eol;
				err << e.printStack() << eol;
				delete controller;
				controller = 0;
				problem = true;
			}

		} else {
			err << "Cannot create controller \"" << descriptor.getName() << "\" (" << descriptor.getType() << ")" << eol;
			problem = true;
		}
	}

	out << "Synchronising drivers" << eol;

	for (s = drivers.begin(); s != drivers.end(); s++) {

		ServantDescriptor descriptor = *s;

		driver::Base* driver = new driver::IcsDriver(descriptor.getID(), HardwareManager::getInstance()->getTestController(descriptor.getName()));

		if (driver != 0) {

			try {

				// synchronise controllers with data provider
				driver->_synchronise(_commandAccessor);

				// add driver to map
				_drivers.insert(descriptor.getID(), driver);

			} catch (Exception& e) {
				err << "Cannot synchronise driver \"" << descriptor.getName() << "\" (" << descriptor.getType() << ")" << eol;
				err << e.printStack() << eol;
				delete driver;
				problem = true;
			}

		} else {
			err << "Cannot create driver \"" << descriptor.getName() << "\" (" << descriptor.getType() << ")" << eol;
			problem = true;
		}
	}

	if (problem) {
		_onError = true;
	}
}

bool ServantManager::initServants(bool verify, bool driversAndProtocolsOnly) {

	if (!driversAndProtocolsOnly) {

		HashMap<unsigned long, controller::Base*>::iterator c;
		for (c = _controllers.begin(); c != _controllers.end(); c++) {

			// set the controller initialised in any case
			_servantAccessor->setServantInitialised(c->second->_getID(), true);
		}
	}

	return true;
}

void ServantManager::reset() {

	// delete controllers
	HashMap<unsigned long, controller::Base*>::iterator c;
	for (c = _controllers.begin(); c != _controllers.end(); c++) {
		delete c->second;
	}

	// delete drivers
	HashMap<unsigned long, driver::Base*>::iterator d;
	for (d = _drivers.begin(); d != _drivers.end(); d++) {
		delete d->second;
	}

}

CommandAccessor* ServantManager::getCommandAccessor() {
	return _commandAccessor;
}

controller::Base* ServantManager::createController(std::string const & controllerType, std::string const & name, unsigned long ID) {

	controller::Base* controller = new controller::IcsController(ID, InstrumentManager::getInstance()->getExperimentController(name));

	if (controller == 0) {
		err << "Cannot create controller \"" << name << "\" (" << controllerType << ")" << eol;
	} else {
		// add controller to map
		_controllers.insert(ID, controller);

		controller->_synchronise(_commandAccessor);
	}

	return controller;
}

driver::Base* ServantManager::createDriver(std::string const & driverType, std::string const & name, unsigned long ID) {

	driver::Base* driver = new driver::IcsDriver(ID, HardwareManager::getInstance()->getTestController(name));

	if (driver == 0) {
		err << "Cannot create driver \"" << name << "\" (" << driverType << ")" << eol;
	} else {
		// add driver to map
		_drivers.insert(ID, driver);

		driver->_synchronise(_commandAccessor);
	}

	return driver;
}

void ServantManager::removeDriver(unsigned long ID) {

	if (_drivers.find(ID)) {
		delete _drivers.get(ID);
		_drivers.erase(ID);
	} else {
		err << "Cannot remove driver \"" << ID << "\"" << eol;
	}
}

void ServantManager::removeController(unsigned long ID) {

	if (_controllers.find(ID)) {
		delete _controllers.get(ID);
		_controllers.erase(ID);
	} else {
		err << "Cannot remove controller \"" << ID << "\"" << eol;
	}
}

Servant* ServantManager::getServant(unsigned long ID) {

	if (_controllers.find(ID)) {
		return _controllers[ID];
	} else if (_drivers.find(ID)) {
		return _drivers[ID];
	}

	return 0;
}

}
