/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_SERVANTMANAGER_H_
#define COMMAND_SERVANTMANAGER_H_

#include <command/exception/TraceableException.h>
#include <common/base/HashMap.h>
#include <command/base/Servant.h>

namespace command {

namespace driver {

	class Base;
}

namespace controller {

	class Base;
}

class ServantAccessor;
class PropertyAccessor;
class CommandAccessor;

/**
 * Singleton class used to obtain/instantiate instances of controllers and drivers.
 * It uses the Factories to perform the real instantiation, the instances are stored
 * in maps here along with their IDs.
 */
class ServantManager {

public:
	/**
	 * Exception for no such servant.
	 */
	class NoSuchServantException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NoSuchServantException();

		/**
		 * Constructor.
		 */
		NoSuchServantException(unsigned long ID);
	};

	/**
	 * returns the instance.
	 */
	static ServantManager* getInstance();

	/**
	 * Creates and initialises the instance with the accessors to the Data Provider.
	 */
	static void initInstance(ServantAccessor* servantAccessor, CommandAccessor* commandAccessor);

	/**
	 * Resets and deletes the instance.
	 */
	static void resetInstance();

	/**
	 * Verifies and initialises the servants, properties and commands.
	 */
	bool initServants(bool verify, bool driversAndProtocolsOnly = false);

	/**
	 * Gets the command accessor.
	 */
	CommandAccessor* getCommandAccessor();

	/**
	 * Creates a Controller.
	 */
	controller::Base* createController(std::string const & controllerType, std::string const & name, unsigned long ID);

	/**
	 * Creates a Driver. Initialises the driver.
	 */
	driver::Base* createDriver(std::string const & driverType, std::string const & name, unsigned long ID);

	/**
	 * Removes the driver with ID.
	 */
	void removeDriver(unsigned long ID);

	/**
	 * Removes the controller with ID.
	 */
	void removeController(unsigned long ID);

	/**
	 * Gets the servant with ID.
	 */
	Servant* getServant(unsigned long ID);

private:
	/**
	 * Constructor with Data Provider accessors.
	 */
	ServantManager();

	/**
	 * Destructor.
	 */
	~ServantManager();

	/**
	 * Initialises the instance.
	 */
	void init(ServantAccessor* servantAccessor, CommandAccessor* commandAccessor);

	/**
	 * Resets the instance.
	 */
	void reset();

	static ServantManager* _instance;

	ServantAccessor* _servantAccessor;
	CommandAccessor* _commandAccessor;

	common::HashMap<unsigned long, controller::Base*> _controllers;
	common::HashMap<unsigned long, driver::Base*> _drivers;

	struct ServantProperty {
		std::string servantType;
		std::string servantName;
		std::string name;
	};

	bool _onError;
};

}

#endif /*COMMAND_SERVANTMANAGER_H_*/
