/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_BASECOMMAND_H_
#define COMMAND_BASECOMMAND_H_

#include "Method.h"
#include <common/base/BaseTypes.h>
#include <common/base/CommonTypes.h>

namespace command {

class CommandAccessor;
class Servant;

/**
 * An abstract class providing a common interface to the template class Command.
 */
class BaseCommand : public Method {

public:
	/**
	 * Constructor.
	 */
	BaseCommand(const std::string& name, Servant* servant);

	/**
	 * Virtual destructor.
	 */
	virtual ~BaseCommand();

	/**
	 * Gets the reference of the object. Used for nomad function calls.
	 */
	const BaseCommand & getReference() const;

	/**
	 * Gets the reference of the object. Used for nomad function calls.
	 */
	BaseCommand & getReference();

	/**
	 * Initialises a Command with its Data Provider ID and accessor.
	 */
	void init(unsigned long ID, CommandAccessor* commandAccessor);

	/**
	 * Sets the state. Calls the Command Accessor.
	 */
	void setState(common::CommandState state);

	/**
	 * Sets active.
	 */
	common::SetCommandActiveResult setActive(const std::string& servantType);

	/**
	 * Returns the ID of the command.
	 */
	unsigned long getID() const;

	/**
	 * Gets the state. Calls the Command Accessor.
	 */
	common::CommandState getState() const;

	/**
	 * Executes the command.
	 */
	virtual void execute(bool queue = false, bool skipIfWait = false) = 0;

	/**
	 * Returns true if the command has executed for longer than the desired timeout.
	 */
	virtual bool timeoutExceeded() = 0;

	/**
	 * Returns the timeout in milliseconds.
	 */
	double getExceededTimeoutMs() const;

	/**
	 * Gets the pre-run calculated duration.
	 */
	double getPreRunCalculatedDurationMs() const;

	/**
	 * Sets the period of calculating the progression.
	 */
	void setProgressionPeriod(unsigned long periodMs);

	/**
	 * Gets the period of calculating the progression.
	 */
	unsigned long getProgressionPeriod() const;

	/**
	 * Calculates the progression.
	 */
	void calculateProgression();

	/**
	 * Returns the elapsed time of the command.
	 */
	double getElapsedTimeMs() const;

	/**
	 * Returns the elapsed time of the command. Nomad function call.
	 */
	double _elapsedTimeMs() const;

	/**
	 * Returns the progression of the command.
	 */
	double getCurrentProgression() const;

	/**
	 * Returns the progression of the command. Nomad function call.
	 */
	double _currentProgression() const;

	/**
	 * Returns the progression of the command. Nomad function call.
	 */
	void _resetProgression();

	/**
	 * Returns a stringified form of the command.
	 */
	virtual std::string toString() const = 0;

protected:
	double _exceededTimeoutMs;
	unsigned long _ID;
	CommandAccessor* _commandAccessor;
};

}

#endif /*COMMAND_BASECOMMAND_H_*/
