/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "LogHandler.h"
#include "DatabaseLog.h"
#include "SimpleLog.h"
#include <command/transportlayer/dataprovider/LogAccessor.h>
#include <command/message/log/WriteLogMessage.h>
#include <command/message/log/DataImageReadyLogMessage.h>
#include <command/message/log/WriteSimpleLogMessage.h>
#include <command/message/ShutdownMessage.h>
#include <command/message/TerminateMessage.h>
#include <command/exception/Failure.h>
#include <command/processor/ThreadProcessorInfo.h>
#include <common/base/Queue.h>
#include <common/output/OutputStream.h>

using namespace std;
using namespace common;

namespace command {

LogHandler* LogHandler::_instance = 0;

LogHandler::LogHandler() :
	_logAccessor(0),
	_running(false),
	_messageQueue(nullptr),
	_thread(nullptr),
	_requestCounter(0) {
}

LogHandler::~LogHandler() {

}


void LogHandler::initInstance(LogAccessor* logAccessor) {
	if (_instance == 0) {
		_instance = new LogHandler();
		_instance->init(logAccessor);
	}
}

void LogHandler::resetInstance() {

	if (_instance != 0) {
		_instance->reset();
	}	

	delete _instance;
	_instance = 0;
}

LogHandler* LogHandler::getInstance() {
	return _instance;
}


void LogHandler::init(LogAccessor* logAccessor) {
	_logAccessor = logAccessor;

	_messageQueue = new Queue<Message>(1024);
	
	_thread = new std::thread(&LogHandler::loop);
}

void LogHandler::reset() {

	_messageQueue->put(new ShutdownMessage());
	
	_thread->join();
	delete _thread;

	delete _messageQueue;
}

void LogHandler::write(const DatabaseLog& log, bool writeLogFile) {
	Queue<Result> waitForReplyQueue(1);
	_messageQueue->put(new WriteLogMessage(log.getType(), log.getLevel(), log.getLineElements(), log.getImageName(), log.getCommandID(), log.getServantID(), writeLogFile, &waitForReplyQueue));
	waitForTermination(waitForReplyQueue);

}

void LogHandler::write(const SimpleLog& log, bool writeLogFile) {
	//Yannick --> now we use simpleLog messages for command line... they are not blocking...
	_messageQueue->put(new WriteSimpleLogMessage(log.getType(), log.getLevel(), log.getMessage(), writeLogFile));
}

void LogHandler::imageDataReady(const std::string& image) {
	// No need for asynchronous queue.
	// This simplification should be applied to write (see core::LogHandler)
	_logAccessor->imageDataReady(image);

}

LogHandler::Result LogHandler::waitForTermination(Queue<Result>& waitForReplyQueue) {

	Result result;

	try {
		Result* waitForResult = waitForReplyQueue.get();
		result = *waitForResult;
		delete waitForResult;
		
	} catch (...) {
		// if an exception occurs, make sure that the operation terminates fully.
		Result* waitForResult = waitForReplyQueue.get();
		result = *waitForResult;
		delete waitForResult;
		
		throw;
	}
	
	if (!result.success) {
		throw Failure(result.message);
	}

	return result;
}

void LogHandler::terminate(unsigned long requestID, bool success, const std::string& message) {
	_messageQueue->put(new TerminateMessage(requestID, success, message, 0));
}


void LogHandler::loop() {

	Message* message = 0;
	_instance->_running = true;
	
	while (_instance->_running) {
		
		try {
			try {
				message = _instance->_messageQueue->get();
				
			} catch (Queue<Message>::UnavailableException& e) {
				break;
				
			} catch (Exception& e) {
				err << e.printStack() << eol;
				break;
			}
			
			try {
				_instance->processMessage(message);
	
			} catch (Exception& e) {
				err << e.printStack() << eol;
			}
				
		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}
		
		delete message;
		message = 0;
 	}
}

void LogHandler::processMessage(Message* message) {

	if (message->getType() == Message::WRITE_LOG) {
		processMessage(static_cast<WriteLogMessage*>(message));

	} else if (message->getType() == Message::WRITE_SIMPLE_LOG) {
		processMessage(static_cast<WriteSimpleLogMessage*>(message));

	} else if (message->getType() == Message::TERMINATE) {
		processMessage(static_cast<TerminateMessage*>(message));

	} else if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<ShutdownMessage*>(message));

	}
}

void LogHandler::processMessage(WriteLogMessage* message) {

	++_requestCounter;
	_waitForReplyQueues.insert(_requestCounter, message->getWaitForReplyQueue());

	// Sending request
	_logAccessor->write(message->getLogType(), message->getLogLevel(), message->getLineElements(), message->getImageName(), message->getCommandID(), message->getServantID(), message->getWriteLogFile(), _requestCounter);

}

void LogHandler::processMessage(WriteSimpleLogMessage* message) {
	_logAccessor->write(message->getLogType(), message->getLogLevel(), message->getMessage(), message->getWriteLogFile());
}

void LogHandler::processMessage(TerminateMessage* message) {
	
	Result* result = new Result;
	result->success = message->isSuccessful();
	result->message = message->getMessage();

	unsigned long requestID = message->getRequestID(); 
	_waitForReplyQueues.get(requestID)->put(result);

	_waitForReplyQueues.erase(requestID);
}

void LogHandler::processMessage(ShutdownMessage* message) {
	
	_running = false;
}


}
