/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_LOGHANDLER_H_
#define COMMAND_LOGHANDLER_H_

#include <string>
#include <common/base/HashMap.h>
#include <thread>
#include <list>

namespace common {

template<typename ItemType>
class Queue;

}

namespace command {

//class BaseLog;
class DatabaseLog;
class SimpleLog;
class LogAccessor;
class Message;
class WriteLogMessage;
class WriteSimpleLogMessage;
class ShutdownMessage;
class TerminateMessage;
class DataImageReadyLogMessage;

/**
 * Class defining a log handler singleton.
 */
class LogHandler {

public:
	/**
	 * Gets the instance. 
	 */
	static LogHandler* getInstance();
	
	/**
	 * Initialises the instance. 
	 */
	static void initInstance(LogAccessor* logAccessor);
	
	/**
	 * Resets the instance. 
	 */
	static void resetInstance();

	/**
	 * Writes the log message (for database).
	 */
	void write(const DatabaseLog& log, bool writeLogFile = true);

	/**
	 * Writes log message coming from CommandLine or related to ASCII/Nexus file
	 */
	void write(const SimpleLog& log, bool writeLogFile);

	void imageDataReady(const std::string& image);

	/**
	 * Notifies that the request of request ID has terminated. 
	 */
	void terminate(unsigned long requestID, bool success, const std::string& message);

	/**
	 * Class defining a result.
	 */
	struct Result {
		bool success;
		std::string message;
	};

private:
	LogHandler();
	virtual ~LogHandler();

	void init(LogAccessor* logAccessor);
	void reset();

	void processMessage(Message* message);
	void processMessage(WriteLogMessage* message);
	void processMessage(WriteSimpleLogMessage* message);
	void processMessage(TerminateMessage* message);
	void processMessage(ShutdownMessage* message);
	Result waitForTermination(common::Queue<Result>& waitForReplyQueue);

	static void loop();

	static LogHandler* _instance;

	LogAccessor* _logAccessor;

	bool _running;	
	common::Queue<Message>* _messageQueue;
	std::thread* _thread;
	
	unsigned long _requestCounter;
	common::HashMap<int, common::Queue<Result>*> _waitForReplyQueues;
	std::list<unsigned long> _completedRequests;

};

}

#endif /*COMMAND_LOGHANDLER_H_*/
