/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Processor.h"
#include <command/message/processor/InitServantsMessage.h>
#include <command/message/processor/CommandMessage.h>
#include <command/message/processor/DurationRequestMessage.h>
#include <command/message/ShutdownMessage.h>
#include <command/transportlayer/dataprovider/ServantAccessor.h>
#include <command/transportlayer/dataprovider/CommandAccessor.h>
#include <command/transportlayer/dataprovider/DatabaseIDKeeper.h>
#include <command/base/method/BaseCommand.h>
#include <command/base/ServantManager.h>
#include <command/base/controller/Base.h>
#include <command/base/driver/Base.h>
#include <command/processor/ThreadProcessorInfo.h>
#include <ics/IcsController.h>
#include <ics/IcsDriver.h>
#include <common/base/Convert.h>
#include <common/base/Date.h>
#include <common/output/OutputStream.h>
#include <string>
#include <iostream>

using namespace common;
using namespace std;

namespace command {

std::mutex Processor::_instancesMutex;
common::HashMap<unsigned long, Processor*> Processor::_instances;

Processor::Processor() :
	_processorID(0),
	_running(false),
	_shutdownCalled(false) {
}

Processor::~Processor() {
}

Processor * Processor::getInstance() {

	std::scoped_lock lock(_instancesMutex);

	unsigned long processorID = ThreadProcessorInfo::getInstance()->getProcessorID();

	if (!_instances.find(processorID)) {
		return 0;
	}

	return _instances.get(processorID);
}

Processor * Processor::getInstance(unsigned long processorID) {

	std::scoped_lock lock(_instancesMutex);

	if (!_instances.find(processorID)) {
		return 0;
	}

	return _instances.get(processorID);
}

void Processor::initInstance(unsigned long processorID, ProcessorManagerProxy* processorManagerProxy, ServantAccessor* servantAccessor, CommandAccessor* commandAccessor) {

	std::scoped_lock lock(_instancesMutex);

	if (!_instances.find(processorID)) {

		Processor * instance = new Processor();
		instance->init(processorID, processorManagerProxy, servantAccessor, commandAccessor);
		_instances.insert(processorID, instance);
	}
}

void Processor::resetInstance() {

	Processor * instance = 0;

	unsigned long processorID = ThreadProcessorInfo::getInstance()->getProcessorID();

	// locking instances
	{
		std::scoped_lock lock(_instancesMutex);

		instance = _instances.get(processorID);
		_instances.erase(processorID);
	}

	if (instance != 0) {
		instance->reset();

		delete instance;
	}

}

void Processor::setProcessorID(unsigned long processorID) {
	_processorID = processorID;
}

unsigned long Processor::getProcessorID() const {
	return _processorID;
}

bool Processor::initServants(unsigned long databaseID, bool verify) {

	Queue<bool> waitingQueue(1);

	// post new message
	_messageQueue->put(new InitServantsMessage(databaseID, verify, false, &waitingQueue));

	// waiting for the termination of the message processing
	bool* result = waitingQueue.get();
	bool value = *result;
	delete result;

	return value;
}

void Processor::initDriversAndProtocols() {

	Queue<bool> waitingQueue(1);

	// post new message
	_messageQueue->put(new InitServantsMessage(0, false, true, &waitingQueue));

	// waiting for the termination of the message processing
	delete waitingQueue.get();

	_processorManagerProxy->initTerminated();
}

void Processor::execute(unsigned long databaseID, unsigned long commandID, bool skipIfWait, unsigned long commandExecutionID) {

	try {
		unsigned long servantID = _commandAccessor->getServantID(commandID);

		try {
			Servant* servant = ServantManager::getInstance()->getServant(servantID);

			// Create new Command Message and put it on the queue
			_messageQueue->put(new CommandMessage(databaseID, servant->_getCommand(commandID), skipIfWait, commandExecutionID));

		} catch (Exception& e) {
			// failure
			if (!_shutdownCalled) {
				_processorManagerProxy->commandFailure(commandID, commandExecutionID, "no such servant");
			}
			return;
		}

	} catch (Exception& e) {

		err << e.printStack() << eol;
		// failure!
		if (!_shutdownCalled) {
			_processorManagerProxy->commandFailure(commandID, commandExecutionID, "cannot execute command");
		}
	}
}

void Processor::stop(unsigned long databaseID, unsigned long commandOrPropertyID) {
}

void Processor::pause(unsigned long databaseID, unsigned long commandOrPropertyID) {
}

void Processor::restart(unsigned long databaseID, unsigned long commandOrPropertyID) {
}

void Processor::executeDurationRequest(unsigned long databaseID, unsigned long commandID, unsigned long durationRequestExecutionID) {

	try {
		unsigned long servantID = _commandAccessor->getServantID(commandID);

		try {
			Servant* servant = ServantManager::getInstance()->getServant(servantID);

			// Create new Command Message and put it on the queue
			_messageQueue->put(new DurationRequestMessage(databaseID, servant->_getCommand(commandID), durationRequestExecutionID));

		} catch (Exception& e) {
			// failure
			if (!_shutdownCalled) {
				_processorManagerProxy->durationFailure(commandID, durationRequestExecutionID, "no such servant");
			}
			return;
		}

	} catch (Exception& e) {

		err << e.printStack() << eol;
		// failure!
		if (!_shutdownCalled) {
			_processorManagerProxy->durationFailure(commandID, durationRequestExecutionID, "cannot execute command");
		}
	}
}

void Processor::shutdown() {

	_shutdownCalled = true;

	// Create new shutdown Message and put it on the queue
	_messageQueue->put(new ShutdownMessage());
}

bool Processor::isShuttingDown() const {

	return _shutdownCalled;
}

std::string Processor::getCurrentStack() const {
	return "";
}

void Processor::init(unsigned long processorID, ProcessorManagerProxy* processorManagerProxy, ServantAccessor* servantAccessor, CommandAccessor* commandAccessor) {

	_processorID = processorID;

	// initialise proxy
	_processorManagerProxy = processorManagerProxy;

	// initialise accessors
	_servantAccessor = servantAccessor;
	_commandAccessor = commandAccessor;

	// initialise the queue with a big size. The queue should never be full!
	_messageQueue = new Queue<Message>(1024);
}

void Processor::reset() {

	delete _messageQueue;
}

void Processor::run() {

	Message* message = 0;

	_running = true;
	while (_running) {

		// Wait for a new message to arrive
		try {
			message = _messageQueue->get();

		} catch (Queue<Message>::UnavailableException& e) {
			break;

		} catch (Exception& e) {
			err << e.printStack() << eol;
			break;
		}

		// Process the message
		try {
			processMessage(message);

		} catch (Exception& e) {
			err << e.printStack() << eol;

		} catch (exception& e) {
			err << "STL exception : " << e.what() << "\n";

		} catch (...) {
			err << "Unknown exception" << "\n";
		}

		// Clean up
		delete message;
		message = 0;
 	}
}

void Processor::processMessage(Message* message) {

	if (message->getType() == Message::INIT_SERVANTS) {
		processMessage(static_cast<InitServantsMessage*>(message));

	} else if (message->getType() == Message::COMMAND) {
		processMessage(static_cast<CommandMessage*>(message));

	} else if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<ShutdownMessage*>(message));

	} else if (message->getType() == Message::ICS_SET_VALUE) {
		processMessage(static_cast<IcsSetValueMessage*>(message));

	} else if (message->getType() == Message::DURATION_REQUEST) {
		processMessage(static_cast<DurationRequestMessage*>(message));
	}

}

void Processor::processMessage(InitServantsMessage* message) {

	bool result = false;

	// set the database ID
	DatabaseIDKeeper::setID(message->getDatabaseID());

	try {
		// execute the initialisation
		result = ServantManager::getInstance()->initServants(message->verify(), message->driversAndProtocolsOnly());

	} catch (...) {

		out << "Stopped execution of initialisation" << eol;
	}

	message->getQueue()->put(new bool(result));
}

void Processor::processMessage(CommandMessage* message) {

	BaseCommand* command = message->getCommand();
	unsigned long commandExecutionID = message->getCommandExecutionID();

	// set the database ID
	DatabaseIDKeeper::setID(message->getDatabaseID());

	try {

		// execute the command
		command->execute(false, message->skipIfWait());

		if (!_shutdownCalled) {
			_processorManagerProxy->commandSuccess(command->getID(), commandExecutionID);
		}

	} catch (TraceableException& e) {

		err << "Error in execution of " << _commandAccessor->what(command->getID()) << eol;

		if (!_shutdownCalled) {
			_processorManagerProxy->commandFailure(command->getID(), commandExecutionID, e.printCallStack());
		}

	} catch (common::Exception& e) {

		err << "Exception in execution of " << _commandAccessor->what(command->getID()) << eol;
		out << "Stack : " << eol << getCurrentStack() << eol;

		if (!_shutdownCalled) {
			_processorManagerProxy->commandFailure(command->getID(), commandExecutionID, e.printStack());
		}

	} catch (exception& e) {

		err << "Exception in execution of " << _commandAccessor->what(command->getID()) << " : " << e.what() << eol;
		out << "Stack : " << eol << getCurrentStack() << eol;

		if (!_shutdownCalled) {
			_processorManagerProxy->commandFailure(command->getID(), commandExecutionID, e.what());
		}

	} catch (...) {

		err << "Exception in execution of " << _commandAccessor->what(command->getID()) << eol;
		out << "Stack : " << eol << getCurrentStack() << eol;

		if (!_shutdownCalled) {
			_processorManagerProxy->commandFailure(command->getID(), commandExecutionID, "");
		}
	}

	// reset the database ID
	DatabaseIDKeeper::setID(0);
}

void Processor::processMessage(DurationRequestMessage* message) {
}

void Processor::processMessage(ShutdownMessage* message) {

	_running = false;
}

void Processor::processMessage(IcsSetValueMessage* message) {

	// set the database ID
	DatabaseIDKeeper::setID(message->getDatabaseID());

	try {
		controller::IcsController* controller = dynamic_cast<controller::IcsController*>(message->getServant());
		if (controller != 0) {
			controller->setValue(message->getPropertyID(), message->getValue());
		}

		driver::IcsDriver* driver = dynamic_cast<driver::IcsDriver*>(message->getServant());

		if (driver != 0) {
			driver->setValue(message->getPropertyID(), message->getValue());
		}

		if (!Processor::getInstance()->isShuttingDown()) {
			_processorManagerProxy->propertySuccess(message->getPropertyID(), message->getPropertySetterID());
		}

	} catch (command::TraceableException& e) {

		out << "Error in execution of setter for ics " << _commandAccessor->what(message->getPropertyID()) << eol;

		if (!_shutdownCalled) {
			_processorManagerProxy->propertyFailure(message->getPropertyID(), message->getPropertySetterID(), e.printCallStack());
		}

	} catch (common::Exception& e) {

		out << "Exception in execution of setter for ics " << _commandAccessor->what(message->getPropertyID()) << " : " << e.getMessage() << eol;

		if (!_shutdownCalled) {
			_processorManagerProxy->propertyFailure(message->getPropertyID(), message->getPropertySetterID(), e.printStack());
		}

	} catch (exception& e) {

		out << "Exception " << typeid(e).name() << " in execution of setter for ics " << _commandAccessor->what(message->getPropertyID()) << " : " << e.what() << eol;

		if (!_shutdownCalled) {
			_processorManagerProxy->propertyFailure(message->getPropertyID(), message->getPropertySetterID(), e.what());
		}

	} catch (...) {

		out << "Exception in execution of setter for ics " << _commandAccessor->what(message->getPropertyID()) << eol;

		if (!_shutdownCalled) {
			_processorManagerProxy->propertyFailure(message->getPropertyID(), message->getPropertySetterID(), "");
		}
	}

	// reset the database ID
	DatabaseIDKeeper::setID(0);
}

}
