/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_PROCESSOR_H_
#define COMMAND_PROCESSOR_H_

#include <common/base/Queue.h>
#include <command/transportlayer/command/ProcessorManagerProxy.h>
#include <command/base/ServantManager.h>
#include <command/base/Servant.h>
#include <command/message/processor/IcsSetValueMessage.h>
#include <common/base/HashMap.h>

namespace command {

class Message;
class InitServantsMessage;
class CommandMessage;
class DurationRequestMessage;
class ShutdownMessage;
class ProcessorManagerProxy;
class ServantAccessor;
class CommandAccessor;
class LogAccessor;

/**
 * Class defining a processor. Is responsible to the execution of a command or a setter. Catches
 * the eventual exceptions and notifies the execution failure or success.
 */
class Processor {

public:
	/**
	 * Initialises the instance.
	 */
	static void initInstance(unsigned long processorID, ProcessorManagerProxy* processorManagerProxy, ServantAccessor* servantAccessor, CommandAccessor* commandAccessor);

	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Gets the instance using the current thread processor ID. The instance must already be initialised by a call to initInstance().
	 */
	static Processor * getInstance();

	/**
	 * Gets the instance using the processorID. The instance must already be initialised by a call to initInstance().
	 */
	static Processor * getInstance(unsigned long processorID);

	/**
	 * Sets the Processor ID.
	 */
	void setProcessorID(unsigned long processorID);

	/**
	 * Gets the Processor ID.
	 */
	unsigned long getProcessorID() const;

	/**
	 * Starts the message queue thread.
	 */
	void run();

	/**
	 * Returns true if the verification successes.
	 */
	bool initServants(unsigned long databaseID, bool verify);

	/**
	 * Initialises the drivers and protocols.
	 */
	void initDriversAndProtocols();

	/**
	 * Executes the command with ID commandID and unique commandExecutionID to allow the Processor Manager
	 * to distinguish between different commands.
	 */
	void execute(unsigned long databaseID, unsigned long commandID, bool skipIfWait, unsigned long commandExecutionID);

	/**
	 * Stops the command with ID commandID.
	 */
	void stop(unsigned long databaseID, unsigned long commandOrPropertyID);

	/**
	 * Pauses the command with ID commandID.
	 */
	void pause(unsigned long databaseID, unsigned long commandOrPropertyID);

	/**
	 * Restarts the command with ID commandID.
	 */
	void restart(unsigned long databaseID, unsigned long commandOrPropertyID);

	/**
	 * Executes the duration request with ID commandID and unique durationRequestExecutionID to allow the Processor Manager
	 * to distinguish between different commands.
	 */
	void executeDurationRequest(unsigned long databaseID, unsigned long commandID, unsigned long durationRequestExecutionID);

	/**
	 * Sets the value. Need to have a template method to set typed values.
	 */
	template<typename ValueType>
	void setValue(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, const ValueType& value);

	/**
	 * Gets the current stack.
	 */
	std::string getCurrentStack() const;

	/**
	 * Shuts down the Processor.
	 */
	void shutdown();

	/**
	 * Returns true if is shutting down.
	 */
	bool isShuttingDown() const;

private:
	Processor();
	~Processor();

	void init(unsigned long processorID, ProcessorManagerProxy* processorManagerProxy, ServantAccessor* servantAccessor, CommandAccessor* commandAccessor);
	void reset();

	void processMessage(Message* message);
	void processMessage(InitServantsMessage* message);
	void processMessage(CommandMessage* message);
	void processMessage(DurationRequestMessage* message);
	void processMessage(ShutdownMessage* message);
	void processMessage(IcsSetValueMessage* message);

	static void processStringBadSize(std::string const & message);
	static void processStringIndexOutOfBounds(std::string const & message);
	static void processStringSTLError(std::string const & message);
	static void processVectorIndexOutOfBounds(std::string const & message);
	static void processVectorCrossProductUndefined(std::string const & message);
	static void processMatrixRowIndexOutOfBounds(std::string const & message);
	static void processMatrixNotInvertible(std::string const & message);

	static std::mutex _instancesMutex;
	static common::HashMap<unsigned long, Processor*> _instances;

	unsigned long _processorID;

	ProcessorManagerProxy* _processorManagerProxy;
	ServantAccessor* _servantAccessor;
	CommandAccessor* _commandAccessor;

	common::Queue<Message>* _messageQueue;
	bool _running;
	bool _shutdownCalled;
};

}

#include "Processor.inl"

#endif /*COMMAND_PROCESSOR_H_*/
