/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_PROCESSORITEM_H_
#define COMMAND_PROCESSORITEM_H_

#include <common/base/Date.h>

namespace command {

class ProcessorProxy;
class OperationItem;

/**
 * Class defining a processor item used by the processor manager to store information
 * relative to the processor.
 */
class ProcessorItem {

public:
	/**
	 * Constructor.
	 */
	ProcessorItem(unsigned long ID);
	
	/**
	 * Copy constructor. The Processor Proxy and CommandItem address are not copied!
	 */
	ProcessorItem(const ProcessorItem& processorItem);
	
	/**
	 * Virtual destructor. The Processor Proxy and CommandItem are not deleted.
	 */
	virtual ~ProcessorItem();

	/**
	 * Gets the ID.
	 */
	unsigned long getID() const;
	
	/**
	 * Gets the PID.
	 */
	unsigned long getPID() const;
	
	/**
	 * Return true if the Processor is initialising.
	 */
	bool isInitialising() const;
	
	/**
	 * Return true if the Processor is ready.
	 */
	bool isReady() const;
	
	/**
	 * Return true if the Processor is on error.
	 */
	bool onError() const;

	/**
	 * Return true if the Processor is shutting down.
	 */
	bool isShuttingDown() const;
	
	/**
	 * Return true if the Processor is active.
	 */
	bool isActive() const;
	
	/**
	 * Gets the Processor Proxy.
	 */
	ProcessorProxy* getProxy() const;
	
	/**
	 * Gets the Operation Item.
	 */
	OperationItem* getOperationItem() const;
	
	/**
	 * Gets the last operation date.
	 */
	const common::Date& getLastOperationDate() const;

	/**
	 * Gets the last inactive date.
	 */
	const common::Date& getLastInactiveDate() const;
	
	/**
	 * Sets the PID.
	 */
	void setPID(unsigned long  PID);
	
	/**
	 * Sets to ready.
	 */
	void setReady();
	
	/**
	 * Sets on error.
	 */
	void setOnError();
	
	/**
	 * Sets the processor shutting down.
	 */
	void setShuttingDown();
	
	/**
	 * Sets to active or not.
	 */
	void setActive(bool isActive);
	
	/**
	 * Sets the Processor Proxy.
	 */
	void setProxy(ProcessorProxy* proxy);
	
	/**
	 * Sets the Operation Item.
	 */
	void setOperationItem(OperationItem* operationItem);
	
	/**
	 * Updates the last operation date.
	 */
	void updateLastOperationDate();

private:
	unsigned long _ID;
	unsigned long _PID;
	
	typedef enum {INIT, READY, ON_ERROR, SHUTTING_DOWN} State;
	State _state;
	bool _isActive;
			
	ProcessorProxy* _proxy;
	OperationItem* _operationItem;
	
	common::Date _lastOperationDate;
	common::Date _lastInactiveDate;
};

}

#endif /*COMMAND_PROCESSORITEM_H_*/