/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ProcessorManager.h"
#include "CommandItem.h"
#include "EnableDisableCommandItem.h"
#include "DurationRequestItem.h"
#include "InitDriversAndProtocolsItem.h"
#include <command/message/processormanager/RegisterProcessorMessage.h>
#include <command/message/processormanager/ExecuteCommandMessage.h>
#include <command/message/processormanager/StopCommandMessage.h>
#include <command/message/processormanager/PauseCommandMessage.h>
#include <command/message/processormanager/RestartCommandMessage.h>
#include <command/message/processormanager/StopSetterMessage.h>
#include <command/message/processormanager/PauseSetterMessage.h>
#include <command/message/processormanager/RestartSetterMessage.h>
#include <command/message/processormanager/CommandSuccessMessage.h>
#include <command/message/processormanager/CommandFailureMessage.h>
#include <command/message/processormanager/PropertySuccessMessage.h>
#include <command/message/processormanager/PropertyFailureMessage.h>
#include <command/message/processormanager/InitDriversAndProtocolsMessage.h>
#include <command/message/processormanager/ExecuteDurationRequestMessage.h>
#include <command/message/processormanager/DurationRequestSuccessMessage.h>
#include <command/message/processormanager/DurationRequestFailureMessage.h>
#include <command/message/processormanager/InitTerminatedMessage.h>
#include <command/message/processormanager/DeleteProcessorMessage.h>
#include <command/message/processormanager/AbortProcessorCreationMessage.h>
#include <command/message/processormanager/FinaliseShutdownMessage.h>
#include <command/message/ShutdownMessage.h>
#include <command/application/ProcessorManagerInterface.h>
#include <command/transportlayer/core/CompletionHandler.h>
#include <command/transportlayer/command/ProcessorManagerMessenger.h>
#include <common/output/OutputStream.h>
#include <common/base/Queue.h>
#include <common/base/ServerProperties.h>
#include <common/base/Convert.h>
#include <sys/types.h>
#include <signal.h>
#include <iostream>
#include <thread>

using namespace common;
using namespace std;

namespace command {

const std::string ProcessorManager::MIN_NUMBER_OF_PROCESSORS = "minNumberOfProcessors";
const std::string ProcessorManager::MAX_NUMBER_OF_PROCESSORS = "maxNumberOfProcessors";

const std::string ProcessorManager::PROCESSOR_LIFE_MANAGER_PERIODMS = "processorLifeManagerPeriodms";
const std::string ProcessorManager::PROCESSOR_CREATION_TIMEOUTMS = "processorCreationTimeoutms";
const std::string ProcessorManager::PROCESSOR_INACTIVITY_TIMEOUTMS = "processorInactivityTimeoutms";

const int ProcessorManager::EXECUTE_MESSAGE_PRIORITY = 2;
const int ProcessorManager::STOP_MESSAGE_PRIORITY = 2;
const int ProcessorManager::PAUSE_MESSAGE_PRIORITY = 2;
const int ProcessorManager::RESTART_MESSAGE_PRIORITY = 2;
const int ProcessorManager::EXECUTE_DURATION_REQUEST_MESSAGE_PRIORITY = 2;
const int ProcessorManager::SET_VALUE_MESSAGE_PRIORITY = 2;
const int ProcessorManager::INIT_DRIVERS_AND_PROTOCOLS_MESSAGE_PRIORITY = 2;
const int ProcessorManager::REGISTER_PROCESSOR_MESSAGE_PRIORITY = 1;
const int ProcessorManager::COMMAND_SUCCESS_MESSAGE_PRIORITY = 1;
const int ProcessorManager::COMMAND_FAILURE_MESSAGE_PRIORITY = 1;
const int ProcessorManager::PROPERTY_SUCCESS_MESSAGE_PRIORITY = 1;
const int ProcessorManager::PROPERTY_FAILURE_MESSAGE_PRIORITY = 1;
const int ProcessorManager::DURATION_REQUEST_SUCCESS_MESSAGE_PRIORITY = 1;
const int ProcessorManager::DURATION_REQUEST_FAILURE_MESSAGE_PRIORITY = 1;
const int ProcessorManager::INIT_TERMINATED_MESSAGE_PRIORITY = 1;
const int ProcessorManager::DELETE_PROCESSOR_MESSAGE_PRIORITY = 1;
const int ProcessorManager::SHUTDOWN_MESSAGE_PRIORITY = 0;
const int ProcessorManager::ABORT_PROCESSOR_CREATION_MESSAGE_PRIORITY = 1;


ProcessorManager * ProcessorManager::_instance = 0;

ProcessorManager::InitFailureException::InitFailureException(const std::string& message) :
	Exception("command::ProcessorManager::InitFailureException", message) {
}

ProcessorManager::NoSuchProcessorException::NoSuchProcessorException(unsigned long processorID) :
	Exception("command::ProcessorManager::NoSuchProcessorException", "unable to find a Processor with id <" + to<string>(processorID) + ">") {
}

ProcessorManager::ProcessorManager() :
	_running(false),
	_currentProcessorID(1),
	_currentCommandExecutionID(0),
	_currentSetterExecutionID(0),
	_currentDurationRequestExecutionID(0),
	_minNumberOfProcessors(0),
	_maxNumberOfProcessors(0),
	_started(false),
	_shuttingDown(false),
	_processorLifeManagerPeriodms(0),
	_processorInactivityTimeoutms(0),
	_processorManagerMessenger(0),
	_processorManagerInterface(0) {
}

ProcessorManager::~ProcessorManager() {
}

void ProcessorManager::initInstance(ProcessorManagerMessenger* processorManagerMessenger, ProcessorManagerInterface* processorManagerInterface) {

	if (_instance == 0) {
		_instance = new ProcessorManager();
		_instance->init(processorManagerMessenger, processorManagerInterface);
	}
}

void ProcessorManager::resetInstance() {

	_instance->reset();
	delete _instance;
}

ProcessorManager * ProcessorManager::getInstance() {

	return _instance;
}


void ProcessorManager::init(ProcessorManagerMessenger* processorManagerMessenger, ProcessorManagerInterface* processorManagerInterface) {

	_processorManagerMessenger = processorManagerMessenger;
	_processorManagerInterface = processorManagerInterface;

	// Initialising the queue with a big size. The queue should never be full!
	_messageQueue = new Queue<Message>(1024);

	// Processors initial instances and max number
	_maxNumberOfProcessors = 64;

	try {
		_maxNumberOfProcessors = ServerProperties::getInstance()->get<unsigned int>(MAX_NUMBER_OF_PROCESSORS);
	} catch (Exception& e) {
		out << "Setting default value 64 to property 'maxNumberOfProcessors'" << eol;
	}

	_minNumberOfProcessors = 8;

	try {
		_minNumberOfProcessors = ServerProperties::getInstance()->get<int>(MIN_NUMBER_OF_PROCESSORS);
	} catch (Exception& e) {
		out << "Setting default value 8 to property 'minNumberOfProcessors'" << eol;
	}

	// All communications initialised correctly. Now create an initial processor.
	createProcessorItem("for the initialisation");
}

void ProcessorManager::run() {

	out << startup << "Processor manager running" << eol;

	Message* message = 0;

	_running = true;

	std::thread* verifyProcessorTimeoutThread = new std::thread(&ProcessorManager::verifyProcessorTimeoutLoop);

	while (_running) {

		try {
			try {
				message = _messageQueue->get();

			} catch (Queue<Message>::UnavailableException& e) {
				break;

			} catch (Exception& e) {
				err << e.printStack() << eol;
				break;
			}

			try {
				processMessage(message);

			} catch (Exception& e) {
				err << e.printStack() << eol;
			}

		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}

		delete message;
		message = 0;
 	}

 	verifyProcessorTimeoutThread->join();
 	delete verifyProcessorTimeoutThread;

}

void ProcessorManager::verifyProcessorTimeoutLoop() {

	_instance->_processorLifeManagerPeriodms = 1000;
	int processorCreationTimeoutms = 10000;
	_instance->_processorInactivityTimeoutms = 100;

	try {
		_instance->_processorLifeManagerPeriodms = ServerProperties::getInstance()->get<int>(PROCESSOR_LIFE_MANAGER_PERIODMS);
	} catch (Exception& e) {
		out << "Setting default value " << _instance->_processorLifeManagerPeriodms << " to property 'processorLifeManagerPeriodms'" << eol;
	}

	try {
		processorCreationTimeoutms = ServerProperties::getInstance()->get<int>(PROCESSOR_CREATION_TIMEOUTMS);
	} catch (Exception& e) {
		out << "Setting default value " << processorCreationTimeoutms << " to property 'processorCreationTimeoutms'" << eol;
	}

	try {
		_instance->_processorInactivityTimeoutms = ServerProperties::getInstance()->get<int>(PROCESSOR_INACTIVITY_TIMEOUTMS);
	} catch (Exception& e) {
		out << "Setting default value " << _instance->_processorInactivityTimeoutms << " to property 'processorInactivityTimeoutms'" << eol;
	}

	bool finalised = false;

	while (_instance->_running && !finalised) {

		try {

			list<ProcessorItem*> processorsToRemove;
			list<ProcessorItem*> processorsToDelete;
			list<unsigned long> processorCreationToAbortIDs;

			Date now;

			// iterating the processor items
			{
				std::scoped_lock lock(_instance->_processorsMutex);

				for (HashMap<unsigned long, ProcessorItem*>::iterator p =_instance-> _processorMap.begin(); p != _instance->_processorMap.end(); p++) {

					ProcessorItem* processorItem = p->second;

					int delay = (int)(now - processorItem->getLastInactiveDate()).getMilliseconds();

					// testing the availability of the processor
					if (processorItem->getProxy() != 0 && !processorItem->getProxy()->isAlive()) {

						// setting the processor item on error
						processorItem->setOnError();

						if (!processorItem->isShuttingDown() && processorItem->isActive()) {
							// the processor was executing a command
							OperationItem* operationItem = processorItem->getOperationItem();

							if (dynamic_cast<CommandItem*>(operationItem) != 0) {

								CommandItem* commandItem = static_cast<CommandItem*>(operationItem);
								commandItem->getHandler()->commandFailure(commandItem->getRequestID(), commandItem->getCommandID(), "the processor is not responding");

							} else if (dynamic_cast<BaseSetterItem*>(operationItem) != 0) {

								BaseSetterItem* setterItem = static_cast<BaseSetterItem*>(operationItem);
								setterItem->getHandler()->propertyFailure(setterItem->getRequestID(), setterItem->getPropertyID(), "the processor is not responding");
							}
						}
					}

					// testing the state of the processor item
					if (processorItem->isReady()) {

						if (!processorItem->isActive() && delay > _instance->_processorInactivityTimeoutms) {
							// a processor has not been used recently
							// deleting only if the current number of ready processors is greater than
							// the min number
							if (_instance->getNumberOfReadyProcessors() > _instance->_minNumberOfProcessors) {

								// setting the processor shutting down
								processorItem->setShuttingDown();
								processorItem->getProxy()->shutdown();
							}
						}

					} else if (processorItem->isShuttingDown()) {

						// testing that the processor can terminate
						if (processorItem->getProxy()->canTerminate()) {
							processorItem->getProxy()->terminate();

							// deleting the ProcessorItem
							processorsToRemove.push_back(processorItem);
						}

					} else if (processorItem->onError()) {

						// deleting the ProcessorItem
						processorsToDelete.push_back(processorItem);

					} else if (processorItem->isInitialising()) {
						// Do nothing.
					}
				}

				// test for shutdown
				if (_instance->_shuttingDown && _instance->_processorMap.getSize() == 0) {
					// stopping the queue wich awakes the message queue and terminates
					_instance->finaliseShutdown();
					finalised = true;
				}
			}

			// Now the mutex is not blocking in case of a queue full
			for (list<ProcessorItem*>::const_iterator p = processorsToRemove.begin(); p != processorsToRemove.end(); p++) {

				ProcessorItem* processorItem = *p;
				{
					std::scoped_lock lock(_instance->_processorsMutex);
					_instance->_processorMap.erase(processorItem->getID());
				}

				// terminate the processor
				_instance->terminateProcessor(processorItem->getPID());
								
				delete processorItem;
			}

			// Now the mutex is not blocking in case of a queue full
			for (list<ProcessorItem*>::const_iterator p = processorsToDelete.begin(); p != processorsToDelete.end(); p++) {

				ProcessorItem* processorItem = *p;
				{
					std::scoped_lock lock(_instance->_processorsMutex);
					_instance->_processorMap.erase(processorItem->getID());
				}

				// terminate the processor
				_instance->terminateProcessor(processorItem->getPID());
				_instance->deleteProcessor(processorItem);
			}

			for (list<unsigned long>::const_iterator i = processorCreationToAbortIDs.begin(); i != processorCreationToAbortIDs.end(); i++) {
				_instance->abortProcessorCreation(*i);
			}

			//out << "waiting " << _instance->_processorLifeManagerPeriodms << eol;

			// Wait until the next cycle
			Date end = now + Duration(_instance->_processorLifeManagerPeriodms, Duration::MILLISECONDS);
			_instance->_lifeManagerCondition.wait(end);

		} catch (exception& e) {
			err << "STL exception : " << e.what() << eol;
		} catch (common::Exception const & e) {
			err << "Exception : " << e.printStack() << eol;
		} catch (...) {
			err << "Unexpected exception" << eol;
		}
	}
}

void ProcessorManager::post(Message* message, int priority) {
	_messageQueue->put(message, priority);
}

void ProcessorManager::executeCommand(unsigned long requestID, unsigned long databaseID, unsigned long commandID, bool skipIfWait, CompletionHandler* completionHandler) {

	post(new ExecuteCommandMessage(requestID, databaseID, commandID, skipIfWait, completionHandler), EXECUTE_MESSAGE_PRIORITY);
}

void ProcessorManager::stopCommand(unsigned long databaseID, unsigned long commandID) {

	post(new StopCommandMessage(databaseID, commandID), STOP_MESSAGE_PRIORITY);
}

void ProcessorManager::pauseCommand(unsigned long databaseID, unsigned long commandID) {

	post(new PauseCommandMessage(databaseID, commandID), PAUSE_MESSAGE_PRIORITY);
}

void ProcessorManager::restartCommand(unsigned long databaseID, unsigned long commandID) {

	post(new RestartCommandMessage(databaseID, commandID), RESTART_MESSAGE_PRIORITY);
}

void ProcessorManager::stopSetter(unsigned long databaseID, unsigned long propertyID) {

	post(new StopSetterMessage(databaseID, propertyID), STOP_MESSAGE_PRIORITY);
}

void ProcessorManager::pauseSetter(unsigned long databaseID, unsigned long propertyID) {

	post(new PauseSetterMessage(databaseID, propertyID), PAUSE_MESSAGE_PRIORITY);
}

void ProcessorManager::restartSetter(unsigned long databaseID, unsigned long propertyID) {

	post(new RestartSetterMessage(databaseID, propertyID), RESTART_MESSAGE_PRIORITY);
}

void ProcessorManager::initDriversAndProtocols() {

	post(new InitDriversAndProtocolsMessage(), INIT_DRIVERS_AND_PROTOCOLS_MESSAGE_PRIORITY);
}

void ProcessorManager::executeDurationRequest(unsigned long requestID, unsigned long databaseID, unsigned long commandID, CompletionHandler* completionHandler) {

	post(new ExecuteDurationRequestMessage(requestID, databaseID, commandID, completionHandler), EXECUTE_DURATION_REQUEST_MESSAGE_PRIORITY);
}

void ProcessorManager::registerProcessor(ProcessorProxy* processorProxy, unsigned long PID) {

	post(new RegisterProcessorMessage(processorProxy, PID), REGISTER_PROCESSOR_MESSAGE_PRIORITY);
}

void ProcessorManager::commandSuccess(unsigned long processorID, unsigned long commandID, unsigned long commandExecutionID) {

	post(new CommandSuccessMessage(processorID, commandID, commandExecutionID), COMMAND_SUCCESS_MESSAGE_PRIORITY);
}

void ProcessorManager::commandFailure(unsigned long processorID, unsigned long commandID, unsigned long commandExecutionID, const string& message) {

	post(new CommandFailureMessage(processorID, commandID, commandExecutionID, message), COMMAND_FAILURE_MESSAGE_PRIORITY);
}

void ProcessorManager::propertySuccess(unsigned long processorID, unsigned long propertyID, unsigned long propertySetterID) {

	post(new PropertySuccessMessage(processorID, propertyID, propertySetterID), PROPERTY_SUCCESS_MESSAGE_PRIORITY);
}

void ProcessorManager::propertyFailure(unsigned long processorID, unsigned long propertyID, unsigned long propertySetterID, const string& message) {

	post(new PropertyFailureMessage(processorID, propertyID, propertySetterID, message), PROPERTY_FAILURE_MESSAGE_PRIORITY);
}

void ProcessorManager::durationRequestSuccess(unsigned long processorID, unsigned long commandID, unsigned long durationRequestExecutionID, double duration) {

	post(new DurationRequestSuccessMessage(processorID, commandID, durationRequestExecutionID, duration), DURATION_REQUEST_SUCCESS_MESSAGE_PRIORITY);
}

void ProcessorManager::durationRequestFailure(unsigned long processorID, unsigned long commandID, unsigned long durationRequestExecutionID, const string& message) {

	post(new DurationRequestFailureMessage(processorID, commandID, durationRequestExecutionID, message), DURATION_REQUEST_FAILURE_MESSAGE_PRIORITY);
}

void ProcessorManager::initTerminated(unsigned long processorID) {

	post(new InitTerminatedMessage(processorID), INIT_TERMINATED_MESSAGE_PRIORITY);
}

void ProcessorManager::deleteProcessor(ProcessorItem* processorItem) {

	post(new DeleteProcessorMessage(processorItem), DELETE_PROCESSOR_MESSAGE_PRIORITY);
}

void ProcessorManager::abortProcessorCreation(unsigned long processorID) {

	post(new AbortProcessorCreationMessage(processorID), ABORT_PROCESSOR_CREATION_MESSAGE_PRIORITY);
}

void ProcessorManager::shutdown() {

	// Sending a message to stop the thread
	post(new ShutdownMessage(), SHUTDOWN_MESSAGE_PRIORITY);
}

void ProcessorManager::finaliseShutdown() {

	post(new FinaliseShutdownMessage(), SHUTDOWN_MESSAGE_PRIORITY);
}

void ProcessorManager::reset() {
	delete _messageQueue;
}

bool ProcessorManager::isProcessorResponding(unsigned long processorID) {

	std::scoped_lock lock(_processorsMutex);

	ProcessorProxy* proxy = 0;

	try {
		proxy = _processorMap.get(processorID)->getProxy();
	} catch (Exception& e) {

		throw link(NoSuchProcessorException(processorID), e);
	}

	if (proxy != 0)
		return proxy->isAlive();

	return false;
}

std::string ProcessorManager::getProcessorCallStack(unsigned long processorID) {

	std::scoped_lock lock(_processorsMutex);

	ProcessorProxy* proxy = 0;

	try {
		proxy = _processorMap.get(processorID)->getProxy();
	} catch (Exception& e) {

		throw link(NoSuchProcessorException(processorID), e);
	}

	if (proxy != 0)
		return proxy->getCurrentStack();

	return "";
}

std::vector<ProcessorItem*> ProcessorManager::getProcessors() {

	vector<ProcessorItem*> result;

	std::scoped_lock lock(_processorsMutex);

	for (HashMap<unsigned long, ProcessorItem*>::const_iterator p = _processorMap.begin(); p != _processorMap.end(); p++) {
		if (p->second->isReady()) {
			result.push_back(new ProcessorItem(*(p->second)));
		}
	}

	return result;
}

OperationItem::Type ProcessorManager::getOperationType(unsigned long processorID) {
	std::scoped_lock lock(_processorsMutex);

	if (_processorMap.find(processorID)) {
		ProcessorItem* processorItem = _processorMap[processorID];
		OperationItem* operationItem = processorItem->getOperationItem();
		if (operationItem != 0) {
			return operationItem->getType();
		}
	}
	return OperationItem::COMMAND;
}

unsigned long ProcessorManager::getOperationID(unsigned long processorID) {
	std::scoped_lock lock(_processorsMutex);

	if (_processorMap.find(processorID)) {
		ProcessorItem* processorItem = _processorMap[processorID];
		OperationItem* operationItem = processorItem->getOperationItem();
		if (operationItem != 0) {

			if (operationItem->getType() == OperationItem::COMMAND) {
				return dynamic_cast<CommandItem*>(operationItem)->getCommandID();

			} else if (operationItem->getType() == OperationItem::PROPERTY){
				return dynamic_cast<BaseSetterItem*>(operationItem)->getPropertyID();
			}
		}
	}

	return 0;

}

void ProcessorManager::processMessage(Message* message) {

	if (message->getType() == Message::EXECUTE_COMMAND) {
		processMessage(static_cast<ExecuteCommandMessage*>(message));

	} else if (message->getType() == Message::STOP_COMMAND) {
		processMessage(static_cast<StopCommandMessage*>(message));

	} else if (message->getType() == Message::PAUSE_COMMAND) {
		processMessage(static_cast<PauseCommandMessage*>(message));

	} else if (message->getType() == Message::RESTART_COMMAND) {
		processMessage(static_cast<RestartCommandMessage*>(message));

	} else if (message->getType() == Message::STOP_SETTER) {
		processMessage(static_cast<StopSetterMessage*>(message));

	} else if (message->getType() == Message::INIT_DRIVERS_AND_PROTOCOLS) {
		processMessage(static_cast<InitDriversAndProtocolsMessage*>(message));

	} else if (message->getType() == Message::REGISTER_PROCESSOR) {
		processMessage(static_cast<RegisterProcessorMessage*>(message));

	} else if (message->getType() == Message::EXECUTE_DURATION_REQUEST) {
		processMessage(static_cast<ExecuteDurationRequestMessage*>(message));

	} else if (message->getType() == Message::COMMAND_SUCCESS) {
		processMessage(static_cast<CommandSuccessMessage*>(message));

	} else if (message->getType() == Message::COMMAND_FAILURE) {
		processMessage(static_cast<CommandFailureMessage*>(message));

	} else if (message->getType() == Message::PROPERTY_SUCCESS) {
		processMessage(static_cast<PropertySuccessMessage*>(message));

	} else if (message->getType() == Message::PROPERTY_FAILURE) {
		processMessage(static_cast<PropertyFailureMessage*>(message));

	} else if (message->getType() == Message::DURATION_REQUEST_SUCCESS) {
		processMessage(static_cast<DurationRequestSuccessMessage*>(message));

	} else if (message->getType() == Message::DURATION_REQUEST_FAILURE) {
		processMessage(static_cast<DurationRequestFailureMessage*>(message));

	} else if (message->getType() == Message::INIT_TERMINATED) {
		processMessage(static_cast<InitTerminatedMessage*>(message));

	} else if (message->getType() == Message::DELETE_PROCESSOR) {
		processMessage(static_cast<DeleteProcessorMessage*>(message));

	} else if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<ShutdownMessage*>(message));

	} else if (message->getType() == Message::FINALISE_SHUTDOWN) {
		processMessage(static_cast<FinaliseShutdownMessage*>(message));

	} else if (message->getType() == Message::ABORT_PROCESSOR_CREATION) {
		processMessage(static_cast<AbortProcessorCreationMessage*>(message));

	} else if (message->getType() == Message::SET_VALUE) {
		processMessage(static_cast<BaseExecuteSetValueMessage*>(message));

	}

}

void ProcessorManager::processMessage(ExecuteCommandMessage* message) {

	if (_shuttingDown) {
		out << "Trying to execute command " << message->getCommandID() << " but refused because server is shutting down" << eol;
		return;
	}

	std::scoped_lock lock(_processorsMutex);

	CommandItem* commandItem = new CommandItem(message->getRequestID(), message->getDatabaseID(), message->getCommandID(), message->skipIfWait(), _currentCommandExecutionID++, message->getCompletionHandler());
	_pendingOperations.push_back(commandItem);

	// Determine if we have an inactive Processor
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;

		// verify that the processor proxy has been set
		if (processorItem->getProxy() != 0) {
			if (processorItem->isReady()) {
				if (processorItem->getProxy()->isAlive()) {
					if (!processorItem->isActive()) {
						break;
					}
				} else {
					out << "Processor with PID " << processorItem->getPID() << " is not responding" << eol;
				}
			}
		}
	}

	if (p != _processorMap.end()) {
		executePendingOperation(p->second);
	} else {
		createProcessorItem("to execute a new command");
	}
}

void ProcessorManager::processMessage(InterruptCommandMessage* message) {

	std::scoped_lock lock(_processorsMutex);

	// finding the Processor which executes the command
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;

		if (processorItem->isActive()) {

			OperationItem* operationItem = processorItem->getOperationItem();
			CommandItem* commandItem = dynamic_cast<CommandItem*>(operationItem);

			if (commandItem != 0) {

				if (commandItem->getCommandID() == message->getCommandID() &&
					commandItem->getDatabaseID() == message->getDatabaseID()) {

					// verify that the processor proxy has been set
					if (processorItem->getProxy() != 0) {
						if (processorItem->isReady()) {
							if (processorItem->getProxy()->isAlive()) {
								if (!processorItem->isActive()) {
									// should not happen!
									err << "Error : processor with PID " << processorItem->getPID() << " is already inactive" << eol;
								} else {
									if (message->getType() == Message::STOP_COMMAND) {
										stopCommand(p->second);
									} else if (message->getType() == Message::PAUSE_COMMAND) {
										pauseCommand(p->second);
									} else if (message->getType() == Message::RESTART_COMMAND) {
										restartCommand(p->second);
									}
								}
							} else {
								out << "Processor with PID " << processorItem->getPID() << " is not responding" << eol;
							}
						}
					}
				}
			}
		}
	}
}

void ProcessorManager::processMessage(InterruptSetterMessage* message) {

	std::scoped_lock lock(_processorsMutex);

	// finding the Processor which executes the command
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;

		if (processorItem->isActive()) {

			OperationItem* operationItem = processorItem->getOperationItem();
			BaseSetterItem* setterItem = dynamic_cast<BaseSetterItem*>(operationItem);

			if (setterItem != 0) {

				if (setterItem->getPropertyID() == message->getPropertyID() &&
					setterItem->getDatabaseID() == message->getDatabaseID()) {

					// verify that the processor proxy has been set
					if (processorItem->getProxy() != 0) {
						if (processorItem->isReady()) {
							if (processorItem->getProxy()->isAlive()) {
								if (!processorItem->isActive()) {
									// should not happen!
									err << "Processor with PID " << processorItem->getPID() << " is already inactive" << eol;
								} else {
									if (message->getType() == Message::STOP_SETTER) {
										stopSetter(p->second);
									} else if (message->getType() == Message::PAUSE_SETTER) {
										pauseSetter(p->second);
									} else if (message->getType() == Message::RESTART_SETTER) {
										restartSetter(p->second);
									}
								}
							} else {
								out << "Processor with PID " << processorItem->getPID() << " is not responding" << eol;
							}
						}
					}
				}
			}
		}
	}
}

void ProcessorManager::processMessage(InitDriversAndProtocolsMessage* message) {

	std::scoped_lock lock(_processorsMutex);

	InitDriversAndProtocolsItem* initDriversAndProtocolsItem = new InitDriversAndProtocolsItem();
	_pendingOperations.push_back(initDriversAndProtocolsItem);

	// Determine if we have an inactive Processor
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;

		// verify that the processor proxy has been set
		if (processorItem->getProxy() != 0) {
			if (processorItem->isReady()) {
				if (processorItem->getProxy()->isAlive()) {
					if (!processorItem->isActive()) {
						break;
					}
				} else {
					out << "Processor with PID " << processorItem->getPID() << " is not responding" << eol;
				}
			}
		}
	}

	if (p != _processorMap.end()) {
		executePendingOperation(p->second);
	} else {
		createProcessorItem("to execute an init drivers and protocols request");
	}
}

void ProcessorManager::processMessage(ExecuteDurationRequestMessage* message) {

	std::scoped_lock lock(_processorsMutex);

	DurationRequestItem* durationRequestItem = new DurationRequestItem(message->getRequestID(), message->getDatabaseID(), message->getCommandID(), _currentDurationRequestExecutionID++, message->getCompletionHandler());
	_pendingOperations.push_back(durationRequestItem);

	// Determine if we have an inactive Processor
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;

		// verify that the processor proxy has been set
		if (processorItem->getProxy() != 0) {
			if (processorItem->isReady()) {
				if (processorItem->getProxy()->isAlive()) {
					if (!processorItem->isActive()) {
						break;
					}
				} else {
					out << "Processor with PID " << processorItem->getPID() << " is not responding" << eol;
				}
			}
		}
	}

	if (p != _processorMap.end()) {
		executePendingOperation(p->second);
	} else {
		createProcessorItem("to execute a duration request");
	}
}

void ProcessorManager::processMessage(BaseExecuteSetValueMessage* message) {

	if (dynamic_cast<ExecuteSetValueMessage<int8>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<int8>::apply(static_cast<ExecuteSetValueMessage<int8>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<uint16>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<uint16>::apply(static_cast<ExecuteSetValueMessage<uint16>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<uint32>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<uint32>::apply(static_cast<ExecuteSetValueMessage<uint32>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<uint64>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<uint64>::apply(static_cast<ExecuteSetValueMessage<uint64>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<int16>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<int16>::apply(static_cast<ExecuteSetValueMessage<int16>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<int32>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<int32>::apply(static_cast<ExecuteSetValueMessage<int32>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<int64>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<int64>::apply(static_cast<ExecuteSetValueMessage<int64>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<float32>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<float32>::apply(static_cast<ExecuteSetValueMessage<float32>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<float64>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<float64>::apply(static_cast<ExecuteSetValueMessage<float64>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<bool>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<bool>::apply(static_cast<ExecuteSetValueMessage<bool>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<string>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<string>::apply(static_cast<ExecuteSetValueMessage<string>*>(message));
	} else if (dynamic_cast<ExecuteSetValueMessage<std::string>*>(message) != 0) {
		ProcessorManagerSetValueMessageHelper<std::string>::apply(static_cast<ExecuteSetValueMessage<std::string>*>(message));
	}

}

void ProcessorManager::processMessage(RegisterProcessorMessage* message) {

	std::scoped_lock lock(_processorsMutex);

	ProcessorProxy* processorProxy = message->getProcessorProxy();
	unsigned long processorID = processorProxy->getProcessorID();

	// Obtain the associated ProcessorItem from the list of Processors
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->getID() == processorID) {
			break;
		}
	}

	if (p == _processorMap.end()) {
		// Create a new Processor item and add it to the map
		ProcessorItem* newProcessorItem = new ProcessorItem(_currentProcessorID);
		newProcessorItem->setActive(false);

		// set the processor ID
		processorProxy->setProcessorID(_currentProcessorID);

		// insert new processor and get iterator relating to it
		_processorMap.insert(_currentProcessorID, newProcessorItem);
		for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
			if (p->second->getID() == _currentProcessorID) {
				break;
			}
		}
		_currentProcessorID++;
	}

	// Processor Item has been found: set it to ready and execute command
	// if one exists on the queue.
	if (p != _processorMap.end()) {

		p->second->setPID(message->getPID());
		p->second->setProxy(processorProxy);
		p->second->setReady();

		// the command processor application is ready
		if (!_started) {
			if (getNumberOfReadyProcessors() == 1) {
				// calling initServants once
				// blocking call
				bool initialised = processorProxy->initServants(0, true);
				if (initialised) {
					// Now create an initial pool of processors.
					for (unsigned int i = 1; i < _minNumberOfProcessors; i++) {
						createProcessorItem("for the initial pool");
					}

				} else {
					err << "Failed to initialise the servants" << eol;
					_processorManagerInterface->setOnError();
					return;
				}
			}

			if (getNumberOfReadyProcessors() >= _minNumberOfProcessors) {
				_started = true;
				_processorManagerInterface->setStarted();
			}
		}

		executePendingOperation(p->second);
	}
}

void ProcessorManager::processMessage(ShutdownMessage* message) {

	// setting the shutting down state and the related attributes
	_shuttingDown = true;

	// shutdown all Processors
	shutdownProcessors();
}

void ProcessorManager::processMessage(FinaliseShutdownMessage* message) {

	// shutdown creator
	try {
		_processorManagerMessenger->shutdown();

	} catch (Exception& e) {
		err << e.printStack() << eol;
		err << "Cannot shutdown the processor creator" << eol;
	}

	_running = false;
}

void ProcessorManager::processMessage(CommandSuccessMessage* message) {

	CommandItem* theCommandItem = 0;
	ProcessorItem* theProcessorItem = 0;

	std::scoped_lock lock(_processorsMutex);

	// Obtain the associated ProcessorItem from the list of Processors
	// that executed the command
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->isActive()) {

			CommandItem* commandItem = dynamic_cast<CommandItem*>(processorItem->getOperationItem());
			if ((commandItem != 0) && (commandItem->getCommandExecutionID() == message->getCommandExecutionID())) {
				theCommandItem = commandItem;
				theProcessorItem = processorItem;
				break;
			}
		}
	}

	// ProcessorItem associated with command has been found: notify scheduler of
	// successful command termination
	if (theCommandItem != 0) {
		theCommandItem->getHandler()->commandSuccess(theCommandItem->getRequestID(), message->getCommandID());
		theProcessorItem->setOperationItem(0);
		theProcessorItem->setActive(false);

		delete theCommandItem;

		executePendingOperation(theProcessorItem);

	} else {
		err << "Got a CommandSuccessMessage but was unable to find associated processor with ID " << message->getProcessorID() << eol;

	}
}

void ProcessorManager::processMessage(CommandFailureMessage* message) {

	CommandItem* theCommandItem = 0;
	ProcessorItem* theProcessorItem = 0;

	std::scoped_lock lock(_processorsMutex);

	// Obtain the associated ProcessorItem from the list of Processors
	// that executed the command
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->isActive()) {

			CommandItem* commandItem = dynamic_cast<CommandItem*>(processorItem->getOperationItem());
			if ((commandItem != 0) && (commandItem->getCommandExecutionID() == message->getCommandExecutionID())) {
				theCommandItem = commandItem;
				theProcessorItem = processorItem;
				break;
			}
		}
	}

	// Processor Item associated with command has been found: notify scheduler of
	// successful command termination
	if (theCommandItem != 0) {
		theCommandItem->getHandler()->commandFailure(theCommandItem->getRequestID(), message->getCommandID(), message->getMessage());
		theProcessorItem->setOperationItem(0);
		theProcessorItem->setActive(false);

		delete theCommandItem;

		// Can launch another pending command
		executePendingOperation(theProcessorItem);

	} else {
		err << "Got a CommandFailureMessage but was unable to find associated processor with ID " << message->getProcessorID() << eol;

	}
}

void ProcessorManager::processMessage(PropertySuccessMessage* message) {

	BaseSetterItem* theSetterItem = 0;
	ProcessorItem* theProcessorItem = 0;

	std::scoped_lock lock(_processorsMutex);

	// Obtain the associated ProcessorItem from the list of Processors
	// that executed the command
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->isActive()) {

			BaseSetterItem* setterItem = dynamic_cast<BaseSetterItem*>(processorItem->getOperationItem());
			if ((setterItem != 0) && (setterItem->getSetterID() == message->getSetterID())) {
				theSetterItem = setterItem;
				theProcessorItem = processorItem;
				break;
			}
		}
	}

	// ProcessorItem associated with command has been found: notify scheduler of
	// successful command termination
	if (theSetterItem != 0) {
		theSetterItem->getHandler()->propertySuccess(theSetterItem->getRequestID(), message->getPropertyID());
		theProcessorItem->setOperationItem(0);
		theProcessorItem->setActive(false);

		delete theSetterItem;

		executePendingOperation(theProcessorItem);

	} else {
		err << "Got a PropertySuccessMessage but was unable to find associated processor with ID " << message->getProcessorID() << eol;

	}
}

void ProcessorManager::processMessage(PropertyFailureMessage* message) {

	BaseSetterItem* theSetterItem = 0;
	ProcessorItem* theProcessorItem = 0;

	std::scoped_lock lock(_processorsMutex);

	// Obtain the associated ProcessorItem from the list of Processors
	// that executed the command
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->isActive()) {

			BaseSetterItem* setterItem = dynamic_cast<BaseSetterItem*>(processorItem->getOperationItem());
			if ((setterItem != 0) && (setterItem->getSetterID() == message->getSetterID())) {
				theSetterItem = setterItem;
				theProcessorItem = processorItem;
				break;
			}
		}
	}

	// ProcessorItem associated with command has been found: notify scheduler of
	// successful command termination
	if (theSetterItem != 0) {
		theSetterItem->getHandler()->propertyFailure(theSetterItem->getRequestID(), message->getPropertyID(), message->getMessage());
		theProcessorItem->setOperationItem(0);
		theProcessorItem->setActive(false);

		delete theSetterItem;

		executePendingOperation(theProcessorItem);

	} else {
		err << "Got a PropertyFailureMessage but was unable to find associated processor with ID " << message->getProcessorID() << eol;

	}
}

void ProcessorManager::processMessage(DurationRequestSuccessMessage* message) {

	DurationRequestItem* theDurationRequestItem = 0;
	ProcessorItem* theProcessorItem = 0;

	std::scoped_lock lock(_processorsMutex);

	// Obtain the associated ProcessorItem from the list of Processors
	// that executed the command
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->isActive()) {

			DurationRequestItem* durationRequestItem = dynamic_cast<DurationRequestItem*>(processorItem->getOperationItem());
			if ((durationRequestItem != 0) && (durationRequestItem->getDurationRequestExecutionID() == message->getDurationRequestExecutionID())) {
				theDurationRequestItem = durationRequestItem;
				theProcessorItem = processorItem;
				break;
			}
		}
	}

	// ProcessorItem associated with command has been found: notify scheduler of
	// successful duration request termination
	if (theDurationRequestItem != 0) {
		theDurationRequestItem->getHandler()->durationRequestSuccess(theDurationRequestItem->getRequestID(), message->getCommandID(), message->getDuration());
		theProcessorItem->setOperationItem(0);
		theProcessorItem->setActive(false);

		delete theDurationRequestItem;

		executePendingOperation(theProcessorItem);

	} else {
		err << "Got a DurationRequestSuccessMessage but was unable to find associated processor with ID " << message->getProcessorID() << eol;

	}
}

void ProcessorManager::processMessage(DurationRequestFailureMessage* message) {

	DurationRequestItem* theDurationRequestItem = 0;
	ProcessorItem* theProcessorItem = 0;

	std::scoped_lock lock(_processorsMutex);

	// Obtain the associated ProcessorItem from the list of Processors
	// that executed the command
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->isActive()) {

			DurationRequestItem* durationRequestItem = dynamic_cast<DurationRequestItem*>(processorItem->getOperationItem());
			if ((durationRequestItem != 0) && (durationRequestItem->getDurationRequestExecutionID() == message->getDurationRequestExecutionID())) {
				theDurationRequestItem = durationRequestItem;
				theProcessorItem = processorItem;
				break;
			}
		}
	}

	// Processor Item associated with command has been found: notify scheduler of
	// successful command termination
	if (theDurationRequestItem != 0) {
		theDurationRequestItem->getHandler()->durationRequestFailure(theDurationRequestItem->getRequestID(), message->getCommandID(), message->getMessage());
		theProcessorItem->setOperationItem(0);
		theProcessorItem->setActive(false);

		delete theDurationRequestItem;

		// Can launch another pending command
		executePendingOperation(theProcessorItem);

	} else {
		err << "Got a DurationRequestFailureMessage but was unable to find associated processor with ID " << message->getProcessorID() << eol;

	}
}

void ProcessorManager::processMessage(InitTerminatedMessage* message) {

	std::scoped_lock lock(_processorsMutex);

	ProcessorItem* theProcessorItem = 0;

	// Obtain the associated ProcessorItem from the list of Processors
	// that executed the command
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->isActive() && p->first == message->getProcessorID()) {
			theProcessorItem = processorItem;
			delete processorItem->getOperationItem();
			theProcessorItem->setOperationItem(0);

			break;
		}
	}

	// Processor Item associated with command has been found
	if (theProcessorItem != 0) {
		theProcessorItem->setActive(false);

		// Can launch another pending command
		executePendingOperation(theProcessorItem);

	} else {
		err << "Got a InitTerminatedMessage but was unable to find associated processor with ID " << message->getProcessorID() << eol;

	}
}

void ProcessorManager::processMessage(DeleteProcessorMessage* message) {

	ProcessorItem* processorItem = message->getProcessorItem();

	if (processorItem->onError()) {
		out << "Processor with PID " << processorItem->getPID() << " is not responding" << eol;
	}

	delete processorItem;

	// creating a processor if necessary
	if (getNumberOfReadyProcessors() < _minNumberOfProcessors) {
		createProcessorItem("to replace a processor on error");
	}
}

void ProcessorManager::processMessage(AbortProcessorCreationMessage* message) {

	std::scoped_lock lock(_processorsMutex);

	// Obtain the associated ProcessorItem from the list of Processors
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->getID() == message->getProcessorID()) {
			break;
		}
	}

	// Processor Item has been found: delete it from the list of available
	// processors if it is still not 'ready' (ie a register message arrived just
	// after the abort message was created).
	if (p != _processorMap.end()) {
		ProcessorItem* processorItem = p->second;
		if (!processorItem->isReady()) {
			_processorMap.erase(p->first);

			delete processorItem;
			out << "Aborted processor creation for ID " << message->getProcessorID() << eol;
		}

	} else {
		err << "Got an AbortProcessorCreationMessage from an unknown processor ID " << message->getProcessorID() << eol;
	}

}


void ProcessorManager::shutdownProcessors() {

	// Each call to proxy->shutdown posts a UnregisterProcessorMessage which could be
	// executed in parallel with this method, BUT the _processorsMutex lock sequentialises these
	// calls.
	std::scoped_lock lock(_processorsMutex);

	list<ProcessorItem*> processorsNotResponding;

	// Send stop message to all Processors.
	HashMap<unsigned long, ProcessorItem*>::iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {

		ProcessorItem* processorItem = p->second;
		ProcessorProxy* proxy = processorItem->getProxy();

		if (proxy != 0 && proxy->isAlive()) {

			// set the processor item shutting down
			processorItem->setShuttingDown();

			// shutdown the proxy
			proxy->shutdown();
		}
	}
}


void ProcessorManager::createProcessorItem(const std::string& reason) {

	if (_processorMap.getSize() < _maxNumberOfProcessors) {

		// For the moment, just creating a new processor
		ProcessorItem* newProcessor = new ProcessorItem(_currentProcessorID);
		newProcessor->setActive(false);

		_processorMap.insert(_currentProcessorID, newProcessor);

		_processorManagerMessenger->sendCreate(_currentProcessorID);
		_currentProcessorID++;

	} else {
		err << "Cannot create a new processor, max number of Processors already reached (" << _maxNumberOfProcessors << ")" << eol;
	}
}

void ProcessorManager::executePendingOperation(ProcessorItem* processorItem) {

	// See if a command is pending and execute it if there is one
	if (!_pendingOperations.empty()) {

		OperationItem* item = _pendingOperations.front();
		_pendingOperations.pop_front();

		if (dynamic_cast<CommandItem*>(item) != 0) {

			CommandItem* commandItem = static_cast<CommandItem*>(item);

			// Launching command
			processorItem->setActive(true);
			processorItem->updateLastOperationDate();
			processorItem->setOperationItem(commandItem);

			processorItem->getProxy()->execute(commandItem->getDatabaseID(), commandItem->getCommandID(), commandItem->skipIfWait(), commandItem->getCommandExecutionID());

		} else if (dynamic_cast<BaseSetterItem*>(item) != 0) {

			if (dynamic_cast<SetterItem<int8>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<int8>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<uint16>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<uint16>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<uint32>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<uint32>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<uint64>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<uint64>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<int16>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<int16>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<int32>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<int32>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<int64>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<int64>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<float32>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<float32>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<float64>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<float64>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<bool>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<bool>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<string>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<string>::apply(processorItem, item);
			} else if (dynamic_cast<SetterItem<std::string>*>(item) != 0) {
				ProcessorManagerExecutePendingSetterHelper<std::string>::apply(processorItem, item);
			}

		} else if (dynamic_cast<DurationRequestItem*>(item) != 0) {

			DurationRequestItem* durationRequestItem = static_cast<DurationRequestItem*>(item);

			// launching request
			processorItem->setActive(true);
			processorItem->updateLastOperationDate();
			processorItem->setOperationItem(durationRequestItem);

			processorItem->getProxy()->executeDurationRequest(durationRequestItem->getDatabaseID(), durationRequestItem->getCommandID(), durationRequestItem->getDurationRequestExecutionID());

		} else if (dynamic_cast<InitDriversAndProtocolsItem*>(item) != 0) {

			InitDriversAndProtocolsItem* initDriversAndProtocolsItem = static_cast<InitDriversAndProtocolsItem*>(item);

			// launching init
			processorItem->setActive(true);
			processorItem->updateLastOperationDate();
			processorItem->setOperationItem(initDriversAndProtocolsItem);

			processorItem->getProxy()->initDriversAndProtocols();
		}
	}
}

void ProcessorManager::stopCommand(ProcessorItem* processorItem) {

	CommandItem* commandItem = static_cast<CommandItem*>(processorItem->getOperationItem());
	processorItem->getProxy()->stop(commandItem->getDatabaseID(), commandItem->getCommandID());
}

void ProcessorManager::pauseCommand(ProcessorItem* processorItem) {

	CommandItem* commandItem = static_cast<CommandItem*>(processorItem->getOperationItem());
	processorItem->getProxy()->pause(commandItem->getDatabaseID(), commandItem->getCommandID());
}

void ProcessorManager::restartCommand(ProcessorItem* processorItem) {

	CommandItem* commandItem = static_cast<CommandItem*>(processorItem->getOperationItem());
	processorItem->getProxy()->restart(commandItem->getDatabaseID(), commandItem->getCommandID());
}

void ProcessorManager::stopSetter(ProcessorItem* processorItem) {

	BaseSetterItem* setterItem = static_cast<BaseSetterItem*>(processorItem->getOperationItem());
	processorItem->getProxy()->stop(setterItem->getDatabaseID(), setterItem->getPropertyID());
}

void ProcessorManager::pauseSetter(ProcessorItem* processorItem) {

	BaseSetterItem* setterItem = static_cast<BaseSetterItem*>(processorItem->getOperationItem());
	processorItem->getProxy()->pause(setterItem->getDatabaseID(), setterItem->getPropertyID());
}

void ProcessorManager::restartSetter(ProcessorItem* processorItem) {

	BaseSetterItem* setterItem = static_cast<BaseSetterItem*>(processorItem->getOperationItem());
	processorItem->getProxy()->restart(setterItem->getDatabaseID(), setterItem->getPropertyID());
}

unsigned int ProcessorManager::getNumberOfReadyProcessors() const {

	unsigned int nReady = 0;

	HashMap<unsigned long, ProcessorItem*>::const_iterator p;
	for (p = _processorMap.begin(); p != _processorMap.end(); p++) {
		ProcessorItem* processorItem = p->second;
		if (processorItem->isReady()) {
			nReady++;
		}
	}

	return nReady;

}

void ProcessorManager::terminateProcessor(unsigned long PID) {

	_processorManagerMessenger->sendTerminate(PID);
}

void ProcessorManagerExecutePendingSetterHelper<int8>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<int8>* setterItem = static_cast<SetterItem<int8>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setInt8Value(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<uint16>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<uint16>* setterItem = static_cast<SetterItem<uint16>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setUint16Value(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<uint32>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<uint32>* setterItem = static_cast<SetterItem<uint32>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setUint32Value(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<uint64>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<uint64>* setterItem = static_cast<SetterItem<uint64>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setUint64Value(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<int16>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<int16>* setterItem = static_cast<SetterItem<int16>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setInt16Value(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<int32>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<int32>* setterItem = static_cast<SetterItem<int32>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setInt32Value(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<int64>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<int64>* setterItem = static_cast<SetterItem<int64>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setInt64Value(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<float32>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<float32>* setterItem = static_cast<SetterItem<float32>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setFloat32Value(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<float64>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<float64>* setterItem = static_cast<SetterItem<float64>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setFloat64Value(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<bool>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<bool>* setterItem = static_cast<SetterItem<bool>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setBooleanValue(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

void ProcessorManagerExecutePendingSetterHelper<std::string>::apply(ProcessorItem* processorItem, OperationItem* item) {

	SetterItem<std::string>* setterItem = static_cast<SetterItem<std::string>*>(item);

	// Launching setter
	processorItem->setActive(true);
	processorItem->updateLastOperationDate();
	processorItem->setOperationItem(setterItem);

	processorItem->getProxy()->setStringValue(setterItem->getDatabaseID(), setterItem->getServantID(), setterItem->getPropertyID(), setterItem->getSetterID(), setterItem->getValue());
}

}
