/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_PROCESSORMANAGER_H_
#define COMMAND_PROCESSORMANAGER_H_

#include "OperationItem.h"
#include <common/base/Exception.h>
#include <common/base/HashMap.h>
#include <common/base/TimeCondition.h>
#include <command/message/processormanager/ExecuteSetValueMessage.h>
#include <vector>
#include <deque>

namespace common {

template<typename ItemType>
class Queue;

}

namespace command {

template<typename ValueType>
struct ProcessorManagerSetValueMessageHelper;

class Message;
class ExecuteCommandMessage;
class InterruptCommandMessage;
class RegisterProcessorMessage;
class InterruptSetterMessage;
class InitDriversAndProtocolsMessage;
class ExecuteDurationRequestMessage;
class CommandSuccessMessage;
class CommandFailureMessage;
class PropertySuccessMessage;
class PropertyFailureMessage;
class DurationRequestSuccessMessage;
class DurationRequestFailureMessage;
class InitTerminatedMessage;
class DeleteProcessorMessage;
class ShutdownMessage;
class FinaliseShutdownMessage;
class AbortProcessorCreationMessage;
class CompletionHandler;
class ProcessorProxy;
class ProcessorManagerMessenger;
class ProcessorItem;
class CommandItem;
class ProcessorManagerInterface;

/**
 * Class defining a processor manager. Receives requests from the Core component like command execution or setter execution.
 * The processor manager must find an inactive processor or create a new one to execute immediately the command.
 */
class ProcessorManager {

	template<typename ValueType>
	friend struct ProcessorManagerSetValueMessageHelper;

public:

	/**
	 * Exception for initialisation failure
	 */
	class InitFailureException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		InitFailureException(const std::string& message);
	};

	/**
	 * Exception for no such Processor.
	 */
	class NoSuchProcessorException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchProcessorException(unsigned long processorID);
	};

	/**
	 * Initialises the instance.
	 * throw InitFailureException
	 */
	static void initInstance(ProcessorManagerMessenger* processorManagerMessenger, ProcessorManagerInterface* processorManagerInterface);

	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Gets the instance. The instance must already be initialised by a call to initInstance().
	 */
	static ProcessorManager * getInstance();

	/**
	 * Runs the message thread and the verify loop thread.
	 */
	void run();


	/**
	 * Posts a message to execute a new command.
	 */
	void executeCommand(unsigned long requestID, unsigned long databaseID, unsigned long commandID, bool skipIfWait, CompletionHandler* completionHandler);

	/**
	 * Posts a message to stop a command.
	 */
	void stopCommand(unsigned long databaseID, unsigned long commandID);

	/**
	 * Posts a message to pause a command.
	 */
	void pauseCommand(unsigned long databaseID, unsigned long commandID);

	/**
	 * Posts a message to restart a command.
	 */
	void restartCommand(unsigned long databaseID, unsigned long commandID);

	/**
	 * Posts a message to stop a setter.
	 */
	void stopSetter(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Posts a message to pause a setter.
	 */
	void pauseSetter(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Posts a message to restart a setter.
	 */
	void restartSetter(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Initialises the drivers and the protocols.
	 */
	void initDriversAndProtocols();

	/**
	 * Posts a message to execute a new duration request.
	 */
	void executeDurationRequest(unsigned long requestID, unsigned long databaseID, unsigned long commandID, CompletionHandler* completionHandler);

	/**
	 * Sets a new value for the Property of ID propertyID.
	 */
	template<typename ValueType>
	void setValue(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, const ValueType& value, CompletionHandler* completionHandler) {

		post(new ExecuteSetValueMessage<ValueType>(requestID, databaseID, servantID, propertyID, value, completionHandler), SET_VALUE_MESSAGE_PRIORITY);
	}

	/**
	 * Posts a message to register a new Processor.
	 */
	void registerProcessor(ProcessorProxy* processorProxy, unsigned long PID);

	/**
	 * Posts a message to indicate the successful termination of a command.
	 */
	void commandSuccess(unsigned long processorID, unsigned long commandID, unsigned long commandExecutionID);

	/**
	 * Posts a message to indicate that a command terminated because of a failure.
	 */
	void commandFailure(unsigned long processorID, unsigned long commandID, unsigned long commandExecutionID, const std::string& message);

	/**
	 * Posts a message to indicate the successful termination of a property setter.
	 */
	void propertySuccess(unsigned long processorID, unsigned long propertyID, unsigned long propertySetterID);

	/**
	 * Posts a message to indicate that a property setter terminated because of a failure.
	 */
	void propertyFailure(unsigned long processorID, unsigned long propertyID, unsigned long propertySetterID, const std::string& message);

	/**
	 * Posts a message to indicate the successful termination of a duration request.
	 */
	void durationRequestSuccess(unsigned long processorID, unsigned long commandID, unsigned long durationRequestExecutionID, double duration);

	/**
	 * Posts a message to indicate that a duration request terminated because of a failure.
	 */
	void durationRequestFailure(unsigned long processorID, unsigned long commandID, unsigned long durationRequestExecutionID, const std::string& message);

	/**
	 * Posts a message to indicate that an initialisation has terminated.
	 */
	void initTerminated(unsigned long processorID);

	/**
	 * Posts a message to abort the creation of a processor.
	 */
	void abortProcessorCreation(unsigned long processorID);

	/**
	 * Send the shutdown message to the ProcessorManager loop of execution and
	 * ensures that all other processes (Processors and Processor Creator) are
	 * shutdown too.
	 */
	void shutdown();

	/**
	 * Gets the Processors.
	 */
	std::vector<ProcessorItem*> getProcessors();

	/**
	 * Gets the type of operation running on a processor.
	 */
	OperationItem::Type getOperationType(unsigned long processorID);

	/**
	 * Gets the type of operation ID of operation running on a processor.
	 */
	unsigned long getOperationID(unsigned long processorID);

	/**
	 * Returns true if the Processor with ID ID is responding.
	 */
	bool isProcessorResponding(unsigned long processorID);

	/**
	 * Gets the processor call stack.
	 */
	std::string getProcessorCallStack(unsigned long processorID);

	/**
	 * Terminates the Processor process.
	 */
	void terminateProcessor(unsigned long PID);

	/**
	 * Gets the number of ready processors.
	 */
	unsigned int getNumberOfReadyProcessors() const;

private:

	static const int EXECUTE_MESSAGE_PRIORITY;
	static const int STOP_MESSAGE_PRIORITY;
	static const int PAUSE_MESSAGE_PRIORITY;
	static const int RESTART_MESSAGE_PRIORITY;
	static const int EXECUTE_DURATION_REQUEST_MESSAGE_PRIORITY;
	static const int SET_VALUE_MESSAGE_PRIORITY;
	static const int INIT_DRIVERS_AND_PROTOCOLS_MESSAGE_PRIORITY;
	static const int REGISTER_PROCESSOR_MESSAGE_PRIORITY;
	static const int COMMAND_SUCCESS_MESSAGE_PRIORITY;
	static const int COMMAND_FAILURE_MESSAGE_PRIORITY;
	static const int PROPERTY_SUCCESS_MESSAGE_PRIORITY;
	static const int PROPERTY_FAILURE_MESSAGE_PRIORITY;
	static const int DURATION_REQUEST_SUCCESS_MESSAGE_PRIORITY;
	static const int DURATION_REQUEST_FAILURE_MESSAGE_PRIORITY;
	static const int INIT_TERMINATED_MESSAGE_PRIORITY;
	static const int DELETE_PROCESSOR_MESSAGE_PRIORITY;
	static const int SHUTDOWN_MESSAGE_PRIORITY;
	static const int ABORT_PROCESSOR_CREATION_MESSAGE_PRIORITY;

	static const std::string MIN_NUMBER_OF_PROCESSORS;
	static const std::string MAX_NUMBER_OF_PROCESSORS;

	static const std::string PROCESSOR_LIFE_MANAGER_PERIODMS;
	static const std::string PROCESSOR_CREATION_TIMEOUTMS;
	static const std::string PROCESSOR_INACTIVITY_TIMEOUTMS;

	static void verifyProcessorTimeoutLoop();
	static ProcessorManager * _instance;

	ProcessorManager();
	~ProcessorManager();

	/**
	 * throw InitFailureException
	 */
	void init(ProcessorManagerMessenger* processorManagerMessenger, ProcessorManagerInterface* processorManagerInterface);
	void reset();

	/**
	 * Posts a message.
	 */
	void post(Message* message, int priority);

	void processMessage(Message* message);
	void processMessage(ExecuteCommandMessage* message);
	void processMessage(InterruptCommandMessage* message);
	void processMessage(BaseExecuteSetValueMessage* message);
	void processMessage(InterruptSetterMessage* message);
	void processMessage(InitDriversAndProtocolsMessage* message);
	void processMessage(ExecuteDurationRequestMessage* message);
	void processMessage(RegisterProcessorMessage* message);
	void processMessage(CommandSuccessMessage* message);
	void processMessage(CommandFailureMessage* message);
	void processMessage(PropertySuccessMessage* message);
	void processMessage(PropertyFailureMessage* message);
	void processMessage(DurationRequestSuccessMessage* message);
	void processMessage(DurationRequestFailureMessage* message);
	void processMessage(InitTerminatedMessage* message);
	void processMessage(ShutdownMessage* message);
	void processMessage(FinaliseShutdownMessage* message);
	void processMessage(DeleteProcessorMessage* message);
	void processMessage(AbortProcessorCreationMessage* message);

	void shutdownProcessors();
	void finaliseShutdown();
	void createProcessorItem(const std::string& reason);
	void deleteProcessor(ProcessorItem* processorItem);

	void executePendingOperation(ProcessorItem* processorItem);

	void stopCommand(ProcessorItem* processorItem);
	void pauseCommand(ProcessorItem* processorItem);
	void restartCommand(ProcessorItem* processorItem);

	void stopSetter(ProcessorItem* processorItem);
	void pauseSetter(ProcessorItem* processorItem);
	void restartSetter(ProcessorItem* processorItem);

	bool _running;
	common::TimeCondition _lifeManagerCondition;

	common::Queue<Message>* _messageQueue;

	unsigned long _currentProcessorID;

	std::deque<OperationItem*> _pendingOperations;
	unsigned long _currentCommandExecutionID;
	unsigned long _currentSetterExecutionID;
	unsigned long _currentDurationRequestExecutionID;

	common::HashMap<unsigned long, ProcessorItem*> _processorMap;
	unsigned int _minNumberOfProcessors;
	unsigned int _maxNumberOfProcessors;
	bool _started;

	bool _shuttingDown;
	int _processorLifeManagerPeriodms;
	int _processorInactivityTimeoutms;

	ProcessorManagerMessenger* _processorManagerMessenger;
	ProcessorManagerInterface* _processorManagerInterface;

	std::mutex _processorsMutex;
};

template<typename ValueType>
struct ProcessorManagerSetValueMessageHelper {

	static void apply(ExecuteSetValueMessage<ValueType>* message);

};

template<typename ValueType>
struct ProcessorManagerExecutePendingSetterHelper;

template<>
struct ProcessorManagerExecutePendingSetterHelper<int8> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<uint16> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<uint32> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<uint64> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<int16> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<int32> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<int64> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<float32> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<float64> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<bool> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

template<>
struct ProcessorManagerExecutePendingSetterHelper<std::string> {

	static void apply(ProcessorItem* processorItem, OperationItem* item);

};

}

#include "ProcessorManager.inl"

#endif /*COMMAND_PROCESSORMANAGER_H_*/
