/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_THREADPROCESSORINFO_H_
#define COMMAND_THREADPROCESSORINFO_H_

#include <common/base/ThreadSpecificPointer.h>
#include <common/base/Exception.h>

namespace command {

/**
 * Class defining a singleton to access the Processor ID through the current thread.
 */
class ThreadProcessorInfo {
	
public:
	/**
	 * Exception for bad processor ID
	 */
	class BadProcessorID : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		BadProcessorID();
	};

	/**
	 * Gets the instance.
	 */
	static ThreadProcessorInfo * getInstance();
	
	/**
	 * Sets the processor ID.
	 */
	void setProcessorID(unsigned long processorID);

	/**
	 * Gets the processor ID.
	 */
	unsigned long getProcessorID();
	
	/**
	 * Releases the associated data.
	 */
	void release();

private:
	struct ProcessorID {
		ProcessorID() : ID(0) {}
		ProcessorID(unsigned long initValue) : ID(initValue) {}

		unsigned long ID;
	};
	
	common::ThreadSpecificPointer<ProcessorID> _processorIDs;
	
	static ThreadProcessorInfo * _instance;
};

/**
 * Class storing the processor info using the scoped object technique.
 */
class ScopedThreadProcessorInfo {

public:
	/**
	 * Constructor.
	 */
	explicit ScopedThreadProcessorInfo(unsigned long processorID);

	/**
	 * Destructor.
	 */
	~ScopedThreadProcessorInfo();
};

}

#endif /*COMMAND_THREADPROCESSORINFO_H_*/