/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ProcessorManagerMessengerDirect.h"
#include <command/message/ShutdownMessage.h>
#include <command/message/processorcreator/CreateProcessorMessage.h>
#include <command/message/processorcreator/TerminateProcessorMessage.h>
#include <command/application/ProcessorThread.h>
#include <common/base/ServerProperties.h>
#include <boost/bind.hpp>
#include <iostream>
#include <functional>

using namespace common;
using namespace std;

namespace command {

ProcessorManagerMessengerDirect::ProcessorManagerMessengerDirect() :
	_messageQueue(nullptr),
	_thread(nullptr),
	_running(false) {
}

ProcessorManagerMessengerDirect::~ProcessorManagerMessengerDirect() {
	delete _messageQueue;
}

void ProcessorManagerMessengerDirect::init() {
	_messageQueue = new Queue<Message>(1024);
}

void ProcessorManagerMessengerDirect::run() {
	
	_thread = new std::thread(std::bind(&ProcessorManagerMessengerDirect::loop, this));
}

void ProcessorManagerMessengerDirect::reset() {
	
	// wait for the server socket thread to terminate
	_thread->join();
	
	delete _thread;
}

void ProcessorManagerMessengerDirect::sendCreate(unsigned long processorID) {
	_messageQueue->put(new CreateProcessorMessage(processorID));
}

void ProcessorManagerMessengerDirect::sendTerminate(unsigned long processorID) {
	_messageQueue->put(new TerminateProcessorMessage(processorID));
}

void ProcessorManagerMessengerDirect::shutdown() {
	_messageQueue->put(new ShutdownMessage());
}

void ProcessorManagerMessengerDirect::loop(ProcessorManagerMessengerDirect* instance) {
	
	Message* message = 0;

	instance->_running = true;
	while (instance->_running) {

		// Wait for a new message to arrive
		try {
			message = instance->_messageQueue->get();

		} catch (Queue<Message>::UnavailableException& e) {
			break;

		} catch (Exception& e) {
			err << e.printStack() << eol;
			break;
		}

		// Process the message
		try {
			instance->processMessage(message);

		} catch (Exception& e) {
			err << e.printStack() << eol;

		} catch (exception& e) {
			err << "STL exception : " << e.what() << "\n";

		} catch (...) {
			err << "Unknown exception" << "\n";
		}

		// Clean up
		delete message;
		message = 0;
	}
}

void ProcessorManagerMessengerDirect::processMessage(Message* message) {

	if (message->getType() == Message::CREATE_PROCESSOR) {
		processMessage(static_cast<CreateProcessorMessage*>(message));

	} else if (message->getType() == Message::TERMINATE_PROCESSOR) {
		processMessage(static_cast<TerminateProcessorMessage*>(message));

	} else if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<ShutdownMessage*>(message));

	}
}

void ProcessorManagerMessengerDirect::processMessage(CreateProcessorMessage* message) {

	unsigned long processorID = message->getProcessorID();

	ProcessorThread * processorThread = new ProcessorThread(processorID);
	_processorThreadMap.insert(processorID, processorThread);

	// start the thread
	processorThread->start();
}

void ProcessorManagerMessengerDirect::processMessage(TerminateProcessorMessage* message) {

	// get the processor thread
	try {
		_processorThreadMap.get(message->getPID())->join();

	} catch (Exception const & e) {
		err << "Unable to find processor thread <" << message->getPID() << "> : " << e.printStack() << eol;
	}
}

void ProcessorManagerMessengerDirect::processMessage(ShutdownMessage* message) {
	_running = false;
}

}
