/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_PROCESSORPROXY_H_
#define COMMAND_PROCESSORPROXY_H_

#include <common/base/Exception.h>
#include <common/base/PropertyType.h>

namespace command {
	
/**
 * Class defining a processor proxy interface.
 */
class ProcessorProxy {
	
public:
	/**
	 * Constructor.
	 */
	ProcessorProxy();
	
	/**
	 * Virtual destructor.
	 */
	virtual ~ProcessorProxy();
	
	/**
	 * Sets the Processor ID.
	 */
	virtual void setProcessorID(unsigned long processorID) = 0;
	
	/**
	 * Gets the Processor ID.
	 */
	virtual unsigned long getProcessorID() = 0;
	
	/**
	 * Returns true if is alive.
	 */
	virtual bool isAlive() = 0;
	
	/**
	 * Returns true if can terminate.
	 */
	virtual bool canTerminate() = 0;
	
	/**
	 * Terminates the processor.
	 */
	virtual void terminate() = 0;
	
	/**
	 * Returns true if the verification successes.
	 */
	virtual bool initServants(unsigned long databaseID, bool verify) = 0;
	
	/**
	 * Initialises the drivers and protocols.
	 */
	virtual void initDriversAndProtocols() = 0;
	
	/**
	 * Gets the current stack.
	 */
	virtual std::string getCurrentStack() = 0;
	
	/**
	 * Executes the Command.
	 */
	virtual void execute(unsigned long databaseID, unsigned long commandID, bool skipIfWait, unsigned long commandExecutionID) = 0;
	
	/**
	 * Stops the Command.
	 */
	virtual void stop(unsigned long databaseID, unsigned long commandID) = 0;
	
	/**
	 * Pauses the Command.
	 */
	virtual void pause(unsigned long databaseID, unsigned long commandID) = 0;

	/**
	 * Restarts the Command.
	 */
	virtual void restart(unsigned long databaseID, unsigned long commandID) = 0;

	/**
	 * Executes the Duration request.
	 */
	virtual void executeDurationRequest(unsigned long databaseID, unsigned long commandID, unsigned long durationRequestExecutionID) = 0;
	
	/**
	 * Shut downs the Processor.
	 */
	virtual void shutdown() = 0;
	
	/**
	 * Sets the int8 value for the property of ID propertyID.
	 */
	virtual void setInt8Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, int8 value) const = 0;

	/**
	 * Sets the int16 value for the property of ID propertyID.
	 */
	virtual void setUint16Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, uint16 value) const = 0;

	/**
	 * Sets the int32 value for the property of ID propertyID.
	 */
	virtual void setUint32Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, uint32 value) const = 0;

	/**
	 * Sets the int64 value for the property of ID propertyID.
	 */
	virtual void setUint64Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, uint64 value) const = 0;
		
	/**
	 * Sets the int16 value for the property of ID propertyID.
	 */
	virtual void setInt16Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, int16 value) const = 0;
		
	/**
	 * Sets the int32 value for the property of ID propertyID.
	 */
	virtual void setInt32Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, int32 value) const = 0;
		
	/**
	 * Sets the int64 value for the property of ID propertyID.
	 */
	virtual void setInt64Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, int64 value) const = 0;
		
	/**
	 * Sets the float32 value for the property of ID propertyID.
	 */
	virtual void setFloat32Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, float32 value) const = 0;
	
	/**
	 * Sets the float64 value for the property of ID propertyID.
	 */
	virtual void setFloat64Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, float64 value) const = 0;
	
	/**
	 * Sets the string value for the property of ID propertyID.
	 */
	virtual void setStringValue(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, const std::string& value) const = 0;
	
	/**
	 * Sets the bool value for the property of ID propertyID.
	 */
	virtual void setBooleanValue(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, bool value) const = 0;

};

}

#endif /*COMMAND_PROCESSORPROXY_H_*/
