/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_PROCESSORPROXYDIRECTIMPL_H_
#define COMMAND_PROCESSORPROXYDIRECTIMPL_H_

#include "ProcessorProxyImpl.h"
#include <boost/thread/mutex.hpp>
#include <common/base/BaseTypes.h>
#include <string>

namespace command {

/**
 * Class defining a direct processor proxy implementation.
 */
class ProcessorProxyDirectImpl : public ProcessorProxyImpl {
	
public:
	/**
	 * Constructor.
	 */
	ProcessorProxyDirectImpl();
	
	/**
	 * Virtual destructor.
	 */
	virtual ~ProcessorProxyDirectImpl();
	
	/**
	 * Sets the Processor ID.
	 */
	virtual void setProcessorID(unsigned long processorID);
	
	/**
	 * Gets the Processor ID.
	 */
	virtual unsigned long getProcessorID();
	
	/**
	 * Returns true if is alive.
	 */
	virtual bool isAlive();
	
	/**
	 * Returns true if can terminate.
	 */
	virtual bool canTerminate();
	
	/**
	 * Terminates the processor.
	 */
	virtual void terminate();
	
	/**
	 * Returns true if the verification successes.
	 */
	virtual bool initServants(unsigned long databaseID, bool verify);
	
	/**
	 * Initialises the drivers and the protocols.
	 */
	virtual void initDriversAndProtocols();
	
	/**
	 * Gets the current stack.
	 */
	virtual std::string getCurrentStack();
	
	/**
	 * Executes the Command.
	 */
	virtual void execute(unsigned long databaseID, unsigned long commandID, bool skipIfWait, unsigned long commandExecutionID);
	
	/**
	 * Stops the Command.
	 */
	virtual void stop(unsigned long databaseID, unsigned long commandOrPropertyID);
	
	/**
	 * Pauses the Command.
	 */
	virtual void pause(unsigned long databaseID, unsigned long commandOrPropertyID);
	
	/**
	 * Restarts the Command.
	 */
	virtual void restart(unsigned long databaseID, unsigned long commandOrPropertyID);
	
	/**
	 * Executes the Command.
	 */
	virtual void executeDurationRequest(unsigned long databaseID, unsigned long commandID, unsigned long durationRequestExecutionID);

	/**
	 * Shutdowns the Processor.
	 */
	virtual void shutdown();
	
	/**
	 * Sets the int8 value.
	 */
	virtual void setInt8Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, int8 value);

	/**
	 * Sets the int16 value.
	 */
	virtual void setUint16Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, uint16 value);

	/**
	 * Sets the int32 value.
	 */
	virtual void setUint32Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, uint32 value);

	/**
	 * Sets the int64 value.
	 */
	virtual void setUint64Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, uint64 value);
		
	/**
	 * Sets the int16 value.
	 */
	virtual void setInt16Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, int16 value);
		
	/**
	 * Sets the int32 value.
	 */
	virtual void setInt32Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, int32 value);
		
	/**
	 * Sets the int64 value.
	 */
	virtual void setInt64Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, int64 value);
		
	/**
	 * Sets the float32 value.
	 */
	virtual void setFloat32Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, float32 value);
		
	/**
	 * Sets the float64 value.
	 */
	virtual void setFloat64Value(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, float64 value);
		
	/**
	 * Sets the string value.
	 */
	virtual void setStringValue(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, const std::string& value);
		
	/**
	 * Sets the boolean value.
	 */
	virtual void setBooleanValue(unsigned long databaseID, unsigned long servantID, unsigned long propertyID, unsigned long propertySetterID, bool value);

	/**
	 * Sets the thread Processor ID.
	 */
	virtual void setThreadProcessorID(unsigned long processorID);

private:
	unsigned long _processorID;
};

}

#endif /*COMMAND_PROCESSORPROXYDIRECTIMPL_H_*/
