/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_COMMANDACCESSOR_H_
#define COMMAND_COMMANDACCESSOR_H_

#include <command/exception/TraceableException.h>
#include <command/base/method/BaseCommand.h>
#include <command/base/DataHolder.h>

#include <string>
#include <vector>
#include <common/utils/CommandDescriptor.h>

namespace command {

/**
 * Abstract class defining the command accessor.
 */
class CommandAccessor {

public:
	/**
	 * Command list creation exception.
	 */
	class CommandListCreationException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		CommandListCreationException(const std::string& type, unsigned long ID);
	};

	/**
	 * No such servant exception.
	 */
	class NoSuchServantException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NoSuchServantException(unsigned long servantID);
	};

	/**
	 * No such command exception.
	 */
	class NoSuchCommandException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NoSuchCommandException(unsigned long commandID);
	};

	/**
	 * Constructor
	 */
	CommandAccessor();

	/**
	 * Virtual destructor.
	 */
	virtual ~CommandAccessor();

	/**
	 * Gets the commands of the controller of controllerID ID.
	 * throw CommandListCreationException, NoSuchServantException
	 */
	virtual std::vector<common::CommandDescriptor> getControllerCommands(unsigned long controllerID) const = 0;

	/**
	 * Gets the commands of the driver of driverID ID.
	 * throw CommandListCreationException, NoSuchServantException
	 */
	virtual std::vector<common::CommandDescriptor> getDriverCommands(unsigned long driverID) const = 0;


	/**
	 * Gets the servant ID corresponding to the commmand ID.
	 * throw NoSuchCommandException
	 */
	virtual unsigned long getServantID(unsigned long commandID) const = 0;

	/**
	 * Sets the command state.
	 * throw NoSuchCommandException
	 */
	virtual void setCommandState(unsigned long servantID, unsigned long commandID, common::CommandState state) = 0;

	/**
	 * Sets the command state.
	 * throw NoSuchCommandException
	 */
	virtual common::SetCommandActiveResult setCommandActive(DataHolder::Type type, unsigned long servantID, unsigned long commandID, const std::string& servantType) = 0;

	/**
	 * Returns true if the command of ID commandID has terminated.
	 */
	virtual bool hasTerminated(unsigned long commandID) = 0;

	/**
	 * Finishes for the termination of the command of ID commandID.
	 */
	virtual void finishTermination(unsigned long commandID) = 0;

	/**
	 * Gets the command state.
	 * throw NoSuchCommandException
	 */
	virtual common::CommandState getCommandState(unsigned long commandID) const = 0;

	/**
	 * Returns the elapsed time of a command.
	 * throw NoSuchCommandException
	 */
	virtual double getCommandElapsedTimeMs(unsigned long commandID) const = 0;

	/**
	 * Gets the command start date.
	 */
	virtual int64 getCommandStartDate(unsigned long commandID) const = 0;

	/**
	 * Adds the command to the servant.
	 * throw NoSuchServantException
	 */
	virtual unsigned long addCommandToServant(unsigned long servantID, const std::string& commandName) = 0;

	/**
	 * Sets the progression of the command.
	 * throw NoSuchCommandException
	 */
	virtual void setProgression(unsigned long commandID, double progression) const = 0;

	/**
	 * Sets the progression unknown.
	 * throw NoSuchCommandException
	 */
	virtual void setUnknownProgression(unsigned long commandID) const = 0;

	/**
	 * Gets the progression of the command.
	 * throw NoSuchCommandException
	 */
	virtual double getProgression(unsigned long commandID) const = 0;

	/**
	 * Sets the pre-run calculated duration of a command.
	 * throw NoSuchCommandException
	 */
	virtual void setCommandPreRunCalculatedDurationMs(unsigned long commandID, double duration) = 0;

	/**
	 * Gets the pre-run calculated duration of a command.
	 * throw NoSuchCommandException
	 */
	virtual double getCommandPreRunCalculatedDurationMs(unsigned long commandID) = 0;

	/**
	 * Gets the description of the element.
	 */
	virtual std::string what(unsigned long elementID) = 0;

	/**
	 * Reports a bug.
	 */
	virtual void reportBug(int64 date) const = 0;
};

}

#endif /*COMMAND_COMMANDACCESSOR_H_*/
