/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DirectCommandAccessor.h"
#include "DatabaseIDKeeper.h"
#include <command/processor/Processor.h>
#include <command/transportlayer/common/CommandStateChangeSubscriberDirectImpl.h>
#include <dataprovider/transportlayer/dataprovider/CommandAccessorDirectImpl.h>
#include <common/output/OutputStream.h>
#include <iostream>

using namespace std;
using namespace common;
using namespace dataprovider;

namespace command {

DirectCommandAccessor::DirectCommandAccessor(dataprovider::CommandAccessorDirectImpl* proxy, DataChangePublisher* dataChangePublisher) :
	_proxy(proxy),
	_dataChangePublisher(dataChangePublisher) {
}

DirectCommandAccessor::~DirectCommandAccessor() {
}

vector<common::CommandDescriptor> DirectCommandAccessor::getControllerCommands(unsigned long controllerID) const {

	try {
		return _proxy->getControllerCommands(DatabaseIDKeeper::getID(), controllerID);

	} catch (CommandListCreationException& e) {
		throw;
	} catch (CommandAccessorDirectImpl::NoSuchServantException& e) {
		throw NoSuchServantException(controllerID);
	}

}

std::vector<common::CommandDescriptor> DirectCommandAccessor::getDriverCommands(unsigned long driverID) const {

	try {
		return _proxy->getDriverCommands(DatabaseIDKeeper::getID(), driverID);

	} catch (CommandListCreationException& e) {
		throw;
	} catch (CommandAccessorDirectImpl::NoSuchServantException&e ) {
		throw NoSuchServantException(driverID);
	}

}

unsigned long DirectCommandAccessor::getServantID(unsigned long commandID) const {

	try {
		return _proxy->getServantID(DatabaseIDKeeper::getID(), commandID);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

void DirectCommandAccessor::setCommandState(unsigned long servantID, unsigned long commandID, common::CommandState state) {

	try {
		_proxy->setCommandState(DatabaseIDKeeper::getID(), servantID, commandID, state);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

common::SetCommandActiveResult DirectCommandAccessor::setCommandActive(DataHolder::Type type, unsigned long servantID, unsigned long commandID, const std::string& servantType) {

	unsigned long threadID = 0;

	CommandStateChangeSubscriberDirectImpl* commandStateChangeSubscriber = new CommandStateChangeSubscriberDirectImpl(_dataChangePublisher, type, servantID, commandID);

	try {
		CommandAccessorDirectImpl::ActiveResult result = _proxy->setCommandActive(DatabaseIDKeeper::getID(), servantID, commandID, servantType, Processor::getInstance()->getProcessorID(), threadID, commandStateChangeSubscriber);

		if (result == CommandAccessorDirectImpl::EXECUTE) {
			// no need to use the subscriber
			delete commandStateChangeSubscriber;

			return EXECUTE;

		} else if (result == CommandAccessorDirectImpl::RECURSIVE) {
			// no need to use the subscriber
			delete commandStateChangeSubscriber;

			return RECURSIVE;

		} else if (result == CommandAccessorDirectImpl::WAIT) {
			// registering into the map
			if (!_commandStateChangeSubscribers.find(commandID)) {
				_commandStateChangeSubscribers.insert(commandID, HashMap<unsigned long, CommandStateChangeSubscriberDirectImpl*>());
			}

			HashMap<unsigned long, CommandStateChangeSubscriberDirectImpl*>& threadCommandStateChangeSubscribers = _commandStateChangeSubscribers.get(commandID);
			threadCommandStateChangeSubscribers.insert(threadID, commandStateChangeSubscriber);

			return WAIT;

		} else if (result == CommandAccessorDirectImpl::SKIP) {
			// no need to use the subscriber
			delete commandStateChangeSubscriber;

			return SKIP;

		} else {
			err << "Unexpected state" << eol;
		}

		return EXECUTE;

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

bool DirectCommandAccessor::hasTerminated(unsigned long commandID) {

	try {
		unsigned long threadID = 0;

		bool found = true;

		if (_commandStateChangeSubscribers.find(commandID)) {

			HashMap<unsigned long, CommandStateChangeSubscriberDirectImpl*>& stateChangeSubscribers = _commandStateChangeSubscribers.get(commandID);

			if (stateChangeSubscribers.find(threadID)) {
				return stateChangeSubscribers.get(threadID)->hasTerminated();
			} else {
				found = false;
			}

		} else {
			found = false;
		}

		// no subscriber found, the command is not listened
		// so the current command waited for termination has NOT terminated!
		if (!found) {
			return false;
		}

	} catch (Exception& e) {
		err << "Exception for command <" << commandID << "> in thread <0>" << eol;
	}

	return true;
}

void DirectCommandAccessor::finishTermination(unsigned long commandID) {

	// unsubscribe, delete, erase the subscriber
	try {
		unsigned long threadID = 0;

		CommandStateChangeSubscriberDirectImpl* subscriber = _commandStateChangeSubscribers.get(commandID).get(threadID);

		// the subscriber has been previously subscribed
		subscriber->unsubscribe();
		delete subscriber;

		_commandStateChangeSubscribers.get(commandID).erase(threadID);

	} catch (Exception& e) {
		err << "Exception for command <" << commandID << "> in thread <0>" << eol;
	}
}

common::CommandState DirectCommandAccessor::getCommandState(unsigned long commandID) const {

	try {
		return _proxy->getCommandState(DatabaseIDKeeper::getID(), commandID);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

double DirectCommandAccessor::getCommandElapsedTimeMs(unsigned long commandID) const {

	try {
		return _proxy->getCommandElapsedTimeMs(DatabaseIDKeeper::getID(), commandID);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

int64 DirectCommandAccessor::getCommandStartDate(unsigned long commandID) const {

	try {
		return _proxy->getCommandStartDate(DatabaseIDKeeper::getID(), commandID);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

unsigned long DirectCommandAccessor::addCommandToServant(unsigned long servantID, const std::string& commandName) {

	try {
		return _proxy->addCommandToServant(DatabaseIDKeeper::getID(), servantID, commandName);

	} catch (CommandAccessorDirectImpl::NoSuchServantException & e) {
		throw NoSuchServantException(servantID);
	}

}

void DirectCommandAccessor::setProgression(unsigned long commandID, double progression) const {

	try {
		_proxy->setProgression(DatabaseIDKeeper::getID(), commandID, progression);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

void DirectCommandAccessor::setUnknownProgression(unsigned long commandID) const {

	try {
		_proxy->setUnknownProgression(DatabaseIDKeeper::getID(), commandID);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

double DirectCommandAccessor::getProgression(unsigned long commandID) const {

	try {
		return _proxy->getProgression(DatabaseIDKeeper::getID(), commandID);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}


void DirectCommandAccessor::setCommandPreRunCalculatedDurationMs(unsigned long commandID, double duration) {

	try {
		_proxy->setCommandPreRunCalculatedDurationMs(DatabaseIDKeeper::getID(), commandID, duration);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

double DirectCommandAccessor::getCommandPreRunCalculatedDurationMs(unsigned long commandID) {

	try {
		return _proxy->getCommandPreRunCalculatedDurationMs(DatabaseIDKeeper::getID(), commandID);

	} catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		// problem occured on the server side
		throw NoSuchCommandException(commandID);
	}
}

std::string DirectCommandAccessor::what(unsigned long elementID) {

	return _proxy->what(DatabaseIDKeeper::getID(), elementID);
}

void DirectCommandAccessor::reportBug(int64 date) const {

	_proxy->reportBug(date);
}

}
