/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_DIRECTSERVANTACCESSOR_H_
#define COMMAND_DIRECTSERVANTACCESSOR_H_

#include "ServantAccessor.h"

namespace dataprovider {

class ServantAccessorDirectImpl;

}

namespace command {

/**
 * Class defining a direct servant accessor.
 */
class DirectServantAccessor : public ServantAccessor {

public:
	/**
 	 * Constructor.
 	 */
	DirectServantAccessor(dataprovider::ServantAccessorDirectImpl* proxy);

	/**
 	 * Virtual destructor.
	 */
	virtual ~DirectServantAccessor();

	/**
	 * Gets the Controllers.
	 * throw ServantsListCreationException
	 */
	virtual std::vector<common::ServantDescriptor> getControllers() const;

	/**
	 * Gets the Drivers.
	 * throw ServantsListCreationException
	 */
	virtual std::vector<common::ServantDescriptor> getDrivers() const;

	/**
	 * Gets the Controller child servants.
	 * throw ChildServantsListCreationException, NoSuchServantException
	 */
	virtual std::vector<common::ChildServantDescriptor> getControllerChildServants(unsigned long controllerID) const;

	/**
	 * Gets the Driver child servants.
	 * throw ChildServantsListCreationException, NoSuchServantException
	 */
	virtual std::vector<common::ChildServantDescriptor> getDriverChildServants(unsigned long driverID) const;

	/**
	 * Returns true if the servant exists.
	 */
	virtual bool servantExists(unsigned long servantID) const;

	/**
	 * Gets the servant type.
	 * throw NoSuchServantException
	 */
	virtual ServantType getServantType(unsigned long servantID) const;

	/**
	 * Gets the controller type.
	 * throw NoSuchServantException
	 */
	virtual std::string getControllerType(unsigned long controllerID) const;

	/**
	 * Returns the mode of execution.
	 * throw NoSuchServantException
	 */
	virtual common::DriverMode getMode(unsigned long driverID) const;

	/**
	 * Returns the ID for the parent of the driver.
	 * throw NoSuchServantException
	 */
	virtual unsigned long getDriverParentID(unsigned long driverID) const;

	/**
	 * Returns the ID for the protocol associated with the driver.
	 * throw NoSuchServantException
	 */
	virtual unsigned long getProtocolIDForDriver(unsigned long driverID) const;

	/**
	 * Returns the servant ID for which we have the parentID, the n name and its child index.
	 * throw NoSuchServantException, ServantDisabledException
	 */
	virtual unsigned long getEnabledOrUniqueServantID(unsigned long parentID, const std::string& nServantName, long childIndex) const;

	/**
	 * Returns the servant IDs for which we have the parentID, the n name and its child index.
	 * throw NoSuchServantException
	 */
	virtual std::vector<unsigned long> getServantIDs(unsigned long parentID, const std::string& nServantName, long childIndex) const;

	/**
	 * Returns the servant name associated with the servant ID.
	 */
	virtual std::string getServantName(unsigned long servantID);

	/**
	 * Adds the servant.
	 * throw NoSuchServantException
	 */
	virtual void addServant(unsigned long parentID, const std::string& nServantName, bool isList) const;

	/**
	 * Returns the servant list size.
	 * throw NoSuchServantException
	 */
	virtual long getServantListSize(unsigned long parentID, const std::string& nServantName) const;

	/**
	 * Resizes the servant list.
	 * throw NoSuchServantException
	 */
	virtual void resizeServantList(unsigned long parentID, const std::string& nServantName, long size) const;

	/**
	 * Adds a servant.
	 * throw NoSuchServantException
	 */
	virtual void assignServant(unsigned long parentID, const std::string& nServantName, long childIndex, unsigned long newParentID, const std::string& newNServantName, long newChildIndex);

	/**
	 * Assigns a servant.
	 * throw NoSuchServantException
	 */
	virtual void assignServant(unsigned long parentID, const std::string& nServantName, long childIndex, unsigned long newServantID);

	/**
	 * Assigns a servant.
	 * throw NoSuchServantException
	 */
	virtual void assignServant(unsigned long servantID, bool newServant, bool localServant);

	/**
	 * Removes a servant.
	 * throw NoSuchServantException, NoSuchServantChildException
	 */
	virtual unsigned long removeServantFromParent(unsigned long parentID, const std::string& nName, long index);

	/**
	 * Removes a servant.
	 * throw NoSuchServantException
	 */
	virtual void removeServant(unsigned long servantID);

	/**
	 * Sets the servant initialised.
	 * throw NoSuchServantException
	 */
	virtual void setServantInitialised(unsigned long servantID, bool initialised) const;

	/**
	 * Returns true if the servant is initialised.
	 * throw NoSuchServantException
	 */
	virtual bool isServantInitialised(unsigned long servantID) const;

	/**
	 * Returns true if the servant is initialised.
	 */
	virtual bool isServantInitialised(unsigned long parentID, const std::string& nName, long childIndex) const;

	/**
	 * Resets the elapsed simulation time.
	 */
	virtual void resetElapsedSimulationTime(unsigned long driverID) const;

	/**
	 * Gets the elapsed simulation time.
	 */
	virtual double getElapsedSimulationTimeMs(unsigned long driverID) const;

	/**
	 * Enables the servant.
	 * throw NoSuchServantException, ServantEnableException
	 */
	virtual void enableServant(unsigned long servantID, bool all);

	/**
	 * Enables the servant.
	 * throw NoSuchServantException, ServantEnableException
	 */
	virtual void enableServant(unsigned long parentID, const std::string& nName, long childIndex);

	/**
	 * Disables the servant.
	 * throw NoSuchServantException, ServantDisableException
	 */
	virtual void disableServant(unsigned long servantID, bool all);

	/**
	 * Disables the servant.
	 * throw NoSuchServantException, ServantDisableException
	 */
	virtual void disableServant(unsigned long parentID, const std::string& nName, long childIndex);

	/**
	 * Returns true if the servant is enabled.
	 * throw NoSuchServantException
	 */
	virtual bool isServantEnabled(unsigned long servantID);

	/**
	 * Returns true if the servant is enabled.
	 * throw NoSuchServantException
	 */
	virtual bool isServantEnabled(unsigned long parentID, const std::string& nName, long childIndex);

	/**
	 * Configures the servant.
	 * throw NoSuchServantException
	 */
	virtual std::vector<unsigned long> configureServant(unsigned long servantID, const std::string& configurationName);

	/**
	 * Gets the description of the element.
	 */
	virtual std::string what(unsigned long elementID);

	/**
	 * Sets the type. Allows developers of N to fake a type, useful for inherited
	 * classes with same view in client
	 */
	virtual void setClientType(unsigned long servantID, const std::string& clientType);

	/**
	 * returns true if the servant is pausable.
	 */
	virtual bool isServantPausable(unsigned long servantID);

	/**
	 * Gets the servant ID from the name.
	 */
	virtual unsigned long getServantID(const std::string& name);

	/**
	 * Sets the servant visible.
	 * throw NoSuchServantException
	 */
	virtual void setServantVisible(unsigned long servantID, bool value);

	/**
	 * Sets the servant visible.
	 * throw NoSuchServantException
	 */
	virtual void setServantFavourite(unsigned long servantID, bool value);

	/**
	 * Sets the servant command.
	 * throw NoSuchServantException
	 */
	virtual void setServantCommand(unsigned long servantID, bool value);

	/**
	 * Enables the configuration changed.
	 */
	virtual void enableConfigurationChanged();

	/**
	 * Disables the configuration changed.
	 */
	virtual void disableConfigurationChanged();

	/**
	 * Notifies a configuration changed for all the servants.
	 */
	virtual void configurationChanged();

	/**
	 * Sets the family to the servant.
	 * throw NoSuchServantException
	 */
	virtual void setServantFamily(unsigned long servantID, const std::string& familyName, const std::string& subfamilyName);

	/**
	 * Gets the family of the servant.
	 * throw NoSuchServantException
	 */
	virtual std::string getServantFamily(unsigned long servantID);

	/**
	 * Gets the family of the servant.
	 * throw NoSuchServantException
	 */
	virtual std::string getServantSubfamily(unsigned long servantID);

	/**
	 * Returns true if the servant is favourite.
	 * throw NoSuchServantException
	 */
	virtual bool isServantFavourite(unsigned long servantID);

private:
	dataprovider::ServantAccessorDirectImpl* _proxy;

};

} // namespace command

#endif /*COMMAND_DIRECTSERVANTACCESSOR_H_*/
