/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_PROPERTYACCESSOR_H_
#define COMMAND_PROPERTYACCESSOR_H_

#include <command/exception/TraceableException.h>
#include <common/base/PropertyType.h>

#include <string>
#include <vector>
#include <set>
#include <common/utils/PropertyDescriptor.h>

namespace command {

/**
 * Abstract class defining the property accessor.
 */
class PropertyAccessor {

public:
	/**
	 * Property list creation exception.
	 */
	class PropertyListCreationException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		PropertyListCreationException(const std::string& type, unsigned long ID);
	};

	/**
	 * No such Servant exception.
	 */
	class NoSuchServantException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NoSuchServantException(unsigned long servantID);
	};

	/**
	 * No such Property exception.
	 */
	class NoSuchPropertyException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NoSuchPropertyException(unsigned long propertyID);

		/**
		 * Constructor.
		 */
		NoSuchPropertyException(const std::string& propertyName);
	};

	class BadPropertyTypeException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		BadPropertyTypeException(unsigned long propertyID);
	};

	class IndexOutOfBoundsException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		IndexOutOfBoundsException(unsigned long propertyID);
	};

	class NotResizableException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NotResizableException(unsigned long propertyID);
	};

	class NullPropertyReferenceException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NullPropertyReferenceException(unsigned long propertyID);
	};

	class NotConfiguredException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NotConfiguredException(unsigned long propertyID);
	};

	class FileNotFoundException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		FileNotFoundException(const std::string& fileName);
	};

	/**
	 * Constructor
	 */
	PropertyAccessor();

	/**
	 * Virtual destructor.
	 */
	virtual ~PropertyAccessor();

	/**
	 * Gets the properties of the controller of controllerID ID.
	 * throw PropertyListCreationException, NoSuchServantException
	 */
	virtual std::vector<common::PropertyDescriptor> getControllerProperties(unsigned long controllerID) const = 0;

	/**
	 * Gets the properties of the driver of driverID ID.
	 * throw PropertyListCreationException, NoSuchServantException
	 */
	virtual std::vector<common::PropertyDescriptor> getDriverProperties(unsigned long driverID) const = 0;

	/**
	 * Gets the property ID.
	 * throw NoSuchPropertyException
	 */
	virtual unsigned long getPropertyID(unsigned long servantID, const std::string& propertyName) const = 0;

	/**
	 * Adds a property to a servant if it isn't already registered in the data provider
	 */
	virtual unsigned long addPropertyToServant(unsigned long servantID, const std::string& propertyName, const std::string& propertyType, bool spy, bool saveBox, bool & added) const = 0;

	/**
	 * Adds a property list to a servant if it isn't already registered in the data provider
	 */
	virtual unsigned long addDynamicPropertyToServant(unsigned long servantID, const std::string& propertyName, const std::string& propertyType, bool spy) const = 0;

	/**
	 * Removes a property from the servant.
	 */
	virtual void removePropertyFromServant(unsigned long servantID, unsigned long propertyID) const = 0;

	/**
	 * Sets a the settings format and save to a property.
	 */
	virtual void setPropertySettings(unsigned long propertyID, const std::string& propertyFormat, bool save, bool spy) const = 0;

	/**
	 * Resizes the dynamic property.
	 * throw NoSuchPropertyException
	 */
	virtual void resizeDynamicProperty(unsigned long propertyID, long size) const = 0;

	/**
	 * Gets the dynamic property size.
	 * throw NoSuchPropertyException
	 */
	virtual long getDynamicPropertySize(unsigned long propertyID) const = 0;

	/**
	 * Gets the dynamic property size.
	 * throw NoSuchPropertyException, IndexOutOfBoundsException
	 */
	virtual unsigned long getDynamicPropertyIDAtIndex(unsigned long propertyID, long index, bool resize) const = 0;

	/**
	 * Gets the dynamic property index.
	 */
	virtual long getIndexOfDynamicPropertyID(unsigned long parentPropertyID, unsigned long propertyID) const = 0;

	/**
	 * Gets the int8 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual int8 getInt8Value(unsigned long propertyID) const = 0;

	/**
	 * Gets the int16 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual int16 getInt16Value(unsigned long propertyID) const = 0;

	/**
	 * Gets the uint16 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual uint16 getUint16Value(unsigned long propertyID) const = 0;

	/**
	 * Gets the int32 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual int32 getInt32Value(unsigned long propertyID) const = 0;

	/**
	 * Gets the uint32 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual uint32 getUint32Value(unsigned long propertyID) const = 0;

	/**
	 * Gets the int64 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual int64 getInt64Value(unsigned long propertyID) const = 0;

	/**
	 * Gets the uint64 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual uint64 getUint64Value(unsigned long propertyID) const = 0;

	/**
	 * Gets the float32 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual float32 getFloat32Value(unsigned long propertyID) const = 0;

	/**
	 * Gets the float64 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual float64 getFloat64Value(unsigned long propertyID) const = 0;

	/**
	 * Gets the string value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual std::string getStringValue(unsigned long propertyID) const = 0;

	/**
	 * Gets the binary string value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual std::string getBinaryStringValue(unsigned long propertyID) const = 0;

	/**
	 * Gets the boolean value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual bool getBooleanValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics boolean value.
	 */
	virtual bool setIcsBooleanValue(unsigned long propertyID, bool value) const = 0;

	/**
	 * Sets the ics boolean value and formatted value.
	 */
	virtual bool setIcsBooleanValueAndFormattedValue(unsigned long propertyID, bool value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics boolean value.
	 */
	virtual bool getIcsBooleanValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int16 value.
	 */
	virtual bool setIcsInt16Value(unsigned long propertyID, int16 value) const = 0;

	/**
	 * Sets the ics int16 value and formatted value.
	 */
	virtual bool setIcsInt16ValueAndFormattedValue(unsigned long propertyID, int16 value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics int16 value.
	 */
	virtual int16 getIcsInt16Value(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics uint16 value.
	 */
	virtual bool setIcsUint16Value(unsigned long propertyID, uint16 value) const = 0;

	/**
	 * Sets the ics uint16 value and formatted value.
	 */
	virtual bool setIcsUint16ValueAndFormattedValue(unsigned long propertyID, uint16 value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics uint16 value.
	 */
	virtual uint16 getIcsUint16Value(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int32 value.
	 */
	virtual bool setIcsInt32Value(unsigned long propertyID, int32 value) const = 0;

	/**
	 * Sets the ics int32 value and formatted value.
	 */
	virtual bool setIcsInt32ValueAndFormattedValue(unsigned long propertyID, int32 value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics int32 value.
	 */
	virtual int32 getIcsInt32Value(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics uint32 value.
	 */
	virtual bool setIcsUint32Value(unsigned long propertyID, uint32 value) const = 0;

	/**
	 * Sets the ics uint32 value and formatted value.
	 */
	virtual bool setIcsUint32ValueAndFormattedValue(unsigned long propertyID, uint32 value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics uint32 value.
	 */
	virtual uint32 getIcsUint32Value(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int64 value.
	 */
	virtual bool setIcsInt64Value(unsigned long propertyID, int64 value) const = 0;

	/**
	 * Sets the ics int64 value and formatted value.
	 */
	virtual bool setIcsInt64ValueAndFormattedValue(unsigned long propertyID, int64 value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics int16 value.
	 */
	virtual int64 getIcsInt64Value(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics uint64 value.
	 */
	virtual bool setIcsUint64Value(unsigned long propertyID, uint64 value) const = 0;

	/**
	 * Sets the ics uint64 value and formatted value.
	 */
	virtual bool setIcsUint64ValueAndFormattedValue(unsigned long propertyID, uint64 value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics uint64 value.
	 */
	virtual uint64 getIcsUint64Value(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics float32value.
	 */
	virtual bool setIcsFloat32Value(unsigned long propertyID, float32 value) const = 0;

	/**
	 * Sets the ics float32 value and formatted value.
	 */
	virtual bool setIcsFloat32ValueAndFormattedValue(unsigned long propertyID, float32 value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics float32 value.
	 */
	virtual float32 getIcsFloat32Value(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics float64 value.
	 */
	virtual bool setIcsFloat64Value(unsigned long propertyID, float64 value) const = 0;

	/**
	 * Sets the ics float64 value and formatted value.
	 */
	virtual bool setIcsFloat64ValueAndFormattedValue(unsigned long propertyID, float64 value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics float64 value.
	 */
	virtual float64 getIcsFloat64Value(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics double value.
	 */
	virtual bool setIcsStringValue(unsigned long propertyID, const std::string& value) const = 0;

	/**
	 * Sets the ics float64 value and formatted value.
	 */
	virtual bool setIcsStringValueAndFormattedValue(unsigned long propertyID, const std::string& value, const std::string& formattedValue) const = 0;

	/**
	 * Gets the ics string value.
	 */
	virtual std::string getIcsStringValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int16 array value.
	 */
	virtual void setIcsInt16ArrayValue(unsigned long propertyID, int16* value) const = 0;

	/**
	 * Gets the ics int16 array value.
	 */
	virtual int16* getIcsInt16ArrayValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics uint16 array value.
	 */
	virtual void setIcsUint16ArrayValue(unsigned long propertyID, uint16* value) const = 0;

	/**
	 * Gets the ics uint16 array value.
	 */
	virtual uint16* getIcsUint16ArrayValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int32 array value.
	 */
	virtual void setIcsInt32ArrayValue(unsigned long propertyID, int32* value) const = 0;

	/**
	 * Gets the ics int32 array value.
	 */
	virtual int32* getIcsInt32ArrayValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics uint32 array value.
	 */
	virtual void setIcsUint32ArrayValue(unsigned long propertyID, uint32* value) const = 0;

	/**
	 * Gets the ics uint32 array value.
	 */
	virtual uint32* getIcsUint32ArrayValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int64 array value.
	 */
	virtual void setIcsInt64ArrayValue(unsigned long propertyID, int64* value) const = 0;

	/**
	 * Gets the ics int64 array value.
	 */
	virtual int64* getIcsInt64ArrayValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics uint64 array value.
	 */
	virtual void setIcsUint64ArrayValue(unsigned long propertyID, uint64* value) const = 0;

	/**
	 * Gets the ics uint64 array value.
	 */
	virtual uint64* getIcsUint64ArrayValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics float32 array value.
	 */
	virtual void setIcsFloat32ArrayValue(unsigned long propertyID, float32* value) const = 0;

	/**
	 * Gets the ics float32 array value.
	 */
	virtual float32* getIcsFloat32ArrayValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics float64 array value.
	 */
	virtual void setIcsFloat64ArrayValue(unsigned long propertyID, float64* value) const = 0;

	/**
	 * Gets the ics float64 array value.
	 */
	virtual float64* getIcsFloat64ArrayValue(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int16 array size.
	 */
	virtual void setIcsInt16ArraySize(unsigned long propertyID, int32 size) const = 0;

	/**
	 * Gets the ics int32 array size.
	 */
	virtual int32 getIcsInt16ArraySize(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics uint16 array size.
	 */
	virtual void setIcsUint16ArraySize(unsigned long propertyID, int32 size) const = 0;

	/**
	 * Gets the ics uint16 array size.
	 */
	virtual int32 getIcsUint16ArraySize(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int32 array size.
	 */
	virtual void setIcsInt32ArraySize(unsigned long propertyID, int32 size) const = 0;

	/**
	 * Gets the ics int32 array size.
	 */
	virtual int32 getIcsInt32ArraySize(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int32 array size.
	 */
	virtual void setIcsUint32ArraySize(unsigned long propertyID, int32 size) const = 0;

	/**
	 * Gets the ics uint32 array size.
	 */
	virtual int32 getIcsUint32ArraySize(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics int64 array size.
	 */
	virtual void setIcsInt64ArraySize(unsigned long propertyID, int32 size) const = 0;

	/**
	 * Gets the ics int64 array size.
	 */
	virtual int32 getIcsInt64ArraySize(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics uint64 array size.
	 */
	virtual void setIcsUint64ArraySize(unsigned long propertyID, int32 size) const = 0;

	/**
	 * Gets the ics uint64 array size.
	 */
	virtual int32 getIcsUint64ArraySize(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics float32 array size.
	 */
	virtual void setIcsFloat32ArraySize(unsigned long propertyID, int32 size) const = 0;

	/**
	 * Gets the ics float32 array size.
	 */
	virtual int32 getIcsFloat32ArraySize(unsigned long propertyID) const = 0;

	/**
	 * Sets the ics float64 array size.
	 */
	virtual void setIcsFloat64ArraySize(unsigned long propertyID, int32 size) const = 0;

	/**
	 * Gets the ics float64 array size.
	 */
	virtual int32 getIcsFloat64ArraySize(unsigned long propertyID) const = 0;

	/**
	 * Notifies the update of the ics property.
	 */
	virtual void updateIcsProperty(unsigned long propertyID) const = 0;

	/**
	 * Imports data from a file.
	 * throw NoSuchPropertyException, BadPropertyTypeException, IndexOutOfBoundsException
	 */
	virtual void importArrayData(unsigned long ID, const std::string& fileName, int8 type) = 0;

	/**
	 * Structure for scannable property.
	 */
	struct PropertyInfoItem {
		std::string actualPropertyName;
		std::string setpointPropertyName;
		std::string offsetPropertyName;
		std::string shortPropertyName;
		bool isScan;
		bool isSpy;
		bool isDynamic;

		PropertyInfoItem() : isScan(false), isSpy(false), isDynamic(false) {}
	};

	/**
	 * Returns true if the property is scannable.
	 */
	virtual bool isScannableProperty(const std::string & controllerType, const std::string & wantedPropertyName) const = 0;

	/**
	 * Gets the scannable property item.
	 */
	virtual PropertyInfoItem getPropertyInfoItem(const std::string & controllerType, const std::string & wantedPropertyName) const = 0;

	/**
	 * Gets the property items.
	 */
	virtual std::vector<PropertyInfoItem> getPropertyInfoItems(const std::string & controllerType) const = 0;

	/**
	 * Adds the item.
	 */
	virtual void addPropertyInfoItem(const std::string & controllerType, const std::string & wantedPropertyName, const std::string & actualPropertyName, const std::string & offsetPropertyName, const std::string & shortPropertyName, bool isScan, bool isSpy, bool isDynamic) const = 0;

	/**
	 * Returns the number of properties for a controller type.
	 */
	virtual int32 getNumberOfScannableProperties(const std::string & controllerType) const = 0;

	/**
	 * Sets the property unit.
	 */
	virtual void setConstantPropertyUnit(unsigned long propertyID, const std::string & value) const = 0;

	/**
	 * Sets the property unit.
	 */
	virtual void setPropertyUnit(unsigned long propertyID, unsigned long unitPropertyID) const = 0;

	/**
	 * Sets the enumerated values with property ID.
	 */
	virtual void setEnumeratedValues(unsigned long propertyID, unsigned long valuesPropertyID) const = 0;

	/**
	 * Sets the enumerated labels with property ID.
	 */
	virtual void setEnumeratedLabels(unsigned long propertyID, unsigned long labelsPropertyID) const = 0;

	/**
	 * Adds the type to the servant.
	 */
	virtual void addServantType(unsigned long servantID, const std::string& type, bool atFront) const = 0;

	/**
	 * Sets the servant in debug.
	 */
	virtual void setServantInDebug(unsigned long servantID) const = 0;

};

}

#endif /*COMMAND_PROPERTYACCESSOR_H_*/
