/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMAND_SERVANTACCESSOR_H_
#define COMMAND_SERVANTACCESSOR_H_

#include <command/exception/TraceableException.h>
#include <common/base/CommonTypes.h>

#include <vector>
#include <common/utils/ChildServantDescriptor.h>
#include <common/utils/ServantDescriptor.h>

namespace command {

class ServantAccessor {

public:
	/**
	 * Exception for servant list creation.
	 */
	class ServantsListCreationException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		ServantsListCreationException(const std::string& type);
	};

	/**
	 * Exception for child servant list creation.
	 */
	class ChildServantsListCreationException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		ChildServantsListCreationException(const std::string& type);
	};

	/**
	 * Exception for servant.
	 */
	class NoSuchServantException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NoSuchServantException(unsigned long servantID);

		/**
		 * Constructor.
		 */
		NoSuchServantException(unsigned long parentID, const std::string& nServantName, long childIndex);
	};

	/**
	 * Exception for servant.
	 */
	class ServantAssignmentException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		ServantAssignmentException(const std::string& message);
	};

	/**
	 * Exception for servant.
	 */
	class CannotConfigureServantException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		CannotConfigureServantException(unsigned long servantID);
	};


	/**
	 * Exception for servant out of bounds.
	 */
	class IndexOutOfBoundsException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		IndexOutOfBoundsException(long size);

		/**
		 * Gets the size.
		 */
		long getSize() const;

	private:
		long _size;
	};

	/**
	 * Exception for servant disabled.
	 */
	class ServantDisabledException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		ServantDisabledException(unsigned long parentID, const std::string nServantName, long childIndex);
	};

	/**
	 * Exception for servant enable.
	 */
	class ServantEnableException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		ServantEnableException(const std::string& message);
	};

	/**
	 * Exception for child servant name.
	 */
	class NoSuchServantChildNameException : public TraceableException {

	public:
		/**
		 * Constructor.
		 */
		NoSuchServantChildNameException(const std::string& nName);
	};

	/**
	 * Servant type.
	 */
	typedef enum {CONTROLLER, DRIVER} ServantType;

	/**
	 * Constructor.
	 */
	ServantAccessor();

	/**
 	 * Virtual destructor.
 	 */
	virtual ~ServantAccessor();

	/**
	 * Gets the Controllers.
	 * throw ServantsListCreationException
	 */
	virtual std::vector<common::ServantDescriptor> getControllers() const = 0;

	/**
	 * Gets the Drivers.
	 * throw ServantsListCreationException
	 */
	virtual std::vector<common::ServantDescriptor> getDrivers() const = 0;

	/**
	 * Gets the Controller child servants.
	 * throw ChildServantsListCreationException
	 */
	virtual std::vector<common::ChildServantDescriptor> getControllerChildServants(unsigned long controllerID) const = 0;

	/**
	 * Gets the Driver child servants.
	 * throw ChildServantsListCreationException
	 */
	virtual std::vector<common::ChildServantDescriptor> getDriverChildServants(unsigned long driverID) const = 0;

	/**
	 * Returns true if the servant exists.
	 */
	virtual bool servantExists(unsigned long servantID) const = 0;

	/**
	 * Gets the servant type.
	 * throw NoSuchServantException
	 */
	virtual ServantType getServantType(unsigned long servantID) const = 0;

	/**
	 * Gets the controller type.
	 * throw NoSuchServantException
	 */
	virtual std::string getControllerType(unsigned long controllerID) const = 0;

	/**
	 * Returns the mode of execution.
	 * throw NoSuchServantException
	 */
	virtual common::DriverMode getMode(unsigned long driverID) const = 0;

	/**
	 * Returns the ID for the parent of the driver.
	 * throw NoSuchServantException
	 */
	virtual unsigned long getDriverParentID(unsigned long driverID) const = 0;

	/**
	 * Returns the ID for the protocol associated with the driver.
	 * throw NoSuchServantException
	 */
	virtual unsigned long getProtocolIDForDriver(unsigned long driverID) const = 0;

	/**
	 * Returns the servant enabled ID for which we have the parentID, the n name and its child index.
	 * throw NoSuchServantException, ServantDisabledException
	 */
	virtual unsigned long getEnabledOrUniqueServantID(unsigned long parentID, const std::string& nServantName, long childIndex) const = 0;

	/**
	 * Returns the servant IDs for which we have the parentID, the n name and its child index.
	 * throw NoSuchServantException
	 */
	virtual std::vector<unsigned long> getServantIDs(unsigned long parentID, const std::string& nServantName, long childIndex) const = 0;

	/**
	 * Returns the servant name associated with the servant ID.
	 */
	virtual std::string getServantName(unsigned long servantID) = 0;

	/**
	 * Adds the servant.
	 * throw NoSuchServantException
	 */
	virtual void addServant(unsigned long parentID, const std::string& nServantName, bool isList) const = 0;

	/**
	 * Returns the servant list size.
	 * throw NoSuchServantException
	 */
	virtual long getServantListSize(unsigned long parentID, const std::string& nServantName) const = 0;

	/**
	 * Resizes the servant list.
	 * throw NoSuchServantException
	 */
	virtual void resizeServantList(unsigned long parentID, const std::string& nServantName, long size) const = 0;

	/**
	 * Assigns a servant.
	 * throw NoSuchServantException
	 */
	virtual void assignServant(unsigned long parentID, const std::string& nServantName, long childIndex, unsigned long newParentID, const std::string& newNServantName, long newChildIndex) = 0;

	/**
	 * Assigns a servant.
	 * throw NoSuchServantException
	 */
	virtual void assignServant(unsigned long parentID, const std::string& nServantName, long childIndex, unsigned long newServantID) = 0;

	/**
	 * Assigns a servant.
	 * throw NoSuchServantException
	 */
	virtual void assignServant(unsigned long servantID, bool newServant, bool localServant) = 0;

	/**
	 * Removes a child servant from a parent. Returns true if child is totally removed.
	 * throw NoSuchServantException, NoSuchServantChildException
	 */
	virtual unsigned long removeServantFromParent(unsigned long parentID, const std::string& nName, long index) = 0;

	/**
	 * Removes a servant.
	 * throw NoSuchServantException
	 */
	virtual void removeServant(unsigned long servantID) = 0;

	/**
	 * Sets the servant initialised.
	 */
	virtual void setServantInitialised(unsigned long servantID, bool initialised) const = 0;

	/**
	 * Returns true if the servant is initialised.
	 */
	virtual bool isServantInitialised(unsigned long servantID) const = 0;

	/**
	 * Returns true if the servant is initialised.
	 */
	virtual bool isServantInitialised(unsigned long parentID, const std::string& nName, long childIndex) const = 0;

	/**
	 * Resets the elapsed simulation time.
	 */
	virtual void resetElapsedSimulationTime(unsigned long driverID) const = 0;

	/**
	 * Gets the elapsed simulation time.
	 */
	virtual double getElapsedSimulationTimeMs(unsigned long driverID) const = 0;

	/**
	 * Enables the servant.
	 * throw NoSuchServantException, ServantEnableException
	 */
	virtual void enableServant(unsigned long servantID, bool all) = 0;

	/**
	 * Enables the servant.
	 * throw NoSuchServantException, ServantEnableException
	 */
	virtual void enableServant(unsigned long parentID, const std::string& nName, long childIndex) = 0;

	/**
	 * Disables the servant.
	 * throw NoSuchServantException, ServantDisableException
	 */
	virtual void disableServant(unsigned long servantID, bool all) = 0;

	/**
	 * Disables the servant.
	 * throw NoSuchServantException, ServantDisableException
	 */
	virtual void disableServant(unsigned long parentID, const std::string& nName, long childIndex) = 0;

	/**
	 * Returns true if the servant is enabled.
	 * throw NoSuchServantException
	 */
	virtual bool isServantEnabled(unsigned long servantID) = 0;

	/**
	 * Returns true if the servant is enabled.
	 * throw NoSuchServantException
	 */
	virtual bool isServantEnabled(unsigned long parentID, const std::string& nName, long childIndex) = 0;

	/**
	 * Configures the servant.
	 */
	virtual std::vector<unsigned long> configureServant(unsigned long servantID, const std::string& configurationName) = 0;

	/**
	 * Gets the description of the element.
	 */
	virtual std::string what(unsigned long elementID) = 0;

	/**
	 * Sets the type. Allows developers of N to fake a type, useful for inherited
	 * classes with same view in client
	 */
	virtual void setClientType(unsigned long servantID, const std::string& clientType) = 0;

	/**
	 * returns true if the servant is pausable.
	 */
	virtual bool isServantPausable(unsigned long servantID) = 0;

	/**
	 * Gets the servant ID from the name.
	 */
	virtual unsigned long getServantID(const std::string& name) = 0;

	/**
	 * Sets the servant visible.
	 * throw NoSuchServantException
	 */
	virtual void setServantVisible(unsigned long servantID, bool value) = 0;

	/**
	 * Sets the servant visible.
	 * throw NoSuchServantException
	 */
	virtual void setServantFavourite(unsigned long servantID, bool value) = 0;

	/**
	 * Sets the servant command.
	 * throw NoSuchServantException
	 */
	virtual void setServantCommand(unsigned long servantID, bool value) = 0;

	/**
	 * Enables the configuration changed.
	 */
	virtual void enableConfigurationChanged() = 0;

	/**
	 * Disables the configuration changed.
	 */
	virtual void disableConfigurationChanged() = 0;

	/**
	 * Notifies a configuration changed for all the servants.
	 */
	virtual void configurationChanged() = 0;

	/**
	 * Sets the family to the servant.
	 * throw NoSuchServantException
	 */
	virtual void setServantFamily(unsigned long servantID, const std::string& familyName, const std::string& subfamilyName) = 0;

	/**
	 * Gets the family of the servant.
	 * throw NoSuchServantException
	 */
	virtual std::string getServantFamily(unsigned long servantID) = 0;

	/**
	 * Gets the subfamily of the servant.
	 * throw NoSuchServantException
	 */
	virtual std::string getServantSubfamily(unsigned long servantID) = 0;

	/**
	 * Returns true if the servant is favourite.
	 * throw NoSuchServantException
	 */
	virtual bool isServantFavourite(unsigned long servantID) = 0;

};

} // namespace command

#endif /*COMMAND_SERVANTACCESSOR_H_*/
