/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*
 *  derived from MultiGaussPeakFinder 
 * F.Cecillon 08-07
 *
 */

#ifndef CENTEROFMASSPEAKFINDER_H
#define CENTEROFMASSPEAKFINDER_H

// This object finds center of mass peaks on 1D data.

#include <iostream>
#include <vector>
#include <fstream>
#include <blitz/array.h>   
#include <math.h>
#include <stdexcept>

#include "MathematicaForm.h"
#include "Moments.h"

// some defines used for calling the function
// standard errors
static const bool CENTEROFMASS_USE_STDERR = true;
static const bool CENTEROFMASS_NO_STDERR = false;

class CenterOfMassPeakFinder {

public:
	/** This constructor is the most general one.
	 * It finds the peaks of the data set placed in y (abscissa are in x).
	 * @param y the values of the curve to fit
	 * @param x the abscissa of the curve to fit. This array should have the same dimension than y.

	 * The values passed in x and y must not be obligatorily in increasing order of abscissa, or have a constant sampling in x. You may for instance
	 * give a logarithmic sampling as for energies.
	 */
	CenterOfMassPeakFinder(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x) // general call
			;

	/** Same constructor as above, but without abscissa values. The values are computed from the bounds of the
	 * y vector. Note that the blitz library manages very well bounds in its data structure so that if you give
	 * a subset of an array by y(Range(3,45)) to the peak finder, the x abscissa used for the fits will take
	 * them into account, meaning that x will contains values going from 3 to 45. Positions will be returned as well
	 * in this referential.
	 */

	CenterOfMassPeakFinder(const blitz::Array<double, 1>& y) // without x (taken as indices of y)
			;

	/**
	 * Gets the center of mass.
	 */
	double getPosition() const; // all peaks that have been found

	/**
	 * Gets the value at the center of mass.
	 */
	double getAmplitude() const;

private:
	void calculate();

	blitz::Array<double, 1> m_vecX;
	blitz::Array<double, 1> m_vecY;

	double m_position;
	double m_amplitude;
};

#endif // PEAKFINDER_H