/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ExponentialFit.h"

using namespace std;
using namespace blitz;

ExponentialFit::ExponentialFit (){
	a0 = 0.;
	a1 = 0.;
	a2 = 0.;
};


void ExponentialFit::calculateFit(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x){
	//fit calculation process is divided into two main steps
	//first one consists: find initial curve parameters
	Array<double, 1> firstGuess(3);
	firstGuess = calculateGuessParameters(y);

	//the second one on calculating the real coefficients by using Least squares method
	fit(y, x, firstGuess);
}

void ExponentialFit::calculateFit(const blitz::Array<double, 1>& y){

	//fit calculation process is divided into two main steps
	//first one consists on finding initial curve parameters
	Array<double, 1> firstGuess(3);
	firstGuess = calculateGuessParameters(y);


	//create x array
	int n = y.size();
	int minIndex = y.lbound(firstDim);
	Array<double, 1> x(n);
	firstIndex i;
	x = i + minIndex;

	//the second one on calculating the real coefficients by using Least squares method
	fit(y, x, firstGuess);
}


Array<double, 1> ExponentialFit::calculateGuessParameters (const blitz::Array<double, 1>& y) {

	//a0 -> amplitude = max (y)
	double maxValue = -9999999;
	for (int i = 0; i< y.size(); i++) {
		if (y(i) > maxValue) {
			maxValue = y(i);
		}
	}
	double la0 = maxValue;

	//a1
	double sum;
	for (int i = 0; i< y.size(); i++) {
		sum += y(i);
	}
	//sign of coefficient a1
	int sign;
	if ((y(0) > y(y.size()/2)) && (y(y.size()/2 +1) > y(y.size()))){
		sign = -1;
	} else {
		sign = 1;
	}
	double la1 = sign *(y.size() / sum);

	//pedestal (a2)
	double minValue = 9999999;
	for (int i = 0; i < y.size(); i++) {
		if (y(i) < minValue) {
			minValue = y(i);
		}
	}
	double la2 = minValue;

	//just to test: firstGuess
	Array<double, 1>firstGuess(3);
	firstGuess(0) = la0;
	firstGuess(1) = la1;
	firstGuess(2) = la2;

	return firstGuess;
}

void ExponentialFit::fit(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x, const blitz::Array<double, 1>& firstGuess){

	unsigned int n = x.size();
	Array<double, 1> sig(n);
	sig = 1.0;

	Array<double, 1> a(3);
	a = firstGuess(0), firstGuess(1), firstGuess(2);
	Array<bool, 1> ia(3);
	ia = true, true, true; // used?

	LMFit fit(x, y, sig, a, ia, LMFIT_EXPONENTIAL);
	a = fit.parameters(); //get final curve coefficients

	//exponential curve formula: y = a0 * exp (a1*x) + a2
	a0 = a(0); // amplitude
	a1 = a(1); // mean
	a2 = a(2); // pedestal

}

double ExponentialFit::amplitude () {
	return a0;
}

double ExponentialFit::mean(){
	return a1;
}

double ExponentialFit::pedestal(){
	return a2;
}
