/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <blitz/array.h>
#include <random/uniform.h>
#include <random/normal.h>
#include <iostream>
#include <cmath>
#include <list>
#include <sstream>
#include <cstdlib>
#include <limits>

#include "common/analysislibrary/MultiGaussPeakFinder.h"
#include "DataReadWrite.h"

using namespace std;
using namespace blitz;

struct PeakInfo
{
  int min_value;
  int max_value;
  double a0;
  double a1;
  double a2;
};

int main(int argc, char *argv[])
{
  if (argc < 11)
  {
    cerr << "usage: generator test file.data samples pedestal N a0_mean a0_dev a2_mean a2_dev noise_amp" << endl;
    return 0;
  }

  int samples = 10000;
  double pedestal = 10.0;
  int number_of_peaks = 10;
  double a0_mean = 50.0;
  double a0_dev = 1.0;
  double a2_mean = 10.0;
  double a2_dev = 1.0;
  double noise_amp = 5.0;
  bool check = false;
  
  string test;
  istringstream stest(argv[1]);
  stest >> test;
  if (test == "test")
    check = true;

  istringstream ssamples(argv[3]);
  ssamples >> samples;
  
  istringstream sped(argv[4]);
  sped >> pedestal;
  
  istringstream speaks(argv[5]);
  speaks >> number_of_peaks;
  
  istringstream sa0m(argv[6]);
  sa0m >> a0_mean;
  
  istringstream sa0d(argv[7]);
  sa0d >> a0_dev;
  
  istringstream sa2m(argv[8]);
  sa2m >> a2_mean;
    
  istringstream sa2d(argv[9]);
  sa2d >> a2_dev;
  
  istringstream sn(argv[10]);
  sn >> noise_amp; 

  Array<double, 1> data(samples);
  
  data = pedestal;
  
  ranlib::Uniform<double> uniform;
  ranlib::Normal<double> a0_normal(a0_mean, a0_dev);
  ranlib::Normal<double> a2_normal(a2_mean, a2_dev);
  
  // intervals already used by peak to avoid overlap
  list<PeakInfo> info;
  
  // determine 10 peaks
  int peaks = 0;
  for (int i = 0; i < number_of_peaks; ++i)
  {
    // Exclude bad peaks, which overlap other peaks
    double a1 = (double)samples*uniform.random();
    double a0 = a0_normal.random();
    double a2 = a2_normal.random();
    
    int begin = (int)(a1 - 3*a2);
    int end = (int)(a1 + 3*a2);
    
    bool in = false;
    
    if ((begin < 0) || (end >= samples))
    {
      in = true;
    }
    else
    {
      // finding if begin or end belongs to an interval
      for (list<PeakInfo>::const_iterator i = info.begin(); i != info.end(); ++i)
      {
        if (((begin >= i->min_value) && (begin < i->max_value)) || ((end >= i->min_value) && (end < i->max_value)))
        {
          in = true;
          break;
        }
      }
    }  
    
    if (!in)
    {
      ++peaks;
      
      for (int i = max(begin, 0); i < min(end + 1, samples); ++i)
        data(i) += a0*exp(0.5*(-pow((i - a1)/a2, 2)));
        
      PeakInfo pinfo;
      
      pinfo.min_value = begin;
      pinfo.max_value = end + 1;
      pinfo.a0 = a0;
      pinfo.a1 = a1;
      pinfo.a2 = a2;
      
      info.push_back(pinfo);
    }    
  }
  
  for (int i = 0; i < samples; ++i)
  {
    data(i) += noise_amp*(uniform.random() - 0.5);
  }

  writeDataFile(argv[2], data);

  ////////////////////////////////////////////////////////////////////////////// 
  // Stats
  if (check)
  {
    try {
      MultiGaussPeakFinder finder(data);

      if (peaks == finder.numberOfPeaks())
        cout << "All peaks found!" << endl;
      else
      {
        if (peaks > finder.numberOfPeaks())
          cout << peaks - finder.numberOfPeaks() << " peak(s) not found!" << endl;
        else
          cout << finder.numberOfPeaks() - peaks << " peak(s) found!" << endl; 
      }
      cout << endl;
  
      double amp_var = 0;
      double pos_var = 0;
      double rms_var = 0;
      double ped_var = 0;
  
      for (int i = 0; i < finder.numberOfPeaks(); ++i)
      {
        Peak* P = finder.peak(i);
    
        // searching among generated peaks
        double min_distance = numeric_limits<double>::max();
        const PeakInfo* fitPeak = 0;
    
        for (list<PeakInfo>::const_iterator p = info.begin(); p != info.end(); ++p)
        {
          if (fabs(P->position() - p->a1) < min_distance)
          {
            min_distance = fabs(P->position() - p->a1);
            fitPeak = &(*p);
          }
        }
      
        cout << "Peak " << i << " :" << endl;
        cout << "a0 = " << P->amplitude() << " - " << fitPeak->a0 << endl;
        cout << "a1 = " << P->position() << " - " << fitPeak->a1 << endl;
        cout << "a2 = " << P->rms() << " - " << fitPeak->a2 << endl;
        cout << "a3 = " << P->pedestal() << " - " << pedestal << endl;
    
        cout << endl;
        
        amp_var += (P->amplitude() - fitPeak->a0)*(P->amplitude() - fitPeak->a0);
        pos_var += (P->position() - fitPeak->a1)*(P->position() - fitPeak->a1);
        rms_var += (P->rms() - fitPeak->a2)*(P->rms() - fitPeak->a2);
        ped_var += (P->pedestal() - pedestal)*(P->pedestal() - pedestal);
      }
      
      amp_var /= (double)finder.numberOfPeaks();
      pos_var /= (double)finder.numberOfPeaks();
      rms_var /= (double)finder.numberOfPeaks();
      ped_var /= (double)finder.numberOfPeaks();
    
      cout << "Amplitude variance = " << amp_var << endl;
      cout << "Position variance = " << pos_var << endl;
      cout << "Rms variance = " << rms_var << endl;
      cout << "Pedestal variance = " << ped_var << endl;
    }
    catch (const logic_error& error)
    {
      cerr << error.what() << endl;
    }
  }

  cout << endl << "To visualize the results, you can launch multiGaussPeakFinderTest " << argv[2] << " Data/test.data" << endl;
  
  return 1;
}
