/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DataReadWrite.h"
#include "common/analysislibrary/MultiGaussPeakFinder.h"
#include "common/analysislibrary/CenterOfMassPeakFinder.h"
#include "common/analysislibrary/SimpleGaussFinder.h"
#include "common/analysislibrary/PolynomialPeakFinder.h"
#include "common/analysislibrary/Polynomial.h"
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/fstream.hpp>
#include <iostream>
#include <list>
#include "common/analysislibrary/Tools.h"

using namespace std;
using namespace blitz;

namespace fs = boost::filesystem;

std::string MULTIGAUSS = "multigauss";
std::string POLYNOMIAL = "polynomial";
std::string GAUSS = "gauss";
std::string ALL = "all";

void testMultiGaussPeakFinder(Array<double, 1> data, ofstream & result, const string& fitFileName, const string& fileName) {

	// Test with multi gauss peak finder
	MultiGaussPeakFinder gaussFinder;

	try {
		gaussFinder.calculatePeaks(data);

		result << gaussFinder.string(true) << endl;

		ofstream file(fitFileName.c_str(), ios::out);

		for (int i = 0; i < gaussFinder.numberOfPeaks(); ++i) {
			Peak* P = gaussFinder.peak(i);

			file << "a0_" << i << " = " << P->amplitude() << endl;
			file << "a1_" << i << " = " << P->position() << endl;
			file << "a2_" << i << " = " << P->rms() << endl;
			file << "a3_" << i << " = " << P->pedestal() << endl;

			file << "e" << i << "(x) = a0_" << i << "*exp((-((x - a1_" << i	<< ")/a2_" << i << ")**2)/2) + a3_" << i << endl;
		}

		file << "plot ";

		for (int i = 0; i < gaussFinder.numberOfPeaks(); ++i) {
			file << "e" << i << "(x), ";
		}

		file << "'" << fileName << "'" << endl;

		file.close();
	}
	catch (const logic_error& error) {
		cerr << error.what() << endl;
		result << "error in peak calculation with multi gauss peak finder" << endl;
		result << gaussFinder.string(true) << endl;
	}
}

void testSimpleGaussFinder(Array<double, 1> data, ofstream & result, const string& fitFileName, const string& fileName) {

	// Test with multi gauss peak finder
	SimpleGaussFinder gaussFinder;

	try {
		gaussFinder.fit(data);

		//result << gaussFinder.string(true) << endl;

		ofstream file(fitFileName.c_str(), ios::out);

		file << "a0" << " = " << gaussFinder.amplitude() << endl;
		file << "a1" << " = " << gaussFinder.position() << endl;
		file << "a2" << " = " << gaussFinder.rms() << endl;
		file << "a3" << " = " << gaussFinder.pedestal() << endl;

		file << "e" << "(x) = a0" << "*exp((-((x - a1" << ")/a2" << ")**2)/2) + a3" << endl;


		file << "plot ";

		file << "e" << "(x), ";

		file << "'" << fileName << "'" << endl;

		file.close();
	}
	catch (const logic_error& error) {
		cerr << error.what() << endl;
		result << "error with simple gauss finder" << endl;
		//result << gaussFinder.string(true) << endl;
	}
}

void testCenterOfMassFinder(Array<double, 1> data, ofstream & result) {

	// Test with center of mass
	try {
		CenterOfMassPeakFinder centerOfMassFinder(data);

		result << "center of mass peak at " << centerOfMassFinder.getPosition() << endl;
	}
	catch (const logic_error& error) {
		cerr << error.what() << endl;
		result << "error in peak calculation with center of mass" << endl;
	}
}

void testPolynomialFitFinder(int order, Array<double, 1> data, ofstream & result, const string& fitFileName, const string& fileName) {

	try {
		PolynomialPeakFinder polynomialFinder(order, data);

		ofstream file(fitFileName.c_str(), ios::out);

		file << "e(x) = " << polynomialFinder.getPolynomial()->string() << endl;
		file << "plot ";
		file << "e(x), ";
		file << "'" << fileName << "'" << endl;

		result << "polynomial fit " << order << " peak at " << polynomialFinder.getMaximumPeakPosition() << " with quality " << polynomialFinder.getQuality() << endl;
	}
	catch (const logic_error& error) {
		cerr << error.what() << endl;
		result << "error in peak calculation with polynomial fit" << endl;
	}
}

/**
 * To visualize the results, you can launch gnuplot and type: load 'file.fit'
 */
int main(int argc, char *argv[]) {

	if (argc < 4)
	{
		cerr << "error : peakFinderTest <type> <input directory> <output directory>" << endl;
		return 1;
	}

	// iterating the directory
	list<fs::path> files;

	string type = argv[1];

	if (type != MULTIGAUSS && type != GAUSS && type != POLYNOMIAL && type != ALL) {
		cerr << "error : type must be multigauss, gauss, polynomial or all" << endl;
		return 1;
	}

	string inputDirectory = argv[2];
	string outputDirectory = argv[3];

	fs::path scanPath = fs::system_complete(fs::path(inputDirectory));

	fs::directory_iterator end_iter;

	for (fs::directory_iterator dir_itr(scanPath); dir_itr != end_iter; ++dir_itr) {
		try {
			if (!fs::is_directory(*dir_itr)) {

				string absoluteFileName = dir_itr->path().string();

				fs::path filePath = dir_itr->path();
				if (filePath.extension() == ".dat") {
					files.push_back(filePath);
				}
			}

		} catch (const std::exception & e) {
			cerr << "cannot list files : " << e.what() << endl;
		}
	}

	string resultFilename = string(outputDirectory) + "/result.txt";
	ofstream result(resultFilename.c_str(), ios::out);

	result << "Testing the peak finders" << endl;

	for (list<fs::path>::const_iterator p = files.begin(); p != files.end(); p++) {

		string fileName = p->filename().string();

		string gfitFileName = string(outputDirectory) + "/" + fileName.substr(0, fileName.size() - 3) + "gfit";
		string sgfitFileName = string(outputDirectory) + "/" + fileName.substr(0, fileName.size() - 3) + "sgfit";
		string p2fitFileName = string(outputDirectory) + "/" + fileName.substr(0, fileName.size() - 3) + "p2fit";
		string p3fitFileName = string(outputDirectory) + "/" + fileName.substr(0, fileName.size() - 3) + "p3fit";
		string p4fitFileName = string(outputDirectory) + "/" + fileName.substr(0, fileName.size() - 3) + "p4fit";
		string p5fitFileName = string(outputDirectory) + "/" + fileName.substr(0, fileName.size() - 3) + "p5fit";

		cout << "------------------------ file " << fileName << endl;

		Array<double, 1> data = readDataFile(p->string());

		result << endl << "file " << p->filename() << endl;

		if (type == MULTIGAUSS || type == ALL) {
			testMultiGaussPeakFinder(data, result, gfitFileName, p->string());
		}

		if (type == GAUSS || type == ALL) {
			testSimpleGaussFinder(data, result, sgfitFileName, p->string());
		}

		if (type == POLYNOMIAL || type == ALL) {
//			testPolynomialFitFinder(2, A, result, p2fitFileName, p->string());
//			testPolynomialFitFinder(3, A, result, p3fitFileName, p->string());
//			testPolynomialFitFinder(4, A, result, p4fitFileName, p->string());
			testPolynomialFitFinder(5, data, result, p5fitFileName, p->string());
		}

		testCenterOfMassFinder(data, result);

		cout << "------------------------ file " << fileName << " done" << endl << endl;
	}

	result.close();

	return 1;
}
