/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "GaussFit.h"
#include "Threshold.h"
#include <algorithm>
#include <vector>

using namespace std;
using namespace blitz;

const double GaussFit::INCLUDED_VALUES_RATIO = 0.5;
const double GaussFit::DELTA_POSITION_MAX = 0.1;
const double GaussFit::DELTA_AMPLITUDE_MAX = 0.2;
const double GaussFit::DISPERSION_RATIO_MIN = 10.0;

double GaussFit::average(const blitz::Array<double, 1>& y) const {

	std::vector<double> cy(y.size());
	for (int i = 0; i < y.size(); i++) {
		cy[i] = y(i);
	}

	std::sort(cy.begin(), cy.end());

	int oldSize = cy.size();
	int newSize = static_cast<int> (oldSize * INCLUDED_VALUES_RATIO);
	cy.resize(newSize);

	double sum = 0.0;

	for (int i = 0; i < newSize; i++) {
		sum += cy[i];
	}

	return sum / newSize;
}

bool greaterThan(double x, double y) {
	return x > y;
}

double GaussFit::dispersion(bool negative, const blitz::Array<double, 1>& y, double value) const {

	std::vector<double> cy(y.size());
	for (int i = 0; i < y.size(); i++) {
		cy[i] = y(i);
	}

	if (negative) {
		std::sort(cy.begin(), cy.end(), greaterThan);
	} else {
		std::sort(cy.begin(), cy.end());
	}

	int oldSize = cy.size();
	int newSize = static_cast<int> (oldSize * INCLUDED_VALUES_RATIO);
	cy.resize(newSize);

	double sum = 0.0;

	for (int i = 0; i < newSize; i++) {
		sum += (cy[i] - value) * (cy[i] - value);
	}

	return sqrt(sum / newSize);
}

//////////////////////////////////////////////////////////////
/// the constructors
GaussFit::GaussFit() :
	m_chiSquare(0),
	m_status(0),
	a0(0), a1(0), a2(0), a3(0), a4(0), a5(0),
	np(0),
	n(0),
	m_minIndex(0), m_maxIndex(0),
	m_dispersionTest(0),
	m_quality(0),
	m_dataDispersionRatio(0) {
}

////////////////////////////////////////////////////
//// this is the call to the LMFit routine
//// this part is called by all other creators
void GaussFit::fit(const Array<double, 1>& y, const Array<double, 1>& x, const Array<double, 1>& firstGuess, unsigned int nbParameters, bool useStandardErrors, bool dispersionTest)  {

	m_minIndex = y.lbound(firstDim);
	m_maxIndex = y.ubound(firstDim);
	m_dispersionTest = dispersionTest;
	m_dataDispersionRatio = 0.0;

	// look at the numner of parameters, verify if it lies between 3 and 6
	if (nbParameters < 3 || nbParameters > 6)
		throw invalid_argument("GaussFit::fit: number of parameters must be greater than 2 and lower than 7");

	// verify that x and y have the same size
	if (x.size() != y.size()) {
		throw domain_error("GaussFit::fit: x size and y size must be equal");
	}

	//cout << "searching y = " << y << endl;

	n = x.size();
	np = nbParameters;

	double sqrt2 = 1.414213562;
	/// NOTE ON SIGMA: The numerical recipe uses sqrt(2)* sigma rather that sigma to perform the gaussian fit
	/// as this is more efficient in terms of computing and convergence. So this routine makes the conversion
	/// here. The rms guess a(2) is multiplied by sqrt(2) before calling the LMFit. The RMS is then redivided
	/// sqrt(2)     after the fit. In brief, GaussFit works with  sigmas and LMFit with sqrt(2)*sigma
	// ANOTHER REMARK: used as it is in LMfit, the sigma can converge towards a negative value, that's normal
	// just make it abs

	// create the standard error, and fill it with the righ values
	Array<double, 1> sig(n);

	if (useStandardErrors)
		sig = standardError(y);
	else
		sig = 1.0;// no error

	try {
		if (nbParameters == 3) {
			// simple gaussian
			// create the first guess
			Array<double, 1> a(3);
			a = firstGuess(0), firstGuess(1), firstGuess(2) * sqrt2;
			Array<bool, 1> ia(3);
			ia = true, true, true;

			LMFit fit(x, y, sig, a, ia, LMFIT_GAUSSIAN); // indicates which variable should be fitted
			m_status = fit.status(); // getting the status
			m_chiSquare = fit.chiSquare(); // the chisquare
			m_quality = fit.quality();
			a = fit.parameters(); // and retrieving the parameters

			a0 = a(0);
			a1 = a(1);
			a2 = std::abs(a(2)) / sqrt2;
			a3 = 0;
			a4 = 0;
			a5 = 0;
		}

		if (nbParameters == 4) {
			// simple gaussian
			// create the first guess
			Array<double, 1> a(6);
			a = firstGuess(0), firstGuess(1), firstGuess(2) * sqrt2, firstGuess(3), 0, 0;
			Array<bool, 1> ia(6);
			ia = true, true, true, true, false, false;

			LMFit fit(x, y, sig, a, ia, LMFIT_GAUSSIAN_PARABOLA); // indicates which variable should be fitted
			m_status = fit.status(); // getting the status
			m_chiSquare = fit.chiSquare(); // the chisquare
			m_quality = fit.quality();
			a = fit.parameters(); // and retrieving the parameters

			a0 = a(0);
			a1 = a(1);
			a2 = std::abs(a(2)) / sqrt2;
			a3 = a(3);
			a4 = 0;
			a5 = 0;
		}

		if (nbParameters == 5) {
			// simple gaussian
			// create the first guess
			Array<double, 1> a(6);
			a = firstGuess(0), firstGuess(1), firstGuess(2) * sqrt2, firstGuess(3), firstGuess(4), 0;
			Array<bool, 1> ia(6);
			ia = true, true, true, true, true, false;

			LMFit fit(x, y, sig, a, ia, LMFIT_GAUSSIAN_PARABOLA); // indicates which variable should be fitted
			m_status = fit.status(); // getting the status
			m_chiSquare = fit.chiSquare(); // the chisquare
			m_quality = fit.quality();
			a = fit.parameters(); // and retrieving the parameters

			a0 = a(0);
			a1 = a(1);
			a2 = std::abs(a(2)) / sqrt2;
			a3 = a(3);
			a4 = a(4);
			a5 = 0;
		}

		if (nbParameters == 6) {
			// simple gaussian
			// create the first guess
			Array<double, 1> a(6);
			a = firstGuess(0), firstGuess(1), firstGuess(2) * sqrt2, firstGuess(3), firstGuess(4), firstGuess(5);
			Array<bool, 1> ia(6);
			ia = true, true, true, true, true, true;

			LMFit fit(x, y, sig, a, ia, LMFIT_GAUSSIAN_PARABOLA); // indicates which variable should be fitted
			m_status = fit.status(); // getting the status
			m_chiSquare = fit.chiSquare(); // the chisquare
			m_quality = fit.quality();
			a = fit.parameters(); // and retrieving the parameters

			a0 = a(0);
			a1 = a(1);
			a2 = std::abs(a(2)) / sqrt2;
			a3 = a(3);
			a4 = a(4);
			a5 = a(5);
		}
	} catch (const logic_error& error) {
		cerr << error.what() << endl;
		throw domain_error("GaussFit::fit: error in computing fit");
	}

	///////////////////////////////////////////////////////////
	// testing the result of the fit with the vertical moments
	///////////////////////////////////////////////////////////
	VerticalMoments moments(x, y);

	double centerOfMass = moments.mmean(); // get the center of mass of the data (vertical mean)
	double dataAmplitude = 0.0;

	//cout << "a0=" << a0 << " a1=" << a1 << " a2=" << a2 << " a3=" << a3 << " a4=" << a4 << " a5=" << a5 << endl;

	bool negative = (a0 < 0.0);

	if (negative) {
		dataAmplitude = min(y) - a3;
	} else {
		dataAmplitude = max(y) - a3;
	}

	double fitAmplitude = a0 + a4 * a1 + a5 * a1 * a1;

	if (isZero(centerOfMass + a1)) {
		throw domain_error("GaussFit::fit: division by zero in delta position calculation");
	}

	if (isZero(dataAmplitude + fitAmplitude)) {
		throw domain_error("GaussFit::fit: division by zero in delta amplitude calculation");
	}

	double deltaAmplitude = abs((dataAmplitude - fitAmplitude) / (dataAmplitude + fitAmplitude));

	double reference = a3;
//	cout << "reference = " << reference << endl;
	double dataDispersion = dispersion(negative, y, reference);
//	cout << "dispersion   = " << dataDispersion << endl;
	m_dataDispersionRatio = abs(dataAmplitude) / dataDispersion;

//	cout << "r1 = " << (dataAmplitude - reference) / dispersion(y, reference) << endl;
//	reference = average(y);
//	cout << "reference = " << reference << endl;
//	cout << "dispersion = " << dispersion(y, reference) << endl;
//	cout << "r2 = " << (dataAmplitude - reference) / dispersion(y, reference) << endl;


//	std::cout << "vertical mean = " << centerOfMass << " fit mean = " << a1 << std::endl;
//	std::cout << "max value = " << dataAmplitude << " fit amplitude = " << fitAmplitude << std::endl;

	bool testAmplitude = (deltaAmplitude > DELTA_AMPLITUDE_MAX);
	bool testDispersion = m_dispersionTest && (m_dataDispersionRatio < DISPERSION_RATIO_MIN);

//	cout << "position is different by " << deltaPosition << " (" << testPosition << ")" << std::endl;
//	cout << "amplitude is different by " << deltaAmplitude << " (" << testAmplitude << ")" << endl;
//	cout << "dispersion ratio = " << m_dataDispersionRatio  << " (" << testDispersion << ")" << endl << endl;
//	cout << "GaussFit " << deltaPosition << " " << deltaAmplitude << endl;

	// if there is a difference of more than 10 % on both values, then say the fit is not correct
	if (testAmplitude || testDispersion) {
		m_status = false;
		//cout << "GaussFit status false " << testAmplitude << " " << testDispersion << endl;

	} else {
		m_status = true;
		//cout << "GaussFit status true" << endl;
	}
}

////////////////////////////////////////////////////
//// this function build the standard error
Array<double, 1> GaussFit::standardError(const Array<double, 1> &y) {
	// the standard error os the square root of y for its valuesgreater or equal to 1, elsewhere 0
	int size = y.size();
	Array<double, 1> errors(size);

	for (int i = 0; i < size; i++)
		errors(i) = (y(i) >= 1) ? std::sqrt(y(i)) : 0;

	return errors;
}

////////////////////////////////////////////////////
//// Functions for fited the interpolated curve
//// The  curve is a 2D array containing first the abscisse, second the fitted values

// this function returns the interpolated curve between indexes.
Array<double, 2> GaussFit::getFittedCurve(unsigned int minIndex, unsigned int maxIndex) const  {
	return getFittedCurve(minIndex, maxIndex, maxIndex - minIndex + 1);
}

// returns the innterpolated curve on the full length of y
Array<double, 2> GaussFit::getFittedCurve() const  {
	return getFittedCurve(m_minIndex, m_maxIndex, m_maxIndex - m_minIndex + 1);
}

// returns the innterpolated curve on the full length of y
Array<double, 2> GaussFit::getFittedCurve(unsigned int nbSamples) const  {
	return getFittedCurve(m_minIndex, m_maxIndex, nbSamples);
}

Array<double, 2> GaussFit::getFittedCurve(unsigned int minIndex, unsigned int maxIndex, unsigned int nbSamples) const  {
	// this function computes the interpolation of the function between the two given indexed and the
	// wanted number of samples. This is thought for a smooth display.
	// make the abscices

	// the number of sample should be greater that 1

	if (nbSamples < 2)
		throw invalid_argument("GaussFit::getFittedCurve: nbSamples must be greater than 1");

	if (maxIndex <= minIndex)
		throw invalid_argument("GaussFit::getFittedCurve: maxIndex must be greater than minIndex");

	Array<double, 1> x(nbSamples);
	Array<double, 1> y(nbSamples);
	firstIndex i;
	x = i;
	x *= 1 / (nbSamples - 1) * (maxIndex - minIndex) + minIndex; // making the x values

	switch (np) {
	case GAUSSFIT_SIMPLE_GAUSSIAN:
		y = a0 * exp(-pow2((x - a1) / a2) / 2);
		break;
	case GAUSSFIT_GAUSSIAN_PEDESTAL:
		y = a0 * exp(-pow2((x - a1) / a2) / 2) + a3;
		break;
	case GAUSSFIT_GAUSSIAN_STRAIGTHLINE:
		y = a0 * exp(-pow2((x - a1) / a2) / 2) + a3 + a4 * x;
		break;
	case GAUSSFIT_GAUSSIAN_PARABOLA:
		y = a0 * exp(-pow2((x - a1) / a2) / 2) + a3 + a4 * x + a5 * x * x;
		break;
	default:
		break;
	}

	/// QUESTION QUAND ON CREE A PARTIR D'AUTRE VECTEURS J'AI L'IMPRESSION QUE
	/// L'ON NE REALLOUE PAS DE NOUVELLE MEMOIRE ... PARCE QUE SI C'EST LE CAS
	//// C'EST LE MOMENT ICI ...

	Array<double, 2> fittedCurve(2, nbSamples);
	fittedCurve(0, Range::all()) = x;
	fittedCurve(1, Range::all()) = y;

	return fittedCurve;
}

///////////////////////////////////////////////////////
///////// fit printout for testing

ostream& operator<<(ostream &os, const GaussFit& fit) {
	double a1 = fit.mean();
	double a2 = fit.rms();
	double a0 = fit.gaussianAmplitude();
	double a3 = fit.pedestalOrder(0);
	double a4 = fit.pedestalOrder(1);
	double a5 = fit.pedestalOrder(2);

	os << "-- Fit result ----------------" << std::endl;

	switch (fit.np) {
	case GAUSSFIT_SIMPLE_GAUSSIAN:
		os << "y=a0 Exp(-((x-a1)/a2)^2 /2) " << std::endl << "with a0 = " << a0 << " a1 = " << a1 << " a2 = " << a2 << std::endl;
		break;
	case GAUSSFIT_GAUSSIAN_PEDESTAL:
		os << "y=a0 Exp(-((x-a1)/a2)^2 /2) + a3" << std::endl << "with a0 = " << a0 << " a1 = " << a1 << " a2 = " << a2 << " a3 = " << a3 << std::endl;
		break;
	case GAUSSFIT_GAUSSIAN_STRAIGTHLINE:
		os << "y=a0 Exp(-((x-a1)/a2)^2 /2) + a3 + a4 x" << std::endl << "with a0 = " << a0 << " a1 = " << a1 << " a2 = " << a2 << " a3 = " << a3 << " a4 = " << a4 << std::endl;
		break;
	case GAUSSFIT_GAUSSIAN_PARABOLA:
		os << "y=a0 Exp(-((x-a1)/a2)^2 /2) + a3 + a4 x +a5 x^2" << std::endl << "with a0 = " << a0 << " a1 = " << a1 << " a2 = " << a2 << " a3 = " << a3 << " a4 = " << a4 << " a5 = " << a5 << std::endl;
		break;
	default:
		break;
	}

	if (fit.fitHasConverged())
		os << "Fit has converged" << std::endl;
	else
		os << "Fit has NOT converged correctly" << std::endl;

	os << "chi square = " << fit.chiSquare() << std::endl;
	os << "------------------------------" << std::endl;

	return os;
}
