/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef GAUSSFIT_H
#define GAUSSFIT_H

#include <iostream>
#include <stdexcept>
#include <blitz/array.h>   
#include <math.h>

#include "Moments.h"
#include "LMFit.h"

// some definitions
static const bool GAUSSFIT_USE_STDERR = true;
static const bool GAUSSFIT_NO_STDERR = false;
static const int GAUSSFIT_SIMPLE_GAUSSIAN = 3;
static const int GAUSSFIT_GAUSSIAN_PEDESTAL = 4;
static const int GAUSSFIT_GAUSSIAN_STRAIGTHLINE = 5;
static const int GAUSSFIT_GAUSSIAN_PARABOLA = 6;

/*! \brief A gaussian fit  */

/** This object performs a gaussian fit with quadratic pedestal on a data set.
 *
 */

class GaussFit {
public:
	/// General constructor fitting y values on x.
	/// @param y the array containing ordinates values
	/// @param x the array containing the abscissa values (same size as x)
	/// @param firstGuess the array initial guess for the fit parameters. This array should be of length \a nbParameters. See note below for initial guess.
	/// @param nbParameters the number of parameters, or said diferently the kind of fit to perform:
	/// - GAUSSFIT_SIMPLE_GAUSSIAN (simple gaussian, amplitude, mean & rms)
	/// - GAUSSFIT_GAUSSIAN_PEDESTAL (simple gaussian, amplitude, mean & rms plus a constant pedestal)
	/// - GAUSSFIT_GAUSSIAN_STRAIGTHLINE (simple gaussian, amplitude, mean & rms plus a straigth line as pedestal)
	/// - GAUSSFIT_GAUSSIAN_PARABOLA (simple gaussian, amplitude, mean & rms plus a quadratic pedestal)
	/// @param useStandardErrors tell the fit to use standard errors in sqrt(N) for N neutrons. This takes the values :
	/// - GAUSSFIT_USE_STDERR use standard errors
	/// - GAUSSFIT_NO_STDERR do not use standard errors
	///
	/// <br>
	/// <b> Note on initial guess</b>.
	/// The fit routine use the Levenberg-Marquardt algorithm provided by LMFit.
	/// The most critical parameters for a good convergence is the rms. Make it smaller that the value you estimate and the convergence will
	/// be easier. <BR>
	/// Position comes next in criticallity. A good way of getting the estimate is to use the mean return by verticalMoments, that can be used
	/// as well for an estimate of the rms.
	/// Amplitude is less important, but it should be of the same order of magnitude than the real amplitude. Using the function max (or min) works perfectly.
	/// Finally, pedestal parameters ar less important and can be set to 0. However a faster convergence will be reached if the pedestal (order 0) is
	/// set to the min (or max) of the data set.
	GaussFit();

	// the fit itself
	void fit(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x, const blitz::Array<double, 1>& firstGuess, unsigned int nbParameters, bool useStandardErrors, bool dispersionTest = true) ;

	// some accessors
	/// Indicates that the fit has converged
	bool fitHasConverged() const; // obvious
	/// Returns the chi square of the fit
	double chiSquare() const;
	/// rReturns the means (position) the peak
	double mean() const; // parameter a1
	/// Returns the rms of the gaussian
	double rms() const; // parameter a2
	/// This returns the amplitude of the fitted function, <b> that without the pedestal value</b>.
	double gaussianAmplitude() const; // parameter a0 (be careful, not a0 + the pedestal, else user amplutide)
	/// Returns the full amplitude of the peak, meaning with the pedestal value. (amplitude+pedestal)
	double amplitude() const; // the amplitude of the peak with the pedestal value
	/// Return the order of the pedestal (0 , 1 or 2)
	double pedestalOrder(unsigned int order) const; // gives a4, a5 and a6 for 0, 1 2 orders

	// getting the interpolated curve
	/// This method returns an 2 dimensions arrays containing abscissa (index 0) and ordinates (index 1) of the fitted curve.
	/// It is essentially thought for data display.
	/// @param minIndex the lower index value for the outputing the curve
	/// @param maxIndex the upper bound
	/// @nbSamples the number of samples used for producing the fitted curves
	blitz::Array<double, 2> getFittedCurve(unsigned int minIndex, unsigned int maxIndex, unsigned int nbSamples) const // same but without the number of samples
			;

	/// Same call as the general call above, but
	/// the x value bounds are derived from the y bounds.
	blitz::Array<double, 2> getFittedCurve(unsigned int nbSamples) const // same as above, but on the full curve
			;
	/// Same call  as the general call , but just with the definition
	/// of bounds.
	blitz::Array<double, 2> getFittedCurve(unsigned int minIndex, unsigned int maxIndex) const // same but without the number of samples
			;

	/// Same call  as the general call , taking the bounds and number of indexes between the bound to produce the curve.
	blitz::Array<double, 2> getFittedCurve() const // taking the min and max index that have been given
			;

	double quality() const;
	double dispersionRatio() const;

	/// Gauss fits cans be printed out for debugging purposes.
	/// it produce displays like this one.<br>
	/// <code>
	///-- Fit result ----------------<br>
	///y=a0 Exp(-((x-a1)/a2)^2 /2) + a3<br>
	///with a0=92.1051 a1=150.792 a2=2.45765 a3=0.657799<br>
	///     ///Fit has converged<br>
	///chi square=38.0078<br>
	///------------------------------<br>
	/// </code>
	friend std::ostream& operator<<(std::ostream &os, const GaussFit& fit);

private:
	// fit results
	double m_chiSquare;
	double m_status;
	double a0, a1, a2, a3, a4, a5; // the parameters

	// remembering some values
	int np; // number of parameters
	unsigned int n; // number of samples in y (and x!)
	unsigned int m_minIndex, m_maxIndex; // the indexes that have been used during the creation

	bool m_dispersionTest;

	double m_quality;
	double m_dataDispersionRatio;

	// a function that creates the standard error array
	blitz::Array<double, 1> standardError(const blitz::Array<double, 1>& y);

	// Calculates the average
	double average(const blitz::Array<double, 1>& data) const;

	// Calculates the dispersion
	double dispersion(bool negative, const blitz::Array<double, 1>& data, double value) const;

	static const double INCLUDED_VALUES_RATIO;
	static const double DELTA_POSITION_MAX;
	static const double DELTA_AMPLITUDE_MAX;
	static const double DISPERSION_RATIO_MIN;
};

inline bool GaussFit::fitHasConverged() const {
	return m_status;
}

inline double GaussFit::chiSquare() const {
	return m_chiSquare;
}

inline double GaussFit::mean() const {
	return a1;
}

inline double GaussFit::rms() const {
	return a2;
}

inline double GaussFit::gaussianAmplitude() const {
	return a0;
}

inline double GaussFit::amplitude() const {
	return a0 + a3 + a4 * a1 + a5 * a1 * a1;
}

inline double GaussFit::pedestalOrder(unsigned int order) const {
	if (order == 0)
		return a3;

	if (order == 1)
		return a4;

	if (order == 2)
		return a5;

	return 0; // by default
}

inline double GaussFit::quality() const {
	return m_quality;
}

inline double GaussFit::dispersionRatio() const {
	return m_dataDispersionRatio;
}

#endif // GAUSSFIT_H
