/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef BZ_BLITZ_H
 #include <blitz/blitz.h>
#endif

#include <blitz/array.h>

namespace blitz {

////////////////////////////////////////////////////////////
/// Main histogramming routine
////////////////////////////////////////////////////////////

template<typename T>
Histogram histogram(const Array<T, 1>& data, double minValue, double maxValue, int nbBins) 
{
  if (nbBins < 2)
    throw invalid_argument("histogram: nbBins must be greater than 1!");

  if (minValue >= maxValue)
    throw invalid_argument("histogram: maxValue must be greater than minValue!");

  // create the histogram
  // the first vector contain the bin start
  // the second vector the bin end
  Array<double, 2> histogram(2, nbBins);

  // create the bin values
  firstIndex i;
  secondIndex j;

  histogram(0, Range::all()) = i;
  histogram(0, Range::all()) = histogram(0, Range::all())/(nbBins - 1)*(maxValue - minValue) + minValue;

  // initializing the histogram to 0
  histogram(1, Range::all()) = 0;

  // fill the histogram
  double factor = (nbBins - 1)/(maxValue - minValue);
  unsigned int underflows = 0;
  unsigned int overflows = 0;

  for (int i = data.lbound(firstDim); i < data.ubound(firstDim); ++i)
  {
    int binIndex;
    double binValue = data(i);

    if (binValue < minValue)
      underflows++; // overflows and underflows
    else
    {
      if (binValue >= maxValue)
        overflows++;
      else
      {
       // compute the bin number
        binIndex = (unsigned int)((binValue - minValue)*factor + 0.5);
        histogram(1, binIndex) += 1; // is that efficient ?
      }
    }
  } // for all values of data

  Histogram result(nbBins);

  result.array = histogram(i, j);
  result.overflows = overflows;
  result.underflows = underflows;

  return result;
}

////////////////////////////////////////////////////////////
/// Other callers
////////////////////////////////////////////////////////////

template<typename T>
Histogram histogram(const Array<T, 1>& data, int nbBins) 
{
  // the bin limits will be computed as the minimum and maximum
  double minValue = min(data);
  double maxValue = max(data);

  return histogram(data, minValue, maxValue, nbBins);
}

}