/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Tools.h"

#include <blitz/array.h>  

extern "C"
{
  void dgesv_(int *n, int *k, double *A, int *lda, int *ipiv, double *X, int *ldx, int *info);
}

namespace blitz {

template<typename T>
Array<T, 2> inverse(Array<T, 2> const & A, double & residual) 
{
  int N = A.rows();
  
  if (N != A.columns())
    throw std::domain_error("inverse: cannot inverse a non-square matrix!");
  
  double* fortranArray = new double[A.size()];
  
  // iterating the array to fill a Fortran array. We fill the fortran array by column,
  // whereas the blitz array is a C-style array.
  typename Array<T, 2>::const_iterator a = A.begin();

  for (int i = 0; i < N; ++i)
  {
    for (int j = 0; j < N; ++j)
    {
      fortranArray[j*N + i] = *a;
      ++a;
    }  
  }

  // using lapack inverse. LU decomposition.
  int* pivot = new int[A.size()];
  int ok;
  
  // creating Identity matrix as result.
  double* resultArray = new double[A.size()];
  
  for (int i = 0; i < N; ++i)
  {
    for (int j = 0; j < N; ++j)
      resultArray[j*N + i] = (i == j ? 1.0 : 0.0);
  }
  
  // lapack call
  dgesv_(&N, &N, fortranArray, &N, pivot, resultArray, &N, &ok);
  
  delete [] fortranArray;
  delete [] pivot;
  
  if (ok == 1)
    throw std::domain_error("inverse: lapack error!");
  
  // result
  Array<T, 2> R(N, N);
  
  typename Array<T, 2>::iterator r = R.begin();

  for (int i = 0; i < N; ++i)
  {
    for (int j = 0; j < N; ++j)
    {
      *r = resultArray[j*N + i];
      ++r;
    }  
  }

  delete [] resultArray;

  // Calculating residual with norm1.
  Array<T, 2> Res(N, N);
  Res = product(A, R) - Identity<double>(N);
  
  residual = norm1(Res);
      
  return R;
}

template<typename T> 
Array<T, 2> inverse(Array<T, 2> const & A) 
{
  double res = 0.0;
  
  return inverse<T>(A, res);
}

}