/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "LMFit.h"

#include <cminpack-1/cminpack.h>
#include <math.h>
#include <sstream>

using namespace std;
using namespace blitz;

double f4(double x, const double * p) {

	double arg = (x - p[1]) / p[2];
	double ex = exp(-arg * arg);
	double y = p[0] * ex + p[3];

	return y;
}

double exponential3(double x, const double *p) {

	double y = p[0]*exp(p[1]*x) + p[2];

	return y;

}

struct Data {
	double * t;
	double * y;
};


int fcn(void *p, int m, int n, const double *x, double *fvec, int iflag) {

	Data * data = (Data *) p;

	double * pt = data->t;
	double * py = data->y;

	for (int i = 0; i < m; i++) {
		fvec[i] = py[i] - f4(pt[i], x);
	}
	return 0;
}


//to calculate error function for exponential fitting
int exponentialFunction(void *p, int m, int n, const double *x, double *fvec, int iflag) {
	Data * data = (Data *) p;

	double * pt = data->t;
	double * py = data->y;

	for (int i = 0; i < m; i++) {
		fvec[i] = py[i] - exponential3(pt[i], x);
	}

	//cout << "exponential function" << endl;

	return 0;
}



void LMFit::calculateLmdif(const blitz::Array<double, 1>& x, const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& a) {

	int n;

	if (function_type != LMFIT_EXPONENTIAL){
		n = 4;
	} else {
		n = 3;
	}

	int m = x.size();
	int info;
	int lwa = m * n + 5 * n + m;
	int *iwa = new int[n];

	double tol = sqrt(dpmpar(1));
	double *p = new double[n];

	for (int i = 0; i < n; i++) {
		p[i] = a(i);
	}

	// test negative peak
	bool negative = false;

	if (function_type != LMFIT_EXPONENTIAL) {
		if (p[0] < 0) {
			negative = true;
			p[0] = -p[0];
			p[3] = -p[3];
		}
	}
	//cout << "lmdif with a = " << p[0] << " " << p[1] << " " << p[2] << " " << p[3] << endl;

	double *fvec = new double[m];
	double *wa = new double[lwa];

	double * pt = new double[m];
	double * py = new double[m];

	for (int i = 0; i < m; i++) {
		pt[i] = x(i);

		if (negative) {
			py[i] = -y(i);
		} else {
			py[i] = y(i);
		}
	}

	Data data;
	data.t = pt;
	data.y = py;

	if (function_type != LMFIT_EXPONENTIAL) {
		info = lmdif1(fcn, (void *) &data, m, n, p, fvec, tol, iwa, wa, lwa);
	} else {
		info = lmdif1(exponentialFunction, (void *) &data, m, n, p, fvec, tol, iwa, wa, lwa);
	}

	double chi = enorm(m, fvec);

//	cout << "lmdif info result = " << info << endl;
//	cout << "lmdif found a = " << p[0] << " " << p[1] << " " << p[2] << " " << p[3] << endl;

	if (negative) {
		p[0] = -p[0];
		p[3] = -p[3];
	}

	if (chi > 0.0) {
		m_quality = m / chi;
	}

	//cout << "lmdif fnorm = " << chi << " quality = " << m_quality << endl;

	for (int i = 0; i < n; i++) {
		m_a(i) = p[i];
	}

	m_chisq = chi * chi;

	// cleaning allocated arrays
	delete [] iwa;
	delete [] p;
	delete [] fvec;
	delete [] wa;
	delete [] pt;
	delete [] py;
}

//////////////////////////////////////////////////////////////////////////
////// the fit wrapper
//////////////////////////////////////////////////////////////////////////
///// this code is essentially a wrapper for the numerical recipe functions for
///// general curve fitting.

LMFit::LMFit(const Array<double, 1>& x, const Array<double, 1>& y, const Array<double, 1>& sig, const Array<double, 1>& a, const Array<bool, 1>& ia, int functionType)  :
	m_a(a.size()), m_status(true), m_quality(0.0) {
	if (x.size() != y.size())
		throw(std::invalid_argument("LMFit::LMFit: x size must equal y size!"));

	//keep function type
	function_type = functionType;

	calculateLmdif(x, y, a);

}
