/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef LMFIT_H
#define LMFIT_H

#include <iostream>
#include <stdexcept>
#include <blitz/array.h>
#include <math.h>

#include "Threshold.h"

// definitions of function types
static const int LMFIT_GAUSSIAN = 1; // simple gaussian
static const int LMFIT_GAUSSIAN_PARABOLA = 2; // simple gaussian + a parabola
static const int LMFIT_EXPONENTIAL = 3; //simple exponential
static const int LMFIT_MAX_ITERATION = 200;
static const double LMFIT_PRECISION = 1e-7;

/*! \brief A Levenberg-Marquardt fitter.  */

/// This object is essentially a wrapper for C++ routines provide by numerical recipes.
/// a complete description of the method can be found <a href='numericalRecipeFit.pdf'> here </a>.
/// No further information is provided here, as this function should never be called directly, but rather
/// throught GaussFit or Lorentzian fit (not yet implemented).
/// Presently only gaussians (up to 6 parameters) are implemented.


class LMFit {
public:
	// call the fit on x and y,
	LMFit(const blitz::Array<double, 1>& x, // the absices
			const blitz::Array<double, 1>& y, // the ordinates
			const blitz::Array<double, 1>& sig, // standard deviations
			const blitz::Array<double, 1>& a, // first guess for parameters
			const blitz::Array<bool, 1>& ia, // indicates which variable should be fitted
			int functionType)// the function type
			;

	// getting the results
	double chiSquare() const; // returns the chi square
	bool status() const;
	blitz::Array<double, 1> parameters() const;
	double quality() const;

private:
	void calculateLmdif(const blitz::Array<double, 1>& x, const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& a);

	blitz::Array<double, 1> m_a;
	double m_chisq;
	bool m_status; // if status is true, the fit has converged correctly
	double m_quality;

protected:
	int function_type;
};

inline double LMFit::chiSquare() const {
	return m_chisq;
}

inline bool LMFit::status() const {
	return m_status;
}

inline blitz::Array<double, 1> LMFit::parameters() const {
	return m_a;
}

inline double LMFit::quality() const {
	return m_quality;
}

#endif // LMFIT_H
