/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef BZ_ARRAY_MOMENTS_H
#define BZ_ARRAY_MOMENTS_H

#include <blitz/array.h>
#include <stdexcept>

#ifndef BZ_ARRAY_H
 #error <blitz/array/myMath.h> must be included after <blitz/array.h>
#endif

#include "Threshold.h"

/*! \file moments.h 
\brief Compute vertical and horizontal moments of 1D data
*/

namespace blitz {

class DirectionMoments
{
public:
  DirectionMoments()
  : m_sum(0), m_mean(0), m_variance(0), m_rms(0)
  {
  }
  
  void setMSum(double sum);
  void setMMean(double mean);
  void setMVariance(double variance);
  void setMRms(double rms);
  
  double msum() const;
  double mmean() const;
  double mvariance() const;
  double mrms() const;
  
private:
  double m_sum;
  double m_mean;
  double m_variance;
  double m_rms;
};

inline double DirectionMoments::msum() const
{
  return m_sum;
}

inline double DirectionMoments::mmean() const
{
  return m_mean;
}

inline double DirectionMoments::mvariance() const
{
  return m_variance;
}

inline double DirectionMoments::mrms() const
{
  return m_rms;
}

class Moments
{
public:
  template<typename T>
  Moments(const Array<T, 1>& y) 
  {
    int lb = y.lbound(firstDim), ub = y.ubound(firstDim);
    Array<T, 1> x(ub - lb + 1);
    firstIndex i;
    x= i + lb;
  
    compute(x, y);
  }

  template<typename T>
  Moments(const Array<T, 1>& x, const Array<T, 1>& y) 
  {
    compute(x, y);
  }

  const DirectionMoments& horizontalMoments() const;
  const DirectionMoments& verticalMoments() const;
  
private:
  DirectionMoments m_horizontalMoments;
  DirectionMoments m_verticalMoments;
    
  template<typename T>
  void compute(const Array<T,1>& x, const Array<T,1>& y) 
  {
    if (x.size() != y.size())
      throw (std::invalid_argument("Moments::compute: x size must equal y size"));
   
    if (x.size() == 0)
      throw (std::invalid_argument("Moments::compute: empty arrays"));

    // compute the standard error weight , if the value of y is equal to 0,
    // then the weigth is equal to 0
 
    // compute the sums
    double n = (double)y.size();
    double sY = sum(y);
    double sY2 = sum(pow2(y));
    double sXY = sum(x*y);
    double sX2Y = sum(x*x*y);

    // compute the moments
    double horizontalMean = sY/n;
    double horizontalVariance = sY2/n - horizontalMean*horizontalMean;
    double horizontalRMS = std::sqrt(horizontalVariance);
    
    m_horizontalMoments.setMSum(sY);
    m_horizontalMoments.setMMean(horizontalMean);
    m_horizontalMoments.setMVariance(horizontalVariance);
    m_horizontalMoments.setMRms(horizontalRMS);
    
    if (isZero(sY))
      throw std::domain_error("Moments::compute: division by zero in vertical mean calculation");
    
    double verticalMean = sXY/sY;
    double verticalVariance = sX2Y/sY - verticalMean*verticalMean;
    double verticalRMS = std::sqrt(verticalVariance);

    m_verticalMoments.setMSum(sXY);
    m_verticalMoments.setMMean(verticalMean);
    m_verticalMoments.setMVariance(verticalVariance);
    m_verticalMoments.setMRms(verticalRMS);
  }
};

inline const DirectionMoments& Moments::verticalMoments() const
{
  return m_verticalMoments;
}

inline const DirectionMoments& Moments::horizontalMoments() const
{
  return m_horizontalMoments;
}

class VerticalMoments : public DirectionMoments
{
public:
  template<typename T>
  VerticalMoments(const Array<T, 1>& y) 
  {
    int lb = y.lbound(firstDim), ub = y.ubound(firstDim);
    Array<T, 1> x(ub - lb + 1);
    firstIndex i;
    x = i + lb;
    
    compute(x, y);
  }

  template<typename T>
  VerticalMoments(const Array<T, 1>& x, const Array<T, 1>& y) 
  {
    compute(x, y);
  }

private:
  
  template<typename T>
  void compute(const Array<T,1>& x, const Array<T,1>& y) 
  {
    if (x.size() != y.size())
      throw (std::invalid_argument("Moments::compute: x size must equal y size"));
   
    if (x.size() == 0)
      throw (std::invalid_argument("Moments::compute: empty arrays"));

    // compute the standard error weight , if the value of y is equal to 0,
    // then the weigth is equal to 0
 
    // compute the sums
    double sY = sum(y);
    double sXY = sum(x*y);
    double sX2Y = sum(x*x*y);

    if (isZero(sY))
      throw std::domain_error("VerticalMoments::compute: division by zero in vertical mean calculation");

    // compute the moments
    double verticalMean = sXY/sY;
    double verticalVariance = sX2Y/sY - verticalMean*verticalMean;
    double verticalRMS = std::sqrt(verticalVariance);

    // returning the results
    Array<double,1> result(4);

    setMSum(sXY);
    setMMean(verticalMean);
    setMVariance(verticalVariance);
    setMRms(verticalRMS);
  }
};

class HorizontalMoments : public DirectionMoments
{
public:
  template<typename T>
  HorizontalMoments(const Array<T, 1>& y) 
  {
    int lb = y.lbound(firstDim), ub = y.ubound(firstDim);
    Array<T, 1> x(ub - lb + 1);
    firstIndex i;
    x = i + lb;
    
    compute(x, y);
  }

  template<typename T>
  HorizontalMoments(const Array<T, 1>& x, const Array<T, 1>& y) 
  {
    compute(x, y);
  }

private:
  
  template<typename T>
  void compute(const Array<T,1>& x, const Array<T,1>& y) 
  {
    if (x.size() != y.size())
      throw (std::invalid_argument("Moments::compute: x size must equal y size"));
   
    if (x.size() == 0)
      throw (std::invalid_argument("Moments::compute: empty arrays"));

    // compute the standard error weight , if the value of y is equal to 0,
    // then the weigth is equal to 0
 
    // compute the sums
    double n=(double)y.size();
    double sY=sum(y);
    double sY2=sum(pow2(y));

    // compute the moments
    double horizontalMean = sY/n;
    double horizontalVariance = sY2/n - horizontalMean*horizontalMean;
    double horizontalRMS = std::sqrt(horizontalVariance);

    setMSum(sY);
    setMMean(horizontalMean);
    setMVariance(horizontalVariance);
    setMRms(horizontalRMS);
  }
};

}

#include "Moments.inl"

#endif // BZ_ARRAY_MOMENTS_H
