/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "MultiGaussPeakFinder.h"
#include "GaussFit.h"
#include <sstream>

using namespace blitz;

////////////////////////////////////////////////////////////////////////////
//// constructors
////////////////////////////////////////////////////////////////////////////

MultiGaussPeakFinder::MultiGaussPeakFinder() :
	m_nbPeaks(0), m_nbGoodPeaks(0), m_nbBadPeaks(0) {
}

MultiGaussPeakFinder::MultiGaussPeakFinder(const Array<double, 1>& y, const Array<double, 1>& x)  {
	// general call
	findPeaks(y, x);
}

MultiGaussPeakFinder::MultiGaussPeakFinder(const Array<double, 1>& y)  {
	// without x (taken as indices of y)
	// creates the x from y indexes
	int n = y.size();
	int minIndex = y.lbound(firstDim);
	Array<double, 1> x(n);
	firstIndex i;
	x = i + minIndex;

	findPeaks(y, x);
}

MultiGaussPeakFinder::~MultiGaussPeakFinder() {
	for (vector<Peak *>::iterator p = peaks.begin(); p != peaks.end(); ++p)
		delete *p;
}

void MultiGaussPeakFinder::calculatePeaks(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x)  {

	findPeaks(y, x);
}

void MultiGaussPeakFinder::calculatePeaks(const blitz::Array<double, 1>& y)  {

	// creates the x from y indexes
	int n = y.size();
	int minIndex = y.lbound(firstDim);
	Array<double, 1> x(n);
	firstIndex i;
	x = i + minIndex;

	findPeaks(y, x);
}

////////////////////////////////////////////////////////////////////////////
//// this is the method where the peak finding takes place
////////////////////////////////////////////////////////////////////////////

void MultiGaussPeakFinder::findPeaks(const Array<double, 1>& y, const Array<double, 1>& x)  {

	if (x.size() != y.size()) {
		throw(std::invalid_argument("MultiGaussPeakFinder::findPeaks: x size must equal y size"));
	}

	if (y.size() == 0 || y.size() == 1) {
		throw(std::invalid_argument("MultiGaussPeakFinder::findPeaks: y must contain data"));
	}

	try {
		Array<double, 2> roughPeaks(roughPeakFinder(y, x));

		// how many peaks have been found ?
		m_nbPeaks = roughPeaks.ubound(firstDim) - roughPeaks.lbound(firstDim) + 1;
		m_nbGoodPeaks = 0;
		m_nbBadPeaks = 0;
		int ny = y.size(); // size of the vector

		/// NOTE: the window used for the fit is four times the rough peak width (which is not the RMS).
		/// Further, a check will be done for avoiding overlapping windows of different peaks. This,
		/// to avoid problems with peaks that are very close to each other.
		/// In case of overlapping, the peak window sides will be set to a mea, value.
		/// So, the peaks are created with rough windowing, refined afterwards.

		//////////////////////////////////////////////////////////////////////
		// create the peaks and fill them with the rough peaks values
		//////////////////////////////////////////////////////////////////////
		Peak *newPeak;

		if (m_nbPeaks == 0) {
			m_nbPeaks = 1;
			// look for max index
			double indexMax = 0.0;
			double maxValue = 0.0;
			for (long k = 0; k < x.size(); ++k) {
				if (y(k) > maxValue)
					indexMax = (double) (k);
			}

			int peakType = PEAK_GAUSSIAN_PEDESTAL;
			double a0 = max(y); // estimate of the amplitude
			double a1 = indexMax; // estimate of the position
			double a2 = x.size() / 3.0; // estimate of the width
			double a3 = min(y); // estimate of the pedestal
			double a4 = 0; // unused
			double a5 = 0; // unused
			double a6 = 0; // unused
			//        int minIndex = (int)(roughPeaks(0, ROUGHPEAK_POSITION) + 0.5 - roughPeaks(0, ROUGHPEAK_WIDTH)*2); // rough windowing, see note above
			//        int maxIndex = (int)(roughPeaks(0, ROUGHPEAK_POSITION) + 0.5 + roughPeaks(0, ROUGHPEAK_WIDTH)*2);
			int minIndex = 0;
			int maxIndex = x.size() - 1;
			// the window used for the gaussian fit is twice the width returned by the rough peak finder.

			newPeak = new Peak(minIndex, maxIndex, peakType, a0, a1, a2, a3, a4, a5, a6, 0); // create the peak
			peaks.push_back(newPeak);
		} else {
			for (int i = 0; i < m_nbPeaks; i++) {
				int peakType = PEAK_GAUSSIAN_PEDESTAL;
				double a0 = roughPeaks(i, ROUGHPEAK_AMPLITUDE); // estimate of the amplitude
				double a1 = roughPeaks(i, ROUGHPEAK_POSITION); // estimate of the mean (as the position)
				double a2 = roughPeaks(i, ROUGHPEAK_WIDTH) / 2; // estimate of the rms as half the width
				double a3 = roughPeaks(i, ROUGHPEAK_PEDESTAL); // estimate of the pedestal
				double a4 = 0; // unused
				double a5 = 0; // unused
				double a6 = 0; // unused
				int minIndex = (int) (roughPeaks(i, ROUGHPEAK_POSITION) + 0.5 - roughPeaks(i, ROUGHPEAK_WIDTH) * 2); // rough windowing, see note above
				int maxIndex = (int) (roughPeaks(i, ROUGHPEAK_POSITION) + 0.5 + roughPeaks(i, ROUGHPEAK_WIDTH) * 2);
				// the window used for the gaussian fit is twice the width returned by the rough peak finder.

				newPeak = new Peak(minIndex, maxIndex, peakType, a0, a1, a2, a3, a4, a5, a6, 0); // create the peak
				peaks.push_back(newPeak);
			}
		}

		//////////////////////////////////////////////////////////////////////
		// checking peak fitting windows
		//////////////////////////////////////////////////////////////////////

		// first checking if the windows do no go over the range of index of y
		std::vector<Peak *>::const_iterator e = peaks.end();
		std::vector<Peak *>::const_iterator b = peaks.begin();
		Peak *f = peaks.front();
		Peak *l = peaks.back();
		for (std::vector<Peak *>::const_iterator peak = b; peak != e; ++peak) {
			if ((*peak)->m_minIndex < 0)
				(*peak)->m_minIndex = 0;
			if ((*peak)->m_maxIndex >= ny)
				(*peak)->m_maxIndex = ny - 1;

			//std::cout<<"indexes "<<(*peak)->m_minIndex<<" "<<(*peak)->m_maxIndex<<std::endl;
		}

		// then checking for overlapping fitting windows
		for (std::vector<Peak *>::const_iterator peak = b; peak != e; ++peak) {
			if (*peak != f && *peak != l) {
				// not the first nor the end peak
				if ((*peak)->m_minIndex < (*(peak - 1))->m_maxIndex) {
					// cheking with the precedent peak
					unsigned int newIndex = (unsigned int) ((*peak)->m_a1 + (*(peak - 1))->m_a1) / 2; // take the middle value
					(*peak)->m_minIndex = newIndex;
					(*(peak - 1))->m_maxIndex = newIndex;
				}
				if ((*peak)->m_maxIndex > (*(peak + 1))->m_minIndex) {
					// cheking with the next peak
					unsigned int newIndex = (unsigned int) ((*peak)->m_a1 + (*(peak + 1))->m_a1) / 2; // take the middle value
					(*peak)->m_maxIndex = newIndex;
					(*(peak + 1))->m_minIndex = newIndex;
				}
			}

			if (*peak == f && *peak != l) {
				// the first not the last peak
				if ((*peak)->m_maxIndex > (*(peak + 1))->m_a1) {
					// cheking with the next peak
					unsigned int newIndex = (unsigned int) ((*peak)->m_a1 + (*(peak + 1))->m_a1) / 2; // take the middle value
					(*peak)->m_maxIndex = newIndex;
					(*(peak + 1))->m_minIndex = newIndex;
				}
			}

			if (*peak != f && *peak == l) {
				// not the first but the last
				if ((*peak)->m_minIndex < (*(peak - 1))->m_maxIndex) {
					// cheking with the precedent peak
					unsigned int newIndex = (unsigned int) ((*peak)->m_a1 + (*(peak - 1))->m_a1) / 2; // take the middle value
					(*peak)->m_minIndex = newIndex;
					(*(peak - 1))->m_maxIndex = newIndex;
				}
			}
			// in there is a single peak, well no check is made
		}

		//////////////////////////////////////////////////////////////////////
		// fitting the peaks now
		//////////////////////////////////////////////////////////////////////

		m_aFinal.resize(0, 8);
		m_aFinal = 0.0;
		e = peaks.end();
		b = peaks.begin();
		for (std::vector<Peak *>::const_iterator peak = peaks.begin(); peak != e; ++peak) {

			//			cout << "peak = " << **peak << endl;

			// calculating peak position and width in coordinate space
			int index = (int) ((*peak)->m_a1);
			double peakPosition = x(index) + (abs(x(index + 1) - x(index)) * abs((*peak)->m_a1 - index));
			//			int width = (*peak)->m_a2;
			//			double peakWidth = fabs(x((*peak)->m_maxIndex) - x((*peak)->m_minIndex)) / 3.0;
			//			double peakWidth = 0.0;
			//			if ((index-width) < x(0) || (index-width) < x(x.size()-1) ||
			//	   		 (index+width) > x(x.size()-1) || (index+width) > x(0)) {
			//			     peakWidth = (max(x) - min(x)) / 4.0;
			//			} else {
			//			  peakWidth = abs(x(index + width) - x(index - width))/2.0;
			//			}
			double peakPedestal = (*peak)->m_a3;

			// prepare the fit parameters from the peak info
			unsigned int minIndex = (*peak)->m_minIndex;
			unsigned int maxIndex = (*peak)->m_maxIndex;

			// Calculate peak width
			long halfMax = floor((*peak)->m_a0 / 2.0);
			long nbPointsInPeak = 0;
			long n = minIndex;

			if (!(*peak)->isNegative()) {
				while (n < y.size() && y(n) < halfMax) {
					n++;
				}
				while (n < y.size() && y(n) >= halfMax) {
					n++;
					nbPointsInPeak++;
				}
			} else {
				while (n < y.size() && y(n) > halfMax) {
					n++;
				}
				while (n < y.size() && y(n) <= halfMax) {
					n++;
					nbPointsInPeak++;
				}
			}

			double delta = x(minIndex + 1) - x(minIndex);
			double peakWidth = 0.0;

			if (nbPointsInPeak > 1) {
				peakWidth = delta * (nbPointsInPeak - 1);
			} else {
				peakWidth = delta;
			}

			m_a.resize(6);
			m_a = (*peak)->m_a0, peakPosition, peakWidth, peakPedestal, 0.0, 0.0;
//			cout << "after conversion fit values" << (*peak)->m_a0 << "/" << peakPosition << "/" << peakWidth << "/" << peakPedestal << endl;
			// perform the fit
			GaussFit fit;
			fit.fit(y(Range(minIndex, maxIndex)), x(Range(minIndex, maxIndex)), m_a, GAUSSFIT_GAUSSIAN_PEDESTAL, GAUSSFIT_NO_STDERR);


			// update the peak info with the fit results
			// update parameters and values
			(*peak)->m_chiSquare = fit.chiSquare();
			(*peak)->m_a0 = fit.gaussianAmplitude();
			(*peak)->m_a1 = fit.mean();
			(*peak)->m_a2 = fit.rms();
			(*peak)->m_a3 = fit.pedestalOrder(0);
			(*peak)->m_a4 = fit.pedestalOrder(1);
			(*peak)->m_quality = fit.quality();
			(*peak)->m_dispersionRatio = fit.dispersionRatio();

//			cout << "peak  position " << peakPosition << " (" << nbPointsInPeak << ")" << endl;
//			cout<< "after FIT, peak position" << fit.mean() << endl;

			if (fit.fitHasConverged()) {
				// fit status for this peak
				(*peak)->m_fitHasConverged = true;
				m_nbGoodPeaks++;

			} else {
				(*peak)->m_fitHasConverged = false;
				m_nbBadPeaks++;
			}

			//			cout << "peak modified = " << **peak << endl;

		} // loop on all peaks for fitting/interpolating

		//cout << "found " << m_nbGoodPeaks << " good peaks" << endl;

		m_aFinal.resize(m_nbGoodPeaks, 8);

		// copying data
		int i = 0;
		for (std::vector<Peak *>::const_iterator peak = peaks.begin(); peak != e; ++peak) {
			if ((*peak)->m_fitHasConverged) {
				// prepare the fit parameters from the peak info
				m_aFinal(i, 0) = (*peak)->m_a0;
				m_aFinal(i, 1) = (*peak)->m_a1;
				m_aFinal(i, 2) = (*peak)->m_a2;
				m_aFinal(i, 3) = (*peak)->m_a3;
				m_aFinal(i, 4) = (*peak)->m_a4;
				m_aFinal(i, 5) = 0.0;
				m_aFinal(i, 6) = (*peak)->m_minIndex;
				m_aFinal(i, 7) = (*peak)->m_maxIndex;
				i++;
			}
		}

		//		cout << "a = " << m_aFinal << endl;

	} catch (const logic_error& error) {
		cerr << error.what() << endl;
		throw domain_error("MultiGaussPeakFinder::findPeaks: error in peak calculation");
	}
}

///////////////////////////////////////////////////////
///////// fit printout for testing

ostream& operator<<(ostream &os, const MultiGaussPeakFinder& fit) {

	os << fit.string(false);

	return os;
}

std::string MultiGaussPeakFinder::string(bool allPeaks) const {

	ostringstream os;

	os << "--- " << m_nbGoodPeaks << " peaks found ----------------------" << std::endl;

	int i = 1;
	os << " #       position    amplitude (pedestal+amplitude)              rms         chi square  quality       ratio         min         max   direction      status" << std::endl;

	std::vector<Peak *>::const_iterator e = peaks.end();
	for (std::vector<Peak *>::const_iterator peak = peaks.begin(); peak != e; ++peak) {
		if ((*peak)->correctlyFitted()) {
			os << std::setw(2) << i++ << ". " << *(*peak);
		} else if (allPeaks) {
			os << std::setw(2) << i++ << ". " << *(*peak);
		}
	}

	os << "--------------------------------------" << std::endl;

	return os.str();
}

///////////////////////////////////////////////////////
/////////Getting the highest peak

Peak* MultiGaussPeakFinder::bestPeak() {
	if (peaks.empty()) {
		return 0;
	}

	// searching for the positive and negative good peaks
	vector<Peak *> positivePeaks;
	vector<Peak *> negativePeaks;
	for (std::vector<Peak *>::const_iterator p = peaks.begin(); p != peaks.end(); p++) {
		if ((*p)->correctlyFitted()) {
			if ((*p)->isNegative()) {
				negativePeaks.push_back(*p);
			} else {
				positivePeaks.push_back(*p);
			}
		}
	}

	// parses the vector of peaks and return the peak with the highest total amplitude
	Peak* maximumPeak = 0;

	double maximum = 0;
	for (std::vector<Peak *>::const_iterator p = positivePeaks.begin(); p != positivePeaks.end(); p++) {
		Peak* peak = *p;
		double peakValue = peak->m_a0 + peak->m_a3;
		if (peakValue > maximum) {
			maximum = peakValue;
			maximumPeak = peak;
		}
	}

	double minimum = std::numeric_limits<double>::max();
	Peak* minimumPeak = 0;
	for (std::vector<Peak *>::const_iterator p = negativePeaks.begin(); p != negativePeaks.end(); p++) {
		Peak* peak = *p;
		double peakValue = peak->m_a0 + peak->m_a3;
		if (peakValue < minimum) {
			minimum = peakValue;
			minimumPeak = peak;
		}
	}

	if (maximumPeak != 0 && minimumPeak != 0) {
		// comparing gaussian amplitude
		if (abs(maximumPeak->m_a0) > abs(minimumPeak->m_a0)) {
			//cout << "found the maximum peak" << endl;
			return maximumPeak;
		} else {
			//cout << "found the minimum peak" << endl;
			return minimumPeak;
		}
	}

	if (maximumPeak != 0) {
		//cout << "found the maximum peak" << endl;
		return maximumPeak;
	}

	if (minimumPeak != 0) {
		//cout << "found the minimum peak" << endl;
		return minimumPeak;
	}

	return 0;
}

Peak *MultiGaussPeakFinder::peak(int peakNumber) {
	// returns a vector of peaks by their position;
	if (peakNumber >= 0 && peakNumber < (int) peaks.size())
		return peaks[peakNumber];

	return 0;
}
