/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef PEAKFINDER_H
#define PEAKFINDER_H

// This object finds gaussian peaks on 1D data.
// It first calls the rough peak finder to find out where peaks are.
// Then it applies a  fit on the data window around the peak.
// The peak finder has the following features
// - Use of the standard error
// - Get the highest peak

#include <iostream>
#include <vector>
#include <fstream>
#include <blitz/array.h>
#include <math.h>
#include <string>

#include "RoughPeakFinder.h"
#include "Peak.h"
#include "MathematicaForm.h"

// some defines used for calling the function
// standard errors
static const bool MULTIGAUSS_USE_STDERR = true;
static const bool MULTIGAUSS_NO_STDERR = false;

/*! \brief A multi gauss peak finder for 1D data sets. */

/** This class finds peaks in one dimensional spectra and returns the fit of each peak.
 *  It performs as follow:
 * -# The peaks are localized using the rough peak finder (see \a roughPeakFinder.h ) . <b> This works only for positive peaks !</b>
 * -# It computes first estimates for the gaussian fit from the values returned by the rough peak finder, meaning:
 *	- The fitting window is twice the width of the rough peak.
 *	- The RMS is half the width of the rough peak.
 *	- The mean value is the position of the peak as returned by the rough peak finder.
 *	- The amplitude of the gaussian is the maximum of y in the fit window.
 *	- The pedestal has no first estimates (set to 0)
 * -# Prior to fit, a check is done in order to avoid overlapping windows. This
 *	 is usefull for peaks that are very close to each other. Presently, when two windows are overlappping
 *    The new window bound is taken as the mean of the position of the ovelapping peaks. This works
 *	 well only if the peaks are of equivalent width. In future version of the peak finder, these window
 *    bounds will take into account the peak widths.
 *	Once the fitting windows have been re-arranged, the fits are applied.
 * -# The gaussian fits are applied with 4 parameters (see  GaussFit), meaning the amplitude, the mean, the RMS and the pedestal.
 *   When the fit has not correctly converged, the peak is not returned. This is quite rare. However, the caller should perform
 *   further checks on the parameters as the fit might converge toward irrealistic values. A simple check is to test that the
 *   gaussian amplitude and pedestal are positive for positive peaks....
 *
 *  The peak finder returns Peak objects. These objects - that should never be created by outside of this class - are simply containers
 *  for the peak information, providing interpolated values useful for data display.
 *  Peaks are accessible  through the \a peak() method. The order of storage is the position in increasing
 *  order. The caller can use as well the \a bestPeak() method that returns the peak with the highest amplitude, useful for instance
 *  to find out the kinematical peak in a ToF spectrum.
 */

class MultiGaussPeakFinder {
public:
	/**
	 * Default constructor.
	 */
	MultiGaussPeakFinder();

	/** This constructor is the most general one.
	 * It finds the peaks of the data set placed in y (abscissa are in x).
	 * @param y the values of the curve to fit
	 * @param x the abscissa of the curve to fit. This array should have the same dimension than y.
	 * @param useStandardErrors a flag telling the gaussian fitter to use standard errors sqrt(N) for N neutrons in a bin. This flag takes the value
	 *  - MULTIGAUSS_USE_STDERR for using standard errors.
	 *  - MULTIGAUSS_NO_STDERR for equal errors over the whole data set.
	 * The values passed in x and y must not be obligatorily in increasing order of abscissa, or have a constant sampling in x. You may for instance
	 * give a logarithmic sampling as for energies.
	 */
	MultiGaussPeakFinder(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x) // general call
			;

	/** Same constructor as above, but without abscissa values. The values are computed from the bounds of the
	 * y vector. Note that the blitz library manages very well bounds in its data structure so that if you give
	 * a subset of an array by y(Range(3,45)) to the peak finder, the x abscissa used for the fits will take
	 * them into account, meaning that x will contains values going from 3 to 45. Positions will be returned as well
	 * in this referential.
	 */

	MultiGaussPeakFinder(const blitz::Array<double, 1>& y) // without x (taken as indices of y)
			;

	/**
	 * Destructor. Deletes all the Peaks.
	 */
	~MultiGaussPeakFinder();

	/**
	 * Calculates the peaks.
	 */
	void calculatePeaks(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x) ;

	/**
	 * Calculates the peaks.
	 */
	void calculatePeaks(const blitz::Array<double, 1>& y) ;

	// accessing the results
	/** returns the number of all peaks that have been found */
	int numberOfPeaks() const; // all peaks that have been found
	/** returns the number of all godd peaks (meaning with converging fits)  that have been found */
	int numberOfGoodPeaks() const; // all peaks that have been correctly fitted
	/** returns the number of all bad peaks (meaning with non converging fits)  that have been found */
	int numberOfBadPeaks() const; // all peaks that have not been correctly fitted
	/** returns the peak at order \a peakNumber, bad and good peaks  */
	Peak *peak(int peakNumber); // returns a vector of peaks by their position;
	/** returns the best peak (maximum or minimum) */
	Peak *bestPeak(); // get the highest peak (absolute value, can be a negative peak)
	blitz::Array<double, 2> getParameters() const;

	/** The peak finder has a text display function intended maily for debugging.
	 * This example has been for instance produced with D1B data, finding diffraction
	 * peaks in D2B data <BR>
	 * <img src="images/d2bdata.jpg">
	 * <CODE>
	 * --- 16 peaks found ----------------------<BR>
	 * #   position  amplitude (pedestal+amplitude )    rms      chi square<BR>
	 *  1.    590.95  22.7299(  7.67729 +   15.0526 )          5.53344   7.8727 <BR>
	 *  2.    685.13  24.7135(  7.51945 +    17.194 )          5.29933  7.19038<BR>
	 *  3.   984.627  60.4958(  7.67869 +   52.8171 )           5.1676  118.878<BR>
	 *  4.   1169.88  23.8048(  7.01359 +   16.7912 )          4.20623  2.58162<BR>
	 *  5.    1227.4  18.9413(   6.9888 +   11.9525 )          3.88139   1.7371<BR>
	 *  6.   1444.43  33.6958(  7.17931 +   26.5165 )          4.02855    13.19<BR>
	 *  7.   1598.51   20.229(  7.13555 +   13.0935 )          3.55262  3.62723<BR>
	 *  8.   1648.79  25.9217(  7.19354 +   18.7281 )          3.64382  4.85919<BR>
	 *  9.   1848.03  58.5691(  7.56179 +   51.0073 )          3.94905  128.961<BR>
	 * 10.   1998.01  21.1386(  7.14208 +   13.9965 )          3.32352  2.97247<BR>
	 * 11.   2258.07  37.9496(  7.43505 +   30.5146 )          3.78424  41.5764<BR>
	 * 12.    2425.8  23.9324(  7.31558 +   16.6168 )          3.76653  7.12853<BR>
	 * 13.   2484.78  23.8851(  7.16897 +   16.7161 )          3.90755  4.33779<BR>
	 * 14.   2748.38  46.8612(  7.28281 +   39.5784 )           5.8117  39.9144<BR>
	 * 15.    3000.8  14.5307(  7.15754 +   7.37319 )          7.71526  3.15858<BR>
	 * 16.   3109.83  16.9955(  7.17782 +   9.81768 )          10.3154  6.76316<BR>
	 * --------------------------------------<BR>
	 *</CODE>
	 * The same output is now applied to look for a kinematical peak of IN4 <BR>
	 * <img src="images/in4data.jpg">
	 * <CODE>
	 *--- 1 peaks found ----------------------<BR>
	 *#   position  amplitude (pedestal+amplitude )    rms      chi square<BR>
	 *1.   150.792  92.7629( 0.657799 +   92.1051 )    2.45765  38.0078<BR>
	 *--------------------------------------<BR>
	 *</CODE>
	 *
	 */
	friend std::ostream& operator<<(std::ostream &os, const MultiGaussPeakFinder& fit);

	/**
	 * Returns the same content than for operator<<
	 */
	std::string string(bool allPeaks = false) const;

private:
	int m_nbPeaks; // number of peaks
	int m_nbGoodPeaks; // number of good peaks
	int m_nbBadPeaks; // number of bad peaks
	std::vector<Peak *> peaks; // allocate a vector of peak pointers (vectors because they will be reordered)
	// the peak finder by itself called by the constructors
	void findPeaks(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x) ;

	blitz::Array<double, 1> m_a;
	blitz::Array<double, 2> m_aFinal;

};

inline int MultiGaussPeakFinder::numberOfPeaks() const {
	return m_nbPeaks;
}

inline int MultiGaussPeakFinder::numberOfGoodPeaks() const {
	return m_nbGoodPeaks;
}

inline int MultiGaussPeakFinder::numberOfBadPeaks() const {
	return m_nbBadPeaks;
}

inline blitz::Array<double, 2> MultiGaussPeakFinder::getParameters() const {
	return m_aFinal;
}

#endif // PEAKFINDER_H
