/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef PEAK_H
#define PEAK_H

#include <iostream>
#include <blitz/array.h>   
#include <math.h>

// it should never be created by the user, but only by the peakFinder object

/// a simple gaussian peak
static const int PEAK_GAUSSIAN = 1; //???
static const int PEAK_SIMPLE_GAUSSIAN = 1;
static const int PEAK_GAUSSIAN_PEDESTAL = 2;
static const int PEAK_GAUSSIAN_STRAIGTHLINE = 3;
static const int PEAK_GAUSSIAN_PARABOLA = 4;
static const int PEAK_PARABOLA = 5;
static const int PEAK_LORENTZIAN = 6; // not yet implemented

/*! \brief Peak information container */ 
/// This is an object that contains peak results.
/// it provides as well some functions to get the points of the peak function(not yet implemented)
/// it should never be created by the user, but only by the MultiGaussPeakFinder object.

class Peak 
{
  friend class MultiGaussPeakFinder;
  
public:
  // constructor of peaks, never call it yourself !
  Peak(int minIndex, int maxIndex, int peakType, double a0, double a1, double a2, double a3, double a4, double a5, double a6, double chiSquare);
  ~Peak();
  /// Peaks come with a text output function
  /// producing outputs like
  /// <br><code>150.792  92.7629( 0.657799 +   92.1051 )    2.45765  38.0078</code><br>
  /// with positions, full amplitude (pedestal + amplitude), rms & chi squares. 
  friend std::ostream& operator<<(std::ostream &os, const Peak& p); // display

  // accessing the parameters
  /// This function indicates if the fit has converged correctly
  bool correctlyFitted() const;
  /// This function returns the type of the fit (look at definitions for the values)
  /// Types are :
  /// - PEAK_GAUSSIAN (simple gaussian, amplitude, mean & rms)
  /// - PEAK_SIMPLE_GAUSSIAN (same as above)
  /// - PEAK_GAUSSIAN_PEDESTAL (simple gaussian, amplitude, mean & rms, plus a constant pedestal)
  /// - PEAK_GAUSSIAN_STRAIGTHLINE (simple gaussian, amplitude, mean & rms, plus a linear pedestal)
  /// - PEAK_GAUSSIAN_PARABOLA (simple gaussian, amplitude, mean & rms, plus a quadratic pedestal)
  /// - PEAK_PARABOLA  (a parabola, 2nd order, 3 parameters)
  /// - PEAK_LORENTZIAN (not yet implemented)
  int peakType() const;
  /// This returns the amplitude of the fitted function, <b> that without the pedestal value</b>.
  double amplitude() const; 
  /// Returns the full amplitude of the peak, meaning with the pedestal value. (amplitude+pedestal)
  double fullAmplitude() const; 
  /// Returns the pedestal of the peak 
  double pedestal() const; // the full amplitude means including the pedestal     
  /// returns the position of the peak, meaning the mean for gaussian peaks
  double position() const;
  /// This returns the RMS of gaussian peaks, if applicable
  double rms() const; 
  /// returns the lower bound of the fitting window
  int minimumWindowIndex() const; 
  /// returns the upper bound of the fitting window
  int maximumWindowIndex() const; 
  /// returns the polynomial factors for peaks that have been fitted with a parabolic component (background).
  /// a call to <code> polynomialFactor(0) </code> is equivalent to <code> pedestal()</code>.
  double polynomialFactor(int order) const; // if applicable
  // Returns the chi square of the fit
  double chiSquare() const; // from the fit
  // Returns the quality
  double quality() const;
  // Returns true if is negative
  bool isNegative() const;

private:
  int m_minIndex;
  int m_maxIndex;
  int m_peakType; 
  double m_chiSquare; // set to -1 if the fit was not correct
  double m_a0,m_a1,m_a2,m_a3,m_a4,m_a5,m_a6; // if they need a peak finder with more than 7 parameters ... call me 7048 ... i'ill talk with them :-)
  bool m_fitHasConverged;
  double m_quality;
  double m_dispersionRatio;
};

inline bool Peak::correctlyFitted() const
{
  return m_fitHasConverged;
}

inline int Peak::peakType() const
{
  return m_peakType;
}

inline double Peak::amplitude() const 
{
  return m_a0;
}

inline double Peak::fullAmplitude() const
{
  return m_a0 + m_a3;
}

inline double Peak::pedestal() const
{
  return m_a3;
}	

inline double Peak::position() const
{
  return m_a1;
}

inline double Peak::rms() const
{
  return m_a2;
}

inline int Peak::minimumWindowIndex() const
{
  return m_minIndex;
}

inline int Peak::maximumWindowIndex() const
{
  return m_maxIndex;
}

inline double Peak::polynomialFactor(int order) const
{
  switch (order)
  {
    case 0:
      return m_a3;
    case 1:
      return m_a4;
    case 2:
      return m_a5;
    default:
      break;
  }
  return 0;
}

inline double Peak::chiSquare() const
{
  return m_chiSquare;
}

inline double Peak::quality() const {
	return m_quality;
}

inline bool Peak::isNegative() const {
	return (m_a0 < 0.0);
}

#endif // PEAK_H
