/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Polynomial.h"
#include "Threshold.h"
#include "Inverse.h"
#include <iostream>

using namespace std;

#include <string>

extern "C" {
void dgeev_(char*, char*, int*, double*, int*, double*, double*, double*, int*,
		double*, int*, double*, int*, int*);
}

using namespace std;
using namespace blitz;

pair<double, double> const Polynomial::NullPoint = pair<double, double> (numeric_limits<double>::min(), numeric_limits<double>::min());

Polynomial::Polynomial() {
}

Polynomial::Polynomial(blitz::Array<double, 1> const & coef)
		 :
	m_coef(coef.size()) {
	// Precondition.
	if (coef.size() == 0)
		throw invalid_argument("Polynomial::Polynomial: coef cannot be empty!");

	// resizing to be sure that the coefficient of the highest order is not null.
	bool begin = true;
	int j = 0;

	for (int i = coef.size() - 1; i != -1; --i) {
		if ((coef(i) != 0) || (!begin)) {
			m_coef(j) = coef(i);
			++j;
		}

		if ((coef(i) != 0) && (begin))
			begin = false;
	}

	m_coef.resizeAndPreserve(j);
}

Polynomial Polynomial::fit(int order, blitz::Array<double, 1> const & y, blitz::Array<double, 1> const & x)  {

	firstIndex i;
	secondIndex j;

	// Defining the Vandermonde matrix for x
	Array<double, 2> X(x.size(), order + 1);
	X = pow(x(i), j);

	// transpose X
	Array<double, 2> Xt(order + 1, x.size());
	Xt = X(j, i);

	// compute Xt x X
	Array<double, 2> XtX = product(Xt, X);

	// invert XtX
	Array<double, 2> XtXi = inverse(XtX);

	// XtXi x Xt
	Array<double, 2> XtXiXt = product(XtXi, Xt);

	// final product
	Array<double, 1> a = product(XtXiXt, y);

	return Polynomial(a);
}

bool Polynomial::isNull() const {
	return (m_coef.size() == 0);
}

double Polynomial::coef(int i) const  {
	if ((i < 0) || (i > m_coef.size())) {
		std::stringstream error;
		error << "Polynomial::coef: cannot access " << i << "th element!";

		throw invalid_argument(error.str());
	}

	return m_coef(m_coef.size() - 1 - i);
}

double Polynomial::eval(double x) const {
	// init
	double res = 0.0;

	for (blitz::Array<double, 1>::const_iterator a = m_coef.begin(); a
			!= m_coef.end(); ++a)
		res = x * res + *a;

	return res;
}

Polynomial Polynomial::derive() const {

	Array<double, 1> coef(m_coef.size() - 1);

	int k = 1;

	for (Array<double, 1>::iterator a = coef.begin(); a != coef.end(); ++a) {
		*a = k * m_coef(m_coef.size() - k - 1);
		++k;
	}

	return Polynomial(coef);
}

blitz::Array<std::complex<double>, 1> Polynomial::roots() const {

	// defining the companion matrix for which we find the eigenvalues.
	int N = m_coef.size() - 1;

	double* M = new double[N * N];

	for (int j = 0; j < N - 1; ++j) {
		M[N * j + 0] = 0.0;

		for (int i = 1; i < N; ++i) {
			M[N * j + i] = (i == j + 1 ? 1.0 : 0.0);
		}
	}

	if (m_coef(0) != 0.0) {
		for (int i = 0; i < N; ++i) {
			M[N * (N - 1) + i] = -m_coef(N - i) / m_coef(0);
		}
	} else {
		for (int i = 0; i < N; ++i) {
			M[N * (N - 1) + i] = -m_coef(N - i);
		}
	}

	// lapack call to find the eigenvalues.
	char c = 'N';
	int S = N;
	int LDVL = 1;
	int LDVR = 1;
	int LWORK = 3 * N;

	double* WR = new double[N];
	double* WI = new double[N];
	double* VL = new double[LDVL * N];
	double* VR = new double[LDVR * N];
	double* WORK = new double[LWORK];
	int info;

	dgeev_(&c, &c, &S, M, &S, WR, WI, VL, &LDVL, VR, &LDVR, WORK, &LWORK, &info);

	delete[] VL;
	delete[] VR;
	delete[] WORK;
	delete[] M;

	Array<complex<double> , 1> R(N);

	for (int i = 0; i < N; ++i) {
		R(i) = complex<double> (WR[i], WI[i]);
	}

	delete[] WR;
	delete[] WI;

	return R;
}

blitz::Array<double, 1> Polynomial::realRoots() const {
	Array<complex<double> , 1> R = roots();
	Array<double, 1> RR(R.size());

	// Getting real roots.
	int n = 0;

	for (int i = 0; i < R.size(); ++i) {
		if (isZero(R(i).imag())) {
			RR(n) = R(i).real();
			++n;
		}
	}

	if (n > 0) {
		RR.resizeAndPreserve(n);
	} else {
		return Array<double, 1>();
	}

	return RR;
}

std::pair<double, double> Polynomial::minimum() const {
	Array<pair<double, double> , 1> E = extrema();

	double min = numeric_limits<double>::max();
	pair<double, double> c = NullPoint;

	for (Array<pair<double, double> , 1>::const_iterator i = E.begin(); i
			!= E.end(); ++i) {
		if (i->second < min) {
			c = *i;
			min = i->second;
		}
	}

	return c;
}

std::pair<double, double> Polynomial::maximum() const {
	Array<pair<double, double> , 1> E = extrema();

	double max = numeric_limits<double>::min();
	pair<double, double> c = NullPoint;

	for (Array<pair<double, double> , 1>::const_iterator i = E.begin(); i
			!= E.end(); ++i) {
		if (i->second > max) {
			c = *i;
			max = i->second;
		}
	}

	return c;
}

blitz::Array<std::pair<double, double>, 1> Polynomial::extrema() const {

	// Getting the real roots of the derivate polynomial.
	Array<double, 1> R = derive().realRoots();

	// filling the result.
	Array<pair<double, double> , 1> E(R.size());

	// Calculating values.
	for (int i = 0; i < R.size(); ++i)
		E(i) = pair<double, double> (R(i), eval(R(i)));

	return E;
}

std::string Polynomial::string() const {

	ostringstream os;

	for (int i = 0; i < order(); ++i) {
		os << coef(i) << "*x**" << i << " + ";
	}

	if (order() >= 0) {
		os << coef(order()) << "*x**" << order();
	}

	return os.str();
}

std::ostream& operator<<(std::ostream &o, Polynomial const & P) {

	o << P.string();

	return o;
}

blitz::Array<double, 1> Polynomial::coefs() const {
	Array<double, 1> r(m_coef.size());

	for (int i = 0; i < m_coef.size(); ++i)
		r(i) = m_coef(m_coef.size() - i - 1);

	return r;
}
