/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef POLYNOMIAL_H
#define POLYNOMIAL_H

#include <stdexcept>
#include <utility>
#include <complex>
#include <blitz/array.h>
#include <iostream>

/**
 * Class providing an useful interface for polynomials. The main goal is to use
 * least squares fitting. The calculus of the roots is made by finding the 
 * eigenvalues of the companion matrix of the polynomial. For double roots, the result
 * is not precise, nevertheless it doesn't matter since in a maximum calculus, they 
 * cannot be used.
 */
class Polynomial {

public:
	/**
	 * Constructor.
	 */
	Polynomial();
	/**
	 * Constructor with a list of coefficients.
	 * Builds coef[0] + coef[1]*X + ... + coef[N]*X^N where N is the maximum
	 * index for which coef[N] is not null.
	 * Precondition: coef is not empty.
	 */
	Polynomial(blitz::Array<double, 1> const & coef) ;

	/**
	 * builds the polynom of degree order fitting the data set using the least squares method.
	 * Precondition: x size equals y size.
	 */
	static Polynomial fit(int order, blitz::Array<double, 1> const & y, blitz::Array<double, 1> const & x) ;

	/**
	 * Return the order.
	 */
	int order() const;

	/**
	 * Return the list of coefficients ordered like this: coef[i] is the coefficient
	 * of the monomial of order i.
	 */
	blitz::Array<double, 1> coefs() const;

	/**
	 * Returns true if is null.
	 */
	bool isNull() const;

	/**
	 * Return the coefficient of the monomial of order i.
	 * Precondition: 0 <= i <= order.
	 */
	double coef(int i) const ;

	/**
	 * Evaluates the polynomial function at x value.
	 */
	double eval(double x) const;

	/**
	 * Return the derivate polynomial.
	 */
	Polynomial derive() const;

	/**
	 * Calculate and return all roots including complex values.
	 */
	blitz::Array<std::complex<double>, 1> roots() const;

	/**
	 * Calculate and return the real roots.
	 */
	blitz::Array<double, 1> realRoots() const;

	/**
	 * Return the (x, y) couple of the minimum of the local extrema.
	 * If there is no minimum, then return NullPoint.
	 */
	std::pair<double, double> minimum() const;

	/**
	 * Return the (x, y) couple of the maximum of the local extrema.
	 * If there is no maximum, then return NullPoint.
	 */
	std::pair<double, double> maximum() const;

	/**
	 * Return the list of (x, y) couples of the local extrema.
	 */
	blitz::Array<std::pair<double, double>, 1> extrema() const;

	/**
	 * Returns the string of the polynomial.
	 */
	std::string string() const;

	/**
	 * NullPoint, returned by minimum() or maximum() functions.
	 */
	static std::pair<double, double> const NullPoint;

private:
	// By convenience, the coefficients are ordered from the highest order to the order 0.
	blitz::Array<double, 1> m_coef;
};

std::ostream& operator<<(std::ostream &o, Polynomial const & P);

inline int Polynomial::order() const {
	return (int) (m_coef.size() - 1);
}

#endif
