/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*
 *  quadraticInterpolator.cpp
 *  caseMath
 *
 *  Created by Fred on 21/10/04.
 *  Copyright 2004 __MyCompanyName__. All rights reserved.
 *
 */

#include "PolynomialInterpolator.h"

using namespace blitz;

ostream& operator<<(ostream& os, const PolynomialInterpolator& q) {
	
		 os <<"y="<< q.a2()<<"x^2+"<<q.a1()<<"x+"<<q.a0()<<std::endl;
		 os <<"chiSquare="<<q.chiSquare()<<std::endl;
		 if(q.isAMaximum())
		 os <<"maximum of y="<<q.optimum()<<" at x="<<q.optimumCoordinates()<<std::endl;
		 else
		 os <<"minimum of y="<<q.optimum()<<" at x="<<q.optimumCoordinates()<<std::endl;
		 if(q.optimumIsInRange()) os<<"optimum is in your coordinates range"<<std::endl;
		 else os<<"optimum is outside of your coordinates range"<<std::endl;

  return os;
}


///////////////////////////////////////////////////////////////////////
// constructors

	PolynomialInterpolator::~PolynomialInterpolator(){}
	
	PolynomialInterpolator::PolynomialInterpolator(const Array<double,1> &y,const Array<double,1> &x,unsigned order){
	m_minIndex=y.lbound(firstDim);
	m_maxIndex=y.ubound(firstDim);
	interpolate(y,x,order,false);
	}
	
	PolynomialInterpolator::PolynomialInterpolator(const Array<double,1> &y,unsigned order){
	// create an x of the size of y
	unsigned n=y.size();
	Array<double,1> x(n);
	firstIndex i; 
	x=i;
	m_minIndex=y.lbound(firstDim);
	m_maxIndex=y.ubound(firstDim);
	interpolate(y,x,order,false);	
	}
	
	PolynomialInterpolator::PolynomialInterpolator(const Array<double,1> &y,unsigned minIndex,unsigned maxIndex,unsigned order){
	// create an x of the size of y
	Array<double,1> x(maxIndex-minIndex+1);	// be carefull, min index and max index are included
	firstIndex i; 
	x=i+(double) minIndex;		
	m_minIndex=minIndex;
	m_maxIndex=maxIndex;	
	interpolate(y(Range(minIndex,maxIndex)),x,order,false);	
	}

	/////////////////////////////////////////////////
	// same constructors with the standard error flag
	/////////////////////////////////////////////////
	
	PolynomialInterpolator::PolynomialInterpolator(const Array<double,1> &y,const Array<double,1> &x,unsigned order,bool useStandardError){
	m_minIndex=y.lbound(firstDim);
	m_maxIndex=y.ubound(firstDim);
	interpolate(y,x,order,useStandardError);
	}
	
	PolynomialInterpolator::PolynomialInterpolator(const Array<double,1> &y,unsigned order,bool useStandardError){
	// create an x of the size of y
	unsigned n=y.size();
	Array<double,1> x(n);
	firstIndex i; 
	x=i;
	m_minIndex=y.lbound(firstDim);
	m_maxIndex=y.ubound(firstDim);
	interpolate(y,x,order,useStandardError);	
	}
	
	PolynomialInterpolator::PolynomialInterpolator(const Array<double,1> &y,unsigned minIndex,unsigned maxIndex,unsigned order,bool useStandardError){
	// create an x of the size of y
	Array<double,1> x(maxIndex-minIndex+1);	// be carefull, min index and max index are included
	firstIndex i; 
	x=i+(double) minIndex;		
	m_minIndex=minIndex;
	m_maxIndex=maxIndex;	
	interpolate(y(Range(minIndex,maxIndex)),x,order,useStandardError);	
	}


///////////////////////////////////////////////////////////////////////
// the main interpolation here
///////////////////////////////////////////////////////////////////////

bool PolynomialInterpolator::interpolate(const Array<double,1> &y,const Array<double,1> &x,unsigned order,bool useStandardError)
throw (BadInput) {



// vectors (arrays) should have the same length, if not exception
if(y.size() != x.size()) throw BadInput();
// testing the order (limited to 2, because of matrix inversion)
if(order >2) throw BadInput();
m_chiSquare=0.0; // a reset of the chisquare
int n=y.size();


// build up the vandermonde matrix X
Array<double,2> X(order+1,n);

firstIndex i;
secondIndex j;
thirdIndex k;
//X = pow(x(j),i);
if(useStandardError)
X =  pow(x(j),i)/y(j);
else X = pow(x(j),i); // standard error is sqrt(y)


// A = (Transp(X).X)^-1.Transp(X).y

// XTX = X *transpose(X)
Array<double,2> XTX(sum(X(j,k) * X(i,k), k));

// compute the parameters
//(Transpose(X) * y  );
//Array<double,1> TXy(sum(X(i,k) * y(k), k));
Array<double,1> TXy(order+1);
if(useStandardError)
	TXy=(sum(X(i,k) * y(k)/y(k), k));
else
	TXy=(sum(X(i,k) * y(k), k));
// Inverse(XTX) * TXy = A = the interpolation result
XTX=inverse(XTX);
Array<double,1> A(sum(XTX(i,k) * TXy(k), k));

/// filling the results
if(order==2) {m_a0=A(0); m_a1=A(1); m_a2=A(2);}
else 
if(order==1) {m_a0=A(0); m_a1=A(1); m_a2=0;}
else 
if(order==0) {m_a0=A(0); m_a1=0; m_a2=0;}

// computing the chi square and optimum stuff
Array<double,1>  yInterpol(m_a0+m_a1*x+m_a2*x*x);
m_chiSquare = sum(pow2(y(i)-yInterpol(i)) );
double optimumCoordinate=optimumCoordinates();
m_optimumIsInRange=(optimumCoordinate >= x(0) &&  optimumCoordinate <= x(n-1));
  
// JUST FOR TESTING
//std::cout<<"y=";mathematicaForm(y);
//std::cout<<"yInterpol=";mathematicaForm(yInterpol);
//std::cout<<"A="<<A<<std::endl;
return true;
}

///////////////////////////////////////////////////////////////////////
// some functions to access to the interpolator values	
///////////////////////////////////////////////////////////////////////
bool PolynomialInterpolator::optimumIsInRange() const {return m_optimumIsInRange;}
double PolynomialInterpolator::optimumCoordinates() const {return -m_a1/(2*m_a2);} // simple parabola mathematics
double PolynomialInterpolator::optimum() const {return m_a0 - m_a1*m_a1/(4*m_a2);} // simple parabola mathematics



// the interpolated curve is a 2D array containing first the abscisse, second the interpolated values

// this function returns the interpolated curve between indexes.
Array<double,2> PolynomialInterpolator::getInterpolatedCurve(int minIndex,int maxIndex) {
return getInterpolatedCurve(minIndex,maxIndex,maxIndex-minIndex+1);
}

// returns the innterpolated curve on the full length of y
Array<double,2> PolynomialInterpolator::getInterpolatedCurve() {
return getInterpolatedCurve(m_minIndex,m_maxIndex,m_maxIndex-m_minIndex+1);
}

// returns the innterpolated curve on the full length of y
Array<double,2> PolynomialInterpolator::getInterpolatedCurve(int nbSamples) {
return getInterpolatedCurve(m_minIndex,m_maxIndex,nbSamples);
}


Array<double,2> PolynomialInterpolator::getInterpolatedCurve(int minIndex,int maxIndex,int nbSamples)
throw (BadInput) {
// this function computes the interpolation of the function between the two given indexed and the 
// wanted number of samples. This is thought for a smooth display.
// make the abscices 

// the number of sample should be greater that 1

if(nbSamples < 2 ) throw BadInput();

Array<double,2> interpolation(2,nbSamples);
firstIndex i;
interpolation(0,Range::all())=i;
interpolation(0,Range::all()) =interpolation(0,Range::all())/(nbSamples-1)*(maxIndex-minIndex)+minIndex;
// make the ordinates
interpolation(1,Range::all()) =(m_a0+m_a1*interpolation(0,Range::all())+m_a2*interpolation(0,Range::all())*interpolation(0,Range::all()));

return interpolation;
}





long PolynomialInterpolator::status() const {return 0;}