/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef POLYNOMIALINTERPOLATOR_H
#define POLYNOMIALINTERPOLATOR_H


#include <iostream>
#include <blitz/array.h>   
#include <math.h>
#include "Inverse.h"
#include "Determinant.h"

// give this parameter to tell the routine it should use the standard error
// thiw should be used when the interpolation is applied to neutron counts.
#define POLYNOMIAL_INTERPOLATOR_USE_STDERR true 

///// DOCUMENTATION, please look at http://mathworld.wolfram.com/LeastSquaresFittingPolynomial.html
///// for the implementation of polynomial fits

/*! \brief A polynomial interpolator that can be used for finding optimum in rocking curves. */ 

/**
This object performs a least-square polynomial approximation of a data set.
Futher documentation can be found here.
http://mathworld.wolfram.com/LeastSquaresFittingPolynomial.html
<br>
Presently, the interpolator is limite to order 2. It will be developped for higher orders.
*/

class PolynomialInterpolator  {
public:   

	 class BadInput {}; // used to throw errors

	// generic constructor, should never be used
	 ~PolynomialInterpolator();
	
	/// Call the interpolator on x and y with an order 0,1 or 2
	PolynomialInterpolator(const blitz::Array<double,1> &y,const blitz::Array<double,1> &x,unsigned order);
	
	/// Call the interpolator for y values, using the index of the y array as abscissa
	PolynomialInterpolator(const blitz::Array<double,1> &y,unsigned order);
	
	///Interpolates y, building the abscisse with values from minIndex to maxIndex (included) This is a partial interpolation
	PolynomialInterpolator(const blitz::Array<double,1> &y,unsigned minIndex,unsigned maxIndex,unsigned order);
	
	
	/// same constructors as above, with the standard error flag POLYNOMIAL_INTERPOLATOR_USE_STDERR
	PolynomialInterpolator(const blitz::Array<double,1> &y,const blitz::Array<double,1> &x,unsigned order, bool useStandardError);
	/// same constructors as above, with the standard error flag POLYNOMIAL_INTERPOLATOR_USE_STDERR
	PolynomialInterpolator(const blitz::Array<double,1> &y,unsigned order, bool useStandardError);
	/// same constructors as above, with the standard error flag POLYNOMIAL_INTERPOLATOR_USE_STDERR
	PolynomialInterpolator(const blitz::Array<double,1> &y,unsigned minIndex,unsigned maxIndex,unsigned order, bool useStandardError);
	

	// some functions to access to the interpolator values	
	bool optimumIsInRange() const; /// tells if the optimum coordinates is in the range (valid only for order 2 polynomes)
	double optimumCoordinates() const; /// the asbissa of the optimum  (valid only for order 2 polynomes)
	double optimum() const; /// the value of the optimum of the interpolation  (valid only for order 2 polynomes)
	bool isAMaximum() const; /// tell if the optimum is a maximum or a minimum
	blitz::Array<double,2>  getInterpolatedCurve(int minIndex,int maxIndex,int nbSamples)  /// get the interpolated (for data display) with specified number of samples
	throw (BadInput);
	blitz::Array<double,2> getInterpolatedCurve(int nbSamples); /// same as above, but on the full curve
	blitz::Array<double,2> getInterpolatedCurve(int minIndex,int maxIndex); /// same but without the number of samples
	blitz::Array<double,2> getInterpolatedCurve(); /// taking the min and max index that have been given for the interpolation
	long status() const; // the status of the interpolator (unused yed)
	double chiSquare() const; 
	double a0() const; // parameter a0 of the polynome a0 + a1 x + a2 x^2 
	double a1() const; // parameter a0 of the polynome a0 + a1 x + a2 x^2 
	double a2() const; // parameter a0 of the polynome a0 + a1 x + a2 x^2 
	
	/// a display 
	friend ostream& operator<<(ostream &os, const PolynomialInterpolator& q);
	

private:
    bool interpolate(const blitz::Array<double,1> &y,const blitz::Array<double,1> &x,unsigned order,bool useStandardError)
	throw (BadInput);	// the interpolation routine called by the constructors

	// attributes
	double m_a0,m_a1,m_a2; //the result of the interpolation
	double m_chiSquare; // the chiSquare-like estimator
	bool m_optimumIsInRange;
	unsigned int m_minIndex,m_maxIndex; // used to remember what the range was
};


inline double PolynomialInterpolator::chiSquare() const{
return m_chiSquare;
}

inline	double PolynomialInterpolator::a0() const{ return m_a0;}
inline	double PolynomialInterpolator::a1() const{ return m_a1;}
inline	double PolynomialInterpolator::a2() const{ return m_a2;}
inline	bool PolynomialInterpolator::isAMaximum() const{
return (m_a2<0);
}

#endif // POLYNOMIALINTERPOLATOR_H