/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "PolynomialPeakFinder.h"
#include <cminpack-1/cminpack.h>

using namespace blitz;

PolynomialPeakFinder::PolynomialPeakFinder(int order, const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x)  :
	_y(y), _x(x), _fittedPolynomial(0), _quality(0.0) {

	if (order < 2) {
		throw domain_error("PolynomialPeakFinder : cannot fit polynomial with order < 2");
	}

	if (x.size() > order + 1) {
		_fittedPolynomial = new Polynomial(Polynomial::fit(order, y, x));
		calculateQuality();
	} else {
		throw domain_error("PolynomialPeakFinder : not enough points");
	}
}

PolynomialPeakFinder::PolynomialPeakFinder(int order, const blitz::Array<double, 1>& y)  :
	_y(y), _fittedPolynomial(0), _quality(0.0) {

	if (order < 2) {
		throw domain_error("PolynomialPeakFinder : cannot fit polynomial with order < 2");
	}

	int minIndex = y.lbound(firstDim);

	_x.resize(y.size());
	firstIndex i;

	_x = i + minIndex;

	if (_x.size() > order + 1) {
		_fittedPolynomial = new Polynomial(Polynomial::fit(order, y, _x));
		calculateQuality();
	} else {
		throw domain_error("PolynomialPeakFinder : not enough points");
	}
}

PolynomialPeakFinder::~PolynomialPeakFinder() {

	delete _fittedPolynomial;
}

double PolynomialPeakFinder::getMaximumPeakPosition() const  {

	if (_fittedPolynomial == 0 || _fittedPolynomial->isNull()) {
		throw domain_error("PolynomialPeakFinder : no maximum peak position");
	}

	// Using the second derivative to determine if it is a maximum or minimum
	Array<pair<double, double>, 1> E = _fittedPolynomial->extrema();

	Polynomial dP2 = _fittedPolynomial->derive().derive();

	bool maxFound = false;
	pair<double, double> maxPoint;

	bool minFound = false;
	pair<double, double> minPoint;

	for (int i = 0; i < E.size(); i++) {

		// checking the bounds and if it is a maximum (second derivative < 0) or minimum (second derivative > 0)
		if (E(i).first >= _x(_x.lbound(firstDim)) && E(i).first <= _x(_x.ubound(firstDim))) {

			if (dP2.eval(E(i).first) < 0.0) {

				if (!maxFound) {
					maxFound = true;
					maxPoint = E(i);
				} else if (E(i).second > maxPoint.second) {
					maxPoint = E(i);
				}
			} else if (dP2.eval(E(i).first) > 0.0) {

				if (!minFound) {
					minFound = true;
					minPoint = E(i);
				} else if (E(i).second < minPoint.second) {
					minPoint = E(i);
				}
			}
		}
	}

	if (maxFound) {
		return maxPoint.first;
	} else if (minFound) {
		return minPoint.first;
	}

	throw domain_error("PolynomialPeakFinder : no peak found");
}

Polynomial * PolynomialPeakFinder::getPolynomial() const {

	return _fittedPolynomial;
}

void PolynomialPeakFinder::calculateQuality() {

	long m = _y.size();
	double* data = new double[m];

	for (int i = 0; i < m; i++) {
		data[i] = _y(i) - _fittedPolynomial->eval(_x(i));
	}

	double norm = enorm(m, data);
	if (norm > 0.0) {
		_quality = m / norm;
	}

	delete [] data;
}

double PolynomialPeakFinder::getQuality() const {
	return _quality;
}
