/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef BZ_ARRAY_ROUGHPEAKFINDER_H
#define BZ_ARRAY_ROUGHPEAKFINDER_H

#include <stdexcept>
#include <blitz/array.h>

#ifndef BZ_ARRAY_H
 #error <blitz/array/myMath.h> must be included after <blitz/array.h>
#endif

static const int ROUGHPEAK_MAX_NB_PEAKS = 20;

// some indexes definitions for the peak finder
static const int ROUGHPEAK_POSITION = 0;
static const int ROUGHPEAK_AMPLITUDE = 1;
static const int ROUGHPEAK_WIDTH = 2;
static const int ROUGHPEAK_PEDESTAL = 3;

namespace blitz {

/*! \brief used for localizing peaks in a rough manner.*/

/// The rough peak finder localizes peaks in a 1D dataset. Its methods are not described as they should
/// not be called directly but rather used through objects like MultiGaussPeakFinder.
/// It is however important to understand how the peak finder makes a locallization.
/// For a given dataset, two values are derived
/// -# an estimate of the pedestal.
/// -# an estimate of the noise.
///
///

/*! \file roughPeakFinder.h
\brief used for localizing peaks in a rough manner.

The rough peak finder localizes peaks in a 1D dataset. Its methods are not described as they should
not be called directly but rather used through objects like MultiGaussPeakFinder.
It is however important to understand how the peak finder makes a locallization.
For a given dataset, two values are derived
-# an estimate of the pedestal.
-# an estimate of the noise.

<b> Pedestal estimation </b><br>
The pedestal is the maximum of the histogram data set. Let's takes an example of "dirty data", meaning data with
poor statistics. The next picture is a kinematic peak from IN4, just at the beginning of the run.<br>
<img src="images/in4bad.jpg"><br>
Running the rough peak finder on this data set will find a pedestal of 0.0545523. It is the maximum of the
histogram of the data. The histogram binning is taken as the tenth of the number of samples for good
statistics, with a lower limit to 10 and upper limit to 100.<br>
<b> Noise estimation </b><br>
The noise level is more critical to find and therefore two attempts are made. The first method scans the histogram
from its maximum in the positive direction. When a value has been found to be less than the half maximum, then the
width at the peak half maximum is computed with a simple linear rule. In the example, a value of  0.00719793 has been
computed.<br>
The second method is more robust and will generally give a higher noise level. It is estimated as the  horizontal rms of the
data derivative. The derivation cancels the pedestal of the data without altering the noise level.<br>

Finally the cut level will be the pedestal plus four time the noise leve (four sigma cut).<br>

Sometimes, high instant noise goes over the threshold (this is the case on this data set). In order to avoid
spurious peaks, the rough peak finder reject peaks with a width of one index, which whould anyway not be suitable
for anything.<br><br>

Once the cut level is computed, peaks are detected (simple scan) and peak width (windows) are taken as the bounds ov the
cross over. Here [148:158].<br><br>

If you have some problems with the peak detection, you might uncomment some printouts in the code<br>
<code>
std::cout<<"pedestal="<<pedestal<<std::endl;<br>
std::cout<<"noiseLevel="<<noiseLevel<<std::endl;<br>
std::cout<<"cut="<<cut<<std::endl;<br>
</code>

*/

template<typename T>
Array<double, 2> roughPeakFinder(const Array<T, 1>& data, const Array<T, 1>& x)
  ;

}

#include "RoughPeakFinder.inl"

#endif // BZ_ARRAY_ROUGHPEAKFINDER_H