/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <cmath>

#ifndef BZ_BLITZ_H
#include <blitz/blitz.h>
#endif

#include "Threshold.h"
#include "Moments.h"
#include "Histogram.h"
#include "GaussFit.h"

#include <blitz/array.h>
#include <blitz/array/convolve.h>
#include <algorithm>
#include <vector>
#include <cassert>

struct Entry {
	double x;
	double y;
};

inline bool compareEntry(const Entry & a, const Entry & b) {
	return a.x < b.x;
}

namespace blitz {

template<typename T>
Array<double, 2> roughPeakFinder(const Array<T, 1>& data, const Array<T, 1>& x)
		 {
	// rough strategy for finding peak
	// step 1 : get the pedestal as the maximum of the histogram of data
	// step 2 : get the noise estimate from the RMS of the derivative of the data
	// step 3 : make cuts at a tenth of the signal to noise ratio.
	// step 4 : get each range and find the maximum
	// step 5 : scan the data till it reaches the half maximum to get there, and get though the width
	// (we should have called it dirty peak finder)

	// Presently this works only for positive peaks (do we have any negative ones ? I do not see any)
	int n = data.size();

	// get the maximum of the histogram as the pedestal mean value
	int nbBins;
	nbBins = n / 10;

	if (nbBins < 10)
		nbBins = 10;

	if (nbBins > 100)
		nbBins = 100;

	// Rebin the data according to smallest increment
	double minDelta = 10.0E20;

	for (long n = 0; n < x.size() - 1; ++n) {
		double delta = ::fabs(x(n) - x(n + 1));
		if (delta < minDelta)
			minDelta = delta;
	}

	if (minDelta < 1E-7) {
		throw std::domain_error("roughPeakFinder: x data must be strictly monotonic");
	}

	int nBins = ::ceil(::fabs(x(x.size() - 1) - x(0)) / minDelta);

	// Check order of data
	if (x.size() > 1) {
		bool order = (x(0) < x(1));

		for (int i = 0; i < x.size() - 1; i++) {
			if (x(i) < x(i + 1) != order) {
				throw std::domain_error("roughPeakFinder: x data must be strictly monotonic");
			}
		}
	}

	// Order arrays
	Array<double, 1> ordered_x(x.size());
	Array<double, 1> ordered_data(x.size());

	std::vector<Entry> entries;
	for (long i = 0; i < x.size(); ++i) {
		Entry entry;
		entry.x = x(i);
		entry.y = data(i);

		entries.push_back(entry);
	}

	std::sort(entries.begin(), entries.end(), compareEntry);
	for (long i = 0; i < x.size(); ++i) {
		ordered_x(i) = entries[i].x;
		ordered_data(i) = entries[i].y;
	}

	Array<double, 1> new_x(nBins);
	Array<double, 1> new_y(nBins);

	firstIndex i;
	new_x(Range::all()) = ordered_x(0) + i * minDelta;

	int index = 0;
	for (long n = 0; n < nBins; ++n) {
		//	cout << "n " << n << "new_x.size() "  << new_x.size() << std::endl;
		//	cout << "index " << index << "xTmp.size() "  <<  xTmp.size() << std::endl;
		//	usleep(2000);
		//	  assert(n >= 0 && n < new_x.size());
		//	  assert(index >= 0 && index < xTmp.size());
		while (new_x(n) >= ordered_x(index)) {
			index++;
		}
		if (index > x.size() - 1)
			index = x.size() - 1;

		new_y(n) = ordered_data(index - 1) + (ordered_data(index) - ordered_data(index - 1))
				* (new_x(n) - ordered_x(index - 1))
				/ (ordered_x(index) - ordered_x(index - 1));
	}

	try {

		Histogram result = histogram(new_y, nbBins);
		Array<double, 2> histo = result.array;

		//mathematicaForm(histo);
		TinyVector<int, 1> tmp = maxIndex(histo(1, Range::all()));
		int maximumIndex = tmp(0);

		double pedestal;

		if (maximumIndex == 0) {
			if (isZero(histo(1, 0) + histo(1, 1)))
				throw std::domain_error("roughPeakFinder: division by zero in pedestal calculus");

			pedestal = (histo(0, 0) * histo(1, 0) + histo(0, 1) * histo(1, 1))
					/ (histo(1, 0) + histo(1, 1));
		} else {
			if (maximumIndex == nbBins - 1) {
				double denom = histo(1, nbBins - 2) + histo(1, nbBins - 1);
				if (isZero(denom))
					throw std::domain_error("roughPeakFinder: division by zero in pedestal calculus");

				pedestal = (histo(0, nbBins - 2) * histo(1, nbBins - 2)
						+ histo(0, nbBins - 1) * histo(1, nbBins - 1)) / denom;
			} else {
				double denom = histo(1, maximumIndex - 1) + histo(1,
						maximumIndex) + histo(1, maximumIndex + 1);
				if (isZero(denom))
					throw std::domain_error("roughPeakFinder: division by zero in pedestal calculus");

				pedestal = (histo(0, maximumIndex - 1) * histo(1, maximumIndex
						- 1) + histo(0, maximumIndex) * histo(1, maximumIndex)
						+ histo(0, maximumIndex + 1) * histo(1, maximumIndex
								+ 1)) / denom;
			}
		}

		// compute now a rough estimate of the noise as the half maximum of the
		// histo maximum
		double noiseLevel = 0;
		int index = maximumIndex;
		double maxHistoValue = histo(1, maximumIndex);

		while (index < nbBins && histo(1, index) > maxHistoValue / 2)
			index++;

		if (index != nbBins) { // it  has worked !
			double X1 = histo(0, maximumIndex);
			double X2 = histo(0, index);
			double Y1 = maxHistoValue;
			double Y2 = histo(1, index);
			//std::cout<<"i="<<index<<" X1="<<X1<<" Y1="<<Y1<<" X2="<<X2<<" Y2="<<Y2<<std::endl;

			if (isZero(Y1 - Y2))
				throw std::domain_error("roughPeakFinder: division by zero in noise level calculus");

			noiseLevel = (X1 * Y1 + X2 * Y1 - 2 * X1 * Y2) / (2. * (Y1 - Y2))
					- X1;
		} else {
			// an alternate solution, less accurate, as the rms of the derivative of the data
			Array<double, 1> derivative(n - 1);
			derivative = data(Range(1, n - 1)) - data(Range(0, n - 2));

			try {
				HorizontalMoments moments(derivative);
				noiseLevel = moments.mrms();
			}
			catch (const std::logic_error& error) {
				std::cerr << error.what() << std::endl;
				throw std::domain_error("roughPeakFinder: error in noise level calculus");
			}
		}

		// Decide if you have a negative or a positive peak
		bool positivePeak = true;
		double dataCenter = ::fabs(max(data) - min(data)) / 2.0;
		if ((dataCenter - pedestal) < 0.0)
			positivePeak = false;

		// compute cuts
		double cut = 0.0;
		if (positivePeak)
			cut = pedestal + 2 * noiseLevel; // changed from x4 to x2
		else
			cut = pedestal - 2 * noiseLevel;

//		std::cout<<"pedestal="<<pedestal<<std::endl;
//		std::cout<<"noiseLevel="<<noiseLevel<<std::endl;
//		std::cout<<"cut="<<cut<<std::endl;

		double peakIndex;
		double maximum;
		double peakWidth;

		// perform maximum finding in the ranges that are over the cuts
		int nbPeaks = 0; // number of found peaks
		Array<double, 2> peaks(ROUGHPEAK_MAX_NB_PEAKS); // storing position, height, width for each peak
		Array<double, 1> moments1(4);
		int i0 = 0; // the index of the last range start
		bool startFound = false; // indicates a range start

		double blurdata = 0.0;
		for (int i = 0; i < n; i++) {
			if (i < n - 2)
				blurdata = (data(i) + data(i + 1) + data(i + 2)) / 3.0;
			else
				blurdata = data(i);

			if (positivePeak) {
				if (!startFound && blurdata >= cut) // starting a range
				{
					i0 = i;
					startFound = true;
				} else {
					if (startFound && blurdata < cut) // stopping a range
					{
						startFound = false;
						// start finding maximum and maximum index
						maximum = max(data(Range(i0, i))); // look at the maximum in the range

						try {
							VerticalMoments moments1(data(Range(i0, i))); // find its index as the vertical mean
							peakIndex = moments1.mmean() + i0;
						}
						catch (const std::logic_error& error) {
							std::cerr << error.what() << std::endl;
							throw std::domain_error("roughPeakFinder: error in peak index calculus");
						}

						peakWidth = (double) (i - i0); // set the width as simply the width of the range (it's a rough peak finder , remember ?)

						//std::cout<<"i="<<i<<" i0="<<i0<<std::endl;
						if (nbPeaks < ROUGHPEAK_MAX_NB_PEAKS && peakWidth > 1) {
							// do not take the peak if the width is equal to 1
							peaks(nbPeaks, ROUGHPEAK_POSITION) = peakIndex; // store it
							peaks(nbPeaks, ROUGHPEAK_AMPLITUDE) = maximum - pedestal;
							peaks(nbPeaks, ROUGHPEAK_WIDTH) = peakWidth;
							peaks(nbPeaks, ROUGHPEAK_PEDESTAL) = pedestal;
							//std::cout << "rough peak " << (maximum - pedestal) << " " << peakWidth << " " << pedestal << std::endl;

							nbPeaks++;
						}
					}
				}
			} else {
				if (!startFound && blurdata <= cut) // starting a range
				{
					i0 = i;
					startFound = true;
				} else {
					if (startFound && blurdata > cut) // stopping a range
					{
						startFound = false;
						// start finding maximum and maximum index
						maximum = min(data(Range(i0, i))); // look at the maximum in the range

						try {
							VerticalMoments moments1(data(Range(i0, i))); // find its index as the vertical mean
							peakIndex = moments1.mmean() + i0;
						}
						catch (const std::logic_error& error) {
							std::cerr << error.what() << std::endl;
							throw std::domain_error("roughPeakFinder: error in peak index calculus");
						}

						peakWidth = (double) (i - i0); // set the width as simply the width of the range (it's a rough peak finder , remember ?)

						//std::cout<<"i="<<i<<" i0="<<i0<<std::endl;
						if (nbPeaks < ROUGHPEAK_MAX_NB_PEAKS && peakWidth > 1) {
							// do not take the peak if the width is equal to 1
							peaks(nbPeaks, ROUGHPEAK_POSITION) = peakIndex; // store it
							peaks(nbPeaks, ROUGHPEAK_AMPLITUDE) = maximum - pedestal;
							peaks(nbPeaks, ROUGHPEAK_WIDTH) = peakWidth;
							peaks(nbPeaks, ROUGHPEAK_PEDESTAL) = pedestal;
							//std::cout << "negative rough peak " << (maximum - pedestal) << " " << peakWidth << " " << pedestal << std::endl;

							nbPeaks++;
						}
					}
				}
			}
		}

		if (nbPeaks > 0) {
			return peaks(Range(0, nbPeaks - 1));
		} else {
			throw std::domain_error("roughPeakFinder: no peak found");
		}
	}
	catch (const logic_error& error) {
		cerr << error.what() << endl;
		throw(std::invalid_argument("roughPeakFinder: bad data"));
	}
}

}
