/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "SimpleGaussFinder.h"

using namespace std;
using namespace blitz;

SimpleGaussFinder::SimpleGaussFinder(SimpleGaussFinder::PeakType type) : m_type(type), _fit(0) {
}

SimpleGaussFinder::~SimpleGaussFinder() {
	delete _fit;
}

void SimpleGaussFinder::fit(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x) {

	if (x.size() < 3) {
		return;
	}

	// find initial curve parameters
	Array<double, 1> firstGuess(4);

	double positiveQuality = 0.0;
	double negativeQuality = 0.0;

	GaussFit * positiveFit = 0;
	GaussFit * negativeFit = 0;

	// test with positive gaussian
	if (m_type == ALL || m_type == MAXIMUM) {

		firstGuess = calculateGuessParameters(y, x, false);

		try {
			positiveFit = new GaussFit();
			positiveFit->fit(y, x, firstGuess, GAUSSFIT_GAUSSIAN_PEDESTAL, true, false);
			positiveQuality = positiveFit->quality();

		} catch (...) {
		}
	}

	// test with negative gaussian
	if (m_type == ALL || m_type == MINIMUM) {

		firstGuess = calculateGuessParameters(y, x, true);

		try {
			negativeFit = new GaussFit();
			negativeFit->fit(y, x, firstGuess, GAUSSFIT_GAUSSIAN_PEDESTAL, true, false);
			negativeQuality = negativeFit->quality();

		} catch (...) {
		}
	}

	//cout << "positive quality = " << positiveQuality << ", negative quality = " << negativeQuality << endl;

	if (m_type == MAXIMUM) {
		_fit = positiveFit;
	} else if (m_type == MINIMUM) {
		_fit = negativeFit;
	} else {
		// compare positive and negative fit
		if (positiveQuality >= negativeQuality) {
			_fit = positiveFit;
			delete negativeFit;
		} else {
			_fit = negativeFit;
			delete positiveFit;
		}
	}
}

void SimpleGaussFinder::fit(const blitz::Array<double, 1>& y) {

	// create x array
	int n = y.size();
	int minIndex = y.lbound(firstDim);
	Array<double, 1> x(n);
	firstIndex i;
	x = i + minIndex;

	fit(y, x);
}

Array<double, 1> SimpleGaussFinder::calculateGuessParameters(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x, bool negative) {

	Array<double, 1> a(4);

	if (negative) {
		a(3) = max(y);
		a(0) = min(y) - a(3);
		a(1) = x(0) + (x(x.size() - 1) - x(0)) / 2.0;
		a(2) = (x(x.size() - 1) - x(0)) / (3.0 * 2.3548);

	} else {
		a(3) = min(y);
		a(0) = max(y) - a(3);
		a(1) = x(0) + (x(x.size() - 1) - x(0)) / 2.0;
		a(2) = (x(x.size() - 1) - x(0)) / (3.0 * 2.3548);
	}

	return a;
}

bool SimpleGaussFinder::fitHasConverged() const {
	if (_fit != 0) {
		if (_fit->fitHasConverged()) {

			// testing peak type
			if (m_type == MAXIMUM) {
				return _fit->gaussianAmplitude() > 0;
			} else if (m_type == MINIMUM) {
				return _fit->gaussianAmplitude() < 0;
			} else {
				return true;
			}
		}
	}
	return false;
}

double SimpleGaussFinder::chiSquare() const {
	if (_fit != 0) {
		return _fit->chiSquare();
	}
	return 0.0;
}

double SimpleGaussFinder::position() const {
	if (_fit != 0) {
		return _fit->mean();
	}
	return 0.0;
}

double SimpleGaussFinder::rms() const {
	if (_fit != 0) {
		return _fit->rms();
	}
	return 0.0;
}

double SimpleGaussFinder::gaussianAmplitude() const {
	if (_fit != 0) {
		return _fit->gaussianAmplitude();
	}
	return 0.0;
}

double SimpleGaussFinder::amplitude() const {
	if (_fit != 0) {
		return _fit->amplitude();
	}
	return 0.0;
}

double SimpleGaussFinder::pedestal() const {
	if (_fit != 0) {
		return _fit->pedestalOrder(0);
	}
	return 0.0;
}

blitz::Array<double, 2> SimpleGaussFinder::getFittedCurve(unsigned int minIndex, unsigned int maxIndex, unsigned int nbSamples) const  {
	if (_fit != 0) {
		return _fit->getFittedCurve(minIndex, maxIndex, nbSamples);
	}
	Array<double, 2> result;
	return result;
}

blitz::Array<double, 2> SimpleGaussFinder::getFittedCurve(unsigned int nbSamples) const  {
	if (_fit != 0) {
		return _fit->getFittedCurve(nbSamples);
	}
	Array<double, 2> result;
	return result;
}

blitz::Array<double, 2> SimpleGaussFinder::getFittedCurve(unsigned int minIndex, unsigned int maxIndex) const  {
	if (_fit != 0) {
		return _fit->getFittedCurve(minIndex, maxIndex);
	}
	Array<double, 2> result;
	return result;
}

blitz::Array<double, 2> SimpleGaussFinder::getFittedCurve() const  {
	if (_fit != 0) {
		return _fit->getFittedCurve();
	}
	Array<double, 2> result;
	return result;
}

double SimpleGaussFinder::quality() const {
	if (_fit != 0) {
		return _fit->quality();
	}
	return 0.0;
}

double SimpleGaussFinder::dispersionRatio() const {
	if (_fit != 0) {
		return _fit->dispersionRatio();
	}
	return 0.0;
}

blitz::Array<double, 1> SimpleGaussFinder::getParameters() const {
	Array<double, 1> result(4);

	if (_fit != 0) {
		result(0) = _fit->gaussianAmplitude();
		result(1) = _fit->mean();
		result(2) = _fit->rms();
		result(3) = _fit->pedestalOrder(0);
	}
	return result;
}
