/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef SIMPLEGAUSSFINDER_H
#define SIMPLEGAUSSFINDER_H

#include "GaussFit.h"

class SimpleGaussFinder {

public:
	enum PeakType {ALL, MAXIMUM, MINIMUM};

	/**
	 * Constructor.
	 */
	SimpleGaussFinder(PeakType type = ALL);

	/**
	 * Destructor.
	 */
	virtual ~SimpleGaussFinder();

	/**
	 * Fits the curve.
	 */
	void fit(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x);

	/**
	 * Fits the curve.
	 */
	void fit(const blitz::Array<double, 1>& y);

	/**
	 * Indicates that the fit has converged
	 */
	bool fitHasConverged() const;

	/**
	 * Returns the chi square of the fit
	 */
	double chiSquare() const;

	/**
	 * Returns the position
	 */
	double position() const;

	/**
	 * Returns the rms of the gaussian
	 */
	double rms() const;

	/**
	 * This returns the amplitude of the fitted function, <b> that without the pedestal value</b>.
	 */
	double gaussianAmplitude() const;

	/**
	 * Returns the full amplitude of the peak, meaning with the pedestal value. (amplitude+pedestal)
	 */
	double amplitude() const;

	/**
	 * Return the pedestal
	 */
	double pedestal() const;

	/**
	 * Getting the interpolated curve
	 */
	blitz::Array<double, 2> getFittedCurve(unsigned int minIndex, unsigned int maxIndex, unsigned int nbSamples) const ;

	/**
	 * Getting the interpolated curve
	 */
	blitz::Array<double, 2> getFittedCurve(unsigned int nbSamples) const ;

	/**
	 * Getting the interpolated curve
	 */
	blitz::Array<double, 2> getFittedCurve(unsigned int minIndex, unsigned int maxIndex) const ;

	/**
	 * Getting the interpolated curve
	 */
	blitz::Array<double, 2> getFittedCurve() const ;

	/**
	 * Getting the parameters
	 */
	blitz::Array<double, 1> getParameters() const;

	/**
	 * Gets the quality.
	 */
	double quality() const;

	/**
	 * Gets the dispersion ratio.
	 */
	double dispersionRatio() const;

private:
	blitz::Array<double, 1> calculateGuessParameters(const blitz::Array<double, 1>& y, const blitz::Array<double, 1>& x, bool negative);

	PeakType m_type;
	GaussFit * _fit;
};

#endif
