/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Convert.h"
#include "PrimitiveConvert.h"
#include <sstream>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <sys/wait.h>
#include <sys/types.h>

using namespace std;

namespace common {

const std::string StringHelper::MAX_NUMBER_OF_PRINTED_ARRAY_ELEMENTS = "maxNumberOfPrintedArrayElements";
unsigned long StringHelper::_maxNumberOfPrintedArrayElements;

void StringHelper::setMaxNumberOfPrintedArrayElements(unsigned long value) {
	_maxNumberOfPrintedArrayElements = value;
}

string ConvertHelper<string, char>::convert(char const & value, std::string const & format) {
	return PrimitiveConvert<string, char>::convert(value, format);
}

string ConvertHelper<string, signed char>::convert(signed char const & value, std::string const & format) {
	return PrimitiveConvert<string, signed char>::convert(value, format);
}

string ConvertHelper<string, short>::convert(short const & value, std::string const & format) {
	return PrimitiveConvert<string, short>::convert(value, format);
}

string ConvertHelper<string, long>::convert(long const & value, std::string const & format) {
	return PrimitiveConvert<string, long>::convert(value, format);
}

string ConvertHelper<string, unsigned long>::convert(unsigned long const & value, std::string const & format) {
	return PrimitiveConvert<string, unsigned long>::convert(value, format);
}

string ConvertHelper<string, int>::convert(int const & value, std::string const & format) {
	return PrimitiveConvert<string, int>::convert(value, format);
}

string ConvertHelper<string, unsigned int>::convert(unsigned int const & value, std::string const & format) {
	return PrimitiveConvert<string, unsigned int>::convert(value, format);
}

string ConvertHelper<string, unsigned short>::convert(unsigned short const & value, std::string const & format) {
	return PrimitiveConvert<string, unsigned short>::convert(value, format);
}

string ConvertHelper<string, long long>::convert(long long const & value, std::string const & format) {
	return PrimitiveConvert<string, long long>::convert(value, format);
}

string ConvertHelper<string, bool>::convert(bool const & value, std::string const & format) {
	return PrimitiveConvert<string, bool>::convert(value, format);
}

string ConvertHelper<string, float32>::convert(float32 const & value, std::string const & format) {
	return PrimitiveConvert<string, float32>::convert(value, format);
}

string ConvertHelper<string, float64>::convert(float64 const & value, std::string const & format) {
	return PrimitiveConvert<string, float64>::convert(value, format);
}

string ConvertHelper<string, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<string, string>::convert(value, format);
}

char ConvertHelper<char, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<char, string>::convert(value, format);
}

short ConvertHelper<short, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<short, string>::convert(value, format);
}

unsigned short ConvertHelper<unsigned short, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<unsigned short, string>::convert(value, format);
}

long ConvertHelper<long, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<long, string>::convert(value, format);
}

unsigned long ConvertHelper<unsigned long, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<unsigned long, string>::convert(value, format);
}

int ConvertHelper<int, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<int, string>::convert(value, format);
}

unsigned int ConvertHelper<unsigned int, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<unsigned int, string>::convert(value, format);
}

long long ConvertHelper<long long, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<long long, string>::convert(value, format);
}

bool ConvertHelper<bool, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<bool, string>::convert(value);
}

float32 ConvertHelper<float32, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<float32, string>::convert(value, format);
}

float64 ConvertHelper<float64, string>::convert(string const & value, std::string const & format) {
	return PrimitiveConvert<float64, string>::convert(value, format);
}

ConversionException::ConversionException(const string& message) :
	Exception("common::ConversionException", message) {
}

void StringHelper::init() {

	try {
		_maxNumberOfPrintedArrayElements = ServerProperties::getInstance()->get<unsigned long>(MAX_NUMBER_OF_PRINTED_ARRAY_ELEMENTS);

	} catch (Exception& e) {
		out << "Setting default value 100 to property 'maxNumberOfPrintedArrayElements'" << eol;
		_maxNumberOfPrintedArrayElements = 100;
	}
}

unsigned long StringHelper::getMaxNumberOfPrintedArrayElements() {

	return _maxNumberOfPrintedArrayElements;
}

list<string> StringHelper::split(const string& input, const string& separator) {

	list<string> result;

	unsigned int index = input.find_first_not_of(separator);
	while(index < input.size() && index != string::npos) {
		unsigned int endPos = input.find_first_of(separator, index);

		if (endPos != string::npos) {
			result.push_back(input.substr(index, endPos - index));
			index = input.find_first_not_of(separator, endPos);
		} else {
			result.push_back(input.substr(index, input.size() - index));
			break;
		}
	}

	return result;
}

list<StringHelper::SplitItem> StringHelper::splitWithSeparator(const string& input, const string& separator) {

	list<SplitItem> result;

	bool isToken = (input.find_first_of(separator) != 0);
	int position = 0;
	int newPosition = 0;

	while (newPosition != string::npos) {

		SplitItem item;
		item.isToken = isToken;
		if (isToken) {
			newPosition = input.find_first_of(separator, position);
		} else {
			newPosition = input.find_first_not_of(separator, position);
		}

		item.value = input.substr(position, newPosition - position);
		result.push_back(item);
		isToken = !isToken;
		position = newPosition;
	}

	return result;
}

StringHelper::Node* StringHelper::tokenise(const string& input) {

	int openingBracket = findNextOpeningBracket(input, 0);
	if (openingBracket > -1) {

		List* list = new List();

		// Sub list
		int position = openingBracket + 1;

		while (position < (int)input.size()) {

			int semi = findNextComma(input, position);

			Node* node = tokenise(input.substr(position, semi - position));
			if (node != 0) {
				list->nodes.push_back(node);
			}
			position = semi + 1;
		}

		return list;

	} else {

		string strippedInput = stripWhitespaces(input);
		if (strippedInput != "") {

			// Element list
			Element* element = new Element();
			element->content = stripWhitespaces(input);

			return element;
		}
	}

	return 0;
}

void StringHelper::release(Node* node) {

	delete node;
}

std::string StringHelper::stripWhitespaces(const std::string& input) {

	string result;
	int begin = 0;
	while (input[begin] == ' ' || input[begin] == '\t' || input[begin] == '\n' || input[begin] == '\r') {
		begin++;
	}

	int end = input.size() - 1;
	while (input[end] == ' ' || input[end] == '\t' || input[end] == '\n' || input[end] == '\r') {
		end--;
	}

	return input.substr(begin, end - begin + 1);
}

int StringHelper::findNextOpeningBracket(const string& input, int position) {

	for (int i = position; i < (int)input.size(); i++) {
		if (input[i] == '[') {
			return i;
		}
	}

	return -1;
}

int StringHelper::findNextClosingBracket(const string& input, int position) {

	int counter = 0;
	for (int i = position; i < (int)input.size(); i++) {
		if (input[i] == '[') {
			counter++;
		}

		if (input[i] == ']') {
			if (counter != 0) {
				counter--;
			} else {
				return i;
			}
		}
	}

	return -1;

}

int StringHelper::findNextComma(const string& input, int position) {

	int counter = 0;
	for (int i = position; i < (int)input.size(); i++) {
		if (input[i] == '[') {
			counter++;
		}

		if (input[i] == ']') {
			counter--;
		}

		if ((input[i] == ',') && (counter == 0)) {
			return i;
		}
	}

	return input.size() - 1;

}

StringHelper::Node::~Node() {

}

StringHelper::List::~List() {

	for (list<Node*>::const_iterator n = nodes.begin(); n != nodes.end(); n++) {
		delete *n;
	}
}

string StringHelper::List::toString() {

	string result;
	result += "[";

	if (!nodes.empty()) {

		list<Node*>::const_iterator n = nodes.begin();
		for (unsigned int i = 0; i < nodes.size() - 1; i++) {
			result += (*n)->toString() + "; ";
			n++;
		}

		result += (*n)->toString();

	}

	result += "]";

	return result;
}

StringHelper::Element::~Element() {

}

string StringHelper::Element::toString() {

	return content;

}

bool StringHelper::isWhitespace(const std::string& input) {

	for (size_t i = 0; i < input.size(); i++) {
		if ((input[i] != ' ') && (input[i] != '\n') && (input[i] != '\t') && (input[i] != '\r')) {
			return false;
		}
	}

	return true;
}

std::string convertToHexString(unsigned long value) {
	ostringstream hexValue;
	hexValue << "0x"<< std::hex << value;

	return hexValue.str();
}

std::string convertSignal(int signal) {

	if (signal == SIGABRT) {
		return "SIGABRT";
	} else if (signal == SIGALRM) {
		return "SIGALRM";
	} else if (signal == SIGFPE) {
		return "SIGFPE";
	} else if (signal == SIGHUP) {
		return "SIGHUP";
	} else if (signal == SIGILL) {
		return "SIGILL";
	} else if (signal == SIGINT) {
		return "SIGINT";
	} else if (signal == SIGKILL) {
		return "SIGKILL";
	} else if (signal == SIGPIPE) {
		return "SIGPIPE";
	} else if (signal == SIGQUIT) {
		return "SIGQUIT";
	} else if (signal == SIGSEGV) {
		return "SIGSEGV";
	} else if (signal == SIGTERM) {
		return "SIGTERM";
	} else if (signal == SIGUSR1) {
		return "SIGUSR1";
	} else if (signal == SIGUSR2) {
		return "SIGUSR2";
	} else if (signal == SIGCHLD) {
		return "SIGCHLD";
	} else if (signal == SIGCONT) {
		return "SIGCONT";
	} else if (signal == SIGSTOP) {
		return "SIGSTOP";
	} else if (signal == SIGTSTP) {
		return "SIGTSTP";
	} else if (signal == SIGTTIN) {
		return "SIGTTIN";
	} else if (signal == SIGTTOU) {
		return "SIGTTOU";
	} else if (signal == SIGBUS) {
		return "SIGBUS";
	} else if (signal == SIGPOLL) {
		return "SIGPOLL";
	} else if (signal == SIGPROF) {
		return "SIGPROF";
	} else if (signal == SIGSYS) {
		return "SIGSYS";
	} else if (signal == SIGTRAP) {
		return "SIGTRAP";
	} else if (signal == SIGURG) {
		return "SIGURG";
	} else if (signal == SIGVTALRM) {
		return "SIGVTALRM";
	} else if (signal == SIGXCPU) {
		return "SIGXCPU";
	} else if (signal == SIGXFSZ) {
		return "SIGXFSZ";
	} else {
		return "" + to<string>(signal);
	}
}

std::string convertStatus(int status) {

	string result;

	if (WIFEXITED(status)) {
		result += "WIFEXITED";
	}

	if (WEXITSTATUS(status)) {
		if (result.length() > 0) {
			result += " ";
		}
		result += "WEXITSTATUS";
	}

	if (WIFSIGNALED(status)) {
		if (result.length() > 0) {
			result += " ";
		}
		result += "WIFSIGNALED with " + convertSignal(WTERMSIG(status));
	}

	if (WIFSTOPPED(status)) {
		if (result.length() > 0) {
			result += " ";
		}
		result += "WIFSTOPPED with " + convertSignal(WSTOPSIG(status));
	}

	return result;
}

std::string convertStringToXML(std::string const & value) {

	string xmlValue;

	for (int i = 0; i < value.length(); i++) {

		// Do not manage ≡ as it seems to be ok to keep it as is (at least in the Nomad GUI log viewer).

		if (value[i] == '&') {
			xmlValue += "&amp;";
		}
		else if (value[i] == '>') {
			xmlValue += "&gt;";
		}
		else if (value[i] == '"') {
			xmlValue += "&quot;";
		}
		else if (value[i] == '<') {
			xmlValue += "&lt;";
		}
		else if (value[i] == '\'') {
			xmlValue += "&apos;";
		}
		else {
			xmlValue += value[i];
		}
	}

	return xmlValue;
}

void replaceString(std::string& value, std::string const & search, std::string const & replace) {

	size_t position(0);

	position = value.find(search, position + 1);

	while (position != string::npos) {
		value.replace(position, search.size(), replace);
		position = value.find(search, position + 1);
	}
}

std::string convertXMLStringToAscii(std::string const & value) {

	string result = value;

	replaceString(result, "&amp;", "&");
	replaceString(result, "&gt;", ">");
	replaceString(result, "quot;", "\"");
	replaceString(result, "&lt;", "<");
	replaceString(result, "&apos;", "\'");

	return result;
}

std::string convertStringToJSON(std::string const & value) {

	string jsonValue;

	for (int i = 0; i < value.length(); i++) {
		if (value[i] == '"') {
			jsonValue += "\\\"";

		}
		else if (value[i] == '\\') {
			jsonValue += "\\\\";

		} else {
			jsonValue += value[i];
		}
	}

	return jsonValue;
}

}
