/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RealArray.h"
#include "ClonedArray.h"
#include <vector>
#include <list>

namespace common {

class StringHelper {

	template<typename Type> struct Loader;

	template<typename ToType, typename FromType>
	friend struct ConvertHelper;

public:

	struct Node;

	static void init();
	static unsigned long getMaxNumberOfPrintedArrayElements();

	/**
 	 * Returns the list of token from the split of the input string with separator.
 	 */
	static std::list<std::string> split(const std::string& input, const std::string& separator = " ");

	struct SplitItem {
		bool isToken;
		std::string value;
	};

	static std::list<SplitItem> splitWithSeparator(const std::string& input, const std::string& separator = " ");


	template<typename Type>
	static Type load(Node* node, std::string const & format) {

		return Loader<Type>::load(node, format);
	}

	struct Node {
		virtual ~Node();
		virtual std::string toString() = 0;
	};

	struct List : public Node {
		std::list<Node*> nodes;

		virtual ~List();
		virtual std::string toString();
	};

	struct Element : public Node {
		std::string content;

		virtual ~Element();
		virtual std::string toString();
	};

	static Node* tokenise(const std::string& input);
	static void release(Node* node);

	static std::string stripWhitespaces(const std::string& input);

	static void setMaxNumberOfPrintedArrayElements(unsigned long value);

private:
	static int findNextOpeningBracket(const std::string& input, int position);
	static int findNextClosingBracket(const std::string& input, int position);
	static int findNextComma(const std::string& input, int position);
	static bool isWhitespace(const std::string& input);

	/**
	 * Default template implementation. Used for primitive types like long long...
	 */
	template<typename Type>
	struct Loader {

		static Type load(Node* node, std::string const & format);
	};

	template<typename ElementType>
	struct Loader<RealArray<ElementType> > {

		static RealArray<ElementType> load(Node* node, std::string const & format);
	};

	static const std::string MAX_NUMBER_OF_PRINTED_ARRAY_ELEMENTS;
	static unsigned long _maxNumberOfPrintedArrayElements;
};

template<typename ToType, typename FromType>
ToType ConvertHelper<ToType, FromType>::convert(FromType const & value, std::string const & format) {
	return static_cast<ToType>(value);
}

template<typename FromType>
struct ConvertHelper<std::string, RealArray<FromType> > {
	static std::string convert(RealArray<FromType> const & value, std::string const & format);
};

template<typename FromType>
struct ConvertHelper<std::string, ClonedArray<FromType> > {
	static std::string convert(ClonedArray<FromType> const & value, std::string const & format);
};

template<typename FromType>
std::string ConvertHelper<std::string, RealArray<FromType> >::convert(RealArray<FromType> const & value, std::string const & format) {

	std::string result;

	result += "[";

	long size = value.getSize();

	if (size > 0) {
		FromType* data = value.getData();

		for (long i = 0; i < size - 1; i++) {
			result += ConvertHelper<std::string, FromType>::convert(data[i], format) + ", ";
			// testing limit value
			if (i == (long)StringHelper::getMaxNumberOfPrintedArrayElements() - 2) {
				result += "... ";
				break;
			}
		}

		result += ConvertHelper<std::string, FromType>::convert(data[size - 1], format);
	}

	result += "]";

	return result;
}

template<typename FromType>
std::string ConvertHelper<std::string, ClonedArray<FromType> >::convert(ClonedArray<FromType> const & value, std::string const & format) {

	std::string result;

	result += "[";

	long size = value.getSize();

	if (size > 0) {
		for (long i = 0; i < size - 1; i++) {
			result += ConvertHelper<std::string, FromType>::convert(FromType(), format) + ", ";
			// testing limit value
			if (i == (long)StringHelper::getMaxNumberOfPrintedArrayElements() - 2) {
				result += "... ";
				break;
			}
		}

		result += ConvertHelper<std::string, FromType>::convert(FromType(), format);
	}

	result += "]";

	return result;
}

template<typename FromType>
std::ostream& operator<< (std::ostream& os, RealArray<FromType> const & A) {
	return os << ConvertHelper<std::string, RealArray<FromType> >::convert(A, "");
}

/**
 * Conversion to string. Template partial specialisation not allowed for functions.
 * Let's enumerate all the types!
 */

template<>
struct ConvertHelper<std::string, char> {
	static std::string convert(char const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, signed char> {
	static std::string convert(signed char const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, short> {
	static std::string convert(short const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, long> {
	static std::string convert(long const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, unsigned long> {
	static std::string convert(unsigned long const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, int> {
	static std::string convert(int const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, unsigned int> {
	static std::string convert(unsigned int const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, unsigned short> {
	static std::string convert(unsigned short const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, long long> {
	static std::string convert(long long const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, bool> {
	static std::string convert(bool const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, float32> {
	static std::string convert(float32 const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, float64> {
	static std::string convert(float64 const & value, std::string const & format);
};

template<>
struct ConvertHelper<std::string, std::string> {
	static std::string convert(std::string const & value, std::string const & format);
};

/**
 * String to type conversion.
 */

template<typename ToType>
struct ConvertHelper<ToType, std::string> {
	static ToType convert(std::string const & value, std::string const & format);
};

template<typename ToType>
ToType ConvertHelper<ToType, std::string>::convert(std::string const & value, std::string const & format) {

	StringHelper::Node* node = StringHelper::tokenise(value);

	ToType result;

	try {
		result = StringHelper::load<ToType>(node, format);
		StringHelper::release(node);

	} catch (Exception&) {
		StringHelper::release(node);
		throw;
	}

	return result;
}

template<>
struct ConvertHelper<char, std::string> {
	static char convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<short, std::string> {
	static short convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<unsigned short, std::string> {
	static unsigned short convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<long, std::string> {
	static long convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<unsigned long, std::string> {
	static unsigned long convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<int, std::string> {
	static int convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<unsigned int, std::string> {
	static unsigned int convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<long long, std::string> {
	static long long convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<bool, std::string> {
	static bool convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<float32, std::string> {
	static float32 convert(std::string const & value, std::string const & format);
};

template<>
struct ConvertHelper<float64, std::string> {
	static float64 convert(std::string const & value, std::string const & format);
};

template<typename FromType>
struct ConvertHelper<std::string, FromType*> {
	static std::string convert(FromType* value, std::string const & format);
};

template<typename FromType>
std::string ConvertHelper<std::string, FromType*>::convert(FromType* value, std::string const & format) {
	return ConvertHelper<std::string, long>::convert(reinterpret_cast<long>(value), format);
}

template<typename ValueType>
struct ConvertHelper<std::vector<ValueType>, std::list<std::string> > {
	static std::vector<ValueType> convert(const std::list<std::string>& values, std::string const & format);
};

template<typename ValueType>
std::vector<ValueType> ConvertHelper<std::vector<ValueType>, std::list<std::string> >::convert(const std::list<std::string>& values, std::string const & format) {
	std::vector<ValueType> returnVector;
	returnVector.reserve(values.size());
	std::list<std::string>::const_iterator v;

	for (v = values.begin(); v != values.end(); v++) {
		returnVector.push_back(to<ValueType>(*v));
	}

	return returnVector;
}

template<typename Type>
Type StringHelper::Loader<Type>::load(StringHelper::Node* node, std::string const & format) {

	StringHelper::Element* element = dynamic_cast<StringHelper::Element*>(node);
	if (element == 0) {
		throw ConversionException("cannot load primitive type");
	}

	Type result = to<Type>(element->content, format);

	return result;
}

template<typename ElementType>
RealArray<ElementType> StringHelper::Loader<RealArray<ElementType> >::load(StringHelper::Node* node, std::string const & format) {

	StringHelper::List* list = dynamic_cast<StringHelper::List*>(node);
	if (list == 0) {
		throw ConversionException("cannot load Array type");
	}

	long size = list->nodes.size();
	RealArray<ElementType> result(size);
	std::list<StringHelper::Node*>::const_iterator n = list->nodes.begin();
	for (long i = 0; i < size; i++) {
		result[i] = StringHelper::Loader<ElementType>::load(*n, format);
		n++;
	}

	return result;
}

}
