/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Date.h"
#include <iostream>
#include <sstream>

using namespace std;
using namespace boost::posix_time;
using namespace boost::local_time;
using namespace boost::gregorian;

namespace common {

time_zone_ptr Date::_zone = time_zone_ptr(new posix_time_zone("UTC-0"));

Date::Date() :
	_time(microsec_clock::local_time()) {
}

Date::Date(int64 timeUs) {
	
	long sec = timeUs/1000000;
	long microsec = timeUs - (int64)sec*(int64)1000000;
	time_duration duration = seconds(sec) + microseconds(microsec);
	_time = ptime(date(1970, Jan, 1)) + duration;
}

Date::Date(boost::posix_time::ptime time) {
	_time = time;
}

Date::~Date() {
}

int64 Date::getTimeUs() const {
	int64 result;
	time_duration duration = _time - ptime(date(1970, Jan, 1));
	result = (int64)duration.total_seconds()*(int64)1000000;
	result += duration.fractional_seconds();
	return result;
}

int64 Date::getTimeMs() const {
	int64 result;
	time_duration duration = _time - ptime(date(1970, Jan, 1));
	result = (int64)duration.total_seconds()*(int64)1000;
	result += (int64) floor(duration.fractional_seconds() / 1000.);
	return result;
}

int64 Date::getDateUs() const {
	int64 result;
	time_duration duration = _time - _time.time_of_day() - ptime(date(1970, Jan, 1));
	result = (int64)duration.total_seconds()*(int64)1000000;
	result += duration.fractional_seconds();
	return result;
}

std::string Date::toString() const {
	return to_iso_extended_string(_time);
}

std::string Date::toString(std::string const & format) const {
	
	local_time_facet* outputFacet = new local_time_facet();
	local_time_input_facet* inputFacet = new local_time_input_facet();
	
	outputFacet->format(format.c_str());
	inputFacet->format(format.c_str());
	
	stringstream ss;
	
	ss.imbue(locale(locale::classic(), outputFacet));
	ss.imbue(locale(ss.getloc(), inputFacet));

	local_date_time localDate(_time, _zone);
	
	ss << localDate;
	return ss.str();
	
}

void Date::operator+= (const Duration& duration) {
	_time += duration._duration;
}

void Date::operator-= (const Duration& duration) {
	_time -= duration._duration;
}

Date operator+ (const Date& date, const Duration& duration) {
	Date result = date;
	result += duration;
	
	return result;
}

Date operator- (const Date& date, const Duration& duration) {
	Date result = date;
	result -= duration;
	
	return result;
}

bool Date::operator< (const Date& date) {
	return _time < date._time;
}

bool Date::operator> (const Date& date) {
	return _time > date._time;
}

bool Date::operator<= (const Date& date) {
	return _time <= date._time;
}

bool Date::operator>= (const Date& date) {
	return _time >= date._time;
}

bool Date::operator== (const Date& date) {
	return _time == date._time;
}

bool Date::operator!= (const Date& date) {
	return _time != date._time;
}

std::ostream& operator<<(std::ostream &o, const Date& date) {
	o << date.toString();
	
	return o;
}

}
