/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_DATE_H_
#define COMMON_DATE_H_

#include "Duration.h"
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/date_time/local_time/local_time.hpp>
#include <common/base/BaseTypes.h>
#include <string>
#include <iostream>
#include <boost/version.hpp>

#if BOOST_VERSION < 105000
#define TIME_UTC_ TIME_UTC
#endif

namespace common {

class Duration;

/**
 * Class defining a date to facilitate calculus with dates and durations.
 */
class Date {

	friend class Duration;

public:
	/**
	 * Constructor. The date is set to the present.
	 */
	Date();
	
	/**
	 * Constructor from the number of microseconds elapsed since 1970, january, 1st.
	 */
	Date(int64 timeUs);
	
	/**
	 * Constructor from posix time.
	 */
	Date(boost::posix_time::ptime time);
	
	/**
	 * Virtual destructor.
	 */
	virtual ~Date();
	
	/**
	 * += operator.
	 */
	void operator+= (const Duration& duration);
	
	/**
	 * -= operator.
	 */
	void operator-= (const Duration& duration);
	
	/**
	 * < operator.
	 */
	bool operator< (const Date& date);
	
	/**
	 * > operator.
	 */
	bool operator> (const Date& date);

	/**
	 * <= operator.
	 */
	bool operator<= (const Date& date);
	
	/**
	 * >= operator.
	 */
	bool operator>= (const Date& date);

	/**
	 * == operator.
	 */
	bool operator== (const Date& date);
	
	/**
	 * != operator.
	 */
	bool operator!= (const Date& date);

	/**
	 * Returns the number of microseconds elapsed since 1970, january, 1st.
	 */
	int64 getTimeUs() const;

	/**
	 * Returns the number of microseconds elapsed since 1970, january, 1st.
	 */
	int64 getTimeMs() const;

	/**
	 * Returns the number of microseconds elapsed since 1970, january, 1st for just day at 0:0:0 time.
	 */
	int64 getDateUs() const;

	/**
	 * Returns an iso formatted date.
	 */
	std::string toString() const;
	
	/**
	 * Returns a formatted date.
	 */
	std::string toString(std::string const & format) const;

private:
	boost::posix_time::ptime _time;
	static boost::local_time::time_zone_ptr _zone;	
};

Date operator+ (const Date& date, const Duration& duration);
Date operator- (const Date& date, const Duration& duration);

std::ostream& operator<<(std::ostream &o, const Date& date);

}

#endif /*COMMON_DATE_H_*/
