/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_DURATION_H_
#define COMMON_DURATION_H_

#include <boost/date_time/posix_time/posix_time.hpp>
#include <string>
#include <iostream>

namespace common {

class Date;

/**
 * Class defining a duration to facilitate calculus with dates and durations. Simple arithmetic operators are provided 
 * and a difference of two Dates is a Duration.
 */
class Duration {
	
	friend class Date;	
public:
	typedef enum {MILLISECONDS, SECONDS} TimeUnit; 

	/**
	 * Constructor.
	 */
	Duration();
	
	/**
	 * Constructor with parameters.
	 */
	Duration(const Date& d0, const Date& d1);
	
	/**
	 * Constructor with parameters.
	 */
	Duration(double time, TimeUnit unit);
	
	/**
	 * Virtual destructor.
	 */
	virtual ~Duration();
	
	/**
	 * += operator.
	 */
	void operator+= (const Duration& duration);
	
	/**
	 * -= operator.
	 */
	void operator-= (const Duration& duration);
	
	/**
	 * Returns the duration value in seconds.
	 */
	double getSeconds() const;
	
	/**
	 * Returns the duration value in milliseconds.
	 */
	double getMilliseconds() const;

	/**
	 * Returns the posix time.
	 */
	const boost::posix_time::time_duration& getPosixTime() const;

	/**
	 * Returns a formatted date.
	 */
	std::string toString() const;

private:
	boost::posix_time::time_duration _duration;

};

Duration operator- (const Duration& d0, const Duration& d1);
Duration operator+ (const Duration& d0, const Duration& d1);
Duration operator- (const Date& d1, const Date& d0);

std::ostream& operator<<(std::ostream &o, const Duration& duration);

}

#endif /*COMMON_DURATION_H_*/
