/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_EXCEPTION_H_
#define COMMON_EXCEPTION_H_

#include <string>
#include <list>

namespace common {

/**
 * Class defining an exception. Here is the way to declare and define
 * an exception:
 *
 * Declaration:
 *
 * class A {
 *
 * public:
 * 		class MyException : public Exception {
 *
 * 		public:
 * 			MyException(std::string const & text);
 * 		};
 * };
 *
 * Definition:
 *
 * A::MyException::MyException(std::string const & text)
 * : Exception("A::MyException", text) {
 * }
 *
 * How to use it:
 *
 * With no child:
 *
 * throw MyException("my text");
 *
 * Thrown from a AnotherException e by using the practical link function:
 *
 * throw link(MyException("my text"), e);
 *
 */
class Exception {

public:
	/**
	 * Constructor.
	 */
	Exception(std::string const & type, std::string const & text);

	/**
	 * Virtual destructor.
	 */
	virtual ~Exception();

	/**
	 * Returns the string of the exception stack.
	 */
	virtual std::string printStack() const;

	/**
	 * Returns the string of the current exception.
	 */
	std::string printCurrentException() const;

	/**
	 * Returns the message of the current exception.
	 */
	std::string const & getMessage() const;

protected:

	struct element {
		std::string type;
		std::string text;
	};

	std::list<element> _exceptionStack;

private:
	// Friend functions.
	template<typename ExceptionType> friend ExceptionType link(ExceptionType const & e, Exception & child);
	template<typename ExceptionType> friend ExceptionType link(ExceptionType const & e, Exception const & child);

	void setChild(Exception & exception);
	void setChild(Exception const & exception);

};

/**
 * Function to facilitate exception stack.
 */
template<typename ExceptionType>
ExceptionType link(ExceptionType const & e, Exception & child) {

	ExceptionType result(e);
	result.setChild(child);

	return result;
}

template<typename ExceptionType>
ExceptionType link(ExceptionType const & e, Exception const & child) {

	ExceptionType result(e);
	result.setChild(child);

	return result;
}

}

#endif /*COMMON_EXCEPTION_H_*/