/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "FileUtilities.h"
#include <boost/filesystem/operations.hpp>
#include <common/output/OutputStream.h>
#include <iostream>
#include <fstream>

using namespace std;
namespace fs = boost::filesystem;

namespace common {


FileUtilities::NoSuchDirectoryException::NoSuchDirectoryException(std::string const & text)
: Exception("common::FileUtilities::NoSuchDirectoryException", text) {
}

FileUtilities::NotRegularFile::NotRegularFile(std::string const & text)
: Exception("common::FileUtilities::NotRegularFile", text) {
}

std::vector<std::string> FileUtilities::getFilesWithExtension(const std::string& pathName, const std::string& extension) {
	vector<string> fileNames;

	fs::path path = fs::system_complete(fs::path(pathName));

	if (fs::exists(path) && fs::is_directory(path)) {
			
		fs::directory_iterator end_iter;

		for (fs::directory_iterator dir_itr(path); dir_itr != end_iter; ++dir_itr) {
			try {
				if (!fs::is_directory(*dir_itr)) {
					
					string fileName = dir_itr->path().string();
					size_t pos = fileName.find_last_of('.') + 1;
					
					if (pos < fileName.length() && fileName.substr(pos, fileName.length() - pos) == extension) {
						fileNames.push_back(fileName);
					}
				}

			} catch (const std::exception & ex) {
				err << nameToString(dir_itr->path().filename()) << " " << ex.what() << eol;
			}
		}

	} else {
		throw NoSuchDirectoryException("cannot read " + pathName);
 	}
 	
 	return fileNames;

}

std::vector<fs::path> FileUtilities::getFilePathsWithExtension(const std::string& pathName, const std::string& extension, bool recursive) {
	vector<fs::path> filePaths;

	fs::path path = fs::system_complete(fs::path(pathName));

	if (fs::exists(path) && fs::is_directory(path)) {

		fs::directory_iterator end_iter;

		for (fs::directory_iterator dir_itr(path); dir_itr != end_iter; ++dir_itr) {
			try {
				string fileName = dir_itr->path().string();

				if (!fs::is_directory(*dir_itr)) {

					size_t pos = fileName.find_last_of('.') + 1;

					if (pos < fileName.length() && fileName.substr(pos, fileName.length() - pos) == extension) {
						filePaths.push_back(*dir_itr);
					}

				} else {
					if (recursive) {
						// recursive exploration
						vector<fs::path> directoryPaths = getFilePathsWithExtension(fileName, extension);
						for (vector<fs::path>::const_iterator p = directoryPaths.begin(); p != directoryPaths.end(); ++p) {
							filePaths.push_back(*p);
						}
					}
				}

			} catch (const std::exception & ex) {
				err << nameToString(dir_itr->path().filename()) << " " << ex.what() << eol;
			}
		}

	} else {
		throw NoSuchDirectoryException("cannot read " + pathName);
 	}

 	return filePaths;

}

std::string FileUtilities::nameToString(std::string const & name) {
	return name;
}

std::string FileUtilities::nameToString(boost::filesystem::path const & name) {
	return name.string();
}

std::string FileUtilities::getFileName(const fs::path& path1, const std::string& name, const std::string& defaultname,
		const std::string& extension)  {
	fs::path path2 = name;
	fs::path file;
	if (path2.is_relative() == true) {
		file = fs::system_complete(path1 / path2).normalize();
	}
	else if (path2.is_absolute() == true) {
		file = path2;
	}

	if (fs::is_directory(file) == true) {
		if (defaultname.empty()) {
			throw NotRegularFile("cannot read " + file.string());
		}
		file /= defaultname;
	}

	if (file.has_extension() == false) {
		file.replace_extension(extension);
	}

	if (fs::is_regular_file(file) == false) {
		throw NotRegularFile("cannot read " + file.string());
	}

	return file.string();
}

std::string FileUtilities::getDirectoryName(const fs::path& path1, const std::string& name)  {
	fs::path path2 = name;
	fs::path path3;
	if (path2.is_relative() == true) {
		path3 = fs::system_complete(path1 / path2).normalize();
	}
	else if (path2.is_absolute() == true) {
		path3 = path2;
	}

	if (fs::is_directory(path3) == false) {
		throw NoSuchDirectoryException("cannot read " + path3.string());
	}

	return path3.string();
}

time_t FileUtilities::getFileLastTime(const std::string& name)  {
	fs::path file(name);
	if (fs::is_regular_file(file) == false) {
		throw NotRegularFile("cannot read " + file.string());
	}
	return fs::last_write_time(file);
}

}
