/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_FUNCTION_H_
#define COMMON_FUNCTION_H_

#include <functional>
#include "ics/Servant.h"

namespace common {

///////////////////////////////////////////////////////////////////////////////
// Function with 0 argument

template<typename ReturnType>
struct BaseFunction {

	virtual ~BaseFunction() {}
	virtual BaseFunction<ReturnType> * clone() const = 0;
	virtual void setObject(Servant * obj) = 0;
	virtual ReturnType execute() = 0;
};

template<typename ReturnType, typename Class>
class Function : public BaseFunction<ReturnType> {

public:
	typedef std::function<ReturnType (Class *)> Type;

	Function(Type f, Class * obj = 0);

	virtual BaseFunction<ReturnType> * clone() const;
	virtual void setObject(Servant * obj);

	virtual ~Function() {}
	virtual ReturnType execute();

private:
	Type m_f;
	Class * m_obj;
};

template<typename ReturnType, typename Class>
Function<ReturnType, Class>::Function(Type f, Class * obj) :
	m_f(f), m_obj(obj) {
}

template<typename ReturnType, typename Class>
ReturnType Function<ReturnType, Class>::execute() {
	return m_f(m_obj);
}

template<typename ReturnType, typename Class>
BaseFunction<ReturnType> * Function<ReturnType, Class>::clone() const {
	return new Function<ReturnType, Class>(m_f, 0);
}

template<typename ReturnType, typename Class>
void Function<ReturnType, Class>::setObject(Servant * obj) {
	m_obj = dynamic_cast<Class *>(obj);
}

///////////////////////////////////////////////////////////////////////////////
// Function with 1 argument

template<typename ReturnType, typename Arg1>
struct BaseFunction1 {

	virtual ~BaseFunction1() {}
	virtual BaseFunction1<ReturnType, Arg1> * clone() const = 0;
	virtual void setObject(Servant * obj) = 0;
	virtual ReturnType execute(Arg1 arg1) = 0;
};

template<typename ReturnType, typename Arg1, typename Class>
class Function1 : public BaseFunction1<ReturnType, Arg1> {

public:
	typedef std::function<ReturnType (Class *, Arg1)> Type;

	Function1(Type f, Class * obj = 0);

	virtual BaseFunction1<ReturnType, Arg1> * clone() const;
	virtual void setObject(Servant * obj);

	virtual ~Function1() {}
	virtual ReturnType execute(Arg1 arg1);

private:
	Type m_f;
	Class * m_obj;
};

template<typename ReturnType, typename Arg1, typename Class>
Function1<ReturnType, Arg1, Class>::Function1(Type f, Class * obj) :
	m_f(f), m_obj(obj) {
}

template<typename ReturnType, typename Arg1, typename Class>
ReturnType Function1<ReturnType, Arg1, Class>::execute(Arg1 arg1) {
	return m_f(m_obj, arg1);
}

template<typename ReturnType, typename Arg1, typename Class>
BaseFunction1<ReturnType, Arg1> * Function1<ReturnType, Arg1, Class>::clone() const {
	return new Function1<ReturnType, Arg1, Class>(m_f, m_obj);
}

template<typename ReturnType, typename Arg1, typename Class>
void Function1<ReturnType, Arg1, Class>::setObject(Servant * obj) {
	m_obj = dynamic_cast<Class *>(obj);
}

///////////////////////////////////////////////////////////////////////////////
// Function with 2 arguments

template<typename ReturnType, typename Arg1, typename Arg2>
struct BaseFunction2 {

	virtual ~BaseFunction2() {}
	virtual BaseFunction2<ReturnType, Arg1, Arg2> * clone() const = 0;
	virtual void setObject(Servant * obj) = 0;
	virtual ReturnType execute(Arg1 arg1, Arg2 arg2) = 0;
};

template<typename ReturnType, typename Arg1, typename Arg2, typename Class>
class Function2 : public BaseFunction2<ReturnType, Arg1, Arg2> {

public:
	typedef std::function<ReturnType (Class *, Arg1, Arg2)> Type;

	Function2(Type f, Class * obj = 0);

	virtual BaseFunction2<ReturnType, Arg1, Arg2> * clone() const;
	virtual void setObject(Servant * obj);

	virtual ~Function2() {}
	virtual ReturnType execute(Arg1 arg1, Arg2 arg2);

private:
	Type m_f;
	Class * m_obj;
};

template<typename ReturnType, typename Arg1, typename Arg2, typename Class>
Function2<ReturnType, Arg1, Arg2, Class>::Function2(Type f, Class * obj) :
	m_f(f), m_obj(obj) {
}

template<typename ReturnType, typename Arg1, typename Arg2, typename Class>
ReturnType Function2<ReturnType, Arg1, Arg2, Class>::execute(Arg1 arg1, Arg2 arg2) {
	return m_f(m_obj, arg1, arg2);
}

template<typename ReturnType, typename Arg1, typename Arg2, typename Class>
BaseFunction2<ReturnType, Arg1, Arg2> * Function2<ReturnType, Arg1, Arg2, Class>::clone() const {
	return new Function2<ReturnType, Arg1, Arg2, Class>(m_f, m_obj);
}

template<typename ReturnType, typename Arg1, typename Arg2, typename Class>
void Function2<ReturnType, Arg1, Arg2, Class>::setObject(Servant * obj) {
	m_obj = dynamic_cast<Class *>(obj);
}

///////////////////////////////////////////////////////////////////////////////
// Function with 3 arguments

template<typename ReturnType, typename Arg1, typename Arg2, typename Arg3>
struct BaseFunction3 {

	virtual ~BaseFunction3() {}
	virtual BaseFunction3<ReturnType, Arg1, Arg2, Arg3> * clone() const = 0;
	virtual void setObject(Servant * obj) = 0;
	virtual ReturnType execute(Arg1 arg1, Arg2 arg2, Arg3 arg3) = 0;
};

template<typename ReturnType, typename Arg1, typename Arg2, typename Arg3, typename Class>
class Function3 : public BaseFunction3<ReturnType, Arg1, Arg2, Arg3> {

public:
	typedef std::function<ReturnType (Class *, Arg1, Arg2, Arg3)> Type;

	Function3(Type f, Class * obj = 0);

	virtual BaseFunction3<ReturnType, Arg1, Arg2, Arg3> * clone() const;
	virtual void setObject(Servant * obj);

	virtual ~Function3() {}
	virtual ReturnType execute(Arg1 arg1, Arg2 arg2, Arg3 arg3);

private:
	Type m_f;
	Class * m_obj;
};

template<typename ReturnType, typename Arg1, typename Arg2, typename Arg3, typename Class>
Function3<ReturnType, Arg1, Arg2, Arg3, Class>::Function3(Type f, Class * obj) :
	m_f(f), m_obj(obj) {
}

template<typename ReturnType, typename Arg1, typename Arg2, typename Arg3, typename Class>
ReturnType Function3<ReturnType, Arg1, Arg2, Arg3, Class>::execute(Arg1 arg1, Arg2 arg2, Arg3 arg3) {
	return m_f(m_obj, arg1, arg2, arg3);
}

template<typename ReturnType, typename Arg1, typename Arg2, typename Arg3, typename Class>
BaseFunction3<ReturnType, Arg1, Arg2, Arg3> * Function3<ReturnType, Arg1, Arg2, Arg3, Class>::clone() const {
	return new Function3<ReturnType, Arg1, Arg2, Arg3, Class>(m_f, m_obj);
}

template<typename ReturnType, typename Arg1, typename Arg2, typename Arg3, typename Class>
void Function3<ReturnType, Arg1, Arg2, Arg3, Class>::setObject(Servant * obj) {
	m_obj = dynamic_cast<Class *>(obj);
}

}

#endif
