/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_PORTABLEBYTEARRAY_H_
#define COMMON_PORTABLEBYTEARRAY_H_

#include "Endian.h"
#include <vector>

namespace common {

enum EncodedType {ERROR_TYPE, NULL_TYPE, FLOAT64_TYPE, INT64_TYPE, INT32_TYPE, FLOAT32_TYPE, UINT64_TYPE, UINT32_TYPE, UINT16_TYPE, INT16_TYPE};

// Specialized for supported types.
template<typename Type>
struct SerializeType;

template<>
struct SerializeType<double> {
	static const EncodedType value = FLOAT64_TYPE;
};

template<>
struct SerializeType<float> {
	static const EncodedType value = FLOAT32_TYPE;
};

template<>
struct SerializeType<int16_t> {
	static const EncodedType value = INT16_TYPE;
};

template<>
struct SerializeType<int32_t> {
	static const EncodedType value = INT32_TYPE;
};

template<>
struct SerializeType<int64_t> {
	static const EncodedType value = INT64_TYPE;
};

template<>
struct SerializeType<uint16_t> {
	static const EncodedType value = UINT16_TYPE;
};

template<>
struct SerializeType<uint32_t> {
	static const EncodedType value = UINT32_TYPE;
};

template<>
struct SerializeType<uint64_t> {
	static const EncodedType value = UINT64_TYPE;
};

struct BasePortableByteArray {

	static inline int16_t reverseInt16(int16_t value) {
		return (int16_t) ((value >> 8 & 0xff) |
						(value >> 0 & 0xff) << 8);

	}

	static inline uint16_t reverseUint16(uint16_t value) {
		return (uint16_t) ((value >> 8 & 0xff) |
						(value >> 0 & 0xff) << 8);

	}

	static inline int32_t reverseInt32(int32_t value) {
		return (int32_t) ((value >> 24 & 0xff) |
						(value >> 16 & 0xff) << 8 |
						(value >> 8 & 0xff) << 16 |
						(value >> 0 & 0xff) << 24);
	}

	static inline uint32_t reverseUint32(uint32_t value) {
		return (uint32_t) ((value >> 24 & 0xff) |
						(value >> 16 & 0xff) << 8 |
						(value >> 8 & 0xff) << 16 |
						(value >> 0 & 0xff) << 24);
	}

	static inline int64_t reverseInt64(int64_t value) {
		return (int64_t) ((value >> 56 & 0xff) |
				(value >> 48 & 0xff) << 8 |
				(value >> 40 & 0xff) << 16 |
				(value >> 32 & 0xff) << 24 |
				((int64_t)(value >> 24 & 0xff) << 0 |
				(value >> 16 & 0xff) << 8 |
				(value >> 8 & 0xff) << 16 |
				(value & 0xff) << 24) << 32 );
	}

	static inline uint64_t reverseUint64(uint64_t value) {
		return (uint64_t) ((value >> 56 & 0xff) |
				(value >> 48 & 0xff) << 8 |
				(value >> 40 & 0xff) << 16 |
				(value >> 32 & 0xff) << 24 |
				((uint64_t)(value >> 24 & 0xff) << 0 |
				(value >> 16 & 0xff) << 8 |
				(value >> 8 & 0xff) << 16 |
				(value & 0xff) << 24) << 32 );
	}

	/**
	 * Detect if we have to reverse bytes when we serialize.
	 */
	static bool needReverse(int endian) {
		bool reverse;
		if (((ENDIANNESS == LITTLE_ENDIAN_ORDER) && endian == 1) ||
				((ENDIANNESS == BIG_ENDIAN_ORDER) && endian != 1)) {
			reverse = false;
		}
		else {
			reverse = true;
		}

		return reverse;
	}
};

/**
 * Class providing a portable byte array to speed up protocol buffers in the transfer of arrays.
 */
template<typename Type>
class PortableByteArray : public BasePortableByteArray {

public:
	void encode(const Type * data, std::size_t size, std::string & encodedString, int outputEndian = LITTLE_ENDIAN_ORDER);
	void encodeError(EncodedType type, std::string & encodedString, int outputEndian = LITTLE_ENDIAN_ORDER);

	std::vector<Type> decode(const std::string & encodedString);
};

template<typename Type>
void PortableByteArray<Type>::encode(const Type * data, std::size_t size, std::string & encodedString, int outputEndian) {

	EncodedType type = SerializeType<Type>::value;
	int16_t byteSize = 0;
	switch(type) {
	case FLOAT64_TYPE:
	case INT64_TYPE:
	case UINT64_TYPE:
		byteSize = 8;
		break;
	case UINT16_TYPE:
	case INT16_TYPE:
		byteSize = 2;
		break;
	default:
		byteSize = 4;
	}

	// Reserve the string.
	encodedString.resize(size * byteSize + 8);

	// Get the allocated array.
	char * array = encodedString.data();

	// Check reverse.
	if (needReverse(outputEndian)) {
		// We have to serialize with reversing bytes.
		((int16_t *)array)[0] = reverseInt16(1);
		((int16_t *)array)[1] = reverseInt16(type);
		((int32_t *)array)[1] = reverseInt32(size);

		if (type == FLOAT32_TYPE || type == INT32_TYPE) {
			for (int i = 0; i < size; i++) {
				int32_t tmp = reverseInt32((((int32_t *)data)[i]));

				// The real array start at byte 8.
				memcpy(&((int32_t *)array)[i + 2], &tmp, sizeof(int32_t));
			}
		}
		else if (type == FLOAT64_TYPE || type == INT64_TYPE) {
			for (int i = 0; i < size; i++) {
				int64_t tmp = reverseInt64((((int64_t *)data)[i]));

				// The real array start at byte 8.
				memcpy(&((int64_t *)array)[i + 1], &tmp, sizeof(int64_t));
			}
		}

	}
	else {
		((int16_t *)array)[0] = 1;
		((int16_t *)array)[1] = type;
		((int32_t *)array)[1] = size;
		memcpy((int8_t *)array + 8, data, size * byteSize);
	}
}

template<typename Type>
void PortableByteArray<Type>::encodeError(EncodedType type, std::string & encodedString, int outputEndian) {

	// Reserve the string.
	encodedString.resize(8);

	// Get the allocated array.
	char * array = encodedString.data();

	// Check reverse.
	if (needReverse(outputEndian)) {
		// We have to serialize with reversing bytes.
		((int16_t *)array)[0] = reverseInt16(1);
		((int16_t *)array)[1] = reverseInt16(type);
		((int32_t *)array)[1] = reverseInt32(0);
	}
	else {
		((int16_t *)array)[0] = 1;
		((int16_t *)array)[1] = type;
	}
}

template<typename Type>
std::vector<Type> PortableByteArray<Type>::decode(const std::string & encodedString) {

	std::vector<Type> array;

	// Get the allocated array.
	const char * data = encodedString.data();

	// The value "1" means that we have the same endianness.
	if (((int16_t *)data)[0] == 1) {
		EncodedType type = (EncodedType)(((int16_t *)data)[1]);

		if (SerializeType<Type>::value != type) {
			std::cerr << "Problem with type" << std::endl;
			return array;
		}

		std::size_t size = ((int32_t *)data)[1];

		int16_t byteSize = 0;
		switch(type) {
		case FLOAT64_TYPE:
		case INT64_TYPE:
		case UINT64_TYPE:
			byteSize = 8;
			break;
		case UINT16_TYPE:
		case INT16_TYPE:
			byteSize = 2;
			break;
		default:
			byteSize = 4;
		}
		array.resize(size);

		memcpy(array.data(), (int8_t *)data + 8, size * byteSize);

	// We have to reverse bytes.
	}
	else {
		EncodedType type = (EncodedType)reverseInt16(((int16_t *)data)[1]);

		if (SerializeType<Type>::value != type) {
			std::cerr << "Problem with type" << std::endl;
			return array;
		}

		std::size_t size = reverseInt32(((int32_t *)data)[1]);

		if (type == FLOAT32_TYPE || type == INT32_TYPE || type == UINT32_TYPE) {
			array.resize(size);
			Type * arrayData = array.data();
			for (int i = 0; i < size; i++) {
				int32_t tmp = reverseInt32(((int32_t *)data)[i + 2]);
				memcpy(&((int32_t *)arrayData)[i], &tmp, sizeof(int32_t));
			}
		}
		else if (type == FLOAT64_TYPE || type == INT64_TYPE || type == UINT64_TYPE) {
			array.resize(size);
			Type * arrayData = array.data();
			for (int i = 0; i < size; i++) {
				int64_t tmp = reverseInt64(((int64_t *)data)[i+1]);
				memcpy(&((int64_t *)arrayData)[i], &tmp, sizeof(int64_t));
			}
		}
		else if (type == UINT16_TYPE || type == INT16_TYPE) {
			array.resize(size);
			Type * arrayData = array.data();
			for (int i = 0; i < size; i++) {
				int16_t tmp = reverseInt16(((int16_t *)data)[i]);
				memcpy(&((int16_t *)arrayData)[i], &tmp, sizeof(int16_t));
			}
		}
	}

	return array;
}

}

#endif
