/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_QUEUE_H_
#define COMMON_QUEUE_H_

#include <common/base/Exception.h>
#include <deque>
#include <mutex>
#include <condition_variable>

namespace common {

/**
 * Template class defining a parallel queue. The access is sequentialised to avoid concurrent access problems. We put or get elements from
 * the queue. The queue is initialised with a size. The call to put is blocking when the queue is full. The call to get is also blocking when 
 * the queue is empty. The queue can be stopped. Any call to put or get lead to the UnavalaibleException.  
 */
template<typename ItemType>
class Queue {
	
public:
	/**
	 * Exception class for unavailable queue.
	 */
	class UnavailableException : public Exception {
	
	public:
		UnavailableException();
	};


	/**
	 * Constructor.
	 */
	Queue(long size);
	
	/**
	 * Virtual destructor. Deletes all the items.
	 */
	virtual ~Queue();

	/**
	 * Puts the item in the correct postion corresponding to its priority. Is blocking when the queue is full. Waits for the queue to be not full. 
	 * throw UnavailableException.
	 */
	void put(ItemType* item, int priority = 0);

	/**
	 * Gets the item. Is blocking when the queue is empty. Waits for the queue to be not empty.
	 * throw UnavailableException.
	 */
	ItemType* get();

	/**
	 * Gets the size.
	 */
	long getNumberOfItems();
	
	/**
	 * Stops the queue. Becomes unavailable.
	 */
	void stop();

private:

	struct Item {
		ItemType* content;
		int priority;
	};

	std::deque<Item> _queue;
	std::mutex _queueMutex;
	std::condition_variable _waitForAvailableSpace;
	std::condition_variable _waitForItemDelivery;
	long _size;
	bool _unavailable;
};

}

#include "Queue.inl"

#endif /*COMMON_QUEUE_H_*/
