/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <string>
#include <iostream>

#include <common/base/Convert.h>

namespace common {
	
template<typename ItemType>
Queue<ItemType>::UnavailableException::UnavailableException()
: Exception("common::Queue<ItemType>::UnavailableException", std::string("queue unavailable for '") + typeid(ItemType).name()  + "'") {
}	
	
template<typename ItemType> 
Queue<ItemType>::Queue(long size) :
	_size(size), _unavailable(false) {
}

template<typename ItemType> 
Queue<ItemType>::~Queue() {
	
	std::unique_lock<std::mutex> lock(_queueMutex);

	for (typename std::deque<Item>::iterator i = _queue.begin(); i != _queue.end(); i++) {
		delete i->content;
	}
}


template<typename ItemType> 
void Queue<ItemType>::put(ItemType* item, int priority) {
	
	std::unique_lock<std::mutex> lock(_queueMutex);
	  
	// waiting until the  condition is fulfilled
	while ((long)_queue.size() == _size && !_unavailable) {
		_waitForAvailableSpace.wait(lock);
	}
	
	// Throwing exception if stopped
	if (_unavailable) {
		throw UnavailableException();
	}
	
	// item to insert
	Item newItem;
	newItem.content = item;
	newItem.priority = priority;
			
	// Searching for the correct position
	typename std::deque<Item>::iterator i;
	for (i = _queue.begin(); i != _queue.end(); i++) {

		if (priority < i->priority) {
			_queue.insert(i, newItem);
			break;
		}
	}
	
	// A greater priority has not been found in the queue, inserting at the end.
	if (i == _queue.end()) {
		_queue.push_back(newItem);
	}
			
	_waitForItemDelivery.notify_one();
}

template<typename ItemType> 
ItemType* Queue<ItemType>::get() {
	
	std::unique_lock<std::mutex> lock(_queueMutex);
	
	// waiting until the wait for item delivery condition is fulfilled
	while (_queue.empty() && !_unavailable) {
		_waitForItemDelivery.wait(lock);
	}
	
	// Throwing exception if stopped
	if (_unavailable) {
		throw UnavailableException();
	}

	// getting the first element
	Item item = _queue.front();
	_queue.pop_front();
	_waitForAvailableSpace.notify_one();
  
	return item.content;
	
}

template<typename ItemType>
long Queue<ItemType>::getNumberOfItems() {
	
	std::unique_lock<std::mutex> lock(_queueMutex);
	
	return _queue.size();
}

template<typename ItemType>
void Queue<ItemType>::stop() {
	
	std::unique_lock<std::mutex> lock(_queueMutex);
	_unavailable = true;
	
	_waitForItemDelivery.notify_all();	
	_waitForAvailableSpace.notify_all();
}

}