/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_REALARRAY_H_
#define COMMON_REALARRAY_H_

#include "Array.h"
#include "BaseTypes.h"
#include <vector>

namespace common {

/**
 * Template class representing an array of values. The array has a fixed size.
 * A value of -1 defines an unfixed size array. Positive values define a fixed size array.
 * The implementation is an encapsulation of a C array with size functions.
 */
template<typename ValueType>
class RealArray : public Array<ValueType> {

	template<typename>
	friend class ClonedArray;

public:
	/**
	 * Default constructor.
	 */
	RealArray();

	/**
	 * Constructor with fixed size parameter.
	 */
	RealArray(long fixedSize);

	/**
	 * Constructor from an std::vector.
	 */
	RealArray(const std::vector<ValueType>& values);

	/**
	 * Constructor from an array and size.
	 */
	RealArray(ValueType * values, long size);

	/**
	 * Copy constructor.
	 */
	RealArray(const RealArray<ValueType>& array);

	/**
	 * Virtual destructor.
	 */
	virtual ~RealArray();

	/**
	 * Assignment operator.
	 */
	RealArray<ValueType>& operator= (const RealArray<ValueType>& array);

	/**
	 * Resizes the array. The array may not be resizable.
	 * throw BaseArray::BadSizeException, BaseArray::NotResizableException
	 */
	virtual void resize(long size);

	/**
	 * Returns the size.
	 */
	virtual long getSize() const;

	/**
	 * Returns the value at index.
	 */
	virtual const ValueType& operator[](long index) const;

	/**
	 * Returns the value at index.
	 */
	virtual ValueType& operator[](long index);

	/**
	 * Sets the array resizable.
	 */
	virtual void setResizable(bool resizable);

	/**
	 * Returns true if the array is resizable.
	 */
	virtual bool isResizable() const;

	/**
	 * Gets the data pointer.
	 */
	ValueType* getData() const;

	/**
	 * Gets the size. For generated code only.
	 */
	virtual long _size() const;

	/**
	 * Copies the array.
	 */
	virtual Array<ValueType> * copy() const;

private:
	long _fixedSize;
	long _realSize;
	ValueType* _data;
};

}

#include "RealArray.inl"

#endif /*COMMON_REALARRAY_H_*/
