/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <common/output/OutputStream.h>

namespace common {

template<typename ValueType>
RealArray<ValueType>::RealArray() :
	_fixedSize(-1), _realSize(0), _data(0) {
}

template<typename ValueType>
RealArray<ValueType>::RealArray(long fixedSize) :
	_fixedSize(fixedSize), _realSize(0), _data(0) {
	
	if (fixedSize > 0) {
		_realSize = fixedSize;
		_data = new ValueType[fixedSize];
	}	
				
	for (long i = 0; i < fixedSize; i++) {
		_data[i] = ValueType();
	}
}

template<typename ValueType>
RealArray<ValueType>::RealArray(const std::vector<ValueType>& values) :
	_fixedSize((long)values.size()),
	_realSize((long)values.size()),
	_data(new ValueType[_realSize]) {

	for (long i = 0; i < _realSize; i++) {
		_data[i] = values[i];
	}
}

template<typename ValueType>
RealArray<ValueType>::RealArray(ValueType * values, long size) :
	_fixedSize(size),
	_realSize(size),
	_data(new ValueType[_realSize]) {

	for (long i = 0; i < _realSize; i++) {
		_data[i] = values[i];
	}
}

template<typename ValueType>
RealArray<ValueType>::RealArray(const RealArray<ValueType>& array) :
	_fixedSize(array._fixedSize),
	_realSize(array._realSize),
	_data(new ValueType[array._realSize]) {

	for (long i = 0; i < _realSize; i++) {
		_data[i] = array._data[i];
	}
}

template<typename ValueType>
RealArray<ValueType>::~RealArray() {
	
	delete [] _data;
}

template<typename ValueType>
RealArray<ValueType>& RealArray<ValueType>::operator= (const RealArray<ValueType>& array) {
	
	_fixedSize = array._fixedSize;
	_realSize = array._realSize;
	_data = new ValueType[array._realSize];

	for (long i = 0; i < _realSize; i++) {
		_data[i] = array._data[i];
	}
	
	return *this;
}

template<typename ValueType>
void RealArray<ValueType>::resize(long size) {
	
	if (_realSize != size) {
		
		if (size < 0) {
			throw BaseArray::BadSizeException(size);
		}
		
		if (_fixedSize == -1) {
			
			ValueType* newData = 0;
				
			try {	
				newData = new ValueType[size];
			} catch (std::exception const & e) {
				err << "Array : allocation exception (" << e.what() << ")" << eol;
				throw BaseArray::NotResizableException(size, _fixedSize);
			} catch (...) {
				err << "Array : allocation exception (?)" << eol;
				throw BaseArray::NotResizableException(size, _fixedSize);
			} 
			
			for (long i = 0; i < std::min(_realSize, size); i++) {
				newData[i] = _data[i];
			}
			
			for (long i = _realSize; i < size; i++) {
				newData[i] = ValueType();
			}
			
			delete [] _data;
			_data = newData;
			_realSize = size;
			
		} else {
			throw BaseArray::NotResizableException(size, _fixedSize);
		}
	}
}

template<typename ValueType>
long RealArray<ValueType>::getSize() const {
	return _realSize;
}

template<typename ValueType>
ValueType* RealArray<ValueType>::getData() const {
	return _data;
}

template<typename ValueType>
const ValueType& RealArray<ValueType>::operator[](long index) const {
	
	if (index < 0 || index >= _realSize) {
		throw BaseArray::IndexOutOfBoundsException(index, _realSize);
	}
	
	return _data[index];
}

template<typename ValueType>
ValueType& RealArray<ValueType>::operator[](long index) {
	
	if (index < 0 || index >= _realSize) {
		throw BaseArray::IndexOutOfBoundsException(index, _realSize);
	}
	
	return _data[index];
}

template<typename ValueType>
void RealArray<ValueType>::setResizable(bool resizable) {
	if (resizable) {
		_fixedSize = -1;
	} else {
		_fixedSize = _realSize;
	}
}

template<typename ValueType>
bool RealArray<ValueType>::isResizable() const {
	return (_fixedSize == -1);
}

template<typename ValueType>
long RealArray<ValueType>::_size() const {
	return getSize();
}


template<typename ValueType>
Array<ValueType> * RealArray<ValueType>::copy() const {
	return new RealArray<ValueType>();
}

}
