/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ServerProperties.h"
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/lexical_cast.hpp>
#include <common/output/OutputStream.h>
#include <common/base/FileUtilities.h>
#include <common/base/Convert.h>
#include "Endian.h"
#include <iostream>
#include <fstream>
#include <cstdlib>

using namespace std;
namespace fs = boost::filesystem;

namespace common {

ServerProperties * ServerProperties::_instance = 0;

ServerProperties::NoSuchDirectoryException::NoSuchDirectoryException(std::string const & text) :
		Exception("common::ServerProperties::NoSuchDirectoryException", text) {
}

ServerProperties::NoSuchPropertyException::NoSuchPropertyException(std::string const & text) :
		Exception("common::ServerProperties::NoSuchPropertyException", text) {
}

ServerProperties * ServerProperties::getInstance() {

	return _instance;
}

void ServerProperties::initInstance(std::string const & configPath) {
	if (_instance == 0) {
		_instance = new ServerProperties();
		_instance->init(configPath);
	}
}

bool ServerProperties::readFile(std::string const & fileName) {

	// Opening the file
	ifstream file(fileName.c_str());

	if (file.is_open()) {
		while (!file.eof()) {
			string line;
			getline(file, line);

			// Analysing the line
			string::size_type equalPosition = line.find_first_of('=');

			if (equalPosition < line.size()) {

				string propertyName = line.substr(0, equalPosition);
				string propertyValue = line.substr(equalPosition + 1, line.size() - equalPosition - 1);

				if (propertyValue.length() > 1 && propertyValue[0] == '$') {
					if (propertyValue.length() > 2 && propertyValue[1] == '{') {
						string::size_type second = propertyValue.find_first_of('}');
						string variable = propertyValue.substr(2, second - 2);
						char * var = getenv(variable.c_str());
						propertyValue = string(var) + propertyValue.substr(second + 1, propertyValue.size() - second - 1);
					}
					else {
						char * var = getenv(propertyValue.substr(1).c_str());
						if (var != 0) {
							propertyValue = string(var);
						}
					}
				}

				// Adding the item
				_propertyMap.insert(propertyName, propertyValue);
			}
		}

		file.close();

		return true;
	}

	return false;
}

void ServerProperties::setConfigPath(std::string const & configPath) {

	// Check default value.
	if (configPath.empty() == true) {
		_configPath = _homePath + "/.nomad";
	}
	else {
		_configPath = _homePath + "/" + configPath;
	}

	string propertyPath = _configPath;

	string extension = ".properties";
	fs::path configPropertiesPath = fs::system_complete(fs::path(propertyPath));

	if (fs::exists(configPropertiesPath)) {
		if (fs::is_directory(configPropertiesPath)) {

			_propertyPath = propertyPath;

			fs::directory_iterator end_iter;

			for (fs::directory_iterator dir_itr(configPropertiesPath); dir_itr != end_iter; ++dir_itr) {
				try {
					if (!fs::is_directory(*dir_itr)) {

						string fileName = dir_itr->path().string();
						size_t pos = fileName.find_last_of('.');
						if (fileName.find("Settings.properties") == string::npos) { // Already did , and could be conflict with local one
							if (pos < fileName.length() && fileName.substr(pos, fileName.length() - pos) == extension) {
								readFile(fileName);
							}
						}
					}
				}
				catch (const std::exception & e) {
					cerr << FileUtilities::nameToString(dir_itr->path().filename()) << " " << e.what() << endl;
				}
			}

		}
		else {
			cerr << "Config path " << propertyPath << " is not a directory" << endl;
			return;
		}
	}
	else {
		throw NoSuchDirectoryException("cannot read \"" + propertyPath + "\"");
	}
}

std::string const & ServerProperties::getHostname() const {
	return _hostname;
}

std::string const & ServerProperties::getConfigPath() const {
	return _configPath;
}

std::string const & ServerProperties::getHomePath() const {
	return _homePath;
}

std::string const & ServerProperties::getNomadHomePath() const {
	return _nomadHomePath;
}

std::string const & ServerProperties::getNomadGUIHomePath() const {
	return _nomadGUIHomePath;
}

std::string const & ServerProperties::getNomadUser() const {
	return _nomadUser;
}

std::string const & ServerProperties::getNomadSenddata() const {
	return _nomadSendata;
}

std::string const & ServerProperties::getMockMode() const {
	return _mockMode;
}


std::string const & ServerProperties::getNomadNotifications() const {
	return _nomadNotifications;
}

std::string const & ServerProperties::getNomadUserNotifications() const {
	return _nomadUserNotifications;
}

std::string const & ServerProperties::getNomadDataPath() const {
	return _nomadDataPath;
}

std::string ServerProperties::getNomadProcessedDataPath() {

	// Create the processed directory if it does not exist.
	fs::path processedPath(_homePath);
	processedPath /= "processed";

	fs::create_directory(processedPath);

	return processedPath.string();
}

void ServerProperties::init(std::string const & configPath) {

	// Test for endianness.
	if (ENDIANNESS == BIG_ENDIAN_ORDER) {
		err << "Big endian platform not supported" << eol;
		exit(EXIT_FAILURE);
	}

	const char* home = getenv("HOME");
	if (home != 0) {
		_homePath = string(home);
	}
	else {
		warn << "$HOME is not set" << eol;
	}

	// Hostname.
	char hostname[1024];
	::gethostname(hostname, 1024);

	_hostname = string(hostname);

	// Set the config path.
	setConfigPath(configPath);

	// Trying to read .nomad/settings.properties.
	const char* nomadsettingsconfig = getenv("NOMAD_SETTINGS_CONFIG");
	string settingsFileName;
	if (nomadsettingsconfig != 0) {
		settingsFileName = string(nomadsettingsconfig);
	}
	else {
		settingsFileName = _configPath + "/Settings.properties";
	}
	bool exists = readFile(settingsFileName);
	if (!exists) {
		warn << "File " << settingsFileName << " should exist" << eol;
	}

	// Nomad home
	string propertyName = "nomadHome";
	if (_propertyMap.find(propertyName)) {
		_nomadHomePath = _propertyMap.get(propertyName);
	}
	else {
		home = getenv("NOMAD_HOME");
		if (home != 0) {
			_nomadHomePath = string(home);
			err << "Use of $NOMAD_HOME is deprecated and the property '" << propertyName << "' should be defined in " << settingsFileName << eol;
		}
		else {
			_nomadHomePath = _homePath + "/nomad/nomad-server";
			warn << "Set the default value " <<_nomadHomePath << " to the property '" << propertyName << "'" << eol;
		}
	}

	// Nomad GUI home.
	propertyName = "nomadGUIHome";
	if (_propertyMap.find(propertyName)) {
		_nomadGUIHomePath = _propertyMap.get(propertyName);
	}
	else {
		home = getenv("NOMADCLIENT_HOME");
		if (home != 0) {
			_nomadGUIHomePath = string(home);
			err << "Use of $NOMADCLIENT_HOME is deprecated and the property '" << propertyName << "' should be defined in " << settingsFileName << eol;
		}
		else {
			_nomadGUIHomePath = _homePath + "/nomad/nomad-gui";
			warn << "Set the default value " <<_nomadGUIHomePath << " to the property '" << propertyName << "'" << eol;
		}
	}

	// Nomad notifications.
	propertyName = "nomadNotifications";
	if (_propertyMap.find(propertyName)) {
		_nomadNotifications = _propertyMap.get(propertyName);
	}
	else {
		home = getenv("NOMAD_NOTIFICATIONS");
		if (home != 0) {
			_nomadNotifications = string(home);
			err << "Use of $NOMAD_NOTIFICATIONS is deprecated and the property '" << propertyName << "' should be defined in " << settingsFileName << eol;
		}
		else {
			_nomadNotifications = "1";
			warn << "Set the default value " <<_nomadNotifications << " to the property '" << propertyName << "'" << eol;
		}
	}

	// Nomad notifications.
	propertyName = "nomadUserNotifications";
	if (_propertyMap.find(propertyName)) {
		_nomadUserNotifications = _propertyMap.get(propertyName);
	}
	else {
		home = getenv("NOMAD_USER_NOTIFICATIONS");
		if (home != 0) {
			_nomadUserNotifications = string(home);
			warn << "Use of $NOMAD_USER_NOTIFICATIONS is deprecated and the property '" << propertyName << "' should be defined in " << settingsFileName << eol;
		}
		else {
			_nomadUserNotifications = "1";
			warn << "Set the default value " <<_nomadUserNotifications << " to the property '" << propertyName << "'" << eol;
		}
	}

	// Nomad senddata.
	propertyName = "nomadSenddata";
	if (_propertyMap.find(propertyName)) {
		_nomadSendata = _propertyMap.get(propertyName);
	}
	else {
		home = getenv("NOMAD_SENDDATA");
		if (home != 0) {
			_nomadSendata = string(home);
			err << "Use of NOMAD_SENDDATA is deprecated and the property '" << propertyName << "' should be defined in " << settingsFileName << eol;
		}
		else {
			_nomadSendata = "-1";
			warn << "Set the default value " <<_nomadSendata << " to the property '" << propertyName << "'" << eol;
		}
	}

	// Mock mode.
	propertyName = "mockMode";
	if (_propertyMap.find(propertyName)) {
		_mockMode = _propertyMap.get(propertyName);
	} else {
		_mockMode = "false";
	}

	// Nomad data path.
	propertyName = "nomadDataPath";
	if (_propertyMap.find(propertyName)) {
		_nomadDataPath = _propertyMap.get(propertyName);
	}
	else {
		_nomadDataPath = "/users";
		warn << "Set the default value " <<_nomadDataPath << " to the property '" << propertyName << "'" << eol;
	}
}

long ServerProperties::getNumberOfProperties() const {

	return _propertyMap.getSize();
}

void ServerProperties::clear() {

	_propertyMap.clear();
}

std::string const & ServerProperties::getString(std::string const & propertyName) {

	try {
		return _propertyMap.get(propertyName);
	}
	catch (HashMap<std::string, std::string>::NoSuchKeyException const & e) {
		throw link(NoSuchPropertyException(string("unknown property \"") + propertyName + "\" in " + _propertyPath), e);
	}
}

void ServerProperties::setString(std::string const & propertyName, std::string const & value) {

	if (_propertyMap.find(propertyName)) {
		_propertyMap[propertyName] = value;
		return;
	}

	throw NoSuchPropertyException(string("unknown property \"") + propertyName + "\" in " + _propertyPath);
}

void ServerProperties::addString(std::string const & propertyName, std::string const & value) {

	if (!_propertyMap.find(propertyName)) {
		_propertyMap.insert(propertyName, value);
	}
}

void ServerProperties::setNomadUser(std::string const & username) {
	_nomadUser = username;
}

}
