/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_SERVERPROPERTIES_
#define COMMON_SERVERPROPERTIES_

#include <string>
#include <common/base/HashMap.h>
#include <common/base/Convert.h>

namespace common {

/**
 * Class defining the server properties read from a file. Singleton class.
 */
class ServerProperties {
	
public:
	/**
	 * Exception class for unaccessible file.
	 */
	class NoSuchDirectoryException : public Exception {
	
	public:
		/**
		 * Constructor.
		 */
		NoSuchDirectoryException(std::string const & fileName);
	};

	/**
	 * Exception class for unaccessible property.
	 */
	class NoSuchPropertyException : public Exception {
	
	public:
		/**
		 * Constructor.
		 */
		NoSuchPropertyException(std::string const & propertyName);
	};

	/**
	 * Gets the instance. the method initInstance must have been called before.
	 */
	static ServerProperties * getInstance();
	
	
	/**
	 * Creates the instance.
	 */
	static void initInstance(std::string const & configPath);
	
	/**
	 * Gets the hostname.
	 */
	std::string const & getHostname() const;

	/**
	 * Gets the config path.
	 */
	std::string const & getConfigPath() const;
	
	/**
	 * Gets the home path.
	 */
	std::string const & getHomePath() const;
	
	/**
	 * Gets the Nomad home path.
	 */
	std::string const & getNomadHomePath() const;

	/**
	 * Gets the Nomad GUI home path.
	 */
	std::string const & getNomadGUIHomePath() const;

	/**
	 * Gets the Nomad user.
	 */
	std::string const & getNomadUser() const;

	/**
	 * Gets the Nomad sendata.
	 */
	std::string const & getNomadSenddata() const;

	/**
	 * Gets the mock mode.
	 */
	std::string const & getMockMode() const;

	/**
	 * Gets the Nomad notifications.
	 */
	std::string const & getNomadNotifications() const;

	/**
	 * Gets the Nomad user notifications.
	 */
	std::string const & getNomadUserNotifications() const;

	/**
	 * Gets the Nomad data path.
	 */
	std::string const & getNomadDataPath() const;

	/**
	 * Gets the Nomad processed data path.
	 * Create the directory if necessary.
	 */
	std::string getNomadProcessedDataPath();

	/**
	 * Gets the number of properties.
	 */
	long getNumberOfProperties() const;
	
	/**
	 * Clears the content of the properties.
	 */
	void clear();
	
	/**
	 * Gets the Nomad user.
	 */
	void setNomadUser(std::string const & username);

	/**
	 * Gets the property named propertyName.
	 * throw NoSuchPropertyException
	 */
	std::string const & getString(std::string const & propertyName);
	
	/**
	 * Gets the property named propertyName and returns a typed value.
	 * throw NoSuchPropertyException
	 */
	template<typename Type>
	Type get(std::string const & propertyName) {

		return to<Type, std::string>(getString(propertyName));
	}
	
	/**
	 * Sets the property named propertyName.
	 * throw NoSuchPropertyException
	 */
	void setString(std::string const & propertyName, std::string const & value);
	
	/**
	 * Adds the property named propertyName with value.
	 */
	void addString(std::string const & propertyName, std::string const & value);

	/**
	 * Sets the property named propertyName and returns a typed value.
	 * throw NoSuchPropertyException
	 */
	template<typename Type>
	void set(std::string const & propertyName, Type value) {

		setString(propertyName, to<std::string, Type>(value));
	}

	/**
	 * Adds the property named propertyName.
	 */
	template<typename Type>
	void add(std::string const & propertyName) {

		addString(propertyName, to<std::string, Type>(Type()));
	}
	
private:
	/**
	 * Initialises the properties from the file called fileName.
	 * throw NoSuchDirectoryException
	 */
	void init(std::string const & configPath);

	/**
	 * Sets the config path.
	 */
	void setConfigPath(std::string const & configPath);
	
	/**
	 * Reads the file and gets the properties.
	 */
	bool readFile(std::string const & fileName);

	static ServerProperties * _instance;

	std::string _hostname;
	std::string _configPath;
	std::string _homePath;
	std::string _nomadHomePath;
	std::string _nomadGUIHomePath;
	std::string _nomadUser;
	std::string _nomadSendata;
	std::string _mockMode;
	std::string _nomadNotifications;
	std::string _nomadUserNotifications;
	std::string _nomadDataPath;
	std::string _propertyPath;
	HashMap<std::string, std::string> _propertyMap;

};

}

#endif /*COMMON_SERVERPROPERTIES_*/
