/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_STANDARDHASHMAP_H_
#define COMMON_STANDARDHASHMAP_H_

#include <common/base/BaseNoSuchKeyException.h>
#include <common/base/Convert.h>
#include <tr1/unordered_map>
#include <vector>

namespace common {

/**
 * Facade class for hiding the non-standard implementation of the hash map.
 */
template<typename Key, typename Data>
class HashMap {

	/**
	 * Implementation type
	 */
	typedef typename std::tr1::unordered_map<Key, Data> ImplementationType;

public:
	/**
	 * Exception.
	 */
	class NoSuchKeyException : public BaseNoSuchKeyException {

	public:
		/**
		 * Constructor.
		 */
		NoSuchKeyException(Key const & key);
	};

	/**
	 * Constructor.
	 */
	inline HashMap() {};

	/**
	 * Typedef for STL interface ease of use.
	 */
	typedef typename ImplementationType::const_iterator const_iterator;
	typedef typename ImplementationType::iterator iterator;

	/**
	 * Returns the const iterator pointing to the first element.
	 */
	inline const_iterator begin() const {
		return _impl.begin();
	}

	/**
	 * Returns the iterator pointing to the first element.
	 */
	inline iterator begin() {
		return _impl.begin();
	}

	/**
	 * Returns the const iterator pointing to the end of the map.
	 */
	inline const_iterator end() const {
		return _impl.end();
	}

	/**
	 * Returns the iterator pointing to the end of the map.
	 */
	inline iterator end() {
		return _impl.end();
	}


	/**
	 * Inserts the data item for the key key. The old value is replaced.
	 */
	inline void insert(Key const & key, Data const & data) {
		// call to the operator[], because insert doesn't replace the old value
		_impl[key] = data;
	}

	/**
	 * Gets the data item reference for the key key.
	 * throw NoSuchKeyException
	 */
	inline Data& operator [] (Key const & key) {
		return get(key);
	}

	/**
	 * Gets the const data item reference for the key key.
	 * throw NoSuchKeyException
	 */
	inline const Data& operator [] (Key const & key) const {
		return get(key);
	}

	/**
	 * Erases the key and its data.
	 * throw NoSuchKeyException
	 */
	void erase(Key const & key);

	/**
	 * Erases the iterator.
	 */
	inline void erase(iterator & it) {
		_impl.erase(it);
	}

	/**
	 * Returns true if key is found.
	 */
	inline bool find(Key const & key) const {
		return (_impl.find(key) != _impl.end());
	}

	/**
	 * Gets the item reference.
	 * throw NoSuchKeyException
	 */
	Data & get(Key const & key);

	/**
	 * Gets the const item reference.
	 * throw NoSuchKeyException
	 */
	Data const & get(Key const & key) const;

	/**
	 * Gets the size of the map.
	 */
	unsigned long getSize() const;

	/**
	 * Gets the vector of data. The data are copied from the map.
	 */
	std::vector<std::pair<Key, Data> > getData() const;

	/**
	 * Clears the map.
	 */
	void clear();

private:
	ImplementationType _impl;
};

}

#include "StandardHashMap.inl"

#endif /*COMMON_STANDARDHASHMAP_H_*/