/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_THREADSPECIFICPOINTER_H_
#define COMMON_THREADSPECIFICPOINTER_H_

#include "HashMap.h"
#include <mutex>

namespace common {

/**
 * Template class used to store specific data per thread. We used the boost::thread_specific_pointer but valgrind detected a memory leak
 * so we coded a simple version.
 */
template <typename ValueType>
class ThreadSpecificPointer {
	
public:
	/**
	 * Constructor.
	 */
	ThreadSpecificPointer();
	
	/**
	 * Virtual destructor.
	 */
	~ThreadSpecificPointer();
	
	/**
	 * Sets the value.
	 */
	void set(const ValueType& value);
	
	/**
	 * Gets the value.
	 */
	const ValueType& get() const;

	/**
	 * Releases the associated data.
	 */
	void release();

	/**
	 * Returns true if the data value has not been set for the thread.
	 */
	bool isNull() const;

private:
	static std::mutex _mutex;

	static HashMap<unsigned long, ValueType> _data;

};

}

#include "ThreadSpecificPointer.inl"

#endif /*COMMON_THREADSPECIFICPOINTER_H_*/
