/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_BASEHTTPCONNECTION_H_
#define COMMON_BASEHTTPCONNECTION_H_

#include <common/base/BaseTypes.h>
#include <common/base/Exception.h>
#include <common/base/Duration.h>
#include <boost/asio.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/date_time/posix_time/posix_time_types.hpp>
#include <vector>
#include <unordered_map>

namespace common {

/**
 * Exception class for connection errors.
 */
class ConnectionException: public common::Exception {

public:
	/**
	 * Constructor.
	 */
	ConnectionException(std::string const & message);
};

/**
 * Exception class for connection errors.
 */
class RequestException: public common::Exception {

public:
	/**
	 * Constructor.
	 */
	RequestException(std::string const & message, uint32 statusCode = 0);

	/**
	 * Returns the status code of the error.
	 */
	uint32 getStatusCode() const;

private:
	uint32 m_statusCode;
};

/**
 * Class managing http connections.
 */
class BaseHttpConnection {

public:
	/**
	 * Constructor.
	 */
	BaseHttpConnection();

	/**
	 * Destructor. Closes the connection.
	 */
	virtual ~BaseHttpConnection();

	/**
	 * Sets the timeout.
	 */
	void setTimeout(const Duration& timeout);

	/**
	 * Opens the connection with url and port.
	 */
	void openUrl(const std::string& url, const std::string& port);

	/**
	 * Sets the method POST or GET.
	 */
	void setMethod(const std::string& type);

	/**
	 * Sets the path.
	 * For instance with http://www.example.com:8080/resources/image
	 * url = www.example.com, port = 8080, path = /resources/image
	 */
	virtual void setPath(const std::string& path);

	/**
	 * Sets the property.
	 * For example User-Agent.
	 */
	void setProperty(const std::string& propertyName, const std::string& value);

	/**
	 * Sets the data.
	 */
	void setData(const std::string& data);

	/**
	 * Typedef for the list of properties.
	 */
	typedef std::unordered_map<std::string, std::string> properties_type;

protected:
	boost::posix_time::time_duration m_timeout;
	std::string m_url;
	std::string m_port;
	std::string m_methodType;
	std::string m_path;
	std::unordered_map<std::string, std::string> m_properties;
	std::string m_data;
};

}

#endif
