/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "HttpSyncConnection.h"
#include <istream>
#include <ostream>
#include <string>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>

using boost::asio::ip::tcp;

namespace common {

/**
 * Implementation of the http client using boost asio.
 */
class HttpSyncClient {

public:
	HttpSyncClient(boost::asio::io_service& ioService, const std::string& server, const std::string& port,const boost::posix_time::time_duration& timeout, bool verbose) :
		m_resolver(ioService), m_socket(ioService), m_deadline(ioService), m_verbose(verbose)
 {

//		m_deadline.expires_from_now(timeout);
//		m_deadline.async_wait(boost::bind(&HttpSyncClient::checkTimeout, this));

		// Start a synchronous resolve to translate the server and service names
		// into a list of endpoints.
		tcp::resolver::query query(server, port);
		boost::system::error_code err;
		const auto results = m_resolver.resolve(query, err);
		if (err) {
			throw boost::system::system_error(err);
		}

		boost::asio::connect(m_socket, results.begin(), results.end(), err);
		if (err) {
			throw boost::system::system_error(err);
		}

//		// The response is received, cancel the deadline.
//		m_deadline.expires_from_now(boost::posix_time::neg_infin);
	}

	~HttpSyncClient() {
		m_socket.close();
	}

	std::string getResponse() const {
		return m_rs.str();
	}

	void writeRequest(const std::string& method, const std::string& server, const std::string& port, const std::string& path, const BaseHttpConnection::properties_type& properties, const std::string& data, const boost::posix_time::time_duration& timeout) {

		m_deadline.expires_from_now(timeout);

		// Form the request. We specify the "Connection: close" header so that the
		// server will close the socket after transmitting the response. This will
		// allow us to treat all data up until the EOF as the content.
		boost::asio::streambuf  request;
		if (m_verbose) {
			std::cout << "Request size : " << request.size() << std::endl;
		}
		std::ostream requestStream(&request);
		requestStream << method << " " << path << " HTTP/1.0\r\n";
		requestStream << "Host: " << server << ":" << port << "\r\n";

		// Add the properties.
		bool accept = false;
		for (auto p : properties) {
			if (p.first == "Accept") {
				accept = true;
			}
			requestStream << p.first << ": " << p.second << "\r\n";
		}

		if (!accept) {
			requestStream << "Accept: */*\r\n";
		}

		requestStream << "Connection: close\r\n\r\n";
		requestStream << data << "\r\n";

		if (m_verbose) {
			std::cout << "Request: "<< std::endl;
			const char* header = boost::asio::buffer_cast<const char*>(request.data());
			std::cout << header << std::endl;
		}

		// Send the request.
		boost::system::error_code err;
		boost::asio::write(m_socket, request, err);
		if (err) {
			throw boost::system::system_error(err);
		}

		boost::asio::read_until(m_socket, m_response, "\r\n", err);
		if (err) {
			throw boost::system::system_error(err);
		}
		// Check that response is OK.
		std::istream responseStream(&m_response);
		std::string httpVersion;
		responseStream >> httpVersion;
		unsigned int statusCode;
		responseStream >> statusCode;
		std::string statusMessage;
		std::getline(responseStream, statusMessage);
		if (!responseStream || httpVersion.substr(0, 5) != "HTTP/") {
			throw RequestException(std::string("Invalid response"), 0);
		}
		if (statusCode != 200) {
			throw RequestException(std::string("response returned with status code ") + boost::lexical_cast<std::string>(statusCode), statusCode);
		}

		if (m_verbose) {
			std::cout << "Status code: " << statusCode << "\n";
		}

		boost::asio::read_until(m_socket, m_response, "\r\n\r\n", err);
		if (err) {
			throw boost::system::system_error(err);
		}

		// Process the response headers.
		std::istream responseStream2(&m_response);
		std::string header;
		while (std::getline(responseStream2, header) && header != "\r") {
			if (m_verbose) {
				std::cout << header << "\n";
			}
		}

		if (m_verbose) {
			std::cout << "\n";
		}

		// Write whatever content we already have to output.
		if (m_response.size() > 0)
			m_rs << &m_response;

		do {
			boost::asio::read(m_socket, m_response, boost::asio::transfer_at_least(1), err);
			if (!err) {
				// Write all of the data that has been read so far.
				m_rs << &m_response;

			} else if (err != boost::asio::error::eof) {

				// To avoid short read in some cases.
				if (m_response.size() != 0) {
					// The response is received, cancel the deadline.
					m_deadline.expires_from_now(boost::posix_time::neg_infin);

					throw boost::system::system_error(err);
				}
				break;
			}
		} while(true);
		m_deadline.expires_from_now(boost::posix_time::neg_infin);
	}

private:

	tcp::resolver m_resolver;
	tcp::socket m_socket;
	boost::asio::streambuf m_response;
	boost::asio::deadline_timer m_deadline;
	std::stringstream m_rs;

	bool m_verbose;
};

HttpSyncConnection::HttpSyncConnection() {
}

HttpSyncConnection::~HttpSyncConnection() {
}

void HttpSyncConnection::openUrl(const std::string& url, const std::string& port) {
	BaseHttpConnection::openUrl(url, port);
}

void HttpSyncConnection::openUrl(const std::string& url) {
	openUrl(url, "80");
}


std::string HttpSyncConnection::send(bool verbose) {

    try
    {
        boost::asio::io_service ioService;
    	HttpSyncClient httpSyncClient(ioService, m_url, m_port, m_timeout, verbose);
    	httpSyncClient.writeRequest(m_methodType, m_url, m_port, m_path, m_properties, m_data, m_timeout);
    	ioService.run();
    	return httpSyncClient.getResponse();
    }
    catch (std::exception& e) {
        throw ConnectionException(e.what());
    }
}

}
