/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "HttpsConnection.h"
#include <istream>
#include <ostream>
#include <string>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/asio/ssl.hpp>
#include <boost/lexical_cast.hpp>

using boost::asio::ip::tcp;

namespace common {

HttpsConnection::HttpsConnection() {
}

HttpsConnection::~HttpsConnection() {
}

void HttpsConnection::openUrl(const std::string& url, const std::string& port) {
	BaseHttpConnection::openUrl(url, port);
}

void HttpsConnection::openUrl(const std::string& url) {
	BaseHttpConnection::openUrl(url, "443");
}

/**
 * Implementation of the https client using boost asio.
 */
class HttpsClient {

public:
	HttpsClient(boost::asio::io_service& ioService, boost::asio::ssl::context& context, const std::string& method, const std::string& server, const std::string& port, const std::string& path, const BaseHttpConnection::properties_type& properties, const std::string& data, const boost::posix_time::time_duration& timeout, bool verbose) :
			m_resolver(ioService), m_socket(ioService, context), m_deadline(ioService), m_verbose(verbose) {

		m_deadline.expires_from_now(timeout);
		m_deadline.async_wait(boost::bind(&HttpsClient::checkTimeout, this));

		// Form the request. We specify the "Connection: close" header so that the
		// server will close the socket after transmitting the response. This will
		// allow us to treat all data up until the EOF as the content.
		std::ostream requestStream(&m_request);
		requestStream << method << " " << path << " HTTP/1.0\r\n";
		requestStream << "Host: " << server << ":" << port << "\r\n";

		// Add the properties.
		bool accept = false;
		for (auto p : properties) {
			if (p.first == "Accept") {
				accept = true;
			}
			requestStream << p.first << ": " << p.second << "\r\n";
		}

		if (!accept) {
			requestStream << "Accept: */*\r\n";
		}

		requestStream << "Connection: close\r\n\r\n";
		requestStream << data << "\r\n";

		// Start an asynchronous resolve to translate the server and service names
		// into a list of endpoints.
		tcp::resolver::query query(server, port);
		m_resolver.async_resolve(query, boost::bind(&HttpsClient::handleResolve, this, boost::asio::placeholders::error, boost::asio::placeholders::iterator));
	}

	std::string getResponse() const {
		return m_rs.str();
	}

private:

	void checkTimeout() {
		// Close the socket if this function is called before receiving the response will cancel the blocking calls.
		m_socket.lowest_layer().close();
	}

	void handleResolve(const boost::system::error_code& err, tcp::resolver::iterator endpointIterator) {
		if (!err) {
			m_socket.set_verify_mode(boost::asio::ssl::verify_peer);
			m_socket.set_verify_callback(boost::bind(&HttpsClient::verifyCertificate, this, _1, _2));

			boost::asio::async_connect(m_socket.lowest_layer(), endpointIterator, boost::bind(&HttpsClient::handleConnect, this, boost::asio::placeholders::error));
		} else {
			throw boost::system::system_error(err);
		}
	}

	bool verifyCertificate(bool preverified, boost::asio::ssl::verify_context& ctx) {
		// The verify callback can be used to check whether the certificate that is
		// being presented is valid for the peer. For example, RFC 2818 describes
		// the steps involved in doing this for HTTPS. Consult the OpenSSL
		// documentation for more details. Note that the callback is called once
		// for each certificate in the certificate chain, starting from the root
		// certificate authority.

		// In this example we will simply print the certificate's subject name.
		char subjectName[256];
		X509* cert = X509_STORE_CTX_get_current_cert(ctx.native_handle());
		X509_NAME_oneline(X509_get_subject_name(cert), subjectName, 256);

		return preverified;
	}

	void handleConnect(const boost::system::error_code& err) {
		if (!err) {
			m_socket.async_handshake(boost::asio::ssl::stream_base::client, boost::bind(&HttpsClient::handleHandshake, this, boost::asio::placeholders::error));
		} else {
			throw boost::system::system_error(err);
		}
	}

	void handleHandshake(const boost::system::error_code& err) {
		if (!err) {

			if (m_verbose) {
				std::cout << "Request: " << "\n";
				const char* header = boost::asio::buffer_cast<const char*>(m_request.data());
				std::cout << header << "\n";
			}

			// The handshake was successful. Send the request.
			boost::asio::async_write(m_socket, m_request, boost::bind(&HttpsClient::handleWriteRequest, this, boost::asio::placeholders::error));
		} else {
			throw boost::system::system_error(err);
		}
	}

	void handleWriteRequest(const boost::system::error_code& err) {
		if (!err) {
			// Read the response status line. The response_ streambuf will
			// automatically grow to accommodate the entire line. The growth may be
			// limited by passing a maximum size to the streambuf constructor.
			boost::asio::async_read_until(m_socket, m_response, "\r\n", boost::bind(&HttpsClient::handleReadStatusLine, this, boost::asio::placeholders::error));
		} else {
			throw boost::system::system_error(err);
		}
	}

	void handleReadStatusLine(const boost::system::error_code& err) {
		if (!err) {
			// Check that response is OK.
			std::istream responseStream(&m_response);
			std::string httpVersion;
			responseStream >> httpVersion;
			unsigned int statusCode;
			responseStream >> statusCode;
			std::string statusMessage;
			std::getline(responseStream, statusMessage);
			if (!responseStream || httpVersion.substr(0, 5) != "HTTP/") {
				throw RequestException(std::string("Invalid response"), 0);
			}
			if (statusCode != 200) {
				throw RequestException(std::string("response returned with status code ") + boost::lexical_cast<std::string>(statusCode), statusCode);
			}

			if (m_verbose) {
				std::cout << "Status code: " << statusCode << "\n";
			}

			// Read the response headers, which are terminated by a blank line.
			boost::asio::async_read_until(m_socket, m_response, "\r\n\r\n", boost::bind(&HttpsClient::handleReadHeaders, this, boost::asio::placeholders::error));
		} else {
			throw boost::system::system_error(err);
		}
	}

	void handleReadHeaders(const boost::system::error_code& err) {
		if (!err) {
			// Process the response headers.
			std::istream responseStream(&m_response);
			std::string header;
			while (std::getline(responseStream, header) && header != "\r") {
				if (m_verbose) {
					std::cout << header << "\n";
				}
			}

			if (m_verbose) {
				std::cout << "\n";
			}

			// Write whatever content we already have to output.
			if (m_response.size() > 0)
				m_rs << &m_response;

			// Start reading remaining data until EOF.
			boost::asio::async_read(m_socket, m_response, boost::asio::transfer_at_least(1), boost::bind(&HttpsClient::handleReadContent, this, boost::asio::placeholders::error));
		} else {
			throw boost::system::system_error(err);
		}
	}

	void handleReadContent(const boost::system::error_code& err) {

		if (!err) {
			// Write all of the data that has been read so far.
			m_rs << &m_response;

			// Continue reading remaining data until EOF.
			boost::asio::async_read(m_socket, m_response, boost::asio::transfer_at_least(1), boost::bind(&HttpsClient::handleReadContent, this, boost::asio::placeholders::error));
		} else if (err != boost::asio::error::eof) {

			// To avoid short read in some cases.
			if (m_response.size() != 0) {
				// The response is received, cancel the deadline.
				m_deadline.expires_from_now(boost::posix_time::neg_infin);

				throw boost::system::system_error(err);
			}
		}

		// The response is received, cancel the deadline.
		m_deadline.expires_from_now(boost::posix_time::neg_infin);
	}

	tcp::resolver m_resolver;
	boost::asio::ssl::stream<boost::asio::ip::tcp::socket> m_socket;
	boost::asio::streambuf m_request;
	boost::asio::streambuf m_response;
	boost::asio::deadline_timer m_deadline;

	std::stringstream m_rs;

	bool m_verbose;
};

std::string HttpsConnection::send(bool verbose) {

    try
    {
        boost::asio::ssl::context ctx(boost::asio::ssl::context::sslv23);
        ctx.set_default_verify_paths();

        boost::asio::io_service ioService;
        HttpsClient client(ioService, ctx, m_methodType, m_url, m_port, m_path, m_properties, m_data, m_timeout, verbose);

        ioService.run();

        return client.getResponse();
    }
    catch (std::exception& e) {
        throw ConnectionException(e.what());
    }
}

}
