/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "WebSocketConnection.h"

#include <boost/beast/core.hpp>
#include <boost/beast/websocket.hpp>
#include <boost/asio/strand.hpp>

#include <iostream>

namespace beast = boost::beast;
// from <boost/beast.hpp>
namespace http = beast::http;
// from <boost/beast/http.hpp>
namespace websocket = beast::websocket;
// from <boost/beast/websocket.hpp>
namespace net = boost::asio;
// from <boost/asio.hpp>
using tcp = boost::asio::ip::tcp;
// from <boost/asio/ip/tcp.hpp>

namespace common {

WebSocketConnectionException::WebSocketConnectionException(std::string const & message) :
	Exception("common::ConnectionException", message) {
}

WebSocketConnection::WebSocketConnection() {
}

WebSocketConnection::~WebSocketConnection() {
}

void WebSocketConnection::openUrl(const std::string &url, const std::string &port) {
	m_url = url;
	m_port = port;
}

void WebSocketConnection::setPath(const std::string& path) {
	m_path = path;
}

void WebSocketConnection::setTimeout(const Duration& timeout) {
	m_timeout = timeout.getPosixTime();
}

// Sends a WebSocket message and prints the response
class session: public std::enable_shared_from_this<session> {
	tcp::resolver resolver_;
	websocket::stream<beast::tcp_stream> ws_;
	beast::flat_buffer buffer_;
	std::string url_;
	std::string path_;
	std::string text_;
	boost::posix_time::time_duration timeout_;
	bool verbose_;

public:
	// Resolver and socket require an io_context
	explicit session(net::io_context &ioc) :
			resolver_(net::make_strand(ioc)), ws_(net::make_strand(ioc)) {
	}

	// Start the asynchronous operation
	void run(char const *url, char const *port, char const *path, char const *text, const boost::posix_time::time_duration& timeout, bool verbose) {
		// Save these for later
		url_ = url;
		path_ = path;
		text_ = text;
		timeout_ = timeout;
		verbose_ = verbose;

		// Look up the domain name
		resolver_.async_resolve(url, port, beast::bind_front_handler(&session::on_resolve, shared_from_this()));
	}

	std::string getResponse() const {
		std::ostringstream os;
		os << beast::make_printable(buffer_.data());
		if (verbose_) {
			std::cout << "buffer_ size" << buffer_.size() << std::endl;
			std::cout << "os : " << os.str().size() << std::endl;
		}
		return os.str();
	}

private:

	void on_resolve(beast::error_code ec, tcp::resolver::results_type results) {
		if (ec)
			throw beast::system_error(ec);

		// Set the timeout for the operation
		beast::get_lowest_layer(ws_).expires_after(std::chrono::seconds(timeout_.seconds()));

		// Make the connection on the IP address we get from a lookup
		beast::get_lowest_layer(ws_).async_connect(results,
				beast::bind_front_handler(&session::on_connect, shared_from_this()));
	}

	void on_connect(beast::error_code ec, tcp::resolver::results_type::endpoint_type ep) {
		if (ec)
			throw beast::system_error(ec);

		// Turn off the timeout on the tcp_stream, because
		// the websocket stream has its own timeout system.
		beast::get_lowest_layer(ws_).expires_never();

		// Set suggested timeout settings for the websocket
		ws_.set_option(websocket::stream_base::timeout::suggested(beast::role_type::client));

		// Set a decorator to change the User-Agent of the handshake
		ws_.set_option(websocket::stream_base::decorator([](websocket::request_type &req) {
			req.set(http::field::user_agent, std::string(BOOST_BEAST_VERSION_STRING) + " websocket-client-async");
		}));

		// Update the host_ string. This will provide the value of the
		// Host HTTP header during the WebSocket handshake.
		// See https://tools.ietf.org/html/rfc7230#section-5.4
		url_ += ':' + std::to_string(ep.port());

		// Perform the websocket handshake
		ws_.async_handshake(url_, path_, beast::bind_front_handler(&session::on_handshake, shared_from_this()));
	}

	void on_handshake(beast::error_code ec) {
		if (ec)
			throw beast::system_error(ec);

		if (verbose_) {
			std::cout << "Request: " << text_ << "\n";
		}

		// Send the message
		ws_.async_write(net::buffer(text_), beast::bind_front_handler(&session::on_write, shared_from_this()));
	}

	void on_write(beast::error_code ec, std::size_t bytes_transferred) {
		boost::ignore_unused(bytes_transferred);

		if (ec)
			throw beast::system_error(ec);

		// Read a message into our buffer
		ws_.async_read(buffer_, beast::bind_front_handler(&session::on_read, shared_from_this()));
	}

	void on_read(beast::error_code ec, std::size_t bytes_transferred) {
		boost::ignore_unused(bytes_transferred);

		if (ec)
			throw beast::system_error(ec);

		// Close the WebSocket connection
		ws_.async_close(websocket::close_code::normal, beast::bind_front_handler(&session::on_close, shared_from_this()));
	}

	void on_close(beast::error_code ec) {
		if (ec)
			throw beast::system_error(ec);

		// If we get here then the connection is closed gracefully

		// The make_printable() function helps print a ConstBufferSequence
		if (verbose_) {
			std::cout << beast::make_printable(buffer_.data()) << std::endl;
		}
	}

};

std::string WebSocketConnection::send(const std::string& text, bool verbose) {

    try {
	   // The io_context is required for all I/O
		net::io_context ioc;

		std::shared_ptr<session> sess(std::make_shared<session>(ioc));
		// Launch the asynchronous operation
		sess->run(m_url.c_str(), m_port.c_str(), m_path.c_str(), text.c_str(), m_timeout, verbose);

		// Run the I/O service. The call will return when
		// the socket is closed.
		ioc.run();

		return sess->getResponse();
    }
	catch (std::exception& e) {
		throw WebSocketConnectionException(e.what());
	}
}

}
