/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "HttpConnection.h"
#include <boost/beast/core.hpp>
#include <boost/beast/http.hpp>
#include <boost/beast/version.hpp>
#include <boost/asio/strand.hpp>
#include <cstdlib>
#include <functional>
#include <iostream>
#include <memory>
#include <string>

namespace beast = boost::beast;
// from <boost/beast.hpp>
namespace http = beast::http;
// from <boost/beast/http.hpp>
namespace net = boost::asio;
// from <boost/asio.hpp>
using tcp = boost::asio::ip::tcp;
// from <boost/asio/ip/tcp.hpp>

//------------------------------------------------------------------------------


// Performs an HTTP GET and prints the response
class HttpSession : public std::enable_shared_from_this<HttpSession> {

	tcp::resolver resolver_;
	beast::tcp_stream stream_;
	beast::flat_buffer buffer_; // (Must persist between reads)
	http::request<http::string_body> req_;
	http::response<http::string_body> res_;
	boost::posix_time::time_duration timeout_;
	bool verbose_;

public:
	// Objects are constructed with a strand to
	// ensure that handlers do not execute concurrently.
	explicit HttpSession(net::io_context &ioc) :
		resolver_(net::make_strand(ioc)), stream_(net::make_strand(ioc)), verbose_(false) {
	}

	// Report a failure
	void fail(beast::error_code ec, char const *what) {
		std::cerr << what << ": " << ec.message() << "\n";
	}

	// Start the asynchronous operation
	void run(char const *host, char const *port, char const *target, int version, const std::string& methodType, const common::BaseHttpConnection::properties_type& properties, const boost::posix_time::time_duration& timeout, const std::string& data, bool verbose) {

		timeout_ = timeout;
		verbose_ = verbose;

		// Set up an HTTP GET request message
		req_.version(version);

		if (methodType == "GET") {
			req_.method(http::verb::get);
		}
		else if (methodType == "POST") {
			req_.method(http::verb::post);
		}
		else {
			req_.method(http::verb::get);
		}

		req_.target(target);
		req_.set(http::field::host, host);
		req_.set(http::field::user_agent, BOOST_BEAST_VERSION_STRING);

		// Set the data in body.
		req_.body() = data;

		// Do not keep the connection.
		req_.keep_alive(false);

		// Set the properties.
		bool accept = false;
		for (auto p : properties) {
			if (p.first == "Accept") {
				accept = true;
			}
			req_.set(p.first, p.second);
		}

		if (!accept) {
			req_.set("Accept", "*/*");
		}

		// Look up the domain name
		resolver_.async_resolve(host, port, beast::bind_front_handler(&HttpSession::on_resolve, shared_from_this()));
	}

	void on_resolve(beast::error_code ec, tcp::resolver::results_type results) {
		if (ec)
			return fail(ec, "resolve");

		// Set a timeout on the operation
		if (timeout_ != boost::posix_time::time_duration()) {
			stream_.expires_after(std::chrono::seconds(timeout_.seconds()));
		}

		// Make the connection on the IP address we get from a lookup
		stream_.async_connect(results, beast::bind_front_handler(&HttpSession::on_connect, shared_from_this()));
	}

	void on_connect(beast::error_code ec, tcp::resolver::results_type::endpoint_type) {
		if (ec)
			return fail(ec, "connect");

		// Set a timeout on the operation
		if (timeout_ != boost::posix_time::time_duration()) {
			stream_.expires_after(std::chrono::seconds(timeout_.seconds()));
		}

		if (verbose_) {
			std::cout << "Request: " << "\n";
			std::cout << req_ << "\n";
		}

		// Send the HTTP request to the remote host
		http::async_write(stream_, req_, beast::bind_front_handler(&HttpSession::on_write, shared_from_this()));
	}

	void on_write(beast::error_code ec, std::size_t bytes_transferred) {
		boost::ignore_unused(bytes_transferred);

		if (ec)
			return fail(ec, "write");

		// Receive the HTTP response
		http::async_read(stream_, buffer_, res_, beast::bind_front_handler(&HttpSession::on_read, shared_from_this()));
	}

	void on_read(beast::error_code ec, std::size_t bytes_transferred) {
		boost::ignore_unused(bytes_transferred);

		if (ec)
			return fail(ec, "read");

		// Write the message to standard out
		if (verbose_) {
			std::cout << res_ << std::endl;
		}

		// Gracefully close the socket
		stream_.socket().shutdown(tcp::socket::shutdown_both, ec);

		// not_connected happens sometimes so don't bother reporting it.
		if (ec && ec != beast::errc::not_connected)
			return fail(ec, "shutdown");

		// If we get here then the connection is closed gracefully
	}

	std::string get_response() {
		return res_.body();
	}
};

namespace common {
namespace exp {

HttpConnection::HttpConnection() {
}

HttpConnection::~HttpConnection() {
}

void HttpConnection::openUrl(const std::string &url, const std::string &port) {
	BaseHttpConnection::openUrl(url, port);
}

void HttpConnection::openUrl(const std::string &url) {
	BaseHttpConnection::openUrl(url, "80");
}

std::string HttpConnection::send(bool verbose) {

	// The io_context is required for all I/O
	net::io_context ioc;

	// Launch the asynchronous operation
    std::shared_ptr<HttpSession> s = std::make_shared<HttpSession>(ioc);
    s->run(m_url.c_str(), m_port.c_str(), m_path.c_str(), 11, m_methodType, m_properties, m_timeout, m_data, verbose);

	// Run the I/O service. The call will return when
	// the get operation is complete.
	ioc.run();

	return s->get_response();
}

}
}
