/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "OutputStream.h"
#include "Color.h"
#include "Debug.h"
#include <common/base/Convert.h>
#include <iostream>

using namespace std;

namespace common {

Out out;
Out slog;
Warn warn;
Err err;
Title title;
Trace trace;

OutputStream<0> & OutputStreamHelper<0, EndOfLine>::apply(OutputStream<0> & os, EndOfLine const & e) {

	lock_guard<mutex> lock(os._lineMutex);

	unsigned long threadID = pthread_self();
	
	OutputStream<0>::Line const & line = os._lineMap[threadID];
	
	// flushing the line
	if (os._stdout) {
		cout << line.value << endl;
	}
	
	if (os._serverLogAccessor != 0 && line.log) {
		os._serverLogAccessor->writeServer(os._processName, threadID, false, line.value);
	}

	os._lineMap.erase(threadID);

	return os;
}

OutputStream<1> & OutputStreamHelper<1, EndOfLine>::apply(OutputStream<1> & os, EndOfLine const & e) {

	lock_guard<mutex> lock(os._lineMutex);

	unsigned long threadID = pthread_self();

	OutputStream<1>::Line const & line = os._lineMap[threadID];

	// flushing the line
	if (os._stdout) {
		cout << color::rize(line.value, "Yellow", "Default", "Default") << endl;
	}

	if (os._serverLogAccessor != 0 && line.log) {
		os._serverLogAccessor->writeServer(os._processName, threadID, false, line.value);
	}

	os._lineMap.erase(threadID);

	return os;
}

OutputStream<2> & OutputStreamHelper<2, EndOfLine>::apply(OutputStream<2> & os, EndOfLine const & e) {
	
	lock_guard<mutex> lock(os._lineMutex);
		
	unsigned long threadID = pthread_self();
	
	OutputStream<2>::Line const & line = os._lineMap[threadID];
		
	// flushing the line
	if (os._stdout) {
		ostringstream sbuf;
		sbuf << line.value << " (" << e.getFile() << " : " << e.getLine() << ")";
		cout << color::rize(sbuf.str(), "Red", "Default", "Default") << endl;
	}
	
	if (os._serverLogAccessor != 0 && line.log) {
		os._serverLogAccessor->writeServer(os._processName, threadID, true, line.value + " (" + e.getFile() + " : " + to<string>(e.getLine()) + ")");
	}	
	os._lineMap.erase(threadID);
	
	return os;
}

OutputStream<3> & OutputStreamHelper<3, EndOfLine>::apply(OutputStream<3> & os, EndOfLine const & e) {

	lock_guard<mutex> lock(os._lineMutex);

	unsigned long threadID = pthread_self();

	OutputStream<3>::Line const & line = os._lineMap[threadID];

	// flushing the line
	if (os._stdout) {
		cout << color::rize(line.value, "Blue", "Default", "Bold") << endl;
	}

	if (os._serverLogAccessor != 0 && line.log) {
		os._serverLogAccessor->writeServer(os._processName, threadID, false, line.value);
	}

	os._lineMap.erase(threadID);

	return os;
}

OutputStream<4> & OutputStreamHelper<4, EndOfLine>::apply(OutputStream<4> & os, EndOfLine const & e) {

	lock_guard<mutex> lock(os._lineMutex);

	unsigned long threadID = pthread_self();

	OutputStream<4>::Line const & line = os._lineMap[threadID];

	// flushing the line
	if (os._stdout) {
		ostringstream sbuf;
		sbuf << Debug::prefix() << "[" << e.getFileName() << " : " << e.getLine() << "] " << line.value;

		cout << sbuf.str() << endl;
	}

	if (os._serverLogAccessor != 0 && line.log) {
		os._serverLogAccessor->writeServer(os._processName, threadID, false, line.value);
	}

	os._lineMap.erase(threadID);

	return os;
}

OutputStream<0> & OutputStreamHelper<0, NoLog>::apply(OutputStream<0> & os, NoLog const &) {

	lock_guard<mutex> lock(os._lineMutex);

	os._lineMap[pthread_self()].log = false;

	return os;
}

OutputStream<1> & OutputStreamHelper<1, NoLog>::apply(OutputStream<1> & os, NoLog const &) {
	
	lock_guard<mutex> lock(os._lineMutex);
		
	os._lineMap[pthread_self()].log = false;
		
	return os;
}

OutputStream<2> & OutputStreamHelper<2, NoLog>::apply(OutputStream<2> & os, NoLog const &) {
	
	lock_guard<mutex> lock(os._lineMutex);
		
	os._lineMap[pthread_self()].log = false;
		
	return os;
}

OutputStream<3> & OutputStreamHelper<3, NoLog>::apply(OutputStream<3> & os, NoLog const &) {

	lock_guard<mutex> lock(os._lineMutex);

	os._lineMap[pthread_self()].log = false;

	return os;
}

OutputStream<4> & OutputStreamHelper<4, NoLog>::apply(OutputStream<4> & os, NoLog const &) {

	lock_guard<mutex> lock(os._lineMutex);

	os._lineMap[pthread_self()].log = false;

	return os;
}

OutputStream<0> & OutputStreamHelper<0, Startup>::apply(OutputStream<0> & os, Startup const &) {

	lock_guard<mutex> lock(os._lineMutex);

	os._lineMap[pthread_self()].type = STARTUP;

	return os;
}

OutputStream<1> & OutputStreamHelper<1, Startup>::apply(OutputStream<1> & os, Startup const &) {
	
	lock_guard<mutex> lock(os._lineMutex);
		
	os._lineMap[pthread_self()].type = STARTUP;
		
	return os;
}

OutputStream<2> & OutputStreamHelper<2, Startup>::apply(OutputStream<2> & os, Startup const &) {
	
	lock_guard<mutex> lock(os._lineMutex);
		
	os._lineMap[pthread_self()].type = STARTUP;
		
	return os;
}

OutputStream<3> & OutputStreamHelper<3, Startup>::apply(OutputStream<3> & os, Startup const &) {

	lock_guard<mutex> lock(os._lineMutex);

	os._lineMap[pthread_self()].type = STARTUP;

	return os;
}

OutputStream<4> & OutputStreamHelper<4, Startup>::apply(OutputStream<4> & os, Startup const &) {

	lock_guard<mutex> lock(os._lineMutex);

	os._lineMap[pthread_self()].type = STARTUP;

	return os;
}

OutputStream<0> & OutputStreamHelper<0, Shutdown>::apply(OutputStream<0> & os, Shutdown const &) {

	lock_guard<mutex> lock(os._lineMutex);

	os._lineMap[pthread_self()].type = SHUTDOWN;

	return os;
}

OutputStream<1> & OutputStreamHelper<1, Shutdown>::apply(OutputStream<1> & os, Shutdown const &) {
	
	lock_guard<mutex> lock(os._lineMutex);
		
	os._lineMap[pthread_self()].type = SHUTDOWN;
		
	return os;
}

OutputStream<2> & OutputStreamHelper<2, Shutdown>::apply(OutputStream<2> & os, Shutdown const &) {
	
	lock_guard<mutex> lock(os._lineMutex);
		
	os._lineMap[pthread_self()].type = SHUTDOWN;
		
	return os;
}

OutputStream<3> & OutputStreamHelper<3, Shutdown>::apply(OutputStream<3> & os, Shutdown const &) {

	lock_guard<mutex> lock(os._lineMutex);

	os._lineMap[pthread_self()].type = SHUTDOWN;

	return os;
}

OutputStream<4> & OutputStreamHelper<4, Shutdown>::apply(OutputStream<4> & os, Shutdown const &) {

	lock_guard<mutex> lock(os._lineMutex);

	os._lineMap[pthread_self()].type = SHUTDOWN;

	return os;
}

OutputStream<0> & OutputStreamHelper<0, char *>::apply(OutputStream<0> & os, char * value) {

	return OutputStreamHelper<0, string>::apply(os, string(value));
}

OutputStream<1> & OutputStreamHelper<1, char *>::apply(OutputStream<1> & os, char * value) {

	return OutputStreamHelper<1, string>::apply(os, string(value));
}

OutputStream<2> & OutputStreamHelper<2, char *>::apply(OutputStream<2> & os, char * value) {
	
	return OutputStreamHelper<2, string>::apply(os, string(value));
}

OutputStream<3> & OutputStreamHelper<3, char *>::apply(OutputStream<3> & os, char * value) {

	return OutputStreamHelper<3, string>::apply(os, string(value));
}

OutputStream<4> & OutputStreamHelper<4, char *>::apply(OutputStream<4> & os, char * value) {

	return OutputStreamHelper<4, string>::apply(os, string(value));
}

OutputStream<0> & OutputStreamHelper<0, const char *>::apply(OutputStream<0> & os, const char * value) {
	
	return OutputStreamHelper<0, string>::apply(os, string(value));
}

OutputStream<1> & OutputStreamHelper<1, const char *>::apply(OutputStream<1> & os, const char * value) {
	
	return OutputStreamHelper<1, string>::apply(os, string(value));
}

OutputStream<2> & OutputStreamHelper<2, const char *>::apply(OutputStream<2> & os, const char * value) {
	
	return OutputStreamHelper<2, string>::apply(os, string(value));
}

OutputStream<3> & OutputStreamHelper<3, const char *>::apply(OutputStream<3> & os, const char * value) {

	return OutputStreamHelper<3, string>::apply(os, string(value));
}

OutputStream<4> & OutputStreamHelper<4, const char *>::apply(OutputStream<4> & os, const char * value) {

	return OutputStreamHelper<4, string>::apply(os, string(value));
}

}
