/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_OUTPUTSTREAM_H
#define COMMON_OUTPUTSTREAM_H

#include "EndOfLine.h"
#include "NoLog.h"
#include "Startup.h"
#include "Shutdown.h"
#include "ServerLogAccessor.h"
#include <common/base/CommonTypes.h>
#include <mutex>
#include <map>
#include <string>

namespace common {

template<int OutType, typename ValueType>
struct OutputStreamHelper;

/**
 * Class defining an output stream. The stream is redirected to the standard output,
 * the dataprovider to write a log file and send to the remote console.
 */
template<int OutType>
class OutputStream {

	template<int, typename>
	friend struct OutputStreamHelper;

public:
	/**
	 * Constructor.
	 */
	OutputStream(bool stdout = false);
	
	/**
	 * Destructor.
	 */
	~OutputStream();

	/**
	 * Sets the server log accessor.
	 */
	void setServerLogAccessor(ServerLogAccessor* serverLogAccessor);
	
	/**
	 * Sets the standard output.
	 */
	void setStdOut(bool stdout);
	
	/**
	 * Sets the process name.
	 */
	void setProcessName(std::string const & name);

private:
	ServerLogAccessor* _serverLogAccessor;
	bool _stdout;
	std::string _processName;
	
	std::mutex _lineMutex;
	
	struct Line {
	
		Line() : log(true), type(NORMAL) {
		}
	
		bool log;
		OutputType type;
		std::string value;
	};
	
	std::map<unsigned long, Line> _lineMap;
};

typedef OutputStream<0> Out;
typedef OutputStream<1> Warn;
typedef OutputStream<2> Err;
typedef OutputStream<3> Title;
typedef OutputStream<4> Trace;

extern Out out;
extern Out slog;
extern Warn warn;
extern Err err;
extern Title title;
extern Trace trace;

template<typename ValueType>
Out & operator<<(Out & os, ValueType const & value)  {
	
	return OutputStreamHelper<0, ValueType>::apply(os, value);
}

template<typename ValueType>
Warn & operator<<(Warn & os, ValueType const & value)  {

	return OutputStreamHelper<1, ValueType>::apply(os, value);
}

template<typename ValueType>
Err & operator<<(Err & os, ValueType const & value)  {
	
	return OutputStreamHelper<2, ValueType>::apply(os, value);
}

template<typename ValueType>
Title & operator<<(Title & os, ValueType const & value)  {

	return OutputStreamHelper<3, ValueType>::apply(os, value);
}

template<typename ValueType>
Trace & operator<<(Trace & os, ValueType const & value)  {
	return OutputStreamHelper<4, ValueType>::apply(os, value);
}

}

#include "OutputStream.inl"

#endif /*COMMON_OUTPUTSTREAM_H*/
