/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "LinearEstimator.h"
#include <iostream>

namespace ba = boost::accumulators;

namespace common {

LinearEstimator::LinearEstimator() {
}

LinearEstimator::~LinearEstimator() {
}

void LinearEstimator::start(double startValue, double endValue) {

	// Init the current command.
	m_currentCommand = {
		startValue,
		endValue,
		std::chrono::system_clock::now()
	};
}

double LinearEstimator::terminate(double actualValue, double elapsedTimeSeconds) {

	// Check current command.
	if (m_currentCommand.has_value()) {

		// Calculation duration for a 1-length command.
		double d = abs(actualValue - m_currentCommand.value().m_startValue);

		// Check d.
		if (d < 1E-7) {
			return 0;
		}

		// Calculate the number of seconds.
		double secondsFor1 = elapsedTimeSeconds / d;

		// Update the accumulator with data of the terminated command.
		m_acc(secondsFor1);

		// Reset the current command.
		m_currentCommand = {};

		return secondsFor1;
	}

	return 0;
}

double LinearEstimator::terminate(double actualValue) {

	// Check current command.
	if (m_currentCommand.has_value()) {

		// Get current time.
		auto endTime = std::chrono::system_clock::now();

		// Calculate duration.
		double seconds = std::chrono::duration<double>(endTime - m_currentCommand.value().m_startTime).count();

		// Call terminate with elapsed time.
		return terminate(actualValue, seconds);
	}

	return 0;
}

double LinearEstimator::terminate() {

	// Check current command.
	if (m_currentCommand.has_value()) {
		// Call terminate with elapsed time.
		return terminate(m_currentCommand.value().m_endValue);
	}

	return 0;
}

double LinearEstimator::getElapsedTime() const {

	// Check current command.
	if (!m_currentCommand.has_value()) {
		return 0;
	}

	// Get current time.
	auto currentTime = std::chrono::system_clock::now();

	// Get elapsed time.
	return std::chrono::duration<double>(currentTime - m_currentCommand.value().m_startTime).count();
}

bool LinearEstimator::hasEstimation() const {

	if (ba::count(m_acc) > 0 or m_value.has_value()) {
		return true;
	}

	return false;
}

double LinearEstimator::getEstimatedTimeFactor() const {

	// Use accumulator if has values.
	if (ba::count(m_acc) > 0) {
		return ba::mean(m_acc);
	}

	// Use value if exists.
	if (m_value.has_value()) {
		return m_value.value();
	}

	// Cannot estimate time.
	return 0;
}

double LinearEstimator::getEstimatedTime(double startValue, double endValue) const {

	// Use the estimated time factor.
	return getEstimatedTimeFactor() * abs(endValue - startValue);
}

double LinearEstimator::getCurrentEstimatedTime() const {

	// Check current command.
	if (!m_currentCommand.has_value()) {
		return 0;
	}

	// Gets the estimated time.
	return getEstimatedTime(m_currentCommand.value().m_startValue, m_currentCommand.value().m_endValue);
}

double LinearEstimator::getCurrentEstimatedRemainingTimeWithProgression(double progression) const {

	// Get elapsed time.
	double elapsedTime = getElapsedTime();

	// Normalise progression.
	double progressionNorm = progression / 100.0;

	// Return the elapsed time.
	if (progressionNorm > 0.0) {
		return elapsedTime * (1.0 / progressionNorm - 1.0);
	}

	// We cannot know.
	return elapsedTime;
}

double LinearEstimator::getCurrentEstimatedRemainingTimeWithoutProgression() const {

	// Check current command.
	if (!m_currentCommand.has_value()) {
		return 0;
	}

	// Get current estimated time.
	double currentEstimatedTime = getCurrentEstimatedTime();

	// Get elapsed time.
	double elapsedTime = getElapsedTime();

	// Check estimated time.
	if (currentEstimatedTime > 0.0 && currentEstimatedTime > elapsedTime) {

		// Return the remaining time.
		return currentEstimatedTime - elapsedTime;
	}

	return 0;
}

double LinearEstimator::getCurrentEstimatedRemainingTime(double progression) const {

	// Check current command.
	if (!m_currentCommand.has_value()) {
		return 0;
	}

	// Prefer the estimation if it exists.
	if (hasEstimation()) {

		// Get the current estimated remaining time without progression first.
		double currentEstimatedRemainingTime = getCurrentEstimatedRemainingTimeWithoutProgression();

		// Check value.
		if (currentEstimatedRemainingTime > 0.0) {

			// Return the remaining time.
			return currentEstimatedRemainingTime;
		}

		return 0;
	}

	// Get the value with progression.
	return getCurrentEstimatedRemainingTimeWithProgression(progression);
}

void LinearEstimator::reset(bool all) {

	// Reset m_value to the mean of the last execution if all is false.
	if (!all and ba::count(m_acc) > 0) {
		m_value = ba::mean(m_acc);
	}
	else {
		m_value = {};
	}

	// Reset the accumulator.
	m_acc = {};
}

}
