/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef COMMON_LINEARESTIMATOR_H_
#define COMMON_LINEARESTIMATOR_H_

#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/stats.hpp>
#include <boost/accumulators/statistics/mean.hpp>
#include <optional>
#include <chrono>

namespace common {

/**
 * Class used to estimate the duration in seconds of a command.
 */
class LinearEstimator {

public:
	/**
	 * Constructor.
	 */
	LinearEstimator();

	/**
	 * Virtual destructor.
	 */
	virtual ~LinearEstimator();

	/**
	 * Starts the command.
	 */
	void start(double startValue, double endValue);

	/**
	 * Terminate the command at actual value which can be different from endValue if a stop occurred.
	 */
	double terminate(double actualValue, double elapsedTime);

	/**
	 * Terminate the command at actual value which can be different from endValue if a stop occurred.
	 */
	double terminate(double actualValue);

	/**
	 * Terminate the command at end value.
	 */
	double terminate();

	/**
	 * Gets the elapsed time.
	 */
	double getElapsedTime() const;

	/**
	 * Returns true if has an estimation.
	 */
	bool hasEstimation() const;

	/**
	 * Gets the estimated time factor.
	 */
	double getEstimatedTimeFactor() const;

	/**
	 * Gets the estimated time.
	 */
	double getEstimatedTime(double startValue, double endValue) const;

	/**
	 * Gets the current estimated time.
	 */
	double getCurrentEstimatedTime() const;

	/**
	 * Gets the estimated remaining time by progression.
	 */
	double getCurrentEstimatedRemainingTimeWithProgression(double progression) const;

	/**
	 * Gets the estimated remaining time without progression.
	 */
	double getCurrentEstimatedRemainingTimeWithoutProgression() const;

	/**
	 * Gets the estimated remaining time.
	 */
	double getCurrentEstimatedRemainingTime(double progression) const;

	/**
	 * Resets the estimator. If value is false, the current mean is used to initiate the next estimation.
	 */
	void reset(bool all = false);

private:
	std::optional<double> m_value;
	boost::accumulators::accumulator_set<double, boost::accumulators::stats<boost::accumulators::tag::mean>> m_acc;

	struct CurrentCommand {
		double m_startValue;
		double m_endValue;
		std::chrono::system_clock::time_point m_startTime;
	};

	std::optional<CurrentCommand> m_currentCommand;
};

}

#endif
