/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DirectCore.h"
#include <common/base/ServerProperties.h>
#include <core/transportlayer/MultiThreadCommunication.h>
#include <core/transportlayer/command/CommandLauncher.h>
#include <core/scheduler/Scheduler.h>
#include <core/message/scheduler/NewCommand.h>
#include <core/message/scheduler/NewEnableDisableCommand.h>
#include <core/scheduler/Sender.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/dataaccess/DataAccess.h>
#include <core/cloner/ServantCloner.h>
#include <core/servantinfo/ServantInfo.h>
#include "variables/UserVariableManager.h"
#include <core/conditions/ConditionManager.h>
#include <core/survey/Survey.h>
#include <core/commandzone/CommandZone.h>
#include <core/commandzone/CommandZoneEventServer.h>
#include <core/commandzone/EstimatedTimeManager.h>
#include <core/commandzone/SaveBoxPropertyManager.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <core/session/SessionManager.h>
#include <core/resource/ResourceManager.h>
#include <dataprovider/transportlayer/MultiThreadCommunication.h>
#include <ics/Scheduler/ThreadPool.h>
#include <ics/DataProvider/Property/AnyProperty.h>
#include <ics/CommandSystem/CommandLine/CLEstimatedTimeManager.h>
#include <common/base/ServerProperties.h>
#include <common/base/Convert.h>
#include <common/base/Real.h>
#include <common/output/OutputStream.h>
#include <cameo/api/cameo.h>
#include <boost/lexical_cast.hpp>
#include <iostream>

using namespace std;
using namespace core;
using namespace common;
using namespace boost;
using namespace cameo;

namespace core {

const std::string DirectCore::ICS = "-ics";
const std::string DirectCore::SIMULATION = "sim";
const std::string DirectCore::SIMULATION3D = "sim3d";
const std::string DirectCore::VISA = "visa";

DirectCore::DirectCore() :
	Application(),
	_onError(false) {
}

DirectCore::~DirectCore() {
}


void DirectCore::init(int argc, char* argv[]) {

	bool ics = false;
	bool visa = false;

	for (int i = 0; i < argc; i++) {

		string oldSimulation = string("-") + SIMULATION;
		if ((string(argv[i]) == SIMULATION || string(argv[i]) == SIMULATION3D || string(argv[i]) == oldSimulation)) {
			_serverId = lexical_cast<string>(This::getId());
		} else if (string(argv[i]) == VISA) {
			visa = true;
		} else if (string(argv[i]) == ICS) {
			ics = true;
		}
	}

	// Default configuration path.
	string configPath = ServerProperties::getInstance()->getConfigPath();

	unsigned long commandZoneID = 0;
	if ((_serverId != "") && (visa ==false)) {
		// converting serverId
		istringstream is(_serverId);
		is >> commandZoneID;
	}

	MultiThreadCommunication::initInstance(argc, argv, false, _serverId);

	// initialising the core components
	Real::init();

	DataNotificationRouter::initInstance();
	DataAccess::initInstance(dataprovider::MultiThreadCommunication::getInstance()->getServantAccessor(), dataprovider::MultiThreadCommunication::getInstance()->getPropertyAccessor(), dataprovider::MultiThreadCommunication::getInstance()->getCommandAccessor());
	LogHandler::initInstance(MultiThreadCommunication::getInstance()->getLogAccessor());
	ServantCloner::initInstance(dataprovider::MultiThreadCommunication::getInstance()->getServantAccessor());
	ServantInfo::initInstance();

	SessionManager::initInstance();
	ResourceManager::initInstance();
	SaveBoxPropertyManager::initInstance();
	CommandZone::initInstance(commandZoneID);
	EstimatedTimeManager::initInstance(MultiThreadCommunication::getInstance()->getCommandZoneEventPublisher());
	CommandZoneEventServer::initInstance(MultiThreadCommunication::getInstance()->getCommandZoneEventPublisher());

	try {
		UserVariableManager::initInstance(configPath);
		ConditionManager::initInstance(configPath);
		Survey::initInstance(configPath, MultiThreadCommunication::getInstance()->getSurveyAccessor());
		Scheduler::initInstance(configPath, MultiThreadCommunication::getInstance()->getCommandLauncher());

	} catch (const Exception& e) {
		_onError = true;
		err << e.printStack() << eol;
	}

}

void DirectCore::run() {

	// Runs the VariableManager
	UserVariableManager::getInstance()->run();

	// Runs the ConditionManager
	ConditionManager::getInstance()->run();

	if (_onError) {
		MultiThreadCommunication::getInstance()->shutdown(false);

	} else {

		MultiThreadCommunication::getInstance()->run();

		string serverName = "Nomad server";
		if (_serverId != "") {
			serverName += " simulated " + _serverId;
		}

		LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, serverName + " is running"));

		Survey::getInstance()->run();

		out << startup << "Core running" << eol;
	}

}

void DirectCore::reset() {

	out << common::shutdown << "Core shutdown" << eol;

	MultiThreadCommunication::getInstance()->stop();

	// resetting the core components
	Survey::getInstance()->stop();

	ConditionManager::getInstance()->stop();
	UserVariableManager::getInstance()->stop();
	Scheduler::getInstance()->stop();

	EstimatedTimeManager::resetInstance();
	CLEstimatedTimeManager::resetInstance();
	CommandZone::resetInstances();
	SessionManager::resetInstance();
	ResourceManager::resetInstance();
	Survey::resetInstance();

	ConditionManager::resetInstance();
	DataNotificationRouter::resetInstance();
	Scheduler::resetInstance();

	string serverName = "Nomad server";
	if (_serverId != "") {
		serverName += " simulated " + _serverId;
	}

	LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, serverName + " has shutdown"));

	LogHandler::resetInstance();

	CommandZoneEventServer::resetInstance();

	// resetting the server log and communication
	MultiThreadCommunication::resetInstance();
}

void DirectCore::shutdown(bool isRemote) {

	MultiThreadCommunication::getInstance()->shutdown(isRemote);
}

bool DirectCore::onError() const {
	return _onError;
}

void DirectCore::show() {
	list<NewCommand> activeCommands = Scheduler::getInstance()->getActiveCommands();

	out << nolog << "active commands :" << eol;

	for (list<NewCommand>::const_iterator c = activeCommands.begin(); c != activeCommands.end(); c++) {
		out << nolog << "\t" << DataAccess::getInstance()->what(c->getDatabaseID(), c->getCommandID()) << " <" << c->getCommandID() << ">" << eol;
	}

	list<NewEnableDisableCommand> enableDisableCommands = Scheduler::getInstance()->getActiveEnableDisableCommands();

	out << nolog << "active enable/disable commands :" << eol;

	for (list<NewEnableDisableCommand>::const_iterator c = enableDisableCommands.begin(); c != enableDisableCommands.end(); c++) {

		if (c->getValue()) {
			out << nolog << "\tcommand <enable> of " << DataAccess::getInstance()->what(c->getDatabaseID(), c->getCommandID()) << " <" << c->getCommandID() << ">" << eol;
		} else {
			out << nolog << "\tcommand <disable> of " << DataAccess::getInstance()->what(c->getDatabaseID(), c->getCommandID()) << " <" << c->getCommandID() << ">" << eol;
		}
	}

	list<NewCommand> pendingCommands = Scheduler::getInstance()->getPendingCommands();

	out << nolog << "pending commands :" << eol;

	for (list<NewCommand>::const_iterator c = pendingCommands.begin(); c != pendingCommands.end(); c++) {
		out << nolog << "\t" << DataAccess::getInstance()->what(c->getDatabaseID(), c->getCommandID()) << " <" << c->getCommandID() << ">" << eol;
	}

	list<BaseSetValue> activeSetValues = Scheduler::getInstance()->getActiveSetValues();

	out << nolog << "active set values :" << eol;

	for (list<BaseSetValue>::const_iterator s = activeSetValues.begin(); s != activeSetValues.end(); s++) {
		out << nolog << "\tset value of " << DataAccess::getInstance()->what(s->getDatabaseID(), s->getPropertyID()) << " <" << s->getPropertyID() << ">" << eol;
	}

	list<BaseSetValue> pendingSetValues = Scheduler::getInstance()->getPendingSetValues();

	out << nolog << "pending set values :" << eol;

	for (list<BaseSetValue>::const_iterator s = pendingSetValues.begin(); s != pendingSetValues.end(); s++) {
		out << nolog << "\tset value of " << DataAccess::getInstance()->what(s->getDatabaseID(), s->getPropertyID()) << " <" << s->getPropertyID() << ">" << eol;
	}
}

void DirectCore::commandZone(unsigned long ID) {

	if (CommandZone::getInstance(ID) != 0) {
		out << nolog << CommandZone::getInstance(ID)->print() << eol;
	}
}

void DirectCore::execute(const std::string& commandName, const std::vector<std::string>& args) {

	if (commandName == "run") {

		int commandID = 2;

		out << nolog << Date().toString("%d-%b-%y %H:%M:%S") << eol;

		if (args.size() > 0) {
			commandID = to<int>(args[0]);
		}

		Sender::newCommand(0, commandID, false);

	} else if (commandName == "stop") {

		if (args.size() < 1) {
			out << nolog << "\tID needed" << eol;

		} else {
			int ID = to<int>(args[0]);
			if (DataAccess::getInstance()->commandExists(ID) || DataAccess::getInstance()->servantExists(ID)) {
				Sender::stopCommand(0, ID, false);
			} else if (DataAccess::getInstance()->propertyExists(ID)) {
				Sender::stopSetter(0, ID);
			}
		}

	} else if (commandName == "pause") {

		if (args.size() < 1) {
			out << nolog << "\tcommand ID needed" << eol;

		} else {
			int ID = to<int>(args[0]);
			if (DataAccess::getInstance()->commandExists(ID)) {
				Sender::pauseCommand(0, ID);
			} else if (DataAccess::getInstance()->propertyExists(ID)) {
				Sender::pauseSetter(0, ID);
			}
		}

	} else if (commandName == "restart") {

		if (args.size() < 1) {
			out << nolog << "\tcommand ID needed" << eol;

		} else {
			int ID = to<int>(args[0]);
			if (DataAccess::getInstance()->commandExists(ID)) {
				Sender::restartCommand(0, ID);
			} else if (DataAccess::getInstance()->propertyExists(ID)) {
				Sender::restartSetter(0, ID);
			}
		}

	} else if (commandName == "duration") {

		if (args.size() < 1) {
			out << nolog << "\tcommand ID needed" << eol;

		} else {
			int commandID = to<int>(args[0]);

			try {
				double duration = Sender::getCommandDuration(0, commandID);
				out << nolog << "command duration is " << duration << eol;
			} catch (Exception& e) {
				out << nolog << e.printStack() << eol;
			}
		}

	} else if (commandName == "set") {

		unsigned long propertyID;

		if (args.size() > 0) {
			propertyID = to<int>(args[0]);
		}

		if (args.size() > 2) {
			set(propertyID, args[1], args[2]);
		} else {
			err << nolog << "set 'propertyID' 'type' 'value'" << eol;
		}

	} else if (commandName == "clone") {

		if (args.size() < 1) {
			out << nolog << "\tcontroller ID needed" << eol;

		} else {
			int servantID = to<int>(args[0]);
			if (DataAccess::getInstance()->servantExists(servantID)) {
				unsigned long cloneID = ServantCloner::getInstance()->cloneController(servantID);
				out << nolog << "\tclone success with ID " << cloneID << eol;
			}
		}

	} else if (commandName == "typemap") {
		ServantInfo::getInstance()->retrieveTypeClassMap();

	} else if (commandName == "typelist") {
		ServantInfo::getInstance()->retrieveControllerTypeList();

	} else if (commandName == "typestats") {
		ServantInfo::getInstance()->displayTypeStats();

	} else if (commandName == "show") {
		show();

	} else if (commandName == "commandzone" || commandName == "cz") {

		unsigned long commandZoneID = 0;

		if (args.size() > 0) {
			commandZoneID = to<int>(args[0]);
		}

		commandZone(commandZoneID);

	} else if (commandName == "showjson") {
		Survey::getInstance()->showJSONMessages(true);

	} else if (commandName == "hidejson") {
		Survey::getInstance()->showJSONMessages(false);

	} else if (commandName == "estim") {
		unsigned long commandZoneID = 0;

		if (args.size() > 0) {
			commandZoneID = to<int>(args[0]);
		}
		EstimatedTimeManager::getInstance()->newEstimatedTime();

	} else if (commandName == "threads") {

		out << nolog << "number of scheduler threads is " << ThreadPool::get().getNumberOfWorkers() << eol;

	} else if (commandName == "help") {

		out << nolog << "commands :" << eol;
		out << nolog << "\trun <ID> : runs the command <ID>" << eol;
		out << nolog << "\tstop <ID> : stops the command <ID>" << eol;
		out << nolog << "\tpause <ID> : pauses the command <ID>" << eol;
		out << nolog << "\trestart <ID> : restarts the command <ID>" << eol;
		out << nolog << "\tduration <ID> : computes the duration of the command <ID>" << eol;
		out << nolog << "\tset <ID> <type> <value> : sets <value> for the property <ID> of <type>" << eol;
		out << nolog << "\tcommandzone (cz) : shows the current command zone" << eol;
		out << nolog << "\tclone <ID> : clones the controller <ID>" << eol;
		out << nolog << "\ttypemap : show the controller type class map" << eol;
		out << nolog << "\ttypelist : show the controller type list" << eol;
		out << nolog << "\ttypestats : show the type stats" << eol;
		out << nolog << "\tshow : shows the current commands" << eol;
		out << nolog << "\tshowjson : shows the json messages" << eol;
		out << nolog << "\thidejson : hides the json messages" << eol;
		out << nolog << "\tquit : quits the application" << eol;

	} else {
		out << nolog << "command '" << commandName << "' unknown" << eol;

	}
}

void DirectCore::quit(const std::vector<std::string>& args, bool isRemote) {

	shutdown(isRemote);
}

void DirectCore::set(unsigned long propertyID, const std::string& type, const std::string& value) {

	if (type == INT8_PROPERTY) {
		int8 newValue = to<int8>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == UINT16_PROPERTY) {
		uint16 newValue = to<uint16>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == UINT32_PROPERTY) {
		uint32 newValue = to<uint32>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == UINT64_PROPERTY) {
		uint64 newValue = to<uint64>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == INT16_PROPERTY) {
		int16 newValue = to<int16>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == INT32_PROPERTY) {
		int32 newValue = to<int32>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == INT64_PROPERTY) {
		int64 newValue = to<int64>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == FLOAT32_PROPERTY) {
		float32 newValue = to<float32>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == FLOAT64_PROPERTY) {
		float64 newValue = to<float64>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == BOOLEAN_PROPERTY) {
		bool newValue = to<bool>(value);
		Sender::setValue(0, propertyID, newValue);
	} else if (type == STRING_PROPERTY) {
		Sender::setValue(0, propertyID, value);
	} else {
		out << nolog << "bad type '" << type << "'" << eol;
	}

}

void stopTestCommand(unsigned long ID, long timeMs) {
	usleep(1000 * timeMs);
	cout << "Stopping command" << endl;
	Sender::stopCommand(0, ID, false);
}

void DirectCore::replacePropertyNamesWithAlias() {

	cout << "Replacing property names in the conditions." << endl;

	ConditionManager::getInstance()->replacePropertyNamesWithAlias();

	cout << "Replaced property names in the conditions." << endl << endl;
}

}
