/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "MathExpression.h"
#include "PropertyWrapper.h"
#include "VariableWrapper.h"
#include "UserVariableWrapper.h"
#include "EnumeratedValueWrapper.h"
#include <common/base/Exception.h>
#include <common/base/Convert.h>
#include <common/base/TypeToString.h>
#include <common/base/StringUtilities.h>
#include <core/variables/UserVariableManager.h>
#include <dataprovider/transportlayer/dataprovider/ServantAccessorDirectImpl.h>
#include <dataprovider/transportlayer/dataprovider/PropertyAccessorDirectImpl.h>
#include <boost/lexical_cast.hpp>
#include <math.h>
#include <string>
#include <iostream>

using namespace std;
using namespace common;
using namespace boost;
using namespace dataprovider;

namespace core {

MathExpression::BadPropertyType::BadPropertyType(unsigned long propertyID) :
	common::Exception("core::MathExpression::BadPropertyType", string("bad type requested for property <") + to<string>(propertyID) + ">") {
}

MathExpression::ParserError::ParserError(const std::string& message) :
	common::Exception("core::MathExpression::ParserError", message) {
}

MathExpression::NoSuchEnumeratedValueError::NoSuchEnumeratedValueError(const std::string& message) :
	common::Exception("core::MathExpression::NoSuchEnumeratedValueError", message) {
}

MathExpression::DisabledControllerError::DisabledControllerError(const std::string& message) :
	common::Exception("core::MathExpression::DisabledControllerError", message) {
}

MathExpression::UndefinedVariableError::UndefinedVariableError(const std::string& message) :
	common::Exception("core::MathExpression::UndefinedVariableError", message) {
}

MathExpression::MathExpression(const std::string& expression) :
	_databaseID(0),	_expression(expression), _parser(new mu::Parser()) {

	// initialise the parser
	_parser->DefineNameChars("0123456789_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ$.#");
	_parser->DefineOprtChars("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+-*^/?<>=!%&|~'_");
	_parser->DefineInfixOprtChars("not+-*^/?<>=!%&|~'_");
	_parser->EnableBuiltInOprt(false);

	_parser->DefineInfixOprt("not", notOperator);
	_parser->DefineInfixOprt("!", notOperator);
	_parser->DefineOprt("or", addOperator, 0);
	_parser->DefineOprt("||", addOperator, 0);
	_parser->DefineOprt("and", multiplyOperator, 1);
	_parser->DefineOprt("&&", multiplyOperator, 1);
	_parser->DefineOprt("<", lesserOperator, 2);
	_parser->DefineOprt("<=", lesserOrEqualsOperator, 2);
	_parser->DefineOprt(">", greaterOperator, 2);
	_parser->DefineOprt(">=", greaterOrEqualsOperator, 2);
	_parser->DefineOprt("==", equalsOperator, 2);
	_parser->DefineOprt("=", equalsOperator, 2);
	_parser->DefineOprt("equal", equalsOperator, 2);
	_parser->DefineOprt("!=", notEqualsOperator, 2);
	_parser->DefineOprt("+", addOperator, 3);
	_parser->DefineOprt("-", lessOperator, 3);
	_parser->DefineOprt("%", modOperator, 3);
	_parser->DefineOprt("/", divideOperator, 4);
	_parser->DefineOprt("*", multiplyOperator, 4);
	_parser->DefineOprt("^", powerOperator, 5);

	_parser->DefineConst("true", 1.0);
	_parser->DefineConst("false", 0.0);

	// Status constants
	_parser->DefineConst("UNKNOWN", 0);
	_parser->DefineConst("IDLE", 1);
	_parser->DefineConst("RUNNING", 2);
	_parser->DefineConst("WARNING", 4);
	_parser->DefineConst("ERROR", 8);

	// Mathematical and physical constants
	_parser->DefineConst("Pi", 3.141592653589793);
	_parser->DefineConst("c", 299792458.0); // m.s-1

	_parser->DefineConst("h", 4.135667662E-15); // eV.s
	_parser->DefineConst("h_js", 6.626070040E-34); // J.s

	_parser->DefineConst("hbar", 6.582119514E-16); // eV.s/rad
	_parser->DefineConst("hbar_js", 1.054571800E-34); // J.s/rad


	_parser->DefineConst("mn", 939.565379); // MeV/c2
	_parser->DefineConst("mn_kg", 1.67493E-27); // kg

	// Set the addValue callback.
	// It will be called in the Eval function and not the SetExpr function.
	_parser->SetVarFactory(addValue, this);

	// Set the expression.
	_parser->SetExpr(expression);
}

MathExpression::~MathExpression() {
	clear();
}

void MathExpression::clear(bool excludeVariables) {

	std::scoped_lock lock(_mutex);

	// Define the map of values to keep.
	map<string, BaseValueWrapper*> valuesToKeep;

	// Delete the values.
	for (map<string, BaseValueWrapper*>::iterator p = _values.begin(); p != _values.end(); ++p) {

		if (excludeVariables && dynamic_cast<VariableWrapper *>(p->second) != nullptr) {
			valuesToKeep.insert(*p);
		}
		else {
			delete p->second;
		}
	}
	_values.clear();

	// Re-add the values to keep.
	for (map<string, BaseValueWrapper*>::iterator p = valuesToKeep.begin(); p != valuesToKeep.end(); ++p) {
		_values.insert(*p);
	}
}

void MathExpression::setDatabaseID(unsigned long databaseID) {

	std::scoped_lock lock(_mutex);

	_databaseID = databaseID;

	// Set the database ID to all the wrappers.
	for (map<string, BaseValueWrapper*>::iterator p = _values.begin(); p != _values.end(); ++p) {

		if (p->second != nullptr) {
			p->second->setDatabaseID(databaseID);
		}
	}
}

void MathExpression::toString() {

	std::scoped_lock lock(_mutex);

	cout << "Wrappers" << endl;
	for (map<string, BaseValueWrapper*>::iterator p = _values.begin(); p != _values.end(); ++p) {
		cout << "  Wrapper for " << p->first << " is " << p->second << endl;
	}
}

void MathExpression::setExpression(const std::string& expression) {

	std::scoped_lock lock(_mutex);

	// The variables are kept because they are global to the box.
	clear(true);

	// Set the expression.
	_expression = expression;

	// Clear all the variables of the parser because they will be re-added at next parsing.
	_parser->ClearVar();
	_parser->SetExpr(expression);
}

std::string MathExpression::getExpression() {

	std::scoped_lock lock(_mutex);

	return _expression;
}

void MathExpression::init() {

	std::scoped_lock lock(_mutex);

	try {
		// Evaluates the expression, that can generate errors.
		_parser->Eval();

		// Do not accept multiple expressions (separated with ,)
		if (_parser->GetNumResults() > 1) {
			throw ParserError("Only single expressions are allowed.");
		}
	}
	catch (mu::Parser::exception_type &e) {
		throw ParserError(e.GetMsg());
	}
}

bool MathExpression::isConstant() {

	std::scoped_lock lock(_mutex);

	// Try to convert the expression to a double number.
	try {
		std::string::size_type pos;
		std::stod(_expression, &pos);

		// The stod function returns the first unconverted character, it is thus necessary to check it in case of expressions like '20 + $i"
		if (pos >= _expression.length()) {
			return true;
		}
		else {
			return false;
		}
	}
	catch (...) {
		// In case the expression cannot be converted by stod.
		return false;
	}
}

std::list<MathExpression::Error> MathExpression::getErrorList() {

	std::scoped_lock lock(_mutex);

	return _errorList;
}

void MathExpression::throwError(MathExpression::Error error) {

	if (error.error == DISABLED_CONTROLLER_ERROR) {
		throw DisabledControllerError(error.message);
	}
	else if (error.error == ENUMERATED_VALUE_ERROR) {
		throw NoSuchEnumeratedValueError(error.message);
	}
	else if (error.error == UNDEFINED_VARIABLE_ERROR) {
		throw UndefinedVariableError(error.message);
	}
	else if (error.error == PARSER_ERROR) {
		throw ParserError(error.message);
	}
	else {
		cout << "No error thrown" << endl;
	}
}

void MathExpression::throwOnError() {

	Error error(common::NO_EXPRESSION_ERROR, "");

	// Iterate the list.
	if (!_errorList.empty()) {
		for (auto e : _errorList) {

			// Set the error if it is worst.
			if (static_cast<int>(error.error) < static_cast<int>(e.error)) {
				error = e;
			}
		}

		// Throw the error.
		throwError(error);
	}
}

void MathExpression::checkValue(bool checkForLoopVariables) {

	std::scoped_lock lock(_mutex);

	// Clear the current list of errors.
	_errorList.clear();

	// Get the values, that can throw errors.
	for (map<string, BaseValueWrapper*>::iterator p = _values.begin(); p != _values.end(); ++p) {
		try {
			UserVariableWrapper * userVariableWrapper = dynamic_cast<UserVariableWrapper *>(p->second);
			bool isForLoopVariable = false;
			if (userVariableWrapper != nullptr) {
				 isForLoopVariable = (userVariableWrapper->getName()[0] == '$');
			}

			// Do not check the for loop variables if requested.
			if (checkForLoopVariables || !isForLoopVariable) {
				p->second->updateValue();
			}
		}
		catch (const BaseValueWrapper::DisabledControllerError & e) {
			_errorList.push_back(Error(DISABLED_CONTROLLER_ERROR, e.getMessage()));
		}
		catch (const BaseValueWrapper::NoSuchEnumeratedValueError & e) {
			_errorList.push_back(Error(ENUMERATED_VALUE_ERROR, e.getMessage()));
		}
		catch (const BaseValueWrapper::UndefinedVariableError & e) {
			_errorList.push_back(Error(UNDEFINED_VARIABLE_ERROR, e.getMessage()));
		}
	}

	for (auto e : _errorList) {
		cout << "  Error " << (int)e.error << " : " << e.message << endl;
	}

	// Throws an error if a previous call to updateValue() threw an error.
	throwOnError();
}

double MathExpression::evaluateValue() {

	std::scoped_lock lock(_mutex);

	try {
		// Checks the value.
		checkValue(true);

		// Evaluates the expression.
		double value = _parser->Eval();

		// Do not accept multiple expressions (separated with ,)
		if (_parser->GetNumResults() > 1) {
			throw ParserError("Only single expressions are allowed.");
		}

		return value;
	}
	catch (const mu::Parser::exception_type & e) {
		throw ParserError(e.GetMsg());
	}
}

void MathExpression::setVariableValue(const std::string& variableName, double value) {

	std::scoped_lock lock(_mutex);

	// searching name
	if (_values.find(variableName) == _values.end()) {
		_values[variableName] = new VariableWrapper();
	}

	dynamic_cast<VariableWrapper*>(_values[variableName])->setValue(value);
}

void MathExpression::changeVariableName(const std::string& formerName, const std::string& newName) {

	std::scoped_lock lock(_mutex);

	// searching name
	if (_values.find(formerName) != _values.end()) {
		VariableWrapper* variableWrapper = dynamic_cast<VariableWrapper*>(_values[formerName]);

		// erasing actual entry
		_values.erase(formerName);

		// adding the new variable
		_values[newName] = variableWrapper;

	} else {
		setVariableValue(newName, 0);
	}

	// Replace into expression.
	_expression = StringUtilities::replaceText(_expression, formerName, newName);

	// Clear all the variables of the parser because they will be re-added at next parsing.
	_parser->ClearVar();
	_parser->SetExpr(_expression);
}

void MathExpression::removeAllVariables() {

	std::scoped_lock lock(_mutex);

	for (auto p = _values.begin(); p != _values.end(); ) {
		VariableWrapper* propertyWrapper = dynamic_cast<VariableWrapper*>(p->second);
		if (propertyWrapper != 0) {
			_parser->RemoveVar(p->first);

			delete propertyWrapper;
			p = _values.erase(p);
		}
		else {
			++p;
		}
	}
}

void MathExpression::updateVariables(const std::vector<std::string>& variableNames) {

	std::scoped_lock lock(_mutex);

	removeAllVariables();

	for (string v : variableNames) {
		setVariableValue(v, 0);
	}
}

std::string MathExpression::replaceVariables(const std::string& expression) {

	std::scoped_lock lock(_mutex);

	string modifiedExpression = expression;

	for (auto v = _values.begin(); v != _values.end(); v++) {

		const string& variableName = v->first;
		VariableWrapper * wrapper = dynamic_cast<VariableWrapper *>(v->second);
		if (wrapper != 0) {
			string variableValueString = to<string>(wrapper->getValue());
			modifiedExpression = StringUtilities::replaceText(modifiedExpression, variableName, variableValueString);
		}
	}

	return modifiedExpression;
}

std::set<unsigned long> MathExpression::getPropertyIDs() {

	std::scoped_lock lock(_mutex);

	set<unsigned long> propertyIDs;

	for (map<string, BaseValueWrapper*>::const_iterator p = _values.begin(); p != _values.end(); ++p) {

		// Get the property IDs.
		set<unsigned long> variablePropertyIDs = p->second->getPropertyIDs();

		for (unsigned long i : variablePropertyIDs) {
			propertyIDs.insert(i);
		}
	}

	return propertyIDs;
}

std::set<std::string> MathExpression::getUserVariableNames() {

	std::scoped_lock lock(_mutex);

	set<string> result;

	for (map<string, BaseValueWrapper*>::const_iterator p = _values.begin(); p != _values.end(); ++p) {
		UserVariableWrapper* userVariableWrapper = dynamic_cast<UserVariableWrapper*>(p->second);
		if (userVariableWrapper != 0) {
			result.insert(userVariableWrapper->getName());
		}
	}

	return result;
}

void MathExpression::checkStaticValue() {

	// Checks the value but do not check the for loop variable because we do not have the context.
	checkValue(false);
}

bool MathExpression::referencesUserVariable(const std::string& variableName) {

	// Gets the list of user variable names.
	set<string> userVariableNames = getUserVariableNames();

	// Find the name.
	return (userVariableNames.find(variableName) != userVariableNames.end());
}

std::ostream& operator<< (std::ostream& stream, const MathExpression::VariableInfos& infos) {

	if (infos.qualifier == MathExpression::VariableInfos::Qualifier::SETPOINT_PROPERTY) {
		stream << "Setpoint " << infos.servant << "." << infos.name;
	}
	else if (infos.qualifier == MathExpression::VariableInfos::Qualifier::OFFSET_PROPERTY) {
		stream << "Offset " << infos.servant << "." << infos.name;
	}
	else if (infos.qualifier == MathExpression::VariableInfos::Qualifier::ACTUAL_PROPERTY) {
		stream << "Actual " << infos.servant << "." << infos.name;
	}
	else if (infos.qualifier == MathExpression::VariableInfos::Qualifier::DYNAMIC_PROPERTY) {
		stream << "Dynamic " << infos.servant << "." << infos.name << "." << infos.index;
	}
	else if (infos.qualifier == MathExpression::VariableInfos::Qualifier::VARIABLE) {
		stream << "Variable " << infos.name;
	}
	else if (infos.qualifier == MathExpression::VariableInfos::Qualifier::ENUMERATED_VALUE) {
		stream << "Enumerated value " << infos.servant << "." << infos.name << "#" << infos.enumeratedValue;
	}

	return stream;
}

std::string MathExpression::aliasInExpression(const DataAccess::PropertyInfoItem& item) {

	string result = item.aliasPropertyName;
	result.erase(remove_if(result.begin(), result.end(), [](char c){ return (c == ' '); }), result.end());

	return result;
}

MathExpression::VariableInfos MathExpression::analyseVariableName(const std::string& variableName) {

	VariableInfos result;

	// Look for .
	string::size_type position = variableName.find('.');
	if (position != string::npos) {

		// The servant name can begin by a '$'. Then it is removed.
		result.servant = variableName.substr(0, position);

		if (result.servant.length() > 0 && result.servant[0] == '$') {
			result.servant = result.servant.substr(1);
		}

		// Start +1 for .
		string propertyName = variableName.substr(position + 1);

		// Look for next .
		position = propertyName.find('.');
		if (position != string::npos) {

			// The name of the property.
			result.name = propertyName.substr(0, position);

			// The last token.
			string lastToken = propertyName.substr(position + 1);

			if (lastToken == "setpoint" || lastToken == "sp") {
				result.qualifier = VariableInfos::Qualifier::SETPOINT_PROPERTY;
			}
			else if (lastToken == "offset" || lastToken == "os") {
				result.qualifier = VariableInfos::Qualifier::OFFSET_PROPERTY;
			}
			else {
				// Test if it is an integer.
				try {
					result.index = std::stoi(lastToken);
					result.qualifier = VariableInfos::Qualifier::DYNAMIC_PROPERTY;
				}
				catch (...) {
					// Bad qualifier.
					string message = "Bad property qualifier '";
					message += lastToken + "'";
					throw ParserError(message);
				}
			}
		}
		else {
			// Look for #.
			position = propertyName.find('#');
			if (position != string::npos) {

				// The name of the property.
				result.name = propertyName.substr(0, position);

				// The last token is the name of the constant.
				result.enumeratedValue = propertyName.substr(position + 1);
				result.qualifier = VariableInfos::Qualifier::ENUMERATED_VALUE;
			}
			else {
				// This is an actual property.
				result.name = propertyName;
				result.qualifier = VariableInfos::Qualifier::ACTUAL_PROPERTY;
			}
		}

		return result;
	}

	// No . found, then this is a variable.
	result.name = variableName;
	result.qualifier = VariableInfos::Qualifier::VARIABLE;

	return result;
}

BaseValueWrapper * MathExpression::createPropertyWrapper(unsigned long propertyID) {

	// Get the type for property ID.
	string propertyType = DataAccess::getInstance()->getPropertyType(_databaseID, propertyID);

	if (propertyType == TypeToString<bool>::convert()) {
		return new PropertyWrapper<bool>(propertyID);
	}
	else if (propertyType == TypeToString<uint16>::convert()) {
		return new PropertyWrapper<uint16>(propertyID);
	}
	else if (propertyType == TypeToString<uint32>::convert()) {
		return new PropertyWrapper<uint32>(propertyID);
	}
	else if (propertyType == TypeToString<uint64>::convert()) {
		return new PropertyWrapper<uint64>(propertyID);
	}
	else if (propertyType == TypeToString<int16>::convert()) {
		return new PropertyWrapper<int16>(propertyID);
	}
	else if (propertyType == TypeToString<int32>::convert()) {
		return new PropertyWrapper<int32>(propertyID);
	}
	else if (propertyType == TypeToString<int64>::convert()) {
		return new PropertyWrapper<int64>(propertyID);
	}
	else if (propertyType == TypeToString<float32>::convert()) {
		return new PropertyWrapper<float32>(propertyID);
	}
	else if (propertyType == TypeToString<float64>::convert()) {
		return new PropertyWrapper<float64>(propertyID);
	}
	else if (propertyType == TypeToString<std::string>::convert()) {
		throw ParserError("String properties are not allowed in expressions.");
	}

	return nullptr;
}

BaseValueWrapper * MathExpression::createEnumeratedValueWrapper(unsigned long propertyID, const std::string& label) {

	// Get the type for property ID.
	string propertyType = DataAccess::getInstance()->getPropertyType(_databaseID, propertyID);

	if (propertyType == TypeToString<bool>::convert()) {
		return new EnumeratedValueWrapper<bool>(propertyID, label);
	}
	else if (propertyType == TypeToString<uint16>::convert()) {
		return new EnumeratedValueWrapper<uint16>(propertyID, label);
	}
	else if (propertyType == TypeToString<uint32>::convert()) {
		return new EnumeratedValueWrapper<uint32>(propertyID, label);
	}
	else if (propertyType == TypeToString<uint64>::convert()) {
		return new EnumeratedValueWrapper<uint64>(propertyID, label);
	}
	else if (propertyType == TypeToString<int16>::convert()) {
		return new EnumeratedValueWrapper<int16>(propertyID, label);
	}
	else if (propertyType == TypeToString<int32>::convert()) {
		return new EnumeratedValueWrapper<int32>(propertyID, label);
	}
	else if (propertyType == TypeToString<int64>::convert()) {
		return new EnumeratedValueWrapper<int64>(propertyID, label);
	}
	else if (propertyType == TypeToString<float32>::convert()) {
		return new EnumeratedValueWrapper<float32>(propertyID, label);
	}
	else if (propertyType == TypeToString<float64>::convert()) {
		return new EnumeratedValueWrapper<float64>(propertyID, label);
	}
	else if (propertyType == TypeToString<std::string>::convert()) {
		throw ParserError("String properties are not allowed in expressions.");
	}

	return nullptr;
}

double * MathExpression::addValue(const char * valueName, void * pUserData) {

	// Get the MathExpression object and its values map.
	MathExpression* expression = static_cast<MathExpression*>(pUserData);

	// Get the mutex from the object expression.
	std::scoped_lock lock(expression->_mutex);

	map<string, BaseValueWrapper*> & values = expression->_values;
	unsigned long databaseID = expression->_databaseID;

	// Create a string from the char *.
	string variableName(valueName);

	// Search the name.
	if (values.find(variableName) == values.end()) {

		// Analyse the string.
		VariableInfos variableInfos = MathExpression::analyseVariableName(variableName);

		if (variableInfos.qualifier == VariableInfos::Qualifier::VARIABLE) {

			// Find the variable in the user variables and create a new wrapper.
			UserVariable * variable = nullptr;
			try {
				variable = UserVariableManager::getInstance()->getUserVariable(variableName);

				// Create the wrapper.
				values[variableName] = new UserVariableWrapper(variableName, variable);
			}
			catch (UserVariableManager::NoSuchUserVariableException const & e) {
				// The variable is nullptr, so that it is a for loop variable.
				values[variableName] = new VariableWrapper();
			}
		}
		else if (variableInfos.qualifier == VariableInfos::Qualifier::ENUMERATED_VALUE) {

			// Find the variable in the enumerated values.
			try {
				// Get the necessary information.
				unsigned long servantID = DataAccess::getInstance()->getServantID(variableInfos.servant);
				string controllerType = DataAccess::getInstance()->getControllerType(databaseID, servantID);

				// Get the property infos.
				vector<DataAccess::PropertyInfoItem> items = DataAccess::getInstance()->getAllPropertyInfoItems(databaseID, controllerType);

				string propertyName = variableInfos.name;

				// Look for the existence of an alias.
				// If the alias is not found, then take the property as is.
				for (auto i : items) {
					// Search for the alias without whitespace.
					if (aliasInExpression(i) == variableInfos.name) {
						propertyName = i.actualPropertyName;
					}
				}

				unsigned long propertyID = DataAccess::getInstance()->getPropertyID(databaseID, servantID, propertyName);

				// Create the wrapper.
				values[variableName] = expression->createEnumeratedValueWrapper(propertyID, variableInfos.enumeratedValue);
			}
			catch (ServantAccessorDirectImpl::NoSuchServantException const & e) {
				string message("Servant with name '");
				message += variableInfos.servant + "' does not exist";
				throw ParserError(message);
			}
			catch (PropertyAccessorDirectImpl::NoSuchPropertyException const & e) {
				string message("Servant with name '");
				message += variableInfos.servant + "' has no property '" + variableInfos.name + "'";
				throw ParserError(message);
			}
			catch (ParserError const & e) {
				throw;
			}
			catch (...) {
				throw ParserError("Unknown error");
			}
		}
		else {

			// The variable is a controller property.
			try {
				unsigned long servantID = DataAccess::getInstance()->getServantID(variableInfos.servant);
				string controllerType = DataAccess::getInstance()->getControllerType(databaseID, servantID);

				// Get the property infos.
				vector<DataAccess::PropertyInfoItem> items = DataAccess::getInstance()->getAllPropertyInfoItems(databaseID, controllerType);

				// Test if is a dynamic property.
				unsigned long propertyID = 0;

				// Finding the associated property name.
				string propertyName = variableInfos.name;

				// Look for the existence of an alias.
				// If the alias is not found, then take the property as is.
				for (auto i : items) {

					// Search for the alias without whitespace.
					if (aliasInExpression(i) == variableInfos.name) {

						// Check if the dynamic property is respected.
						if (i.isDynamic && variableInfos.qualifier != VariableInfos::Qualifier::DYNAMIC_PROPERTY) {
							// Error, the property should be accessed as a dynamic property.
							string message("Property '");
							message += variableInfos.name + "' of servant " + variableInfos.servant + "' is dynamic";
							throw ParserError(message);
						}

						if (variableInfos.qualifier == VariableInfos::Qualifier::ACTUAL_PROPERTY) {
							propertyName = i.actualPropertyName;
							break;
						}
						else if (variableInfos.qualifier == VariableInfos::Qualifier::SETPOINT_PROPERTY) {
							propertyName = i.setpointPropertyName;
							break;
						}
						else if (variableInfos.qualifier == VariableInfos::Qualifier::OFFSET_PROPERTY) {
							propertyName = i.offsetPropertyName;
							break;
						}
						else if (variableInfos.qualifier == VariableInfos::Qualifier::DYNAMIC_PROPERTY) {

							// Check if the simple property is respected.
							if (!i.isDynamic) {
								// Error, the property should not be accessed as a dynamic property.
								string message("Property '");
								message += variableInfos.name + "' of servant " + variableInfos.servant + "' is not dynamic";
								throw ParserError(message);
							}

							propertyName = i.actualPropertyName;
							break;
						}
					}
				}

				// Get the property ID.
				if (variableInfos.qualifier == VariableInfos::Qualifier::DYNAMIC_PROPERTY) {
					// This is a dynamic property.
					unsigned long dynamicPropertyID = DataAccess::getInstance()->getPropertyID(databaseID, servantID, propertyName);
					propertyID = DataAccess::getInstance()->getDynamicPropertyIDAtIndex(databaseID, dynamicPropertyID, variableInfos.index, false);
				}
				else {
					// Get the property ID.
					propertyID = DataAccess::getInstance()->getPropertyID(databaseID, servantID, propertyName);
				}

				// Create the wrapper.
				values[variableName] = expression->createPropertyWrapper(propertyID);
			}
			catch (ServantAccessorDirectImpl::NoSuchServantException const & e) {
				string message("Servant '");
				message += variableInfos.servant + "' does not exist";
				throw ParserError(message);
			}
			catch (PropertyAccessorDirectImpl::NoSuchPropertyException const & e) {
				string message("Servant '");
				message += variableInfos.servant + "' has no property '" + variableInfos.name + "'";
				throw ParserError(message);
			}
			catch (PropertyAccessorDirectImpl::IndexOutOfBoundsException const & e) {
				string message("Servant '");
				message += variableInfos.servant + "' has no dynamic property '" + variableInfos.name + "' at index " + to_string(variableInfos.index);
				throw ParserError(message);
			}
			catch (ParserError const & e) {
				throw;
			}
			catch (Exception const & e) {
				cerr << e.getMessage() << endl;
				throw ParserError("Unknown error");
			}
			catch (...) {
				throw ParserError("Unknown error");
			}
		}
	}

	return &values[variableName]->getValue();
}

double MathExpression::notOperator(double a) {
	if (a > 0.0) {
		return 0.0;
	}
	return 1.0;
}

double MathExpression::multiplyOperator(double a, double b) {
	return a * b;
}

double MathExpression::addOperator(double a, double b) {
	return a + b;
}

double MathExpression::divideOperator(double a, double b) {
	return a / b;
}

double MathExpression::lessOperator(double a, double b) {
	return a - b;
}

double MathExpression::lesserOperator(double a, double b) {
	return a < b;
}

double MathExpression::lesserOrEqualsOperator(double a, double b) {
	return a <= b;
}

double MathExpression::greaterOperator(double a, double b) {
	return a > b;
}

double MathExpression::greaterOrEqualsOperator(double a, double b) {
	return a >= b;
}

double MathExpression::equalsOperator(double a, double b) {
	return a == b;
}

double MathExpression::notEqualsOperator(double a, double b) {
	return a != b;
}

double MathExpression::modOperator(double a, double b) {
	return fmod(a, b);
}

double MathExpression::powerOperator(double a, double b) {
	return std::pow(a, b);
}

}
