/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_MATHEXPRESSION_H_
#define CORE_MATHEXPRESSION_H_

#include <common/base/Exception.h>
#include <common/base/CommonTypes.h>
#include <core/dataaccess/DataAccess.h>
#include <muParser.h>
#include <string>
#include <map>
#include <set>
#include <memory>
#include <mutex>

namespace core {

class BaseValueWrapper;

/**
 * Class defining a dynamic math expression based on string expressions.
 */
class MathExpression {

public:
	/**
	 * Bad property type exception.
	 */
	class BadPropertyType : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		BadPropertyType(unsigned long propertyID);
	};

	/**
	 * Parser exception.
	 */
	class ParserError : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		ParserError(const std::string& message);
	};

	/**
	 * Exception when the enumerated value does not exist.
	 */
	class NoSuchEnumeratedValueError : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchEnumeratedValueError(const std::string& message);
	};

	/**
	 * Exception when the controller is disabled.
	 */
	class DisabledControllerError : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		DisabledControllerError(const std::string& message);
	};

	/**
	 * Exception when the variable is undefined.
	 */
	class UndefinedVariableError : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		UndefinedVariableError(const std::string& message);
	};

	/**
	 * Constructor with an expression.
	 */
	explicit MathExpression(const std::string& expression);

	/**
	 * Virtual destructor.
	 */
	virtual ~MathExpression();

	/**
	 * Sets the database ID.
	 */
	void setDatabaseID(unsigned long databaseID);

	/**
	 * Sets the expression.
	 */
	void setExpression(const std::string& expression);

	/**
	 * Gets the expression.
	 */
	std::string getExpression();

	/**
	 * Initialises the expression.
	 */
	void init();

	/**
	 * Returns true if it is a constant expression.
	 */
	bool isConstant();

	/**
	 * Checks the value.
	 */
	void checkValue(bool checkForLoopVariables);

	/**
	 * Evaluates the value of the expression.
	 */
	double evaluateValue();

	/**
	 * Sets the variable value.
	 */
	void setVariableValue(const std::string& variableName, double value);

	/**
	 * Changes the variable name.
	 */
	void changeVariableName(const std::string& formerName, const std::string& newName);

	/**
	 * Removes all the variables.
	 */
	void removeAllVariables();

	/**
	 * Updates the variables.
	 */
	void updateVariables(const std::vector<std::string>& variableNames);

	/**
	 * Replaces the variables in the expression.
	 */
	std::string replaceVariables(const std::string& expression);

	/**
	 * Returns the list of property IDs. Init must have been called before.
	 */
	std::set<unsigned long> getPropertyIDs();

	/**
	 * Returns the set of user variable names.
	 */
	std::set<std::string> getUserVariableNames();

	/**
	 * Returns true if the expression references a user variable.
	 */
	bool referencesUserVariable(const std::string& variableName);

	/**
	 * Error structure.
	 */
	struct Error {
		common::ExpressionError error;
		std::string message;

		Error(common::ExpressionError anError, const std::string& aMessage) : error(anError), message(aMessage) {}
	};

	/**
	 * Returns the current list of errors.
	 */
	std::list<Error> getErrorList();

	/**
	 * Throws an error if the error list is not empty.
	 */
	void throwOnError();

	/**
	 * Checks the user variables.
	 */
	void checkStaticValue();

	/**
	 * Prints the variables.
	 */
	void toString();

protected:
	/**
	 * Clears the values.
	 */
	void clear(bool excludeVariables = false);

	/**
	 * Throws the error.
	 */
	void throwError(Error error);

	/**
	 * Structure used for analysing the variable.
	 */
	struct VariableInfos {

		enum class Qualifier {VARIABLE, ACTUAL_PROPERTY, SETPOINT_PROPERTY, OFFSET_PROPERTY, DYNAMIC_PROPERTY, ENUMERATED_VALUE};
		Qualifier qualifier;

		std::string servant;
		std::string name;
		std::string enumeratedValue;
		int index = 0;
	};

	friend std::ostream& operator<< (std::ostream& stream, const VariableInfos& infos);

	/**
	 * Analyses the variable name.
	 */
	static VariableInfos analyseVariableName(const std::string& variableName);

	/**
	 * Returns the alias without whitespace characters.
	 */
	static std::string aliasInExpression(const DataAccess::PropertyInfoItem& item);

	/**
	 * Creates a property wrapper.
	 */
	BaseValueWrapper * createPropertyWrapper(unsigned long propertyID);

	/**
	 * Creates an enumerated value wrapper.
	 */
	BaseValueWrapper * createEnumeratedValueWrapper(unsigned long propertyID, const std::string& label);

	/**
	 * Callback for mu::Parser.
	 */
	static double * addValue(const char * valueName, void * pUserData);

	/**
	 * ! operator.
	 */
	static double notOperator(double a);

	/**
	 * * operator.
	 */
	static double multiplyOperator(double a, double b);

	/**
	 * + operator.
	 */
	static double addOperator(double a, double b);

	/**
	 * / operator.
	 */
	static double divideOperator(double a, double b);

	/**
	 * - operator.
	 */
	static double lessOperator(double a, double b);

	/**
	 * < operator.
	 */
	static double lesserOperator(double a, double b);

	/**
	 * <= operator.
	 */
	static double lesserOrEqualsOperator(double a, double b);

	/**
	 * > operator.
	 */
	static double greaterOperator(double a, double b);

	/**
	 * >= operator.
	 */
	static double greaterOrEqualsOperator(double a, double b);

	/**
	 * == operator.
	 */
	static double equalsOperator(double a, double b);

	/**
	 * != operator.
	 */
	static double notEqualsOperator(double a, double b);

	/**
	 * % operator.
	 */
	static double modOperator(double a, double b);

	/**
	 * power operator.
	 */
	static double powerOperator(double a, double b);

	unsigned long _databaseID;
	std::string _expression;
	std::unique_ptr<mu::Parser> _parser;
	std::map<std::string, BaseValueWrapper*> _values;
	std::list<Error> _errorList;

	// The mutex is necessary for a safe copy that can occur in parallel with the setting of expressions.
	std::recursive_mutex _mutex;

};


std::ostream& operator<< (std::ostream& stream, const MathExpression::VariableInfos& infos);

}

#endif
