/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "MathExpressionRenamer.h"
#include <core/dataaccess/DataAccess.h>
#include <common/base/StringUtilities.h>
#include <algorithm>
#include <iostream>

using namespace std;
using namespace common;

namespace core {

MathExpressionRenamer::MathExpressionRenamer(const std::string& expression) :
	MathExpression(expression) {

	// Set the addValue callback.
	// It will be called in the Eval function and not the SetExpr function.
	_parser->SetVarFactory(addValue, this);

	// Set the expression.
	_parser->SetExpr(expression);
}

MathExpressionRenamer::~MathExpressionRenamer() {

	// Delete the name pairs.
	for (auto n : m_namePairs) {
		delete n;
	}
}

double * MathExpressionRenamer::addValue(const char * valueName, void * pUserData) {

	MathExpressionRenamer * renamer = static_cast<MathExpressionRenamer *>(pUserData);

	NamePair * namePair = new NamePair();
	namePair->name = valueName;
	namePair->value = 0.0;

	// Analyse the variable name.
	string variableName(valueName);
	VariableInfos variableInfos = MathExpression::analyseVariableName(variableName);

	// Get the new name.
	namePair->newName = renamer->getNewVariableName(variableName, variableInfos);

	// Add to the list.
	renamer->m_namePairs.push_back(namePair);

	// Return the address of the value so that the parser can evaluate the expression.
	return &namePair->value;
}

std::string MathExpressionRenamer::getNewVariableName(const std::string& variableName, const VariableInfos& variableInfos) {

	unsigned long servantID = DataAccess::getInstance()->getServantID(variableInfos.servant);
	string controllerType = DataAccess::getInstance()->getControllerType(0, servantID);

	// Test if is a dynamic property.
	unsigned long propertyID = 0;

	if (variableInfos.qualifier == VariableInfos::Qualifier::DYNAMIC_PROPERTY) {
		// This is a dynamic property.
		unsigned long dynamicPropertyID = DataAccess::getInstance()->getPropertyID(0, servantID, variableInfos.name);

		// Get the property infos.
		vector<DataAccess::PropertyInfoItem> items = DataAccess::getInstance()->getAllPropertyInfoItems(0, controllerType);

		// Look for the existence of an alias.
		// If the alias is not found, then take the property as is.
		for (auto i : items) {

			// The variable name is already the alias.
			if (aliasInExpression(i) == variableInfos.name) {
				return variableName;
			}
		}

		// Search for the alias.
		for (auto i : items) {

			// In case the item has an alias, we can replace if we find the name.
			if (i.aliasPropertyName != "") {

				if (i.actualPropertyName == variableInfos.name) {
					return variableInfos.servant + "."  + aliasInExpression(i) + "." + to_string(variableInfos.index);
				}
			}
		}

		return variableName;
	}
	else {
		// Get the property infos.
		vector<DataAccess::PropertyInfoItem> items = DataAccess::getInstance()->getAllPropertyInfoItems(0, controllerType);

		// Look for the existence of an alias.
		// If the alias is not found, then take the property as is.
		for (auto i : items) {

			// The variable name is already the alias.
			if (aliasInExpression(i) == variableInfos.name) {
				return variableName;
			}
		}

		// Search for the alias.
		for (auto i : items) {

			// In case the item has an alias, we can replace if we find the name.
			if (i.aliasPropertyName != "") {

				if (i.actualPropertyName == variableInfos.name) {
					return variableInfos.servant + "."  + aliasInExpression(i);
				}
				else if (i.setpointPropertyName == variableInfos.name) {
					return variableInfos.servant + "."  + aliasInExpression(i) + ".sp";
				}
				else if (i.offsetPropertyName == variableInfos.name) {
					return variableInfos.servant + "."  + aliasInExpression(i) + ".os";
				}
			}
		}

		cout << "Cannot find an alias for " << variableName << "." << endl;

		return variableName;
	}
}

bool MathExpressionRenamer::compareNamePairs(MathExpressionRenamer::NamePair * p1, MathExpressionRenamer::NamePair * p2) {
	return p1->name.length() > p2->name.length();
}

std::string MathExpressionRenamer::execute() {

	try {
		// Evaluates the expression, that can generate errors.
		_parser->Eval();
	}
	catch (mu::Parser::exception_type &e) {
		throw ParserError(e.GetMsg());
	}

	string expression = _parser->GetExpr();

	// Sort from the longest to the shortest name so that there is no risk to replace a substring of another property name e.g. A1.actual_position and A1.actual_position_long.
	sort(m_namePairs.begin(), m_namePairs.end(), compareNamePairs);

	for (auto n : m_namePairs) {
		expression = StringUtilities::replaceText(expression, n->name, n->newName);
	}

	return expression;
}

}
