/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DataAccessHelper.h"
#include <common/base/Convert.h>
#include <common/base/TypeToString.h>
#include <core/scheduler/Sender.h>
#include <core/dataaccess/DataAccess.h>

namespace core {

template<typename ValueType>
MathProperty<ValueType>::MathProperty(unsigned long propertyID) :
	BaseMathProperty(propertyID) {
}

template<typename ValueType>
MathProperty<ValueType>::~MathProperty() {
}

template<typename ValueType>
std::string MathProperty<ValueType>::getType() const {
	return common::TypeToString<ValueType>::convert();	
}

template<typename ValueType>
void MathProperty<ValueType>::setServerValue(unsigned long databaseID) {

	ValueType valueCopy;
	{
		std::scoped_lock lock(_mutex);
		valueCopy = _value;
	}

	// blocking call to Sender
	bool success = Sender::setValueOrReset(databaseID, propertyID(), valueCopy, true);
	if (!success) {
		std::string message = std::string("failed to set value ") + common::to<std::string>(valueCopy) + " to " + DataAccess::getInstance()->what(databaseID, propertyID());
		throw BaseProperty::SetValueFailure(message);
	}
}

template<typename ValueType>
void MathProperty<ValueType>::getServerValue(unsigned long databaseID) {

	std::scoped_lock lock(_mutex);
	_value = DataAccessHelper<ValueType>::getValue(databaseID, propertyID());
	_formattedValue = DataAccess::getInstance()->getFormattedValue(databaseID, propertyID());
}

template<typename ValueType>
void MathProperty<ValueType>::init(unsigned long databaseID) {
	getServerValue(databaseID);
}

template<typename ValueType>
void MathProperty<ValueType>::setValue(ValueType value) {

	std::scoped_lock lock(_mutex);
	_value = value;
}

template<typename ValueType>
ValueType MathProperty<ValueType>::getValue() {
	std::scoped_lock lock(_mutex);
	return _value;	
}

template<typename ValueType>
std::string MathProperty<ValueType>::getFormattedValue() {

	std::scoped_lock lock(_mutex);
	return _formattedValue;
}

template<typename ValueType>
std::string MathProperty<ValueType>::getStringValue() {

	std::scoped_lock lock(_mutex);
	return common::to<std::string>(_value);
}

}
