/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "fr_ill_ics_bridge_ClientCalculator.h"
#include "common/analysislibrary/MultiGaussPeakFinder.h"
#include "common/analysislibrary/ExponentialFit.h"
#include "common/analysislibrary/SimpleGaussFinder.h"
#include <iostream>
#include <blitz/array.h>

using namespace std;
jint numberOfPeaksFound;

JNIEXPORT jdoubleArray JNICALL Java_fr_ill_ics_bridge_ClientCalculator_multiGaussianFit(JNIEnv * env, jobject obj, jdoubleArray yData) {

	//get pointer to yData parameter
	jsize yLength = env->GetArrayLength(yData);
	jboolean isYDataCopied;
	jdouble *yDataPointer = env->GetDoubleArrayElements(yData, &isYDataCopied);

	//create vector and copy yData into it
	blitz::Array<double, 1> yDataVector;
	yDataVector.resize(yLength);
	for (int i = 0; i < yLength; i++) {
		yDataVector(i) = yDataPointer[i];

	}

	//release pointer
	if(isYDataCopied == JNI_TRUE){
		env->ReleaseDoubleArrayElements(yData, yDataPointer, JNI_ABORT);
	}


	MultiGaussPeakFinder gaussFinder;

	//be careful with possible exceptions when calling calculatePeaks() method
	try {
		//gaussFinder.calculatePeaks(yDataVector, xDataVector);
		gaussFinder.calculatePeaks(yDataVector);

	} catch (std::exception& e) {
		jclass exceptionClass = env ->FindClass("java/lang/IllegalArgumentException");

		if(exceptionClass){
			env -> ThrowNew( exceptionClass, "exception thrown");
			env ->DeleteLocalRef(exceptionClass);
			return 0;
		}
	}

	//get values
	numberOfPeaksFound = gaussFinder.numberOfPeaks();
	jboolean areResultsCopied;
	jdoubleArray resultsArray = env->NewDoubleArray(numberOfPeaksFound*4);
	jdouble* finalValue = (jdouble*) env ->GetPrimitiveArrayCritical(resultsArray, &areResultsCopied);

	for (int i = 0; i < gaussFinder.numberOfPeaks(); ++i) {
		Peak* P = gaussFinder.peak(i);

		finalValue[i] = P->amplitude();
		finalValue[i + 1] = P->position();
		finalValue[i + 2] = P->rms();
		finalValue[i + 3] = P->pedestal();
	}

	//release pointer
	env->ReleasePrimitiveArrayCritical(resultsArray, finalValue, 0);

	return resultsArray;
}

JNIEXPORT jint JNICALL Java_fr_ill_ics_bridge_ClientCalculator_getNumberOfPeaks (JNIEnv *env, jobject obj){

	//jint numberOfPeaks = 1;
	return numberOfPeaksFound;
}

JNIEXPORT jdoubleArray JNICALL Java_fr_ill_ics_bridge_ClientCalculator_exponentialFit(
		JNIEnv *env, jobject obj, jdoubleArray yData, jdoubleArray xData) {

	//get pointer to yData parameter
	jsize yLength = env->GetArrayLength(yData);
	jboolean isYDataCopied;
	jdouble *yDataPointer = env->GetDoubleArrayElements(yData, &isYDataCopied);

	//create vector and copy yData into it
	blitz::Array<double, 1> yDataVector;
	yDataVector.resize(yLength);
	for (int i = 0; i < yLength; i++) {
		yDataVector(i) = yDataPointer[i];

	}

	//release pointer
	if (isYDataCopied == JNI_TRUE) {
		env->ReleaseDoubleArrayElements(yData, yDataPointer, JNI_ABORT);
	}

	//same for xData array
	jsize xLength = env->GetArrayLength(xData);
	jboolean isXDataCopied;
	jdouble *xDataPointer = env->GetDoubleArrayElements(xData, &isXDataCopied);

	blitz::Array<double, 1> xDataVector;
	xDataVector.resize(xLength);
	for (int i = 0; i < xLength; i++) {
		xDataVector(i) = xDataPointer[i];
	}

	if (isXDataCopied == JNI_TRUE) {
		env->ReleaseDoubleArrayElements(xData, xDataPointer, JNI_ABORT);
	}

	//ATTENTION: I only use y values to do the fit!
	ExponentialFit exponentialFit;
//	exponentialFit.calculateFit(yDataVector, xDataVector);
	exponentialFit.calculateFit(yDataVector);

	jboolean areResultsCopied;
	jdoubleArray resultsArray = env->NewDoubleArray(4);
	jdouble* finalValue = (jdouble*) env ->GetPrimitiveArrayCritical(resultsArray, &areResultsCopied);

	finalValue[0] = exponentialFit.amplitude();
	finalValue[1] = exponentialFit.mean();
	finalValue[2] = exponentialFit.pedestal();
	finalValue[3] = 0;

	//release pointer
	env->ReleasePrimitiveArrayCritical(resultsArray, finalValue, 0);

	return resultsArray;

}

JNIEXPORT jdoubleArray JNICALL Java_fr_ill_ics_bridge_ClientCalculator_simpleGaussianFit(
		JNIEnv *env, jobject obj, jdoubleArray yData) {

	//get pointer to yData parameter
	jsize yLength = env->GetArrayLength(yData);
	jboolean isYDataCopied;
	jdouble *yDataPointer = env->GetDoubleArrayElements(yData, &isYDataCopied);

	//create vector and copy yData into it
	blitz::Array<double, 1> yDataVector;
	yDataVector.resize(yLength);
	for (int i = 0; i < yLength; i++) {
		yDataVector(i) = yDataPointer[i];

	}

	//release pointer
	if (isYDataCopied == JNI_TRUE) {
		env->ReleaseDoubleArrayElements(yData, yDataPointer, JNI_ABORT);
	}

	SimpleGaussFinder simpleGaussFinder;
	simpleGaussFinder.fit(yDataVector);

	jboolean areResultsCopied;
	jdoubleArray resultsArray = env->NewDoubleArray(4);
	jdouble* finalValue = (jdouble*) env ->GetPrimitiveArrayCritical(resultsArray, &areResultsCopied);

	finalValue[0] = simpleGaussFinder.gaussianAmplitude();
	finalValue[1] = simpleGaussFinder.position();
	finalValue[2] = simpleGaussFinder.rms();
	finalValue[3] = simpleGaussFinder.pedestal();

	//release pointer
	env->ReleasePrimitiveArrayCritical(resultsArray, finalValue, 0);

	return resultsArray;

}
