/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ServantCloner.h"
#include <core/dataaccess/DataAccess.h>
#include <command/base/ServantManager.h>
#include <ics/InstrumentManager/HardwareManager.h>
#include <ics/InstrumentManager/InstrumentManager.h>
#include <dataprovider/transportlayer/dataprovider/ServantAccessorDirectImpl.h>
#include <common/output/OutputStream.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <iostream>

using namespace std;
using namespace common;

namespace core {

ServantCloner* ServantCloner::_instance = 0;

ServantCloner::ServantCloner(dataprovider::ServantAccessorDirectImpl* proxy) :
	_proxy(proxy) {
}

ServantCloner::~ServantCloner() {
}

ServantCloner* ServantCloner::getInstance() {
	return _instance;
}

void ServantCloner::initInstance(dataprovider::ServantAccessorDirectImpl* proxy) {

	if (_instance == 0) {
		_instance = new ServantCloner(proxy);
	}
}

unsigned long ServantCloner::cloneController(unsigned long controllerID) {

	// test ics
	if (InstrumentManager::getInstance() != 0) {

		// Get the name of the controller.
		string controllerName = DataAccess::getInstance()->getServantName(0, controllerID);
		string clonedControllerName = controllerName + ".clone";
		string controllerType = DataAccess::getInstance()->getControllerType(0, controllerID);

		// Verify if the cloned controller exists.
		try {
			return DataAccess::getInstance()->getServantID(clonedControllerName);

		} catch (...) {
			// The controller does not exist, so we create it.
		}

		// Disable the configuration changed at this point and not at the beginning of the function.
		dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

		// Create the controller in the data provider.
		unsigned long clonedControllerID = _proxy->cloneController(0, controllerID);

		ExperimentController* controller = 0;

		try {
			controller = InstrumentManager::getInstance()->cloneController(controllerName);

		} catch (InstrumentManager::InstrumentManagerError const & e) {
			// Error.
		}

		// Check result.
		if (controller == nullptr) {
			// Re-enable the configuration changed.
			dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);

			// Remove the controller from data provider.
			_proxy->removeServant(0, clonedControllerID);
			return 0;
		}

		// Add the controller in the command processor.
		command::ServantManager::getInstance()->createController(controllerType, clonedControllerName, clonedControllerID);

		// Format all the values.
		controller->formatAllValues();

		// Set the controller initialised.
		_proxy->setServantInitialised(0, clonedControllerID, true);

		// Re-enable the configuration changed.
		dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
		dataprovider::NotificationServer::getInstance()->configurationChanged(0, 0, common::TREE);

		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Created clone controller " + clonedControllerName));

		return clonedControllerID;

	} else {
		err << "ServantCloner::cloneController not yet implemented for non-ics controllers" << eol;
	}
	return 0;
}

unsigned long ServantCloner::cloneOrGetController(unsigned long controllerID) {
	unsigned long id;
	map<unsigned long, unsigned long>::iterator iter = m_SaveCloneIDs.find(controllerID);
	if (iter == m_SaveCloneIDs.end()) {
		id = cloneController(controllerID);
		m_SaveCloneIDs[controllerID] = id;
		m_SaveCloneIDsCounter[controllerID] = 1;
	}
	else {
		id = iter->second;
		m_SaveCloneIDsCounter[controllerID] = m_SaveCloneIDsCounter[controllerID] + 1;
	}
	return id;
}

bool ServantCloner::removeCloneController(unsigned long cloneID) {
	bool removeec = false;
	for(map<unsigned long, unsigned long>::iterator iter = m_SaveCloneIDs.begin(); iter !=  m_SaveCloneIDs.end(); ++iter) {
		if (iter->second == cloneID) {
			m_SaveCloneIDsCounter[iter->first] = m_SaveCloneIDsCounter[iter->first] - 1;
			if (m_SaveCloneIDsCounter[iter->first] == 0) {
				_proxy->removeServant(0, cloneID);
				m_SaveCloneIDs.erase(iter);
				map<unsigned long, unsigned long>::iterator iter2 = m_SaveCloneIDsCounter.find(iter->first);
				m_SaveCloneIDsCounter.erase(iter2);
				removeec = true;
			}
			break;
		}
	}
	return removeec;
}

void ServantCloner::removeCloneControllers() {
	for(auto iter : m_SaveCloneIDs) {
		_proxy->removeServant(0, iter.second);
	}
	m_SaveCloneIDs.clear();
}

}
