/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "AtomicCommandBox.h"
#include "PropertyState.h"
#include "CommandSequencer.h"
#include "CommandBoxContainer.h"
#include "CommandZoneEventServer.h"
#include "CommandZone.h"

#include <core/scheduler/Sender.h>
#include <core/scheduler/Scheduler.h>
#include <common/base/Convert.h>
#include <core/scheduler/Scheduler.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/dataaccess/DataAccess.h>
#include <core/log/LogHandler.h>
#include <core/log/CommandZoneContentLog.h>
#include <core/log/SimpleLog.h>
#include <thread>
#include <ics/InstrumentManager/InstrumentManager.h>
#include <core/cloner/ServantCloner.h>

using namespace std;
using namespace common;

namespace core {

AtomicCommandBox::AtomicCommandBox(unsigned long commandID, bool isSettings, bool isEstimated) :
	PropertyCommandBox(isEstimated),
	_commandID(commandID),
	_isSettings(isSettings),
	_waitQueue(new Queue<bool>(1)),
	_waitingForTerminate(false),
	_isStarting(false),
	_onError(false),
	_firstCommandTerminatedNotification(false) {

	if (_forEstimatedTime == false) {
		// Subscribe to command state changes and command terminated.
		DataNotificationRouter::getInstance()->addCommandStateChangeSubscriber(_commandID, this);
		DataNotificationRouter::getInstance()->addCommandProgressChangeSubscriber(_commandID, this);
		DataNotificationRouter::getInstance()->addCommandTerminatedSubscriber(_commandID, this);
	}

	_servantID = DataAccess::getInstance()->getServantIDFromCommandID(getExecutionDatabaseID(), _commandID);
	if (DataAccess::getInstance()->isServantInDebug(getExecutionDatabaseID(), _servantID)) {
		_propertyState->setDebug();
	}
}

AtomicCommandBox::~AtomicCommandBox() {

	if (_forEstimatedTime == false) {
		// Unsubscribe from property changes.
		DataNotificationRouter::getInstance()->removePropertyChangeSubscriber(_propertyState->getPropertyIDs(), this);

		// Unsubscribe from command state changes and command terminated.
		DataNotificationRouter::getInstance()->removeCommandStateChangeSubscriber(_commandID, this);
		DataNotificationRouter::getInstance()->removeCommandProgressChangeSubscriber(_commandID, this);
		DataNotificationRouter::getInstance()->removeCommandTerminatedSubscriber(_commandID, this);
	}

	delete _waitQueue;

}


AtomicCommandBox* AtomicCommandBox::clone() const {

	AtomicCommandBox* acb = new AtomicCommandBox(_commandID, isSettings(), true);

	acb->_waitingForTerminate = _waitingForTerminate;
	acb->_isStarting = _isStarting;
	acb->_onError = _onError;
	acb->_firstCommandTerminatedNotification = _firstCommandTerminatedNotification;
	acb->_servantID = _servantID;
	acb->PropertyCommandBox::copy(this);

	ExperimentController* ec = InstrumentManager::getInstance()->getExperimentController(getServantName());
	ec->resetEstimatedTime();
	return acb;
}

unsigned long AtomicCommandBox::getCommandID() const {
	return _commandID;
}

bool AtomicCommandBox::isSettings() const {
	return _isSettings;
}

unsigned long AtomicCommandBox::getExecutionDatabaseID() const {
	return _propertyState->getExecutionDatabaseID();
}

void AtomicCommandBox::setSettingsFileName(const std::string& filename) {
	_settingsFileName = filename;
}

const std::string& AtomicCommandBox::getSettingsFileName() const {
	return _settingsFileName;
}

CommandBox::CommandBoxType AtomicCommandBox::getType() const {
	return CommandBox::ATOMIC;
}

void AtomicCommandBox::addVariableNamesToList(std::vector<string>& variableNames) {
	parent()->addVariableNamesToList(variableNames);
}

std::string AtomicCommandBox::getServantName() const {
	return DataAccess::getInstance()->getServantName(getExecutionDatabaseID(), _servantID);
}

bool AtomicCommandBox::canRun() const {
	bool canRun = Scheduler::getInstance()->canRun(getExecutionDatabaseID(), _commandID);
	if (!canRun) {
		CommandZone::getInstance()->addProblemCommand(getServantName(), DataAccess::getInstance()->getCommandName(getExecutionDatabaseID(), _commandID));
	}

	return canRun;
}

void AtomicCommandBox::setExecutionDatabaseID(unsigned long databaseID) {
	_propertyState->setExecutionDatabaseID(databaseID);
}

void AtomicCommandBox::execute() {

	initialiseExecution();

	// Launch if not on error.
	if (!_onError) {
		launchCommand();

	} else {
		// Force setRunning.
		setRunning();
		onCommandInactive();
	}
}


double AtomicCommandBox::computeEstimatedTime(bool doprogression) {

	map<string, string> buf;
	MathParserCommandPropertyState *nmath = dynamic_cast<MathParserCommandPropertyState*>(_propertyState.get());
	try {
		// Set all properties.
		nmath->setPropertiesForEstimatedTime();
		for (auto p : nmath->getPropertiesToSet()) {
			pair<string, long> prop = DataAccess::getInstance()->getCompletePropertyName(0, _servantID, p.first);
			ostringstream sbuf;
			sbuf << prop.first;
			if (prop.second >= 0) {
				sbuf << "." << prop.second;
			}
			buf[sbuf.str()] = p.second->getStringValue();
		}

	} catch (const Exception& e) {
		_onError = true;
		err << e.getMessage() << eol;
	}

	ExperimentController* ec = InstrumentManager::getInstance()->getExperimentController(getServantName());
	double value = ec->computeEstimatedTime(buf);
	if (isRunning() && doprogression) {
		if (ec->commandStatus.isRunning()) {
			value *= (100 - ec->commandProgression()) / 100.;
		}
	}
	return value;
}

void AtomicCommandBox::initialiseExecution() {

	_onError = false;

	// Sets the database ID for the subscribers.
	setDatabaseID(getExecutionDatabaseID());

	// Sets the content of the command zone to the data provider.
	LogHandler::getInstance()->write(CommandZoneContentLog(toXMLString()));

	// Temp : prevents the client from deleting the command box
	// Scenario :
	// - The command is already running
	// - The AtomicCommandBox is stopping it
	// - If the stop is not immediate, the command box must not be deleted by the client
	// We force the command box to be started
	// Question : are there some cases where we do not have to set the command box started at this point?
	// Another state STARTING would be more precise (as well as STOPPING)
	setStarting();

	// Stop command if the command is already running.
	common::CommandState state = DataAccess::getInstance()->getCommandState(getExecutionDatabaseID(), _commandID);
	if (state != common::INACTIVE) {

		string message = string("Box ") + to_string(getID()) + " (" + to_string(getCommandID()) + ") is stopping before starting, state is " + to_string((int)state);
		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, message));

		_waitingForTerminate = true;

		// It is mandatory to wait for the stop command to be terminated before launching the new command.
		// Indeed, if there is no wait, the stop can be executed after the new command is launched causing its immediate stop.
		// Without wait, we encountered the problem with a sample environment command running in background which just finished without the stop command.
		Sender::stopCommand(getExecutionDatabaseID(), _commandID, false, true);

		// Wait for the previous command termination.
		waitForCommandToTerminate();
	}

	// Update the log ID in the database.
	DataAccess::getInstance()->setCommandLogID(getExecutionDatabaseID(), _commandID, logID());

	// Initialisation can fail.
	try {
		// Set all properties.
		_propertyState->setProperties();

	} catch (const Exception& e) {
		_onError = true;
		err << e.getMessage() << eol;

		LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, getServantName() + " skipped execution : " + e.getMessage()));
	}
}

void AtomicCommandBox::launchCommand() {

	// Start the command.
	if (Scheduler::getInstance()->canRun(getExecutionDatabaseID(), _commandID)) {
		//cout << getID() << " (" << getCommandID() << ") is starting" << endl;
		_isStarting = true;
		RequestReply reply = Sender::newCommandFromCommandZone(getExecutionDatabaseID(), _commandID, this);
		if (reply.getValue() == RequestReply::REJECT) {
			common::err << DataAccess::getInstance()->what(getExecutionDatabaseID(), _commandID) << " rejected : " << reply.getReason() << common::eol;
		}

	} else {
		string servantName = getServantName();
		string commandName = DataAccess::getInstance()->getCommandName(getExecutionDatabaseID(), _commandID);
		CommandZone::getInstance()->errorOccurred(string("Cannot run commandbox for servant '") + servantName + "' and command '" + commandName + "'");
	}
}

void AtomicCommandBox::stop() {

	if (isRunning() || isPaused() || _isStarting) {
		// Stop the command.
		Sender::stopCommand(getExecutionDatabaseID(), _commandID, false);
		_isStarting = false;
	}
}

void AtomicCommandBox::stopAtEnd() {

	if (isRunning() || isPaused() || _isStarting) {
		// Stop the command.
		Sender::stopCommand(getExecutionDatabaseID(), _commandID, true);
		_isStarting = false;

		setStoppingAtEnd();
	}
}

void AtomicCommandBox::pause() {
	if (isRunning() && !isPaused()) {
		// Pause the command if possible, otherwise just send event and change state.
		if (DataAccess::getInstance()->isServantPausable(getExecutionDatabaseID(), _commandID)) {
			Sender::pauseCommand(getExecutionDatabaseID(), _commandID);
		} else {
			setPaused();
		}
	}
}

void AtomicCommandBox::pauseAtEnd() {

	if (isRunning() && !isPaused()) {
		// stop the command
		//Sender::stopCommand(getExecutionDatabaseID(), _commandID, true);

		setPausingAtEnd();
	}
}

void AtomicCommandBox::restart() {
	if (isPaused()) {
		// Restart the command if possible, otherwise just send event and change state.
		if (DataAccess::getInstance()->isServantPausable(getExecutionDatabaseID(), _commandID)) {
			Sender::restartCommand(getExecutionDatabaseID(), _commandID);
		} else {
			setRunning();
		}
	}
}

void AtomicCommandBox::reset() {

	if (_waitingForTerminate) {
		// Prevent from resetting while waiting for termination.
		return;
	}

	setConstruction();
	try {
		_propertyState->reset();

	} catch (Exception& e) {
		out << "cannot reset the property state : " << e.printStack() << eol;
	}
}

void AtomicCommandBox::changeVariableName(const std::string& formerName, const std::string& newName) {
	_propertyState->changeVariableName(formerName, newName);
}

void AtomicCommandBox::setVariable(const std::string& variableName, double value) {
	_propertyState->setVariable(variableName, value);
}

void AtomicCommandBox::setInsideForLoop() {
	_propertyState->setInsideForLoop();
}

void AtomicCommandBox::commandStateChanged(unsigned long databaseID, unsigned long commandID, common::CommandState state) {

	// A mutex is used because commandStateChanged() and commandTerminated() can execute in concurrence.
	std::scoped_lock lock(_commandMessageMutex);

	// The command execution can be terminated when receiving the notification.
	if (isRunning() || _isStarting || isPaused() || _waitingForTerminate) {
		if (/*databaseID == getExecutionDatabaseID() && */commandID == _commandID) {

			if (state == common::ACTIVE) {
				_isStarting = false;
				onCommandActive();

			} else if (state == common::INACTIVE) {

				// To call the onCommandInactive event callback, the commandStateChanged with INACTIVE
				// state and the commandTerminated messages must have been processed.
				if (_firstCommandTerminatedNotification) {
					onCommandInactive();
				} else {
					_firstCommandTerminatedNotification = true;
				}

			} else if (state == common::PAUSING) {
				setPaused();

			} else if (state == common::RESUMING) {
				setRunning();

			}
		}
	}
}

void AtomicCommandBox::commandProgressChanged(unsigned long databaseID, unsigned long commandID, double progression) {

	if ((isRunning() || isPaused()) /*&& databaseID == getExecutionDatabaseID()*/ && commandID == _commandID) {
		onProgressChanged(progression);
	}
}

void AtomicCommandBox::commandTerminated(unsigned long databaseID, unsigned long commandID, unsigned long requestID, bool sameRequest) {

	// A mutex is used because commandStateChanged() and commandTerminated() can execute in concurrence.
	std::scoped_lock lock(_commandMessageMutex);

	if (isRunning() || _isStarting || isPaused() || _waitingForTerminate) {
		if (/*databaseID == getExecutionDatabaseID() &&*/ commandID == _commandID) {

			// To call the onCommandInactive event callback, the commandStateChanged with INACTIVE
			// state and the commandTerminated messages must have been processed.
			if (_firstCommandTerminatedNotification) {
				onCommandInactive();
			} else {
				_firstCommandTerminatedNotification = true;
			}
		}
	}
}

void AtomicCommandBox::propertyChanged(unsigned long databaseID, unsigned long propertyID) {
	if ((isRunning() || isPaused()) /*&& databaseID == getExecutionDatabaseID()*/ && _propertyState->hasProperty(propertyID)) {
		_propertyState->refreshProperty(propertyID);
		CommandZoneEventServer::getInstance()->commandBoxPropertyChanged(getID(), propertyID);
	}
}


void AtomicCommandBox::onCommandActive() {
	//cout << getID() << " (" << getCommandID() << ") onCommandActive " << (int)_executionState << endl;
	if (!isRunning() && !isFinished()) {
		_propertyState->setToRunningMode();
		setRunning();
	}
}

void AtomicCommandBox::onCommandInactive() {
	//cout << getID() << " (" << getCommandID() << ") onCommandInactive" << endl;
	if (_waitingForTerminate) {
		_waitingForTerminate = false;
		_waitQueue->put(new bool());
	}
	//cout << getID() << " (" << getCommandID() << ") onCommandInactive " << isRunning() << " " << isPaused() << " " << _onError << endl;

	if (isRunning() || isPaused() || _onError) {
		_propertyState->setToFinishedMode();
		setFinished();
	}

	_firstCommandTerminatedNotification = false;

	DataAccess::getInstance()->setCommandLogID(getExecutionDatabaseID(), _commandID, "");
}


void AtomicCommandBox::waitForCommandToTerminate() {
	delete _waitQueue->get();
}

std::string AtomicCommandBox::toXMLString(const std::string& space) const {

	string controllerName = getServantName();

	return space + "<atomic controller=\"" + controllerName + "\" id=\"" + to<string>(getID()) + "\" logId=\"" + to<string>(logID()) + "\" commandId=\"" + to<string>(_commandID) + "\" isSettings=\"" + to<string>(_isSettings) + "\" settingsFile=\"" + to<string>(_settingsFileName) + "\"/>";
}

std::string AtomicCommandBox::toString(const std::string& space) const {

	string controllerName = getServantName();

	return space + controllerName + " ID = " + _propertyState->toString() + " " + std::to_string(getID());
}

}
