/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandBox.h"
#include "CommandBoxContainer.h"
#include "CommandBoxParent.h"
#include "CommandZoneEventServer.h"
#include <core/log/LogHandler.h>
#include <core/log/CommandLog.h>
#include <common/base/Date.h>
#include <common/output/OutputStream.h>

using namespace std;
using namespace common;

namespace core {

unsigned long CommandBox::_counter = 1;

CommandBox::CommandBox(bool forEstimatedTime) :
	_ID(_counter++),
	_executionType(SEQUENTIAL),
	_executionState(CONSTRUCTION),
	_isModifiable(true),
	_parent(0),
	_progression(0),
	_forEstimatedTime(forEstimatedTime) {

	// The counter starts with 1, the command box itself.
	_commandBoxCounter.store(1);
}

CommandBox::~CommandBox() {
}


void CommandBox::copy(const CommandBox* commandBox) {
	_executionType = commandBox->_executionType;
	_executionState = commandBox->_executionState;
	_isModifiable = commandBox->_isModifiable;
	_parent = commandBox->_parent;
	_progression = commandBox->_progression;
}

unsigned long CommandBox::getID() const {
	return _ID;
}

void CommandBox::toggleBackground() {
	if (_isModifiable) {
		if (_executionState == CONSTRUCTION) {
			if (_executionType == BACKGROUND) {
				_executionType = SEQUENTIAL;
			} else {
				_executionType = BACKGROUND;
			}
		}
	}
}

void CommandBox::toggleParallel() {
	if (_isModifiable) {
		if (_executionState == CONSTRUCTION) {
			if (_executionType == PARALLEL) {
				_executionType = SEQUENTIAL;
			} else {
				_executionType = PARALLEL;
				// notifying the parent
				_parent->onCommandBoxParallel(this);
			}
		}
	}
}

bool CommandBox::isBackground() const {
	return (_executionType == BACKGROUND);
}

bool CommandBox::isParallel() const {
	return (_executionType == PARALLEL);
}

CommandBox::ExecutionState CommandBox::getExecutionState() const {
	return _executionState;
}

void CommandBox::setExecutionState(ExecutionState state) {
	_executionState = state;
}


void CommandBox::setModifiable(bool isModifiable) {
	_isModifiable = isModifiable;
}

bool CommandBox::isModifiable() const {
	return _isModifiable;
}

double CommandBox::getProgression() const {
	return _progression;
}

void CommandBox::setParent(CommandBoxContainer* parent) {
	_parent = parent;
}

bool CommandBox::canBeDeleted() const {
	return (_executionState != RUNNING);
}

bool CommandBox::isConstruction() const {
	return (_executionState == CONSTRUCTION);
}

bool CommandBox::isRunning() const {
	return (_executionState == RUNNING || _executionState == STOPPING_AT_END || _executionState == PAUSING_AT_END);
}

bool CommandBox::isPaused() const {
	return (_executionState == PAUSED);
}

bool CommandBox::isPausingAtEnd() const {
	return (_executionState == PAUSING_AT_END);
}

bool CommandBox::isStoppingAtEnd() const {
	return (_executionState == STOPPING_AT_END);
}

bool CommandBox::isFinished() const {
	return (_executionState == FINISHED);
}

void CommandBox::setConstruction() {
	_executionState = CONSTRUCTION;
	CommandZoneEventServer::getInstance()->commandBoxReset(_ID);

	_progression = 0.;
}

void CommandBox::setRunning() {

	_executionState = RUNNING;

	_parent->onCommandBoxRunning(this);

	CommandZoneEventServer::getInstance()->commandBoxStarted(_ID);

}

void CommandBox::setStarting() {

	CommandZoneEventServer::getInstance()->commandBoxStarted(_ID);
}

void CommandBox::setPaused() {

	_executionState = PAUSED;

	CommandZoneEventServer::getInstance()->commandBoxPaused(_ID);
}

void CommandBox::setPausingAtEnd() {

	_executionState = PAUSING_AT_END;

	CommandZoneEventServer::getInstance()->commandBoxPausingAtEnd(_ID);
}

void CommandBox::setStoppingAtEnd() {

	_executionState = STOPPING_AT_END;

	CommandZoneEventServer::getInstance()->commandBoxStoppingAtEnd(_ID);
}

void CommandBox::setFinished() {

	// write log
	LogHandler::getInstance()->write(CommandLog(common::END, common::LOG_INFO, "", logID()));

	_executionState = FINISHED;
	CommandZoneEventServer::getInstance()->commandBoxFinished(_ID);

	// notify container
	_progression = 100.;
	_parent->onProgressChanged(this);
	_parent->onCommandBoxFinished(this);


}

void CommandBox::onProgressChanged(double progression) {

	// first see what execution state we have
	if (isRunning() || isPaused()) {
		_progression = progression;
	}

	// next modify the progression depending on the type of execution
	if (isBackground()) {
		// always send 100% progression
		_progression = 100;
	}

	// notify container
	_parent->onProgressChanged(this);
}

long CommandBox::getProgression() {
	// first see what execution state we have
	if (isRunning() || isPaused()) {
		return (long)_progression;

	} else if (isFinished()) {
		return 100;
	}

	return 0;
}

CommandBoxContainer* CommandBox::parent() {
	return _parent;
}

void CommandBox::setLogID() {
	_logID = common::Date().toString();
}

const std::string& CommandBox::logID() const {
	return _logID;
}

int CommandBox::getCommandBoxCounter() const {
	return _commandBoxCounter.load();
}

void CommandBox::updateCommandBoxCounter(int count) {
	_commandBoxCounter += count;

	if (_parent != nullptr) {
		CommandBox* commandBoxParent = dynamic_cast<CommandBox*>(_parent->getParent());
		if (commandBoxParent != nullptr) {
			commandBoxParent->updateCommandBoxCounter(count);
		}
	}
}

}
