/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_COMMANDBOX_H_
#define CORE_COMMANDBOX_H_

#include <string>
#include <vector>
#include <set>
#include <atomic>

namespace core {

class CommandBoxContainer;

/**
 * Class defining a command box.
 */
class CommandBox {

	friend class CommandBoxContainer;

public:

	/**
	 * Command box type.
	 */
	typedef enum {ATOMIC, SCAN, FOR_LOOP, GENERIC, IF, BREAK, IF_ELSE} CommandBoxType;
	
	/**
	 * Execution type.
	 */
	typedef enum {PARALLEL, BACKGROUND, SEQUENTIAL} ExecutionType;
	
	/**
	 * Execution state.
	 */
	typedef enum {CONSTRUCTION, RUNNING, PAUSED, STOPPING_AT_END, PAUSING_AT_END, FINISHED} ExecutionState;

	/**
	 * Constructor.
	 */
	CommandBox(bool forEstimatedTime = false);

	/**
	 * Virtual destructor.
	 */
	virtual ~CommandBox();

	/**
	 * clone.
	 */
	void copy(const CommandBox* commandBox);

	/**
	 * Returns the ID of the command box.
	 */
	unsigned long getID() const;

	/**
	 * Returns the type of command box.
	 */
	virtual CommandBoxType getType() const = 0;

	/**
	 * Toggle the background execution type.
	 */
	void toggleBackground();
	
	/**
	 * Toggle the parallel execution type.
	 */
	void toggleParallel();

	/**
	 * Returns true if running in background.
	 */
	bool isBackground() const;

	/**
	 * Returns true if running in parallel.
	 */
	bool isParallel() const;
	
	/**
	 * Returns the execution state.
	 */
	ExecutionState getExecutionState() const;
	
	/**
	 * Set the possibility to modify or not the command box.
	 */
	void setModifiable(bool isModifiable);

	/**
	 * Returns true if the command box is modifiable.
	 */
	bool isModifiable() const;

	/**
	 * Sets the parent command box container.
	 */
	void setParent(CommandBoxContainer* parent);

	/**
	 * Returns true if the command box can run.
	 */
	virtual bool canRun() const = 0;

	/**
	 * Returns true if the command box can be deleted.
	 */
	bool canBeDeleted() const;

	/**
	 * Sets the database ID to be used during the execution.
	 */
	virtual void setExecutionDatabaseID(unsigned long databaseID) = 0;

	/**
	 * Execute the command box.
	 */
	virtual void execute() = 0;

	/**
	 * Initialises execution.
	 */
	virtual void initialiseExecution() = 0;

	/**
	 * Launches the command associated with the command box.
	 */
	virtual void launchCommand() = 0;

	/**
	 * Stops the command box.
	 */
	virtual void stop() = 0;

	/**
	 * Stops the command box at end.
	 */
	virtual void stopAtEnd() = 0;

	/**
	 * Pauses the command box.
	 */
	virtual void pause() = 0;

	/**
	 * Pauses the command box at end.
	 */
	virtual void pauseAtEnd() = 0;

	/**
	 * Restarts the command box.
	 */
	virtual void restart() = 0;

	/**
	 * Resets the command box.
	 */
	virtual void reset() = 0;

	/**
	 * Returns the progression of the command box.
	 */
	long getProgression();

	/**
	 * Returns true if construction.
	 */
	bool isConstruction() const;

	/**
	 * Returns true if running.
	 */
	bool isRunning() const;

	/**
	 * Returns true if paused.
	 */
	bool isPaused() const;
	
	/**
	 * Returns true if pausing at end.
	 */
	bool isPausingAtEnd() const;

	/**
	 * Returns true if stopping at end.
	 */
	bool isStoppingAtEnd() const;

	/**
	 * Returns true if finished.
	 */
	bool isFinished() const;
	
	/**
	 * Returns the progression.
	 */
	double getProgression() const;

	/**
	 * Adds the variable names to the list.
	 */
	virtual void addVariableNamesToList(std::vector<std::string>& variableNames) = 0;

	/**
	 * Needed for for loops. sets the given variable name with the value.
	 * Iterates through command containers if necessary.
	 */
	virtual void setVariable(const std::string& variableName, double value) = 0; 

	/**
	 * Changes the variable name and updates any expressions that use it.
	 */
	virtual void changeVariableName(const std::string& formerName, const std::string& newName) = 0;

	/**
	 * Returns true if the box references a user variable.
	 */
	virtual bool referencesUserVariable(const std::string& variableName) = 0;

	/**
	 * Returns the list of referenced user variables.
	 */
	virtual std::set<std::string> getReferencedUserVariables() = 0;

	/**
	 * Sets inside for loop.
	 */
	virtual void setInsideForLoop() = 0;

	/**
	 * Sets a new log ID.
	 */
	virtual void setLogID();
	
	/**
	 * Prints the content of the box.
	 */
	virtual std::string toXMLString(const std::string& space = std::string()) const = 0;

	/**
	 * Prints the content of the box.
	 */
	virtual std::string toString(const std::string& space = std::string()) const = 0;

	/**
	 * Gets the box counter.
	 */
	int getCommandBoxCounter() const;

	/**
	 * Compute the estimated time of the command box.
	 */
	virtual double computeEstimatedTime(bool doprogression) = 0;

protected:
	/**
	 * Sets the execution state.
	 */
	void setExecutionState(ExecutionState state);

	/**
	 * Sets the command box to construction and executes necessary operations.
	 */
	void setConstruction();

	/**
	 * Sets the command box started.
	 */
	void setStarting();

	/**
	 * Sets the command box to running and executes necessary operations.
	 */
	void setRunning();

	/**
	 * Sets the command box to paused and executes necessary operations.
	 */
	void setPaused();
	
	/**
	 * Sets the command box to pausing at end and executes necessary operations.
	 */
	void setPausingAtEnd();

	/**
	 * Sets the command box to stopping at end and executes necessary operations.
	 */
	void setStoppingAtEnd();

	/**
	 * Sets the command box to finished and executes necessary operations.
	 */
	void setFinished();
	
	/**
	 * Does the necessary operations when the progress changes.
	 */
	void onProgressChanged(double progression);

	/**
	 * Gets the parent.
	 */
	CommandBoxContainer* parent();

	/**
	 * Returns the log ID.
	 */
	const std::string& logID() const;

	/**
	 * Updates the command box counter by count that can be negative if a command box is deleted.
	 */
	void updateCommandBoxCounter(int count);

private:
	static unsigned long _counter;
	unsigned long _ID;
	
	ExecutionType _executionType;

protected:
	ExecutionState _executionState;
	
	bool _isModifiable;
	CommandBoxContainer* _parent;
	double _progression;
	std::string _logID;
	std::atomic_int _commandBoxCounter; // The number of command boxes contained including itself.
	bool _forEstimatedTime;

};

}


#endif /*CORE_COMMANDBOX_H_*/
