/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandBoxContainer.h"
#include "CommandBoxMap.h"
#include "CommandBox.h"
#include "CommandSequencer.h"
#include "CommandBoxParent.h"
#include "CommandBoxFactory.h"
#include "ParallelGroup.h"
#include "CommandZoneEventServer.h"
#include <common/output/OutputStream.h>
#include <iostream>
#include <Utilities/Counter.h>

using namespace std;
using namespace boost;
using namespace common;

namespace core {

std::recursive_mutex CommandBoxContainer::_onCommandBoxMutex;

CommandBoxContainer::CannotRunException::CannotRunException() :
	Exception("core::CommandBoxContainer::CannotRunException", string("a command cannot run")) {
}

CommandBoxContainer::CommandBoxContainer(CommandBoxParent* parent) :
	_parent(parent),
	_parallelGroup(new ParallelGroup()),
	_isStopped(false),
	_isStoppingAtEnd(false),
	_isRunning(false),
	_isPaused(false),
	_isPausingAtEnd(false),
	_insideForLoop(false) {
}

CommandBoxContainer::~CommandBoxContainer() {

	// Set the mutex so that it is unlocked when arriving here when the command box is removed by the automatic removal.
	// Otherwise the mutex is destroyed at the end of the destructor that is not allowed if it is already locked.
	std::scoped_lock lock(_onCommandBoxMutex);

	for (vector<CommandBox*>::iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		CommandBoxMap::getInstance()->remove(*b);
	}

	delete _parallelGroup;
}

void CommandBoxContainer::copy(const CommandBoxContainer* commandBoxContainer) {

	_isStopped = commandBoxContainer->_isStopped;
	_isStoppingAtEnd = commandBoxContainer->_isStoppingAtEnd;
	_isRunning = commandBoxContainer->_isRunning;
	_isPaused = commandBoxContainer->_isPaused;
	_isPausingAtEnd = commandBoxContainer->_isPausingAtEnd;
	_insideForLoop = commandBoxContainer->_insideForLoop;

	for(auto box : commandBoxContainer->_commandBoxes) {
		Counter c1;
		CommandBox* clonebox = nullptr;
		if (box->isBackground() == false) {
			clonebox = CommandBoxFactory::getInstance()->cloneAtomicCommandBox(box);
			if (clonebox == nullptr) {
				clonebox = CommandBoxFactory::getInstance()->cloneScanCommandBox(box);
				if (clonebox == nullptr) {
					clonebox = CommandBoxFactory::getInstance()->cloneForLoopCommandBox(box);
					if (clonebox == nullptr) {
						clonebox = CommandBoxFactory::getInstance()->cloneGenericCommandBox(box);
						if (clonebox == nullptr) {
							clonebox = CommandBoxFactory::getInstance()->cloneControlCommandBox(box);
						}
					}
				}
			}
			clonebox->setParent(this);
			_commandBoxes.push_back(clonebox);
			CommandBoxMap::getInstance()->add(clonebox);
		}
	}
}

CommandBoxParent* CommandBoxContainer::getParent() const {
	return _parent;
}

const std::vector<CommandBox*>& CommandBoxContainer::getCommandBoxes() const {
	return _commandBoxes;
}

void CommandBoxContainer::updateCommandBoxCounter(int count) {

	CommandBox* commandBoxParent = dynamic_cast<CommandBox*>(_parent);
	if (commandBoxParent != nullptr) {
		commandBoxParent->updateCommandBoxCounter(count);
	}
}

unsigned long CommandBoxContainer::addCommandBoxAtEnd(CommandBox* commandBox) {
	_commandBoxes.push_back(commandBox);

	if (_insideForLoop) {
		commandBox->setInsideForLoop();
	}

	commandBox->setParent(this);
	CommandBoxMap::getInstance()->add(commandBox);

	// Increase the counter.
	updateCommandBoxCounter(commandBox->getCommandBoxCounter());

	return commandBox->getID();
}

std::vector<CommandBox*>& CommandBoxContainer::commandBoxes() {
	return _commandBoxes;
}

bool CommandBoxContainer::isInsideForLoop() const {
	return _insideForLoop;
}

bool CommandBoxContainer::canRun() const {
	for (vector<CommandBox*>::const_iterator c = _commandBoxes.begin(); c != _commandBoxes.end(); c++) {
		if (!(*c)->canRun()) {
			return false;
		}
	}
	return true;
}

void CommandBoxContainer::setExecutionDatabaseID(unsigned long databaseID) {
	for (vector<CommandBox*>::const_iterator c = _commandBoxes.begin(); c != _commandBoxes.end(); c++) {
		(*c)->setExecutionDatabaseID(databaseID);
	}
}

void CommandBoxContainer::startExecution() {

	// Verify that the command container is not running.
	if (hasCommandBoxRunningIgnoringBackground() && !_isPaused) {
		return;
	}

	// Find first command to run.
	vector<CommandBox*>::const_iterator it;
	for (it = _commandBoxes.begin(); it != _commandBoxes.end(); it++) {

		if ((*it)->isBackground()) {
			if ((*it)->isConstruction()) {
				break;
			}
			// ... ?
		}

		if (!(*it)->isBackground()) {
			if (!(*it)->isFinished()) {
				break;
			}
		}
	}

	if (it == _commandBoxes.end()) {
		return;
	}

	// Verify for command conflicts.
	bool canRun = true;
	for (vector<CommandBox*>::const_iterator rit = it; rit != _commandBoxes.end(); rit++) {
		canRun &= (*rit)->canRun();
	}
	if (!canRun) {
		throw CannotRunException();
	}

	// Launch parallel or sequential.
	CommandBox* commandBox = *it;

	_isStopped = false;
	_isStoppingAtEnd = false;
	_isPaused = false;
	_isPausingAtEnd = false;

	if (commandBox->isParallel()) {

		fillParallelGroupFromCommandBox(commandBox, _parallelGroup);
		_parallelGroup->execute();

	} else {
		CommandSequencer::getInstance()->executeCommandBox(commandBox);
	}
}

bool CommandBoxContainer::executeNext(CommandBox* commandBox) {

	// Find iterator associated with commandBox and start the next one.
	vector<CommandBox*>::const_iterator b;
	for (b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		if ((*b)->getID() == commandBox->getID()) {
			break;
		}
	}

	// Execute next command box if it wasn't the last one.
	b++;
	if (b != _commandBoxes.end()) {

		CommandBox* nextCommandBox = *(b);

		// Check if the next command box is executed now or postponed.
		if (!_isStopped && !_isPaused) {

			// Check if execution is parallel or not.
			if (nextCommandBox->isParallel()) {

				// Add the command box and the following parallel command boxes.
				fillParallelGroupFromCommandBox(nextCommandBox, _parallelGroup);
				// Execute the added command boxes.
				_parallelGroup->execute();

			} else {
				CommandSequencer::getInstance()->executeCommandBox(nextCommandBox);
			}
		}

		// There is a new command to execute now or later.
		return true;
	}

	// There is no command to execute.
	return false;
}

void CommandBoxContainer::stop() {

	_isStopped = true;
	_isStoppingAtEnd = false;
	_isRunning = false;
	_isPaused = false;
	_isPausingAtEnd = false;

	if (hasCommandBoxRunningOrPaused()) {
		for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {

			// Stopping a container sets the command boxes to finished
			// even those that don't run
			//if (!(*b)->isRunning() && !(*b)->isPaused()) {
			//	(*b)->setFinished();
			//}
			(*b)->stop();
		}
	} else {
		_parent->onExecutionEnd();
	}
}

void CommandBoxContainer::stopAtEnd() {

	_isStopped = true;
	_isStoppingAtEnd = true;
	_isRunning = false;
	_isPaused = false;

	if (hasCommandBoxRunningOrPaused()) {
		for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
			(*b)->stopAtEnd();
		}
	} else {
		_parent->onExecutionEnd();
	}
}

void CommandBoxContainer::pause() {

	_isPaused = true;
	_isPausingAtEnd = false;

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		// Do not pause background commands.
		if (!(*b)->isBackground()) {
			(*b)->pause();
		}
	}

}

void CommandBoxContainer::pauseAtEnd() {

	_isPaused = true;
	_isPausingAtEnd = true;

	if (hasCommandBoxRunningOrPaused()) {
		for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
			(*b)->pauseAtEnd();
		}
	} else {
		// ???
		//_parent->onExecutionEnd();
	}
}

void CommandBoxContainer::restart() {

	_isPaused = false;
	_isPausingAtEnd = false;

	if (hasCommandBoxPausedIgnoringBackground()) {
		// Iterate over command list and continue all (finds which ones were previously running).
		for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
			(*b)->restart();
		}

	} else if (!allCommandsFinishedIgnoringBackground()) {
		startExecution();

	} else {
		_parent->onExecutionEnd();
	}
}

void CommandBoxContainer::reset() {
	if (!hasCommandBoxRunningIgnoringBackground() && !hasCommandBoxPausedIgnoringBackground()) {
		for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
			(*b)->reset();
		}
	}
	_isRunning = false;
}

void CommandBoxContainer::onCommandBoxRunning(CommandBox* commandBox) {

	std::scoped_lock lock(_onCommandBoxMutex);

	// Determine if its the first command box to start.
	if (!_isRunning) {
		_isRunning = true;
		_parent->onExecutionBegin();
	}

	if (commandBox->isBackground()) {
		// Background command so execute the next command.
		executeNext(commandBox);
	}
}

void CommandBoxContainer::onCommandBoxPaused(CommandBox* commandBox) {

	_parent->onExecutionPause();
}

void CommandBoxContainer::onCommandBoxFinished(CommandBox* commandBox) {

	std::scoped_lock lock(_onCommandBoxMutex);

	bool nextCommandExists = false;

	if (commandBox->isParallel()) {
		_parallelGroup->onCommandBoxFinished(commandBox);

		if (_parallelGroup->isCompleted()) {
			CommandBox* lastCommandBox = _parallelGroup->getLastCommandBox();
			 _parallelGroup->reset();
			 nextCommandExists = executeNext(lastCommandBox);
		}

	} else if (!commandBox->isBackground()) {
		// Sequential command so execute the next command.
		nextCommandExists = executeNext(commandBox);
	}

	//cout << "onCommandBoxFinished " << commandBox->getID() << " : " << hasCommandBoxRunningIgnoringBackground() << " " << allCommandsFinishedIgnoringBackground() << " " << _isStopped << " " << _isPausingAtEnd << " " << _isPaused << " " << nextCommandExists << " " << _parent->hasEnded() << endl;

	// The next command can exist in case of a for loop for which the container has finished (nextCommandExists is false) but there remains some iterations.
	nextCommandExists = nextCommandExists || !_parent->hasEnded();

	// Determine if it is the last command box to end.
	// No test for allCommandsFinishedIgnoringBackground, because of the non started commands.
	bool executionStopped = !hasCommandBoxRunningIgnoringBackground() && (_isStopped || _isPaused || allCommandsFinishedIgnoringBackground());

	// Check if execution is stopped.
	if (executionStopped) {

		// Check if is paused and there is not a next command to execute.
		// If the next command does not exist, then the pause is considered a stop.
		if (_isPaused && nextCommandExists) {
			_parent->onExecutionPause();
			_isRunning = false;
			_isPausingAtEnd = false;
		}
		else {
			_parent->onExecutionEnd();
			_isRunning = false;
			_isPaused = false;
			_isStoppingAtEnd = false;
			_isPausingAtEnd = false;
		}
	}

	_parent->onCommandBoxFinished(commandBox);
}

void CommandBoxContainer::onCommandBoxParallel(CommandBox* commandBox) {

	std::scoped_lock lock(_onCommandBoxMutex);

	// Add the command box to the group if it is next to the last command box
	// of the group.
	if (!_parallelGroup->isCompleted()) {

		CommandBox* lastCommandBox = _parallelGroup->getLastCommandBox();
		bool next = false;
		for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
			if (!next) {
				if (*b == lastCommandBox) {
					next = true;
				}
			} else {
				if (*b == commandBox) {
					// Add the command box and the following parallel command boxes.
					fillParallelGroupFromCommandBox(commandBox, _parallelGroup);
					// Execute the added command boxes.
					_parallelGroup->execute();

				} else {
					break;
				}
			}
		}
	}

}

void CommandBoxContainer::onProgressChanged(CommandBox* commandBox) {

	std::scoped_lock lock(_onCommandBoxMutex);

	double progression = commandBox->getProgression();

	// Calculates the progression for the parallel group and notifies
	// the clients of command box progression change.
	if (commandBox->isParallel()) {

		// Test the existence of the last command box.
		// Indeed this call can come from a deeper command box running in background.
		// The container may have finished (this parallel group has been reset) but the command box running in background is still sending progress events.
		if (_parallelGroup->getLastCommandBox() != nullptr) {
			progression = _parallelGroup->calculateMinimumProgression();
			CommandZoneEventServer::getInstance()->commandBoxProgressionChanged(_parallelGroup->getLastCommandBox()->getID(), progression);
		}

	} else {
		CommandZoneEventServer::getInstance()->commandBoxProgressionChanged(commandBox->getID(), progression);
	}

	// If stopped, do not send 100% to the parent as it was formerly done.

	double totalProgression = 0;
	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		totalProgression += (*b)->getProgression();
	}

	totalProgression = totalProgression / _commandBoxes.size();

	_parent->onContainerProgressChanged(totalProgression);
}

void CommandBoxContainer::addVariableNamesToList(std::vector<string>& variableNames) {
	_parent->addVariableNamesToList(variableNames);
}

bool CommandBoxContainer::isRunning() const {
	return _isRunning;
}

bool CommandBoxContainer::isPaused() const {
	return _isPaused;
}

bool CommandBoxContainer::isPausingAtEnd() const {
	return _isPausingAtEnd;
}

bool CommandBoxContainer::isStopped() const {
	return _isStopped;
}

bool CommandBoxContainer::isStoppingAtEnd() const {
	return _isStoppingAtEnd;
}

bool CommandBoxContainer::hasCommandBoxRunningOrPaused() const {

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		if ((*b)->isRunning() || (*b)->isPaused()) {
			return true;
		}
	}

	return false;
}

bool CommandBoxContainer::hasCommandBoxRunningIgnoringBackground() const {

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		if ((*b)->isRunning() && !(*b)->isBackground()) {
			return true;
		}
	}

	return false;
}

bool CommandBoxContainer::hasCommandBoxPausedIgnoringBackground() const {

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		if ((*b)->isPaused() && !(*b)->isBackground()) {
			return true;
		}
	}

	return false;
}

bool CommandBoxContainer::allCommandsFinishedIgnoringBackground() const {

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		if (!(*b)->isFinished() && !(*b)->isBackground()) {
			return false;
		}
	}

	return true;
}

void CommandBoxContainer::fillParallelGroupFromCommandBox(CommandBox* commandBox, ParallelGroup* parallelGroup) {

	bool fill = false;
	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {

		if (!fill) {
			if (*b == commandBox) {
				fill = true;
				parallelGroup->addCommandBox(*b);
			}
		} else {
			if ((*b)->isParallel()) {
				parallelGroup->addCommandBox(*b);
			} else {
				break;
			}
		}
	}
}

void CommandBoxContainer::setVariable(const std::string& variableName, double value) {

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		(*b)->setVariable(variableName, value);
	}
}

void CommandBoxContainer::changeVariableName(const std::string& formerName, const std::string& newName) {

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		(*b)->changeVariableName(formerName, newName);
	}
}

bool CommandBoxContainer::referencesUserVariable(const std::string& variableName) {

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		if ((*b)->referencesUserVariable(variableName)) {
			return true;
		}
	}

	return false;
}

std::set<std::string> CommandBoxContainer::getReferencedUserVariables() {

	set<string> result;

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		set<string> commandBoxVariables = (*b)->getReferencedUserVariables();
		result.insert(commandBoxVariables.begin(), commandBoxVariables.end());
	}

	return result;
}

void CommandBoxContainer::setInsideForLoop() {
	_insideForLoop = true;
}

void CommandBoxContainer::setLogIDs() {
	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		(*b)->setLogID();
	}
}

std::string CommandBoxContainer::toXMLString(const std::string& space) const {

	std::string result;

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		result += (*b)->toXMLString(space + "\t") + "\n";
	}

	return result;
}

std::string CommandBoxContainer::toString(const std::string& space) const {

	std::string result;

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		result += (*b)->toString(space + "\t") + "\n";
	}
	return result;
}

std::string CommandBoxContainer::print() const {

	std::string result;

	for (vector<CommandBox*>::const_iterator b = _commandBoxes.begin(); b != _commandBoxes.end(); b++) {
		result += (*b)->toString() + "\n";
	}

	return result;
}

double CommandBoxContainer::computeEstimatedTime(bool doprogression) {
	double value = 0.;
	// find first command to run
	vector<double> parrallelTimes;
	for (auto commandBox : _commandBoxes) {
		if (commandBox->isBackground()) {
			// No estimated on background
		}
		else {
			if (!commandBox->isFinished()) {
				if (commandBox->isParallel()) {
					parrallelTimes.push_back(commandBox->computeEstimatedTime(doprogression));
				}
				else {
					if (parrallelTimes.empty() == false) {
						double maxtime = -numeric_limits<float64>::max();
						for(auto partime : parrallelTimes) {
							if (partime > maxtime) {
								maxtime = partime;
							}
						}
						value += maxtime;
						parrallelTimes.clear();
					}
					value += commandBox->computeEstimatedTime(doprogression);
				}
			}
		}
	}
	// Check if parallel isn't empty
	if (parrallelTimes.empty() == false) {
		double maxtime = -numeric_limits<float64>::max();
		for(auto partime : parrallelTimes) {
			if (partime > maxtime) {
				maxtime = partime;
			}
		}
		value += maxtime;
	}
	return value;
}


}
