/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_COMMANDBOXCONTAINER_H_
#define CORE_COMMANDBOXCONTAINER_H_

#include <common/base/Exception.h>
#include <mutex>
#include <vector>
#include <set>
#include <string>

namespace core {

class CommandBox;
class CommandBoxParent;
class ParallelGroup;

/**
 * Class defining a command box container.
 */
class CommandBoxContainer {

public:
	
	/**
	 * Cannot run exception.
	 */
	class CannotRunException : public common::Exception {
	
	public:
		/**
		 * Constructor.
		 */
		CannotRunException();
	};
	
	/**
	 * Constructor.
	 */
	CommandBoxContainer(CommandBoxParent* parent);

	/**
	 * Virtual destructor. Deletes all the command boxes.
	 */
	virtual ~CommandBoxContainer();
	
	/**
	 * copy.
	 */
	void copy(const CommandBoxContainer* commandBoxContainer);

	/**
	 * Returns the parent.
	 */
	CommandBoxParent* getParent() const;

	/**
	 * Returns the command boxes.
	 */
	const std::vector<CommandBox*>& getCommandBoxes() const;
	
	/**
	 * Adds a new command box to the end of the container. 
	 * Returns the ID of the created command box.
	 */
	unsigned long addCommandBoxAtEnd(CommandBox* commandBox);

	/**
	 * Returns true if all of the contents can run.
	 */
	bool canRun() const;

	/**
	 * Sets the database ID to be used during the execution.
	 */
	void setExecutionDatabaseID(unsigned long databaseID);

	/**
	 * Starts the execution by executing the first command box.
	 */
	void startExecution();
	
	/**
	 * Executes the next command box.
	 * Returns true if there is a new command to execute.
	 */
	bool executeNext(CommandBox* commandBox);

	/**
	 * Stops the execution.
	 */
	void stop();

	/**
	 * Stops the execution at end.
	 */
	void stopAtEnd();

	/**
	 * Pauses the execution.
	 */
	void pause();

	/**
	 * Pauses the execution at end.
	 */
	void pauseAtEnd();

	/**
	 * Restarts the execution.
	 */
	void restart();

	/**
	 * Resets the content.
	 */
	void reset();

	/**
	 * Notifies the container that one if its command boxes has started.
	 */
	void onCommandBoxRunning(CommandBox* commandBox);
	
	/**
	 * Notifies the container that one if its command boxes has been paused.
	 */
	void onCommandBoxPaused(CommandBox* commandBox);
	
	/**
	 * Notifies the container that one if its command boxes has finished.
	 */
	void onCommandBoxFinished(CommandBox* commandBox);

	/**
	 * Notifies the container that one if its command boxes has toggled parallel.
	 */
	void onCommandBoxParallel(CommandBox* commandBox);

	/**
	 * Notifies the container that one if its command boxes has changed progress.
	 */
	void onProgressChanged(CommandBox* commandBox);

	/**
	 * Returns true if is running.
	 */
	bool isRunning() const;

	/**
	 * Returns true if is paused.
	 */
	bool isPaused() const;

	/**
	 * Returns true if is pausing at end.
	 */
	bool isPausingAtEnd() const;

	/**
	 * Returns true if is stopped.
	 */
	bool isStopped() const;

	/**
	 * Returns true if is stopping at end.
	 */
	bool isStoppingAtEnd() const;

	/**
	 * Adds the variable names to the list.
	 */
	void addVariableNamesToList(std::vector<std::string>& variableNames);

	/**
	 * Changes the variable name and updates any expressions that use it.
	 */
	void changeVariableName(const std::string& formerName, const std::string& newName);

	/**
	 * Returns true if the box references a user variable.
	 */
	virtual bool referencesUserVariable(const std::string& variableName);

	/**
	 * Returns the list of referenced user variables.
	 */
	virtual std::set<std::string> getReferencedUserVariables();

	/**
	 * Sets inside for loop.
	 */
	void setInsideForLoop();

	/**
	 * Needed for for loops. Iterates through command setting the 
	 * given variable name with the value.
	 */
	void setVariable(const std::string& variableName, double value);

	/**
	 * Sets a new log IDs for all command boxes.
	 */
	void setLogIDs();

	/**
	 * Prints the content of the container.
	 */
	std::string toXMLString(const std::string& space = std::string()) const;

	/**
	 * Prints the content of the container.
	 */
	std::string toString(const std::string& space = std::string()) const;

	/**
	 * Prints the content of the container.
	 */
	std::string print() const;

	/**
	 * Compute Estimated Time.
	 */
	double computeEstimatedTime(bool doprogression);

protected:
	std::vector<CommandBox*>& commandBoxes();
	bool isInsideForLoop() const;

	/**
	 * Updates the command box counter by count that can be negative if a command box is deleted.
	 */
	void updateCommandBoxCounter(int count);

	// The mutex is static so that there is only one mutex for all the containers.
	//
	// Indeed having only one mutex solves potential deadlocks when we have two nested containers.
	// Suppose we have one mutex per container and the CommandZone:
	// Forloop 1
	//   Forloop 2
	//     Count
	// The call AtomicCommandBox::onCommandInactive() goes up in the CommandZone hierarchy. Then mutex 2 of Forloop 2 is locked before mutex 1 of Forloop 1.
	// When we clone the CommandZone, mutex 1 is locked before mutex 2. In parallel, we can have a deadlock.
	static std::recursive_mutex _onCommandBoxMutex;
	
private:
	bool hasCommandBoxRunningOrPaused() const;
	bool hasCommandBoxRunningIgnoringBackground() const;
	bool hasCommandBoxPausedIgnoringBackground() const;
	bool allCommandsFinishedIgnoringBackground() const;
	void fillParallelGroupFromCommandBox(CommandBox* commandBox, ParallelGroup* parallelGroup);

	CommandBoxParent* _parent;
	std::vector<CommandBox*> _commandBoxes;

	ParallelGroup* _parallelGroup;

	bool _isStopped;
	bool _isStoppingAtEnd;

	bool _isRunning;
	bool _isPaused;
	bool _isPausingAtEnd;

	bool _insideForLoop;
};

}


#endif /*CORE_COMMANDBOXCONTAINER_H_*/
