/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandBoxFactory.h"
#include <core/commandzone/generic/FixCommandBox.h>
#include <core/commandzone/generic/ClearCommandBox.h>
#include <core/commandzone/generic/SetZeroCommandBox.h>
#include <core/commandzone/generic/PrintCommandBox.h>
#include <core/commandzone/generic/SetCommandBox.h>
#include <core/commandzone/generic/VariableCommandBox.h>
#include <core/commandzone/control/IfCommandBox.h>
#include <core/commandzone/control/BreakCommandBox.h>
#include <core/dataaccess/DataAccess.h>
#include <common/base/Convert.h>
#include <common/output/OutputStream.h>

using namespace common;
using namespace std;

namespace core {

CommandBoxFactory::NoSuchCommandException::NoSuchCommandException(unsigned long commandID) :
	Exception("core::CommandBoxFactory::NoSuchCommandException", "no such command with ID <" + to<string>(commandID) + ">") {
}

CommandBoxFactory* CommandBoxFactory::_instance = 0;

CommandBoxFactory::CommandBoxFactory() {
}

CommandBoxFactory::~CommandBoxFactory() {
}

CommandBoxFactory* CommandBoxFactory::getInstance() {
	if (_instance == 0) {
		_instance = new CommandBoxFactory();
	}
	return _instance;
}

AtomicCommandBox* CommandBoxFactory::createAtomicCommandBox(const std::string& controllerName, bool isSettings) {
	
	// Search for the command ID of the start command.
	unsigned long controllerID = DataAccess::getInstance()->getServantID(controllerName);
	unsigned long commandID = DataAccess::getInstance()->getCommandID(0, controllerID, "start");

	if (DataAccess::getInstance()->commandExists(commandID)) {
		return new AtomicCommandBox(commandID, isSettings, false);
	}
	
	throw NoSuchCommandException(commandID);
}

ScanCommandBox* CommandBoxFactory::createScanCommandBox(const std::string& scanName, const std::string& fullpath) {
	return new ScanCommandBox(scanName, fullpath);
}

ForLoopCommandBox* CommandBoxFactory::createForLoopCommandBox(const std::string& variableName) {
	return new ForLoopCommandBox(variableName);
}

CommandBox* CommandBoxFactory::createGenericCommandBox(PropertyGenericCommandBox::BoxType type) {

	switch (type) {
	case GenericCommandBox::FIX:
		return new FixCommandBox();
	case GenericCommandBox::CLEAR:
		return new ClearCommandBox();
	case GenericCommandBox::SETZERO:
		return new SetZeroCommandBox();
	case GenericCommandBox::PRINT:
		return new PrintCommandBox();
	case GenericCommandBox::SET:
		return new SetCommandBox();
	case GenericCommandBox::VARIABLE:
		return new VariableCommandBox();
	default:
		err << "Cannot create generic command box with type " << (long)type << eol;
	}

	return 0;
}

ControlCommandBox* CommandBoxFactory::createControlCommandBox(ControlCommandBox::BoxType type) {

	switch (type) {
	case ControlCommandBox::IF:
		return new IfCommandBox(ControlCommandBox::IF);
	case ControlCommandBox::IF_ELSE:
		return new IfCommandBox(ControlCommandBox::IF_ELSE);
	case ControlCommandBox::BREAK:
		return new BreakCommandBox();
	default:
		err << "Cannot create control command box with type " << (long)type << eol;
	}

	return 0;
}

AtomicCommandBox* CommandBoxFactory::cloneAtomicCommandBox(const CommandBox* commandBox) {
	const AtomicCommandBox* boxe = dynamic_cast<const AtomicCommandBox*>(commandBox);
	if (boxe != nullptr) {
		return boxe->clone();
	}
	else {
		return nullptr;
	}
}

ScanCommandBox* CommandBoxFactory::cloneScanCommandBox(const CommandBox* commandBox) {
	const ScanCommandBox* boxe = dynamic_cast<const ScanCommandBox*>(commandBox);
	if (boxe != nullptr) {
		return boxe->clone();
	}
	else {
		return nullptr;
	}
}

ForLoopCommandBox* CommandBoxFactory::cloneForLoopCommandBox(const CommandBox* commandBox) {
	const ForLoopCommandBox* boxe = dynamic_cast<const ForLoopCommandBox*>(commandBox);
	if (boxe != nullptr) {
		return boxe->clone();
	}
	else {
		return nullptr;
	}
}

CommandBox* CommandBoxFactory::cloneGenericCommandBox(const CommandBox* commandBox) {
	const GenericCommandBox* boxe = dynamic_cast<const GenericCommandBox*>(commandBox);
	if (boxe != nullptr) {
		switch (boxe->getBoxType()) {
		case GenericCommandBox::FIX:
			return dynamic_cast<const FixCommandBox*>(commandBox)->clone();
		case GenericCommandBox::CLEAR:
			return dynamic_cast<const ClearCommandBox*>(commandBox)->clone();
		case GenericCommandBox::SETZERO:
			return dynamic_cast<const SetZeroCommandBox*>(commandBox)->clone();
		case GenericCommandBox::PRINT:
			return dynamic_cast<const PrintCommandBox*>(commandBox)->clone();
		case GenericCommandBox::SET:
			return dynamic_cast<const SetCommandBox*>(commandBox)->clone();
		case GenericCommandBox::VARIABLE:
			return dynamic_cast<const VariableCommandBox*>(commandBox)->clone();
		default:
			err << "Cannot clone generic command box with type " << (long)commandBox->getType() << eol;
		}
	}
	return nullptr;
}

ControlCommandBox* CommandBoxFactory::cloneControlCommandBox(const CommandBox* commandBox) {
	const ControlCommandBox* boxe = dynamic_cast<const ControlCommandBox*>(commandBox);
	if (boxe != nullptr) {
		switch (boxe->getBoxType()) {
		case ControlCommandBox::IF:
			return dynamic_cast<const IfCommandBox*>(commandBox)->clone();
		case ControlCommandBox::IF_ELSE:
			return dynamic_cast<const IfCommandBox*>(commandBox)->clone();
		case ControlCommandBox::BREAK:
			return dynamic_cast<const BreakCommandBox*>(commandBox)->clone();
		default:
			err << "Cannot create control command box with type " << (long)boxe->getType() << eol;
		}
	}
	return nullptr;
}

}
