/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandBoxMap.h"
#include "CommandBox.h"
#include <common/base/Queue.h>
#include <string>

using namespace std;
using namespace common;

namespace core {

CommandBoxMap* CommandBoxMap::_instance = 0;

CommandBoxMap::CommandBoxNotFound::CommandBoxNotFound(unsigned long commandBoxID) :
	Exception("core::CommandBoxMap::CommandBoxNotFound", "cannot find the command box <" + to<string>(commandBoxID) + ">") {
}


CommandBoxMap::CommandBoxMap() {

	_commandBoxToDeleteQueue.reset(new Queue<CommandBox*>(1024));

	_commandBoxToDeleteThread.reset(new thread([this]() {

		// When the server is stopping, running becomes false.
		// Then the thread executes as long as there are some items in the queue.
		// They have been put by the destructor of CommandZone (recursively, a container puts some items when it is deleted here).
		while (true) {

			try {
				// Get the item of the queue.
				CommandBoxPtr* commandBoxPtr = _commandBoxToDeleteQueue->get();

				lock_guard<recursive_mutex> lock(_commandBoxToDeleteMutex);

				// Get the id.
				int id = (*commandBoxPtr)->getID();

				// Delete the pointed command box.
				delete *commandBoxPtr;

				// Delete the item of the queue.
				delete commandBoxPtr;
			}
			catch (Queue<CommandBox*>::UnavailableException& e) {
				break;
			}
		}
	}));
}

CommandBoxMap::~CommandBoxMap() {

	_commandBoxToDeleteQueue->stop();
	_commandBoxToDeleteThread->join();

	for (CommandBoxPtr commandBoxPtr : _remainingCommandBoxToDeleteList) {
		delete commandBoxPtr;
	}
}

CommandBoxMap* CommandBoxMap::getInstance() {
	if (_instance == 0) {
		_instance = new CommandBoxMap();
	}
	return _instance;
}

void CommandBoxMap::resetInstance() {
	delete _instance;
}

void CommandBoxMap::add(CommandBox* commandBox) {
	
	lock_guard<recursive_mutex> lock(_commandBoxMapMutex);

	_commandBoxMap.insert(commandBox->getID(), commandBox);
}

void CommandBoxMap::remove(CommandBox* commandBox) {

	lock_guard<recursive_mutex> lock(_commandBoxMapMutex);

	if (_commandBoxMap.find(commandBox->getID())) {
		_commandBoxMap.erase(commandBox->getID());

		try {
			_commandBoxToDeleteQueue->put(new CommandBoxPtr(commandBox));
		}
		catch (Queue<CommandBox*>::UnavailableException& e) {
			_remainingCommandBoxToDeleteList.push_back(commandBox);
		}
	}
}

CommandBox* CommandBoxMap::get(unsigned long commandBoxID) {

	lock_guard<recursive_mutex> lock(_commandBoxMapMutex);

	try {
		return _commandBoxMap.get(commandBoxID);
	}
	catch (Exception& e) {
		throw link(CommandBoxNotFound(commandBoxID), e);		
	}
}


CommandBoxMap::Lock::Lock() :
	_deleteMutex(CommandBoxMap::getInstance()->_commandBoxToDeleteMutex) {

	_deleteMutex.lock();
}

CommandBoxMap::Lock::~Lock() {

	_deleteMutex.unlock();
}

}
