/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_COMMANDBOXMAP_H_
#define CORE_COMMANDBOXMAP_H_

#include <common/base/HashMap.h>
#include <memory>
#include <thread>
#include <mutex>
#include <list>

namespace common {

template<typename ItemType>
class Queue;

}

namespace core {

class CommandBox;

/**
 * Class defining a map for the command boxes. The class is a singleton encapsulating
 * a hash map.
 */
class CommandBoxMap {

	friend class Lock;

public:
	/**
	 * Command box not found exception.
	 */
	class CommandBoxNotFound : public common::Exception {
	
	public:
		/**
		 * Constructor.
		 */
		CommandBoxNotFound(unsigned long commandBoxID);
	};

	/**
	 * Gets the instance.
	 */
	static CommandBoxMap* getInstance();

	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Adds a command box.
	 */
	void add(CommandBox* commandBox);
	
	/**
	 * Removes a command box.
	 */
	void remove(CommandBox* commandBox);
	
	/**
	 * Gets a command box from its ID.
	 * throw CommandBoxNotFound.
	 */
	CommandBox* get(unsigned long commandBoxID);

	/**
	 * Class locking the delete mutex.
	 * Only the delete mutex must be locked.
	 */
	class Lock {

	public:
		/**
		 * Constructor.
		 */
		Lock();

		/**
		 * Destructor.
		 */
		~Lock();

	private:
		std::recursive_mutex& _deleteMutex;
	};

private:
	CommandBoxMap();
	~CommandBoxMap();

	static CommandBoxMap* _instance;

	// Mutex for the command box map data structure.
	std::recursive_mutex _commandBoxMapMutex;
	common::HashMap<unsigned long, CommandBox*> _commandBoxMap;

	// Mutex for the deletion of command boxes.
	std::recursive_mutex _commandBoxToDeleteMutex;
	typedef CommandBox* CommandBoxPtr;
	std::unique_ptr<common::Queue<CommandBoxPtr>> _commandBoxToDeleteQueue;
	std::unique_ptr<std::thread> _commandBoxToDeleteThread;
	std::list<CommandBoxPtr> _remainingCommandBoxToDeleteList;
};

}


#endif /*CORE_COMMANDBOXMAP_H_*/
