/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandSequencer.h"
#include "CommandBox.h"
#include <common/base/Convert.h>
#include <common/base/Queue.h>
#include <core/message/Message.h>
#include <core/message/Shutdown.h>
#include <core/message/commandzone/ExecuteCommandBox.h>
#include <common/output/OutputStream.h>

using namespace std;
using namespace common;

namespace core {

CommandSequencer* CommandSequencer::_instance = 0;

CommandSequencer::CommandSequencer() :
	_running(false),
	_thread(nullptr) {
		
	_messageQueue = new Queue<Message>(1024);
}

CommandSequencer::~CommandSequencer() {
	
	delete _messageQueue;
	
	if (_thread != 0) {
		delete _thread;
	}
}

CommandSequencer* CommandSequencer::getInstance() {
		
	return _instance;
}

void CommandSequencer::init() {
	_thread = new std::thread(&CommandSequencer::loop);
}

void CommandSequencer::reset() {
	_messageQueue->put(new Shutdown(true));
	_thread->join();
}

void CommandSequencer::initInstance() {
	
	if (_instance == 0) {
		_instance = new CommandSequencer();
		_instance->init();
	}
}

void CommandSequencer::resetInstance() {
	
	_instance->reset();
	
	delete _instance;
	_instance = 0;
}

void CommandSequencer::executeCommandBox(CommandBox* commandBox) {
	_messageQueue->put(new ExecuteCommandBox(commandBox, ExecuteCommandBox::ALL));
}

void CommandSequencer::initialiseCommandBoxExecution(CommandBox* commandBox) {
	_messageQueue->put(new ExecuteCommandBox(commandBox, ExecuteCommandBox::INIT));
} 

void CommandSequencer::launchCommandBox(CommandBox* commandBox) {
	_messageQueue->put(new ExecuteCommandBox(commandBox, ExecuteCommandBox::COMMAND));
} 



void CommandSequencer::loop() {
	
	Message* message = 0;
		
	_instance->_running = true;
	while (_instance->_running) {

		try {
			try {		
				message = _instance->_messageQueue->get();
				
			} catch (Queue<Message>::UnavailableException& e) {
				break;
				
			} catch (Exception& e) {
				err << e.printStack() << eol;
				break;
			}
			
			try {
				_instance->processMessage(message);
	
			} catch (Exception& e) {
				err << e.printStack() << eol;
			}
			
		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}	
		
		delete message;
		message = 0;
 	}
}

void CommandSequencer::processMessage(Message* message) {
	
	if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<Shutdown*>(message));
		
	} else if (message->getType() == Message::EXECUTE_COMMAND_BOX) {
		processMessage(static_cast<ExecuteCommandBox*>(message));
		
	} 
}

void CommandSequencer::processMessage(Shutdown* message) {

	_running = false;
}

void CommandSequencer::processMessage(ExecuteCommandBox* message) {
	if (message->getActionType() == ExecuteCommandBox::INIT) {
		message->getCommandBox()->initialiseExecution();
		
	} else if (message->getActionType() == ExecuteCommandBox::COMMAND) {
		message->getCommandBox()->launchCommand();
		
	} else if (message->getActionType() == ExecuteCommandBox::ALL) {
		message->getCommandBox()->execute();
	}
}

}
