/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandZone.h"
#include "ModifiableCommandBoxContainer.h"
#include "CommandSequencer.h"
#include "CommandBoxMap.h"
#include "CommandZoneEventServer.h"
#include "AtomicCommandBox.h"
#include "EstimatedTimeManager.h"
#include <core/log/LogHandler.h>
#include <core/log/CommandZoneContentLog.h>
#include <core/log/CommandZoneEventLog.h>
#include <core/log/SimpleLog.h>
#include <core/base/Hostname.h>
#include <core/dataaccess/DataAccess.h>
#include <common/base/Convert.h>
#include <common/output/OutputStream.h>
#include <common/base/CommonTypes.h>
#include <common/base/ServerProperties.h>
#include <ics/IcsServerManager.h>
#include <dataprovider/notification/NotificationServer.h>
#include <core/cloner/ServantCloner.h>
#include <Utilities/Counter.h>

using namespace std;
using namespace common;

namespace core {

const std::string CommandZone::COUNT_REPEATER = "Count Repeater";
const std::string CommandZone::START = "start";
const std::string CommandZone::ACTIVE = "active";
const std::string CommandZone::MAX_FINISHED_COMMAND_BOXES = "maxFinishedCommandBoxes";

CommandZone::CommandZoneNotFound::CommandZoneNotFound(unsigned long commandZoneID) :
	Exception("core::CommandZone::CommandZoneNotFound", "cannot find the command zone <" + to<string>(commandZoneID) + ">") {
}

CommandZone::CommandConflicts::CommandConflicts(unsigned long commandZoneID) :
	Exception("core::CommandZone::CommandConflicts", "content conflicts with running commands for command zone <" + to<string>(commandZoneID) + ">") {
}

map<unsigned long, CommandZone*> CommandZone::_instances;

CommandZone::CommandZone(unsigned long ID) :
	_container(new ModifiableCommandBoxContainer(this)),
	_ID(ID),
	_progression(0) {

	try {
		_maxFinishedCommandBoxes = ServerProperties::getInstance()->get<int>(MAX_FINISHED_COMMAND_BOXES);

	} catch (const Exception& e) {
		out << "Setting default value 100 to property '" << MAX_FINISHED_COMMAND_BOXES << "'" << eol;
		_maxFinishedCommandBoxes = 100;
	}

	_finishedCommandBoxCounter.store(0);
}

CommandZone::~CommandZone() {
	delete _container;
}

CommandZone* CommandZone::getInstance(unsigned long ID) {

	map<unsigned long, CommandZone*>::iterator c = _instances.find(ID);

	if (c != _instances.end()) {
		return c->second;
	}

	throw CommandZoneNotFound(ID);
}

CommandZone* CommandZone::getInstance() {

	if (_instances.size() == 1) {
		return _instances.begin()->second;
	}

	throw CommandZoneNotFound(0);
}

void CommandZone::initInstance(unsigned long ID) {

	map<unsigned long, CommandZone*>::iterator c = _instances.find(ID);

	if (c == _instances.end()) {
		_instances[ID] = new CommandZone(ID);
	}

	// initialising the CommandSequencer once.
	CommandSequencer::initInstance();
}

void CommandZone::resetInstance(unsigned long ID) {

	if (_instances[ID] != 0) {
		delete _instances[ID];
		_instances[ID] = 0;
	}
}

void CommandZone::resetInstances() {

	CommandSequencer::resetInstance();

	for (map<unsigned long, CommandZone*>::iterator c = _instances.begin(); c != _instances.end(); c++) {
		delete c->second;
	}

	_instances.clear();

	CommandBoxMap::resetInstance();
}

unsigned long CommandZone::getID() const {
	return _ID;
}

unsigned long CommandZone::addCommandBoxAtEnd(CommandBox* commandBox) {
	std::scoped_lock lock(_containerMutex);
	unsigned long id =  _container->addCommandBoxAtEnd(commandBox);
	return id;
}

unsigned long CommandZone::addCommandBox(CommandBox* commandBox, unsigned long commandBoxToShiftID, bool above) {
	std::scoped_lock lock(_containerMutex);
	unsigned long id = _container->addCommandBox(commandBox, commandBoxToShiftID, above);
	return id;
}

void CommandZone::moveCommandBoxToEnd(unsigned long commandBoxID) {
	std::scoped_lock lock(_containerMutex);
	_container->moveCommandBoxToEnd(commandBoxID);
}

void CommandZone::moveCommandBox(unsigned long commandBoxToMoveID, unsigned long commandBoxToShiftID, bool above) {
	std::scoped_lock lock(_containerMutex);
	_container->moveCommandBox(commandBoxToMoveID, commandBoxToShiftID, above);
}

void CommandZone::deleteCommandBox(unsigned long commandBoxID) {
	std::scoped_lock lock(_containerMutex);
	_container->deleteCommandBox(commandBoxID);
	LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, string("Deleted command box ") + to_string(commandBoxID)));
}

const std::vector<CommandBox*>& CommandZone::getCommandBoxes() {
	std::scoped_lock lock(_containerMutex);
	EstimatedTimeManager::getInstance()->getEstimatedTime();
	return _container->getCommandBoxes();
}

void CommandZone::execute(const std::string& endpoint, unsigned long databaseID) {

	// set the database ID for the command zone
	_container->setExecutionDatabaseID(databaseID);

	// Give all command boxes a unique ID for the log.
	_container->setLogIDs();

	// Sets the content of the command zone to the data provider
	LogHandler::getInstance()->write(CommandZoneContentLog(toXMLString(), true));

	// clear conflicting commands
	clearCommandConflicts();

	// notifying ics server manager
	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::getInstance()->onCommandZoneStarted();
	}

	LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, "Client " + hostname(endpoint) + " started launch pad"));

	try {
		// start the execution
		_container->startExecution();

	} catch (CommandBoxContainer::CannotRunException& e) {
		throw link(CommandConflicts(_ID), e);
	}
}

ModifiableCommandBoxContainer* CommandZone::clone() {

	return _container->clone(getInstance());
}

void CommandZone::stop(const std::string& endpoint) {

	_container->stop();

	if (endpoint == "") {
		LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, "Server stopped launch pad"));
	}
	else {
		LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, "Client " + hostname(endpoint) + " stopped launch pad"));
	}
}

void CommandZone::stopAtEnd(const std::string& endpoint) {

	_container->stopAtEnd();
	CommandZoneEventServer::getInstance()->commandZoneStoppingAtEnd(_ID);

	LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, "Client " + hostname(endpoint) + " stopped launch pad at end of command box"));
}

void CommandZone::restart(const std::string& endpoint) {

	try {
		// restart the execution
		_container->restart();
		if (_container->isStoppingAtEnd()) {
			CommandZoneEventServer::getInstance()->commandZoneStoppingAtEnd(_ID);
		}
		else {
			CommandZoneEventServer::getInstance()->commandZoneStarted(_ID);
		}

		LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, "Client " + hostname(endpoint) + " restarted launch pad"));

	} catch (CommandBoxContainer::CannotRunException& e) {
		throw link(CommandConflicts(_ID), e);
	}
}

void CommandZone::pause(const std::string& endpoint) {

	_container->pause();
	CommandZoneEventServer::getInstance()->commandZonePaused(_ID);

	LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, "Client " + hostname(endpoint) + " paused launch pad"));
}

void CommandZone::pauseAtEnd(const std::string& endpoint) {

	_container->pauseAtEnd();
	CommandZoneEventServer::getInstance()->commandZonePausingAtEnd(_ID);

	LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, "Client " + hostname(endpoint) + " paused launch pad at end of command box"));
}

void CommandZone::resetCommandZone(unsigned long databaseID) {

	std::scoped_lock lock(_containerMutex);

	_container->reset();

	// Reset the counter.
	_finishedCommandBoxCounter.store(0);

	CommandZoneEventServer::getInstance()->commandZoneReset(_ID);

	EstimatedTimeManager::getInstance()->newEstimatedTime();
}

bool CommandZone::deleteFinishedCommandBoxes() {

	std::scoped_lock lock(_containerMutex);

	if (isRunning() || isPaused()) {
		return false;
	}

	// Get the command box ids of the command zone container i.e. we de not return the inner command boxes, only those at the root of the Command Zone.
	vector<unsigned long> deletedCommandBoxIds = _container->deleteFinishedCommandBoxes();

	// Reset the counter.
	_finishedCommandBoxCounter.store(0);

	CommandZoneEventServer::getInstance()->commandBoxesDeleted(_ID, deletedCommandBoxIds);

	LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Deleted finished command boxes"));

	return true;
}

void CommandZone::onExecutionBegin() {
	_progression = 0.0;
	CommandZoneEventServer::getInstance()->commandZoneStarted(_ID);
}

void CommandZone::onExecutionEnd() {
	_progression = 100.0;
	CommandZoneEventServer::getInstance()->commandZoneFinished(_ID);
	// notifying ics server manager
	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::getInstance()->onCommandZoneStopped();
	}
}

void CommandZone::onExecutionPause() {
	CommandZoneEventServer::getInstance()->commandZonePaused(_ID);
}

void CommandZone::onCommandBoxFinished(CommandBox* commandBox) {

	int finishedCommandBoxCounterBefore = _finishedCommandBoxCounter.load();

	// Update the counter.
	_finishedCommandBoxCounter += commandBox->getCommandBoxCounter();

	//cout << "Number of finished command boxes : " << _finishedCommandBoxCounter.load() << endl;

	// Check if it is necessary to delete some finished command boxes.
	int commandBoxesToDeleteCounter = _finishedCommandBoxCounter.load() - _maxFinishedCommandBoxes;
	if (commandBoxesToDeleteCounter > 0) {

		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, string("Deleting ") + std::to_string(commandBoxesToDeleteCounter) + " finished command boxes"));

		int deletedCommandBoxCounter = 0;

		// Get the command box ids of the command zone container i.e. we de not return the inner command boxes, only those at the root of the Command Zone.
		vector<unsigned long> deletedCommandBoxIds = _container->deleteFinishedCommandBoxes(commandBoxesToDeleteCounter, deletedCommandBoxCounter);

		// Update the counter.
		_finishedCommandBoxCounter -= deletedCommandBoxCounter;

		CommandZoneEventServer::getInstance()->commandBoxesDeleted(_ID, deletedCommandBoxIds);
	}
	else {

		// Log when the number of command boxes changes its 50.
		int before = finishedCommandBoxCounterBefore / 50;
		int after = _finishedCommandBoxCounter.load() / 50;

		if (after != before) {
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, string("Number of finished command boxes ") + std::to_string(_finishedCommandBoxCounter.load())));
		}
	}

	EstimatedTimeManager::getInstance()->newEstimatedTime();
	_estimatedTimeRefreshCounter.clear();
}

void CommandZone::addVariableNamesToList(std::vector<std::string>& variableNames) {
	// do nothing
}

void CommandZone::onContainerProgressChanged(double progression) {
	_progression = progression;
	CommandZoneEventServer::getInstance()->commandZoneProgressionChanged(_ID, progression);
	if (_estimatedTimeRefreshCounter.getTime() > 10.) {
		EstimatedTimeManager::getInstance()->newEstimatedTime();
		_estimatedTimeRefreshCounter.clear();
	}
}

bool CommandZone::hasEnded() {
	return true;
}

bool CommandZone::isRunning() const {
	return _container->isRunning();
}

bool CommandZone::isPaused() const {
	return _container->isPaused();
}

bool CommandZone::isStoppingAtEnd() const {
	return _container->isStoppingAtEnd();
}

bool CommandZone::isPausingAtEnd() const {
	return _container->isPausingAtEnd();
}

CommandZone::ExecutionState CommandZone::getExecutionState() const {

	if (_container->isRunning()) {
		return ExecutionState::STARTED;
	}
	else if (_container->isPaused()) {
		return ExecutionState::PAUSED;
	}
	else if (_container->isStoppingAtEnd()) {
		return ExecutionState::STOPPING_AT_END;
	}
	return ExecutionState::FINISHED;
}

double CommandZone::getProgression() const {
	return _progression;
}

void CommandZone::clearCommandConflicts() {
	_problemCommands.clear();
}

void CommandZone::addProblemCommand(const std::string& servantName, const std::string& commandName) {
	_problemCommands.push_back(std::pair<string, string>(servantName, commandName));
}

string CommandZone::getConflictingCommands() const {
	string result;
	for (std::vector<std::pair<std::string, std::string> >::const_iterator i = _problemCommands.begin(); i != _problemCommands.end(); i++) {
		result += "servant '" + i->first + "', command '" + i->second + "'\n";
	}
	return result;
}

void CommandZone::errorOccurred(const std::string& message) {
	CommandZoneEventServer::getInstance()->commandZoneErrorOccurred(_ID, message);

	// stop the command zone to make sure that nothing is running
	_container->stop();

	// make sure that an event is sent to the client
	_progression = 100.0;
	CommandZoneEventServer::getInstance()->commandZoneProgressionChanged(_ID, _progression);
	CommandZoneEventServer::getInstance()->commandZoneFinished(_ID);
}

bool CommandZone::referencesUserVariable(const std::string& variableName) {
	return _container->referencesUserVariable(variableName);
}

std::string CommandZone::toXMLString() const {

	std::string result = "<zone>\n";
	result += _container->toXMLString();
	result += "</zone>\n";

	return result;
}

std::string CommandZone::print() const {

	std::string result;
	result += _container->print();

	return result;
}

}
