/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_COMMANDZONE_H_
#define CORE_COMMANDZONE_H_

#include "CommandBoxParent.h"
#include <common/base/Exception.h>
#include <Utilities/Counter.h>
#include <vector>
#include <map>
#include <atomic>
#include <mutex>

namespace core {

class ModifiableCommandBoxContainer;
class CommandBox;

/**
 * Class defining a command zone. This is the class that references all the content of
 * the command zone. We can have several command zones identified with a unique ID. The
 * class is a multiton.
 */
class CommandZone : public CommandBoxParent {

public:
	enum class ExecutionState {STARTED, PAUSED, STOPPING_AT_END, PAUSING_AT_END, FINISHED};

	/**
	 * Command zone not found.
	 */
	class CommandZoneNotFound : public common::Exception {

	public:
		CommandZoneNotFound(unsigned long commandZoneID);
	};

	/**
	 * Command conflicts.
	 */
	class CommandConflicts : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		CommandConflicts(unsigned long commandZoneID);
	};

	/**
	 * Returns the instance with ID.
	 */
	static CommandZone* getInstance(unsigned long ID);

	/**
	 * Returns the first instance.
	 */
	static CommandZone* getInstance();

	/**
	 * Initialises the instance with ID = 0.
	 */
	static void initInstance(unsigned long ID);

	/**
	 * Resets a particular instance.
	 */
	static void resetInstance(unsigned long ID);

	/**
	 * Resets all the instances.
	 */
	static void resetInstances();

	/**
	 * Gets the ID of the command zone.
	 */
	unsigned long getID() const;

	/**
	 * Adds a new command box to the end of the container.
	 * Returns the ID of the created command box.
	 */
	unsigned long addCommandBoxAtEnd(CommandBox* commandBox);

	/**
	 * Adds a new command box above or below another one.
	 * Returns the ID of the created command box.
	 */
	unsigned long addCommandBox(CommandBox* commandBox, unsigned long commandBoxToShiftID, bool above);

	/**
	 * Moves a command box to the end.
	 */
	void moveCommandBoxToEnd(unsigned long commandBoxID);

	/**
	 * Moves a command box above or below another one.
	 */
	void moveCommandBox(unsigned long commandBoxToMoveID, unsigned long commandBoxToShiftID, bool above);

	/**
	 * Deletes the command box.
	 */
	void deleteCommandBox(unsigned long commandBoxID);

	/**
	 * Returns the command boxes.
	 */
	const std::vector<CommandBox*>& getCommandBoxes();

	/**
	 * Executes the command zone.
	 */
	void execute(const std::string& endpoint, unsigned long databaseID);

	/**
	 * Clones the command zone.
	 */
	ModifiableCommandBoxContainer* clone();

	/**
	 * Stops the command zone.
	 */
	void stop(const std::string& endpoint);

	/**
	 * Stops the command zone at end (terminates the current commands).
	 */
	void stopAtEnd(const std::string& endpoint);

	/**
	 * Pauses the command zone.
	 */
	void pause(const std::string& endpoint = "");

	/**
	 * Pauses the command zone at end (pauses the current commands).
	 */
	void pauseAtEnd(const std::string& endpoint);

	/**
	 * Restarts the command zone.
	 */
	void restart(const std::string& endpoint = "");

	/**
	 * Resets the command zone.
	 */
	void resetCommandZone(unsigned long databaseID);

	/**
	 * Deletes the finished command boxes.
	 * Does nothing when the command zone is running or paused.
	 */
	bool deleteFinishedCommandBoxes();

	/**
	 * Notifies the container the execution has begun.
	 */
	virtual void onExecutionBegin();

	/**
	 * Notifies the container that the execution has ended.
	 */
	virtual void onExecutionEnd();

	/**
	 * Notifies the container that the execution has been paused.
	 */
	virtual void onExecutionPause();

	/**
	 * Notifies the container that one if its command boxes has finished.
	 */
	virtual void onCommandBoxFinished(CommandBox* commandBox);

	/**
	 * Adds the variable names to the list.
	 */
	virtual void addVariableNamesToList(std::vector<std::string>& variableNames);

	/**
	 * Notifies the container that the execution progress has changed.
	 */
	virtual void onContainerProgressChanged(double progression);

	/**
	 * Returns true if there is no command to execute anymore.
	 * Only ForLoopCommandBox is able to return true.
	 */
	virtual bool hasEnded();

	/**
	 * Returns true if is running.
	 */
	bool isRunning() const;

	/**
	 * Returns true if is paused.
	 */
	bool isPaused() const;

	/**
	 * Returns true if is pausing at end.
	 */
	bool isPausingAtEnd() const;

	/**
	 * Returns true if is stopping at end.
	 */
	bool isStoppingAtEnd() const;

	/**
	 * Returns the execution state.
	 */
	ExecutionState getExecutionState() const;

	/**
	 * Returns the progression of the command zone.
	 */
	double getProgression() const;

	/**
	 * Returns true if problems exist.
	 */
	bool commandConflicts() const;

	/**
	 * Returns a string identifying the conflicting commands.
	 */
	std::string getConflictingCommands() const;

	/**
	 * Adds a command to the list of commands that cannot run at the start
	 * of execution.
	 */
	void addProblemCommand(const std::string& servantName, const std::string& commandName);

	/**
	 * Notifies the client that an error occurred.
	 */
	void errorOccurred(const std::string& message);

	/**
	 * Returns true if the box references a user variable.
	 */
	bool referencesUserVariable(const std::string& variableName);

	/**
	 * Prints the content of the command zone in XML format.
	 */
	std::string toXMLString() const;

	/**
	 * Prints the content of the command zone.
	 */
	std::string print() const;

private:
	CommandZone(unsigned long ID);
	~CommandZone();
	void clearCommandConflicts();

	static std::map<unsigned long, CommandZone*> _instances;

	ModifiableCommandBoxContainer* _container;
	std::recursive_mutex _containerMutex;

	unsigned long _ID;
	double _progression;
	std::vector<std::pair<std::string, std::string> > _problemCommands;

	std::atomic_int _finishedCommandBoxCounter;
	int _maxFinishedCommandBoxes;

	Counter _estimatedTimeRefreshCounter;

	static const std::string COUNT_REPEATER;
	static const std::string START;
	static const std::string ACTIVE;

	static const std::string MAX_FINISHED_COMMAND_BOXES;
};

}


#endif /*CORE_COMMANDZONE_H_*/
