/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandZoneEventServer.h"
#include <core/message/Shutdown.h>
#include <core/transportlayer/core/commandzone/CommandZoneEventPublisher.h>
#include <core/message/commandzone/CommandBoxExecutionEvent.h>
#include <core/message/commandzone/CommandBoxProgressionEvent.h>
#include <core/message/commandzone/CommandBoxResetEvent.h>
#include <core/message/commandzone/CommandBoxPropertyEvent.h>
#include <core/message/commandzone/CommandZoneExecutionEvent.h>
#include <core/message/commandzone/CommandZoneProgressionEvent.h>
#include <core/message/commandzone/CommandZoneResetEvent.h>
#include <core/message/commandzone/CommandZoneErrorEvent.h>
#include <core/message/commandzone/ForLoopCommandBoxCurrentValueEvent.h>
#include <core/message/commandzone/CommandBoxExpressionChangedEvent.h>
#include <core/message/commandzone/CommandBoxExpressionValueChangedEvent.h>
#include <core/message/commandzone/CommandBoxesDeletedEvent.h>
#include <core/message/commandzone/CommandZoneSyncEvent.h>
#include <core/message/commandzone/NewEstimatedTimeMessage.h>
#include <common/base/Queue.h>
#include <common/output/OutputStream.h>
#include <boost/bind.hpp>
#include <iostream>

using namespace std;
using namespace common;

namespace core {

CommandZoneEventServer* CommandZoneEventServer::_instance = 0;
bool CommandZoneEventServer::_running = false;

CommandZoneEventServer::CommandZoneEventServer(CommandZoneEventPublisher* commandZoneEventPublisher) :
	_commandZoneEventPublisher(commandZoneEventPublisher),
	_queue(new common::Queue<Message>(1024)),
	_thread(nullptr) {
}

CommandZoneEventServer::~CommandZoneEventServer() {
	delete _queue;
}

void CommandZoneEventServer::initInstance(CommandZoneEventPublisher* commandZoneEventPublisher) {
	if (_instance == 0) {
		_instance = new CommandZoneEventServer(commandZoneEventPublisher);
		_instance->init();
	}
}

void CommandZoneEventServer::resetInstance() {
	_instance->reset();
}

CommandZoneEventServer* CommandZoneEventServer::getInstance() {
	return _instance;
}


void CommandZoneEventServer::init() {
	_running = true;
	_thread = new std::thread(&CommandZoneEventServer::run);
}

void CommandZoneEventServer::reset() {
	_queue->put(new Shutdown(false));

	_thread->join();
	delete _thread;

}

void CommandZoneEventServer::commandBoxStarted(unsigned long commandBoxID) {
	_queue->put(new CommandBoxExecutionEvent(commandBoxID, CommandBoxExecutionEvent::STARTED));
}

void CommandZoneEventServer::commandBoxPaused(unsigned long commandBoxID) {
	_queue->put(new CommandBoxExecutionEvent(commandBoxID, CommandBoxExecutionEvent::PAUSED));
}

void CommandZoneEventServer::commandBoxPausingAtEnd(unsigned long commandBoxID) {
	_queue->put(new CommandBoxExecutionEvent(commandBoxID, CommandBoxExecutionEvent::PAUSING_AT_END));
}

void CommandZoneEventServer::commandBoxStoppingAtEnd(unsigned long commandBoxID) {
	_queue->put(new CommandBoxExecutionEvent(commandBoxID, CommandBoxExecutionEvent::STOPPING_AT_END));
}

void CommandZoneEventServer::commandBoxFinished(unsigned long commandBoxID) {
	_queue->put(new CommandBoxExecutionEvent(commandBoxID, CommandBoxExecutionEvent::FINISHED));
}

void CommandZoneEventServer::commandBoxProgressionChanged(unsigned long commandBoxID, double progression) {
	_queue->put(new CommandBoxProgressionEvent(commandBoxID, progression));
}

void CommandZoneEventServer::commandBoxReset(unsigned long commandBoxID) {
	_queue->put(new CommandBoxResetEvent(commandBoxID));
}

void CommandZoneEventServer::commandBoxPropertyChanged(unsigned long commandBoxID, unsigned long propertyID) {
	_queue->put(new CommandBoxPropertyEvent(commandBoxID, propertyID));
}

void CommandZoneEventServer::commandZoneStarted(unsigned long commandZoneID) {
	_queue->put(new CommandZoneExecutionEvent(commandZoneID, CommandZoneExecutionEvent::STARTED));
}

void CommandZoneEventServer::commandZonePaused(unsigned long commandZoneID) {
	_queue->put(new CommandZoneExecutionEvent(commandZoneID, CommandZoneExecutionEvent::PAUSED));
}

void CommandZoneEventServer::commandZonePausingAtEnd(unsigned long commandZoneID) {
	_queue->put(new CommandZoneExecutionEvent(commandZoneID, CommandZoneExecutionEvent::PAUSING_AT_END));
}

void CommandZoneEventServer::commandZoneStoppingAtEnd(unsigned long commandZoneID) {
	_queue->put(new CommandZoneExecutionEvent(commandZoneID, CommandZoneExecutionEvent::STOPPING_AT_END));
}

void CommandZoneEventServer::commandZoneFinished(unsigned long commandZoneID) {
	_queue->put(new CommandZoneExecutionEvent(commandZoneID, CommandZoneExecutionEvent::FINISHED));
}

void CommandZoneEventServer::commandZoneProgressionChanged(unsigned long commandZoneID, double progression) {
	_queue->put(new CommandZoneProgressionEvent(commandZoneID, progression));
}

void CommandZoneEventServer::commandZoneReset(unsigned long commandZoneID) {
	_queue->put(new CommandZoneResetEvent(commandZoneID));
}

void CommandZoneEventServer::commandZoneErrorOccurred(unsigned long commandZoneID, const std::string& message) {
	_queue->put(new CommandZoneErrorEvent(commandZoneID, message));
}

void CommandZoneEventServer::forLoopCurrentValueChanged(unsigned long forLoopCommandBoxID, long index, double currentValue) {
	_queue->put(new ForLoopCommandBoxCurrentValueEvent(forLoopCommandBoxID, index, currentValue));
}

void CommandZoneEventServer::commandBoxExpressionChanged(unsigned long commandBoxID) {
	_queue->put(new CommandBoxExpressionChangedEvent(commandBoxID));
}

void CommandZoneEventServer::commandBoxExpressionValueChanged(unsigned long commandBoxID, bool currentValue) {
	_queue->put(new CommandBoxExpressionValueChangedEvent(commandBoxID, currentValue));
}

void CommandZoneEventServer::commandBoxesDeleted(unsigned long commandZoneID, const std::vector<unsigned long>& commandBoxIDs) {
	_queue->put(new CommandBoxesDeletedEvent(commandZoneID, commandBoxIDs));
}

void CommandZoneEventServer::commandZoneSync(const std::string& serializedMessageType, const std::string& serializedMessageContent, const std::string& serializedResponse) {
	_queue->put(new CommandZoneSyncEvent(serializedMessageType, serializedMessageContent, serializedResponse));
}

/**
 * Main loop
 */
void CommandZoneEventServer::run() {

	_running = true;
	while (_running) {

		Message* message = 0;

		try {
			message = _instance->_queue->get();

			_instance->processMessage(message);

		} catch (common::Queue<Message>::UnavailableException& e) {
			_running = false;
		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}

		delete message;
	}

}

void CommandZoneEventServer::processMessage(Message* message) {

	if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<Shutdown*>(message));

	} else if (message->getType() == Message::COMMAND_ZONE_EVENT) {
		processMessage(static_cast<CommandZoneEvent*>(message));

	}
}

void CommandZoneEventServer::processMessage(Shutdown* message) {

	_running = false;
}

void CommandZoneEventServer::processMessage(CommandZoneEvent* commandZoneEvent) {

	_commandZoneEventPublisher->notify(commandZoneEvent);
}

}
