/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_COMMANDZONEEVENTSERVER_H_
#define CORE_COMMANDZONEEVENTSERVER_H_

#include <vector>
#include <thread>

namespace common {
	template<typename ValueType>
	class Queue;
}

namespace core {

class Message;
class Shutdown;
class CommandZoneEvent;
class CommandZoneEventPublisher;

/**
 * Class defining a command zone event server.
 */
class CommandZoneEventServer {

public:
	/**
	 * Gets the instance.
	 */
	static CommandZoneEventServer* getInstance();

	/**
	 * Initialises the instance.
	 */
	static void initInstance(CommandZoneEventPublisher* commandZoneEventPublisher);

	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Notifies that the command box has started.
	 */
	void commandBoxStarted(unsigned long commandBoxID);

	/**
	 * Notifies that the command box has paused.
	 */
	void commandBoxPaused(unsigned long commandBoxID);

	/**
	 * Notifies that the command box is pausing at end.
	 */
	void commandBoxPausingAtEnd(unsigned long commandBoxID);

	/**
	 * Notifies that the command box is stopping at end.
	 */
	void commandBoxStoppingAtEnd(unsigned long commandBoxID);

	/**
	 * Notifies that the command box has finished.
	 */
	void commandBoxFinished(unsigned long commandBoxID);

	/**
	 * Notifies that the command box progression has changed.
	 */
	void commandBoxProgressionChanged(unsigned long commandBoxID, double progression);

	/**
	 * Notifies that the command box has reset.
	 */
	void commandBoxReset(unsigned long commandBoxID);

	/**
	 * Notifies that the command box property has changed.
	 */
	void commandBoxPropertyChanged(unsigned long commandBoxID, unsigned long propertyID);

	/**
	 * Notifies that the command zone has started.
	 */
	void commandZoneStarted(unsigned long commandZoneID);

	/**
	 * Notifies that the command zone has paused.
	 */
	void commandZonePaused(unsigned long commandZoneID);

	/**
	 * Notifies that the command zone is pausing at end.
	 */
	void commandZonePausingAtEnd(unsigned long commandZoneID);

	/**
	 * Notifies that the command zone is stopping at end.
	 */
	void commandZoneStoppingAtEnd(unsigned long commandZoneID);

	/**
	 * Notifies that the command zone has finished.
	 */
	void commandZoneFinished(unsigned long commandZoneID);

	/**
	 * Notifies that the command zone progression has changed.
	 */
	void commandZoneProgressionChanged(unsigned long commandZoneID, double progression);

	/**
	 * Notifies that the command zone has reset.
	 */
	void commandZoneReset(unsigned long commandZoneID);

	/**
	 * Notifies that an error occurred in the command zone.
	 */
	void commandZoneErrorOccurred(unsigned long commandZoneID, const std::string& message);

	/**
	 * Notifies that the value has changed in the current for loop.
	 */
	void forLoopCurrentValueChanged(unsigned long forLoopCommandBoxID, long index, double currentValue);

	/**
	 * Notifies that the expression has changed.
	 */
	void commandBoxExpressionChanged(unsigned long commandBoxID);

	/**
	 * Notifies that the expression value has changed.
	 */
	void commandBoxExpressionValueChanged(unsigned long commandBoxID, bool currentValue);

	/**
	 * Notifies that the command boxes have been deleted.
	 */
	void commandBoxesDeleted(unsigned long commandZoneID, const std::vector<unsigned long>& commandBoxIDs);

	/**
	 * Notifies that the command zone with a sync event.
	 */
	void commandZoneSync(const std::string& serializedMessageType, const std::string& serializedMessageContent, const std::string& serializedResponse);

private:
	CommandZoneEventServer(CommandZoneEventPublisher* commandZoneEventPublisher);
	virtual ~CommandZoneEventServer();

	void init();
	void reset();

	static void run();

	void processMessage(Message* message);
	void processMessage(Shutdown* message);
	void processMessage(CommandZoneEvent* commandZoneEvent);

	static CommandZoneEventServer* _instance;

	CommandZoneEventPublisher* _commandZoneEventPublisher;
	common::Queue<Message>* _queue;
	std::thread* _thread;
	static bool _running;

};


} // namespace core

#endif /*CORE_COMMANDZONEEVENTSERVER_H_*/
