/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "EstimatedTimeManager.h"
#include <core/message/Shutdown.h>
#include <core/message/commandzone/NewEstimatedTimeMessage.h>
#include "CommandZoneEventServer.h"
#include <common/base/Queue.h>
#include <common/output/OutputStream.h>
#include "CommandZone.h"
#include "ModifiableCommandBoxContainer.h"
#include <common/base/ServerProperties.h>
#include <dataprovider/notification/NotificationServer.h>
#include <ics/IcsServerManager.h>
#include <CommandZoneRequests.pb.h>
#include <iostream>

using namespace std;
using namespace common;

namespace core {

EstimatedTimeManager* EstimatedTimeManager::_instance = 0;
bool EstimatedTimeManager::_running= false;

const std::string EstimatedTimeManager::USE_ESTIMATED_TIME = "useEstimatedTime";

EstimatedTimeManager::EstimatedTimeManager(CommandZoneEventPublisher* commandZoneEventPublisher) :
	_commandZoneEventPublisher(commandZoneEventPublisher),
	_queue(new common::Queue<Message>(1024)),
	_thread(nullptr),
	_estimatedRunning(false),
	_activated(true),
	_estimatedTime(0.) {
	try {
		_useEstimatedTime = common::ServerProperties::getInstance()->get<bool>(USE_ESTIMATED_TIME);
	} catch (const Exception& e) {
		out << "Setting default value true to property '" << USE_ESTIMATED_TIME << "'" << eol;
		_useEstimatedTime = true;
	}

}

EstimatedTimeManager::~EstimatedTimeManager() {
	delete _queue;
}

void EstimatedTimeManager::initInstance(CommandZoneEventPublisher* commandZoneEventPublisher) {
	if (_instance == 0) {
		_instance = new EstimatedTimeManager(commandZoneEventPublisher);
		_instance->init();
	}
}

void EstimatedTimeManager::resetInstance() {
	_instance->reset();
	delete _instance;
}

EstimatedTimeManager* EstimatedTimeManager::getInstance() {
	return _instance;
}


void EstimatedTimeManager::init() {
	_running = true;
	_thread = new std::thread(&EstimatedTimeManager::run);

	// Create the requester.
	std::unique_ptr<cameo::App> app = cameo::This::getServer().connect(cameo::This::getId());

	_requester = cameo::coms::Requester::create(*app, "command_box");
	_requester->init();

	cout << "Created command zone requester for estimated time" << endl;
}

void EstimatedTimeManager::reset() {
	_queue->put(new Shutdown(false));

	_thread->join();
	delete _thread;

}

void EstimatedTimeManager::setActivationState(bool state) {
	if (_useEstimatedTime == false) {
		_activated = false;
	}
	else {
		_activated = state;
	}
	if (_activated) {
		newEstimatedTime();
	}
}

void EstimatedTimeManager::newEstimatedTime() {
	if (_activated) {
		if (_estimatedRunning == false) {
			_queue->put(new NewEstimatedTimeMessage(0, 0, true));
		}
	}
}

void EstimatedTimeManager::getEstimatedTime() {
	if (_activated) {
		if (_estimatedRunning == false) {
			_queue->put(new NewEstimatedTimeMessage(0, 0, false));
		}
	}
}

void EstimatedTimeManager::computeEstimatedTime() {

	//Counter c1;
	//cout << "EstimatedTimeManager clone request" << endl;

	// Send a request for cloning the command zone. The cloning is called by the command zone responder.
	// The goal is to ensure that the command zone cannot be modified by another request during the cloning.
	commandzone::Message messageType;
	messageType.set_type(commandzone::Message::CloneForEstimatedTime);

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.set_clientid(0);
	messageRequest.set_commandzoneid(0);

	_requester->sendTwoParts(messageType.SerializeAsString(), messageRequest.SerializeAsString());

	optional<string> responseMessage = _requester->receive();

	common::BooleanResponse response;
	response.ParseFromString(responseMessage.value());

	//cout << "EstimatedTimeManager clone request received response after " << c1.getTime() << endl;

	_cloneContainer->setExecutionDatabaseID(0);
	try {
		// Start the computation.
		_estimatedTime = _cloneContainer->computeEstimatedTime(true);
		//cout << "CommandZone::computeEstimatedTime " << c1.getTime() << endl;
	}
	catch (...) {
		_estimatedTime = -1.0;
	}

	// Delete the command zone.
	_cloneContainer.reset();

	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::getInstance()->onEstimatedTimeChanged(_estimatedTime);
		ics::ServerManager::getInstance()->onEstimatedTimeComputingStatus(0);
	}

	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);

	//cout << "EstimatedTimeManager finished computing time " << c1.getTime() << endl;
}

void EstimatedTimeManager::cloneCommandZone() {

	// The function cloneCommandZone() is called by the command zone responder.
	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::getInstance()->onEstimatedTimeComputingStatus(1);
	}
	_cloneContainer.reset(CommandZone::getInstance()->clone());
}

/**
 * Main loop
 */
void EstimatedTimeManager::run() {

	_running = true;
	while (_running) {

		Message* message = 0;

		try {
			message = _instance->_queue->get();

			_instance->processMessage(message);

		} catch (common::Queue<Message>::UnavailableException& e) {
			_running = false;
		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}

		delete message;
	}

}

void EstimatedTimeManager::processMessage(Message* message) {
	_estimatedRunning = true;
	if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<Shutdown*>(message));

	} else if (message->getType() == Message::COMMAND_ZONE_EVENT) {
		processMessage(static_cast<CommandZoneEvent*>(message));

	}
	_estimatedRunning = false;
}

void EstimatedTimeManager::processMessage(Shutdown* message) {

	_running = false;
}

void EstimatedTimeManager::processMessage(CommandZoneEvent* commandZoneEvent) {
	_estimatedRunning = true;
	if (commandZoneEvent->getEventType() == CommandZoneEvent::NEW_ESTIMATED_TIME) {
		if (static_cast<NewEstimatedTimeMessage*>(commandZoneEvent)->getDoCompute() == true) {
			computeEstimatedTime();
		}

	}
	_estimatedRunning = false;
}

}
