/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ForLoopCommandBox.h"
#include "ModifiableCommandBoxContainer.h"
#include "CommandZoneEventServer.h"
#include "DataProvider/Property/AnyProperty.h"
#include <core/log/LogHandler.h>
#include <core/log/CommandZoneContentLog.h>
#include <core/log/SimpleLog.h>
#include <core/log/VariableChangeLog.h>
#include <core/log/ForLoopStartLog.h>
#include <core/log/ForLoopEndLog.h>
#include <common/base/Convert.h>
#include "EstimatedTimeManager.h"
#include <errno.h>
#include <cmath>

using namespace std;
using namespace common;

namespace core {

ForLoopCommandBox::ValuesSyntaxErrorException::ValuesSyntaxErrorException(const std::string allValues, const std::string valueWithError) :
	Exception("core::ForLoopCommandBox::ValuesSyntaxErrorException", string("syntax error in values : \'") + allValues + "\', specifically \'" + valueWithError + "\'") {
}

const std::string ForLoopCommandBox::SEPARATORS = ",;/ ";

ForLoopCommandBox::ForLoopCommandBox(const std::string& variableName) :
	_container(new ModifiableCommandBoxContainer(this)),
	_iStep(0) {

	// No new line created
	ForLoopCommandBox::setInsideForLoop();
}

ForLoopCommandBox::~ForLoopCommandBox() {
	delete _container;
}

ForLoopCommandBox* ForLoopCommandBox::clone() const {
	ForLoopCommandBox* fclb = new ForLoopCommandBox("");
	fclb->copy(this);
	delete fclb->_container;
	fclb->_container = _container->clone(fclb);
	fclb->_forLoopLines = _forLoopLines;
	fclb->_iStep = _iStep;
	fclb->_name = _name;
	fclb->_variablesField = _variablesField;
	return fclb;
}

CommandBox::CommandBoxType ForLoopCommandBox::getType() const {
	return CommandBox::FOR_LOOP;
}

void ForLoopCommandBox::setVariableName(long lineIndex, const std::string& variableName) {
	ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	changeVariableName(forLoopLine.variableName, variableName);
	forLoopLine.variableName = variableName;
	constructName();
	EstimatedTimeManager::getInstance()->newEstimatedTime();
}

const std::string& ForLoopCommandBox::getVariableName(long lineIndex) const {
	const ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	return forLoopLine.variableName;
}

void ForLoopCommandBox::setStartValue(long lineIndex, float64 startValue) {
	ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	forLoopLine.startValue = startValue;
	constructName();
	EstimatedTimeManager::getInstance()->newEstimatedTime();
}

void ForLoopCommandBox::setEndValue(long lineIndex, float64 endValue) {
	ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	forLoopLine.endValue = endValue;
	constructName();
	EstimatedTimeManager::getInstance()->newEstimatedTime();
}

void ForLoopCommandBox::setStepSize(long lineIndex, float64 stepSize) {
	ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	forLoopLine.stepSize = stepSize;
	constructName();
	EstimatedTimeManager::getInstance()->newEstimatedTime();
}

float64 ForLoopCommandBox::getStartValue(long lineIndex) const {
	const ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	return forLoopLine.startValue;
}

float64 ForLoopCommandBox::getEndValue(long lineIndex) const {
	const ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	return forLoopLine.endValue;
}

float64 ForLoopCommandBox::getStepSize(long lineIndex) const {
	const ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	return forLoopLine.stepSize;
}

float64 ForLoopCommandBox::getCurrentValue(long lineIndex) const {
	const ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	return forLoopLine.currentValue;
}

void ForLoopCommandBox::setForLoopType(long lineIndex, ForLoopType type) {
	ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	forLoopLine.forLoopType = type;
}

ForLoopCommandBox::ForLoopType ForLoopCommandBox::getForLoopType(long lineIndex) const {
	const ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	return forLoopLine.forLoopType;
}

void ForLoopCommandBox::setForLoopStepType(long lineIndex, ForLoopStepType type) {
	ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	forLoopLine.forLoopStepType = type;
}

ForLoopCommandBox::ForLoopStepType ForLoopCommandBox::getForLoopStepType(long lineIndex) const {
	const ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	return forLoopLine.forLoopStepType;
}

void ForLoopCommandBox::setValues(long lineIndex, const std::string& valuesString) {
	ForLoopLine& forLoopLine = _forLoopLines[lineIndex];

	// erase all old values
	forLoopLine.values.clear();

	vector<float64> values;

	// Go through string converting into double values
	string currentValue = "";
	unsigned int position = valuesString.find_first_not_of(SEPARATORS);
	while (position < valuesString.length()) {

		unsigned int nextPosition = valuesString.find_first_of(SEPARATORS, position);

		if (nextPosition != string::npos) {
			currentValue = valuesString.substr(position, nextPosition - position);

			if (checkValue(currentValue)) {
				float64 value = strtod(currentValue.c_str(), NULL);
				values.push_back(value);

			} else {
				constructName();
				throw ValuesSyntaxErrorException(valuesString, currentValue);
			}

			// refresh position
			position = valuesString.find_first_not_of(SEPARATORS, nextPosition);
			if (position == string::npos) {
				position = valuesString.length();
			}

		} else {
			currentValue = valuesString.substr(position);

			if (checkValue(currentValue)) {
				float64 value = strtod(currentValue.c_str(), 0);
				values.push_back(value);
			} else {
				throw ValuesSyntaxErrorException(valuesString, currentValue);
			}

			position = valuesString.length();
		}
	}

	vector<float64>::iterator v;
	for (v = values.begin(); v != values.end(); v++) {
		forLoopLine.values.push_back(*v);
	}

	forLoopLine.valuesString = valuesString;
	constructName();
}

const std::string& ForLoopCommandBox::getValues(long lineIndex) const {
	const ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	return forLoopLine.valuesString;
}

bool ForLoopCommandBox::checkValue(const string& valueString) const {
	const char* currentValueChars = valueString.c_str();
	char* lastCharPtr;
	errno = 0;    // To distinguish success/failure after call
	float64 value = strtod(currentValueChars, &lastCharPtr);

	if (lastCharPtr == currentValueChars || lastCharPtr != &currentValueChars[valueString.length()]) {
		return false;
	} else {
		if ((errno == ERANGE && (value == HUGE_VALF || value == HUGE_VALL)) || (errno != 0 && value == 0)) {
			return false;
		}
	}
	return true;
}

bool ForLoopCommandBox::addForLoopLine(long lineIndex) {

	if (_forLoopLines.size() <= (size_t)lineIndex) {
		ForLoopLine forLoopLine;
		_forLoopLines.push_back(forLoopLine);

		constructName();

		return true;
	}
	return false;
}

void ForLoopCommandBox::removeForLoopLine(long lineIndex) {

	if (lineIndex < 0 || (size_t)lineIndex >= _forLoopLines.size()) {
		return;
	}

	ForLoopLine& forLoopLine = _forLoopLines[lineIndex];
	vector<ForLoopLine>::iterator it = _forLoopLines.begin();
	while (it != _forLoopLines.end()) {
		if ((*it) == forLoopLine) {
			_forLoopLines.erase(it);
			break;
		}
		it++;
	}
	constructName();

}

long ForLoopCommandBox::getNumberOfLines() const {
	return _forLoopLines.size();
}


const std::vector<CommandBox*>& ForLoopCommandBox::getCommandBoxes() const {
	return _container->getCommandBoxes();
}

unsigned long ForLoopCommandBox::addCommandBoxAtEnd(CommandBox* commandBox) {
	return _container->addCommandBoxAtEnd(commandBox);
}

unsigned long ForLoopCommandBox::addCommandBox(CommandBox* commandBox, unsigned long commandBoxToShiftID, bool above) {
	return _container->addCommandBox(commandBox, commandBoxToShiftID, above);
}

void ForLoopCommandBox::moveCommandBoxToEnd(unsigned long commandBoxID) {
	_container->moveCommandBoxToEnd(commandBoxID);
}

void ForLoopCommandBox::moveCommandBox(unsigned long commandBoxToMoveID, unsigned long commandBoxToShiftID, bool above) {
	_container->moveCommandBox(commandBoxToMoveID, commandBoxToShiftID, above);
}

void ForLoopCommandBox::deleteCommandBox(unsigned long commandBoxID) {
	_container->deleteCommandBox(commandBoxID);
	EstimatedTimeManager::getInstance()->newEstimatedTime();
}

void ForLoopCommandBox::constructName() {

	_name = "For ";

	for (unsigned int i = 0; i < _forLoopLines.size(); i++) {
		_name = _name + "{";
		ForLoopLine& forLoopLine = _forLoopLines[i];

		if (forLoopLine.forLoopType == RANGE) {
			_name = _name + forLoopLine.variableName + " = " + to<string>(forLoopLine.startValue) + " to " + to<string>(forLoopLine.endValue) + " step " + to<string>(forLoopLine.stepSize);

		} else {
			_name = _name + forLoopLine.variableName + " = {";
			for (unsigned int j = 0; j < forLoopLine.values.size(); j++) {

				_name += to<string>(forLoopLine.values[j]);

				if (j < forLoopLine.values.size() - 1) {
					_name += ", ";
				} else {
					_name += "}";
				}
			}
		}
		_name = _name + "}";
		if (i < _forLoopLines.size() - 1) {
			_name = _name + ", ";
		}

	}

	contructVariablesLogField();
}

void ForLoopCommandBox::contructVariablesLogField() {

	_variablesField = "";
	for (unsigned int i = 0; i < _forLoopLines.size(); i++) {
			//_valuesField = _valuesField + "{";
			ForLoopLine& forLoopLine = _forLoopLines[i];
			std::string valuesField = "";
			if (forLoopLine.forLoopType == RANGE) {
				valuesField = valuesField + to<string>(forLoopLine.startValue) + " to " + to<string>(forLoopLine.endValue) + " step " + to<string>(forLoopLine.stepSize);

			} else {
				valuesField = valuesField + "{";
				for (unsigned int j = 0; j < forLoopLine.values.size(); j++) {
					valuesField += to<string>(forLoopLine.values[j]);

					if (j < forLoopLine.values.size() - 1) {
						valuesField += ",";
					} else {
						valuesField += "}";
					}
				}
			}

			std::string name = Log::NAME + "\"" +forLoopLine.variableName + "\"" + ",";
			std::string type = Log::TYPE + "\"range\",";
			std::string value = Log::VALUE + "\"" + valuesField + "\"";
			std::string field = "{" + name + type + value + "}";

			if (i < _forLoopLines.size() - 1) {
				field = field + ",";
			}
			_variablesField = _variablesField + field;
		}

	_variablesField = "[" + _variablesField + "]";
}

bool ForLoopCommandBox::canRun() const {
	return _container->canRun();
}

void ForLoopCommandBox::setExecutionDatabaseID(unsigned long databaseID) {
	_container->setExecutionDatabaseID(databaseID);
}

void ForLoopCommandBox::execute() {

	initialiseExecution();
	launchCommand();
}

double ForLoopCommandBox::computeEstimatedTime(bool doprogression) {
	double estimetime = 0;
	bool ended = true;
	uint32 step = 0;
	do {
		ended = true;
		for (unsigned int i = 0; i < _forLoopLines.size(); i++) {

			ForLoopLine &forLoopLine = _forLoopLines[i];

			if (forLoopLine.forLoopType == RANGE) {
				if (step == 0) {
					forLoopLine.currentValue = forLoopLine.startValue;
					// Check if not too much steps , too long for computing
					uint32 nbsteps = (uint32) fabs(forLoopLine.endValue - forLoopLine.startValue) / forLoopLine.stepSize;
					if (nbsteps >= 1000) {
						throw EstimatedTimeManager::ComputeEstimatedTimeError();
					}
					ended = false;
				}
				else {
					if ((forLoopLine.stepSize > 0
							&& (forLoopLine.currentValue + forLoopLine.stepSize) <= forLoopLine.endValue + 0.00001)
							|| (forLoopLine.stepSize < 0
									&& (forLoopLine.currentValue + forLoopLine.stepSize) >= forLoopLine.endValue - 0.00001)) {
						forLoopLine.currentValue = forLoopLine.currentValue + forLoopLine.stepSize;
						ended = false;
					}
				}
			} else {
				if (step == 0) {
					// Check if not too much steps , too long for computing
					if (forLoopLine.values.size() >= 1000) {
						throw EstimatedTimeManager::ComputeEstimatedTimeError();
					}
				}
				if (step < forLoopLine.values.size()) {
					forLoopLine.currentValue = forLoopLine.values[step];
					ended = false;
				}
			}

			if (!ended) {
				// set new variable
				setVariable(forLoopLine.variableName, forLoopLine.currentValue);
			}

		}
		if (!ended) {
			if (_iStep == step) {
				estimetime += _container->computeEstimatedTime(true);
			}
			else if (step > _iStep){
				estimetime += _container->computeEstimatedTime(false);
			}
		}
		step++;
	}
	while(ended == false);

	return estimetime;
}

void ForLoopCommandBox::initialiseExecution() {

	// Sets the content of the command zone to the data provider
	LogHandler::getInstance()->write(CommandZoneContentLog(toXMLString()));
	LogHandler::getInstance()->write(ForLoopStartLog(common::LOG_INFO, _variablesField, _name, CommandBox::getID()));
}

void ForLoopCommandBox::launchCommand() {

	for (unsigned int i = 0; i < _forLoopLines.size(); i++) {

		ForLoopLine& forLoopLine = _forLoopLines[i];

		if (forLoopLine.forLoopType == RANGE) {
			// start of for loop
			forLoopLine.currentValue = forLoopLine.startValue;

		} else {
			if (forLoopLine.values.size() > 0) {
				// set to initial value if values exist
				forLoopLine.currentValue = forLoopLine.values[0];

			} else {
				// terminate immediately and exit
				setFinished();
				return;
			}
		}

		//Create log entry for first loop iteration ($i = 0)
		if (i == 0) {
			string logMessage = forLoopLine.variableName + " := " + to<string>(forLoopLine.currentValue);
			LogHandler::getInstance()->write(VariableChangeLog(common::LOG_INFO, forLoopLine.variableName, FLOAT64_PROPERTY, to<string>(forLoopLine.currentValue), logMessage, CommandBox::getID()));
		}

		setVariable(forLoopLine.variableName, forLoopLine.currentValue);

		// send event for change in current value
		CommandZoneEventServer::getInstance()->forLoopCurrentValueChanged(getID(), i, forLoopLine.currentValue);
	}

	_iStep = 0;
	_container->startExecution();
}

void ForLoopCommandBox::stop() {
	_container->stop();

	// Do not set finished, otherwise the scan command box will finish before the sub-commands
}

void ForLoopCommandBox::stopAtEnd() {
	if (isRunning()) {
		_container->stopAtEnd();
		setStoppingAtEnd();
	}

	// Do not set finished, otherwise the scan command box will finish before the sub-commands
}

void ForLoopCommandBox::pause() {
	if (isRunning() && !isPaused()) {
		_container->pause();
		setPaused();
	}
}

void ForLoopCommandBox::pauseAtEnd() {
	if (isRunning()) {
		_container->pauseAtEnd();
		setPausingAtEnd();
	}
}

void ForLoopCommandBox::restart() {
	if (isPaused()) {
		setRunning();
		_container->restart();
	}
}

void ForLoopCommandBox::reset() {
	_container->reset();
	setConstruction();
	_iStep = 0;
}

void ForLoopCommandBox::onExecutionBegin() {
	if (!isRunning() && !isFinished()) {
		setRunning();
	}
}

bool ForLoopCommandBox::hasEnded() {

	bool ended = true;

	for (unsigned int i = 0; i < _forLoopLines.size(); i++) {

		ForLoopLine& forLoopLine = _forLoopLines[i];

		if (forLoopLine.forLoopType == RANGE) {
			if ((forLoopLine.stepSize > 0 && (forLoopLine.currentValue + forLoopLine.stepSize) <= forLoopLine.endValue + 0.00001) || (forLoopLine.stepSize < 0 && (forLoopLine.currentValue + forLoopLine.stepSize) >= forLoopLine.endValue - 0.00001)) {
				ended = false;
			}

		} else {
			if (_iStep < forLoopLine.values.size()) {
				ended = false;
			}
		}
	}

	return ended;
}

void ForLoopCommandBox::onExecutionEnd() {

	if (isRunning() || isPaused()) {

		// need to be careful with somethingRunning : between commands something running = false (there is a pause before the
		// command is executed) so we need to verify that all have finished as well.
		if (!_container->isStopped() && !_container->isPaused()) {

			_iStep++;

			bool ended = hasEnded();

			for (unsigned int i = 0; i < _forLoopLines.size(); i++) {

				ForLoopLine& forLoopLine = _forLoopLines[i];

				if (forLoopLine.forLoopType == RANGE) {
					if ((forLoopLine.stepSize > 0 && (forLoopLine.currentValue + forLoopLine.stepSize) <= forLoopLine.endValue + 0.00001) || (forLoopLine.stepSize < 0 && (forLoopLine.currentValue + forLoopLine.stepSize) >= forLoopLine.endValue - 0.00001)) {
						forLoopLine.currentValue = forLoopLine.currentValue + forLoopLine.stepSize;
					}

				} else {
					if (_iStep < forLoopLine.values.size()) {
						forLoopLine.currentValue = forLoopLine.values[_iStep];
					}
				}

				if (!ended) {
					// set new variable
					setVariable(forLoopLine.variableName, forLoopLine.currentValue);

					// send event for change in current value
					CommandZoneEventServer::getInstance()->forLoopCurrentValueChanged(getID(), i, forLoopLine.currentValue);

					// log current value
					string logMessage = forLoopLine.variableName + " := " + to<string>(forLoopLine.currentValue);
					LogHandler::getInstance()->write(VariableChangeLog(common::LOG_INFO, forLoopLine.variableName, FLOAT64_PROPERTY, to<string>(forLoopLine.currentValue), logMessage, CommandBox::getID()));
				}
			}

			if (!ended) {
				// reset the container commands
				_container->reset();

				// restart the list of commands
				_container->startExecution();

				// return without setting the command finished
				return;

			} else {
				LogHandler::getInstance()->write(ForLoopEndLog(common::LOG_INFO, _name, CommandBox::getID()));
			}
		}

		setFinished();
	}
}

void ForLoopCommandBox::onExecutionPause() {

	if (isRunning()) {
		setPaused();
	}

	_parent->onCommandBoxPaused(this);
}

void ForLoopCommandBox::onCommandBoxFinished(CommandBox* commandBox) {
	// Do nothing
}

void ForLoopCommandBox::onContainerProgressChanged(float64 progression) {

	// calculate number of steps: it is the same for all lines, so use first one
	ForLoopLine& forLoopLine = _forLoopLines[0];

	int nSteps = 1;

	// calculate number of steps
	if (forLoopLine.forLoopType == RANGE) {
		nSteps = (int)((forLoopLine.endValue - forLoopLine.startValue) / forLoopLine.stepSize) + 1;
	} else if (forLoopLine.forLoopType == VALUES) {
		nSteps = forLoopLine.values.size();
	}
	float64 progressPerStep = 100.0 / nSteps;
	// calculate total progress
	float64 currentProgress = (_iStep * progressPerStep + progression / nSteps);
	onProgressChanged(currentProgress);
}

void ForLoopCommandBox::addVariableNamesToList(std::vector<std::string>& variableNames) {
	// add the variable name corresponding to this for loop
	// add the variable names corresponding to this for loop
	for (unsigned int i = 0; i < _forLoopLines.size(); i++) {
		ForLoopLine& forLoopLine = _forLoopLines[i];
		variableNames.push_back(forLoopLine.variableName);
	}
	// get any variables from parent (in case it is a for loop, or contained in one)
	parent()->addVariableNamesToList(variableNames);
}

void ForLoopCommandBox::changeVariableName(const std::string& formerName, const std::string& newName) {
	_container->changeVariableName(formerName, newName);
}

bool ForLoopCommandBox::referencesUserVariable(const std::string& variableName) {
	return _container->referencesUserVariable(variableName);
}

std::set<std::string> ForLoopCommandBox::getReferencedUserVariables() {
	return _container->getReferencedUserVariables();
}

void ForLoopCommandBox::setInsideForLoop() {
	_container->setInsideForLoop();
}

void ForLoopCommandBox::setVariable(const std::string& variableName, float64 value) {
	_container->setVariable(variableName, value);
}

void ForLoopCommandBox::setLogID() {
	CommandBox::setLogID();

	_container->setLogIDs();
}

std::string ForLoopCommandBox::toXMLString(const std::string& space) const {

	string result = space + "<for_loop id=\"" + to<string>(getID()) + "\" logId=\"" + to<string>(logID()) + "\">\n";
	for (unsigned int i = 0; i < _forLoopLines.size(); i++) {
		const ForLoopLine& forLoopLine = _forLoopLines[i];

		string line = space + "\t<for_loop_line ";

		string type;
		string values;

		if (forLoopLine.forLoopType == RANGE) {
			type = "range";
			values = "startValue=\"" + to<string>(forLoopLine.startValue) + "\" endValue=\"" + to<string>(forLoopLine.endValue) + "\" step=\"" + to<string>(forLoopLine.stepSize) + "\"";
		} else {
			type = "values";
			values = "values=\"{";
			for (unsigned int j = 0; j < forLoopLine.values.size(); j++) {
				values += to<string>(forLoopLine.values[j]);

				if (j < forLoopLine.values.size() - 1) {
					values += ", ";
				} else {
					values += "}\"";
				}
			}
		}

		line = line + " type=\"" + type + "\" variableName=\"" + forLoopLine.variableName + "\" " + values + "/>\n";

		result = result + line;

	}

	result += _container->toXMLString(space);

	result += space + "</for_loop>";

	return result;
}

std::string ForLoopCommandBox::toString(const std::string& space) const {

	string result = space + "for ";
	for (unsigned int i = 0; i < _forLoopLines.size(); i++) {
		const ForLoopLine& forLoopLine = _forLoopLines[i];
		string values;

		if (forLoopLine.forLoopType == RANGE) {

			string type = (forLoopLine.forLoopStepType == ForLoopCommandBox::DELTA ? "Delta":"NbPoints");

			values = "from " + to<string>(forLoopLine.startValue) + " to " + to<string>(forLoopLine.endValue) + " step " + to<string>(forLoopLine.stepSize) + " type " + type;
		} else {
			values = "in {";
			for (unsigned int j = 0; j < forLoopLine.values.size(); j++) {
				values += to<string>(forLoopLine.values[j]);

				if (j < forLoopLine.values.size() - 1) {
					values += ", ";
				} else {
					values += "}";
				}
			}
		}

		if (i > 0) {
			result += ", ";
		}
		result += forLoopLine.variableName + " " + values;
	}

	result += "\n";
	result += _container->toString(space);

	result += space + "end";

	return result;
}

}
