/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_FORLOOPCOMMANDBOX_H_
#define CORE_FORLOOPCOMMANDBOX_H_

#include <string>
#include <vector>
#include <common/base/Exception.h>
#include "CommandBox.h"
#include "CommandBoxParent.h"

namespace core {

class ModifiableCommandBoxContainer;

/**
 * Class defining a for loop command box.
 */
class ForLoopCommandBox : public CommandBox, public CommandBoxParent {

public:
	/**
	 * For loop type.
	 */
	typedef enum {RANGE, VALUES} ForLoopType;

	/**
	 * For loop step type.
	 */
	typedef enum {DELTA, NBPOINTS} ForLoopStepType;

	/**
	 * Sets to running mode.
	 */
	class ValuesSyntaxErrorException : public common::Exception {
	
	public:
		/**
		 * Constructor.
		 */
		ValuesSyntaxErrorException(const std::string allValues, const std::string valueWithError);
	};

	/**
	 * Constructor.
	 */
	ForLoopCommandBox(const std::string& variableName);
	
	/**
	 * destructor.
	 */
	virtual ~ForLoopCommandBox();

	/**
	 * clone.
	 */
	ForLoopCommandBox* clone() const;

	/**
	 * Returns the type of command box.
	 */
	virtual CommandBoxType getType() const;

	/**
	 * Sets the variable name.
	 */
	void setVariableName(long lineIndex, const std::string& variableName);

	/**
	 * Gets the variable name.
	 */
	const std::string& getVariableName(long lineIndex) const;

	/**
	 * Sets the start value.
	 */
	void setStartValue(long lineIndex, double startValue);

	/**
	 * Sets the end value.
	 */
	void setEndValue(long lineIndex, double endValue);

	/**
	 * Sets the step size.
	 */
	void setStepSize(long lineIndex, double stepSize);

	/**
	 * Gets the start value.
	 */
	double getStartValue(long lineIndex) const;

	/**
	 * Gets the end value.
	 */
	double getEndValue(long lineIndex) const;

	/**
	 * Gets the step size.
	 */
	double getStepSize(long lineIndex) const;

	/**
	 * Gets the current value.
	 */
	double getCurrentValue(long lineIndex) const;

	/**
	 * Sets the for loop type.
	 */
	void setForLoopType(long lineIndex, ForLoopType type);

	/**
	 * Gets the for loop type.
	 */
	ForLoopType getForLoopType(long lineIndex) const;

	/**
	 * Sets the for loop step type.
	 */
	void setForLoopStepType(long lineIndex, ForLoopStepType type);

	/**
	 * Gets the for loop step type.
	 */
	ForLoopStepType getForLoopStepType(long lineIndex) const;

	/**
	 * Sets the values of the for loop.
	 */
	void setValues(long lineIndex, const std::string& valuesString);

	/**
	 * Gets the values of the for loop.
	 */
	const std::string& getValues(long lineIndex) const;

	/**
	 * Adds a new for loop line.
	 */
	bool addForLoopLine(long lineIndex);
	
	/**
	 * Removes the for loop line at the given index.
	 */
	void removeForLoopLine(long lineIndex);
	
	/**
	 * Returns the number of for loop lines.
	 */
	long getNumberOfLines() const;

	/**
	 * Returns the command boxes.
	 */
	const std::vector<CommandBox*>& getCommandBoxes() const;
	
	/**
	 * Adds a new command box to the end of the container. 
	 * Returns the ID of the created command box.
	 */
	unsigned long addCommandBoxAtEnd(CommandBox* commandBox);
	
	/**
	 * Adds a new command box above or below another one. 
	 * Returns the ID of the created command box.
	 */
	unsigned long addCommandBox(CommandBox* commandBox, unsigned long commandBoxToShiftID, bool above);

	/**
	 * Moves a command box to the end.
	 */
	void moveCommandBoxToEnd(unsigned long commandBoxID);
	
	/**
	 * Moves a command box above or below another one.
	 */
	void moveCommandBox(unsigned long commandBoxToMoveID, unsigned long commandBoxToShiftID, bool above);

	/**
	 * Deletes the command box.
	 */
	void deleteCommandBox(unsigned long commandBoxID);

	/**
	 * Returns true if the command box can run.
	 */
	virtual bool canRun() const;

	/**
	 * Sets the database ID to be used during the execution.
	 */
	virtual void setExecutionDatabaseID(unsigned long databaseID);

	/**
	 * Execute the command box.
	 */
	virtual void execute();

	/**
	 * Initialises execution.
	 */
	virtual void initialiseExecution();

	/**
	 * Launches the command associated with the command box.
	 */
	virtual void launchCommand();

	/**
	 * Stops the command box.
	 */
	virtual void stop();

	/**
	 * Stops the command box at end.
	 */
	virtual void stopAtEnd();

	/**
	 * Pauses the command box.
	 */
	virtual void pause();

	/**
	 * Pauses the command box at end.
	 */
	virtual void pauseAtEnd();

	/**
	 * Restarts the command box.
	 */
	virtual void restart();

	/**
	 * Resets the command box.
	 */
	virtual void reset();

	/**
	 * Notifies the container the execution has begun.
	 */
	virtual void onExecutionBegin();
	
	/**
	 * Notifies the container that the execution has ended.
	 */
	virtual void onExecutionEnd();
	
	/**
	 * Notifies the container that the execution has been paused.
	 */
	virtual void onExecutionPause();

	/**
	 * Notifies the container that one if its command boxes has finished.
	 */
	virtual void onCommandBoxFinished(CommandBox* commandBox);

	/**
	 * Notifies the container that the execution progress has changed.
	 */
	virtual void onContainerProgressChanged(double progression);

	/**
	 * Adds the variable names to the list.
	 */
	virtual void addVariableNamesToList(std::vector<std::string>& variableNames);

	/**
	 * Returns true if there is no command to execute anymore.
	 * Only ForLoopCommandBox is able to return true.
	 */
	virtual bool hasEnded();

	/**
	 * Changes the variable name and updates any expressions that use it.
	 */
	virtual void changeVariableName(const std::string& formerName, const std::string& newName);

	/**
	 * Returns true if the box references a user variable.
	 */
	virtual bool referencesUserVariable(const std::string& variableName);

	/**
	 * Returns the list of referenced user variables.
	 */
	virtual std::set<std::string> getReferencedUserVariables();

	/**
	 * Sets inside for loop.
	 */
	virtual void setInsideForLoop();

	/**
	 * Needed for for loops. sets the given variable name with the value.
	 * Iterates through command containers if necessary.
	 */
	virtual void setVariable(const std::string& variableName, double value); 
	
	/**
	 * Prints the content of the container.
	 */
	virtual std::string toXMLString(const std::string& space = "") const;

	/**
	 * Prints the content of the container.
	 */
	virtual std::string toString(const std::string& space = "") const;

	/**
	 * Sets a new log ID.
	 */
	virtual void setLogID();

	/**
	 * Compute the estimated time of the command box.
	 */
	virtual double computeEstimatedTime(bool doprogression);

private:

	void constructName();
	void contructVariablesLogField();
	bool checkValue(const std::string& valueString) const;

	ModifiableCommandBoxContainer* _container;
	
	struct ForLoopLine {

		ForLoopLine() :
			startValue(0),
			endValue(0),
			stepSize(1),
			currentValue(0),
			forLoopType(RANGE),
			forLoopStepType(DELTA) {
		}
	
		std::string variableName;
		double startValue;
		double endValue;
		double stepSize;
		double currentValue;
		ForLoopType forLoopType;
		ForLoopStepType forLoopStepType;
		std::string valuesString;
		std::vector<double> values;
		
		inline bool operator== (const ForLoopLine& f) const {
			return variableName == f.variableName;
		}
		inline ForLoopLine& operator= (const ForLoopLine& f) {
			variableName = f.variableName;
			startValue = f.startValue;
			endValue = f.endValue;
			stepSize = f.stepSize;
			currentValue = f.currentValue;
			forLoopType = f.forLoopType;
			forLoopStepType = f.forLoopStepType;
			valuesString = f.valuesString;
			values = f.values;
			return *this;
		}
	};
	
	std::vector <ForLoopLine> _forLoopLines;

	unsigned long _iStep;
	bool _isStopped;
	
	std::string _name;
	std::string _variablesField;
	
	static const std::string SEPARATORS;	
};

}

#endif /*CORE_FORLOOPCOMMANDBOX_H_*/
