/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "MathParserCommandPropertyState.h"
#include "PropertyCommandBox.h"
#include "CommandZoneEventServer.h"
#include <core/base/property/MathProperty.h>
#include <core/base/MathExpression.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/base/VariableWrapper.h>
#include <common/base/PropertyType.h>
#include <common/base/TypeToString.h>
#include <common/base/StringUtilities.h>
#include <common/output/OutputStream.h>
#include <core/dataaccess/DataAccess.h>
#include <core/log/LogHandler.h>
#include <core/log/VariableChangeLog.h>
#include <core/log/SimpleLog.h>

#include <algorithm>

using namespace common;
using namespace std;

namespace core {

MathParserCommandPropertyState::MathParserCommandPropertyState(PropertyCommandBox* commandBox) :
	PropertyState(commandBox),
	_databaseID(0),
	_mathExpression(unique_ptr<MathExpression>(new MathExpression(""))),
	_isRunning(false), _debug(false) {
}

MathParserCommandPropertyState::~MathParserCommandPropertyState() {

	removeAllVariables();

	for (auto p = _properties.begin(); p != _properties.end(); p++) {
		delete p->second;
	}
}

PropertyState::State MathParserCommandPropertyState::getType() const {

	return PropertyState::MATH_PARSER;
}

void MathParserCommandPropertyState::copy(MathParserCommandPropertyState* mathParserCommandPropertyState) {
	_databaseID = mathParserCommandPropertyState->_databaseID;

	vector<pair<unsigned long, string>> expressions = mathParserCommandPropertyState->getPropertiesToSetExpressions();
	for(auto p : expressions) {
		setPropertyExpression(p.first, p.second);
	}
}

void MathParserCommandPropertyState::setExecutionDatabaseID(unsigned long databaseID) {
	_databaseID = databaseID;

	// Set the parser database ID.
	_mathExpression->setDatabaseID(databaseID);
}

unsigned long MathParserCommandPropertyState::getExecutionDatabaseID() const {
	return _databaseID;
}


bool MathParserCommandPropertyState::hasPropertyExpression(unsigned long propertyID) {

	// Search for the property.
	if (_properties.find(propertyID) != _properties.end()) {
		BaseMathProperty* property = _properties.at(propertyID);

		return property->isExpression();
	}

	// Property is not found, create a new one.
	BaseMathProperty* property = createMathProperty(propertyID);
	property->init(_databaseID);

	_properties[propertyID] = property;

	return property->isExpression();
}

const std::string& MathParserCommandPropertyState::getPropertyExpression(unsigned long propertyID) {

	// Search for the property.
	if (_properties.find(propertyID) != _properties.end()) {
		BaseMathProperty* property = _properties.at(propertyID);

		return property->getExpression();
	}

	// Property is not found, create a new one.
	BaseMathProperty* property = createMathProperty(propertyID);
	property->init(_databaseID);

	_properties[propertyID] = property;

	return property->getExpression();
}

std::string MathParserCommandPropertyState::getStringValue(unsigned long propertyID) {

	// Search for the property.
	if (_properties.find(propertyID) != _properties.end()) {
		BaseMathProperty* property = _properties.at(propertyID);

		return property->getStringValue();
	}

	// Property is not found, create a new one.
	BaseMathProperty* property = createMathProperty(propertyID);
	property->init(_databaseID);

	_properties[propertyID] = property;

	return property->getStringValue();
}

std::string MathParserCommandPropertyState::getFormattedValue(unsigned long propertyID) {

	// Search for the property.
	if (_properties.find(propertyID) != _properties.end()) {
		BaseMathProperty* property = _properties.at(propertyID);

		return property->getFormattedValue();
	}

	// Property is not found, create a new one.
	BaseMathProperty* property = createMathProperty(propertyID);
	property->init(_databaseID);

	_properties[propertyID] = property;

	return property->getFormattedValue();
}

common::ExpressionStatus MathParserCommandPropertyState::checkPropertyExpression(unsigned long propertyID, const std::string& expression, common::ExpressionError& error, std::string& message) {

	string propertyType = DataAccess::getInstance()->getPropertyType(_databaseID, propertyID);

	// Special case for the string and boolean expressions.
	if (propertyType == TypeToString<string>::convert() || propertyType == TypeToString<bool>::convert()) {
		return common::VALID_EXPRESSION;
	}

	try {
		// Update the variables if necessary.
		updateVariables();

		_mathExpression->setExpression(expression);
		_mathExpression->init();
		_mathExpression->evaluateValue();

		return common::VALID_EXPRESSION;
	}
	catch (const MathExpression::ParserError & e) {
		error = PARSER_ERROR;
		message = e.getMessage();
	}
	catch (const MathExpression::NoSuchEnumeratedValueError & e) {
		error = ENUMERATED_VALUE_ERROR;
		message = e.getMessage();
	}
	catch (const MathExpression::DisabledControllerError & e) {
		error = DISABLED_CONTROLLER_ERROR;
		message = e.getMessage();
	}
	catch (const MathExpression::UndefinedVariableError & e) {
		error = UNDEFINED_VARIABLE_ERROR;
		message = e.getMessage();
	}

	return common::INVALID_EXPRESSION;
}

common::ExpressionStatus MathParserCommandPropertyState::checkStaticPropertyExpression(unsigned long propertyID, const std::string& expression, common::ExpressionError& error, std::string& message) {

	string propertyType = DataAccess::getInstance()->getPropertyType(0, propertyID);

	// Special case for the string and boolean expressions.
	if (propertyType == TypeToString<string>::convert() || propertyType == TypeToString<bool>::convert()) {
		return common::VALID_EXPRESSION;
	}

	try {
		unique_ptr<MathExpression> mathExpression(new MathExpression(expression));

		mathExpression->init();
		mathExpression->checkStaticValue();

		return common::VALID_EXPRESSION;
	}
	catch (const MathExpression::ParserError & e) {
		error = PARSER_ERROR;
		message = e.getMessage();
	}
	catch (const MathExpression::UndefinedVariableError & e) {
		error = UNDEFINED_VARIABLE_ERROR;
		message = e.getMessage();
	}
	catch (const MathExpression::DisabledControllerError & e) {
		error = DISABLED_CONTROLLER_ERROR;
		message = e.getMessage();
	}
	catch (const MathExpression::NoSuchEnumeratedValueError & e) {
		error = ENUMERATED_VALUE_ERROR;
		message = e.getMessage();
	}

	return common::INVALID_EXPRESSION;
}

void MathParserCommandPropertyState::setPropertyExpression(unsigned long propertyID, const std::string& expression) {

	std::scoped_lock lock(_mutex);

	// Do not perform a check of the expression.
	// It should be done before on request to control the level of errors.

	string propertyType = DataAccess::getInstance()->getPropertyType(_databaseID, propertyID);

	BaseMathProperty* property = 0;

	// Search for the property.
	if (_properties.find(propertyID) != _properties.end()) {
		property = _properties.at(propertyID);

		if (property != 0) {
			property->setExpression(expression);
		}
	}
	else {
		// Property is not found, create a new one.
		property = createMathProperty(propertyID);
		property->init(_databaseID);
		property->setExpression(expression);

		_properties[propertyID] = property;
	}

	// Searching for the property.
	if (!hasPropertyToSet(propertyID)) {
		_propertiesToSet.push_back(pair<unsigned long, BaseMathProperty*>(propertyID, property));
	}

	if (_debug) {
		string message = to_string((long long)this) + string(" set property expression ") + to_string(propertyID) + " with " + property->getExpression();
		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, message));
	}
}

void MathParserCommandPropertyState::setProperties() {

	// Copy the properties to set.
	vector<pair<unsigned long, BaseMathProperty*> > currentPropertiesToSet;
	{
		std::scoped_lock lock(_mutex);
		currentPropertiesToSet = _propertiesToSet;
	}

	// Try the properties.
	// At each step we count the success and stop when the counter remains constant.
	// We must do that because some controllers require a different sequence of set values to avoid "cannot set value" errors.
	//
	// Exceptions can be of different types:
	// - Cannot set value
	// - Math parser error
	// - Controller disabled in case of expression
	// - Enumerated value does not exist in case of expression
	// In case of controller disabled or enumerated value error or math parser error, the loop is stopped immediately and properties are not retried.
	while (true) {

		// Define the set of failed properties.
		vector<pair<unsigned long, BaseMathProperty*> > failedPropertiesToSet;

		int propertySuccess = 0;
		for (vector<pair<unsigned long, BaseMathProperty*> >::iterator p = currentPropertiesToSet.begin(); p != currentPropertiesToSet.end(); p++) {
			try {
				if (p->second != 0) {
					setPropertyValueWithExpressionEvaluation(p->second);
					p->second->setServerValue(_databaseID);
				}
				propertySuccess++;
			}
			catch (const MathExpression::ParserError & e) {
				// Immediate break.
				throw SetPropertiesFailure();
			}
			catch (const MathExpression::DisabledControllerError & e) {
				// Immediate break.
				throw SetPropertiesFailure();
			}
			catch (const MathExpression::NoSuchEnumeratedValueError & e) {
				// Immediate break.
				throw SetPropertiesFailure();
			}
			catch (const MathExpression::UndefinedVariableError & e) {
				// Immediate break.
				throw SetPropertiesFailure();
			}
			catch (...) {
				// Failed to set the property. Retry at next iteration.
				failedPropertiesToSet.push_back(*p);

				if (_debug) {
					string message = to_string((long long)this) + string(" set property value ") + to_string(p->second->getPropertyID()) + " failed";
					LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, message));
				}
			}
		}

		// Copy the set of failed properties into the current set.
		currentPropertiesToSet = failedPropertiesToSet;
		if (propertySuccess == 0) {
			break;
		}
	}

	// Check whether the setting is successful.
	if (!currentPropertiesToSet.empty()) {
		throw SetPropertiesFailure();
	}
}

void MathParserCommandPropertyState::setPropertiesForEstimatedTime() {

	for (vector<pair<unsigned long, BaseMathProperty*> >::iterator p = _propertiesToSet.begin(); p != _propertiesToSet.end(); p++) {
		try {
			if (p->second != 0) {
				setPropertyValueWithExpressionEvaluation(p->second, false);
			}
		}
		catch (const MathExpression::ParserError & e) {
			// Immediate break.
			throw SetPropertiesFailure();
		}
		catch (const MathExpression::DisabledControllerError & e) {
			// Immediate break.
			throw SetPropertiesFailure();
		}
		catch (const MathExpression::NoSuchEnumeratedValueError & e) {
			// Immediate break.
			throw SetPropertiesFailure();
		}
		catch (const MathExpression::UndefinedVariableError & e) {
			// Immediate break.
			throw SetPropertiesFailure();
		}
	}
}

void MathParserCommandPropertyState::clearProperties() {

	std::scoped_lock lock(_mutex);

	_properties.clear();
	_propertiesToSet.clear();
}

void MathParserCommandPropertyState::clearPropertiesToSet() {

	std::scoped_lock lock(_mutex);

	_propertiesToSet.clear();

	if (_debug) {
		string message = to_string((long long)(this)) + string(" clear properties to set");
		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, message));
	}
}

void MathParserCommandPropertyState::refreshProperty(unsigned long propertyID) {

	if (!hasPropertyToSet(propertyID) || _isRunning) {
		_properties.at(propertyID)->getServerValue(_databaseID);
	}
}

void MathParserCommandPropertyState::reset() {

	for (auto p = _properties.begin(); p != _properties.end(); p++) {
		// If not found in the properties to set.
		if (!hasPropertyToSet(p->first)) {
			p->second->getServerValue(_databaseID);
			CommandZoneEventServer::getInstance()->commandBoxPropertyChanged(commandBox()->getID(), p->first);
		}
	}
}

void MathParserCommandPropertyState::setVariable(const std::string& variableName, double value) {

	_mathExpression->setVariableValue(variableName, value);
}

void MathParserCommandPropertyState::changeVariableName(const std::string& formerName, const std::string& newName) {

	for (auto p = _properties.begin(); p != _properties.end(); p++) {

		// Do not replace variable in a non-expression otherwise a non-expression becomes an expression which causes troubles.
		if (p->second->isExpression()) {

			// Get the expression.
			string expression = p->second->getExpression();

			// Replace the occurrences of former name by new name.
			expression = StringUtilities::replaceText(expression, formerName, newName);

			// Set the new expression.
			p->second->setExpression(expression);

			// Notify.
			CommandZoneEventServer::getInstance()->commandBoxPropertyChanged(commandBox()->getID(), p->first);
		}
	}
}

bool MathParserCommandPropertyState::referencesUserVariable(const std::string& variableName) {

	// Copy the properties to set.
	vector<pair<unsigned long, BaseMathProperty*> > currentPropertiesToSet;
	{
		std::scoped_lock lock(_mutex);
		currentPropertiesToSet = _propertiesToSet;
	}

	// Iterate on the properties and check them.
	for (vector<pair<unsigned long, BaseMathProperty*> >::iterator p = currentPropertiesToSet.begin(); p != currentPropertiesToSet.end(); p++) {
		try {
			if (p->second != 0) {
				// Get the property.
				BaseMathProperty* property = p->second;

				// Set the expression and parse it.
				_mathExpression->setExpression(property->getExpression());

				// Init the expression so that all the values are initialised and able to throw an exception.
				_mathExpression->init();

				// Check if the current expression references the variable name.
				if (_mathExpression->referencesUserVariable(variableName)) {
					return true;
				}
			}
		}
		catch (...) {
		}
	}

	return false;
}

std::set<std::string> MathParserCommandPropertyState::getReferencedUserVariables() {

	// Copy the properties to set.
	vector<pair<unsigned long, BaseMathProperty*> > currentPropertiesToSet;
	{
		std::scoped_lock lock(_mutex);
		currentPropertiesToSet = _propertiesToSet;
	}

	std::set<std::string> result;

	// Iterate on the properties and check them.
	for (vector<pair<unsigned long, BaseMathProperty*> >::iterator p = currentPropertiesToSet.begin(); p != currentPropertiesToSet.end(); p++) {
		try {
			if (p->second != 0) {
				// Get the property.
				BaseMathProperty* property = p->second;

				// Do not take string properties.
				// TODO: We should analyse the content of the string to get any variable identified with {}.
				if (property->getType() != TypeToString<string>::convert()) {

					unique_ptr<MathExpression> mathExpression(new MathExpression(property->getExpression()));

					// Init the expression so that all the values are initialised and able to throw an exception.
					mathExpression->init();

					set<string> propertyVariables = mathExpression->getUserVariableNames();
					result.insert(propertyVariables.begin(), propertyVariables.end());
				}
			}
		}
		catch (...) {
		}
	}

	return result;
}


void MathParserCommandPropertyState::setInsideForLoop() {
	// Do nothing
}

void MathParserCommandPropertyState::setToRunningMode() {

	// Copy the properties to set.
	vector<pair<unsigned long, BaseMathProperty*> > currentPropertiesToSet;
	{
		std::scoped_lock lock(_mutex);
		currentPropertiesToSet = _propertiesToSet;
	}

	// Save all the parser expressions, use actual property values for propertiesToSet and allow updates.
	_isRunning = true;

	for (vector<pair<unsigned long, BaseMathProperty*> >::iterator p = currentPropertiesToSet.begin(); p != currentPropertiesToSet.end(); p++) {
		p->second->getServerValue(_databaseID);
		CommandZoneEventServer::getInstance()->commandBoxPropertyChanged(commandBox()->getID(), p->first);
	}
}

void MathParserCommandPropertyState::setToFinishedMode() {

	// Copy the properties to set.
	vector<pair<unsigned long, BaseMathProperty*> > currentPropertiesToSet;
	{
		std::scoped_lock lock(_mutex);
		currentPropertiesToSet = _propertiesToSet;
	}

	_isRunning = false;
	for (vector<pair<unsigned long, BaseMathProperty*> >::iterator p = currentPropertiesToSet.begin(); p != currentPropertiesToSet.end(); p++) {
		CommandZoneEventServer::getInstance()->commandBoxPropertyChanged(commandBox()->getID(), p->first);
	}
}

std::set<unsigned long> MathParserCommandPropertyState::getPropertyIDs() {
	std::set<unsigned long> propertyIDs;
	for (auto it = _properties.begin(); it != _properties.end(); it++) {
		propertyIDs.insert(it->first);
	}
	return propertyIDs;
}

bool MathParserCommandPropertyState::hasProperty(unsigned long propertyID) {
	return _properties.find(propertyID) != _properties.end();
}


BaseMathProperty* MathParserCommandPropertyState::createMathProperty(unsigned long propertyID) {

	// Get the type.
	string propertyType = DataAccess::getInstance()->getPropertyType(_databaseID, propertyID);

	if (propertyType == TypeToString<int8>::convert()) {
		return new MathProperty<int8>(propertyID);
	}
	else if (propertyType == TypeToString<uint16>::convert()) {
		return new MathProperty<uint16>(propertyID);
	}
	else if (propertyType == TypeToString<uint32>::convert()) {
		return new MathProperty<uint32>(propertyID);
	}
	else if (propertyType == TypeToString<uint64>::convert()) {
		return new MathProperty<uint64>(propertyID);
	}
	else if (propertyType == TypeToString<int16>::convert()) {
		return new MathProperty<int16>(propertyID);
	}
	else if (propertyType == TypeToString<int32>::convert()) {
		return new MathProperty<int32>(propertyID);
	}
	else if (propertyType == TypeToString<int64>::convert()) {
		return new MathProperty<int64>(propertyID);
	}
	else if (propertyType == TypeToString<float32>::convert()) {
		return new MathProperty<float32>(propertyID);
	}
	else if (propertyType == TypeToString<float64>::convert()) {
		return new MathProperty<float64>(propertyID);
	}
	else if (propertyType == TypeToString<bool>::convert()) {
		return new MathProperty<bool>(propertyID);

	} else if (propertyType == TypeToString<string>::convert()) {
		return new MathProperty<string>(propertyID);

	}

	return 0;
}

std::string MathParserCommandPropertyState::getServantPropertyName(unsigned long propertyID) const {

	unsigned long servantID = DataAccess::getInstance()->getServantIDFromPropertyID(_databaseID, propertyID);
	string servantName = DataAccess::getInstance()->getServantName(_databaseID, servantID);
	string controllerType = DataAccess::getInstance()->getControllerType(_databaseID, servantID);

	DataAccess::PropertyInfo propertyInfo = DataAccess::getInstance()->getPropertyInfo(_databaseID, controllerType, propertyID);

	string suffix = "";
	if (propertyInfo.isSetpoint) {
		suffix = ".sp";
	}
	else if (propertyInfo.isOffset) {
		suffix = ".os";
	}

	return servantName + "." + propertyInfo.aliasPropertyName + suffix;
}

void MathParserCommandPropertyState::logError(const std::string& servantPropertyName, const std::string& expression, const std::string& reason) {

	string logMessage = servantPropertyName + " <- " + expression + " := error : " + reason;
	LogHandler::getInstance()->write(VariableChangeLog(common::LOG_ERROR, servantPropertyName, "expr", "", logMessage, commandBox()->getID()));
}

void MathParserCommandPropertyState::setBooleanPropertyValueWithExpressionEvaluation(BaseMathProperty* property) {

	MathProperty<bool>* booleanProperty = dynamic_cast<MathProperty<bool>*>(property);

	if (property->getExpression() == "true") {
		booleanProperty->setValue(true);
	}
	else {
		booleanProperty->setValue(false);
	}
}

void MathParserCommandPropertyState::setStringPropertyValueWithExpressionEvaluation(BaseMathProperty* property) {

	MathProperty<string>* stringProperty = dynamic_cast<MathProperty<string>*>(property);

	// Search for braces.
	string result;
	string expression = property->getExpression();

	// Iterate on the {...} sub-expressions.
	size_t pos = 0;
	while (true) {

		// Find the first { character from pos.
		size_t prev = pos;
		pos = expression.find_first_of('{', pos);

		// Find the first } character from pos + 1.
		size_t endPos = expression.find_first_of('}', pos + 1);

		// Break the loop if there is no { or }.
		if (pos == string::npos || endPos == string::npos) {
			result += expression.substr(prev);
			break;
		}

		// Add the first part of the expression.
		result += expression.substr(prev, pos - prev);

		// Define the sub expression.
		string subExpression = expression.substr(pos + 1, endPos - pos - 1);

		// Set the expression and parse it.
		_mathExpression->setExpression(subExpression);

		// Check the expression.
		try {
			// Init the expression so that all the values are initialised and able to throw an exception.
			_mathExpression->init();

			// Evaluate the expression.
			double expressionValue = _mathExpression->evaluateValue();

			// Add the result of the evaluation if there is no error.
			// Use std::ostringstream for better formatting.
			std::ostringstream os;
			os << expressionValue;
			result += os.str();
		}
		catch (...) {
			// Add the sub expression string if the expression is not valid.
			result += "{" + subExpression + "}";
		}

		pos = endPos + 1;
	}

	string value = replaceVariables(result);

	// Replace the loop variables.
	stringProperty->setValue(value);

	if (_debug) {
		string message = to_string((long long)this) + string(" set property value ") + to_string(property->getPropertyID()) + " with " + value;
		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, message));
	}
}

void MathParserCommandPropertyState::setNumericPropertyValueWithExpressionEvaluation(BaseMathProperty* property, bool log) {

	// Set the expression and parse it.
	_mathExpression->setExpression(property->getExpression());

	double expressionValue = 0;
	string servantPropertyName = getServantPropertyName(property->getPropertyID());

	try {
		// Init the expression so that all the values are initialised and able to throw an exception.
		_mathExpression->init();

		// Evaluate the expression.
		expressionValue = _mathExpression->evaluateValue();

		// Log the result of the expression if it is not constant.
		if (!_mathExpression->isConstant()) {
			string expressionValueString = to_string(expressionValue);
			string logMessage = servantPropertyName + " <- " + property->getExpression() + " := " + expressionValueString;
			if (log) {
				LogHandler::getInstance()->write(VariableChangeLog(common::LOG_INFO, servantPropertyName, "expr", expressionValueString, logMessage, commandBox()->getID()));
			}
		}

	}
	catch (const MathExpression::ParserError & e) {
		if (log) {
			logError(servantPropertyName, property->getExpression(), e.getMessage());
		}
		throw;
	}
	catch (const MathExpression::DisabledControllerError & e) {
		if (log) {
			logError(servantPropertyName, property->getExpression(), e.getMessage());
		}
		throw;
	}
	catch (const MathExpression::NoSuchEnumeratedValueError & e) {
		if (log) {
			logError(servantPropertyName, property->getExpression(), e.getMessage());
		}
		throw;
	}
	catch (const MathExpression::UndefinedVariableError & e) {
		if (log) {
			logError(servantPropertyName, property->getExpression(), e.getMessage());
		}
		throw;
	}
	catch (...) {
		if (log) {
			logError(servantPropertyName, property->getExpression(), "Unknown");
		}
		throw;
	}

	// Set the value depending on the type.
	string propertyType = property->getType();

	if (propertyType == TypeToString<int8>::convert()) {
		MathProperty<int8>* int8Property = dynamic_cast<MathProperty<int8>*>(property);
		int8Property->setValue((int8)expressionValue);

	} else if (propertyType == TypeToString<uint16>::convert()) {
		MathProperty<uint16>* uint16Property = dynamic_cast<MathProperty<uint16>*>(property);
		uint16Property->setValue((uint16)expressionValue);

	} else if (propertyType == TypeToString<uint32>::convert()) {
		MathProperty<uint32>* uint32Property = dynamic_cast<MathProperty<uint32>*>(property);
		uint32Property->setValue((uint32)expressionValue);

	} else if (propertyType == TypeToString<uint64>::convert()) {
		MathProperty<uint64>* uint64Property = dynamic_cast<MathProperty<uint64>*>(property);
		uint64Property->setValue((uint64)expressionValue);

	} else if (propertyType == TypeToString<int16>::convert()) {
		MathProperty<int16>* int16Property = dynamic_cast<MathProperty<int16>*>(property);
		int16Property->setValue((int16)expressionValue);

	} else if (propertyType == TypeToString<int32>::convert()) {
		MathProperty<int32>* int32Property = dynamic_cast<MathProperty<int32>*>(property);
		int32Property->setValue((int32)expressionValue);

	} else if (propertyType == TypeToString<int64>::convert()) {
		MathProperty<int64>* int64Property = dynamic_cast<MathProperty<int64>*>(property);
		int64Property->setValue((int64)expressionValue);

	} else if (propertyType == TypeToString<float32>::convert()) {
		MathProperty<float32>* float32Property = dynamic_cast<MathProperty<float32>*>(property);
		float32Property->setValue((float32)expressionValue);

	} else if (propertyType == TypeToString<float64>::convert()) {
		MathProperty<float64>* float64Property = dynamic_cast<MathProperty<float64>*>(property);
		float64Property->setValue((float64)expressionValue);
	}

	if (_debug) {
		string message = to_string((long long)this) + string(" set property value ") + to_string(property->getPropertyID()) + " with " + to_string(expressionValue);
		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, message));
	}
}

void MathParserCommandPropertyState::setPropertyValueWithExpressionEvaluation(BaseMathProperty* property, bool log) {

	string propertyType = property->getType();

	// Special case for boolean properties.
	if (propertyType == common::TypeToString<bool>::convert()) {
		setBooleanPropertyValueWithExpressionEvaluation(property);
	}
	// Special case for string properties.
	else if (propertyType == common::TypeToString<string>::convert()) {
		setStringPropertyValueWithExpressionEvaluation(property);
	}
	// Case for numeric properties.
	else {
		setNumericPropertyValueWithExpressionEvaluation(property, log);
	}
}

std::string MathParserCommandPropertyState::replaceVariables(const std::string& expression) {

	return _mathExpression->replaceVariables(expression);
}

void MathParserCommandPropertyState::removeAllVariables() {

	_mathExpression->removeAllVariables();
}

void MathParserCommandPropertyState::initialiseVariables() {

	// Get the full list of variables available.
	vector<string> variableNames;
	commandBox()->addVariableNamesToList(variableNames);

	// Iterate through the variable names and initialise them.
	vector<string>::iterator it = variableNames.begin();
	while (it != variableNames.end()) {
		setVariable(*it, 0);
		it++;
	}
}

void MathParserCommandPropertyState::updateVariables() {

	// Get the full list of variables available.
	vector<string> variableNames;
	commandBox()->addVariableNamesToList(variableNames);

	_mathExpression->updateVariables(variableNames);
}

bool MathParserCommandPropertyState::isRunning() const {
	return _isRunning;
}

bool MathParserCommandPropertyState::hasPropertyToSet(unsigned long propertyID) {

	std::scoped_lock lock(_mutex);

	for (auto p = _propertiesToSet.begin(); p != _propertiesToSet.end(); p++) {
		if (p->first == propertyID) {
			return true;
		}
	}

	return false;
}

std::string MathParserCommandPropertyState::getFormattedPropertyValue(unsigned long propertyID) {

	if (isRunning() || !hasPropertyExpression(propertyID)) {
		return getFormattedValue(propertyID);
	}

	return getPropertyExpression(propertyID);
}

void MathParserCommandPropertyState::setDebug() {
	_debug = true;
}

std::string MathParserCommandPropertyState::toString() {

	std::scoped_lock lock(_mutex);

	string result;

	for (auto p = _propertiesToSet.begin(); p != _propertiesToSet.end(); ++p) {
		unsigned long propertyID = p->first;
		BaseMathProperty* property = p->second;

		string alias = getServantPropertyName(propertyID);

		if (alias != "") {
			result += alias + " " + property->getExpression() + " ";
		}
	}

	return result;
}

std::vector<std::pair<unsigned long, std::string>> MathParserCommandPropertyState::getPropertiesToSetExpressions() {

	std::vector<std::pair<unsigned long, std::string>> result;

	std::scoped_lock lock(_mutex);

	for (auto p = _propertiesToSet.begin(); p != _propertiesToSet.end(); ++p) {
		result.push_back(std::make_pair(p->first, p->second->getExpression()));
	}

	return result;
}

std::vector<std::pair<unsigned long, BaseMathProperty*>>& MathParserCommandPropertyState::getPropertiesToSet() {
	return _propertiesToSet;
}

}
