/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_MATHPARSERCOMMANDPROPERTYSTATE_H_
#define CORE_MATHPARSERCOMMANDPROPERTYSTATE_H_

#include "PropertyState.h"
#include <core/base/property/BaseMathProperty.h>
#include <common/base/CommonTypes.h>
#include <unordered_map>
#include <string>
#include <vector>
#include <memory>

namespace core {

class MathExpression;

/**
 * Class defining a math parser command property state.
 */
class MathParserCommandPropertyState : public PropertyState {

public:
	/**
	 * Constructor.
	 */
	MathParserCommandPropertyState(PropertyCommandBox* commandBox);

	/**
	 * Virtual destructor.
	 */
	virtual ~MathParserCommandPropertyState();
	
	/**
	 * Gets the type.
	 */
	virtual State getType() const;
	
	/**
	 * Sets the execution database ID.
	 */
	virtual void setExecutionDatabaseID(unsigned long databaseID);

	/**
	 * Gets the execution database ID.
	 */
	virtual unsigned long getExecutionDatabaseID() const;

	/**
	 * Returns true if has a property expression.
	 */
	bool hasPropertyExpression(unsigned long propertyID);

	/**
	 * Gets the property expression.
	 */
	const std::string& getPropertyExpression(unsigned long propertyID);
	
	/**
	 * Gets the string value.
	 */
	std::string getStringValue(unsigned long propertyID);
	
	/**
	 * Gets the formatted string value.
	 */
	std::string getFormattedValue(unsigned long propertyID);

	/**
	 * Sets the property expression.
	 */
	common::ExpressionStatus checkPropertyExpression(unsigned long propertyID, const std::string& expression, common::ExpressionError& error, std::string& message);

	/**
	 * Checks the expression statically. The loop variables are not taken into account.
	 */
	static common::ExpressionStatus checkStaticPropertyExpression(unsigned long propertyID, const std::string& expression, common::ExpressionError& error, std::string& message);

	/**
	 * Sets the property expression.
	 */
 	void setPropertyExpression(unsigned long propertyID, const std::string& expression);

 	/**
	 * Sets the properties.
	 */
	virtual void setProperties();

 	/**
	 * Sets the properties.
	 */
	virtual void setPropertiesForEstimatedTime();
	
	/**
	 * Clears the properties.
	 */
	virtual void clearProperties();

	/**
	 * Clears the properties to set.
	 */
	virtual void clearPropertiesToSet();

	/**
	 * Refreshes the property.
	 */
 	virtual void refreshProperty(unsigned long propertyID);
 	
 	/**
	 * Resets the state.
	 */
 	virtual void reset();

	/**
	 * Sets the variable.
	 */
	virtual void setVariable(const std::string& variableName, double value);
	
	/**
	 * Changes the variable name.
	 */
	virtual void changeVariableName(const std::string& formerName, const std::string& newName);
	
	/**
	 * Returns true if the box references a user variable.
	 */
	virtual bool referencesUserVariable(const std::string& variableName);

	/**
	 * Returns the list of referenced user variables.
	 */
	virtual std::set<std::string> getReferencedUserVariables();

	/**
	 * Sets the state inside the for loop.
	 */
	virtual void setInsideForLoop();
	
	/**
	 * Sets to running mode.
	 */
	virtual void setToRunningMode();
	
	/**
	 * Sets to finished mode.
	 */
	virtual void setToFinishedMode();
	
	/**
	 * Gets the property IDs.
	 */
	virtual std::set<unsigned long> getPropertyIDs();
	
	/**
	 * Returns true if has the property with ID propertyID.
	 */
	virtual bool hasProperty(unsigned long propertyID);
	
	/**
	 * Returns true if is running.
	 */
	bool isRunning() const;
	
	/**
	 * Gets the formatted property value.
	 */
	virtual std::string getFormattedPropertyValue(unsigned long propertyID);

	/**
	 * Sets in debug.
	 */
	virtual void setDebug();

	/**
	 * Returns the formatted string content of the properties.
	 */
	virtual std::string toString();

	/**
	 * Returns true if the property has to be set.
	 */
	bool hasPropertyToSet(unsigned long propertyID);

	/**
	 * Gets the properties to set.
	 */
	std::vector<std::pair<unsigned long, BaseMathProperty*>>& getPropertiesToSet();

	/**
	 * Gets the properties expressions to set.
	 */
	std::vector<std::pair<unsigned long, std::string>> getPropertiesToSetExpressions();

	/**
	 * Copies the object.
	 */
	void copy(MathParserCommandPropertyState* mathParserCommandPropertyState);

private:
	BaseMathProperty* createMathProperty(unsigned long propertyID);
	std::string getServantPropertyName(unsigned long propertyID) const;

	void setBooleanPropertyValueWithExpressionEvaluation(BaseMathProperty* property);
	void setStringPropertyValueWithExpressionEvaluation(BaseMathProperty* property);
	void setNumericPropertyValueWithExpressionEvaluation(BaseMathProperty* property, bool log = true);
	void setPropertyValueWithExpressionEvaluation(BaseMathProperty* property, bool log = true);

	std::string replaceVariables(const std::string& expression);
	void removeAllVariables();
	void initialiseVariables();
	void updateVariables();
	void logError(const std::string& servantPropertyName, const std::string& expression, const std::string& reason);

	// The mutex is necessary for a safe copy that can occur in parallel with the setting of expressions.
	std::recursive_mutex _mutex;

	std::unordered_map<unsigned long, BaseMathProperty*> _properties;
	std::vector<std::pair<unsigned long, BaseMathProperty*> > _propertiesToSet;

	unsigned long _databaseID;

	// The parser is unique and global to the command box.
	// It is shared among all the property expressions so it is necessary to clear and add the "variables" at each call.
	// It could be optimized but keeping a state is more subject to errors.
	std::unique_ptr<MathExpression> _mathExpression;
	bool _isRunning;
	bool _debug;
};

}

#endif
