/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ModifiableCommandBoxContainer.h"
#include "CommandBoxMap.h"
#include "CommandBox.h"
#include <common/base/Convert.h>
#include <string>

using namespace std;
using namespace common;

namespace core {

ModifiableCommandBoxContainer::CommandBoxNotFound::CommandBoxNotFound(unsigned long commandBoxID) :
	Exception("core::ModifiableCommandBoxContainer::CommandBoxNotFound", "cannot find the command box <" + to<string>(commandBoxID) + ">") {
}

ModifiableCommandBoxContainer::ModifiableCommandBoxContainer(CommandBoxParent* parent) :
	CommandBoxContainer(parent) {
}

ModifiableCommandBoxContainer::~ModifiableCommandBoxContainer() {
}

ModifiableCommandBoxContainer* ModifiableCommandBoxContainer::clone(CommandBoxParent* parent) {
	std::scoped_lock lock(_onCommandBoxMutex);
	ModifiableCommandBoxContainer* cbc =  new ModifiableCommandBoxContainer(parent);
	cbc->CommandBoxContainer::copy(this);
	return cbc;
}

CommandBox* ModifiableCommandBoxContainer::getFirstCommandBox() {

	if (!commandBoxes().empty()) {
		return commandBoxes().front();
	}

	return 0;
}

unsigned long ModifiableCommandBoxContainer::addCommandBoxAtBeginning(CommandBox* commandBox) {

	commandBoxes().insert(commandBoxes().begin(), commandBox);
	commandBox->setParent(this);
	CommandBoxMap::getInstance()->add(commandBox);

	// Increase the counter.
	updateCommandBoxCounter(commandBox->getCommandBoxCounter());

	return commandBox->getID();
}

unsigned long ModifiableCommandBoxContainer::addCommandBox(CommandBox* commandBox, unsigned long commandBoxToShiftID, bool above) {

	bool inserted = false;

	for (vector<CommandBox*>::iterator b = commandBoxes().begin(); b != commandBoxes().end(); b++) {

		if ((*b)->getID() == commandBoxToShiftID) {

			if (above) {
				commandBoxes().insert(b, commandBox);
			} else {
				b++;
				commandBoxes().insert(b, commandBox);
			}

			inserted = true;

			if (isInsideForLoop()) {
				commandBox->setInsideForLoop();
			}

			break;
		}
	}

	if (inserted) {
		commandBox->setParent(this);
		CommandBoxMap::getInstance()->add(commandBox);
	} else {
		throw CommandBoxNotFound(commandBoxToShiftID);
	}

	// Increase the counter.
	updateCommandBoxCounter(commandBox->getCommandBoxCounter());

	return commandBox->getID();
}

void ModifiableCommandBoxContainer::moveCommandBoxToEnd(unsigned long commandBoxID) {

	for (vector<CommandBox*>::iterator b = commandBoxes().begin(); b != commandBoxes().end(); b++) {

		if ((*b)->getID() == commandBoxID) {
			CommandBox* commandBox = *b;
			commandBoxes().erase(b);
			commandBoxes().push_back(commandBox);
			return;
		}
	}

	throw CommandBoxNotFound(commandBoxID);
}

void ModifiableCommandBoxContainer::moveCommandBox(unsigned long commandBoxToMoveID, unsigned long commandBoxToShiftID, bool above) {

	if (commandBoxToMoveID == commandBoxToShiftID) {
		return;
	}

	// 1st pass : get command box to move and remove it from vector
	CommandBox* commandBoxToMove = 0;
	vector<CommandBox*>::iterator originalPosition = commandBoxes().begin();
	for (vector<CommandBox*>::iterator bm = commandBoxes().begin(); bm != commandBoxes().end(); bm++) {
		if ((*bm)->getID() == commandBoxToMoveID) {
			commandBoxToMove = *bm;
			// erase: modifies bm
			commandBoxes().erase(bm);
			originalPosition = bm;
			break;
		}
	}

	if (commandBoxToMove != 0) {

		// 2nd pass : find command box to shift
		vector<CommandBox*>::iterator bs;
		for (bs = commandBoxes().begin(); bs != commandBoxes().end(); bs++) {

			if ((*bs)->getID() == commandBoxToShiftID) {
				break;
			}
		}

		// 3rd pass : insert command box to move
		if (bs != commandBoxes().end()) {
			if (above) {
				commandBoxes().insert(bs, commandBoxToMove);
			} else {
				bs++;
				commandBoxes().insert(bs, commandBoxToMove);
			}

		} else {
			// put command box back to its original position
			commandBoxes().insert(originalPosition, commandBoxToMove);

			throw CommandBoxNotFound(commandBoxToShiftID);
		}

	} else {
		throw CommandBoxNotFound(commandBoxToMoveID);
	}

}

void ModifiableCommandBoxContainer::deleteCommandBox(unsigned long commandBoxID) {

	vector<CommandBox*>::iterator bs;
	for (bs = commandBoxes().begin(); bs != commandBoxes().end(); bs++) {

		CommandBox* commandBox = *bs;

		if ((commandBox)->getID() == commandBoxID) {
			if ((commandBox)->canBeDeleted()) {

				// Decrease the counter.
				updateCommandBoxCounter(-commandBox->getCommandBoxCounter());

				// The deletion of the command box is made in a dedicated thread in CommandBoxMap.
				CommandBoxMap::getInstance()->remove(commandBox);

				commandBoxes().erase(bs);
			}
			return;
		}
	}

	// command box not found
	throw CommandBoxNotFound(commandBoxID);
}

std::vector<unsigned long> ModifiableCommandBoxContainer::deleteFinishedCommandBoxes() {

	std::vector<unsigned long> deletedCommandBoxIds;

	std::scoped_lock lock(_onCommandBoxMutex);

	vector<CommandBox*>::iterator bs = commandBoxes().begin();

	while (bs != commandBoxes().end()) {

		CommandBox* commandBox = *bs;

		if (commandBox->isFinished()) {
			deletedCommandBoxIds.push_back(commandBox->getID());

			// The deletion of the command box is made in a dedicated thread in CommandBoxMap.
			CommandBoxMap::getInstance()->remove(commandBox);

			//delete commandBox;
			bs = commandBoxes().erase(bs);

		} else {
			bs++;
		}
	}

	return deletedCommandBoxIds;
}

std::vector<unsigned long> ModifiableCommandBoxContainer::deleteFinishedCommandBoxes(int commandBoxesToDeleteCounter, int& deletedCommandBoxesCounter) {

	std::vector<unsigned long> deletedCommandBoxIds;

	std::scoped_lock lock(_onCommandBoxMutex);

	vector<CommandBox*>::iterator bs = commandBoxes().begin();

	// Iterate until the number of deleted command boxes is reached.
	while (commandBoxesToDeleteCounter > deletedCommandBoxesCounter && bs != commandBoxes().end()) {

		CommandBox* commandBox = *bs;

		if (commandBox->isFinished()) {

			// Increase the counter of deleted command boxes.
			deletedCommandBoxesCounter += commandBox->getCommandBoxCounter();

			deletedCommandBoxIds.push_back(commandBox->getID());

			// The deletion of the command box is made in a dedicated thread in CommandBoxMap.
			// Otherwise a deadlock because this function call is made in a CommandZone::onCommandBoxFinished() coming from DataNotificationRouter::processMessage().
			// Delete a command box performs a subscription on DataNotificationRouter that is blocked if made in this thread.
			CommandBoxMap::getInstance()->remove(commandBox);

			bs = commandBoxes().erase(bs);

		} else {
			bs++;
		}
	}

	return deletedCommandBoxIds;
}

}
