/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_PARALLELGROUP_H_
#define CORE_PARALLELGROUP_H_

#include <common/base/Exception.h>
#include <map>

namespace core {

class CommandBox;

/**
 * Class defining a parallel group. The parallel group adds commands boxes to be executed
 * in parallel and executes them. It receives notifications of their termination and is 
 * completed when all the command boxes have finished.
 *
 */
class ParallelGroup {

public:
	/**
	 * Constructor.
	 */
	ParallelGroup();
	
	/**
	 * Virtual destructor. Deletes all the command boxes.
	 */
	virtual ~ParallelGroup();

	/**
	 * Adds a new command box to the end of the container. 
	 */
	void addCommandBox(CommandBox* commandBox);

	/**
	 * Returns the last command box of the container. 
	 */
	CommandBox* getLastCommandBox();

	/**
	 * Resets the group.
	 */
	void reset();
	
	/**
	 * Executes the group.
	 */
	void execute();

	/**
	 * Notifies the container that one if its command boxes has finished.
	 */
	void onCommandBoxFinished(CommandBox* commandBox);
	
	/**
	 * Returns true if the group is completed.
	 */
	bool isCompleted() const;
	
	/**
	 * Returns the minimum progression for all commands in group.
	 */
	double calculateMinimumProgression();
		
private:
	bool allCommandsFinished() const;
	
	struct CommandBoxStatus {
		bool started;
		bool terminated;
	};
	
	std::map<CommandBox*, CommandBoxStatus> _commandBoxes;
	CommandBox* _lastCommandBox;
	
	bool _stopped;
	bool _isCompleted;
};

}


#endif /*CORE_PARALLELGROUP_H_*/
