/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "PropertyCommandBox.h"
#include "MathParserCommandPropertyState.h"
#include <core/dataaccess/DataNotificationRouter.h>
#include <thread>
#include "EstimatedTimeManager.h"
#include "SaveBoxPropertyManager.h"

using namespace std;

namespace core {

PropertyCommandBox::PropertyCommandBox(bool isEstimated) :
	CommandBox(isEstimated),
	_propertyState(std::unique_ptr<PropertyState>(new MathParserCommandPropertyState(this))) {
}

PropertyCommandBox::~PropertyCommandBox() {
	// Do not unsubscribe from property changed here, because this can lead to a crash when the command box is automatically removed.
	// Indeed a message can be sent to the command box that is being destroyed: a pure virtual method called can happen.
}

void PropertyCommandBox::copy(const PropertyCommandBox* propertyCommandBox) {
	CommandBox::copy(propertyCommandBox);
	dynamic_cast<MathParserCommandPropertyState*>(_propertyState.get())->copy(dynamic_cast<MathParserCommandPropertyState*>(propertyCommandBox->_propertyState.get()));
}

void PropertyCommandBox::setPropertyState(PropertyState* state) {
	_propertyState.reset(state);
}

std::string PropertyCommandBox::getFormattedPropertyValue(unsigned long propertyID) {

	try {
		if (!_propertyState->hasProperty(propertyID)) {
			// subscribe to property changes
			if (_forEstimatedTime == false) {
				DataNotificationRouter::getInstance()->addPropertyChangeSubscriber(propertyID, this);
			}
		}

		return _propertyState->getFormattedPropertyValue(propertyID);

	} catch (PropertyAccessor::NullPropertyReferenceException& e) {
		throw link(NullPropertyReferenceException(), e);
	} catch (PropertyAccessor::NoSuchPropertyException& e) {
		throw link(NoSuchPropertyException(), e);
	} catch (PropertyAccessor::BadPropertyTypeException& e) {
		throw link(BadPropertyType(), e);
	} catch (PropertyState::BadPropertyType& e) {
		throw link(BadPropertyType(), e);
	}
}

common::ExpressionStatus PropertyCommandBox::checkPropertyExpression(unsigned long propertyID, const std::string & expression, common::ExpressionError& error, std::string& message) {
	return dynamic_cast<MathParserCommandPropertyState*>(_propertyState.get())->checkPropertyExpression(propertyID, expression, error, message);
}

common::ExpressionStatus PropertyCommandBox::checkStaticPropertyExpression(unsigned long propertyID, const std::string & expression, common::ExpressionError& error, std::string& message) {
	return MathParserCommandPropertyState::checkStaticPropertyExpression(propertyID, expression, error, message);
}

void PropertyCommandBox::setPropertyExpression(unsigned long propertyID, const std::string & expression) {

	try {
		if (!_propertyState->hasProperty(propertyID)) {
			// subscribe to property changes
			if (_forEstimatedTime == false) {
				DataNotificationRouter::getInstance()->addPropertyChangeSubscriber(propertyID, this);
			}
		}

		// Direct call to setPropertyExpression.
		dynamic_cast<MathParserCommandPropertyState*>(_propertyState.get())->setPropertyExpression(propertyID, expression);

		CommandZoneEventServer::getInstance()->commandBoxPropertyChanged(getID(), propertyID);
		EstimatedTimeManager::getInstance()->newEstimatedTime();

	} catch (PropertyAccessor::NullPropertyReferenceException& e) {
		throw link(NullPropertyReferenceException(), e);
	} catch (PropertyAccessor::NoSuchPropertyException& e) {
		throw link(NoSuchPropertyException(), e);
	} catch (PropertyAccessor::BadPropertyTypeException& e) {
		throw link(BadPropertyType(), e);
	} catch (PropertyState::BadPropertyType& e) {
		throw link(BadPropertyType(), e);
	}
}

std::string PropertyCommandBox::getPropertyExpression(unsigned long propertyID) {

	try {
		if (!_propertyState->hasProperty(propertyID)) {
			// subscribe to property changes
			if (_forEstimatedTime == false) {
				DataNotificationRouter::getInstance()->addPropertyChangeSubscriber(propertyID, this);
			}
		}

		// Direct call to setPropertyExpression.
		return dynamic_cast<MathParserCommandPropertyState*>(_propertyState.get())->getPropertyExpression(propertyID);

	} catch (PropertyAccessor::NullPropertyReferenceException& e) {
		throw link(NullPropertyReferenceException(), e);
	} catch (PropertyAccessor::NoSuchPropertyException& e) {
		throw link(NoSuchPropertyException(), e);
	} catch (PropertyAccessor::BadPropertyTypeException& e) {
		throw link(BadPropertyType(), e);
	} catch (PropertyState::BadPropertyType& e) {
		throw link(BadPropertyType(), e);
	}
}

bool PropertyCommandBox::referencesUserVariable(const std::string& variableName) {
	return _propertyState->referencesUserVariable(variableName);
}

std::set<std::string> PropertyCommandBox::getReferencedUserVariables() {
	return _propertyState->getReferencedUserVariables();
}

vector<pair<unsigned long,string>> PropertyCommandBox::getPropertyExpressions(PropertyState* state) {
	vector<pair<unsigned long,string>> properties;
	set<unsigned long> propertyIDs = _propertyState->getPropertyIDs();
	MathParserCommandPropertyState* math = dynamic_cast<MathParserCommandPropertyState*>(_propertyState.get());
	for(auto p : propertyIDs) {
		properties.push_back(make_pair(p, math->getPropertyExpression(p)));
	}
	return properties;
}


void PropertyCommandBox::initPropertyExpression(unsigned int propertyID) {
	string savedValue = "";
	try {
		savedValue = SaveBoxPropertyManager::getInstance()->getPropertySavedInBoxValue(propertyID);
	} catch (SaveBoxPropertyManager::NoSuchPropertyException & e) {
		savedValue = getPropertyValue<string>(propertyID);
	}
	setPropertyExpression(propertyID, savedValue);
}

}
