/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_PROPERTYCOMMANDBOX_H_
#define CORE_PROPERTYCOMMANDBOX_H_

#include "CommandBox.h"
#include <core/dataaccess/PropertyChangeSubscriber.h>
#include <common/base/Exception.h>
#include <common/base/CommonTypes.h>
#include <memory>

namespace core {

class PropertyState;

/**
 * Class defining a property command box.
 */
class PropertyCommandBox : public CommandBox, public PropertyChangeSubscriber {

public:
	/**
	 * No such property exception.
	 */
	class NoSuchPropertyException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchPropertyException();
	};

	/**
	 * Null property reference exception.
	 */
	class NullPropertyReferenceException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NullPropertyReferenceException();
	};

	/**
	 * Bad property type exception.
	 */
	class BadPropertyType : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		BadPropertyType();
	};


	/**
	 * Constructor.
	 */
	PropertyCommandBox(bool isEstimated = false);

	/**
	 * Virtual destructor.
	 */
	virtual ~PropertyCommandBox();

	/**
	 * Constructor.
	 */
	void copy(const PropertyCommandBox* propertyCommandBox);

	/**
	 * Sets the property state
	 */
	void setPropertyState(PropertyState* state);


	/**
	 * Sets the property value.
	 */
	template<typename ValueType>
	void setPropertyValue(unsigned long propertyID, ValueType value);

	/**
	 * Gets the property value.
	 */
	template<typename ValueType>
	ValueType getPropertyValue(unsigned long propertyID);

	/**
	 * Gets the formatted property value.
	 */
	std::string getFormattedPropertyValue(unsigned long propertyID);

	/**
	 * Checks the property expression and sets the message if there is an error.
	 */
	common::ExpressionStatus checkPropertyExpression(unsigned long propertyID, const std::string & expression, common::ExpressionError& error, std::string& message);

	/**
	 * Checks the property expression statically and sets the message if there is an error.
	 */
	static common::ExpressionStatus checkStaticPropertyExpression(unsigned long propertyID, const std::string & expression, common::ExpressionError& error, std::string& message);

	/**
	 * Sets the property expression.
	 */
	void setPropertyExpression(unsigned long propertyID, const std::string & expression);

	/**
	 * Gets the property expression.
	 */
	std::string getPropertyExpression(unsigned long propertyID);

	std::vector<std::pair<unsigned long,std::string>> getPropertyExpressions(PropertyState* state);

	/**
	 * Returns true if the box references a user variable.
	 */
	virtual bool referencesUserVariable(const std::string& variableName);

	/**
	 * Returns the list of referenced user variables.
	 */
	virtual std::set<std::string> getReferencedUserVariables();

	void initPropertyExpression(unsigned int propertyID);


protected:
	std::unique_ptr<PropertyState> _propertyState;
};

}

#include "PropertyCommandBox.inl"

#endif /*CORE_PROPERTYCOMMANDBOX_H_*/
