/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_PROPERTYSTATE_H_
#define CORE_PROPERTYSTATE_H_

#include <common/base/Exception.h>
#include <common/base/HashMap.h>
#include <core/base/property/BaseProperty.h>
#include <string>
#include <set>

namespace core {

class PropertyCommandBox;

/**
 * Class defining a base class for the property state.
 */
class PropertyState {

public:
	/**
	 * Bad property type exception.
	 */
	class BadPropertyType : public common::Exception {
	
	public:
		/**
		 * Constructor.
		 */
		BadPropertyType(unsigned long propertyID);
	};
	
	/**
	 * Set properties failure exception.
	 */
	class SetPropertiesFailure : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		SetPropertiesFailure();
	};

	/**
	 * State type.
	 */
	typedef enum {DIRECT, MATH_PARSER} State;

	/**
	 * Constructor.
	 */
	PropertyState(PropertyCommandBox* commandBox);

	/**
	 * Virtual destructor.
	 */
	virtual ~PropertyState();

	/**
	 * Gets the type.
	 */
	virtual State getType() const = 0;
	
	/**
	 * Sets the properties.
	 */
	virtual void setProperties() = 0;

	/**
	 * Sets the properties.
	 */
	virtual void setPropertiesForEstimatedTime() = 0;

	/**
	 * Refreshes the property.
	 */
 	virtual void refreshProperty(unsigned long propertyID) = 0;
 	
 	/**
	 * Resets the state.
	 */
 	virtual void reset() = 0;

	/**
	 * Sets the variable.
	 */
	virtual void setVariable(const std::string& variableName, double value) = 0;
	
	/**
	 * Changes the variable name.
	 */
	virtual void changeVariableName(const std::string& formerName, const std::string& newName) = 0;
	
	/**
	 * Returns true if the box references a user variable.
	 */
	virtual bool referencesUserVariable(const std::string& variableName) = 0;

	/**
	 * Returns the list of referenced user variables.
	 */
	virtual std::set<std::string> getReferencedUserVariables() = 0;

	/**
	 * Sets the state inside the for loop.
	 */
	virtual void setInsideForLoop() = 0;
	
	/**
	 * Sets to running mode.
	 */
	virtual void setToRunningMode() = 0;
	
	/**
	 * Sets to finished mode.
	 */
	virtual void setToFinishedMode() = 0;
	
	/**
	 * Gets the property IDs.
	 */
	virtual std::set<unsigned long> getPropertyIDs() = 0;

	/**
	 * Clears the properties.
	 */
	virtual void clearProperties() = 0;

	/**
	 * Clears the properties to set.
	 */
	virtual void clearPropertiesToSet() = 0;
	
	/**
	 * Returns true if has the property.
	 */
	virtual bool hasProperty(unsigned long propertyID) = 0;
	
	/**
	 * Sets the execution database ID.
	 */
	virtual void setExecutionDatabaseID(unsigned long databaseID) = 0;
	
	/**
	 * Gets the execution database ID.
	 */
	virtual unsigned long getExecutionDatabaseID() const = 0;
	
	/**
	 * Gets the formatted property value.
	 */
	virtual std::string getFormattedPropertyValue(unsigned long propertyID) = 0;

	/**
	 * Sets in debug.
	 */
	virtual void setDebug() = 0;

	/**
	 * Returns the formatted string content of the properties.
	 */
	virtual std::string toString() = 0;

protected:
	PropertyCommandBox* commandBox();
	
private:
	PropertyCommandBox* _commandBox;
};

}

#endif /*CORE_PROPERTYSTATE_H_*/
