/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ScanCommandBox.h"
#include "ScanCommandBoxContainer.h"
#include <common/base/Convert.h>
#include <core/log/LogHandler.h>
#include <core/log/CommandLog.h>
#include <core/log/CommandZoneContentLog.h>
#include <core/log/UserScanStartLog.h>
#include <ics/IcsServerManager.h>

using namespace std;
using namespace common;

namespace core {

ScanCommandBox::ScanCommandBox(const std::string& name, const std::string& fullpath) :
	_name(name), _fullpath(fullpath),
	_container(new ScanCommandBoxContainer(this)) {
}

ScanCommandBox::~ScanCommandBox() {
	delete _container;
}

ScanCommandBox* ScanCommandBox::clone() const  {
	ScanCommandBox* scb = new ScanCommandBox(_name, _fullpath);
	scb->_container->copy(_container);
	return scb;
}

const std::string& ScanCommandBox::getName() const {
	return _name;
}

const std::string& ScanCommandBox::getFullPath() const {
	return _fullpath;
}

CommandBox::CommandBoxType ScanCommandBox::getType() const {
	return CommandBox::SCAN;
}

const std::vector<CommandBox*>& ScanCommandBox::getCommandBoxes() const {
	return _container->getCommandBoxes();
}

unsigned long ScanCommandBox::addCommandBoxAtEnd(CommandBox* commandBox) {
	return _container->addCommandBoxAtEnd(commandBox);
}

bool ScanCommandBox::canRun() const {
	return _container->canRun();
}

void ScanCommandBox::setExecutionDatabaseID(unsigned long databaseID) {
	_container->setExecutionDatabaseID(databaseID);
}

void ScanCommandBox::execute() {

	initialiseExecution();
	launchCommand();
}


double ScanCommandBox::computeEstimatedTime(bool doprogression) {
	double tst = _container->computeEstimatedTime(doprogression);
	return _container->computeEstimatedTime(doprogression);
}

void ScanCommandBox::initialiseExecution() {

	// Sets the content of the command zone to the data provider
	LogHandler::getInstance()->write(CommandZoneContentLog(toXMLString()));
}

void ScanCommandBox::launchCommand() {
	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::getInstance()->onScanCommandBoxLaunch( _fullpath);
	}

	std::string message = string("Users'scripts     --> ") + getName();

	// write log
	LogHandler::getInstance()->write(CommandLog(common::START, common::LOG_INFO, message, logID()));

	// New log message
	LogHandler::getInstance()->write(UserScanStartLog(common::LOG_INFO, _name, message));

	_container->startExecution();
}

void ScanCommandBox::stop() {
	_container->stop();

	// Do not set finished, otherwise the scan command box will finish before the sub-commands
}

void ScanCommandBox::stopAtEnd() {
	if (isRunning()) {
		_container->stopAtEnd();
	}

	// Do not set finished, otherwise the scan command box will finish before the sub-commands
}

void ScanCommandBox::pause() {
	if (isRunning() && !isPaused()) {
		_container->pause();
		setPaused();
	}
}

void ScanCommandBox::pauseAtEnd() {
	if (isRunning()) {
		_container->pauseAtEnd();
	}
}

void ScanCommandBox::restart() {
	if (isPaused()) {
		setRunning();
		_container->restart();
	}
}

void ScanCommandBox::reset() {
	_container->reset();
	setConstruction();
}

void ScanCommandBox::onExecutionBegin() {

	if (!isRunning() && !isFinished()) {
		setRunning();
	}
}

void ScanCommandBox::onExecutionEnd() {

	if (isRunning() || isPaused()) {
		setFinished();
	}
}

void ScanCommandBox::onExecutionPause() {

	if (isRunning()) {
		setPaused();
	}

	_parent->onCommandBoxPaused(this);
}

void ScanCommandBox::onCommandBoxFinished(CommandBox* commandBox) {
	// Do nothing
}

void ScanCommandBox::onContainerProgressChanged(double progression) {
	onProgressChanged(progression);
}

bool ScanCommandBox::hasEnded() {
	return true;
}

void ScanCommandBox::addVariableNamesToList(std::vector<std::string>& variableNames) {
	// get any variables from parent (in case it is a for loop, or contained in one)
	parent()->addVariableNamesToList(variableNames);
}

void ScanCommandBox::changeVariableName(const std::string& formerName, const std::string& newName) {
	_container->changeVariableName(formerName, newName);
}

bool ScanCommandBox::referencesUserVariable(const std::string& variableName) {
	return _container->referencesUserVariable(variableName);
}

std::set<std::string> ScanCommandBox::getReferencedUserVariables() {
	return _container->getReferencedUserVariables();
}

void ScanCommandBox::setInsideForLoop() {
	_container->setInsideForLoop();
}

void ScanCommandBox::setVariable(const std::string& variableName, double value) {
	_container->setVariable(variableName, value);
}

void ScanCommandBox::setLogID() {
	CommandBox::setLogID();

	_container->setLogIDs();
}

std::string ScanCommandBox::toXMLString(const std::string& space) const {

	string result = space + "<scan id=\"" + to<string>(getID()) + "\" logId=\"" + to<string>(logID()) + "\" name=\"" + _name + "\">\n";

	result += _container->toXMLString(space);

	result += space + "</scan>";

	return result;
}

std::string ScanCommandBox::toString(const std::string& space) const {

	string result = space + "scan " + _name + "\n";

	result += _container->toString(space);

	result += space + "end";

	return result;
}

}
