/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_SCANCOMMANDBOX_H_
#define CORE_SCANCOMMANDBOX_H_

#include <string>
#include <vector>
#include "CommandBox.h"
#include "CommandBoxParent.h"

namespace core {

class ScanCommandBoxContainer;

/**
 * Class defining a scan command box.
 */
class ScanCommandBox : public CommandBox, public CommandBoxParent {

public:
	/**
	 * Constructor.
	 */
	ScanCommandBox(const std::string& name, const std::string& fullpath);

	/**
	 * Virtual destructor.
	 */
	virtual ~ScanCommandBox();

	/**
	 * Constructor.
	 */
	ScanCommandBox* clone() const;

	/**
	 * Gets the scan name.
	 */
	const std::string& getName() const;

	/**
	 * Gets the full path name.
	 */
	const std::string& getFullPath() const;

	/**
	 * Returns the type of command box.
	 */
	virtual CommandBoxType getType() const;
		
	/**
	 * Returns the command boxes.
	 */
	const std::vector<CommandBox*>& getCommandBoxes() const;
	
	/**
	 * Adds a new command box to the end of the container. 
	 * Returns the ID of the created command box.
	 */
	unsigned long addCommandBoxAtEnd(CommandBox* commandBox);

	/**
	 * Returns true if the command box can run.
	 */
	virtual bool canRun() const;

	/**
	 * Sets the database ID to be used during the execution.
	 */
	virtual void setExecutionDatabaseID(unsigned long databaseID);

	/**
	 * Execute the command box.
	 */
	virtual void execute();

	/**
	 * Initialises execution.
	 */
	virtual void initialiseExecution();

	/**
	 * Launches the command associated with the command box.
	 */
	virtual void launchCommand();

	/**
	 * Stops the command box.
	 */
	virtual void stop();

	/**
	 * Stops the command box at end.
	 */
	virtual void stopAtEnd();

	/**
	 * Pauses the command box.
	 */
	virtual void pause();

	/**
	 * Pauses the command box at end.
	 */
	virtual void pauseAtEnd();

	/**
	 * Restarts the command box.
	 */
	virtual void restart();

	/**
	 * Resets the command box.
	 */
	virtual void reset();

	/**
	 * Notifies the container the execution has begun.
	 */
	virtual void onExecutionBegin();
	
	/**
	 * Notifies the container that the execution has ended.
	 */
	virtual void onExecutionEnd();
	
	/**
	 * Notifies the container that the execution has been paused.
	 */
	virtual void onExecutionPause();

	/**
	 * Notifies the container that one if its command boxes has finished.
	 */
	virtual void onCommandBoxFinished(CommandBox* commandBox);

	/**
	 * Notifies the container that the execution progress has changed.
	 */
	virtual void onContainerProgressChanged(double progression);
	
	/**
	 * Returns true if there is no command to execute anymore.
	 * Only ForLoopCommandBox is able to return true.
	 */
	virtual bool hasEnded();

	/**
	 * Adds the variable names to the list.
	 */
	virtual void addVariableNamesToList(std::vector<std::string>& variableNames);

	/**
	 * Changes the variable name and updates any expressions that use it.
	 */
	virtual void changeVariableName(const std::string& formerName, const std::string& newName);
	
	/**
	 * Returns true if the box references a user variable.
	 */
	virtual bool referencesUserVariable(const std::string& variableName);

	/**
	 * Returns the list of referenced user variables.
	 */
	virtual std::set<std::string> getReferencedUserVariables();

	/**
	 * Sets inside for loop.
	 */
	virtual void setInsideForLoop();
	
	/**
	 * Needed for for loops. sets the given variable name with the value.
	 * Iterates through command containers if necessary.
	 */
	virtual void setVariable(const std::string& variableName, double value); 

	/**
	 * Prints the content of the box.
	 */
	virtual std::string toXMLString(const std::string& space = std::string()) const;

	/**
	 * Prints the content of the box.
	 */
	virtual std::string toString(const std::string& space = std::string()) const;

	/**
	 * Sets a new log ID.
	 */
	virtual void setLogID();

	/**
	 * Compute the estimated time of the command box.
	 */
	virtual double computeEstimatedTime(bool doprogression);

private:
	std::string _name;
	std::string _fullpath;
	ScanCommandBoxContainer* _container;
	
};

}

#endif /*CORE_SCANCOMMANDBOX_H_*/
