/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "BreakCommandBox.h"
#include <core/commandzone/ForLoopCommandBox.h>
#include <core/commandzone/CommandBoxContainer.h>
#include <core/commandzone/CommandZoneEventServer.h>
#include <core/log/LogHandler.h>
#include <core/log/CommandZoneContentLog.h>
#include <core/log/SimpleLog.h>
#include <core/log/BreakEvaluationLog.h>
#include <common/output/OutputStream.h>
#include <common/base/StringUtilities.h>
#include <iostream>

using namespace common;
using namespace std;

namespace core {

BreakCommandBox::BreakCommandBox() :
	ControlCommandBox(ControlCommandBox::BREAK) {
}

BreakCommandBox::~BreakCommandBox() {
}

BreakCommandBox* BreakCommandBox::clone() const {
	BreakCommandBox* bcb = new BreakCommandBox();
	bcb->ControlCommandBox::copy(this);
	return bcb;
}

bool BreakCommandBox::canRun() const {
	return true;
}

void BreakCommandBox::setExecutionDatabaseID(unsigned long databaseID) {
}

void BreakCommandBox::execute() {

	// testing command box place
	if (getForLoopParent() == 0) {
		err << "Break command must have a ForLoop command box as parent" << eol;
		setFinished();
		return;
	}

	initialiseExecution();

	// launch if not on error
	if (!_onError) {
		launchCommand();

	} else {
		// forcing setRunning
		setRunning();
		setFinished();
	}
}

std::string BreakCommandBox::getName() {

	if (_expression != "") {
		return string("Break ") + _expression;
	}

	return string("Break ?");
}

ForLoopCommandBox* BreakCommandBox::getForLoopParent() {

	if (parent() != 0) {
		return dynamic_cast<ForLoopCommandBox*>(parent()->getParent());
	}

	return 0;
}

void BreakCommandBox::initialiseExecution() {

	_onError = false;

	// Sets the content of the command zone to the data provider
	LogHandler::getInstance()->write(CommandZoneContentLog(toXMLString()));

	// Initialisation can fail.
	try {
		// Evaluate the condition.
		_conditionValue = evaluateCondition();

		// notifying clients
		CommandZoneEventServer::getInstance()->commandBoxExpressionValueChanged(getID(), _conditionValue);
	}
	catch (const Exception& e) {
		_onError = true;
		err << e.getMessage() << eol;

		std::string outputMessage = getName() + " := error : " + e.getMessage();
		LogHandler::getInstance()->write(BreakEvaluationLog(common::LOG_ERROR, _expression, "expr", "", outputMessage, getForLoopParent()->getID()));

		// Stop the execution.
		getForLoopParent()->stop();
		LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, getName() + " skipped execution"));
	}
}

void BreakCommandBox::launchCommand() {

	// forcing setRunning
	setRunning();

	if (_conditionValue) {
		// stopping the for loop
		getForLoopParent()->stop();

		//New Break log message
		std::string value = "true";
		std::string outputMessage = getName() + " := " + value;
		LogHandler::getInstance()->write(BreakEvaluationLog(common::LOG_INFO, _expression, "expr", value, outputMessage, getForLoopParent()->getID()));
	}
	else {
		//New Break log message
		std::string value = "false";
		std::string outputMessage = getName() + " := " + value;
		LogHandler::getInstance()->write(BreakEvaluationLog(common::LOG_INFO, _expression, "expr", value, outputMessage, getForLoopParent()->getID()));
	}

	setFinished();
}

void BreakCommandBox::stop() {

}

void BreakCommandBox::stopAtEnd() {

}

void BreakCommandBox::pause() {

}

void BreakCommandBox::pauseAtEnd() {

}

void BreakCommandBox::restart() {

}

void BreakCommandBox::reset() {

	setConstruction();
}

void BreakCommandBox::addVariableNamesToList(std::vector<std::string>& variableNames) {
	// get any variables from parent (in case it is a for loop, or contained in one)
	parent()->addVariableNamesToList(variableNames);
}

void BreakCommandBox::setVariable(const std::string& variableName, double value) {

	_booleanExpression->setVariableValue(variableName, value);
}

void BreakCommandBox::changeVariableName(const std::string& formerName, const std::string& newName) {

	_expression = StringUtilities::replaceText(_expression, formerName, newName);

	CommandZoneEventServer::getInstance()->commandBoxExpressionChanged(getID());
}

void BreakCommandBox::setInsideForLoop() {

}

std::string BreakCommandBox::toXMLString(const std::string& space) const {

	string result = space + "<break id=\"" + to<string>(getID()) + "\" logId=\"" + to<string>(logID()) + "\" condition=\"" + convertStringToXML(_expression) + "\"/>";

	return result;
}

std::string BreakCommandBox::toString(const std::string& space) const {

	string result = space + "break " + convertStringToXML(_expression);

	return result;
}

}
