/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ControlCommandBox.h"
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <core/dataaccess/DataAccess.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <common/base/Convert.h>
#include <common/output/OutputStream.h>

using namespace std;
using namespace common;

namespace core {

ControlCommandBox::EvaluateConditionFailure::EvaluateConditionFailure() :
	Exception("core::ControlCommandBox::EvaluateExpressionFailure", string("Failed to evaluate the condition")) {
}

ControlCommandBox::ControlCommandBox(BoxType type) :
	_boxType(type),
	_booleanExpression(std::unique_ptr<BooleanExpression>(new BooleanExpression(""))),
	_onError(false),
	_conditionValue(false) {
}

ControlCommandBox::~ControlCommandBox() {
}

void ControlCommandBox::copy(const ControlCommandBox* ControlCommandBox) {
	CommandBox::copy(ControlCommandBox);
	_boxType = ControlCommandBox->_boxType;
	_onError = ControlCommandBox->_onError;
	_conditionValue = ControlCommandBox->_conditionValue;
}

ControlCommandBox::BoxType ControlCommandBox::getBoxType() const {
	return _boxType;
}

CommandBox::CommandBoxType ControlCommandBox::getType() const {

	switch (_boxType) {
	case IF:
		return CommandBox::IF;
	case BREAK:
		return CommandBox::BREAK;
	case IF_ELSE:
		return CommandBox::IF_ELSE;
	}

	return CommandBox::IF;
}

bool ControlCommandBox::referencesUserVariable(const std::string& variableName) {

	// Set the expression and parse it.
	_booleanExpression->setExpression(_expression);

	// Init the expression so that all the values are initialised and able to throw an exception.
	try {
		_booleanExpression->init();
	}
	catch (...) {
		// Should not happen.
		return false;
	}

	return _booleanExpression->referencesUserVariable(variableName);
}

std::set<std::string> ControlCommandBox::getReferencedUserVariables() {
	return _booleanExpression->getUserVariableNames();
}

void ControlCommandBox::updateVariables() {

	// Get the full list of variables available.
	vector<string> variableNames;
	addVariableNamesToList(variableNames);

	_booleanExpression->updateVariables(variableNames);
}

common::ExpressionStatus ControlCommandBox::checkExpression(const std::string& expression, common::ExpressionError& error, std::string& message) {

	try {
		// Update the variables if necessary.
		updateVariables();

		_booleanExpression->setExpression(expression);
		_booleanExpression->init();
		_booleanExpression->evaluateValue();

		return common::VALID_EXPRESSION;
	}
	catch (const MathExpression::ParserError & e) {
		error = common::PARSER_ERROR;
		message = e.getMessage();
	}
	catch (const MathExpression::NoSuchEnumeratedValueError & e) {
		error = common::ENUMERATED_VALUE_ERROR;
		message = e.getMessage();
	}
	catch (const MathExpression::DisabledControllerError & e) {
		error = common::DISABLED_CONTROLLER_ERROR;
		message = e.getMessage();
	}
	catch (const MathExpression::UndefinedVariableError & e) {
		error = common::UNDEFINED_VARIABLE_ERROR;
		message = e.getMessage();
	}

	return common::INVALID_EXPRESSION;
}

void ControlCommandBox::setExpression(const std::string& expression) {
	_expression = expression;

	slog << "Control command box " << getID() << " sets expression " << expression << eol;
}

const std::string& ControlCommandBox::getExpression() {
	return _expression;
}

bool ControlCommandBox::evaluateCondition() {

	// Set the expression and parse it.
	_booleanExpression->setExpression(_expression);

	try {
		// Init the expression so that all the values are initialised and able to throw an exception.
		_booleanExpression->init();

		// Evaluate the expression.
		return _booleanExpression->evaluateValue();
	}
	catch (const MathExpression::ParserError & e) {
		throw;
	}
	catch (const MathExpression::DisabledControllerError & e) {
		throw;
	}
	catch (const MathExpression::NoSuchEnumeratedValueError & e) {
		throw;
	}
	catch (const MathExpression::UndefinedVariableError & e) {
		throw;
	}
	catch (...) {
		throw EvaluateConditionFailure();
	}
}

}
