/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_CONTROLCOMMANDBOX_H_
#define CORE_CONTROLCOMMANDBOX_H_

#include "../CommandBox.h"
#include "../CheckExpressionCommandBox.h"
#include <core/base/BooleanExpression.h>
#include <core/dataaccess/ConfigurationChangeSubscriber.h>
#include <common/base/Exception.h>
#include <memory>

namespace core {

/**
 * Class defining a control command box.
 */
class ControlCommandBox : public CommandBox, public CheckExpressionCommandBox {

public:
	/**
	 * Evaluate expression failure exception.
	 */
	class EvaluateConditionFailure : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		EvaluateConditionFailure();
	};

	enum BoxType {
		IF, BREAK, IF_ELSE
	};

	/**
	 * Constructor.
	 */
	ControlCommandBox(BoxType type);

	/**
	 * Virtual destructor.
	 */
	virtual ~ControlCommandBox();

	/**
	 * copy.
	 */
	void copy(const ControlCommandBox* ControlCommandBox);

	/**
	 * Gets the type.
	 */
	BoxType getBoxType() const;

	/**
	 * Returns the type of command box.
	 */
	virtual CommandBoxType getType() const;

	/**
	 * Returns true if the box references a user variable.
	 */
	virtual bool referencesUserVariable(const std::string& variableName);

	/**
	 * Returns the list of referenced user variables.
	 */
	virtual std::set<std::string> getReferencedUserVariables();

	/**
	 * Verifies the expression.
	 */
	common::ExpressionStatus checkExpression(const std::string& expression, common::ExpressionError& error, std::string& message);

	/**
	 * Sets the expression for the if condition.
	 */
	void setExpression(const std::string& expression);

	/**
	 * Gets the expression for the if condition.
	 */
	const std::string& getExpression();

	/**
	 * Evaluates the condition.
	 */
	bool evaluateCondition();

	/**
	 * Compute the estimated time of the command box.
	 */
	virtual double computeEstimatedTime(bool doprogression) {
		return 0.;
	}

protected:
	void updateVariables();

	BoxType _boxType;
	std::string _expression;
	std::unique_ptr<BooleanExpression> _booleanExpression;
	bool _onError;
	bool _conditionValue;
};

}

#endif
