/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "IfCommandBox.h"
#include <core/commandzone/ModifiableCommandBoxContainer.h>
#include <core/commandzone/CommandZoneEventServer.h>
#include <core/log/LogHandler.h>
#include <core/log/CommandZoneContentLog.h>
#include <core/log/SimpleLog.h>
#include <core/log/IfEvaluationLog.h>
#include <core/dataaccess/DataAccess.h>
#include <common/base/StringUtilities.h>
#include <common/base/Convert.h>

using namespace common;
using namespace std;

namespace core {

IfCommandBox::IfCommandBox(BoxType type) :
	ControlCommandBox(type),
	_thenContainer(new ModifiableCommandBoxContainer(this)),
	_elseContainer(new ModifiableCommandBoxContainer(this)) {
}

IfCommandBox::~IfCommandBox() {

	delete _thenContainer;
	delete _elseContainer;
}

IfCommandBox* IfCommandBox::clone() const {
	IfCommandBox* icb = new IfCommandBox(_boxType);
	icb->ControlCommandBox::copy(this);
	delete icb->_thenContainer;
	delete icb->_elseContainer;
	icb->_thenContainer = _thenContainer->clone(icb);
	icb->_elseContainer = _elseContainer->clone(icb);
	return icb;
}

const std::vector<CommandBox*>& IfCommandBox::getCommandBoxes(bool thenBlock) const {
	if (thenBlock) {
		return _thenContainer->getCommandBoxes();
	}
	return _elseContainer->getCommandBoxes();
}

std::string IfCommandBox::getName() {

	if (_expression != "") {
		return string("If ") + _expression;
	}

	return string("If ?");
}

unsigned long IfCommandBox::addCommandBoxAtEnd(bool thenBlock, CommandBox* commandBox) {
	if (thenBlock) {
		return _thenContainer->addCommandBoxAtEnd(commandBox);
	}
	return _elseContainer->addCommandBoxAtEnd(commandBox);
}

unsigned long IfCommandBox::addCommandBox(bool thenBlock, CommandBox* commandBox, unsigned long commandBoxToShiftID, bool above) {
	if (thenBlock) {
		return _thenContainer->addCommandBox(commandBox, commandBoxToShiftID, above);
	}
	return _elseContainer->addCommandBox(commandBox, commandBoxToShiftID, above);
}

void IfCommandBox::moveCommandBoxToEnd(bool thenBlock, unsigned long commandBoxID) {
	if (thenBlock) {
		_thenContainer->moveCommandBoxToEnd(commandBoxID);
	} else {
		_elseContainer->moveCommandBoxToEnd(commandBoxID);
	}
}

void IfCommandBox::moveCommandBox(bool thenBlock, unsigned long commandBoxToMoveID, unsigned long commandBoxToShiftID, bool above) {
	if (thenBlock) {
		_thenContainer->moveCommandBox(commandBoxToMoveID, commandBoxToShiftID, above);
	} else {
		_elseContainer->moveCommandBox(commandBoxToMoveID, commandBoxToShiftID, above);
	}
}

void IfCommandBox::deleteCommandBox(bool thenBlock, unsigned long commandBoxID) {
	if (thenBlock) {
		_thenContainer->deleteCommandBox(commandBoxID);
	} else {
		_elseContainer->deleteCommandBox(commandBoxID);
	}
}

bool IfCommandBox::canRun() const {
	return _thenContainer->canRun() && _elseContainer->canRun();
}

void IfCommandBox::setExecutionDatabaseID(unsigned long databaseID) {
	_thenContainer->setExecutionDatabaseID(databaseID);
	_elseContainer->setExecutionDatabaseID(databaseID);
}

void IfCommandBox::execute() {

	initialiseExecution();

	// launch if not on error
	if (!_onError) {
		launchCommand();

	} else {
		// forcing setRunning
		setRunning();
		setFinished();
	}
}

void IfCommandBox::initialiseExecution() {

	_onError = false;

	// Sets the content of the command zone to the data provider
	LogHandler::getInstance()->write(CommandZoneContentLog(toXMLString()));

	// Initialisation can fail.
	try {
		// Evaluate the condition.
		_conditionValue = evaluateCondition();

		// notifying clients
		CommandZoneEventServer::getInstance()->commandBoxExpressionValueChanged(getID(), _conditionValue);
	}
	catch (const Exception& e) {
		_onError = true;
		err << e.getMessage() << eol;

		std::string outputMessage = getName() + " := error : " + e.getMessage();
		LogHandler::getInstance()->write(IfEvaluationLog(common::LOG_ERROR, _expression, "expr", "", outputMessage));
		LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, getName() + " skipped execution"));
	}
}

void IfCommandBox::launchCommand() {

	// test boolean expression
	if (_conditionValue) {
		//New IF log message
		std::string value = "true";
		std::string outputMessage = getName() + " := " + value;
		LogHandler::getInstance()->write(IfEvaluationLog(common::LOG_INFO, _expression, "expr", value, outputMessage));

		_thenContainer->startExecution();
	}
	else {
		//New If log message
		std::string value = "false";
		std::string outputMessage = getName() + " := " + value;
		LogHandler::getInstance()->write(IfEvaluationLog(common::LOG_INFO, _expression, "expr", value, outputMessage));

		if (_elseContainer->getFirstCommandBox() != 0) {
			_elseContainer->startExecution();
		}
		else {
			// the execution is immediately terminated.
			setRunning();
			setFinished();
		}
	}
}

void IfCommandBox::stop() {
	if (_conditionValue) {
		_thenContainer->stop();
	} else {
		_elseContainer->stop();
	}
}

void IfCommandBox::stopAtEnd() {
	if (isRunning()) {
		if (_conditionValue) {
			_thenContainer->stopAtEnd();
		} else {
			_elseContainer->stopAtEnd();
		}
		setStoppingAtEnd();
	}
}

void IfCommandBox::pause() {
	if (isRunning() && !isPaused()) {
		if (_conditionValue) {
			_thenContainer->pause();
		} else {
			_elseContainer->pause();
		}
		setPaused();
	}
}

void IfCommandBox::pauseAtEnd() {
	if (isRunning()) {
		if (_conditionValue) {
			_thenContainer->pauseAtEnd();
		} else {
			_elseContainer->pauseAtEnd();
		}
		setPausingAtEnd();
	}
}

void IfCommandBox::restart() {
	if (isPaused()) {
		setRunning();
		if (_conditionValue) {
			_thenContainer->restart();
		} else {
			_elseContainer->restart();
		}
	}
}

void IfCommandBox::reset() {
	_thenContainer->reset();
	_elseContainer->reset();

	setConstruction();
}

void IfCommandBox::addVariableNamesToList(std::vector<std::string>& variableNames) {
	// get any variables from parent (in case it is a for loop, or contained in one)
	parent()->addVariableNamesToList(variableNames);
}

void IfCommandBox::setVariable(const std::string& variableName, double value) {

	_booleanExpression->setVariableValue(variableName, value);

	_thenContainer->setVariable(variableName, value);
	_elseContainer->setVariable(variableName, value);
}

void IfCommandBox::changeVariableName(const std::string& formerName, const std::string& newName) {

	_expression = StringUtilities::replaceText(_expression, formerName, newName);

	CommandZoneEventServer::getInstance()->commandBoxExpressionChanged(getID());

	_thenContainer->changeVariableName(formerName, newName);
	_elseContainer->changeVariableName(formerName, newName);
}

void IfCommandBox::setInsideForLoop() {
	_thenContainer->setInsideForLoop();
	_elseContainer->setInsideForLoop();
}

void IfCommandBox::onExecutionBegin() {
	if (!isRunning() && !isFinished()) {
		setRunning();
	}
}

void IfCommandBox::onExecutionEnd() {
	if (isRunning() || isPaused()) {
		setFinished();
	}
}

void IfCommandBox::onExecutionPause() {

	if (isRunning()) {
		setPaused();
	}

	_parent->onCommandBoxPaused(this);
}

void IfCommandBox::onCommandBoxFinished(CommandBox* commandBox) {
	// Do nothing
}

void IfCommandBox::onContainerProgressChanged(double progression) {
	onProgressChanged(progression);
}

bool IfCommandBox::hasEnded() {
	return true;
}

void IfCommandBox::setLogID() {
	CommandBox::setLogID();

	_thenContainer->setLogIDs();
	_elseContainer->setLogIDs();
}

double IfCommandBox::computeEstimatedTime(bool doprogression) {
	double thentime = _thenContainer->computeEstimatedTime(doprogression);
	double elsetime = _elseContainer->computeEstimatedTime(doprogression);
	if (thentime > elsetime) {
		return thentime;
	}
	else {
		return elsetime;
	}
}

std::string IfCommandBox::toXMLString(const std::string& space) const {

	string result = space + "<if id=\"" + to<string>(getID()) + "\" logId=\"" + to<string>(logID()) + "\" condition=\"" + convertStringToXML(_expression) + "\">\n";
	result += space + "\t" + "<then>\n";
	result += _thenContainer->toXMLString(space + "\t");
	result += space + "\t" + "</then>\n";
	result += space + "\t" + "<else>\n";
	result += _elseContainer->toXMLString(space + "\t");
	result += space + "\t" + "</else>\n";
	result += space + "</if>";

	return result;
}

std::string IfCommandBox::toString(const std::string& space) const {

	string result = space + "if " + convertStringToXML(_expression) + "\n";
	result += space + "then\n";
	result += _thenContainer->toString(space + "\t");

	string elseString = _elseContainer->toString(space + "\t");

	if (elseString != "") {
		result += space + "else\n";
		result += elseString;
	}
	result += space + "end";

	return result;
}

}
