/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_IFCOMMANDBOX_H_
#define CORE_IFCOMMANDBOX_H_

#include "ControlCommandBox.h"
#include <core/commandzone/CommandBoxParent.h>

namespace core {

class ModifiableCommandBoxContainer;

/**
 * Class defining an if command box.
 */
class IfCommandBox : public ControlCommandBox, public CommandBoxParent {

public:
	/**
	 * Constructor.
	 */
	IfCommandBox(BoxType type);

	/**
	 * Virtual destructor.
	 */
	virtual ~IfCommandBox();

	/**
	 * Constructor.
	 */
	IfCommandBox* clone() const;

	/**
	 * Returns the command boxes.
	 */
	const std::vector<CommandBox*>& getCommandBoxes(bool thenBlock) const;

	/**
	 * Adds a new command box to the end of the container.
	 * Returns the ID of the created command box.
	 */
	unsigned long addCommandBoxAtEnd(bool thenBlock, CommandBox* commandBox);

	/**
	 * Adds a new command box above or below another one.
	 * Returns the ID of the created command box.
	 */
	unsigned long addCommandBox(bool thenBlock, CommandBox* commandBox, unsigned long commandBoxToShiftID, bool above);

	/**
	 * Moves a command box to the end.
	 */
	void moveCommandBoxToEnd(bool thenBlock, unsigned long commandBoxID);

	/**
	 * Moves a command box above or below another one.
	 */
	void moveCommandBox(bool thenBlock, unsigned long commandBoxToMoveID, unsigned long commandBoxToShiftID, bool above);

	/**
	 * Deletes the command box.
	 */
	void deleteCommandBox(bool thenBlock, unsigned long commandBoxID);

	/**
	 * Returns true if the command box can run.
	 */
	virtual bool canRun() const;

	/**
	 * Sets the database ID to be used during the execution.
	 */
	virtual void setExecutionDatabaseID(unsigned long databaseID);

	/**
	 * Execute the command box.
	 */
	virtual void execute();

	/**
	 * Initialises execution.
	 */
	virtual void initialiseExecution();

	/**
	 * Launches the command associated with the command box.
	 */
	virtual void launchCommand();

	/**
	 * Stops the command box.
	 */
	virtual void stop();

	/**
	 * Stops the command box at end.
	 */
	virtual void stopAtEnd();

	/**
	 * Pauses the command box.
	 */
	virtual void pause();

	/**
	 * Pauses the command box at end.
	 */
	virtual void pauseAtEnd();

	/**
	 * REstarts the command box.
	 */
	virtual void restart();

	/**
	 * Resets the command box.
	 */
	virtual void reset();

	/**
	 * Adds the variable names to the list.
	 */
	virtual void addVariableNamesToList(std::vector<std::string>& variableNames);

	/**
	 * Needed for for loops. sets the given variable name with the value.
	 * Iterates through command containers if necessary.
	 */
	virtual void setVariable(const std::string& variableName, double value);

	/**
	 * Changes the variable name and updates any expressions that use it.
	 */
	virtual void changeVariableName(const std::string& formerName, const std::string& newName);

	/**
	 * Sets inside for loop.
	 */
	virtual void setInsideForLoop();

	/**
	 * Notifies the container the execution has begun.
	 */
	virtual void onExecutionBegin();

	/**
	 * Notifies the container that the execution has ended.
	 */
	virtual void onExecutionEnd();

	/**
	 * Notifies the container that the execution has been paused.
	 */
	virtual void onExecutionPause();

	/**
	 * Notifies the container that one if its command boxes has finished.
	 */
	virtual void onCommandBoxFinished(CommandBox* commandBox);

	/**
	 * Notifies the container that the execution progress has changed.
	 */
	virtual void onContainerProgressChanged(double progression);

	/**
	 * Returns true if there is no command to execute anymore.
	 * Only ForLoopCommandBox is able to return true.
	 */
	virtual bool hasEnded();

	/**
	 * Prints the content of the box.
	 */
	virtual std::string toXMLString(const std::string& space = std::string()) const;

	/**
	 * Prints the content of the box.
	 */
	virtual std::string toString(const std::string& space = std::string()) const;

	/**
	 * Sets a new log ID.
	 */
	virtual void setLogID();

	/**
	 * Compute the estimated time of the command box.
	 */
	virtual double computeEstimatedTime(bool doprogression);

private:
	std::string getName();

	ModifiableCommandBoxContainer* _thenContainer;
	ModifiableCommandBoxContainer* _elseContainer;
};

}

#endif /*CORE_IFCOMMANDBOX_H_*/
