/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "../PropertyState.h"
#include "../CommandSequencer.h"
#include "../CommandBoxContainer.h"
#include "../CommandZoneEventServer.h"
#include "../CommandZone.h"
#include <core/scheduler/Sender.h>
#include <core/scheduler/Scheduler.h>
#include <common/base/Convert.h>
#include <common/base/TypeToString.h>
#include <common/base/StringUtilities.h>
#include <core/scheduler/Scheduler.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/dataaccess/DataAccess.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <core/log/CommandZoneContentLog.h>
#include <ics/InstrumentManager/InstrumentManager.h>
#include "PropertyGenericCommandBox.h"
#include "GenericCommandBoxManager.h"

using namespace std;
using namespace common;

namespace core {

PropertyCommandBox::NoSuchPropertyException::NoSuchPropertyException() :
	Exception("core::PropertyCommandBox::NoSuchPropertyException", "no such property") {
}

PropertyCommandBox::NullPropertyReferenceException::NullPropertyReferenceException() :
	Exception("core::PropertyCommandBox::NullPropertyReferenceException", "null property reference") {
}

PropertyCommandBox::BadPropertyType::BadPropertyType() :
	Exception("core::PropertyCommandBox::BadPropertyType", "bad property type") {
}

void PropertyCommandBoxHelper<std::string>::setMathParserPropertyValue(MathParserCommandPropertyState* propertyState, unsigned long propertyID, std::string value) {

	propertyState->setPropertyExpression(propertyID, value);
}

std::string PropertyCommandBoxHelper<std::string>::getMathParserPropertyValue(MathParserCommandPropertyState* propertyState, unsigned long propertyID) {

	if (propertyState->isRunning() || !propertyState->hasPropertyExpression(propertyID)) {
		return string(propertyState->getStringValue(propertyID));
	}

	return string(propertyState->getPropertyExpression(propertyID));
}

PropertyGenericCommandBox::PropertyGenericCommandBox(PropertyGenericCommandBox::BoxType type, bool forEstimatedTime) :
	GenericCommandBox(type),
	PropertyCommandBox(forEstimatedTime),
	_controllerID(0) {
}

PropertyGenericCommandBox::~PropertyGenericCommandBox() {
	// Unsubscribe from property changes.
	if (_forEstimatedTime == false) {
		DataNotificationRouter::getInstance()->removePropertyChangeSubscriber(_propertyState->getPropertyIDs(), this);
	}
}

void PropertyGenericCommandBox::copy(const PropertyGenericCommandBox* genericCommandBox) {
	GenericCommandBox::copy(genericCommandBox);
	PropertyCommandBox::copy(genericCommandBox);
	_controllerName = genericCommandBox->_controllerName;
	_propertyName = genericCommandBox->_propertyName;
	_value = genericCommandBox->_value;
	_controllerID = genericCommandBox->_controllerID;
}

unsigned long PropertyGenericCommandBox::getExecutionDatabaseID() const {
	return _propertyState->getExecutionDatabaseID();
}

CommandBox::CommandBoxType PropertyGenericCommandBox::getType() const {
	return CommandBox::GENERIC;
}

void PropertyGenericCommandBox::setControllerName(const std::string& name) {
	_controllerName = name;
	try {
		_controllerID = DataAccess::getInstance()->getServantID(_controllerName);
		GenericCommandBoxManager::getInstance()->setLastControllerName(getBoxType(), _controllerName);
	}
	catch (...) {
		// Do nothing.
	}
}

const std::string& PropertyGenericCommandBox::getControllerName() {
	return _controllerName;
}

void PropertyGenericCommandBox::setPropertyName(const std::string& name) {
	_propertyName = name;
	GenericCommandBoxManager::getInstance()->setLastSetPropertyName(_propertyName);
}

const std::string& PropertyGenericCommandBox::getPropertyName() {
	return _propertyName;
}

std::tuple<std::string, int> PropertyGenericCommandBox::findPropertyName() const {

	std::string propertyName = _propertyName;
	int index = -1;
	std::size_t position = _propertyName.find('.');
	if (position != string::npos) {

		// The name of the property.
		propertyName = _propertyName.substr(0, position);

		// The last token.
		string indexToken = _propertyName.substr(position + 1);

		// Test if it is an integer.
		try {
			index = std::stoi(indexToken);
		}
		catch (...) {
			LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, "Cannot set " + _controllerName + " with a bad property name " + _propertyName));
		}
	}

	string controllerType = DataAccess::getInstance()->getControllerType(getExecutionDatabaseID(), _controllerID);

	// Get the property infos.
	vector<DataAccess::PropertyInfoItem> items = DataAccess::getInstance()->getAllPropertyInfoItems(getExecutionDatabaseID(), controllerType);

	// Look for the existence of an alias.
	// If the alias is not found, then take the property as is.
	for (auto i : items) {
		// Search for the alias without whitespace.
		if (i.aliasPropertyName == _propertyName) {

			// Select the setpoint or actual property.
			if (i.setpointPropertyName.empty()) {
				propertyName = i.actualPropertyName;
			}
			else {
				propertyName = i.setpointPropertyName;
			}
		}
	}

	return std::make_tuple(propertyName, index);
}

unsigned long PropertyGenericCommandBox::findPropertyID() const {

	auto res = findPropertyName();
	string propertyName = std::get<0>(res);
	int index = std::get<1>(res);

	// Set the property command box.
	unsigned long propertyID = 0;

	try {
		// Check dynamic property.
		if (index != -1) {
			unsigned long dynamicPropertyID = core::DataAccess::getInstance()->getPropertyID(getExecutionDatabaseID(), _controllerID, propertyName);
			propertyID = core::DataAccess::getInstance()->getDynamicPropertyIDAtIndex(0, dynamicPropertyID, index, false);
		} else {
			propertyID = DataAccess::getInstance()->getPropertyID(getExecutionDatabaseID(), _controllerID, propertyName);
		}
	} catch (...) {

	}

	return propertyID;
}

void PropertyGenericCommandBox::setValue(const std::string& value) {
	_value = value;

	if (_controllerID != 0) {

		if (_propertyName != "") {

			// Find the property ID.
			unsigned long propertyID = findPropertyID();

			try {
				PropertyCommandBox::setPropertyExpression(propertyID, value);
			}
			catch (...) {
				LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, "Cannot set value of " + _controllerName + " " + _propertyName));
			}
		}
		else {
			LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, "Cannot set " + _controllerName + " with an empty property"));
		}
	}
}

const std::string& PropertyGenericCommandBox::getValue() {
	return _value;
}

void PropertyGenericCommandBox::addVariableNamesToList(std::vector<string>& variableNames) {
	parent()->addVariableNamesToList(variableNames);
}

bool PropertyGenericCommandBox::canRun() const {
	return true;
}

void PropertyGenericCommandBox::setExecutionDatabaseID(unsigned long databaseID) {
	_propertyState->setExecutionDatabaseID(databaseID);
}

void PropertyGenericCommandBox::setProperty() {

	if (_propertyName != "") {

		_propertyState->clearProperties();

		// Getting the property ID, checking if it exists
		try {
			// Find the property ID.
			unsigned long propertyID = findPropertyID();

			if (_propertyState->getType() == PropertyState::DIRECT) {

				string propertyType = DataAccess::getInstance()->getPropertyType(getExecutionDatabaseID(), propertyID);

				if (propertyType == TypeToString<int8>::convert()) {
					setPropertyValue<int8>(propertyID, to<int8>(_value));

				} else if (propertyType == TypeToString<uint16>::convert()) {
					setPropertyValue<uint16>(propertyID, to<uint16>(_value));

				} else if (propertyType == TypeToString<uint32>::convert()) {
					setPropertyValue<uint32>(propertyID, to<uint32>(_value));

				} else if (propertyType == TypeToString<uint64>::convert()) {
					setPropertyValue<uint64>(propertyID, to<uint64>(_value));

				} else if (propertyType == TypeToString<int16>::convert()) {
					setPropertyValue<int16>(propertyID, to<int16>(_value));

				} else if (propertyType == TypeToString<int32>::convert()) {
					setPropertyValue<int32>(propertyID, to<int32>(_value));

				} else if (propertyType == TypeToString<int64>::convert()) {
					setPropertyValue<int64>(propertyID, to<int64>(_value));

				} else if (propertyType == TypeToString<float32>::convert()) {
					setPropertyValue<float32>(propertyID, to<float32>(_value));

				} else if (propertyType == TypeToString<float64>::convert()) {
					setPropertyValue<float64>(propertyID, to<float64>(_value));

				} else if (propertyType == TypeToString<bool>::convert()) {
					setPropertyValue<bool>(propertyID, to<bool>(_value));

				} else if (propertyType == TypeToString<string>::convert()) {
					setPropertyValue<string>(propertyID, string(_value));

				} else if (propertyType == TypeToString<std::string>::convert()) {
					setPropertyValue<string>(propertyID, string(_value));
				}

			} else {
				setPropertyValue<string>(propertyID, string(_value));
			}

		} catch (...) {
			// write log
			LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, "Cannot set property " + _controllerName + " " + _propertyName));
		}
	}
	else {
		LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, "Cannot set " + _controllerName + " with an empty property"));
	}
}

void PropertyGenericCommandBox::execute() {

	initialiseExecution();
	launchCommand();
}

void PropertyGenericCommandBox::initialiseExecution() {

	// Sets the database ID for the subscribers
	setDatabaseID(getExecutionDatabaseID());

	// Sets the content of the command zone to the data provider
	LogHandler::getInstance()->write(CommandZoneContentLog(toXMLString()));

	// log ID?

	// set the property
	setProperty();
	_propertyState->setProperties();
}

void PropertyGenericCommandBox::launchCommand() {
	_propertyState->setToRunningMode();
	setRunning();

	executeCommand();

	_propertyState->setToFinishedMode();
	setFinished();
}

void PropertyGenericCommandBox::stop() {
}

void PropertyGenericCommandBox::stopAtEnd() {
}

void PropertyGenericCommandBox::pause() {
}

void PropertyGenericCommandBox::pauseAtEnd() {
}

void PropertyGenericCommandBox::restart() {
}

void PropertyGenericCommandBox::reset() {
	setConstruction();
	try {
		_propertyState->reset();

	} catch (Exception& e) {
		out << "cannot reset the property state : " << e.printStack() << eol;
	}
}

void PropertyGenericCommandBox::changeVariableName(const std::string& formerName, const std::string& newName) {

	// Change value.
	_value = StringUtilities::replaceText(_value, formerName, newName);

	// Change the variable name in the state which sends a commandBoxPropertyChanged event which is not listened by the client.
	_propertyState->changeVariableName(formerName, newName);

	// Also send a commandBoxExpressionChanged event which is listened by the client to update its views.
	CommandZoneEventServer::getInstance()->commandBoxExpressionChanged(getID());
}

void PropertyGenericCommandBox::setVariable(const std::string& variableName, double value) {
	_propertyState->setVariable(variableName, value);
}

void PropertyGenericCommandBox::setInsideForLoop() {
	_propertyState->setInsideForLoop();
}

void PropertyGenericCommandBox::commandStateChanged(unsigned long databaseID, unsigned long commandID, common::CommandState state) {
}

void PropertyGenericCommandBox::commandProgressChanged(unsigned long databaseID, unsigned long commandID, double progression) {
}

void PropertyGenericCommandBox::commandTerminated(unsigned long databaseID, unsigned long commandID, unsigned long requestID, bool sameRequest) {
}

void PropertyGenericCommandBox::propertyChanged(unsigned long databaseID, unsigned long propertyID) {
}

}
