/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_PROPERTYGENERICCOMMANDBOX_H_
#define CORE_PROPERTYGENERICCOMMANDBOX_H_

#include "../PropertyCommandBox.h"
#include "GenericCommandBox.h"
#include <core/dataaccess/CommandStateChangeSubscriber.h>
#include <core/dataaccess/CommandProgressChangeSubscriber.h>
#include <core/dataaccess/CommandTerminatedSubscriber.h>
#include <core/dataaccess/PropertyChangeSubscriber.h>
#include <core/dataaccess/DataAccess.h>
#include <common/base/Exception.h>
#include <common/base/Queue.h>
#include <boost/thread/recursive_mutex.hpp>
#include <vector>
#include <string>

namespace core {

class PropertyState;

/**
 * Class defining a generic command box.
 */
class PropertyGenericCommandBox : public PropertyCommandBox,
	public GenericCommandBox,
	public CommandStateChangeSubscriber,
	public CommandProgressChangeSubscriber,
	public CommandTerminatedSubscriber {

public:
	/**
	 * Constructor.
	 */
	PropertyGenericCommandBox(BoxType type, bool forEstimatedTime = false);

	/**
	 * Virtual destructor.
	 */
	virtual ~PropertyGenericCommandBox();

	/**
	 * Constructor.
	 */
	void copy(const PropertyGenericCommandBox* genericCommandBox);

	/**
	 * Returns the current database ID.
	 */
	unsigned long getExecutionDatabaseID() const;

	/**
	 * Returns the type of command box.
	 */
	virtual CommandBoxType getType() const;

	/**
	 * Sets the controller name.
	 */
	void setControllerName(const std::string& name);

	/**
	 * Gets the controller name.
	 */
	const std::string& getControllerName();

	/**
	 * Sets the property name.
	 */
	void setPropertyName(const std::string& name);

	/**
	 * Gets the controller name.
	 */
	const std::string& getPropertyName();

	/**
	 * Sets the value.
	 */
	void setValue(const std::string& value);

	/**
	 * Gets the value.
	 */
	const std::string& getValue();

	/**
	 * Returns true if the command box can run.
	 */
	virtual bool canRun() const;

	/**
	 * Sets the database ID to be used during the execution.
	 */
	virtual void setExecutionDatabaseID(unsigned long databaseID);

	/**
	 * Execute the command box.
	 */
	virtual void execute();

	/**
	 * Initialises execution. Sends the associated command box properties.
	 */
	virtual void initialiseExecution();

	/**
	 * Launches the command associated with the command box.
	 */
	virtual void launchCommand();

	/**
	 * Stops the command box.
	 */
	virtual void stop();

	/**
	 * Stops the command box at end.
	 */
	virtual void stopAtEnd();

	/**
	 * Pauses the command box.
	 */
	virtual void pause();

	/**
	 * Pauses the command box at end.
	 */
	virtual void pauseAtEnd();

	/**
	 * Restarts the command box.
	 */
	virtual void restart();

	/**
	 * Resets the command box.
	 */
	virtual void reset();

	/**
	 * Adds the variable names to the list.
	 */
	virtual void addVariableNamesToList(std::vector<std::string>& variableNames);

	/**
	 * Needed for for loops. sets the given variable name with the value.
	 * Iterates through command containers if necessary.
	 */
	virtual void setVariable(const std::string& variableName, double value);

	/**
	 * Changes the variable name and updates any expressions that use it.
	 */
	virtual void changeVariableName(const std::string& formerName, const std::string& newName);

	/**
	 * Sets inside for loop.
	 */
	virtual void setInsideForLoop();

	/**
	 * Notifies command box that the command state has changed.
	 */
	virtual void commandStateChanged(unsigned long databaseID, unsigned long commandID, common::CommandState state);

	/**
	 * Notifies command box that the command progress has changed.
	 */
	virtual void commandProgressChanged(unsigned long databaseID, unsigned long commandID, double progression);

	/**
	 * Notifies command box that the command has terminated. Receives the notification from the Scheduler.
	 */
	virtual void commandTerminated(unsigned long databaseID, unsigned long commandID, unsigned long requestID, bool sameRequest);

	/**
	 * Notifies command box that a property has changed.
	 */
	virtual void propertyChanged(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Compute the estimated time of the command box.
	 */
	virtual double computeEstimatedTime(bool doprogression) {
		return 0.;
	}

protected:
	void setProperty();
	virtual void executeCommand() = 0;

	std::tuple<std::string, int> findPropertyName() const;
	unsigned long findPropertyID() const;

	std::string _controllerName;
	std::string _propertyName;
	std::string _value;

	unsigned long _controllerID;
};

}

#endif /*CORE_GENERICCOMMANDBOX_H_*/
