/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Alarm.h"
#include <common/output/OutputStream.h>
#include <common/base/TimeCondition.h>
#include <common/base/Convert.h>
#include <core/commandzone/CommandZone.h>
#include <ics/CommandSystem/CommandLine/CommandLineManager.h>
#include <boost/bind.hpp>

using namespace common;
using namespace std;

namespace core {

Alarm::Alarm(const std::string& name) :
	Condition(common::CONDITION_ALARM, name),
	_activationDelay(0),
	_stopCommandZone(false),
	_delayCondition(0),
	_delayThread(nullptr) {
}

Alarm::~Alarm() {

	// for notifying the condition
	{
		std::scoped_lock lock(_mutex);

		// deleting the thread
		if (_delayThread != nullptr) {

			if (_delayCondition != 0) {
				_delayCondition->notify();
			}
		}
	}

	// delete at last
	if (_delayThread != nullptr) {
		_delayThread->join();
		delete _delayThread;
	}
}

void Alarm::setOn(bool value) {

	std::scoped_lock lock(_mutex);

	Condition::setOn(value);

	if (!value) {
		if (getState() == CONDITION_ON_ACTIVATION_DELAY) {
			_delayCondition->notify();
		} else {
			setState(CONDITION_INACTIVE);
		}
	}
}

void Alarm::setActivationDelay(int delay) {

	std::scoped_lock lock(_mutex);

	_activationDelay = delay;
}


int Alarm::getActivationDelay() {

	std::scoped_lock lock(_mutex);

	return _activationDelay;
}

void Alarm::setStopCommandZone(bool value) {

	std::scoped_lock lock(_mutex);

	_stopCommandZone = value;
}

bool Alarm::isStopCommandZone() {

	std::scoped_lock lock(_mutex);

	return _stopCommandZone;
}

void Alarm::check() {

	std::scoped_lock lock(_mutex);

	// only check when value changes
	if (isFirst() || actualValue() != oldValue()) {

		if (getState() == CONDITION_INACTIVE && actualValue()) {

			setState(CONDITION_ON_ACTIVATION_DELAY);

			_delayCondition = new TimeCondition();

			// initialising the thread
			if (_delayThread != nullptr) {
				_delayThread->join();
				delete _delayThread;
			}
			_delayThread = new std::thread(std::bind(&Alarm::delay, this));

		} else if (getState() == CONDITION_ON_ACTIVATION_DELAY && !actualValue()) {

			// stop time condition
			_delayCondition->notify();

		} else if (getState() == CONDITION_ACTIVE && !actualValue()) {

			setState(CONDITION_INACTIVE);

			out << "Alarm : " << getMessage() << " : inactive" << eol;
			notify();
		}

	}
}

void Alarm::delay() {

	out << "Alarm : " << getMessage() << " : testing delay " << _activationDelay << "s" << eol;

	// first starting the delay
	bool notified = _delayCondition->wait(_activationDelay * 1000);

	std::scoped_lock lock(_mutex);
	if (notified) {
		out << "Alarm : " << getMessage() << " : not confirmed" << eol;
		setState(CONDITION_INACTIVE);

	} else {
		setState(CONDITION_ACTIVE);

		out << "Alarm : " << getMessage() << " : confirmed, active" << eol;

		// stop command zone if required
		if (_stopCommandZone) {
			out << "Alarm : " << getMessage() << " : stops command zone" << eol;
			CommandZone::getInstance()->stop("");
			CommandLineManager::getInstance()->stopAtEndCommands();
		}

		notify();
	}

	delete _delayCondition;
	_delayCondition = 0;
}

std::string Alarm::getNotificationSubject() {

	if (getState() == CONDITION_ACTIVE) {
		return string("Alarm ") + getName();
	} else {
		return string("End of alarm ") + getName();
	}
}

std::string Alarm::getStringType() {
	return "Alarm";
}

std::string Alarm::toString(const std::string& space) {

	string result = space + "<alarm ";

	result += "name=\"" + getName() + "\" on=\"" + (isOn() ? "true" : "false") + "\" activation_delay=\"" + to<string>(_activationDelay) + "\" stop_command_zone=\"" + (isStopCommandZone() ? "true" : "false") + "\">\n";
	result += space + "  <expression value=\"" + getExpression() + "\"/>\n";
	result += space + "  <message value=\"" + getMessage() + "\"/>\n";
	result += space + "  <mail recipient=\"" + getMailRecipients() + "\"/>\n";
	result += space + "  <sms number=\"" + getSmsNumber() + "\"/>\n";

	result += space + "</alarm>";

	return result;
}

std::string Alarm::toXMLString(const std::string& space) {

	string result = space + "<alarm ";

	result += "id=\"" + to<string>(getConditionId()) + "\" ";
	result += "name=\"" + getName() + "\" on=\"" + (isOn() ? "true" : "false") + "\" activation_delay=\"" + to<string>(_activationDelay) + "\" stop_command_zone=\"" + (isStopCommandZone() ? "true" : "false") + "\">\n";
	result += space + "  <expression value=\"" + convertStringToXML(getExpression()) + "\"/>\n";
	result += space + "  <message value=\"" + convertStringToXML(getMessage()) + "\"/>\n";
	result += space + "  <mail recipient=\"" + getMailRecipients() + "\"/>\n";
	result += space + "  <sms number=\"" + getSmsNumber() + "\"/>\n";

	result += space + "</alarm>";

	return result;
}

ConditionElement * Alarm::toConditionElement() {

	ConditionElement * element = new ConditionElement(common::CONDITION_ALARM, getName(), getActivationDelay(), isOn(), isStopCommandZone());

	element->setExpression(getExpression());
	element->setMessage(getMessage());
	element->setMailRecipient(getMailRecipients());
	element->setSmsNumber(getSmsNumber());

	return element;
}

}
