/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandZoneTermination.h"
#include <common/output/OutputStream.h>
#include <core/transportlayer/MultiThreadCommunication.h>
#include <core/message/commandzone/CommandZoneExecutionEvent.h>
#include <core/transportlayer/core/commandzone/CommandZoneEventPublisher.h>

using namespace common;
using namespace std;

namespace core {

const std::string CommandZoneTermination::NAME = "End of Launch Pad";

CommandZoneTermination::CommandZoneTermination() :
	Warning(common::CONDITION_COMMAND_ZONE_TERMINATION, NAME), _ID(0) {

	_ID = MultiThreadCommunication::getInstance()->getCommandZoneEventPublisher()->subscribeDirect(this);
}

CommandZoneTermination::~CommandZoneTermination() {
	MultiThreadCommunication::getInstance()->getCommandZoneEventPublisher()->unsubscribeDirect(_ID);
}

bool CommandZoneTermination::isRemovable() const {
	return false;
}

void CommandZoneTermination::setOn(bool value) {

	std::scoped_lock lock(_mutex);

	Warning::setOn(value);
}

void CommandZoneTermination::check() {
}

std::string CommandZoneTermination::getNotificationSubject() {
	return NAME;
}

unsigned long CommandZoneTermination::getID() {
	return _ID;
}

void CommandZoneTermination::notify(CommandZoneEvent* commandZoneEvent) {
	CommandZoneEvent::EventType eventType = commandZoneEvent->getEventType();

	if (eventType == CommandZoneEvent::COMMAND_ZONE_EXECUTION) {
		notify(static_cast<CommandZoneExecutionEvent*>(commandZoneEvent));

	}
}

void CommandZoneTermination::notify(CommandZoneExecutionEvent* commandZoneEvent) {

	if (commandZoneEvent->getExecutionState() == CommandZoneExecutionEvent::FINISHED && isOn()) {
		Condition::notify();
	}
}

std::string CommandZoneTermination::toString(const std::string& space) {

	string result = space + "<command_zone_termination ";

	result += "name=\"" + getName() + "\" ";
	result += string("on=\"") + (isOn() ? "true" : "false") + "\">\n";
	result += space + "  <message value=\"" + getMessage() + "\"/>\n";
	result += space + "  <mail recipient=\"" + getMailRecipients() + "\"/>\n";
	result += space + "  <sms number=\"" + getSmsNumber() + "\"/>\n";

	result += space + "</command_zone_termination>";

	return result;
}

std::string CommandZoneTermination::toXMLString(const std::string& space) {

	string result = space + "<command_zone_termination ";

	result += "id=\"" + to<string>(getConditionId()) + "\" ";
	result += "name=\"" + getName() + "\" ";
	result += string("on=\"") + (isOn() ? "true" : "false") + "\">\n";
	result += space + "  <message value=\"" + convertStringToXML(getMessage()) + "\"/>\n";
	result += space + "  <mail recipient=\"" + getMailRecipients() + "\"/>\n";
	result += space + "  <sms number=\"" + getSmsNumber() + "\"/>\n";

	result += space + "</command_zone_termination>";

	return result;
}

ConditionElement * CommandZoneTermination::toConditionElement() {

	ConditionElement * element = new ConditionElement(common::CONDITION_COMMAND_ZONE_TERMINATION, getName(), 0, isOn(), false);

	element->setMailRecipient(getMailRecipients());
	element->setSmsNumber(getSmsNumber());

	return element;
}

}
