/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Condition.h"
#include "ConditionManager.h"
#include <common/base/ServerProperties.h>
#include <core/base/BooleanExpression.h>
#include <core/base/MathExpressionRenamer.h>
#include <core/base/BaseValueWrapper.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/dataaccess/DataAccess.h>
#include <core/transportlayer/MultiThreadCommunication.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <iostream>
#include <algorithm>
#include <boost/tokenizer.hpp>
#include <boost/algorithm/string.hpp>
#include "../transportlayer/dataprovider/StateAccessor.h"

using namespace std;

namespace core {

const std::string Condition::SMS_MAIL_RECIPIENT = "notill@ill.fr";

unsigned long Condition::_index = 0;


Condition::Condition(common::ConditionType type, const std::string& name) :
	_type(type),
	_name(name),
	_on(false),
	_state(common::CONDITION_INACTIVE),
	_booleanExpression(0),
	_actualValue(false),
	_oldValue(false),
	_first(true) {

	_conditionId = _index;
	_index = _index + 1;
	_stateAccessor = MultiThreadCommunication::getInstance()->getStateAccessor();
}

Condition::~Condition() {
	if (_booleanExpression != 0) {
		// removing the subscribers
		set<unsigned long> propertyIDs = _booleanExpression->getPropertyIDs();
		for (set<unsigned long>::const_iterator p = propertyIDs.begin(); p != propertyIDs.end(); p++) {
			DataNotificationRouter::getInstance()->removePropertyChangeSubscriber(*p, this);
		}

		delete _booleanExpression;
	}
}

common::ConditionType Condition::getType() const {
	return _type;
}

void Condition::verify(const std::string& expression) {

	BooleanExpression booleanExpression(expression);
	booleanExpression.init();
}

void Condition::init() {

	// Protect from concurrent access but release the lock after the block.
	{
		std::scoped_lock lock(_mutex);

		_booleanExpression->init();

		// adding the subscribers
		_propertyIDs = _booleanExpression->getPropertyIDs();

		// Reset the servant IDs before re-inserting the updated ones.
		_servantIDs.clear();
	}

	// The lock is released.
	for (set<unsigned long>::const_iterator p = _propertyIDs.begin(); p != _propertyIDs.end(); p++) {

		// Add the subscriber without locking because Condition::propertyChanged can be called in parallel and block.
		DataNotificationRouter::getInstance()->addPropertyChangeSubscriber(*p, this);

		// Protect _servantIDs.
		std::scoped_lock lock(_mutex);

		// get the servant ID
		_servantIDs.insert(DataAccess::getInstance()->getServantIDFromPropertyID(0, *p));
	}
}

void Condition::reset() {

	try {
		if (_on && _booleanExpression != 0) {
			_oldValue = _actualValue;
			_actualValue = _booleanExpression->evaluateBoolean();
		}
	}
	catch (const MathExpression::ParserError & e) {
		_on = false;
		LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, getStringType() + " " + _name + " cannot be set ON : " + e.getMessage()));
	}
	catch (const MathExpression::DisabledControllerError & e) {
		_on = false;
		LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, getStringType() + " " + _name + " cannot be set ON : " + e.getMessage()));
	}
	catch (const MathExpression::NoSuchEnumeratedValueError & e) {
		_on = false;
		LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, getStringType() + " " + _name + " cannot be set ON : " + e.getMessage()));
	}
	catch (const MathExpression::UndefinedVariableError & e) {
		_on = false;
		LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, getStringType() + " " + _name + " cannot be set ON : " + e.getMessage()));
	}
	catch (...) {
		_on = false;
		LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, getStringType() + " " + _name + " cannot be set ON"));
	}
}

void Condition::resetAndCheck() {

	// only reset if on
	if (_on) {
		reset();
	}

	// on may have changed in reset if a servant is disabled
	if (_on) {
		check();
		_first = false;
	}

	// notification
	_stateAccessor->notifyConditionState(_conditionId, _on, _state);

	LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, getStringType() + " " + _name + " is " + (_on ? "ON" : "OFF")));
}

unsigned long Condition::getConditionId() const {
	return _conditionId;
}

void Condition::setName(const std::string& name) {

	std::scoped_lock lock(_mutex);

	_name = name;
}


const std::string& Condition::getName() const {
	return _name;
}

void Condition::setExpression(const std::string& expression) {

	std::scoped_lock lock(_mutex);

	delete _booleanExpression;
	_booleanExpression = new BooleanExpression(expression);

	resetAndCheck();
}

std::string Condition::getExpression() {

	std::scoped_lock lock(_mutex);

	return _booleanExpression->getExpression();
}

void Condition::setMessage(const std::string& message) {

	std::scoped_lock lock(_mutex);

	_message = message;
}

const std::string& Condition::getMessage() {

	std::scoped_lock lock(_mutex);

	return _message;
}

void Condition::setMailRecipients(const std::string& recipients) {

	std::scoped_lock lock(_mutex);

	_mailRecipients = recipients;
}

const std::string& Condition::getMailRecipients() {

	std::scoped_lock lock(_mutex);

	return _mailRecipients;
}

void Condition::setSmsNumber(const std::string& number) {

	std::scoped_lock lock(_mutex);

	_smsNumber = number;
}

const std::string& Condition::getSmsNumber() {

	std::scoped_lock lock(_mutex);

	return _smsNumber;
}

std::set<unsigned long> Condition::getPropertyIDs() {

	std::scoped_lock lock(_mutex);

	return _propertyIDs;
}

std::set<unsigned long> Condition::getServantIDs() {

	std::scoped_lock lock(_mutex);

	return _servantIDs;
}

void Condition::setOn(bool value) {

	std::scoped_lock lock(_mutex);

	// resetting if the condition changes from OFF to ON
	if (!_on && value) {
		_first = true;
	}

	_on = value;


	resetAndCheck();
}

bool Condition::isOn() const {
	return _on;
}

void Condition::setState(common::ConditionState state) {

	common::ConditionState oldState = _state;
	_state = state;

	if (state == common::CONDITION_ACTIVE) {
		LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, getStringType() + " " + _name + " is active"));
	} else if (state == common::CONDITION_INACTIVE) {
		LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, getStringType() + " " + _name + " is inactive"));
	} else if (state == common::CONDITION_ON_ACTIVATION_DELAY) {
		// no log
	}

	if (oldState != _state) {
		_stateAccessor->notifyConditionState(_conditionId, _on, _state);
	}

}

bool Condition::actualValue() const {
	return _actualValue;
}

bool Condition::oldValue() const {
	return _oldValue;
}

bool Condition::isFirst() const {
	return _first;
}

void Condition::notify() {

	// mail
	if (_mailRecipients != "") {
		string mailCommand("echo \"");
		mailCommand += _message + "\" | mail -r nomadmail@ill.fr -s \"";
		mailCommand += getNotificationSubject() + " on " + common::ServerProperties::getInstance()->getHostname() + "\" ";

		string recipients = _mailRecipients;
		replace(recipients.begin(), recipients.end(), ',', ' ');

		mailCommand += recipients;

		cout << "Mail command = " << mailCommand << endl;

		ConditionManager::getInstance()->mailCommand(mailCommand);
	}

	// sms
	if (_smsNumber != "") {

		// get the list of numbers
		vector<string> numbers;

		typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
		boost::char_separator<char> sep(", ");
		tokenizer tokens(_smsNumber, sep);
		for (tokenizer::iterator t = tokens.begin(); t != tokens.end(); ++t) {
			numbers.push_back(*t);
		}

		// iterate the numbers
		for (vector<string>::const_iterator n = numbers.begin(); n != numbers.end(); ++n) {

			string mailCommand("echo \"");
			mailCommand += getNotificationSubject() + " on " + common::ServerProperties::getInstance()->getHostname() + ":\n" + _message + "\" | mail -r nomadmail@ill.fr -s \"";
			mailCommand += *n + "\"";

			// temp
			mailCommand += " " + SMS_MAIL_RECIPIENT;

			cout << "Mail command = " << mailCommand << endl;

			ConditionManager::getInstance()->mailCommand(mailCommand);
		}
	}
}

bool Condition::isRemovable() const {
	return true;
}

common::ConditionState Condition::getState() {

	std::scoped_lock lock(_mutex);

	return _state;
}

void Condition::propertyChanged(unsigned long databaseID, unsigned long propertyID) {

	std::scoped_lock lock(_mutex);

	//cout << "Alarm propertyChanged " << propertyID << endl;
	if (_on) {
		_oldValue = _actualValue;
		_actualValue = _booleanExpression->evaluateBoolean();

		check();
		_first = false;
	}
}

void Condition::replacePropertyNamesWithAlias() {

	std::scoped_lock lock(_mutex);

	cout << "Processing " << _name;

	if (_booleanExpression != 0) {
		string expression = _booleanExpression->getExpression();
		cout << " : " << expression << endl;

		MathExpressionRenamer renamer(expression);
		string newExpression = renamer.execute();

		_booleanExpression->setExpression(newExpression);
	}
	cout << endl << endl;
}

}
