/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_CONDITION_H_
#define CORE_CONDITION_H_

#include <common/base/CommonTypes.h>
#include <core/base/BooleanExpression.h>
#include <core/dataaccess/PropertyChangeSubscriber.h>
#include <core/parser/conditions/ConditionElement.h>
#include <mutex>
#include <string>

namespace core {

class ConditionElement;
class StateAccessor;

/**
 * Base class defining a Condition for the ConditionManager.
 */
class Condition : public PropertyChangeSubscriber {

public:
	/**
	 * Constructor.
	 */
	Condition(common::ConditionType type, const std::string& name);

	/**
	 * Destructor.
	 */
	virtual ~Condition();

	/**
	 * Gets the condition ID.
	 */
	unsigned long getConditionId() const;


	/**
	 * Gets the alarm type.
	 */
	common::ConditionType getType() const;

	/**
	 * Verifies the expression.
	 */
	static void verify(const std::string& expression);

	/**
	 * Initialises the condition.
	 */
	void init();

	/**
	 * Resets the condition.
	 */
	void reset();

	/**
	 * Sets the name.
	 */
	void setName(const std::string& name);

	/**
	 * Gets the name.
	 */
	const std::string& getName() const;

	/**
	 * Sets the expression.
	 */
	void setExpression(const std::string& expression);

	/**
	 * Gets the expression.
	 */
	std::string getExpression();

	/**
	 * Sets the message.
	 */
	void setMessage(const std::string& message);

	/**
	 * Gets the message.
	 */
	const std::string& getMessage();

	/**
	 * Sets the mail recipients.
	 */
	void setMailRecipients(const std::string& recipients);

	/**
	 * Gets the mail recipients.
	 */
	const std::string& getMailRecipients();

	/**
	 * Sets the sms number.
	 */
	void setSmsNumber(const std::string& number);

	/**
	 * Gets the sms number.
	 */
	const std::string& getSmsNumber();

	/**
	 * Returns the list of property IDs. Init must have been called before.
	 */
	std::set<unsigned long> getPropertyIDs();

	/**
	 * Returns the list of servant IDs. Init must have been called before.
	 */
	std::set<unsigned long> getServantIDs();

	/**
	 * Sets on or off. To be redefined in subclasses.
	 */
	virtual void setOn(bool value) = 0;

	/**
	 * Returns true if is on.
	 */
	bool isOn() const;

	/**
	 * Returns true if is removable.
	 */
	virtual bool isRemovable() const;

	/**
	 * Gets the state.
	 */
	common::ConditionState getState();

	/**
	 * Notifies that the property has changed.
	 */
	virtual void propertyChanged(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Checks the alarm.
	 */
	virtual void check() = 0;

	/**
	 * Gets the notification subject.
	 */
	virtual std::string getNotificationSubject() = 0;

	/**
	 * Gets the string type.
	 */
	virtual std::string getStringType() = 0;

	/**
	 * Returns the stringified content of the condition.
	 */
	virtual std::string toString(const std::string& space) = 0;

	/**
	 * Returns the xml stringified content of the condition.
	 */
	virtual std::string toXMLString(const std::string& space) = 0;

	/**
	 * Converts the condition to an element.
	 */
	virtual ConditionElement * toConditionElement() = 0;

	/**
	 * Notifies by sending emails and sms.
	 */
	void notify();

	/**
	 * Replaces the property names with their alias.
	 */
	void replacePropertyNamesWithAlias();

protected:
	void setState(common::ConditionState state);
	bool actualValue() const;
	bool oldValue() const;
	bool isFirst() const;
	void resetAndCheck();

	std::recursive_mutex _mutex;

private:
	static unsigned long _index;
	common::ConditionType _type;
	unsigned long _conditionId;
	std::string _name;
	std::string _message;

	std::set<unsigned long> _propertyIDs;
	std::set<unsigned long> _servantIDs;

	bool _on;
	common::ConditionState _state;
	std::string _mailRecipients;
	std::string _smsNumber;
	BooleanExpression* _booleanExpression;
	bool _actualValue;
	bool _oldValue;
	bool _first;

	StateAccessor* _stateAccessor;

	static const std::string SMS_MAIL_RECIPIENT;
};

}

#endif /* CORE_CONDITION_H_ */
