/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ConditionManager.h"
#include "Alarm.h"
#include "Warning.h"
#include "Watchdog.h"
#include "CommandZoneTermination.h"
#include <Utilities/Simulated.h>
#include <core/dataaccess/DataAccess.h>
#include <core/parser/conditions/ConditionsParser.h>
#include <core/parser/conditions/ConditionElement.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <boost/filesystem.hpp>
#include <iostream>
#include <fstream>
#include <algorithm>

using namespace std;
using namespace common;
using namespace boost::filesystem;

namespace core {

ConditionManager* ConditionManager::_instance = 0;
const std::string ConditionManager::CONDITION_PATH = "/Conditions.xml";
const int ConditionManager::WRITE_PERIOD_S = 60;

ConditionManager::ConditionExistsException::ConditionExistsException() :
	Exception("core::ConditionManager::NameAlreadyExistsException", "name already exists") {
};

ConditionManager::NoSuchConditionException::NoSuchConditionException() :
	Exception("core::ConditionManager::NoSuchConditionException", "name does not exist") {
};

ConditionManager::UnremovableConditionException::UnremovableConditionException() :
	Exception("core::ConditionManager::UnremovableConditionException", "condition is not removable") {
};

ConditionManager::UnrenamableConditionException::UnrenamableConditionException() :
	Exception("core::ConditionManager::UnrenamableConditionException", "condition is not renamable") {
};

ConditionManager::InvalidExpressionException::InvalidExpressionException(const std::string& message) :
	Exception("core::ConditionManager::InvalidExpressionException", message) {
};

ConditionManager::ConditionManager() :
	_writeThread(nullptr),
	_running(false),
	_notificationsEnabled(true) {

	DataNotificationRouter::getInstance()->addConfigurationChangeSubscriber(this);
}

ConditionManager::~ConditionManager() {

	DataNotificationRouter::getInstance()->removeConfigurationChangeSubscriber(this);

	if (_writeThread != nullptr) {
		_writeThread->join();
		delete _writeThread;
	}

	for (vector<Condition*>::iterator c = _conditions.begin(); c != _conditions.end(); c++) {
		delete *c;
	}
}

ConditionManager* ConditionManager::getInstance() {
	return _instance;
}

void ConditionManager::initInstance(const std::string& configPath) {

	if (_instance == 0) {
		_instance = new ConditionManager();
		_instance->init(configPath);
	}
}

void ConditionManager::resetInstance() {
	delete _instance;
	_instance = 0;
}

std::vector<Condition*>::iterator ConditionManager::findCondition(const std::string& conditionName) {
	std::vector<Condition*>::iterator c;

	for (c = _conditions.begin(); c != _conditions.end(); c++) {
		if ((*c)->getName() == conditionName) {
			return c;
		}
	}

	return c;
}

std::vector<Condition*>::iterator ConditionManager::findConditionWithId(unsigned long conditionId) {
	std::vector<Condition*>::iterator c;

	for (c = _conditions.begin(); c != _conditions.end(); c++) {
		if ((*c)->getConditionId() == conditionId) {
			return c;
		}
	}

	return c;
}

int ConditionManager::findConditionId(const std::string& conditionName) {
	std::vector<Condition*>::iterator c;

	for (c = _conditions.begin(); c != _conditions.end(); c++) {
		if ((*c)->getName() == conditionName) {
			return (*c)->getConditionId();
		}
	}

	throw NoSuchConditionException();
}

bool ConditionManager::conditionExists(const std::string& conditionName) const {

	for (vector<Condition*>::const_iterator c = _conditions.begin(); c != _conditions.end(); c++) {
		if ((*c)->getName() == conditionName) {
			return true;
		}
	}
	return false;
}


bool ConditionManager::conditionWithIdExists(unsigned long conditionId) const {
	for (vector<Condition*>::const_iterator c = _conditions.begin(); c != _conditions.end(); c++) {
		if ((*c)->getConditionId() == conditionId) {
			return true;
		}
	}
	return false;
}


void ConditionManager::init(const std::string& configPath) {

	_conditionManagerConfigFileName = configPath + CONDITION_PATH;

	// test environment variable
	string notificationsString = ServerProperties::getInstance()->getNomadNotifications();
	if ((notificationsString == "0") || (notificationsString == "-1")) {
		_notificationsEnabled = false;
	}
}

void ConditionManager::run() {
	readConditionManagerFile();

	_running = true;

	if ((Simulated::activated == false) && (Simulated::for3d == false)) {
		_writeThread = new std::thread(&loopWrite);
	}
}

void ConditionManager::stop() {

	std::unique_lock<std::mutex> lock(_mutex);
	_writeCondition.notify_all();
	_running = false;
}

void ConditionManager::verifyExpression(const std::string& expression) {

	try {
		Condition::verify(expression);

	} catch (const Exception& e) {
		throw InvalidExpressionException(e.getMessage());
	}
}

void ConditionManager::addAlarm(const std::string& name, const std::string& expression, const std::string& message, int activationDelay, bool stopCommandZone, bool on, const std::string& mailRecipients, const std::string& smsNumber) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (conditionExists(name)) {
		throw ConditionExistsException();
	}

	Alarm* condition = new Alarm(name);
	condition->setExpression(expression);
	condition->setMessage(message);
	condition->setMailRecipients(mailRecipients);
	condition->setSmsNumber(smsNumber);
	condition->setActivationDelay(activationDelay);
	condition->setStopCommandZone(stopCommandZone);

	try {
		condition->init();

	} catch (const Exception& e) {
		throw InvalidExpressionException(e.getMessage());
	}

	condition->setOn(on);
	_conditions.push_back(condition);

	cout << "Added alarm " << name << " - " << expression << " : " << message << endl;
}

void ConditionManager::addWarning(const std::string& name, const std::string& expression, const std::string& message, bool on, const std::string& mailRecipients, const std::string& smsNumber) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (conditionExists(name)) {
		throw ConditionExistsException();
	}

	Warning* condition = new Warning(name);
	condition->setExpression(expression);
	condition->setMessage(message);
	condition->setMailRecipients(mailRecipients);
	condition->setSmsNumber(smsNumber);

	try {
		condition->init();

	} catch (const Exception& e) {
		throw InvalidExpressionException(e.getMessage());
	}

	condition->setOn(on);
	_conditions.push_back(condition);

	cout << "Added warning " << name << " - " << expression << " : " << message << endl;
}

void ConditionManager::addWatchdog(const std::string& name, const std::string& expression, bool on, const std::string& actions, const std::string& mailRecipients, const std::string& smsNumber) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (conditionExists(name)) {
		throw ConditionExistsException();
	}

	Watchdog* condition = new Watchdog(name);
	condition->setExpression(expression);
	condition->setActions(actions);
	condition->setMailRecipients(mailRecipients);
	condition->setSmsNumber(smsNumber);

	try {
		condition->init();

	} catch (const Exception& e) {
		throw InvalidExpressionException(e.getMessage());
	}

	condition->setOn(on);
	_conditions.push_back(condition);

	cout << "Added watchdog " << name << " - " << expression << " : " << actions << endl;
}

void ConditionManager::addCommandZoneTermination(const std::string& message, bool on, const std::string& mailRecipients, const std::string& smsNumber) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (conditionExists(CommandZoneTermination::NAME)) {
		throw ConditionExistsException();
	}

	CommandZoneTermination* condition = new CommandZoneTermination();
	condition->setMessage(message);
	condition->setMailRecipients(mailRecipients);
	condition->setSmsNumber(smsNumber);

	condition->setOn(on);
	_conditions.push_back(condition);

	cout << "Added command zone termination" << endl;
}

void ConditionManager::removeCondition(unsigned long conditionId) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (!conditionWithIdExists(conditionId)) {
		throw NoSuchConditionException();
	}

	vector<Condition*>::iterator c = findConditionWithId(conditionId);

	if (c == _conditions.end()) {
		throw NoSuchConditionException();
	}

	Condition* condition = *c;

	if (!condition->isRemovable()) {
		throw UnremovableConditionException();
	}

	// we can delete the condition
	delete condition;
	_conditions.erase(c);
}

void ConditionManager::setConditionOn(unsigned long conditionId, bool value) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (!conditionWithIdExists(conditionId)) {
		throw NoSuchConditionException();
	}

	(*findConditionWithId(conditionId))->setOn(value);
}

void ConditionManager::renameCondition(unsigned long conditionId, const std::string& newName) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (!conditionWithIdExists(conditionId)) {
		throw NoSuchConditionException();
	}
	if (conditionExists(newName)) {
		throw UnrenamableConditionException();
	}

	vector<Condition*>::iterator condition = findConditionWithId(conditionId);
	(*condition)->setName(newName);
}

void ConditionManager::editAlarm(unsigned long conditionId, const std::string& expression, const std::string& message, int activationDelay, bool stopCommandZone, const std::string& mailRecipients, const std::string& smsNumber) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (!conditionWithIdExists(conditionId)) {
		throw NoSuchConditionException();
	}

	vector<Condition*>::iterator c = findConditionWithId(conditionId);

	if (c == _conditions.end()) {
		throw NoSuchConditionException();
	}

	Alarm* condition = dynamic_cast<Alarm*>(*c);

	condition->setMessage(message);
	condition->setMailRecipients(mailRecipients);
	condition->setSmsNumber(smsNumber);
	condition->setActivationDelay(activationDelay);
	condition->setStopCommandZone(stopCommandZone);

	try {
		condition->setExpression(expression);
		condition->init();

	} catch (const Exception& e) {
		throw InvalidExpressionException(e.getMessage());
	}
}

void ConditionManager::editWarning(unsigned long conditionId, const std::string& expression, const std::string& message, const std::string& mailRecipients, const std::string& smsNumber) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (!conditionWithIdExists(conditionId)) {
		throw NoSuchConditionException();
	}

	vector<Condition*>::iterator c = findConditionWithId(conditionId);

	if (c == _conditions.end()) {
		throw NoSuchConditionException();
	}

	Warning* condition = dynamic_cast<Warning*>(*c);

	condition->setMessage(message);
	condition->setMailRecipients(mailRecipients);
	condition->setSmsNumber(smsNumber);

	try {
		condition->setExpression(expression);
		condition->init();

	} catch (const Exception& e) {
		throw InvalidExpressionException(e.getMessage());
	}
}

void ConditionManager::editWatchdog(unsigned long conditionId, const std::string& expression, const std::string& actions, const std::string& mailRecipients, const std::string& smsNumber) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (!conditionWithIdExists(conditionId)) {
		throw NoSuchConditionException();
	}

	vector<Condition*>::iterator c = findConditionWithId(conditionId);

	if (c == _conditions.end()) {
		throw NoSuchConditionException();
	}

	Watchdog* condition = dynamic_cast<Watchdog*>(*c);

	condition->setActions(actions);
	condition->setMailRecipients(mailRecipients);
	condition->setSmsNumber(smsNumber);

	try {
		condition->setExpression(expression);
		condition->init();

	} catch (const Exception& e) {
		throw InvalidExpressionException(e.getMessage());
	}
}

void ConditionManager::editCommandZoneTermination(const std::string& message, const std::string& mailRecipients, const std::string& smsNumber) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (!conditionExists(CommandZoneTermination::NAME)) {
		throw NoSuchConditionException();
	}

	vector<Condition*>::iterator c = findCondition(CommandZoneTermination::NAME);

	if (c == _conditions.end()) {
		throw NoSuchConditionException();
	}

	CommandZoneTermination* condition = dynamic_cast<CommandZoneTermination*>(*c);

	condition->setMessage(message);
	condition->setMailRecipients(mailRecipients);
	condition->setSmsNumber(smsNumber);
}

common::ConditionState ConditionManager::getConditionState(unsigned long conditionId) {

	std::unique_lock<std::mutex> lock(_mutex);

	if (!conditionWithIdExists(conditionId)) {
		throw NoSuchConditionException();
	}

	vector<Condition*>::iterator c = findConditionWithId(conditionId);

	if (c == _conditions.end()) {
		throw NoSuchConditionException();
	}

	Condition* condition = *c;

	return condition->getState();
}

void ConditionManager::configurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState enable) {

	std::unique_lock<std::mutex> lock(_mutex);

	// iterating the conditions
	for (vector<Condition*>::iterator c = _conditions.begin(); c != _conditions.end(); c++) {
		Condition * condition = *c;

		// Copy the servantIDs set because it is returned by value.
		set<unsigned long> servantIDs = condition->getServantIDs();

		// test if the condition depends on the servant
		if (servantIDs.find(servantID) != servantIDs.end()) {
			// sets the condition off if it is on
			// the condition is not set on automatically
			if (enable == common::DISABLED) {
				if (condition->isOn()) {
					condition->setOn(false);
				}
			}
		}
	}
}

void ConditionManager::readConditionManagerFile() {

	path filePath(_conditionManagerConfigFileName);

	if (exists(filePath)) {

		ConditionsParser* parser = new ConditionsParser();
		parser->init();
		parser->parseXMLFile(_conditionManagerConfigFileName);

		const vector<ConditionElement*>& elements = parser->getConditionElements();
	
		for (vector<ConditionElement*>::const_iterator e = elements.begin(); e != elements.end(); e++) {

			try {

				if ((*e)->getType() == common::CONDITION_ALARM) {
					addAlarm((*e)->getName(), (*e)->getExpression(), (*e)->getMessage(), (*e)->getActivationDelay(), (*e)->isStopCommandZone(), (*e)->isOn(), (*e)->getMailRecipient(), (*e)->getSmsNumber());

				} else if ((*e)->getType() == common::CONDITION_WARNING) {
					addWarning((*e)->getName(), (*e)->getExpression(), (*e)->getMessage(), (*e)->isOn(), (*e)->getMailRecipient(), (*e)->getSmsNumber());

				} else if ((*e)->getType() == common::CONDITION_WATCHDOG) {
					addWatchdog((*e)->getName(), (*e)->getExpression(), (*e)->isOn(), (*e)->getActions(), (*e)->getMailRecipient(), (*e)->getSmsNumber());

				} else if ((*e)->getType() == common::CONDITION_COMMAND_ZONE_TERMINATION) {
					addCommandZoneTermination((*e)->getMessage(), (*e)->isOn(), (*e)->getMailRecipient(), (*e)->getSmsNumber());

				}

			} catch (common::Exception const & ex) {
				string message = "Condition '";
				message += (*e)->getName() + "' cannot be defined : " + ex.getMessage();
				LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, message));
			}
		}

		delete parser;

	} else {
		err << "Cannot find Conditions.xml file" << eol;
	}
}

std::string ConditionManager::toXMLString() {

	string result = "<conditions>\n";

	for (vector<Condition*>::const_iterator c = _conditions.begin(); c != _conditions.end(); c++) {
		result += (*c)->toXMLString("  ") + "\n";
	}

	result += "</conditions>";

	return result;
}

std::string ConditionManager::toString() {

	string result = "<conditions>\n";

	for (vector<Condition*>::const_iterator c = _conditions.begin(); c != _conditions.end(); c++) {
		result += (*c)->toString("  ") + "\n";
	}

	result += "</conditions>";

	return result;
}

void ConditionManager::writeConditionManagerFile() {

	// write the file
	std::ofstream conditionFile(_conditionManagerConfigFileName.c_str(), ios::out);
	conditionFile << toXMLString() << endl;
	conditionFile.close();

	// copy a backup file
	path filePath(_conditionManagerConfigFileName);
	path backupFilePath(_conditionManagerConfigFileName + ".bak");

	try {
		remove(backupFilePath);
		copy_file(filePath, backupFilePath);
	} catch (const filesystem_error & e) {
		out << "Cannot backup Conditions.xml file" << eol;
	}
}

void ConditionManager::loopWrite() {

	while (true) {

		std::unique_lock<std::mutex> lock(_instance->_mutex);

		if (!_instance->_running) {
			break;
		}

		// waiting
		_instance->_writeCondition.wait_for(lock, std::chrono::seconds(WRITE_PERIOD_S));
		_instance->writeConditionManagerFile();
	}

	_instance->writeConditionManagerFile();
}

void ConditionManager::mailCommand(const std::string& command) {

	if (_notificationsEnabled) {
		::system(command.c_str());
	}
}

void ConditionManager::replacePropertyNamesWithAlias() {

	std::unique_lock<std::mutex> lock(_instance->_mutex);

	for (vector<Condition*>::const_iterator c = _conditions.begin(); c != _conditions.end(); c++) {
		(*c)->replacePropertyNamesWithAlias();
	}

	_instance->writeConditionManagerFile();
}

}
