/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_CONDITIONMANAGER_H_
#define CORE_CONDITIONMANAGER_H_

#include <common/base/CommonTypes.h>
#include <common/base/Exception.h>
#include <core/dataaccess/ConfigurationChangeSubscriber.h>
#include <thread>
#include <string>
#include <vector>
#include <string>
#include <condition_variable>

namespace core {

class Condition;
class ConditionElement;

/**
 * Class defining an alarm manager.
 */
class ConditionManager : public ConfigurationChangeSubscriber {

public:
	/**
	 * Exception when creating a condition with existent name.
	 */
	class ConditionExistsException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		ConditionExistsException();
	};

	/**
	 * Exception when removing a condition with non-existent name.
	 */
	class NoSuchConditionException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchConditionException();
	};

	/**
	 * Exception when removing an unremovable condition.
	 */
	class UnremovableConditionException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		UnremovableConditionException();
	};

	/**
	 * Exception when renaming an unrenamable condition.
	 */
	class UnrenamableConditionException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		UnrenamableConditionException();
	};

	/**
	 * Invalid expression exception.
	 */
	class InvalidExpressionException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		InvalidExpressionException(const std::string& message);
	};

	/**
	 * Returns the instance.
	 */	
	static ConditionManager* getInstance();
	
	/**
	 * Initialises the instance.
	 */
	static void initInstance(const std::string& configPath);
	
	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Starts the survey.
	 */
	void run();
	
	/**
	 * Stops the survey.
	 */
	void stop();

	/**
	 * Verifies the expression.
	 * throws InvalidExpressionException.
	 */
	void verifyExpression(const std::string& expression);

	/**
	 * Adds an alarm to the manager.
	 * throws ConditionExistsException, InvalidExpressionException.
	 */
	void addAlarm(const std::string& name, const std::string& expression, const std::string& message, int activationDelay, bool stopCommandZone, bool on, const std::string& mailRecipients, const std::string& smsNumber);

	/**
	 * Adds a warning to the manager.
	 * throws ConditionExistsException, InvalidExpressionException.
	 */
	void addWarning(const std::string& name, const std::string& expression, const std::string& message, bool on, const std::string& mailRecipients, const std::string& smsNumber);

	/**
	 * Adds a watchdog to the manager.
	 * throws ConditionExistsException, InvalidExpressionException.
	 */
	void addWatchdog(const std::string& name, const std::string& expression, bool on, const std::string& actions, const std::string& mailRecipients, const std::string& smsNumber);

	/**
	 * Adds a command zone termination to the manager.
	 * throws ConditionExistsException.
	 */
	void addCommandZoneTermination(const std::string& message, bool on, const std::string& mailRecipients, const std::string& smsNumber);

	/**
	 * Removes a condition with id.
	 * throws NoSuchConditionException, UnremovableConditionException.
	 */
	void removeCondition(unsigned long conditionId);

	/**
	 * Activates the condition with id.
	 * throws NoSuchConditionException.
	 */
	void setConditionOn(unsigned long conditionId, bool value);

	/**
	 * Renames a condition with id.
	 * throws NoSuchConditionException, UnrenamableConditionException.
	 */
	void renameCondition(unsigned long conditionId, const std::string& newName);

	/**
	 * Edits an alarm.
	 * throws NoSuchConditionException, InvalidExpressionException.
	 */
	void editAlarm(unsigned long conditionId, const std::string& expression, const std::string& message, int activationDelay, bool stopCommandZone, const std::string& mailRecipients, const std::string& smsNumber);

	/**
	 * Edits a warning.
	 * throws NoSuchConditionException, InvalidExpressionException.
	 */
	void editWarning(unsigned long conditionId, const std::string& expression, const std::string& message, const std::string& mailRecipients, const std::string& smsNumber);

	/**
	 * Edits a watchdog.
	 * throws NoSuchConditionException, InvalidExpressionException.
	 */
	void editWatchdog(unsigned long conditionId, const std::string& expression, const std::string& actions, const std::string& mailRecipients, const std::string& smsNumber);

	/**
	 * Edits a command zone termination.
	 * throws NoSuchConditionException.
	 */
	void editCommandZoneTermination(const std::string& message, const std::string& mailRecipients, const std::string& smsNumber);

	/**
	 * Returns the condition state.
	 * throws NoSuchConditionException.
	 */
	common::ConditionState getConditionState(unsigned long conditionId);

	/**
	 * Returns the condition id if it exists.
	 * throws NoSuchConditionException.
	 */
	int findConditionId(const std::string& conditionName);

	/**
	 * Notifies that the configuration has changed.
	 */
	virtual void configurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState enable);

	/**
	 * Executes the mail command (not if NOMAD_NOTIFICATIONS equals 0).
	 */
	void mailCommand(const std::string& command);

	/**
	 * Replaces the property names with their alias.
	 */
	void replacePropertyNamesWithAlias();

	/**
	 * Gets the conditions with xml content.
	 */
	std::string toXMLString();

	/**
	 * Returns the stringified content of the condition manager.
	 */
	std::string toString();

private:
	ConditionManager();
	~ConditionManager();

	std::vector<Condition*>::iterator findCondition(const std::string& conditionName);
	std::vector<Condition*>::iterator findConditionWithId(unsigned long conditionId);
	bool conditionExists(const std::string& conditionName) const;
	bool conditionWithIdExists(unsigned long conditionId) const;

	void init(const std::string& configPath);
	void readConditionManagerFile();
	void writeConditionManagerFile();

	static void loopWrite();

	static ConditionManager* _instance;
	static const std::string CONDITION_PATH;
	
	std::string _conditionManagerConfigFileName;
	std::mutex _mutex;
	std::vector<Condition*> _conditions;

	std::condition_variable _writeCondition;
	std::thread * _writeThread;
	
	bool _running;
	bool _notificationsEnabled;

	static const int WRITE_PERIOD_S;
};

}

#endif /* CORE_CONDITIONMANAGER_H_ */
