/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Warning.h"
#include <common/output/OutputStream.h>

using namespace common;
using namespace std;

namespace core {

Warning::Warning(const std::string& name) :
	Condition(common::CONDITION_WARNING, name) {
}

Warning::Warning(common::ConditionType type, const std::string& name) :
	Condition(type, name) {
}

Warning::~Warning() {
}

void Warning::setOn(bool value) {

	std::scoped_lock lock(_mutex);

	Condition::setOn(value);

	if (!value) {
		setState(CONDITION_INACTIVE);
	}
}

void Warning::check() {

	std::scoped_lock lock(_mutex);

	// only check when value changes
	if (isFirst() || actualValue() != oldValue()) {

		// state becomes active
		if (actualValue()) {
			setState(CONDITION_ACTIVE);
			out << "Condition warning: " << getMessage() << eol;
			notify();
		} else {
			setState(CONDITION_INACTIVE);
		}
	}

}

std::string Warning::getNotificationSubject() {
	return string("Warning ") + getName();
}

std::string Warning::getStringType() {
	return "Warning";
}

std::string Warning::toString(const std::string& space) {

	string result = space + "<warning ";

	result += "name=\"" + getName() + "\" on=\"" + (isOn() ? "true" : "false") + "\">\n";
	result += space + "  <expression value=\"" + getExpression() + "\"/>\n";
	result += space + "  <message value=\"" + getMessage() + "\"/>\n";
	result += space + "  <mail recipient=\"" + getMailRecipients() + "\"/>\n";
	result += space + "  <sms number=\"" + getSmsNumber() + "\"/>\n";

	result += space + "</warning>";

	return result;
}

std::string Warning::toXMLString(const std::string& space) {

	string result = space + "<warning ";

	result += "id=\"" + to<string>(getConditionId()) + "\" ";
	result += "name=\"" + getName() + "\" on=\"" + (isOn() ? "true" : "false") + "\">\n";
	result += space + "  <expression value=\"" + convertStringToXML(getExpression()) + "\"/>\n";
	result += space + "  <message value=\"" + convertStringToXML(getMessage()) + "\"/>\n";
	result += space + "  <mail recipient=\"" + getMailRecipients() + "\"/>\n";
	result += space + "  <sms number=\"" + getSmsNumber() + "\"/>\n";

	result += space + "</warning>";

	return result;
}

ConditionElement * Warning::toConditionElement() {

	ConditionElement * element = new ConditionElement(common::CONDITION_WARNING, getName(), 0, isOn(), false);

	element->setExpression(getExpression());
	element->setMessage(getMessage());
	element->setMailRecipient(getMailRecipients());
	element->setSmsNumber(getSmsNumber());

	return element;
}


}
