/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Watchdog.h"
#include <common/output/OutputStream.h>
#include <ics/CommandSystem/CommandLine/CommandLineManager.h>
#include <boost/bind.hpp>

using namespace common;
using namespace std;

namespace core {

const std::string Watchdog::WATCHDOG_COMMANDLINE = "watchdog";

Watchdog::Watchdog(const std::string& name) :
	Condition(common::CONDITION_WATCHDOG, name),
	_actionsState(CONDITION_INACTIVE),
	_actionsThread(nullptr) {
}

Watchdog::~Watchdog() {
	if (_actionsThread != nullptr) {
		_actionsThread->join();
		delete _actionsThread;
	}
}

void Watchdog::setOn(bool value) {

	std::scoped_lock lock(_mutex);

	Condition::setOn(value);

	if (!value) {
		setState(CONDITION_INACTIVE);
	}
}

void Watchdog::setActions(const std::string& actions) {

	std::scoped_lock lock(_mutex);

	_actions = actions;
}

const std::string& Watchdog::getActions() {

	std::scoped_lock lock(_mutex);

	return _actions;
}

void Watchdog::setActionsState(common::ConditionState state) {

	std::scoped_lock lock(_mutex);

	_actionsState = state;
}

common::ConditionState Watchdog::getActionsState() {

	std::scoped_lock lock(_mutex);

	return _actionsState;
}

void Watchdog::check() {

	std::scoped_lock lock(_mutex);

	// only check when value changes
	if (isFirst() || actualValue() != oldValue()) {

		if (actualValue()) {

			setState(CONDITION_ACTIVE);

			if (getActionsState() == CONDITION_INACTIVE) {

				setActionsState(CONDITION_ACTIVE);
				out << "Watchdog : " << getName() << " : executes " << _actions << eol;
				if (_actionsThread != nullptr) {
					_actionsThread->join();
					delete _actionsThread;
				}
				_actionsThread = new std::thread(std::bind(&Watchdog::executeActions, this));

				notify();

			} else {
				// do nothing, the actions are already executing
			}

		} else {
			setState(CONDITION_INACTIVE);
		}
	}
}

void Watchdog::executeActions() {

	CommandLineManager::getInstance(WATCHDOG_COMMANDLINE)->executeCommand(_actions, true, true);
	setActionsState(CONDITION_INACTIVE);
}

std::string Watchdog::getNotificationSubject() {
	return string("Watchdog ") + getName();
}

std::string Watchdog::getStringType() {
	return "Watchdog";
}

std::string Watchdog::toString(const std::string& space) {

	string result = space + "<watchdog ";

	result += "name=\"" + getName() + "\" on=\"" + (isOn() ? "true" : "false") + "\">\n";
	result += space + "  <expression value=\"" + getExpression() + "\"/>\n";
	result += space + "  <actions content=\"" + _actions + "\"/>\n";
	result += space + "  <mail recipient=\"" + getMailRecipients() + "\"/>\n";
	result += space + "  <sms number=\"" + getSmsNumber() + "\"/>\n";

	result += space + "</watchdog>";

	return result;
}

std::string Watchdog::toXMLString(const std::string& space) {

	string result = space + "<watchdog ";

	result += "id=\"" + to<string>(getConditionId()) + "\" ";
	result += "name=\"" + getName() + "\" on=\"" + (isOn() ? "true" : "false") + "\">\n";
	result += space + "  <expression value=\"" + convertStringToXML(getExpression()) + "\"/>\n";
	result += space + "  <actions content=\"" + _actions + "\"/>\n";
	result += space + "  <mail recipient=\"" + getMailRecipients() + "\"/>\n";
	result += space + "  <sms number=\"" + getSmsNumber() + "\"/>\n";

	result += space + "</watchdog>";

	return result;
}

ConditionElement * Watchdog::toConditionElement() {

	ConditionElement * element = new ConditionElement(common::CONDITION_WATCHDOG, getName(), 0, isOn(), false);

	element->setExpression(getExpression());
	element->setActions(_actions);

	return element;
}


}
