/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_DATAACCESS_H_
#define CORE_DATAACCESS_H_

#include <common/base/BaseTypes.h>
#include <common/base/CommonTypes.h>
#include <string>
#include <vector>

namespace dataprovider {

class ServantAccessorDirectImpl;
class CommandAccessorDirectImpl;
class PropertyAccessorDirectImpl;

}

namespace core {

class PropertyAccessor;
class CommandAccessor;

/**
 * Singleton class providing an access to the data provider. It encapsulates the different accessors.
 */
class DataAccess {

public :
	/**
	 * Servant type.
	 */
	typedef enum {CONTROLLER, DRIVER} ServantType;

	/**
	 * Gets the instance.
	 */
	static DataAccess* getInstance();

	/**
	 * Initialises the instance.
	 */
	static void initInstance(dataprovider::ServantAccessorDirectImpl* servantAccessor, dataprovider::PropertyAccessorDirectImpl* propertyAccessor, dataprovider::CommandAccessorDirectImpl* commandAccessor);

	/**
	 * Gets the servant type.
	 */
	ServantType getServantType(unsigned long databaseID, const std::string& servantName) const;

	/**
	 * Gets the controller type
	 */
	std::string getControllerType(unsigned long databaseID, unsigned long controllerID) const;

	/**
	 * Gets the controller type
	 */
	std::string getDriverType(unsigned long databaseID, unsigned long driverID) const;

	/**
	 * Returns the ID corresponding to the servant name.
	 */
	unsigned long getServantID(const std::string& servantName) const;

	/**
	 * Returns the ID corresponding to the property name of servant with ID.
	 */
	unsigned long getPropertyID(unsigned long databaseID, unsigned long servantID, const std::string& propertyName) const;

	/**
	 * Gets the dynamic property size.
	 */
	long getDynamicPropertySize(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Returns the ID corresponding to the dynamic property at given index.
	 */
	unsigned long getDynamicPropertyIDAtIndex(unsigned long databaseID, unsigned long propertyID, long index, bool resize) const;

	/**
	 * Returns the ID corresponding to the command name of servant with ID.
	 */
	unsigned long getCommandID(unsigned long databaseID, unsigned long servantID, const std::string& commandName) const;

	/**
	 * Gets the sister command ID.
	 */
	unsigned long getSisterCommandID(unsigned long databaseID, unsigned long commandID, const std::string& commandName) const;

	/**
	 * Returns true if the command belongs to a ics servant
	 */
	bool isCommandFromIcsServant(unsigned long databaseID, unsigned long commandID) const;

	/**
	 * Returns the ID of the servant that contains the property ID.
	 */
	unsigned long getServantIDFromPropertyID(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Returns the ID of the servant that contains the command ID.
	 */
	unsigned long getServantIDFromCommandID(unsigned long databaseID, unsigned long commandID) const;

	/**
	 * Returns the servant name corresponding to the servantID.
	 */
	std::string getServantName(unsigned long databaseID, unsigned long servantID) const;

	/**
	 * Returns the servant name corresponding to the servantID.
	 */
	std::string getCommandName(unsigned long databaseID, unsigned long commandID) const;

	/**
	 * Gets the command elapsed time.
	 */
	double getCommandElapsedTimeMs(unsigned long databaseID, unsigned long commandID) const;

	/**
	 * Returns true if the servant is enabled.
	 */
	bool isServantEnabledAndInitialised(unsigned long databaseID, unsigned long servantID) const;

	/**
	 * Returns the type of property.
	 */
	std::string getPropertyType(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Returns the name of property.
	 */
	std::string getPropertyName(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Returns the name of property.
	 */
	std::pair<std::string, long> getCompletePropertyName(unsigned long databaseID, unsigned long servantID, unsigned long propertyID) const;

	/**
	 * Structure for the result of the getPropertyInfo function.
	 */
	struct PropertyInfo {
		std::string aliasPropertyName;
		bool isSetpoint = false;
		bool isOffset = false;
		bool isDynamic = false;
	};

	/**
	 * Returns the info of the property ID.
	 * throw NoSuchPropertyException
	 */
	PropertyInfo getPropertyInfo(unsigned long databaseID, const std::string& controllerType, unsigned long propertyID);

	/**
	 * Gets the int8 value for the property of ID propertyID.
	 */
	int8 getInt8Value(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the int16 value for the property of ID propertyID.
	 */
	uint16 getUint16Value(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the int32 value for the property of ID propertyID.
	 */
	uint32 getUint32Value(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the int64 value for the property of ID propertyID.
	 */
	uint64 getUint64Value(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the int16 value for the property of ID propertyID.
	 */
	int16 getInt16Value(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the int32 value for the property of ID propertyID.
	 */
	int32 getInt32Value(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the int64 value for the property of ID propertyID.
	 */
	int64 getInt64Value(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the float32 value for the property of ID propertyID.
	 */
	float32 getFloat32Value(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the float64 value for the property of ID propertyID.
	 */
	float64 getFloat64Value(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the string value for the property of ID propertyID.
	 */
	std::string getStringValue(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the boolean value for the property of ID propertyID.
	 */
	bool getBooleanValue(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the formatted value for the property of ID propertyID.
	 */
	std::string getFormattedValue(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Sets the state of the command.
	 */
	void setCommandState(unsigned long databaseID, unsigned long servantID, unsigned long commandID, common::CommandState state);

	/**
	 * Returns the state of the command.
	 */
	common::CommandState getCommandState(unsigned long databaseID, unsigned long commandID) const;

	/**
	 * Sets the command from the command zone.
	 */
	void setCommandFromCommandZone(unsigned long databaseID, unsigned long commandID, bool fromCommandZone);

	/**
	 * Returns true if the command exists.
	 */
	bool commandExists(unsigned long commandID);

	/**
	 * Returns true if the command exists.
	 */
	bool propertyExists(unsigned long propertyID);

	/**
	 * Returns true if the servant exists.
	 */
	bool servantExists(unsigned long servantID);

	/**
	 * Sets the command log ID.
	 */
	void setCommandLogID(unsigned long databaseID, unsigned long commandID, const std::string& logID);

	/**
	 * Gets the description of the element.
	 */
	std::string what(unsigned long databaseID, unsigned long elementID);

	/**
	 * Returns true if the servant associated with the command ID is pausable.
	 */
	bool isServantPausable(unsigned long databaseID, unsigned long commandID);

	/**
	 * Structure for property info.
	 */
	struct PropertyInfoItem {
		std::string actualPropertyName;
		std::string setpointPropertyName;
		std::string offsetPropertyName;
		std::string aliasPropertyName;
		bool isScan;
		bool isSpy;
		bool isDynamic;

		PropertyInfoItem() : isScan(false), isSpy(false), isDynamic(false) {}
	};

	/**
	 * Gets the property info items.
	 */
	std::vector<PropertyInfoItem> getPropertyInfoItems(unsigned long databaseID, const std::string& controllerType);

	/**
	 * Gets the property info items.
	 */
	std::vector<PropertyInfoItem> getAllPropertyInfoItems(unsigned long databaseID, const std::string& controllerType);

	/**
	 * Returns true if property is saved in command box.
	 */
	bool isPropertySavedInCommandBox(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the enumerated values.
	 */
	unsigned long getEnumeratedValues(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the enumerated labels.
	 */
	unsigned long getEnumeratedLabels(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Returns the ID of the unit property.
	 */
	unsigned long getUnitPropertyID(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Returns true if the servant is in debug.
	 */
	bool isServantInDebug(unsigned long databaseID, unsigned long servantID);


	std::vector<unsigned long> getPropertiesSavedInBox(unsigned long databaseID);


private :
	DataAccess();
	~DataAccess();
	void init(dataprovider::ServantAccessorDirectImpl* servantAccessor, dataprovider::PropertyAccessorDirectImpl* propertyAccessor, dataprovider::CommandAccessorDirectImpl* commandAccessor);

	static DataAccess* _instance;

	dataprovider::ServantAccessorDirectImpl* _servantAccessor;
	dataprovider::PropertyAccessorDirectImpl* _propertyAccessor;
	dataprovider::CommandAccessorDirectImpl* _commandAccessor;
};


}

#endif /*CORE_DATAACCESS_H_*/
