/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DataNotificationRouter.h"
#include "PropertyChangeSubscriber.h"
#include "CommandStateChangeSubscriber.h"
#include "CommandProgressChangeSubscriber.h"
#include "CommandTerminatedSubscriber.h"
#include "PropertySetterTerminatedSubscriber.h"
#include "ConfigurationChangeSubscriber.h"
#include <core/message/datanotificationrouter/PropertyChangedMessage.h>
#include <core/message/datanotificationrouter/CommandStateChangedMessage.h>
#include <core/message/datanotificationrouter/CommandProgressChangedMessage.h>
#include <core/message/datanotificationrouter/CommandTerminatedMessage.h>
#include <core/message/datanotificationrouter/PropertySetterTerminatedMessage.h>
#include <core/message/datanotificationrouter/ConfigurationChangedMessage.h>
#include <core/message/datanotificationrouter/SubscriptionMessage.h>
#include <core/message/datanotificationrouter/DataNotificationShutdownMessage.h>
#include <common/base/Queue.h>
#include <common/output/OutputStream.h>
#include <boost/bind.hpp>
#include <iostream>

using namespace common;
using namespace std;

namespace core {

DataNotificationRouter* DataNotificationRouter::_instance = 0;
unsigned long DataNotificationRouter::SUBSCRIPTION_PRIORITY = 1;
unsigned long DataNotificationRouter::NOTIFICATION_PRIORITY = 2;

DataNotificationRouter::DataNotificationRouter() :
	_queue(new common::Queue<DataNotificationMessage>(1024)),
	_thread(nullptr) {
}

DataNotificationRouter::~DataNotificationRouter() {
	delete _queue;
}

void DataNotificationRouter::initInstance() {
	if (_instance == nullptr) {
		_instance = new DataNotificationRouter();
		_instance->start();
	}
}

void DataNotificationRouter::resetInstance() {
	_instance->stop();
	delete _instance;
}

DataNotificationRouter* DataNotificationRouter::getInstance() {
	return _instance;
}


void DataNotificationRouter::start() {
	_thread = new std::thread(&DataNotificationRouter::run);
}

void DataNotificationRouter::stop() {

	if (_thread != 0) {

		Queue<bool> shutdownQueue(1);

		// Create new shutdown Message and put it on the queue
		_queue->put(new DataNotificationShutdownMessage(&shutdownQueue));

		delete shutdownQueue.get();

		_queue->stop();

		_thread->join();
		delete _thread;
	}
}

void DataNotificationRouter::propertyChanged(unsigned long databaseID, unsigned long propertyID) {
	_queue->put(new PropertyChangedMessage(databaseID, propertyID), NOTIFICATION_PRIORITY);
}

void DataNotificationRouter::commandStateChanged(unsigned long databaseID, unsigned long commandID, common::CommandState state) {
	_queue->put(new CommandStateChangedMessage(databaseID, commandID, state), NOTIFICATION_PRIORITY);
}

void DataNotificationRouter::commandProgressChanged(unsigned long databaseID, unsigned long commandID, double progression) {
	_queue->put(new CommandProgressChangedMessage(databaseID, commandID, progression), NOTIFICATION_PRIORITY);
}

void DataNotificationRouter::commandTerminated(unsigned long databaseID, unsigned long commandID, unsigned long requestID) {
	_queue->put(new CommandTerminatedMessage(databaseID, commandID, requestID), NOTIFICATION_PRIORITY);
}

void DataNotificationRouter::propertySetterTerminated(unsigned long databaseID, unsigned long propertyID, unsigned long requestID, bool successful) {
	_queue->put(new PropertySetterTerminatedMessage(databaseID, propertyID, requestID, successful), NOTIFICATION_PRIORITY);
}

void DataNotificationRouter::configurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState enable) {
	_queue->put(new ConfigurationChangedMessage(databaseID, servantID, enable), NOTIFICATION_PRIORITY);
}

void DataNotificationRouter::addPropertyChangeSubscriber(unsigned long propertyID, PropertyChangeSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(propertyID, subscriber, true), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::removePropertyChangeSubscriber(unsigned long propertyID, PropertyChangeSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(propertyID, subscriber, false), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::removePropertyChangeSubscriber(std::set<unsigned long> propertyIDs, PropertyChangeSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(propertyIDs, subscriber, false), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::addCommandStateChangeSubscriber(unsigned long commandID, CommandStateChangeSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(commandID, subscriber, true), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::removeCommandStateChangeSubscriber(unsigned long commandID, CommandStateChangeSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(commandID, subscriber, false), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::addCommandProgressChangeSubscriber(unsigned long commandID, CommandProgressChangeSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(commandID, subscriber, true), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::removeCommandProgressChangeSubscriber(unsigned long commandID, CommandProgressChangeSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(commandID, subscriber, false), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::addCommandTerminatedSubscriber(unsigned long commandID, CommandTerminatedSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(commandID, subscriber, true), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::removeCommandTerminatedSubscriber(unsigned long commandID, CommandTerminatedSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(commandID, subscriber, false), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::addPropertySetterTerminatedSubscriber(unsigned long propertyID, PropertySetterTerminatedSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(propertyID, subscriber, true), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::removePropertySetterTerminatedSubscriber(unsigned long propertyID, PropertySetterTerminatedSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(propertyID, subscriber, false), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::addConfigurationChangeSubscriber(ConfigurationChangeSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(0, subscriber, true), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

void DataNotificationRouter::removeConfigurationChangeSubscriber(ConfigurationChangeSubscriber* subscriber) {
	_queue->put(new SubscriptionMessage(0, subscriber, false), SUBSCRIPTION_PRIORITY);
	subscriber->waitForSubscriptionCompletion();
}

/**
 * Main loop
 */
void DataNotificationRouter::run() {

	bool running = true;
	while (running) {

		DataNotificationMessage* message = 0;

		try {
			message = _instance->_queue->get();

			_instance->processMessage(message);

		} catch (common::Queue<DataNotificationMessage>::UnavailableException& e) {
			running = false;
		} catch (common::Exception& e) {
			err << "Exception in " << message->getTypeString() << " : " << e.getMessage() << eol;
		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}

		delete message;
	}

}

void DataNotificationRouter::processMessage(DataNotificationMessage* message) {

	if (message->getType() == DataNotificationMessage::PROPERTY_CHANGED) {
		processMessage(static_cast<PropertyChangedMessage*>(message));

	} else if (message->getType() == DataNotificationMessage::COMMAND_STATE_CHANGED) {
		processMessage(static_cast<CommandStateChangedMessage*>(message));

	} else if (message->getType() == DataNotificationMessage::COMMAND_PROGRESS_CHANGED) {
		processMessage(static_cast<CommandProgressChangedMessage*>(message));

	} else if (message->getType() == DataNotificationMessage::COMMAND_TERMINATED) {
		processMessage(static_cast<CommandTerminatedMessage*>(message));

	} else if (message->getType() == DataNotificationMessage::PROPERTY_SETTER_TERMINATED) {
		processMessage(static_cast<PropertySetterTerminatedMessage*>(message));

	} else if (message->getType() == DataNotificationMessage::CONFIGURATION_CHANGED) {
		processMessage(static_cast<ConfigurationChangedMessage*>(message));

	} else if (message->getType() == DataNotificationMessage::SUBSCRIPTION) {
		processMessage(static_cast<SubscriptionMessage*>(message));

	} else if (message->getType() == DataNotificationMessage::SHUTDOWN) {
		processMessage(static_cast<DataNotificationShutdownMessage*>(message));
	}
}

void DataNotificationRouter::processMessage(PropertyChangedMessage* message) {

	unsigned long databaseID = message->getDatabaseID();
	unsigned long ID = message->getPropertyID();

	if (_propertySubscribers.find(ID) != _propertySubscribers.end()) {
		set<PropertyChangeSubscriber*>::iterator it;
		for (it = _propertySubscribers[ID].begin(); it != _propertySubscribers[ID].end(); it++) {
			//if ((*it)->getDatabaseID() == databaseID) {
				(*it)->propertyChanged(databaseID, ID);
			//}
		}
	}
}

void DataNotificationRouter::processMessage(CommandStateChangedMessage* message) {

	unsigned long databaseID = message->getDatabaseID();
	unsigned long ID = message->getCommandID();
	common::CommandState state = message->getCommandState();

	// notify subscribers
	if (_commandStateSubscribers.find(ID) != _commandStateSubscribers.end()) {
		set<CommandStateChangeSubscriber*>::iterator it;

		for (it = _commandStateSubscribers[ID].begin(); it != _commandStateSubscribers[ID].end(); it++) {
			//if ((*it)->getDatabaseID() == databaseID) {
				(*it)->commandStateChanged(databaseID, ID, state);
			//}
		}
	}
}

void DataNotificationRouter::processMessage(CommandProgressChangedMessage* message) {

	unsigned long databaseID = message->getDatabaseID();
	unsigned long ID = message->getCommandID();
	double progression = message->getProgression();

	// notify subscribers
	if (_commandProgressSubscribers.find(ID) != _commandProgressSubscribers.end()) {
		set<CommandProgressChangeSubscriber*>::iterator it;
		for (it = _commandProgressSubscribers[ID].begin(); it != _commandProgressSubscribers[ID].end(); it++) {
			//if ((*it)->getDatabaseID() == databaseID) {
				(*it)->commandProgressChanged(databaseID, ID, progression);
			//}
		}
	}
}

void DataNotificationRouter::processMessage(CommandTerminatedMessage* message) {

	unsigned long databaseID = message->getDatabaseID();
	unsigned long ID = message->getCommandID();
	unsigned long requestID = message->getRequestID();

	// notify subscribers
	if (_commandTerminatedSubscribers.find(ID) != _commandTerminatedSubscribers.end()) {
		set<CommandTerminatedSubscriber*>::iterator it;
		for (it = _commandTerminatedSubscribers[ID].begin(); it != _commandTerminatedSubscribers[ID].end(); it++) {
			//if ((*it)->getDatabaseID() == databaseID) {
				(*it)->commandTerminated(databaseID, ID, requestID);
			//}
		}
	}
}

void DataNotificationRouter::processMessage(PropertySetterTerminatedMessage* message) {

	unsigned long databaseID = message->getDatabaseID();
	unsigned long ID = message->getPropertyID();
	unsigned long requestID = message->getRequestID();
	bool successful = message->isSuccessful();

	// notify subscribers
	if (_propertySetterTerminatedSubscribers.find(ID) != _propertySetterTerminatedSubscribers.end()) {
		set<PropertySetterTerminatedSubscriber*>::iterator it;
		for (it = _propertySetterTerminatedSubscribers[ID].begin(); it != _propertySetterTerminatedSubscribers[ID].end(); it++) {
			//if ((*it)->getDatabaseID() == databaseID) {
				(*it)->propertySetterTerminated(databaseID, ID, requestID, successful);
			//}
		}
	}
}

void DataNotificationRouter::processMessage(ConfigurationChangedMessage* message) {

	unsigned long databaseID = message->getDatabaseID();
	unsigned long servantID = message->getServantID();
	common::ConfigurationState state = message->getState();

	set<ConfigurationChangeSubscriber*>::iterator it;
	for (it = _configurationSubscribers.begin(); it != _configurationSubscribers.end(); it++) {
		//if ((*it)->getDatabaseID() == databaseID) {
			(*it)->configurationChanged(databaseID, servantID, state);
		//}
	}
}

void DataNotificationRouter::processMessage(SubscriptionMessage* message) {

	bool isSubscribe = message->isSubscriber();
	DataChangeSubscriber* subscriber = message->getSubscriber();

	if (message->isBulk()) {
		const std::set<unsigned long>& dataIDs = message->getDataIDs();
		for (std::set<unsigned long>::const_iterator it = dataIDs.begin(); it != dataIDs.end(); it++) {
			performSubscriptionOperation(*it, isSubscribe, subscriber);
		}

	} else {
		unsigned long dataID = message->getDataID();
		performSubscriptionOperation(dataID, isSubscribe, subscriber);
	}

	subscriber->setSubscriptionCompleted();
}

void DataNotificationRouter::processMessage(DataNotificationShutdownMessage* message) {

	// do nothing
	message->getQueue()->put(new bool(true));
}

void DataNotificationRouter::performSubscriptionOperation(unsigned long dataID, bool isSubscribe, DataChangeSubscriber* subscriber) {

	if (subscriber->getSubscriberType() == DataChangeSubscriber::PROPERTY_CHANGE) {
		if (isSubscribe == true) {
			// subscriber property change subscriber
			_propertySubscribers[dataID].insert(static_cast<PropertyChangeSubscriber*>(subscriber));

		} else {
			// unsubscriber property change subscriber
			_propertySubscribers[dataID].erase(static_cast<PropertyChangeSubscriber*>(subscriber));

			if (_propertySubscribers[dataID].size() == 0) {
				_propertySubscribers.erase(dataID);
			}
		}

	} else if (subscriber->getSubscriberType() == DataChangeSubscriber::COMMAND_STATE_CHANGE) {
		if (isSubscribe == true) {
			// subscriber command state change subscriber
			_commandStateSubscribers[dataID].insert(static_cast<CommandStateChangeSubscriber*>(subscriber));

		} else {
			// unsubscriber command state change subscriber
			_commandStateSubscribers[dataID].erase(static_cast<CommandStateChangeSubscriber*>(subscriber));

			if (_commandStateSubscribers[dataID].size() == 0) {
				_commandStateSubscribers.erase(dataID);
			}
		}

	} else if (subscriber->getSubscriberType() == DataChangeSubscriber::COMMAND_PROGRESS_CHANGE) {
		if (isSubscribe == true) {
			// subscriber command state change subscriber
			_commandProgressSubscribers[dataID].insert(static_cast<CommandProgressChangeSubscriber*>(subscriber));

		} else {
			// unsubscriber command state change subscriber
			_commandProgressSubscribers[dataID].erase(static_cast<CommandProgressChangeSubscriber*>(subscriber));

			if (_commandProgressSubscribers[dataID].size() == 0) {
				_commandProgressSubscribers.erase(dataID);
			}
		}

	} else if (subscriber->getSubscriberType() == DataChangeSubscriber::COMMAND_TERMINATED) {
		if (isSubscribe == true) {
			// subscriber
			_commandTerminatedSubscribers[dataID].insert(static_cast<CommandTerminatedSubscriber*>(subscriber));

		} else {
			// unsubscriber
			_commandTerminatedSubscribers[dataID].erase(static_cast<CommandTerminatedSubscriber*>(subscriber));

			if (_commandTerminatedSubscribers[dataID].size() == 0) {
				_commandTerminatedSubscribers.erase(dataID);
			}
		}

	} else if (subscriber->getSubscriberType() == DataChangeSubscriber::PROPERTY_SETTER_TERMINATED) {
		if (isSubscribe == true) {
			// subscribe
			_propertySetterTerminatedSubscribers[dataID].insert(static_cast<PropertySetterTerminatedSubscriber*>(subscriber));

		} else {
			// unsubscribe
			_propertySetterTerminatedSubscribers[dataID].erase(static_cast<PropertySetterTerminatedSubscriber*>(subscriber));

			if (_propertySetterTerminatedSubscribers[dataID].size() == 0) {
				_propertySetterTerminatedSubscribers.erase(dataID);
			}
		}

	} else if (subscriber->getSubscriberType() == DataChangeSubscriber::CONFIGURATION_CHANGE) {
		if (isSubscribe == true) {
			// subscribe
			_configurationSubscribers.insert(static_cast<ConfigurationChangeSubscriber*>(subscriber));

		} else {
			// unsubscribe
			_configurationSubscribers.erase(static_cast<ConfigurationChangeSubscriber*>(subscriber));
		}
	}
}

}
