/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_DATANOTIFICATIONROUTER_H_
#define CORE_DATANOTIFICATIONROUTER_H_

#include "DataAccess.h"
#include <common/base/CommonTypes.h>
#include <thread>
#include <map>
#include <set>

namespace common {
	template<typename ValueType>
	class Queue;
}

namespace core {

class DataNotificationMessage;
class PropertyChangedMessage;
class CommandStateChangedMessage;
class CommandProgressChangedMessage;
class CommandTerminatedMessage;
class PropertySetterTerminatedMessage;
class ConfigurationChangedMessage;
class SubscriptionMessage;
class DataNotificationShutdownMessage;
class PropertyChangeSubscriber;
class CommandStateChangeSubscriber;
class CommandProgressChangeSubscriber;
class CommandTerminatedSubscriber;
class PropertySetterTerminatedSubscriber;
class ConfigurationChangeSubscriber;
class DataChangeSubscriber;

/**
 * Class defining a data notification router that forwards the notifications to all the
 * subscribers. The class is a singleton which processes the messages sequentially thanks to
 * a message queue.
 */
class DataNotificationRouter {

public :
	/**
	 * Gets the instance.
	 */
	static DataNotificationRouter* getInstance();
	
	/**
	 * Initialises the instance.
	 */
	static void initInstance();
	
	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Starts the instance.
	 */
	void start();
	
	/**
	 * Stops the instance.
	 */
	void stop();
	
	/**
	 * Notifies that the property has changed.
	 */
	void propertyChanged(unsigned long databaseID, unsigned long propertyID);
	
	/**
	 * Notifies that the command state has changed.
	 */
	void commandStateChanged(unsigned long databaseID, unsigned long commandID, common::CommandState state);
	
	/**
	 * Notifies that the command progress has changed.
	 */
	void commandProgressChanged(unsigned long databaseID, unsigned long commandID, double progression);
	
	/**
	 * Notifies that the command has terminated.
	 */
	void commandTerminated(unsigned long databaseID, unsigned long commandID, unsigned long requestID);
	
	/**
	 * Notifies that the property setter is terminated.
	 */
	void propertySetterTerminated(unsigned long databaseID, unsigned long propertyID, unsigned long requestID, bool successful);

	/**
	 * Notifies that the configuration has changed.
	 */
	void configurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState enable);

	/**
	 * Adds a property change subscriber.
	 */
	void addPropertyChangeSubscriber(unsigned long propertyID, PropertyChangeSubscriber* subscriber);
	
	/**
	 * Removes a property change subscriber.
	 */
	void removePropertyChangeSubscriber(unsigned long propertyID, PropertyChangeSubscriber* subscriber);
	
	/**
	 * Removes a property change subscriber from a list of property IDs.
	 */
	void removePropertyChangeSubscriber(std::set<unsigned long> propertyIDs, PropertyChangeSubscriber* subscriber);

	/**
	 * Adds a command state change subscriber.
	 */
	void addCommandStateChangeSubscriber(unsigned long commandID, CommandStateChangeSubscriber* subscriber);
	
	/**
	 * Removes a command state change subscriber.
	 */
	void removeCommandStateChangeSubscriber(unsigned long commandID, CommandStateChangeSubscriber* subscriber);

	/**
	 * Adds a command progress change subscriber.
	 */
	void addCommandProgressChangeSubscriber(unsigned long commandID, CommandProgressChangeSubscriber* subscriber);
	
	/**
	 * Removes a command progress change subscriber.
	 */
	void removeCommandProgressChangeSubscriber(unsigned long commandID, CommandProgressChangeSubscriber* subscriber);

	/**
	 * Adds a command progress change subscriber.
	 */
	void addCommandTerminatedSubscriber(unsigned long commandID, CommandTerminatedSubscriber* subscriber);
	
	/**
	 * Removes a command terminated subscriber.
	 */
	void removeCommandTerminatedSubscriber(unsigned long commandID, CommandTerminatedSubscriber* subscriber);
	
	/**
	 * Adds a property setter terminated subscriber.
	 */
	void addPropertySetterTerminatedSubscriber(unsigned long propertyID, PropertySetterTerminatedSubscriber* subscriber);
	
	/**
	 * Removes a property setter terminated subscriber.
	 */
	void removePropertySetterTerminatedSubscriber(unsigned long propertyID, PropertySetterTerminatedSubscriber* subscriber);

	/**
	 * Adds a configuration change subscriber.
	 */
	void addConfigurationChangeSubscriber(ConfigurationChangeSubscriber* subscriber);

	/**
	 * Removes a configuration change subscriber.
	 */
	void removeConfigurationChangeSubscriber(ConfigurationChangeSubscriber* subscriber);


private :
	DataNotificationRouter();
	virtual ~DataNotificationRouter();
	static void run();

	void processMessage(DataNotificationMessage* message);
	void processMessage(PropertyChangedMessage* message);
	void processMessage(CommandStateChangedMessage* message);
	void processMessage(CommandProgressChangedMessage* message);
	void processMessage(CommandTerminatedMessage* message);
	void processMessage(PropertySetterTerminatedMessage* message);
	void processMessage(ConfigurationChangedMessage* message);
	void processMessage(SubscriptionMessage* message);
	void processMessage(DataNotificationShutdownMessage* message);

	void performSubscriptionOperation(unsigned long dataID, bool isSubscribe, DataChangeSubscriber* subscriber);
	
	static DataNotificationRouter* _instance;

	static unsigned long SUBSCRIPTION_PRIORITY;
	static unsigned long NOTIFICATION_PRIORITY;

	common::Queue<DataNotificationMessage>* _queue;
	std::thread* _thread;

	std::map<unsigned long, std::set<PropertyChangeSubscriber*> > _propertySubscribers;
	std::map<unsigned long, std::set<CommandStateChangeSubscriber*> > _commandStateSubscribers;
	std::map<unsigned long, std::set<CommandProgressChangeSubscriber*> > _commandProgressSubscribers;
	std::map<unsigned long, std::set<CommandTerminatedSubscriber*> > _commandTerminatedSubscribers;
	std::map<unsigned long, std::set<PropertySetterTerminatedSubscriber*> > _propertySetterTerminatedSubscribers;
	std::set<ConfigurationChangeSubscriber*> _configurationSubscribers;
};


} // namespace dataprovider

#endif /*CORE_DATANOTIFICATIONROUTER_H_*/
