/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ConditionsParser.h"
#include "ConditionElement.h"
#include <core/conditions/CommandZoneTermination.h>
#include <common/output/OutputStream.h>
#include <common/base/Convert.h>
#include <iostream>


using namespace std;
using namespace common;

namespace core {

const std::string ConditionsParser::CONDITIONS_TAG = "conditions";
const std::string ConditionsParser::ALARM_TAG = "alarm";
const std::string ConditionsParser::WARNING_TAG = "warning";
const std::string ConditionsParser::WATCHDOG_TAG = "watchdog";
const std::string ConditionsParser::COMMAND_ZONE_TERMINATION_TAG = "command_zone_termination";
const std::string ConditionsParser::MAIL_TAG = "mail";
const std::string ConditionsParser::SMS_TAG = "sms";

const std::string ConditionsParser::EXPRESSION_TAG = "expression";
const std::string ConditionsParser::MESSAGE_TAG = "message";
const std::string ConditionsParser::ACTIONS_TAG = "actions";

const std::string ConditionsParser::NAME_ATTRIBUTE = "name";
const std::string ConditionsParser::ON_ATTRIBUTE = "on";
const std::string ConditionsParser::ACTIVATION_DELAY_ATTRIBUTE = "activation_delay";
const std::string ConditionsParser::STOP_COMMAND_ZONE_ATTRIBUTE = "stop_command_zone";
const std::string ConditionsParser::RECIPIENT_ATTRIBUTE = "recipient";
const std::string ConditionsParser::NUMBER_ATTRIBUTE = "number";
const std::string ConditionsParser::VALUE_ATTRIBUTE = "value";
const std::string ConditionsParser::CONTENT_ATTRIBUTE = "content";

using namespace std;

ConditionsParser::ConditionsParser() :
	common::GenericParser(),
	_currentCondition(0) {
}

ConditionsParser::~ConditionsParser() {
	vector<ConditionElement*>::iterator it;
	for (it = _conditions.begin(); it < _conditions.end(); it++) {
		delete (*it);
	}
	_conditions.clear();
}


void ConditionsParser::startElement(const std::string& qName) {

	if (qName == CONDITIONS_TAG) {
		// do nothing

	} else if (qName == ALARM_TAG) {

		string timeString = getAttribute(ACTIVATION_DELAY_ATTRIBUTE);
		int delay = 0;
		if (timeString != "") {
			delay = to<int>(timeString);
		}

		_currentCondition = new ConditionElement(common::CONDITION_ALARM, getAttribute(NAME_ATTRIBUTE), delay, getAttribute(ON_ATTRIBUTE) == "true", getAttribute(STOP_COMMAND_ZONE_ATTRIBUTE) == "true");

	} else if (qName == WARNING_TAG) {
		_currentCondition = new ConditionElement(common::CONDITION_WARNING, getAttribute(NAME_ATTRIBUTE), 0, getAttribute(ON_ATTRIBUTE) == "true", getAttribute(STOP_COMMAND_ZONE_ATTRIBUTE) == "true");
		
	} else if (qName == WATCHDOG_TAG) {
		_currentCondition = new ConditionElement(common::CONDITION_WATCHDOG, getAttribute(NAME_ATTRIBUTE), 0, getAttribute(ON_ATTRIBUTE) == "true", getAttribute(STOP_COMMAND_ZONE_ATTRIBUTE) == "true");

	} else if (qName == COMMAND_ZONE_TERMINATION_TAG) {
		_currentCondition = new ConditionElement(common::CONDITION_COMMAND_ZONE_TERMINATION, CommandZoneTermination::NAME, 0, getAttribute(ON_ATTRIBUTE) == "true", false);

	} else if (qName == EXPRESSION_TAG) {
		_currentCondition->setExpression(getAttribute(VALUE_ATTRIBUTE));

	} else if (qName == MESSAGE_TAG) {
		_currentCondition->setMessage(getAttribute(VALUE_ATTRIBUTE));

	} else if (qName == MAIL_TAG) {
		_currentCondition->setMailRecipient(getAttribute(RECIPIENT_ATTRIBUTE));

	} else if (qName == SMS_TAG) {
		_currentCondition->setSmsNumber(getAttribute(NUMBER_ATTRIBUTE));

	} else if (qName == ACTIONS_TAG) {
		_currentCondition->setActions(getAttribute(CONTENT_ATTRIBUTE));
		
	} else {
		throw UnknownTagException(qName, _fileName);
	}
	
}

void ConditionsParser::endElement(const std::string& qName) {

	if (qName == CONDITIONS_TAG) {
		// do nothing
		
	} else if (qName == ALARM_TAG) {
		// verify that we have a condition and an action
		_conditions.push_back(_currentCondition);

	} else if (qName == WARNING_TAG) {
		// verify that we have a condition and an action
		_conditions.push_back(_currentCondition);
		
	} else if (qName == WATCHDOG_TAG) {
		// verify that we have a condition and an action
		_conditions.push_back(_currentCondition);

	} else if (qName == COMMAND_ZONE_TERMINATION_TAG) {
		// verify that we have a condition and an action
		_conditions.push_back(_currentCondition);

	} else if (qName == EXPRESSION_TAG) {
		// do nothing

	} else if (qName == MESSAGE_TAG) {
		// do nothing

	} else if (qName == MAIL_TAG) {
		// do nothing

	} else if (qName == SMS_TAG) {
		// do nothing

	} else if (qName == ACTIONS_TAG) {
		// do nothing

	} else {
		throw UnknownTagException(qName, _fileName);
	}
}

const std::vector<ConditionElement*>& ConditionsParser::getConditionElements() const {
	return _conditions;
}

}
