/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Scheduler.h"
#include "SchedulerArbitrator.h"
#include <core/parser/schedulerrules/SchedulerRulesParser.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/message/Message.h>
#include <core/message/scheduler/NewCommand.h>
#include <core/message/scheduler/NewEnableDisableCommand.h>
#include <core/message/scheduler/StopCommand.h>
#include <core/message/scheduler/PauseCommand.h>
#include <core/message/scheduler/RestartCommand.h>
#include <core/message/scheduler/StopSetter.h>
#include <core/message/scheduler/PauseSetter.h>
#include <core/message/scheduler/RestartSetter.h>
#include <core/message/scheduler/InitDriversAndProtocols.h>
#include <core/message/scheduler/CommandTerminated.h>
#include <core/message/scheduler/PropertySetterTerminated.h>
#include <core/message/scheduler/DurationRequest.h>
#include <core/message/scheduler/DurationResult.h>
#include <core/message/Shutdown.h>
#include <common/base/Queue.h>
#include <common/base/Date.h>
#include <common/base/PropertyType.h>
#include <core/transportlayer/command/CommandLauncher.h>
#include <core/dataaccess/CommandTerminatedSubscriber.h>
#include <core/dataaccess/PropertySetterTerminatedSubscriber.h>
#include <boost/bind.hpp>
#include <common/output/OutputStream.h>
#include <iostream>

using namespace std;
using namespace common;

namespace core {

Scheduler* Scheduler::_instance = 0;

void Scheduler::initInstance(const std::string& configPath, CommandLauncher* commandLauncher) {

	if (_instance == 0) {
		_instance = new Scheduler();
		_instance->init(configPath, commandLauncher);
	}
}

Scheduler* Scheduler::getInstance() {

	return _instance;
}

void Scheduler::resetInstance() {

	if (_instance != 0) {
		_instance->reset();
	}

	delete _instance;
}

Scheduler::Scheduler() :
	_runningCore(false),
	_runningCommand(false),
	_commandLauncher(nullptr),
	_coreThread(nullptr),
	_commandProcessorThread(nullptr),
	_coreMessageQueue(nullptr),
	_commandProcessorMessageQueue(nullptr),
	_requestCounter(0),
	_arbitrator(new SchedulerArbitrator()) {
}

Scheduler::~Scheduler() {

	if (_coreThread != nullptr) {
		delete _coreThread;
	}

	if (_commandProcessorThread != nullptr) {
		delete _commandProcessorThread;
	}

	delete _coreMessageQueue;
	delete _commandProcessorMessageQueue;
	delete _arbitrator;
}

void Scheduler::init(const std::string& configPath, CommandLauncher* commandLauncher) {

	_configPath = configPath;

	_commandLauncher = commandLauncher;

//	_arbitrator->loadRules(configPath);

	// Initialising the queue with a big size. The queue should never be full!
	_coreMessageQueue = new Queue<Message>(1024);
	_commandProcessorMessageQueue = new Queue<Message>(1024);

	_coreThread = new std::thread(&Scheduler::coreLoop);
	_commandProcessorThread = new std::thread(&Scheduler::commandLoop);
}

void Scheduler::reset() {

	if (_coreThread != nullptr && _commandProcessorThread != nullptr) {

		postToCoreMessageQueue(new Shutdown(true));
		postToCommandProcessorMessageQueue(new Shutdown(false));

		_coreThread->join();
		_commandProcessorThread->join();

		delete _coreThread;
		_coreThread = nullptr;

		delete _commandProcessorThread;
		_commandProcessorThread = nullptr;
	}
}

void Scheduler::stop() {

	// locking the commands list
	{
		std::scoped_lock lock(_operationsMutex);

		// empty pending operations
		_pendingOperations.clear();

		// stopping active operations
		for (list<Operation*>::iterator o = _activeOperations.begin(); o != _activeOperations.end(); o++) {

			NewCommand* command = dynamic_cast<NewCommand*>(*o);
			if (command != 0) {
				_commandLauncher->stop(command->getDatabaseID(), command->getCommandID());
			}

			NewEnableDisableCommand* newEnableDisableCommand = dynamic_cast<NewEnableDisableCommand*>(*o);
			if (newEnableDisableCommand != 0) {
				_commandLauncher->stop(newEnableDisableCommand->getDatabaseID(), newEnableDisableCommand->getCommandID());
			}

			BaseSetValue* setter = dynamic_cast<BaseSetValue*>(*o);
			if (setter != 0) {
				_commandLauncher->stopSetter(setter->getDatabaseID(), setter->getPropertyID());
			}
		}
	}

	// waiting for the active operations be empty
	while (true) {

		usleep(100000);

		std::scoped_lock lock(_operationsMutex);
		if (_activeOperations.empty()) {
			break;
		}
	}
}

void Scheduler::coreLoop() {

	Message* message = 0;

	_instance->_runningCore = true;
	while (_instance->_runningCore) {

		try {
			try {
				message = _instance->_coreMessageQueue->get();

			} catch (Queue<Message>::UnavailableException& e) {
				break;

			} catch (Exception& e) {
				err << e.printStack() << eol;
				break;
			}

			try {
				_instance->processMessage(message);

			} catch (Exception& e) {
				err << e.printStack() << eol;
			}

		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}

		delete message;
		message = 0;
 	}
}

void Scheduler::commandLoop() {

	Message* message = 0;

	_instance->_runningCommand = true;
	while (_instance->_runningCommand) {

		try {
			try {
				message = _instance->_commandProcessorMessageQueue->get();

			} catch (Queue<Message>::UnavailableException& e) {
				break;

			} catch (Exception& e) {
				err << e.printStack() << eol;
				break;
			}

			try {
				_instance->processMessage(message);

			} catch (Exception& e) {
				err << e.printStack() << eol;
			}

		} catch (exception& e) {
			err << "STL exception : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception" << eol;
		}

		delete message;
		message = 0;
 	}
}

int Scheduler::getNextRequestID() {

	// locking the request counter
	std::scoped_lock lock(_requestCounterMutex);
	++_requestCounter;

	return _requestCounter;
}

RequestReply Scheduler::requestNewCommand(unsigned long databaseID, unsigned long commandID, bool skipIfWait, bool fromCommandZone, CommandTerminatedSubscriber* subscriber, int& requestID) {

	Queue<RequestReply> waitForReplyQueue(1);

	requestID = getNextRequestID();
	Message* message = new NewCommand(databaseID, DataAccess::getInstance()->getServantIDFromCommandID(databaseID, commandID), commandID, fromCommandZone, skipIfWait, subscriber, &waitForReplyQueue);
	message->setRequestID(requestID);

	postToCoreMessageQueue(message);

	RequestReply* replyPtr = waitForReplyQueue.get();
	RequestReply reply = *replyPtr;
	delete replyPtr;

	return reply;
}

void Scheduler::requestStopCommand(unsigned long databaseID, unsigned long commandID, int& requestID) {

	requestID = getNextRequestID();
	Message* message = new StopCommand(databaseID, commandID);
	message->setRequestID(requestID);

	postToCoreMessageQueue(message);
}

void Scheduler::requestPauseCommand(unsigned long databaseID, unsigned long commandID, int& requestID) {

	requestID = getNextRequestID();
	Message* message = new PauseCommand(databaseID, commandID);
	message->setRequestID(requestID);

	postToCoreMessageQueue(message);
}

void Scheduler::requestRestartCommand(unsigned long databaseID, unsigned long commandID, int& requestID) {

	requestID = getNextRequestID();
	Message* message = new RestartCommand(databaseID, commandID);
	message->setRequestID(requestID);

	postToCoreMessageQueue(message);
}

double Scheduler::requestDuration(unsigned long databaseID, unsigned long commandID) {

	Queue<double> waitForReplyQueue(1);

	int requestID = getNextRequestID();
	Message* message = new DurationRequest(databaseID, DataAccess::getInstance()->getServantIDFromCommandID(databaseID, commandID), commandID, &waitForReplyQueue);
	message->setRequestID(requestID);

	postToCoreMessageQueue(message);

	double* valuePtr = waitForReplyQueue.get();
	double value = *valuePtr;
	delete valuePtr;

	return value;
}

void Scheduler::requestStopSetter(unsigned long databaseID, unsigned long propertyID, int& requestID) {

	requestID = getNextRequestID();
	Message* message = new StopSetter(databaseID, propertyID);
	message->setRequestID(requestID);

	postToCoreMessageQueue(message);
}

void Scheduler::requestPauseSetter(unsigned long databaseID, unsigned long propertyID, int& requestID) {

	requestID = getNextRequestID();
	Message* message = new PauseSetter(databaseID, propertyID);
	message->setRequestID(requestID);

	postToCoreMessageQueue(message);
}

void Scheduler::requestRestartSetter(unsigned long databaseID, unsigned long propertyID, int& requestID) {

	requestID = getNextRequestID();
	Message* message = new RestartSetter(databaseID, propertyID);
	message->setRequestID(requestID);

	postToCoreMessageQueue(message);
}

void Scheduler::initDriversAndProtocols(int& requestID) {

	requestID = getNextRequestID();
	Message* message = new InitDriversAndProtocols();
	message->setRequestID(requestID);

	postToCoreMessageQueue(message);
}

void Scheduler::commandSuccess(unsigned long requestID, unsigned long commandID, const std::string& message) {
	postToCommandProcessorMessageQueue(new CommandTerminated(requestID, commandID, message, true));
}

void Scheduler::commandFailure(unsigned long requestID, unsigned long commandID, const std::string& message) {
	postToCommandProcessorMessageQueue(new CommandTerminated(requestID, commandID, message, false));
}

void Scheduler::propertySuccess(unsigned long requestID, unsigned long propertyID, const std::string& message) {
	postToCommandProcessorMessageQueue(new PropertySetterTerminated(requestID, propertyID, message, true));
}

void Scheduler::propertyFailure(unsigned long requestID, unsigned long propertyID, const std::string& message) {
	postToCommandProcessorMessageQueue(new PropertySetterTerminated(requestID, propertyID, message, false));
}

void Scheduler::durationRequestSuccess(unsigned long requestID, unsigned long commandID, double result, const std::string& message) {
	postToCommandProcessorMessageQueue(new DurationResult(requestID, commandID, result, message, true));
}

void Scheduler::durationRequestFailure(unsigned long requestID, unsigned long commandID, const std::string& message) {
	postToCommandProcessorMessageQueue(new DurationResult(requestID, commandID, 0, message, false));
}

void Scheduler::postToCoreMessageQueue(Message* message) {
	_coreMessageQueue->put(message);
}

void Scheduler::postToCommandProcessorMessageQueue(Message* message) {
	_commandProcessorMessageQueue->put(message);
}

void Scheduler::processMessage(Message* message) {

	if (message->getType() == Message::NEW_COMMAND) {
		processMessage(static_cast<NewCommand*>(message));

	} else if (message->getType() == Message::STOP_COMMAND) {
		processMessage(static_cast<StopCommand*>(message));

	} else if (message->getType() == Message::PAUSE_COMMAND) {
		processMessage(static_cast<PauseCommand*>(message));

	} else if (message->getType() == Message::RESTART_COMMAND) {
		processMessage(static_cast<RestartCommand*>(message));

	} else if (message->getType() == Message::STOP_SETTER) {
		processMessage(static_cast<StopSetter*>(message));

	} else if (message->getType() == Message::PAUSE_SETTER) {
		processMessage(static_cast<PauseSetter*>(message));

	} else if (message->getType() == Message::RESTART_SETTER) {
		processMessage(static_cast<RestartSetter*>(message));

	} else if (message->getType() == Message::INIT_DRIVERS_AND_PROTOCOLS) {
		processMessage(static_cast<InitDriversAndProtocols*>(message));

	} else if (message->getType() == Message::COMMAND_TERMINATED) {
		processMessage(static_cast<CommandTerminated*>(message));

	} else if (message->getType() == Message::SET_VALUE) {
		processMessage(static_cast<BaseSetValue*>(message));

	} else if (message->getType() == Message::PROPERTY_SETTER_TERMINATED) {
		processMessage(static_cast<PropertySetterTerminated*>(message));

	} else if (message->getType() == Message::DURATION_REQUEST) {
		processMessage(static_cast<DurationRequest*>(message));

	} else if (message->getType() == Message::DURATION_RESULT) {
		processMessage(static_cast<DurationResult*>(message));

	} else if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<Shutdown*>(message));

	}
}

void Scheduler::processMessage(NewCommand* message) {

	// Locking the commands list
	std::scoped_lock lock(_operationsMutex);

	// setting the subscriber
	CommandTerminatedSubscriber* subscriber = message->getSubscriber();

	if (subscriber != 0) {
		subscriber->setCommandActive(message->getRequestID());
	}

	RequestReply reply = _arbitrator->requestExecution(message->getDatabaseID(), message->getCommandID(), _activeOperations);

	message->getWaitForReplyQueue()->put(new RequestReply(reply));

	if (reply.getValue() == RequestReply::EXECUTE) {

		// copying the command and executing
		_activeOperations.push_back(message->clone());
		execute(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getCommandID(), message->fromCommandZone(), message->skipIfWait());

	} else if (reply.getValue() == RequestReply::QUEUE) {

		// copying the command and adding to the pending list
		_pendingOperations.push_back(message->clone());

		// Setting the PENDING state to the command
		DataAccess::getInstance()->setCommandState(message->getDatabaseID(), message->getServantID(), message->getCommandID(), common::PENDING);

	} else {
		// Notifying the listeners
		DataNotificationRouter::getInstance()->commandTerminated(message->getDatabaseID(), message->getCommandID(), message->getRequestID());
	}

}

void Scheduler::processMessage(StopCommand* message) {

	stopCommand(message);
}

void Scheduler::processMessage(PauseCommand* message) {

	pauseCommand(message);
}

void Scheduler::processMessage(RestartCommand* message) {

	restartCommand(message);
}

void Scheduler::processMessage(StopSetter* message) {

	stopSetter(message);
}

void Scheduler::processMessage(PauseSetter* message) {

	pauseSetter(message);
}

void Scheduler::processMessage(RestartSetter* message) {

	restartSetter(message);
}

void Scheduler::processMessage(InitDriversAndProtocols* message) {

	// Calling directly the command launcher. more synchronisation can be added if necessary
	_commandLauncher->initDriversAndProtocols();
}

void Scheduler::processMessage(CommandTerminated* message) {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	// Erasing the command and getting the database ID
	unsigned long databaseID = 0;
	for (list<Operation*>::iterator c = _activeOperations.begin(); c != _activeOperations.end(); c++) {

		NewCommand* newCommand = dynamic_cast<NewCommand*>(*c);
		if (newCommand != 0) {
			databaseID = newCommand->getDatabaseID();
		}

		NewEnableDisableCommand* newEnableDisableCommand = dynamic_cast<NewEnableDisableCommand*>(*c);
		if (newEnableDisableCommand != 0) {
			databaseID = newEnableDisableCommand->getDatabaseID();
		}

		if ((*c)->getRequestID() == message->getRequestID()) {
			delete *c;
			c = _activeOperations.erase(c);
		}
	}

	// Notifying the listeners
	DataNotificationRouter::getInstance()->commandTerminated(databaseID, message->getCommandID(), message->getRequestID());

	launchPendingOperation();
}

void Scheduler::processMessage(BaseSetValue* message) {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	// getting the subscriber
	PropertySetterTerminatedSubscriber* subscriber = message->getSubscriber();
	if (subscriber != 0) {
		subscriber->setSetterActive(message->getRequestID());
	}

	RequestReply reply =_arbitrator->requestPropertySetter(message->getDatabaseID(), message->getPropertyID(), _activeOperations);

	message->getWaitForReplyQueue()->put(new RequestReply(reply));

	if (reply.getValue() == RequestReply::EXECUTE) {

		// copying the command
		_activeOperations.push_back(message->clone());
		executeSetter(message);

	} else if (reply.getValue() == RequestReply::QUEUE) {

		// copying the command and adding to the pending list
		_pendingOperations.push_back(message->clone());

	} else {
		// Notifying the listeners
		DataNotificationRouter::getInstance()->propertySetterTerminated(message->getDatabaseID(), message->getPropertyID(), message->getRequestID(), false);
	}

}

void Scheduler::processMessage(PropertySetterTerminated* message) {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	// Erasing the command
	unsigned long databaseID = 0;
	for (list<Operation*>::iterator c = _activeOperations.begin(); c != _activeOperations.end(); c++) {

		BaseSetValue* setValue = dynamic_cast<BaseSetValue*>(*c);
		if (setValue != 0) {
			if (setValue->getRequestID() == message->getRequestID()) {
				databaseID = setValue->getDatabaseID();
				delete setValue;
				c = _activeOperations.erase(c);
			}
		}
	}

	// Notifying the listeners
	DataNotificationRouter::getInstance()->propertySetterTerminated(databaseID, message->getPropertyID(), message->getRequestID(), message->isSuccessful());

	launchPendingOperation();
}

void Scheduler::processMessage(DurationRequest* message) {

	// Locking the commands list
	std::scoped_lock lock(_operationsMutex);

	RequestReply reply =_arbitrator->requestDuration(message->getDatabaseID(), message->getCommandID(), _activeOperations);

	if (reply.getValue() == RequestReply::EXECUTE) {

		// copying the request and executing
		_activeOperations.push_back(message->clone());
		executeDurationRequest(message->getRequestID(), message->getDatabaseID(), message->getCommandID());

	} else {
		message->getWaitForReplyQueue()->stop();
	}

}

void Scheduler::processMessage(DurationResult* message) {

	// Locking the commands list
	std::scoped_lock lock(_operationsMutex);

	for (list<Operation*>::iterator c = _activeOperations.begin(); c != _activeOperations.end(); c++) {

		DurationRequest* request = dynamic_cast<DurationRequest*>(*c);
		if (request != 0) {
			if (request->getRequestID() == message->getRequestID()) {

				if (message->isSuccessful()) {
					request->getWaitForReplyQueue()->put(new double(message->getResult()));

				} else {
					// stop the queue to signal a failure to the waiting thread
					request->getWaitForReplyQueue()->stop();
				}

				delete request;
				c = _activeOperations.erase(c);

			}
		}
	}
}

void Scheduler::processMessage(Shutdown* message) {

	if (message->isCore()) {
		_runningCore = false;
	} else {
		_runningCommand = false;
	}
}

void Scheduler::launchPendingOperation() {

	// No need to lock the operations mutex because launchPendingOperation is always called in the scope of an operations mutex.

	// Launching a command
	if (!_pendingOperations.empty()) {

		list<Operation*>::iterator o = _pendingOperations.begin();

		while (o != _pendingOperations.end()) {

			// New command
			NewCommand* command = dynamic_cast<NewCommand*>(*o);
			if (command != 0) {

				RequestReply reply =_arbitrator->requestExecution(command->getDatabaseID(), command->getCommandID(), _activeOperations);

				if (reply.getValue() == RequestReply::EXECUTE) {

					// copying the command and executing
					_activeOperations.push_back(command->clone());
					execute(command->getRequestID(), command->getDatabaseID(), command->getServantID(), command->getCommandID(), command->fromCommandZone(), command->skipIfWait());
					o = _pendingOperations.erase(o);

				} else if (reply.getValue() == RequestReply::QUEUE) {

					// do nothing
					o++;

				} else {

					o = _pendingOperations.erase(o);
				}

			}

			// Setter
			BaseSetValue* setter = dynamic_cast<BaseSetValue*>(*o);
			if (setter != 0) {

				RequestReply reply =_arbitrator->requestPropertySetter(setter->getDatabaseID(), setter->getPropertyID(), _activeOperations);

				if (reply.getValue() == RequestReply::EXECUTE) {

					// copying the property setter and executing
					_activeOperations.push_back(setter->clone());
					executeSetter(setter);
					o = _pendingOperations.erase(o);

				} else if (reply.getValue() == RequestReply::QUEUE) {

					// do nothing
					o++;

				} else {

					o = _pendingOperations.erase(o);
				}
			}

			// Duration request
			DurationRequest* durationRequest = dynamic_cast<DurationRequest*>(*o);
			if (durationRequest != 0) {

				RequestReply reply =_arbitrator->requestDuration(durationRequest->getDatabaseID(), durationRequest->getCommandID(), _activeOperations);

				if (reply.getValue() == RequestReply::EXECUTE) {

					// copying the duration request and executing
					_activeOperations.push_back(durationRequest->clone());
					executeDurationRequest(durationRequest->getRequestID(), durationRequest->getDatabaseID(), durationRequest->getCommandID());
					o = _pendingOperations.erase(o);

				} else if (reply.getValue() == RequestReply::QUEUE) {

					// do nothing
					o++;

				} else {

					o = _pendingOperations.erase(o);
				}
			}
		}
	}
}

void Scheduler::execute(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long commandID, bool fromCommandZone, bool skipIfWait) {

	// Setting the STARTING state to the command
	DataAccess::getInstance()->setCommandState(databaseID, servantID, commandID, common::STARTING);
	DataAccess::getInstance()->setCommandFromCommandZone(databaseID, commandID, fromCommandZone);

	_commandLauncher->execute(requestID, databaseID, commandID, skipIfWait);
}

void Scheduler::executeDurationRequest(unsigned long requestID, unsigned long databaseID, unsigned long commandID) {
	_commandLauncher->executeDurationRequest(requestID, databaseID, commandID);
}

void Scheduler::stopCommand(StopCommand* command) {
	_commandLauncher->stop(command->getDatabaseID(), command->getCommandID());
}

void Scheduler::pauseCommand(PauseCommand* command) {
	_commandLauncher->pause(command->getDatabaseID(), command->getCommandID());
}

void Scheduler::restartCommand(RestartCommand* command) {
	_commandLauncher->restart(command->getDatabaseID(), command->getCommandID());
}

void Scheduler::stopSetter(StopSetter* setter) {
	_commandLauncher->stopSetter(setter->getDatabaseID(), setter->getPropertyID());
}

void Scheduler::pauseSetter(PauseSetter* setter) {
	_commandLauncher->pauseSetter(setter->getDatabaseID(), setter->getPropertyID());
}

void Scheduler::restartSetter(RestartSetter* setter) {
	_commandLauncher->restartSetter(setter->getDatabaseID(), setter->getPropertyID());
}

std::list<NewCommand> Scheduler::getActiveCommands() {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	// copy to avoid blocking the Scheduler
	std::list<NewCommand> result;

	for (list<Operation*>::const_iterator c = _activeOperations.begin(); c != _activeOperations.end(); c++) {
		NewCommand* command = dynamic_cast<NewCommand*>(*c);
		if (command != 0) {
			result.push_back(*command);
		}
	}

	return result;
}

std::list<NewEnableDisableCommand> Scheduler::getActiveEnableDisableCommands() {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	// copy to avoid blocking the Scheduler
	std::list<NewEnableDisableCommand> result;

	for (list<Operation*>::const_iterator c = _activeOperations.begin(); c != _activeOperations.end(); c++) {
		NewEnableDisableCommand* command = dynamic_cast<NewEnableDisableCommand*>(*c);
		if (command != 0) {
			result.push_back(*command);
		}
	}

	return result;
}

std::list<NewCommand> Scheduler::getPendingCommands() {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	// copy to avoid blocking the Scheduler
	std::list<NewCommand> result;

	for (list<Operation*>::const_iterator c = _pendingOperations.begin(); c != _pendingOperations.end(); c++) {
		NewCommand* command = dynamic_cast<NewCommand*>(*c);
		if (command != 0) {
			result.push_back(*command);
		}
	}

	return result;
}

std::list<BaseSetValue> Scheduler::getActiveSetValues() {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	// copy to avoid blocking the Scheduler
	std::list<BaseSetValue> result;

	for (list<Operation*>::const_iterator c = _activeOperations.begin(); c != _activeOperations.end(); c++) {
		BaseSetValue* setValue = dynamic_cast<BaseSetValue*>(*c);
		if (setValue != 0) {
			result.push_back(*setValue);
		}
	}

	return result;
}

std::list<BaseSetValue> Scheduler::getPendingSetValues() {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	// copy to avoid blocking the Scheduler
	std::list<BaseSetValue> result;

	for (list<Operation*>::const_iterator c = _pendingOperations.begin(); c != _pendingOperations.end(); c++) {
		BaseSetValue* setValue = dynamic_cast<BaseSetValue*>(*c);
		if (setValue != 0) {
			result.push_back(*setValue);
		}
	}

	return result;
}

std::list<std::pair<Operation*, bool> > Scheduler::getOperations() {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	// copy to avoid blocking the Scheduler
	std::list<std::pair<Operation*, bool> > result;

	for (list<Operation*>::const_iterator c = _activeOperations.begin(); c != _activeOperations.end(); c++) {
		result.push_back(pair<Operation*, bool>((*c)->clone(), true));
	}

	for (list<Operation*>::const_iterator c = _pendingOperations.begin(); c != _pendingOperations.end(); c++) {
		result.push_back(pair<Operation*, bool>((*c)->clone(), false));
	}

	return result;
}


void Scheduler::resetSchedulerRules() {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	delete _arbitrator;

	_arbitrator = new SchedulerArbitrator();

	_arbitrator->loadRules(_configPath);

}

bool Scheduler::canRun(unsigned long databaseID, unsigned long commandID) {

	// Locking the active commands list
	std::scoped_lock lock(_operationsMutex);

	return (_arbitrator->requestExecution(databaseID, commandID, _activeOperations, false).getValue() != RequestReply::REJECT);
}

void Scheduler::executeSetter(BaseSetValue* message) {

	if (dynamic_cast<SetValue<int8>*>(message) != 0) {
		SetValueHelper<int8>::apply(_commandLauncher, static_cast<SetValue<int8>*>(message));
	} else if (dynamic_cast<SetValue<uint16>*>(message) != 0) {
		SetValueHelper<uint16>::apply(_commandLauncher, static_cast<SetValue<uint16>*>(message));
	} else if (dynamic_cast<SetValue<uint32>*>(message) != 0) {
		SetValueHelper<uint32>::apply(_commandLauncher, static_cast<SetValue<uint32>*>(message));
	} else if (dynamic_cast<SetValue<uint64>*>(message) != 0) {
		SetValueHelper<uint64>::apply(_commandLauncher, static_cast<SetValue<uint64>*>(message));
	} else if (dynamic_cast<SetValue<int16>*>(message) != 0) {
		SetValueHelper<int16>::apply(_commandLauncher, static_cast<SetValue<int16>*>(message));
	} else if (dynamic_cast<SetValue<int32>*>(message) != 0) {
		SetValueHelper<int32>::apply(_commandLauncher, static_cast<SetValue<int32>*>(message));
	} else if (dynamic_cast<SetValue<int64>*>(message) != 0) {
		SetValueHelper<int64>::apply(_commandLauncher, static_cast<SetValue<int64>*>(message));
	} else if (dynamic_cast<SetValue<float32>*>(message) != 0) {
		SetValueHelper<float32>::apply(_commandLauncher, static_cast<SetValue<float32>*>(message));
	} else if (dynamic_cast<SetValue<float64>*>(message) != 0) {
		SetValueHelper<float64>::apply(_commandLauncher, static_cast<SetValue<float64>*>(message));
	} else if (dynamic_cast<SetValue<bool>*>(message) != 0) {
		SetValueHelper<bool>::apply(_commandLauncher, static_cast<SetValue<bool>*>(message));
	} else if (dynamic_cast<SetValue<string>*>(message) != 0) {
		SetValueHelper<string>::apply(_commandLauncher, static_cast<SetValue<string>*>(message));
	} else {
		err << "Cannot process the message " << message->getTypeString() << " type " << typeid(*message).name() << eol;
	}
}

void SetValueHelper<bool>::apply(CommandLauncher* commandLauncher, SetValue<bool>* message) {

	commandLauncher->setBooleanValue(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<std::string>::apply(CommandLauncher* commandLauncher, SetValue<std::string>* message) {

	commandLauncher->setStringValue(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<int8>::apply(CommandLauncher* commandLauncher, SetValue<int8>* message) {

	commandLauncher->setInt8Value(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<uint16>::apply(CommandLauncher* commandLauncher, SetValue<uint16>* message) {

	commandLauncher->setUint16Value(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<uint32>::apply(CommandLauncher* commandLauncher, SetValue<uint32>* message) {

	commandLauncher->setUint32Value(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<uint64>::apply(CommandLauncher* commandLauncher, SetValue<uint64>* message) {

	commandLauncher->setUint64Value(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<int16>::apply(CommandLauncher* commandLauncher, SetValue<int16>* message) {

	commandLauncher->setInt16Value(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<int32>::apply(CommandLauncher* commandLauncher, SetValue<int32>* message) {

	commandLauncher->setInt32Value(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<int64>::apply(CommandLauncher* commandLauncher, SetValue<int64>* message) {

	commandLauncher->setInt64Value(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<float32>::apply(CommandLauncher* commandLauncher, SetValue<float32>* message) {

	commandLauncher->setFloat32Value(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

void SetValueHelper<float64>::apply(CommandLauncher* commandLauncher, SetValue<float64>* message) {

	commandLauncher->setFloat64Value(message->getRequestID(), message->getDatabaseID(), message->getServantID(), message->getPropertyID(), message->getValue());
}

}
