/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_SCHEDULER_H_
#define CORE_SCHEDULER_H_

#include "RequestReply.h"
#include <common/base/PropertyType.h>
#include <core/message/scheduler/SetValue.h>
#include <core/dataaccess/DataAccess.h>
#include <thread>
#include <list>
#include <mutex>

namespace common {

template<typename ItemType>
class Queue;

}

namespace core {

class CommandLauncher;
class Message;
class NewCommand;
class NewEnableDisableCommand;
class StopCommand;
class PauseCommand;
class RestartCommand;
class StopSetter;
class PauseSetter;
class RestartSetter;
class InitDriversAndProtocols;
class CommandTerminated;
class PropertySetterTerminated;
class DurationRequest;
class DurationResult;
class Shutdown;
class CommandShutdown;
class SchedulerArbitrator;
class CommandTerminatedSubscriber;
class PropertySetterTerminatedSubscriber;

template<typename ValueType>
struct SetValueHelper;

/**
 * Class defining a scheduler. It is implemented with a single thread that processes messages
 * from a message queue. There are two types of message: new command and command terminated.
 * The scheduler allows only a single to be executed at a time. That is why it is active
 * when a command is executing and inactive otherwise. When a new command message is processed,
 * if the scheduler is inactive, then it launches the command and puts it into the active command list
 * or puts it into the pending command list if it is active. When a command terminated message is
 * processed the scheduler first erases the command from the active command list and then if it is
 * inactive tries to launch a command from the pending command list.
 */
class Scheduler {

public:
	/**
	 * Initialises the instance.
	 */
	static void initInstance(const std::string& configPath, CommandLauncher* commandLauncher);

	/**
	 * Gets the instance.
	 */
	static Scheduler* getInstance();

	/**
	 * Resets the instance.
	 */
	static void resetInstance();


	/**
	 * Requests the execution of a new Command of ID commandID.
	 */
	RequestReply requestNewCommand(unsigned long databaseID, unsigned long commandID, bool skipIfWait, bool fromCommandZone, CommandTerminatedSubscriber* subscriber, int& requestID);

	/**
	 * Requests the stop of the execution of a Command of ID commandID.
	 */
	void requestStopCommand(unsigned long databaseID, unsigned long commandID, int& requestID);

	/**
	 * Requests the pause of the execution of a Command of ID commandID.
	 */
	void requestPauseCommand(unsigned long databaseID, unsigned long commandID, int& requestID);

	/**
	 * Requests the restart of the execution of a Command of ID commandID.
	 */
	void requestRestartCommand(unsigned long databaseID, unsigned long commandID, int& requestID);

	/**
	 * Requests the duration of a Command of ID commandID.
	 */
	double requestDuration(unsigned long databaseID, unsigned long commandID);

	/**
	 * Requests the stop of a setter.
	 */
	void requestStopSetter(unsigned long databaseID, unsigned long propertyID, int& requestID);

	/**
	 * Requests the pause of a setter.
	 */
	void requestPauseSetter(unsigned long databaseID, unsigned long propertyID, int& requestID);

	/**
	 * Requests the restart of a setter.
	 */
	void requestRestartSetter(unsigned long databaseID, unsigned long propertyID, int& requestID);

	/**
	 * Initialises the drivers and the protocols.
	 */
	void initDriversAndProtocols(int& requestID);

	/**
	 * Notifies the termination success of a Command of ID commandID.
	 */
	void commandSuccess(unsigned long requestID, unsigned long commandID, const std::string& message);

	/**
	 * Notifies the termination failure of a Command of ID commandID.
	 */
	void commandFailure(unsigned long requestID, unsigned long commandID, const std::string& message);

	/**
	 * Notifies the termination success of a Property of ID propertyID.
	 */
	void propertySuccess(unsigned long requestID, unsigned long propertyID, const std::string& message);

	/**
	 * Notifies the termination failure of a Property of ID propertyID.
	 */
	void propertyFailure(unsigned long requestID, unsigned long propertyID, const std::string& message);

	/**
	 * Notifies the termination success of a duration request of ID commandID.
	 */
	void durationRequestSuccess(unsigned long requestID, unsigned long commandID, double result, const std::string& message);

	/**
	 * Notifies the termination success of a duration request of ID commandID.
	 */
	void durationRequestFailure(unsigned long requestID, unsigned long commandID, const std::string& message);

	/**
	 * Requests the set of a Property.
	 */
	template<typename ValueType>
	RequestReply requestSetValue(unsigned long databaseID, unsigned long propertyID, const ValueType& value, PropertySetterTerminatedSubscriber* subscriber, int& requestID) {

		common::Queue<RequestReply> waitForReplyQueue(1);

		requestID = getNextRequestID();

		unsigned long servantID = DataAccess::getInstance()->getServantIDFromPropertyID(databaseID, propertyID);
		if (servantID == 0) {
			RequestReply reply(RequestReply::FAIL, "cannot find associated servant");
			return reply;
		}

		Message* message = new SetValue<ValueType>(databaseID, servantID, propertyID, value, subscriber, &waitForReplyQueue);
		message->setRequestID(requestID);

		postToCoreMessageQueue(message);

		RequestReply* replyPtr = waitForReplyQueue.get();
		RequestReply reply = *replyPtr;
		delete replyPtr;

		return reply;
	}

	/**
	 * Gets the active commands. Copy the original list.
	 */
	std::list<NewCommand> getActiveCommands();

	/**
	 * Gets the enable disable active commands. Copy the original list.
	 */
	std::list<NewEnableDisableCommand> getActiveEnableDisableCommands();

	/**
	 * Gets the pending commands. Copy the original list.
	 */
	std::list<NewCommand> getPendingCommands();

	/**
	 * Gets the active set values. Copy the original list.
	 */
	std::list<BaseSetValue> getActiveSetValues();

	/**
	 * Gets the pending set values. Copy the original list.
	 */
	std::list<BaseSetValue> getPendingSetValues();

	/**
	 * Gets all the operations. Copy the original list.
	 */
	std::list<std::pair<Operation*, bool> > getOperations();

	/**
	 * Reinitialises the scheduler rules.
	 */
	void resetSchedulerRules();

	/**
	 * Returns true if the command can run.
	 */
	bool canRun(unsigned long databaseID, unsigned long commandID);

	/**
	 * Stops the scheduler.
	 */
	void stop();

private:
	Scheduler();
	virtual ~Scheduler();
	void init(const std::string& configPath, CommandLauncher* commandLauncher);
	void reset();

	/**
	 * Posts to the core message queue. The core message queue is used by one thread and
	 * the command message queue is used by another one. They concurrently access to the
	 * active operations list, pending operations list so that we use a mutex to protect
	 * the shared data.
	 */
	void postToCoreMessageQueue(Message* message);
	void postToCommandProcessorMessageQueue(Message* message);

	void launchPendingOperation();

	void processMessage(Message* message);
	void processMessage(NewCommand* message);
	void processMessage(StopCommand* message);
	void processMessage(PauseCommand* message);
	void processMessage(RestartCommand* message);
	void processMessage(StopSetter* message);
	void processMessage(PauseSetter* message);
	void processMessage(RestartSetter* message);
	void processMessage(InitDriversAndProtocols* message);
	void processMessage(CommandTerminated* message);
	void processMessage(BaseSetValue* message);
	void processMessage(PropertySetterTerminated* message);
	void processMessage(DurationRequest* message);
	void processMessage(DurationResult* message);
	void processMessage(Shutdown* message);

	void executeSetter(BaseSetValue* message);

	void launchCommand(NewCommand* command);
	void stopCommand(StopCommand* command);
	void pauseCommand(PauseCommand* command);
	void restartCommand(RestartCommand* command);
	void stopSetter(StopSetter* setter);
	void pauseSetter(PauseSetter* setter);
	void restartSetter(RestartSetter* setter);

	void execute(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long commandID, bool fromCommandZone, bool skipIfWait);
	void executeDurationRequest(unsigned long requestID, unsigned long databaseID, unsigned long commandID);
	int getNextRequestID();

	static void coreLoop();
	static void commandLoop();

	static Scheduler* _instance;

	bool _runningCore;
	bool _runningCommand;

	CommandLauncher* _commandLauncher;

	std::thread* _coreThread;
	std::thread* _commandProcessorThread;

	std::mutex _operationsMutex;
	std::list<Operation*> _activeOperations;
	std::list<Operation*> _pendingOperations;

	common::Queue<Message>* _coreMessageQueue;
	common::Queue<Message>* _commandProcessorMessageQueue;

	std::mutex _requestCounterMutex;
	int _requestCounter;

	std::string _configPath;
	SchedulerArbitrator* _arbitrator;


};

// specialising SetValueMessageHelper

template<>
struct SetValueHelper<bool> {

	static void apply(CommandLauncher* commandLauncher, SetValue<bool>* message);
};

template<>
struct SetValueHelper<std::string> {

	static void apply(CommandLauncher* commandLauncher, SetValue<std::string>* message);
};

template<>
struct SetValueHelper<int8> {

	static void apply(CommandLauncher* commandLauncher, SetValue<int8>* message);
};

template<>
struct SetValueHelper<uint16> {

	static void apply(CommandLauncher* commandLauncher, SetValue<uint16>* message);
};

template<>
struct SetValueHelper<uint32> {

	static void apply(CommandLauncher* commandLauncher, SetValue<uint32>* message);
};

template<>
struct SetValueHelper<uint64> {

	static void apply(CommandLauncher* commandLauncher, SetValue<uint64>* message);
};

template<>
struct SetValueHelper<int16> {

	static void apply(CommandLauncher* commandLauncher, SetValue<int16>* message);
};

template<>
struct SetValueHelper<int32> {

	static void apply(CommandLauncher* commandLauncher, SetValue<int32>* message);
};

template<>
struct SetValueHelper<int64> {

	static void apply(CommandLauncher* commandLauncher, SetValue<int64>* message);
};

template<>
struct SetValueHelper<float32> {

	static void apply(CommandLauncher* commandLauncher, SetValue<float32>* message);
};

template<>
struct SetValueHelper<float64> {

	static void apply(CommandLauncher* commandLauncher, SetValue<float64>* message);
};

}

#endif /*CORE_SCHEDULER_H_*/
