/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "SchedulerArbitrator.h"
#include <core/parser/schedulerrules/SchedulerRulesParser.h>
#include <core/parser/schedulerrules/ParallelExclusionElement.h>
#include <core/parser/schedulerrules/CommandExclusionElement.h>
#include <core/dataaccess/DataAccess.h>
#include <core/message/scheduler/NewCommand.h>
#include <core/message/scheduler/BaseSetValue.h>
#include <core/message/scheduler/DurationRequest.h>
#include <common/base/ServerProperties.h>
#include <common/base/Convert.h>
#include <common/output/OutputStream.h>

using namespace std;
using namespace common;

namespace core {

const std::string SchedulerArbitrator::SCHEDULER_RULES_PATH = "schedulerRulesPath";

SchedulerArbitrator::SchedulerArbitrator() {
}

SchedulerArbitrator::~SchedulerArbitrator() {
}

void SchedulerArbitrator::loadRules(const std::string& configPath) {

	// Initialise scheduling rules
	string fileName = configPath + "/" + ServerProperties::getInstance()->getString(SCHEDULER_RULES_PATH);
	SchedulerRulesParser* parser = new SchedulerRulesParser();
	cout << "FileName : " << fileName << endl;
	parser->init();
	parser->parseXMLFile(fileName);


	// get not parallelisables
	vector<CommandExclusionElement*> notParallelisables = parser->getNotParallelisables();
	vector<CommandExclusionElement*>::iterator npit;
	for (npit = notParallelisables.begin(); npit != notParallelisables.end(); npit++) {

		CommandExclusionElement* command = *npit;

		try {
			// get command ID from servant name and command name
			unsigned long servantID = DataAccess::getInstance()->getServantID(command->getServantName());
			unsigned long commandID = DataAccess::getInstance()->getCommandID(0, servantID, command->getCommandName());

			_notParallelisableCommandIDs.insert(commandID, command->toBeQueued());

		} catch (Exception& e) {
			err << "Cannot create 'Not Parallelisable' rule for servant '" << command->getServantName() << "' and command name '" << command->getCommandName() << "'" << eol;
			err << e.printStack();
		}
	}

	// get exclusions
	vector<ParallelExclusionElement*> exclusions = parser->getExclusions();

	// iterate over exclusions
	vector<ParallelExclusionElement*>::iterator pit;
	for (pit = exclusions.begin(); pit != exclusions.end(); pit++) {
		ParallelExclusionElement* exclusionElement = *pit;

		vector<CommandExclusionElement*> commands = exclusionElement->getCommands();
		vector<CommandExclusionElement*>::iterator cit;

		vector<pair<unsigned long, bool> > commandGroup;

		// iterate over commands for a given parallel exclusion
		for (cit = commands.begin(); cit != commands.end(); cit++) {
			CommandExclusionElement* command = *cit;

			try {

				// get command ID from servant name and command name
				unsigned long servantID = DataAccess::getInstance()->getServantID(command->getServantName());
				unsigned long commandID = DataAccess::getInstance()->getCommandID(0, servantID, command->getCommandName());

				commandGroup.push_back(pair<unsigned long, bool>(commandID, command->toBeQueued()));

			} catch (Exception& e) {
				err << "Cannot create parallelisation rule for servant '" << command->getServantName() << "' and command name '" << command->getCommandName() << "'" << eol;
				err << e.printStack();
			}

		}

		// Iterate of primary IDs
		vector<pair<unsigned long, bool> >::const_iterator primaryIt;
		for (primaryIt = commandGroup.begin(); primaryIt != commandGroup.end(); primaryIt++) {

			unsigned long primaryID = primaryIt->first;
			bool toBeQueued = primaryIt->second;

			// Iterate over secondary IDs
			vector<pair<unsigned long, bool> >::const_iterator secondaryIt;
			for (secondaryIt = commandGroup.begin(); secondaryIt != commandGroup.end(); secondaryIt++) {

				unsigned long secondaryID = secondaryIt->first;

				// construct rules associated with primary ID
				if (secondaryID != primaryID) {
					if (!_rules.find(primaryID)) {
						_rules.insert(primaryID, vector<pair<unsigned long, bool> >());
					}

					_rules.get(primaryID).push_back(pair<unsigned long, bool>(secondaryID, toBeQueued));

				}
			}
		}

	}

	delete parser;
}

RequestReply SchedulerArbitrator::requestExecution(unsigned long databaseID, unsigned long commandID, const std::list<Operation*>& activeOperations, bool rejectSameCommand) const {

	bool found = false;
	unsigned long conflictingCommandID = 0;
	bool queued = true;

	unsigned long servantID = DataAccess::getInstance()->getServantIDFromCommandID(databaseID, commandID);

	// Determine if the servant is enabled or not
	if (!DataAccess::getInstance()->isServantEnabledAndInitialised(databaseID, servantID)) {
		return RequestReply(RequestReply::REJECT, DataAccess::getInstance()->what(databaseID, servantID) + " is disabled and cannot execute commands");
	}

	// Iterating over the active operations to check whether the requested command is already running
	for (list<Operation*>::const_iterator o = activeOperations.begin(); o != activeOperations.end(); o++) {

		// searching for the command
		NewCommand* runningCommand = dynamic_cast<NewCommand*>(*o);
		if (runningCommand != 0) {
			if (commandID == runningCommand->getCommandID() && rejectSameCommand) {
				found = true;
				conflictingCommandID = commandID;
				queued = false;
			}
		}

		// searching for the set value
		BaseSetValue* setter = dynamic_cast<BaseSetValue*>(*o);
		if (setter != 0) {
			if (servantID == setter->getServantID()) {
				found = true;
				queued = true;
			}
		}

		// searching for the duration request
		DurationRequest* durationRequest = dynamic_cast<DurationRequest*>(*o);
		if (durationRequest != 0) {
			if (servantID == durationRequest->getServantID()) {
				found = true;
				queued = true;
			}
		}
	}

	// Determine if the command is parallelisable
	if (!found) {
		if (_notParallelisableCommandIDs.find(commandID)) {

			// commandID CANNOT be run in parallel : ensure that no other
			// command is running.
			for (list<Operation*>::const_iterator o = activeOperations.begin(); o != activeOperations.end(); o++) {
				NewCommand* runningCommand = dynamic_cast<NewCommand*>(*o);
				if (runningCommand != 0) {
					found = true;
					conflictingCommandID = runningCommand->getCommandID();
					queued = _notParallelisableCommandIDs[commandID];
					break;
				}
			}

		} else {

			// Verify that a nonparallelisable command is not running.
			for (list<Operation*>::const_iterator o = activeOperations.begin(); o != activeOperations.end(); o++) {
				NewCommand* runningCommand = dynamic_cast<NewCommand*>(*o);
				if (runningCommand != 0 && _notParallelisableCommandIDs.find(runningCommand->getCommandID())) {
					found = true;
					conflictingCommandID = runningCommand->getCommandID();
					queued = _notParallelisableCommandIDs[commandID];
					break;
				}
			}

		}
	}

	// Iterating over the active operations to check whether a rule exists
	if (!found && _rules.find(commandID)) {
		const vector<pair<unsigned long, bool> >& refusedParallelCommandIDs = _rules.get(commandID);

		for (list<Operation*>::const_iterator o = activeOperations.begin(); o != activeOperations.end(); o++) {
			NewCommand* runningCommand = dynamic_cast<NewCommand*>(*o);
			if (runningCommand != 0) {
				for (vector<pair<unsigned long, bool> >::const_iterator c = refusedParallelCommandIDs.begin(); c != refusedParallelCommandIDs.end(); c++) {
					if (c->first == runningCommand->getCommandID()) {
						found = true;
						conflictingCommandID = runningCommand->getCommandID();
						queued &= c->second;
					}
				}
			}
		}

	}

	if (found && queued) {
		return RequestReply(RequestReply::QUEUE);

	} else if (found && !queued) {
		return RequestReply(RequestReply::REJECT, DataAccess::getInstance()->what(databaseID, commandID) + " is rejected due to conflict with " + DataAccess::getInstance()->what(databaseID, conflictingCommandID));
	}

	return RequestReply(RequestReply::EXECUTE);
}

RequestReply SchedulerArbitrator::requestEnableDisableExecution(unsigned long databaseID, unsigned long commandID, const std::list<Operation*>& activeOperations) const {

	// can only be executed alone
	//if (activeOperations.empty()) {
	//
	//}

	//return RequestReply(RequestReply::REJECT, DataAccess::getInstance()->what(databaseID, commandID) + " is rejected because it only can be executed alone");

	// execute
	return RequestReply(RequestReply::EXECUTE);
}

RequestReply SchedulerArbitrator::requestDuration(unsigned long databaseID, unsigned long commandID, const std::list<Operation*>& activeOperations) const {

	unsigned long servantID = DataAccess::getInstance()->getServantIDFromCommandID(databaseID, commandID);

	// Iterating over the active operations to check whether the associated command is running
	for (list<Operation*>::const_iterator o = activeOperations.begin(); o != activeOperations.end(); o++) {

		// searching for the command
		NewCommand* runningCommand = dynamic_cast<NewCommand*>(*o);
		if (runningCommand != 0) {
			if (commandID == runningCommand->getCommandID()) {
				return RequestReply(RequestReply::REJECT, DataAccess::getInstance()->what(databaseID, commandID) + " cannot calculate its progression because it is still executed");
			}
		}

		// searching for the set value
		BaseSetValue* setter = dynamic_cast<BaseSetValue*>(*o);
		if (setter != 0) {
			if (servantID == setter->getServantID()) {
				return RequestReply(RequestReply::QUEUE);
			}
		}

		// searching for the duration request
		DurationRequest* durationRequest = dynamic_cast<DurationRequest*>(*o);
		if (durationRequest != 0) {
			if (servantID == durationRequest->getServantID()) {
				return RequestReply(RequestReply::QUEUE);
			}
		}
	}

	return RequestReply(RequestReply::EXECUTE);
}

RequestReply SchedulerArbitrator::requestPropertySetter(unsigned long databaseID, unsigned long propertyID, const std::list<Operation*>& activeOperations) const {

	unsigned long requestedServantID = DataAccess::getInstance()->getServantIDFromPropertyID(0, propertyID);

	// Determine if the servant is enabled or not
	if (!DataAccess::getInstance()->isServantEnabledAndInitialised(databaseID, requestedServantID)) {
		return RequestReply(RequestReply::REJECT, DataAccess::getInstance()->what(databaseID, requestedServantID) + " is disabled and cannot set properties");
	}

	// Iterating over the active operations to check whether the associated servant has a command running
	for (list<Operation*>::const_iterator o = activeOperations.begin(); o != activeOperations.end(); o++) {


// Removed because some controllers need to be updated when it is active
//		// searching for the command
//		NewCommand* runningCommand = dynamic_cast<NewCommand*>(*o);
//		if (runningCommand != 0) {
//			unsigned long servantID = DataAccess::getInstance()->getServantIDFromCommandID(0, runningCommand->getCommandID());
//
//			if (requestedServantID == servantID) {
//				return RequestReply(RequestReply::REJECT, DataAccess::getInstance()->what(databaseID, servantID) + " is still active and cannot set properties");
//			}
//		}

		// searching for the set value
		BaseSetValue* runningSetter = dynamic_cast<BaseSetValue*>(*o);
		if (runningSetter != 0) {
			unsigned long servantID = DataAccess::getInstance()->getServantIDFromPropertyID(0, runningSetter->getPropertyID());

			if (requestedServantID == servantID) {
				return RequestReply(RequestReply::QUEUE);
			}
		}

		// searching for the duration request
		DurationRequest* durationRequest = dynamic_cast<DurationRequest*>(*o);
		if (durationRequest != 0) {
			if (requestedServantID == durationRequest->getServantID()) {
				return RequestReply(RequestReply::QUEUE);
			}
		}
	}

	return RequestReply(RequestReply::EXECUTE);
}


void SchedulerArbitrator::printRules() const {

	HashMap<unsigned long, std::vector<std::pair<unsigned long, bool> > >::const_iterator r;
	for (r = _rules.begin(); r != _rules.end(); r++) {
		out << "Rules for command " << r->first << " : " << eol;
		for (size_t i = 0; i < r->second.size(); i++) {
			unsigned long ID = r->second[i].first;
			bool toBeQueued = r->second[i].second;
			out << "\tCommand " << ID << " queued = " << to<string>(toBeQueued) << eol;
		}
	}
}

}
