/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Sender.h"
#include "SenderCommandTerminatedSubscriber.h"
#include "SenderCommandStateChangeSubscriber.h"
#include <common/base/Convert.h>
#include <common/base/TimeCondition.h>
#include <common/output/OutputStream.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/dataaccess/DataAccess.h>
#include <boost/bind.hpp>

using namespace std;
using namespace common;

namespace core {

Sender::DurationRequestException::DurationRequestException(unsigned long databaseID, unsigned long commandID) 
	: Exception("core::Sender::DurationRequestException", string("cannot get duration for ") + DataAccess::getInstance()->what(databaseID, commandID)) {
}

RequestReply Sender::newCommand(unsigned long databaseID, unsigned long commandID, bool skipIfWait, bool blocking, long timeoutMs) {
	
	SenderCommandTerminatedSubscriber* subscriber = 0;
	
	if (blocking) {
		subscriber = new SenderCommandTerminatedSubscriber(databaseID, commandID);
		DataNotificationRouter::getInstance()->addCommandTerminatedSubscriber(commandID, subscriber);
	}
	
	int requestID = 0;
	RequestReply reply = Scheduler::getInstance()->requestNewCommand(databaseID, commandID, skipIfWait, false, subscriber, requestID);
	
	// waiting for the end of the command if subscriber
	if (blocking) {
	
		std::thread* timeoutThread = nullptr;
		TimeCondition condition;
		
		if (timeoutMs > 0) {
			timeoutThread = new std::thread(std::bind(&Sender::timeout, databaseID, commandID, requestID, &condition, timeoutMs));
		}	
	
		subscriber->waitForTermination();
		
		if (timeoutThread != 0) {
			condition.notify();
		}
		
		DataNotificationRouter::getInstance()->removeCommandTerminatedSubscriber(commandID, subscriber);
		delete subscriber;
		
		if (timeoutThread != 0) {
			timeoutThread->join();
			delete timeoutThread;
		}	
	}
	
	return reply;
}

RequestReply Sender::newCommandFromCommandZone(unsigned long databaseID, unsigned long commandID, CommandTerminatedSubscriber* subscriber) {
	
	int requestID = 0;
	return Scheduler::getInstance()->requestNewCommand(databaseID, commandID, false, true, subscriber, requestID);
}

void Sender::stopCommand(unsigned long databaseID, unsigned long commandID, bool atEnd, bool blocking) {
	
	int requestID = 0;

	string stopCommandName = "stop";
	if (atEnd) {
		stopCommandName = "stopatend";
	}

	try {

		if (DataAccess::getInstance()->isCommandFromIcsServant(databaseID, commandID)) {
			// Special code for ics commands
			unsigned long stopCommandID = DataAccess::getInstance()->getSisterCommandID(databaseID, commandID, stopCommandName);
			if (stopCommandID != 0) {
				// starting the associated stop command
				newCommand(databaseID, stopCommandID, false, blocking);
			}

		} else {
			Scheduler::getInstance()->requestStopCommand(databaseID, commandID, requestID);
		}
	} catch (Exception & e) {
		err << "Cannot stop command : " << e.printStack() << eol;
	}
}

void Sender::pauseCommand(unsigned long databaseID, unsigned long commandID) {

	int requestID = 0;

	try {

		if (DataAccess::getInstance()->isCommandFromIcsServant(databaseID, commandID)) {
			// Special code for ics commands
			unsigned long pauseCommandID = DataAccess::getInstance()->getSisterCommandID(databaseID, commandID, "pause");

			// starting the associated stop command
			if (pauseCommandID != 0) {
				newCommand(databaseID, pauseCommandID, false);
			} else {
				cout << "Cannot pause command <" << commandID << ">" << endl;
			}

		} else {
			Scheduler::getInstance()->requestPauseCommand(databaseID, commandID, requestID);
		}
	} catch (Exception & e) {
		err << "Cannot pause command <" << commandID << "> : " << e.printStack() << eol;
	}
}

void Sender::restartCommand(unsigned long databaseID, unsigned long commandID) {
	
	int requestID = 0;

	try {

		if (DataAccess::getInstance()->isCommandFromIcsServant(databaseID, commandID)) {
			// Special code for ics commands
			unsigned long restartCommandID = DataAccess::getInstance()->getSisterCommandID(databaseID, commandID, "continue");

			// starting the associated stop command
			if (restartCommandID != 0) {
				newCommand(databaseID, restartCommandID, false);
			} else {
				cout << "Cannot restart command <" << commandID << ">" << endl;
			}

		} else {
			Scheduler::getInstance()->requestRestartCommand(databaseID, commandID, requestID);
		}
	} catch (Exception & e) {
		err << "Cannot restart command <" << commandID << "> : " << e.printStack() << eol;
	}
}

double Sender::getCommandDuration(unsigned long databaseID, unsigned long commandID) {
	try {
		return Scheduler::getInstance()->requestDuration(databaseID, commandID);
	} catch (Exception& e) {
		throw link(DurationRequestException(databaseID, commandID), e);
	}	
}

void Sender::waitForTermination(unsigned long databaseID, unsigned long commandID) {

	SenderCommandTerminatedSubscriber* subscriber = new SenderCommandTerminatedSubscriber(databaseID, commandID);
	DataNotificationRouter::getInstance()->addCommandTerminatedSubscriber(commandID, subscriber);

	// testing command state
	if (DataAccess::getInstance()->getCommandState(databaseID, commandID) == common::ACTIVE) {
		subscriber->waitForTermination();
	}

	DataNotificationRouter::getInstance()->removeCommandTerminatedSubscriber(commandID, subscriber);
	delete subscriber;
}

void Sender::stopCommandAndWaitForTermination(unsigned long databaseID, unsigned long commandID, bool atEnd) {

	SenderCommandStateChangeSubscriber* subscriber = new SenderCommandStateChangeSubscriber(databaseID, commandID);
	DataNotificationRouter::getInstance()->addCommandStateChangeSubscriber(commandID, subscriber);

	// testing command state
	bool waitForTermination = (DataAccess::getInstance()->getCommandState(databaseID, commandID) == common::ACTIVE);

	int requestID = 0;

	string stopCommandName = "stop";
	if (atEnd) {
		stopCommandName = "stopatend";
	}

	try {

		if (DataAccess::getInstance()->isCommandFromIcsServant(databaseID, commandID)) {
			// Special code for ics commands
			unsigned long stopCommandID = DataAccess::getInstance()->getSisterCommandID(databaseID, commandID, stopCommandName);
			if (stopCommandID != 0) {
				// starting the associated stop command
				newCommand(databaseID, stopCommandID, false);
			}

		} else {
			Scheduler::getInstance()->requestStopCommand(databaseID, commandID, requestID);
		}

		// now wait for the termination
		if (waitForTermination) {
			subscriber->waitForTermination();
		}

		DataNotificationRouter::getInstance()->removeCommandStateChangeSubscriber(commandID, subscriber);
		delete subscriber;

	} catch (Exception & e) {
		err << "Cannot stop command : " << e.printStack() << eol;
	}
}

void Sender::timeout(unsigned long databaseID, unsigned long commandID, unsigned long requestID, TimeCondition* condition, int timeoutMs) {

	bool notified = condition->wait(timeoutMs);
	if (!notified) {
		out << "Timeout in execution of " << DataAccess::getInstance()->what(databaseID, commandID) << eol;
		DataNotificationRouter::getInstance()->commandTerminated(databaseID, commandID, requestID);
	}
}

void Sender::stopSetter(unsigned long databaseID, unsigned long propertyID) {
	
	int requestID = 0;
	Scheduler::getInstance()->requestStopSetter(databaseID, propertyID, requestID);
}

void Sender::pauseSetter(unsigned long databaseID, unsigned long propertyID) {
	
	int requestID = 0;
	Scheduler::getInstance()->requestPauseSetter(databaseID, propertyID, requestID);
}

void Sender::restartSetter(unsigned long databaseID, unsigned long propertyID) {
	
	int requestID = 0;
	Scheduler::getInstance()->requestRestartSetter(databaseID, propertyID, requestID);
}

void Sender::initDriversAndProtocols() {
	
	int requestID = 0;
	Scheduler::getInstance()->initDriversAndProtocols(requestID);
}

}
