/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_SENDER_H_
#define CORE_SENDER_H_

#include "Scheduler.h"
#include <common/base/Exception.h>

namespace common {
	
class TimeCondition;
 
}

namespace core {

class CommandTerminatedSubscriber;

/**
 * Adapter class to provide an interface to the Scheduler.
 */
class Sender {

public:
	class DurationRequestException : public common::Exception {
	
	public:
		/**
		 * Constructor.
		 */
		DurationRequestException(unsigned long databaseID, unsigned long commandID);
	};
	
	/**
	 * Posts the new command request with ID commandID.
	 */
	static RequestReply newCommand(unsigned long databaseID, unsigned long commandID, bool skipIfWait, bool blocking = false, long timeoutMs = -1);
	
	/**
	 * Posts the new command request with ID commandID.
	 */
	static RequestReply newCommandFromCommandZone(unsigned long databaseID, unsigned long commandID, CommandTerminatedSubscriber* subscriber);
	
	/**
	 * Posts the stop command request with ID commandID.
	 */
	static void stopCommand(unsigned long databaseID, unsigned long commandID, bool atEnd, bool blocking = false);
	
	/**
	 * Posts the pause command request with ID commandID.
	 */
	static void pauseCommand(unsigned long databaseID, unsigned long commandID);
	
	/**
	 * Posts the restart command request with ID commandID.
	 */
	static void restartCommand(unsigned long databaseID, unsigned long commandID);
	
	/**
	 * Posts and waits for return value for the command duration for ID commandID.
	 */
	static double getCommandDuration(unsigned long databaseID, unsigned long commandID);
	
	/**
	 * Waits for the termination of the command with ID commandID.
	 */
	static void waitForTermination(unsigned long databaseID, unsigned long commandID);

	/**
	 * Posts the stop command request with ID commandID.
	 */
	static void stopCommandAndWaitForTermination(unsigned long databaseID, unsigned long commandID, bool atEnd = false);

	/**
	 * Sets a typed value.
	 */
	template<typename ValueType>
	static bool setValue(unsigned long databaseID, unsigned long propertyID, const ValueType& value, bool blocking = false);

	/**
	 * Sets a typed value. If the set value fails, the old value is reset.
	 */
	template<typename ValueType>
	static bool setValueOrReset(unsigned long databaseID, unsigned long propertyID, const ValueType& value, bool blocking = false);

	/**
	 * Posts the stop setter request with ID propertyID.
	 */
	static void stopSetter(unsigned long databaseID, unsigned long propertyID);
	
	/**
	 * Posts the pause setter request with ID propertyID.
	 */
	static void pauseSetter(unsigned long databaseID, unsigned long propertyID);
	
	/**
	 * Posts the restart setter request with ID propertyID.
	 */
	static void restartSetter(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Initialises the drivers and the protocols.
	 */
	static void initDriversAndProtocols();

private:
	static void timeout(unsigned long databaseID, unsigned long commandID, unsigned long requestID, common::TimeCondition* condition, int timeoutMs);
};

}

#include "Sender.inl"

#endif /*CORE_SENDER_H_*/
