/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "SenderPropertySetterTerminatedSubscriber.h"
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/dataaccess/DataAccess.h>
#include <core/base/property/DataAccessHelper.h>
#include <common/output/OutputStream.h>

namespace core {

template<typename ValueType>
bool Sender::setValue(unsigned long databaseID, unsigned long propertyID, const ValueType& value, bool blocking) {

	SenderPropertySetterTerminatedSubscriber* subscriber = 0;
	
	if (blocking) {
		subscriber = new SenderPropertySetterTerminatedSubscriber(databaseID, propertyID);
		DataNotificationRouter::getInstance()->addPropertySetterTerminatedSubscriber(propertyID, subscriber);
	}

	int requestID = 0;
	RequestReply reply = Scheduler::getInstance()->template requestSetValue<ValueType>(databaseID, propertyID, value, subscriber, requestID);

	bool success = true;
	if (reply.getValue() == RequestReply::REJECT) {
		common::err << "Setter for " << DataAccess::getInstance()->what(databaseID, propertyID) << " rejected : " << reply.getReason() << common::eol;
		success = false;
				
	} else if (reply.getValue() == RequestReply::FAIL) {
		common::err << "Setter for " << DataAccess::getInstance()->what(databaseID, propertyID) << " ID <" << propertyID << "> failed : " << reply.getReason() << common::eol;
		success = false;
	}
	
	// waiting for the end of the command if subscriber
	if (blocking) {
		if (success) {
			success = subscriber->waitForTermination();
		}
		DataNotificationRouter::getInstance()->removePropertySetterTerminatedSubscriber(propertyID, subscriber);
		delete subscriber;
	}

	return success;
}

template<typename ValueType>
bool Sender::setValueOrReset(unsigned long databaseID, unsigned long propertyID, const ValueType& value, bool blocking) {

	// get the current value
	ValueType currentValue = DataAccessHelper<ValueType>::getValue(databaseID, propertyID);

	// try to set the value
	bool success = setValue(databaseID, propertyID, value, blocking);

	// test the success
	if (!success) {
		// in case of failure, the property can be set (e.g. an updater throws an exception)
		// the property must be reset
		// if an exception is thrown in a refresher, this call is useless
		Sender::setValue(databaseID, propertyID, currentValue, true);
	}

	return success;
}

}
