/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "SessionManager.h"
#include <common/output/OutputStream.h>
#include <common/base/ServerProperties.h>
#include <core/transportlayer/MultiThreadCommunication.h>
#include <core/transportlayer/dataprovider/StateAccessor.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <core/base/Hostname.h>
#include "Utilities/Simulated.h"
#include <iostream>
#include <chrono>

using namespace std;
using namespace common;

namespace core {

const std::string SessionManager::MULTIPLE_CLIENTS_PROPERTY = "multipleClients";
const std::string SessionManager::MAIN_TYPE = "main";
const std::string SessionManager::REMOTE_TYPE = "remote";
const std::string SessionManager::VIRTUAL_TYPE = "virtual";
const std::string SessionManager::TABLET_TYPE = "tablet";
const std::string SessionManager::TOKEN_TIMEOUT_PROPERTY = "tokenTimeout";
const std::string SessionManager::CLIENT_SERVER_TIMEOUT_PROPERTY = "connectServerTimeout";

SessionManager* SessionManager::m_instance = nullptr;

std::string toString(common::CriticalState state) {

	if (state == CriticalState::DISABLED) {
		return "DISABLED";
	}
	else if (state == CriticalState::ENABLED) {
		return "ENABLED";
	}
	return "?";
}

std::string toString(common::RemoteControlState state) {

	if (state == RemoteControlState::DISABLED) {
		return "DISABLED";
	}
	else if (state == RemoteControlState::ENABLED) {
		return "ENABLED";
	}
	return "?";
}

std::string toString(common::TokenState state) {

	if (state == TokenState::FREE) {
		return "FREE";
	}
	else if (state == TokenState::OWNED) {
		return "OWNED";
	}
	return "?";
}

RequesterApplication::RequesterApplication() : id(0) {}

RequesterApplication::RequesterApplication(RequesterApplication && application)
	: serverAndApp(std::move(application.serverAndApp)), id(application.id), user(application.user), clientType(application.clientType) {
}

void RequesterApplication::reset() {
}

RequesterApplication & RequesterApplication::operator= (RequesterApplication && application) {
	serverAndApp = std::move(application.serverAndApp);
	id = application.id;
	user = application.user;
	clientType = application.clientType;

	return *this;
}

SessionManager::SessionManager() :
	m_notificationsEnabled(true), m_tokenTimeout(300), m_clientServerTimeout(10000), m_running(false), m_multipleClients(0), m_currentId(0), m_tokenClientID(0) {

	m_stateAccessor = MultiThreadCommunication::getInstance()->getStateAccessor();
}

SessionManager::~SessionManager() {

	// Join the thread if started.
	if (m_instance != 0 && m_checkThread.get() != nullptr) {
		m_running.store(false);
		m_checkThread->join();
	}
}

SessionManager* SessionManager::getInstance() {
	return m_instance;
}

void SessionManager::initInstance() {

	if (m_instance == nullptr) {
		m_instance = new SessionManager();
		m_instance->init();
	}
}

void SessionManager::resetInstance() {
	
	delete m_instance;
	m_instance = nullptr;
}

void SessionManager::init() {

	// Get the multiple clients configuration.
	try {
		m_multipleClients = ServerProperties::getInstance()->get<int>(MULTIPLE_CLIENTS_PROPERTY);

	}
	catch (...) {
		m_multipleClients = 1;
		out << "Setting default value 1 to property 'multipleClients'" << eol;
	}

	try {
		m_tokenTimeout = ServerProperties::getInstance()->get<int>(TOKEN_TIMEOUT_PROPERTY);

	}
	catch (...) {
		out << "Setting default value 300s to property 'tokenTimeout'" << eol;
	}

	try {
		m_clientServerTimeout = ServerProperties::getInstance()->get<int>(CLIENT_SERVER_TIMEOUT_PROPERTY);

	}
	catch (...) {
		out << "Setting default value 1000ms to property 'clientServerTimeout'" << eol;
	}

	// Test environment variable.
	string notificationsString = ServerProperties::getInstance()->getNomadNotifications();
	if ((notificationsString == "0") || (notificationsString == "-1")) {
		m_notificationsEnabled = false;
	}

	m_running.store(true);

	// Start the thread.
	m_checkThread.reset(new std::thread([this] {

		while (m_running.load()) {

			// Sleep for 1s.
			this_thread::sleep_for(chrono::seconds(1));

			// Get the duration since last activity.
			int duration = chrono::duration_cast<chrono::seconds>(chrono::system_clock::now() - lastTokenActivityDate()).count();

			// Get the mutex.
			{
				std::scoped_lock lock(m_mutex);

				// Set the token free if the timeout occurs.
				// Do not change the owner because it still owns the token.
				if (m_tokenState == TokenState::OWNED && duration > m_tokenTimeout) {
					m_tokenState = TokenState::FREE;

					// Notify the token state.
					notifyTokenState();

					LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + m_tokenUser + " on client " + hostname(m_tokenOwner) + " releases the token after timeout"));
				}
			}

			// Check the applications.
			checkRequesterApplications();
		}
	}));
}

RequesterApplication SessionManager::getRequesterApplication(cameo::coms::basic::Request * request, const std::string& clientType, const std::string& user) {

	RequesterApplication application;

	// The application has terminated and can be replaced.
	// Use the proxy depending on the client type.
	int options = cameo::option::USE_PROXY;
	if (clientType == TABLET_TYPE) {
		options = 0;
	}

	// Notice that the connection can fail, in that case the app is null.
	application.serverAndApp = request->connectToRequester(options, m_clientServerTimeout);

	// Store the user. Used for remote applications.
	application.user = user;

	// Store the client type.
	application.clientType = clientType;

	// Return the application.
	return application;
}

void SessionManager::checkRequesterApplications() {

	// Get the mutex.
	std::scoped_lock lock(m_mutex);

	// Get the iterator.
	vector<RequesterApplication>::iterator a = m_applications.begin();

	// Iterate over the applications.
	while (a != m_applications.end()) {

		cameo::state::Value state = cameo::state::NIL;

		// In case of timeout, getActualState() throws an exception.
		bool timeout = false;
		try {
			// Check the state if it is terminal.
			state = a->serverAndApp->getApp().getActualState();
		}
		catch (...) {
			timeout = true;
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + a->user + " on client " + a->serverAndApp->getServer().getEndpoint().getAddress() + " is unreachable"));
		}

		if (state == cameo::state::NIL) {

			// Get the last state: should be a terminal state.
			//cout << "Client " << a->instance->getId() << " has finished with last state " << cameo::application::toString(a->instance->getLastState()) << endl;

			// Release the token if necessary.
			if (a->serverAndApp->getServer().getEndpoint().toString() == m_tokenOwner) {
				// Change the state.
				m_tokenState = TokenState::FREE;
				m_tokenClientID = 0;
				m_tokenOwner = "";
				m_tokenUser = "";

				// Notify the token state.
				notifyTokenState();

				LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + a->user + " on client " + a->serverAndApp->getServer().getEndpoint().getAddress() + " released the token after logging out"));
			}
			else {
				LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + a->user + " on client " + a->serverAndApp->getServer().getEndpoint().getAddress() + " logged out"));
			}

			string message = "User " + a->user + " with client id " + to_string(a->id) + " is disconnected";
			if (timeout) {
				message += " after timeout";
			}

			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, message));

			// Erase the client id.
			uint32 clientId = a->id;
			m_clientTypes.erase(clientId);
			m_clientDisconnections[clientId].timeout = timeout;
			m_clientDisconnections[clientId].notified = false;

			// Delete the instance and server objects.
			// The cameo Server is deleted i.e. the zeromq context is deleted.
			// In case of timeout in Instance::getActualState(), the linger period is used not to block the zeromq context indefinitely.
			a->reset();

			// Erase from the vector.
			a = m_applications.erase(a);

			// Notify.
			notifyConnectedUsersChanged();
		}
		else {
			// Continue the iteration.
			++a;
		}
	}
}

void SessionManager::disconnectRemoteUsers() {

	// Get the mutex.
	std::scoped_lock lock(m_mutex);

	// Get the iterator.
	vector<RequesterApplication>::iterator a = m_applications.begin();

	// Iterate over the applications.
	// Delete all the remote clients.
	while (a != m_applications.end()) {

		if (a->clientType == REMOTE_TYPE) {

			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + a->user + " on client " + a->serverAndApp->getServer().getEndpoint().getAddress() + " logged out after proposal change"));

			// Erase the client id.
			uint32 clientId = a->id;
			m_clientTypes.erase(clientId);
			m_clientDisconnections[clientId].timeout = false;
			m_clientDisconnections[clientId].notified = false;

			// Delete the instance and server objects.
			// The cameo Server is deleted i.e. the zeromq context is deleted.
			// In case of timeout in Instance::getActualState(), the linger period is used not to block the zeromq context indefinitely.
			a->reset();

			// Erase from the vector.
			a = m_applications.erase(a);

			// Notify.
			notifyConnectedUsersChanged();
		}
		else {
			// Continue the iteration.
			++a;
		}
	}
}

int SessionManager::nextId() {
	m_currentId++;
	return m_currentId;
}

std::string SessionManager::getClientType(uint32 clientID) {

	try {
		return m_clientTypes[clientID];
	}
	catch (...) {
		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + to_string(clientID) + " has no entry"));
	}

	return "";
}

std::chrono::system_clock::time_point SessionManager::lastTokenActivityDate() {

	// Get the mutex as it is a shared resource.
	std::scoped_lock lock(m_mutex);

	return m_lastTokenActivityDate;
}

int SessionManager::login(cameo::coms::basic::Request * request, const std::string& clientType, const std::string& user) {

	// Check the requester applications.
	checkRequesterApplications();

	// Get the mutex.
	std::scoped_lock lock(m_mutex);

	try {
		// Get the application.
		RequesterApplication application = getRequesterApplication(request, clientType, user);

		// Check the existence of the application.
		if (!application.serverAndApp || !application.serverAndApp->hasApp()) {
			// The connection was not possible.
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(request->getRequesterEndpoint()) + " cannot be reached while logging"));

			return CONNECTION_ERROR;
		}

		if ((Simulated::activated) || (Simulated::for3d)) {
			// Allow all clients
		}
		else if (m_multipleClients == SINGLE_CLIENT && !m_applications.empty()) {
			// Only one client accepted.
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " refused from client " + hostname(request->getRequesterEndpoint()) + ": multiple clients per endpoint not allowed"));

			return MULTIPLE_CLIENTS_PER_ENDPOINT_NOT_ALLOWED;
		}
		else if (m_multipleClients == SINGLE_CLIENT_PER_ENDPOINT) {

			// Get the requester application endpoint.
			string applicationEndpoint = application.serverAndApp->getServer().getEndpoint().toString();

			// Only clients with different endpoints accepted.
			for (vector<RequesterApplication>::iterator a = m_applications.begin(); a != m_applications.end(); ++a) {

				// Check if user is not already connected from another computer.
				if (user == a->user) {
					LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " refused from client " + hostname(request->getRequesterEndpoint()) + ": user already connected from " + hostname(a->serverAndApp->getServer().getEndpoint().toString())));

					return MULTIPLE_CONNECTIONS_PER_USER_NOT_ALLOWED;
				}

				// Check applications that have the same endpoint.
				if (applicationEndpoint == a->serverAndApp->getServer().getEndpoint().toString()) {

					// The application is not accepted.
					LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " refused from client " + hostname(request->getRequesterEndpoint()) + ": user already connected from the endpoint"));

					return MULTIPLE_CLIENTS_PER_ENDPOINT_NOT_ALLOWED;
				}
			}
		}
		else {
			// Every application is allowed.
		}

		// Get the next id.
		uint32 clientId = nextId();

		// Assign the id.
		application.id = clientId;

		// Get the requester application endpoint.
		string applicationAddress = application.serverAndApp->getServer().getEndpoint().getAddress();

		// Add the app in the list of client apps.
		m_applications.push_back(std::move(application));

		// Update the client types.
		m_clientTypes[clientId] = clientType;

		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " with client id " + to_string(clientId) + " logged from client " + applicationAddress));

		// Notify.
		notifyConnectedUsersChanged();

		// Return the id.
		return clientId;
	}
	catch (const cameo::ConnectionTimeout& e) {
		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(request->getRequesterEndpoint()) + " is unreachable while logging"));
	}

	return CONNECTION_ERROR;
}

void SessionManager::loginRefused(cameo::coms::basic::Request * request, const std::string& reason, const std::string& user) {

	LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " refused from client " + hostname(request->getRequesterEndpoint()) + ": " + reason));
}

void SessionManager::enableRemoteControl(cameo::coms::basic::Request * request, uint32 clientID, const std::string& user) {

	// Get the mutex as it is a shared resource.
	std::scoped_lock lock(m_mutex);

	// Get the client type.
	string clientType = getClientType(clientID);

	// There is no check of the client type because the rule is on the role of the user known by the client.
	// Get the token in any case.
	m_remoteControlState = RemoteControlState::ENABLED;
	m_tokenState = TokenState::OWNED;
	m_tokenClientID = clientID;
	m_tokenOwner = request->getRequesterEndpoint();
	m_tokenUser = user;

	// Notify the token state.
	notifyTokenState();
	updateTokenActivity();

	LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " on client " + hostname(request->getRequesterEndpoint()) + " enables remote control and has the token"));
}

void SessionManager::disableRemoteControl(cameo::coms::basic::Request * request, uint32 clientID, const std::string& user) {

	// Get the mutex as it is a shared resource.
	std::scoped_lock lock(m_mutex);

	// Get the client type.
	string clientType = getClientType(clientID);

	// There is no check of the client type because the rule is on the role of the user known by the client.
	// Get the token in any case.
	m_remoteControlState = RemoteControlState::DISABLED;
	m_tokenState = TokenState::OWNED;
	m_tokenClientID = clientID;
	m_tokenOwner = request->getRequesterEndpoint();
	m_tokenUser = user;

	// Notify the token state.
	notifyTokenState();
	updateTokenActivity();

	LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " on client " + hostname(request->getRequesterEndpoint()) + " disables remote control and has the token"));
}

void SessionManager::setCriticalState(bool enabled) {

	// Get the mutex as it is a shared resource.
	std::scoped_lock lock(m_mutex);

	// Set the state.
	m_criticalState = (enabled ? common::CriticalState::ENABLED : common::CriticalState::DISABLED);

	// Notify the token state but do not update the token activity.
	notifyTokenState();

	LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Critical state " + toString(m_criticalState)));
}

bool SessionManager::authorizeControl(uint32 clientID) {

	// Get the mutex as it is a shared resource.
	std::scoped_lock lock(m_mutex);

	// Check critical state.
	if (m_criticalState == common::CriticalState::ENABLED) {
		return false;
	}

	// Get the client type.
	string clientType = getClientType(clientID);

	// In case of tablet, the control is authorized if remote control is disabled.
	if (((clientType == TABLET_TYPE) && m_remoteControlState == (common::RemoteControlState::DISABLED)) || (Simulated::activated) || (Simulated::for3d)) {
		return true;
	}

	return (m_tokenClientID == clientID);
}

common::TokenAuthorizationStatus SessionManager::checkClientTokenAuthorization(uint32 clientID) {

	// Get the mutex as it is a shared resource.
	{
		std::scoped_lock lock(m_mutex);

		// In case the client type is empty, it means that the client has been disconnected.
		if ((getClientType(clientID) == "") && (!Simulated::activated) && (!Simulated::for3d)) {
			string message = "User with client id " + to_string(clientID) + " is sending a request but has been disconnected";
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, message));

			// Alarm only once when a disconnected user is coming back.
			if (m_clientDisconnections.find(clientID) != m_clientDisconnections.end()) {

				// Send a notification it it is not already done.
				if (!m_clientDisconnections[clientID].notified) {

					string subject = "User disconnected";

					if (m_clientDisconnections[clientID].timeout) {
						subject += " after timeout";
					}

					// Send a mail.
					sendMailAlarm(subject, message);

					// Memorize the notification.
					m_clientDisconnections[clientID].notified = true;
				}
			}

			return common::TokenAuthorizationStatus::CLIENT_DISCONNECTED;
		}
	}

	// Check authorization.
	if (authorizeControl(clientID)) {
		return common::TokenAuthorizationStatus::OK;
	}
	else {
		return common::TokenAuthorizationStatus::TOKEN_NOT_OWNED;
	}
}

void SessionManager::requestToken(cameo::coms::basic::Request * request, uint32 clientID, const std::string& user) {

	// Get the mutex as it is a shared resource.
	std::scoped_lock lock(m_mutex);

	// Get the client type.
	string clientType = getClientType(clientID);

	// Test the client type.
	if ((clientType == MAIN_TYPE) && (Simulated::visa == false)) {

		// Get the token in any case.
		m_tokenState = TokenState::OWNED;
		m_tokenClientID = clientID;
		m_tokenOwner = request->getRequesterEndpoint();
		m_tokenUser = user;

		// Notify the token state.
		notifyTokenState();
		updateTokenActivity();

		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " on client " + hostname(request->getRequesterEndpoint()) + " has the token"));
	}
	else if (clientType == REMOTE_TYPE || clientType == VIRTUAL_TYPE) {
		// It is not possible to request the token when remote control is disabled.
		if (m_remoteControlState == RemoteControlState::DISABLED) {
			return;
		}

		// Get the token only if free.
		if (m_tokenState == TokenState::FREE) {

			// Change the state.
			m_tokenState = TokenState::OWNED;
			m_tokenClientID = clientID;
			m_tokenOwner = request->getRequesterEndpoint();
			m_tokenUser = user;

			// Notify the token state.
			notifyTokenState();
			updateTokenActivity();

			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " on client " + hostname(request->getRequesterEndpoint()) + " has the token"));
		}
	}
	else if ((clientType == MAIN_TYPE) && (Simulated::visa == true)) {
		// Get the token in any case.
		m_tokenState = TokenState::OWNED;
		m_tokenClientID = clientID;
		m_tokenOwner = request->getRequesterEndpoint();
		m_tokenUser = user;

		// Notify the token state.
		notifyTokenState();
		updateTokenActivity();
	}
	else {
		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " on client " + hostname(request->getRequesterEndpoint()) + " tries to request token but fails, client type is " + clientType + ", client ID is " + to_string(clientID)));
	}
}

void SessionManager::releaseToken(cameo::coms::basic::Request * request, uint32 clientID, const std::string& user) {

	// Get the mutex as it is a shared resource.
	std::scoped_lock lock(m_mutex);

	// Verify that this is the same client endpoint.
	if (request->getRequesterEndpoint() == m_tokenOwner) {

		// Change the state.
		m_tokenState = TokenState::FREE;
		m_tokenClientID = 0;
		m_tokenOwner = "";
		m_tokenUser = "";

		// Notify the token state.
		notifyTokenState();

		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + user + " on client " + hostname(request->getRequesterEndpoint()) + " released the token"));
	}
}

void SessionManager::notifyTokenState() {

	// Do not send the token event in case of a simulated server that has side-effects on the client.
	// Indeed the client cannot distinguish a token event from the real server or a simulated server.
	if ((!Simulated::activated) && (!Simulated::for3d)) {
		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Notify token state " + toString(m_remoteControlState) + " " + toString(m_tokenState) + " " + m_tokenOwner + " " + m_tokenUser));

		// Notifies the data change publisher that will send an event.
		m_stateAccessor->notifyTokenStateChanged(m_criticalState, m_remoteControlState, m_tokenState, m_tokenClientID, m_tokenOwner, m_tokenUser);
	}
	else if (Simulated::visa == true) {
		// Notifies the data change publisher that will send an event.
		m_stateAccessor->notifyTokenStateChanged(m_criticalState, m_remoteControlState, m_tokenState, m_tokenClientID, m_tokenOwner, m_tokenUser);
	}
}

void SessionManager::notifyTokenActivity() {

	// Get the mutex as it is a shared resource.
	std::scoped_lock lock(m_mutex);

	// Change the token only if free.
	if (m_tokenState == TokenState::FREE) {

		// Keep the current owner by changing only the state.
		m_tokenState = TokenState::OWNED;

		// Notify the token state.
		notifyTokenState();

		LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "User " + m_tokenUser + " on client " + hostname(m_tokenOwner) + " has the token by activity"));
	}

	// Update the activity in any case.
	updateTokenActivity();
}


void SessionManager::updateTokenActivity() {

	// Change the last activity date.
	m_lastTokenActivityDate = chrono::system_clock::now();
}

void SessionManager::sendChatMessage(cameo::coms::basic::Request * request, uint32 clientID, const std::string& sender, const std::string& message, const std::string& fileName) {

	// Get the current date.
	const auto now = chrono::system_clock::now();
	int64 seconds = chrono::duration_cast<chrono::seconds>(now.time_since_epoch()).count();

	// Notify the other clients.
	m_stateAccessor->notifyChatMessage(clientID, seconds, sender, message, fileName);
}

void SessionManager::sendCrashMail(cameo::coms::basic::Request * request, const std::string& content) {

	string mailCommand("echo \"");
	mailCommand += content + "\" | mail -r nomadmail@ill.fr -s \"";
	mailCommand += "[" + ServerProperties::getInstance()->getHostname() + "][remote client]\" ";
	string recipients = "nomad_l@ill.fr";
	mailCommand += recipients;

	cout << "Crash mail command = " << mailCommand << endl;

	if (m_notificationsEnabled) {
		::system(mailCommand.c_str());
	}
}

std::vector<SessionManager::ConnectedUser> SessionManager::getConnectedUsers() {

	// Get the mutex.
	std::scoped_lock lock(m_mutex);

	vector<ConnectedUser> result;
	result.reserve(m_applications.size());

	// Get the iterator.
	vector<RequesterApplication>::iterator a = m_applications.begin();

	// Iterate over the applications.
	while (a != m_applications.end()) {

		ConnectedUser user;

		user.id = a->id;
		user.user = a->user;
		user.clientType = m_clientTypes[a->id];
		user.endpoint = a->serverAndApp->getServer().getEndpoint().toString();

		result.push_back(user);

		++a;
	}

	return result;
}

void SessionManager::notifyConnectedUsersChanged() {

	// Do not send the token event in case of a simulated server that has side-effects on the client.
	// Indeed the client cannot distinguish a token event from the real server or a simulated server.
	if ((!Simulated::activated) && (!Simulated::activated)) {
		m_stateAccessor->notifyConnectedUsersChanged();
	}
}

void SessionManager::print() {

	std::scoped_lock lock(m_mutex);

	cout << "Token state " << toString(m_remoteControlState) << " " << toString(m_tokenState) << " " << m_tokenOwner << " " << m_tokenUser << endl;
}

void SessionManager::sendMailAlarm(const std::string& subject, const std::string& content) {

	// test environment variable
	string notificationsString = ServerProperties::getInstance()->getNomadNotifications();
	if ((notificationsString == "0") || (notificationsString == "-1")) {
		return;
	}

	// Temporary contacts.
	string contacts = "legoc@ill.fr locatelli@ill.fr";

	string subjectText = "[" + ServerProperties::getInstance()->getNomadUser() + "][" + ServerProperties::getInstance()->getHostname() + "][warning] " + subject;

	string mailCommand("echo \"");
	mailCommand += content + "\" | mail -r nomadmail@ill.fr -s \"";
	mailCommand += subjectText + "\" " + contacts;

	::system(mailCommand.c_str());

	//cout << "Mail command = " << mailCommand << endl;
}

}
