/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_SESSIONMANAGER_H_
#define CORE_SESSIONMANAGER_H_

#include <common/base/CommonTypes.h>
#include <common/base/Exception.h>
#include <common/base/TimeCondition.h>
#include <cameo/api/cameo.h>
#include <unordered_map>
#include <unordered_set>
#include <string>
#include <vector>
#include <memory>
#include <thread>
#include <atomic>

namespace core {

class StateAccessor;

/**
 * Class helping to manage the server/instance pair.
 */
struct RequesterApplication {

	std::unique_ptr<cameo::ServerAndApp> serverAndApp;
	uint32 id;
	std::string user;
	std::string clientType;

	/**
	 * Constructor.
	 */
	RequesterApplication();

	/**
	 * Move constructor.
	 */
	RequesterApplication(RequesterApplication && application);

	/**
	 * Resets the object.
	 */
	void reset();

	/**
	 * Operator=.
	 */
	RequesterApplication & operator= (RequesterApplication && application);
};

/**
 * Class defining the singleton session manager.
 */
class SessionManager {

public:
	/**
	 * Returns the instance.
	 */	
	static SessionManager* getInstance();
	
	/**
	 * Initialises the instance.
	 */
	static void initInstance();

	/**
	 * resets the instance.
	 */
	static void resetInstance();

	/**
	 * Logins a session. Returns the id of the session if accepted.
	 * Otherwise MULTIPLE_CLIENTS_PER_ENDPOINT_NOT_ALLOWED or MULTIPLE_CONNECTION_PER_USER_NOT_ALLOWED.
	 */
	int login(cameo::coms::basic::Request * request, const std::string& clientType, const std::string& user = "?");

	/**
	 * Logs the refusal of login from the client.
	 */
	void loginRefused(cameo::coms::basic::Request * request, const std::string& reason, const std::string& user = "?");

	/**
	 * Enables the remote control depending on conditions.
	 */
	void enableRemoteControl(cameo::coms::basic::Request * request, uint32 clientID, const std::string& user);

	/**
	 * Disables the remote control depending on conditions.
	 */
	void disableRemoteControl(cameo::coms::basic::Request * request, uint32 clientID, const std::string& user);

	/**
	 * Sets the critical state, true for enabled, false for disabled.
	 */
	void setCriticalState(bool enabled);

	/**
	 * Returns true if the client with ID clientID can control the server.
	 * In case of tablet, the control is authorized if and only if remote control is disabled.
	 * For other client types, the control is authorized if clientID equals the clientID of the token owner.
	 */
	bool authorizeControl(uint32 clientID);

	/**
	 * Checks the client token authorization.
	 */
	common::TokenAuthorizationStatus checkClientTokenAuthorization(uint32 clientID);

	/**
	 * Requests the token.
	 */
	void requestToken(cameo::coms::basic::Request * request, uint32 clientID, const std::string& user);

	/**
	 * Releases the token.
	 */
	void releaseToken(cameo::coms::basic::Request * request, uint32 clientID, const std::string& user);

	/**
	 * Requests a synchronization so that a token event is sent.
	 */
	void notifyTokenState();

	/**
	 * Notifies an activity using the token. The timeout is reset.
	 */
	void notifyTokenActivity();

	/**
	 * Sends a chat message.
	 */
	void sendChatMessage(cameo::coms::basic::Request * request, uint32 clientID, const std::string& sender, const std::string& message, const std::string& fileName);

	/**
	 * Sends a mail of crash.
	 */
	void sendCrashMail(cameo::coms::basic::Request * request, const std::string& content);

	/**
	 * Connected user.
	 */
	struct ConnectedUser {
		uint32 id;
		std::string user;
		std::string clientType;
		std::string endpoint;
	};

	/**
	 * Gets the connected users.
	 */
	std::vector<ConnectedUser> getConnectedUsers();

	/**
	 * Notifies that the connected users have changed.
	 */
	void notifyConnectedUsersChanged();

	/**
	 * Removes all the remote clients.
	 */
	void disconnectRemoteUsers();

	/**
	 * Prints the current state of the session.
	 */
	void print();

	/**
	 * Constant results in login().
	 */
	static const int MULTIPLE_CLIENTS_PER_ENDPOINT_NOT_ALLOWED = -1;
	static const int MULTIPLE_CONNECTIONS_PER_USER_NOT_ALLOWED = -2;
	static const int CONNECTION_ERROR = -3;

private:
	SessionManager();
	~SessionManager();

	void init();

	RequesterApplication getRequesterApplication(cameo::coms::basic::Request * request, const std::string& clientType, const std::string& user);
	void checkRequesterApplications();
	int nextId();
	std::string getClientType(uint32 clientID);
	std::chrono::system_clock::time_point lastTokenActivityDate();
	void updateTokenActivity();
	void sendMailAlarm(const std::string& subject, const std::string& content);

	static SessionManager* m_instance;

	bool m_notificationsEnabled;
	int m_tokenTimeout;
	int m_clientServerTimeout;
	std::atomic_bool m_running;
	std::unique_ptr<std::thread> m_checkThread;

	int m_multipleClients;
	int m_currentId;
	std::vector<RequesterApplication> m_applications;
	std::unordered_map<uint32, std::string> m_clientTypes;

	struct DisconnectionInfo {
		bool timeout;
		bool notified;

		DisconnectionInfo() : timeout(false), notified(false) {}
	};

	std::unordered_map<uint32, DisconnectionInfo> m_clientDisconnections;

	common::CriticalState m_criticalState = common::CriticalState::DISABLED;
	common::RemoteControlState m_remoteControlState = common::RemoteControlState::DISABLED;
	common::TokenState m_tokenState = common::TokenState::FREE;
	uint32 m_tokenClientID;
	std::string m_tokenOwner;
	std::string m_tokenUser;

	std::recursive_mutex m_mutex;
	std::chrono::system_clock::time_point m_lastTokenActivityDate;

	StateAccessor* m_stateAccessor;

	static const std::string MULTIPLE_CLIENTS_PROPERTY;
	static const int SINGLE_CLIENT = 0;
	static const int SINGLE_CLIENT_PER_ENDPOINT = 1;
	static const int MULTIPLE_CLIENTS = 2;

	static const std::string MAIN_TYPE;
	static const std::string REMOTE_TYPE;
	static const std::string VIRTUAL_TYPE;
	static const std::string TABLET_TYPE;

	static const std::string TOKEN_TIMEOUT_PROPERTY;
	static const std::string CLIENT_SERVER_TIMEOUT_PROPERTY;
};

}

#endif
