/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */
#include "FluentdSurveyWriter.h"

#include "PropertySurveyElement.h"
#include "NumorSurveyElement.h"
#include "DataProvider/Property/AnyProperty.h"
#include <iostream>
#include <common/base/Date.h>
#include <cstdlib>
#include <dataprovider/database/Database.h>
#include <dataprovider/iconandfamily/ServantIconAndFamilyMaps.h>
#include <common/connection/RestJsonHttpSyncConnection.h>
#include <common/base/Exception.h>
#include <common/base/ServerProperties.h>
#include <common/base/Convert.h>
#include <boost/algorithm/string.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <fstream>
#include <unistd.h>

using namespace std;
using namespace common;
using namespace boost;

using namespace rapidjson;

namespace core {

const std::string FluentdSurveyWriter::VERSION_STR = "version";
const int32 FluentdSurveyWriter::VERSION_VALUE = 2;

const std::string FluentdSurveyWriter::CTYPE = "ctype";
const std::string FluentdSurveyWriter::CNAME = "cname";
const std::string FluentdSurveyWriter::PNAME = "pname";
const std::string FluentdSurveyWriter::ALIAS = "alias";
const std::string FluentdSurveyWriter::UNIT = "unit";
const std::string FluentdSurveyWriter::DAY = "day";
const std::string FluentdSurveyWriter::PROPOSALID = "propid";
const std::string FluentdSurveyWriter::NVALUES = "nvalues";
const std::string FluentdSurveyWriter::VALUES = "values";
const std::string FluentdSurveyWriter::VAL = "val";
const std::string FluentdSurveyWriter::TIMESTAMP = "timestamp";

const std::string FluentdSurveyWriter::NNUMORS = "nnumors";
const std::string FluentdSurveyWriter::NUMORS = "numors";
const std::string FluentdSurveyWriter::NUMOR = "numor";
const std::string FluentdSurveyWriter::START = "start";
const std::string FluentdSurveyWriter::END = "end";

using namespace common;

FluentdSurveyWriter::SurveyException::SurveyException(const std::string& message) :
		common::Exception("dataprovider::FluentdSurveyWriter::SurveyException", message) {
}

FluentdSurveyWriter::FluentdSurveyWriter() :
		m_propertyIdOfProposalId(0), m_initDone(false), m_showMessages(false) {

	m_nomadSendData = ServerProperties::getInstance()->getNomadSenddata();
	to_lower(m_nomadSendData);
}

FluentdSurveyWriter::~FluentdSurveyWriter() {

}

void FluentdSurveyWriter::write(SurveyElement* element) {

	if (m_nomadSendData == "-1") {
		return;
	}

	if (m_initDone == false) {
		try {
			init();
		} catch (Exception&) {
			// cerr << "In Fluentd: Servant Name not correct" << endl;
			return;
		}
	}

	m_properties.clear();

	rapidjson:: StringBuffer s;
	PrettyWriter<StringBuffer> writer(s);

//	Start new message
	writer.StartObject();

	try {

		string typeforward;
		if (dynamic_cast<PropertySurveyElement*>(element) != 0) {
			createValuesMessageUsingRapidJson(dynamic_cast<PropertySurveyElement*>(element), writer);
			typeforward = "histo";
		}
		else if (dynamic_cast<NumorSurveyElement*>(element) != 0) {
			createValuesMessageUsingRapidJson(dynamic_cast<NumorSurveyElement*>(element), writer);
			typeforward = "numor";
		}

		writer.EndObject();

		// only show messages in case it was activated
		if (m_showMessages) {
			cout << "Fluentd log message: " << s.GetString() << endl << endl;
		}

		// Trying to detect bad JSON messages (e.g. driver errors with bad characters) using rapidjson::Document.Parse(string).HasParseError()
		// does not provide errors.

		RestJsonHttpSyncConnection con;

		con.openUrl("localhost", "8888");

		//New environment variable
		if (m_nomadSendData == "0" || m_nomadSendData == "-2") {

			//if 0 send data to dev database
			con.setPath(string("/mongo.") + typeforward + ".DEV");

		} else if ((m_nomadSendData == "1")|| (m_nomadSendData == "2")) {
			//if 1 send data to prod database+
			con.setPath(string("/mongo.") + typeforward) ;// + instrumentCode);
		}

		con.setData(s.GetString());
		boost::property_tree::ptree response = con.send(false);

	} catch (const Exception& e) {
		// log to server
		cerr << e.printStack() << endl;
		slog << e.printStack() << eol;
	}
}

void FluentdSurveyWriter::showJSONMessages(bool value) {
	m_showMessages = value;
}

void FluentdSurveyWriter::init() {
	dataprovider::Database* database = dataprovider::Database::getInstance();
	unsigned long servantID = database->getServantID("Title");
	std::string controllerType = dataprovider::Database::getInstance()->getControllerType(servantID);
	dataprovider::Controller* experimentDataController = dataprovider::Database::getInstance()->getController(controllerType);

	//Save ID of all necessary properties:
	//It is missing: instrument ID, sample ID, Sample Name.
	m_propertyIdOfProposalId = experimentDataController->getPropertyID("proposalId");
	unsigned long readyId = experimentDataController->getPropertyID("ready");
	m_initDone = dataprovider::Database::getInstance()->getValue<bool>(readyId);

}

void FluentdSurveyWriter::createValuesMessageUsingRapidJson(PropertySurveyElement* element, rapidjson::PrettyWriter<rapidjson::StringBuffer> &writer) {

	writer.Key(VERSION_STR.c_str());
	writer.Int(VERSION_VALUE);

	writer.Key(CTYPE.c_str());
	writer.String(element->getServantType().c_str());

	writer.Key(CNAME.c_str());
	writer.String(element->getServantName().c_str());

	writer.Key(PNAME.c_str());
	writer.String(element->getPropertyName().c_str());

	writer.Key(ALIAS.c_str());
	writer.String(element->getAlias().c_str());

	writer.Key(UNIT.c_str());
	writer.String(element->getPropertyUnit().c_str());

	writer.Key(DAY.c_str());
	writer.Int64(element->getDate().getDateUs() / 1000); // In ms for fluentd, fluend convert it to date type for mongodb

	writer.Key(PROPOSALID.c_str());
	istringstream sbuf(dataprovider::Database::getInstance()->getValue<std::string>(m_propertyIdOfProposalId));
	int32 num;
	sbuf >> num;
	writer.Int(num);

	writer.Key(NVALUES.c_str());
	writer.Int(1);

	writer.Key(VALUES.c_str());
	writer.StartArray();

	writer.StartObject();
	writer.Key(TIMESTAMP.c_str());
	writer.Int((int32) (element->getDate().getTimeUs() / 1000000)); // Save seconds for time stamp, it's enough

	double value = 0.0;
	string ptype = element->getPropertyType();
	try {
		if (ptype == BOOLEAN_PROPERTY) {
			if ((element->getLastValue() == "1") || (element->getLastValue() == "true") || (element->getLastValue() == "TRUE")) {
				value = 1.0;
			}
			else {
				value = 0.0;
			}
		} else if (ptype == STRING_PROPERTY) {
			value = 0.0;
		} else {
			value = to<float64>(element->getLastValue());
		}
	} catch (Exception& e) {
		value = 0.0;
	}

	writer.Key(VAL.c_str());
	writer.Double(value);
	writer.EndObject();

	writer.EndArray();
}

void FluentdSurveyWriter::createValuesMessageUsingRapidJson(NumorSurveyElement* element, rapidjson::PrettyWriter<rapidjson::StringBuffer> &writer) {

	writer.Key(VERSION_STR.c_str());
	writer.Int(VERSION_VALUE);

	int32 time = (int32) (element->getDate().getTimeUs() / 1000000);

	writer.Key(DAY.c_str());
	writer.Int64(element->getDate().getDateUs() / 1000); // In ms for fluentd, fluend convert it to date type for mongodb

	writer.Key(PROPOSALID.c_str());
	istringstream sbuf(dataprovider::Database::getInstance()->getValue<std::string>(m_propertyIdOfProposalId));
	int32 num;
	sbuf >> num;
	writer.Int(num);

	writer.Key(NNUMORS.c_str());
	writer.Int(1);

	writer.Key(NUMORS.c_str());
	writer.StartArray();

	writer.StartObject();

	writer.Key(NUMOR.c_str());
	writer.Int((int32) element->getNumor());

	if (element->isStart()) {
		writer.Key(START.c_str());
		writer.Int(time);
		writer.Key(END.c_str());
		writer.Int(-1);
	}
	else {
		writer.Key(START.c_str());
		writer.Int(-1);
		writer.Key(END.c_str());
		writer.Int(time);
	}

	writer.EndObject();
	writer.EndArray();
}

}
