/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_PROPERTYSURVEYELEMENT_H_
#define CORE_PROPERTYSURVEYELEMENT_H_

#include "SurveyElement.h"
#include <common/base/Date.h>
#include <core/dataaccess/CommandStateChangeSubscriber.h>
#include <string>

namespace core {

/**
 * Class defining a property survey element.
 */
class PropertySurveyElement : public SurveyElement, public CommandStateChangeSubscriber {

public:
	/**
	 * Constructor
	 */
	PropertySurveyElement(unsigned long servantID, unsigned long propertyID, long index, const std::string& servantName, const std::string& propertyName, const std::string& alias);
	
	/**
	 * Destructor
	 */
	virtual ~PropertySurveyElement();

	/**
	 * Implements the CommandStateChangeSubscriber interface.
	 */
	virtual void commandStateChanged(unsigned long databaseID, unsigned long commandID, common::CommandState state);

	/**
	 * Returns the servant ID.
	 */
	unsigned long getServantID() const;

	/**
	 * Returns the servant name.
	 */
	std::string getServantName() const;
	
	/**
	 * Returns the servant type.
	 */
	std::string getServantType() const;

	/**
	 * Returns the property ID.
	 */
	unsigned long getPropertyID() const;

	/**
	 * Returns the property index.
	 */
	long getIndex() const;
	
	/**
	 * Returns the property name.
	 */
	std::string getPropertyName() const;
	
	/**
	 * Returns the alias.
	 */
	std::string getAlias() const;
	
	/**
	 * Returns the property type.
	 */
	std::string getPropertyType() const;
	
	/**
	 * Sets the active write time.
	 */
	void setActiveWriteTime(int seconds);

	/**
	 * Gets the active write time.
	 */
	int getActiveWriteTime() const;

	/**
	 * Returns the date of the last save.
	 */
	common::Date getLastDate() const;
	
	/**
	 * Sets the date of the last save.
	 */
	 void setLastDate(const common::Date& lastDate);
	
	/**
	 * Sets the value at the last save.
	 */
	void setLastValue(const std::string& lastValue);
	
	/**
	 * Gets the value at the last save.
	 */
	const std::string& getLastValue() const;

	/**
	 * Gets the value at the last save.
	 */
	const std::string& getPropertyUnit() const;

	/**
	 * Returns true if is active.
	 */
	bool isActive() const;

	/**
	 * Returns true if the element hasn't been written to the file yet.
	 */
	bool isFirst() const;

	/**
	 * Sets the flag to indicate if the survey element has run or not yet.
	 */
	void setFirst(bool isFirst);

private:
	unsigned long m_servantID;
	unsigned long m_propertyID;
	unsigned long m_startCommandID = 0;
	long m_index;
	std::string m_servantName;
	std::string m_propertyName;
	std::string m_alias;
	std::string m_servantType;
	std::string m_propertyType;
	int m_activeWriteTime = 1;
	common::Date m_lastDate;
	std::string m_lastValue;
	std::string m_propertyUnit;
	bool m_active = false;
	bool m_first = true;
};

}

#endif /* CORE_PROPERTYSURVEYELEMENT_H_ */
