/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Survey.h" 
#include "PropertySurveyElement.h"
#include "NumorSurveyElement.h"
#include "SurveyWriter.h"
#include <core/parser/survey/SurveyParser.h>
#include <core/dataaccess/DataAccess.h>
#include <core/transportlayer/dataprovider/SurveyAccessor.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <boost/lexical_cast.hpp>
#include <iostream>

using namespace std;
using namespace common;

namespace core {

Survey* Survey::m_instance = 0; 

const std::string Survey::DEFAULT_TIME = "default";
const std::string Survey::SURVEY_CYCLE_DURATIONMS = "surveyCycleDurationms";
const std::string Survey::ACTIVE_WRITE_TIMES = "surveyActiveWriteTimes";
const std::string Survey::INACTIVE_WRITE_TIMES = "surveyInactiveWriteTimes";
const std::string Survey::SURVEY_PATH = "surveyPath";
const std::string Survey::HISTOGRAM_CONFIG_NAME = "HistoFileConfig.txt";

Survey::Survey() :
	m_surveyAccessor(0),
	m_running(false),
	m_activeWriteTime(0),
	m_inactiveWriteTime(0) {

	m_fluentdwriter.showJSONMessages(false);
}

Survey::~Survey() {
	for (auto i = m_elements.begin(); i != m_elements.end(); i++) {
		delete (*i);
	}
}

Survey* Survey::getInstance() {
	return m_instance;
}

void Survey::initInstance(const std::string& configPath, SurveyAccessor * surveyAccessor) {

	if (m_instance == 0) {
		m_instance = new Survey();
		m_instance->init(configPath, surveyAccessor);
	}
}

void Survey::resetInstance() {
	delete m_instance;
	m_instance = 0;
}

void Survey::init(const std::string& configPath, SurveyAccessor * surveyAccessor) {

	try {
		m_activeWriteTime = ServerProperties::getInstance()->get<int>(ACTIVE_WRITE_TIMES);

	} catch (const Exception& e) {
		out << "Setting default value 10s to property 'surveyActiveWriteTimes'" << eol;
		m_activeWriteTime = 10;
	}

	try {
		m_inactiveWriteTime = ServerProperties::getInstance()->get<int>(INACTIVE_WRITE_TIMES);

	} catch (const Exception& e) {
		out << "Setting default value 60s to property 'surveyInactiveWriteTimes'" << eol;
		m_inactiveWriteTime = 60;
	}

	m_surveyAccessor = surveyAccessor;
}

void Survey::run() {
	
	m_running = true;
	m_thread.reset(new std::thread(&Survey::loop));

	// Test save.
	//saveConfig();
}

void Survey::stop() {
	
	if (m_instance != nullptr && m_thread != nullptr) {
	
		m_running = false;
		m_condition.notify();
		
		m_thread->join();
		m_thread.reset();
	}	
}

bool Survey::findElement(const std::string& servantName, const std::string& propertyName, int index) {

	// Iterate the elements.
	for (auto e = m_elements.begin(); e != m_elements.end(); e++) {

		if ((*e)->getServantName() == servantName and (*e)->getPropertyName() == propertyName and (*e)->getIndex() == index) {
			// Found.
			return true;
		}
	}

	return false;
}

bool Survey::addElement(const std::string& servantName, const std::string& propertyName, const std::optional<int>& activeWriteTime) {

	unsigned long servantID = ::core::DataAccess::getInstance()->getServantID(servantName);
	unsigned long propertyID = ::core::DataAccess::getInstance()->getPropertyID(0, servantID, propertyName);
	std::string controllerType = ::core::DataAccess::getInstance()->getControllerType(0, servantID);
	::core::DataAccess::PropertyInfo propertyInfo = ::core::DataAccess::getInstance()->getPropertyInfo(0, controllerType, propertyID);

	//trace << "Added histo " << servantName << " " << propertyName << " : " << servantID << " " << propertyID << " " << controllerType << " " << propertyInfo.aliasPropertyName << eol;

	PropertySurveyElement* element = new PropertySurveyElement(servantID, propertyID, -1, servantName, propertyName, propertyInfo.aliasPropertyName);

	// Set the active write time if exists.
	if (activeWriteTime.has_value()) {
		element->setActiveWriteTime(activeWriteTime.value());
	}
	else {
		element->setActiveWriteTime(10);
	}

	// Lock the elements.
	std::scoped_lock lock(m_mutex);

	// Check existence.
	if (findElement(servantName, propertyName)) {
		return false;
	}

	m_elements.push_back(element);
	saveConfig();

	return true;
}

bool Survey::addElement(const std::string& servantName, const std::string& propertyName, int index, const std::optional<int>& activeWriteTime) {

	unsigned long servantID = ::core::DataAccess::getInstance()->getServantID(servantName);
	unsigned long dynamicPropertyID = ::core::DataAccess::getInstance()->getPropertyID(0, servantID, propertyName);
	unsigned long propertyID = ::core::DataAccess::getInstance()->getDynamicPropertyIDAtIndex(0, dynamicPropertyID, index, false);
	std::string controllerType = ::core::DataAccess::getInstance()->getControllerType(0, servantID);
	::core::DataAccess::PropertyInfo propertyInfo = ::core::DataAccess::getInstance()->getPropertyInfo(0, controllerType, propertyID);

	string propertyNameWithIndex = propertyName;
	propertyNameWithIndex += ".";
	propertyNameWithIndex += boost::lexical_cast<string>(index);

	string propertyAliasWithIndex = propertyInfo.aliasPropertyName;

	// Complete with the index if the alias exists.
	if (propertyAliasWithIndex != "") {
		propertyAliasWithIndex += ".";
		propertyAliasWithIndex += boost::lexical_cast<string>(index);
	}

	//trace << "Added histo dynamic " << servantName << " " << propertyName << " " << index << " : " << servantID << " " << propertyID << " " << controllerType << " " << propertyInfo.aliasPropertyName << eol;

	PropertySurveyElement* element = new PropertySurveyElement(servantID, propertyID, index, servantName, propertyName, propertyInfo.aliasPropertyName);

	// Set the active write time if exists.
	if (activeWriteTime.has_value()) {
		element->setActiveWriteTime(activeWriteTime.value());
	}
	else {
		element->setActiveWriteTime(10);
	}

	// Lock the elements.
	std::scoped_lock lock(m_mutex);

	// Check existence.
	if (findElement(servantName, propertyName)) {
		return false;
	}

	m_elements.push_back(element);
	saveConfig();

	return true;
}

bool Survey::removeElement(const std::string& servantName, const std::string& propertyName, int index) {

	// Lock the elements.
	std::scoped_lock lock(m_mutex);

	// Iterate the elements.
	for (auto e = m_elements.begin(); e != m_elements.end(); e++) {

		if ((*e)->getServantName() == servantName and (*e)->getPropertyName() == propertyName and (*e)->getIndex() == index) {
			// Found.
			m_elements.erase(e);

			//trace << "Removed histo " << servantName << " " << propertyName << " " << index << eol;

			saveConfig();

			// Exit.
			return true;
		}
	}

	return false;
}

bool Survey::setActiveWriteTime(const std::string& servantName, const std::string& propertyName, int index, int seconds) {

	// Lock the elements.
	std::scoped_lock lock(m_mutex);

	// Iterate the elements.
	for (auto e = m_elements.begin(); e != m_elements.end(); e++) {

		if ((*e)->getServantName() == servantName and (*e)->getPropertyName() == propertyName and (*e)->getIndex() == index) {
			// Found.
			(*e)->setActiveWriteTime(seconds);

			//trace << "Set active time write " << servantName << " " << propertyName << " " << seconds << eol;

			saveConfig();

			// Exit.
			return true;
		}
	}

	return false;
}

bool Survey::setActiveWriteTime(const std::string& servantName, const std::string& propertyName, int seconds) {
	return setActiveWriteTime(servantName, propertyName, -1, seconds);
}

int Survey::getActiveWriteTime(const std::string& servantName, const std::string& propertyName, int index) {

	// Lock the elements.
	std::scoped_lock lock(m_mutex);

	// Iterate the elements.
	for (auto e = m_elements.begin(); e != m_elements.end(); e++) {

		if ((*e)->getServantName() == servantName and (*e)->getPropertyName() == propertyName and (*e)->getIndex() == index) {
			// Found.
			return (*e)->getActiveWriteTime();
		}
	}

	return -1;
}

std::vector<std::tuple<std::string, std::string, int, int>> Survey::getElements() {

	std::vector<std::tuple<std::string, std::string, int, int>> result;

	// Lock the elements.
	std::scoped_lock lock(m_mutex);

	// Iterate the elements.
	for (auto e = m_elements.begin(); e != m_elements.end(); e++) {
		result.push_back(std::make_tuple((*e)->getServantName(), (*e)->getPropertyName(), (*e)->getIndex(), (*e)->getActiveWriteTime()));
	}

	return result;
}

void Survey::showJSONMessages(bool value) {
	m_fluentdwriter.showJSONMessages(value);
}


void Survey::loop() {
	
	int cycleDurationMs = 5000;
	try {
		cycleDurationMs = ServerProperties::getInstance()->get<int>(SURVEY_CYCLE_DURATIONMS);

	} catch (const Exception& e) {
		out << "Setting default value 5000ms to property 'surveyCycleDurationms'" << eol;
	}
	
	while (m_instance->m_running) {
		
		try {
		
			Date end;
			end += Duration(cycleDurationMs, Duration::MILLISECONDS);

			m_instance->executeSurvey();
			m_instance->m_condition.wait(end);
			
		} catch (Exception& e) {
			err << e.printStack() << eol;
		} catch (exception& e) {
			err << "STL exception : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception" << eol;
		}
	}
}

void Survey::executeSurvey() {

	Date currentDate;
	string xmlMessage;

	bool newFile = false;
	if (SurveyWriter::getInstance()->isNewFile()) {
		newFile = true;
	}

	// Lock the elements.
	std::scoped_lock lock(m_mutex);

	// Iterate the elements.
	for (auto e = m_elements.begin(); e != m_elements.end(); e++) {
		
		PropertySurveyElement* element = *e;

		// Get write time.
		int writeTime = 0;
		if (element->isActive()) {

			int activeWriteTime = element->getActiveWriteTime();
			if (activeWriteTime == -1) {
				// Default value.
				writeTime = m_activeWriteTime;
			}
			else {
				writeTime = activeWriteTime;
			}
		}
		else {
			writeTime = m_inactiveWriteTime;
		}

		// Check write condition.
		if ((((currentDate - element->getDate()).getSeconds() - static_cast<double>(writeTime)) > -0.1) || element->isFirst() || newFile) {

			if (!DataAccess::getInstance()->isServantEnabledAndInitialised(0, element->getServantID())) {
				element->setDate(currentDate);
			}
			else {
				// Check for change of day for element and append to end of yesterday's file.
				if (currentDate.toString("%y%m%d") != element->getLastDate().toString("%y%m%d")) {
					string isoDateMidnight = element->getDate().toString("%Y%m%d") + "T235959";

					element->setDate(Date(boost::posix_time::from_iso_string(isoDateMidnight)));

					xmlMessage += SurveyWriter::getInstance()->write(element, true);
				}
				
				element->setDate(currentDate);

				//out << "writing element " << element->getPropertyName() << eol;
				xmlMessage += SurveyWriter::getInstance()->write(element, false);

				// Send JSON
				m_fluentdwriter.write(element);

				//trace << common::Date().toString() << " Write histo " << element->getServantName() << " " << element->getPropertyName() << eol;
			}
		}
		
	}

	// Send XML message.
	if (!xmlMessage.empty()) {
		m_surveyAccessor->notifyXML(xmlMessage);
	}

	// Exit if instance is stopped.
	if (!m_running) {
		return;
	}
}

void Survey::resetSurvey() {
	
	// Lock the elements.
	std::scoped_lock lock(m_mutex);

	// Destroy the elements.
	for (auto e = m_elements.begin(); e != m_elements.end(); e++) {
		delete *e;
	}

	m_elements.clear();
}


void Survey::writeAndSend(NumorSurveyElement* element) {
	string xmlMessage = SurveyWriter::getInstance()->write(element, false);
	if (!xmlMessage.empty()) {
		m_surveyAccessor->notifyXML(xmlMessage);
	}
	m_fluentdwriter.write(element);
}

void Survey::saveConfig() {

	// Do not save while loading the config.
	if (!m_running) {
		return;
	}

	string home;
	home.append(common::ServerProperties::getInstance()->getConfigPath());
	home.append("/");

	string filename;
	filename.append(home);
	filename.append(HISTOGRAM_CONFIG_NAME);

	//trace << "Saving config " << filename << eol;

	ofstream configFile(filename.c_str());

	// Iterate the elements.
	for (auto e = m_elements.begin(); e != m_elements.end(); e++) {

		int index = (*e)->getIndex();

		string timeString;
		int writeTime = (*e)->getActiveWriteTime();

		if (writeTime == -1) {
			timeString = DEFAULT_TIME;
		}
		else {
			timeString = std::to_string(writeTime);
		}

		if (index != -1) {
			configFile << (*e)->getServantName() << "," << (*e)->getPropertyName() << "." << index << "," << timeString << endl;
		}
		else {
			configFile << (*e)->getServantName() << "," << (*e)->getPropertyName() << "," << timeString << endl;
		}
	}

	configFile.close();
}

}
