/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_SURVEY_H_
#define CORE_SURVEY_H_

#include "FluentdSurveyWriter.h"
#include <common/base/Date.h>
#include <common/base/TimeCondition.h>
#include <string>
#include <vector>
#include <thread>
#include <mutex>
#include <memory>
#include <optional>

namespace core {

class PropertySurveyElement;
class NumorSurveyElement;
class SurveyAccessor;

/**
 * Class defining a singleton survey.
 */
class Survey {

public:
	/**
	 * Returns the instance.
	 */	
	static Survey* getInstance();
	
	/**
	 * Initialises the instance.
	 */
	static void initInstance(const std::string& configPath, SurveyAccessor * surveyAccessor);
	
	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Starts the survey.
	 */
	void run();
	
	/**
	 * Stops the survey.
	 */
	void stop();

	/**
	 * Resets the survey.
	 */
	void resetSurvey();

	/**
	 * Writes the element and send it to survey listeners.
	 * Returns the formatted XML line.
	 */
	void writeAndSend(NumorSurveyElement* element);

	/**
	 * Adds an element to the survey.
	 */
	bool addElement(const std::string& servantName, const std::string& propertyName, const std::optional<int>& activeWriteTime);

	/**
	 * Adds an element to the survey.
	 */
	bool addElement(const std::string& servantName, const std::string& propertyName, int index, const std::optional<int>& activeWriteTime);

	/**
	 * Removes an element.
	 */
	bool removeElement(const std::string& servantName, const std::string& propertyName, int index = -1);

	/**
	 * Sets the active write time.
	 */
	bool setActiveWriteTime(const std::string& servantName, const std::string& propertyName, int seconds);

	/**
	 * Sets the active write time.
	 */
	bool setActiveWriteTime(const std::string& servantName, const std::string& propertyName, int index, int seconds);

	/**
	 * Gets the active write time.
	 */
	int getActiveWriteTime(const std::string& servantName, const std::string& propertyName, int index = -1);

	/**
	 * Gets the elements. Tuple is servantName, propertyName, index, active write time.
	 */
	std::vector<std::tuple<std::string, std::string, int, int>> getElements();

	/**
	 * Shows or hides the JSON messages.
	 */
	void showJSONMessages(bool value);

	/**
	 * Default time value.
	 */
	static const std::string DEFAULT_TIME;

private:
	Survey();
	~Survey();

	void init(const std::string& configPath, SurveyAccessor * surveyAccessor);
	static void loop();	
	void executeSurvey();

	bool findElement(const std::string& servantName, const std::string& propertyName, int index = -1);
	void saveConfig();

	static Survey* m_instance;
	SurveyAccessor* m_surveyAccessor;

	FluentdSurveyWriter m_fluentdwriter;
	
	std::mutex m_mutex;
	std::vector<PropertySurveyElement*> m_elements;
	
	std::unique_ptr<std::thread> m_thread;
	bool m_running;
	
	common::TimeCondition m_condition;
	
	int m_activeWriteTime;
	int m_inactiveWriteTime;

	static const std::string SURVEY_CYCLE_DURATIONMS;
	static const std::string INACTIVE_WRITE_TIMES;
	static const std::string ACTIVE_WRITE_TIMES;
	static const std::string SURVEY_PATH;
	static const std::string HISTOGRAM_CONFIG_NAME;
};

}

#endif /* CORE_SURVEY_H_ */
