/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "SurveyWriter.h" 
#include "PropertySurveyElement.h"
#include "NumorSurveyElement.h"
#include "ics/DataProvider/Property/AnyProperty.h"
#include <core/dataaccess/DataAccess.h>
#include <core/transportlayer/dataprovider/PropertyAccessor.h>
#include <common/base/Exception.h> 
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h> 
#include <boost/format.hpp>
#include <iostream>

using namespace std;
using namespace common;

namespace core {

SurveyWriter* SurveyWriter::m_instance = 0;
const std::string SurveyWriter::SURVEY_PATH = "histo";

// Template function to convert floating values
template<typename T>
string convertFloatValue(T value) {

	if (value == T(0)) {
		return "0";
	}

	// Scientific format in any case.
	return (boost::format("%E") % value).str();
}

SurveyWriter::SurveyWriter() {
}

SurveyWriter::~SurveyWriter() {
}

SurveyWriter* SurveyWriter::getInstance() {
	if (m_instance == 0) {
		m_instance = new SurveyWriter();
	}
	return m_instance;
}

std::string SurveyWriter::write(SurveyElement* element, bool useLast) {
	
	std::scoped_lock lock(m_mutex);

	Date date = element->getDate();
	string surveyFileName = ServerProperties::getInstance()->getHomePath() + "/" + SURVEY_PATH + "/" + date.toString("%y%m%d") + ".xml";
	
	// Open file, create if it doesn't exist
	fstream surveyOutput(surveyFileName.c_str(), ios::out | ios::in); 
	surveyOutput.seekp(0, ios::end);
	
	// Determine if the file is new or not
	long position = surveyOutput.tellp();
	if (position == -1) {
		// open a new file
		surveyOutput.open(surveyFileName.c_str(), ios::out); 
		if(!surveyOutput.is_open()) { 
	    	err << "Cannot create survey file " << surveyFileName << eol;
	    	return "";
		}
		surveyOutput << createSurveyHeader(date);
	} else {
		// move to position before closing log tag
		surveyOutput.seekp(-9, ios::end);
	}

	string line;

	// Append the output
	try {
		line = format(element, useLast);
		surveyOutput << line;
		surveyOutput << "</histo>\n";
		
	} catch (Exception& e) {
		err << e.printStack() << eol;
	}	
	
	// close log file
	surveyOutput.close();
	
	return line;
}


std::string SurveyWriter::format(SurveyElement* element, bool useLast) const {

	PropertySurveyElement* propertyElement = dynamic_cast<PropertySurveyElement*>(element);

	if (propertyElement != 0) {
		return format(propertyElement, useLast);
	}

	NumorSurveyElement* numorElement = dynamic_cast<NumorSurveyElement*>(element);

	if (numorElement != 0) {
		return format(numorElement, useLast);
	}
	return "";
}

std::string SurveyWriter::format(PropertySurveyElement* element, bool useLast) const {
	string time = element->getDate().toString("%H:%M:%S");
	string ctype = element->getServantType();
	string cname = element->getServantName();
	string ptype = element->getPropertyType();
	string pname = element->getPropertyName();
	string alias = element->getAlias();
	unsigned long propertyID = element->getPropertyID();
	long index = element->getIndex();
	string pvalue;

	// not get the value if the servant is not ready	
	if (!DataAccess::getInstance()->isServantEnabledAndInitialised(0, element->getServantID())) {
		// Set the last date
		element->setLastDate(element->getDate());

		return "";
	}
	try {
		if (ptype == INT8_PROPERTY) {
			pvalue = to<string>(DataAccess::getInstance()->getInt8Value(0, propertyID));
		} else if (ptype == UINT16_PROPERTY) {
			pvalue = to<string>(DataAccess::getInstance()->getUint16Value(0, propertyID));
		} else if (ptype == UINT32_PROPERTY) {
			pvalue = to<string>(DataAccess::getInstance()->getUint32Value(0, propertyID));
		} else if (ptype == UINT64_PROPERTY) {
			pvalue = to<string>(DataAccess::getInstance()->getUint64Value(0, propertyID));
		} else if (ptype == INT16_PROPERTY) {
			pvalue = to<string>(DataAccess::getInstance()->getInt16Value(0, propertyID));
		} else if (ptype == INT32_PROPERTY) {
			pvalue = to<string>(DataAccess::getInstance()->getInt32Value(0, propertyID));
		} else if (ptype == INT64_PROPERTY) {
			pvalue = to<string>(DataAccess::getInstance()->getInt64Value(0, propertyID));
		} else if (ptype == FLOAT32_PROPERTY) {
			// special conversion
			pvalue = convertFloatValue(DataAccess::getInstance()->getFloat32Value(0, propertyID));
		} else if (ptype == FLOAT64_PROPERTY) {
			// special conversion
			pvalue = convertFloatValue(DataAccess::getInstance()->getFloat64Value(0, propertyID));
		} else if (ptype == BOOLEAN_PROPERTY) {
			pvalue = to<string>(DataAccess::getInstance()->getBooleanValue(0, propertyID));
		} else if (ptype == STRING_PROPERTY) {
			pvalue = DataAccess::getInstance()->getStringValue(0, propertyID);
		}
		
		element->setFirst(false);

	} catch (PropertyAccessor::NullPropertyReferenceException& e) {
		pvalue = "null";
		return "";
		
	} catch (Exception& e) {
		err << "Cannot get value of " << DataAccess::getInstance()->what(0, propertyID) << "\n" << e.printStack() << eol;
		pvalue = "?";
		return "";
	}

	// save the last value as well
	string returnValue;
	if (useLast) {
		if ((ptype != STRING_PROPERTY && element->getLastValue() != "") || ptype == STRING_PROPERTY) {
			returnValue = formatData(time, ctype, cname, pname, alias, index, element->getLastValue());
		}	
	} else {
		if ((ptype != STRING_PROPERTY && pvalue != "") || ptype == STRING_PROPERTY) {
			returnValue = formatData(time, ctype, cname, pname, alias, index, pvalue);
		}
	}
	
	element->setLastValue(pvalue);
	element->setLastDate(element->getDate());
	
	return returnValue;
}

std::string SurveyWriter::format(NumorSurveyElement* element, bool useLast) const {
	string time = element->getDate().toString("%H:%M:%S");

	string formattedData;

	if (element->isStart()) {
		formattedData = string("\t<numor_start time=\"") + time + "\" numor=\"" + to<string>(element->getNumor()) + "\"/>\n";

	} else {
		formattedData = string("\t<numor_end time=\"") + time + "\" numor=\"" + to<string>(element->getNumor()) + "\"/>\n";
	}

	return formattedData;
}

std::string SurveyWriter::formatData(const std::string& time, const std::string& ctype, const std::string& cname, const std::string& pname, const std::string& alias, long index, const std::string& pvalue) const {
	string formattedData = string("\t<entry time=\"") + time + "\" ctype=\"" + ctype + "\" cname=\"" + cname + "\" pname=\"" + pname + "\"";

	if (alias != "") {
		formattedData = formattedData + " alias=\"" + alias + "\"";
	}
	
	if (index != -1) {
		formattedData = formattedData + " index=\"" + to<string>(index) + "\"";
	}
	
	formattedData = formattedData + " pvalue=\"" + pvalue + "\"/>\n";
	return formattedData;
}

std::string SurveyWriter::createSurveyHeader(const Date& date) const {
	return string("<histo date=\"") + date.toString("%y%m%d") + "\">\n";
}

bool SurveyWriter::isNewFile() {
	Date date = Date();
	string surveyFileName = ServerProperties::getInstance()->getHomePath() + "/" + SURVEY_PATH + "/" + date.toString("%y%m%d") + ".xml";
	
	// Open file, don't create if it doesn't exist
	ofstream surveyOutput(surveyFileName.c_str(), ios::in); 
	if (surveyOutput.is_open()) {
		return false;
	}
	return true;
	
}


}
