/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "MultiThreadCommunication.h"
#include <core/transportlayer/command/DirectCommandLauncher.h>
#include <core/transportlayer/core/CompletionHandlerDirectImpl.h>

#include <core/transportlayer/core/SessionResponder.h>
#include <core/transportlayer/core/DatabaseResponder.h>
#include <core/transportlayer/core/ServantManagerResponder.h>
#include <core/transportlayer/core/commandline/CommandLineResponder.h>
#include <core/transportlayer/core/ConditionManagerResponder.h>
#include <core/transportlayer/core/AnalysisResponder.h>
#include <core/transportlayer/core/SurveyResponder.h>
#include <core/transportlayer/core/UserVariableManagerResponder.h>
#include <core/transportlayer/core/commandzone/CommandZoneEventPublisher.h>
#include <core/transportlayer/common/DataChangeSubscriberDirectImpl.h>
#include <core/transportlayer/dataprovider/DirectDataChangePublisher.h>
#include <core/transportlayer/dataprovider/DirectLogAccessor.h>
#include <core/transportlayer/dataprovider/DirectServerLogAccessor.h>
#include <core/transportlayer/dataprovider/DirectSurveyAccessor.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <command/transportlayer/MultiThreadProcessorManagerCommunication.h>
#include <dataprovider/transportlayer/MultiThreadCommunication.h>
#include <iostream>
#include <core/transportlayer/core/commandzone/CommandZoneResponder.h>
#include "dataprovider/DirectStateAccessor.h"

using namespace common;
using namespace std;

namespace core {

MultiThreadCommunication * MultiThreadCommunication::m_instance = 0;

MultiThreadCommunication::InitFailureException::InitFailureException(const std::string& problem) :
	Exception("core::MultiThreadCommunication::InitFailureException", problem) {
}

MultiThreadCommunication::DestructionFailureException::DestructionFailureException(const string& problem) :
	Exception("core::MultiThreadCommunication::DestructionFailureException", problem) {
}


MultiThreadCommunication::MultiThreadCommunication() {
}

MultiThreadCommunication::~MultiThreadCommunication() {
}

void MultiThreadCommunication::initInstance(int argc, char* argv[], bool remote, const std::string& serverID) {

	if (m_instance == 0) {
		m_instance = new MultiThreadCommunication();
		m_instance->init(argc, argv, remote, serverID);
	}
}

MultiThreadCommunication * MultiThreadCommunication::getInstance() {

	return m_instance;
}

void MultiThreadCommunication::resetInstance() {

	if (m_instance != 0) {
		m_instance->reset();
		delete m_instance;
	}
}

void MultiThreadCommunication::init(int argc, char* argv[], bool remote, const std::string& serverID) {

	try {
		// Create the session responder.
		m_sessionResponder.reset(new SessionResponder());
		m_sessionResponder->init();

		// Create the database responder.
		m_databaseResponder.reset(new DatabaseResponder(dataprovider::MultiThreadCommunication::getInstance()->getPropertyAccessor(), dataprovider::MultiThreadCommunication::getInstance()->getCommandAccessor()));
		m_databaseResponder->init();

		// Create the servant manager responder.
		m_servantManagerResponder.reset(new ServantManagerResponder(dataprovider::MultiThreadCommunication::getInstance()->getServantAccessor()));
		m_servantManagerResponder->init();

		// Create the command line responder.
		m_commandLineResponder.reset(new CommandLineResponder());
		m_commandLineResponder->init();

		// Create the condition manager responder.
		m_conditionManagerResponder.reset(new ConditionManagerResponder());
		m_conditionManagerResponder->init();

		// Create the command zone responder.
		m_commandBoxResponder.reset(new CommandZoneResponder());
		m_commandBoxResponder->init();

		// Create the variable manager responder.
		m_variableManagerResponder.reset(new UserVariableManagerResponder());
		m_variableManagerResponder->init();

		// Create the analysis responder.
		m_analysisResponder.reset(new AnalysisResponder());
		m_analysisResponder->init();

		// Create the survey responder.
		m_surveyResponder.reset(new SurveyResponder());
		m_surveyResponder->init();

		// Initialise the command zone and line event publishers.
		m_commandZoneEventPublisher.reset(new CommandZoneEventPublisher());

		// create CompletionHandler
		m_completionHandler = new CompletionHandlerDirectImpl();

		// create the property change subscriber
		m_dataChangePublisher = new DirectDataChangePublisher(dataprovider::MultiThreadCommunication::getInstance()->getDataChangePublisher());
		m_dataChangeSubscriber = new DataChangeSubscriberDirectImpl(m_dataChangePublisher);
		m_dataChangeSubscriber->subscribe();


		// direct objects
		m_commandLauncher = new DirectCommandLauncher(command::MultiThreadProcessorManagerCommunication::getInstance()->getCommandLauncher(), m_completionHandler);
		m_logAccessor = new DirectLogAccessor(dataprovider::MultiThreadCommunication::getInstance()->getLogAccessor());
		m_surveyAccessor = new DirectSurveyAccessor(dataprovider::MultiThreadCommunication::getInstance()->getSurveyAccessor());
		m_stateAccessor = new DirectStateAccessor(dataprovider::MultiThreadCommunication::getInstance()->getPropertyExpressionAccessor());

		out << "Core communications ready" << eol;

	} catch (Exception const & e) {
		throw link(InitFailureException("cannot initialise"), e);
	}
}

void MultiThreadCommunication::reset() {

	// remove subscriber for properties
	m_dataChangeSubscriber->unsubscribe();

	// Delete the objects
	delete m_commandLauncher;

	m_sessionResponder->stop();
	m_databaseResponder->stop();
	m_servantManagerResponder->stop();
	m_commandLineResponder->stop();
	m_conditionManagerResponder->stop();
	m_commandBoxResponder->stop();
	m_variableManagerResponder->stop();
	m_analysisResponder->stop();
	m_surveyResponder->stop();

	delete m_dataChangePublisher;

	delete m_logAccessor;
	delete m_surveyAccessor;
	delete m_stateAccessor;

	delete m_completionHandler;
	delete m_dataChangeSubscriber;

	m_commandZoneEventPublisher.reset(0);
}

void MultiThreadCommunication::run() {
}

void MultiThreadCommunication::stop() {
}

void MultiThreadCommunication::shutdown(bool isRemote) {
}

CommandLauncher* MultiThreadCommunication::getCommandLauncher() const {
	return m_commandLauncher;
}

LogAccessor* MultiThreadCommunication::getLogAccessor() const {
	return m_logAccessor;
}

SurveyAccessor* MultiThreadCommunication::getSurveyAccessor() const {
	return m_surveyAccessor;
}

StateAccessor* MultiThreadCommunication::getStateAccessor() const {
	return m_stateAccessor;
}

CommandZoneEventPublisher* MultiThreadCommunication::getCommandZoneEventPublisher() const {
	return m_commandZoneEventPublisher.get();
}

}
