/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_MULTITHREADCOMMUNICATION_H_
#define CORE_MULTITHREADCOMMUNICATION_H_

#include <common/base/Exception.h>
#include <common/base/TimeCondition.h>

namespace common {

	class ServerLogAccessor;
}

namespace core {

class CommandLauncher;
class CompletionHandlerDirectImpl;

class SessionResponder;
class DatabaseResponder;
class ServantManagerResponder;
class CommandLineResponder;
class ConditionManagerResponder;
class CommandZoneResponder;
class UserVariableManagerResponder;
class AnalysisResponder;
class SurveyResponder;

class DataChangeSubscriberDirectImpl;
class CommandZoneEventPublisher;
class DataChangePublisher;
class LogAccessor;
class SurveyAccessor;
class StateAccessor;

/**
 * Class defining the thread communication manager of the Core application.
 */
class MultiThreadCommunication {

public:
	/**
	 * Exception for initialisation failure.
	 */
	class InitFailureException  : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		InitFailureException(const std::string& problem);
	};

	/**
	 * Exception for destruction failure.
	 */
	class DestructionFailureException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		DestructionFailureException(const std::string& problem);
	};

	/**
	 * Initialises the instance.
	 * throw InitFailureException
	 */
	static void initInstance(int argc, char* argv[], bool remote, const std::string& serverID);

	/**
	 * Gets the instance.
	 */
	static MultiThreadCommunication * getInstance();

	/**
	 * Resets the instance.
	 * throw DestructionFailureException
	 */
	static void resetInstance();

	/**
	 * Runs the communication.
	 */
	void run();

	/**
	 * Stops the communication.
	 */
	void stop();

	/**
	 *  stops the main loop.
	 */
	void shutdown(bool isRemote);

	/**
	 * Gets the processor proxy.
	 */
	CommandLauncher* getCommandLauncher() const;

	/**
	 * Gets the log accessor.
	 */
	LogAccessor* getLogAccessor() const;

	/**
	 * Gets the survey accessor.
	 */
	SurveyAccessor* getSurveyAccessor() const;

	/**
	 * Gets the condition accessor.
	 */
	StateAccessor* getStateAccessor() const;

	/**
	 * Gets the command zone event publisher.
	 */
	CommandZoneEventPublisher* getCommandZoneEventPublisher() const;

private:
	/**
	 * Constructor.
	 */
	MultiThreadCommunication();

	/**
	 * Virtual destructor.
	 */
	virtual ~MultiThreadCommunication();

	/**
	 * Initialises the object.
	 * throw InitFailureException
	 */
	void init(int argc, char* argv[], bool remote, const std::string& serverID);

	/**
	 * Resets the communication objects.
	 */
	void reset();

	CommandLauncher * m_commandLauncher;
	CompletionHandlerDirectImpl* m_completionHandler;

	std::unique_ptr<SessionResponder> m_sessionResponder;
	std::unique_ptr<DatabaseResponder> m_databaseResponder;
	std::unique_ptr<ServantManagerResponder> m_servantManagerResponder;
	std::unique_ptr<CommandLineResponder> m_commandLineResponder;
	std::unique_ptr<ConditionManagerResponder> m_conditionManagerResponder;
	std::unique_ptr<CommandZoneResponder> m_commandBoxResponder;
	std::unique_ptr<CommandZoneEventPublisher> m_commandZoneEventPublisher;
	std::unique_ptr<UserVariableManagerResponder> m_variableManagerResponder;
	std::unique_ptr<AnalysisResponder> m_analysisResponder;
	std::unique_ptr<SurveyResponder> m_surveyResponder;

	LogAccessor* m_logAccessor;
	SurveyAccessor* m_surveyAccessor;
	StateAccessor* m_stateAccessor;

	DataChangePublisher* m_dataChangePublisher;
	DataChangeSubscriberDirectImpl* m_dataChangeSubscriber;

	bool m_running;
	common::TimeCondition m_condition;

	static MultiThreadCommunication * m_instance;
};

}

#endif /*CORE_MULTITHREADCOMMUNICATION_H_*/
