/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_COMMANDLAUNCHER_H_
#define CORE_COMMANDLAUNCHER_H_

#include <common/base/PropertyType.h>

namespace core {

/**
 * Client class for the command launcher.
 */
class CommandLauncher {
	
public:
	/**
	 * Constructor.
	 */
	CommandLauncher();
	
	/**
	 * Virtual destructor.
	 */
	virtual ~CommandLauncher();
	
	/**
	 * Executes the command with ID commandID.
	 */
	virtual void execute(unsigned long requestID, unsigned long databaseID, unsigned long commandID, bool skipIfWait) const = 0;
	
	/**
	 * Stops the command with ID commandID.
	 */
	virtual void stop(unsigned long databaseID, unsigned long commandID) const = 0;
	
	/**
	 * Pauses the command with ID commandID.
	 */
	virtual void pause(unsigned long databaseID, unsigned long commandID) const = 0;
	
	/**
	 * Restarts the command with ID commandID.
	 */
	virtual void restart(unsigned long databaseID, unsigned long commandID) const = 0;
	
	/**
	 * Stops the setter with ID propertyID.
	 */
	virtual void stopSetter(unsigned long databaseID, unsigned long propertyID) const = 0;
	
	/**
	 * Pauses the setter with ID propertyID.
	 */
	virtual void pauseSetter(unsigned long databaseID, unsigned long propertyID) const = 0;
	
	/**
	 * Restarts the setter with ID propertyID.
	 */
	virtual void restartSetter(unsigned long databaseID, unsigned long propertyID) const = 0;
	
	/**
	 * Initialises the drivers and the protocols
	 */
	virtual void initDriversAndProtocols() const = 0;
	
	/**
	 * Requests the duration of the command ID commandID.
	 */
	virtual void executeDurationRequest(unsigned long requestID, unsigned long databaseID, unsigned long commandID) const = 0;
	
	/**
	 * Sets the int8 value for the property of ID propertyID.
	 */
	virtual void setInt8Value(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, int8 value) const = 0;

	/**
	 * Sets the int16 value for the property of ID propertyID.
	 */
	virtual void setUint16Value(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, uint16 value) const = 0;

	/**
	 * Sets the int32 value for the property of ID propertyID.
	 */
	virtual void setUint32Value(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, uint32 value) const = 0;

	/**
	 * Sets the int64 value for the property of ID propertyID.
	 */
	virtual void setUint64Value(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, uint64 value) const = 0;
		
	/**
	 * Sets the int16 value for the property of ID propertyID.
	 */
	virtual void setInt16Value(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, int16 value) const = 0;
		
	/**
	 * Sets the int32 value for the property of ID propertyID.
	 */
	virtual void setInt32Value(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, int32 value) const = 0;
		
	/**
	 * Sets the int64 value for the property of ID propertyID.
	 */
	virtual void setInt64Value(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, int64 value) const = 0;
		
	/**
	 * Sets the float32 value for the property of ID propertyID.
	 */
	virtual void setFloat32Value(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, float32 value) const = 0;
	
	/**
	 * Sets the float64 value for the property of ID propertyID.
	 */
	virtual void setFloat64Value(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, float64 value) const = 0;
	
	/**
	 * Sets the string value for the property of ID propertyID.
	 */
	virtual void setStringValue(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, const std::string& value) const = 0;
	
	/**
	 * Sets the bool value for the property of ID propertyID.
	 */
	virtual void setBooleanValue(unsigned long requestID, unsigned long databaseID, unsigned long servantID, unsigned long propertyID, bool value) const = 0;

};

}

#endif /*CORE_COMMANDLAUNCHER_H_*/
