/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "AnalysisResponder.h"
#include <AnalysisRequests.pb.h>
#include <common/base/BaseTypes.h>
#include "common/analysislibrary/SimpleGaussFinder.h"
#include "common/analysislibrary/MultiGaussPeakFinder.h"
#include "common/analysislibrary/CenterOfMassPeakFinder.h"
#include <numeric>

using namespace std;
using namespace cameo;

namespace core {

/*
 * constructor
 */
AnalysisResponder::AnalysisResponder() {
}

/*
 * destructor
 */
AnalysisResponder::~AnalysisResponder() {
}

/*
 * init
 */
void AnalysisResponder::init() {

	// Create the responder.
	m_responder = coms::basic::Responder::create("analysis");
	m_responder->init();

	// Start the thread.
	m_thread.reset(new std::thread(std::bind(&AnalysisResponder::loop, this)));

	cout << "Initialised AnalysisResponder" << endl;
}

/*
 * stop
 */
void AnalysisResponder::stop() {

	// Cancel the responder so that unblocks the waiting thread.
	m_responder->cancel();

	// Join the thread.
	m_thread->join();

	cout << "Stopped AnalysisResponder" << endl;
}

/*
 * initProcessFunctions
 */
void AnalysisResponder::initProcessFunctions() {

	m_processFunctions[analysis::Message::GetFit] = &AnalysisResponder::processGetFit;
}

/*
 * processGetFit
 */
void AnalysisResponder::processGetFit(const std::string & message, std::string & response) {

	analysis::DataToFitMessage fitMessage;
	fitMessage.ParseFromString(message);

	if (fitMessage.type() == analysis::GaussianFit) {
		simpleGaussianFit(fitMessage, response);
	}
	else if (fitMessage.type() == analysis::MultiGaussianFit) {
		multiGaussianFit(fitMessage, response);
	}
	else if (fitMessage.type() == analysis::CenterOfMass) {
		centerOfMassFit(fitMessage, response);
	}
}

/*
 * simpleGaussianFit
 */
void AnalysisResponder::simpleGaussianFit(const analysis::DataToFitMessage& data, std::string & response) {
	analysis::FitDataMessage fitMessage;
	fitMessage.set_type(analysis::GaussianFit);
	try {
		if ((data.xdata_size() > 0) && ((data.ydata_size() > 0))) {
			blitz::Array<float64, 1> y;
			blitz::Array<float64, 1> x;

			x.resizeAndPreserve(data.xdata_size());
			y.resizeAndPreserve(data.ydata_size());

			for(uint32 i = 0;i < data.xdata_size(); ++i) {
				x(i) = data.xdata(i);
				y(i) = data.ydata(i);
			}

//			ostringstream os;
//			os << "y = " << y << endl << "x = " << x;
//			cout << "gaussian fit with " << os.str() << endl;

			SimpleGaussFinder finder(SimpleGaussFinder::ALL);
			finder.fit(y, x);

			blitz::Array<float64, 1> parameters = finder.getParameters();

//			for (int32 i = 0; i < parameters.size(); i++) {
//				cout << "parameters[" << i << "] = " << parameters(i) << endl;
//			}

			fitMessage.set_converged(finder.fitHasConverged());
			if (finder.fitHasConverged()) {
				fitMessage.set_position(finder.position());
				fitMessage.set_amplitude(finder.gaussianAmplitude());
				fitMessage.set_fwhm(2. * sqrt(2. * log(2.)) * parameters(2));
				for (int32 i = 0; i < x.size(); i++) {
					if (parameters.size() >= 4) {
						float64 ly = parameters(0) * exp(-pow((x(i) - parameters(1)) / parameters(2), 2.) / 2.) + parameters(3);
						if (ly < 0.0) {
							ly = 0.0;
						}
//						cout << "y = " << y << endl;
						fitMessage.add_fitydata(ly);
					}
				}
				fitMessage.set_min(*std::min_element(fitMessage.fitydata().begin(), fitMessage.fitydata().end()));
				fitMessage.set_max(*std::max_element(fitMessage.fitydata().begin(), fitMessage.fitydata().end()));
				fitMessage.set_integral(std::accumulate(fitMessage.fitydata().begin(), fitMessage.fitydata().end(), 0.0));

//				cout << "found position " << finder.position() << endl;

			} else {
//				cout << "SimpleGaussFinderFunction::calculate: no gaussian found" << endl;
			}
		}
		else {
			fitMessage.set_converged(false);

		}
	}
	catch (logic_error const & error) {
		cerr << error.what() << endl;
		fitMessage.set_converged(false);
	}
	fitMessage.SerializeToString(&response);
}

/*
 * multiGaussianFit
 */
void AnalysisResponder::multiGaussianFit(const analysis::DataToFitMessage& data, std::string & response) {
	analysis::FitDataMessage fitMessage;
	fitMessage.set_type(analysis::MultiGaussianFit);
	try {
		if ((data.xdata_size() > 0) && ((data.ydata_size() > 0))) {
			blitz::Array<float64, 1> y;
			blitz::Array<float64, 1> x;

			x.resizeAndPreserve(data.xdata_size());
			y.resizeAndPreserve(data.ydata_size());

			for(uint32 i = 0;i < data.xdata_size(); ++i) {
				x(i) = data.xdata(i);
				y(i) = data.ydata(i);
			}

			ostringstream os;
			os << "y = " << y << endl << "x = " << x;
			cout << "Multi gaussian fit with " << os.str() << endl;

			MultiGaussPeakFinder finder(y, x);

			uint32 nbPeaks = finder.numberOfGoodPeaks();
			if (nbPeaks > 0 && finder.getParameters().size() != 0) {
				blitz::Array<float64, 2> parameters = finder.getParameters();
				// Getting the highest peak.
				Peak* peak = finder.bestPeak();
				if (peak != 0) {
					fitMessage.set_converged(true);
					fitMessage.set_position(peak->position());
					fitMessage.set_amplitude(peak->amplitude());
					int32 peakindex = 0;
					for(uint32 i = 0;i < finder.numberOfPeaks();++i) {
						if (finder.peak(i) == peak) {
							peakindex = i;
							break;
						}
					}
					fitMessage.set_fwhm(2. * sqrt(2. * log(2.)) * (float64) parameters(peakindex, 2));
					for (int32 i = 0; i < x.size(); i++) {
						float64 ly = parameters(peakindex, 0) * exp(-pow((x(i) - parameters(peakindex, 1)) / parameters(peakindex, 2), 2.) / 2.) + parameters(peakindex, 3);
						if (ly < 0.0) {
							ly = 0.0;
						}
						cout << "y = " << ly << endl;
						fitMessage.add_fitydata(ly);
					}
					fitMessage.set_min(*std::min_element(fitMessage.fitydata().begin(), fitMessage.fitydata().end()));
					fitMessage.set_max(*std::max_element(fitMessage.fitydata().begin(), fitMessage.fitydata().end()));
					fitMessage.set_integral(std::accumulate(fitMessage.fitydata().begin(), fitMessage.fitydata().end(), 0.0));
				}
				else {
					fitMessage.set_converged(false);
					cout << "SimpleGaussFinderFunction::calculate: no good peak found" << endl;
				}
			}
			else {
				fitMessage.set_converged(false);
				cout << "SimpleGaussFinderFunction::calculate: no gaussian found" << endl;
			}
		}
		else {
			fitMessage.set_converged(false);
		}
	}
	catch (logic_error const & error) {
		cerr << error.what() << endl;
		fitMessage.set_converged(false);
	}
	fitMessage.SerializeToString(&response);
}

/*
 * centerOfMassFit
 */
void AnalysisResponder::centerOfMassFit(const analysis::DataToFitMessage& data, std::string & response) {
	analysis::FitDataMessage fitMessage;
	fitMessage.set_type(analysis::CenterOfMass);
	try {
		if ((data.xdata_size() > 0) && ((data.ydata_size() > 0))) {
			blitz::Array<float64, 1> y;
			blitz::Array<float64, 1> x;

			x.resizeAndPreserve(data.xdata_size());
			y.resizeAndPreserve(data.ydata_size());

			for(uint32 i = 0;i < data.xdata_size(); ++i) {
				x(i) = data.xdata(i);
				y(i) = data.ydata(i);
			}

			CenterOfMassPeakFinder finderPeak(y, x);

			float64 centerOfMass = finderPeak.getPosition();

			if (centerOfMass != 0) {
				fitMessage.set_converged(true);
				fitMessage.set_position(centerOfMass);
				float64 amplitude = finderPeak.getAmplitude();
				fitMessage.set_amplitude(amplitude);
				for (int32 i = 0; i < x.size(); i++) {
					fitMessage.add_fitydata(amplitude);
				}
				fitMessage.set_min(*std::min_element(fitMessage.fitydata().begin(), fitMessage.fitydata().end()));
				fitMessage.set_max(*std::max_element(fitMessage.fitydata().begin(), fitMessage.fitydata().end()));
				fitMessage.set_integral(std::accumulate(fitMessage.fitydata().begin(), fitMessage.fitydata().end(), 0.0));
			} else {
				fitMessage.set_converged(false);
			}
		}
		else {
			fitMessage.set_converged(false);
		}
	}
	catch (logic_error const & error) {
		cerr << error.what() << endl;
		fitMessage.set_converged(false);
	}
	fitMessage.SerializeToString(&response);
}

/*
 * loop
 */
void AnalysisResponder::loop() {

	// Initialize the map of functions.
	initProcessFunctions();

	// Loop until cancel is done.
	while (true) {
		// Wait for a new request.
		unique_ptr<coms::basic::Request> request = m_responder->receive();

		// Stop the loop if cancel is called.
		if (request.get() == 0) {
			break;
		}

		// Get the request type.
		string firstPart = request->get();

		// Get the request message.
		string secondPart = request->getSecondPart();

		analysis::Message messageType;
		messageType.ParseFromString(firstPart);

		string response;

		// Get the process function and apply it.
		ProcessFunctionType functionPointer = m_processFunctions[messageType.type()];

		if (functionPointer != 0) {
			(this->*functionPointer)(secondPart, response);
		}
		else {
			cerr << "analysis responder cannot process " << messageType.type() << endl;
		}

		request->reply(response);
	}
}


}
