/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ConditionManagerResponder.h"
#include <CommonResponses.pb.h>
#include <ConditionManagerRequests.pb.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <core/conditions/ConditionManager.h>
#include <core/session/SessionManager.h>

using namespace std;
using namespace cameo;
using namespace common;

namespace core {

ConditionManagerResponder::ConditionManagerResponder() {
}

ConditionManagerResponder::~ConditionManagerResponder() {
}

void ConditionManagerResponder::init() {

	// Create the responder.
	m_responder = coms::basic::Responder::create("condition_manager");
	m_responder->init();

	// Start the thread.
	m_thread.reset(new std::thread(std::bind(&ConditionManagerResponder::loop, this)));

	cout << "Initialised ConditionManagerResponder" << endl;
}

void ConditionManagerResponder::stop() {

	// Cancel the responder so that unblocks the waiting thread.
	m_responder->cancel();

	// Join the thread.
	m_thread->join();

	cout << "Stopped ConditionManagerResponder" << endl;
}

void ConditionManagerResponder::initProcessFunctions() {

	m_processFunctions[::condition::Message::VerifyExpression] = ProcessFunction(&ConditionManagerResponder::processVerifyExpression, false);
	m_processFunctions[::condition::Message::AddAlarm] = ProcessFunction(&ConditionManagerResponder::processAddAlarm, true);
	m_processFunctions[::condition::Message::EditAlarm] = ProcessFunction(&ConditionManagerResponder::processEditAlarm, true);
	m_processFunctions[::condition::Message::AddWarning] = ProcessFunction(&ConditionManagerResponder::processAddWarning, true);
	m_processFunctions[::condition::Message::EditWarning] = ProcessFunction(&ConditionManagerResponder::processEditWarning, true);
	m_processFunctions[::condition::Message::AddWatchdog] = ProcessFunction(&ConditionManagerResponder::processAddWatchdog, true);
	m_processFunctions[::condition::Message::EditWatchdog] = ProcessFunction(&ConditionManagerResponder::processEditWatchdog, true);
	m_processFunctions[::condition::Message::AddCommandZoneTermination] = ProcessFunction(&ConditionManagerResponder::processAddCommandZoneTermination, true);
	m_processFunctions[::condition::Message::EditCommandZoneTermination] = ProcessFunction(&ConditionManagerResponder::processEditCommandZoneTermination, true);
	m_processFunctions[::condition::Message::RemoveCondition] = ProcessFunction(&ConditionManagerResponder::processRemoveCondition, true);
	m_processFunctions[::condition::Message::SetConditionOn] = ProcessFunction(&ConditionManagerResponder::processSetConditionOn, true);
	m_processFunctions[::condition::Message::RenameCondition] = ProcessFunction(&ConditionManagerResponder::processRenameCondition, true);
	m_processFunctions[::condition::Message::GetConditionState] = ProcessFunction(&ConditionManagerResponder::processGetConditionState, false);
	m_processFunctions[::condition::Message::GetConditionsXMLContent] = ProcessFunction(&ConditionManagerResponder::processGetConditionsXMLContent, false);
}

bool ConditionManagerResponder::processVerifyExpression(const std::string & message, std::string & response) {

	::condition::VerifyExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	try {
		core::ConditionManager::getInstance()->verifyExpression(messageRequest.expression());
	}
	catch (core::ConditionManager::InvalidExpressionException const & e) {
		responseMessage.set_error(common::Error::INVALID_EXPRESSION);
		responseMessage.set_message(e.getMessage());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ConditionManagerResponder::processAddAlarm(const std::string & message, std::string & response) {

	::condition::AddAlarmRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->addAlarm(messageRequest.name(), messageRequest.expression(), messageRequest.message(), messageRequest.activationdelay(), messageRequest.stopcommandzone(), messageRequest.on(), messageRequest.mailrecipients(), messageRequest.smsnumber());
		}
		catch (core::ConditionManager::ConditionExistsException const & e) {
			responseMessage.set_error(common::Error::CONDITION_EXISTS);
		}
		catch (core::ConditionManager::InvalidExpressionException const & e) {
			responseMessage.set_error(common::Error::INVALID_EXPRESSION);
			responseMessage.set_message(e.getMessage());
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processEditAlarm(const std::string & message, std::string & response) {

	::condition::EditAlarmRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->editAlarm(messageRequest.conditionid(), messageRequest.expression(), messageRequest.message(), messageRequest.activationdelay(), messageRequest.stopcommandzone(), messageRequest.mailrecipients(), messageRequest.smsnumber());
		}
		catch (core::ConditionManager::NoSuchConditionException const & e) {
			responseMessage.set_error(common::Error::NO_SUCH_CONDITION);
		}
		catch (core::ConditionManager::InvalidExpressionException const & e) {
			responseMessage.set_error(common::Error::INVALID_EXPRESSION);
			responseMessage.set_message(e.getMessage());
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processAddWarning(const std::string & message, std::string & response) {

	::condition::AddWarningRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->addWarning(messageRequest.name(), messageRequest.expression(), messageRequest.message(), messageRequest.on(), messageRequest.mailrecipients(), messageRequest.smsnumber());
		}
		catch (core::ConditionManager::ConditionExistsException const & e) {
			responseMessage.set_error(common::Error::CONDITION_EXISTS);
		}
		catch (core::ConditionManager::InvalidExpressionException const & e) {
			responseMessage.set_error(common::Error::INVALID_EXPRESSION);
			responseMessage.set_message(e.getMessage());
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processEditWarning(const std::string & message, std::string & response) {

	::condition::EditWarningRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->editWarning(messageRequest.conditionid(), messageRequest.expression(), messageRequest.message(), messageRequest.mailrecipients(), messageRequest.smsnumber());
		}
		catch (core::ConditionManager::NoSuchConditionException const & e) {
			responseMessage.set_error(common::Error::NO_SUCH_CONDITION);
		}
		catch (core::ConditionManager::InvalidExpressionException const & e) {
			responseMessage.set_error(common::Error::INVALID_EXPRESSION);
			responseMessage.set_message(e.getMessage());
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processAddWatchdog(const std::string & message, std::string & response) {

	::condition::AddWatchdogRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->addWatchdog(messageRequest.name(), messageRequest.expression(), messageRequest.on(), messageRequest.actions(), messageRequest.mailrecipients(), messageRequest.smsnumber());
		}
		catch (core::ConditionManager::ConditionExistsException const & e) {
			responseMessage.set_error(common::Error::CONDITION_EXISTS);
		}
		catch (core::ConditionManager::InvalidExpressionException const & e) {
			responseMessage.set_error(common::Error::INVALID_EXPRESSION);
			responseMessage.set_message(e.getMessage());
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processEditWatchdog(const std::string & message, std::string & response) {

	::condition::EditWatchdogRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->editWatchdog(messageRequest.conditionid(), messageRequest.expression(), messageRequest.actions(), messageRequest.mailrecipients(), messageRequest.smsnumber());
		}
		catch (core::ConditionManager::NoSuchConditionException const & e) {
			responseMessage.set_error(common::Error::NO_SUCH_CONDITION);
		}
		catch (core::ConditionManager::InvalidExpressionException const & e) {
			responseMessage.set_error(common::Error::INVALID_EXPRESSION);
			responseMessage.set_message(e.getMessage());
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processAddCommandZoneTermination(const std::string & message, std::string & response) {

	::condition::AddCommandZoneTerminationRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->addCommandZoneTermination(messageRequest.message(), messageRequest.on(), messageRequest.mailrecipients(), messageRequest.smsnumber());
		}
		catch (core::ConditionManager::ConditionExistsException const & e) {
			responseMessage.set_error(common::Error::CONDITION_EXISTS);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processEditCommandZoneTermination(const std::string & message, std::string & response) {

	::condition::EditCommandZoneTerminationRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->editCommandZoneTermination(messageRequest.message(), messageRequest.mailrecipients(), messageRequest.smsnumber());
		}
		catch (core::ConditionManager::NoSuchConditionException const & e) {
			responseMessage.set_error(common::Error::NO_SUCH_CONDITION);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processRemoveCondition(const std::string & message, std::string & response) {

	::condition::RemoveConditionRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->removeCondition(messageRequest.conditionid());
		}
		catch (core::ConditionManager::NoSuchConditionException const & e) {
			responseMessage.set_error(common::Error::NO_SUCH_CONDITION);
		}
		catch (core::ConditionManager::UnremovableConditionException const & e) {
			responseMessage.set_error(common::Error::UNREMOVABLE_CONDITION);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processSetConditionOn(const std::string & message, std::string & response) {

	::condition::SetConditionOnRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->setConditionOn(messageRequest.conditionid(), messageRequest.value());
		}
		catch (core::ConditionManager::NoSuchConditionException const & e) {
			responseMessage.set_error(common::Error::NO_SUCH_CONDITION);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processRenameCondition(const std::string & message, std::string & response) {

	::condition::RenameConditionRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::ConditionManager::getInstance()->renameCondition(messageRequest.conditionid(), messageRequest.newname());
		}
		catch (core::ConditionManager::NoSuchConditionException const & e) {
			responseMessage.set_error(common::Error::NO_SUCH_CONDITION);
		}
		catch (core::ConditionManager::UnrenamableConditionException const & e) {
			responseMessage.set_error(common::Error::UNRENAMABLE_CONDITION);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processGetConditionState(const std::string & message, std::string & response) {

	::condition::GetConditionStateRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::ConditionStateResponse responseMessage;

	try {
		common::ConditionStateType::Type responseState = common::ConditionStateType::CONDITION_INACTIVE;
		common::ConditionState state = core::ConditionManager::getInstance()->getConditionState(messageRequest.conditionid());

		if (state == common::CONDITION_ACTIVE) {
			responseState = common::ConditionStateType::CONDITION_ACTIVE;
		} else if (state == common::CONDITION_INACTIVE) {
			responseState = common::ConditionStateType::CONDITION_INACTIVE;
		} else if (state == common::CONDITION_ON_ACTIVATION_DELAY) {
			responseState = common::ConditionStateType::CONDITION_ON_ACTIVATION_DELAY;
		}

		responseMessage.set_value(responseState);
	}
	catch (core::ConditionManager::NoSuchConditionException const & e) {
		responseMessage.set_error(common::Error::NO_SUCH_CONDITION);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ConditionManagerResponder::processGetConditionsXMLContent(const std::string & message, std::string & response) {

	::condition::GetConditionsXMLContentRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	responseMessage.set_value(core::ConditionManager::getInstance()->toXMLString());

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

void ConditionManagerResponder::loop() {

	// Initialize the map of functions.
	initProcessFunctions();

	// Loop until cancel is done.
	while (true) {
		// Wait for a new request.
		unique_ptr<coms::basic::Request> request = m_responder->receive();

		// Stop the loop if cancel is called.
		if (request.get() == 0) {
			break;
		}

		// Get the request type.
		string firstPart = request->get();

		// Get the request message.
		string secondPart = request->getSecondPart();

		::condition::Message messageType;
		messageType.ParseFromString(firstPart);

		string response;
		bool onError = false;

		// Get the process function and apply it.
		ProcessFunction & processFunction = m_processFunctions[messageType.type()];

		if (processFunction.functionPtr != nullptr) {
			onError = (this->*processFunction.functionPtr)(secondPart, response);
		}
		else {
			cerr << "Condition manager responder cannot process " << messageType.type() << endl;
		}

		request->reply(response);

		// Check error.
		if (!onError) {

			// Notify the clients in case the function modifies the command zone.
			if (processFunction.activity) {
				// Notify the token activity.
				SessionManager::getInstance()->notifyTokenActivity();
			}
		}
	}
}

}
