/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DatabaseResponder.h"
#include <dataprovider/transportlayer/dataprovider/PropertyAccessorDirectImpl.h>
#include <dataprovider/transportlayer/dataprovider/CommandAccessorDirectImpl.h>
#include <dataprovider/database/Database.h>
#include <CommonResponses.pb.h>
#include <DatabaseRequests.pb.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <core/scheduler/Sender.h>
#include <core/session/SessionManager.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <core/base/Hostname.h>

using namespace std;
using namespace boost;
using namespace cameo;
using namespace common;
using namespace dataprovider;

namespace core {

DatabaseResponder::DatabaseResponder(dataprovider::PropertyAccessorDirectImpl* propertyAccessor, dataprovider::CommandAccessorDirectImpl* commandAccessor) :
	m_propertyAccessor(propertyAccessor), m_commandAccessor(commandAccessor) {
}

DatabaseResponder::~DatabaseResponder() {
}

void DatabaseResponder::init() {

	// Create the responder.
	m_responder = coms::basic::Responder::create("database");
	m_responder->init();

	// Start the thread.
	m_thread.reset(new std::thread(std::bind(&DatabaseResponder::loop, this)));

	cout << "Initialised DatabaseResponder" << endl;
}

void DatabaseResponder::stop() {

	// Cancel the responder so that unblocks the waiting thread.
	m_responder->cancel();

	// Join the thread.
	m_thread->join();

	cout << "Stopped DatabaseResponder" << endl;
}

void DatabaseResponder::initProcessFunctions() {

	m_processFunctions[database::Message::GetUint16Value] = ProcessFunction(&DatabaseResponder::processGetUint16Value, false);
	m_processFunctions[database::Message::GetUint32Value] = ProcessFunction(&DatabaseResponder::processGetUint32Value, false);
	m_processFunctions[database::Message::GetUint64Value] = ProcessFunction(&DatabaseResponder::processGetUint64Value, false);
	m_processFunctions[database::Message::GetInt16Value] = ProcessFunction(&DatabaseResponder::processGetInt16Value, false);
	m_processFunctions[database::Message::GetInt32Value] = ProcessFunction(&DatabaseResponder::processGetInt32Value, false);
	m_processFunctions[database::Message::GetInt64Value] = ProcessFunction(&DatabaseResponder::processGetInt64Value, false);
	m_processFunctions[database::Message::GetBooleanValue] = ProcessFunction(&DatabaseResponder::processGetBooleanValue, false);
	m_processFunctions[database::Message::GetFloat32Value] = ProcessFunction(&DatabaseResponder::processGetFloat32Value, false);
	m_processFunctions[database::Message::GetFloat64Value] = ProcessFunction(&DatabaseResponder::processGetFloat64Value, false);
	m_processFunctions[database::Message::GetStringValue] = ProcessFunction(&DatabaseResponder::processGetStringValue, false);
	m_processFunctions[database::Message::SetUint16Value] = ProcessFunction(&DatabaseResponder::processSetUint16Value, true);
	m_processFunctions[database::Message::SetUint32Value] = ProcessFunction(&DatabaseResponder::processSetUint32Value, true);
	m_processFunctions[database::Message::SetUint64Value] = ProcessFunction(&DatabaseResponder::processSetUint64Value, true);
	m_processFunctions[database::Message::SetInt16Value] = ProcessFunction(&DatabaseResponder::processSetInt16Value, true);
	m_processFunctions[database::Message::SetInt32Value] = ProcessFunction(&DatabaseResponder::processSetInt32Value, true);
	m_processFunctions[database::Message::SetInt64Value] = ProcessFunction(&DatabaseResponder::processSetInt64Value, true);
	m_processFunctions[database::Message::SetFloat32Value] = ProcessFunction(&DatabaseResponder::processSetFloat32Value, true);
	m_processFunctions[database::Message::SetFloat64Value] = ProcessFunction(&DatabaseResponder::processSetFloat64Value, true);
	m_processFunctions[database::Message::SetStringValue] = ProcessFunction(&DatabaseResponder::processSetStringValue, true);
	m_processFunctions[database::Message::SetBooleanValue] = ProcessFunction(&DatabaseResponder::processSetBooleanValue, true);
	m_processFunctions[database::Message::GetArraySize] = ProcessFunction(&DatabaseResponder::processGetArraySize, false);
	m_processFunctions[database::Message::GetUint16Array] = ProcessFunction(&DatabaseResponder::processGetUint16Array, false);
	m_processFunctions[database::Message::GetUint32Array] = ProcessFunction(&DatabaseResponder::processGetUint32Array, false);
	m_processFunctions[database::Message::GetUint64Array] = ProcessFunction(&DatabaseResponder::processGetUint64Array, false);
	m_processFunctions[database::Message::GetInt16Array] = ProcessFunction(&DatabaseResponder::processGetInt16Array, false);
	m_processFunctions[database::Message::GetInt32Array] = ProcessFunction(&DatabaseResponder::processGetInt32Array, false);
	m_processFunctions[database::Message::GetInt64Array] = ProcessFunction(&DatabaseResponder::processGetInt64Array, false);
	m_processFunctions[database::Message::GetFloat32Array] = ProcessFunction(&DatabaseResponder::processGetFloat32Array, false);
	m_processFunctions[database::Message::GetFloat64Array] = ProcessFunction(&DatabaseResponder::processGetFloat64Array, false);
	m_processFunctions[database::Message::SetUint16Array] = ProcessFunction(&DatabaseResponder::processSetUint16Array, true);
	m_processFunctions[database::Message::SetUint32Array] = ProcessFunction(&DatabaseResponder::processSetUint32Array, true);
	m_processFunctions[database::Message::SetUint64Array] = ProcessFunction(&DatabaseResponder::processSetUint64Array, true);
	m_processFunctions[database::Message::SetInt16Array] = ProcessFunction(&DatabaseResponder::processSetInt16Array, true);
	m_processFunctions[database::Message::SetInt32Array] = ProcessFunction(&DatabaseResponder::processSetInt32Array, true);
	m_processFunctions[database::Message::SetInt64Array] = ProcessFunction(&DatabaseResponder::processSetInt64Array, true);
	m_processFunctions[database::Message::SetFloat32Array] = ProcessFunction(&DatabaseResponder::processSetFloat32Array, true);
	m_processFunctions[database::Message::SetFloat64Array] = ProcessFunction(&DatabaseResponder::processSetFloat64Array, true);
	m_processFunctions[database::Message::GetCommandState] = ProcessFunction(&DatabaseResponder::processGetCommandState, false);
	m_processFunctions[database::Message::GetCommandProgression] = ProcessFunction(&DatabaseResponder::processGetCommandProgression, false);
	m_processFunctions[database::Message::ExecuteCommand] = ProcessFunction(&DatabaseResponder::processExecuteCommand, true);
	m_processFunctions[database::Message::StopCommand] = ProcessFunction(&DatabaseResponder::processStopCommand, true);
	m_processFunctions[database::Message::GetWholeScannableProperties] = ProcessFunction(&DatabaseResponder::processGetWholeScannableProperties, false);
	m_processFunctions[database::Message::GetFormattedValue] = ProcessFunction(&DatabaseResponder::processGetFormattedValue, false);
	m_processFunctions[database::Message::GetPropertyId] = ProcessFunction(&DatabaseResponder::processGetPropertyIDRequest, false);
	m_processFunctions[database::Message::GetWholeDynamicProperties] = ProcessFunction(&DatabaseResponder::processGetWholeDynamicProperties, false);
	m_processFunctions[database::Message::GetPropertyType] = ProcessFunction(&DatabaseResponder::processGetProperyTypeRequest, false);
	m_processFunctions[database::Message::GetDynamicPropertyIds] = ProcessFunction(&DatabaseResponder::processGetDynamicPropertyIdsRequest, false);
	m_processFunctions[database::Message::GetServantId] = ProcessFunction(&DatabaseResponder::processGetServantIDRequest, false);
}

bool DatabaseResponder::processGetUint16Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Uint16Response responseMessage;

	try {
		uint16 value = m_propertyAccessor->getUint16Value(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(to<string>(value));
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetUint32Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Uint32Response responseMessage;

	try {
		uint32 value = m_propertyAccessor->getUint32Value(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(to<string>(value));
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetUint64Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Uint64Response responseMessage;

	try {
		uint64 value = m_propertyAccessor->getUint64Value(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(to<string>(value));
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}


bool DatabaseResponder::processGetInt16Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int16Response responseMessage;

	try {
		int16 value = m_propertyAccessor->getInt16Value(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(to<string>(value));
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetInt32Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	try {
		int32 value = m_propertyAccessor->getInt32Value(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(value);
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetInt64Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int64Response responseMessage;

	try {
		int64 value = m_propertyAccessor->getInt64Value(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(value);
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetFloat32Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Float32Response responseMessage;

	try {
		float32 value = m_propertyAccessor->getFloat32Value(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(value);
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetFloat64Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Float64Response responseMessage;

	try {
		float64 value = m_propertyAccessor->getFloat64Value(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(value);
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetStringValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	try {
		string value = m_propertyAccessor->getStringValue(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(value);
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetBooleanValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	try {
		bool value = m_propertyAccessor->getBooleanValue(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(value);
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processSetUint16Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetUint16ValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();
		bool result = Sender::setValueOrReset<uint16>(messageRequest.databaseid(), messageRequest.propertyid(), to<uint16>(messageRequest.value()) , true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + messageRequest.value()));
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processSetUint32Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetUint32ValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		bool result = Sender::setValueOrReset<uint32>(messageRequest.databaseid(), messageRequest.propertyid(), to<uint32>(messageRequest.value()), true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + messageRequest.value()));
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processSetUint64Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetUint64ValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		bool result = Sender::setValueOrReset<uint64>(messageRequest.databaseid(), messageRequest.propertyid(), to<uint64>(messageRequest.value()), true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + messageRequest.value()));
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processSetInt16Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetInt16ValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();
		bool result = Sender::setValueOrReset<int16>(messageRequest.databaseid(), messageRequest.propertyid(), to<int16>(messageRequest.value()) , true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + messageRequest.value()));
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processSetInt32Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetInt32ValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		bool result = Sender::setValueOrReset<int32>(messageRequest.databaseid(), messageRequest.propertyid(), messageRequest.value(), true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + to_string(messageRequest.value())));
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processSetInt64Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetInt64ValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		bool result = Sender::setValueOrReset<int64>(messageRequest.databaseid(), messageRequest.propertyid(), messageRequest.value(), true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + to_string(messageRequest.value())));
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processSetFloat32Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetFloat32ValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		bool result = Sender::setValueOrReset<float32>(messageRequest.databaseid(), messageRequest.propertyid(), messageRequest.value(), true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + to_string(messageRequest.value())));
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processSetFloat64Value(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetFloat64ValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		bool result = Sender::setValueOrReset<float64>(messageRequest.databaseid(), messageRequest.propertyid(), messageRequest.value(), true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + to_string(messageRequest.value())));
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processSetStringValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetStringValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		bool result = Sender::setValueOrReset<std::string>(messageRequest.databaseid(), messageRequest.propertyid(), std::string(messageRequest.value()), true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());

			// Do not log credentials property.
			if (propertyName.find("credentials") == std::string::npos) {
				LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + messageRequest.value()));
			}
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processSetBooleanValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::SetBooleanValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		bool result = Sender::setValueOrReset<bool>(messageRequest.databaseid(), messageRequest.propertyid(), messageRequest.value(), true);

		if (result) {
			string propertyName = Database::getInstance()->what(messageRequest.propertyid());
			string valueString = (messageRequest.value() ? "true" : "false");
			LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " sets " + propertyName + " to " + valueString));
		}

		responseMessage.set_value(result);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processGetArraySize(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	try {
		int32 value = m_propertyAccessor->getArraySize(messageRequest.databaseid(), messageRequest.propertyid());
		responseMessage.set_value(value);
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetUint16Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_propertyAccessor->getEncodedUint16ArrayValue(messageRequest.databaseid(), messageRequest.propertyid(), response);

	return false;
}

bool DatabaseResponder::processGetUint32Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_propertyAccessor->getEncodedUint32ArrayValue(messageRequest.databaseid(), messageRequest.propertyid(), response);

	return false;
}

bool DatabaseResponder::processGetUint64Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_propertyAccessor->getEncodedUint64ArrayValue(messageRequest.databaseid(), messageRequest.propertyid(), response);

	return false;
}

bool DatabaseResponder::processGetInt16Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_propertyAccessor->getEncodedInt16ArrayValue(messageRequest.databaseid(), messageRequest.propertyid(), response);

	return false;
}

bool DatabaseResponder::processGetInt32Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_propertyAccessor->getEncodedInt32ArrayValue(messageRequest.databaseid(), messageRequest.propertyid(), response);

	return false;
}

bool DatabaseResponder::processGetInt64Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_propertyAccessor->getEncodedInt64ArrayValue(messageRequest.databaseid(), messageRequest.propertyid(), response);

	return false;
}

bool DatabaseResponder::processGetFloat32Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_propertyAccessor->getEncodedFloat32ArrayValue(messageRequest.databaseid(), messageRequest.propertyid(), response);

	return false;
}

bool DatabaseResponder::processGetFloat64Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_propertyAccessor->getEncodedFloat64ArrayValue(messageRequest.databaseid(), messageRequest.propertyid(), response);

	return false;
}

bool DatabaseResponder::processSetUint16Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {
	cout << "Process of SetUint16ArrayRequest not yet implemented" << endl;
	return false;
}

bool DatabaseResponder::processSetUint32Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {
	cout << "Process of SetUint32ArrayRequest not yet implemented" << endl;
	return false;
}

bool DatabaseResponder::processSetUint64Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {
	cout << "Process of SetUint64ArrayRequest not yet implemented" << endl;
	return false;
}

bool DatabaseResponder::processSetInt16Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {
	cout << "Process of SetInt16ArrayRequest not yet implemented" << endl;
	return false;
}

bool DatabaseResponder::processSetInt32Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {
	cout << "Process of SetInt32ArrayRequest not yet implemented" << endl;
	return false;
}

bool DatabaseResponder::processSetInt64Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {
	cout << "Process of SetInt64ArrayRequest not yet implemented" << endl;
	return false;
}

bool DatabaseResponder::processSetFloat32Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {
	cout << "Process of SetFloat32ArrayRequest not yet implemented" << endl;
	return false;
}

bool DatabaseResponder::processSetFloat64Array(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {
	cout << "Process of SetFloat64ArrayRequest not yet implemented" << endl;
	return false;
}

bool DatabaseResponder::processGetCommandState(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetCommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::CommandStateResponse responseMessage;

	try {
		common::CommandState commandState = m_commandAccessor->getCommandState(messageRequest.databaseid(), messageRequest.commandid());

		if (commandState == common::ACTIVE) {
			responseMessage.set_value(common::CommandStateType::ACTIVE);
		}
		else if (commandState == common::PAUSED) {
			responseMessage.set_value(common::CommandStateType::PAUSED);
		}
		else if (commandState == common::PAUSING) {
			responseMessage.set_value(common::CommandStateType::PAUSING);
		}
		else if (commandState == common::STOPPED) {
			responseMessage.set_value(common::CommandStateType::STOPPED);
		}
		else if (commandState == common::STOPPING) {
			responseMessage.set_value(common::CommandStateType::STOPPING);
		}
		else if (commandState == common::RESUMING) {
			responseMessage.set_value(common::CommandStateType::RESTARTING);
		}
		else if (commandState == common::PENDING) {
			responseMessage.set_value(common::CommandStateType::PENDING);
		}
		else if (commandState == common::STARTING) {
			responseMessage.set_value(common::CommandStateType::STARTING);
		}
		else {
			responseMessage.set_value(common::CommandStateType::INACTIVE);
		}
	}
	catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_COMMAND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetCommandProgression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetCommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Float64Response responseMessage;

	try {
		double value = m_commandAccessor->getProgression(messageRequest.databaseid(), messageRequest.commandid());
		responseMessage.set_value(value);
	}
	catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_COMMAND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processExecuteCommand(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::CommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		try {
			// Call to generate the exception.
			m_commandAccessor->getCommandState(messageRequest.databaseid(), messageRequest.commandid());

			RequestReply reply = Sender::newCommand(messageRequest.databaseid(), messageRequest.commandid(), false);
			if (reply.getValue() == RequestReply::REJECT) {
				common::out << reply.getReason() << common::eol;
			}
			else {
				string commandName = Database::getInstance()->what(messageRequest.commandid());
				LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " executes " + commandName));
			}

			responseMessage.set_value(true);
		}
		catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processStopCommand(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::CommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		try {
			// Call to generate the exception.
			m_commandAccessor->getCommandState(messageRequest.databaseid(), messageRequest.commandid());

			Sender::stopCommand(messageRequest.databaseid(), messageRequest.commandid(), false);

			string commandName = Database::getInstance()->what(messageRequest.commandid());
						LogHandler::getInstance()->write(SimpleLog(common::LOG_DEBUG, "Client " + hostname(endpoint) + " stops " + commandName));

			responseMessage.set_value(true);
		}
		catch (CommandAccessorDirectImpl::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool DatabaseResponder::processGetWholeScannableProperties(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetWholePropertiesRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	// Call to generate the exception.
	std::string result = Database::getInstance(messageRequest.databaseid())->getWholeScannableProperties();
	responseMessage.set_value(result);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetFormattedValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	try {
		responseMessage.set_value(Database::getInstance(messageRequest.databaseid())->getFormattedValue(messageRequest.propertyid()));
	}
	catch (PropertyAccessorDirectImpl::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (PropertyAccessorDirectImpl::BadPropertyTypeException& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetPropertyIDRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetPropertyIDRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	try {
		unsigned long servantID = Database::getInstance(messageRequest.databaseid())->getServantID(messageRequest.servantname());
		unsigned long propertyID = Database::getInstance(messageRequest.databaseid())->getPropertyID(servantID, messageRequest.propertyname());

		responseMessage.set_value(propertyID);
	}
	catch (Database::NoSuchServantException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_SERVANT);
	}
	catch (Database::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetWholeDynamicProperties(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetWholePropertiesRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	// Call to generate the exception.
	std::string result = Database::getInstance(messageRequest.databaseid())->getWholeDynamicProperties();
	responseMessage.set_value(result);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetProperyTypeRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetPropertyTypeRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	try {
		std::string propertyType = Database::getInstance(messageRequest.databaseid())->getPropertyType(messageRequest.propertyid());

		responseMessage.set_value(propertyType);
	}
	catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetDynamicPropertyIdsRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetDynamicPropertyIdsRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::DynamicPropertyIds responseMessage;

	// Call to generate the exception.
	std::vector<unsigned long> result = Database::getInstance(messageRequest.databaseid())->getDynamicPropertyIDs(messageRequest.propertyid());
	for(auto id : result) {
		responseMessage.add_id(id);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool DatabaseResponder::processGetServantIDRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	database::GetServantIDRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	try {
		unsigned long servantID = Database::getInstance(messageRequest.databaseid())->getServantID(messageRequest.servantname());

		responseMessage.set_value(servantID);
	}
	catch (Database::NoSuchServantException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_SERVANT);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

void DatabaseResponder::loop() {

	// Initialize the map of functions.
	initProcessFunctions();

	// Loop until cancel is done.
	while (true) {
		// Wait for a new request.
		unique_ptr<coms::basic::Request> request = m_responder->receive();

		// Stop the loop if cancel is called.
		if (request.get() == 0) {
			break;
		}

		// Get the request type.
		string firstPart = request->get();

		// Get the request message.
		string secondPart = request->getSecondPart();

		database::Message messageType;
		messageType.ParseFromString(firstPart);

		string response;
		bool onError = false;

		// Get the process function and apply it.
		ProcessFunction & processFunction = m_processFunctions[messageType.type()];

		if (processFunction.functionPtr != nullptr) {
			onError = (this->*processFunction.functionPtr)(request.get(), secondPart, response);
		}
		else {
			cerr << "Database manager responder cannot process " << messageType.type() << endl;
		}

		request->reply(response);

		// Check error.
		if (!onError) {

			// Notify the clients in case the function modifies the command zone.
			if (processFunction.activity) {
				// Notify the token activity.
				SessionManager::getInstance()->notifyTokenActivity();
			}
		}
	}
}

}
