/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ServantManagerResponder.h"
#include <dataprovider/transportlayer/dataprovider/ServantAccessorDirectImpl.h>
#include <dataprovider/iconandfamily/ServantIconAndFamilyMaps.h>
#include <CommonResponses.pb.h>
#include <ServantManagerRequests.pb.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <core/cloner/ServantCloner.h>
#include <core/dataaccess/DataAccess.h>
#include <command/base/ServantManager.h>
#include <ics/InstrumentManager/HardwareManager.h>
#include <ics/InstrumentManager/InstrumentManager.h>
#include <ics/Module/ModuleManager.h>
#include <ics/DeviceContainer/Spy.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <core/session/SessionManager.h>
#include <core/scheduler/Scheduler.h>
#include <core/message/scheduler/NewCommand.h>
#include <../../core/commandzone/SaveBoxPropertyManager.h>

using namespace std;
using namespace boost;
using namespace cameo;
using namespace common;
using namespace dataprovider;

namespace core {

ServantManagerResponder::ServantManagerResponder(dataprovider::ServantAccessorDirectImpl* servantAccessor) :
	m_servantAccessor(servantAccessor) {
}

ServantManagerResponder::~ServantManagerResponder() {
}

void ServantManagerResponder::init() {

	// Create the responder.
	m_responder = coms::basic::Responder::create("servant_manager");
	m_responder->init();

	// Start the thread.
	m_thread.reset(new std::thread(std::bind(&ServantManagerResponder::loop, this)));

	cout << "Initialised ServantManagerResponder" << endl;
}

void ServantManagerResponder::stop() {

	// Cancel the responder so that unblocks the waiting thread.
	m_responder->cancel();

	// Join the thread.
	m_thread->join();

	cout << "Stopped ServantManagerResponder" << endl;
}

void ServantManagerResponder::initProcessFunctions() {

	m_processFunctions[servantmanager::Message::GetWholeConfiguration] = ProcessFunction(&ServantManagerResponder::processGetWholeConfiguration, false);
	m_processFunctions[servantmanager::Message::GetWholeConfigurationUpdate] = ProcessFunction(&ServantManagerResponder::processGetWholeConfigurationUpdate, false);
	m_processFunctions[servantmanager::Message::GetControllerConfiguration] = ProcessFunction(&ServantManagerResponder::processGetControllerConfiguration, false);
	m_processFunctions[servantmanager::Message::GetServantSpyProperties] = ProcessFunction(&ServantManagerResponder::processGetServantSpyProperties, false);
	m_processFunctions[servantmanager::Message::GetServantDynamicProperties] = ProcessFunction(&ServantManagerResponder::processGetServantDynamicProperties, false);
	m_processFunctions[servantmanager::Message::GetServantSpyDynamicProperties] = ProcessFunction(&ServantManagerResponder::processGetServantSpyDynamicProperties, false);
	m_processFunctions[servantmanager::Message::CloneController] = ProcessFunction(&ServantManagerResponder::processCloneController, false);
	m_processFunctions[servantmanager::Message::RemoveController] = ProcessFunction(&ServantManagerResponder::processRemoveController, true);
	m_processFunctions[servantmanager::Message::GetDriverChannel] = ProcessFunction(&ServantManagerResponder::processGetDriverChannel, false);
	m_processFunctions[servantmanager::Message::ReconnectDriver] = ProcessFunction(&ServantManagerResponder::processReconnectDriver, true);
	m_processFunctions[servantmanager::Message::RemoveDriver] = ProcessFunction(&ServantManagerResponder::processRemoveDriver, true);
	m_processFunctions[servantmanager::Message::PauseSpies] = ProcessFunction(&ServantManagerResponder::processPauseSpies, true);
	m_processFunctions[servantmanager::Message::GetPropertiesForGeneric] = ProcessFunction(&ServantManagerResponder::processGetPropertiesForGeneric, false);
	m_processFunctions[servantmanager::Message::GetWebSpyProperties] = ProcessFunction(&ServantManagerResponder::processGetWebSpyProperties, false);
	m_processFunctions[servantmanager::Message::GetIcsExperimentControllersConfiguration] = ProcessFunction(&ServantManagerResponder::processGetIcsExperimentControllersConfiguration, false);
	m_processFunctions[servantmanager::Message::GetIcsTestControllersConfiguration] = ProcessFunction(&ServantManagerResponder::processGetIcsTestControllersConfiguration, false);
	m_processFunctions[servantmanager::Message::GetResourceFilesByExtension] = ProcessFunction(&ServantManagerResponder::processGetResourceFilesByExtension, false);
	m_processFunctions[servantmanager::Message::ReadBinaryResourceFile] = ProcessFunction(&ServantManagerResponder::processReadBinaryResourceFile, false);
	m_processFunctions[servantmanager::Message::ReadResourceFile] = ProcessFunction(&ServantManagerResponder::processReadResourceFile, false);
	m_processFunctions[servantmanager::Message::WriteResourceFile] = ProcessFunction(&ServantManagerResponder::processWriteResourceFile, false);
	m_processFunctions[servantmanager::Message::GetResourceFileLastTime] = ProcessFunction(&ServantManagerResponder::processGetResourceFileLastTime, false);
	m_processFunctions[servantmanager::Message::GetServantPropertiesValue] = ProcessFunction(&ServantManagerResponder::processGetServantPropertiesValue, false);
	m_processFunctions[servantmanager::Message::GetServantCommandsState] = ProcessFunction(&ServantManagerResponder::processGetServantCommandsState, false);
	m_processFunctions[servantmanager::Message::GetFamilyImageKeys] = ProcessFunction(&ServantManagerResponder::processGetFamilyImageKeys, false);
	m_processFunctions[servantmanager::Message::GetControllerIconKeys] = ProcessFunction(&ServantManagerResponder::processGetControllerIconKeys, false);
	m_processFunctions[servantmanager::Message::GetControllerTypesInheritingType] = ProcessFunction(&ServantManagerResponder::processGetControllerTypesInheritingType, false);
	m_processFunctions[servantmanager::Message::UserLog] = ProcessFunction(&ServantManagerResponder::processUserLog, false);
	m_processFunctions[servantmanager::Message::HasRunningCommands] = ProcessFunction(&ServantManagerResponder::processHasRunningCommands, false);
	m_processFunctions[servantmanager::Message::GetSpyPeriod] = ProcessFunction(&ServantManagerResponder::processGetSpyPeriod, false);
	m_processFunctions[servantmanager::Message::GetPropertiesSavedInBox] = ProcessFunction(&ServantManagerResponder::processGetPropertiesSavedInBox, false);
	m_processFunctions[servantmanager::Message::GetPropertySavedInBoxValue] = ProcessFunction(&ServantManagerResponder::processGetPropertySavedInBoxValue, false);
	m_processFunctions[servantmanager::Message::SetPropertySavedInBoxValue] = ProcessFunction(&ServantManagerResponder::processSetPropertySavedInBoxValue, false);


}

bool ServantManagerResponder::processGetWholeConfiguration(const std::string & message, std::string & response) {

	servantmanager::GetWholeConfigurationRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_servantAccessor->getWholeConfiguration(messageRequest.databaseid(), response);

	return false;
}

bool ServantManagerResponder::processGetWholeConfigurationUpdate(const std::string & message, std::string & response) {

	servantmanager::GetWholeConfigurationRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_servantAccessor->getWholeConfigurationUpdate(messageRequest.databaseid(), response);

	return false;
}

bool ServantManagerResponder::processGetControllerConfiguration(const std::string & message, std::string & response) {

	servantmanager::ServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_servantAccessor->getControllerConfiguration(messageRequest.databaseid(), messageRequest.servantid(), response);

	return false;
}

bool ServantManagerResponder::processGetServantSpyProperties(const std::string & message, std::string & response) {

	servantmanager::ServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_servantAccessor->getServantSpyProperties(messageRequest.databaseid(), messageRequest.servantid(), response);

	return false;
}

bool ServantManagerResponder::processGetServantDynamicProperties(const std::string & message, std::string & response) {

	servantmanager::ServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_servantAccessor->getServantDynamicProperties(messageRequest.databaseid(), messageRequest.servantid(), response);

	return false;
}

bool ServantManagerResponder::processGetServantSpyDynamicProperties(const std::string & message, std::string & response) {

	servantmanager::ServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_servantAccessor->getServantSpyDynamicProperties(messageRequest.databaseid(), messageRequest.servantid(), response);

	return false;
}

bool ServantManagerResponder::processCloneController(const std::string & message, std::string & response) {

	servantmanager::ServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	unsigned long cloneControllerID = ServantCloner::getInstance()->cloneController(messageRequest.servantid());

	common::Int32Response responseMessage;

	if (cloneControllerID == 0) {
		responseMessage.set_error(common::Error::NO_SUCH_SERVANT);
	}

	responseMessage.set_value(cloneControllerID);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processRemoveController(const std::string & message, std::string & response) {

	servantmanager::ModifyServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// test ics
		if (InstrumentManager::getInstance() != 0) {

		// get the name of the controller
		string controllerName = DataAccess::getInstance()->getServantName(0, messageRequest.servantid());

		// remove the controller from InstrumentManager
		try {
			InstrumentManager::getInstance()->removeController(controllerName);

		} catch (InstrumentManager::InstrumentManagerError const & e) {
			responseMessage.set_error(common::Error::NO_SUCH_SERVANT);
		}

		// remove the controller in the command processor
		command::ServantManager::getInstance()->removeController(messageRequest.servantid());

		// remove the controller in data provider
		m_servantAccessor->removeServant(0, messageRequest.servantid());
		}

		responseMessage.set_value(true);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ServantManagerResponder::processGetDriverChannel(const std::string & message, std::string & response) {

	servantmanager::ServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	try {
		responseMessage.set_value(m_servantAccessor->getDriverChannel(messageRequest.databaseid(), messageRequest.servantid()));

	} catch (...) {
		responseMessage.set_error(common::Error::NO_SUCH_SERVANT);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processReconnectDriver(const std::string & message, std::string & response) {

	servantmanager::ModifyServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// test ics
		if (HardwareManager::getInstance() != 0) {
			// get the name of the test controller
			string driverName = DataAccess::getInstance()->getServantName(messageRequest.databaseid(), messageRequest.servantid());

			// enable the test controller
			try {
				HardwareManager::getInstance()->getTestController(driverName)->forceEnabled(true);
			} catch (...) {
				//HardwareManager::getInstance()->getTestController(driverName)->getDeviceDriver()->setDisconnected();
				cerr << "Failed to reconnect " << driverName << endl;
			}
		}

		responseMessage.set_value(true);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ServantManagerResponder::processRemoveDriver(const std::string & message, std::string & response) {

	servantmanager::ModifyServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// test ics
		if (HardwareManager::getInstance() != 0) {

			// get the name of the driver
			string driverName = DataAccess::getInstance()->getServantName(0, messageRequest.servantid());

			// remove the driver from HardwareManager
			try {
				HardwareManager::getInstance()->removeDriver(driverName);

			} catch (HardwareManager::HardwareManagerError const & e) {
				responseMessage.set_error(common::Error::NO_SUCH_SERVANT);
			}

			// remove the driver in the command processor
			command::ServantManager::getInstance()->removeDriver(messageRequest.servantid());

			// remove the driver in data provider
			m_servantAccessor->removeServant(0, messageRequest.servantid());
		}
	}

	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ServantManagerResponder::processPauseSpies(const std::string & message, std::string & response) {

	servantmanager::PauseSpiesRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// test ics
		if (HardwareManager::getInstance() != 0) {
			HardwareManager::getInstance()->setPaused(messageRequest.ispaused());
		}

		responseMessage.set_value(true);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool ServantManagerResponder::processGetPropertiesForGeneric(const std::string & message, std::string & response) {

	common::StringResponse responseMessage;

	// test ics
	if (InstrumentManager::getInstance() != 0) {
		responseMessage.set_value(InstrumentManager::getInstance()->getPropertiesForGeneric());
	}
	else {
		responseMessage.set_value("");
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetWebSpyProperties(const std::string & message, std::string & response) {

	common::StringResponse responseMessage;

	// test ics
	if (InstrumentManager::getInstance() != 0) {
		responseMessage.set_value(InstrumentManager::getInstance()->getWebSpyProperties());
	}
	else {
		responseMessage.set_value("");
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetIcsExperimentControllersConfiguration(const std::string & message, std::string & response) {

	common::StringResponse responseMessage;

	// test ics
	if (InstrumentManager::getInstance() != 0) {
		responseMessage.set_value(InstrumentManager::getInstance()->getExperimentControllersConfig());
	}
	else {
		responseMessage.set_value("");
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetIcsTestControllersConfiguration(const std::string & message, std::string & response) {

	common::StringResponse responseMessage;

	// test ics
	if (InstrumentManager::getInstance() != 0) {
		responseMessage.set_value(HardwareManager::getInstance()->getTestControllersConfig());
	}
	else {
		responseMessage.set_value("");
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetResourceFilesByExtension(const std::string & message, std::string & response) {

	servantmanager::GetResourceFilesByExtensionRequest messageRequest;
	messageRequest.ParseFromString(message);

	vector<string> list = ModuleManager::getInstance()->getResourceFilesByExtension(messageRequest.extension());

	common::StringArrayResponse responseMessage;

	// Copy the array elements
	vector<string>::const_iterator it;
	for (it = list.begin(); it != list.end(); it++) {
		responseMessage.add_value(*it);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processReadBinaryResourceFile(const std::string & message, std::string & response) {

	servantmanager::ReadResourceRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BinaryResponse responseMessage;

	try {
		string content = ModuleManager::getInstance()->readResourceFile(messageRequest.type(), messageRequest.filename());
		responseMessage.set_value(content);

	} catch (const NoSuchFile& e) {
		responseMessage.set_error(common::Error::NO_SUCH_RESOURCE);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processReadResourceFile(const std::string & message, std::string & response) {

	servantmanager::ReadResourceRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	try {
		string content = ModuleManager::getInstance()->readResourceFile(messageRequest.type(), messageRequest.filename());
		responseMessage.set_value(content);

	} catch (const NoSuchFile& e) {
		responseMessage.set_error(common::Error::NO_SUCH_RESOURCE);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processWriteResourceFile(const std::string & message, std::string & response) {

	servantmanager::WriteResourceRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// use module manager
	ModuleManager::getInstance()->writeResourceFile(messageRequest.type(), messageRequest.filename(), messageRequest.content());
	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetResourceFileLastTime(const std::string & message, std::string & response) {

	servantmanager::ReadResourceRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	try {
		time_t date = ModuleManager::getInstance()->getResourceFileLastTime(messageRequest.type(), messageRequest.filename());
		responseMessage.set_value((int32) date);
	} catch (const NoSuchFile& e) {
		responseMessage.set_error(common::Error::NO_SUCH_RESOURCE);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetServantPropertiesValue(const std::string & message, std::string & response) {

	servantmanager::ServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_servantAccessor->getServantPropertiesValue(messageRequest.databaseid(), messageRequest.servantid(), response);

	return false;
}

bool ServantManagerResponder::processGetServantCommandsState(const std::string & message, std::string & response) {

	servantmanager::ServantRequest messageRequest;
	messageRequest.ParseFromString(message);

	m_servantAccessor->getServantCommandsState(messageRequest.databaseid(), messageRequest.servantid(), response);

	return false;
}

bool ServantManagerResponder::processGetFamilyImageKeys(const std::string & message, std::string & response) {

	servantmanager::EmptyRequest emptyRequest;

	map<string, string> values = dataprovider::ServantIconAndFamilyMaps::getInstance()->getFamilyImageKeys();

	common::StringMapResponse responseMessage;

	map<string, string>::const_iterator it;
	for (it = values.begin(); it != values.end(); it++) {
		common::StringAndKey* avalue = responseMessage.add_map();
		avalue->set_key(it->first);
		avalue->set_value(it->second);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetControllerIconKeys(const std::string & message, std::string & response) {

	servantmanager::EmptyRequest emptyRequest;

	map<string, string> values = dataprovider::ServantIconAndFamilyMaps::getInstance()->getControllerIconKeys();

	common::StringMapResponse responseMessage;

	map<string, string>::const_iterator it;
	for (it = values.begin(); it != values.end(); it++) {
		common::StringAndKey* avalue = responseMessage.add_map();
		avalue->set_key(it->first);
		avalue->set_value(it->second);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processUserLog(const std::string & message, std::string & response) {

	servantmanager::UserLogRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Log the message.
	string logMessage = "";
	if (!messageRequest.username().empty()) {
		logMessage += messageRequest.username() + " : ";
	}
	logMessage += messageRequest.message();

	LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, logMessage));

	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetControllerTypesInheritingType(const std::string & message, std::string & response) {

	servantmanager::GetControllerTypesInheritingTypeRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringArrayResponse responseMessage;

	set<string> controllerTypes = Database::getInstance(messageRequest.databaseid())->getControllerTypesInheritingType(messageRequest.type());

	for (auto it = controllerTypes.begin(); it != controllerTypes.end(); it++) {
		responseMessage.add_value(*it);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processHasRunningCommands(const std::string & message, std::string & response) {

	common::BooleanResponse responseMessage;

	// Check the scheduler.
	bool runningCommands = !Scheduler::getInstance()->getActiveCommands().empty();

	responseMessage.set_value(runningCommands);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetSpyPeriod(const std::string & message, std::string & response) {

	servantmanager::ReadResourceRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	responseMessage.set_value(Spy::getInstance()->getCycleDurationMs());

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool ServantManagerResponder::processGetPropertiesSavedInBox(const std::string & message, std::string & response) {

	servantmanager::GetPropertiesSavedInBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	vector<unsigned long> list = m_servantAccessor->getPropertiesSavedInBox(messageRequest.databaseid());

	common::Int32ArrayResponse responseMessage;

	// Copy the array elements
	vector<unsigned long>::const_iterator it;
	for (it = list.begin(); it != list.end(); it++) {
		responseMessage.add_value(*it);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}


bool ServantManagerResponder::processGetPropertySavedInBoxValue(const std::string & message, std::string & response) {
	servantmanager::GetPropertySavedInBoxValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	try {
		string value = SaveBoxPropertyManager::getInstance()->getPropertySavedInBoxValue(messageRequest.propertyid());
		responseMessage.set_value(value);

	} catch (const Exception& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}


bool ServantManagerResponder::processSetPropertySavedInBoxValue(const std::string & message, std::string & response) {
	servantmanager::SetPropertySavedInBoxValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	try {
		SaveBoxPropertyManager::getInstance()->setPropertySavedInBoxValue(messageRequest.propertyid(), messageRequest.value());

	} catch (const NoSuchFile& e) {
		responseMessage.set_error(common::Error::NO_SUCH_RESOURCE);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}


void ServantManagerResponder::loop() {

	// Initialize the map of functions.
	initProcessFunctions();

	// Loop until cancel is done.
	while (true) {
		// Wait for a new request.
		unique_ptr<coms::basic::Request> request = m_responder->receive();

		// Stop the loop if cancel is called.
		if (request.get() == 0) {
			break;
		}

		// Get the request type.
		string firstPart = request->get();

		// Get the request message.
		string secondPart = request->getSecondPart();

		servantmanager::Message messageType;
		messageType.ParseFromString(firstPart);

		string response;
		bool onError = false;

		// Get the process function and apply it.
		ProcessFunction & processFunction = m_processFunctions[messageType.type()];

		if (processFunction.functionPtr != nullptr) {
			onError = (this->*processFunction.functionPtr)(secondPart, response);
		}
		else {
			cerr << "Servant manager responder cannot process " << messageType.type() << endl;
		}

		request->reply(response);

		// Check error.
		if (!onError) {

			// Notify the clients in case the function modifies the command zone.
			if (processFunction.activity) {
				// Notify the token activity.
				SessionManager::getInstance()->notifyTokenActivity();
			}
		}
	}
}


}
