/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "SessionResponder.h"
#include <CommonResponses.pb.h>
#include <SessionRequests.pb.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>

using namespace std;
using namespace cameo;
using namespace common;

namespace core {

SessionResponder::SessionResponder() {
}

SessionResponder::~SessionResponder() {
}

void SessionResponder::init() {

	// Create the responder.
	m_responder = coms::basic::Responder::create("session");
	m_responder->init();

	// Start the thread.
	m_thread.reset(new std::thread(std::bind(&SessionResponder::loop, this)));

	cout << "Initialised SessionResponder" << endl;
}

void SessionResponder::stop() {

	// Cancel the responder so that unblocks the waiting thread.
	m_responder->cancel();

	// Join the thread.
	m_thread->join();

	cout << "Stopped SessionResponder" << endl;
}

void SessionResponder::initProcessFunctions() {

	m_processFunctions[session::Message::Login] = &SessionResponder::processLogin;
	m_processFunctions[session::Message::LoginRefused] = &SessionResponder::processLoginRefused;
	m_processFunctions[session::Message::EnableRemoteControl] = &SessionResponder::processEnableRemoteControlRequest;
	m_processFunctions[session::Message::DisableRemoteControl] = &SessionResponder::processDisableRemoteControlRequest;
	m_processFunctions[session::Message::GetToken] = &SessionResponder::processGetTokenRequest;
	m_processFunctions[session::Message::ReleaseToken] = &SessionResponder::processReleaseTokenRequest;
	m_processFunctions[session::Message::Sync] = &SessionResponder::processSync;
	m_processFunctions[session::Message::Chat] = &SessionResponder::processChatMessage;
	m_processFunctions[session::Message::CrashMail] = &SessionResponder::processCrashMailRequest;
	m_processFunctions[session::Message::ConnectedUsers] = &SessionResponder::processConnectedUsersRequest;
}

void SessionResponder::processLogin(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	session::LoginRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// The request should contain a user but it is optional.
	// Forwards the call to the session manager.
	int result = 0;

	if (messageRequest.has_user()) {
		result = SessionManager::getInstance()->login(request, messageRequest.clienttype(), messageRequest.user());
	}
	else {
		result = SessionManager::getInstance()->login(request, messageRequest.clienttype());
	}

	// Add the application if it is accepted.
	if (result > 0) {
		// Assign the next id.
		responseMessage.set_value(result);
	}
	else {
		// Process the error.
		if (result == SessionManager::MULTIPLE_CLIENTS_PER_ENDPOINT_NOT_ALLOWED) {
			responseMessage.set_error(common::Error::MULTIPLE_CLIENTS_PER_ENDPOINT_NOT_ALLOWED);
		}
		else if (result == SessionManager::MULTIPLE_CONNECTIONS_PER_USER_NOT_ALLOWED) {
			responseMessage.set_error(common::Error::MULTIPLE_CONNECTIONS_PER_USER_NOT_ALLOWED);
		}
		else {
			responseMessage.set_error(common::Error::CONNECTION_REFUSED);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}

void SessionResponder::processLoginRefused(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	session::LoginRefusedRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Forwards to the session manager.
	if (messageRequest.has_user()) {
		SessionManager::getInstance()->loginRefused(request, messageRequest.reason(), messageRequest.user());
	}
	else {
		SessionManager::getInstance()->loginRefused(request, messageRequest.reason());
	}

	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}


void SessionResponder::processEnableRemoteControlRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	session::RemoteControlRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
		responseMessage.set_value(false);
	}
	else {
		// Forwards to the session manager.
		SessionManager::getInstance()->enableRemoteControl(request, messageRequest.clientid(), messageRequest.user());

		responseMessage.set_value(true);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}

void SessionResponder::processDisableRemoteControlRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	session::RemoteControlRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
		responseMessage.set_value(false);
	}
	else {
		// Forwards to the session manager.
		SessionManager::getInstance()->disableRemoteControl(request, messageRequest.clientid(), messageRequest.user());

		responseMessage.set_value(true);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}

void SessionResponder::processGetTokenRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	session::GetTokenRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
		responseMessage.set_value(false);
	}
	else {
		// Forwards to the session manager.
		SessionManager::getInstance()->requestToken(request, messageRequest.clientid(), messageRequest.user());

		responseMessage.set_value(true);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}

void SessionResponder::processReleaseTokenRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	session::ReleaseTokenRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
		responseMessage.set_value(false);
	}
	else {
		// Forwards to the session manager.
		SessionManager::getInstance()->releaseToken(request, messageRequest.clientid(), messageRequest.user());

		responseMessage.set_value(true);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}

void SessionResponder::processSync(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	common::BooleanResponse responseMessage;

	// Forwards to the session manager.
	if (SessionManager::getInstance()) {
		SessionManager::getInstance()->notifyTokenState();
	}

	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}

void SessionResponder::processChatMessage(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	session::ChatMessage messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
		responseMessage.set_value(false);
	}
	else {
		// Forwards to the session manager.
		SessionManager::getInstance()->sendChatMessage(request, messageRequest.clientid(), messageRequest.sender(), messageRequest.message(), messageRequest.filename());

		responseMessage.set_value(true);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}

void SessionResponder::processCrashMailRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	common::BooleanResponse responseMessage;

	session::CrashMailRequest messageRequest;
	messageRequest.ParseFromString(message);

	// Forwards to the session manager.
	SessionManager::getInstance()->sendCrashMail(request, messageRequest.content());

	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}

void SessionResponder::processConnectedUsersRequest(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	session::ConnectedUsersResponse responseMessage;

	vector<SessionManager::ConnectedUser> connectedUsers = SessionManager::getInstance()->getConnectedUsers();

	for (auto u : connectedUsers) {
		session::ConnectedUser* user = responseMessage.add_users();
		user->set_clientid(u.id);
		user->set_user(u.user);
		user->set_clienttype(u.clientType);
		user->set_endpoint(u.endpoint);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);
}

void SessionResponder::loop() {

	// Initialize the map of functions.
	initProcessFunctions();

	// Loop until cancel is done.
	while (true) {
		// Wait for a new request.
		unique_ptr<coms::basic::Request> request = m_responder->receive();

		// Stop the loop if cancel is called.
		if (request.get() == 0) {
			break;
		}

		// Get the request type.
		string firstPart = request->get();

		// Get the request message.
		string secondPart = request->getSecondPart();

		session::Message messageType;
		messageType.ParseFromString(firstPart);

		string response;

		// Get the process function and apply it.
		ProcessFunctionType functionPointer = m_processFunctions[messageType.type()];

		if (functionPointer != 0) {
			(this->*functionPointer)(request.get(), secondPart, response);
		}
		else {
			cerr << "Session manager responder cannot process " << messageType.type() << endl;
		}

		request->reply(response);
	}
}

}
