/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "SurveyResponder.h"
#include <SurveyRequests.pb.h>
#include <CommonResponses.pb.h>
#include <common/output/OutputStream.h>
#include <core/survey/Survey.h>
#include <core/session/SessionManager.h>

using namespace std;
using namespace cameo;
using namespace common;

namespace core {

SurveyResponder::SurveyResponder() {
}

SurveyResponder::~SurveyResponder() {
}

void SurveyResponder::init() {

	// Create the responder.
	m_responder = coms::basic::Responder::create("survey");
	m_responder->init();

	// Start the thread.
	m_thread.reset(new std::thread(std::bind(&SurveyResponder::loop, this)));

	cout << "Initialised SurveyResponder" << endl;
}

void SurveyResponder::stop() {

	// Cancel the responder so that unblocks the waiting thread.
	m_responder->cancel();

	// Join the thread.
	m_thread->join();

	cout << "Stopped SurveyResponder" << endl;
}

void SurveyResponder::initProcessFunctions() {

	m_processFunctions[::survey::Message::AddProperty] = ProcessFunction(&SurveyResponder::processAddProperty, true);
	m_processFunctions[::survey::Message::RemoveProperty] = ProcessFunction(&SurveyResponder::processRemoveProperty, true);
	m_processFunctions[::survey::Message::SetActiveWriteTime] = ProcessFunction(&SurveyResponder::processSetActiveWriteTime, true);
	m_processFunctions[::survey::Message::GetProperties] = ProcessFunction(&SurveyResponder::processGetProperties, false);
}

bool SurveyResponder::processAddProperty(const std::string & message, std::string & response) {

	::survey::PropertyRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		bool result = core::Survey::getInstance()->addElement(messageRequest.controllername(), messageRequest.propertyname(), messageRequest.index(), {});

		if (!result) {
			responseMessage.set_error(common::Error::PROPERTY_EXISTS);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool SurveyResponder::processRemoveProperty(const std::string & message, std::string & response) {

	::survey::PropertyRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		bool result = core::Survey::getInstance()->removeElement(messageRequest.controllername(), messageRequest.propertyname(), messageRequest.index());

		if (!result) {
			responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool SurveyResponder::processSetActiveWriteTime(const std::string & message, std::string & response) {

	::survey::SetActiveWriteTimeRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		bool result = core::Survey::getInstance()->setActiveWriteTime(messageRequest.controllername(), messageRequest.propertyname(), messageRequest.index(), messageRequest.seconds());

		if (!result) {
			responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool SurveyResponder::processGetProperties(const std::string & message, std::string & response) {

	::survey::EmptyRequest messageRequest;
	messageRequest.ParseFromString(message);

	::survey::GetPropertiesResponse responseMessage;

	auto elements = core::Survey::getInstance()->getElements();

	for (auto const& e : elements) {
		std::string controllerName = std::get<0>(e);
		std::string propertyName = std::get<1>(e);
		int index = std::get<2>(e);
		int seconds = std::get<3>(e);

		::survey::PropertyResponse* property = responseMessage.add_properties();
		property->set_controllername(controllerName);
		property->set_propertyname(propertyName);
		property->set_index(index);
		property->set_seconds(seconds);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

void SurveyResponder::loop() {

	// Initialize the map of functions.
	initProcessFunctions();

	// Loop until cancel is done.
	while (true) {
		// Wait for a new request.
		unique_ptr<coms::basic::Request> request = m_responder->receive();

		// Stop the loop if cancel is called.
		if (request.get() == 0) {
			break;
		}

		// Get the request type.
		string firstPart = request->get();

		// Get the request message.
		string secondPart = request->getSecondPart();

		::survey::Message messageType;
		messageType.ParseFromString(firstPart);

		string response;
		bool onError = false;

		// Get the process function and apply it.
		ProcessFunction & processFunction = m_processFunctions[messageType.type()];

		if (processFunction.functionPtr != nullptr) {
			onError = (this->*processFunction.functionPtr)(secondPart, response);
		}
		else {
			cerr << "Survey responder cannot process " << messageType.type() << endl;
		}

		request->reply(response);

		// Check error.
		if (!onError) {
			// Notify the clients in case the function modifies the command zone.
			if (processFunction.activity) {
				// Notify the token activity.
				SessionManager::getInstance()->notifyTokenActivity();
			}
		}
	}
}

}
