/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "UserVariableManagerResponder.h"
#include <CommonResponses.pb.h>
#include <VariableManagerRequests.pb.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <core/commandzone/CommandZone.h>
#include <core/variables/UserVariableManager.h>
#include <core/variables/UserVariable.h>
#include <core/session/SessionManager.h>

using namespace std;
using namespace cameo;
using namespace common;

namespace core {

UserVariableManagerResponder::UserVariableManagerResponder() {
}

UserVariableManagerResponder::~UserVariableManagerResponder() {
}

void UserVariableManagerResponder::init() {

	// Create the responder.
	m_responder = coms::basic::Responder::create("variable_manager");
	m_responder->init();

	// Start the thread.
	m_thread.reset(new std::thread(std::bind(&UserVariableManagerResponder::loop, this)));

	cout << "Initialised VariableManagerResponder" << endl;
}

void UserVariableManagerResponder::stop() {

	// Cancel the responder so that unblocks the waiting thread.
	m_responder->cancel();

	// Join the thread.
	m_thread->join();

	cout << "Stopped VariableManagerResponder" << endl;
}

void UserVariableManagerResponder::initProcessFunctions() {

	m_processFunctions[::variable::Message::VerifyExpression] = ProcessFunction(&UserVariableManagerResponder::processVerifyExpression, false);

	m_processFunctions[::variable::Message::AddVariable] = ProcessFunction(&UserVariableManagerResponder::processAddVariable, true);
	m_processFunctions[::variable::Message::EditVariable] = ProcessFunction(&UserVariableManagerResponder::processEditVariable, true);
	m_processFunctions[::variable::Message::RemoveVariable] = ProcessFunction(&UserVariableManagerResponder::processRemoveVariable, true);

	m_processFunctions[::variable::Message::GetVariableValue] = ProcessFunction(&UserVariableManagerResponder::processGetVariableValue, false);
	m_processFunctions[::variable::Message::GetVariablesXMLContent] = ProcessFunction(&UserVariableManagerResponder::processGetVariablesXMLContent, false);
	m_processFunctions[::variable::Message::GetUsableVariableList] = ProcessFunction(&UserVariableManagerResponder::processGetUsableVariableListValue, false);
}

bool UserVariableManagerResponder::processVerifyExpression(const std::string & message, std::string & response) {

	::variable::VerifyExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::VariableResponse responseMessage;

	try {
		core::UserVariableManager::getInstance()->verifyExpression(messageRequest.expression(), messageRequest.name());
	}
	catch (core::UserVariableManager::InvalidExpressionException const & e) {
		responseMessage.set_error(common::Error::INVALID_EXPRESSION);
		responseMessage.set_message(e.getMessage());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool UserVariableManagerResponder::processAddVariable(const std::string & message, std::string & response) {

	::variable::AddVariableRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::VariableResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::UserVariableManager::getInstance()->addUserVariable(messageRequest.name(), messageRequest.expression(), false);
		}
		catch (core::UserVariableManager::UserVariableExistsException const & e) {
			responseMessage.set_error(common::Error::VARIABLE_EXISTS);
		}
		catch (core::UserVariableManager::InvalidExpressionException const & e) {
			responseMessage.set_error(common::Error::INVALID_EXPRESSION);
			responseMessage.set_message(e.getMessage());
		}
		catch (core::UserVariableManager::DisabledUserVariableException const & e) {
			responseMessage.set_error(common::Error::DISABLED_VARIABLE);
			responseMessage.set_message(e.getMessage());
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool UserVariableManagerResponder::processEditVariable(const std::string & message, std::string & response) {

	::variable::EditVariableRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::VariableResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			core::UserVariableManager::getInstance()->editUserVariable(messageRequest.name(), messageRequest.expression());
		}
		catch (core::UserVariableManager::NoSuchUserVariableException const & e) {
			responseMessage.set_error(common::Error::NO_SUCH_VARIABLE);
		}
		catch (core::UserVariableManager::InvalidExpressionException const & e) {
			responseMessage.set_error(common::Error::INVALID_EXPRESSION);
			responseMessage.set_message(e.getMessage());
		}
		catch (core::UserVariableManager::DisabledUserVariableException const & e) {
			responseMessage.set_error(common::Error::DISABLED_VARIABLE);
			responseMessage.set_message(e.getMessage());
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool UserVariableManagerResponder::processRemoveVariable(const std::string & message, std::string & response) {

	::variable::RemoveVariableRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::VariableResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Check if the variable is referenced in the command zone.
		if (core::CommandZone::getInstance()->referencesUserVariable(messageRequest.name())) {
			responseMessage.set_error(common::Error::UNREMOVABLE_VARIABLE);
			responseMessage.set_message("Variable is referenced in the Launch Pad");
		}
		else {
			// Remove the user variable.
			try {
				core::UserVariableManager::getInstance()->removeUserVariable(messageRequest.name());
			}
			catch (core::UserVariableManager::NoSuchUserVariableException const & e) {
				responseMessage.set_error(common::Error::NO_SUCH_VARIABLE);
			}
			catch (core::UserVariableManager::UnremovableUserVariableException const & e) {
				responseMessage.set_error(common::Error::UNREMOVABLE_VARIABLE);
				responseMessage.set_message(e.getMessage());
			}
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool UserVariableManagerResponder::processGetVariableValue(const std::string & message, std::string & response) {

	::variable::GetVariableValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Float64Response responseMessage;

	try {
		double value = core::UserVariableManager::getInstance()->getUserVariable(messageRequest.name())->getValue();
		responseMessage.set_value(value);
	}
	catch (core::UserVariableManager::NoSuchUserVariableException const & e) {
		responseMessage.set_error(common::Error::NO_SUCH_VARIABLE);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool UserVariableManagerResponder::processGetVariablesXMLContent(const std::string & message, std::string & response) {

	::variable::GetVariablesXMLContentRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	responseMessage.set_value(core::UserVariableManager::getInstance()->toXMLString(true));

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

bool UserVariableManagerResponder::processGetUsableVariableListValue(const std::string & message, std::string & response) {

	::variable::GetUsableVariableListRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::VariableListResponse responseMessage;

	std::set<std::string> usableVariableNames = core::UserVariableManager::getInstance()->getUsableUserVariableNames(messageRequest.name());

	for (auto n : usableVariableNames) {
		responseMessage.add_name(n);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	return false;
}

void UserVariableManagerResponder::loop() {

	// Initialize the map of functions.
	initProcessFunctions();

	// Loop until cancel is done.
	while (true) {
		// Wait for a new request.
		unique_ptr<coms::basic::Request> request = m_responder->receive();

		// Stop the loop if cancel is called.
		if (request.get() == 0) {
			break;
		}

		// Get the request type.
		string firstPart = request->get();

		// Get the request message.
		string secondPart = request->getSecondPart();

		::variable::Message messageType;
		messageType.ParseFromString(firstPart);

		string response;
		bool onError = false;

		// Get the process function and apply it.
		ProcessFunction & processFunction = m_processFunctions[messageType.type()];

		if (processFunction.functionPtr != nullptr) {
			onError = (this->*processFunction.functionPtr)(secondPart, response);
		}
		else {
			cerr << "User variable manager responder cannot process " << messageType.type() << endl;
		}

		request->reply(response);

		// Check error.
		if (!onError) {
			// Notify the clients in case the function modifies the command zone.
			if (processFunction.activity) {
				// Notify the token activity.
				SessionManager::getInstance()->notifyTokenActivity();
			}
		}
	}
}

}
