/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandLineResponder.h"
#include <CommonResponses.pb.h>
#include <CommandLineRequests.pb.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <ics/CommandSystem/CommandLine/CommandLineManager.h>
#include <core/session/SessionManager.h>
#include <string>

using namespace std;
using namespace boost;
using namespace cameo;
using namespace common;

namespace core {

CommandLineResponder::CommandLineResponder() {
}

CommandLineResponder::~CommandLineResponder() {
}

void CommandLineResponder::init() {

	// Create the responder.
	m_responder = coms::basic::Responder::create("command_line");
	m_responder->init();

	// Start the thread.
	m_thread.reset(new std::thread(std::bind(&CommandLineResponder::loop, this)));

	cout << "Initialised CommandLineResponder" << endl;
}

void CommandLineResponder::stop() {

	// Cancel the responder so that unblocks the waiting thread.
	m_responder->cancel();

	// Join the thread.
	m_thread->join();

	cout << "Stopped CommandLineResponder" << endl;
}

void CommandLineResponder::initProcessFunctions() {

	m_processFunctions[commandline::Message::ExecuteCommand] = ProcessFunction(&CommandLineResponder::processExecuteCommand, true);
	m_processFunctions[commandline::Message::PauseCommands] = ProcessFunction(&CommandLineResponder::processPauseCommands, true);
	m_processFunctions[commandline::Message::RestartCommands] = ProcessFunction(&CommandLineResponder::processRestartCommands, true);
	m_processFunctions[commandline::Message::StopCommands] = ProcessFunction(&CommandLineResponder::processStopCommands, true);
	m_processFunctions[commandline::Message::StopAtEndCommands] = ProcessFunction(&CommandLineResponder::processStopAtEndCommands, true);
	m_processFunctions[commandline::Message::Quit] = ProcessFunction(&CommandLineResponder::processQuit, true);

	m_processFunctions[commandline::Message::IsStarted] = ProcessFunction(&CommandLineResponder::processIsStarted, false);
	m_processFunctions[commandline::Message::IsPaused] = ProcessFunction(&CommandLineResponder::processIsPaused, false);

	m_processFunctions[commandline::Message::SetCommandLineState] = ProcessFunction(&CommandLineResponder::processSetCommandLineState, true);

	m_processFunctions[commandline::Message::GetParserType] = ProcessFunction(&CommandLineResponder::processGetParserType, false);

	m_processFunctions[commandline::Message::SimulateCommand] = ProcessFunction(&CommandLineResponder::processSimulateCommand, true);
}

bool CommandLineResponder::processExecuteCommand(const std::string & message, std::string & response) {

	commandline::CommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandLineManager::getInstance()->executeCommand(messageRequest.text(), false);
	}

	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool CommandLineResponder::processPauseCommands(const std::string & message, std::string & response) {

	commandline::CommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandLineManager::getInstance()->pauseCommands();
	}

	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool CommandLineResponder::processRestartCommands(const std::string & message, std::string & response) {

	commandline::CommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandLineManager::getInstance()->restartCommands();
	}

	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool CommandLineResponder::processStopCommands(const std::string & message, std::string & response) {

	commandline::CommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandLineManager::getInstance()->stopCommands();
	}

	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool CommandLineResponder::processStopAtEndCommands(const std::string & message, std::string & response) {

	commandline::CommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandLineManager::getInstance()->stopAtEndCommands();
	}

	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool CommandLineResponder::processQuit(const std::string & message, std::string & response) {

	commandline::CommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandLineManager::getInstance()->reset();
	}

	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool CommandLineResponder::processIsStarted(const std::string & message, std::string & response) {

	common::BooleanResponse responseMessage;

	responseMessage.set_value(CommandLineManager::getInstance()->isStarted());

	responseMessage.SerializeToString(&response);

	return false;
}

bool CommandLineResponder::processIsPaused(const std::string & message, std::string & response) {

	common::BooleanResponse responseMessage;

	responseMessage.set_value(CommandLineManager::getInstance()->isPaused());

	responseMessage.SerializeToString(&response);

	return false;
}

bool CommandLineResponder::processSetCommandLineState(const std::string & message, std::string & response) {

	commandline::CommandLineStateRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandLineManager::getInstance()->setCommandLineState(messageRequest.state());
	}

	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

bool CommandLineResponder::processGetParserType(const std::string & message, std::string & response) {

	common::StringResponse responseMessage;

	responseMessage.set_value(CommandLineManager::getInstance()->getParserType());

	responseMessage.SerializeToString(&response);

	return false;
}

bool CommandLineResponder::processSimulateCommand(const std::string & message, std::string & response) {

	commandline::CommandRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandLineManager::getInstance()->simulateCommand(messageRequest.text());
	}

	responseMessage.SerializeToString(&response);

	return responseMessage.has_error();
}

void CommandLineResponder::loop() {

	// Initialize the map of functions.
	initProcessFunctions();

	// Loop until cancel is done.
	while (true) {
		// Wait for a new request.
		unique_ptr<coms::basic::Request> request = m_responder->receive();

		// Stop the loop if cancel is called.
		if (request.get() == 0) {
			break;
		}

		// Get the request type.
		string firstPart = request->get();

		// Get the request message.
		string secondPart = request->getSecondPart();

		commandline::Message messageType;
		messageType.ParseFromString(firstPart);

		string response;
		bool onError = false;

		// Get the process function and apply it.
		ProcessFunction & processFunction = m_processFunctions[messageType.type()];

		if (processFunction.functionPtr != nullptr) {
			onError = (this->*processFunction.functionPtr)(secondPart, response);
		}
		else {
			cerr << "Command line responder cannot process " << messageType.type() << endl;
		}

		request->reply(response);

		// Check error.
		if (!onError) {

			// Notify the clients in case the function modifies the command zone.
			if (processFunction.activity) {

				// Notify the token activity.
				SessionManager::getInstance()->notifyTokenActivity();
			}
		}
	}
}

}
