/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <core/message/commandzone/CommandBoxExecutionEvent.h>
#include <core/message/commandzone/CommandBoxProgressionEvent.h>
#include <core/message/commandzone/CommandBoxResetEvent.h>
#include <core/message/commandzone/CommandBoxPropertyEvent.h>
#include <core/message/commandzone/CommandZoneExecutionEvent.h>
#include <core/message/commandzone/CommandZoneProgressionEvent.h>
#include <core/message/commandzone/CommandZoneResetEvent.h>
#include <core/message/commandzone/CommandZoneErrorEvent.h>
#include <core/message/commandzone/ForLoopCommandBoxCurrentValueEvent.h>
#include <core/message/commandzone/CommandBoxExpressionChangedEvent.h>
#include <core/message/commandzone/CommandBoxExpressionValueChangedEvent.h>
#include <core/message/commandzone/CommandBoxesDeletedEvent.h>
#include <core/message/commandzone/CommandZoneSyncEvent.h>
#include <core/message/commandzone/NewEstimatedTimeMessage.h>
#include <core/commandzone/CommandZone.h>
#include <core/transportlayer/core/commandzone/CommandZoneEventSubscriber.h>
#include <NotificationMessages.pb.h>
#include <iostream>
#include <vector>
#include "CommandZoneEventPublisher.h"

using namespace std;
using namespace common;
using namespace cameo;

namespace core {

unsigned long CommandZoneEventPublisher::_counter = 0;

CommandZoneEventPublisher::CommandZoneEventPublisher() {

	string commandZonePublisherName = "command_zone_publisher";

	// Create the publisher.
	m_publisher = coms::Publisher::create(commandZonePublisherName);
	m_publisher->init();

	cout << "Created command zone publisher " << commandZonePublisherName << endl;
}

CommandZoneEventPublisher::~CommandZoneEventPublisher() {

	cout << "Deleted command zone publisher" << endl;
}

unsigned long CommandZoneEventPublisher::subscribeDirect(CommandZoneEventSubscriber* subscriber) {

	std::scoped_lock lock(_mutex);

	unsigned long subscriberID = subscriber->getID();

	// check to see if subscriber already exists
	if (_subscribers.find(subscriberID) != _subscribers.end()) {
		delete subscriber;
		return subscriberID;
	}

	unsigned long nextID = ++_counter;
	_subscribers[nextID] = subscriber;

	return nextID;
}

void CommandZoneEventPublisher::unsubscribeDirect(unsigned long subscriberID) {
	
	std::scoped_lock lock(_mutex);
	
	if (_subscribers.find(subscriberID) != _subscribers.end()) {
		// do not delete the subscriber (responsibility of its own)
		_subscribers.erase(subscriberID);
	}
}

void CommandZoneEventPublisher::notify(CommandZoneEvent* commandZoneEvent) {

	{
		std::scoped_lock lock(_mutex);

		// Notify all subscribers
		std::map<unsigned long, CommandZoneEventSubscriber*>::iterator it;
		for (it = _subscribers.begin(); it != _subscribers.end(); it++) {
			it->second->notify(commandZoneEvent);
		}
	}

	CommandZoneEvent::EventType eventType = commandZoneEvent->getEventType();

	if (eventType == CommandZoneEvent::COMMAND_BOX_EXECUTION) {
		notify(static_cast<CommandBoxExecutionEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_BOX_PROGRESSION) {
		notify(static_cast<CommandBoxProgressionEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_BOX_RESET) {
		notify(static_cast<CommandBoxResetEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_BOX_PROPERTY) {
		notify(static_cast<CommandBoxPropertyEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_ZONE_EXECUTION) {
		notify(static_cast<CommandZoneExecutionEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_ZONE_PROGRESSION) {
		notify(static_cast<CommandZoneProgressionEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_ZONE_RESET) {
		notify(static_cast<CommandZoneResetEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_ZONE_ERROR) {
		notify(static_cast<CommandZoneErrorEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::FORLOOP_CURRENT_VALUE_CHANGED) {
		notify(static_cast<ForLoopCommandBoxCurrentValueEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_BOX_EXPRESSION_CHANGED) {
		notify(static_cast<CommandBoxExpressionChangedEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_BOX_EXPRESSION_VALUE_CHANGED) {
		notify(static_cast<CommandBoxExpressionValueChangedEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_BOXES_DELETED) {
		notify(static_cast<CommandBoxesDeletedEvent*>(commandZoneEvent));

	} else if (eventType == CommandZoneEvent::COMMAND_ZONE_SYNC) {
		notify(static_cast<CommandZoneSyncEvent*>(commandZoneEvent));
	}
}

void CommandZoneEventPublisher::notify(CommandBoxExecutionEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;


	if (commandZoneEvent->getExecutionState() == CommandBoxExecutionEvent::STARTED) {
//		cout << "Notify CommandBoxStarted " << commandZoneEvent->getCommandBoxID() << endl;
		messageType.set_type(notification::Message::CommandBoxStarted);

	} else if (commandZoneEvent->getExecutionState() == CommandBoxExecutionEvent::PAUSED) {
//		cout << "Notify CommandBoxPaused " << commandZoneEvent->getCommandBoxID() << endl;
		messageType.set_type(notification::Message::CommandBoxPaused);

	} else if (commandZoneEvent->getExecutionState() == CommandBoxExecutionEvent::STOPPING_AT_END) {
//		cout << "Notify CommandBoxStoppingAtEnd " << commandZoneEvent->getCommandBoxID() << endl;
		messageType.set_type(notification::Message::CommandBoxStoppingAtEnd);

	} else if (commandZoneEvent->getExecutionState() == CommandBoxExecutionEvent::PAUSING_AT_END) {
//		cout << "Notify CommandBoxPausingAtEnd " << commandZoneEvent->getCommandBoxID() << endl;
		messageType.set_type(notification::Message::CommandBoxPausingAtEnd);

	} else {
//		cout << "Notify CommandBoxTerminated " << commandZoneEvent->getCommandBoxID() << endl;
		messageType.set_type(notification::Message::CommandBoxTerminated);
	}

	// Create and send the message.
	notification::CommandBoxChanged message;

	message.set_databaseid(CommandZone::getInstance()->getID());
	message.set_commandboxid(commandZoneEvent->getCommandBoxID());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandBoxProgressionEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandBoxProgressionChanged);

	// Create and send the message.
	notification::CommandBoxProgressionChanged message;

	message.set_databaseid(CommandZone::getInstance()->getID());
	message.set_commandboxid(commandZoneEvent->getCommandBoxID());
	message.set_progression(commandZoneEvent->getProgression());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandBoxResetEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandBoxReset);

	// Create and send the message.
	notification::CommandBoxChanged message;

	message.set_databaseid(CommandZone::getInstance()->getID());
	message.set_commandboxid(commandZoneEvent->getCommandBoxID());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandBoxPropertyEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandBoxPropertyChanged);

	// Create and send the message.
	notification::CommandBoxPropertyChanged message;

	message.set_databaseid(CommandZone::getInstance()->getID());
	message.set_commandboxid(commandZoneEvent->getCommandBoxID());
	message.set_propertyid(commandZoneEvent->getPropertyID());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandZoneExecutionEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	if (commandZoneEvent->getExecutionState() == CommandZoneExecutionEvent::STARTED) {
//		cout << "Notify CommandZoneStarted" << endl;
		messageType.set_type(notification::Message::CommandZoneStarted);

	} else if (commandZoneEvent->getExecutionState() == CommandZoneExecutionEvent::PAUSED) {
//		cout << "Notify CommandZonePaused" << endl;
		messageType.set_type(notification::Message::CommandZonePaused);

	} else if (commandZoneEvent->getExecutionState() == CommandZoneExecutionEvent::STOPPING_AT_END) {
//		cout << "Notify CommandZoneStoppingAtEnd" << endl;
		messageType.set_type(notification::Message::CommandZoneStoppingAtEnd);

	} else if (commandZoneEvent->getExecutionState() == CommandZoneExecutionEvent::PAUSING_AT_END) {
//		cout << "Notify CommandZonePausingAtEnd" << endl;
		messageType.set_type(notification::Message::CommandZonePausingAtEnd);

	} else {
//		cout << "Notify CommandZoneTerminated" << endl;
		messageType.set_type(notification::Message::CommandZoneTerminated);
	}

	// Create and send the message.
	notification::CommandZoneChanged message;

	message.set_commandzoneid(commandZoneEvent->getCommandZoneID());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandZoneProgressionEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandZoneProgressionChanged);

	// Create and send the message.
	notification::CommandZoneProgressionChanged message;

	message.set_commandzoneid(commandZoneEvent->getCommandZoneID());
	message.set_progression(commandZoneEvent->getProgression());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandZoneResetEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandZoneReset);

	// Create and send the message.
	notification::CommandZoneChanged message;

	message.set_commandzoneid(commandZoneEvent->getCommandZoneID());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandZoneErrorEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandZoneErrorOccurred);

	// Create and send the message.
	notification::CommandZoneErrorOccurred message;

	message.set_commandzoneid(commandZoneEvent->getCommandZoneID());
	message.set_message(commandZoneEvent->getMessage());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(ForLoopCommandBoxCurrentValueEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::ForLoopCurrentValueChanged);

	// Create and send the message.
	notification::ForLoopCurrentValueChanged message;

	message.set_databaseid(CommandZone::getInstance()->getID());
	message.set_commandboxid(commandZoneEvent->getCommandBoxID());
	message.set_index(commandZoneEvent->getIndex());
	message.set_currentvalue(commandZoneEvent->getCurrentValue());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandBoxExpressionChangedEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandBoxExpressionChanged);

	// Create and send the message.
	notification::CommandBoxChanged message;

	message.set_databaseid(CommandZone::getInstance()->getID());
	message.set_commandboxid(commandZoneEvent->getCommandBoxID());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandBoxExpressionValueChangedEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandBoxExpressionValueChanged);

	// Create and send the message.
	notification::CommandBoxExpressionValueChanged message;

	message.set_databaseid(CommandZone::getInstance()->getID());
	message.set_commandboxid(commandZoneEvent->getCommandBoxID());
	message.set_currentvalue(commandZoneEvent->getCurrentValue());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandBoxesDeletedEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandBoxesDeleted);

	// Create and send the message.
	notification::CommandBoxesDeleted message;

	message.set_commandzoneid(CommandZone::getInstance()->getID());

	for (unsigned long id : commandZoneEvent->getCommandBoxIDs()) {
		message.add_commandboxid(id);
	}

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void CommandZoneEventPublisher::notify(CommandZoneSyncEvent* commandZoneEvent) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::CommandZoneSync);

	// Create and send the message.
	notification::CommandZoneSync message;

	message.set_serializedtype(commandZoneEvent->getSerializedMessageType());
	message.set_serializedcontent(commandZoneEvent->getSerializedMessageContent());
	message.set_serializedresponse(commandZoneEvent->getSerializedResponse());

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

}
