/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandZoneResponder.h"
#include <CommonResponses.pb.h>
#include <CommandZoneRequests.pb.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <core/commandzone/CommandBox.h>
#include <core/commandzone/AtomicCommandBox.h>
#include <core/commandzone/generic/VariableCommandBox.h>
#include <core/commandzone/control/IfCommandBox.h>
#include <core/commandzone/control/ControlCommandBox.h>
#include <core/commandzone/control/BreakCommandBox.h>
#include <core/commandzone/ForLoopCommandBox.h>
#include <core/commandzone/ScanCommandBox.h>
#include <core/commandzone/ModifiableCommandBoxContainer.h>
#include <core/commandzone/CommandBoxMap.h>
#include <core/commandzone/CommandBoxFactory.h>
#include <core/commandzone/CommandZone.h>
#include <core/commandzone/CommandZoneEventServer.h>
#include <core/commandzone/EstimatedTimeManager.h>
#include <core/session/SessionManager.h>
#include <ics/PersistenceServices/ClientFiles/FileManager.h>
#include "../../../commandzone/generic/PropertyGenericCommandBox.h"
#include <core/commandzone/generic/GenericCommandBoxManager.h>
#include <core/commandzone/generic/GenericCommandBox.h>

using namespace std;
using namespace boost;
using namespace cameo;
using namespace common;

namespace core {

CommandZoneResponder::CommandZoneResponder() {
}

CommandZoneResponder::~CommandZoneResponder() {
}

void CommandZoneResponder::init() {

	// Create the responder.
	m_responder = coms::basic::Responder::create("command_box");
	m_responder->init();

	// Start the thread.
	m_thread.reset(new std::thread(std::bind(&CommandZoneResponder::loop, this)));

	cout << "Initialised CommandZoneResponder" << endl;
}

void CommandZoneResponder::stop() {

	// Cancel the responder so that unblocks the waiting thread.
	m_responder->cancel();

	// Join the thread.
	m_thread->join();

	cout << "Stopped CommandZoneResponder" << endl;
}

void CommandZoneResponder::initProcessFunctions() {

	m_processFunctions[commandzone::Message::ToggleBackground] = ProcessFunction(&CommandZoneResponder::processToggleBackground, true, true);
	m_processFunctions[commandzone::Message::ToggleParallel] = ProcessFunction(&CommandZoneResponder::processToggleParallel, true, true);
	m_processFunctions[commandzone::Message::IsBackground] = ProcessFunction(&CommandZoneResponder::processIsBackground, false, false);
	m_processFunctions[commandzone::Message::IsParallel] = ProcessFunction(&CommandZoneResponder::processIsParallel, false, false);
	m_processFunctions[commandzone::Message::GetExecutionState] = ProcessFunction(&CommandZoneResponder::processGetExecutionState, false, false);
	m_processFunctions[commandzone::Message::SetModifiable] = ProcessFunction(&CommandZoneResponder::processSetModifiable, true, true);
	m_processFunctions[commandzone::Message::IsModifiable] = ProcessFunction(&CommandZoneResponder::processIsModifiable, false, false);
	m_processFunctions[commandzone::Message::StopCommandBox] = ProcessFunction(&CommandZoneResponder::processStopCommandBox, true, true);
	m_processFunctions[commandzone::Message::GetCommandBoxProgression] = ProcessFunction(&CommandZoneResponder::processGetCommandBoxProgression, false, false);
	m_processFunctions[commandzone::Message::GetCommandID] = ProcessFunction(&CommandZoneResponder::processGetCommandID, false, false);
	m_processFunctions[commandzone::Message::IsSettings] = ProcessFunction(&CommandZoneResponder::processIsSettings, false, false);
	m_processFunctions[commandzone::Message::GetDatabaseID] = ProcessFunction(&CommandZoneResponder::processGetDatabaseID, false, false);
	m_processFunctions[commandzone::Message::GetSettingsFileName] = ProcessFunction(&CommandZoneResponder::processGetSettingsFileName, false, false);
	m_processFunctions[commandzone::Message::GetStringValue] = ProcessFunction(&CommandZoneResponder::processGetStringValue, false, false);
	m_processFunctions[commandzone::Message::GetFormattedValue] = ProcessFunction(&CommandZoneResponder::processGetFormattedValue, false, false);
	m_processFunctions[commandzone::Message::SetSettingsFileName] = ProcessFunction(&CommandZoneResponder::processSetSettingsFileName, true, true);
	m_processFunctions[commandzone::Message::SetControllerName] = ProcessFunction(&CommandZoneResponder::processSetControllerName, true, true);
	m_processFunctions[commandzone::Message::SetPropertyName] = ProcessFunction(&CommandZoneResponder::processSetPropertyName, true, true);
	m_processFunctions[commandzone::Message::SetValue] = ProcessFunction(&CommandZoneResponder::processSetValue, true, true);
	m_processFunctions[commandzone::Message::SetUserVariableName] = ProcessFunction(&CommandZoneResponder::processSetUserVariableName, true, true);
	m_processFunctions[commandzone::Message::SetUserVariableExpression] = ProcessFunction(&CommandZoneResponder::processSetUserVariableExpression, true, true);
	m_processFunctions[commandzone::Message::GetUserVariableName] = ProcessFunction(&CommandZoneResponder::processGetUserVariableName, false, false);
	m_processFunctions[commandzone::Message::GetUserVariableExpression] = ProcessFunction(&CommandZoneResponder::processGetUserVariableExpression, false, false);
	m_processFunctions[commandzone::Message::GetControllerName] = ProcessFunction(&CommandZoneResponder::processGetControllerName, false, false);
	m_processFunctions[commandzone::Message::GetPropertyName] = ProcessFunction(&CommandZoneResponder::processGetPropertyName, false, false);
	m_processFunctions[commandzone::Message::GetValue] = ProcessFunction(&CommandZoneResponder::processGetValue, false, false);
	m_processFunctions[commandzone::Message::GetBoxType] = ProcessFunction(&CommandZoneResponder::processGetBoxType, false, false);
	m_processFunctions[commandzone::Message::GetCommandList] = ProcessFunction(&CommandZoneResponder::processGetCommandList, false, false);
	m_processFunctions[commandzone::Message::AddNewAtomicCommandBoxAtEnd] = ProcessFunction(&CommandZoneResponder::processAddNewAtomicCommandBoxAtEnd, true, true);
	m_processFunctions[commandzone::Message::AddNewAtomicCommandBox] = ProcessFunction(&CommandZoneResponder::processAddNewAtomicCommandBox, true, true);
	m_processFunctions[commandzone::Message::AddNewScanCommandBoxAtEnd] = ProcessFunction(&CommandZoneResponder::processAddNewScanCommandBoxAtEnd, true, true);
	m_processFunctions[commandzone::Message::AddNewScanCommandBox] = ProcessFunction(&CommandZoneResponder::processAddNewScanCommandBox, true, true);
	m_processFunctions[commandzone::Message::AddNewForLoopCommandBoxAtEnd] = ProcessFunction(&CommandZoneResponder::processAddNewForLoopCommandBoxAtEnd, true, true);
	m_processFunctions[commandzone::Message::AddNewForLoopCommandBox] = ProcessFunction(&CommandZoneResponder::processAddNewForLoopCommandBox, true, true);
	m_processFunctions[commandzone::Message::AddNewGenericCommandBoxAtEnd] = ProcessFunction(&CommandZoneResponder::processAddNewGenericCommandBoxAtEnd, true, true);
	m_processFunctions[commandzone::Message::AddNewGenericCommandBox] = ProcessFunction(&CommandZoneResponder::processAddNewGenericCommandBox, true, true);
	m_processFunctions[commandzone::Message::AddNewControlCommandBoxAtEnd] = ProcessFunction(&CommandZoneResponder::processAddNewControlCommandBoxAtEnd, true, true);
	m_processFunctions[commandzone::Message::AddNewControlCommandBox] = ProcessFunction(&CommandZoneResponder::processAddNewControlCommandBox, true, true);
	m_processFunctions[commandzone::Message::MoveCommandBox] = ProcessFunction(&CommandZoneResponder::processMoveCommandBox, true, true);
	m_processFunctions[commandzone::Message::MoveCommandBoxToEnd] = ProcessFunction(&CommandZoneResponder::processMoveCommandBoxToEnd, true, true);
	m_processFunctions[commandzone::Message::DeleteCommandBox] = ProcessFunction(&CommandZoneResponder::processDeleteCommandBox, true, true);
	m_processFunctions[commandzone::Message::CheckExpression] = ProcessFunction(&CommandZoneResponder::processCheckExpression, false, false);
	m_processFunctions[commandzone::Message::CheckStaticExpression] = ProcessFunction(&CommandZoneResponder::processCheckStaticExpression, false, false);
	m_processFunctions[commandzone::Message::SetExpression] = ProcessFunction(&CommandZoneResponder::processSetExpression, true, true);
	m_processFunctions[commandzone::Message::GetExpression] = ProcessFunction(&CommandZoneResponder::processGetExpression, false, false);
	m_processFunctions[commandzone::Message::GetScanName] = ProcessFunction(&CommandZoneResponder::processGetScanName, false, false);
	m_processFunctions[commandzone::Message::GetNumberOfLines] = ProcessFunction(&CommandZoneResponder::processGetNumberOfLines, false, false);
	m_processFunctions[commandzone::Message::SetVariableName] = ProcessFunction(&CommandZoneResponder::processSetVariableName, true, true);
	m_processFunctions[commandzone::Message::SetType] = ProcessFunction(&CommandZoneResponder::processSetType, true, true);
	m_processFunctions[commandzone::Message::SetValues] = ProcessFunction(&CommandZoneResponder::processSetValues, true, true);
	m_processFunctions[commandzone::Message::SetStartValue] = ProcessFunction(&CommandZoneResponder::processSetStartValue, true, true);
	m_processFunctions[commandzone::Message::SetEndValue] = ProcessFunction(&CommandZoneResponder::processSetEndValue, true, true);
	m_processFunctions[commandzone::Message::SetStepSize] = ProcessFunction(&CommandZoneResponder::processSetStepSize, true, true);
	m_processFunctions[commandzone::Message::SetStepType] = ProcessFunction(&CommandZoneResponder::processSetStepType, true, true);
	m_processFunctions[commandzone::Message::GetVariableName] = ProcessFunction(&CommandZoneResponder::processGetVariableName, false, false);
	m_processFunctions[commandzone::Message::GetStartValue] = ProcessFunction(&CommandZoneResponder::processGetStartValue, false, false);
	m_processFunctions[commandzone::Message::GetEndValue] = ProcessFunction(&CommandZoneResponder::processGetEndValue, false, false);
	m_processFunctions[commandzone::Message::GetStepSize] = ProcessFunction(&CommandZoneResponder::processGetStepSize, false, false);
	m_processFunctions[commandzone::Message::GetStepType] = ProcessFunction(&CommandZoneResponder::processGetStepType, false, false);
	m_processFunctions[commandzone::Message::GetCurrentValue] = ProcessFunction(&CommandZoneResponder::processGetCurrentValue, false, false);
	m_processFunctions[commandzone::Message::GetType] = ProcessFunction(&CommandZoneResponder::processGetType, false, false);
	m_processFunctions[commandzone::Message::GetValues] = ProcessFunction(&CommandZoneResponder::processGetValues, false, false);
	m_processFunctions[commandzone::Message::AddForLoopLine] = ProcessFunction(&CommandZoneResponder::processAddForLoopLine, true, true);
	m_processFunctions[commandzone::Message::RemoveForLoopLine] = ProcessFunction(&CommandZoneResponder::processRemoveForLoopLine, true, true);
	m_processFunctions[commandzone::Message::GetScanCommandBoxUserVariableNames] = ProcessFunction(&CommandZoneResponder::processGetScanCommandBoxUserVariableNames, false, false);
	m_processFunctions[commandzone::Message::Create] = ProcessFunction(&CommandZoneResponder::processCreate, true, true);
	m_processFunctions[commandzone::Message::Reset] = ProcessFunction(&CommandZoneResponder::processReset, true, true);
	m_processFunctions[commandzone::Message::Execute] = ProcessFunction(&CommandZoneResponder::processExecute, true, true);
	m_processFunctions[commandzone::Message::Stop] = ProcessFunction(&CommandZoneResponder::processStop, true, true);
	m_processFunctions[commandzone::Message::StopAtEnd] = ProcessFunction(&CommandZoneResponder::processStopAtEnd, true, true);
	m_processFunctions[commandzone::Message::Pause] = ProcessFunction(&CommandZoneResponder::processPause, true, true);
	m_processFunctions[commandzone::Message::PauseAtEnd] = ProcessFunction(&CommandZoneResponder::processPauseAtEnd, true, true);
	m_processFunctions[commandzone::Message::Restart] = ProcessFunction(&CommandZoneResponder::processRestart, true, true);
	m_processFunctions[commandzone::Message::IsRunning] = ProcessFunction(&CommandZoneResponder::processIsRunning, false, false);
	m_processFunctions[commandzone::Message::IsPaused] = ProcessFunction(&CommandZoneResponder::processIsPaused, false, false);
	m_processFunctions[commandzone::Message::IsPausingAtEnd] = ProcessFunction(&CommandZoneResponder::processIsPausingAtEnd, false, false);
	m_processFunctions[commandzone::Message::IsStoppingAtEnd] = ProcessFunction(&CommandZoneResponder::processIsStoppingAtEnd, false, false);
	m_processFunctions[commandzone::Message::GetCommandZoneExecutionState] = ProcessFunction(&CommandZoneResponder::processGetCommandZoneExecutionState, false, false);
	m_processFunctions[commandzone::Message::GetProgression] = ProcessFunction(&CommandZoneResponder::processGetProgression, false, false);
	m_processFunctions[commandzone::Message::GetConflictingCommands] = ProcessFunction(&CommandZoneResponder::processGetConflictingCommands, false, false);
	m_processFunctions[commandzone::Message::Print] = ProcessFunction(&CommandZoneResponder::processPrint, false, false);
	m_processFunctions[commandzone::Message::CheckPropertyExpression] = ProcessFunction(&CommandZoneResponder::processCheckPropertyExpression, false, false);
	m_processFunctions[commandzone::Message::CheckStaticPropertyExpression] = ProcessFunction(&CommandZoneResponder::processCheckStaticPropertyExpression, false, false);
	m_processFunctions[commandzone::Message::SetPropertyExpression] = ProcessFunction(&CommandZoneResponder::processSetPropertyExpression, true, true);
	m_processFunctions[commandzone::Message::GetPropertyExpression] = ProcessFunction(&CommandZoneResponder::processGetPropertyExpression, false, false);
	m_processFunctions[commandzone::Message::GetVariableList] = ProcessFunction(&CommandZoneResponder::processGetVariableList, false, false);
	m_processFunctions[commandzone::Message::GetFileContent] = ProcessFunction(&CommandZoneResponder::processGetFileContent, false, false);
	m_processFunctions[commandzone::Message::GetFileContentAbsolutePath] = ProcessFunction(&CommandZoneResponder::processGetFileContentAbsolutePath, false, false);
	m_processFunctions[commandzone::Message::GetTree] = ProcessFunction(&CommandZoneResponder::processGetTree, false, false);

	// The SaveFile and DeleteFile generate a notification to other clients but do not generate an activity.
	m_processFunctions[commandzone::Message::SaveFile] = ProcessFunction(&CommandZoneResponder::processSaveFile, true, false);
	m_processFunctions[commandzone::Message::DeleteFile] = ProcessFunction(&CommandZoneResponder::processDeleteFile, true, false);

	m_processFunctions[commandzone::Message::CreateDirectory] = ProcessFunction(&CommandZoneResponder::processCreateDirectory, false, false);
	m_processFunctions[commandzone::Message::AppendToFile] = ProcessFunction(&CommandZoneResponder::processAppendToFile, false, false);

	m_processFunctions[commandzone::Message::SaveBinaryFile] = ProcessFunction(&CommandZoneResponder::processSaveBinaryFile, false, false);
	m_processFunctions[commandzone::Message::GetBinaryFileContent] = ProcessFunction(&CommandZoneResponder::processGetBinaryFileContent, false, false);

	m_processFunctions[commandzone::Message::CloneForEstimatedTime] = ProcessFunction(&CommandZoneResponder::processCloneForEstimatedTime, false, false);
	m_processFunctions[commandzone::Message::ComputeEstimatedTimeState] = ProcessFunction(&CommandZoneResponder::processComputeEstimatedTimeState, false, false);

	m_processFunctions[commandzone::Message::FileExists] = ProcessFunction(&CommandZoneResponder::processFileExists, false, false);

	m_processFunctions[commandzone::Message::GetLastControllerName] = ProcessFunction(&CommandZoneResponder::processGetLastControllerName, false, false);
	m_processFunctions[commandzone::Message::GetLastSetPropertyName] = ProcessFunction(&CommandZoneResponder::processGetLastSetPropertyName, false, false);

	m_processFunctions[commandzone::Message::InitPropertyExpression] = ProcessFunction(&CommandZoneResponder::processInitPropertyExpression, false, false);
}

CommandBox* CommandZoneResponder::getCommandBox(unsigned long commandBoxID) {

	try {
		return CommandBoxMap::getInstance()->get(commandBoxID);

	} catch (CommandBoxMap::CommandBoxNotFound& e) {
		return 0;
	}
}

AtomicCommandBox* CommandZoneResponder::getAtomicCommandBox(unsigned long commandBoxID) {

	return dynamic_cast<AtomicCommandBox*>(getCommandBox(commandBoxID));
}

GenericCommandBox* CommandZoneResponder::getGenericCommandBox(unsigned long commandBoxID) {

	return dynamic_cast<GenericCommandBox*>(getCommandBox(commandBoxID));
}

PropertyGenericCommandBox* CommandZoneResponder::getPropertyGenericCommandBox(unsigned long commandBoxID) {

	return dynamic_cast<PropertyGenericCommandBox*>(getCommandBox(commandBoxID));
}

VariableCommandBox* CommandZoneResponder::getVariableCommandBox(unsigned long commandBoxID) {

	return dynamic_cast<VariableCommandBox*>(getCommandBox(commandBoxID));
}

IfCommandBox* CommandZoneResponder::getIfCommandBox(unsigned long commandBoxID) {

	return dynamic_cast<IfCommandBox*>(getCommandBox(commandBoxID));
}

ScanCommandBox* CommandZoneResponder::getScanCommandBox(unsigned long commandBoxID) {

	return dynamic_cast<ScanCommandBox*>(getCommandBox(commandBoxID));
}

ForLoopCommandBox* CommandZoneResponder::getForLoopCommandBox(unsigned long commandBoxID) {

	return dynamic_cast<ForLoopCommandBox*>(getCommandBox(commandBoxID));
}

ControlCommandBox* CommandZoneResponder::getControlCommandBox(unsigned long commandBoxID) {

	return dynamic_cast<ControlCommandBox*>(getCommandBox(commandBoxID));
}

BreakCommandBox* CommandZoneResponder::getBreakCommandBox(unsigned long commandBoxID) {

	return dynamic_cast<BreakCommandBox*>(getCommandBox(commandBoxID));
}

template<typename Response>
void processStatusAndErrorResponse(Response & responseMessage, ExpressionStatus status, common::ExpressionError error, const std::string & message) {

	if (status == common::VALID_EXPRESSION) {
		responseMessage.set_status(commandzone::ExpressionStatus::VALID_EXPRESSION);
	}
	else if (status == common::INVALID_EXPRESSION) {
		responseMessage.set_status(commandzone::ExpressionStatus::INVALID_EXPRESSION);

		if (error == common::PARSER_ERROR) {
			responseMessage.set_expressionerror(commandzone::ExpressionStatus::PARSER_ERROR);
		}
		else if (error == common::ENUMERATED_VALUE_ERROR) {
			responseMessage.set_expressionerror(commandzone::ExpressionStatus::ENUMERATED_VALUE_ERROR);
		}
		else if (error == common::DISABLED_CONTROLLER_ERROR) {
			responseMessage.set_expressionerror(commandzone::ExpressionStatus::DISABLED_CONTROLLER_ERROR);
		}
		else if (error == common::UNDEFINED_VARIABLE_ERROR) {
			responseMessage.set_expressionerror(commandzone::ExpressionStatus::UNDEFINED_VARIABLE_ERROR);
		}

		responseMessage.set_message(message);
	}
}

bool checkBreakCommandBox(BreakCommandBox* commandBox, common::BooleanResponse & responseMessage, std::string & response) {

	if (commandBox) {
		responseMessage.set_error(common::Error::COMMAND_BOX_FORBIDDEN);

		// Serialize the response.
		responseMessage.SerializeToString(&response);
		return true;
	}

	return false;
}

bool checkAddBreakCommandBox(int32 boxType, common::Int32Response & responseMessage, std::string & response) {

	// check for Break command
	if ((ControlCommandBox::BoxType)boxType == ControlCommandBox::BREAK) {
		responseMessage.set_error(common::Error::COMMAND_BOX_FORBIDDEN);

		// Serialize the response.
		responseMessage.SerializeToString(&response);
		return true;
	}

	return false;
}

bool CommandZoneResponder::processToggleBackground(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->toggleBackground();
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processToggleParallel(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->toggleParallel();
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processIsBackground(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->isBackground());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processIsParallel(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->isParallel());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetExecutionState(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ExecutionStateResponse responseMessage;

	CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		CommandBox::ExecutionState state = commandBox->getExecutionState();

		if (state == CommandBox::RUNNING) {
			responseMessage.set_state(commandzone::ExecutionStateResponse::RUNNING);
		}
		else if (state == CommandBox::PAUSED) {
			responseMessage.set_state(commandzone::ExecutionStateResponse::PAUSED);
		}
		else if (state == CommandBox::STOPPING_AT_END) {
			responseMessage.set_state(commandzone::ExecutionStateResponse::STOPPING_AT_END);
		}
		else if (state == CommandBox::PAUSING_AT_END) {
			responseMessage.set_state(commandzone::ExecutionStateResponse::PAUSING_AT_END);
		}
		else if (state == CommandBox::FINISHED) {
			responseMessage.set_state(commandzone::ExecutionStateResponse::FINISHED);
		}
		else {
			responseMessage.set_state(commandzone::ExecutionStateResponse::CONSTRUCTION);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetModifiable(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxSetModifiableRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->setModifiable(messageRequest.modifiable());
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processIsModifiable(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->isModifiable());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processStopCommandBox(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->stop();
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetCommandBoxProgression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else {
		CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			responseMessage.set_value(commandBox->getProgression());
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetCommandID(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getCommandID());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processIsSettings(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->isSettings());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetDatabaseID(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getExecutionDatabaseID());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetSettingsFileName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getSettingsFileName());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetStringValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxPropertyRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else {
		AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			responseMessage.set_value(commandBox->getPropertyValue<string>(messageRequest.propertyid()));
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetFormattedValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxPropertyRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getFormattedPropertyValue(messageRequest.propertyid()));
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetSettingsFileName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetSettingsFileNameRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->setSettingsFileName(messageRequest.filename());
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetControllerName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetStringRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		PropertyGenericCommandBox* commandBox = getPropertyGenericCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {

			try {
				commandBox->setControllerName(messageRequest.value());
				responseMessage.set_value(true);
			}
			catch (common::Exception& e) {
				err << e.printStack() << eol;
			}
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetPropertyName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetStringRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		PropertyGenericCommandBox* commandBox = getPropertyGenericCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {

			try {
				commandBox->setPropertyName(messageRequest.value());
				responseMessage.set_value(true);
			}
			catch (common::Exception& e) {
				err << e.printStack() << eol;
			}
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetStringRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		PropertyGenericCommandBox* commandBox = getPropertyGenericCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {

			try {
				commandBox->setValue(messageRequest.value());
				responseMessage.set_value(true);
			}
			catch (common::Exception& e) {
				err << e.printStack() << eol;
			}
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetUserVariableName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetUserVariableNameRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		VariableCommandBox* commandBox = getVariableCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {

			try {
				commandBox->setVariableName(messageRequest.name());
				//commandBox->setExpression(messageRequest.expression());
				responseMessage.set_value(true);
			}
			catch (common::Exception& e) {
				err << e.printStack() << eol;
			}
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetUserVariableExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetUserVariableExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		VariableCommandBox* commandBox = getVariableCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {

			try {
				commandBox->setExpression(messageRequest.expression());
				responseMessage.set_value(true);
			}
			catch (common::Exception& e) {
				err << e.printStack() << eol;
			}
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetUserVariableName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	VariableCommandBox* commandBox = getVariableCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getVariableName());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetUserVariableExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	VariableCommandBox* commandBox = getVariableCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getExpression());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetControllerName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	PropertyGenericCommandBox* commandBox = getPropertyGenericCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getControllerName());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetPropertyName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	PropertyGenericCommandBox* commandBox = getPropertyGenericCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getPropertyName());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	PropertyGenericCommandBox* commandBox = getPropertyGenericCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getValue());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetBoxType(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;
	{
		GenericCommandBox* commandBox = getGenericCommandBox(messageRequest.commandboxid());
		if (commandBox != 0) {
			responseMessage.set_value(commandBox->getBoxType());
			responseMessage.SerializeToString(&response);
			return false;
		}
	}
	{
		ControlCommandBox* commandBox = getControlCommandBox(messageRequest.commandboxid());
		if (commandBox != 0) {
			responseMessage.set_value(commandBox->getBoxType());
			responseMessage.SerializeToString(&response);
			return false;
		}
	}

	responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

void fillCommandList(commandzone::CommandListResponse& responseMessage, const vector<CommandBox*>& commandBoxes) {

	for (vector<CommandBox*>::const_iterator b = commandBoxes.begin(); b != commandBoxes.end(); b++) {

		CommandBox* commandBox = *b;
		commandzone::CommandBoxDescriptor * descriptor = responseMessage.add_descriptor();
		descriptor->set_commandboxid(commandBox->getID());

		if (commandBox->getType() == CommandBox::ATOMIC) {
			descriptor->set_type(commandzone::CommandBoxDescriptor::ATOMIC);
		}
		else if (commandBox->getType() == CommandBox::SCAN) {
			descriptor->set_type(commandzone::CommandBoxDescriptor::SCAN);
		}
		else if (commandBox->getType() == CommandBox::FOR_LOOP) {
			descriptor->set_type(commandzone::CommandBoxDescriptor::FOR_LOOP);
		}
		else if (commandBox->getType() == CommandBox::GENERIC) {
			descriptor->set_type(commandzone::CommandBoxDescriptor::GENERIC);
		}
		else if (commandBox->getType() == CommandBox::IF) {
			descriptor->set_type(commandzone::CommandBoxDescriptor::IFTHEN);
		}
		else if (commandBox->getType() == CommandBox::BREAK) {
			descriptor->set_type(commandzone::CommandBoxDescriptor::BREAK);
		}
		else if (commandBox->getType() == CommandBox::IF_ELSE) {
			descriptor->set_type(commandzone::CommandBoxDescriptor::IFTHENELSE);
		}
	}
}

bool CommandZoneResponder::processGetCommandList(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetCommandListRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::CommandListResponse responseMessage;
	commandzone::Container::Type type = messageRequest.type();

	if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

		IfCommandBox* commandBox = getIfCommandBox(messageRequest.containerid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
			fillCommandList(responseMessage, commandBox->getCommandBoxes(thenBlock));
		}
	}
	else if (type == commandzone::Container::ScanCommandBox) {

		ScanCommandBox* commandBox = getScanCommandBox(messageRequest.containerid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			fillCommandList(responseMessage, commandBox->getCommandBoxes());
		}
	}
	else if (type == commandzone::Container::ForLoopCommandBox) {

		ForLoopCommandBox* commandBox = getForLoopCommandBox(messageRequest.containerid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			fillCommandList(responseMessage, commandBox->getCommandBoxes());
		}
	}
	else if (type == commandzone::Container::CommandZone) {

		fillCommandList(responseMessage, CommandZone::getInstance(messageRequest.containerid())->getCommandBoxes());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewAtomicCommandBoxAtEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewAtomicCommandBoxAtEndRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createAtomicCommandBox(messageRequest.controllername(), messageRequest.issettings());
					responseMessage.set_value(ifCommandBox->addCommandBoxAtEnd(thenBlock, commandBox));
				}
			}
			else if (type == commandzone::Container::ScanCommandBox) {

				ScanCommandBox* scanCommandBox = getScanCommandBox(messageRequest.containerid());

				if (scanCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createAtomicCommandBox(messageRequest.controllername(), messageRequest.issettings());
					responseMessage.set_value(scanCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createAtomicCommandBox(messageRequest.controllername(), messageRequest.issettings());
					responseMessage.set_value(forLoopCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createAtomicCommandBox(messageRequest.controllername(), messageRequest.issettings());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBoxAtEnd(commandBox));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewAtomicCommandBox(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewAtomicCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createAtomicCommandBox(messageRequest.controllername(), messageRequest.issettings());
					responseMessage.set_value(ifCommandBox->addCommandBox(thenBlock, commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createAtomicCommandBox(messageRequest.controllername(), messageRequest.issettings());
					responseMessage.set_value(forLoopCommandBox->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createAtomicCommandBox(messageRequest.controllername(), messageRequest.issettings());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewScanCommandBoxAtEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewScanCommandBoxAtEndRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createScanCommandBox(messageRequest.scanname(), messageRequest.fullpath());
					responseMessage.set_value(ifCommandBox->addCommandBoxAtEnd(thenBlock, commandBox));
				}
			}
			else if (type == commandzone::Container::ScanCommandBox) {

				ScanCommandBox* scanCommandBox = getScanCommandBox(messageRequest.containerid());

				if (scanCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createScanCommandBox(messageRequest.scanname(), messageRequest.fullpath());
					responseMessage.set_value(scanCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createScanCommandBox(messageRequest.scanname(), messageRequest.fullpath());
					responseMessage.set_value(forLoopCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createScanCommandBox(messageRequest.scanname(), messageRequest.fullpath());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBoxAtEnd(commandBox));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewScanCommandBox(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewScanCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createScanCommandBox(messageRequest.scanname(), messageRequest.fullpath());
					responseMessage.set_value(ifCommandBox->addCommandBox(thenBlock, commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createScanCommandBox(messageRequest.scanname(), messageRequest.fullpath());
					responseMessage.set_value(forLoopCommandBox->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createScanCommandBox(messageRequest.scanname(), messageRequest.fullpath());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewForLoopCommandBoxAtEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewForLoopCommandBoxAtEndRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createForLoopCommandBox(messageRequest.variablename());
					responseMessage.set_value(ifCommandBox->addCommandBoxAtEnd(thenBlock, commandBox));
				}
			}
			else if (type == commandzone::Container::ScanCommandBox) {

				ScanCommandBox* scanCommandBox = getScanCommandBox(messageRequest.containerid());

				if (scanCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createForLoopCommandBox(messageRequest.variablename());
					responseMessage.set_value(scanCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createForLoopCommandBox(messageRequest.variablename());
					responseMessage.set_value(forLoopCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createForLoopCommandBox(messageRequest.variablename());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBoxAtEnd(commandBox));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewForLoopCommandBox(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewForLoopCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createForLoopCommandBox(messageRequest.variablename());
					responseMessage.set_value(ifCommandBox->addCommandBox(thenBlock, commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createForLoopCommandBox(messageRequest.variablename());
					responseMessage.set_value(forLoopCommandBox->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createForLoopCommandBox(messageRequest.variablename());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewGenericCommandBoxAtEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewGenericCommandBoxAtEndRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createGenericCommandBox((PropertyGenericCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(ifCommandBox->addCommandBoxAtEnd(thenBlock, commandBox));
				}
			}
			else if (type == commandzone::Container::ScanCommandBox) {

				ScanCommandBox* scanCommandBox = getScanCommandBox(messageRequest.containerid());

				if (scanCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createGenericCommandBox((PropertyGenericCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(scanCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createGenericCommandBox((PropertyGenericCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(forLoopCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createGenericCommandBox((PropertyGenericCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBoxAtEnd(commandBox));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewGenericCommandBox(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewGenericCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createGenericCommandBox((PropertyGenericCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(ifCommandBox->addCommandBox(thenBlock, commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createGenericCommandBox((PropertyGenericCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(forLoopCommandBox->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createGenericCommandBox((PropertyGenericCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewControlCommandBoxAtEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewControlCommandBoxAtEndRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				if (checkAddBreakCommandBox(messageRequest.boxtype(), responseMessage, response)) {
					return true;
				}

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createControlCommandBox((ControlCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(ifCommandBox->addCommandBoxAtEnd(thenBlock, commandBox));
				}
			}
			else if (type == commandzone::Container::ScanCommandBox) {

				ScanCommandBox* scanCommandBox = getScanCommandBox(messageRequest.containerid());

				if (scanCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createControlCommandBox((ControlCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(scanCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createControlCommandBox((ControlCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(forLoopCommandBox->addCommandBoxAtEnd(commandBox));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				if (checkAddBreakCommandBox(messageRequest.boxtype(), responseMessage, response)) {
					return true;
				}

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createControlCommandBox((ControlCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBoxAtEnd(commandBox));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddNewControlCommandBox(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AddNewControlCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				if (checkAddBreakCommandBox(messageRequest.boxtype(), responseMessage, response)) {
					return true;
				}

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createControlCommandBox((ControlCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(ifCommandBox->addCommandBox(thenBlock, commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createControlCommandBox((ControlCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(forLoopCommandBox->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				if (checkAddBreakCommandBox(messageRequest.boxtype(), responseMessage, response)) {
					return true;
				}

				try {
					CommandBox* commandBox = CommandBoxFactory::getInstance()->createControlCommandBox((ControlCommandBox::BoxType)messageRequest.boxtype());
					responseMessage.set_value(CommandZone::getInstance(messageRequest.containerid())->addCommandBox(commandBox, messageRequest.commandboxidtomove(), messageRequest.above()));

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (CommandBoxFactory::NoSuchCommandException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_COMMAND);

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			// The command box may have been deleted. However this should not happen.
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processMoveCommandBox(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::MoveCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				// Do not authorize to move a break command box here.
				if (checkBreakCommandBox(getBreakCommandBox(messageRequest.commandboxidtomove()), responseMessage, response)) {
					return true;
				}

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					ifCommandBox->moveCommandBox(thenBlock, messageRequest.commandboxidtomove(), messageRequest.commandboxidtoshift(), messageRequest.above());
					responseMessage.set_value(true);
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					forLoopCommandBox->moveCommandBox(messageRequest.commandboxidtomove(), messageRequest.commandboxidtoshift(), messageRequest.above());
					responseMessage.set_value(true);
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				// Do not authorize to move a break command box here.
				if (checkBreakCommandBox(getBreakCommandBox(messageRequest.commandboxidtomove()), responseMessage, response)) {
					return true;
				}

				try {
					CommandZone::getInstance(messageRequest.containerid())->moveCommandBox(messageRequest.commandboxidtomove(), messageRequest.commandboxidtoshift(), messageRequest.above());
					responseMessage.set_value(true);

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processMoveCommandBoxToEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::MoveCommandBoxToEndRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				// Do not authorize to move a break command box here.
				if (checkBreakCommandBox(getBreakCommandBox(messageRequest.commandboxidtomove()), responseMessage, response)) {
					return true;
				}

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					ifCommandBox->moveCommandBoxToEnd(thenBlock, messageRequest.commandboxidtomove());
					responseMessage.set_value(true);
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					forLoopCommandBox->moveCommandBoxToEnd(messageRequest.commandboxidtomove());
					responseMessage.set_value(true);
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				// Do not authorize to move a break command box here.
				if (checkBreakCommandBox(getBreakCommandBox(messageRequest.commandboxidtomove()), responseMessage, response)) {
					return true;
				}

				try {
					CommandZone::getInstance(messageRequest.containerid())->moveCommandBoxToEnd(messageRequest.commandboxidtomove());
					responseMessage.set_value(true);

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processDeleteCommandBox(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::DeleteCommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		commandzone::Container::Type type = messageRequest.type();

		try {
			if (type == commandzone::Container::IfThenCommmandBox || type == commandzone::Container::IfElseCommmandBox) {

				IfCommandBox* ifCommandBox = getIfCommandBox(messageRequest.containerid());

				if (ifCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					bool thenBlock = (type == commandzone::Container::IfThenCommmandBox);
					ifCommandBox->deleteCommandBox(thenBlock, messageRequest.commandboxidtodelete());
					responseMessage.set_value(true);
				}
			}
			else if (type == commandzone::Container::ForLoopCommandBox) {

				ForLoopCommandBox* forLoopCommandBox = getForLoopCommandBox(messageRequest.containerid());

				if (forLoopCommandBox == 0) {
					responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
				}
				else {
					forLoopCommandBox->deleteCommandBox(messageRequest.commandboxidtodelete());
					responseMessage.set_value(true);
				}
			}
			else if (type == commandzone::Container::CommandZone) {

				try {
					CommandZone::getInstance(messageRequest.containerid())->deleteCommandBox(messageRequest.commandboxidtodelete());
					responseMessage.set_value(true);

				} catch (CommandZone::CommandZoneNotFound& e) {
					responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
				}
			}

		} catch (ModifiableCommandBoxContainer::CommandBoxNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processCheckExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::ExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ExpressionResponse responseMessage;

	ControlCommandBox * commandBox = getControlCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {

		try {
			// Check the expression.
			common::ExpressionError error;
			std::string lmessage;

			ExpressionStatus status = commandBox->checkExpression(messageRequest.expression(), error, lmessage);
			processStatusAndErrorResponse(responseMessage, status, error, lmessage);
		}
		catch (common::Exception& e) {
			err << e.printStack() << eol;
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processCheckStaticExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::StaticExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ExpressionResponse responseMessage;

	try {
		// Check the expression.
		common::ExpressionError error;
		std::string message;

		ExpressionStatus status = ControlCommandBox::checkStaticExpression(messageRequest.expression(), error, message);
		processStatusAndErrorResponse(responseMessage, status, error, message);
	}
	catch (AtomicCommandBox::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (AtomicCommandBox::BadPropertyType& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (common::Exception& e) {
		err << e.printStack() << eol;
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processSetExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::ExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ExpressionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ControlCommandBox * commandBox = getControlCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {

			try {
				// Check the expression.
				common::ExpressionError error;
				std::string lmessage;

				ExpressionStatus estatus = commandBox->checkExpression(messageRequest.expression(), error, lmessage);
				if (estatus == common::VALID_EXPRESSION) {
					responseMessage.set_status(commandzone::ExpressionStatus::VALID_EXPRESSION);

					// Set the expression.
					commandBox->setExpression(messageRequest.expression());
				}
				else if (estatus == common::INVALID_EXPRESSION) {

					// The only errors the server does not accept are the parser errors.
					// The other errors can be "corrected" at runtime before the execution of the box.
					// The status of the expression is "changed" for the client.
					if (error == common::PARSER_ERROR) {
						responseMessage.set_status(commandzone::ExpressionStatus::INVALID_EXPRESSION);
						responseMessage.set_expressionerror(commandzone::ExpressionStatus::PARSER_ERROR);

						responseMessage.set_message(lmessage);
					}
					else {
						responseMessage.set_status(commandzone::ExpressionStatus::VALID_EXPRESSION);

						commandBox->setExpression(messageRequest.expression());
					}
				}
			}
			catch (AtomicCommandBox::NoSuchPropertyException& e) {
				responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
			}
			catch (AtomicCommandBox::BadPropertyType& e) {
				responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
			}
			catch (common::Exception& e) {
				err << e.printStack() << eol;
			}
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;
	ControlCommandBox * commandBox = getControlCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getExpression());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetScanName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;
	ScanCommandBox * commandBox = getScanCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getName());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetNumberOfLines(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Int32Response responseMessage;
	ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getNumberOfLines());
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetVariableName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetForLoopStringValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->setVariableName(messageRequest.lineindex(), messageRequest.value());
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetType(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetForLoopTypeRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			if (messageRequest.type() == commandzone::ForLoopType::FOR_RANGE) {
				commandBox->setForLoopType(messageRequest.lineindex(), ForLoopCommandBox::RANGE);
			}
			else if (messageRequest.type() == commandzone::ForLoopType::FOR_VALUES) {
				commandBox->setForLoopType(messageRequest.lineindex(), ForLoopCommandBox::VALUES);
			}

			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetValues(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetForLoopStringValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			try {
				commandBox->setValues(messageRequest.lineindex(), messageRequest.value());
				responseMessage.set_value(true);
			}
			catch (ForLoopCommandBox::ValuesSyntaxErrorException& e) {
				err << e.printStack() << eol;
				responseMessage.set_error(common::Error::VALUES_SYNTAX_ERROR);
			}
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetStartValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->setStartValue(messageRequest.lineindex(), messageRequest.value());
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetEndValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->setEndValue(messageRequest.lineindex(), messageRequest.value());
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetStepSize(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->setStepSize(messageRequest.lineindex(), messageRequest.value());
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetStepType(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SetForLoopStepTypeRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			if (messageRequest.type() == commandzone::ForLoopStepType::DELTA) {
//				cout << "CommandZoneResponder::processSetStepType DELTA" << endl;
				commandBox->setForLoopStepType(messageRequest.lineindex(), ForLoopCommandBox::DELTA);
			}
			else if (messageRequest.type() == commandzone::ForLoopStepType::NBPOINTS) {
//				cout << "CommandZoneResponder::processSetStepType NBPOINTS" << endl;
				commandBox->setForLoopStepType(messageRequest.lineindex(), ForLoopCommandBox::NBPOINTS);
			}

			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetVariableName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;
	ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getVariableName(messageRequest.lineindex()));
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetStartValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Float64Response responseMessage;
	ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getStartValue(messageRequest.lineindex()));
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetEndValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Float64Response responseMessage;
	ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getEndValue(messageRequest.lineindex()));
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetStepSize(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Float64Response responseMessage;
	ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getStepSize(messageRequest.lineindex()));
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetStepType(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ForLoopStepTypeResponse responseMessage;
	ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		ForLoopCommandBox::ForLoopStepType stepType = commandBox->getForLoopStepType(messageRequest.lineindex());
		if (stepType == ForLoopCommandBox::DELTA) {
			responseMessage.set_value(commandzone::ForLoopStepType::DELTA);
		}
		else {
			responseMessage.set_value(commandzone::ForLoopStepType::NBPOINTS);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetCurrentValue(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Float64Response responseMessage;
	ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getCurrentValue(messageRequest.lineindex()));
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetType(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ForLoopTypeResponse responseMessage;
	ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		ForLoopCommandBox::ForLoopType type = commandBox->getForLoopType(messageRequest.lineindex());
		if (type == ForLoopCommandBox::RANGE) {
			responseMessage.set_value(commandzone::ForLoopType::FOR_RANGE);
		}
		else {
			responseMessage.set_value(commandzone::ForLoopType::FOR_VALUES);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetValues(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;
	ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		responseMessage.set_value(commandBox->getValues(messageRequest.lineindex()));
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processAddForLoopLine(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			responseMessage.set_value(commandBox->addForLoopLine(messageRequest.lineindex()));
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processRemoveForLoopLine(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetForLoopValueRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		ForLoopCommandBox * commandBox = getForLoopCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {
			commandBox->removeForLoopLine(messageRequest.lineindex());
			responseMessage.set_value(true);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetScanCommandBoxUserVariableNames(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringArrayResponse responseMessage;
	ScanCommandBox * commandBox = getScanCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		set<string> userVariables = commandBox->getReferencedUserVariables();

		for (const string& v : userVariables) {
			responseMessage.add_value(v);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processCreate(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::EmptyRequest messageRequest;
	messageRequest.ParseFromString(message);

	err << "Creation of a new command zone is not yet implemented" << eol;

	common::Int32Response responseMessage;

	responseMessage.set_value(0);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processReset(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::DatabaseCommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		try {
			CommandZone::getInstance(messageRequest.commandzoneid())->resetCommandZone(messageRequest.databaseid());
			responseMessage.set_value(true);

		} catch (CommandZone::CommandZoneNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processExecute(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::DatabaseCommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		try {
			CommandZone::getInstance(messageRequest.commandzoneid())->execute(endpoint, messageRequest.databaseid());
			responseMessage.set_value(true);

		} catch (CommandZone::CommandZoneNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
		} catch (CommandZone::CommandConflicts& e) {
			err << "Cannot start Launch Pad" << eol;
			responseMessage.set_error(common::Error::COMMAND_CONFLICTS);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processStop(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		try {
			CommandZone::getInstance(messageRequest.commandzoneid())->stop(endpoint);
			responseMessage.set_value(true);

		} catch (CommandZone::CommandZoneNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processStopAtEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		try {
			CommandZone::getInstance(messageRequest.commandzoneid())->stopAtEnd(endpoint);
			responseMessage.set_value(true);

		} catch (CommandZone::CommandZoneNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processPause(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		try {
			CommandZone::getInstance(messageRequest.commandzoneid())->pause(endpoint);
			responseMessage.set_value(true);

		} catch (CommandZone::CommandZoneNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processPauseAtEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		try {
			CommandZone::getInstance(messageRequest.commandzoneid())->pauseAtEnd(endpoint);
			responseMessage.set_value(true);

		} catch (CommandZone::CommandZoneNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processRestart(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		// Get the requester endpoint.
		string endpoint = request->getRequesterEndpoint();

		try {
			CommandZone::getInstance(messageRequest.commandzoneid())->restart(endpoint);
			responseMessage.set_value(true);

		} catch (CommandZone::CommandZoneNotFound& e) {
			responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
		} catch (CommandZone::CommandConflicts& e) {
			err << "Cannot start Launch Pad" << eol;
			responseMessage.set_error(common::Error::COMMAND_CONFLICTS);
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processIsRunning(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	try {
		responseMessage.set_value(CommandZone::getInstance(messageRequest.commandzoneid())->isRunning());

	} catch (CommandZone::CommandZoneNotFound& e) {
		responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processIsPaused(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	try {
		responseMessage.set_value(CommandZone::getInstance(messageRequest.commandzoneid())->isPaused());

	} catch (CommandZone::CommandZoneNotFound& e) {
		responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processIsPausingAtEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	try {
		responseMessage.set_value(CommandZone::getInstance(messageRequest.commandzoneid())->isPausingAtEnd());

	} catch (CommandZone::CommandZoneNotFound& e) {
		responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processIsStoppingAtEnd(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	try {
		responseMessage.set_value(CommandZone::getInstance(messageRequest.commandzoneid())->isStoppingAtEnd());

	} catch (CommandZone::CommandZoneNotFound& e) {
		responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetCommandZoneExecutionState(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ExecutionStateResponse responseMessage;

	try {
		CommandZone::ExecutionState state = CommandZone::getInstance(messageRequest.commandzoneid())->getExecutionState();

		if (state == CommandZone::ExecutionState::STARTED) {
			responseMessage.set_state(commandzone::ExecutionStateResponse::RUNNING);
		}
		else if (state == CommandZone::ExecutionState::PAUSED) {
			responseMessage.set_state(commandzone::ExecutionStateResponse::PAUSED);
		}
		else if (state == CommandZone::ExecutionState::STOPPING_AT_END) {
			responseMessage.set_state(commandzone::ExecutionStateResponse::STOPPING_AT_END);
		}
		else if (state == CommandZone::ExecutionState::FINISHED) {
			responseMessage.set_state(commandzone::ExecutionStateResponse::FINISHED);
		}

	} catch (CommandZone::CommandZoneNotFound& e) {
		responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetProgression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::Float64Response responseMessage;

	try {
		responseMessage.set_value(CommandZone::getInstance(messageRequest.commandzoneid())->getProgression());

	} catch (CommandZone::CommandZoneNotFound& e) {
		responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetConflictingCommands(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	try {
		responseMessage.set_value(CommandZone::getInstance(messageRequest.commandzoneid())->getConflictingCommands());

	} catch (CommandZone::CommandZoneNotFound& e) {
		responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processPrint(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	try {
		responseMessage.set_value(CommandZone::getInstance(messageRequest.commandzoneid())->print());

	} catch (CommandZone::CommandZoneNotFound& e) {
		responseMessage.set_error(common::Error::COMMAND_ZONE_NOT_FOUND);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processCheckPropertyExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::PropertyExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ExpressionResponse responseMessage;

	AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {

		try {
			// Check the expression.
			common::ExpressionError error;
			std::string lmessage;

			ExpressionStatus status = commandBox->checkPropertyExpression(messageRequest.propertyid(), messageRequest.expression(), error, lmessage);
			//ExpressionStatus status = commandBox->checkStaticPropertyExpression(messageRequest.propertyid(), messageRequest.expression(), error, message);
			processStatusAndErrorResponse(responseMessage, status, error, lmessage);
		}
		catch (AtomicCommandBox::NoSuchPropertyException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
		}
		catch (AtomicCommandBox::BadPropertyType& e) {
			responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
		}
		catch (common::Exception& e) {
			err << e.printStack() << eol;
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processCheckStaticPropertyExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::StaticPropertyExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ExpressionResponse responseMessage;

	try {
		// Check the expression.
		common::ExpressionError error;
		std::string lmessage;

		ExpressionStatus status = PropertyCommandBox::checkStaticPropertyExpression(messageRequest.propertyid(), messageRequest.expression(), error, lmessage);
		processStatusAndErrorResponse(responseMessage, status, error, lmessage);
	}
	catch (AtomicCommandBox::NoSuchPropertyException& e) {
		responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
	}
	catch (AtomicCommandBox::BadPropertyType& e) {
		responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
	}
	catch (common::Exception& e) {
		err << e.printStack() << eol;
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processSetPropertyExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::PropertyExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::ExpressionResponse responseMessage;

	// Check if the client has the token.
	common::TokenAuthorizationStatus status = SessionManager::getInstance()->checkClientTokenAuthorization(messageRequest.clientid());

	if (status == common::TokenAuthorizationStatus::CLIENT_DISCONNECTED) {
		responseMessage.set_error(common::Error::CLIENT_DISCONNECTED);
	}
	else if (status == common::TokenAuthorizationStatus::TOKEN_NOT_OWNED) {
		responseMessage.set_error(common::Error::TOKEN_NOT_OWNED);
	}
	else {
		AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

		if (commandBox == 0) {
			responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
		}
		else {

			try {
				// Check the expression.
				common::ExpressionError error;
				std::string lmessage;

				ExpressionStatus estatus = commandBox->checkPropertyExpression(messageRequest.propertyid(), messageRequest.expression(), error, lmessage);
				if (estatus == common::VALID_EXPRESSION) {
					responseMessage.set_status(commandzone::ExpressionStatus::VALID_EXPRESSION);

					// Set the expression.
					commandBox->setPropertyExpression(messageRequest.propertyid(), string(messageRequest.expression()));
				}
				else if (estatus == common::INVALID_EXPRESSION) {

					// The only errors the server does not accept are the parser errors.
					// The other errors can be "corrected" at runtime before the execution of the box.
					// The status of the expression is "changed" for the client.
					if (error == common::PARSER_ERROR) {
						responseMessage.set_status(commandzone::ExpressionStatus::INVALID_EXPRESSION);
						responseMessage.set_expressionerror(commandzone::ExpressionStatus::PARSER_ERROR);

						responseMessage.set_message(lmessage);
					}
					else {
						responseMessage.set_status(commandzone::ExpressionStatus::VALID_EXPRESSION);

						commandBox->setPropertyExpression(messageRequest.propertyid(), string(messageRequest.expression()));
					}
				}
			}
			catch (AtomicCommandBox::NoSuchPropertyException& e) {
				responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
			}
			catch (AtomicCommandBox::BadPropertyType& e) {
				responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
			}
			catch (common::Exception& e) {
				err << e.printStack() << eol;
			}
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetPropertyExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AtomicCommandBoxPropertyRequest messageRequest;
	messageRequest.ParseFromString(message);

	commandzone::GetExpressionResponse responseMessage;

	AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {

		try {
			// Get the expression.
			std:string expression = commandBox->getPropertyExpression(messageRequest.propertyid());
			responseMessage.set_expression(expression);

			// Check the expression.
			common::ExpressionError error;
			string lmessage;

			ExpressionStatus status = commandBox->checkPropertyExpression(messageRequest.propertyid(), expression, error, lmessage);
			processStatusAndErrorResponse(responseMessage, status, error, lmessage);
		}
		catch (AtomicCommandBox::NoSuchPropertyException& e) {
			responseMessage.set_error(common::Error::NO_SUCH_PROPERTY);
		}
		catch (AtomicCommandBox::BadPropertyType& e) {
			responseMessage.set_error(common::Error::BAD_PROPERTY_TYPE);
		}
		catch (common::Exception& e) {
			err << e.printStack() << eol;
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetVariableList(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandBoxRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringArrayResponse responseMessage;

	CommandBox* commandBox = getCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {

		try {
			// Get the variables.
			std::vector<std::string> variableNames;
			commandBox->addVariableNamesToList(variableNames);

			// Copy the values.
			vector<string>::const_iterator it;
			for (it = variableNames.begin(); it != variableNames.end(); it++) {
				responseMessage.add_value(*it);
			}
		}
		catch (common::Exception& e) {
			err << e.printStack() << eol;
		}
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processSaveFile(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SaveFileRequest messageRequest;

	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	FileManager::getInstance()->saveFile(messageRequest.filename(), messageRequest.content(), messageRequest.addserverhome());
	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processSaveBinaryFile(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::SaveBinaryFileRequest messageRequest;

	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	FileManager::getInstance()->saveBinaryFile(messageRequest.filename(), messageRequest.content(), messageRequest.addserverhome());
	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processGetFileContent(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetFileContentRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	string content = FileManager::getInstance()->getFileContent(messageRequest.filename());
	responseMessage.set_value(content);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processGetBinaryFileContent(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetBinaryFileContentRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	string content = FileManager::getInstance()->getBinaryFileContent(messageRequest.filename());
	responseMessage.set_value(content);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processGetFileContentAbsolutePath(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetFileContentAbsolutePathRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	string content = FileManager::getInstance()->getFileContentAbsolutePath(messageRequest.filename());
	responseMessage.set_value(content);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processGetTree(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetTreeRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;

	string content = FileManager::getInstance()->getTree(messageRequest.rootdirectory(), messageRequest.extension(), messageRequest.recurse());
	responseMessage.set_value(content);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processDeleteFile(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::DeleteFileRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	FileManager::getInstance()->deleteFile(messageRequest.filename());
	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processCreateDirectory(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CreateDirectoryRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	FileManager::getInstance()->createDirectory(messageRequest.parentdirectory(), messageRequest.filename());
	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processAppendToFile(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::AppendToFileRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	FileManager::getInstance()->appendToFile(messageRequest.filename(), messageRequest.stringtoappend());
	responseMessage.set_value(true);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processCloneForEstimatedTime(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::CommandZoneRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	EstimatedTimeManager::getInstance()->cloneCommandZone();

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processComputeEstimatedTimeState(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::ComputeEstimatedTimeStateRequest messageRequest;
	messageRequest.ParseFromString(message);

	EstimatedTimeManager::getInstance()->setActivationState(messageRequest.state());

	common::Int32Response responseMessage;
	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

void CommandZoneResponder::loop() {

	// Initialize the map of functions.
	initProcessFunctions();

	// Loop until cancel is done.
	while (true) {
		// Wait for a new request.
		unique_ptr<coms::basic::Request> request = m_responder->receive();

		// Stop the loop if cancel is called.
		if (request.get() == 0) {
			break;
		}

		// Get the request type.
		string firstPart = request->get();

		// Get the request message.
		string secondPart = request->getSecondPart();

		commandzone::Message messageType;
		messageType.ParseFromString(firstPart);

		string response;
		bool onError = false;

		// Get the process function and apply it.
		ProcessFunction & processFunction = m_processFunctions[messageType.type()];

		if (processFunction.functionPtr != nullptr) {

			// It is necessary to protect the entire process to avoid this race condition:
			// - Process command box <id> by getting the pointer from CommandBoxMap.
			// - Then in parallel, the command box is removed from the CommandBoxMap in the execution/event thread.
			// - The process is not finished and tries to access to the command box that has been deleted.
			CommandBoxMap::Lock lock;

			// Process the request.
			onError = (this->*processFunction.functionPtr)(request.get(), secondPart, response);
		}
		else {
			cerr << "Command zone responder cannot process " << messageType.type() << endl;
		}

		request->reply(response);

		// Check error.
		if (!onError) {

			// Notify the clients in case the function modifies the command zone.
			if (processFunction.notification) {
				// The clients command zones are notified only if there is no error.
				CommandZoneEventServer::getInstance()->commandZoneSync(firstPart, secondPart, response);
			}

			if (processFunction.activity) {
				// Notify the token activity.
				SessionManager::getInstance()->notifyTokenActivity();
			}
		}


	}
}


bool CommandZoneResponder::processFileExists(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::FileExistsRequest messageRequest;
	messageRequest.ParseFromString(message);
//	cout << "FILENAME " << messageRequest.filename() << endl;

	common::BooleanResponse responseMessage;

	bool exists = FileManager::getInstance()->fileExists(messageRequest.filename());
	responseMessage.set_value(exists);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

bool CommandZoneResponder::processGetLastControllerName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetLastControllerNameRequest messageRequest;
	messageRequest.ParseFromString(message);

	GenericCommandBox::BoxType type;
	switch (messageRequest.boxtype()) {
	case 0:
		type = GenericCommandBox::FIX;
		break;
	case 1:
		type = GenericCommandBox::CLEAR;
		break;
	case 2:
		type = GenericCommandBox::SETZERO;
		break;
	case 4:
		type = GenericCommandBox::SET;
		break;
	}

	common::StringResponse responseMessage;
	std::string controllerName = GenericCommandBoxManager::getInstance()->getlastControllerName(type);
	responseMessage.set_value(controllerName);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}


bool CommandZoneResponder::processGetLastSetPropertyName(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::GetLastSetPropertyNameRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::StringResponse responseMessage;
	std::string propertyName = GenericCommandBoxManager::getInstance()->getlastSetPropertyName();
	responseMessage.set_value(propertyName);

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}



bool CommandZoneResponder::processInitPropertyExpression(cameo::coms::basic::Request * request, const std::string & message, std::string & response) {

	commandzone::InitPropertyExpressionRequest messageRequest;
	messageRequest.ParseFromString(message);

	common::BooleanResponse responseMessage;

	AtomicCommandBox* commandBox = getAtomicCommandBox(messageRequest.commandboxid());

	if (commandBox == 0) {
		responseMessage.set_error(common::Error::COMMAND_BOX_NOT_FOUND);
	}
	else {
		commandBox->initPropertyExpression(messageRequest.propertyid());
		responseMessage.set_value(true);
	}

	// Serialize the response.
	responseMessage.SerializeToString(&response);

	// Return true if there are errors.
	return responseMessage.has_error();
}

}
