/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_COMMANDACCESSOR_H_
#define CORE_COMMANDACCESSOR_H_

#include <common/base/Exception.h>
#include <common/base/CommonTypes.h>
#include <string>

namespace core {

/**
 * Abstract class defining the command accessor.
 */
class CommandAccessor {

public:
	/**
	 * Command list creation exception.
	 */
	class CommandListCreationException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		CommandListCreationException(unsigned long ID);
	};

	/**
	 * No such servant exception.
	 */
	class NoSuchServantException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchServantException(unsigned long servantID);
	};

	/**
	 * No such command exception.
	 */
	class NoSuchCommandException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchCommandException(unsigned long commandID);

		/**
		 * Constructor.
		 */
		NoSuchCommandException(unsigned long servantID, const std::string& commandName);
	};

	/**
	 * Constructor
	 */
	CommandAccessor();

	/**
	 * Virtual destructor.
	 */
	virtual ~CommandAccessor();

	/**
	 * Returns the ID of the command name of given servant.
	 * throw NoSuchServantException, NoSuchCommandException
	 */
	virtual unsigned long getCommandID(unsigned long databaseID, unsigned long servantID, const std::string& propertyName) = 0;

	/**
	 * Returns the command name associated with the command ID.
	 */
	virtual std::string getCommandName(unsigned long databaseID, unsigned long commandID) = 0;

	/**
	 * Sets the state of the command
	 */
	virtual void setCommandState(unsigned long databaseID, unsigned long servantID, unsigned long commandID, common::CommandState state) = 0;

	/**
	 * Returns the state of the command
	 */
	virtual common::CommandState getCommandState(unsigned long databaseID, unsigned long commandID) const = 0;

	/**
	 * Sets the command from the command zone.
	 * throw NoSuchCommandException
	 */
	virtual void setCommandFromCommandZone(unsigned long databaseID, unsigned long commandID, bool fromCommandZone) = 0;

	/**
	 * Returns the ID of the servant that contains the command ID.
	 */
	virtual unsigned long getServantIDFromCommandID(unsigned long databaseID, unsigned long propertyID) const = 0;

	/**
	 * Sets the log ID of the command.
	 */
	virtual void setCommandLogID(unsigned long databaseID, unsigned long commandID, const std::string& logID) = 0;

	/**
	 * Returns true if is a command.
	 */
	virtual bool commandExists(unsigned long databaseID, unsigned long commandID) const = 0;

	/**
	 * Gets the sister command ID.
	 */
	virtual unsigned long getSisterCommandID(unsigned long databaseID, unsigned long commandID, const std::string& commandName) const = 0;

	/**
	 * Returns true if the command belongs to a ics servant
	 */
	virtual bool isCommandFromIcsServant(unsigned long databaseID, unsigned long commandID) const = 0;
};

}

#endif /*CORE_COMMANDACCESSOR_H_*/
