/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_PROPERTYACCESSOR_H_
#define CORE_PROPERTYACCESSOR_H_

#include <common/base/Exception.h>
#include <common/base/BaseTypes.h>

#include <string>

namespace core {

/**
 * Abstract class defining the property accessor.
 */
class PropertyAccessor {
	
public:
	/**
	 * Property list creation exception.
	 */
	class PropertyListCreationException : public common::Exception {
	
	public:
		/**
		 * Constructor.
		 */
		PropertyListCreationException(unsigned long ID); 
	};

	/**
	 * No such servant exception.
	 */
	class NoSuchServantException : public common::Exception {
	
	public:
		/**
		 * Constructor.
		 */
		NoSuchServantException(long servantID); 
	};
	
	/**
	 * No such property exception.
	 */
	class NoSuchPropertyException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchPropertyException();
	
		/**
		 * Constructor.
		 */
		NoSuchPropertyException(long propertyID);
		
		/**
		 * Constructor.
		 */ 
		NoSuchPropertyException(const std::string& propertyName); 
	};
	
	/**
	 * Bad property type exception.
	 */
	class BadPropertyTypeException : public common::Exception {
		
	public:
		/**
		 * Constructor.
		 */
		BadPropertyTypeException(long propertyID); 
	};
	
	/**
	 * Null property reference exception.
	 */
	class NullPropertyReferenceException : public common::Exception {
		
	public:
		/**
		 * Constructor.
		 */
		NullPropertyReferenceException(long propertyID); 
	};
	
	class IndexOutOfBoundsException : public common::Exception {
		
	public:
		/**
		 * Constructor.
		 */
		IndexOutOfBoundsException(unsigned long propertyID); 
	};
		
	/**
	 * Constructor
	 */
	PropertyAccessor();
	
	/**
	 * Virtual destructor.
	 */
	virtual ~PropertyAccessor();
	
	/**
	 * Returns the ID of the property name of given servant.
	 * throw NoSuchServantException, NoSuchPropertyException
	 */
	virtual unsigned long getPropertyID(unsigned long databaseID, unsigned long servantID, const std::string& propertyName) = 0;
	
	/**
	 * Gets the dynamic property size.
	 * throw NoSuchPropertyException, IndexOutOfBoundsException
	 */
	virtual unsigned long getDynamicPropertyIDAtIndex(unsigned long databaseID, unsigned long propertyID, long index, bool resize) const = 0;
	
	/**
	 * Returns the ID of the servant that contains the property ID.
	 */
	virtual unsigned long getServantIDFromPropertyID(unsigned long databaseID, unsigned long propertyID) const = 0;
	
	/**
	 * Returns the type of property.
	 * throw NoSuchPropertyException
	 */
	virtual std::string getPropertyType(unsigned long databaseID, unsigned long propertyID) = 0;

	/**
	 * Returns true if is a property.
	 */
	virtual bool propertyExists(unsigned long databaseID, unsigned long propertyID) const = 0;
	
	/**
	 * Gets the int8 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual int8 getInt8Value(unsigned long databaseID, unsigned long propertyID) const = 0;
		
	/**
	 * Gets the int16 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual int16 getInt16Value(unsigned long databaseID, unsigned long propertyID) const = 0;

	/**
	 * Gets the int32 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual int32 getInt32Value(unsigned long databaseID, unsigned long propertyID) const = 0;

	/**
	 * Gets the int64 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual int64 getInt64Value(unsigned long databaseID, unsigned long propertyID) const = 0;

	/**
	 * Gets the float32 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual float32 getFloat32Value(unsigned long databaseID, unsigned long propertyID) const = 0;

	/**
	 * Gets the float64 value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual float64 getFloat64Value(unsigned long databaseID, unsigned long propertyID) const = 0;

	/**
	 * Gets the string value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual std::string getStringValue(unsigned long databaseID, unsigned long propertyID) const = 0;

	/**
	 * Gets the boolean value for the property of ID propertyID.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	virtual bool getBooleanValue(unsigned long databaseID, unsigned long propertyID) const = 0;
	
	/**
	 * Gets the formatted value for the property of ID propertyID.
	 * throw NoSuchPropertyException
	 */
	virtual std::string getFormattedValue(unsigned long databaseID, unsigned long propertyID) const = 0;

};

}

#endif /*CORE_PROPERTYACCESSOR_H_*/
