/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_SERVANTACCESSOR_H_
#define CORE_SERVANTACCESSOR_H_

#include <common/base/Exception.h>


namespace core {

class ServantAccessor {

public:

	/**
	 * Exception for servant.
	 */
	class NoSuchServantException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchServantException(const std::string servantName);

		/**
		 * Constructor.
		 */
		NoSuchServantException(unsigned long servantID);
	};

	/**
	 * Initialisation exception.
	 */
	class InitialisationException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		InitialisationException();
	};

	/**
	 * Servant type.
	 */
	typedef enum {CONTROLLER, DRIVER} ServantType;

	/**
	 * Constructor.
	 */
	ServantAccessor();

	/**
 	 * Virtual destructor.
 	 */
	virtual ~ServantAccessor();

	/**
	 * Initialises.
	 */
	virtual void init() = 0;

	/**
	 * Returns the ID corresponding to the servant name.
	 * throw NoSuchServantException
	 */
	virtual unsigned long getServantID(const std::string& servantName) const = 0;

	/**
	 * Returns the servant name associated with the servant ID.
	 */
	virtual std::string getServantName(unsigned long databaseID, unsigned long servantID) = 0;

	/**
	 * Gets the servant type.
	 * throw NoSuchServantException
	 */
	virtual ServantType getServantType(unsigned long databaseID, unsigned long servantID) const = 0;

	/**
	 * Gets the servant type.
	 * throw NoSuchServantException
	 */
	virtual ServantType getServantType(unsigned long databaseID, const std::string& servantName) const = 0;

	/**
	 * Gets the controller type.
	 * throw NoSuchServantException
	 */
	virtual std::string getControllerType(unsigned long databaseID, unsigned long controllerID) const = 0;

	/**
	 * Gets the driver type.
	 * throw NoSuchServantException
	 */
	virtual std::string getDriverType(unsigned long databaseID, unsigned long driverID) const = 0;

	/**
	 * Returns true if the servant is enabled.
	 */
	virtual bool isServantEnabledAndInitialised(unsigned long databaseID, unsigned long servantID) const = 0;

	/**
	 * Gets the description of the element.
	 */
	virtual std::string what(unsigned long databaseID, unsigned long elementID) const = 0;

	/**
	 * Returns true if is a servant.
	 */
	virtual bool servantExists(unsigned long databaseID, unsigned long servantID) const = 0;

	/**
	 * Returns true if servant is pausable.
	 */
	virtual bool isServantPausable(unsigned long databaseID, unsigned long servantID) const = 0;

	/**
	 * Returns the driver channel of a driver for a given ID.
	 * throw NoSuchServantException
	 */
	virtual long getDriverChannel(unsigned long databaseID, unsigned long servantID) const = 0;
};

} // namespace command

#endif /*CORE_SERVANTACCESSOR_H_*/