/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "UserVariable.h"
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/dataaccess/DataAccess.h>
#include <core/transportlayer/MultiThreadCommunication.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <iostream>
#include <algorithm>
#include <boost/tokenizer.hpp>
#include <boost/algorithm/string.hpp>
#include "../transportlayer/dataprovider/StateAccessor.h"

using namespace std;

namespace core {

UserVariable::UserVariable(int id, const std::string& name) :
	_id(id), _name(name), _enabled(true), _value(0.0) {

	_stateAccessor = MultiThreadCommunication::getInstance()->getStateAccessor();
}

UserVariable::~UserVariable() {

	if (_mathExpression != 0) {
		// removing the subscribers
		set<unsigned long> propertyIDs = _mathExpression->getPropertyIDs();
		for (set<unsigned long>::const_iterator p = propertyIDs.begin(); p != propertyIDs.end(); p++) {
			DataNotificationRouter::getInstance()->removePropertyChangeSubscriber(*p, this);
		}
	}
}

void UserVariable::verify(const std::string& expression) {

	// Create a new math expression.
	MathExpression mathExpression(expression);
	mathExpression.init();

	// Check the value to generate an exception if there is an error.
	// Do not check for loop variables as we do not have the command box context here.
	mathExpression.checkValue(false);
}

void UserVariable::init() {

	// Protect from concurrent access but release the lock after the block.
	{
		std::scoped_lock lock(_mutex);

		_mathExpression->init();

		// adding the subscribers
		_propertyIDs = _mathExpression->getPropertyIDs();

		for (unsigned long i : _propertyIDs) {
			cout << "Variable " << _name << " depends on property " << i << endl;
		}

		// Reset the servant IDs before re-inserting the updated ones.
		_servantIDs.clear();
	}

	// The lock is released.
	for (set<unsigned long>::const_iterator p = _propertyIDs.begin(); p != _propertyIDs.end(); p++) {

		// Add the subscriber without locking because Condition::propertyChanged can be called in parallel and block.
		DataNotificationRouter::getInstance()->addPropertyChangeSubscriber(*p, this);

		// Protect _servantIDs.
		std::scoped_lock lock(_mutex);

		// get the servant ID
		unsigned long servantID = DataAccess::getInstance()->getServantIDFromPropertyID(0, *p);
		_servantIDs.insert(servantID);

		cout << "Variable " << _name << " depends on servant " << servantID << endl;
	}

	// Update the value to notify the clients.
	updateValue();
}

void UserVariable::reset() {
}

int UserVariable::getId() const {
	return _id;
}

const std::string& UserVariable::getName() const {
	return _name;
}

void UserVariable::setExpression(const std::string& expression) {

	std::scoped_lock lock(_mutex);

	_mathExpression.reset(new MathExpression(expression));
}

std::string UserVariable::getExpression() {

	std::scoped_lock lock(_mutex);

	return _mathExpression->getExpression();
}

bool UserVariable::isRemovable() {
	return true;
}

std::set<unsigned long> UserVariable::getPropertyIDs() {

	std::scoped_lock lock(_mutex);

	return _propertyIDs;
}

std::set<unsigned long> UserVariable::getServantIDs() {

	std::scoped_lock lock(_mutex);

	return _servantIDs;
}

std::set<std::string> UserVariable::getUserVariableNames() {

	std::scoped_lock lock(_mutex);

	return _mathExpression->getUserVariableNames();
}

void UserVariable::setEnabled(bool value) {

	std::scoped_lock lock(_mutex);

	_enabled = value;

	cout << "Variable " << _name << " is enabled? " << value << endl;

	_stateAccessor->notifyUserVariableStateChanged(0, _name, value);
}

bool UserVariable::isEnabled() {

	std::scoped_lock lock(_mutex);

	return _enabled;
}

bool UserVariable::evaluateValue() {

	std::scoped_lock lock(_mutex);

	// Evaluate the expression.
	double value = _mathExpression->evaluateValue();

	// Change the value and notify if the value has changed.
	if (value != _value) {
		_value = value;

		return true;
	}

	return false;
}

void UserVariable::updateValue() {

	// Evaluate the value and notify the value has changed.
	if (evaluateValue()) {
		_stateAccessor->notifyUserVariableValueChanged(0, _name, _value);
	}
}

double UserVariable::getValue() {

	std::scoped_lock lock(_mutex);

	return _value;
}

void UserVariable::propertyChanged(unsigned long databaseID, unsigned long propertyID) {

	updateValue();
}

std::string UserVariable::toString(bool saveEnable, const std::string& space) {

	std::scoped_lock lock(_mutex);

	string result = space + "<variable ";

	result += "name=\"" + getName() + "\"";

	if (saveEnable) {
		string enabledString = " enabled=\"";
		if (isEnabled()) {
			enabledString += "true\"";
		}
		else {
			enabledString += "false\"";
		}

		result += enabledString;
	}

	result += ">\n";

	result += space + "  <expression value=\"" + getExpression() + "\"/>\n";
	result += space + "</variable>";

	return result;
}

std::string UserVariable::toXMLString(bool saveEnable, const std::string& space) {
	return toString(saveEnable, space);
}

}
