/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_USERVARIABLE_H_
#define CORE_USERVARIABLE_H_

#include <common/base/CommonTypes.h>
#include <core/base/MathExpression.h>
#include <core/dataaccess/PropertyChangeSubscriber.h>
#include <mutex>
#include <memory>
#include <string>

namespace core {

class StateAccessor;

/**
 * Class defining a user variable for the UserVariableManager.
 */
class UserVariable : public PropertyChangeSubscriber {

public:
	/**
	 * Constructor.
	 */
	explicit UserVariable(int id, const std::string& name);

	/**
	 * Destructor.
	 */
	virtual ~UserVariable();

	/**
	 * Verifies the expression.
	 */
	static void verify(const std::string& expression);

	/**
	 * Initialises the condition.
	 */
	void init();

	/**
	 * Resets the condition.
	 */
	void reset();

	/**
	 * Gets the id.
	 */
	int getId() const;

	/**
	 * Gets the name.
	 */
	const std::string& getName() const;

	/**
	 * Sets the expression.
	 */
	void setExpression(const std::string& expression);

	/**
	 * Gets the expression.
	 */
	std::string getExpression();

	/**
	 * Returns true if is removable.
	 */
	bool isRemovable();

	/**
	 * Returns the list of all the property IDs it depends on. Init must have been called before.
	 */
	std::set<unsigned long> getPropertyIDs();

	/**
	 * Returns the list of all the servant IDs it depends on. Init must have been called before.
	 */
	std::set<unsigned long> getServantIDs();

	/**
	 * Returns the set of user variable names.
	 */
	std::set<std::string> getUserVariableNames();

	/**
	 * Sets enabled.
	 */
	void setEnabled(bool value);

	/**
	 * Returns true if is enabled.
	 */
	bool isEnabled();

	/**
	 * Updates the value.
	 */
	void updateValue();

	/**
	 * Gets the value.
	 */
	double getValue();

	/**
	 * Notifies that the property has changed.
	 */
	virtual void propertyChanged(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Returns the stringified content of the condition.
	 */
	virtual std::string toString(bool saveEnable, const std::string& space);

	/**
	 * Returns the xml stringified content of the condition.
	 */
	virtual std::string toXMLString(bool saveEnable, const std::string& space);

protected:
	std::recursive_mutex _mutex;

private:
	bool evaluateValue();

	int _id;
	std::string _name;

	std::set<unsigned long> _propertyIDs;
	std::set<unsigned long> _servantIDs;

	std::unique_ptr<MathExpression> _mathExpression;
	bool _enabled;
	double _value;

	StateAccessor* _stateAccessor;
};

}

#endif
