/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR VariableS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "UserVariableManager.h"
#include "UserVariable.h"
#include <core/dataaccess/DataAccess.h>
#include <core/parser/variables/VariablesParser.h>
#include <core/parser/variables/VariableElement.h>
#include <core/dataaccess/DataNotificationRouter.h>
#include <core/base/BaseValueWrapper.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <core/log/LogHandler.h>
#include <core/log/SimpleLog.h>
#include <boost/filesystem.hpp>
#include <iostream>
#include <fstream>
#include <algorithm>

using namespace std;
using namespace common;
using namespace boost::filesystem;

namespace core {

UserVariableManager* UserVariableManager::_instance = 0;
const std::string UserVariableManager::VARIABLE_PATH = "/Variables.xml";

UserVariableManager::UserVariableExistsException::UserVariableExistsException() :
	Exception("core::VariableManager::NameAlreadyExistsException", "name already exists") {
};

UserVariableManager::NoSuchUserVariableException::NoSuchUserVariableException() :
	Exception("core::VariableManager::NoSuchVariableException", "name does not exist") {
};

UserVariableManager::UnremovableUserVariableException::UnremovableUserVariableException(const std::string& message) :
	Exception("core::VariableManager::UnremovableVariableException", message) {
};

UserVariableManager::DisabledUserVariableException::DisabledUserVariableException(const std::string& message) :
	Exception("core::VariableManager::DisabledUserVariableException", message) {
};

UserVariableManager::InvalidExpressionException::InvalidExpressionException(const std::string& message) :
	Exception("core::VariableManager::InvalidExpressionException", message) {
};

UserVariableManager::UserVariableManager() {

	DataNotificationRouter::getInstance()->addConfigurationChangeSubscriber(this);
}

UserVariableManager::~UserVariableManager() {

	DataNotificationRouter::getInstance()->removeConfigurationChangeSubscriber(this);

	for (auto v : _variables) {
		delete v.second;
	}
}

UserVariableManager* UserVariableManager::getInstance() {
	return _instance;
}

void UserVariableManager::initInstance(const std::string& configPath) {

	if (_instance == 0) {
		_instance = new UserVariableManager();
		_instance->init(configPath);
	}
}

void UserVariableManager::resetInstance() {
	delete _instance;
	_instance = 0;
}

bool UserVariableManager::userVariableExists(const std::string& variableName) const {
	for (auto v : _variables) {
		if (v.first == variableName) {
			return true;
		}
	}
	return false;
}

void UserVariableManager::init(const std::string& configPath) {

	_userVariableManagerConfigFileName = configPath + VARIABLE_PATH;

	// Create the parent directory if it does not exist.
	path filePath(_userVariableManagerConfigFileName);
	create_directory(filePath.parent_path());
}

void UserVariableManager::run() {
	readUserVariableManagerFile();
}

void UserVariableManager::stop() {
}

void UserVariableManager::verifyExpression(const std::string& expression, const std::string& name) {

	try {
		UserVariable::verify(expression);

	} catch (const Exception& e) {
		throw InvalidExpressionException(e.getMessage());
	}
}

void UserVariableManager::addUserVariable(const std::string& name, const std::string& expression, bool fromFile) {

	std::scoped_lock lock(_mutex);

	if (userVariableExists(name)) {
		throw UserVariableExistsException();
	}

	UserVariable* variable = new UserVariable(_variables.size() + 1, name);
	variable->setExpression(expression);

	// Check if the error is accepted.
	common::ExpressionError error = common::NO_EXPRESSION_ERROR;
	string errorMessage;

	try {
		variable->init();
	}
	catch (const MathExpression::ParserError & e) {
		throw InvalidExpressionException(e.getMessage());
	}
	catch (const MathExpression::DisabledControllerError & e) {
		error = common::DISABLED_CONTROLLER_ERROR;
		errorMessage = e.getMessage();
	}
	catch (const MathExpression::NoSuchEnumeratedValueError & e) {
		error = common::ENUMERATED_VALUE_ERROR;
		errorMessage = e.getMessage();
	}
	catch (const MathExpression::UndefinedVariableError & e) {
		// Undefined variable errors are not accepted.
		throw InvalidExpressionException(e.getMessage());
	}
	catch (const InvalidExpressionException & e) {
		throw;
	}
	catch (...) {
		throw InvalidExpressionException("Unknown error");
	}

	// Add the variable.
	_variables.push_back(make_pair(name, variable));

	// Disable the variable in case of error.
	if (error != common::NO_EXPRESSION_ERROR) {
		variable->setEnabled(false);
		string message = "Variable '";
		message += name + "' is disabled : " + errorMessage;
		LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, message));
	}

	// Do not save the file if the variable is added from reading the file.
	if (!fromFile) {
		// Save the file.
		writeUserVariableManagerFile();

		// Disable the variable in case of error.
		if (error != common::NO_EXPRESSION_ERROR) {
			variable->setEnabled(false);
			throw DisabledUserVariableException(errorMessage);
		}
	}
}

bool UserVariableManager::isUserVariableReferenced(const std::string& name) const {

	// Iterate and verify each variable.
	for (auto v : _variables) {
		// Check if name belongs to the dependent variable names.
		set<string> variableNames = v.second->getUserVariableNames();

		if (variableNames.find(name) != variableNames.end()) {
			return true;
		}
	}

	return false;
}

void UserVariableManager::removeUserVariable(const std::string& name) {

	std::scoped_lock lock(_mutex);

	UserVariable* variable = getUserVariable(name);

	// The variable must be removable and not referenced by another variable.
	if (!variable->isRemovable()) {
		throw UnremovableUserVariableException("Variable is not removable");
	}

	if (isUserVariableReferenced(name)) {
		throw UnremovableUserVariableException("Variable is referenced and cannot be removed");
	}

	// We can delete the variable.
	delete variable;

	// Erase from the vector.
	auto it = _variables.begin();
	while (it != _variables.end()) {
		if (it->first == name) {
			break;
		}
		++it;
	}
	_variables.erase(it);

	// Save file.
	writeUserVariableManagerFile();
}

void UserVariableManager::editUserVariable(const std::string& name, const std::string& expression) {

	std::scoped_lock lock(_mutex);

	if (!userVariableExists(name)) {
		throw NoSuchUserVariableException();
	}

	// Get the existing variable.
	UserVariable* variable = getUserVariable(name);

	// First define a temporary expression to check circularities and disallowed errors.
	unique_ptr<MathExpression> mathExpression;

	try {
		mathExpression.reset(new MathExpression(expression));
		mathExpression->init();

		// Call evaluateValue to call addValue.
		mathExpression->evaluateValue();
	}
	catch (const MathExpression::ParserError & e) {
		throw InvalidExpressionException(e.getMessage());
	}
	catch (const MathExpression::DisabledControllerError & e) {
		// Is processed afterwards.
	}
	catch (const MathExpression::NoSuchEnumeratedValueError & e) {
		// Is processed afterwards.
	}
	catch (const MathExpression::UndefinedVariableError & e) {
		// Undefined variable errors are not accepted.
		throw InvalidExpressionException(e.getMessage());
	}
	catch (const InvalidExpressionException & e) {
		throw;
	}
	catch (...) {
		throw;
	}

	// Check the circularity.
	set<string> variableNames = mathExpression->getUserVariableNames();

	try {
		// Iterate the dependent variable names.
		for (const string& variableName : variableNames) {
			checkDependencies(variable, getUserVariable(variableName));
		}
	}
	catch (const InvalidExpressionException & e) {
		// If a circularity is detected.
		throw;
	}

	// The temporary expression is no longer needed.
	mathExpression.reset();

	// The variable can be updated.
	// Check the expression and check if the error is accepted.
	common::ExpressionError error = common::NO_EXPRESSION_ERROR;
	string errorMessage;

	try {
		variable->setExpression(expression);
		variable->init();
	}
	catch (const MathExpression::DisabledControllerError & e) {
		error = common::DISABLED_CONTROLLER_ERROR;
		errorMessage = e.getMessage();
	}
	catch (const MathExpression::NoSuchEnumeratedValueError & e) {
		error = common::ENUMERATED_VALUE_ERROR;
		errorMessage = e.getMessage();
	}
	catch (...) {
		throw InvalidExpressionException("Unknown error");
	}

	// Disable the variable in case of error.
	if (error != common::NO_EXPRESSION_ERROR) {
		variable->setEnabled(false);
		string message = "Variable '";
		message += name + "' is disabled : " + errorMessage;
		LogHandler::getInstance()->write(SimpleLog(common::LOG_INFO, message));
	}

	// Save the file.
	writeUserVariableManagerFile();

	// Disable the variable in case of error.
	if (error != common::NO_EXPRESSION_ERROR) {
		variable->setEnabled(false);
		throw DisabledUserVariableException(errorMessage);
	}

}

UserVariable* UserVariableManager::getUserVariable(const std::string& variableName) {

	for (auto v : _variables) {
		if (variableName == v.first) {
			return v.second;
		}
	}

	throw NoSuchUserVariableException();
}

std::set<std::string> UserVariableManager::getUsableUserVariableNames(const std::string& variableName) {

	set<string> result;

	// Iterate and take all the variables defined before the variable with name variableName.
	for (auto v : _variables) {
		if (variableName == v.first) {
			break;
		}
		else {
			result.insert(v.first);
		}
	}

	return result;
}

void UserVariableManager::checkDependencies(UserVariable * variableToEdit, UserVariable * variable) {

//	cout << "Check dependencies " << variableToEdit->getName() << " (" << variableToEdit->getId()
//					<< ") with " << variable->getName() << " (" << variable->getId() << ")" << endl;

	if (variableToEdit->getName() == variable->getName()) {
		throw InvalidExpressionException("Circular definition");
	}

	if (variableToEdit->getId() < variable->getId()) {
		throw InvalidExpressionException("A variable can only depend on previously defined variables");
	}

	// Check if name belongs to the dependent variable names.
	set<string> variableNames = variable->getUserVariableNames();

	// Iterate the variable with variable names.
	for (const string& variableName : variableNames) {
		checkDependencies(variableToEdit, getUserVariable(variableName));
	}
}

void UserVariableManager::configurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState enable) {

	std::scoped_lock lock(_mutex);

	// iterating the conditions
	for (auto v : _variables) {

		UserVariable * variable = v.second;

		// Copy the servantIDs set because it is returned by value.
		set<unsigned long> servantIDs = variable->getServantIDs();

		// test if the variable depends on the servant
		if (servantIDs.find(servantID) != servantIDs.end()) {

			// sets the variable off if it is on
			// the variable is not set on automatically
			if ((enable == common::DISABLED) && variable->isEnabled()) {
				variable->setEnabled(false);
			}
			else if ((enable == common::ENABLED) && !variable->isEnabled()) {
				variable->setEnabled(true);
			}
		}
	}
}

void UserVariableManager::readUserVariableManagerFile() {

	path filePath(_userVariableManagerConfigFileName);

	if (exists(filePath)) {

		VariablesParser* parser = new VariablesParser();
		parser->init();
		parser->parseXMLFile(_userVariableManagerConfigFileName);

		const vector<VariableElement*>& elements = parser->getVariableElements();
	
		for (vector<VariableElement*>::const_iterator e = elements.begin(); e != elements.end(); e++) {

			try {
				addUserVariable((*e)->getName(), (*e)->getExpression(), true);
			}
			catch (common::Exception const & ex) {
				string message = "Variable '";
				message += (*e)->getName() + "' cannot be defined : " + ex.getMessage();
				LogHandler::getInstance()->write(SimpleLog(common::LOG_ERROR, message));
			}
		}

		delete parser;
	}
	else {
		err << "Cannot find Variables.xml file" << eol;
	}
}

std::string UserVariableManager::toXMLString(bool saveEnable) {

	string result = "<variables>\n";

	for (auto v : _variables) {
		result += v.second->toXMLString(saveEnable, "  ") + "\n";
	}

	result += "</variables>";

	return result;
}

std::string UserVariableManager::toString() {

	string result = "<variables>\n";

	for (auto v : _variables) {
		result += v.second->toString(true, "  ") + "\n";
	}

	result += "</variables>";

	return result;
}

void UserVariableManager::writeUserVariableManagerFile() {

	// write the file
	std::ofstream VariableFile(_userVariableManagerConfigFileName.c_str(), ios::out);
	VariableFile << toXMLString(false) << endl;
	VariableFile.close();

	// copy a backup file
	path filePath(_userVariableManagerConfigFileName);
	path backupFilePath(_userVariableManagerConfigFileName + ".bak");

	try {
		remove(backupFilePath);
		copy_file(filePath, backupFilePath);
	} catch (const filesystem_error & e) {
		out << "Cannot backup Variables.xml file" << eol;
	}
}

}
