/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR VariableS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CORE_USERVARIABLEMANAGER_H_
#define CORE_USERVARIABLEMANAGER_H_

#include <common/base/CommonTypes.h>
#include <common/base/Exception.h>
#include <core/dataaccess/ConfigurationChangeSubscriber.h>
#include <mutex>
#include <string>
#include <vector>
#include <set>

namespace core {

class UserVariable;

/**
 * Class defining a user variable manager.
 */
class UserVariableManager : public ConfigurationChangeSubscriber {

public:
	/**
	 * Exception when creating a user variable with existent name.
	 */
	class UserVariableExistsException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		UserVariableExistsException();
	};

	/**
	 * Exception when removing a user variable with non-existent name.
	 */
	class NoSuchUserVariableException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchUserVariableException();
	};

	/**
	 * Exception when removing an unremovable user variable.
	 */
	class UnremovableUserVariableException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		UnremovableUserVariableException(const std::string& message);
	};

	/**
	 * Exception when a user variable is disabled.
	 */
	class DisabledUserVariableException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		DisabledUserVariableException(const std::string& message);
	};

	/**
	 * Invalid expression exception.
	 */
	class InvalidExpressionException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		InvalidExpressionException(const std::string& message);
	};

	/**
	 * Returns the instance.
	 */	
	static UserVariableManager* getInstance();
	
	/**
	 * Initialises the instance.
	 */
	static void initInstance(const std::string& configPath);
	
	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Starts the survey.
	 */
	void run();
	
	/**
	 * Stops the survey.
	 */
	void stop();

	/**
	 * Verifies the expression for a variable with name.
	 * throws InvalidExpressionException.
	 */
	void verifyExpression(const std::string& expression, const std::string& name);

	/**
	 * Adds a variable to the manager.
	 * throws VariableExistsException, InvalidExpressionException.
	 */
	void addUserVariable(const std::string& name, const std::string& expression, bool fromFile);

	/**
	 * Removes a user variable with name.
	 * throws NoSuchVariableException, UnremovableVariableException.
	 */
	void removeUserVariable(const std::string& name);

	/**
	 * Edits a user variable.
	 * throws NoSuchVariableException, InvalidExpressionException.
	 */
	void editUserVariable(const std::string& name, const std::string& expression);

	/**
	 * Gets the user variable.
	 * throws NoSuchUserVariableException.
	 */
	UserVariable* getUserVariable(const std::string& name);

	/**
	 * Gets the usable user variables for the definition of the variable.
	 */
	std::set<std::string> getUsableUserVariableNames(const std::string& variableName);

	/**
	 * Notifies that the configuration has changed.
	 */
	virtual void configurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState enable);

	/**
	 * Gets the user variables with xml content.
	 */
	std::string toXMLString(bool saveEnable);

	/**
	 * Returns the stringified content of the user variable manager.
	 */
	std::string toString();

private:
	UserVariableManager();
	~UserVariableManager();

	bool isUserVariableReferenced(const std::string& name) const;
	bool userVariableExists(const std::string& name) const;
	void checkDependencies(UserVariable * variableToEdit, UserVariable * variable);

	void init(const std::string& configPath);
	void readUserVariableManagerFile();
	void writeUserVariableManagerFile();

	static UserVariableManager* _instance;
	static const std::string VARIABLE_PATH;
	
	std::string _userVariableManagerConfigFileName;
	std::recursive_mutex _mutex;
	std::vector<std::pair<std::string, UserVariable*>> _variables;
};

}

#endif
