/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DirectDataProvider.h"
#include <common/base/ServerProperties.h>
#include <dataprovider/transportlayer/MultiThreadCommunication.h>
#include <dataprovider/database/Database.h>
#include <dataprovider/database/Lock.h>
#include <dataprovider/log/LogHandler.h>
#include <dataprovider/log/ServerLogHandler.h>
#include <dataprovider/notification/NotificationServer.h>
#include <dataprovider/iconandfamily/ServantIconAndFamilyMaps.h>
#include <common/base/ServerProperties.h>
#include <common/base/Convert.h>
#include <common/output/OutputStream.h>
#include <ics/ServantConfiguration.h>
#include <boost/lexical_cast.hpp>
#include <cameo/api/cameo.h>
#include <iostream>
#include <signal.h>
#include <cstdlib>

using namespace std;
using namespace boost;
using namespace dataprovider;
using namespace common;
using namespace cameo;

namespace dataprovider {

const std::string DirectDataProvider::ICS = "-ics";
const std::string DirectDataProvider::SIMULATION = "sim";
const std::string DirectDataProvider::SIMULATION3D = "sim3d";

DirectDataProvider::DirectDataProvider() :
	Application(),
	_onError(false) {
}

DirectDataProvider::~DirectDataProvider() {
}


void DirectDataProvider::init(int argc, char* argv[]) {

	string serverId;

	out.setStdOut(true);
	err.setStdOut(true);

	for (int i = 0; i < argc; i++) {

		string oldSimulation = string("-") + SIMULATION;
		if ((string(argv[i]) == SIMULATION || string(argv[i]) == SIMULATION3D || string(argv[i]) == oldSimulation)) {
			serverId = lexical_cast<string>(This::getId());
			out << "Server simulated " << serverId << eol;
		}
	}

	// Default configuration path.
	string configPath = ServerProperties::getInstance()->getConfigPath() + "/Server";

	// initialising the server log and communication components
	ServerLogHandler::initInstance();

	MultiThreadCommunication::initInstance(argc, argv, false, serverId);

	// initialising the data provider components
	IndexGenerator::initInstance();
	LogHandler::initInstance();

	if (serverId != "") {
		LogHandler::getInstance()->setOutputName(string("sim") + serverId);
	}

	NotificationServer::initInstance(serverId);

	try {
		ics::ServantConfiguration::initInstance();
	} catch (const Exception& e) {
		_onError = true;
		err << e.printStack() << eol;
		return;
	}

	try {
		Database::initInstance("", false);

	} catch (const Exception& e) {
		_onError = true;
		err << e.printStack() << eol;
	}

	// converting serverId
	unsigned long databaseID = 0;
	if (serverId != "") {

		istringstream is(serverId);
		is >> databaseID;

		Database::getInstance()->setID(databaseID);
	}
}

void DirectDataProvider::run() {

	if (_onError) {
		MultiThreadCommunication::getInstance()->shutdown(false);

	} else {

		MultiThreadCommunication::getInstance()->run();

		out << startup << "Dataprovider running" << eol;
	}

}

void DirectDataProvider::reset() {

	out << common::shutdown << "Dataprovider shutdown" << eol;

	MultiThreadCommunication::getInstance()->stop();

	// stopping the event service
	NotificationServer::resetInstance();

	try {
		// resetting the data provider components
		LogHandler::resetInstance();
		Database::resetInstances();
		IndexGenerator::resetInstances();

		// resetting the server log and communication
		ServerLogHandler::resetInstance();
		MultiThreadCommunication::resetInstance();

	} catch (MultiThreadCommunication::DestructionFailureException & e) {
		cerr << e.printStack() << endl;
	}
}

void DirectDataProvider::shutdown(bool isRemote) {

	MultiThreadCommunication::getInstance()->shutdown(isRemote);
}

void DirectDataProvider::show() {

	out << nolog << "database :" << eol;
	out << nolog << Database::getInstance()->toString() << eol;
}

void DirectDataProvider::execute(const std::string& commandName, const std::vector<std::string>& args) {

	if (commandName == "show") {
		if (args.empty()) {
			show();
		} else {
			if (args.size() >= 1) {

				// first testing name
				string name = args[0];
				for (int i = 1; i < args.size(); ++i) {
					name += " " + args[i];
				}

				bool success(true);
				try {
					int elementID = Database::getInstance()->getServantID(name);
					out << nolog << Database::getInstance()->toString(elementID) << eol;
				} catch (...) {
					success = false;
				}

				if (!success) {
					try {
						int elementID = lexical_cast<int>(args[0]);
						out << nolog << Database::getInstance()->toString(elementID) << eol;

					} catch (...) {
						out << nolog << "no servant for " << name << eol;
					}
				}
			}
		}

	} else if (commandName == "help") {
		out << nolog << "commands :" << eol;
		out << nolog  << "\tshow : prints the content of the database" << eol;
		out << nolog  << "\tshow <ID> : prints the content of the servant <ID>" << eol;
		out << nolog  << "\twhat <ID> : prints the element <ID>" << eol;
		out << nolog  << "\tlocks <seconds> : verify the locks with timeout in <seconds>" << eol;
		out << nolog  << "\tshowjson : shows the json messages" << eol;
		out << nolog  << "\thidejson : hides the json messages" << eol;
		out << nolog  << "\tquit : quits the application" << eol;

	} else if (commandName == "what") {
		if (args.size() >= 1) {
			int elementID = to<int>(args[0]);
			out << nolog << Database::getInstance()->what(elementID) << eol;
		}

	} else if (commandName == "lock") {
		if (args.size() >= 1) {
			int elementID = to<int>(args[0]);
			_lockMap.insert(elementID, Database::getInstance()->lock(elementID));
		}

	} else if (commandName == "unlock") {
		if (args.size() >= 1) {
			int elementID = to<int>(args[0]);
			try {
				delete _lockMap.get(elementID);
				_lockMap.erase(elementID);

			} catch (Exception& e) {
				Database::getInstance()->unlock(elementID);
			}
		}

	} else if (commandName == "locks") {
		if (args.size() >= 1) {
			int seconds = to<int>(args[0]);
			Database::getInstance()->verifyLocks(seconds);
		}

	} else if (commandName == "showjson") {
		LogHandler::getInstance()->showJSONMessages(true);

	} else if (commandName == "hidejson") {
		LogHandler::getInstance()->showJSONMessages(false);

	} else {
		out << nolog << "command '" << commandName << "' unknown" << eol;
	}
}

void DirectDataProvider::quit(const std::vector<std::string>& args, bool isRemote) {

	shutdown(isRemote);

}

}
