/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Database.h"

#include <boost/lexical_cast.hpp>
#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <deque>
#include <iostream>
#include <iterator>
#include <list>
#include <map>
#include <utility>

#include "../../common/base/BaseTypes.h"
#include "../../common/base/Convert.h"
#include "../../common/base/Real.h"
#include "../../common/base/ServerProperties.h"
#include "../../common/base/StandardHashMap.h"
#include "../../common/base/StringUtilities.h"
#include "../../common/base/TypeToString.h"
#include "../../common/output/EndOfLine.h"
#include "../../common/output/OutputStream.h"
#include "../../ics/DataProvider/Property/AnyProperty.h"
#include "../dataelement/Controller.h"
#include "../dataelement/Driver.h"
#include "../dataelement/DynamicProperty.h"
#include "../dataelement/ElementaryProperty.h"
#include "../dataelement/IcsArrayProperty.h"
#include "../dataelement/Servant.h"
#include "../notification/DataChangePublisher.h"
#include "../notification/NotificationServer.h"
#include "../transportlayer/common/DataChangeSubscriber.h"
#include "DatabaseLoader.h"
#include "ElementLockManager.h"
#include "ScopedLock.h"
#include "StandardLockableHashMap.h"

using namespace common;
using namespace std;
using namespace boost;

namespace dataprovider {

const std::string Database::LOCK_MANAGER_TYPE = "lockManagerType";
const std::string Database::NO_LOCK_MANAGER = "NoLockManager";
const std::string Database::SINGLE_LOCK_MANAGER = "SingleLockManager";
const std::string Database::ELEMENT_LOCK_MANAGER = "ElementLockManager";
const std::string Database::MAX_TRANSFER_SIZE_BYTES = "maxTransferSizeBytes";

Database::NoSuchServantException::NoSuchServantException(unsigned long servantID)
: Exception("dataprovider::Database::NoSuchServantException", string("unable to find servant <") + to<string>(servantID) + ">") {
}

Database::NoSuchServantException::NoSuchServantException(unsigned long parentID, const std::string& servantName, long childIndex)
: Exception("dataprovider::Database::NoSuchServantException", string("unable to find servant '") + servantName + "' of parent <" + to<string>(parentID) + ">" + (childIndex != -1 ? (" at index " + to<string>(childIndex)) : "")) {
}

Database::NoSuchServantException::NoSuchServantException(const std::string& servantName)
: Exception("dataprovider::Database::NoSuchServantException", string("unable to find servant '") + servantName + "'") {
}

Database::ServantDisabledException::ServantDisabledException(unsigned long parentID, const std::string& servantName, long childIndex)
: Exception("dataprovider::Database::ServantDisabledException", string("disabled servant '") + servantName + "' of parent <" + to<string>(parentID) + ">" + (childIndex != -1 ? (" at index " + to<string>(childIndex)) : "")) {
}

Database::ServantEnableException::ServantEnableException(const std::string& servantName, unsigned long servantID)
: Exception("dataprovider::Database::ServantEnableException", string("cannot enable servant '") + to<string>(servantName) + "' with ID <" + to<string>(servantID) + ">") {
}

Database::ServantEnableException::ServantEnableException(const std::string& parentName, unsigned long parentID, const std::string& servantName, long childIndex)
: Exception("dataprovider::Database::ServantEnableException", string("cannot enable servant '") + servantName + "' of parent '" + to<string>(parentName) + "' with ID <" + to<string>(parentID) + ">" + (childIndex != -1 ? (" at index " + to<string>(childIndex)) : "")) {
}

Database::ServantAssignmentException::ServantAssignmentException(unsigned long parentID, const std::string& servantName, long childIndex)
: Exception("dataprovider::Database::ServantAssignmentException", string("cannot assign servant '") + servantName + "' of parent with ID <" + to<string>(parentID) + ">" + (childIndex != -1 ? (" at index " + to<string>(childIndex)) : "")) {
}

Database::NoSuchPropertyException::NoSuchPropertyException(unsigned long propertyID)
: Exception("dataprovider::Database::NoSuchPropertyException", string("unable to find property <") + to<string>(propertyID) + ">") {
}

Database::NoSuchPropertyException::NoSuchPropertyException(unsigned long servantID, std::string const & propertyName)
: Exception("dataprovider::Database::NoSuchPropertyException", string("unable to find property '") + propertyName + "' of servant <" + to<string>(servantID) + ">") {
}

Database::NoSuchPropertyNameInControllerException::NoSuchPropertyNameInControllerException(unsigned long controllerID, std::string const & propertyName)
: Exception("dataprovider::Database::NoSuchPropertyNameInControllerException", string("unable to find property '") + propertyName + "' in controller <" + to<string>(controllerID) + ">") {
}

Database::NoSuchPropertyNameInDriverException::NoSuchPropertyNameInDriverException(unsigned long driverID, std::string const & propertyName)
: Exception("dataprovider::Database::NoSuchPropertyNameInDriverException", string("unable to find property '") + propertyName + "' in driver <" + to<string>(driverID) + ">") {
}

Database::NoSuchCommandException::NoSuchCommandException(unsigned long commandID)
: Exception("dataprovider::Database::NoSuchCommandException", string("unable to find command <") + to<string>(commandID) + ">") {
}

Database::BadPropertyTypeException::BadPropertyTypeException(unsigned long propertyID, std::string const & type, std::string const & realType)
: Exception("dataprovider::Database::BadPropertyTypeException", string("requesting ") + type + " for property <" + to<std::string>(propertyID) + "> (" + realType + ")") {
}

Database::BadPropertyTypeException::BadPropertyTypeException(std::string const & propertyName, std::string const & type, std::string const & realType)
: Exception("dataprovider::Database::BadPropertyTypeException", string("requesting ") + type + " for property '" + propertyName + "' (" + realType + ")") {
}

Database::NullPropertyReferenceException::NullPropertyReferenceException()
: Exception("dataprovider::Database::NullPropertyReferenceException", "null reference") {
}

Database::InitialisationException::InitialisationException()
: Exception("dataprovider::Database::InitialisationException", string("problem occured while initialising Database")) {
}

Database::IndexOutOfBoundsException::IndexOutOfBoundsException(unsigned long propertyID)
: Exception("dataprovider::Database::IndexOutOfBoundsException", "requesting bad index for property <" + to<string>(propertyID) + ">") {
}

Database::IndexOutOfBoundsException::IndexOutOfBoundsException(unsigned long parentID, const std::string& nName, long size)
: Exception("dataprovider::Database::IndexOutOfBoundsException", "requesting bad index for '" + nName + "' child from servant <" + to<string>(parentID) + ">"),
	_size(size) {
}

long Database::IndexOutOfBoundsException::getSize() const {
	return _size;
}

Database::NotConfiguredException::NotConfiguredException()
: Exception("dataprovider::Database::NotConfiguredException", "configuration error") {
}

Database::ArrayTransferFailureException::ArrayTransferFailureException(unsigned long propertyID)
: Exception("dataprovider::Database::ArrayTransferFailureException", string("Failed to transfer array for property <") + to<string>(propertyID) +">") {
}

Database::BadServantTypeException::BadServantTypeException(unsigned long propertyID)
: Exception("dataprovider::Database::BadServantTypeException", string("Property does not belong to a servant for property <") + to<string>(propertyID) +">") {
}

Database::FileNotFoundException::FileNotFoundException(const std::string& fileName)
: Exception("dataprovider::Database::FileNotFoundException", string("File not found : ") + fileName) {
}

Database::NotResizableException::NotResizableException(unsigned long propertyID) :
	Exception("dataprovider::Database::NotResizableException", string("array property <") + to<string>(propertyID) + "> is not resizable") {
}

Database::NoSuchConfigurationException::NoSuchConfigurationException(const std::string& servantName, const std::string& configurationName) :
	Exception("dataprovider::Database::NoSuchConfigurationException", string("no such configuration '") + configurationName + "' for servant '" + servantName + "'") {
}

Database::NoPropertyInfo::NoPropertyInfo(const std::string & controllerType, const std::string & wantedPropertyName) :
	Exception("dataprovider::Database::NoPropertyInfo", string("the property ") + controllerType + " " + wantedPropertyName + " has no info") {
}

vector<Database *> Database::_instances;

void Database::initInstance(std::string const & configPath, bool clean) {

	if (_instances.empty() || _instances[0] == 0) {
		_instances.push_back(new Database());
		_instances[0]->init(configPath, clean);
	}
}

void Database::resetInstance(unsigned long ID) {

	if (_instances[ID] != 0) {
		_instances[ID]->reset();
		delete _instances[ID];
		_instances[ID] = 0;
	}
}

void Database::resetInstances() {

	unsigned long index;
	for (index = 0; index < _instances.size(); index++) {
		if (_instances[index] != 0) {
			resetInstance(index);
		}
	}

	_instances.clear();
}

Database * Database::getInstance(unsigned long ID) {
	return _instances[ID];
}

unsigned long Database::clone() {

	// cloning is temporarily disabled
	return 1;
}

void Database::setID(unsigned long ID) {
	_ID = ID;
}

unsigned long Database::getID() const {
	return _ID;
}

void Database::addController(Controller * controller) {

	// Locking the database for writing
	ScopedLock servantLock(_lockManager->lockServantForReading(controller->getID(), eol));

	_IDControllerMap.insert(controller->getID(), controller);
	_lockManager->createServant(controller->getID());
}

void Database::removeCommandsAndPropertiesFromServant(Servant* servant) {

	// Iterating the commands to erase from the map and delete them.
	HashMap<string, unsigned long> const & commands = servant->getCommands();

	for (HashMap<string, unsigned long>::const_iterator c = commands.begin(); c != commands.end(); ++c) {

		unsigned long commandID = c->second;

		ScopedLock commandLock(_lockManager->lockCommandForDeleting(commandID, eol));

		try {
			delete _IDCommandMap.get(commandID);
			_IDCommandMap.erase(commandID);

		} catch (Exception const & e) {
			err << "Command map and servant object not synchronised" << eol;
			err << e.printStack() << eol;
		}
	}

	// Iterating the properties to erase from the map and delete them.
	HashMap<string, unsigned long> const & properties = servant->getProperties();

	for (HashMap<string, unsigned long>::const_iterator p = properties.begin(); p != properties.end(); ++p) {
		removeProperty(p->second);
	}

	unsigned long servantID = servant->getID();

	// Deleting the controller
	_lockManager->removeServant(servantID);

}

void Database::removeController(unsigned long controllerID) {

	//out << "Database::removeController removing controller " << controllerID << eol;

	// lock
	{
		ScopedLock servantLock(_lockManager->lockServantForDeleting(controllerID, eol));

		Controller * controller = 0;

		try {
			controller = _IDControllerMap.get(controllerID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(controllerID), e);
		}

		removeCommandsAndPropertiesFromServant(controller);

		// Getting the children
		HashMap<string, pair<bool, vector<Servant::Children> > > const & servantMap = controller->getServantMap();
		for (HashMap<string, pair<bool, vector<Servant::Children> > >::const_iterator s = servantMap.begin(); s != servantMap.end(); s++) {

			vector<Servant::Children> const & servants = s->second.second;

			if (s->second.first) {

				for (unsigned int i = 0; i < servants.size(); i++) {
					// only a single controller is supposed to exist
					if (servants[i].children.size() > 0) {

						unsigned long servantID = servants[i].children.front();

						if (servantID != 0) {

							ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

							Controller* childController = dynamic_cast<Controller*>(getServant(servantID));
							if (childController != 0) {
								if (childController->isGenerated()) {
									// decrementing the reference counter
									childController->decrementReferenceCounter();

									// if controller is no longer used, remove it completely from the database.
									if (childController->getReferenceCounter() == 0) {
										removeController(childController->getID());
									}
								}
							}
						}
					}
				}

			} else {

				unsigned long servantID = servants[0].children.front();

				if (servantID != 0) {

					ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

					Controller* childController = dynamic_cast<Controller*>(getServant(servantID));

					if (childController != 0) {
						if (childController->isGenerated()) {
							// decrementing the reference counter
							childController->decrementReferenceCounter();

							// if controller is no longer used, remove it completely from the database.
							if (childController->getReferenceCounter() == 0) {
								removeController(childController->getID());
							}
						}
					}
				}
			}
		}

		_IDControllerMap.erase(controllerID);
		delete controller;
	}

	NotificationServer::getInstance()->configurationChanged(_ID, controllerID, common::CONTROLLER_REMOVED);
}

void Database::renameController(unsigned long controllerID, const std::string& newName) {
	ScopedLock servantLock(_lockManager->lockServantForDeleting(controllerID, eol));

	Controller * controller = 0;

	try {
		controller = _IDControllerMap.get(controllerID);
		controller->setName(newName);
	} catch (Exception const & e) {
		throw link(NoSuchServantException(controllerID), e);
	}
}

void Database::addDriver(Driver * driver) {

	// Locking the database for writing
	ScopedLock servantLock(_lockManager->lockServantForReading(driver->getID(), eol));

	_IDDriverMap.insert(driver->getID(), driver);
	_lockManager->createServant(driver->getID());
}

void Database::removeDriver(unsigned long driverID) {

	unsigned long fatherID = 0;

	// lock
	{
		ScopedLock servantLock(_lockManager->lockServantForDeleting(driverID, eol));

		Driver * driver = 0;

		try {
			driver = _IDDriverMap.get(driverID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(driverID), e);
		}

		removeCommandsAndPropertiesFromServant(driver);

		_IDDriverMap.erase(driverID);

		fatherID = driver->getFatherID();
		delete driver;
	}

	removeParentFromServant(fatherID, driverID);

	// lock
	{
		ScopedLock servantLock(_lockManager->lockServantForWriting(fatherID, eol));

		Driver * father = 0;

		try {
			father = _IDDriverMap.get(fatherID);
			if (father != 0) {

				HashMap<string, pair<bool, vector<Servant::Children> > > & childMap = father->getServantMap();

				for (HashMap<std::string, std::pair<bool, std::vector<Servant::Children> > >::iterator c = childMap.begin(); c != childMap.end(); c++) {
					vector<Servant::Children> & children = c->second.second;

					for (vector<Servant::Children>::iterator cc = children.begin(); cc != children.end(); cc++) {

						vector<unsigned long>& childChildren = cc->children;

						bool found = false;
						long index = 0;
						for (vector<unsigned long>::iterator ccc = childChildren.begin(); ccc != childChildren.end(); ccc++) {
							if (*ccc == driverID) {
								childChildren.erase(ccc);
								found = true;
								break;
							}
							index++;
						}

						if (found && index == cc->lastEnabled) {
							cc->lastEnabled = -1;
						}
					}
				}
			}

		} catch (Exception const & e) {
			//throw link(NoSuchServantException(fatherID), e);
		}

	}


	NotificationServer::getInstance()->configurationChanged(_ID, driverID, common::TREE);
}

unsigned long Database::createGeneratedController(const std::string& controllerType, unsigned long parentID, const std::string& n_name, long childIndex) {

	unsigned long ID = 0;

	// lock
	{
		ScopedLock parentLock(_lockManager->lockServantForWriting(parentID, eol));

		Controller* parent = 0;

		try {
			parent = _IDControllerMap.get(parentID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(parentID), e);
		}

		Controller* controller = 0;

		if (controllerType != "") {

			// creating the controller and incrementing the counter
			controller = new Controller(controllerType, "", false, false, false, false, true);
			controller->incrementReferenceCounter();

			// Adding driver to the database and setting the ID.
			ID = IndexGenerator::getInstance()->getNextIndex();
			controller->setID(ID);

			string name = "servant$";
			name += to<string>(ID);
			controller->setName(name);

			controller->enable();
			_IDControllerMap.insert(ID, controller);
			_lockManager->createServant(ID);
		}

		if (n_name != "") {
			if (childIndex == -1) {
				parent->addServant(n_name, controller, false);
			} else {
				parent->addServantToList(n_name, childIndex, controller);
			}
		}

		//out << "added controller '" << n_name << "' (" << controllerType << ") with ID <" << ID << ">" << eol;
	}

	// update the configuration
	NotificationServer::getInstance()->configurationChanged(_ID, ID, common::TREE);

	return ID;
}

unsigned long Database::createController(const std::string& controllerName, const std::string& controllerType, const std::string& family, bool isCommand, const common::HashMap<std::string,  std::pair<std::string , std::string>>& childrenMap) {

	unsigned long ID = 0;

	// lock
	{
		Controller* controller = 0;

		if (controllerType != "") {

			bool isVisible = (family != "hidden" && family != "");

			// creating the controller and incrementing the counter
			controller = new Controller(controllerType, controllerName, isCommand, isVisible, false, false, false);
			controller->setFamily(family, "none");

			warn << "Created controller with subfamily \"none\"" << eol;

			// Adding driver to the database and setting the ID.
			ID = IndexGenerator::getInstance()->getNextIndex();
			controller->setID(ID);

			controller->enable();
			_IDControllerMap.insert(ID, controller);
			_lockManager->createServant(ID);
		}

		// add children
		for (HashMap<string,  std::pair<string , string>>::const_iterator c = childrenMap.begin(); c != childrenMap.end(); c++) {
			string n_name = c->first;
			string servantName = c->second.first;

			Servant* servant = findServant(servantName);
			if (servant != 0) {
				controller->addServant(n_name, servant, true);
			} else {
				err << "Cannot find servant with name '" << servantName << "'" << eol;
			}
		}
	}

	// update the configuration
	NotificationServer::getInstance()->configurationChanged(_ID, ID, common::TREE);

	return ID;
}

unsigned long Database::createDriver(const std::string& driverName, const std::string& driverType, const std::string& protocolType, common::DriverMode mode, unsigned long parentID, const std::string& n_name, long childIndex) {

	unsigned long ID = 0;

	// lock
	{
		ScopedLock parentLock(_lockManager->lockServantForWriting(parentID, eol));

		Driver* parent = 0;

		try {
			parent = _IDDriverMap.get(parentID);

		} catch (Exception const & e) {
//			throw link(NoSuchServantException(parentID), e);
		}

		Driver* driver = 0;

		if (driverType != "") {

			// creating the driver and incrementing the counter
			driver = new Driver(driverType, driverName, false, true, false);

			// Does not exist.
			//driver->incrementReferenceCounter();

			// Adding driver to the database and setting the ID.
			ID = IndexGenerator::getInstance()->getNextIndex();
			driver->setID(ID);
			driver->setMode(mode);

			//driver->enable();
			_IDDriverMap.insert(ID, driver);
			_lockManager->createServant(ID);
		}

		if (n_name != "") {

			driver->setFatherInfo(parentID, n_name, childIndex);

			if (childIndex == -1) {
				parent->addServant(n_name, driver, false);
			} else {
				parent->addServantToList(n_name, childIndex, driver);
			}
		}

		//out << "added driver '" << driverName << "' (" << driverType << ") with ID <" << ID << ">" << eol;
	}

	// update the configuration
	NotificationServer::getInstance()->configurationChanged(_ID, ID, common::TREE);

	return ID;
}

unsigned long Database::cloneController(unsigned long controllerID) {

	unsigned long ID = 0;

	std::string controllerName;
	std::string controllerType;
	std::string family;
	bool isCommand;
	common::HashMap<std::string, std::pair<bool, std::vector<Servant::Children> > > servantMap;

	// lock
	{
		ScopedLock servantLock(_lockManager->lockServantForReading(controllerID, eol));

		Controller* servant = 0;

		try {
			servant = _IDControllerMap.get(controllerID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(controllerID), e);
		}

		ID = IndexGenerator::getInstance()->getNextIndex();
		controllerName = servant->getName() + ".clone";
		controllerType = servant->getControllerType();
		family = "hidden";
		isCommand = servant->isCommand();
		servantMap = servant->getServantMap();
	}

	Controller * controller = new Controller(controllerType, controllerName, isCommand, false, false, false, false);
	controller->setFamily(family, "none");

	// Adding controller to the database and setting the ID.
	controller->setID(ID);

	controller->enable();
	_IDControllerMap.insert(ID, controller);
	_lockManager->createServant(ID);

	// add children
	controller->setServantMap(servantMap);

	// update the configuration
	NotificationServer::getInstance()->configurationChanged(_ID, ID, common::TREE);

	return ID;
}

void Database::addServantToParent(unsigned long parentID, const std::string& nName, bool isList) {

	// lock
	{
		ScopedLock parentLock(_lockManager->lockServantForWriting(parentID, eol));

		Controller* parent = 0;

		try {
			parent = _IDControllerMap.get(parentID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(parentID), e);
		}

		// Add child to parent
		if (isList) {
			parent->addServantList(nName, 0);
		} else {
			parent->addServant(nName, 0, false);
		}
	}
}

unsigned long Database::removeServantFromParent(unsigned long parentID, const std::string& nName, long index) {

	//out << "Database::removeServantFromParent removing servant " << nName << " at " << index << " from " << parentID << eol;

	unsigned long childID = 0;

	// lock
	{
		ScopedLock parentLock(_lockManager->lockServantForWriting(parentID, eol));

		Controller* parent = 0;

		try {
			parent = _IDControllerMap.get(parentID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(parentID), e);
		}

		// Remove child from parent
		childID = parent->removeServant(nName, index);

		// nothing to do
		if (childID == 0) {
			return 0;
		}

		ScopedLock childLock(_lockManager->lockServantForDeleting(childID, eol));

		Controller* controller = dynamic_cast<Controller*>(getServant(childID));

		// keeping the servant if it is not generated
		if (controller != 0 && controller->isGenerated()) {

			//out << "controller " << controller->getID() << " has " << controller->getReferenceCounter() << " references" << eol;

			controller->decrementReferenceCounter();

			// if controller is no longer used, remove it completely from the database.
			if (controller->getReferenceCounter() == 0) {
				childID = controller->getID();
				removeController(childID);
			}
		}
	}

	// servantID is 0, forcing to update
	NotificationServer::getInstance()->configurationChanged(_ID, 0, common::TREE);

	return childID;
}

void Database::removeServant(unsigned long servantID) {

	//out << "Database::removeServant removing servant " << servantID << eol;

	// servant lock
	{
		ScopedLock servantLock(_lockManager->lockServantForDeleting(servantID, eol));

		Servant* servant = getServant(servantID);
		Controller* controller = dynamic_cast<Controller*>(servant);
		Driver* driver = dynamic_cast<Driver*>(servant);

		if (controller != 0) {

			// keeping the servant if it is not generated
			if (controller->isGenerated()) {

				controller->decrementReferenceCounter();

				// if controller is no longer used, remove it completely from the database.
				if (controller->getReferenceCounter() == 0) {
					removeController(servantID);
				}

			} else {
				removeController(servantID);
			}

		} else if (driver != 0) {

			removeDriver(servantID);
		}
	}

	// servantID is 0, forcing to update
	NotificationServer::getInstance()->configurationChanged(_ID, 0, common::TREE);
}

void Database::assignServant(unsigned long parentID, const std::string& nName, long childIndex, unsigned long newParentID, const std::string& newNName, long newChildIndex) {

	//out << "Database::assignServant assign servant " << parentID << " " << nName << " at " << childIndex << " with " << newParentID << " " << newNName << " at " << newChildIndex << eol;

	// removing the servant from the parent
	removeServantFromParent(parentID, nName, childIndex);

	unsigned long newServantID = 0;

	// locking the new parent
	{
		Controller* newParent = 0;

		ScopedLock newParentLock(_lockManager->lockServantForReading(newParentID, eol));

		try {
			newParent = _IDControllerMap.get(newParentID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(newParentID), e);
		}

		vector<unsigned long> servantList = newParent->getServant(newNName, newChildIndex);
		if (servantList.size() != 0) {
			newServantID = servantList.front();
		}
	}

	bool isEnabled = false;

	// locking the parent
	{
		ScopedLock parentLock(_lockManager->lockServantForWriting(parentID, eol));

		Controller* parent = 0;

		try {
			parent = _IDControllerMap.get(parentID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(parentID), e);
		}

		parent->assignServant(nName, childIndex, newServantID);
		isEnabled = parent->isEnabled();

		try {
			_servantOperations.verifyCircularity(parent);

		} catch (ServantOperations::CircularServantException& e) {
			// setting the child null
			parent->removeServant(nName, childIndex);
			throw ServantAssignmentException(parentID, nName, childIndex);
		}

		// locking the new servant
		ScopedLock newServantLock(_lockManager->lockServantForWriting(newServantID, eol));

		try {
			Controller* newServant = _IDControllerMap.get(newServantID);

			if (newServant->isGenerated()) {
				newServant->incrementReferenceCounter();
			}

		} catch (Exception const & e) {
		}
	}

	NotificationServer::getInstance()->configurationChanged(_ID, parentID, common::TREE);
}

void Database::assignServant(unsigned long parentID, const std::string& nName, long childIndex, unsigned long newServantID) {

	//out << "Database::assignServant assign servant " << parentID << " " << nName << " at " << childIndex << " with " << newServantID << eol;

	// removing the servant from the parent
	removeServantFromParent(parentID, nName, childIndex);

	bool isEnabled = false;

	// lock
	{
		ScopedLock parentLock(_lockManager->lockServantForWriting(parentID, eol));

		Controller* parent = 0;

		try {
			parent = _IDControllerMap.get(parentID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(parentID), e);
		}

		parent->assignServant(nName, childIndex, newServantID);
		isEnabled = parent->isEnabled();

		try {
			_servantOperations.verifyCircularity(parent);

		} catch (ServantOperations::CircularServantException& e) {
			// setting the child null
			parent->removeServant(nName, childIndex);
			throw ServantAssignmentException(parentID, nName, childIndex);
		}

		ScopedLock newServantLock(_lockManager->lockServantForWriting(newServantID, eol));

		try {
			Controller* newServant = _IDControllerMap.get(newServantID);

			if (newServant->isGenerated()) {
				newServant->incrementReferenceCounter();
			}

		} catch (Exception const & e) {
		}
	}

	NotificationServer::getInstance()->configurationChanged(_ID, parentID, common::TREE);
}

void Database::assignServant(unsigned long servantID, bool newServant, bool localServant) {

	//out << "Database::assignServant assign servant " << servantID << eol;

	// lock
	{
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

		try {
			Controller* servant = _IDControllerMap.get(servantID);

			// not incrementing in case of new servant, because it already has been incremented at its creation
			if (servant->isGenerated() && !newServant) {
				servant->incrementReferenceCounter();
			}

		} catch (Exception const & e) {
		}
	}

	NotificationServer::getInstance()->configurationChanged(_ID, 0, common::TREE);
}

void Database::addProperty(BaseProperty* property, unsigned long servantID) {
	_keyIDPropertyMap.insert(property->getID(), property);
	_IDPropertyIDServantMap.insert(property->getID(), servantID);
	_lockManager->createProperty(property->getID());
}

void Database::removeProperty(unsigned long propertyID) {

	//out << "Database::removeProperty " << propertyID << eol;

	ScopedLock lock(_lockManager->lockPropertyForDeleting(propertyID, eol));

	unsigned long propertyKeyID = getPropertyKeyID(propertyID);

	// remove the actual property only if it is not a reference (both IDs the same)
	if (propertyKeyID == propertyID) {

		if (_keyIDPropertyMap.find(propertyKeyID)) {
			BaseProperty * p = _keyIDPropertyMap.get(propertyID);

			DataProperty * ep = dynamic_cast<DataProperty *>(p);
			if (ep != 0) {
				_keyIDPropertyMap.erase(propertyID);
				delete ep;
				_IDPropertyIDServantMap.erase(propertyID);
				_lockManager->removeProperty(propertyID);
			}
			else {
				BaseDynamicProperty * pl = dynamic_cast<BaseDynamicProperty *>(p);

				if (pl != 0) {
					std::vector<unsigned long> const & propertyIDs = pl->getPropertyIDs();

					for (std::vector<unsigned long>::const_iterator c = propertyIDs.begin(); c != propertyIDs.end(); ++c) {
						removeProperty(*c);
					}
					_keyIDPropertyMap.erase(propertyID);
					delete pl;
					_IDPropertyIDServantMap.erase(propertyID);
					_lockManager->removeProperty(propertyID);
				}
			}
		}
	}
}

BaseProperty* Database::getPropertyFromKeyID(unsigned long propertyKeyID) {

	// returning 0 if the property is a null reference
	if (propertyKeyID != 0) {
		try {
			return _keyIDPropertyMap.get(propertyKeyID);
		} catch (const Exception& e) {
			throw NoSuchPropertyException(propertyKeyID);
		}
	}

	throw NullPropertyReferenceException();
}

unsigned long Database::getPropertyKeyID(unsigned long propertyID) {

	return propertyID;
}

std::list<unsigned long> Database::getPropertyIDs(unsigned long propertyKeyID) {

	list<unsigned long> result;
	result.push_back(propertyKeyID);

	return result;
}

unsigned long Database::addPropertyToServant(unsigned long servantID, const std::string& propertyName, const std::string& propertyType, bool spy, bool saveBox, bool & added) {

	unsigned long index = 0;

	{
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

		PropertyHolder* propertyHolder = 0;

		try {
			propertyHolder = getPropertyHolder(servantID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(servantID), e);
		}

		unsigned long propertyID = propertyHolder->getPropertyID(propertyName);
		if (propertyID != 0) {

			try {
				unsigned long propertyKeyID = getPropertyKeyID(propertyID);
				ScopedLock propertyLock(_lockManager->lockPropertyForReading(propertyKeyID, eol));

				BaseProperty* property = dynamic_cast<BaseProperty*>(getPropertyFromKeyID(propertyKeyID));
				if (property != 0) {
					property->setSpy(spy);
					property->setSaveBox(saveBox);
					added = false;
					return propertyID;
				}
			} catch (Exception& e) {
				err << "No reference found for property '" << propertyName << "' of servant <" << servantID << ">" << eol;
			}

			removeProperty(propertyID);
		}

		BaseProperty* property = 0;

		// create the correct property content type
		if (propertyType == TypeToString<int8>::convert()) {
			property = new ElementaryProperty<int8>(propertyName, true, false);

		} else if (propertyType == TypeToString<int16>::convert()) {
			property = new ElementaryProperty<int16>(propertyName, true, false);

		} else if (propertyType == TypeToString<uint16>::convert()) {
			property = new ElementaryProperty<uint16>(propertyName, true, false);

		} else if (propertyType == TypeToString<int32>::convert()) {
			property = new ElementaryProperty<int32>(propertyName, true, false);

		} else if (propertyType == TypeToString<uint32>::convert()) {
			property = new ElementaryProperty<uint32>(propertyName, true, false);

		} else if (propertyType == TypeToString<int64>::convert()) {
			property = new ElementaryProperty<int64>(propertyName, true, false);

		} else if (propertyType == TypeToString<uint64>::convert()) {
			property = new ElementaryProperty<uint64>(propertyName, true, false);

		} else if (propertyType == TypeToString<float32>::convert()) {
			property = new ElementaryProperty<float32>(propertyName, true, false);

		} else if (propertyType == TypeToString<float64>::convert()) {
			property = new ElementaryProperty<float64>(propertyName, true, false);

		} else if (propertyType == TypeToString<bool>::convert()) {
			property = new ElementaryProperty<bool>(propertyName, true, false);

		} else if (propertyType == TypeToString<string>::convert()) {
			property = new ElementaryProperty<string>(propertyName, true, false);

		} else if (propertyType == INT16ARRAY_PROPERTY) {
			property = new IcsArrayProperty<int16>(propertyName, INT16ARRAY_PROPERTY, false, "");

		} else if (propertyType == INT32ARRAY_PROPERTY) {
			property = new IcsArrayProperty<int32>(propertyName, INT32ARRAY_PROPERTY, false, "");

		} else if (propertyType == INT64ARRAY_PROPERTY) {
			property = new IcsArrayProperty<int64>(propertyName, INT64ARRAY_PROPERTY, false, "");

		} else if (propertyType == UINT16ARRAY_PROPERTY) {
			property = new IcsArrayProperty<uint16>(propertyName, UINT16ARRAY_PROPERTY, false, "");

		} else if (propertyType == UINT32ARRAY_PROPERTY) {
			property = new IcsArrayProperty<uint32>(propertyName, UINT32ARRAY_PROPERTY, false, "");

		} else if (propertyType == UINT64ARRAY_PROPERTY) {
			property = new IcsArrayProperty<uint64>(propertyName, UINT64ARRAY_PROPERTY, false, "");

		} else if (propertyType == FLOAT32ARRAY_PROPERTY) {
			property = new IcsArrayProperty<float32>(propertyName, FLOAT32ARRAY_PROPERTY, false, "");

		} else if (propertyType == FLOAT64ARRAY_PROPERTY) {
			property = new IcsArrayProperty<float64>(propertyName, FLOAT64ARRAY_PROPERTY, false, "");

		} else {
			err << "Cannot find a type for '" << propertyType << "'" << eol;
		}

		index = IndexGenerator::getInstance()->getNextIndex();
		property->setID(index);
		property->setSpy(spy);
		property->setSaveBox(saveBox);

		propertyHolder->addProperty(propertyName, property);
		addProperty(property, servantID);

		_lockManager->createProperty(index);
	}

	NotificationServer::getInstance()->configurationChanged(_ID, servantID, common::PROPERTY);

	added = true;

	return index;
}

unsigned long Database::addDynamicPropertyToServant(unsigned long servantID, const std::string& propertyName, const std::string& propertyType, bool spy) {

	unsigned long index = 0;

	{
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

		PropertyHolder* propertyHolder = 0;

		try {
			propertyHolder = getPropertyHolder(servantID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(servantID), e);
		}

		unsigned long propertyID = propertyHolder->getPropertyID(propertyName);
		if (propertyID != 0) {

			try {
				unsigned long propertyKeyID = getPropertyKeyID(propertyID);
				ScopedLock propertyLock(_lockManager->lockPropertyForReading(propertyKeyID, eol));

				BaseProperty* property = dynamic_cast<BaseDynamicProperty*>(getPropertyFromKeyID(propertyKeyID));
				if (property != 0) {
					property->setSpy(spy);
					return propertyID;
				}
			} catch (Exception& e) {
				err << "No reference found for property '" << propertyName << "' of servant <" << servantID << ">" << eol;
			}

			removeProperty(propertyID);
		}

		BaseDynamicProperty* property = 0;

		// create the correct property content type
		if (propertyType == TypeToString<int8>::convert()) {
			property = new DynamicProperty<int8>(propertyName, true, false);

		} else if (propertyType == TypeToString<int16>::convert()) {
			property = new DynamicProperty<int16>(propertyName, true, false);

		} else if (propertyType == TypeToString<uint16>::convert()) {
			property = new DynamicProperty<uint16>(propertyName, true, false);

		} else if (propertyType == TypeToString<int32>::convert()) {
			property = new DynamicProperty<int32>(propertyName, true, false);

		} else if (propertyType == TypeToString<uint32>::convert()) {
			property = new DynamicProperty<uint32>(propertyName, true, false);

		} else if (propertyType == TypeToString<int64>::convert()) {
			property = new DynamicProperty<int64>(propertyName, true, false);

		} else if (propertyType == TypeToString<uint64>::convert()) {
			property = new DynamicProperty<uint64>(propertyName, true, false);

		} else if (propertyType == TypeToString<float32>::convert()) {
			property = new DynamicProperty<float32>(propertyName, true, false);

		} else if (propertyType == TypeToString<float64>::convert()) {
			property = new DynamicProperty<float64>(propertyName, true, false);

		} else if (propertyType == TypeToString<bool>::convert()) {
			property = new DynamicProperty<bool>(propertyName, true, false);

		} else if (propertyType == TypeToString<string>::convert()) {
			property = new DynamicProperty<string>(propertyName, true, false);

		} else if (propertyType == INT16ARRAY_PROPERTY) {
			property = new DynamicProperty<int16*>(propertyName, true, false);

		} else if (propertyType == INT32ARRAY_PROPERTY) {
			property = new DynamicProperty<int32*>(propertyName, true, false);

		} else if (propertyType == INT64ARRAY_PROPERTY) {
			property = new DynamicProperty<int64*>(propertyName, true, false);

		} else if (propertyType == UINT16ARRAY_PROPERTY) {
			property = new DynamicProperty<uint16*>(propertyName, true, false);

		} else if (propertyType == UINT32ARRAY_PROPERTY) {
			property = new DynamicProperty<uint32*>(propertyName, true, false);

		} else if (propertyType == UINT64ARRAY_PROPERTY) {
			property = new DynamicProperty<uint64*>(propertyName, true, false);

		} else if (propertyType == FLOAT32ARRAY_PROPERTY) {
			property = new DynamicProperty<float32*>(propertyName, true, false);

		} else if (propertyType == FLOAT64ARRAY_PROPERTY) {
			property = new DynamicProperty<float64*>(propertyName, true, false);

		} else {
			err << "Cannot find a type for '" << propertyType << "'" << eol;
		}

		index = IndexGenerator::getInstance()->getNextIndex();
		property->setID(index);
		property->setSpy(spy);

		propertyHolder->addProperty(propertyName, property);
		addProperty(property, servantID);

		_lockManager->createProperty(index);
	}

	NotificationServer::getInstance()->configurationChanged(_ID, servantID, common::PROPERTY);

	return index;
}

void Database::removePropertyFromServant(unsigned long servantID, unsigned long propertyID) {

	{
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

		PropertyHolder* propertyHolder = 0;

		try {
			propertyHolder = getPropertyHolder(servantID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(servantID), e);
		}

		unsigned long propertyKeyID = getPropertyKeyID(propertyID);
		ScopedLock propertyLock(_lockManager->lockPropertyForReading(propertyKeyID, eol));

		BaseProperty* property = getPropertyFromKeyID(propertyKeyID);
		if (property != 0) {
			propertyHolder->removeProperty(property->getName());

			removeProperty(propertyID);
		}
	}

	NotificationServer::getInstance()->configurationChanged(_ID, servantID, common::PROPERTY);

}

void Database::setPropertySettings(unsigned long propertyID, const std::string& propertyFormat, bool save, bool spy) {

	// Locking the property. No need to get the key ID since a dynamic property cannot be referenced.
	ScopedLock lock(_lockManager->lockPropertyForWriting(propertyID, eol));

	// Get property
	BaseProperty * property = getPropertyFromKeyID(propertyID);

	property->setFormat(propertyFormat);
	property->setSaveValue(save);
	property->setSpy(spy);
}

void Database::resizeDynamicProperty(unsigned long propertyID, long size) {

	unsigned long servantID = 0;

	{
		// Locking the property. No need to get the key ID since a dynamic property cannot be referenced.
		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyID, eol));

		// Get property
		BaseProperty * property = getPropertyFromKeyID(propertyID);
		BaseDynamicProperty * dynamicProperty = dynamic_cast<BaseDynamicProperty*>(property);

		// verify property is of correct type
		if (dynamicProperty != 0) {

			servantID = dynamicProperty->getParentID();

			if (size < dynamicProperty->getSize()) {

				// resizing the dynamic property and getting the list of removed properties
				vector<unsigned long> removedPropertyIDs = dynamicProperty->downsize(size);

				// removing the properties from the database
				for (vector<unsigned long>::const_iterator p = removedPropertyIDs.begin(); p != removedPropertyIDs.end(); p++) {
					removeProperty(*p);
				}

			} else {

				// adding new properties
				for (long i = dynamicProperty->getSize(); i < size; i++) {

					DataProperty* newProperty = dynamicProperty->createProperty(true, true);
					// new ID
					unsigned long newPropertyID = IndexGenerator::getInstance()->getNextIndex();
					newProperty->setID(newPropertyID);
					Database::getInstance()->addProperty(newProperty, servantID);

					dynamicProperty->addProperty(newProperty);
				}
			}
		}
	}

	NotificationServer::getInstance()->configurationChanged(_ID, servantID, common::PROPERTY);
}

long Database::getDynamicPropertySize(unsigned long propertyID) {

	// Locking the property. No need to get the key ID since a dynamic property cannot be referenced.
	ScopedLock lock(_lockManager->lockPropertyForWriting(propertyID, eol));

	try {
		// Get property
		BaseProperty * property = getPropertyFromKeyID(propertyID);
		BaseDynamicProperty * dynamicProperty = dynamic_cast<BaseDynamicProperty*>(property);

		// verify property is of correct type
		if (dynamicProperty != 0) {
			return dynamicProperty->getSize();
		}
	}
	catch(common::Exception& e) {

	}

	return 0;
}

unsigned long Database::getDynamicPropertyIDAtIndex(unsigned long propertyID, long index, bool resize) {

	unsigned long newPropertyID = 0;
	unsigned long servantID = 0;
	bool resized = false;

	{
		// Locking the property. No need to get the key ID since a dynamic property cannot be referenced.
		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyID, eol));

		try {
			// Get property
			BaseProperty * property = getPropertyFromKeyID(propertyID);
			BaseDynamicProperty * dynamicProperty = dynamic_cast<BaseDynamicProperty*>(property);

			// verify property is of correct type
			if (dynamicProperty != 0) {

				servantID = dynamicProperty->getParentID();

				if (index < dynamicProperty->getSize()) {

					// returning the property ID
					return dynamicProperty->getPropertyIDs()[index];

				} else if (resize) {

					// adding new properties
					for (long i = dynamicProperty->getSize(); i < index + 1; i++) {

						DataProperty* newProperty = dynamicProperty->createProperty(true, true);
						// new ID
						newPropertyID = IndexGenerator::getInstance()->getNextIndex();
						newProperty->setID(newPropertyID);
						Database::getInstance()->addProperty(newProperty, servantID);

						dynamicProperty->addProperty(newProperty);

						resized = true;
					}

				} else {
					// index out of bounds
					throw IndexOutOfBoundsException(propertyID);
				}
			}
		}
		catch(common::Exception& e) {

		}
	}

	if (resized) {
		NotificationServer::getInstance()->configurationChanged(_ID, servantID, common::PROPERTY);
	}

	return newPropertyID;
}

long Database::getIndexOfDynamicPropertyID(unsigned long parentPropertyID, unsigned long propertyID) {

	// Locking the property. No need to get the key ID since a dynamic property cannot be referenced.
	ScopedLock lock(_lockManager->lockPropertyForWriting(propertyID, eol));

	try {

		// Get property
		BaseProperty * property = getPropertyFromKeyID(parentPropertyID);
		BaseDynamicProperty * dynamicProperty = dynamic_cast<BaseDynamicProperty*>(property);

		// verify property is of correct type
		if (dynamicProperty != 0) {
			long index = 0;
			vector<unsigned long> const & properties = dynamicProperty->getPropertyIDs();
			vector<unsigned long>::const_iterator it;
			for (it = properties.begin(); it != properties.end(); it++) {

				if ((*it) == propertyID) {
					return index;
				}
				index++;
			}
		}
	}
	catch(common::Exception& e) {

	}

	return -1;
}

vector<unsigned long> Database::getDynamicPropertyIDs(unsigned long parentPropertyID) {
	// Locking the property. No need to get the key ID since a dynamic property cannot be referenced.
	ScopedLock lock(_lockManager->lockPropertyForWriting(parentPropertyID, eol));

	try {
		// Get property
		BaseProperty * property = getPropertyFromKeyID(parentPropertyID);
		BaseDynamicProperty * dynamicProperty = dynamic_cast<BaseDynamicProperty*>(property);

		// verify property is of correct type
		if (dynamicProperty != 0) {
			long index = 0;
			return dynamicProperty->getPropertyIDs();
		}
	}
	catch(common::Exception& e) {

	}
	return vector<unsigned long>();
}

void Database::addCommand(Command* command) {

	_IDCommandMap.insert(command->getID(), command);
}

unsigned long Database::addCommandToServant(unsigned long servantID, const std::string& commandName) {

	ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

	Servant* servant = 0;

	try {
		servant = getServant(servantID);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

	unsigned long commandID = servant->getCommand(commandName);
	if (commandID != 0) {
		return commandID;
	}

	Command * command = new Command(commandName);
	unsigned long index = IndexGenerator::getInstance()->getNextIndex();
	command->setID(index);

	servant->addCommand(command);
	_IDCommandMap.insert(index, command);

	return index;
}

Database::PropertyMap Database::getControllerProperties(unsigned long controllerID) {

	ScopedLock lock(_lockManager->lockServantForReading(controllerID, eol));

	Controller * controller = 0;

	try {
		controller = _IDControllerMap.get(controllerID);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(controllerID), e);
	}

	return createPropertyItems(controller->getProperties());
}

Database::PropertyMap Database::getDriverProperties(unsigned long driverID) {

	ScopedLock lock(_lockManager->lockServantForReading(driverID, eol));

	Driver * driver = 0;

	try {
		driver = _IDDriverMap.get(driverID);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(driverID), e);
	}

	return createPropertyItems(driver->getProperties());
}

Database::PropertyMap Database::createPropertyItems(const HashMap<string, unsigned long>& properties) {

	PropertyMap result;

	// Iterating the properties
	for (HashMap<string, unsigned long>::const_iterator p = properties.begin(); p != properties.end(); ++p) {

		const string& propertyName = p->first;
		unsigned long propertyID = p->second;
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		PropertyItem item;

		ScopedLock lock(_lockManager->lockPropertyForReading(propertyKeyID, eol));

		BaseProperty* property = 0;

		try {
			property = getPropertyFromKeyID(propertyKeyID);
			item.isSpy = property->isSpy();

		} catch (NullPropertyReferenceException&) {
			// do nothing
		} catch (Exception&) {
			throw;
		}

		if (property != 0) {

			BaseDynamicProperty * dynamicProperty = dynamic_cast<BaseDynamicProperty *>(property);

			if (dynamicProperty != 0) {

				item.itemType = PropertyItem::DYNAMIC;
				item.IDs.push_back(dynamicProperty->getID());

				// Iterating the property contents
				vector<unsigned long> const & propertyIDs = dynamicProperty->getPropertyIDs();

				for (vector<unsigned long>::const_iterator i = propertyIDs.begin(); i != propertyIDs.end(); ++i) {
					item.IDs.push_back(*i);
				}

				item.propertyType = dynamicProperty->getType();

				item.enumeratedValuesID = dynamicProperty->getEnumeratedValuesPropertyID();
				item.enumeratedLabelsID = dynamicProperty->getEnumeratedLabelsPropertyID();
			}

			DataProperty * dataProperty = dynamic_cast<DataProperty *>(property);

			if (dataProperty != 0) {

				item.itemType = PropertyItem::ELEMENTARY;
				item.IDs.push_back(propertyID);
				item.propertyType = dataProperty->getType();

				item.enumeratedValuesID = dataProperty->getEnumeratedValuesPropertyID();
				item.enumeratedLabelsID = dataProperty->getEnumeratedLabelsPropertyID();
			}
		}

		result.insert(propertyName, item);
	}

	return result;

}

Database::CommandMap Database::getControllerCommands(unsigned long controllerID) {

	try {
		ScopedLock lock(_lockManager->lockServantForReading(controllerID, eol));

		return _IDControllerMap.get(controllerID)->getCommands();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(controllerID), e);
	}
}

Database::CommandMap Database::getDriverCommands(unsigned long driverID) {

	try {
		ScopedLock lock(_lockManager->lockServantForReading(driverID, eol));

		return _IDDriverMap.get(driverID)->getCommands();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(driverID), e);
	}
}

Database::ChildServantMap Database::getControllerChildServants(unsigned long controllerID) {

	ScopedLock lock(_lockManager->lockServantForReading(controllerID, eol));

	Controller * controller = 0;

	try {
		controller = _IDControllerMap.get(controllerID);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(controllerID), e);
	}

	const HashMap<string, pair<bool, vector<Servant::Children> > >& servants = controller->getServantMap();

	ChildServantMap result;

	// loop over all children
	for (HashMap<std::string, pair<bool, vector<Servant::Children> > >::const_iterator s = servants.begin(); s != servants.end(); ++s) {

		ChildServant childServant;

		childServant.isList = s->second.first;
		childServant.size = s->second.second.size();

		// loop over child list (all children stored as lists in parser however single children stored as a vector of size 1)
		long childIndex = 0;
		for (vector<Servant::Children>::const_iterator csl = s->second.second.begin(); csl != s->second.second.end(); csl++) {
			for (vector<unsigned long>::const_iterator cs = csl->children.begin(); cs != csl->children.end(); cs++) {

				unsigned long currentServantID = *cs;

				// check that a child is not null (possible for noncontiguous child arrays)
				if (currentServantID != 0) {

					ScopedLock lock(_lockManager->lockServantForReading(currentServantID, eol));

					Servant* currentServant = getServant(currentServantID);

					childServant.children.push_back(pair<long, string>(childIndex, currentServant->getName()));
				}
			}

			childIndex++;
		}

		// only inserting the servant if not null
		if (childServant.isList || (!childServant.isList && childServant.children.size() != 0)) {
			result.insert(s->first, childServant);
		}
	}

	return result;
}

Database::ChildServantMap Database::getDriverChildServants(unsigned long driverID) {

	ScopedLock lock(_lockManager->lockServantForReading(driverID, eol));

	Driver * driver = 0;

	try {
		driver = _IDDriverMap.get(driverID);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(driverID), e);
	}

	ChildServantMap result;

	const HashMap<string, pair<bool, vector<Servant::Children> > >& servants = driver->getServantMap();

	// loop over all children
	for (HashMap<std::string, pair<bool, vector<Servant::Children> > >::const_iterator s = servants.begin(); s != servants.end(); ++s) {

		ChildServant childServant;

		childServant.isList = s->second.first;
		childServant.size = s->second.second.size();

		// loop over child list (all children stored as lists in parser however single children stored as a vector of size 1)
		long childIndex = 0;
		for (vector<Servant::Children>::const_iterator csl = s->second.second.begin(); csl != s->second.second.end(); csl++) {

			for (vector<unsigned long>::const_iterator cs = csl->children.begin(); cs != csl->children.end(); cs++) {

				unsigned long currentServantID = *cs;

				// check that a child is not null (possible for noncontiguous child arrays)
				if (currentServantID != 0) {

					ScopedLock lock(_lockManager->lockServantForReading(currentServantID, eol));

					Servant* currentServant = getServant(currentServantID);

					childServant.children.push_back(pair<long, string>(childIndex, currentServant->getName()));
				}
			}

			childIndex++;
		}

		result.insert(s->first, childServant);
	}

	return result;
}

void Database::enableServant(unsigned long servantID) {

	try {
		unsigned long firstDriverID = 0;

		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));
		{
			set<unsigned long> servantsToEnable = _servantOperations.getServantsForEnable(getServant(servantID), firstDriverID);

			// verifying that the servant can be enabled and setting it
			setServantEnabled(servantsToEnable);
		}

		if (firstDriverID != 0) {
			set<unsigned long> servantsToEnable = _servantOperations.getServantsForEnableDriver(dynamic_cast<Driver*>(getServant(firstDriverID)));
			setServantEnabled(servantsToEnable);
		}

	} catch (ServantEnableException const & e) {
		throw;

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

}

void Database::enableServant(unsigned long parentID, const std::string& nName, long childIndex) {

	ScopedLock parentLock(_lockManager->lockServantForReading(parentID, eol));

	Servant* parent = getServant(parentID);

	vector<unsigned long> children = parent->getServant(nName, childIndex);

	unsigned long servantID = 0;
	if (children.size() == 0) {
		throw NoSuchServantException(parentID, nName, childIndex);

	} else if (children.size() == 1) {
		servantID = children[0];

	} else {
		long lastEnabledServant = parent->getLastEnabledServant(nName, childIndex);
		if (lastEnabledServant != -1) {
			servantID = children[lastEnabledServant];
		} else {
			throw ServantEnableException(parent->getName(), parentID, nName, childIndex);
		}
	}

	if (servantID == 0) {
		throw ServantEnableException(parent->getName(), parentID, nName, childIndex);
	}

	enableServant(servantID);
}

void Database::enableOnlyServant(unsigned long servantID) {

	try {
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));
		{
			getServant(servantID)->enable();
		}

	} catch (ServantEnableException const & e) {
		throw;

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

	NotificationServer::getInstance()->configurationChanged(_ID, servantID, common::ENABLED);
}

void Database::disableServant(unsigned long servantID) {

	try {
		unsigned long firstDriverID = 0;

		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));
		{
			set<unsigned long> servantsToDisable = _servantOperations.getServantsForDisable(getServant(servantID), firstDriverID);

			setServantDisabled(servantsToDisable);
			setServantInitialised(servantID, false);
		}

		if (firstDriverID != 0) {
			set<unsigned long> servantsToDisable = _servantOperations.getServantsForDisableDriver(dynamic_cast<Driver*>(getServant(firstDriverID)));
			setServantDisabled(servantsToDisable);
		}

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}
}

void Database::disableServant(unsigned long parentID, const std::string& nName, long childIndex) {

	ScopedLock parentLock(_lockManager->lockServantForReading(parentID, eol));

	Servant* parent = getServant(parentID);

	vector<unsigned long> children = parent->getServant(nName, childIndex);

	unsigned long servantID = 0;
	if (children.size() == 0) {
		throw NoSuchServantException(parentID, nName, childIndex);

	} else if (children.size() == 1) {
		servantID = children[0];

	} else {
		long lastEnabledServant = parent->getLastEnabledServant(nName, childIndex);
		if (lastEnabledServant != -1) {
			servantID = children[lastEnabledServant];
		} else {
			return;
		}
	}

	if (servantID == 0) {
		// no error, does nothing
		return;
	}

	disableServant(servantID);
}

void Database::disableOnlyServant(unsigned long servantID) {

	try {
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));
		{
			getServant(servantID)->disable();
		}

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

	NotificationServer::getInstance()->configurationChanged(_ID, servantID, common::DISABLED);
}

bool Database::isServantEnabled(unsigned long servantID) {

	try {
		ScopedLock lock(_lockManager->lockServantForWriting(servantID, eol));

		return getServant(servantID)->isEnabled();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}
}

bool Database::isServantEnabled(unsigned long parentID, const std::string& nName, long childIndex) {

	ScopedLock parentLock(_lockManager->lockServantForReading(parentID, eol));

	Servant* parent = getServant(parentID);

	vector<unsigned long> children = parent->getServant(nName, childIndex);

	unsigned long servantID = 0;
	if (children.size() == 0) {
		throw NoSuchServantException(parentID, nName, childIndex);

	} else if (children.size() == 1) {
		servantID = children[0];

	} else {
		long lastEnabledServant = parent->getLastEnabledServant(nName, childIndex);
		if (lastEnabledServant != -1) {
			servantID = children[lastEnabledServant];
		} else {
			return false;
		}
	}

	if (servantID == 0) {
		return false;
	}

	ScopedLock servantLock(_lockManager->lockServantForReading(servantID, eol));

	return getServant(servantID)->isEnabled();
}

void Database::setServantEnabled(const std::set<unsigned long>& servantsToEnable) {

	try {
		// iterating the controllers and drivers
		vector<unsigned long> controllerIDs = _IDControllerMap.getKeySet();
		for (vector<unsigned long>::const_iterator c = controllerIDs.begin(); c != controllerIDs.end(); c++) {

			ScopedLock servantLock(_lockManager->lockServantForReading(*c, eol));

			try {
				_servantOperations.verifyServantEnabled(_IDControllerMap.get(*c), servantsToEnable);

			} catch (ServantOperations::ServantEnableException&) {
				throw;
			} catch (Exception&) {
			}
		}

		vector<unsigned long> driverIDs = _IDDriverMap.getKeySet();
		for (vector<unsigned long>::const_iterator d = driverIDs.begin(); d != driverIDs.end(); d++) {

			ScopedLock servantLock(_lockManager->lockServantForReading(*d, eol));

			try {
				_servantOperations.verifyServantEnabled(_IDDriverMap.get(*d), servantsToEnable);

			} catch (ServantOperations::ServantEnableException&) {
				throw;
			} catch (Exception&) {
			}
		}

		// if no exception has been thrown, the servants can be enabled
		for (set<unsigned long>::iterator s = servantsToEnable.begin(); s != servantsToEnable.end(); s++) {

			unsigned long currentServantID = *s;

			// lock
			{
				ScopedLock servantLock(_lockManager->lockServantForWriting(currentServantID, eol));

				getServant(currentServantID)->enable();
			}

			NotificationServer::getInstance()->configurationChanged(_ID, currentServantID, common::ENABLED);
		}

		// updating the last enabled
		for (vector<unsigned long>::const_iterator c = controllerIDs.begin(); c != controllerIDs.end(); c++) {

			ScopedLock servantLock(_lockManager->lockServantForWriting(*c, eol));

			try {
				_IDControllerMap.get(*c)->updateLastEnabled(servantsToEnable);

			} catch (Exception&) {
			}
		}

		for (vector<unsigned long>::const_iterator d = driverIDs.begin(); d != driverIDs.end(); d++) {

			ScopedLock servantLock(_lockManager->lockServantForWriting(*d, eol));

			try {
				_IDDriverMap.get(*d)->updateLastEnabled(servantsToEnable);

			} catch (Exception&) {
			}
		}

	} catch (ServantOperations::ServantEnableException const & e) {
		throw ServantEnableException(e.getName(), e.getID(), e.getNName(), e.getIndex());
	}
}

void Database::setServantDisabled(const std::set<unsigned long>& servantsToDisable) {

	for (set<unsigned long>::iterator s = servantsToDisable.begin(); s != servantsToDisable.end(); s++) {

		unsigned long currentServantID = *s;

		// lock
		{
			ScopedLock servantLock(_lockManager->lockServantForWriting(currentServantID, eol));

			getServant(currentServantID)->disable();
		}

		NotificationServer::getInstance()->configurationChanged(_ID, currentServantID, common::DISABLED);
	}
}

void Database::addParentToServant(unsigned long parentID, unsigned long servantID) {

	try {
		if (!_parentIDsMap.find(servantID)) {
			_parentIDsMap.insert(servantID, set<unsigned long>());
		}

		_parentIDsMap.get(servantID).insert(parentID);

	} catch (Exception const & e) {
		out << "Cannot add parent " << parentID << " to " << servantID << eol;
	}
}

void Database::removeParentFromServant(unsigned long parentID, unsigned long servantID) {

	try {
		_parentIDsMap.get(servantID).erase(parentID);

		if (_parentIDsMap.get(servantID).empty()) {
			_parentIDsMap.erase(servantID);
		}

	} catch (Exception const & e) {
		out << "Cannot remove parent " << parentID << " from " << servantID << eol;
	}
}

void Database::setServantVisible(unsigned long servantID, bool value) {

	try {
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

		getServant(servantID)->setVisible(value);

	} catch (ServantEnableException const & e) {
		throw;

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

	NotificationServer::getInstance()->configurationChanged(_ID, servantID, common::TREE);
}

void Database::setServantFavourite(unsigned long servantID, bool value) {

	try {
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

		getServant(servantID)->setFavourite(value);

	} catch (ServantEnableException const & e) {
		throw;

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

	// forcing to update the tree
	NotificationServer::getInstance()->configurationChanged(_ID, 0, common::TREE);
}

void Database::setServantCommand(unsigned long servantID, bool value) {

	try {
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

		getServant(servantID)->setCommand(value);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

	NotificationServer::getInstance()->configurationChanged(_ID, servantID, common::TREE);
}

vector<Database::ServantItem> Database::getControllers() {

	vector<ServantItem> result;

	// Iterating the controllers.
	vector<unsigned long> controllerIDs = _IDControllerMap.getKeySet();
	for (vector<unsigned long>::const_iterator c = controllerIDs.begin(); c != controllerIDs.end(); ++c) {

		ScopedLock lock(_lockManager->lockServantForReading(*c, eol));

		try {
			Controller* controller = _IDControllerMap.get(*c);

			ServantItem item;

			item.ID = *c;
			item.name = controller->getName();
			item.type = controller->getControllerType();
			item.clientType = controller->getClientType();
			item.family = controller->getFamily();
			item.subfamily = controller->getSubfamily();
			item.isCommand = controller->isCommand();
			item.isVisible = controller->isVisible();
			item.isEnabled = controller->isEnabled();
			item.isFavourite = controller->isFavourite();
			item.isGenerated = controller->isGenerated();
			item.isRoot = false;

			result.push_back(item);

		} catch (Exception&) {
		}
	}

	return result;
}

Database::ServantItem Database::getController(unsigned long servantID) {

	Controller* controller = _IDControllerMap.get(servantID);

	ServantItem item;

	item.ID = servantID;
	item.name = controller->getName();
	item.type = controller->getControllerType();
	item.clientType = controller->getClientType();
	item.family = controller->getFamily();
	item.subfamily = controller->getSubfamily();
	item.isCommand = controller->isCommand();
	item.isVisible = controller->isVisible();
	item.isEnabled = controller->isEnabled();
	item.isFavourite = controller->isFavourite();
	item.isGenerated = controller->isGenerated();
	item.isRoot = false;

	return item;
}

Controller* Database::getController(const std::string& type) {
	vector<unsigned long> controllerIDs = _IDControllerMap.getKeySet();
	for (vector<unsigned long>::iterator c = controllerIDs.begin(); c != controllerIDs.end(); ++c) {

		Controller* controller = _IDControllerMap.get(*c);

		if (controller->getControllerType() == type) {
			return controller;
		}
	}

	return 0;
}

vector<Database::ServantItem> Database::getDrivers() {

	vector<ServantItem> result;

	// Iterating the drivers.
	vector<unsigned long> driverIDs = _IDDriverMap.getKeySet();
	for (vector<unsigned long>::const_iterator d = driverIDs.begin(); d != driverIDs.end(); ++d) {

		ScopedLock lock(_lockManager->lockServantForReading(*d, eol));

		try {
			Driver* driver = _IDDriverMap.get(*d);

			ServantItem item;

			item.ID = *d;
			item.name = driver->getName();
			item.type = driver->getDriverType();
			item.clientType = driver->getClientType();
			item.isCommand = driver->isCommand();
			item.isVisible = driver->isVisible();
			item.isEnabled = driver->isEnabled();
			item.isRoot = driver->isRoot();

			result.push_back(item);

		} catch (Exception&) {
		}
	}

	return result;
}

Database::ServantItem Database::getDriver(unsigned long servantID) {

	Driver* driver = _IDDriverMap.get(servantID);

	ServantItem item;

	item.ID = servantID;
	item.name = driver->getName();
	item.type = driver->getDriverType();
	item.clientType = driver->getClientType();
	item.isCommand = driver->isCommand();
	item.isVisible = driver->isVisible();
	item.isEnabled = driver->isEnabled();
	item.isRoot = driver->isRoot();

	return item;
}

unsigned long Database::getServantID(unsigned long commandID) {

	try {
		ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

		return getCommand(commandID)->getServantID();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

std::string Database::getCommandName(unsigned long commandID) {

	string commandName;

	try {
		ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

		commandName = getCommand(commandID)->getName();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}

	return commandName;
}

std::string Database::getServantName(unsigned long servantID) {

	try {
		ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

		return getServant(servantID)->getName();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}
}

unsigned long Database::getServantID(const std::string& servantName) {

	// We search for the compact servant name: without any whitespace.
	// The string servantName can already be compact if it comes from an expression (no whitespace allowed).
	string compactServantName = common::StringUtilities::stripWhitespaces(servantName);

	// Search among the controllers.
	vector<unsigned long> controllerIDs = _IDControllerMap.getKeySet();
	for (vector<unsigned long>::const_iterator c = controllerIDs.begin(); c != controllerIDs.end(); c++) {

		ScopedLock lock(_lockManager->lockServantForReading(*c, eol));

		try {
			// Compact the controller name.
			string controllerName = common::StringUtilities::stripWhitespaces(_IDControllerMap.get(*c)->getName());

			if (controllerName == compactServantName) {
				return *c;
			}

		} catch (Exception&) {
		}
	}

	// Search among the drivers.
	vector<unsigned long> driverIDs = _IDDriverMap.getKeySet();
	for (vector<unsigned long>::const_iterator d = driverIDs.begin(); d != driverIDs.end(); d++) {

		ScopedLock lock(_lockManager->lockServantForReading(*d, eol));

		try {
			// Compact the driver name.
			string driverName = common::StringUtilities::stripWhitespaces(_IDDriverMap.get(*d)->getName());

			if (driverName == compactServantName) {
				return *d;
			}

		} catch (Exception&) {
		}
	}

	throw NoSuchServantException(servantName);
}

long Database::getServantListSize(unsigned long servantID, const std::string& nServantName) {

	try {
		ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

		return getServant(servantID)->getServantListSize(nServantName);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}
}

void Database::resizeServantList(unsigned long servantID, const std::string& nServantName, long size) {

	ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

	Servant* servant = 0;

	try {
		servant = getServant(servantID);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

	// getting the servants
	long previousSize = servant->getServantListSize(nServantName);

	for (long index = size; index < previousSize; index++) {
		removeServantFromParent(servantID, nServantName, index);
	}

	// finally resizing the list
	servant->resizeServantList(nServantName, size);
}

std::string Database::getControllerType(unsigned long controllerID) {

	ScopedLock lock(_lockManager->lockServantForReading(controllerID, eol));

	try {
		return _IDControllerMap.get(controllerID)->getControllerType();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(controllerID), e);
	}
}

std::string Database::getDriverType(unsigned long driverID) {

	ScopedLock lock(_lockManager->lockServantForReading(driverID, eol));

	try {
		return _IDDriverMap.get(driverID)->getDriverType();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(driverID), e);
	}
}

std::string Database::getServantType(unsigned long servantID) {

	string type;

	try {
		type = "controller." + getControllerType(servantID);

	} catch (NoSuchServantException&) {
	}

	try {
		type = "driver." + getDriverType(servantID);

	} catch (NoSuchServantException&) {
	}

	return type;
}

std::string Database::getShortServantType(unsigned long servantID) {

	string type;

	try {
		type = getControllerType(servantID);

	} catch (NoSuchServantException&) {
	}

	try {
		type = getDriverType(servantID);

	} catch (NoSuchServantException&) {
	}

	return type;
}

std::deque<std::string> Database::getControllerTypes(unsigned long controllerID) {

	try {
		string servantType = getShortServantType(controllerID);
		return _controllerTypeInheritance.get(servantType);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(controllerID), e);
	}
}

std::set<std::string> Database::getControllerTypesInheritingType(const std::string& type) {

	std::set<std::string> result;

	for (HashMap<string, deque<string> >::const_iterator i = _controllerTypeInheritance.begin(); i != _controllerTypeInheritance.end(); ++i) {

		string servantType = i->first;
		deque<string> const & types = i->second;

		if (find(types.begin(), types.end(), type) != types.end()) {
			result.insert(servantType);
		}
	}

	return result;
}

void Database::setServantFamily(unsigned long servantID, const std::string& familyName, const std::string& subfamilyName) {

	try {
		ScopedLock lock(_lockManager->lockServantForWriting(servantID, eol));

		getServant(servantID)->setFamily(familyName, subfamilyName);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

	// forcing to update the tree
	NotificationServer::getInstance()->configurationChanged(_ID, 0, common::TREE);
}

std::string Database::getServantFamily(unsigned long servantID) {

	try {
		ScopedLock lock(_lockManager->lockServantForWriting(servantID, eol));

		return getServant(servantID)->getFamily();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}
}

std::string Database::getServantSubfamily(unsigned long servantID) {

	try {
		ScopedLock lock(_lockManager->lockServantForWriting(servantID, eol));

		return getServant(servantID)->getSubfamily();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}
}

bool Database::isServantFavourite(unsigned long servantID) {
	try {
		ScopedLock lock(_lockManager->lockServantForWriting(servantID, eol));

		return getServant(servantID)->isFavourite();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}
}

void Database::addServantType(unsigned long servantID, const std::string& type, bool atFront) {
	try {
		ScopedLock lock(_lockManager->lockServantForWriting(servantID, eol));

		bool controller = isController(servantID);

		// add in the inheritance map
		string servantType = getShortServantType(servantID);

		if (controller) {
			if (!_controllerTypeInheritance.find(servantType)) {
				_controllerTypeInheritance.insert(servantType, deque<string>());
			}
		} else {
			if (!_driverTypeInheritance.find(servantType)) {
				_driverTypeInheritance.insert(servantType, deque<string>());
			}
		}

		deque<string>& types = (controller ? _controllerTypeInheritance.get(servantType) : _driverTypeInheritance.get(servantType));

		if (!atFront) {
			deque<std::string>::const_iterator t = types.begin();
			deque<std::string>::const_iterator end = types.end();
			while (t != end) {
				if (type == *t) {
					return;
				}
				t++;
			}

			types.push_front(type);

		} else {

			if (types.empty() || type != types.front()) {
				types.push_front(type);
			}
		}

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}
}

void Database::setCommandState(unsigned long servantID, unsigned long commandID, common::CommandState state) {

	// locking servant, command
	{
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));
		ScopedLock commandLock(_lockManager->lockCommandForWriting(commandID, eol));

		Command* command = 0;

		try {
			command = getCommand(commandID);

		} catch (Exception const & e) {
			throw link(NoSuchCommandException(commandID), e);
		}

		if (command->getState() != state) {
			if (command->getState() == common::INACTIVE && (state == common::STOPPING || state == common::PAUSING)) {
				// do not change INACTIVE state
			} else {
				command->setState(state);
			}

			// special case for drivers
			Driver* driver = dynamic_cast<Driver*>(getServant(servantID));
			if (driver != 0) {
				// do not set the states only coming from the scheduler
				// to avoid replacing ACTIVE and other states
				if (state != common::STARTING && state != common::PENDING) {
					driver->setState(state);
				}
			}
		}
	}

	NotificationServer::getInstance()->commandStateChanged(_ID, servantID, commandID, state);
}

common::SetCommandActiveResult Database::setCommandActive(unsigned long servantID, unsigned long commandID, const std::string servantType, unsigned long processorID, unsigned long threadID, DataChangeSubscriber* subscriber) {

	// The subscriber is only used in case of wait

	common::SetCommandActiveResult result;
	bool changeToActiveState = false;

	// locking servant, command
	{
		ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));
		ScopedLock commandLock(_lockManager->lockCommandForWriting(commandID, eol));

		Command* command = 0;

		try {
			command = getCommand(commandID);

		} catch (Exception const & e) {
			throw link(NoSuchCommandException(commandID), e);
		}

		bool done = false;

		// special case for a driver, verify the state of the driver
		Driver* driver = dynamic_cast<Driver*>(getServant(servantID));
		if (driver != 0) {

			if (driver->getServerType() == common::ICS) {
				done = false;

			} else {

				common::CommandState driverState = driver->getState();
				if (driverState == common::ACTIVE || driverState == common::STOPPING || driverState == common::RESUMING) {
					// return WAIT if the request is made in another processors or threads
					// except for STOPPING for which the onStop is executed in the main thread (thread test has not to be done!)
					if (processorID != driver->getRunningProcessorID()
						|| (driverState != common::STOPPING && threadID != driver->getRunningThreadID())) {
						// registering the subscriber
						unsigned long subscriberID = NotificationServer::getInstance()->getDataChangePublisher()->subscribe(subscriber);
						subscriber->setID(subscriberID);

						//out << "command " << commandID << " waits because processorID = " << processorID << " != driver " << driver->getRunningProcessorID()
						//	<< " threadID = " << threadID << " != driver " << driver->getRunningThreadID()
						//	<< " driver state = " << (int)driverState << eol;

						done = true;
						result = WAIT;
					}

				} else if (driverState == common::PAUSED || driverState == common::PAUSING) {
					// no need to keep the subscriber
					delete subscriber;

					done = true;
					result = SKIP;
				}
			}
		}

		// continue if not already processed
		if (!done) {

			// in case of driver, the test can continue to the command
			common::CommandState currentState = command->getState();

			if (currentState != common::ACTIVE) {
				if (currentState == common::INACTIVE || currentState == common::PENDING || currentState == common::STARTING) {

					// the servant is an inactive driver or a controller
					command->setServantType(servantType);
					command->setRunningProcessorID(processorID);
					command->setRunningThreadID(threadID);
					command->setState(common::ACTIVE);

					// special case for drivers
					if (driver != 0) {
						driver->setState(common::ACTIVE);
						driver->setRunningProcessorID(processorID);
						driver->setRunningThreadID(threadID);

						//out << "setting state ACTIVE to driver for command " << command->getID() << " processorID = " << processorID << " threadID = " << threadID << eol;
					}

					changeToActiveState = true;

					// no need to keep the subscriber
					delete subscriber;

					result = EXECUTE;

				} else {
					// no need to keep the subscriber
					delete subscriber;

					//out << "skipping " << commandID << " because command is in state " << currentState << eol;

					result = SKIP;
				}

			} else if (command->getServantType() != servantType) {
				// super call
				// no need to keep the subscriber
				delete subscriber;

				result = EXECUTE;

			} else if (processorID == command->getRunningProcessorID() && threadID == command->getRunningThreadID()) {
				// no need to keep the subscriber
				delete subscriber;

				result = RECURSIVE;

			} else {
				// registering the subscriber
				unsigned long subscriberID = NotificationServer::getInstance()->getDataChangePublisher()->subscribe(subscriber);
				subscriber->setID(subscriberID);

				result = WAIT;
			}
		}
	}

	// notification
	if (changeToActiveState) {
		NotificationServer::getInstance()->commandStateChanged(_ID, servantID, commandID, common::ACTIVE);
	}

	return result;
}

common::CommandState Database::getCommandState(unsigned long commandID) {

	try {
		ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

		return getCommand(commandID)->getState();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

double Database::getCommandElapsedTimeMs(unsigned long commandID) {

	try {
		ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

		return getCommand(commandID)->getElapsedTimeMs();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

void Database::resetElapsedSimulationTime(unsigned long driverID) {

	try {
		ScopedLock lock(_lockManager->lockServantForWriting(driverID, eol));

		_IDDriverMap.get(driverID)->resetElapsedSimulationTime();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(driverID), e);
	}
}

double Database::getElapsedSimulationTimeMs(unsigned long driverID) {

	try {
		ScopedLock lock(_lockManager->lockServantForReading(driverID, eol));

		return _IDDriverMap.get(driverID)->getElapsedSimulationTimeMs();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(driverID), e);
	}
}

void Database::setProgression(unsigned long commandID, double progression) {

	// locking command
	{
		ScopedLock lock(_lockManager->lockCommandForWriting(commandID, eol));

		Command* command = 0;

		try {
			command = getCommand(commandID);

		} catch (Exception const & e) {
			throw link(NoSuchCommandException(commandID), e);
		}

		command->setProgression(progression);
	}

	NotificationServer::getInstance()->commandProgressionChanged(_ID, commandID, progression);
}

void Database::setUnknownProgression(unsigned long commandID) {

	// locking command
	{
		ScopedLock lock(_lockManager->lockCommandForWriting(commandID, eol));

		try {
			getCommand(commandID)->setProgression(-1);


		} catch (Exception const & e) {
			throw link(NoSuchCommandException(commandID), e);
		}
	}

	NotificationServer::getInstance()->commandProgressionChanged(_ID, commandID, -1);
}

double Database::getProgression(unsigned long commandID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		return getCommand(commandID)->getProgression();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

double Database::getCommandRemainingTimeMs(unsigned long commandID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		return getCommand(commandID)->getRemainingTimeMs();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

double Database::getCommandDurationMs(unsigned long commandID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		return getCommand(commandID)->getDurationMs();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

int64 Database::getCommandStartDate(unsigned long commandID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		return getCommand(commandID)->getStartDate();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

void Database::setCommandPreRunCalculatedDurationMs(unsigned long commandID, double duration) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		getCommand(commandID)->setPreRunCalculatedDurationMs(duration);

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

double Database::getCommandPreRunCalculatedDurationMs(unsigned long commandID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		return getCommand(commandID)->getPreRunCalculatedDurationMs();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

void Database::setCommandFromCommandZone(unsigned long commandID, bool fromCommandZone) {

	ScopedLock lock(_lockManager->lockCommandForWriting(commandID, eol));

	try {
		getCommand(commandID)->setFromCommandZone(fromCommandZone);

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

bool Database::isCommandFromCommandZone(unsigned long commandID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		return getCommand(commandID)->isFromCommandZone();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

bool Database::isController(unsigned long servantID) {
	return _IDControllerMap.find(servantID);
}

bool Database::isDriver(unsigned long servantID) {
	return _IDDriverMap.find(servantID);
}

bool Database::isServant(unsigned long servantID) {
	return (isController(servantID) || isDriver(servantID));
}

bool Database::isIcsServant(unsigned long servantID) {

	ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

	Servant * servant = getServant(servantID);

	if (servant != 0) {
		return (servant->getServerType() == common::ICS);
	}

	return false;
}

bool Database::isCommand(unsigned long commandID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		_IDCommandMap.get(commandID);
		return true;

	} catch (Exception const & e) {
		// do nothing
	}

	return false;
}

bool Database::isProperty(unsigned long propertyID) {

	ScopedLock lock(_lockManager->lockPropertyForReading(propertyID, eol));

	try {
		_keyIDPropertyMap.get(propertyID);
		return true;

	} catch (Exception const & e) {
		// do nothing
	}

	return false;
}

std::string Database::getPropertyType(unsigned long propertyID) {

	ScopedLock lock(_lockManager->lockPropertyForReading(propertyID, eol));
	std::string propertyType;

	try {
		propertyType = _keyIDPropertyMap.get(propertyID)->getType();
	} catch (Exception const & e) {
		throw NoSuchPropertyException(propertyID);
	}

	return propertyType;
}

PropertyHolder* Database::getPropertyHolder(unsigned long servantID) {

	if (isController(servantID)) {
		return _IDControllerMap.get(servantID);
	}

	return _IDDriverMap.get(servantID);
}

Servant* Database::getServant(unsigned long servantID) {

	if (isController(servantID)) {
		return static_cast<Servant*>(_IDControllerMap.get(servantID));
	} else if (isDriver(servantID)) {
		return static_cast<Servant*>(_IDDriverMap.get(servantID));
	}

	throw NoSuchServantException(servantID);
}

std::string Database::getFormattedValue(unsigned long propertyID) {

	try {
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		ScopedLock lock(_lockManager->lockPropertyForReading(propertyKeyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(propertyKeyID);
		BaseElementaryProperty * typedProperty = dynamic_cast<BaseElementaryProperty *>(property);

		if (typedProperty == 0) {
			return "";
		}

		return typedProperty->getFormattedValue();

	} catch (common::Exception const & e) {
		throw;
	}
}

void Database::updateIcsProperty(unsigned long propertyID) {

	//out << "Database::updateIcsProperty " << propertyID << eol;

	NotificationServer::getInstance()->icsPropertyChanged(_ID, propertyID);
}

Lock* Database::beginArrayTransfer(unsigned long propertyID) {

	return _lockManager->lockPropertyForReading(getPropertyKeyID(propertyID), eol);
}

Database::Database() :
	_ID(0) {

	try {
		string lockManagerType = ServerProperties::getInstance()->get<string>(LOCK_MANAGER_TYPE);

		if (lockManagerType == NO_LOCK_MANAGER) {
			_lockManager = new NoLockManager();
		} else if (lockManagerType == SINGLE_LOCK_MANAGER) {
			_lockManager = new SingleLockManager();
		} else if (lockManagerType == ELEMENT_LOCK_MANAGER) {
			_lockManager = new ElementLockManager();
		} else {
			err << lockManagerType << " is not a valid type!" << eol;
			_lockManager = new SingleLockManager();
		}
	} catch (ServerProperties::NoSuchPropertyException const & e) {
		warn << "No lock manager type: use " << ELEMENT_LOCK_MANAGER << " by default" << eol;
		_lockManager = new ElementLockManager();
	}

	_servantOperations.setLockManager(_lockManager);

	_IDControllerMap.setLockManager(_lockManager);
	_IDDriverMap.setLockManager(_lockManager);
	_IDCommandMap.setLockManager(_lockManager);
	_keyIDPropertyMap.setLockManager(_lockManager);
	_IDPropertyIDServantMap.setLockManager(_lockManager);
	_parentIDsMap.setLockManager(_lockManager);
	_unitIDPropertyMap.setLockManager(_lockManager);

}

Database::~Database() {

	delete _lockManager;
}

void Database::setMode(unsigned long driverID, common::DriverMode mode) {

	ScopedLock lock(_lockManager->lockServantForWriting(driverID, eol));

	try {
		_IDDriverMap.get(driverID)->setMode(mode);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(driverID), e);
	}
}

common::DriverMode Database::getMode(unsigned long driverID) {

	ScopedLock lock(_lockManager->lockServantForReading(driverID, eol));

	try {
		return _IDDriverMap.get(driverID)->getMode();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(driverID), e);
	}
}

unsigned long Database::getDriverParentID(unsigned long driverID) {

	ScopedLock lock(_lockManager->lockServantForReading(driverID, eol));

	try {
		return _IDDriverMap.get(driverID)->getFatherID();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(driverID), e);
	}
}

unsigned long Database::getEnabledOrUniqueServantID(unsigned long parentID, const std::string& nServantName, long childIndex) {

	ScopedLock lock(_lockManager->lockServantForReading(parentID, eol));

	try {
		Servant* servant = getServant(parentID);

		vector<unsigned long> children = servant->getServant(nServantName, childIndex);

		if (children.empty()) {
			throw NoSuchServantException(parentID, nServantName, childIndex);

		} else {
			bool nullServants = true;

			// see if there is a unique servant
			if (children.size() == 1) {
				unsigned long currentServantID = *(children.begin());

				if (currentServantID != 0) {
					nullServants = false;

					return currentServantID;
				}

			// obtain currently enabled servant
			} else {

				for (vector<unsigned long>::const_iterator c = children.begin(); c != children.end(); c++) {

					unsigned long currentServantID = *c;

					if (currentServantID != 0) {
						nullServants = false;

						ScopedLock lock(_lockManager->lockServantForReading(currentServantID, eol));

						if (getServant(currentServantID)->isEnabled()) {
							return currentServantID;
						}
					}
				}
			}

			if (nullServants) {
				throw NoSuchServantException(parentID, nServantName, childIndex);

			} else {
				throw ServantDisabledException(parentID, nServantName, childIndex);
			}
		}

	} catch (Servant::ServantIndexException const & e) {
		throw IndexOutOfBoundsException(parentID, nServantName, e.getSize());

	} catch (ServantDisabledException const & e) {
		throw;

	} catch (NoSuchServantException const & e) {
		throw;

	} catch (Exception const & e) {
		throw link(NoSuchServantException(parentID, nServantName, childIndex), e);
	}
}

std::vector<unsigned long> Database::getServantIDs(unsigned long parentID, const std::string& nServantName, long childIndex) {

	ScopedLock lock(_lockManager->lockServantForReading(parentID, eol));

	try {
		Servant* servant = getServant(parentID);

		vector<unsigned long> children = servant->getServant(nServantName, childIndex);

		vector<unsigned long> result;
		for (vector<unsigned long>::const_iterator c = children.begin(); c != children.end(); c++) {
			if (*c != 0) {
				result.push_back(*c);
			}
		}

		return result;

	} catch (Servant::ServantIndexException const & e) {
		throw IndexOutOfBoundsException(parentID, nServantName, e.getSize());

	} catch (NoSuchServantException const & e) {
		throw;

	} catch (Exception const & e) {
		throw link(NoSuchServantException(parentID, nServantName, childIndex), e);
	}
}

unsigned long Database::getServantIDFromPropertyID(unsigned long propertyID) {

	ScopedLock lock(_lockManager->lockPropertyForReading(propertyID, eol));

	if (_IDPropertyIDServantMap.find(propertyID)) {
		return _IDPropertyIDServantMap.get(propertyID);
	}

	return 0;
}

unsigned long Database::getPropertyID(unsigned long servantID, const std::string& propertyName) {

	ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

	try {
		return getServant(servantID)->getProperties()[propertyName];

	} catch (Exception&) {
		throw NoSuchPropertyException(servantID, propertyName);
	}
}

std::string Database::getPropertyName(unsigned long propertyID) {

	ScopedLock lock(_lockManager->lockPropertyForReading(propertyID, eol));

	try {
		return getPropertyFromKeyID(propertyID)->getName();

	} catch (Exception&) {
		throw NoSuchPropertyException(propertyID);
	}
}


std::pair<std::string, long> Database::getCompletePropertyName(unsigned long servantID, unsigned long propertyID) {

	ScopedLock lock(_lockManager->lockPropertyForReading(propertyID, eol));

	try {
		Controller * controller = 0;

		try {
			controller = _IDControllerMap.get(servantID);

		} catch (Exception const & e) {
			throw link(NoSuchServantException(servantID), e);
		}
		controller->getProperties();
		unsigned long parentpropid = 0;
		unsigned long index = 0;
		HashMap<string, unsigned long> const & properties = controller->getProperties();
		for (HashMap<string, unsigned long>::const_iterator p = properties.begin(); p != properties.end(); ++p) {
			BaseDynamicProperty* parentProperty = dynamic_cast<BaseDynamicProperty*>(_keyIDPropertyMap.get(p->second));
			if (parentProperty != nullptr) {
				bool foundit = false;
				index = 0;
				for(auto i : parentProperty->getPropertyIDs()) {
					if (i == propertyID) {
						parentpropid = p->second;
						foundit = true;
						break;
					}
					++index;
				}
				if (foundit == true) {
					break;
				}
			}
		}
		if (parentpropid != 0) {
			return make_pair(_keyIDPropertyMap[propertyID]->getName(), (long) index);
		}
		else {
			return make_pair(_keyIDPropertyMap[propertyID]->getName(), (long) -1);
		}

	} catch (Exception&) {
		throw NoSuchPropertyException(propertyID);
	}
}

unsigned long Database::getUnitPropertyID(unsigned long propertyID) {

	if (_unitIDPropertyMap.find(propertyID)) {
		return _unitIDPropertyMap.get(propertyID);
	}

	return 0;
}

Command* Database::getCommand(unsigned long commandID) {

	// the commandID can be either the ID of a real command, or a property ID
	// in case of a property setter
	// searching into the key ID property map to access directly the property
	// and not the eventual reference.
	// this is made for preventing a concurrent access from two servants using the
	// same property
	if (_keyIDPropertyMap.find(commandID)) {
		return _keyIDPropertyMap.get(commandID)->getCommand();
	}

	// searching into the command map
	return _IDCommandMap.get(commandID);
}

unsigned long Database::getServantIDFromCommandID(unsigned long commandID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		return getCommand(commandID)->getServantID();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

unsigned long Database::getCommandID(unsigned long servantID, const std::string& commandName) {

	ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

	if (isController(servantID)) {
		return _IDControllerMap.get(servantID)->getCommand(commandName);
	} else if (isDriver(servantID)) {
		return _IDDriverMap.get(servantID)->getCommand(commandName);
	} else {
		throw NoSuchServantException(servantID);
	}
}

void Database::setServantInitialised(unsigned long servantID, bool initialised) {

	ScopedLock lock(_lockManager->lockServantForWriting(servantID, eol));

	if (isController(servantID)) {
		_IDControllerMap.get(servantID)->setInitialised(initialised);
	} else if (isDriver(servantID)) {
		_IDDriverMap.get(servantID)->setInitialised(initialised);
	} else {
		throw NoSuchServantException(servantID);
	}

}

bool Database::isServantInitialised(unsigned long servantID) {

	ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

	bool result = true;
	if (isController(servantID)) {
		Controller* controller = _IDControllerMap.get(servantID);
		if (controller->getServerType() == common::NOMAD) {
			result = controller->isInitialised();
		}
	} else if (isDriver(servantID)) {
		Driver* driver = _IDDriverMap.get(servantID);
		if (driver->getServerType() == common::NOMAD) {
			result = driver->isInitialised();
		}
	} else {
		throw NoSuchServantException(servantID);
	}

	return result;
}

bool Database::isServantInitialised(unsigned long parentID, const std::string& nName, long childIndex) {

	ScopedLock parentLock(_lockManager->lockServantForReading(parentID, eol));

	Servant* parent = getServant(parentID);

	vector<unsigned long> children = parent->getServant(nName, childIndex);

	unsigned long servantID = 0;
	if (children.size() == 0) {
		throw NoSuchServantException(parentID, nName, childIndex);

	} else if (children.size() == 1) {
		servantID = children[0];

	} else {
		long lastEnabledServant = parent->getLastEnabledServant(nName, childIndex);
		if (lastEnabledServant != -1) {
			servantID = children[lastEnabledServant];
		} else {
			return false;
		}
	}

	if (servantID == 0) {
		return false;
	}

	ScopedLock servantLock(_lockManager->lockServantForReading(servantID, eol));

	return getServant(servantID)->isInitialised();
}

long Database::getDriverChannel(unsigned long servantID) {

	ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

	try {
		return _IDDriverMap.get(servantID)->getChannel();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}
}

void Database::setServantInDebug(unsigned long servantID) {

	ScopedLock lock(_lockManager->lockServantForWriting(servantID, eol));

	_debugIDServants.push_back(servantID);

	cout << "Servant " << servantID << " in debug" << endl;
}

bool Database::isServantInDebug(unsigned long servantID) {

	ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

	auto it = std::find(_debugIDServants.begin(), _debugIDServants.end(), servantID);
	return (it != _debugIDServants.end());
}

void Database::init(std::string const & configPath, bool clean) {

	StringHelper::init();
	Real::init();

	DatabaseLoader loader;

	// Parse the config files
	try {
		loader.loadServants(configPath, clean);

		// create the servants
		loader.createControllers();
		loader.createDrivers();

		// configure the servants
		loader.connectControllerServants();
		loader.connectDriverServants();

	} catch (Exception& e) {
		throw link(InitialisationException(), e);
	}

	// error
	bool onError = false;

	// Verifying the servants have no circularity
	vector<unsigned long> controllerIDs = _IDControllerMap.getKeySet();
	for (vector<unsigned long>::iterator c = controllerIDs.begin(); c != controllerIDs.end(); ++c) {

		try {
			_servantOperations.verifyCircularity(_IDControllerMap.get(*c));

		} catch (ServantOperations::CircularServantException& e) {

			onError = true;
			err << _IDControllerMap.get(*c)->getName() << " has circular servants" << eol;
		}
	}

	vector<unsigned long> driverIDs = _IDDriverMap.getKeySet();
	for (vector<unsigned long>::iterator d = driverIDs.begin(); d != driverIDs.end(); ++d) {

		try {
			_servantOperations.verifyCircularity(_IDDriverMap.get(*d));

		} catch (ServantOperations::CircularServantException& e) {

			onError = true;
			err << _IDDriverMap.get(*d)->getName() << " has circular servants" << eol;
		}
	}

	// exiting if error
	if (onError) {
		exit(EXIT_FAILURE);
	}

	// Verifying the servants are well enabled
	for (vector<unsigned long>::iterator c = controllerIDs.begin(); c != controllerIDs.end(); ++c) {

		try {
			_servantOperations.verifyServantsEnabled(_IDControllerMap.get(*c));

		} catch (ServantOperations::MultipleServantsEnabledException& e) {

			onError = true;

			string index;
			if (e.getIndex() != -1) {
				index = string(" at index <") + to<string>(e.getIndex()) + ">";
			}

			err << _IDControllerMap.get(*c)->getName() << " has multiple servants enabled for the servant '" << e.getNName() << "'" << index << eol;
		}
	}

	for (vector<unsigned long>::iterator d = driverIDs.begin(); d != driverIDs.end(); ++d) {

		try {
			_servantOperations.verifyServantsEnabled(_IDDriverMap.get(*d));

		} catch (ServantOperations::MultipleServantsEnabledException& e) {

			onError = true;

			string index;
			if (e.getIndex() != -1) {
				index = string(" at index <") + to<string>(e.getIndex()) + ">";
			}

			err << _IDDriverMap.get(*d)->getName() << " has multiple servants enabled for the servant '" << e.getNName() << "'" << index << eol;
		}
	}

	// exiting if error
	if (onError) {
		exit(EXIT_FAILURE);
	}

}

void Database::reset() {

	vector<unsigned long> controllerIDs = _IDControllerMap.getKeySet();
	for (vector<unsigned long>::iterator c = controllerIDs.begin(); c != controllerIDs.end(); ++c) {
		delete _IDControllerMap.get(*c);
	}

	_IDControllerMap.clear();

	vector<unsigned long> driverIDs = _IDDriverMap.getKeySet();
	for (vector<unsigned long>::iterator d = driverIDs.begin(); d != driverIDs.end(); ++d) {
		delete _IDDriverMap.get(*d);
	}

	_IDDriverMap.clear();

	vector<unsigned long> commandIDs = _IDCommandMap.getKeySet();
	for (vector<unsigned long>::iterator c = commandIDs.begin(); c != commandIDs.end(); ++c) {
		delete _IDCommandMap.get(*c);
	}

	_IDCommandMap.clear();

	vector<unsigned long> propertyIDs = _keyIDPropertyMap.getKeySet();
	for (vector<unsigned long>::iterator p = propertyIDs.begin(); p != propertyIDs.end(); ++p) {
		delete _keyIDPropertyMap.get(*p);
	}

	_keyIDPropertyMap.clear();
}

Controller* Database::findController(const std::string& name) {

	vector<unsigned long> controllerIDs = _IDControllerMap.getKeySet();
	for (vector<unsigned long>::iterator c = controllerIDs.begin(); c != controllerIDs.end(); ++c) {

		Controller* controller = _IDControllerMap.get(*c);

		if (controller->getName() == name) {
			return controller;
		}
	}

	return 0;
}

Driver* Database::findDriver(const std::string& name) {

	vector<unsigned long> driverIDs = _IDDriverMap.getKeySet();
	for (vector<unsigned long>::iterator d = driverIDs.begin(); d != driverIDs.end(); ++d) {

		Driver* driver = _IDDriverMap.get(*d);

		if (driver->getName() == name) {
			return driver;
		}
	}

	return 0;
}

Servant* Database::findServant(const std::string& name) {

	Controller* controller = findController(name);

	if (controller != 0) {
		return controller;
	}

	return findDriver(name);
}

void Database::setCommandLogID(unsigned long commandID, const std::string& logID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		getCommand(commandID)->setLogID(logID);

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}

}

std::string Database::getCommandLogID(unsigned long commandID) {

	ScopedLock lock(_lockManager->lockCommandForReading(commandID, eol));

	try {
		return getCommand(commandID)->getLogID();

	} catch (Exception const & e) {
		throw link(NoSuchCommandException(commandID), e);
	}
}

std::string Database::what(unsigned long elementID, bool withReference) {

	if (_IDControllerMap.find(elementID)) {

		ScopedLock lock(_lockManager->lockServantForReading(elementID, eol));

		Controller* controller = _IDControllerMap.get(elementID);
		if (controller != 0) {
			return controller->toString();
		}
	}

	if (_IDDriverMap.find(elementID)) {

		ScopedLock lock(_lockManager->lockServantForReading(elementID, eol));

		Driver* driver = _IDDriverMap.get(elementID);
		if (driver != 0) {
			return driver->toString();
		}
	}

	if (_IDCommandMap.find(elementID)) {

		// First getting the servant ID to order the locks.
		int servantID = 0;
		{
			ScopedLock lock(_lockManager->lockCommandForReading(elementID, eol));

			Command* command = _IDCommandMap.get(elementID);
			if (command != 0) {
				servantID = command->getServantID();
			}
			else {
				return "?";
			}
		}

		ScopedLock servantLock(_lockManager->lockServantForReading(servantID, eol));
		ScopedLock commandLock(_lockManager->lockCommandForReading(elementID, eol));

		Command* command = _IDCommandMap.get(elementID);
		if (command != 0) {
			return command->toString();
		}
	}

	if (_keyIDPropertyMap.find(elementID)) {

		// First getting the servant ID to order the locks.
		int servantID = 0;
		{
			ScopedLock lock(_lockManager->lockPropertyForReading(elementID, eol));

			BaseProperty* property = _keyIDPropertyMap.get(elementID);
			if (property != 0) {
				servantID = property->getParentID();
			}
			else {
				return "?";
			}
		}

		ScopedLock servantLock(_lockManager->lockServantForReading(servantID, eol));
		ScopedLock lock(_lockManager->lockPropertyForReading(elementID, eol));

		BaseProperty* property = _keyIDPropertyMap.get(elementID);
		if (property != 0) {
			return property->toString();
		}
	}

	return "";
}

void Database::reportBug(int64 date) {

	err << "database : \n" << toString() << eol;
}

void Database::verifyLocks(int seconds) {

	_lockManager->verifyLocks(IndexGenerator::getInstance()->getCurrentIndex(), seconds);
}

Lock* Database::lock(unsigned long ID) {

	// implemented identically for all elements
	return _lockManager->lockPropertyForReading(ID, eol);
}

void Database::unlock(unsigned long ID) {

	_lockManager->unlockPropertyForReading(ID);
}

void Database::setClientType(unsigned long servantID, const std::string& clientType) {

	ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

	Servant* servant = getServant(servantID);

	servant->setClientType(clientType);

}

bool Database::isServantPausable(unsigned long servantID) {
	ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));

	try {
		Servant* servant = getServant(servantID);
		return servant->isPausable();

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

}

void Database::setConfigServantNames(vector<std::string>& driverNames, vector<std::string>& controllerNames) {
	for (vector<std::string>::iterator it = driverNames.begin(); it < driverNames.end(); it++) {
		_driverConfigNames.push_back(*it);
	}
	for (vector<std::string>::iterator it = controllerNames.begin(); it < controllerNames.end(); it++) {
		_controllerConfigNames.push_back(*it);
	}
}

bool Database::PropertyInfoComparator::operator() (const PropertyInfoKey & a, const PropertyInfoKey & b) const {

	string as = a.controllerType + "." + a.propertyName;
	string bs = b.controllerType + "." + b.propertyName;

	return strcmp(as.c_str(), bs.c_str()) < 0;
}

bool Database::isScannableProperty(const std::string & controllerType, const std::string & wantedPropertyName) const {

	try {
		PropertyInfoItem item = getSetpointPropertyInfoItem(controllerType, wantedPropertyName);
		return item.isScan;

	} catch (const NoPropertyInfo & e) {
		return false;
	}
}

Database::PropertyInfoItem Database::getSetpointPropertyInfoItem(const std::string & servantType, const std::string & wantedPropertyName) const {

	if (_controllerTypeInheritance.find(servantType)) {
		deque<string> types = _controllerTypeInheritance.get(servantType);

		if (types.front() != servantType) {
			types.push_front(servantType);
		}

		for (deque<string>::const_iterator t = types.begin(); t != types.end(); ++t) {

			const string& inheritedType = *t;

			map<PropertyInfoKey, PropertyInfoItem, PropertyInfoComparator>::const_iterator p = _setpointPropertyInfoMap.find(PropertyInfoKey(inheritedType, wantedPropertyName));

			if (p != _setpointPropertyInfoMap.end()) {
				return p->second;
			}
		}

		throw NoPropertyInfo(servantType, wantedPropertyName);

	} else {

		map<PropertyInfoKey, PropertyInfoItem, PropertyInfoComparator>::const_iterator p = _setpointPropertyInfoMap.find(PropertyInfoKey(servantType, wantedPropertyName));

		if (p != _setpointPropertyInfoMap.end()) {
			return p->second;
		}

		throw NoPropertyInfo(servantType, wantedPropertyName);
	}
}

Database::PropertyInfoItem Database::getNoSetpointPropertyInfoItem(const std::string & servantType, const std::string & propertyName) const {

	if (_controllerTypeInheritance.find(servantType)) {
		deque<string> types = _controllerTypeInheritance.get(servantType);

		if (types.front() != servantType) {
			types.push_front(servantType);
		}

		for (deque<string>::const_iterator t = types.begin(); t != types.end(); ++t) {

			const string& inheritedType = *t;

			map<PropertyInfoKey, PropertyInfoItem, PropertyInfoComparator>::const_iterator p = _noSetpointPropertyInfoMap.find(PropertyInfoKey(inheritedType, propertyName));

			if (p != _noSetpointPropertyInfoMap.end()) {
				return p->second;
			}
		}

		throw NoPropertyInfo(servantType, propertyName);

	} else {

		map<PropertyInfoKey, PropertyInfoItem, PropertyInfoComparator>::const_iterator p = _noSetpointPropertyInfoMap.find(PropertyInfoKey(servantType, propertyName));

		if (p != _noSetpointPropertyInfoMap.end()) {
			return p->second;
		}

		throw NoPropertyInfo(servantType, propertyName);
	}
}

std::vector<Database::PropertyInfoItem> Database::getPropertyInfoItems(const std::string & servantType) const {

	vector<PropertyInfoItem> result;

	// Iterate on setpoint properties.
	for (auto p = _setpointPropertyInfoMap.begin(); p != _setpointPropertyInfoMap.end(); p++) {

		if (p->first.controllerType == servantType) {
			result.push_back(p->second);
		}
	}

	// Iterate on no setpoint properties.
	for (auto p = _noSetpointPropertyInfoMap.begin(); p != _noSetpointPropertyInfoMap.end(); p++) {

		if (p->first.controllerType == servantType) {
			result.push_back(p->second);
		}
	}

	return result;
}

std::vector<Database::PropertyInfoItem> Database::getAllPropertyInfoItems(const std::string & servantType) const {

	vector<PropertyInfoItem> result;

	if (_controllerTypeInheritance.find(servantType)) {
		deque<string> types = _controllerTypeInheritance.get(servantType);

		if (types.front() != servantType) {
			types.push_front(servantType);
		}

		for (deque<string>::const_iterator t = types.begin(); t != types.end(); ++t) {

			const string& inheritedType = *t;

			// Iterate on setpoint properties.
			for (auto p = _setpointPropertyInfoMap.begin(); p != _setpointPropertyInfoMap.end(); p++) {

				if (p->first.controllerType == inheritedType) {
					result.push_back(p->second);
				}
			}

			// Iterate on no setpoint properties.
			for (auto p = _noSetpointPropertyInfoMap.begin(); p != _noSetpointPropertyInfoMap.end(); p++) {

				if (p->first.controllerType == inheritedType) {
					result.push_back(p->second);
				}
			}
		}
	}

	return result;
}

void Database::addPropertyInfoItem(const std::string & servantType, const std::string & setpointPropertyName, const std::string & actualPropertyName, const std::string & offsetPropertyName, const std::string & aliasPropertyName, bool isScan, bool isSpy, bool isDynamic) {

	// We separate between setpoint properties and no setpoint properties.
	if (!setpointPropertyName.empty()) {

		PropertyInfoKey key(servantType, setpointPropertyName);
		if (_setpointPropertyInfoMap.find(key) == _setpointPropertyInfoMap.end()) {
			_setpointPropertyInfoMap[key] = PropertyInfoItem(actualPropertyName, setpointPropertyName, offsetPropertyName, aliasPropertyName, isScan, isSpy, isDynamic);
		}

	} else {

		PropertyInfoKey key(servantType, actualPropertyName);
		if (_noSetpointPropertyInfoMap.find(key) == _noSetpointPropertyInfoMap.end()) {
			_noSetpointPropertyInfoMap[key] = PropertyInfoItem(actualPropertyName, setpointPropertyName, offsetPropertyName, aliasPropertyName, isScan, isSpy, isDynamic);
		}

	}
}

int32 Database::getNumberOfScannableProperties(const std::string & controllerType) const {

	int32 result = 0;

	for (map<PropertyInfoKey, PropertyInfoItem, PropertyInfoComparator>::const_iterator p = _setpointPropertyInfoMap.begin(); p != _setpointPropertyInfoMap.end(); p++) {

		if (p->first.controllerType == controllerType && p->second.isScan) {
			result++;
		}
	}

	return result;
}

void Database::addPropertyInfoString(const std::string& servantType, const Database::PropertyInfoItem& item, std::string& output, bool dynamic) const {

	// Only add a property that has an alias.
	if (item.aliasPropertyName.empty()) {
		return;
	}

	// Select the dynamic properties or not depending on the dynamic value.
	if (item.isDynamic != dynamic) {
		return;
	}

	// Set dynamic.
	string dynamicString = "false";
	if (item.isDynamic) {
		dynamicString = "true";
	}

	// Set scannable.
	string scannable = "false";
	if (item.isScan) {
		scannable = "true";
	}

	// Create the XML tags.
	output += string("\t<property_info controller_type='") + servantType + "' alias='" + item.aliasPropertyName + "' scannable='" + scannable + "' dynamic='" + dynamicString + "'>\n";
	output += string("\t\t<actual_property name='") + item.actualPropertyName + "'/>\n";
	output += string("\t\t<setpoint_property name='") + item.setpointPropertyName + "'/>\n";
	output += string("\t\t<offset_property name='") + item.offsetPropertyName + "'/>\n";
	output += string("\t</property_info>\n");
}


std::string Database::getWholeScannableProperties() const {

	string result("<properties>\n");

	// iterate the controller type inheritance
	for (HashMap<string, deque<string> >::const_iterator i = _controllerTypeInheritance.begin(); i != _controllerTypeInheritance.end(); ++i) {

		string servantType = i->first;
		deque<string> const & types = i->second;

		// test if servantType is not types (seems to happen only for legacy controllers)
		if (types.front() != servantType) {

			// Get the setpoints.
			vector<PropertyInfoItem> items = getPropertyInfoItems(servantType);

			for (auto p = items.begin(); p != items.end(); ++p) {
				addPropertyInfoString(servantType, *p, result, false);
			}

			for (auto p = items.begin(); p != items.end(); ++p) {
				addPropertyInfoString(servantType, *p, result, false);
			}
		}

		for (deque<string>::const_iterator t = types.begin(); t != types.end(); ++t) {

			vector<PropertyInfoItem> items = getPropertyInfoItems(*t);

			for (auto p = items.begin(); p != items.end(); ++p) {
				addPropertyInfoString(servantType, *p, result, false);
			}
		}
	}

	result += "</properties>";

	return result;
}

std::string Database::getWholeDynamicProperties() const {

	string result("<properties>\n");

	// iterate the controller type inheritance
	for (HashMap<string, deque<string> >::const_iterator i = _controllerTypeInheritance.begin(); i != _controllerTypeInheritance.end(); ++i) {

		string servantType = i->first;
		deque<string> const & types = i->second;

		// test if servantType is not types (seems to happen only for legacy controllers)
		if (types.front() != servantType) {

			// Get the setpoints.
			vector<PropertyInfoItem> items = getPropertyInfoItems(servantType);

			for (auto p = items.begin(); p != items.end(); ++p) {
				addPropertyInfoString(servantType, *p, result, true);
			}

			for (auto p = items.begin(); p != items.end(); ++p) {
				addPropertyInfoString(servantType, *p, result, true);
			}
		}

		for (deque<string>::const_iterator t = types.begin(); t != types.end(); ++t) {

			vector<PropertyInfoItem> items = getPropertyInfoItems(*t);

			for (auto p = items.begin(); p != items.end(); ++p) {
				addPropertyInfoString(servantType, *p, result, true);
			}
		}
	}

	result += "</properties>";

	return result;
}

void Database::setConstantPropertyUnit(unsigned long propertyID, const std::string & value) {

	unsigned long unitPropertyID = 0;
	unsigned long servantID = getServantIDFromPropertyID(propertyID);

	ScopedLock servantLock(_lockManager->lockServantForWriting(servantID, eol));
	PropertyHolder* propertyHolder = 0;

	try {
		propertyHolder = getPropertyHolder(servantID);

	} catch (Exception const & e) {
		throw link(NoSuchServantException(servantID), e);
	}

	// we need to have a property
	// first we test if it exists
	if (_unitIDPropertyMap.find(propertyID)) {
		unitPropertyID = _unitIDPropertyMap.get(propertyID);

	} else {
		// generate a property name
		string propertyName = "unit";
		propertyName += lexical_cast<string>(propertyID);

		// create the property
		BaseProperty* property = new ElementaryProperty<string>("unit", true, false);

		unitPropertyID = IndexGenerator::getInstance()->getNextIndex();
		property->setID(unitPropertyID);

		propertyHolder->addProperty(propertyName, property);
		addProperty(property, servantID);

		_unitIDPropertyMap.insert(propertyID, unitPropertyID);
	}

	// finally set the value
	setIcsValue(unitPropertyID, value);
}

void Database::setPropertyUnit(unsigned long propertyID, unsigned long unitPropertyID) {

	_unitIDPropertyMap.insert(propertyID, unitPropertyID);
}

void Database::setEnumeratedValuesProperty(unsigned long propertyID, unsigned long valuesPropertyID) {

	// sets the property
	ScopedLock lock(_lockManager->lockPropertyForWriting(propertyID, common::eol));

	BaseProperty * property = getPropertyFromKeyID(propertyID);

	if (property == 0) {
		throw BadPropertyTypeException(propertyID, "?", "?");
	}

	property->setEnumeratedValuesPropertyID(valuesPropertyID);
}

void Database::setEnumeratedLabelsProperty(unsigned long propertyID, unsigned long labelsPropertyID) {

	// sets the property
	ScopedLock lock(_lockManager->lockPropertyForWriting(propertyID, common::eol));

	BaseProperty * property = getPropertyFromKeyID(propertyID);

	if (property == 0) {
		throw BadPropertyTypeException(propertyID, "?", "?");
	}

	property->setEnumeratedLabelsPropertyID(labelsPropertyID);
}

unsigned long Database::getEnumeratedValuesProperty(unsigned long propertyID) {

	// Gets the property
	ScopedLock lock(_lockManager->lockPropertyForReading(propertyID, common::eol));

	BaseProperty * property = getPropertyFromKeyID(propertyID);

	if (property == 0) {
		throw BadPropertyTypeException(propertyID, "?", "?");
	}

	return property->getEnumeratedValuesPropertyID();
}

unsigned long Database::getEnumeratedLabelsProperty(unsigned long propertyID) {

	// Gets the property
	ScopedLock lock(_lockManager->lockPropertyForReading(propertyID, common::eol));

	BaseProperty * property = getPropertyFromKeyID(propertyID);

	if (property == 0) {
		throw BadPropertyTypeException(propertyID, "?", "?");
	}

	return property->getEnumeratedLabelsPropertyID();
}

bool Database::isPropertySavedInCommandBox(unsigned long propertyID) {

	// Gets the property
	ScopedLock lock(_lockManager->lockPropertyForReading(propertyID, common::eol));

	BaseProperty * property = getPropertyFromKeyID(propertyID);

	if (property == 0) {
		throw BadPropertyTypeException(propertyID, "?", "?");
	}

	return property->isSaveBox();
}


vector<unsigned long> Database::getPropertiesSavedInBox() {

	vector<unsigned long> propertyIDs = _keyIDPropertyMap.getKeySet();
	vector<unsigned long> propertyIDsToSave;


	for (unsigned long propertyID : propertyIDs) {
		bool saved = isPropertySavedInCommandBox(propertyID);
		if (saved) {
			propertyIDsToSave.push_back(propertyID);
		}
	}
	return propertyIDsToSave;
}

}
