/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_DATABASE_H_
#define DATAPROVIDER_DATABASE_H_

#include "IndexGenerator.h"
#include "NoLockManager.h"
#include "SingleLockManager.h"
#include "ServantOperations.h"
#include "LockableHashMap.h"

#include <vector>
#include <set>
#include <string>
#include <common/base/HashMap.h>
#include <common/base/Exception.h>
#include <common/base/CommonTypes.h>
#include <dataprovider/dataelement/Command.h>


namespace common {
template<typename ValueType> class ArrayTransferService;
}

namespace dataprovider {

/**
 * Forward declarations.
 */
class Command;
class BaseProperty;
class DataProperty;
class BaseDynamicProperty;
class PropertyContentElement;
class Driver;
class Controller;
class PropertyHolder;
template<typename Type> class PropertyContent;
template<typename ValueType> class ArrayCopier;
class LockManager;
class DataChangeSubscriber;
class Lock;

/**
 * Class defining the database. Stores all the DataElement objects.
 * Implemented by a singleton class. To secure the access, the data that are
 * returned are unshared data. They are copied from the shared data. That
 * is why no reference is returned. For instance getStringProperty returns a string
 * and not a reference to a string. The LockManager is used internally to
 * sequentialise the access.
 */
class Database {

	friend class BaseAccessor;
	friend class DatabaseLoader;
	friend class ServantOperations;
	friend class Servant;

public:
	/**
	 * Exception for servant.
	 */
	class NoSuchServantException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchServantException(unsigned long servantID);

		/**
		 * Constructor.
		 */
		NoSuchServantException(unsigned long parentID, const std::string& servantName, long childIndex);

		/**
		 * Constructor.
		 */
		NoSuchServantException(const std::string& servantName);
	};

	/**
	 * Exception for servant disabled.
	 */
	class ServantDisabledException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		ServantDisabledException(unsigned long parentID, const std::string& servantName, long childIndex);
	};

	/**
	 * Exception for servant enable.
	 */
	class ServantEnableException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		ServantEnableException(const std::string& servantName, unsigned long servantID);

		/**
		 * Constructor.
		 */
		ServantEnableException(const std::string& parentName, unsigned long parentID, const std::string& servantName, long childIndex);
	};

	/**
	 * Exception for circular servant.
	 */
	class ServantAssignmentException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		ServantAssignmentException(unsigned long parentID, const std::string& servantName, long childIndex);
	};

	/**
	 * Exception for property.
	 */
	class NoSuchPropertyException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchPropertyException(unsigned long propertyID);

		/**
		 * Constructor.
		 */
		NoSuchPropertyException(unsigned long servantID, std::string const & propertyName);
	};

	/**
	 * Exception for property.
	 */
	class NoSuchPropertyNameInControllerException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchPropertyNameInControllerException(unsigned long controllerID, std::string const & propertyName);
	};

	/**
	 * Exception for property.
	 */
	class NoSuchPropertyNameInDriverException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchPropertyNameInDriverException(unsigned long driverID, std::string const & propertyName);
	};

	/**
	 * Exception for command.
	 */
	class NoSuchCommandException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchCommandException(unsigned long commandID);
	};

	/**
	 * Exception for not resizable.
	 */
	class NotResizableException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NotResizableException(unsigned long propertyID);
	};

	/**
	 * Exception for property type.
	 */
	class BadPropertyTypeException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		BadPropertyTypeException(unsigned long propertyID, std::string const & type, std::string const & realType);
		BadPropertyTypeException(std::string const & propertyName, std::string const & type, std::string const & realType);
	};

	/**
	 * Exception for null reference.
	 */
	class NullPropertyReferenceException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NullPropertyReferenceException();
	};

	/**
	 * Initialisation exception.
	 */
	class InitialisationException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		InitialisationException();
	};

	/**
	 * Index out of bounds exception.
	 */
	class IndexOutOfBoundsException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		IndexOutOfBoundsException(unsigned long propertyID);

		/**
		 * Constructor.
		 */
		IndexOutOfBoundsException(unsigned long parentID, const std::string& nName, long size);

		/**
		 * Gets the size.
		 */
		long getSize() const;

	private:
		long _size;
	};

	/**
	 * Not configured exception.
	 */
	class NotConfiguredException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NotConfiguredException();
	};

	/**
	 * Index out of bounds exception.
	 */
	class ArrayTransferFailureException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		ArrayTransferFailureException(unsigned long propertyID);
	};

	/**
	 * Bad servant type exception.
	 */
	class BadServantTypeException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		BadServantTypeException(unsigned long propertyID);
	};

	/**
	 * File not found exception.
	 */
	class FileNotFoundException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		FileNotFoundException(const std::string& fileName);
	};

	/**
	 * No such configuration exception.
	 */
	class NoSuchConfigurationException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchConfigurationException(const std::string& servantName, const std::string& configurationName);
	};

	/**
	 * Class defining a scannable property exception.
	 */
	class NoPropertyInfo : public common::Exception {

	public :
		/**
		 * Constructor.
		 */
		NoPropertyInfo(const std::string & controllerType, const std::string & wantedPropertyName);
	};

	/**
	 * Class used for returning servant information.
	 */
	struct ServantItem {
		unsigned long ID;
		std::string name;
		std::string type;
		std::string clientType;
		std::string family;
		std::string subfamily;
		bool isCommand;
		bool isVisible;
		bool isEnabled;
		bool isFavourite;
		bool isGenerated;
		bool isRoot;
	};


	/**
	 * Class used in the property map.
	 */
	struct PropertyItem {

		PropertyItem() :
			itemType(ELEMENTARY),
			isSpy(false),
			undefined(false),
			enumeratedValuesID(0),
			enumeratedLabelsID(0) {
		}

		typedef enum {ELEMENTARY, DYNAMIC} Type;

		Type itemType;
		std::vector<unsigned long> IDs;
		std::string propertyType;
		bool isSpy;
		bool undefined;
		unsigned long enumeratedValuesID;
		unsigned long enumeratedLabelsID;
	};

	/**
	 * Class used in the scannable property map.
	 */
	struct PropertyInfoItem {
		std::string actualPropertyName;
		std::string setpointPropertyName;
		std::string offsetPropertyName;
		std::string aliasPropertyName;
		bool isScan;
		bool isSpy;
		bool isDynamic;

		PropertyInfoItem() : isScan(false), isSpy(false), isDynamic(false) {}
		PropertyInfoItem(const std::string & actualName, const std::string & setpointName, const std::string & offsetName, const std::string & aliasName, bool scan, bool spy, bool dynamic)
		: actualPropertyName(actualName), setpointPropertyName(setpointName), offsetPropertyName(offsetName), aliasPropertyName(aliasName), isScan(scan), isSpy(spy), isDynamic(dynamic) {}
	};

	/**
	 * Typedef for property map.
	 */
	typedef common::HashMap<std::string, PropertyItem> PropertyMap;

	/**
	 * Typedef for command map.
	 */
	typedef common::HashMap<std::string, unsigned long> CommandMap;

	/**
	 * Typedef for child servant map. first = n_name, second = name.
	 */
	struct ChildServant {
		bool isList;
		long size;
		std::vector<std::pair<long, std::string> > children;
	};

	typedef common::HashMap<std::string, ChildServant> ChildServantMap;

	/**
	 * Initialises the instance. Reads the XML configuration file to fill the database.
	 */
	static void initInstance(std::string const & configPath, bool clean);

	/**
	 * Resets the instance.
	 */
	static void resetInstance(unsigned long ID = 0);

	/**
	 * Resets the instances.
	 */
	static void resetInstances();

	/**
	 * Gets the instance. The instance must already be initialised by a call to initInstance().
	 */
	static Database * getInstance(unsigned long ID = 0);


	/**
	 * Clones the instance. Returns the index of the cloned instance.
	 */
	unsigned long clone();

	/**
	 * Sets the ID.
	 */
	void setID(unsigned long ID);

	/**
	 * Gets the ID.
	 */
	unsigned long getID() const;

	// Add servants ////////////////////////////////////////////////////////////////////////////

	/**
	 * Adds a Controller and stores it. The controller is cloned. The controller object must be destroyed.
	 */
	void addController(Controller * controller);

	/**
	 * Removes the controller of ID controllerID.
	 * throw NoSuchServantException
	 */
	void removeController(unsigned long controllerID);

	/**
	 * Renames the controller of ID controllerID.
	 * throw NoSuchServantException
	 */
	void renameController(unsigned long controllerID, const std::string& newName);

	/**
	 * Adds a Driver and stores it. The driver is cloned. The driver object must be destroyed.
	 */
	void addDriver(Driver * driver);

	// Create/add servant ///////////////////////////////////////////////////////////////////////

	/**
	 * Creates a generated Controller from a name and sets it as servant to the controller parent. Returns the new ID.
	 */
	unsigned long createGeneratedController(const std::string& controllerType, unsigned long parentID, const std::string& n_name, long childIndex);

	/**
	 * Creates a Controller from a name and type. Returns the new ID.
	 */
	unsigned long createController(const std::string& controllerName, const std::string& controllerType, const std::string& family, bool isCommand, const common::HashMap<std::string,  std::pair<std::string , std::string>>& childrenMap);

	/**
	 * Creates a Driver from a name and sets it as servant to the controller parent. Returns the new ID.
	 */
	unsigned long createDriver(const std::string& driverName, const std::string& driverType, const std::string& protocolType, common::DriverMode mode, unsigned long parentID, const std::string& n_name, long childIndex);

	/**
	 * Clones a Controller from the controllerID. Returns the new ID.
	 */
	unsigned long cloneController(unsigned long controllerID);

	/**
	 * Adds the servant to the parent
	 * throw NoSuchServantException
	 */
	void addServantToParent(unsigned long parentID, const std::string& nName, bool isList);

	/**
	 * Removes a given servant from the list of servants of a parent controller.
	 * Returns the ID of the controller if it is removed from the database.
	 * throw NoSuchServantException, NoSuchChildException
	 */
	unsigned long removeServantFromParent(unsigned long parentID, const std::string& nName, long index);

	/**
	 * Removes a servant
	 * throw NoSuchServantException
	 */
	void removeServant(unsigned long servantID);

	/**
	 * Adds an existing servant.
	 * throw ServantAssignmentException
	 */
	void assignServant(unsigned long parentID, const std::string& nName, long childIndex, unsigned long newParentID, const std::string& newNName, long newChildIndex);

	/**
	 * Adds an existing servant.
	 * throw ServantAssignmentException
	 */
	void assignServant(unsigned long parentID, const std::string& nName, long childIndex, unsigned long newServantID);

	/**
	 * Adds an existing servant.
	 * throw ServantAssignmentException
	 */
	void assignServant(unsigned long servantID, bool newServant, bool localServant);

	// Add properties //////////////////////////////////////////////////////////////////////////

	/**
	 * Adds a property
	 */
	void addProperty(BaseProperty* property, unsigned long servantID);

	/**
	 * Removes a property
	 */
	void removeProperty(unsigned long propertyID);

	/**
	 * Adds a property to a servant if it isn't already registered in the data provider
	 */
	unsigned long addPropertyToServant(unsigned long servantID, const std::string& propertyName, const std::string& propertyType, bool spy, bool saveBox, bool & added);

	/**
	 * Adds a property to a servant if it isn't already registered in the data provider
	 */
	unsigned long addDynamicPropertyToServant(unsigned long servantID, const std::string& propertyName, const std::string& propertyType, bool spy);

	/**
	 * Removes the property from the servant.
	 */
	void removePropertyFromServant(unsigned long servantID, unsigned long propertyID);

	/**
	 * Sets the property settings.
	 */
	void setPropertySettings(unsigned long propertyID, const std::string& propertyFormat, bool save, bool spy);

	/**
	 * Resizes the dynamic property.
	 * throw NoSuchPropertyException
	 */
	void resizeDynamicProperty(unsigned long propertyID, long size);

	/**
	 * Gets the dynamic property size.
	 * throw NoSuchPropertyException
	 */
	long getDynamicPropertySize(unsigned long propertyID);

	/**
	 * Gets the dynamic property index. The resize attribute indicates if the container is resized if necessary.
	 * throw NoSuchPropertyException, IndexOutOfBoundsException
	 */
	unsigned long getDynamicPropertyIDAtIndex(unsigned long propertyID, long index, bool resize);

	/**
	 * Gets the dynamic property index. Returns -1 if property does not belong to this parent.
	 */
	long getIndexOfDynamicPropertyID(unsigned long parentPropertyID, unsigned long propertyID);

	/**
	 * Gets the list of Id of a dynamic property.
	 */
	std::vector<unsigned long> getDynamicPropertyIDs(unsigned long parentPropertyID);

	// add command ////////////////////////////////////////////////////////////////////////////////

	/**
	 * Adds a command.
	 */
	void addCommand(Command* command);

	/**
	 * Adds a command to a servant if it isn't already registered in the data provider
	 */
	unsigned long addCommandToServant(unsigned long servantID, const std::string& commandName);

	// Get properties and commands ////////////////////////////////////////////////////////////////

	/**
	 * Gets the properties of controller controllerID.
	 * throw NoSuchServantException
	 */
	PropertyMap getControllerProperties(unsigned long controllerID);

	/**
	 * Gets the properties of driver driverID.
	 * throw NoSuchServantException
	 */
	PropertyMap getDriverProperties(unsigned long driverID);

	/**
	 * Gets the commands of controller controllerID.
	 * throw NoSuchServantException
	 */
	CommandMap getControllerCommands(unsigned long controllerID);

	/**
	 * Gets the commands of driver driverID.
	 * throw NoSuchServantException
	 */
	CommandMap getDriverCommands(unsigned long driverID);

	/**
	 * Gets the child servants of controller controllerID.
	 * throw NoSuchServantException
	 */
	ChildServantMap getControllerChildServants(unsigned long controllerID);

	/**
	 * Gets the child servants of driver driverID.
	 * throw NoSuchServantException
	 */
	ChildServantMap getDriverChildServants(unsigned long driverID);

	// Set and get values /////////////////////////////////////////////////////////////////////////

	/**
	 * Sets the value.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	template<typename ValueType>
	bool setIcsValue(unsigned long propertyID, ValueType value);

	/**
	 * Sets the value and the formatted value.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	template<typename ValueType>
	bool setIcsValueAndFormattedValue(unsigned long propertyID, ValueType value, const std::string& formattedValue);

	/**
	 * Sets the ics array value.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	template<typename ValueType>
	void setIcsArrayValue(unsigned long propertyID, ValueType* value);

	/**
	 * Sets the ics array size.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	template<typename ValueType>
	void setIcsArraySize(unsigned long propertyID, long size);

	/**
	 * Gets the value.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	template<typename ValueType>
	ValueType getValue(unsigned long propertyID);

	/**
	 * Gets the ics array value.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	template<typename ValueType>
	ValueType* getIcsArrayValue(unsigned long propertyID);

	/**
	 * Gets the ics array value.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	template<typename ValueType>
	void getIcsArrayValue(unsigned long propertyID, std::vector<ValueType> & array);

	/**
	 * Gets the ics array size.
	 * throw NoSuchPropertyException, BadPropertyTypeException
	 */
	template<typename ValueType>
	long getIcsArraySize(unsigned long propertyID);

	/**
	 * Gets the encoded array.
	 */
	template<typename ValueType>
	void getEncodedArrayValue(unsigned long propertyID, std::string & encodedString);

	/**
	 * Gets the formatted value.
	 * throw NoSuchPropertyException
	 */
	std::string getFormattedValue(unsigned long propertyID);

	/**
	 * Notifies the update of the ics property.
	 */
	void updateIcsProperty(unsigned long propertyID);

	/**
	 * Get the array
	 */
	template<typename ValueType>
	void getArray(unsigned long propertyID, common::ArrayTransferService<ValueType>* arrayTransferService);

	/**
	 * Set the array
	 */
	template<typename ValueType>
	void setArray(unsigned long propertyID, ArrayCopier<ValueType>* arrayCopier);

	/**
	 * Begins the transfer of an array. Used for string arrays.
	 */
	Lock* beginArrayTransfer(unsigned long propertyID);

	// Enable, disable, configure servant

	/**
	 * Enables the servant.
	 * throw NoSuchServantException, ServantEnableException
	 */
	void enableServant(unsigned long servantID);

	/**
	 * Enables the servant.
	 * throw NoSuchServantException, ServantEnableException
	 */
	void enableServant(unsigned long parentID, const std::string& nName, long childIndex);

	/**
	 * Enables only the servant.
	 * throw NoSuchServantException, ServantEnableException
	 */
	void enableOnlyServant(unsigned long servantID);

	/**
	 * Disables the servant.
	 * throw NoSuchServantException, ServantDisableException
	 */
	void disableServant(unsigned long servantID);

	/**
	 * Disables the servant.
	 * throw NoSuchServantException, ServantDisableException
	 */
	void disableServant(unsigned long parentID, const std::string& nName, long childIndex);

	/**
	 * Disables only the servant.
	 * throw NoSuchServantException, ServantDisableException
	 */
	void disableOnlyServant(unsigned long servantID);

	/**
	 * Returns true if the servant is enabled and initialised.
	 * throw NoSuchServantException
	 */
	bool isServantEnabled(unsigned long servantID);

	/**
	 * Returns true if the servant is enabled and initialised.
	 * throw NoSuchServantException
	 */
	bool isServantEnabled(unsigned long parentID, const std::string& nName, long childIndex);

	/**
	 * Sets the servant visible.
	 * throw NoSuchServantException
	 */
	void setServantVisible(unsigned long servantID, bool value);

	/**
	 * Sets the servant favourite.
	 * throw NoSuchServantException
	 */
	void setServantFavourite(unsigned long servantID, bool value);

	/**
	 * Sets the servant command true or false.
	 */
	void setServantCommand(unsigned long servantID, bool value);

	// Get servants ///////////////////////////////////////////////////////////////////////////////

	/**
	 * Gets the controllers id, name and type.
	 */
	std::vector<ServantItem> getControllers();

	/**
	 * Gets the controller.
	 */
	ServantItem getController(unsigned long servantID);

	/**
	 * Get the controller instance from its type
	 */

	Controller* getController(const std::string& controllerType);

	/**
	 * Gets the drivers id, name and type.
	 */
	std::vector<ServantItem> getDrivers();

	/**
	 * Gets the drivers id
	 */
	ServantItem getDriver(unsigned long servantID);

/**
	 * Gets the servantID from a commandID.
	 * throw NoSuchCommandException
	 */
	unsigned long getServantID(unsigned long commandID);

	/**
	 * Gets the command name from the command ID.
	 * throw NoSuchCommandException
	 */
	std::string getCommandName(unsigned long commandID);

	/**
	 * Gets the servant name from the servant ID.
	 * throw NoSuchServantException
	 */
	std::string getServantName(unsigned long servantID);

	/**
	 * Gets the servant ID from the servant name.
	 * throw NoSuchServantException
	 */
	unsigned long getServantID(const std::string& servantName);

	/**
	 * Gets the size of the servant list.
	 */
	long getServantListSize(unsigned long servantID, const std::string& nServantName);

	/**
	 * Resizes the servant list.
	 */
	void resizeServantList(unsigned long servantID, const std::string& nServantName, long size);

	/**
	 * Gets the controller type from the controller ID.
	 * throw NoSuchServantException
	 */
	std::string getControllerType(unsigned long controllerID);

	/**
	 * Gets the driver type from the driver ID.
	 * throw NoSuchServantException
	 */
	std::string getDriverType(unsigned long driverID);

	/**
	 * Gets the servant type from the servant ID.
	 * throw NoSuchServantException
	 */
	std::string getServantType(unsigned long servantID);

	/**
	 * Gets the short servant type from the servant ID.
	 * throw NoSuchServantException
	 */
	std::string getShortServantType(unsigned long servantID);

	/**
	 * Gets the controller types from the controller ID.
	 * throw NoSuchServantException
	 */
	std::deque<std::string> getControllerTypes(unsigned long controllerID);

	/**
	 * Gets the controller types inheriting the type.
	 */
	std::set<std::string> getControllerTypesInheritingType(const std::string& type);

	/**
	 * Sets the servant family.
	 */
	void setServantFamily(unsigned long servantID, const std::string& familyName, const std::string& subfamilyName);

	/**
	 * Gets the servant family.
	 */
	std::string getServantFamily(unsigned long servantID);

	/**
	 * Gets the servant subfamily.
	 */
	std::string getServantSubfamily(unsigned long servantID);

	/**
	 * Returns true if the servant is favourite.
	 */
	bool isServantFavourite(unsigned long servantID);

	/**
	 * Adds the servant type.
	 */
	void addServantType(unsigned long servantID, const std::string& type, bool atFront);

	// Command states /////////////////////////////////////////////////////////////////////////////

	/**
	 * Sets the command state
	 * throw NoSuchCommandException
	 */
	void setCommandState(unsigned long servantID, unsigned long commandID, common::CommandState state);

	/**
	 * Sets the command active
	 * throw NoSuchCommandException
	 */
	common::SetCommandActiveResult setCommandActive(unsigned long servantID, unsigned long commandID, const std::string servantType, unsigned long processorID, unsigned long threadID, DataChangeSubscriber* subscriber);

	/**
	 * Gets the command state
	 * throw NoSuchCommandException
	 */
	common::CommandState getCommandState(unsigned long commandID);

	/**
	 * Gets the elapsed time of the command in milliseconds.
	 * throw NoSuchCommandException
	 */
	double getCommandElapsedTimeMs(unsigned long commandID);

	/**
	 * Resets the elapsed time of the driver.
	 * throw NoSuchServantException
	 */
	void resetElapsedSimulationTime(unsigned long driverID);

	/**
	 * Resets the elapsed time of the driver.
	 * throw NoSuchServantException
	 */
	double getElapsedSimulationTimeMs(unsigned long driverID);

	/**
	 * Sets the progression.
	 */
	void setProgression(unsigned long commandID, double progression);

	/**
	 * Sets the progression unknown.
	 */
	void setUnknownProgression(unsigned long commandID);

	/**
	 * Gets the progression.
	 */
	double getProgression(unsigned long commandID);

	/**
	 * Gets the remaining time of a command.
	 */
	double getCommandRemainingTimeMs(unsigned long commandID);

	/**
	 * Gets the duration of a command.
	 */
	double getCommandDurationMs(unsigned long commandID);

	/**
	 * Gets the start date of the command.
	 */
	int64 getCommandStartDate(unsigned long commandID);

	/**
	 * Sets the command's pre-run calculated duration.
	 */
	void setCommandPreRunCalculatedDurationMs(unsigned long commandID, double duration);

	/**
	 * Gets the command's pre-run calculated duration.
	 */
	double getCommandPreRunCalculatedDurationMs(unsigned long commandID);

	/**
	 * Sets the command from the command zone.
	 */
	void setCommandFromCommandZone(unsigned long commandID, bool fromCommandZone);

	/**
	 * Returns true if the command is from the command zone.
	 */
	bool isCommandFromCommandZone(unsigned long commandID);

	// queries ////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Returns true if servantID is contained within the HashMap of controllers.
	 */
	bool isController(unsigned long servantID);

	/**
	 * Returns true if servantID is contained within the HashMap of drivers.
	 */
	bool isDriver(unsigned long servantID);

	/**
	 * Returns true if servantID is associated with a servant.
	 */
	bool isServant(unsigned long servantID);

	/**
	 * Returns true if servantID is associated with a ics servant.
	 */
	bool isIcsServant(unsigned long servantID);

	/**
	 * Returns true if is a command.
	 */
	bool isCommand(unsigned long commandID);

	/**
	 * Returns true if is a property.
	 */
	bool isProperty(unsigned long propertyID);

	/**
	 * Returns the type of property.
	 */
	std::string getPropertyType(unsigned long propertyID);

	/**
	 * Sets the mode of execution of a driver.
	 * throw NoSuchServantException
	 */
	void setMode(unsigned long driverID, common::DriverMode mode);

	/**
	 * Returns the mode of execution of a driver.
	 * throw NoSuchServantException
	 */
	common::DriverMode getMode(unsigned long driverID);

	/**
	 * Returns the driver ID for the parent driver of the driver with driverID.
	 * throw NoSuchServantException
	 */
	unsigned long getDriverParentID(unsigned long driverID);

	/**
	 * Returns the ID of the servant enabled for which we have the parentID, the n name and the childIndex.
	 *
	 */
	unsigned long getEnabledOrUniqueServantID(unsigned long parentID, const std::string& nServantName, long childIndex);

	/**
	 * Returns the ID of the servants for which we have the parentID, the n name and the childIndex.
	 *
	 */
	std::vector<unsigned long> getServantIDs(unsigned long parentID, const std::string& nServantName, long childIndex);

	/**
	 * Returns the ID of the servant that contains the property ID.
	 */
	unsigned long getServantIDFromPropertyID(unsigned long propertyID);

	/**
	 * Returns the ID of the property.
	 */
	unsigned long getPropertyID(unsigned long servantID, const std::string& propertyName);

	/**
	 * Returns the name of the property.
	 */
	std::string getPropertyName(unsigned long propertyID);

	/**
	 * Returns the name of the property.
	 */
	std::pair<std::string, long> getCompletePropertyName(unsigned long servantID, unsigned long propertyID);

	/**
	 * Returns the ID of the unit property.
	 */
	unsigned long getUnitPropertyID(unsigned long propertyID);

	/**
	 * Returns the ID of the servant that contains the command ID.
	 */
	unsigned long getServantIDFromCommandID(unsigned long commandID);

	/**
	 * Gets the ID of the command of the servant.
	 */
	unsigned long getCommandID(unsigned long servantID, const std::string& commandName);

	/**
	 * Sets the servant initialised.
	 */
	void setServantInitialised(unsigned long servantID, bool initialised);

	/**
	 * Returns true the servant initialised.
	 */
	bool isServantInitialised(unsigned long servantID);

	/**
	 * Returns true if the servant is enabled and initialised.
	 * throw NoSuchServantException
	 */
	bool isServantInitialised(unsigned long parentID, const std::string& nName, long childIndex);

	/**
	 * Returns the channel for the driver with servantID.
	 * throw NoSuchServantException
	 */
	long getDriverChannel(unsigned long servantID);

	/**
	 * Sets the servant in debug.
	 */
	void setServantInDebug(unsigned long servantID);

	/**
	 * Returns true if servant is in debug.
	 */
	bool isServantInDebug(unsigned long servantID);

	// to string //////////////////////////////////////////////////////////////////////////////////

	/**
	 * Prints the content of the database.
	 */
	std::string toString();

	/**
	 * Prints the content of the servant with ID servantID.
	 */
	std::string toString(unsigned long servantID);

	/**
	 * Gets the controllers to string.
	 */
	std::string getControllersToString(bool save);

	/**
	 * Gets the drivers to string.
	 */
	std::string getDriversToString(bool save);

	/**
	 * Sets the log ID of a command.
	 */
	void setCommandLogID(unsigned long commandID, const std::string& logID);

	/**
	 * Returns the log ID of a command.
	 */
	std::string getCommandLogID(unsigned long commandID);

	/**
	 * Returns the string describing what is the element with ID elementID.
	 */
	std::string what(unsigned long elementID, bool withReference = true);

	/**
	 * Reports a bug at date.
	 */
	void reportBug(int64 date);

	/**
	 * Typedef for a predefined config property.
	 */
	struct PredefinedConfigProperty {
		std::string propertyName;
		std::string propertyValue;
		std::string format;
		std::string index;
		bool dynamic;
	};

	/**
	 * Typedef for a predefined config.
	 */
	struct PredefinedConfig {
		std::string description;
		std::vector<PredefinedConfigProperty> properties;
	};

	/**
	 * Locks the element ID
	 */
	Lock* lock(unsigned long ID);

	/**
	 * Unlocks the element ID
	 */
	void unlock(unsigned long ID);

	/**
	 * Verify the locks
	 */
	void verifyLocks(int seconds);

	/**
	 * Sets the type. Allows developers of N to fake a type, useful for inherited
	 * classes with same view in client
	 */
	void setClientType(unsigned long servantID, const std::string& clientType);

	/**
	 * Returns true if servant is pausable
	 */
	bool isServantPausable(unsigned long servantID);

	/**
	 * Sets the controller and driver names for config file
	 */
	void setConfigServantNames(std::vector<std::string>& driverNames, std::vector<std::string>& controllerNames);

	// Scannable properties ///////////////////////////////////////////////////////////////////////

	/**
	 * Returns true if the property is scannable.
	 */
	bool isScannableProperty(const std::string & controllerType, const std::string & wantedPropertyName) const;

	/**
	 * Gets the setpoint property info item.
	 * The item exists if there is an association between the setpoint and the actual property.
	 * throw NoPropertyInfo
	 */
	PropertyInfoItem getSetpointPropertyInfoItem(const std::string & servantType, const std::string & wantedPropertyName) const;

	/**
	 * Gets the no setpoint property info item.
	 * The item exists if there is no association between the setpoint and the actual property.
	 * throw NoPropertyInfo
	 */
	PropertyInfoItem getNoSetpointPropertyInfoItem(const std::string & servantType, const std::string & propertyName) const;

	/**
	 * Gets the property info items.
	 */
	std::vector<PropertyInfoItem> getPropertyInfoItems(const std::string & servantType) const;

	/**
	 * Gets all the property info items including the inherited ones.
	 */
	std::vector<PropertyInfoItem> getAllPropertyInfoItems(const std::string & servantType) const;

	/**
	 * Adds the property info item.
	 */
	void addPropertyInfoItem(const std::string & servantType, const std::string & setpointPropertyName, const std::string & actualPropertyName, const std::string & offsetPropertyName, const std::string & aliasPropertyName, bool isScan, bool isSpy, bool isDynamic);

	/**
	 * Returns the number of scannable properties for a controller type.
	 */
	int32 getNumberOfScannableProperties(const std::string & controllerType) const;

	/**
	 * Returns the model of all the scannable properties.
	 * The result is an XML string with all possible entries. The controller inheritance is not used but the properties are copied if inherited.
	 */
	std::string getWholeScannableProperties() const;

	/**
	 * Returns the model of all the dynamic properties.
	 * The result is an XML string with all possible entries. The controller inheritance is not used but the properties are copied if inherited.
	 */
	std::string getWholeDynamicProperties() const;

	// Property unit //////////////////////////////////////////////////////////////////////////////

	/**
	 * Sets the constant property unit.
	 */
	void setConstantPropertyUnit(unsigned long propertyID, const std::string & value);

	/**
	 * Sets the property unit.
	 */
	void setPropertyUnit(unsigned long propertyID, unsigned long unitPropertyID);

	// Property enumerated values /////////////////////////////////////////////////////////////////

	/**
	 * Sets the enumerated values property.
	 */
	void setEnumeratedValuesProperty(unsigned long propertyID, unsigned long valuesPropertyID);

	/**
	 * Sets the enumerated labels property.
	 */
	void setEnumeratedLabelsProperty(unsigned long propertyID, unsigned long labelsPropertyID);

	/**
	 * Gets the enumerated values property.
	 */
	unsigned long getEnumeratedValuesProperty(unsigned long propertyID);

	/**
	 * Gets the enumerated labels property.
	 */
	unsigned long getEnumeratedLabelsProperty(unsigned long propertyID);

	/**
	 * Returns true if property is saved in box.
	 */
	bool isPropertySavedInCommandBox(unsigned long propertyID);

	std::vector<unsigned long> getPropertiesSavedInBox();


private:
	static const std::string LOCK_MANAGER_TYPE;
	static const std::string NO_LOCK_MANAGER;
	static const std::string SINGLE_LOCK_MANAGER;
	static const std::string ELEMENT_LOCK_MANAGER;
	static const std::string MAX_TRANSFER_SIZE_BYTES;

	void init(std::string const & configPath, bool clean);
	void reset();
	PropertyMap createPropertyItems(const common::HashMap<std::string, unsigned long>& properties);

	BaseProperty* getPropertyFromKeyID(unsigned long propertyKeyID);
	unsigned long getPropertyKeyID(unsigned long propertyID);
	std::list<unsigned long> getPropertyIDs(unsigned long propertyKeyID);

	Controller* findController(const std::string& name);
	Driver* findDriver(const std::string& name);
	Servant* findServant(const std::string& name);
	PropertyHolder* getPropertyHolder(unsigned long servantID);
	Servant* getServant(unsigned long servantID);
	Command* getCommand(unsigned long commandID);

	void setServantEnabled(const std::set<unsigned long>& servantsToEnable);
	void setServantDisabled(const std::set<unsigned long>& servantsToDisable);

	void addParentToServant(unsigned long parentID, unsigned long servantID);
	void removeParentFromServant(unsigned long parentID, unsigned long servantID);
	void removeCommandsAndPropertiesFromServant(Servant* servant);
	void removeDriver(unsigned long driverID);

	std::set<std::string> getKeys(const common::HashMap<std::string, unsigned long> & map);
	std::string typeToString(const std::string& type);
	std::string IDToString(unsigned long ID, bool save);
	std::string propertyIDToString(unsigned long ID, bool save);
	std::string DataPropertyToString(const std::string& name, DataProperty * property, bool save, const std::string& space);
	std::string DynamicPropertyToString(const std::string& DynamicPropertyName, BaseDynamicProperty * DynamicProperty, bool save, const std::string& space);

	void controllerToString(Controller* controller, bool save, std::string& result);
	void driverToString(Driver* driver, bool save, std::string& result);

	void addPropertyInfoString(const std::string& servantType, const PropertyInfoItem& item, std::string& output, bool dynamic) const;

	/**
	 * Checks the value against the list of enumerated values.
	 */
	template<typename ValueType>
	bool checkValue(unsigned long enumeratedValuesPropertyID, ValueType value);

	Database();
	~Database();

	void printDriverHierarchy(Driver * driver, int channel, std::string const & offset, std::string & output);

	static std::vector<Database *> _instances;
	unsigned long _ID;
	LockManager * _lockManager;

	LockableHashMap<unsigned long, Controller *> _IDControllerMap;
	LockableHashMap<unsigned long, Driver *> _IDDriverMap;
	LockableHashMap<unsigned long, Command *> _IDCommandMap;
	LockableHashMap<unsigned long, BaseProperty *> _keyIDPropertyMap;
	LockableHashMap<unsigned long, std::set<unsigned long> > _parentIDsMap;
	LockableHashMap<unsigned long, unsigned long> _unitIDPropertyMap;
	LockableHashMap<unsigned long, unsigned long> _IDPropertyIDServantMap;

	common::HashMap<std::string, std::deque<std::string> > _driverTypeInheritance;
	common::HashMap<std::string, std::deque<std::string> > _controllerTypeInheritance;

	std::vector<unsigned long> _debugIDServants;

	ServantOperations _servantOperations;

	std::vector<std::string> _controllerConfigNames;
	std::vector<std::string> _driverConfigNames;

	struct PropertyInfoKey {
		PropertyInfoKey(const std::string & type, const std::string & name) :
			controllerType(type), propertyName(name) {
		}

		std::string controllerType;
		std::string propertyName;
	};

	struct PropertyInfoComparator {

		bool operator() (const PropertyInfoKey & a, const PropertyInfoKey & b) const;
	};

	std::map<PropertyInfoKey, PropertyInfoItem, PropertyInfoComparator> _setpointPropertyInfoMap;
	std::map<PropertyInfoKey, PropertyInfoItem, PropertyInfoComparator> _noSetpointPropertyInfoMap;

	template<typename>
	friend struct DatabaseGetter;
};

}

#include "Database.inl"

#endif /*DATAPROVIDER_DATABASE_H_*/
