/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "LockManager.h"
#include <common/base/RealArray.h>
#include <common/base/PropertyType.h>
#include <common/base/TypeToString.h>
#include <common/base/PortableByteArray.h>
#include <dataprovider/dataelement/Command.h>
#include <dataprovider/dataelement/Controller.h>
#include <dataprovider/dataelement/Driver.h>
#include <dataprovider/dataelement/ElementaryProperty.h>
#include <dataprovider/dataelement/DynamicProperty.h>
#include <dataprovider/dataelement/IcsArrayProperty.h>
#include <dataprovider/notification/NotificationServer.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include "ScopedLock.h"

namespace dataprovider {

template<typename ValueType>
bool Database::checkValue(unsigned long enumeratedValuesPropertyID, ValueType value) {

	std::vector<unsigned long> propertyIDs;

	// find the property list from the dynamic property
	{
		ScopedLock lock(_lockManager->lockPropertyForReading(enumeratedValuesPropertyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(enumeratedValuesPropertyID);
		DynamicProperty<ValueType> * dynamicProperty = dynamic_cast<DynamicProperty<ValueType> *>(property);

		if (dynamicProperty == 0) {
			throw BadPropertyTypeException(enumeratedValuesPropertyID, common::TypeToString<ValueType>::convert(), property->getType());
		}

		propertyIDs = dynamicProperty->getPropertyIDs();
	}

	// iterate the property IDs
	for (std::vector<unsigned long>::const_iterator p = propertyIDs.begin(); p != propertyIDs.end(); ++p) {

		unsigned long propertyID = *p;

		BaseProperty * property = getPropertyFromKeyID(propertyID);
		ElementaryProperty<ValueType> * typedProperty = dynamic_cast<ElementaryProperty<ValueType> *>(property);

		if (value == typedProperty->getValue()) {
			return true;
		}
	}

	return false;
}

template<typename ValueType>
bool Database::setIcsValue(unsigned long propertyID, ValueType value) {

	bool result = false;

	try {
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyKeyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(propertyKeyID);
		ElementaryProperty<ValueType> * typedProperty = dynamic_cast<ElementaryProperty<ValueType> *>(property);

		if (typedProperty == 0) {
			throw BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
		}

		if (typedProperty->getEnumeratedValuesPropertyID() != 0) {
			if (!checkValue(typedProperty->getEnumeratedValuesPropertyID(), value)) {
				return false;
			}
		}

		typedProperty->getValue() = value;
		result = true;

	} catch (common::Exception const & e) {
		throw;
	}

	if (result) {
		NotificationServer::getInstance()->icsPropertyChanged(_ID, propertyID);
	}

	return result;
}

template<typename ValueType>
bool Database::setIcsValueAndFormattedValue(unsigned long propertyID, ValueType value, const std::string& formattedValue) {

	bool result = false;

	try {
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyKeyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(propertyKeyID);
		ElementaryProperty<ValueType> * typedProperty = dynamic_cast<ElementaryProperty<ValueType> *>(property);

		if (typedProperty == 0) {
			throw BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
		}

		if (typedProperty->getEnumeratedValuesPropertyID() != 0) {
			if (!checkValue(typedProperty->getEnumeratedValuesPropertyID(), value)) {
				return false;
			}
		}

		typedProperty->getValue() = value;
		result = true;

		// set formatted value, because it can be different
		typedProperty->setFormattedValue(formattedValue);

	} catch (common::Exception const & e) {
		throw;
	}

	if (result) {
		NotificationServer::getInstance()->icsPropertyChanged(_ID, propertyID);
	}

	return result;
}

template<typename ValueType>
void Database::setIcsArrayValue(unsigned long propertyID, ValueType* value) {

	try {
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyKeyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(propertyKeyID);
		IcsArrayProperty<ValueType> * typedProperty = dynamic_cast<IcsArrayProperty<ValueType> *>(property);

		if (typedProperty == 0) {
			throw BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
		}

		typedProperty->setSizeValue(0);
		typedProperty->setPointerValue(value);

		// no notification

	} catch (NoSuchPropertyException const & e) {
		throw;
	} catch (BadPropertyTypeException & e) {
		throw;
	}
}

template<typename ValueType>
void Database::setIcsArraySize(unsigned long propertyID, long size) {

	try {
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyKeyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(propertyKeyID);
		IcsArrayProperty<ValueType> * typedProperty = dynamic_cast<IcsArrayProperty<ValueType> *>(property);

		if (typedProperty == 0) {
			throw BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
		}

		typedProperty->setSizeValue(size);

		// no notification

	} catch (NoSuchPropertyException const & e) {
		throw;
	} catch (BadPropertyTypeException & e) {
		throw;
	}
}

/**
 * Template class to manage special case:
 * A client can request a string value whereas the real type is not string.
 * In that case, we return the converted string value.
 */
template<typename ValueType>
struct DatabaseGetter {

	static ValueType get(Database & db, unsigned long propertyID) {

		try {
			unsigned long propertyKeyID = db.getPropertyKeyID(propertyID);

			ScopedLock lock(db._lockManager->lockPropertyForReading(propertyKeyID, common::eol));

			BaseProperty * property = db.getPropertyFromKeyID(propertyKeyID);

			ElementaryProperty<ValueType> * typedProperty = dynamic_cast<ElementaryProperty<ValueType> *>(property);

			if (typedProperty == 0) {
				throw Database::BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
			}

			return typedProperty->getValue();

		} catch (common::Exception const & e) {
			throw;
		}
	}
};

/**
 * Specialisation of the DatabaseGetter<> class for std::string.
 */
template<>
struct DatabaseGetter<std::string> {

	typedef std::string ValueType;

	static ValueType get(Database & db, unsigned long propertyID) {

		try {
			unsigned long propertyKeyID = db.getPropertyKeyID(propertyID);

			ScopedLock lock(db._lockManager->lockPropertyForReading(propertyKeyID, common::eol));

			BaseProperty * property = db.getPropertyFromKeyID(propertyKeyID);

			// Special case for
			if (property->getType() != common::TypeToString<std::string>::convert()) {
				DataProperty * dataProperty = dynamic_cast<DataProperty *>(property);
				if (dataProperty == 0) {
					throw Database::BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
				}
				bool overflow;

				return dataProperty->getStringValue(overflow);
			}

			ElementaryProperty<ValueType> * typedProperty = dynamic_cast<ElementaryProperty<ValueType> *>(property);

			if (typedProperty == 0) {
				throw Database::BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
			}

			return typedProperty->getValue();

		} catch (common::Exception const & e) {
			throw;
		}
	}
};

template<typename ValueType>
ValueType Database::getValue(unsigned long propertyID) {
	return DatabaseGetter<ValueType>::get(*this, propertyID);
}

template<typename ValueType>
ValueType* Database::getIcsArrayValue(unsigned long propertyID) {

	try {
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyKeyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(propertyKeyID);
		IcsArrayProperty<ValueType> * typedProperty = dynamic_cast<IcsArrayProperty<ValueType> *>(property);

		if (typedProperty == 0) {
			throw BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
		}

		return typedProperty->getPointerValue();

	} catch (NoSuchPropertyException const & e) {
		throw;
	} catch (BadPropertyTypeException & e) {
		throw;
	}
}

template<typename ValueType>
void Database::getIcsArrayValue(unsigned long propertyID, std::vector<ValueType> & array) {

	try {
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyKeyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(propertyKeyID);
		IcsArrayProperty<ValueType> * typedProperty = dynamic_cast<IcsArrayProperty<ValueType> *>(property);

		if (typedProperty == 0) {
			throw BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
		}

		long size = typedProperty->getSizeValue();
		array.reserve(size);

		ValueType * data = typedProperty->getPointerValue();

		for (long i = 0; i < size; ++i) {
			array.push_back(data[i]);
		}

	} catch (NoSuchPropertyException const & e) {
		throw;
	} catch (BadPropertyTypeException & e) {
		throw;
	}
}

template<typename ValueType>
long Database::getIcsArraySize(unsigned long propertyID) {

	try {
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyKeyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(propertyKeyID);
		IcsArrayProperty<ValueType> * typedProperty = dynamic_cast<IcsArrayProperty<ValueType> *>(property);

		if (typedProperty == 0) {
			throw BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
		}

		return typedProperty->getSizeValue();

	} catch (NoSuchPropertyException const & e) {
		throw;
	} catch (BadPropertyTypeException & e) {
		throw;
	}
}

template<typename ValueType>
void Database::getEncodedArrayValue(unsigned long propertyID, std::string & encodedString) {

	try {
		unsigned long propertyKeyID = getPropertyKeyID(propertyID);

		ScopedLock lock(_lockManager->lockPropertyForWriting(propertyKeyID, common::eol));

		BaseProperty * property = getPropertyFromKeyID(propertyKeyID);
		IcsArrayProperty<ValueType> * typedProperty = dynamic_cast<IcsArrayProperty<ValueType> *>(property);

		if (typedProperty == 0) {
			throw BadPropertyTypeException(propertyID, common::TypeToString<ValueType>::convert(), property->getType());
		}

		long size = typedProperty->getSizeValue();
		ValueType * data = typedProperty->getPointerValue();

		common::PortableByteArray<ValueType> portableByteArray;
		portableByteArray.encode(data, size, encodedString, common::LITTLE_ENDIAN_ORDER);

	} catch (NoSuchPropertyException const & e) {

		common::PortableByteArray<ValueType> portableByteArray;
		portableByteArray.encodeError(common::NULL_TYPE, encodedString, common::LITTLE_ENDIAN_ORDER);

	} catch (BadPropertyTypeException const & e) {

		common::PortableByteArray<ValueType> portableByteArray;
		portableByteArray.encodeError(common::ERROR_TYPE, encodedString, common::LITTLE_ENDIAN_ORDER);
	}
}

template<typename ValueType>
void Database::getArray(unsigned long propertyID, common::ArrayTransferService<ValueType>* arrayTransferService) {

	std::cerr << "cannot get an array" << std::endl;
}


template<typename ValueType>
void Database::setArray(unsigned long propertyID, ArrayCopier<ValueType>* arrayCopier) {

	std::cerr << "cannot set an array from a client" << std::endl;
}

}
