/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DatabaseLoader.h"
#include "Database.h"
#include <dataprovider/dataelement/Command.h>
#include <dataprovider/dataelement/ElementaryProperty.h>
#include <dataprovider/dataelement/DynamicProperty.h>
#include <dataprovider/dataelement/IcsArrayProperty.h>
#include <dataprovider/dataelement/Controller.h>
#include <dataprovider/dataelement/Driver.h>

#include <dataprovider/parser/config/element/ControllerElement.h>
#include <dataprovider/parser/config/element/DriverElement.h>
#include <dataprovider/parser/config/element/PropertyElement.h>
#include <dataprovider/parser/config/element/DynamicPropertyElement.h>
#include <dataprovider/parser/config/element/PropertyContentElement.h>
#include <dataprovider/parser/config/element/ElementaryPropertyElement.h>
#include <dataprovider/parser/config/element/IcsArrayPropertyElement.h>
#include <dataprovider/parser/config/element/CommandElement.h>

#include <dataprovider/ics/ServantLoader.h>

#include <common/base/ServerProperties.h>
#include <common/base/FileUtilities.h>
#include <common/base/Convert.h>
#include <common/base/StringUtilities.h>
#include <common/base/PropertyType.h>
#include <common/base/Real.h>
#include <common/base/RealArray.h>
#include <common/output/OutputStream.h>

#include <DataProvider/Property/AnyProperty.h>

#include <fstream>
#include <iostream>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/fstream.hpp>

using namespace std;
using namespace common;
namespace fs = boost::filesystem;


namespace dataprovider {

const std::string DatabaseLoader::CONFIG = "/config/";
const std::string DatabaseLoader::CONFIG_SAVE = "/config/save/";
const std::string DatabaseLoader::CONTROLLER_CONFIG = "ControllerConfig";
const std::string DatabaseLoader::CONTROLLER_CLEAN_CONFIG = "ControllerCleanConfig";
const std::string DatabaseLoader::DRIVER_CONFIG = "DriverConfig";
const std::string DatabaseLoader::DRIVER_CLEAN_CONFIG = "DriverCleanConfig";
const std::string DatabaseLoader::PREDEFINED_CONFIGS_PATH = "predefinedConfigsPath";
const std::string DatabaseLoader::ICS_NAMESPACE = "ics";

DatabaseLoader::ParserException::ParserException()
: Exception("dataprovider::DatabaseLoader::ParserException", string("Problem occurred while parsing configuration file")) {
}

DatabaseLoader::ServantAssignmentException::ServantAssignmentException(std::string const & parentName, std::string const & n_name, std::string const & name)
: Exception("dataprovider::DatabaseLoader::ServantAssignmentException", string("cannot assign servant '") + name + "' to the child '" + n_name + "' of '" + parentName + "'") {
}

DatabaseLoader::UnknownServantException::UnknownServantException(std::string const & name)
: Exception("dataprovider::DatabaseLoader::UnknownServantException", string("cannot find servant '") + name + "'") {
}

DatabaseLoader::LoadConfigurationException::LoadConfigurationException(std::string const & message)
: Exception("dataprovider::DatabaseLoader::LoadConfigurationException", message) {
}

DatabaseLoader::DatabaseLoader() {
}

DatabaseLoader::~DatabaseLoader() {
}

void DatabaseLoader::loadServants(std::string const & configPath, bool clean) {

	// loading ics servants
	ics::ServantLoader servantLoader;
	_controllers = servantLoader.loadControllers();
	_drivers = servantLoader.loadDrivers();
}

void DatabaseLoader::createControllers() {

	// iterate over the controller elements and create new controllers
	for (vector<ControllerElement*>::const_iterator c = _controllers.begin(); c != _controllers.end(); c++) {
		ControllerElement* controllerElement = *c;

		// create a new controller
		Controller* controller = new Controller(controllerElement->getType(), controllerElement->getName(), controllerElement->isCommand(), controllerElement->isVisible(), controllerElement->isPausable(), false);
		controller->setID(IndexGenerator::getInstance()->getNextIndex());
		controller->setFamily(controllerElement->getFamily(), "none");
		controller->setServerType(controllerElement->getServerType());
		if (controllerElement->isEnabled()) {
			controller->enable();
		} else {
			controller->disable();
		}
		createProperties(controller, controllerElement->getProperties(), controller->getID());
		createCommands(controller, controllerElement->getCommands());

		// add controller to the data base
		Database::getInstance()->addController(controller);
	}
}

void DatabaseLoader::createDrivers() {

	// iterate over drivers from xml file
	for (vector<DriverElement*>::const_iterator d = _drivers.begin(); d != _drivers.end(); d++) {
		DriverElement* driverElement = *d;

		// create a new driver
		Driver* driver = new Driver(driverElement->getType(), driverElement->getName(), false, true, driverElement->isPausable());
		driver->setID(IndexGenerator::getInstance()->getNextIndex());
		driver->setMode(driverElement->getMode());
		driver->setServerType(driverElement->getServerType());
		if (driverElement->isEnabled()) {
			driver->enable();
		} else {
			driver->disable();
		}

		createProperties(driver, driverElement->getProperties(), driver->getID());
		createCommands(driver, driverElement->getCommands());

		// add driver to the data base
		Database::getInstance()->addDriver(driver);
	}
}

unsigned long DatabaseLoader::createController(ControllerElement* controllerElement) {

	// create a new controller
	Controller* controller = new Controller(controllerElement->getType(), controllerElement->getName(), controllerElement->isCommand(), controllerElement->isVisible(), controllerElement->isPausable(), false);
	controller->setID(IndexGenerator::getInstance()->getNextIndex());
	controller->setFamily(controllerElement->getFamily(), "none");
	controller->setServerType(controllerElement->getServerType());
	if (controllerElement->isEnabled()) {
		controller->enable();
	} else {
		controller->disable();
	}
	createProperties(controller, controllerElement->getProperties(), controller->getID());
	createCommands(controller, controllerElement->getCommands());

	// add controller to the data base
	Database::getInstance()->addController(controller);

	return controller->getID();
}


void DatabaseLoader::destroyController(const std::string & name) {
	// add controller to the data base
	Database::getInstance()->removeController(Database::getInstance()->getServantID(name));
}

unsigned long DatabaseLoader::renameController(const std::string & oldname, const std::string & newname) {
	// rename controller to the data base
	unsigned long id = Database::getInstance()->getServantID(oldname);
	Database::getInstance()->renameController(id, newname);
	return id;
}

void DatabaseLoader::createProperties(PropertyHolder* element, const vector<PropertyElement*>& properties, uint32 servantID) {

	// Iterate over PropertyElements and create real Properties
	for (vector<PropertyElement*>::const_iterator p = properties.begin(); p != properties.end(); p++) {
		PropertyElement* propertyElement = (*p);

		ElementaryPropertyElement* elementary = dynamic_cast<ElementaryPropertyElement*>(propertyElement);

		if (elementary != 0) {

			// get the property content from xml file
			DataProperty* property = dynamic_cast<DataProperty *>(createProperty(elementary, servantID));

			if (property == 0) {
				err << "Bad format in \'" << elementary->getName() << "\' content (" << elementary->getContent()->getType() << ") : " << elementary->getContent()->getValue() << eol;

			} else {
				PropertyContentElement* content = elementary->getContent();
				property->setValueFromString(elementary->getContent()->getValue(), property->getFormat());
			}

			element->addProperty(propertyElement->getName(), property);
		}

		DynamicPropertyElement* list = dynamic_cast<DynamicPropertyElement*>(propertyElement);

		if (list != 0) {
			BaseDynamicProperty* dynamicProperty = createDynamicProperty(list, servantID);

			const std::vector<PropertyContentElement*>& content = list->getContent();

			for (std::vector<PropertyContentElement*>::const_iterator c = content.begin(); c != content.end(); c++) {
				PropertyContentElement* contentElement = *c;

				DataProperty* property = dynamicProperty->createProperty(false, contentElement->saveValue());
				// new ID
				unsigned long index = IndexGenerator::getInstance()->getNextIndex();
				property->setID(index);
			 	Database::getInstance()->addProperty(property, servantID);

				dynamicProperty->addProperty(property);

				property->setValueFromString(contentElement->getValue(), property->getFormat());
			}

			element->addProperty(propertyElement->getName(), dynamicProperty);
		}

		// Ics array properties
		IcsArrayPropertyElement* icsArray = dynamic_cast<IcsArrayPropertyElement*>(propertyElement);

		if (icsArray != 0) {
			BaseIcsArrayProperty* property = 0;
			if (icsArray->getContent()->getType() == INT16ARRAY_PROPERTY) {
				property = new IcsArrayProperty<int16>(icsArray->getName(), icsArray->getContent()->getType(), icsArray->getContent()->saveValue(), icsArray->getContent()->getFormat());
			} else	if (icsArray->getContent()->getType() == INT32ARRAY_PROPERTY) {
				property = new IcsArrayProperty<int32>(icsArray->getName(), icsArray->getContent()->getType(), icsArray->getContent()->saveValue(), icsArray->getContent()->getFormat());
			} else	if (icsArray->getContent()->getType() == INT64ARRAY_PROPERTY) {
				property = new IcsArrayProperty<int64>(icsArray->getName(), icsArray->getContent()->getType(), icsArray->getContent()->saveValue(), icsArray->getContent()->getFormat());
			} else if (icsArray->getContent()->getType() == UINT16ARRAY_PROPERTY) {
				property = new IcsArrayProperty<uint16>(icsArray->getName(), icsArray->getContent()->getType(), icsArray->getContent()->saveValue(), icsArray->getContent()->getFormat());
			} else	if (icsArray->getContent()->getType() == UINT32ARRAY_PROPERTY) {
				property = new IcsArrayProperty<uint32>(icsArray->getName(), icsArray->getContent()->getType(), icsArray->getContent()->saveValue(), icsArray->getContent()->getFormat());
			} else	if (icsArray->getContent()->getType() == UINT64ARRAY_PROPERTY) {
				property = new IcsArrayProperty<uint64>(icsArray->getName(), icsArray->getContent()->getType(), icsArray->getContent()->saveValue(), icsArray->getContent()->getFormat());
			} else if (icsArray->getContent()->getType() == FLOAT32ARRAY_PROPERTY) {
				property = new IcsArrayProperty<float64>(icsArray->getName(), icsArray->getContent()->getType(), icsArray->getContent()->saveValue(), icsArray->getContent()->getFormat());
			} else if (icsArray->getContent()->getType() == FLOAT64ARRAY_PROPERTY) {
				property = new IcsArrayProperty<float64>(icsArray->getName(), icsArray->getContent()->getType(), icsArray->getContent()->saveValue(), icsArray->getContent()->getFormat());
			}

			// new ID
			unsigned long index = IndexGenerator::getInstance()->getNextIndex();
			property->setID(index);
			Database::getInstance()->addProperty(property, servantID);

			element->addProperty(icsArray->getName(), property);

			property->setValueFromString(icsArray->getArrayValue());
		}

	}
}

void DatabaseLoader::createCommands(Servant* servant, const vector<CommandElement*>& commands) {

	// iterate over commands in the xml file for the controller
	for (vector<CommandElement*>::const_iterator ci = commands.begin(); ci != commands.end(); ci++) {

		// create a new command and add it to the controllers
		Command* command = new Command((*ci)->getName());
		command->setID(IndexGenerator::getInstance()->getNextIndex());
		Database::getInstance()->addCommand(command);

		servant->addCommand(command);
	}
}

BaseProperty* DatabaseLoader::createProperty(ElementaryPropertyElement* propertyElement, uint32 servantID) {

	if (propertyElement->getContent()->isArray()) {
		cerr << "cannot create an array property" << endl;
		return 0;
	}

	BaseProperty* result = 0;

	try {

		// create the correct property content typeDiagram
		if (propertyElement->getContent()->getType() == TypeToString<int8>::convert()) {
			result = new ElementaryProperty<int8>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<uint16>::convert()) {
			result = new ElementaryProperty<uint16>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<uint32>::convert()) {
			result = new ElementaryProperty<uint32>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<uint64>::convert()) {
			result = new ElementaryProperty<uint64>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<int16>::convert()) {
			result = new ElementaryProperty<int16>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<int32>::convert()) {
			result = new ElementaryProperty<int32>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<int64>::convert()) {
			result = new ElementaryProperty<int64>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<float32>::convert()) {
			result = new ElementaryProperty<float32>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<float64>::convert()) {
			result = new ElementaryProperty<float64>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<bool>::convert()) {
			result = new ElementaryProperty<bool>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else if (propertyElement->getContent()->getType() == TypeToString<string>::convert()) {
			result = new ElementaryProperty<string>(propertyElement->getName(), false, propertyElement->getContent()->saveValue(), propertyElement->getContent()->getFormat());

		} else {
			err << "DatabaseLoader::createPropertyContent : cannot find a type for " << propertyElement->getContent()->getType() << eol;
		}

	} catch (Exception& e) {
		err << e.printStack() << eol;
	}

	// new ID
	unsigned long index = IndexGenerator::getInstance()->getNextIndex();
	result->setID(index);
 	Database::getInstance()->addProperty(result, servantID);

    return result;
}


BaseDynamicProperty* DatabaseLoader::createDynamicProperty(DynamicPropertyElement* listElement, uint32 servantID) {

	if (listElement->isArray()) {
		cerr << "cannot create dynamic property for array" << endl;
		return 0;
	}

	BaseDynamicProperty* result = 0;

	// create the correct property list
	if (listElement->getType() == TypeToString<int8>::convert()) {
		result = new DynamicProperty<int8>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<uint16>::convert()) {
		result = new DynamicProperty<uint16>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<uint32>::convert()) {
		result = new DynamicProperty<uint32>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<uint64>::convert()) {
		result = new DynamicProperty<uint64>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<int16>::convert()) {
		result = new DynamicProperty<int16>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<int32>::convert()) {
		result = new DynamicProperty<int32>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<int64>::convert()) {
		result = new DynamicProperty<int64>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<float32>::convert()) {
		result = new DynamicProperty<float32>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<float64>::convert()) {
		result = new DynamicProperty<float64>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<string>::convert()) {
		result = new DynamicProperty<string>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else if (listElement->getType() == TypeToString<bool>::convert()) {
		result = new DynamicProperty<bool>(listElement->getName(), false, listElement->saveValue(), listElement->getFormat());

	} else {
		err << "Cannot find a type for " << listElement->getType() << eol;
	}

	// new ID
	unsigned long index = IndexGenerator::getInstance()->getNextIndex();
	result->setID(index);
	Database::getInstance()->addProperty(result, servantID);

	return result;
}


void DatabaseLoader::connectControllerServants() {
	// Configure servants for Controllers
	for (vector<ControllerElement*>::const_iterator c = _controllers.begin(); c != _controllers.end(); c++) {
		ControllerElement* controllerElement = *c;

		// get the newly created Controller
		Controller* controller = Database::getInstance()->findController(controllerElement->getName());

		HashMap<string, pair<bool, vector<ControllerElement::ControllerChild> > > const & servants = controllerElement->getServants();

		for (HashMap<string, pair<bool, vector<ControllerElement::ControllerChild> > >::const_iterator s = servants.begin(); s != servants.end(); s++) {
			//cout << "n name = " << s->first;

			vector<ControllerElement::ControllerChild> const & children = s->second.second;

			if (s->second.first) {
				controller->addServantList(s->first, children.size());

				for (unsigned int i = 0; i < children.size(); i++) {

					if (!children[i].isNull) {

						Servant* servant = Database::getInstance()->findServant(children[i].name);
						if (servant != 0) {
							if (children[i].isController && dynamic_cast<Controller*>(servant) != 0) {
								controller->addServantToList(s->first, i, servant);
							} else if (!children[i].isController && dynamic_cast<Driver*>(servant) != 0) {
								controller->addServantToList(s->first, i, servant);
							} else {
								throw ServantAssignmentException(controllerElement->getName(), s->first, children[i].name);
							}

						} else {
							throw UnknownServantException(children[i].name);
						}
					}
				}
			} else {
				Servant* servant = Database::getInstance()->findServant(children[0].name);
				if (servant != 0) {
					if (children[0].isController && dynamic_cast<Controller*>(servant) != 0) {
						controller->addServant(s->first, servant, true);
					} else if (!children[0].isController && dynamic_cast<Driver*>(servant) != 0) {
						controller->addServant(s->first, servant, true);
					} else {
						throw ServantAssignmentException(controllerElement->getName(), s->first, children[0].name);
					}

				} else {
					throw UnknownServantException(children[0].name);
				}
			}
		}
	}
}

void DatabaseLoader::connectDriverServants() {

	for (vector<DriverElement*>::const_iterator d = _drivers.begin(); d != _drivers.end(); d++) {

		DriverElement* driverElement = *d;

		// get the newly created Driver
		Driver* driver = Database::getInstance()->findDriver(driverElement->getName());

		// iterate over servants in the xml file for the driver
		vector<pair<string, pair<bool, vector<list<DriverElement*> > > > > children = driverElement->getChildren().getData();

		for (vector<pair<string, pair<bool, vector<list<DriverElement*> > > > >::const_iterator c = children.begin(); c != children.end(); c++) {

			string const & name = c->first;
			bool isList = c->second.first;
			vector<list<DriverElement*> > const & drivers = c->second.second;

			if (!isList) {

				for (list<DriverElement*>::const_iterator de = drivers[0].begin(); de != drivers[0].end(); de++) {

					DriverElement* childDriverElement = *de;

					if (childDriverElement != 0) {

						//cout << name << " - " << drivers[i]->getName() << eol;

						Driver * childDriver = Database::getInstance()->findDriver(childDriverElement->getName());

						if (childDriver != 0) {
							//cout << "child driver " << drivers[i]->getName() << " = " << childDriver << eol;

							driver->addServant(name, childDriver, true);
							childDriver->setFatherInfo(driver->getID(), name, 0);

						} else {
							throw UnknownServantException(childDriverElement->getName());
						}
					}
				}

			} else {

				driver->addServantList(name, drivers.size());

				for (unsigned int i = 0; i < drivers.size(); i++) {

					for (list<DriverElement*>::const_iterator de = drivers[i].begin(); de != drivers[i].end(); de++) {

						DriverElement* childDriverElement = *de;

						if (childDriverElement != 0) {

							//cout << name << " - " << drivers[i]->getName() << eol;

							Driver * childDriver = Database::getInstance()->findDriver(childDriverElement->getName());

							if (childDriver != 0) {
								//cout << "child driver " << drivers[i]->getName() << " = " << childDriver << endl;

								driver->addServantToList(name, i, childDriver);
								childDriver->setFatherInfo(driver->getID(), name, i);

							} else {
								throw UnknownServantException(childDriverElement->getName());
							}
						}
					}
				}
			}
		}
	}

	//cout << "connectDriverServants ok" << eol;
}

void DatabaseLoader::connectControllerServant(ControllerElement* controllerElement) {

	// get the newly created Controller
	Controller* controller = Database::getInstance()->findController(controllerElement->getName());

	HashMap<string, pair<bool, vector<ControllerElement::ControllerChild> > > const & servants = controllerElement->getServants();

	for (HashMap<string, pair<bool, vector<ControllerElement::ControllerChild> > >::const_iterator s = servants.begin(); s != servants.end(); s++) {
		//cout << "n name = " << s->first;

		vector<ControllerElement::ControllerChild> const & children = s->second.second;

		if (s->second.first) {
			controller->addServantList(s->first, children.size());

			for (unsigned int i = 0; i < children.size(); i++) {

				if (!children[i].isNull) {

					Servant* servant = Database::getInstance()->findServant(children[i].name);
					if (servant != 0) {
						if (children[i].isController && dynamic_cast<Controller*>(servant) != 0) {
							controller->addServantToList(s->first, i, servant);
						} else if (!children[i].isController && dynamic_cast<Driver*>(servant) != 0) {
							controller->addServantToList(s->first, i, servant);
						} else {
							throw ServantAssignmentException(controllerElement->getName(), s->first, children[i].name);
						}

					} else {
						throw UnknownServantException(children[i].name);
					}
				}
			}
		} else {
			Servant* servant = Database::getInstance()->findServant(children[0].name);
			if (servant != 0) {
				if (children[0].isController && dynamic_cast<Controller*>(servant) != 0) {
					controller->addServant(s->first, servant, true);
				} else if (!children[0].isController && dynamic_cast<Driver*>(servant) != 0) {
					controller->addServant(s->first, servant, true);
				} else {
					throw ServantAssignmentException(controllerElement->getName(), s->first, children[0].name);
				}

			} else {
				throw UnknownServantException(children[0].name);
			}
		}
	}
}

}
