/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_DATABASELOADER_H_
#define DATAPROVIDER_DATABASELOADER_H_

#include <common/base/Exception.h>
#include <common/base/BaseTypes.h>
#include <vector>

namespace dataprovider {

class PropertyHolder;
class Servant;
class ControllerElement;
class DriverElement;
class BaseDynamicProperty;
class BaseProperty;
class PropertyElement;
class DynamicPropertyElement;
class ElementaryPropertyElement;
class CommandElement;

/**
 * Class defining a database loader for loading the servants from the configuration file.
 * The class is based upon an XML parser.
 */
class DatabaseLoader {

public:
	/**
	 * Parser exception.
	 */
	class ParserException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		ParserException();
	};

	/**
	 * Servant assignment exception.
	 */
	class ServantAssignmentException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		ServantAssignmentException(std::string const & parentName, std::string const & n_name, std::string const & name);
	};

	/**
	 * Unknown servant exception.
	 */
	class UnknownServantException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		UnknownServantException(std::string const & name);
	};

	/**
	 * load configuration exception.
	 */
	class LoadConfigurationException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		LoadConfigurationException(std::string const & message);
	};

	/**
	 * Constructor.
	 */
	DatabaseLoader();

	/**
	 * Virtual destructor.
	 */
	virtual ~DatabaseLoader();

	/**
	 * Returns true if is test config.
	 */
	bool isTestConfig() const;

	/**
	 * Loads the servants from a config path with the clean option.
	 */
	void loadServants(std::string const & configPath, bool clean);

	/**
	 * Creates and allocates the controllers.
	 */
	void createControllers();

	/**
	 * Creates and allocates the drivers.
	 */
	void createDrivers();

	/**
	 * Creates and allocates the controllers.
	 */
	unsigned long createController(ControllerElement* controllerElement);

	/**
	 * Destroy controller
	 */
	void destroyController(const std::string & name);

	/**
	 * Rename controller
	 */
	unsigned long renameController(const std::string & oldname, const std::string & newname);

	/**
	 * Connects the controllers to the servants.
	 */
	void connectControllerServants();

	/**
	 * Connects the drivers to the servants.
	 */
	void connectDriverServants();

	/**
	 * Connects the controllers to the servants.
	 */
	void connectControllerServant(ControllerElement* controllerElement);

private:
	static const std::string CONFIG;
	static const std::string CONFIG_SAVE;
	static const std::string CONTROLLER_CONFIG;
	static const std::string CONTROLLER_CLEAN_CONFIG;
	static const std::string DRIVER_CONFIG;
	static const std::string DRIVER_CLEAN_CONFIG;
	static const std::string PREDEFINED_CONFIGS_PATH;
	static const std::string ICS_NAMESPACE;

	void createProperties(PropertyHolder* element, const std::vector<PropertyElement*>& properties, uint32 servantID);
	void createCommands(Servant* servant, const std::vector<CommandElement*>& commands);
	BaseProperty* createProperty(ElementaryPropertyElement* contentElement, uint32 servantID);
	BaseDynamicProperty* createDynamicProperty(DynamicPropertyElement* listElement, uint32 servantID);

	std::vector<ControllerElement*> _controllers;
	std::vector<DriverElement*> _drivers;

};

}

#endif /*DATAPROVIDER_DATABASELOADER_H_*/
