/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Database.h"
#include <dataprovider/dataelement/Command.h>
#include <dataprovider/dataelement/ElementaryProperty.h>
#include <dataprovider/dataelement/BaseDynamicProperty.h>
#include <dataprovider/dataelement/Controller.h>
#include <dataprovider/dataelement/Driver.h>
#include <dataprovider/dataelement/BaseProperty.h>
#include <common/base/ServerProperties.h>
#include <common/base/Convert.h>
#include <common/base/PropertyType.h>
#include <iostream>

using namespace common;
using namespace std;

namespace dataprovider {

std::set<std::string> Database::getKeys(const common::HashMap<std::string, unsigned long> & map) {
	set<string> result;

	vector<pair<string, unsigned long> > data = map.getData();

	for (vector<pair<string, unsigned long> >::const_iterator p = data.begin(); p != data.end(); ++p) {
		result.insert(p->first);
	}

	return result;
}

std::string Database::typeToString(const std::string& type) {

	string result;
	result += " type=\"";
	result += type + "\"";

	if (result.size() < 22) {
		result.resize(22, ' ');
	}

	return result;
}

std::string Database::IDToString(unsigned long ID, bool save) {

	string result;

	result += " id=\"";
	result += to<string>(ID) + "\"";

	if (result.size() < 15) {
		result.resize(15, ' ');
	}

	return result;
}

std::string Database::propertyIDToString(unsigned long ID, bool save) {

	return IDToString(ID, save);
}

std::string Database::DataPropertyToString(const std::string& propertyName, DataProperty * property, bool save, const std::string& space) {

	if (property == 0 || (save && property->isGenerated())) {
		return "";
	}

	if (property->getName() == "credentials") {
		return "";
	}

	string result;
	string value;
	bool overflow;
	if (!save) {
		value = string("   value=\"") + property->getStringValue(overflow) + "\"";
	} else if (save && property->saveValue()) {
		string v = property->getStringValue(overflow);
		if (!overflow) {
			value = string("   value=\"") + v + "\"";
		}

		if (property->getFormat() != "") {
			value += string(" format=\"") + property->getFormat() + "\"";
		}
	}
	string attributes;

	if (save && !property->saveValue()) {
		attributes = " save_value=\"false\"";
	}

	BaseElementaryProperty* elementaryProperty = dynamic_cast<BaseElementaryProperty*>(property);

	if (elementaryProperty != 0) {
		value += string(" formattedValue=\"") + elementaryProperty->getFormattedValue() + "\"";
	}

	result.append(space + "<property" + propertyIDToString(property->getID(), save) + typeToString(property->getType()) + " name=\"" + propertyName + "\"" + value + attributes + "/>\n");

	return result;
}

std::string Database::DynamicPropertyToString(const std::string& dynamicPropertyName, BaseDynamicProperty * dynamicProperty, bool save, const std::string& space) {

	if (dynamicProperty == 0 || (save && dynamicProperty->isGenerated())) {
		return "";
	}

	string result;
	string attributes;
	if (save && !dynamicProperty->saveValue()) {
		attributes = " save_value=\"false\"";
	}

	if (dynamicProperty->getFormat() != "") {
		attributes += string(" format=\"") + dynamicProperty->getFormat() + "\"";
	}

	result.append(space + "<property" + propertyIDToString(dynamicProperty->getID(), save) + typeToString(dynamicProperty->getType()) + " name=\"" + dynamicPropertyName + "\"" + attributes + " dynamic=\"true\">\n");

	if (!save || (save && dynamicProperty->saveValue())) {

		// list Properties
		vector<unsigned long> const & properties = dynamicProperty->getPropertyIDs();

		for (vector<unsigned long>::const_iterator p = properties.begin(); p != properties.end(); ++p) {

			unsigned long propertyID = *p;
			DataProperty* property = dynamic_cast<DataProperty*>(_keyIDPropertyMap.get(propertyID));
			if (property != 0) {

				string value;
				bool overflow;
				if (!save) {
					value = string("value=\"") + property->getStringValue(overflow) + "\"";
				} else if (save && dynamicProperty->saveValue() && property->saveValue()) {
					string v = property->getStringValue(overflow);
					if (!overflow) {
						value = string("value=\"") + v + "\"";
					}
				}

				BaseElementaryProperty* elementaryProperty = dynamic_cast<BaseElementaryProperty*>(property);

				if (elementaryProperty != 0) {
					value += string(" formattedValue=\"") + elementaryProperty->getFormattedValue() + "\"";
				}

				result.append(space + "      <element" + propertyIDToString(property->getID(), save) + value + "/>\n");
			}

		}
	}

	result.append(space + "</property>\n");

	return result;
}

void Database::controllerToString(Controller* controller, bool save, std::string& result) {

	if ((save && !controller->isGenerated()) || !save) {

		string commandZone;
		if (save) {
			commandZone = string(" is_command=\"") + to<string>(controller->isCommand()) + "\" is_visible=\"" + to<string>(controller->isVisible()) + "\"";
			if (controller->isPausable()) {
				commandZone = commandZone + string(" pausable=\"true\"");
			}
		}

		string servantType = getShortServantType(controller->getID());
		deque<string> types = _controllerTypeInheritance.get(servantType);

		string typesString = "{";

		for (deque<string>::const_iterator t = types.begin(); t != types.end(); ++t) {
			typesString += *t;
			if ((t + 1) != types.end()) {
				typesString += ", ";
			}
		}
		typesString += "}";


		result.append(string("            <controller" + IDToString(controller->getID(), save) + " type=\"") + controller->getControllerType() + "\" types=\"" + typesString + "\" name=\"" + controller->getName() + "\" is_enabled=\"" + to<string>(controller->isEnabled()) + "\"" + commandZone + ">\n");
		result.append(string("                  <family name=\"") + controller->getFamily() + "\"/>\n");
		result.append(string("                  <subfamily name=\"") + controller->getSubfamily() + "\"/>\n");

		result.append(string("                  <properties>\n"));

		// Properties
		HashMap<string, unsigned long> const & properties = controller->getProperties();
		set<string> propertyNames = getKeys(properties);

		for (set<string>::const_iterator p = propertyNames.begin(); p != propertyNames.end(); ++p) {

			const string& propertyName = *p;

			try {
				BaseProperty* property = _keyIDPropertyMap.get(properties.get(propertyName));

				//ScopedLock propertyLock(_lockManager->lockPropertyForReading(propertyID, eol));

				result.append(DataPropertyToString(propertyName, dynamic_cast<DataProperty *>(property), save, "                        "));
				result.append(DynamicPropertyToString(propertyName, dynamic_cast<BaseDynamicProperty *>(property), save, "                        "));
			} catch (Exception& e) {
				// do nothing
				err << e.printStack() << eol;
			}
		}

		result.append(string("                  </properties>\n"));

		result.append(string("                  <commands>\n"));

		HashMap<string, unsigned long> const & commands = controller->getCommands();

		for (HashMap<string, unsigned long>::const_iterator c = commands.begin(); c != commands.end(); ++c) {

			//ScopedLock commandLock(_lockManager->lockCommandForReading(c->second, eol));

			Command* command = _IDCommandMap.get(c->second);

			string state;
			if (!save) {
				state = string(" state=\"") + command->getStateString() + "\"";
			}

			result.append(string("                        <command") + IDToString(c->second, save) + " name=\"" + command->getName() + "\"" + state + "/>\n");

		}

		result.append(string("                  </commands>\n"));


		result.append(string("                  <servants>\n"));

		HashMap<string, pair<bool, vector<Servant::Children> > > const & servantMap = controller->getServantMap();
		for (HashMap<string, pair<bool, vector<Servant::Children> > >::const_iterator s = servantMap.begin(); s != servantMap.end(); s++) {

			vector<Servant::Children> const & servants = s->second.second;

			if (controller->servantToBeSaved(s->first)) {

				if (s->second.first) {
					result.append(string("                        <servant_list name=\"") + s->first + "\" size=\"" + to<string>(servants.size()) + "\">\n");

					for (unsigned int i = 0; i < servants.size(); i++) {
						// only a single controller is supposed to exist
						if (servants[i].children.size() > 0) {

							unsigned long servantID = servants[i].children.front();

							//ScopedLock servantLock(_lockManager->lockServantForReading(servantID, eol));

							Controller* lcontroller = dynamic_cast<Controller*>(getServant(servantID));
							if (controller != 0) {
								if ((save && !lcontroller->isGenerated()) || !save) {
									result.append(string("                              <controller name=\"") + lcontroller->getName() + "\" position=\"" + to<string>(i) + "\"/>\n");
								}
							} else {
								Driver* driver = dynamic_cast<Driver*>(getServant(servantID));
								if (driver != 0) {
									result.append(string("                              <driver name=\"") + driver->getName() + "\" position=\"" + to<string>(i) + "\"/>\n");
								}
							}
						}
					}

					result.append(string("                        </servant_list>\n"));
				} else {

					unsigned long servantID = servants[0].children.front();

					if (servantID != 0) {
						//ScopedLock servantLock(_lockManager->lockServantForReading(servantID, eol));

						Controller* lcontroller = dynamic_cast<Controller*>(getServant(servantID));
						if (lcontroller != 0) {
							if ((save && !lcontroller->isGenerated()) || !save) {
								result.append(string("                        <servant name=\"") + s->first + "\">\n");
								result.append(string("                              <controller name=\"") + lcontroller->getName() + "\"/>\n");
								result.append(string("                        </servant>\n"));
							}
						} else {
							Driver* driver = dynamic_cast<Driver*>(getServant(servantID));
							if (driver != 0) {
								result.append(string("                        <servant name=\"") + s->first + "\">\n");
								result.append(string("                              <driver name=\"") + driver->getName() + "\"/>\n");
								result.append(string("                        </servant>\n"));
							}
						}
					}
				}
			}
		}

		result.append(string("                  </servants>\n"));

		result.append(string("            </controller>\n\n"));
	}
}

std::string Database::getControllersToString(bool save) {

	string result;

	result.append("<config>\n");
	result.append("      <controllers>\n");

	vector<unsigned long> controllerIDs;
	if (save) {
		for (vector<string>::iterator it = _controllerConfigNames.begin(); it < _controllerConfigNames.end(); it++) {
			controllerIDs.push_back(getServantID(*it));
		}
	} else {
		controllerIDs = _IDControllerMap.getKeySet();
	}

	for (vector<unsigned long>::const_iterator c = controllerIDs.begin(); c != controllerIDs.end(); ++c) {

		try {
			//ScopedLock servantLock(_lockManager->lockServantForReading(*c, eol));

			Controller * controller = _IDControllerMap.get(*c);

			controllerToString(controller, save, result);

		} catch (const Exception& e) {
			out << "Database::getControllersToString : " << e.printStack() << eol;
		}
	}

	result.append("      </controllers>\n");
	result.append("</config>\n");

	return result;
}

void Database::driverToString(Driver* driver, bool save, std::string& result) {

	string mode = "";
	if (driver->getMode() == common::REAL_MODE) {
		mode = "real";
	} else if (driver->getMode() == common::SIMULATED_MODE) {
		mode = "simulated";
	} else {
		mode = "perfect";
	}

	string pausable = "";
	if (driver->isPausable()) {
		pausable = " pausable=\"true\"";
	}

	string servantType = getShortServantType(driver->getID());
	deque<string> types = _driverTypeInheritance.get(servantType);

	string typesString = "{";

	for (deque<string>::const_iterator t = types.begin(); t != types.end(); ++t) {
		typesString += *t;
		if ((t + 1) != types.end()) {
			typesString += ", ";
		}
	}
	typesString += "}";

	result.append(string("            <driver") + IDToString(driver->getID(), save) + " type=\"" + driver->getDriverType() + "\" types=\"" + typesString + "\" name=\"" + driver->getName() + "\" mode=\"" + mode + "\" is_enabled=\"" + to<string>(driver->isEnabled()) + "\"" + pausable + ">\n");
	result.append(string("                  <properties>\n"));

	// Properties
	HashMap<string, unsigned long> const & properties = driver->getProperties();
	set<string> propertyNames = getKeys(properties);

	for (set<string>::const_iterator p = propertyNames.begin(); p != propertyNames.end(); ++p) {

		const string& propertyName = *p;

		try {
			BaseProperty* property = _keyIDPropertyMap.get(properties.get(propertyName));

			//ScopedLock propertyLock(_lockManager->lockPropertyForReading(propertyID, eol));
			result.append(DataPropertyToString(propertyName, dynamic_cast<DataProperty *>(property), save, "                        "));
			result.append(DynamicPropertyToString(propertyName, dynamic_cast<BaseDynamicProperty *>(property), save, "                        "));
		} catch (Exception& e) {
			// do nothing
			err << e.printStack() << eol;
		}
	}

	result.append(string("                  </properties>\n"));

	result.append(string("                  <commands>\n"));

	HashMap<string, unsigned long> const & commands = driver->getCommands();

	for (HashMap<string, unsigned long>::const_iterator c = commands.begin(); c != commands.end(); ++c) {

		//ScopedLock commandLock(_lockManager->lockCommandForReading(c->second, eol));

		Command* command = _IDCommandMap.get(c->second);

		string state;
		if (!save) {
			state = string(" state=\"") + command->getStateString() + "\"";
		}

		result.append(string("                        <command") + IDToString(c->second, save) + " name=\"" + command->getName() + "\"" + state + "/>\n");

	}

	result.append(string("                  </commands>\n"));

	result.append(string("            </driver>\n\n"));
}

std::string Database::getDriversToString(bool save) {

	string result;

	result.append("<config>\n");
	result.append("      <drivers>\n");

	vector<unsigned long> driverIDs;
	if (save) {
		for (vector<string>::iterator it = _driverConfigNames.begin(); it < _driverConfigNames.end(); it++) {
			driverIDs.push_back(getServantID(*it));
		}
	} else {
		driverIDs = _IDDriverMap.getKeySet();
	}

	for (vector<unsigned long>::const_iterator d = driverIDs.begin(); d != driverIDs.end(); ++d) {

		try {
			//ScopedLock servantLock(_lockManager->lockServantForReading(*d, eol));

			Driver * driver = _IDDriverMap.get(*d);

			driverToString(driver, save, result);

		} catch (Exception&) {
		}
	}

	result.append("      </drivers>\n");


	result.append("      <hierarchy>\n");

	for (vector<unsigned long>::const_iterator d = driverIDs.begin(); d != driverIDs.end(); ++d) {

		try {
			//ScopedLock servantLock(_lockManager->lockServantForReading(*d, eol));

			Driver * driver = _IDDriverMap.get(*d);
			if (driver->isRoot()) {
				printDriverHierarchy(driver, -1, "            ", result);
			}

		} catch (Exception&) {
		}
	}

	result.append("      </hierarchy>\n");
	result.append("</config>\n");

	return result;
}

std::string Database::toString() {

	string result;

	result.append(getDriversToString(false));
	result.append(getControllersToString(false));

	return result;
}

std::string Database::toString(unsigned long servantID) {

	if (isController(servantID)) {

		ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

		Controller* controller = _IDControllerMap.get(servantID);
		if (controller != 0) {
			string result;
			controllerToString(controller, false, result);
			return result;
		}

		return "unknown controller\n";
	}

	if (isDriver(servantID)) {

		ScopedLock lock(_lockManager->lockServantForReading(servantID, eol));

		Driver* driver = _IDDriverMap.get(servantID);
		if (driver != 0) {
			string result;
			driverToString(driver, false, result);
			return result;
		}

		return "unknown driver\n";
	}

	return "unknown servant\n";
}

void Database::printDriverHierarchy(Driver * driver, int channel, std::string const & offset, std::string & output) {

	if (driver != 0) {

		if (channel != -1) {
			output.append(offset + "<driver name=\"" + driver->getName() + "\" channel=\"" + to<string>(channel) + "\"");
		} else {
			output.append(offset + "<driver name=\"" + driver->getName() + "\"");
		}

		if (driver->getServantMap().getData().empty()) {
			output.append("/>\n");

		} else {
			output.append(">\n");

			vector<pair<string, pair<bool, vector<Servant::Children> > > > children = driver->getServantMap().getData();
			for (vector<pair<string, pair<bool, vector<Servant::Children> > > >::const_iterator c = children.begin(); c != children.end(); c++) {

				string const & name = c->first;
				bool isList = c->second.first;
				vector<Servant::Children> const & servants = c->second.second;

				if (!isList) {

					output.append(offset + "      <child name=\"" + name + "\">\n");

					for (vector<unsigned long>::const_iterator d = servants[0].children.begin(); d != servants[0].children.end(); d++) {

						unsigned long driverID = *d;

						//ScopedLock servantLock(_lockManager->lockServantForReading(driverID, eol));

						Driver* childDriver = static_cast<Driver *>(getServant(driverID));
						printDriverHierarchy(childDriver, -1, offset + "            ", output);
					}

					output.append(offset + "      </child>\n");

				} else {

					output.append(offset + "      <child_list name=\"" + name + "\" size=\"" + to<string>(servants.size()) + "\">\n");

					for (unsigned long i = 0; i < servants.size(); i++) {
						for (vector<unsigned long>::const_iterator d = servants[i].children.begin(); d != servants[i].children.end(); d++) {

							unsigned long driverID = *d;

							//ScopedLock servantLock(_lockManager->lockServantForReading(driverID, eol));

							Driver* childDriver = static_cast<Driver *>(getServant(driverID));

							printDriverHierarchy(childDriver, i, offset + "            ", output);
						}
					}

					output.append(offset + "      </child_list>\n");
				}
			}

			output.append(offset + "</driver>\n");
		}
	}
}

}
