/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ElementLockManager.h"
#include "Lock.h"
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <common/base/Date.h>
#include <boost/thread/xtime.hpp>
#include <iostream>
#include <time.h>
#include <errno.h>

using namespace std;
using namespace common;

namespace dataprovider {

const std::string ElementLockManager::ELEMENT_LOCK_MANAGER_MUTEX_POOL_SIZE = "elementLockManagerMutexPoolSize";

ElementLockManager::ElementLockManager() {
	
	pthread_mutexattr_t mutexAttributes;
	pthread_mutexattr_init(&mutexAttributes);
	pthread_mutexattr_settype(&mutexAttributes, PTHREAD_MUTEX_RECURSIVE);
	
	pthread_mutex_init(&_databaseMutex.mutex, &mutexAttributes);
	
	try {
		_mutexPoolSize = ServerProperties::getInstance()->get<unsigned long>(ELEMENT_LOCK_MANAGER_MUTEX_POOL_SIZE);
		
	} catch (ServerProperties::NoSuchPropertyException const & e) {
		
		_mutexPoolSize = 8192;
		out << "Setting default value 8192 to property 'elementLockManagerMutexPoolSize'" << eol;

	}
	
	if (_mutexPoolSize == 0) {
		_mutexPoolSize = 8192;
		out << "Setting default value 8192 to property 'elementLockManagerMutexPoolSize'" << eol;
	}
	
	_mutexPool = new Mutex[_mutexPoolSize];
	
	for (unsigned long i = 0; i < _mutexPoolSize; i++) {
		pthread_mutex_init(&(_mutexPool[i].mutex), &mutexAttributes);
	}	
	
}
		
ElementLockManager::~ElementLockManager() {
	
	delete [] _mutexPool;
}

void ElementLockManager::createServant(unsigned long servantID) {
	
}
	
void ElementLockManager::removeServant(unsigned long servantID) {

}
	
Lock* ElementLockManager::lockServantForReading(unsigned long servantID, common::EndOfLine const & endOfLine) {

	if (servantID != 0) {
		lock(servantID, endOfLine);
	}
	return new Lock(this, Lock::READ_SERVANT, servantID);
}
	
void ElementLockManager::unlockServantForReading(unsigned long servantID) {
	
	if (servantID != 0) {
		unlock(servantID);
	}	
}
	
Lock* ElementLockManager::lockServantForWriting(unsigned long servantID, common::EndOfLine const & endOfLine) {
	
	if (servantID != 0) {
		lock(servantID, endOfLine);
	}	
	return new Lock(this, Lock::WRITE_SERVANT, servantID);
}
	
void ElementLockManager::unlockServantForWriting(unsigned long servantID) {
	
	if (servantID != 0) {
		unlock(servantID);
	}	
}

Lock* ElementLockManager::lockServantForDeleting(unsigned long servantID, common::EndOfLine const & endOfLine) {

	if (servantID != 0) {
		lock(servantID, endOfLine);
	}	
	return new Lock(this, Lock::DELETE_SERVANT, servantID);
}
	
void ElementLockManager::unlockServantForDeleting(unsigned long servantID) {

	if (servantID != 0) {
		unlock(servantID);
	}	
}
	
void ElementLockManager::createProperty(unsigned long propertyID) {

}
	
void ElementLockManager::removeProperty(unsigned long propertyID) {

}

Lock* ElementLockManager::lockPropertyForReading(unsigned long propertyID, common::EndOfLine const & endOfLine) {

	if (propertyID != 0) {
		lock(propertyID, endOfLine);
	}	
	return new Lock(this, Lock::READ_PROPERTY, propertyID);
}
	
void ElementLockManager::unlockPropertyForReading(unsigned long propertyID) {
	
	if (propertyID != 0) {
		unlock(propertyID);
	}	
}

Lock* ElementLockManager::lockPropertyForWriting(unsigned long propertyID, common::EndOfLine const & endOfLine) {

	if (propertyID != 0) {
		lock(propertyID, endOfLine);
	}		
	return new Lock(this, Lock::WRITE_PROPERTY, propertyID);
}
	
void ElementLockManager::unlockPropertyForWriting(unsigned long propertyID) {
	
	if (propertyID != 0) {
		unlock(propertyID);
	}	
}

Lock* ElementLockManager::lockPropertyForDeleting(unsigned long propertyID, common::EndOfLine const & endOfLine) {
	
	if (propertyID != 0) {
		lock(propertyID, endOfLine);
	}	
	return new Lock(this, Lock::DELETE_PROPERTY, propertyID);
}
	
void ElementLockManager::unlockPropertyForDeleting(unsigned long propertyID) {

	if (propertyID != 0) {
		unlock(propertyID);
	}	
}

Lock* ElementLockManager::lockCommandForReading(unsigned long commandID, common::EndOfLine const & endOfLine) {

	if (commandID != 0) {
		lock(commandID, endOfLine);
	}
	return new Lock(this, Lock::READ_COMMAND, commandID);
}
	
void ElementLockManager::unlockCommandForReading(unsigned long commandID) {
	
	if (commandID != 0) {
		unlock(commandID);
	}	
}

Lock* ElementLockManager::lockCommandForWriting(unsigned long commandID, common::EndOfLine const & endOfLine) {

	if (commandID != 0) {
		lock(commandID, endOfLine);
	}		
	return new Lock(this, Lock::WRITE_COMMAND, commandID);
}
	
void ElementLockManager::unlockCommandForWriting(unsigned long commandID) {
	
	if (commandID != 0) {
		unlock(commandID);
	}	
}

Lock* ElementLockManager::lockCommandForDeleting(unsigned long commandID, common::EndOfLine const & endOfLine) {
	
	if (commandID != 0) {
		lock(commandID, endOfLine);
	}	
	return new Lock(this, Lock::DELETE_COMMAND, commandID);
}
	
void ElementLockManager::unlockCommandForDeleting(unsigned long commandID) {

	if (commandID != 0) {
		unlock(commandID);
	}	
}
	
Lock* ElementLockManager::lockMaps(common::EndOfLine const & endOfLine) {

	pthread_mutex_lock(&_databaseMutex.mutex);
	_databaseMutex.line = endOfLine;
	return new Lock(this, Lock::DATABASE, 0);
}
	
void ElementLockManager::unlockMaps() {

	pthread_mutex_unlock(&_databaseMutex.mutex);
}

inline void ElementLockManager::lock(unsigned long ID, common::EndOfLine const & endOfLine) {
	
	Mutex& mutex = _mutexPool[(ID - 1) % _mutexPoolSize];
	
	pthread_mutex_lock(&(mutex.mutex));
	mutex.line = endOfLine;
}

inline void ElementLockManager::unlock(unsigned long ID) {

	pthread_mutex_unlock(&(_mutexPool[(ID - 1) % _mutexPoolSize].mutex));
}

timespec ElementLockManager::currentTimespec() {
	
	boost::xtime xt;
#if (BOOST_VERSION > 105000)
	xtime_get(&xt, boost::TIME_UTC_);
#else
	xtime_get(&xt, boost::TIME_UTC);
#endif
	
	timespec ts;

	ts.tv_sec = static_cast<int>(xt.sec);
    ts.tv_nsec = static_cast<int>(xt.nsec);
    if(ts.tv_nsec >= NANOSECONDS_PER_SECOND) {
        ts.tv_sec += ts.tv_nsec / NANOSECONDS_PER_SECOND;
        ts.tv_nsec %= NANOSECONDS_PER_SECOND;
    }
    
    return ts;
}

void ElementLockManager::verifyLocks(unsigned long maxID, int seconds) {
	
	timespec ts = currentTimespec();
	ts.tv_sec += seconds;

    int res = pthread_mutex_timedlock(&_databaseMutex.mutex, &ts);
    if (res == ETIMEDOUT) {
		out << nolog << "database is locked (" << _databaseMutex.line.getFile() << " : " << _databaseMutex.line.getLine() << ")" << eol;
	} else if (res == 0) {
		pthread_mutex_unlock(&_databaseMutex.mutex);
	} else {
		out << nolog << "error while verifying database, errno = " << res << eol;
	}
	
	for (unsigned long id = 1; id <= maxID; id++) {
		
		unsigned long i = (id - 1) % _mutexPoolSize;
		
		timespec ts = currentTimespec();
		ts.tv_sec += seconds;

        res = pthread_mutex_timedlock(&_mutexPool[i].mutex, &ts);
        if (res == ETIMEDOUT) {
			out << nolog << "element <" << id << "> is locked (" << _mutexPool[i].line.getFile() << " : " << _mutexPool[i].line.getLine() << ")" << eol;
		} else if (res == 0) {
			pthread_mutex_unlock(&_mutexPool[i].mutex);
		} else {
			out << nolog << "error while verifying element <" << id << ">, errno = " << res << eol;
		}
	}
	
	out << nolog << "locks verified" << eol;
}

}
