/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_ELEMENTLOCKMANAGER_H_
#define DATAPROVIDER_ELEMENTLOCKMANAGER_H_

#include "LockManager.h"

#include <string>
#include <pthread.h>

namespace dataprovider {

/**
 * Class defining a lock manager for the database. A mutex is used per element.
 */	
class ElementLockManager : public LockManager {

public:
	/**
	 * Constructor.
	 */
	ElementLockManager();
		
	/**
	 * Virtual destructor.
	 */
	virtual ~ElementLockManager();
	
	/**
	 * Creates a servant.
	 */
	virtual void createServant(unsigned long servantID);

	/**
	 * Removes a servant.
	 */
	virtual void removeServant(unsigned long servantID);
	
	/**
	 * Locks the servant for reading.
	 */
	virtual Lock* lockServantForReading(unsigned long servantID, common::EndOfLine const & endOfLine);
	
	/**
	 * Unlocks the servant for reading.
	 */
	virtual void unlockServantForReading(unsigned long servantID);
	
	/**
	 * Locks the servant for writing.
	 */
	virtual Lock* lockServantForWriting(unsigned long servantID, common::EndOfLine const & endOfLine);
	
	/**
	 * Unlocks the servant for writing.
	 */
	virtual void unlockServantForWriting(unsigned long servantID);

	/**
	 * Locks the servant for deleting.
	 */
	virtual Lock* lockServantForDeleting(unsigned long servantID, common::EndOfLine const & endOfLine);
	
	/**
	 * Unlocks the servant for deleting.
	 */
	virtual void unlockServantForDeleting(unsigned long servantID);
	
	/**
	 * Creates a property.
	 */
	virtual void createProperty(unsigned long propertyID);
	
	/**
	 * Removes a property.
	 */
	virtual void removeProperty(unsigned long propertyID);

	/**
	 * Locks the property for reading.
	 */
	virtual Lock* lockPropertyForReading(unsigned long propertyID, common::EndOfLine const & endOfLine);
	
	/**
	 * Unlocks the property for reading.
	 */
	virtual void unlockPropertyForReading(unsigned long propertyID);

	/**
	 * Locks the property for writing.
	 */
	virtual Lock* lockPropertyForWriting(unsigned long propertyID, common::EndOfLine const & endOfLine);
	
	/**
	 * Unlocks the property for writing.
	 */
	virtual void unlockPropertyForWriting(unsigned long propertyID);

	/**
	 * Locks the property for deleting.
	 */
	virtual Lock* lockPropertyForDeleting(unsigned long propertyID, common::EndOfLine const & endOfLine);
	
	/**
	 * Unlocks the property for deleting.
	 */
	virtual void unlockPropertyForDeleting(unsigned long propertyID);

	/**
	 * Locks the command for reading.
	 */
	virtual Lock* lockCommandForReading(unsigned long commandID, common::EndOfLine const & endOfLine);

	/**
	 * Unlocks the command for reading.
	 */
	virtual void unlockCommandForReading(unsigned long commandID);
	
	/**
	 * Locks the command for writing.
	 */
	virtual Lock* lockCommandForWriting(unsigned long commandID, common::EndOfLine const & endOfLine);

	/**
	 * Unlocks the command for writing.
	 */
	virtual void unlockCommandForWriting(unsigned long commandID);

	/**
	 * Locks the command for deleting.
	 */
	virtual Lock* lockCommandForDeleting(unsigned long commandID, common::EndOfLine const & endOfLine);
	
	/**
	 * Unlocks the command for deleting.
	 */
	virtual void unlockCommandForDeleting(unsigned long commandID);
	
	/**
	 * Locks the database.
	 */
	virtual Lock* lockMaps(common::EndOfLine const & endOfLine);
	
	/**
	 * Unlocks the database.
	 */
	virtual void unlockMaps();

	/**
	 * Verifies the locks
	 */
	virtual void verifyLocks(unsigned long maxID, int seconds);

private:
	void lock(unsigned long ID, common::EndOfLine const & endOfLine);
	void unlock(unsigned long ID);
	static timespec currentTimespec();

	static const int NANOSECONDS_PER_SECOND = 1000000000;
	
	struct Mutex {
		pthread_mutex_t mutex;
		common::EndOfLine line;
	};
	
	Mutex _databaseMutex;
	Mutex* _mutexPool;

	static const std::string ELEMENT_LOCK_MANAGER_MUTEX_POOL_SIZE;
	unsigned long _mutexPoolSize;
};

}

#endif /*DATAPROVIDER_ELEMENTLOCKMANAGER_H_*/