/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_GNULOCKABLEHASHMAP_H_
#define DATAPROVIDER_GNULOCKABLEHASHMAP_H_

#include <common/base/BaseNoSuchKeyException.h>
#include <common/base/Convert.h>
#include <common/base/Hash.h>
#include <string>
#include <ext/hash_map>
#include <vector>
#include "LockManager.h"
#include "ScopedLock.h"

namespace dataprovider {

/**
 * Facade class for hiding the non-standard implementation of the hash map. Concurrent version using a mutex
 * to prevent concurrent access and removing iterators.
 */
template<typename Key, typename Data, typename HashFun = common::Hash<Key> >
class LockableHashMap {

	/**
	 * Implementation type
	 */
	typedef typename __gnu_cxx::hash_map<Key, Data, HashFun> ImplementationType;

public:
	/**
	 * Exception.
	 */
	class NoSuchKeyException : public common::BaseNoSuchKeyException {

	public:
		/**
		 * Constructor.
		 */
		NoSuchKeyException(Key const & key);
	};

	/**
	 * Constructor.
	 */
	inline LockableHashMap() {}

	/**
	 * Sets the lock manager.
	 */
	void setLockManager(LockManager * lockManager) {
		_lockManager = lockManager;
	}

	/**
	 * Copy constructor.
	 */
	inline LockableHashMap(LockableHashMap<Key, Data, HashFun> const & map) {
		ScopedLock lock(_lockManager->lockMaps(common::eol));
		_impl = map._impl;
	}

	/**
	 * Operator=.
	 */
	inline LockableHashMap<Key, Data, HashFun> const & operator= (LockableHashMap<Key, Data, HashFun> & map) {
		ScopedLock lock(_lockManager->lockMaps(common::eol));
		_impl = map._impl;

		return *this;
	}

	/**
	 * Inserts the data item for the key key. The old value is replaced.
	 */
	inline void insert(Key const & key, Data const & data) {
		ScopedLock lock(_lockManager->lockMaps(common::eol));
		// call to the operator[], because insert doesn't replace the old value
		_impl[key] = data;
	}

	/**
	 * Gets the data item reference for the key key.
	 * throw NoSuchKeyException
	 */
	inline Data& operator [] (Key const & key) {
		return get(key);
	}

	/**
	 * Erases the key and its data.
	 * throw NoSuchKeyException
	 */
	void erase(Key const & key);

	/**
	 * Returns true if key is found.
	 */
	inline bool find(Key const & key) {
		ScopedLock lock(_lockManager->lockMaps(common::eol));
		return (_impl.find(key) != _impl.end());
	}

	/**
	 * Gets the item reference.
	 * throw NoSuchKeyException
	 */
	Data & get(Key const & key);

	/**
	 * Gets the size of the map.
	 */
	unsigned long getSize();

	/**
	 * Gets the vector of Key.
	 */
	std::vector<Key> getKeySet();

	/**
	 * Clears the map.
	 */
	void clear();

private:
	ImplementationType _impl;
	LockManager* _lockManager;
};

}

#include "GnuLockableHashMap.inl"

#endif /*DATAPROVIDER_GNULOCKABLEHASHMAP_H_*/