/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_LOCKMANAGER_H_
#define DATAPROVIDER_LOCKMANAGER_H_

#include <pthread.h>
#include <common/output/EndOfLine.h>

namespace dataprovider {

class Lock;
	
/**
 * Abstract class defining a Lock manager for the database.
 */	
class LockManager {

public:
	/**
	 * Constructor.
	 */
	LockManager();

	/**
	 * Virtual destructor.
	 */
	virtual ~LockManager();
	
	/**
	 * Creates a servant.
	 */
	virtual void createServant(unsigned long servantID) = 0;
	
	/**
	 * Removes a servant.
	 */
	virtual void removeServant(unsigned long servantID) = 0;

	/**
	 * Locks the servant for reading.
	 */
	virtual Lock* lockServantForReading(unsigned long servantID, common::EndOfLine const & endOfLine) = 0;

	/**
	 * Unlocks the servant for reading.
	 */
	virtual void unlockServantForReading(unsigned long servantID) = 0;
	
	/**
	 * Locks the servant for writing.
	 */
	virtual Lock* lockServantForWriting(unsigned long servantID, common::EndOfLine const & endOfLine) = 0;
	
	/**
	 * Unlocks the servant for writing.
	 */
	virtual void unlockServantForWriting(unsigned long servantID) = 0;

	/**
	 * Locks the servant for deleting.
	 */
	virtual Lock* lockServantForDeleting(unsigned long servantID, common::EndOfLine const & endOfLine) = 0;
	
	/**
	 * Unlocks the servant for deleting.
	 */
	virtual void unlockServantForDeleting(unsigned long servantID) = 0;
	
	/**
	 * Creates a property.
	 */
	virtual void createProperty(unsigned long propertyID) = 0;
	
	/**
	 * Removes a property.
	 */
	virtual void removeProperty(unsigned long propertyID) = 0;
	
	/**
	 * Locks the property for reading.
	 */
	virtual Lock* lockPropertyForReading(unsigned long propertyID, common::EndOfLine const & endOfLine) = 0;

	/**
	 * Unlocks the property for reading.
	 */
	virtual void unlockPropertyForReading(unsigned long propertyID) = 0;
	
	/**
	 * Locks the property for writing.
	 */
	virtual Lock* lockPropertyForWriting(unsigned long propertyID, common::EndOfLine const & endOfLine) = 0;

	/**
	 * Unlocks the property for writing.
	 */
	virtual void unlockPropertyForWriting(unsigned long propertyID) = 0;

	/**
	 * Locks the property for deleting.
	 */
	virtual Lock* lockPropertyForDeleting(unsigned long propertyID, common::EndOfLine const & endOfLine) = 0;
	
	/**
	 * Unlocks the property for deleting.
	 */
	virtual void unlockPropertyForDeleting(unsigned long propertyID) = 0;
	
	/**
	 * Locks the command for reading.
	 */
	virtual Lock* lockCommandForReading(unsigned long commandID, common::EndOfLine const & endOfLine) = 0;

	/**
	 * Unlocks the command for reading.
	 */
	virtual void unlockCommandForReading(unsigned long commandID) = 0;
	
	/**
	 * Locks the command for writing.
	 */
	virtual Lock* lockCommandForWriting(unsigned long commandID, common::EndOfLine const & endOfLine) = 0;

	/**
	 * Unlocks the command for writing.
	 */
	virtual void unlockCommandForWriting(unsigned long commandID) = 0;

	/**
	 * Locks the command for deleting.
	 */
	virtual Lock* lockCommandForDeleting(unsigned long commandID, common::EndOfLine const & endOfLine) = 0;
	
	/**
	 * Unlocks the command for deleting.
	 */
	virtual void unlockCommandForDeleting(unsigned long commandID) = 0;
	
	/**
	 * Locks the maps.
	 */
	virtual Lock* lockMaps(common::EndOfLine const & endOfLine) = 0;
	
	/**
	 * Unlocks the maps.
	 */
	virtual void unlockMaps() = 0;

	/**
	 * Verifies the locks
	 */
	virtual void verifyLocks(unsigned long maxID, int seconds) = 0;
};

}

#endif /*DATAPROVIDER_LOCKMANAGER_H_*/