/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ServantOperations.h"
#include "Database.h"
#include "ScopedLock.h"
#include <dataprovider/dataelement/Servant.h>
#include <dataprovider/dataelement/Driver.h>
#include <common/base/Convert.h>
#include <string>

using namespace common;
using namespace std;

namespace dataprovider {

ServantOperations::CircularServantException::CircularServantException()
 : Exception("ServantOperations::CircularServantException::CircularServantException", string("circular servant")) {
}

ServantOperations::MultipleServantsEnabledException::MultipleServantsEnabledException(std::string const & name, long index)
: Exception("ServantOperations::MultipleServantsEnabledException::MultipleServantsEnabledException", string("multiple servants enabled for '") + name + "'" + (index != -1 ? (" at index " + to<string>(index)) : "")),
	_name(name),
	_index(index) {
}

std::string const & ServantOperations::MultipleServantsEnabledException::getNName() const {
	return _name;
}

long ServantOperations::MultipleServantsEnabledException::getIndex() const {
	return _index;
}

ServantOperations::ServantChildException::ServantChildException(std::string const & name)
 : Exception("ServantOperations::ServantChildException::ServantChildException", string("bad type for child '") + name + "'") {
}

ServantOperations::ServantIndexException::ServantIndexException(std::string const & name, long index, long size)
 : Exception("ServantOperations::ServantIndexException::ServantIndexException", string("bad index <") + to<string>(index) + ("> for child '") + name + "'"),
 	_size(size) {
}

ServantOperations::ServantEnableException::ServantEnableException(std::string const & name, unsigned long ID, std::string const & nName, long index)
: Exception("ServantOperations::ServantEnableException::ServantEnableException", name + string(" cannot enable '") + nName + "'" + (index != -1 ? (" at index " + to<string>(index)) : "")),
	_name(name),
	_ID(ID),
	_nName(nName),
	_index(index) {
}

std::string const & ServantOperations::ServantEnableException::getName() const {
	return _name;
}

unsigned long ServantOperations::ServantEnableException::getID() const {
	return _ID;
}

std::string const & ServantOperations::ServantEnableException::getNName() const {
	return _nName;
}

long ServantOperations::ServantEnableException::getIndex() const {
	return _index;
}

ServantOperations::ServantOperations() {
}

void ServantOperations::setLockManager(LockManager* lockManager) {

	_lockManager = lockManager;
}

void ServantOperations::verifyCircularity(Servant* servant) {

	std::set<Servant*> servantsVisited;

	verifyCircularity(servant, servantsVisited, servant);
}

void ServantOperations::verifyCircularity(Servant* servant, std::set<Servant*>& servantsVisited, Servant* servantToVerify) {

	// return if already visited
	if (servantsVisited.find(servant) != servantsVisited.end()) {
		return;
	}

	servantsVisited.insert(servant);

	// iterating all the children
	HashMap<string, pair<bool, vector<Servant::Children> > > const & servantMap = servant->getServantMap();

	for (HashMap<string, pair<bool, vector<Servant::Children> > >::const_iterator sl = servantMap.begin(); sl != servantMap.end(); sl++) {
		for (vector<Servant::Children>::const_iterator l = sl->second.second.begin(); l != sl->second.second.end(); l++) {
			for (vector<unsigned long>::const_iterator s = l->children.begin(); s != l->children.end(); s++) {

				unsigned long currentServantID = *s;

				if (currentServantID != 0) {

					ScopedLock servantLock(_lockManager->lockServantForReading(currentServantID, eol));

					Servant* currentServant = Database::getInstance()->getServant(currentServantID);

					// circularity
					if (currentServant == servantToVerify) {
						throw CircularServantException();
					}

					verifyCircularity(currentServant, servantsVisited, servantToVerify);
				}
			}
		}
	}
}

void ServantOperations::verifyServantsEnabled(Servant* servant) {

	// iterating the servants
	HashMap<string, pair<bool, vector<Servant::Children> > > & servantMap = servant->getServantMap();

	for (HashMap<string, pair<bool, vector<Servant::Children> > >::iterator sl = servantMap.begin(); sl != servantMap.end(); sl++) {

		string nName = sl->first;
		bool isList = sl->second.first;

		int i = 0;
		for (vector<Servant::Children>::iterator l = sl->second.second.begin(); l != sl->second.second.end(); l++) {

			int counter = 0;

			// in case of list, the last enabled is set to -1, otherwise a single element.
			if (isList) {
				l->lastEnabled = -1;
			} else {
				l->lastEnabled = 0;
			}

			// iterating the potential servants
			int j = 0;
			for (vector<unsigned long>::iterator s = l->children.begin(); s != l->children.end(); s++) {

				unsigned long currentServantID = *s;

				if (currentServantID != 0) {

					ScopedLock servantLock(_lockManager->lockServantForReading(currentServantID, eol));

					Servant* currentServant = Database::getInstance()->getServant(currentServantID);

					if (currentServant->isEnabled()) {
						counter++;
						l->lastEnabled = j;
					}
				}

				j++;
			}

			// multiple servants enabled
			if (counter > 1) {
				long index = -1;
				if (isList) {
					index = i;
				}

				throw MultipleServantsEnabledException(nName, index);
			}

			i++;
		}
	}
}

unsigned long ServantOperations::getEnabledServantID(Servant* servant, std::string const & nName, long childIndex) {

	HashMap<string, pair<bool, vector<Servant::Children> > > & servantMap = servant->getServantMap();

	if (servantMap.find(nName)) {

		pair<bool, vector<Servant::Children> > & servants = servantMap.get(nName);

		if (!servants.first) {
			// not a list, verify that index is -1 and return the enabled child
			if (childIndex != -1) {
				throw ServantChildException(nName);
			} else {
				vector<unsigned long>& potentialServants = servantMap.get(nName).second.front().children;
				for (vector<unsigned long>::iterator s = potentialServants.begin(); s != potentialServants.end(); s++) {

					unsigned long currentServantID = *s;

					if (currentServantID != 0) {

						ScopedLock currentServantLock(_lockManager->lockServantForReading(currentServantID, eol));

						Servant* currentServant = Database::getInstance()->getServant(currentServantID);

						if (currentServant->isEnabled()) {
							return currentServantID;
						}
					}
				}
				// nothing enabled
				return 0;
			}

		} else {
			// is a list, verify that the index is coherent
			if ((long)servants.second.size() <= childIndex) {
				throw ServantIndexException(nName, childIndex, servants.second.size());
			}

			vector<unsigned long>& potentialServants = servantMap.get(nName).second[childIndex].children;
			for (vector<unsigned long>::iterator s = potentialServants.begin(); s != potentialServants.end(); s++) {

				unsigned long currentServantID = *s;

				if (currentServantID != 0) {

					ScopedLock currentServantLock(_lockManager->lockServantForReading(currentServantID, eol));

					Servant* currentServant = Database::getInstance()->getServant(currentServantID);

					if (currentServant->isEnabled()) {
						return currentServantID;
					}
				}
			}
		}
	}

	return 0;
}

std::set<unsigned long> ServantOperations::getServantsForEnable(Servant* servant, unsigned long& firstDriverID) {

	std::set<Servant*> servantsVisited;
	std::set<unsigned long> result;

	if (!servant->isEnabled()) {
		result.insert(servant->getID());
	}

	getServantsForEnable(servant, servantsVisited, result, firstDriverID);

	return result;
}

std::set<unsigned long> ServantOperations::getServantsForDisable(Servant* servant, unsigned long& firstDriverID) {

	std::set<Servant*> servantsVisited;
	std::set<unsigned long> result;

	result.insert(servant->getID());

	getServantsForDisable(servant, servantsVisited, result, firstDriverID);

	return result;
}

void ServantOperations::getServantsForEnable(Servant* servant, std::set<Servant*>& servantsVisited, std::set<unsigned long>& result, unsigned long& firstDriverID) {

	// exiting immediately if already visited
	if (servantsVisited.find(servant) != servantsVisited.end()) {
		return;
	}

	servantsVisited.insert(servant);

	// test first driver ID
	if (servant->getServantType() == Servant::DRIVER && firstDriverID == 0) {
		firstDriverID = servant->getID();
	}

	// iterating all the children
	HashMap<string, pair<bool, vector<Servant::Children> > > & servantMap = servant->getServantMap();

	for (HashMap<string, pair<bool, vector<Servant::Children> > >::const_iterator sl = servantMap.begin(); sl != servantMap.end(); sl++) {

		// memorising the child index
		int childIndex = 0;
		if (!sl->second.first) {
			childIndex = -1;
		}

		for (vector<Servant::Children>::const_iterator l = sl->second.second.begin(); l != sl->second.second.end(); l++) {

			// getting the last enabled
			if (l->children.size() > 1) {
				if (l->lastEnabled == -1) {
					// we cannot enable
					throw ServantEnableException(servant->getName(), servant->getID(), sl->first, childIndex);

				} else {

					unsigned long currentServantID = l->children[l->lastEnabled];

					if (currentServantID != 0) {

						ScopedLock servantLock(_lockManager->lockServantForReading(currentServantID, eol));

						Servant* currentServant = Database::getInstance()->getServant(currentServantID);

						if (!currentServant->isEnabled()) {
							result.insert(currentServantID);
						}
					}
				}

			} else if (l->children.size() == 1) {

				unsigned long currentServantID = l->children[0];

				if (currentServantID != 0) {

					ScopedLock servantLock(_lockManager->lockServantForReading(currentServantID, eol));

					Servant* currentServant = Database::getInstance()->getServant(currentServantID);

					if (!currentServant->isEnabled()) {
						result.insert(currentServantID);
					}
				}
			}

			// iterating the children
			for (vector<unsigned long>::const_iterator s = l->children.begin(); s != l->children.end(); s++) {

				unsigned long currentServantID = *s;

				if (currentServantID != 0) {

					ScopedLock servantLock(_lockManager->lockServantForReading(currentServantID, eol));

					Servant* currentServant = Database::getInstance()->getServant(currentServantID);

					getServantsForEnable(currentServant, servantsVisited, result, firstDriverID);
				}
			}

			if (sl->second.first) {
				childIndex++;
			}
		}
	}
}

std::set<unsigned long> ServantOperations::getServantsForEnableDriver(Driver* driver) {
	std::set<unsigned long> result;

	getServantsForEnableDriver(driver, result);

	return result;
}

void ServantOperations::getServantsForEnableDriver(Driver* driver, std::set<unsigned long>& result) {

	// getting the parent of servant
	unsigned long fatherID = driver->getFatherID();

	if (fatherID != 0) {

		Driver* father = dynamic_cast<Driver*>(Database::getInstance()->getServant(fatherID));
		if (father != 0) {
			// iterating the children to find an enabled child
			HashMap<string, pair<bool, vector<Servant::Children> > > & servantMap = father->getServantMap();

			result.insert(fatherID);
			getServantsForEnableDriver(father, result);

		} else {
			err << "Cannot get the driver father" << eol;
		}
	}
}

void ServantOperations::getServantsForDisable(Servant* servant, std::set<Servant*>& servantsVisited, std::set<unsigned long>& result, unsigned long& firstDriverID) {

	// exiting immediately if already visited
	if (servantsVisited.find(servant) != servantsVisited.end()) {
		return;
	}

	servantsVisited.insert(servant);

	if (servant->isEnabled()) {
		result.insert(servant->getID());
	}

	// test first driver ID
	if (servant->getServantType() == Servant::DRIVER && firstDriverID == 0) {
		firstDriverID = servant->getID();
	}

	// iterating all the children
	HashMap<string, pair<bool, vector<Servant::Children> > > & servantMap = servant->getServantMap();

	for (HashMap<string, pair<bool, vector<Servant::Children> > >::const_iterator sl = servantMap.begin(); sl != servantMap.end(); sl++) {

		for (vector<Servant::Children>::const_iterator l = sl->second.second.begin(); l != sl->second.second.end(); l++) {

			// iterating the children
			for (vector<unsigned long>::const_iterator s = l->children.begin(); s != l->children.end(); s++) {

				unsigned long currentServantID = *s;

				if (currentServantID != 0) {

					ScopedLock servantLock(_lockManager->lockServantForReading(currentServantID, eol));

					Servant* currentServant = Database::getInstance()->getServant(currentServantID);

					getServantsForDisable(currentServant, servantsVisited, result, firstDriverID);
				}
			}
		}
	}
}

std::set<unsigned long> ServantOperations::getServantsForDisableDriver(Driver* driver) {
	std::set<unsigned long> result;

	getServantsForDisableDriver(driver, result);

	return result;
}

void ServantOperations::getServantsForDisableDriver(Driver* driver, std::set<unsigned long>& result) {

	// getting the IDs
	unsigned long driverID = driver->getID();
	unsigned long fatherID = driver->getFatherID();

	if (fatherID != 0) {

		Driver* father = dynamic_cast<Driver*>(Database::getInstance()->getServant(fatherID));
		if (father != 0) {
			// iterating the children to find an enabled child
			HashMap<string, pair<bool, vector<Servant::Children> > > & servantMap = father->getServantMap();

			bool disable = true;

			for (HashMap<string, pair<bool, vector<Servant::Children> > >::const_iterator sl = servantMap.begin(); sl != servantMap.end(); sl++) {

				for (vector<Servant::Children>::const_iterator l = sl->second.second.begin(); l != sl->second.second.end(); l++) {

					// searching for servant
					for (vector<unsigned long>::const_iterator s = l->children.begin(); s != l->children.end(); s++) {

						unsigned long currentServantID = *s;

						// testing only the other servants than driver
						if (currentServantID != 0 && currentServantID != driverID) {

							ScopedLock servantLock(_lockManager->lockServantForReading(currentServantID, eol));

							Servant* currentServant = Database::getInstance()->getServant(currentServantID);

							if (currentServant->isEnabled()) {
								disable = false;
								break;
							}
						}
					}
				}
			}

			// if no child is enabled, then the driver can be disabled
			if (disable) {
				result.insert(fatherID);
				getServantsForDisableDriver(father, result);
			}

		} else {
			err << "Cannot get the driver father" << eol;
		}
	}
}

void ServantOperations::verifyServantEnabled(Servant* servant, const std::set<unsigned long>& servantsToEnable) {

	// iterating all the children
	HashMap<string, pair<bool, vector<Servant::Children> > > & servantMap = servant->getServantMap();

	for (HashMap<string, pair<bool, vector<Servant::Children> > >::const_iterator sl = servantMap.begin(); sl != servantMap.end(); sl++) {

		int childIndex = 0;
		if (!sl->second.first) {
			childIndex = -1;
		}

		for (vector<Servant::Children>::const_iterator l = sl->second.second.begin(); l != sl->second.second.end(); l++) {

			int counter = 0;

			// searching for servant
			for (vector<unsigned long>::const_iterator s = l->children.begin(); s != l->children.end(); s++) {

				if (servantsToEnable.find(*s) != servantsToEnable.end()) {
					counter++;
				}
			}

			// problem if more than one servant must be enabled
			if (counter > 1) {
				throw ServantEnableException(servant->getName(), servant->getID(), sl->first, childIndex);
			}

			// then only the servant to enable can be in enabled state
			if (counter == 1) {
				for (vector<unsigned long>::const_iterator s = l->children.begin(); s != l->children.end(); s++) {

					unsigned long currentServantID = *s;

					if (currentServantID != 0) {

						ScopedLock servantLock(_lockManager->lockServantForReading(currentServantID, eol));

						Servant* currentServant = Database::getInstance()->getServant(currentServantID);

						if (currentServant->isEnabled() && servantsToEnable.find(currentServantID) == servantsToEnable.end()) {
							throw ServantEnableException(servant->getName(), servant->getID(), sl->first, childIndex);
						}
					}
				}
			}

			if (sl->second.first) {
				childIndex++;
			}
		}

	}
}

}
