/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_SERVANTOPERATIONS_H_
#define DATAPROVIDER_SERVANTOPERATIONS_H_

#include <common/base/Exception.h>
#include <set>

namespace dataprovider {

class Servant;
class Driver;
class LockManager;

/**
 * Class defining servant operations.
 */
class ServantOperations {

public:
	/**
	 * Exception for circular servant.
	 */
	class CircularServantException : public common::Exception {

	public:
		/**
	 	 * Constructor.
	 	 */
		CircularServantException();
	};

	/**
	 * Exception for multiple servant enabled.
	 */
	class MultipleServantsEnabledException : public common::Exception {

	public:
		/**
	 	 * Constructor.
	 	 */
		MultipleServantsEnabledException(std::string const & name, long index);

		/**
	 	 * Gets the n-name.
	 	 */
		std::string const & getNName() const;

		/**
	 	 * Gets the index.
	 	 */
		long getIndex() const;

	private:
		std::string _name;
		long _index;
	};

	/**
	 * Exception for servant child.
	 */
	class ServantChildException : public common::Exception {

	public:
		/**
	 	 * Constructor.
	 	 */
		ServantChildException(std::string const & name);
	};

	/**
	 * Exception for servant index.
	 */
	class ServantIndexException : public common::Exception {

	public:
		/**
	 	 * Constructor.
	 	 */
		ServantIndexException(std::string const & name, long index, long size);

		long getSize() const;

	private:
		long _size;
	};

	/**
	 * Exception for servant enable.
	 */
	class ServantEnableException : public common::Exception {

	public:
		/**
	 	 * Constructor.
	 	 */
		ServantEnableException(std::string const & name, unsigned long ID, std::string const & nName, long index);

		/**
	 	 * Gets the name.
	 	 */
		std::string const & getName() const;

		/**
	 	 * Gets the ID.
	 	 */
		unsigned long getID() const;

		/**
	 	 * Gets the n-name.
	 	 */
		std::string const & getNName() const;

		/**
	 	 * Gets the index.
	 	 */
		long getIndex() const;

	private:
		std::string _name;
		unsigned long _ID;
		std::string _nName;
		long _index;
	};

	/**
 	 * Constructor.
 	 */
	ServantOperations();

	/**
	 * Sets the lock manager.
	 */
	void setLockManager(LockManager* lockManager);

	/**
	 * Verifies the circular servants.
	 * throw CircularServantException
	 */
	void verifyCircularity(Servant* servant);

	/**
	 * Verifies that the servants are correctly enabled.
	 * throw MultipleServantsEnabledException
	 */
	void verifyServantsEnabled(Servant* servant);

	/**
	 * Returns the enabled servant from the vector of potential ones.
	 * throw ServantChildException, ServantIndexException
	 */
	unsigned long getEnabledServantID(Servant* servant, std::string const & nName, long childIndex);

	/**
	 * Gets the servants for enable.
	 * throw ServantEnableException
	 */
	std::set<unsigned long> getServantsForEnable(Servant* servant, unsigned long& firstDriverID);

	/**
	 * Gets the servants for enable.
	 */
	std::set<unsigned long> getServantsForEnableDriver(Driver* driver);

	/**
	 * Gets the servants for disable.
	 */
	std::set<unsigned long> getServantsForDisable(Servant* servant, unsigned long& firstDriverID);

	/**
	 * Gets the servants for disable.
	 */
	std::set<unsigned long> getServantsForDisableDriver(Driver* driver);

	/**
	 * Returns true if the servant can be enabled
	 * throw ServantEnableException
	 */
	void verifyServantEnabled(Servant* servant, const std::set<unsigned long>& servantsToEnable);

private:
	void verifyCircularity(Servant* servant, std::set<Servant*>& servantsVisited, Servant* servantToVerify);
	void getServantsForEnable(Servant* servant, std::set<Servant*>& servantsVisited, std::set<unsigned long>& result, unsigned long& firstDriverID);
	void getServantsForEnableDriver(Driver* driver, std::set<unsigned long>& result);
	void getServantsForDisable(Servant* servant, std::set<Servant*>& servantsVisited, std::set<unsigned long>& result, unsigned long& firstDriverID);
	void getServantsForDisableDriver(Driver* driver, std::set<unsigned long>& result);

	LockManager * _lockManager;
};

}

#endif