/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Command.h"
#include "Servant.h"
#include "Driver.h"
#include <dataprovider/database/Database.h>

using namespace std;
using namespace common;

namespace dataprovider {

Command::Command(std::string const & name) :
	DataElement(),
	_name(name),
	_servantID(0),
	_state(INACTIVE),
	_runningProcessorID(0),
	_runningThreadID(0),
	_progression(0),
	_preRunCalculatedDuration(0),
	_logID(""),
	_fromCommandZone(false) {
}

Command::Command(Command const & content) :
	DataElement(content),
	_name(content._name),
	_servantID(content._servantID),
	_state(content._state),
	_runningProcessorID(content._runningProcessorID),
	_runningThreadID(content._runningThreadID),
	_startDate(content._startDate),
	_lastActive(content._lastActive),
	_duration(content._duration),
	_progression(content._progression),
	_preRunCalculatedDuration(content._preRunCalculatedDuration),
	_logID(content._logID),
	_fromCommandZone(content._fromCommandZone) {
}

Command::~Command() {
}

string const & Command::getName() const {

	return _name;
}

DataElement * Command::clone() const {

	return new Command(*this);
}

void Command::setServantID(unsigned long servantID) {

	_servantID = servantID;
}

void Command::setState(common::CommandState state) {

	// 3 cases to test:
	// - start -> stop
	// - start -> pause -> restart
	// - start -> pause -> stop

	// It is necessary for some state to check the previous state.
	if (state == ACTIVE) {

		// If previous state is RESUMING then the command was paused so that there is no reset of the start date and duration.
		if (_state != RESUMING) {
			_startDate = Date();
			_duration = {};
		}

		// Set the last active date in any case.
		_lastActive = {};
	}
	else if (state == PAUSED) {
		// Update the duration.
		_duration = _duration + (Date() - _lastActive);
	}
	else if (state == STOPPING) {
		// If state PAUSED is followed with STOPPING then this is a tip to avoid to count time during the pause.
		if (_state == PAUSED) {
			_lastActive = {};
		}
	}
	else if (state == INACTIVE) {
		// Update the duration.
		_duration = _duration + (Date() - _lastActive);
	}

	_state = state;

	//common::trace << "COMMAND " << _servantID << " " << _name << " sets state " << getStateString() << " duration " << _duration.getMilliseconds() << common::eol;
}

void Command::setRunningProcessorID(unsigned long processorID) {
	_runningProcessorID = processorID;
}

void Command::setRunningThreadID(unsigned long threadID) {
	_runningThreadID = threadID;
}

void Command::setServantType(std::string const & type) {
	_servantType = type;
}

unsigned long Command::getServantID() const {

	return _servantID;
}

common::CommandState Command::getState() const {

	return _state;
}

unsigned long Command::getRunningProcessorID() const {

	return _runningProcessorID;
}

unsigned long Command::getRunningThreadID() const {

	return _runningThreadID;
}

std::string const & Command::getServantType() const {

	return _servantType;
}

double Command::getElapsedTimeMs() const {
	//common::trace << "COMMAND " << _servantID << " " << _name << " get elapsed time " << _duration.getMilliseconds() << common::eol;

	return _duration.getMilliseconds();
}

string Command::getStateString() const {

	if (_state == ACTIVE) {
		return "active";

	} else if (_state == INACTIVE) {
		return "inactive";

	} else if (_state == PAUSED) {
		return "paused";

	} else if (_state == PAUSING) {
		return "pausing";

	} else if (_state == STOPPED) {
		return "stopped";

	} else if (_state == STOPPING) {
		return "stopping";

	} else if (_state == RESUMING) {
		return "restarting";

	} else if (_state == PENDING) {
		return "pending";

	} else if (_state == STARTING) {
		return "starting";
	}

	return "unknown";
}

void Command::setProgression(double progression) {
	_progression = progression;
}

double Command::getProgression() const {
	return _progression;
}

double Command::getRemainingTimeMs() const {
	double rate = _progression / getElapsedTimeMs();
	return (100 - _progression) / rate;
}

double Command::getDurationMs() const {
	return getElapsedTimeMs() + getRemainingTimeMs();
}

void Command::setPreRunCalculatedDurationMs(double duration) {
	_preRunCalculatedDuration = duration;
}

double Command::getPreRunCalculatedDurationMs() const {
	return _preRunCalculatedDuration;
}

int64 Command::getStartDate() const {
	return _startDate.getTimeUs();
}

void Command::setLogID(const std::string& logID) {
	_logID = logID;
}

const std::string& Command::getLogID() const {
	return _logID;
}

void Command::setFromCommandZone(bool fromCommandZone) {
	_fromCommandZone = fromCommandZone;
}

bool Command::isFromCommandZone() const {
	return _fromCommandZone;
}

std::string Command::toString() const {

	return string("command '") + getName() + "' of " + Database::getInstance()->what(_servantID);
}

}
