/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_COMMAND_H_
#define DATAPROVIDER_COMMAND_H_

#include "DataElement.h"

#include <string>
#include <common/base/Date.h>

namespace dataprovider {

class Servant;

/**
 * Class defining a command.
 */
class Command : public DataElement {

public:
	/**
	 * Constructor with name.
	 */
	Command(std::string const & name);

	/**
	 * Copy constructor.
	 */
	Command(Command const & content);

	/**
	 * Virtual destructor.
	 */
	virtual ~Command();

	/**
	 * Clones the element by keeping the references to the sub-elements.
	 */
	virtual DataElement * clone() const;

	/**
	 * Sets the servant.
	 */
	void setServantID(unsigned long servantID);

	/**
	 * Sets the state.
	 */
	void setState(common::CommandState state);

	/**
	 * Sets the running processor ID.
	 */
	void setRunningProcessorID(unsigned long processorID);

	/**
	 * Sets the thread ID.
	 */
	void setRunningThreadID(unsigned long threadID);

	/**
	 * Sets the servant type.
	 */
	void setServantType(std::string const & type);

	/**
	 * Gets the name.
	 */
	std::string const & getName() const;

	/**
	 * Gets the servant.
	 */
	unsigned long getServantID() const;

	/**
	 * Gets the state.
	 */
	common::CommandState getState() const;

	/**
	 * Gets the running processor ID.
	 */
	unsigned long getRunningProcessorID() const;

	/**
	 * Gets the running thread ID.
	 */
	unsigned long getRunningThreadID() const;

	/**
	 * Gets the servant type.
	 */
	std::string const & getServantType() const;

	/**
	 * Returns the elapsed time of the command in milliseconds.
	 */
	double getElapsedTimeMs() const;

	/**
	 * Returns stringified form of the state.
	 */
	std::string getStateString() const;

	/**
	 * Sets the progression.
	 */
	void setProgression(double progression);

	/**
	 * Gets the progression.
	 */
	double getProgression() const;

	/**
	 * Gets the remaining time.
	 */
	double getRemainingTimeMs() const;

	/**
	 * Gets the duration.
	 */
	double getDurationMs() const;

	/**
	 * Sets the pre-run calculated duration.
	 */
	void setPreRunCalculatedDurationMs(double duration);

	/**
	 * Gets the pre-run calculated duration.
	 */
	double getPreRunCalculatedDurationMs() const;

	/**
	 * Gets the start date.
	 */
	int64 getStartDate() const;

	/**
	 * Sets the log ID.
	 */
	void setLogID(const std::string& logID);

	/**
	 * Gets the log ID.
	 */
	const std::string& getLogID() const;

	/**
	 * Sets the command from the command zone.
	 */
	void setFromCommandZone(bool fromCommandZone);

	/**
	 * Returns true if the command comes from the command zone.
	 */
	bool isFromCommandZone() const;

	/**
	 * Prints the servant.
	 */
	virtual std::string toString() const;

private:
	std::string _name;
	unsigned long _servantID;
	common::CommandState _state;
	unsigned long _runningProcessorID;
	unsigned long _runningThreadID;
	std::string _servantType;
	common::Date _startDate;
	common::Date _lastActive;
	common::Duration _duration;
	double _progression;
	double _preRunCalculatedDuration;
	std::string _logID;
	bool _fromCommandZone;
};

}

#endif /*DATAPROVIDER_COMMAND_H_*/
