/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_DRIVER_H_
#define DATAPROVIDER_DRIVER_H_

#include "Servant.h"
#include <common/base/CommonTypes.h>
#include <common/base/Date.h>

namespace dataprovider {

class Protocol;

/**
 * Class defining a Driver.
 */
class Driver : public Servant {

public:
	/**
	 * Exception for multiple parent.
	 */
	class MultipleParentException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		MultipleParentException(std::string const & name);
	};

	/**
	 * Constructor.
	 */
	Driver(std::string const & driverType, std::string const & name, bool isCommand, bool isVisible, bool isPausable);

	/**
	 * Copy constructor.
	 */
	Driver(Driver const & content);

	/**
	 * Virtual destructor.
	 */
	virtual ~Driver();

	/**
	 * Gets the servant type.
	 */
	virtual Type getServantType() const;

	/**
	 * Gets the driver type.
	 */
	virtual std::string const & getDriverType() const;

	/**
	 * Sets the driver father.
	 */
	void setFatherInfo(unsigned long fatherID, std::string const & childName, long channel);

	/**
	 * Gets the father ID.
	 */
	unsigned long getFatherID();

	/**
	 * Gets the father child name.
	 */
	std::string const & getFatherChildName() const;

	/**
	 * Gets the channel.
	 */
	long getChannel();

	/**
	 * Returns true if it has no father.
	 */
	bool isRoot() const;

	/**
	 * Sets the simulated state.
	 */
	void setMode(common::DriverMode mode);

	/**
	 * Returns true if is simulated.
	 */
	common::DriverMode getMode() const;

	/**
	 * Sets the driver state.
	 */
	void setState(common::CommandState state);

	/**
	 * Sets the running processor ID.
	 */
	void setRunningProcessorID(unsigned long processorID);

	/**
	 * Sets the thread ID.
	 */
	void setRunningThreadID(unsigned long threadID);

	/**
	 * Returns the state of the driver.
	 */
	common::CommandState getState() const;

	/**
	 * Gets the running processor ID.
	 */
	unsigned long getRunningProcessorID() const;

	/**
	 * Gets the running thread ID.
	 */
	unsigned long getRunningThreadID() const;

	/**
	 * Resets the elapsed simulation time.
	 */
	void resetElapsedSimulationTime();

	/**
	 * Gets the elapsed simulation time.
	 */
	double getElapsedSimulationTimeMs() const;

	/**
	 * Clones the element by keeping the references to the sub-elements.
	 */
	virtual DataElement * clone() const;

	/**
	 * Prints the servant.
	 */
	virtual std::string toString() const;

private:
	std::string _driverType;
	bool _fatherInitialised;
	unsigned long _fatherID;
	std::string _fatherChildName;
	long _channel;
	common::DriverMode _mode;
	common::Date _beginSimulationDate;
	common::CommandState _state;
	unsigned long _runningProcessorID;
	unsigned long _runningThreadID;
};

}

#endif /*DATAPROVIDER_DRIVER_H_*/
