/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DynamicProperty.h"
#include "ElementaryProperty.h"
#include "IcsArrayProperty.h"
#include "DataProvider/Property/AnyProperty.h"
#include <dataprovider/database/IndexGenerator.h>
#include <common/base/TypeToString.h>
#include <common/output/OutputStream.h>

namespace dataprovider {

template<typename ValueType>
DynamicProperty<ValueType>::DynamicProperty(std::string const & name, bool generated, bool saveValue) :
	BaseDynamicProperty(name, generated, saveValue, "") {
}

template<typename ValueType>
DynamicProperty<ValueType>::DynamicProperty(std::string const & name, bool generated, bool saveValue, std::string const & format) :
	BaseDynamicProperty(name, generated, saveValue, format) {
}

template<typename ValueType>
DynamicProperty<ValueType>::DynamicProperty(DynamicProperty const & content) :
	BaseDynamicProperty(content) {
}

template<typename ValueType>
DynamicProperty<ValueType>::~DynamicProperty() {
}

template<typename ValueType>
DataElement * DynamicProperty<ValueType>::clone() const {
	
	return new DynamicProperty<ValueType>(*this);
}


template<typename ValueType>
DataProperty * DynamicProperty<ValueType>::createProperty(bool generated, bool saveValue) const {
	
	DataProperty* property = 0;
	
	if (typeid(ValueType) == typeid(uint16*)) {
		property = new IcsArrayProperty<uint16>(getName(), UINT16ARRAY_PROPERTY, false, "");
	} else if (typeid(ValueType) == typeid(int16*)) {
		property = new IcsArrayProperty<int16>(getName(), INT16ARRAY_PROPERTY, false, "");
	} else if (typeid(ValueType) == typeid(uint32*)) {
		property = new IcsArrayProperty<uint32>(getName(), UINT32ARRAY_PROPERTY, false, "");
	} else if (typeid(ValueType) == typeid(int32*)) {
		property = new IcsArrayProperty<int32>(getName(), INT32ARRAY_PROPERTY, false, "");
	} else if (typeid(ValueType) == typeid(uint64*)) {
		property = new IcsArrayProperty<uint64>(getName(), UINT64ARRAY_PROPERTY, false, "");
	} else if (typeid(ValueType) == typeid(int64*)) {
		property = new IcsArrayProperty<int64>(getName(), INT64ARRAY_PROPERTY, false, "");
	} else if (typeid(ValueType) == typeid(float32*)) {
		property = new IcsArrayProperty<float32>(getName(), FLOAT32ARRAY_PROPERTY, false, "");
	} else if (typeid(ValueType) == typeid(float64*)) {
		property = new IcsArrayProperty<float64>(getName(), FLOAT64ARRAY_PROPERTY, false, "");
	} else {
		property = new ElementaryProperty<ValueType>(getName(), generated, BaseProperty::saveValue() && saveValue, getFormat());
	}

	return property;
}

template<typename ValueType>
std::vector<unsigned long> DynamicProperty<ValueType>::downsize(long size) {

	std::vector<unsigned long> removedPropertyIDs;

	// getting the removed property IDs
	if (size < (long)_propertyIDs.size()) {
		for (long i = size; i < (long)_propertyIDs.size(); i++) {
			removedPropertyIDs.push_back(_propertyIDs[i]);
		}
		
	} else {
		common::err << "Cannot downsize the dynamic property to a greater size" << common::eol;
	}
	
	// resizing and inserting 0 if necessary
	_propertyIDs.resize(size, 0);
	
	return removedPropertyIDs;
}

template<typename ValueType>
std::string DynamicProperty<ValueType>::getType() const {
	return common::TypeToString<ValueType>::convert();
}

template<typename ValueType>
std::string DynamicProperty<ValueType>::toString() const {

	return std::string("property '") + getName() + "' (dynamic <" + getType() + ">) of " + getParentString();
}

}
