/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Servant.h"
#include "Command.h"
#include <dataprovider/database/Database.h>
#include <common/base/Convert.h>
#include <common/output/OutputStream.h>

using namespace std;
using namespace common;

namespace dataprovider {

Servant::ServantIndexException::ServantIndexException(std::string const & name, long index, long size)
 : Exception("Servant::ServantIndexException::ServantIndexException", string("bad index <") + to<string>(index) + ("> for child '") + name + "'"),
 	_size(size) {
}

long Servant::ServantIndexException::getSize() const {
	return _size;
}

Servant::NoSuchServantChildException::NoSuchServantChildException(std::string const & name)
 : Exception("Servant::NoSuchServantChildException::NoSuchServantChildException", string("no such servant child having name '") + name + "'") {
}

Servant::Servant(std::string const & name, const std::string& clientType, bool isCommand, bool isVisible, bool isPausable, bool isFavourite)
: PropertyHolder(),
	_serverType(NOMAD),
	_name(name),
	_isCommand(isCommand),
	_isVisible(isVisible),
	_enabled(false),
	_isPausable(isPausable),
	_isFavourite(isFavourite),
	_clientType(clientType) {
}

Servant::Servant(Servant const & content) :
	PropertyHolder(content),
	_serverType(content._serverType),
	_name(content._name),
	_servantMap(content._servantMap),
	_family(content._family),
	_subfamily(content._subfamily),
	_isCommand(content._isCommand),
	_isVisible(content._isVisible),
	_commands(content._commands),
	_enabled(content._enabled),
	_isPausable(content._isPausable),
	_isFavourite(content._isFavourite),
	_clientType(content._clientType) {
}

Servant::~Servant() {
	// No need to iterate the properties to delete them.
}

void Servant::setServerType(common::ServerType serverType) {
	_serverType = serverType;
}

void Servant::setName(std::string const & name) {

	_name = name;
}

void Servant::setID(unsigned long ID) {
	DataElement::setID(ID);
}

common::ServerType Servant::getServerType() const {
	return _serverType;
}

std::string const & Servant::getName() const {

	return _name;
}

void Servant::setClientType(const std::string & type) {

	_clientType = type;
}

std::string const & Servant::getClientType() const {
	return _clientType;
}

void Servant::addCommand(Command * command) {

	_commands.insert(command->getName(), command->getID());
	command->setServantID(getID());
}

common::HashMap<std::string, unsigned long> const & Servant::getCommands() const {

	return _commands;
}

unsigned long Servant::getCommand(std::string const & commandName) const {

	try {
		return _commands.get(commandName);

	} catch (Exception & e) {
		return 0;
	}
}

void Servant::addServantList(std::string const & nName, long size) {

	if (!_servantMap.find(nName)) {
		pair<bool, std::vector<Children> > item;
		item.first = true;
		item.second.resize(size);
		_servantMap.insert(nName, item);

		_servantsToSave.insert(nName);
	}
}

void Servant::addServantToList(std::string const & nName, int position, Servant * servant) {

	std::vector<Children> & servants = _servantMap[nName].second;
	unsigned long servantID = 0;
	if (servant != 0) {
		servantID = servant->getID();
	}
	servants[position].children.push_back(servantID);

	// adding servant in the map
	if (servantID != 0) {
		Database::getInstance()->addParentToServant(getID(), servantID);
	}
}

void Servant::addServant(std::string const & nName, Servant * servant, bool toBeSaved) {

	if (!_servantMap.find(nName)) {
		pair<bool, std::vector<Children> > item;
		item.first = false;
		_servantMap.insert(nName, item);

		if (toBeSaved) {
			_servantsToSave.insert(nName);
		}
	}

	std::vector<Children> & servants = _servantMap[nName].second;
	servants.resize(1);

	unsigned long servantID = 0;
	if (servant != 0) {
		servantID = servant->getID();
	}

	if (servants[0].children.size() > 0 && servants[0].children.front() == 0) {
		// replacing null controller
		servants[0].children.front() = servantID;
	} else {
		// adding servant
		servants[0].children.push_back(servantID);
	}

	// adding servant in the map
	if (servantID != 0) {
		Database::getInstance()->addParentToServant(getID(), servantID);
	}
}

bool Servant::servantToBeSaved(const std::string nName) const {
	return _servantsToSave.find(nName) != _servantsToSave.end();
}

void Servant::assignServant(std::string const & nName, int position, unsigned long servantID) {

	if (position == -1) {

		if (!_servantMap.find(nName)) {
			pair<bool, std::vector<Children> > item;
			item.first = false;
			_servantMap.insert(nName, item);
		}

		std::vector<Children> & servants = _servantMap[nName].second;
		servants.resize(1);
		servants[0].children.clear();
		servants[0].children.push_back(servantID);

	} else {

		if (!_servantMap.find(nName)) {
			pair<bool, std::vector<Children> > item;
			item.first = true;
			_servantMap.insert(nName, item);
		}
		std::vector<Children> & servants = _servantMap[nName].second;

		if (position >= servants.size()) {
			servants.resize(position + 1);
		}

		servants[position].children.clear();
		servants[position].children.push_back(servantID);
	}

	// adding servant in the map
	if (servantID != 0) {
		Database::getInstance()->addParentToServant(getID(), servantID);
	}
}

unsigned long Servant::removeServant(const std::string& nName, long index) {

	if (_servantMap.find(nName)) {

		std::vector<Children> & servants = _servantMap.get(nName).second;

		if (index >= (long)servants.size()) {

			long previousSize = servants.size();
			servants.resize(index + 1);

			for (int i = previousSize; i < index + 1; i++) {
				servants[previousSize].lastEnabled = -1;
			}
		}

		// if index is -1 then we get the element at 0
		long servantsIndex = 0;
		if (index >= 0) {
			servantsIndex = index;
		}

		vector<unsigned long> & children = servants[servantsIndex].children;

		// only controllers can be removed, so the size is 1
		if (children.size() == 1) {

			unsigned long servantID = children.front();
			children.front() = 0;

			// removing servant in the map
			if (servantID != 0) {
				Database::getInstance()->removeParentFromServant(getID(), servantID);
			}

			return servantID;
		}
	}

	return 0;
}

common::HashMap<std::string, std::pair<bool, std::vector<Servant::Children> > > & Servant::getServantMap() {

	return _servantMap;
}

void Servant::setServantMap(common::HashMap<std::string, std::pair<bool, std::vector<Children> > > const & servantMap) {

	_servantMap = servantMap;

	// iterating all the children
	for (HashMap<string, pair<bool, vector<Children> > >::iterator sl = _servantMap.begin(); sl != _servantMap.end(); sl++) {

		for (vector<Children>::iterator l = sl->second.second.begin(); l != sl->second.second.end(); l++) {

			// searching for servant
			for (vector<unsigned long>::iterator s = l->children.begin(); s != l->children.end(); s++) {
				Database::getInstance()->addParentToServant(getID(), *s);
			}
		}
	}
}

std::vector<unsigned long> Servant::getServant(std::string const & nName, long childIndex) {

	if (_servantMap.find(nName)) {

		pair<bool, std::vector<Children> > & servants = _servantMap.get(nName);

		if (!servants.first) {
			if (childIndex != -1) {
				err << "Cannot request index != -1 for a elementary child" << eol;
			} else {
				return _servantMap.get(nName).second.front().children;
			}

		} else {

			if ((long)servants.second.size() <= childIndex) {
				throw ServantIndexException(nName, childIndex, servants.second.size());
			}
		}

		return servants.second[childIndex].children;
	}

	return vector<unsigned long>();
}

long Servant::getServantListSize(std::string const & nName) const {

	if (_servantMap.find(nName)) {

		pair<bool, std::vector<Children> > const & servants = _servantMap.get(nName);

		if (!servants.first) {
			err << "Cannot request index != -1 for a elementary child" << eol;

		} else {
			return servants.second.size();
		}
	}

	return 0;
}

void Servant::resizeServantList(std::string const & nName, long size) {

	if (_servantMap.find(nName)) {

		pair<bool, std::vector<Children> > & servants = _servantMap.get(nName);

		if (!servants.first) {
			err << "Cannot request index != -1 for a elementary child" << eol;

		} else {
			servants.second.resize(size);
		}
	}
}

bool Servant::isCommand() const {
	return _isCommand;
}

void Servant::setCommand(bool value) {
	_isCommand = value;
}

void Servant::setVisible(bool value) {

	if (_family != "hidden" && _family != "") {
		_isVisible = value;
	}
}

bool Servant::isVisible() const {
	return _isVisible;
}

void Servant::setFamily(const std::string& family, const std::string& subfamily) {
	_family = family;
	_subfamily = subfamily;

	if (family != "hidden" && family != "") {
		_isVisible = true;
	} else {
		_isVisible = false;
	}
}

const std::string& Servant::getFamily() const {
	return _family;
}

const std::string& Servant::getSubfamily() const {
	return _subfamily;
}

void Servant::setInitialised(bool initialised) {

	if (!_enabled && initialised) {
		//err << "warning : initialising servant '" << _name << "' already disabled!" << eol;
	} else {
		PropertyHolder::setInitialised(initialised);
	}
}

void Servant::enable() {
	_enabled = true;
}

void Servant::disable() {
	setInitialised(false);
	_enabled = false;
}

bool Servant::isEnabled() const {
	return _enabled;
}

bool Servant::isPausable() const {
	return _isPausable;
}

void Servant::setFavourite(bool value) {
	_isFavourite = value;
}

bool Servant::isFavourite() const {
	return _isFavourite;
}

long Servant::getLastEnabledServant(const std::string& nName, long childIndex) const {

	if (_servantMap.find(nName)) {

		pair<bool, std::vector<Children> > const & servants = _servantMap.get(nName);

		if (!servants.first) {
			if (childIndex != -1) {
				err << "Cannot request index != -1 for a elementary child" << eol;
			} else {
				return servants.second.front().lastEnabled;
			}

		}

		return servants.second[childIndex].lastEnabled;
	}

	return -1;
}

void Servant::updateLastEnabled(const std::set<unsigned long>& servantsToEnable) {

	// iterating all the children
	for (HashMap<string, pair<bool, vector<Children> > >::iterator sl = _servantMap.begin(); sl != _servantMap.end(); sl++) {

		for (vector<Children>::iterator l = sl->second.second.begin(); l != sl->second.second.end(); l++) {

			// searching for servant
			int i = 0;
			for (vector<unsigned long>::iterator s = l->children.begin(); s != l->children.end(); s++) {
				if (servantsToEnable.find(*s) != servantsToEnable.end()) {
					l->lastEnabled = i;
				}

				i++;
			}
		}
	}
}

}
