/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_SERVANT_H_
#define DATAPROVIDER_SERVANT_H_

#include "PropertyHolder.h"
#include "Command.h"
#include <common/base/HashMap.h>
#include <common/base/CommonTypes.h>
#include <vector>
#include <string>
#include <set>
#include <deque>

namespace dataprovider {

/**
 * Class defining the servant element in the Data Provider database.
 */
class Servant : public PropertyHolder {

public:
	/**
	 * Exception for servant index.
	 */
	class ServantIndexException : public common::Exception {

	public:
		/**
	 	 * Constructor.
	 	 */
		ServantIndexException(std::string const & name, long index, long size);

		long getSize() const;

	private:
		long _size;
	};

	/**
	 * Exception for no such servant child.
	 */
	class NoSuchServantChildException : public common::Exception {

	public:
		/**
	 	 * Constructor.
	 	 */
		NoSuchServantChildException(std::string const & name);
	};

	/**
	 * Servant type.
	 */
	typedef enum {DRIVER, CONTROLLER} Type;

	struct Children {
		std::vector<unsigned long> children;
		long lastEnabled;
	};

	/**
	 * Constructor.
	 */
	Servant(std::string const & name, std::string const & clientType, bool isCommand, bool isVisible, bool isPausable, bool isFavourite);

	/**
	 * Copy constructor.
	 */
	Servant(Servant const & content);

	/**
	 * Virtual destructor. Deletes all the ElementaryProperty and
	 * PropertyArray objects.
	 */
	virtual ~Servant();

	/**
	 * Sets the server type (ICS or NOMAD).
	 */
	void setServerType(common::ServerType serverType);

	/**
	 * Sets the name.
	 */
	void setName(std::string const & name);

	/**
	 * Sets the ID.
	 */
	virtual void setID(unsigned long ID);

	/**
	 * Gets the servant type.
	 */
	virtual Type getServantType() const = 0;

	/**
	 * Sets the driver/controller type as seen by the client.
	 */
	virtual void setClientType(const std::string & type);

	/**
	 * Gets the driver/controller type as seen by the client.
	 */
	virtual std::string const & getClientType() const;

	/**
	 * Gets the server type (ICS or NOMAD).
	 */
	common::ServerType getServerType() const;

	/**
	 * Gets the name.
	 */
	virtual std::string const & getName() const;

	/**
	 * Adds the command and adds the servant to the command.
	 */
	void addCommand(Command * command);

	/**
	 * Gets the list of commands.
	 */
	common::HashMap<std::string, unsigned long> const & getCommands() const;

	/**
	 * Gets a Command from its name.
	 */
	unsigned long getCommand(std::string const & commandName) const;

	/**
	 * Adds the servant list.
	 */
	void addServantList(std::string const & nName, long size);

	/**
	 * Adds the servant to the list.
	 */
	void addServantToList(std::string const & nName, int position, Servant * servant);

	/**
	 * Adds the servant.
	 */
	void addServant(std::string const & nName, Servant * servant, bool toBeSaved);

	/**
	 * Assigns a servant.
	 */
	void assignServant(std::string const & nName, int position, unsigned long servantID);

	/**
	 * Removes a servant with nName.
	 * throw ServantIndexException
	 */
	unsigned long removeServant(std::string const & nName, long index);

	/**
	 * Gets the list of servants.
	 */
	common::HashMap<std::string, std::pair<bool, std::vector<Children> > > & getServantMap();

	/**
	 * Sets the list of servants.
	 */
	void setServantMap(common::HashMap<std::string, std::pair<bool, std::vector<Children> > > const & servantMap);

	/**
	 * Gets the servant. The servant list is resized if needed.
	 */
	std::vector<unsigned long> getServant(std::string const & nName, long childIndex);

	/**
	 * Gets the servant list size.
	 */
	long getServantListSize(std::string const & nName) const;

	/**
	 * Resizes the servant list. Does not perform operations onto the servants.
	 */
	void resizeServantList(std::string const & nName, long size);

	/**
	 * Sets command.
	 */
	void setCommand(bool value);

	/**
	 * returns true if is a command.
	 */
	bool isCommand() const;

	/**
	 * Sets visible or not.
	 */
	void setVisible(bool value);

	/**
	 * returns true if is visible.
	 */
	bool isVisible() const;

	/**
	 * Sets the family
	 */
	void setFamily(const std::string& family, const std::string& subfamily);

	/**
	 * Returns the family.
	 */
	const std::string& getFamily() const;

	/**
	 * Returns the subfamily.
	 */
	const std::string& getSubfamily() const;

	/**
	 * Sets initialised.
	 */
	virtual void setInitialised(bool initialised);

	/**
	 * Enables the servant.
	 */
	void enable();

	/**
	 * Disables the servant.
	 */
	void disable();

	/**
	 * Returns true if the servant is enabled.
	 */
	bool isEnabled() const;

	/**
	 * Returns true if the servant is pausable.
	 */
	bool isPausable() const;

	/**
	 * Sets the servant favourite.
	 */
	void setFavourite(bool value);

	/**
	 * Returns true if the servant is favourite.
	 */
	bool isFavourite() const;

	/**
	 * Gets the last enabled servant.
	 */
	long getLastEnabledServant(const std::string& nName, long childIndex) const;

	/**
	 * Sets the servant enabled.
	 */
	void updateLastEnabled(const std::set<unsigned long>& servantsToEnable);

	/**
	 * Returns true if the servant is to be saved
	 */
	bool servantToBeSaved(const std::string nName) const;

private:
	common::ServerType _serverType;
	std::string _name;
	common::HashMap<std::string, std::pair<bool, std::vector<Children> > > _servantMap;
	std::set<std::string> _servantsToSave;
	std::string _family;
	std::string _subfamily;
	bool _isCommand;
	bool _isVisible;
	common::HashMap<std::string, unsigned long> _commands;
	bool _enabled;
	bool _isPausable;
	bool _isFavourite;
	std::string _clientType;
};

}

#endif /*DATAPROVIDER_SERVANT_H_*/
