/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ServantIconAndFamilyMaps.h"
#include "Module/ModuleManager.h"
#include "common/base/ServerProperties.h"
#include <boost/filesystem/operations.hpp>
#include <boost/foreach.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/property_tree/exceptions.hpp>

using namespace boost;
using namespace std;
namespace fs = boost::filesystem;
namespace pt = boost::property_tree;

namespace dataprovider {

ServantIconAndFamilyMaps* ServantIconAndFamilyMaps::m_instance = 0;

const std::string ServantIconAndFamilyMaps::TAG_XMLATTR = "<xmlattr>"; //without points!
const std::string ServantIconAndFamilyMaps::TAG_CONTROLLER_CONFIG_PLUGIN = "controller_plugin_config";
const std::string ServantIconAndFamilyMaps::TAG_FAMILY_CONFIG_PLUGIN = "family_plugin_config";
const std::string ServantIconAndFamilyMaps::TAG_FAMILY_IMAGES = "family_images";
const std::string ServantIconAndFamilyMaps::TAG_FAMILY = "family";
const std::string ServantIconAndFamilyMaps::TAG_IMAGE = "image";

const std::string ServantIconAndFamilyMaps::TYPE_ATTRIBUTE = "type";
const std::string ServantIconAndFamilyMaps::KEY_ATTRIBUTE = "key";
const std::string ServantIconAndFamilyMaps::NAME_ATTRIBUTE = "name";
const std::string ServantIconAndFamilyMaps::FAMILY_ATTRIBUTE = "family";

ServantIconAndFamilyMaps::ServantIconAndFamilyMaps() {
}

ServantIconAndFamilyMaps::~ServantIconAndFamilyMaps() {

}

ServantIconAndFamilyMaps* ServantIconAndFamilyMaps::getInstance() {
	if (m_instance == 0) {
		m_instance = new ServantIconAndFamilyMaps();
	}
	return m_instance;
}

void ServantIconAndFamilyMaps::readClientPluginFiles() {
	readServantPluginFiles();
}

void ServantIconAndFamilyMaps::readFamilyImageKeys(const std::string& XMLFileName) {
	readFamilyImagesXMLFile(XMLFileName);
}

void ServantIconAndFamilyMaps::readServantPluginFiles() {

	string pluginConfigPath = common::ServerProperties::getInstance()->getNomadGUIHomePath() + "/conf/server/controllers";

	//Iterate over all plugin.xml files to retrieve required info
	fs::path pluginDirectory(pluginConfigPath);
	fs::directory_iterator end_iter;

	if (fs::exists(pluginDirectory) && fs::is_directory(pluginDirectory)) {
		for (fs::directory_iterator dir_iter(pluginDirectory); dir_iter != end_iter; ++dir_iter) {

			if (fs::is_directory(*dir_iter)) {
				std::string fileName = dir_iter->path().string();

				//Read content of each directory contained in controllers folder
				fs::path pluginFile = dir_iter->path();
				fs::directory_iterator end_iter;

				if (fs::exists(pluginFile) && fs::is_directory(pluginFile)) {
					for (fs::directory_iterator dir_iter(pluginFile); dir_iter != end_iter; ++dir_iter) {
						if (!fs::is_directory(*dir_iter)) {
							std::string fileName = dir_iter->path().string();
							if (fileName.find("Plugin") != std::string::npos) {
								readServantXMLFile(fileName);
							}
						}
					}
				}
			}
		}
	}

	// iterate over module resources
	const multimap<string, string>& resourceFiles = ModuleManager::getInstance()->getResourceFiles();
	for (multimap<string, string>::const_iterator f = resourceFiles.begin(); f != resourceFiles.end(); f++) {
		string fileName = f->first;
		if (fileName.find("Plugin.xml") != std::string::npos) {
			readServantXMLFile(f->second);
		}
	}
}

void ServantIconAndFamilyMaps::readFamilyPluginFiles() {

	string familyPluginPath = common::ServerProperties::getInstance()->getNomadGUIHomePath() + "/conf/server/families";

	//Iterate over all .xml files in families folder to retrieve required info
	fs::path familiesDirectory(familyPluginPath);
	fs::directory_iterator end_iter;

	if (fs::exists(familiesDirectory) && fs::is_directory(familiesDirectory)) {
		for (fs::directory_iterator dir_iter(familiesDirectory); dir_iter != end_iter; ++dir_iter) {

			if (!fs::is_directory(*dir_iter)) {
				std::string fileName = dir_iter->path().string();

				if (fileName.find("_familyImages.xml") != std::string::npos) {
					readFamilyImagesXMLFile(fileName);

				} else {
					readFamilyPluginXMLFile(fileName);
				}
			}
		}
	}

}

void ServantIconAndFamilyMaps::readServantXMLFile(const std::string& XMLFileName) {

	pt::ptree pt;

	//read xml file
	read_xml(XMLFileName, pt);

	//Get info: controllerType and imageKey -> they are both attributes of a specific xml tag
	std::string controllerAttributeName = TAG_CONTROLLER_CONFIG_PLUGIN + "." + TAG_XMLATTR + "." + TYPE_ATTRIBUTE;
	std::string servantType = pt.get<std::string>(controllerAttributeName);

	std::string imageAttributeName = TAG_CONTROLLER_CONFIG_PLUGIN + "." + TAG_IMAGE + "." + TAG_XMLATTR + "." + KEY_ATTRIBUTE;
	std::string imageKey;
	try {
		imageKey = pt.get<std::string>(imageAttributeName);

	} catch (pt::ptree_bad_path& e) {
		imageKey = ""; //Maybe I should use default imageKey
	}
	m_controllerIcons[servantType] = imageKey;
//	std::cout << "servantName: " << servantType << " imageKey: " << imageKey << std::endl;
}

void ServantIconAndFamilyMaps::readFamilyImagesXMLFile(const std::string& XMLFileName) {
	pt::ptree pt;
	std::string familyName;
	std::string imageKey;

	//read xml file
	read_xml(XMLFileName, pt);

	//In this case we must iterate over all the <family> elements (surrounded by tag <family_images>)
	static pt::ptree emptyTree;
	boost::property_tree::ptree childTree = pt.get_child(TAG_FAMILY_IMAGES, emptyTree);
	BOOST_FOREACH(const pt::ptree::value_type &f, childTree) {
		std::string attributesField = f.first + "." + TAG_XMLATTR;
		static pt::ptree emptyTreeBis;
		const pt::ptree& attributes = f.second.get_child(TAG_XMLATTR, emptyTreeBis);
		BOOST_FOREACH(const pt::ptree::value_type &v, attributes) {

			std::string attributeName = v.first.data();
			if (attributeName.find(NAME_ATTRIBUTE) != std::string::npos) {
				familyName = v.second.data();
			} else if (attributeName.find(KEY_ATTRIBUTE) != std::string::npos) {
				imageKey = v.second.data();
			}
		}

		//Add values to familyImages map:
		m_familyImages[familyName] = imageKey;
		//std::cout << "familyName: " << familyName << " imageKey: " << imageKey << std::endl;
	}
}


void ServantIconAndFamilyMaps::readFamilyPluginXMLFile(const std::string& XMLFileName) {
	pt::ptree pt;

	//read xml file
	read_xml(XMLFileName, pt);

	//Get info: controllerType and imageKey -> they are both attributes of a specific xml tag
	std::string familyAttributeName = TAG_FAMILY_CONFIG_PLUGIN + "." + TAG_XMLATTR + "." + FAMILY_ATTRIBUTE;
	std::string familyName = pt.get<std::string>(familyAttributeName);

	std::string imageAttributeName = TAG_FAMILY_CONFIG_PLUGIN + "." + TAG_IMAGE + "." + TAG_XMLATTR + "." + KEY_ATTRIBUTE;
	std::string imageKey;
	try {
		imageKey = pt.get<std::string>(imageAttributeName);

	} catch (pt::ptree_bad_path& e) {
		imageKey = ""; //Maybe I should use default imageKey
	}

	//Add to familyImages map
	m_familyImages[familyName] = imageKey;
//	std::cout << "(2) familyName: " << familyName << " imageKey: " << imageKey << std::endl;
}

std::string ServantIconAndFamilyMaps::getControllerIconKey(const std::string& controllerType) {
	return m_controllerIcons[controllerType];
}

std::string ServantIconAndFamilyMaps::getFamilyImageKey(const std::string& familyName) {
	return m_familyImages[familyName];
}

const std::map<std::string, std::string>& ServantIconAndFamilyMaps::getFamilyImageKeys() const {
	return m_familyImages;
}

const std::map<std::string, std::string>& ServantIconAndFamilyMaps::getControllerIconKeys() const {
	return m_controllerIcons;
}

}
