/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ServantLoader.h"
#include <dataprovider/dataelement/Command.h>
#include <dataprovider/dataelement/ElementaryProperty.h>
#include <dataprovider/dataelement/DynamicProperty.h>
#include <dataprovider/dataelement/Controller.h>
#include <dataprovider/dataelement/Driver.h>

#include <dataprovider/parser/config/element/ControllerElement.h>
#include <dataprovider/parser/config/element/DriverElement.h>
#include <dataprovider/parser/config/element/PropertyElement.h>
#include <dataprovider/parser/config/element/DynamicPropertyElement.h>
#include <dataprovider/parser/config/element/PropertyContentElement.h>
#include <dataprovider/parser/config/element/ElementaryPropertyElement.h>
#include <dataprovider/parser/config/element/IcsArrayPropertyElement.h>
#include <dataprovider/parser/config/element/CommandElement.h>

#include <common/base/ServerProperties.h>
#include <common/base/FileUtilities.h>
#include <common/base/Convert.h>
#include <common/base/StringUtilities.h>
#include <common/base/PropertyType.h>
#include <common/base/Real.h>
#include <common/base/HashMap.h>
#include <common/output/OutputStream.h>
#include <ics/ServantConfiguration.h>
#include <ics/TypeConverter.h>
#include <ics/DeviceContainer/DeviceDriverCreator.h>
#include <ics/InstrumentAbstraction/ExperimentControllerCreator.h>
#include <ics/InstrumentAbstraction/ExperimentController.h>

#include <iostream>
#include <map>

using namespace std;
using namespace common;
using namespace dataprovider;

namespace ics {

ServantLoader::ServantLoader() {
}

ServantLoader::~ServantLoader() {
}

std::vector<ControllerElement*> ServantLoader::loadControllers() {

	vector<ExperimentControllerConfig> experimentControllerConfigs = ServantConfiguration::getInstance()->getExperimentControllerList();

	HashMap<string, ControllerElement*> controllerMap;

	vector<ControllerElement*> result;

	// iterating the configs
	for (vector<ExperimentControllerConfig>::const_iterator c = experimentControllerConfigs.begin(); c != experimentControllerConfigs.end(); c++) {

		string family = c->getFamily();

		// Getting the properties
		vector<PropertyConfig> properties = c->getProperties();
		bool isCommand = false;
		for (vector<PropertyConfig>::const_iterator p = properties.begin(); p != properties.end(); p++) {
			if (p->getName() == "isCommand") {
				isCommand = (p->getValue() == "true");
			}
		}

		bool isVisible = false;
		if (family != "hidden" && family != "") {
			isVisible = true;
		}

		ControllerElement* controller = new ControllerElement(c->getName(), c->getType(), isCommand, isVisible, c->isEnabled(), true);
		controller->setServerType(common::ICS);

		// Set the first family
		controller->setFamily(family);

		// iterating the roles
		vector<RoleConfig> roles = c->getRoles();
		for (vector<RoleConfig>::const_iterator r = roles.begin(); r != roles.end(); r++) {

			if ((r->getWho() != "not configured") && (r->getWho() != "<not configured>")) {
				controller->addServant(r->getName(), controllerMap.find(r->getWho()), r->getWho());
			}
		}

		result.push_back(controller);
		controllerMap.insert(controller->getName(), controller);
	}

	return result;
}

std::vector<DriverElement*> ServantLoader::loadDrivers() {

	vector<TestControllerConfig> testControllerConfigs = ServantConfiguration::getInstance()->getTestControllerList();

	HashMap<string, DriverElement*> driverMap;

	vector<DriverElement*> result;

	// iterating the configs
	for (vector<TestControllerConfig>::const_reverse_iterator c = testControllerConfigs.rbegin(); c != testControllerConfigs.rend(); c++) {

		DriverElement* driver = new DriverElement(c->getName(), c->getType(), c->isEnabled(), true);
		driver->setServerType(common::ICS);
		driver->addChildList("children");

		// Getting the properties
		vector<PropertyConfig> properties = c->getProperties();
		common::DriverMode mode = common::SIMULATED_MODE;
		for (vector<PropertyConfig>::const_iterator p = properties.begin(); p != properties.end(); p++) {
			if (p->getName() == "state") {
				if (p->getValue() == "real") {
					mode = common::REAL_MODE;
				} else if (p->getValue() == "perfect") {
					mode = common::PERFECT_MODE;
				} else if (p->getValue() == "simulated") {
					mode = common::SIMULATED_MODE;
				} else {
					err << "Error : mode " << p->getValue() << " not managed" << eol;
				}
			}
		}

		driver->setMode(mode);

		// setting the child relation
		if (driverMap.find(c->getParentName())) {

			DriverElement* parent = driverMap.get(c->getParentName());
			parent->addChildToList("children", driver, c->getParentChannel());

		} else {
			if (c->getParentName() != "") {
				err << "Cannot find driver parent " << c->getParentName() << eol;
			}
		}

		result.push_back(driver);
		driverMap.insert(driver->getName(), driver);
	}


	return result;
}

dataprovider::ControllerElement* ServantLoader::createController(const ExperimentControllerConfig& c) {
	string family = c.getFamily();

	// Getting the properties
	vector<PropertyConfig> properties = c.getProperties();
	bool isCommand = false;
	for (vector<PropertyConfig>::const_iterator p = properties.begin(); p != properties.end(); p++) {
		if (p->getName() == "isCommand") {
			isCommand = (p->getValue() == "true");
		}
	}

	bool isVisible = false;
	if (family != "hidden" && family != "") {
		isVisible = true;
	}

	ControllerElement* controller = new ControllerElement(c.getName(), c.getType(), isCommand, isVisible, c.isEnabled(), true);
	controller->setServerType(common::ICS);

	// Set the first family
	controller->setFamily(family);

	// iterating the roles
	vector<RoleConfig> roles = c.getRoles();
	for (vector<RoleConfig>::const_iterator r = roles.begin(); r != roles.end(); r++) {

		if ((r->getWho() != "not configured") && (r->getWho() != "<not configured>")) {
			controller->addServant(r->getName(), r->getRoleType() == EXPERIMENT_CONTROLLER_TYPE, r->getWho());
		}
	}
	return controller;
}

}
