/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */
#include "ControllerLog.h"
#include "VariableChangeLog.h"
#include "ForLoopStartLog.h"
#include "ForLoopEndLog.h"
#include "IfEvaluationLog.h"
#include "BreakEvaluationLog.h"
#include "UserScanStartLog.h"
#include "SimpleLog.h"
#include "CommandLog.h"
#include "CommandZoneContentLog.h"
#include "CommandZoneEventLog.h"
#include "DirectLogPublisher.h"
#include "FluentdLogWriter.h"
#include <iostream>
#include <common/base/Date.h>
#include <cstdlib>
#include <dataprovider/database/Database.h>
#include <dataprovider/iconandfamily/ServantIconAndFamilyMaps.h>
#include <common/connection/RestJsonHttpSyncConnection.h>
#include <common/base/Exception.h>
#include <common/base/ServerProperties.h>
#include <boost/algorithm/string.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <rapidjson/document.h>
#include <unordered_map>
#include <fstream>
#include <unistd.h>

#include <boost/date_time/local_time/local_time.hpp>


using namespace std;
using namespace common;
using namespace boost;
using namespace boost::posix_time;
using namespace rapidjson;

namespace dataprovider {

const std::string FluentdLogWriter::VERSION_STR = "version";
const int32 FluentdLogWriter::VERSION_VALUE = 1;

const std::string FluentdLogWriter::CONTROLLER_MSG_TYPE = "controller";
const std::string FluentdLogWriter::DRIVER_MSG_TYPE = "driver";
const std::string FluentdLogWriter::SIMPLE_MSG_TYPE = "simple";
const std::string FluentdLogWriter::VAR_MSG_TYPE = "var";
const std::string FluentdLogWriter::FORLOOP_MSG_TYPE = "forloop";
const std::string FluentdLogWriter::IF_MSG_TYPE = "if";
const std::string FluentdLogWriter::BREAK_MSG_TYPE = "break";
const std::string FluentdLogWriter::SCAN_MSG_TYPE = "scan";

const std::string FluentdLogWriter::FROM = "from";
const std::string FluentdLogWriter::ID = "id";
const std::string FluentdLogWriter::DATE = "date";
const std::string FluentdLogWriter::DATE_ID = "date_id";
const std::string FluentdLogWriter::PID = "pid";
const std::string FluentdLogWriter::MESSAGE_TYPE = "mt";
const std::string FluentdLogWriter::NAME = "name";
const std::string FluentdLogWriter::TYPE = "type";
const std::string FluentdLogWriter::PROPERTIES = "properties";
const std::string FluentdLogWriter::PNAME = "pname";
const std::string FluentdLogWriter::PTYPE = "ptype";
const std::string FluentdLogWriter::PQUALIFIER = "pqualifier";
const std::string FluentdLogWriter::PUNIT = "punit";
const std::string FluentdLogWriter::PVALUE = "pvalue";
const std::string FluentdLogWriter::LEVEL = "level";
const std::string FluentdLogWriter::PROPOSAL_ID = "propid";
const std::string FluentdLogWriter::PROPOSAL = "proposal";
const std::string FluentdLogWriter::INSTRUMENT_ID = "instrid";
const std::string FluentdLogWriter::INSTRUMENT = "instrument";
const std::string FluentdLogWriter::SAMPLE_ID = "sampid";
const std::string FluentdLogWriter::SAMPLE = "sample";
const std::string FluentdLogWriter::VARIABLES = "variables";
const std::string FluentdLogWriter::VALUE = "value";
const std::string FluentdLogWriter::QUALIFIER = "qualifier";
const std::string FluentdLogWriter::UNIT = "unit";
const std::string FluentdLogWriter::CURLY_BRACES_OPENING = "{";
const std::string FluentdLogWriter::CURLY_BRACES_ENDING = "}";
const std::string FluentdLogWriter::SQUARE_BRACES_OPENING = "[";
const std::string FluentdLogWriter::SQUARE_BRACES_ENDING = "]";
const std::string FluentdLogWriter::INFO = "info";
const std::string FluentdLogWriter::OUTPUT = "output";
const std::string FluentdLogWriter::CYCLE_ID = "cycleid";
const std::string FluentdLogWriter::ICON = "icon";
const std::string FluentdLogWriter::FAMILY = "family";
const std::string FluentdLogWriter::IMAGE = "imagefile";

const std::string FluentdLogWriter::HARDWARE = "Hardware";

const std::string FluentdLogWriter::RJSON_ID = "id";
const std::string FluentdLogWriter::RJSON_N_ID = "n_id";
const std::string FluentdLogWriter::RJSON_OUTPUT = "output";
const std::string FluentdLogWriter::RJSON_TYPE = "type";
const std::string FluentdLogWriter::RJSON_ID_CYCLE = "id_cycle";
const std::string FluentdLogWriter::RJSON_CYCLE_YEAR = "cycle_year";
const std::string FluentdLogWriter::RJSON_LEVEL = "level";
const std::string FluentdLogWriter::RJSON_ID_INVESTIGATION = "id_investigation"; 
const std::string FluentdLogWriter::RJSON_ID_INSTRUMENT = "id_instrument";
const std::string FluentdLogWriter::RJSON_FAMILY = "family";
const std::string FluentdLogWriter::RJSON_FAMILY_KEY = "family_key";
const std::string FluentdLogWriter::RJSON_IMAGE = "image";
const std::string FluentdLogWriter::RJSON_ACQ_TYPE = "a_type";
const std::string FluentdLogWriter::RJSON_CONTROLLER = "controller";
const std::string FluentdLogWriter::RJSON_NUMOR = "numor";
const std::string FluentdLogWriter::RJSON_R_CONF = "r_conf";
const std::string FluentdLogWriter::RJSON_R_COUNT = "r_count";
const std::string FluentdLogWriter::RJSON_SUBTITLE = "subtitle";
const std::string FluentdLogWriter::RJSON_TIME = "time";
const std::string FluentdLogWriter::RJSON_TIME_UNIT = "time_unit";
const std::string FluentdLogWriter::RJSON_DET_SUM = "det_sum";
const std::string FluentdLogWriter::RJSON_DET_RATE = "det_rate";
const std::string FluentdLogWriter::RJSON_DET_RATE_UNIT = "det_rate_unit";
const std::string FluentdLogWriter::RJSON_DET1_SUM = "det1_sum";
const std::string FluentdLogWriter::RJSON_DET1_RATE = "det1_rate";
const std::string FluentdLogWriter::RJSON_DET1_RATE_UNIT = "det1_rate_unit";
const std::string FluentdLogWriter::RJSON_DET2_SUM = "det2_sum";
const std::string FluentdLogWriter::RJSON_DET2_RATE = "det2_rate";
const std::string FluentdLogWriter::RJSON_DET2_RATE_UNIT = "det2_rate_unit";
const std::string FluentdLogWriter::RJSON_DET3_SUM = "det3_sum";
const std::string FluentdLogWriter::RJSON_DET3_RATE = "det3_rate";
const std::string FluentdLogWriter::RJSON_DET3_RATE_UNIT = "det3_rate_unit";
const std::string FluentdLogWriter::RJSON_DET4_SUM = "det4_sum";
const std::string FluentdLogWriter::RJSON_DET4_RATE = "det4_rate";
const std::string FluentdLogWriter::RJSON_DET4_RATE_UNIT = "det4_rate_unit";
const std::string FluentdLogWriter::RJSON_DET5_SUM = "det5_sum";
const std::string FluentdLogWriter::RJSON_DET5_RATE = "det5_rate";
const std::string FluentdLogWriter::RJSON_DET5_RATE_UNIT = "det5_rate_unit";
const std::string FluentdLogWriter::RJSON_MON1_SUM = "mon1_sum";
const std::string FluentdLogWriter::RJSON_MON1_RATE = "mon1_rate";
const std::string FluentdLogWriter::RJSON_MON1_RATE_UNIT = "mon1_rate_unit";
const std::string FluentdLogWriter::RJSON_MON2_SUM = "mon2_sum";
const std::string FluentdLogWriter::RJSON_MON2_RATE = "mon2_rate";
const std::string FluentdLogWriter::RJSON_MON2_RATE_UNIT = "mon2_rate_unit";
const std::string FluentdLogWriter::RJSON_MON3_SUM = "mon3_sum";
const std::string FluentdLogWriter::RJSON_MON3_RATE = "mon3_rate";
const std::string FluentdLogWriter::RJSON_MON3_RATE_UNIT = "mon3_rate_unit";
const std::string FluentdLogWriter::RJSON_MAIN_SAMPLE_TEMP = "main_sample_temp";
const std::string FluentdLogWriter::RJSON_MAIN_SAMPLE_TEMP_UNIT = "main_sample_temp_unit";
const std::string FluentdLogWriter::RJSON_SEC_SAMPLE_TEMP = "secondary_sample_temp";
const std::string FluentdLogWriter::RJSON_SEC_SAMPLE_TEMP_UNIT = "secondary_sample_temp_unit";
const std::string FluentdLogWriter::RJSON_MAGNETIC_FIELD = "magnetic_field";
const std::string FluentdLogWriter::RJSON_MAGNETIC_FIELD_UNIT = "magnetic_field_unit";
const std::string FluentdLogWriter::RJSON_ACTION = "action";
const std::string FluentdLogWriter::RJSON_COUNTS = "counts";
const std::string FluentdLogWriter::RJSON_CONTROLLERS = "controllers";
const std::string FluentdLogWriter::RJSON_POINTS = "points";
const std::string FluentdLogWriter::RJSON_OPTIMUM = "optimum";
const std::string FluentdLogWriter::RJSON_CREATED_AT = "created_at";

const std::string FluentdLogWriter::PROPERTY_ACQTYPE = "AcqType";
const std::string FluentdLogWriter::PROPERTY_DATANUMOR = "Numor";
const std::string FluentdLogWriter::PROPERTY_REPEATERCONF = "RepeaterConf";
const std::string FluentdLogWriter::PROPERTY_REPEATERCOUNT = "RepeaterCount";
const std::string FluentdLogWriter::PROPERTY_DATASUBTITLE = "Subtitle";
const std::string FluentdLogWriter::PROPERTY_DATATIME = "DataTime";
const std::string FluentdLogWriter::PROPERTY_DETSUM = "DetSum";
const std::string FluentdLogWriter::PROPERTY_DETRATE = "DetRate";
const std::string FluentdLogWriter::PROPERTY_DET1SUM = "DetSum1";
const std::string FluentdLogWriter::PROPERTY_DET1RATE = "DetRate1";
const std::string FluentdLogWriter::PROPERTY_DET2SUM = "DetSum2";
const std::string FluentdLogWriter::PROPERTY_DET2RATE = "DetRate2";
const std::string FluentdLogWriter::PROPERTY_DET3SUM = "DetSum3";
const std::string FluentdLogWriter::PROPERTY_DET3RATE = "DetRate3";
const std::string FluentdLogWriter::PROPERTY_DET4SUM = "DetSum4";
const std::string FluentdLogWriter::PROPERTY_DET4RATE = "DetRate4";
const std::string FluentdLogWriter::PROPERTY_DET5SUM = "DetSum5";
const std::string FluentdLogWriter::PROPERTY_DET5RATE = "DetRate5";
const std::string FluentdLogWriter::PROPERTY_MON1SUM = "Mon1Sum";
const std::string FluentdLogWriter::PROPERTY_MON1RATE = "Mon1Rate";
const std::string FluentdLogWriter::PROPERTY_MON2SUM = "Mon2Sum";
const std::string FluentdLogWriter::PROPERTY_MON2RATE = "Mon2Rate";
const std::string FluentdLogWriter::PROPERTY_MON3SUM = "Mon3Sum";
const std::string FluentdLogWriter::PROPERTY_MON3RATE = "Mon3Rate";
const std::string FluentdLogWriter::PROPERTY_MAIN_SAMPLE_TEMP = "MainSampleTemp";
const std::string FluentdLogWriter::PROPERTY_SEC_SAMPLE_TEMP = "SecSampleTemp";
const std::string FluentdLogWriter::PROPERTY_MAGNETIC_FIELD = "MagField";
const std::string FluentdLogWriter::PROPERTY_SCANACTION = "ScanAction";
const std::string FluentdLogWriter::PROPERTY_SCANCOUNTS = "ScanCounts";
const std::string FluentdLogWriter::PROPERTY_CONTROLLERS = "Controllers";
const std::string FluentdLogWriter::PROPERTY_NBPOINTS = "NbPoints";
const std::string FluentdLogWriter::PROPERTY_SCANOPTIMUM = "ScanOptimum";

const std::string FluentdLogWriter::DEFAULT_ICON_NAME = "INFO";

const std::string FluentdLogWriter::COLON = ":";
const std::string FluentdLogWriter::COMMA = ",";
const std::string FluentdLogWriter::QUOTATION_MARK = "\"";

const std::string FluentdLogWriter::MAINTENANCE_FAMILY = "Maintenance";
const std::string FluentdLogWriter::INTERNAL_USE_PROPOSAL_ID = "0";

const std::string FluentdLogWriter::LOGS_LIVE_UPDATE = "logsLiveUpdate";
const std::string FluentdLogWriter::LOGS_DEV_HOST = "logsDevHost";
const std::string FluentdLogWriter::LOGS_PROD_HOST = "logsProdHost";
const std::string FluentdLogWriter::LOGS_PROD_HOST_SEC = "logsProdHostSec";

const std::string FluentdLogWriter::EXECUTION_CONTROL = "Execution control";
const std::string FluentdLogWriter::USER_SCAN = "User Scan";
const std::string FluentdLogWriter::OTHER = "Other";

using namespace common;

FluentdLogWriter::LogException::LogException(const std::string& message) :
	common::Exception("dataprovider::FluentdLogWriter::LogException", message) {
}

FluentdLogWriter::FluentdLogWriter() :
	m_propertyIdOfProposalId(0),
	m_propertyIdOfProposalName(0),
	m_propertyIdOfInstrumentId(0),
	m_propertyIdOfInstrumentCode(0),
	m_propertyIdOfCycleId(0),
	m_propertyIdOfSampleId(0),
	m_propertyIdOfSampleName(0),
	m_initDone(false),
	m_isMaintenanceLog(false),
	m_showMessages(false),
	m_n(0) {

	m_nomadSendData = ServerProperties::getInstance()->getNomadSenddata();
	to_lower(m_nomadSendData);

	size_t length = 255;
	char *hostname = (char *) malloc(length * sizeof(char));
	int res = gethostname(hostname, length);

	m_hostname = hostname;
	free(hostname);
}

FluentdLogWriter::~FluentdLogWriter() {
}

void FluentdLogWriter::write(Log* log, int n) {

	if (m_nomadSendData == "-1") {
		return;
	}

	m_n = n;

	// Do not log CommandLog, CommandZoneContentLog and CommandZoneEventLog messages.
	if ((dynamic_cast<CommandLog*>(log) != 0)
	  || (dynamic_cast<CommandZoneContentLog*>(log) != 0)
	  || (dynamic_cast<CommandZoneEventLog*>(log) != 0)) {
		return;
	}

	//Check it is not a debug log line
	if (log->getLevel() != common::LOG_DEBUG) {

		if (dynamic_cast<ForLoopEndLog*>(log) != 0) {
			//Remove entry in parentId map
			m_parentID.erase((dynamic_cast<ForLoopEndLog*>(log))->getMessageId());

		} else {

			if (m_initDone == false) {
				try {
					init();
				} catch (Exception&) {
					// cerr << "In Fluentd: Servant Name not correct" << endl;
					return;
				}
			}

			string newjsonMessage;

			// Clean writer and stringbuffer.
			PrettyWriter<StringBuffer> newwriter(news);
			news.Clear();

			// Start new message
			newwriter.StartObject();

			newwriter.Key(VERSION_STR.c_str());
			newwriter.Int(VERSION_VALUE);

			try {
				if (dynamic_cast<ControllerLog*>(log) != 0) { //CONTROLLER (and driver?) CASE
					createControllerMessageUsingRapidJson(dynamic_cast<ControllerLog*>(log), newwriter);

				} else if (dynamic_cast<VariableChangeLog*>(log) != 0) {
					createVariableChangeMessageUsingRapidJson(dynamic_cast<VariableChangeLog*>(log), newwriter);

				} else if (dynamic_cast<ForLoopStartLog*>(log) != 0) {
					createForLoopStartMessageUsingRapidJson(dynamic_cast<ForLoopStartLog*>(log), newwriter);

				} else if (dynamic_cast<IfEvaluationLog*>(log) != 0) {
					createIfEvaluationMessageUsingRapidJson(dynamic_cast<IfEvaluationLog*>(log), newwriter);

				} else if (dynamic_cast<BreakEvaluationLog*>(log) != 0) {
					createBreakEvaluationMessageUsingRapidJson(dynamic_cast<BreakEvaluationLog*>(log), newwriter);

				} else if (dynamic_cast<UserScanStartLog*>(log) != 0) {
					createUserScanStartMessageUsingRapidJson(dynamic_cast<UserScanStartLog*>(log), newwriter);

				} else if (dynamic_cast<SimpleLog*>(log) != 0) {
					createSimpleMessageUsingRapidJson(dynamic_cast<SimpleLog*>(log), newwriter);
				}

				addMessageEndingUsingRapidJson(log, newwriter);
				newwriter.EndObject();

				// only show messages in case it was activated
				if (m_showMessages) {
					cout << "Fluentd mongo log message: " << news.GetString() << endl << endl;
				}

				// Trying to detect bad JSON messages (e.g. driver errors with bad characters) using rapidjson::Document.Parse(string).HasParseError()
				// does not provide errors.

				std::string instrumentCode = Database::getInstance()->getValue<std::string>(m_propertyIdOfInstrumentCode);
				boost::to_upper(instrumentCode);
				RestJsonHttpSyncConnection newcon;

				newcon.openUrl("localhost", "8888");

				//New environment variable
				if ((m_nomadSendData == "0") || (m_nomadSendData == "-2")) {
					//if 0 send data to dev database
					newcon.setPath(string("/mongo.forward.") + instrumentCode + "_DEV" + ".log" + Database::getInstance()->getValue<std::string>(m_propertyIdOfProposalId));

				} else if ((m_nomadSendData == "1") || (m_nomadSendData == "2")) {
					//if 1 send data to prod database
					newcon.setPath(string("/mongo.forward.") + instrumentCode + ".log" + Database::getInstance()->getValue<std::string>(m_propertyIdOfProposalId));
				}
				newjsonMessage = news.GetString();
				// Send to the td agent.
				newcon.setData(newjsonMessage);
				boost::property_tree::ptree response = newcon.send(false);

			} catch (const Exception& e) {
				// log to server
				cerr << e.printStack() << endl;
				slog << e.printStack() << eol;
			}

			// Publish the message.
			if (m_directLogPublisher || m_directLogPublisherSecondary) {
				publishLog(newjsonMessage);
			}

			//Once it is sent clean local variables
			m_outputMessage = "null";
		}
	}
}

void FluentdLogWriter::showJSONMessages(bool value) {
	m_showMessages = value;
}

void FluentdLogWriter::init() {
	Database* database = Database::getInstance();
	unsigned long servantID = database->getServantID("Title");
	std::string controllerType = Database::getInstance()->getControllerType(servantID);
	Controller* experimentDataController = Database::getInstance()->getController(controllerType);

	//Save ID of all necessary properties:
	//It is missing: instrument ID, sample ID, Sample Name.
	m_propertyIdOfProposalId = experimentDataController->getPropertyID("proposalId");
	m_propertyIdOfProposalName = experimentDataController->getPropertyID("experiment_number"); //proposal name == experiment number
	m_propertyIdOfInstrumentId = experimentDataController->getPropertyID("instrumentId");
	m_propertyIdOfInstrumentCode = experimentDataController->getPropertyID("instrumentCode");
	m_propertyIdOfCycleId = experimentDataController->getPropertyID("cycleId");
	m_propertyIdOfSampleId = experimentDataController->getPropertyID("sampleId");
	m_propertyIdOfSampleName = experimentDataController->getPropertyID("chemicalFormula"); //sampleName = chemicalFormula
	unsigned long readyId = experimentDataController->getPropertyID("ready");
	m_initDone = Database::getInstance()->getValue<bool>(readyId);

	// Direct log publisher.
	string logsLiveUpdate = "true";

	try {
		logsLiveUpdate = ServerProperties::getInstance()->getString(LOGS_LIVE_UPDATE);
	} catch (...) {
	}

	if (logsLiveUpdate == "true") {

		string host = "localhost";
		string hostsec;

		try {
			if ((m_nomadSendData == "0")  || (m_nomadSendData == "-2")) {
				host = ServerProperties::getInstance()->getString(LOGS_DEV_HOST);
			} else if ((m_nomadSendData == "1") || (m_nomadSendData == "2")) {
				host = ServerProperties::getInstance()->getString(LOGS_PROD_HOST);
				hostsec = ServerProperties::getInstance()->getString(LOGS_PROD_HOST_SEC);
			}
		} catch (...) {
		}

		out << "Direct log publisher connects to " << host;
		if (hostsec.empty() == false) {
			out << " and " << hostsec;
		}
		out << eol;

		// Initialise the direct log publisher.
		m_directLogPublisher.reset(new DirectLogPublisher(host));
		if (hostsec.empty() == false) {
			m_directLogPublisherSecondary.reset(new DirectLogPublisher(hostsec));
		}
	}
}

void FluentdLogWriter::printAllControllersTypes() {
	vector<Database::ServantItem> controllers = Database::getInstance()->getControllers();
	//Iterate through controllers and get info
	for (std::vector<Database::ServantItem>::iterator it = controllers.begin(); it != controllers.end(); it++) {
		cout << "Controller type: " << it->type << endl;
	}
}

std::string FluentdLogWriter::getId(const Date& allDate) {
	return allDate.toString();
}

std::string FluentdLogWriter::getMessageType(unsigned long servantID) {
	//Either a controller or a driver
	std::string servantType = Database::getInstance()->getServantType(servantID);
	if (servantType.find(CONTROLLER_MSG_TYPE) != string::npos) {
		return CONTROLLER_MSG_TYPE;

	} else if (servantType.find(DRIVER_MSG_TYPE) != string::npos) {
		return DRIVER_MSG_TYPE;
	}

	// throw an exception
	throw LogException(string("servant <" + boost::lexical_cast<string>(servantID) + "> does not exist"));
}

std::string FluentdLogWriter::getType(unsigned long servantID) {
	//From servantID "type" fields
	std::string servantType = Database::getInstance()->getServantType(servantID);
	std::string controller = CONTROLLER_MSG_TYPE;
	unsigned mtPosition = servantType.find(controller) + controller.length();
	std::string type = servantType.substr(mtPosition + 1);
	return type;
}

std::string FluentdLogWriter::getName(unsigned long servantID) {
	std::string name = Database::getInstance()->getServantName(servantID);
	return name;
}

std::string FluentdLogWriter::createLevelString(common::LogLevel level) {
	if (level == common::LOG_INFO) {
		return "info";
	} else if (level == common::LOG_WARNING) {
		return "warning";
	} else if (level == common::LOG_ERROR) {
		return "error";
	}
	return "";
}

std::string FluentdLogWriter::getOutputMessage(Log* log) {

	string result;

	if (dynamic_cast<ControllerLog*>(log) != 0) {
		result = m_outputMessage;
	} else {
		result = log->getMessage();
	}

	if (m_n > 0) {
		result += " (x " + std::to_string(m_n) + ")";
	}

	return result;
}

std::string FluentdLogWriter::getControllerIconKey(std::string controllerType) {
	return ServantIconAndFamilyMaps::getInstance()->getControllerIconKey(controllerType);
}

std::string FluentdLogWriter::getFamilyName(long servantID) {
	std::string servantType = Database::getInstance()->getServantType(servantID);
	if (servantType.find(DRIVER_MSG_TYPE) != string::npos) {
		return HARDWARE;
	}
	return Database::getInstance()->getServantFamily(servantID);
}

void FluentdLogWriter::createControllerMessageUsingRapidJson(ControllerLog* controllerLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter) {

	long servantID = controllerLog->getServantID();

	int64 time = controllerLog->getDate().getTimeUs();
	newwriter.Key(DATE.c_str());
	newwriter.Int64(time / 1000);
	newwriter.Key(DATE_ID.c_str());
	newwriter.Int64(time);

	if (servantID != 0) {
		calculateFieldsUsingRapidJson(controllerLog);

		newwriter.Key(MESSAGE_TYPE.c_str());
		newwriter.String(getMessageType(servantID).c_str());

		newwriter.Key(NAME.c_str());
		newwriter.String(getName(servantID).c_str());

		newwriter.Key(TYPE.c_str());
		newwriter.String(getType(servantID).c_str());

		//Missing properties here
		newRetrieveProperties(controllerLog, newwriter);
	} else {
		newwriter.Key(MESSAGE_TYPE.c_str());
		newwriter.String(SIMPLE_MSG_TYPE.c_str());
	}
}

void FluentdLogWriter::createVariableChangeMessageUsingRapidJson(VariableChangeLog* variableChangeLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter) {

	int64 time = variableChangeLog->getDate().getTimeUs();
	newwriter.Key(DATE.c_str());
	newwriter.Int64(time / 1000);
	newwriter.Key(DATE_ID.c_str());
	newwriter.Int64(time);

	newwriter.Key(PID.c_str());
	newwriter.String(m_parentID[variableChangeLog->getMessageParentId()].c_str());

	newwriter.Key(MESSAGE_TYPE.c_str());
	newwriter.String(VAR_MSG_TYPE.c_str());

	newwriter.Key(NAME.c_str());
	newwriter.String(variableChangeLog->getName().c_str());

	newwriter.Key(TYPE.c_str());
	newwriter.String(variableChangeLog->getType().c_str());

	newwriter.Key(VALUE.c_str());
	newwriter.String(variableChangeLog->getValue().c_str());

}

void FluentdLogWriter::createForLoopStartMessageUsingRapidJson(ForLoopStartLog* forLoopStartLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter) {

	int64 time = forLoopStartLog->getDate().getTimeUs();
	newwriter.Key(DATE.c_str());
	newwriter.Int64(time / 1000);
	newwriter.Key(DATE_ID.c_str());
	newwriter.Int64(time);

	newwriter.Key(MESSAGE_TYPE.c_str());
	newwriter.String(FORLOOP_MSG_TYPE.c_str());

	newwriter.Key(VARIABLES.c_str());
	newwriter.String(forLoopStartLog->getVariablesField().c_str());

	//Save parentID for both break and variableChange log messages
	m_parentID[forLoopStartLog->getMessageId()] = getId(forLoopStartLog->getDate());

	newwriter.Key(PID.c_str());
	newwriter.String(m_parentID[forLoopStartLog->getMessageId()].c_str());
}

void FluentdLogWriter::createIfEvaluationMessageUsingRapidJson(IfEvaluationLog* ifEvaluationLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter) {

	int64 time = ifEvaluationLog->getDate().getTimeUs();
	newwriter.Key(DATE.c_str());
	newwriter.Int64(time / 1000);
	newwriter.Key(DATE_ID.c_str());
	newwriter.Int64(time);

	newwriter.Key(MESSAGE_TYPE.c_str());
	newwriter.String(IF_MSG_TYPE.c_str());

	newwriter.Key(NAME.c_str());
	newwriter.String(ifEvaluationLog->getName().c_str());

	newwriter.Key(TYPE.c_str());
	newwriter.String(ifEvaluationLog->getType().c_str());

	newwriter.Key(VALUE.c_str());
	newwriter.String(ifEvaluationLog->getValue().c_str());

}

void FluentdLogWriter::createBreakEvaluationMessageUsingRapidJson(BreakEvaluationLog* breakEvaluationLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter) {

	int64 time = breakEvaluationLog->getDate().getTimeUs();
	newwriter.Key(DATE.c_str());
	newwriter.Int64(time / 1000);
	newwriter.Key(DATE_ID.c_str());
	newwriter.Int64(time);

	newwriter.Key(PID.c_str());
	newwriter.String(m_parentID[breakEvaluationLog->getMessageParentId()].c_str());

	newwriter.Key(MESSAGE_TYPE.c_str());
	newwriter.String(BREAK_MSG_TYPE.c_str());

	newwriter.Key(NAME.c_str());
	newwriter.String(breakEvaluationLog->getName().c_str());

	newwriter.Key(TYPE.c_str());
	newwriter.String(breakEvaluationLog->getType().c_str());

	newwriter.Key(VALUE.c_str());
	newwriter.String(breakEvaluationLog->getValue().c_str());

}

void FluentdLogWriter::createUserScanStartMessageUsingRapidJson(UserScanStartLog* userScanStartLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter) {

	int64 time = userScanStartLog->getDate().getTimeUs();
	newwriter.Key(DATE.c_str());
	newwriter.Int64(time / 1000);
	newwriter.Key(DATE_ID.c_str());
	newwriter.Int64(time);

	newwriter.Key(MESSAGE_TYPE.c_str());
	newwriter.String(SCAN_MSG_TYPE.c_str());

	newwriter.Key(NAME.c_str());
	newwriter.String(userScanStartLog->getName().c_str());

}

void FluentdLogWriter::createSimpleMessageUsingRapidJson(SimpleLog* simpleDatabaseLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter) {

	int64 time = simpleDatabaseLog->getDate().getTimeUs();
	newwriter.Key(DATE.c_str());
	newwriter.Int64(time / 1000);
	newwriter.Key(DATE_ID.c_str());
	newwriter.Int64(time);

	newwriter.Key(MESSAGE_TYPE.c_str());
	newwriter.String(SIMPLE_MSG_TYPE.c_str());

}

void FluentdLogWriter::addMessageEndingUsingRapidJson(Log* log, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter) {

	//The end of the message is composed by fields: Level, proposal id, proposal, instrument id, instrument, sample id, sample and output.
	newwriter.Key(LEVEL.c_str());
	newwriter.String(createLevelString(log->getLevel()).c_str());

	newwriter.Key(PROPOSAL_ID.c_str());
	if (m_isMaintenanceLog) {
		// All log messages concerning maintenance must be attached to "internal use"
		newwriter.Int(std::stol(INTERNAL_USE_PROPOSAL_ID));
	} else {
		newwriter.Int(std::stol(Database::getInstance()->getValue<std::string>(m_propertyIdOfProposalId)));
	}

	int32 sampleid = 0;
	try {
		sampleid = std::stod(Database::getInstance()->getValue<std::string>(m_propertyIdOfSampleId));
	}
	catch(...) {

	}
	newwriter.Key(SAMPLE_ID.c_str());
	newwriter.Int(sampleid);

	newwriter.Key(SAMPLE.c_str());
	newwriter.String(Database::getInstance()->getValue<std::string>(m_propertyIdOfSampleName).c_str());

	int32 cycleid = 0;
	try {
		cycleid = std::stod(Database::getInstance()->getValue<std::string>(m_propertyIdOfCycleId));
	}
	catch(...) {

	}
	newwriter.Key(CYCLE_ID.c_str());
	newwriter.Int(cycleid);

	newwriter.Key(OUTPUT.c_str());
	newwriter.String(getOutputMessage(log).c_str());

	if (dynamic_cast<ControllerLog*>(log) != 0) {
		//Add special fields for controllerLog messages
		long servantID = dynamic_cast<ControllerLog*>(log)->getServantID();
		if (servantID != 0) {
			std::string controllerType = getType(servantID);

			newwriter.Key(FAMILY.c_str());
			newwriter.String(getFamilyName(servantID).c_str());

			newwriter.Key(IMAGE.c_str());
			newwriter.String(dynamic_cast<ControllerLog*>(log)->getImageName().c_str());

		} else {
			throw LogException("servant Id cannot be 0 in a controller log message");
		}
	} else if (dynamic_cast<VariableChangeLog*>(log) != 0) {
		newwriter.Key(FAMILY.c_str());
		newwriter.String(OTHER.c_str());
	} else if (dynamic_cast<ForLoopStartLog*>(log) != 0) {
		newwriter.Key(FAMILY.c_str());
		newwriter.String(EXECUTION_CONTROL.c_str());
	} else if (dynamic_cast<IfEvaluationLog*>(log) != 0) {
		newwriter.Key(FAMILY.c_str());
		newwriter.String(EXECUTION_CONTROL.c_str());
	} else if (dynamic_cast<BreakEvaluationLog*>(log) != 0) {
		newwriter.Key(FAMILY.c_str());
		newwriter.String(EXECUTION_CONTROL.c_str());
	} else if (dynamic_cast<UserScanStartLog*>(log) != 0) {
		newwriter.Key(FAMILY.c_str());
		newwriter.String(USER_SCAN.c_str());
	} else if (dynamic_cast<SimpleLog*>(log) != 0) {
		newwriter.Key(FAMILY.c_str());
		newwriter.String(OTHER.c_str());
	}
}

void FluentdLogWriter::calculateFieldsUsingRapidJson(ControllerLog* controllerLog) {

	std::string propertyField;
	std::string textMessage;

	string name;
	if (controllerLog->getServantID() != 0) {
		name = Database::getInstance()->getServantName(controllerLog->getServantID());
	} else {
		err << "Error related to controller log message: servant ID is 0!" << eol;
	}
	m_outputMessage = name;

	for (int i = m_outputMessage.size(); i < 16; ++i) {
		m_outputMessage += ' ';
	}

	int32 index = 0;
	std::list<LogElement> propertiesList = controllerLog->getLineElements();

	for (std::list<LogElement>::const_iterator it = propertiesList.begin(); it != propertiesList.end(); it++) {
		LogElement logElement = *it;
		if (!logElement.isHidden()) {
			// Create at the same time the "output" field
			m_outputMessage += logElement.getValue() + logElement.getUnit();
		}
	}
}

void FluentdLogWriter::retrieveProperties(ControllerLog* controllerLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &writer) {

//	std::string propertyField;
//	std::string outputMessage;
//	std::string textMessage;

//	int32 index = 0;
	std::list<LogElement> propertiesList = controllerLog->getLineElements();

	for (std::list<LogElement>::const_iterator it = propertiesList.begin(); it != propertiesList.end(); it++) {
		LogElement logElement = *it;

		if (logElement.isPropertyLog()) {

			writer.StartObject();

			writer.Key(NAME.c_str());
			writer.String(logElement.getName().c_str());

			writer.Key(TYPE.c_str());
			writer.String(logElement.getType().c_str());

			writer.Key(QUALIFIER.c_str());
			writer.String(logElement.getQualifier().c_str());

			writer.Key(UNIT.c_str());
			writer.String(logElement.getUnit().c_str());

			writer.Key(VALUE.c_str());
			writer.String(logElement.getValue().c_str());

			writer.EndObject();

//			stringstream indexValue; //Convert the index into a string...
//			indexValue << index;

//			if (!logElement.isHidden()) {
//				textMessage += "$" + indexValue.str() + logElement.getUnit();
//			}
//			index++;
		}
	}
}

void FluentdLogWriter::newRetrieveProperties(ControllerLog* controllerLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &writer) {

//	std::string propertyField;
//	std::string outputMessage;
//	std::string textMessage;

//	int32 index = 0;
	std::list<LogElement> propertiesList = controllerLog->getLineElements();

	if (propertiesList.empty() == false) {
		bool islog = false;
		for (std::list<LogElement>::const_iterator it = propertiesList.begin(); it != propertiesList.end(); it++) {
			LogElement logElement = *it;
			if (logElement.isPropertyLog()) {
				islog = true;
				break;
			}
		}
		if (islog == false) {
			return;
		}
	}

	writer.Key(PROPERTIES.c_str());

	writer.StartArray();

	for (std::list<LogElement>::const_iterator it = propertiesList.begin(); it != propertiesList.end(); it++) {
		LogElement logElement = *it;

		if (logElement.isPropertyLog()) {

			writer.StartObject();

			writer.Key(PNAME.c_str());
			writer.String(logElement.getName().c_str());

			writer.Key(PTYPE.c_str());
			writer.String(logElement.getType().c_str());

			writer.Key(PQUALIFIER.c_str());
			writer.String(logElement.getQualifier().c_str());

			writer.Key(PUNIT.c_str());
			writer.String(logElement.getUnit().c_str());

			writer.Key(PVALUE.c_str());
			writer.String(logElement.getValue().c_str());

			writer.EndObject();

//			stringstream indexValue; //Convert the index into a string...
//			indexValue << index;

//			if (!logElement.isHidden()) {
//				textMessage += "$" + indexValue.str() + logElement.getUnit();
//			}
//			index++;
		}
	}

	writer.EndArray();
}

std::unordered_map<std::string, std::pair<std::string, std::string>> getProperties(const rapidjson::Document& jsonObject) {

	unordered_map<string,  pair<string, string>> result;

	if (jsonObject.HasMember(FluentdLogWriter::PROPERTIES.c_str())) {

		rapidjson::Value::ConstArray array = jsonObject[FluentdLogWriter::PROPERTIES.c_str()].GetArray();

		for (int i = 0; i < array.Size(); ++i) {
			rapidjson::Value::ConstObject object = array[i].GetObject();
			result.insert(std::pair<string, pair<string, string>>(object[FluentdLogWriter::PNAME.c_str()].GetString(), make_pair(object[FluentdLogWriter::PVALUE.c_str()].GetString(), object[FluentdLogWriter::PUNIT.c_str()].GetString())));
		}
	}

	return result;
}

std::string FluentdLogWriter::convertMessageForPublisher(const std::string& message) {

	rapidjson::Document jsonObject;
	jsonObject.Parse(static_cast<const char *>(message.c_str()), message.size());

	// Get the properties of jsonObject.
	unordered_map<string, pair<string, string>> properties = getProperties(jsonObject);

	rapidjson::StringBuffer buffer;
	rapidjson::Writer<rapidjson::StringBuffer> writer(buffer);
	writer.StartObject();

//  Example message:

//	{
//	    "from": "gamma34",
//	    "id": "2020-10-28T09:27:42.975181",
//	    "mt": "controller",
//	    "name": "Wait",
//	    "type": "wait",
//	    "text": "$0s",
//	    "properties": [
//	        {
//	            "name": "Time",
//	            "type": "float64",
//	            "qualifier": "setpoint",
//	            "unit": "s",
//	            "value": "3"
//	        }
//	    ],
//	    "level": "info",
//	    "propid": "0",
//	    "instrid": "53",
//	    "sampid": "",
//	    "sample": "",
//	    "cycleid": "706",
//	    "output": "3s",
//	    "icon": "WAIT",
//	    "family": "Tools",
//	    "imagefile": ""
//	}


//	Expected conversion:

//	{   "id":5110589,                              = null
//	    "output":"Wait            3s",             = name 16 chars output
//	    "type":"controller",                       = mt
//	    "id_cycle":706,                            = cycleid
//	    "cycle_year":2020,                         = get from date
//	    "icon":null,                               = null
//	    "level":"info",                            = level
//	    "id_investigation":0,                      = propid
//	    "id_instrument":53,                        = instrid
//	    "family":"Tools",                          = family
//	    "family_key":"TOOLS",                      = family upper
//	    "image":null,                              = imagefile (check "" -> null)
//	    "a_type":null,                             = property AcqType
//	    "controller":"Wait",                       = name if present
//	    "numor":null,                              = property DataNumor
//	    "r_conf":null,                             = property RepeaterConf (can be "null", convert to null in that case)
//	    "r_count":null,                            = property RepeaterCount (can be "null", convert to null in that case)
//	    "subtitle":null,                           = property DataSubtitle
//	    "time":null,                               = property DataTime
//	    "det_sum":null,                            = property DetSum (not present in scans)
//	    "det_rate":null,                           = property DetRate (not present in scans)
//	    "mon1_sum":null,                           = property Mon1Sum (not present in scans)
//	    "mon1_rate":null,                          = property Mon1Rate (not present in scans)
//	    "mon2_sum":null,                           = property Mon2Sum (not present in scans)
//	    "mon2_rate":null,                          = property Mon2Rate (not present in scans)
//	    "action":null,                             = property ScanAction
//	    "counts":null,                             = property ScanCounts (can be "null", convert to null in that case)
//	    "controllers":null,                        = property Controllers
//	    "points":null,                             = property NbPoints
//	    "optimum":null,                            = property ScanOptimum
//	    "created_at":1603576879000                 = date
//	}:

	// Get the message type.
	string messageType = jsonObject[MESSAGE_TYPE.c_str()].GetString();

//	// The direct log messages have no id that formerly came from the database.
//	writer.Key(RJSON_ID.c_str());
//	writer.Null();

	// The n_id i.e. nomad id is the number of microseconds, that should be unique per instrument and define an id for the annotations.
	if (jsonObject.HasMember(DATE_ID.c_str())) {
		writer.Key(RJSON_N_ID.c_str());
		writer.Int64(jsonObject[DATE_ID.c_str()].GetInt64());
	}
	// Create the output field.
	writer.Key(RJSON_OUTPUT.c_str());

	string output;

	output += string(jsonObject[OUTPUT.c_str()].GetString());

	writer.String(output.c_str());

	// Create the type field.
	writer.Key(RJSON_TYPE.c_str());
	writer.String(messageType.c_str());

	// Create the id_cycle field.
	writer.Key(RJSON_ID_CYCLE.c_str());
	writer.Int(jsonObject[CYCLE_ID.c_str()].GetInt());

	// Create the cycle_year field from the current date.
	writer.Key(RJSON_CYCLE_YEAR.c_str());
	int year = second_clock::local_time().date().year();
	writer.Int(year);

//	// No icon.
//	writer.Key("icon");
//	writer.Null();

	// Create the level field.
	writer.Key(RJSON_LEVEL.c_str());
	writer.String(jsonObject[LEVEL.c_str()].GetString());

	// Create the id_investigation field.
	writer.Key(RJSON_ID_INVESTIGATION.c_str());
	writer.Int(jsonObject[PROPOSAL_ID.c_str()].GetInt());

	// Create the id_instrument field.
	string instrumentId = Database::getInstance()->getValue<std::string>(m_propertyIdOfInstrumentId);

	if (instrumentId == "0") {
		throw LogException(string("instrument id cannot be 0"));
	}

	writer.Key(RJSON_ID_INSTRUMENT.c_str());
	writer.Int(stoi(instrumentId));


	if (jsonObject.HasMember(FAMILY.c_str())) {
		// Create the family field.
		writer.Key(RJSON_FAMILY.c_str());
		writer.String(jsonObject[FAMILY.c_str()].GetString());
//	} else {
//		writer.Null();
	}


	if (jsonObject.HasMember(FAMILY.c_str())) {
		// Create the family_key field.
		writer.Key(RJSON_FAMILY_KEY.c_str());
		string familyKey = jsonObject[FAMILY.c_str()].GetString();
		boost::to_upper(familyKey);
		writer.String(familyKey.c_str());
//	} else {
//		writer.Null();
	}


	if (jsonObject.HasMember(IMAGE.c_str())) {
		// Create the image field.
		string image = jsonObject[IMAGE.c_str()].GetString();
		if (image != "") {
			writer.Key(RJSON_IMAGE.c_str());
			writer.String(image.c_str());
//		} else {
//			writer.Null();
		}
//	} else {
//		writer.Null();
	}

	auto it = properties.find(PROPERTY_ACQTYPE);
	if (it != properties.end()) {
		// Create the a_type field.
		writer.Key(RJSON_ACQ_TYPE.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	if (jsonObject.HasMember(NAME.c_str())) {
		// Create the controller field.
		writer.Key(RJSON_CONTROLLER.c_str());
		writer.String(jsonObject[NAME.c_str()].GetString());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DATANUMOR);
	if (it != properties.end()) {
		// Create the numor field.
		writer.Key(RJSON_NUMOR.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_REPEATERCONF);
	if (it != properties.end()) {
		string value = it->second.first;
		if (value != "null") {
			// Create the r_conf field.
			writer.Key(RJSON_R_CONF.c_str());
			writer.String(value.c_str());
//		} else {
//			writer.Null();
		}
//	} else {
//		writer.Null();
	}


	it = properties.find(PROPERTY_REPEATERCOUNT);
	if (it != properties.end()) {
		string value = it->second.first;
		if (value != "null") {
			// Create the r_count field.
			writer.Key(RJSON_R_COUNT.c_str());
			writer.String(value.c_str());
//		} else {
//			writer.Null();
		}
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DATASUBTITLE);
	if (it != properties.end()) {
		// Create the subtitle field.
		writer.Key(RJSON_SUBTITLE.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DATATIME);
	if (it != properties.end()) {
		// Create the time field.
		writer.Key(RJSON_TIME.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_TIME_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DETSUM);
	if (it != properties.end()) {
		// Create the det_sum field.
		writer.Key(RJSON_DET_SUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DETRATE);
	if (it != properties.end()) {
		// Create the det_rate field.
		writer.Key(RJSON_DET_RATE.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_DET_RATE_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET1SUM);
	if (it != properties.end()) {
		// Create the det_sum field.
		writer.Key(RJSON_DET1_SUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET1RATE);
	if (it != properties.end()) {
		// Create the det_rate field.
		writer.Key(RJSON_DET1_RATE.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_DET1_RATE_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET2SUM);
	if (it != properties.end()) {
		// Create the det_sum field.
		writer.Key(RJSON_DET2_SUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET2RATE);
	if (it != properties.end()) {
		// Create the det_sum field.
		writer.Key(RJSON_DET2_RATE.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_DET2_RATE_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET3SUM);
	if (it != properties.end()) {
		// Create the det_sum field.
		writer.Key(RJSON_DET3_SUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET3RATE);
	if (it != properties.end()) {
		// Create the det_rate field.
		writer.Key(RJSON_DET3_RATE.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_DET3_RATE_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET4SUM);
	if (it != properties.end()) {
		// Create the det_sum field.
		writer.Key(RJSON_DET4_SUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET4RATE);
	if (it != properties.end()) {
		// Create the det_rate field.
		writer.Key(RJSON_DET4_RATE.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_DET4_RATE_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET5SUM);
	if (it != properties.end()) {
		// Create the det_sum field.
		writer.Key(RJSON_DET5_SUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_DET5RATE);
	if (it != properties.end()) {
		// Create the det_rate field.
		writer.Key(RJSON_DET5_RATE.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_DET5_RATE_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_MON1SUM);
	if (it != properties.end()) {
		// Create the mon1_sum field.
		writer.Key(RJSON_MON1_SUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_MON1RATE);
	if (it != properties.end()) {
		// Create the mon1_rate field.
		writer.Key(RJSON_MON1_RATE.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_MON1_RATE_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_MON2SUM);
	if (it != properties.end()) {
		// Create the mon2_sum field.
		writer.Key(RJSON_MON2_SUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_MON2RATE);
	if (it != properties.end()) {
		// Create the mon2_rate field.
		writer.Key(RJSON_MON2_RATE.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_MON2_RATE_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_MON3SUM);
	if (it != properties.end()) {
		// Create the mon3_sum field.
		writer.Key(RJSON_MON3_SUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_MON3RATE);
	if (it != properties.end()) {
		// Create the mon3_rate field.
		writer.Key(RJSON_MON3_RATE.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_MON3_RATE_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_MAIN_SAMPLE_TEMP);
	if (it != properties.end()) {
		// Create the main sample temp
		writer.Key(RJSON_MAIN_SAMPLE_TEMP.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_MAIN_SAMPLE_TEMP_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_SEC_SAMPLE_TEMP);
	if (it != properties.end()) {
		// Create the sec sample temp
		writer.Key(RJSON_SEC_SAMPLE_TEMP.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_SEC_SAMPLE_TEMP_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_MAGNETIC_FIELD);
	if (it != properties.end()) {
		// Create the sec sample temp
		writer.Key(RJSON_MAGNETIC_FIELD.c_str());
		writer.String(it->second.first.c_str());
		writer.Key(RJSON_MAGNETIC_FIELD_UNIT.c_str());
		writer.String(it->second.second.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_SCANACTION);
	if (it != properties.end()) {
		// Create the action field.
		writer.Key(RJSON_ACTION.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_SCANCOUNTS);
	if (it != properties.end()) {
		string value = it->second.first;
		if (value != "null") {
			// Create the counts field.
			writer.Key(RJSON_COUNTS.c_str());
			writer.String(value.c_str());
//		} else {
//			writer.Null();
		}
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_CONTROLLERS);
	if (it != properties.end()) {
		// Create the controllers field.
		writer.Key(RJSON_CONTROLLERS.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_NBPOINTS);
	if (it != properties.end()) {
		// Create the points field.
		writer.Key(RJSON_POINTS.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	it = properties.find(PROPERTY_SCANOPTIMUM);
	if (it != properties.end()) {
		// Create the optimum field.
		writer.Key(RJSON_OPTIMUM.c_str());
		writer.String(it->second.first.c_str());
//	} else {
//		writer.Null();
	}

	// Create the created_at field.
	// Set the number of milliseconds since 1970, january, 1st.
	writer.Key(RJSON_CREATED_AT.c_str());
	writer.Int64(jsonObject[DATE.c_str()].GetInt64());

	writer.EndObject();

	if (m_showMessages) {
		cout << "Live log message: " << buffer.GetString() << endl << endl;
	}

	return buffer.GetString();
}

void FluentdLogWriter::publishLog(const std::string& message) {
	if (!message.empty()) {
		string pmessage = convertMessageForPublisher(message);
		if (m_directLogPublisher) {
			m_directLogPublisher->publish(pmessage);
		}
		if (m_directLogPublisherSecondary) {
			m_directLogPublisherSecondary->publish(pmessage);
		}
	}
}

}

