/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef FLUENTDLOGWRITER_H_
#define FLUENTDLOGWRITER_H_

#include <common/base/Exception.h>
#include <string>
#include "LogWriter.h"
#include "Log.h"
#include <boost/regex.hpp>
#include <boost/property_tree/ptree.hpp>

#include <rapidjson/writer.h>
#include <rapidjson/stringbuffer.h>
#include <rapidjson/prettywriter.h>

namespace common {

class Date;

}


namespace dataprovider {

class VariableChangeLog;
class ForLoopStartLog;
class ForLoopEndLog;
class IfEvaluationLog;
class BreakEvaluationLog;
class UserScanStartLog;
class SimpleLog;
class ControllerLog;

class DirectLogPublisher;


class FluentdLogWriter : public LogWriter {

public:
	/**
	 * Log exception.
	 */
	class LogException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		LogException(const std::string& message);
	};


	FluentdLogWriter();
	virtual ~FluentdLogWriter();
	virtual void write(Log* log, int n = 0);

	void showJSONMessages(bool value);

	static const std::string PROPERTIES;
	static const std::string PNAME;
	static const std::string PVALUE;
	static const std::string PUNIT;

private:
	std::string getId(const common::Date& date);
	std::string getMessageType(unsigned long servantID);
	std::string getType(unsigned long servantID);
	std::string getName(unsigned long servantID);
	std::string createLevelString(common::LogLevel level);
	std::string getControllerIconKey(std::string type);
	std::string getFamilyName(long servantID);

	void init();

	std::string getOutputMessage(Log* log);

	void printAllControllersTypes();

	void createControllerMessageUsingRapidJson(ControllerLog* controllerLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter);
	void createVariableChangeMessageUsingRapidJson(VariableChangeLog* variableChangeLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter);
	void createForLoopStartMessageUsingRapidJson(ForLoopStartLog* forLoopStartLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter);
	void createIfEvaluationMessageUsingRapidJson(IfEvaluationLog* ifEvaluationLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter);
	void createBreakEvaluationMessageUsingRapidJson(BreakEvaluationLog* breakEvaluationLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter);
	void createUserScanStartMessageUsingRapidJson(UserScanStartLog* userScanStartLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter);
	void createSimpleMessageUsingRapidJson(SimpleLog* simpleDatabaseLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter);

	void addMessageEndingUsingRapidJson(Log* log, rapidjson::PrettyWriter<rapidjson::StringBuffer> &newwriter);

	void calculateFieldsUsingRapidJson(ControllerLog* controllerLog);
	void retrieveProperties(ControllerLog* controllerLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &writer);
	void newRetrieveProperties(ControllerLog* controllerLog, rapidjson::PrettyWriter<rapidjson::StringBuffer> &writer);

	std::string convertMessageForPublisher(const std::string& message);
	void publishLog(const std::string& message);

	std::string m_nomadSendData;
	std::string m_hostname;

	int m_n;
	std::string m_outputMessage;

	//Properties IDs
	unsigned long m_propertyIdOfProposalId;
	unsigned long m_propertyIdOfProposalName;
	unsigned long m_propertyIdOfInstrumentId;
	unsigned long m_propertyIdOfInstrumentCode;
	unsigned long m_propertyIdOfCycleId;
	unsigned long m_propertyIdOfSampleId;
	unsigned long m_propertyIdOfSampleName;

	bool m_initDone;
	bool m_isMaintenanceLog;

	std::map<unsigned long, std::string> m_parentID;

//	rapidjson:: StringBuffer s;
	rapidjson:: StringBuffer news;
	rapidjson::Writer<rapidjson::StringBuffer> test;

	std::unique_ptr<DirectLogPublisher> m_directLogPublisher;
	std::unique_ptr<DirectLogPublisher> m_directLogPublisherSecondary;

	bool m_showMessages;

	static const std::string CONTROLLER_MSG_TYPE;
	static const std::string DRIVER_MSG_TYPE;
	static const std::string SIMPLE_MSG_TYPE;
	static const std::string VAR_MSG_TYPE;
	static const std::string FORLOOP_MSG_TYPE;
	static const std::string IF_MSG_TYPE;
	static const std::string BREAK_MSG_TYPE;
	static const std::string SCAN_MSG_TYPE;

	static const std::string VERSION_STR;
	static const int32 VERSION_VALUE;
	static const std::string FROM;
	static const std::string ID;
	static const std::string DATE;
	static const std::string DATE_ID;
	static const std::string PID;
	static const std::string MESSAGE_TYPE;
	static const std::string NAME;
	static const std::string TYPE;
	static const std::string PTYPE;
	static const std::string PQUALIFIER;
	static const std::string LEVEL;
	static const std::string PROPOSAL_ID;
	static const std::string PROPOSAL;
	static const std::string INSTRUMENT_ID;
	static const std::string INSTRUMENT;
	static const std::string SAMPLE_ID;
	static const std::string SAMPLE;
	static const std::string VARIABLES;
	static const std::string VALUE;
	static const std::string QUALIFIER;
	static const std::string UNIT;
	static const std::string CURLY_BRACES_OPENING;
	static const std::string CURLY_BRACES_ENDING;
	static const std::string SQUARE_BRACES_OPENING;
	static const std::string SQUARE_BRACES_ENDING;
	static const std::string INFO;
	static const std::string OUTPUT;
	static const std::string CYCLE_ID;
	static const std::string ICON;
	static const std::string FAMILY;
	static const std::string IMAGE;

	static const std::string HARDWARE;

	static const std::string RJSON_ID;
	static const std::string RJSON_N_ID;
	static const std::string RJSON_OUTPUT;
	static const std::string RJSON_TYPE;
	static const std::string RJSON_ID_CYCLE;
	static const std::string RJSON_CYCLE_YEAR;
	static const std::string RJSON_LEVEL;
	static const std::string RJSON_ID_INVESTIGATION;
	static const std::string RJSON_ID_INSTRUMENT;
	static const std::string RJSON_FAMILY;
	static const std::string RJSON_FAMILY_KEY;
	static const std::string RJSON_IMAGE;
	static const std::string RJSON_ACQ_TYPE;
	static const std::string RJSON_CONTROLLER;
	static const std::string RJSON_NUMOR;
	static const std::string RJSON_R_CONF;
	static const std::string RJSON_R_COUNT;
	static const std::string RJSON_SUBTITLE;
	static const std::string RJSON_TIME;
	static const std::string RJSON_TIME_UNIT;
	static const std::string RJSON_DET_SUM;
	static const std::string RJSON_DET_RATE;
	static const std::string RJSON_DET_RATE_UNIT;
	static const std::string RJSON_DET1_SUM;
	static const std::string RJSON_DET1_RATE;
	static const std::string RJSON_DET1_RATE_UNIT;
	static const std::string RJSON_DET2_SUM;
	static const std::string RJSON_DET2_RATE;
	static const std::string RJSON_DET2_RATE_UNIT;
	static const std::string RJSON_DET3_SUM;
	static const std::string RJSON_DET3_RATE;
	static const std::string RJSON_DET3_RATE_UNIT;
	static const std::string RJSON_DET4_SUM;
	static const std::string RJSON_DET4_RATE;
	static const std::string RJSON_DET4_RATE_UNIT;
	static const std::string RJSON_DET5_SUM;
	static const std::string RJSON_DET5_RATE;
	static const std::string RJSON_DET5_RATE_UNIT;
	static const std::string RJSON_MON1_SUM;
	static const std::string RJSON_MON1_RATE;
	static const std::string RJSON_MON1_RATE_UNIT;
	static const std::string RJSON_MON2_SUM;
	static const std::string RJSON_MON2_RATE;
	static const std::string RJSON_MON2_RATE_UNIT;
	static const std::string RJSON_MON3_SUM;
	static const std::string RJSON_MON3_RATE;
	static const std::string RJSON_MON3_RATE_UNIT;
	static const std::string RJSON_MAIN_SAMPLE_TEMP;
	static const std::string RJSON_MAIN_SAMPLE_TEMP_UNIT;
	static const std::string RJSON_SEC_SAMPLE_TEMP;
	static const std::string RJSON_SEC_SAMPLE_TEMP_UNIT;
	static const std::string RJSON_MAGNETIC_FIELD;
	static const std::string RJSON_MAGNETIC_FIELD_UNIT;
	static const std::string RJSON_ACTION;
	static const std::string RJSON_COUNTS;
	static const std::string RJSON_CONTROLLERS;
	static const std::string RJSON_POINTS;
	static const std::string RJSON_OPTIMUM;
	static const std::string RJSON_CREATED_AT;

	static const std::string PROPERTY_ACQTYPE;
	static const std::string PROPERTY_DATANUMOR;
	static const std::string PROPERTY_REPEATERCONF;
	static const std::string PROPERTY_REPEATERCOUNT;
	static const std::string PROPERTY_DATASUBTITLE;
	static const std::string PROPERTY_DATATIME;
	static const std::string PROPERTY_DETSUM;
	static const std::string PROPERTY_DETRATE;
	static const std::string PROPERTY_DET1SUM;
	static const std::string PROPERTY_DET1RATE;
	static const std::string PROPERTY_DET2SUM;
	static const std::string PROPERTY_DET2RATE;
	static const std::string PROPERTY_DET3SUM;
	static const std::string PROPERTY_DET3RATE;
	static const std::string PROPERTY_DET4SUM;
	static const std::string PROPERTY_DET4RATE;
	static const std::string PROPERTY_DET5SUM;
	static const std::string PROPERTY_DET5RATE;
	static const std::string PROPERTY_MON1SUM;
	static const std::string PROPERTY_MON1RATE;
	static const std::string PROPERTY_MON2SUM;
	static const std::string PROPERTY_MON2RATE;
	static const std::string PROPERTY_MON3SUM;
	static const std::string PROPERTY_MON3RATE;
	static const std::string PROPERTY_MAIN_SAMPLE_TEMP;
	static const std::string PROPERTY_SEC_SAMPLE_TEMP;
	static const std::string PROPERTY_MAGNETIC_FIELD;
	static const std::string PROPERTY_SCANACTION;
	static const std::string PROPERTY_SCANCOUNTS;
	static const std::string PROPERTY_CONTROLLERS;
	static const std::string PROPERTY_NBPOINTS;
	static const std::string PROPERTY_SCANOPTIMUM;

	static const std::string DEFAULT_ICON_NAME;

	static const std::string COLON;
	static const std::string COMMA;
	static const std::string QUOTATION_MARK;

	static const std::string MAINTENANCE_FAMILY;
	static const std::string INTERNAL_USE_PROPOSAL_ID;

	static const std::string LOGS_LIVE_UPDATE;
	static const std::string LOGS_DEV_HOST;
	static const std::string LOGS_PROD_HOST;
	static const std::string LOGS_PROD_HOST_SEC;

	static const std::string EXECUTION_CONTROL;
	static const std::string USER_SCAN;
	static const std::string OTHER;

};

}

#endif /* FLUENTDLOGWRITER_H_ */
