/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "LogHandler.h"
#include "Log.h"
#include "SimpleLog.h"
#include "ControllerLog.h"
#include "XMLLogWriter.h"
#include "TextLogWriter.h"
#include "FluentdLogWriter.h"
#include <common/base/Queue.h>
#include <dataprovider/transportlayer/command/LogCompletionHandler.h>
#include <dataprovider/message/log/WriteLogMessage.h>
#include <dataprovider/message/ShutdownMessage.h>
#include <common/output/OutputStream.h>
#include <common/base/Convert.h>

using namespace common;
using namespace std;

namespace dataprovider {

LogHandler* LogHandler::m_instance = 0;

LogHandler* LogHandler::getInstance() {
	return m_instance;
}

LogHandler::LogHandler() :
	m_running(false),
	m_messageQueue(nullptr),
	m_thread(nullptr),
	m_xmlLogWriter(new XMLLogWriter()),
	m_textLogWriter(new TextLogWriter()),
	m_fluentdLogWriter(new FluentdLogWriter()) {
}

LogHandler::~LogHandler() {

	delete m_messageQueue;
	delete m_xmlLogWriter;
	delete m_textLogWriter;
	delete m_fluentdLogWriter;
}
	
void LogHandler::initInstance() {
	if (m_instance == nullptr) {
		m_instance = new LogHandler();
		m_instance->init();
	}
}

void LogHandler::resetInstance() {

	if (m_instance != nullptr) {
		m_instance->reset();
	}
	
	delete m_instance;
	m_instance = nullptr;
}
	
void LogHandler::setOutputName(std::string const &name) {
	TextLogWriter* writer = dynamic_cast<TextLogWriter*>(m_textLogWriter);
	if (writer != nullptr) {
		writer->setOutputName(name);
	}
}

void LogHandler::writeLog(Log* log, unsigned long requestID, LogCompletionHandler* logCompletionHandler) {
	m_messageQueue->put(new WriteLogMessage(log, requestID, logCompletionHandler));
}

void LogHandler::writeLog(Log* log) {
	m_messageQueue->put(new WriteLogMessage(log, 0, 0));
}

void LogHandler::showJSONMessages(bool value) {
	m_fluentdLogWriter->showJSONMessages(value);
}

void LogHandler::init() {

	m_messageQueue = new Queue<Message>(1024);
	m_thread = new std::thread(&LogHandler::loop);
}

void LogHandler::reset() {
	
	if (m_thread != nullptr) {
		m_messageQueue->put(new ShutdownMessage());
		m_thread->join();
		delete m_thread;
	}	
}

void LogHandler::loop() {
	
	Message* message = nullptr;
	m_instance->m_running = true;
	while (m_instance->m_running) {
		
		try {
			try {		
				message = m_instance->m_messageQueue->get();
				
			} catch (Queue<Message>::UnavailableException& e) {
				break;
				
			} catch (Exception& e) {
				err << e.printStack() << eol;
				break;
			}
			
			try {
				m_instance->processMessage(message);
	
			} catch (Exception& e) {
				err << e.printStack() << eol;
			}
		
		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}
		
		delete message;
		message = 0;
 	}
}

void LogHandler::processMessage(Message* message) {
	
	if (message->getType() == Message::WRITE_LOG) {
		processMessage(static_cast<WriteLogMessage*>(message));

	} else if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<ShutdownMessage*>(message));

	} 
}

void LogHandler::processMessage(WriteLogMessage* message) {

	std::shared_ptr<Log> logMessage(message->getLog());

	try {
		writeLogMessage(logMessage);
		
		if (message->getLogCompletionHandler() != nullptr) {
			message->getLogCompletionHandler()->terminate(message->getRequestID(), true, "");
		}
				
	} catch (Exception const & e) {
		if (message->getLogCompletionHandler() != nullptr) {
			message->getLogCompletionHandler()->terminate(message->getRequestID(), false, e.getMessage());
		}
	} catch (exception& e) {
		err << "STL exception in dataprovider::LogHandler thread: " << e.what() << eol;
		message->getLogCompletionHandler()->terminate(message->getRequestID(), false, e.what());
	} catch (...) {
		err << "Unexpected exception in dataprovider::LogHandler thread" << eol;
		message->getLogCompletionHandler()->terminate(message->getRequestID(), false, "?");
	}
}

void LogHandler::processMessage(ShutdownMessage* message) {
	
	m_running = false;
}

void LogHandler::writeLogMessage(std::shared_ptr<Log> logMessage) {

	WritableLog result = checkMessage(logMessage);

	//cout << "Check: " << (result.write ? "write" : "!write") << " : " << (result.message ? result.message->getMessage() : "") << " x " << result.n << endl;

	if (result.message) {
		m_fluentdLogWriter->write(result.message.get(), result.n);
		m_xmlLogWriter->write(result.message.get(), result.n);
		m_textLogWriter->write(result.message.get(), result.n);
	}

	if (result.write) {
		m_fluentdLogWriter->write(logMessage.get());
		m_xmlLogWriter->write(logMessage.get());
		m_textLogWriter->write(logMessage.get());
	}
}

LogHandler::WritableLog LogHandler::checkMessage(std::shared_ptr<Log> logMessage) {

	WritableLog result;

	ControllerLog* controllerLog = dynamic_cast<ControllerLog*>(logMessage.get());
	if (controllerLog == nullptr) {
		result.write = true;
		return result;
	}
	else {
		// Check the number of line elements.
		if (controllerLog->getLineElements().size() > 1) {
			result.write = true;
			return result;
		}

		// Here, we have a controller log with only 1 line element.

		// Check message with last.
		std::shared_ptr<Log> lastLogMessage = m_logMessagesMap[controllerLog->getServantID()].message;

		// The message is not written if log level is error and message equals the last message.
		if (lastLogMessage
			&& lastLogMessage->getMessage() == logMessage->getMessage()) {

			// Do not write if error.
			if (logMessage->getLevel() == common::LOG_ERROR) {

				// Do not write the current message but set write to true for the next time.
				result.write = false;
				m_logMessagesMap[controllerLog->getServantID()].write = true;
				// Replace the last message.
				m_logMessagesMap[controllerLog->getServantID()].message = logMessage;
				// Increase the number of occurrences.
				m_logMessagesMap[controllerLog->getServantID()].n++;
			}
			else {
				// Write the current message but set write to false for the next time.
				result.write = true;
				m_logMessagesMap[controllerLog->getServantID()].write = false;
				// Replace the last message.
				m_logMessagesMap[controllerLog->getServantID()].message = logMessage;
			}
		}
		else {
			// Write.
			result.write = true;

			// Set the result message if last write was true (i.e. message has not been written).
			if (m_logMessagesMap[controllerLog->getServantID()].write) {
				result.message = m_logMessagesMap[controllerLog->getServantID()].message;

				// Copy the date to have sequential dates.
				result.message->setDate(logMessage->getDate());

				result.n = m_logMessagesMap[controllerLog->getServantID()].n;
			}

			// Replace the last message.
			m_logMessagesMap[controllerLog->getServantID()].write = false;
			m_logMessagesMap[controllerLog->getServantID()].message = logMessage;
			m_logMessagesMap[controllerLog->getServantID()].n = 0;
		}

		return result;
	}
}

}
