/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_LOGHANDLER_H_
#define DATAPROVIDER_LOGHANDLER_H_

#include <thread>
#include <unordered_map>
#include <memory>

namespace common {

template<typename ItemType>
class Queue;

}

namespace dataprovider {

class Log;
class LogCompletionHandler;
class LogWriter;
class Message;
class WriteLogMessage;
class ShutdownMessage;

class TextLogWriter;
class XMLLogWriter;
class FluentdLogWriter;

/**
 * Class defining a log handler. The class is a singleton that processes log messages.
 * The writeLog methods are not blocking but the caller can be synchronised using a completion
 * handler
 */
class LogHandler {

public:
	/**
	 * Gets the instance.
	 */
	static LogHandler* getInstance();
	
	/**
	 * Initialises the instance.
	 */
	static void initInstance();
	
	/**
	 * Resets the instance.
	 */
	static void resetInstance();
	
	/**
	 * Sets the output name.
	 */
	void setOutputName(std::string const &name);

	/**
	 * Writes the log with completion handler.
	 */
	void writeLog(Log* log, unsigned long requestID, LogCompletionHandler* logCompletionHandler);
	
	/**
	 * Writes the log.
	 */
	void writeLog(Log* log);
	
	/**
	 * Shows or hides the JSON messages.
	 */
	void showJSONMessages(bool value);

private:
	LogHandler();
	~LogHandler();
	
	void init();
	void reset();

	void processMessage(Message* message);
	void processMessage(WriteLogMessage* message);
	void processMessage(ShutdownMessage* message);

	void writeLogMessage(std::shared_ptr<Log> logMessage);
	
	struct WritableLog {
		bool write;
		std::shared_ptr<Log> message;
		int n;

		WritableLog() : write(false), message(nullptr), n(0) {}
	};

	WritableLog checkMessage(std::shared_ptr<Log> logMessage);

	static void loop();

	static LogHandler * m_instance;
	
	bool m_running;
	common::Queue<Message>* m_messageQueue;
	std::thread* m_thread;

	XMLLogWriter* m_xmlLogWriter;
	TextLogWriter* m_textLogWriter;
	FluentdLogWriter* m_fluentdLogWriter;

	std::unordered_map<unsigned long, WritableLog> m_logMessagesMap;
};

}

#endif
