/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ServerLogHandler.h"
#include <common/base/Queue.h>
#include <dataprovider/message/log/WriteServerLogMessage.h>
#include <dataprovider/message/ShutdownMessage.h>
#include <common/base/Date.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <boost/filesystem/operations.hpp>

using namespace common;
using namespace std;

namespace dataprovider {

const std::string ServerLogHandler::SERVER_LOG_PATH = "serverLogPath";

ServerLogHandler* ServerLogHandler::_instance = 0;

ServerLogHandler* ServerLogHandler::getInstance() {
	return _instance;
}

ServerLogHandler::ServerLogHandler() :
	_running(false),
	_messageQueue(nullptr),
	_thread(nullptr) {
}

ServerLogHandler::~ServerLogHandler() {

	delete _messageQueue;
}
	
void ServerLogHandler::initInstance() {
	if (_instance == 0) {
		_instance = new ServerLogHandler();
		_instance->init();
	}
}

void ServerLogHandler::resetInstance() {

	if (_instance != 0) {
		_instance->reset();
	}
	
	delete _instance;
	_instance = 0;
}
	
void ServerLogHandler::writeLog(const std::string& processName, unsigned long threadID, common::OutputType type, const std::string& line) {

	_messageQueue->put(new WriteServerLogMessage(processName, threadID, type, line));
}

void ServerLogHandler::init() {

	_messageQueue = new Queue<Message>(1024);
	_thread = new std::thread(&ServerLogHandler::loop);
}

void ServerLogHandler::reset() {
	
	_messageQueue->put(new ShutdownMessage());
	
	_thread->join();
	delete _thread;
}

void ServerLogHandler::loop() {
	
	Message* message = 0;
	_instance->_running = true;
	while (_instance->_running) {
		
		try {
			try {		
				message = _instance->_messageQueue->get();
				
			} catch (Queue<Message>::UnavailableException& e) {
				break;
				
			} catch (Exception& e) {
				err << e.printStack() << eol;
				break;
			}
			
			try {
				_instance->processMessage(message);
	
			} catch (Exception& e) {
				err << e.printStack() << eol;
			}
		
		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}
		
		delete message;
		message = 0;
 	}
}

void ServerLogHandler::processMessage(Message* message) {
	
	if (message->getType() == Message::WRITE_SERVER_LOG) {
		processMessage(static_cast<WriteServerLogMessage*>(message));

	} else if (message->getType() == Message::SHUTDOWN) {
		processMessage(static_cast<ShutdownMessage*>(message));

	} 
}

void ServerLogHandler::processMessage(WriteServerLogMessage* message) {

	Date date;
	string logFileName = ServerProperties::getInstance()->getHomePath() + ServerProperties::getInstance()->getString(SERVER_LOG_PATH) + "/nomad" + date.toString("%y%m%d") + ".log";
	
	// Open file, create if it doesn't exist
	ofstream logOutput(logFileName.c_str(), ios::app); 

	// Append the output
	try {
		string type;
		
		if (message->getLogType() == common::NORMAL) {
			type = "out";
		} else if (message->getLogType() == common::ERROR) {
			type = "err";
		} else if (message->getLogType() == common::DEBUGGING) {
			type = "dbg";
		}

		logOutput << message->getProcessName() << " - " << message->getThreadID() << " " << date.toString("%H:%M:%S") << " " << type << " : " << message->getLine() << endl;
		
	} catch (Exception& e) {
		cerr << "ServerLogHandler : an error occurred" << endl;
		cerr << e.printStack() << endl;
	}	
	
	// close log file
	logOutput.close();
}

void ServerLogHandler::processMessage(ShutdownMessage* message) {
	
	_running = false;
}

}
