/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "TextLogWriter.h"
#include "ControllerLog.h"
#include "CommandLog.h"
#include "SimpleLog.h"
#include "CommandZoneContentLog.h"
#include "CommandZoneEventLog.h"
#include "ForLoopStartLog.h"
#include "VariableChangeLog.h"
#include "BreakEvaluationLog.h"
#include "IfEvaluationLog.h"
#include "ServerLogHandler.h"
#include <dataprovider/database/Database.h>
#include <dataprovider/dataelement/Controller.h>
#include <dataprovider/dataelement/Driver.h>
#include <dataprovider/parser/commandzone/CommandZoneParser.h>
#include <dataprovider/parser/commandzone/element/ForLoopElement.h>
#include <dataprovider/parser/commandzone/element/ForLoopLineElement.h>
#include <dataprovider/parser/commandzone/element/Start.h>
#include <common/base/Date.h>
#include <common/base/ServerProperties.h>
#include <common/base/Convert.h>
#include <common/output/OutputStream.h>
#include <boost/filesystem/operations.hpp>
#include <boost/format.hpp>
#include <iostream>

using namespace common;
using namespace std;
using namespace boost::filesystem;

namespace dataprovider {

const std::string TextLogWriter::LOG_PATH = "logfiles";
const std::string TextLogWriter::DATE_FORMAT = "%-12s";
const std::string TextLogWriter::LEVEL_FORMAT = "%-8s";
const std::string TextLogWriter::DATA_FORMAT = "%-15s";

TextLogWriter::TextLogWriter() :
	_outputName("main"),
	_parser(new CommandZoneParser()) {

	_parser->init();
}

TextLogWriter::~TextLogWriter() {
	delete _parser;
}

void TextLogWriter::setOutputName(std::string const &name) {
	_outputName = name;
}

void TextLogWriter::write(Log* log, int n) {

	Date date = log->getDate();
	string logFileName = ServerProperties::getInstance()->getHomePath() + "/" + LOG_PATH + "/nomad" + date.toString("%y%m%d") + ".log";
//	string temporaryLogFileName = logFileName + ".temp";
//
//	// test if file already exists, othereles copy_file doesn't worked
//	if (exists(temporaryLogFileName) == true) {
//		remove(temporaryLogFileName);
//	}
//	if (exists(logFileName)) {
//		copy_file(logFileName, temporaryLogFileName);
//	}

	// Open file, create if it doesn't exist
	ofstream logOutput(logFileName.c_str(), ios::app);

	string logMessage = log->getMessage();
	if (n > 0) {
		logMessage += " (x " + std::to_string(n) + ")";
	}

	// Append the output
	try {
		if (log->getLevel() != common::LOG_DEBUG) {

			// converting XML character-sequences
			string formattedLog = convertXMLStringToAscii(format(log));

			if (formattedLog != "") {
				if (n > 0) {
					logOutput << formattedLog << " (x " << n << ")" << endl;
				}
				else {
					logOutput << formattedLog << endl;
				}
			}

			// Also redirect to server log
			if (log->getLevel() == common::LOG_ERROR) {
				// error log
				ServerLogHandler::getInstance()->writeLog(_outputName, pthread_self(), common::ERROR, logMessage);

			} else {
				if (dynamic_cast<ControllerLog*> (log) != 0 || dynamic_cast<SimpleLog*> (log) != 0) {
					ServerLogHandler::getInstance()->writeLog(_outputName, pthread_self(), common::NORMAL, logMessage);
				}
			}

		} else {
			// only redirect to server log in case of debug
			ServerLogHandler::getInstance()->writeLog(_outputName, pthread_self(), common::DEBUGGING, logMessage);
		}

	} catch (Exception& e) {
		err << "TextLogWriter : an error occurred" << eol;
		err << e.printStack() << eol;
	}

	// close log file
	logOutput.close();

//	// Set temporary file to used log file
//	if (rename(temporaryLogFileName.c_str(), logFileName.c_str()) != 0) {
//		cout << "Error renaming text log file" << eol;
//	}

}

std::string TextLogWriter::format(Log* log) {

	if (dynamic_cast<ControllerLog*>(log) != 0) {
		return format(dynamic_cast<ControllerLog*>(log));

	} else if (dynamic_cast<CommandLog*>(log) != 0) {
		return format(dynamic_cast<CommandLog*>(log));

	} else if (dynamic_cast<SimpleLog*>(log) != 0) {
		return format(dynamic_cast<SimpleLog*>(log));

	} else if (dynamic_cast<CommandZoneContentLog*>(log) != 0) {
		return format(dynamic_cast<CommandZoneContentLog*>(log));

	} else if (dynamic_cast<CommandZoneEventLog*>(log) != 0) {
		return format(dynamic_cast<CommandZoneEventLog*>(log));

	} else if (dynamic_cast<ControllerLog*>(log) != 0) {
		return format(dynamic_cast<ControllerLog*>(log));

	} else if (isASimpleTextLogMessage(log)) {
		//Process the rest of messages as a simple message!
		return formatOtherTypeOfMessages(log);
	}

	return "";
}

std::string TextLogWriter::format(ControllerLog* log) {

	string output = "";
	string name;
	if (log->getServantID() != 0) {
		name = Database::getInstance()->getServantName(log->getServantID());
	} else {
		std::cerr << "Error related to controller log message: servant ID is 0!" << std::endl;
	}
	string time = log->getDate().toString("%H:%M:%S");
	string level = createLevelString(log->getLevel());

//	output += formatColumn(time, DATE_FORMAT) + formatColumn(level, LEVEL_FORMAT) + formatColumn(name) + log->getMessage();
	output += formatColumn(time, DATE_FORMAT) + formatColumn(level, LEVEL_FORMAT) + log->getMessage();

	return output;
}

std::string TextLogWriter::format(CommandLog* log) {

	if (log->getType() == common::END) {
		return "";
	}

	string output = "";
	string time = log->getDate().toString("%H:%M:%S");
	string level = createLevelString(log->getLevel());

	output += formatColumn(time, DATE_FORMAT) + formatColumn(level, LEVEL_FORMAT) + log->getMessage();

	return output;
}

//NOTE: this method is not used. Replaced by "formatOtherTypeOfMessages"
std::string TextLogWriter::format(SimpleLog* log) {

	string output = "";
	string time = log->getDate().toString("%H:%M:%S");
	string level = createLevelString(log->getLevel());

	output += formatColumn(time, DATE_FORMAT) + formatColumn(level, LEVEL_FORMAT) + log->getMessage();

	return output;
}

std::string TextLogWriter::format(CommandZoneContentLog* log) {

//	_parser->parseXMLString(log->getMessage());
//	CommandZoneElement* content = _parser->getContent();
//	string commandZoneContent = formatCommandZoneContent(content);
//	string output = "";
//
//	if (commandZoneContent != "") {
//
//		string time = log->getDate().toString("%H:%M:%S");
//		string level = createLevelString(log->getLevel());
//
//		delete content;
//
//		output += formatColumn(time, DATE_FORMAT) + formatColumn(level, LEVEL_FORMAT) + commandZoneContent;
//	}
//	return output;

	return "";
}

std::string TextLogWriter::format(CommandZoneEventLog* log) {

	string output = "";
	string time = log->getDate().toString("%H:%M:%S");
	string level = createLevelString(log->getLevel());

	output += formatColumn(time, DATE_FORMAT) + formatColumn(level, LEVEL_FORMAT) + log->getMessage();

	return output;
}

std::string TextLogWriter::formatOtherTypeOfMessages(Log* log) {

	string output = "";
	string time = log->getDate().toString("%H:%M:%S");
	string level = createLevelString(log->getLevel());

	output += formatColumn(time, DATE_FORMAT) + formatColumn(level, LEVEL_FORMAT) + log->getMessage();

	return output;
}

bool TextLogWriter::isASimpleTextLogMessage(Log* log) {
	//Careful!! We take into account not only SimpleLog messages but also ForLoopStartLog, VariableChangeLog, BreakEvaluationLog and IfEvaluationLog
	if ((dynamic_cast<SimpleLog*>(log) != 0) || (dynamic_cast<ForLoopStartLog*>(log) != 0) || (dynamic_cast<VariableChangeLog*>(log) != 0) || (dynamic_cast<BreakEvaluationLog*>(log) != 0) || (dynamic_cast<IfEvaluationLog*>(log) != 0)) {
		return true;
	}
	return false;
}

std::string TextLogWriter::createLevelString(common::LogLevel level) {
	if (level == common::LOG_INFO) {
		return "info";
	} else if (level == common::LOG_WARNING) {
		return "warning";
	} else if (level == common::LOG_ERROR) {
		return "error";
	} else if (level == common::LOG_DEBUG) {
		return "debug";
	}
	return "";
}

std::string TextLogWriter::formatColumn(const std::string& value, const std::string& format) {
	return boost::str(boost::format(format) % value);
}

std::string TextLogWriter::formatCommandZoneContent(CommandZoneElement* content) {

	Start* start = dynamic_cast<Start*>(content);

	if (start != 0) {

		if (start->getChildren().size() == 1) {

			CommandZoneElement* element = start->getChildren()[0];

			ForLoopElement* forLoopElement = dynamic_cast<ForLoopElement*>(element);

			if (forLoopElement != 0) {

				vector<string> lines;
				string content;

				// iterating the children of the ForLoopElement that can be either of any CommandZoneElement type or ForLoopLineElement describing the dimension
				// of the loop
				for (vector<CommandZoneElement*>::const_iterator e = forLoopElement->getChildren().begin(); e != forLoopElement->getChildren().end(); e++) {

					ForLoopLineElement* lineElement = dynamic_cast<ForLoopLineElement*>(*e);
					if (lineElement != 0) {

						if (lineElement->getType() == "range") {
							lines.push_back(string("{") + lineElement->getVariableName() + " = " + lineElement->getStartValue() + " to " + lineElement->getEndValue() + " step " + lineElement->getStep() + "}");
						} else {
							lines.push_back(string("{") + lineElement->getVariableName() + " = " + lineElement->getValues() + "}");
						}

					}
				}

				string name = "For";

				for (unsigned int i = 0; i < lines.size() - 1; i++) {
					name += " " + lines[i] + ",";
				}

				name += " " + lines.back();

				return name;
			}
		}
	}

	return "";
}

}
