/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "XMLLogWriter.h"
#include "ControllerLog.h"
#include "CommandLog.h"
#include "SimpleLog.h"
#include "VariableChangeLog.h"
#include "ForLoopStartLog.h"
#include "BreakEvaluationLog.h"
#include "IfEvaluationLog.h"
#include "CommandZoneContentLog.h"
#include "CommandZoneEventLog.h"
#include <dataprovider/database/Database.h>
#include <dataprovider/dataelement/Controller.h>
#include <dataprovider/dataelement/Driver.h>
#include <dataprovider/parser/commandzone/CommandZoneParser.h>
#include <dataprovider/parser/commandzone/element/AtomicElement.h>
#include <dataprovider/parser/commandzone/element/ScanElement.h>
#include <dataprovider/parser/commandzone/element/ForLoopElement.h>
#include <dataprovider/parser/commandzone/element/ForLoopLineElement.h>
#include <dataprovider/parser/commandzone/element/GenericElement.h>
#include <dataprovider/parser/commandzone/element/CommandZone.h>
#include <dataprovider/parser/commandzone/element/Start.h>
#include <dataprovider/parser/commandzone/element/IfElement.h>
#include <dataprovider/parser/commandzone/element/BreakElement.h>
#include <common/base/Date.h>
#include <common/base/Convert.h>
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <boost/filesystem/operations.hpp>
#include <iostream>

using namespace common;
using namespace std;
using namespace boost::filesystem;

namespace dataprovider {

const std::string XMLLogWriter::LOG_PATH = "logfiles";


XMLLogWriter::XMLLogWriter() :
	_parser(new CommandZoneParser()) {

	_parser->init();
}

XMLLogWriter::~XMLLogWriter() {

	delete _parser;
}

void XMLLogWriter::write(Log* log, int n) {

	Date date = log->getDate();
	string logFileName = ServerProperties::getInstance()->getHomePath() + "/" + LOG_PATH + "/nomad" + date.toString("%y%m%d") + ".log.xml";
//	string temporaryLogFileName = logFileName + ".temp";
//
//	// test if file already exists, othereles copy_file doesn't worked
//	if (exists(temporaryLogFileName) == true) {
//		remove(temporaryLogFileName);
//	}
//	if (exists(logFileName)) {
//		copy_file(logFileName, temporaryLogFileName);
//	}

	// Open file, create if it doesn't exist
	fstream logOutput(logFileName.c_str(), ios::out | ios::in);
	logOutput.seekp(0, ios::end);

	// Determine if the file is new or not
	long position = logOutput.tellp();
	if (position == -1) {
		// open a new file
		logOutput.open(logFileName.c_str(), ios::out);
		if(!logOutput.is_open()) {
	    	err << "XMLLogWriter : cannot create file " << logFileName << "!" << eol;
	    	return;
		}
		logOutput << createLogHeader(date);
	} else {
		// move to position before closing log tag
		logOutput.seekp(-7, ios::end);
	}

	string message;

	// Append the output
	try {
		message = format(log, n);
		logOutput << message;
		logOutput << "</log>\n";

	} catch (Exception& e) {
		err << "XMLLogWriter : an error occurred" << eol;
		err << e.printStack() << eol;
	}

	// close log file
	logOutput.close();

//	// Set temporary file to used log file
//	if (rename(temporaryLogFileName.c_str(), logFileName.c_str()) != 0) {
//		cout << "Error renaming xml log file" << eol;
//	}

	// send XML log to subscribers
	if (message != "") {
		NotificationServer::getInstance()->logXML(message);
	}
}

std::string XMLLogWriter::format(Log* log, int n) {

	// Only printing non-debugging logs
	if (log->getLevel() != common::LOG_DEBUG) {

		if (dynamic_cast<ControllerLog*>(log) != 0) {
			return format(dynamic_cast<ControllerLog*>(log), n);

		} else if (dynamic_cast<CommandLog*>(log) != 0) {
			return format(dynamic_cast<CommandLog*>(log));

		} else if (dynamic_cast<CommandZoneContentLog*>(log) != 0) {
			return format(dynamic_cast<CommandZoneContentLog*>(log));

		} else if (dynamic_cast<CommandZoneEventLog*>(log) != 0) {
			return format(dynamic_cast<CommandZoneEventLog*>(log));

		} else if (isASimpleTextLogMessage(log)) {
			//Process the rest of messages as a simple message!
			return formatOtherTypeOfMessages(log);
		}
	}
	return "";
}

std::string XMLLogWriter::format(ControllerLog* log, int n) {

	string output = "";
	string time = log->getDate().toString("%H:%M:%S");
	string numor;
	string level = createLevelString(log->getLevel());

	if (log->getType() == common::ACQ_START) {
		output += createNumorStart(time, numor);
	}

	string logMessage = log->getMessage();
	if (n > 0) {
		logMessage += " (x " + std::to_string(n) + ")";
	}

	if (log->getCommandID() != 0 && Database::getInstance()->isCommandFromCommandZone(log->getCommandID())) {
		string logID = Database::getInstance()->getCommandLogID(log->getCommandID());
		output += string("\t<command_entry id=\"") + logID + "\" time=\"" + time + "\">\n";
		output += string("\t\t<") + level + " value=\"" + convertStringToXML(logMessage) + "\"/>\n";
		output += string("\t</command_entry>\n\n");

	} else {

		if (log->getServantID() != 0) {
			string type = getServantType(log->getServantID());
			string name = getServantName(log->getServantID());
			output += string("\t<controller_entry time=\"") + time + "\" ctype=\"" + type + "\" cname=\"" + name + "\">\n";

		} else {
			output += string("\t<controller_entry time=\"") + time + "\">\n";
			std::cerr << "Error related to controller log message: servant ID is 0!" << std::endl;
		}

		output += string("\t\t<") + level + " value=\"" + convertStringToXML(logMessage) + "\"/>\n";
		output += string("\t</controller_entry>\n\n");
	}
	if (log->getType() == common::ACQ_END) {
		output += createNumorEnd(time);
	}

	return output;
}

std::string XMLLogWriter::format(CommandLog* log) {

	string output = "";
	string time = log->getDate().toString("%H:%M:%S");
	string level = createLevelString(log->getLevel());

	if (log->getType() == common::START) {
		output += string("\t<command_entry id=\"") + log->getLogID() + "\" time=\"" + time + "\">\n";
		output += string("\t\t<") + level + " value=\"" + convertStringToXML(log->getMessage()) + "\"/>\n";
		output += string("\t</command_entry>\n\n");

	} else if (log->getType() == common::END) {
		output += string("\t<command_end id=\"") + log->getLogID() + "\" time=\"" + time + "\"/>\n\n";
	}

	return output;
}

//NOTE: this method is not used. Replaced by "formatOtherTypeOfMessages"
std::string XMLLogWriter::format(SimpleLog* log) {

	string output = "";
	string time = log->getDate().toString("%H:%M:%S");
	string level = createLevelString(log->getLevel());

	output += string("\t<simple_entry time=\"") + time + "\">\n";
	output += string("\t\t<") + level + " value=\"" + convertStringToXML(log->getMessage()) + "\"/>\n";
	output += string("\t</simple_entry>\n\n");

	return output;
}

std::string XMLLogWriter::format(CommandZoneContentLog* log) {

	_parser->parseXMLString(log->getMessage());
	CommandZoneElement* content = _parser->getContent();

	string output = formatCommandZoneContent(content) + "\n";

	delete content;

	return output;
}

std::string XMLLogWriter::format(CommandZoneEventLog* log) {

	string output = "";
	string time = log->getDate().toString("%H:%M:%S");
	string level = createLevelString(log->getLevel());

	output += string("\t<command_zone time=\"") + time + "\">\n";
	output += string("\t\t<") + level + " value=\"" + convertStringToXML(log->getMessage()) + "\"/>\n";
	output += string("\t</command_zone>\n\n");

	return output;
}

std::string XMLLogWriter::formatOtherTypeOfMessages(Log* log) {

	string output = "";
	string time = log->getDate().toString("%H:%M:%S");
	string level = createLevelString(log->getLevel());

	output += string("\t<simple_entry time=\"") + time + "\">\n";
	output += string("\t\t<") + level + " value=\"" + convertStringToXML(log->getMessage()) + "\"/>\n";
	output += string("\t</simple_entry>\n\n");

	return output;
}

std::string XMLLogWriter::createNumorStart(const std::string& time, const std::string& numor) {
	return string("\t<numor_start time=\"") + time + "\" numor=\"" + numor + "\"/>\n\n";
}

std::string XMLLogWriter::createNumorEnd(const std::string& time) {
	return string("\t<numor_end time=\"") + time + "\"/>\n\n";
}

std::string XMLLogWriter::createLevelString(common::LogLevel level) {
	if (level == common::LOG_INFO) {
		return "info";
	} else if (level == common::LOG_WARNING) {
		return "warning";
	} else if (level == common::LOG_ERROR) {
		return "error";
	} else if (level == common::LOG_DEBUG) {
		return "debug";
	}

	return "";
}

std::string XMLLogWriter::createLogHeader(const Date& date) {
	return string("<log date=\"") + date.toString("%m/%d/%y") + "\" version=\"1.0\">\n";
}

std::string XMLLogWriter::getServantName(unsigned long servantID) const {

	return Database::getInstance()->getServantName(servantID);
}

std::string XMLLogWriter::getServantType(unsigned long servantID) const {

	return Database::getInstance()->getServantType(servantID);
}

bool XMLLogWriter::isASimpleTextLogMessage(Log* log) {
	//Careful!! We take into account not only SimpleLog messages but also ForLoopStartLog, VariableChangeLog, BreakEvaluationLog and IfEvaluationLog
	if ((dynamic_cast<SimpleLog*> (log) != 0) || (dynamic_cast<ForLoopStartLog*> (log) != 0) || (dynamic_cast<VariableChangeLog*> (log) != 0) || (dynamic_cast<BreakEvaluationLog*> (log) != 0) || (dynamic_cast<IfEvaluationLog*> (log) != 0)) {
		return true;
	}
	return false;
}

std::string XMLLogWriter::formatCommandZoneContent(CommandZoneElement* content, const std::string& indent) {

	AtomicElement* atomicElement = dynamic_cast<AtomicElement*>(content);

	if (atomicElement != 0) {
		return formatCommandZoneContent(atomicElement, indent);
	}

	GenericElement* genericElement = dynamic_cast<GenericElement*>(content);

	if (genericElement != 0) {
		return formatCommandZoneContent(genericElement, indent);
	}

	ScanElement* scanElement = dynamic_cast<ScanElement*>(content);

	if (scanElement != 0) {
		return formatCommandZoneContent(scanElement, indent);
	}

	ForLoopElement* forLoopElement = dynamic_cast<ForLoopElement*>(content);

	if (forLoopElement != 0) {
		return formatCommandZoneContent(forLoopElement, indent);
	}

	IfElement* ifElement = dynamic_cast<IfElement*>(content);

	if (ifElement != 0) {
		return formatCommandZoneContent(ifElement, indent);
	}

	CommandZone* commandZone = dynamic_cast<CommandZone*>(content);

	if (commandZone != 0) {
		return formatCommandZoneContent(commandZone, indent);
	}

	Start* start = dynamic_cast<Start*>(content);

	if (start != 0) {
		return formatCommandZoneContent(start, indent);
	}

	BreakElement* breakElement = dynamic_cast<BreakElement*>(content);

	if (breakElement != 0) {
		return formatCommandZoneContent(breakElement, indent);
	}

	return "";
}

std::string XMLLogWriter::formatCommandZoneContent(AtomicElement* element, const std::string& indent) {

	string cname = getServantName(Database::getInstance()->getServantID(element->getCommandID()));
	string ctype = getServantType(Database::getInstance()->getServantID(element->getCommandID()));

	if (element->isSettings()) {
		return indent + "<settings id=\"" + element->getLogID() + "\" name=\"" + cname + "\" settingsFile=\"" + element->getSettingsFileName() + "\"/>\n";
	} else {
		return indent + "<atomic id=\"" + element->getLogID() + "\" ctype=\"" + ctype + "\" cname=\"" + cname + "\"/>\n";
	}
}

std::string XMLLogWriter::formatCommandZoneContent(GenericElement* element, const std::string& indent) {

	string cname = getServantName(element->getControllerID());
	string ctype = getServantType(element->getControllerID());

	if (element->getBoxType() == GenericElement::FIX) {
		return indent + "<fix id=\"" + element->getLogID() + "\" ctype=\"" + ctype + "\" cname=\"" + cname + "\"/>\n";
	} else if (element->getBoxType() == GenericElement::CLEAR) {
		return indent + "<clear id=\"" + element->getLogID() + "\" ctype=\"" + ctype + "\" cname=\"" + cname + "\"/>\n";
	} else if (element->getBoxType() == GenericElement::SET) {
		return indent + "<set id=\"" + element->getLogID() + "\" ctype=\"" + ctype + "\" cname=\"" + cname + "\"/>\n";
	} else if (element->getBoxType() == GenericElement::SETZERO) {
		return indent + "<set_zero id=\"" + element->getLogID() + "\" ctype=\"" + ctype + "\" cname=\"" + cname + "\"/>\n";
	}

	return "";
}

std::string XMLLogWriter::formatCommandZoneContent(ScanElement* element, const std::string& indent) {

	string result = indent + "<scan id=\"" + element->getLogID() + "\" name=\"" + element->getName() + "\">\n";

	for (vector<CommandZoneElement*>::const_iterator e = element->getChildren().begin(); e != element->getChildren().end(); e++) {
		result += formatCommandZoneContent(*e, indent + "\t");
	}

	result += indent + "</scan>\n";

	return result;
}

std::string XMLLogWriter::formatCommandZoneContent(ForLoopElement* element, const std::string& indent) {

	vector<string> lines;
	string content;

	// iterating the children of the ForLoopElement that can be either of any CommandZoneElement type or ForLoopLineElement describing the dimension
	// of the loop
	for (vector<CommandZoneElement*>::const_iterator e = element->getChildren().begin(); e != element->getChildren().end(); e++) {

		ForLoopLineElement* lineElement = dynamic_cast<ForLoopLineElement*>(*e);
		if (lineElement != 0) {

			if (lineElement->getType() == "range") {
				lines.push_back(string("{") + lineElement->getVariableName() + " = " + lineElement->getStartValue() + " to " + lineElement->getEndValue() + " step " + lineElement->getStep() + "}");
			} else {
				lines.push_back(string("{") + lineElement->getVariableName() + " = " + lineElement->getValues() + "}");
			}

		} else {
			content += formatCommandZoneContent(*e, indent + "\t");
		}
	}

	string name = "loop for";

	for (unsigned int i = 0; i < lines.size() - 1; i++) {
		name += " " + lines[i] + ",";
	}

	name += " " + lines.back();

	string result = indent + "<forloop id=\"" + element->getLogID() + "\" name=\"" + name + "\">\n";
	result += content;
	result += indent + "</forloop>\n";

	return result;
}

std::string XMLLogWriter::formatCommandZoneContent(CommandZone* element, const std::string& indent) {

	string result = indent + "<command_zone>\n";

	for (vector<CommandZoneElement*>::const_iterator e = element->getChildren().begin(); e != element->getChildren().end(); e++) {
		result += formatCommandZoneContent(*e, indent + "\t");
	}

	result += indent + "</command_zone>\n";

	return result;
}

std::string XMLLogWriter::formatCommandZoneContent(Start* element, const std::string& indent) {

	string result = indent + "<command_start>\n";

	for (vector<CommandZoneElement*>::const_iterator e = element->getChildren().begin(); e != element->getChildren().end(); e++) {
		result += formatCommandZoneContent(*e, indent + "\t");
	}

	result += indent + "</command_start>\n";

	return result;
}

std::string XMLLogWriter::formatCommandZoneContent(IfElement* element, const std::string& indent) {

	string content;

	if (element->getChildren().size() > 0) {

		CommandZoneElement* thenElement = element->getChildren()[0];

		content += indent + "\t<then>\n";
		for (vector<CommandZoneElement*>::const_iterator e = thenElement->getChildren().begin(); e != thenElement->getChildren().end(); e++) {
			content += formatCommandZoneContent(*e, indent + "\t\t");
		}
		content += indent + "\t</then>\n";
	}

	if (element->getChildren().size() > 1) {

		CommandZoneElement* elseElement = element->getChildren()[1];

		content += indent + "\t<else>\n";
		for (vector<CommandZoneElement*>::const_iterator e = elseElement->getChildren().begin(); e != elseElement->getChildren().end(); e++) {
			content += formatCommandZoneContent(*e, indent + "\t\t");
		}
		content += indent + "\t</else>\n";
	}

	string condition = "if ";
	condition += element->getCondition();

	string result = indent + "<if id=\"" + element->getLogID() + "\" condition=\"" + convertStringToXML(condition) + "\">\n";
	result += content;
	result += indent + "</if>\n";

	return result;
}

std::string XMLLogWriter::formatCommandZoneContent(BreakElement* element, const std::string& indent) {

	string condition = "if ";
	condition += element->getCondition();

	return indent + "<break id=\"" + element->getLogID() + "\" condition=\"" + convertStringToXML(condition) + "\"/>\n";
}

}
