/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DataChangePublisher.h"
#include <dataprovider/transportlayer/common/DataChangeSubscriber.h>
#include <NotificationMessages.pb.h>
#include <common/output/OutputStream.h>
#include <vector>
#include <iostream>

using namespace std;
using namespace common;
using namespace cameo;

namespace dataprovider {

unsigned long DataChangePublisher::m_counter = 0;

DataChangePublisher::DataChangePublisher() {

	string dataChangePublisherName = "data_change_publisher";

	// Create the publisher.
	m_publisher = coms::Publisher::create(dataChangePublisherName);
	m_publisher->init();

	cout << "Created data change publisher " << dataChangePublisherName << endl;
}

DataChangePublisher::~DataChangePublisher() {
	std::map<unsigned long, DataChangeSubscriber*>::iterator it;
	for (it = m_subscribers.begin(); it != m_subscribers.end(); it++) {
		delete it->second;
	}

	cout << "Deleted data change publisher" << endl;
}

unsigned long DataChangePublisher::subscribe(DataChangeSubscriber* subscriber) {

	std::scoped_lock lock(m_mutex);

	unsigned long subscriberID = subscriber->getID();

	// check to see if subscriber already exists
	if (m_subscribers.find(subscriberID) != m_subscribers.end()) {
		delete subscriber;
		return subscriberID;
	}

	unsigned long nextID = ++m_counter;
	m_subscribers[nextID] = subscriber;

	//slog << "added subscriber <" << nextID << ">" << eol;

	return nextID;

}

void DataChangePublisher::unsubscribe(unsigned long subscriberID) {

	std::scoped_lock lock(m_mutex);

	if (m_subscribers.find(subscriberID) != m_subscribers.end()) {
		delete m_subscribers[subscriberID];
		m_subscribers.erase(subscriberID);
	}
}


void DataChangePublisher::notifyPropertyChanged(unsigned long databaseID, unsigned long propertyID) {

	{
		std::scoped_lock lock(m_mutex);

		// Notify all subscribers
		std::map<unsigned long, DataChangeSubscriber*>::iterator it;
		for (it = m_subscribers.begin(); it != m_subscribers.end(); it++) {

			it->second->propertyChanged(databaseID, propertyID);
		}
	}

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::PropertyChanged);

	// Create and send the message.
	notification::PropertyChanged message;

	message.set_databaseid(databaseID);
	message.set_propertyid(propertyID);

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyCommandStateChanged(unsigned long databaseID, unsigned long servantID, unsigned long commandID, common::CommandState state) {

	{
		std::scoped_lock lock(m_mutex);

		// Notify all subscribers
		std::map<unsigned long, DataChangeSubscriber*>::iterator it;
		for (it = m_subscribers.begin(); it != m_subscribers.end(); it++) {

			it->second->commandStateChanged(databaseID, servantID, commandID, state);
		}
	}

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::CommandStateChanged);

	// Create and send the message.
	notification::CommandStateChanged message;

	message.set_databaseid(databaseID);
	message.set_servantid(servantID);
	message.set_commandid(commandID);

	if (state == common::INACTIVE) {
		message.set_state(common::CommandStateType::INACTIVE);
	}
	else if (state == common::PENDING) {
		message.set_state(common::CommandStateType::PENDING);
	}
	else if (state == common::STARTING) {
		message.set_state(common::CommandStateType::STARTING);
	}
	else if (state == common::ACTIVE) {
		message.set_state(common::CommandStateType::ACTIVE);
	}
	else if (state == common::PAUSED) {
		message.set_state(common::CommandStateType::PAUSED);
	}
	else if (state == common::PAUSING) {
		message.set_state(common::CommandStateType::PAUSING);
	}
	else if (state == common::STOPPED) {
		message.set_state(common::CommandStateType::STOPPED);
	}
	else if (state == common::STOPPING) {
		message.set_state(common::CommandStateType::STOPPING);
	}
	else if (state == common::RESUMING) {
		message.set_state(common::CommandStateType::RESTARTING);
	}
	else {
		message.set_state(common::CommandStateType::INACTIVE);
	}

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyCommandInterruptChanged(unsigned long databaseID, unsigned long servantID, unsigned long commandID, bool value) {

	std::scoped_lock lock(m_mutex);

	// Notify all subscribers
	std::map<unsigned long, DataChangeSubscriber*>::iterator it;
	for (it = m_subscribers.begin(); it != m_subscribers.end(); it++) {

		it->second->commandInterruptChanged(databaseID, servantID, commandID, value);
	}
}

void DataChangePublisher::notifyCommandProgressionChanged(unsigned long databaseID, unsigned long commandID, double progression) {

	{
		std::scoped_lock lock(m_mutex);

		// Notify all subscribers
		std::map<unsigned long, DataChangeSubscriber*>::iterator it;
		for (it = m_subscribers.begin(); it != m_subscribers.end(); it++) {

			it->second->commandProgressionChanged(databaseID, commandID, progression);
		}
	}

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::CommandProgressionChanged);

	// Create and send the message.
	notification::CommandProgressionChanged message;

	message.set_databaseid(databaseID);
	message.set_commandid(commandID);
	message.set_progression(progression);

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyConfigurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState state, bool clientNotification) {

	{
		std::scoped_lock lock(m_mutex);

		// Notify all subscribers
		std::map<unsigned long, DataChangeSubscriber*>::iterator it;
		for (it = m_subscribers.begin(); it != m_subscribers.end(); it++) {

			it->second->configurationChanged(databaseID, servantID, state);
		}
	}

	// If the notification is also done for the clients.
	if (clientNotification) {
		// Create the type.
		notification::Message messageType;
		messageType.set_type(notification::Message::ConfigurationChanged);

		// Create and send the message.
		notification::ConfigurationChanged message;

		message.set_databaseid(databaseID);
		message.set_servantid(servantID);

		if (state == common::ENABLED) {
			message.set_enable(common::ConfigurationStateType::ENABLED);
		}
		else if (state == common::DISABLED) {
			message.set_enable(common::ConfigurationStateType::DISABLED);
		}
		else if (state == common::PROPERTY) {
			message.set_enable(common::ConfigurationStateType::PROPERTY);
		}
		else if (state == common::TREE) {
			message.set_enable(common::ConfigurationStateType::TREE);
		}
		else if (state == common::CONTROLLER_ADDED) {
			message.set_enable(common::ConfigurationStateType::CONTROLLER_ADDED);
		}
		else if (state == common::CONTROLLER_REMOVED) {
			message.set_enable(common::ConfigurationStateType::CONTROLLER_REMOVED);
		}
		else {
			message.set_enable(common::ConfigurationStateType::TREE);
		}

		// Publish the message
		m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());

		//slog << "Published configuration changed " << servantID << eol;
	}
}

void DataChangePublisher::notifyConditionChanged(unsigned long conditionId, bool on, common::ConditionState state) {

	{
		std::scoped_lock lock(m_mutex);

		// Notify all subscribers
		std::map<unsigned long, DataChangeSubscriber*>::iterator it;
		for (it = m_subscribers.begin(); it != m_subscribers.end(); it++) {

			it->second->conditionStateChanged(conditionId, on, state);
		}
	}

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::ConditionStateChanged);

	// Create and send the message.
	notification::ConditionStateChanged message;

	message.set_conditionid(conditionId);
	message.set_on(on);

	if (state == common::CONDITION_ACTIVE) {
		message.set_state(common::ConditionStateType::CONDITION_ACTIVE);
	}
	else if (state == common::CONDITION_INACTIVE) {
		message.set_state(common::ConditionStateType::CONDITION_INACTIVE);
	}
	else if (state == common::CONDITION_ON_ACTIVATION_DELAY) {
		message.set_state(common::ConditionStateType::CONDITION_ON_ACTIVATION_DELAY);
	}
	else {
		message.set_state(common::ConditionStateType::CONDITION_INACTIVE);
	}

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyConditionActionsChanged(unsigned long conditionId, common::ConditionState state) {

	{
		std::scoped_lock lock(m_mutex);

		// Notify all subscribers
		std::map<unsigned long, DataChangeSubscriber*>::iterator it;
		for (it = m_subscribers.begin(); it != m_subscribers.end(); it++) {

			it->second->conditionActionsStateChanged(conditionId, state);
		}
	}

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::ConditionActionsStateChanged);

	// Create and send the message.
	notification::ConditionActionsStateChanged message;

	message.set_conditionid(conditionId);

	if (state == common::CONDITION_ACTIVE) {
		message.set_state(common::ConditionStateType::CONDITION_ACTIVE);
	}
	else if (state == common::CONDITION_INACTIVE) {
		message.set_state(common::ConditionStateType::CONDITION_INACTIVE);
	}
	else if (state == common::CONDITION_ON_ACTIVATION_DELAY) {
		message.set_state(common::ConditionStateType::CONDITION_ON_ACTIVATION_DELAY);
	}
	else {
		message.set_state(common::ConditionStateType::CONDITION_INACTIVE);
	}

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyUserVariableValueChanged(unsigned long databaseID, const std::string& name, double value) {

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::UserVariableValueChanged);

	// Create and send the message.
	notification::UserVariableValueChanged message;

	message.set_databaseid(databaseID);
	message.set_variablename(name);
	message.set_value(value);

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyUserVariableStateChanged(unsigned long databaseID, const std::string& name, bool enabled) {

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::UserVariableStateChanged);

	// Create and send the message.
	notification::UserVariableStateChanged message;

	message.set_databaseid(databaseID);
	message.set_variablename(name);

	if (enabled) {
		message.set_state(common::ConfigurationStateType::ENABLED);
	}
	else {
		message.set_state(common::ConfigurationStateType::DISABLED);
	}

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyTokenStateChanged(common::CriticalState criticalState, common::RemoteControlState remoteControlState, common::TokenState state, uint32 ownerClientID, const std::string& ownerEndpoint, const std::string& user) {

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::TokenState);

	// Create and send the message.
	notification::TokenState message;

	if (criticalState == common::CriticalState::DISABLED) {
		message.set_criticalstate(notification::TokenState::DISABLED);
	}
	else if (criticalState == common::CriticalState::ENABLED) {
		message.set_criticalstate(notification::TokenState::ENABLED);
	}

	if (remoteControlState == common::RemoteControlState::DISABLED) {
		message.set_remotecontrolstate(notification::TokenState::DISABLED);
	}
	else if (remoteControlState == common::RemoteControlState::ENABLED) {
		message.set_remotecontrolstate(notification::TokenState::ENABLED);
	}

	if (state == common::TokenState::FREE) {
		message.set_state(notification::TokenState::FREE);
	}
	else if (state == common::TokenState::OWNED) {
		message.set_state(notification::TokenState::OWNED);
	}
	else {
		message.set_state(notification::TokenState::FREE);
	}

	message.set_ownerclientid(ownerClientID);
	message.set_ownerendpoint(ownerEndpoint);
	message.set_user(user);

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyChatMessage(uint32 clientID, int64 date, const std::string& sender, const std::string& chatMessage, const std::string& fileName) {

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::Chat);

	// Create and send the message.
	notification::ChatMessage message;

	message.set_clientid(clientID);
	message.set_date(date);
	message.set_sender(sender);
	message.set_message(chatMessage);
	message.set_filename(fileName);

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyConnectedUsersChanged() {

	// Create the type.
	notification::Message messageType;
	messageType.set_type(notification::Message::ConnectedUsersChanged);

	// Create and send the message.
	notification::EmptyMessage message;

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

void DataChangePublisher::notifyNewFileCreated(common::FileType fileType, const std::string& fileName) {

	// Create the type.
	notification::Message messageType;

	messageType.set_type(notification::Message::NewFileCreated);

	// Create and send the message.
	notification::NewFileCreated message;

	switch (fileType) {
	case FileType::SCRIPT_FILE:
		message.set_type(notification::NewFileCreated::SCRIPT_FILE);
		break;
	case FileType::SETTINGS_FILE:
		message.set_type(notification::NewFileCreated::SETTINGS_FILE);
		break;
	case FileType::PAL_FILE:
		message.set_type(notification::NewFileCreated::PAL_FILE);
		break;
	case FileType::XBU_FILE:
		message.set_type(notification::NewFileCreated::XBU_FILE);
		break;
	case FileType::ATTACHED_FILE:
		message.set_type(notification::NewFileCreated::ATTACHED_FILE);
		break;
	}

	message.set_filename(fileName);

	// Publish the message
	m_publisher->sendTwoParts(messageType.SerializeAsString(), message.SerializeAsString());
}

}
