/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_DATACHANGEPUBLISHER_H_
#define DATAPROVIDER_DATACHANGEPUBLISHER_H_

#include <dataprovider/dataelement/Command.h>
#include <common/base/CommonTypes.h>
#include <cameo/api/cameo.h>
#include <map>
#include <mutex>

namespace dataprovider {

class DataChangeSubscriber;

/**
 * Class defining a data change publisher. The publisher adds subscriber for which it delivers notifications.
 */
class DataChangePublisher {

public:
	/**
	 * Constructor.
	 */
	DataChangePublisher();

	/**
	 * Virtual destructor.
	 */
	virtual ~DataChangePublisher();

	/**
	 * Adds a subscriber.
	 */
	unsigned long subscribe(DataChangeSubscriber* subscriber);

	/**
	 * Removes a subscriber.
	 */
	virtual void unsubscribe(unsigned long subscriberID);

	/**
	 * Notifies the subscribers that a property has changed.
	 */
	void notifyPropertyChanged(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Notifies the subscribers that a command state has changed.
	 */
	void notifyCommandStateChanged(unsigned long databaseID, unsigned long servantID, unsigned long commandID, common::CommandState state);

	/**
	 * Notifies the subscribers that a command interrupt has changed.
	 */
	void notifyCommandInterruptChanged(unsigned long databaseID, unsigned long servantID, unsigned long commandID, bool value);

	/**
	 * Notifies the subscribers that a command progression has changed.
	 */
	void notifyCommandProgressionChanged(unsigned long databaseID, unsigned long commandID, double progression);

	/**
	 * Notifies the subscribers that a configuration has changed.
	 */
	void notifyConfigurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState state, bool clientNotification);

	/**
	 * Notifies the condition state.
	 */
	void notifyConditionChanged(unsigned long conditionId, bool on, common::ConditionState state);

	/**
	 * Notifies the condition actions state.
	 */
	void notifyConditionActionsChanged(unsigned long conditionId, common::ConditionState state);

	/**
	 * Notifies the user variable value changed.
	 */
	void notifyUserVariableValueChanged(unsigned long databaseID, const std::string& name, double value);

	/**
	 * Notifies the user variable value changed.
	 */
	void notifyUserVariableStateChanged(unsigned long databaseID, const std::string& name, bool enabled);

	/**
	 * Notifies the token changed.
	 */
	void notifyTokenStateChanged(common::CriticalState criticalState, common::RemoteControlState remoteControlState, common::TokenState state, uint32 ownerClientID, const std::string& ownerEndpoint, const std::string& user);

	/**
	 * Notifies a chat message.
	 */
	void notifyChatMessage(uint32 clientID, int64 date, const std::string& sender, const std::string& chatMessage, const std::string& fileName);

	/**
	 * Notifies connected users changed.
	 */
	void notifyConnectedUsersChanged();

	/**
	 * Notifies that a new file has been created.
	 */
	void notifyNewFileCreated(common::FileType fileType, const std::string& fileName);

private:
	// We keep now the DataChangeSubscriber pattern for the internal server updates.
	static unsigned long m_counter;
	std::map<unsigned long, DataChangeSubscriber*> m_subscribers;

	std::mutex m_mutex;

	// Publisher for the client classes.
	std::unique_ptr<cameo::coms::Publisher> m_publisher;
};

}

#endif /*DATAPROVIDER_DATACHANGEPUBLISHERIMPL_H_*/
