/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "NotificationServer.h"
#include "DataChangePublisher.h"
#include "LogPublisher.h"
#include "SurveyPublisher.h"
#include <dataprovider/message/notification/PropertyChangedMessage.h>
#include <dataprovider/message/notification/CommandStateChangedMessage.h>
#include <dataprovider/message/notification/CommandInterruptChangedMessage.h>
#include <dataprovider/message/notification/CommandProgressionChangedMessage.h>
#include <dataprovider/message/notification/ConfigurationChangedMessage.h>
#include <dataprovider/message/notification/EnableConfigurationChangedMessage.h>
#include <dataprovider/message/notification/IcsPropertyChangedMessage.h>
#include <dataprovider/message/notification/ConditionStateChangedMessage.h>
#include <dataprovider/message/notification/ConditionActionsStateChangedMessage.h>
#include <dataprovider/message/notification/UserVariableValueChangedMessage.h>
#include <dataprovider/message/notification/UserVariableStateChangedMessage.h>
#include <dataprovider/message/notification/LogMessage.h>
#include <dataprovider/message/notification/LogXMLMessage.h>
#include <dataprovider/message/notification/SurveyXMLMessage.h>
#include <dataprovider/message/notification/NewFileMessage.h>
#include <dataprovider/message/notification/LogDataImageReadyMessage.h>
#include <dataprovider/message/notification/TokenStateChangedMessage.h>
#include <dataprovider/message/notification/ChatMessage.h>
#include <dataprovider/message/notification/ConnectedUsersChangedMessage.h>
#include <dataprovider/message/notification/NewFileCreatedMessage.h>
#include <dataprovider/database/Database.h>
#include <common/base/Queue.h>
#include <common/output/OutputStream.h>
#include <boost/bind.hpp>
#include <iostream>

using namespace std;
using namespace common;

namespace dataprovider {

NotificationServer* NotificationServer::m_instance = 0;

NotificationServer::NotificationServer(const std::string& serverId) :
	_dataChangePublisher(new DataChangePublisher()),
	_logPublisher(new LogPublisher(serverId)),
	_surveyPublisher(new SurveyPublisher(serverId)),
	_queue(new common::Queue<NotificationMessage>(1024)),
	_thread(nullptr),
	_clientNotification(true) {
}

NotificationServer::~NotificationServer() {
	delete _queue;
	delete _dataChangePublisher;
	delete _logPublisher;
	delete _surveyPublisher;
}

void NotificationServer::initInstance(const std::string& serverId) {
	if (m_instance == 0) {
		m_instance = new NotificationServer(serverId);
		m_instance->start();
	}
}

void NotificationServer::resetInstance() {

	if (m_instance != 0) {
		m_instance->stop();
		delete m_instance;
		m_instance = 0;
	}
}

NotificationServer* NotificationServer::getInstance() {
	return m_instance;
}

DataChangePublisher* NotificationServer::getDataChangePublisher() {
	return _dataChangePublisher;
}

LogPublisher* NotificationServer::getLogPublisher() {
	return _logPublisher;
}

SurveyPublisher* NotificationServer::getSurveyPublisher() {
	return _surveyPublisher;
}

void NotificationServer::start() {
	_thread = new std::thread(&NotificationServer::run);
}

void NotificationServer::stop() {

	if (_thread != nullptr) {
		_queue->stop();
		_thread->join();
		delete _thread;
	}
}

void NotificationServer::propertyChanged(unsigned long databaseID, unsigned long propertyID) {
	_queue->put(new PropertyChangedMessage(databaseID, propertyID));
}

void NotificationServer::commandStateChanged(unsigned long databaseID, unsigned long servantID, unsigned long commandID, common::CommandState state)  {
	_queue->put(new CommandStateChangedMessage(databaseID, servantID, commandID, state));
}

void NotificationServer::commandInterruptChanged(unsigned long databaseID, unsigned long servantID, unsigned long commandID, bool value) {
	_queue->put(new CommandInterruptChangedMessage(databaseID, servantID, commandID, value));
}

void NotificationServer::commandProgressionChanged(unsigned long databaseID, unsigned long commandID, double progression) {
	_queue->put(new CommandProgressionChangedMessage(databaseID, commandID, progression));
}

void NotificationServer::configurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState enable) {
	_queue->put(new ConfigurationChangedMessage(databaseID, servantID, enable));
}

void NotificationServer::enableConfigurationChanged(unsigned long databaseID) {
	_queue->put(new EnableConfigurationChangedMessage(databaseID, true));
}

void NotificationServer::disableConfigurationChanged(unsigned long databaseID) {
	_queue->put(new EnableConfigurationChangedMessage(databaseID, false));
}

void NotificationServer::icsPropertyChanged(unsigned long databaseID, unsigned long propertyID) {
	_queue->put(new IcsPropertyChangedMessage(databaseID, propertyID));
}

void NotificationServer::conditionStateChanged(unsigned long conditionId, bool on, common::ConditionState state) {
	_queue->put(new ConditionStateChangedMessage(conditionId, on, state));
}

void NotificationServer::conditionActionsStateChanged(unsigned long conditionId, common::ConditionState state) {
	_queue->put(new ConditionActionsStateChangedMessage(conditionId, state));
}

void NotificationServer::userVariableValueChanged(unsigned long databaseId, const std::string& name, double value) {
	_queue->put(new UserVariableValueChangedMessage(databaseId, name, value));
}

void NotificationServer::userVariableStateChanged(unsigned long databaseId, const std::string& name, bool enabled) {
	_queue->put(new UserVariableStateChangedMessage(databaseId, name, enabled));
}

void NotificationServer::log(common::LogType type, common::LogLevel level, const std::string& message, unsigned long commandID, unsigned long servantID) {
	_queue->put(new LogMessage(type, level, message, commandID, servantID));
}

void NotificationServer::imageDataReady(const std::string& image) {
	_queue->put(new LogDataImageReadyMessage(image));
}

void NotificationServer::logXML(const std::string& message) {
	_queue->put(new LogXMLMessage(message));
}

void NotificationServer::surveyXML(const std::string& message) {
	_queue->put(new SurveyXMLMessage(message));
}

void NotificationServer::newFile(const std::string& message) {
	_queue->put(new NewFileMessage(message));
}

void NotificationServer::tokenStateChanged(common::CriticalState criticalState, common::RemoteControlState remoteControlState, common::TokenState state, uint32 ownerClientID, const std::string& ownerEndpoint, const std::string& user) {
	_queue->put(new TokenStateChangedMessage(criticalState, remoteControlState, state, ownerClientID, ownerEndpoint, user));
}

void NotificationServer::chatMessage(uint32 clientID, int64 date, const std::string& sender, const std::string& message, const std::string& fileName) {
	_queue->put(new ChatMessage(clientID, date, sender, message, fileName));
}

void NotificationServer::notifyConnectedUsersChanged() {
	_queue->put(new ConnectedUsersChangedMessage());
}

void NotificationServer::notifyNewFileCreated(common::FileType fileType, const std::string& fileName) {
	_queue->put(new NewFileCreatedMessage(fileType, fileName));
}

/**
 * Main loop
 */
void NotificationServer::run() {

	bool running = true;
	while (running) {

		NotificationMessage* message = 0;

		try {
			message = m_instance->_queue->get();

			m_instance->processMessage(message);

		} catch (common::Queue<NotificationMessage>::UnavailableException& e) {
			running = false;
		} catch (Exception& e) {
			err << "Error in " << message->getTypeString() << eol;
			err << e.printStack() << eol;
		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}

		delete message;
	}
}

void NotificationServer::processMessage(NotificationMessage* message) {

	if (message->getType() == NotificationMessage::PROPERTY_CHANGED) {
		processMessage(static_cast<PropertyChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::COMMAND_STATE_CHANGED) {
		processMessage(static_cast<CommandStateChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::COMMAND_INTERRUPT_CHANGED) {
		processMessage(static_cast<CommandInterruptChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::COMMAND_PROGRESSION_CHANGED) {
		processMessage(static_cast<CommandProgressionChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::CONFIGURATION_CHANGED) {
		processMessage(static_cast<ConfigurationChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::ENABLE_CONFIGURATION_CHANGED) {
		processMessage(static_cast<EnableConfigurationChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::ICS_PROPERTY_CHANGED) {
		processMessage(static_cast<IcsPropertyChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::CONDITION_STATE_CHANGED) {
		processMessage(static_cast<ConditionStateChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::CONDITION_ACTIONS_STATE_CHANGED) {
		processMessage(static_cast<ConditionActionsStateChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::USER_VARIABLE_VALUE_CHANGED) {
		processMessage(static_cast<UserVariableValueChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::USER_VARIABLE_STATE_CHANGED) {
		processMessage(static_cast<UserVariableStateChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::LOG) {
		processMessage(static_cast<LogMessage*>(message));

	} else if (message->getType() == NotificationMessage::LOG_XML) {
		processMessage(static_cast<LogXMLMessage*>(message));

	} else if (message->getType() == NotificationMessage::SURVEY_XML) {
		processMessage(static_cast<SurveyXMLMessage*>(message));

	} else if (message->getType() == NotificationMessage::LOG_DATAREADY) {
		processMessage(static_cast<LogDataImageReadyMessage*>(message));

	} else if (message->getType() == NotificationMessage::TOKEN_STATE_CHANGED) {
		processMessage(static_cast<TokenStateChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::CHAT_MESSAGE) {
		processMessage(static_cast<ChatMessage*>(message));

	} else if (message->getType() == NotificationMessage::CONNECTED_USERS_CHANGED) {
		processMessage(static_cast<ConnectedUsersChangedMessage*>(message));

	} else if (message->getType() == NotificationMessage::NEW_FILE_CREATED) {
		processMessage(static_cast<NewFileCreatedMessage*>(message));
	}
}

void NotificationServer::processMessage(PropertyChangedMessage* message) {

	try {
		_dataChangePublisher->notifyPropertyChanged(message->getDatabaseID(), message->getPropertyID());

	} catch (Database::NoSuchPropertyException&) {
		// the property doesn't exist anymore, but has existed
		// this is because there is no synchronisation between the DataNotificationServer and the removeProperty
	}
}

void NotificationServer::processMessage(CommandStateChangedMessage* message) {

	_dataChangePublisher->notifyCommandStateChanged(message->getDatabaseID(), message->getServantID(), message->getCommandID(), message->getCommandState());
}

void NotificationServer::processMessage(CommandInterruptChangedMessage* message) {

	_dataChangePublisher->notifyCommandInterruptChanged(message->getDatabaseID(), message->getServantID(), message->getCommandID(), message->getValue());
}

void NotificationServer::processMessage(CommandProgressionChangedMessage* message) {

	_dataChangePublisher->notifyCommandProgressionChanged(message->getDatabaseID(), message->getCommandID(), message->getProgression());
}

void NotificationServer::processMessage(ConfigurationChangedMessage* message) {

	_dataChangePublisher->notifyConfigurationChanged(message->getDatabaseID(), message->getServantID(), message->getState(), _clientNotification);
}

void NotificationServer::processMessage(EnableConfigurationChangedMessage* message) {

	_clientNotification = message->getValue();
}

void NotificationServer::processMessage(IcsPropertyChangedMessage* message) {

	_dataChangePublisher->notifyPropertyChanged(message->getDatabaseID(), message->getPropertyID());
}

void NotificationServer::processMessage(ConditionStateChangedMessage* message) {

	_dataChangePublisher->notifyConditionChanged(message->getConditionId(), message->isOn(), message->getState());
}

void NotificationServer::processMessage(ConditionActionsStateChangedMessage* message) {

	_dataChangePublisher->notifyConditionActionsChanged(message->getConditionId(), message->getState());
}

void NotificationServer::processMessage(UserVariableValueChangedMessage* message) {

	_dataChangePublisher->notifyUserVariableValueChanged(message->getDatabaseID(), message->getName(), message->getValue());
}

void NotificationServer::processMessage(UserVariableStateChangedMessage* message) {

	_dataChangePublisher->notifyUserVariableStateChanged(message->getDatabaseID(), message->getName(), message->isEnabled());
}

void NotificationServer::processMessage(LogMessage* message) {

	_logPublisher->notifyLog(message->getLogType(), message->getLogLevel(), message->getMessage(), message->getCommandID(), message->getServantID());
}

void NotificationServer::processMessage(LogXMLMessage* message) {

	_logPublisher->notifyLogXML(message->getMessage());
}

void NotificationServer::processMessage(SurveyXMLMessage* message) {

	_surveyPublisher->notifySurveyXML(message->getMessage());
}

void NotificationServer::processMessage(LogDataImageReadyMessage* message) {

	_logPublisher->notifyLogImageDataReady(message->getImage());
}

void NotificationServer::processMessage(TokenStateChangedMessage* message) {

	_dataChangePublisher->notifyTokenStateChanged(message->getCriticalState(), message->getRemoteControlState(), message->getState(), message->getOwnerClientID(), message->getOwnerEndpoint(), message->getUser());
}

void NotificationServer::processMessage(ChatMessage* message) {

	_dataChangePublisher->notifyChatMessage(message->getClientID(), message->getDate(), message->getSender(), message->getMessage(), message->getFileName());
}

void NotificationServer::processMessage(ConnectedUsersChangedMessage* message) {

	_dataChangePublisher->notifyConnectedUsersChanged();
}

void NotificationServer::processMessage(NewFileCreatedMessage* message) {

	_dataChangePublisher->notifyNewFileCreated(message->getFileType(), message->getFileName());
}

}
