/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_NOTIFICATIONSERVER_H_
#define DATAPROVIDER_NOTIFICATIONSERVER_H_

#include <dataprovider/dataelement/Command.h>
#include <dataprovider/log/Log.h>
#include <common/base/CommonTypes.h>
#include <common/base/HashMap.h>
#include <thread>

namespace common {
	template<typename ValueType>
	class Queue;
}

namespace dataprovider {

class NotificationMessage;
class PropertyChangedMessage;
class CommandStateChangedMessage;
class CommandInterruptChangedMessage;
class CommandProgressionChangedMessage;
class ConfigurationChangedMessage;
class EnableConfigurationChangedMessage;
class IcsPropertyChangedMessage;
class ConditionStateChangedMessage;
class ConditionActionsStateChangedMessage;
class UserVariableValueChangedMessage;
class UserVariableStateChangedMessage;
class LogMessage;
class LogXMLMessage;
class SurveyXMLMessage;
class DataChangePublisher;
class LogPublisher;
class SurveyPublisher;
class LogDataImageReadyMessage;
class TokenStateChangedMessage;
class ChatMessage;
class ConnectedUsersChangedMessage;
class NewFileCreatedMessage;

/**
 * Class defining the data notification server. The class is singleton and processes messages sequentially
 * through a message queue to forward the notification to the data change publisher.
 */
class NotificationServer {

public:
	/**
	 * Gets the instance.
	 */
	static NotificationServer* getInstance();

	/**
	 * Initialises the instance.
	 */
	static void initInstance(const std::string& serverId);

	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Gets the data change publisher.
	 */
	DataChangePublisher* getDataChangePublisher();

	/**
	 * Gets the log publisher.
	 */
	LogPublisher* getLogPublisher();

	/**
	 * Gets the log publisher.
	 */
	SurveyPublisher* getSurveyPublisher();

	/**
	 * Starts the data notification server.
	 */
	void start();

	/**
	 * Stops the data notification server.
	 */
	void stop();

	/**
	 * Notifies that a property has changed.
	 */
	void propertyChanged(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Notifies that a command state has changed.
	 */
	void commandStateChanged(unsigned long databaseID, unsigned long servantID, unsigned long commandID, common::CommandState state);

	/**
	 * Notifies that a command interrupt has changed.
	 */
	void commandInterruptChanged(unsigned long databaseID, unsigned long servantID, unsigned long commandID, bool value);

	/**
	 * Notifies that a command progression has changed.
	 */
	void commandProgressionChanged(unsigned long databaseID, unsigned long commandID, double progression);

	/**
	 * Notifies that a configuration has changed.
	 */
	void configurationChanged(unsigned long databaseID, unsigned long servantID, common::ConfigurationState enable);

	/**
	 * Enables the configuration changed.
	 */
	void enableConfigurationChanged(unsigned long databaseID);

	/**
	 * Disables the configuration changed.
	 */
	void disableConfigurationChanged(unsigned long databaseID);

	/**
	 * Notifies that a property has changed.
	 */
	void icsPropertyChanged(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Notifies the condition state.
	 */
	void conditionStateChanged(unsigned long conditionId, bool on, common::ConditionState state);

	/**
	 * Notifies the condition actions state.
	 */
	void conditionActionsStateChanged(unsigned long conditionId, common::ConditionState state);

	/**
	 * Notifies the user variable value changed.
	 */
	void userVariableValueChanged(unsigned long databaseId, const std::string& name, double value);

	/**
	 * Notifies the user variable state changed.
	 */
	void userVariableStateChanged(unsigned long databaseId, const std::string& name, bool enabled);

	/**
	 * Logs the message.
	 */
	void log(common::LogType type, common::LogLevel level, const std::string& message, unsigned long commandID = 0, unsigned long servantID = 0);


	/**
	 * Data image ready log message
	 */
	void imageDataReady(const std::string& image);

	/**
	 * Logs the XML message.
	 */
	void logXML(const std::string& message);

	/**
	 * Notifies a property survey.
	 */
	void surveyXML(const std::string& message);


	/**
	 * Notifies a new file in client side (user script, settings, xbu or pal file)
	 */
	void newFile(const std::string& message);

	/**
	 * Notifies a token state change.
	 */
	void tokenStateChanged(common::CriticalState criticalState, common::RemoteControlState remoteControlState, common::TokenState state, uint32 ownerClientID, const std::string& ownerEndpoint, const std::string& user);

	/**
	 * Notifies a chat message.
	 */
	void chatMessage(uint32 clientID, int64 date, const std::string& sender, const std::string& message, const std::string& fileName);

	/**
	 * Notifies connected users changed.
	 */
	void notifyConnectedUsersChanged();

	/**
	 * Notifies that a new file has been created.
	 */
	void notifyNewFileCreated(common::FileType fileType, const std::string& fileName);

private:
	NotificationServer(const std::string& serverId);
	virtual ~NotificationServer();
	static void run();

	void processMessage(NotificationMessage* message);
	void processMessage(PropertyChangedMessage* message);
	void processMessage(CommandStateChangedMessage* message);
	void processMessage(CommandInterruptChangedMessage* message);
	void processMessage(CommandProgressionChangedMessage* message);
	void processMessage(ConfigurationChangedMessage* message);
	void processMessage(EnableConfigurationChangedMessage* message);
	void processMessage(IcsPropertyChangedMessage* message);
	void processMessage(ConditionStateChangedMessage* message);
	void processMessage(ConditionActionsStateChangedMessage* message);
	void processMessage(UserVariableValueChangedMessage* message);
	void processMessage(UserVariableStateChangedMessage* message);
	void processMessage(LogMessage* message);
	void processMessage(LogXMLMessage* message);
	void processMessage(SurveyXMLMessage* message);
	void processMessage(LogDataImageReadyMessage* message);
	void processMessage(TokenStateChangedMessage* message);
	void processMessage(ChatMessage* message);
	void processMessage(ConnectedUsersChangedMessage* message);
	void processMessage(NewFileCreatedMessage* message);

	static NotificationServer* m_instance;

	DataChangePublisher* _dataChangePublisher;
	LogPublisher* _logPublisher;
	SurveyPublisher* _surveyPublisher;

	common::Queue<NotificationMessage>* _queue;
	std::thread* _thread;

	bool _clientNotification;
};


} // namespace dataprovider

#endif /*DATAPROVIDER_NOTIFICATIONSERVER_H_*/
