/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandZoneParser.h"
#include <dataprovider/parser/commandzone/element/AtomicElement.h>
#include <dataprovider/parser/commandzone/element/ScanElement.h>
#include <dataprovider/parser/commandzone/element/ForLoopElement.h>
#include <dataprovider/parser/commandzone/element/ForLoopLineElement.h>
#include <dataprovider/parser/commandzone/element/GenericElement.h>
#include <dataprovider/parser/commandzone/element/CommandZone.h>
#include <dataprovider/parser/commandzone/element/Start.h>
#include <dataprovider/parser/commandzone/element/IfElement.h>
#include <dataprovider/parser/commandzone/element/Then.h>
#include <dataprovider/parser/commandzone/element/Else.h>
#include <dataprovider/parser/commandzone/element/BreakElement.h>

#include <common/base/Convert.h>
#include <iostream>

using namespace std;
using namespace common;

namespace dataprovider {

const std::string CommandZoneParser::ZONE_TAG = "zone";
const std::string CommandZoneParser::START_TAG = "start";

const std::string CommandZoneParser::SCAN_TAG = "scan";
const std::string CommandZoneParser::ATOMIC_TAG = "atomic";
const std::string CommandZoneParser::FOR_LOOP_TAG = "for_loop";
const std::string CommandZoneParser::FOR_LOOP_LINE_TAG = "for_loop_line";

const std::string CommandZoneParser::FIX_TAG = "fix";
const std::string CommandZoneParser::CLEAR_TAG = "clear";
const std::string CommandZoneParser::SET_TAG = "set";
const std::string CommandZoneParser::SET_ZERO_TAG = "set_zero";

const std::string CommandZoneParser::IF_TAG = "if";
const std::string CommandZoneParser::THEN_TAG = "then";
const std::string CommandZoneParser::ELSE_TAG = "else";
const std::string CommandZoneParser::BREAK_TAG = "break";

const std::string CommandZoneParser::NAME_ATTRIBUTE = "name";
const std::string CommandZoneParser::ID_ATTRIBUTE = "id";
const std::string CommandZoneParser::LOG_ID_ATTRIBUTE = "logId";

const std::string CommandZoneParser::COMMAND_ID_ATTRIBUTE = "commandId";
const std::string CommandZoneParser::CONTROLLER_ID_ATTRIBUTE = "controllerId";
const std::string CommandZoneParser::IS_SETTINGS_ATTRIBUTE = "isSettings";
const std::string CommandZoneParser::SETTINGS_FILE_NAME_ATTRIBUTE = "settingsFile";


const std::string CommandZoneParser::VARIABLE_NAME_ATTRIBUTE = "variableName";
const std::string CommandZoneParser::TYPE_ATTRIBUTE = "type";
const std::string CommandZoneParser::START_VALUE_ATTRIBUTE = "startValue";
const std::string CommandZoneParser::END_VALUE_ATTRIBUTE = "endValue";
const std::string CommandZoneParser::STEP_ATTRIBUTE = "step";
const std::string CommandZoneParser::VALUES_ATTRIBUTE = "values";
const std::string CommandZoneParser::CONDITION_ATTRIBUTE = "condition";

CommandZoneParser::CommandZoneParser() :
	common::GenericParser(),
	_root(0) {
}

CommandZoneParser::~CommandZoneParser() {
}


void CommandZoneParser::startElement(const std::string& qName) {

	if (qName == ZONE_TAG) {
		_root = new CommandZone();
		_currentElementStack.push_back(_root);

	} else if (qName == START_TAG) {
		_root = new Start();
		_currentElementStack.push_back(_root);

	} else if (qName == ATOMIC_TAG) {
		CommandZoneElement* element = new AtomicElement(to<unsigned long>(getAttribute(ID_ATTRIBUTE)), getAttribute(LOG_ID_ATTRIBUTE), to<unsigned long>(getAttribute(COMMAND_ID_ATTRIBUTE)), getAttribute(IS_SETTINGS_ATTRIBUTE) == "true", getAttribute(SETTINGS_FILE_NAME_ATTRIBUTE));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == SCAN_TAG) {
		CommandZoneElement* element = new ScanElement(to<unsigned long>(getAttribute(ID_ATTRIBUTE)), getAttribute(LOG_ID_ATTRIBUTE), getAttribute(NAME_ATTRIBUTE));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == FOR_LOOP_TAG) {
		CommandZoneElement* element = new ForLoopElement(to<unsigned long>(getAttribute(ID_ATTRIBUTE)), getAttribute(LOG_ID_ATTRIBUTE));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == FOR_LOOP_LINE_TAG) {
		CommandZoneElement* element = new ForLoopLineElement(getAttribute(VARIABLE_NAME_ATTRIBUTE), getAttribute(TYPE_ATTRIBUTE), getAttribute(START_VALUE_ATTRIBUTE), getAttribute(END_VALUE_ATTRIBUTE), getAttribute(STEP_ATTRIBUTE), getAttribute(VALUES_ATTRIBUTE));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == FIX_TAG) {
		CommandZoneElement* element = new GenericElement(to<unsigned long>(getAttribute(ID_ATTRIBUTE)), getAttribute(LOG_ID_ATTRIBUTE), GenericElement::FIX, to<unsigned long>(getAttribute(CONTROLLER_ID_ATTRIBUTE)));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == CLEAR_TAG) {
		CommandZoneElement* element = new GenericElement(to<unsigned long>(getAttribute(ID_ATTRIBUTE)), getAttribute(LOG_ID_ATTRIBUTE), GenericElement::CLEAR, to<unsigned long>(getAttribute(CONTROLLER_ID_ATTRIBUTE)));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == SET_TAG) {
		CommandZoneElement* element = new GenericElement(to<unsigned long>(getAttribute(ID_ATTRIBUTE)), getAttribute(LOG_ID_ATTRIBUTE), GenericElement::SET, to<unsigned long>(getAttribute(CONTROLLER_ID_ATTRIBUTE)));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == SET_ZERO_TAG) {
		CommandZoneElement* element = new GenericElement(to<unsigned long>(getAttribute(ID_ATTRIBUTE)), getAttribute(LOG_ID_ATTRIBUTE), GenericElement::SETZERO, to<unsigned long>(getAttribute(CONTROLLER_ID_ATTRIBUTE)));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == IF_TAG) {
		CommandZoneElement* element = new IfElement(to<unsigned long>(getAttribute(ID_ATTRIBUTE)), getAttribute(LOG_ID_ATTRIBUTE), getAttribute(CONDITION_ATTRIBUTE));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == THEN_TAG) {
		CommandZoneElement* element = new Then();
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == ELSE_TAG) {
		CommandZoneElement* element = new Else();
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else if (qName == BREAK_TAG) {
		CommandZoneElement* element = new BreakElement(to<unsigned long>(getAttribute(ID_ATTRIBUTE)), getAttribute(LOG_ID_ATTRIBUTE), getAttribute(CONDITION_ATTRIBUTE));
		_currentElementStack.back()->addChild(element);
		_currentElementStack.push_back(element);

	} else {
		throw UnknownTagException(qName, _fileName);
	}

}

void CommandZoneParser::endElement(const std::string& qName) {

	_currentElementStack.pop_back();
}

CommandZoneElement* CommandZoneParser::getContent() const {

	return _root;
}

}
