/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_SERVANTELEMENT_H_
#define DATAPROVIDER_SERVANTELEMENT_H_

#include <common/base/Exception.h>
#include <common/base/CommonTypes.h>
#include <vector>
#include <string>

namespace dataprovider {

class PropertyElement;
class CommandElement;

/**
 * Class defining a Servant element for the ConfigurationParser. 
 */
class ServantElement {

public:
	/**
	 * Exception for child position failure.
	 */
	class ChildPositionException : public common::Exception {
	
	public:
		/**
		 * Constructor.
		 */
		ChildPositionException();
	};

	/**
	 * Exception for child name.
	 */
	class ChildNameException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		ChildNameException(const std::string parentName, const std::string nName, const std::string childName);
	};
	
	/**
	 * Constructor with the name and the type.
	 */
	ServantElement(const std::string& name, const std::string& type, bool isEnabled, bool isPausable);
	
	/**
	 * Virtual destructor.
	 */	
	virtual ~ServantElement();

	/**
	 * Sets the server type (ICS or NOMAD).
	 */
	void setServerType(common::ServerType serverType);

	/**
	 * Gets the server type (ICS or NOMAD).
	 */
	common::ServerType getServerType() const;

	/**
	 * Gets the name.
	 */
	const std::string& getName() const;
	
	/**
	 * Gets the type.
	 */
	const std::string& getType() const;

	/**
	 * Adds a property.
	 */
	void addProperty(PropertyElement* property);
	
	/**
	 * Gets the properties.
	 */
	const std::vector<PropertyElement*>& getProperties() const;

	/**
	 * Adds a command.
	 */
	void addCommand(CommandElement* command);

	/**
	 * Gets the commands.
	 */
	const std::vector<CommandElement*>& getCommands() const;

	/**
	 * Returns true if enabled.
	 */
	bool isEnabled() const;

	/**
	 * Returns true if pausable.
	 */
	bool isPausable() const;

private:
	common::ServerType _serverType;
	std::string _name;
	std::string _type;
	std::vector<PropertyElement*> _properties;
	std::vector<CommandElement*> _commands;
	bool _isEnabled;
	bool _isPausable;
};

}

#endif /*DATAPROVIDER_SERVANTELEMENT_H_*/