/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "MultiThreadCommunication.h"
#include <common/base/ServerProperties.h>
#include <common/output/OutputStream.h>
#include <dataprovider/transportlayer/dataprovider/PropertyAccessorDirectImpl.h>
#include <dataprovider/transportlayer/dataprovider/ServantAccessorDirectImpl.h>
#include <dataprovider/transportlayer/dataprovider/CommandAccessorDirectImpl.h>
#include <dataprovider/transportlayer/dataprovider/LogAccessorDirectImpl.h>
#include <dataprovider/transportlayer/dataprovider/SurveyAccessorDirectImpl.h>
#include <dataprovider/transportlayer/dataprovider/DataChangePublisherDirectImpl.h>
#include <dataprovider/transportlayer/dataprovider/DirectServerLogAccessor.h>
#include <iostream>
#include "dataprovider/StateAccessorDirectImpl.h"

using namespace std;
using namespace common;

namespace dataprovider {

MultiThreadCommunication * MultiThreadCommunication::_instance = 0;

MultiThreadCommunication::InitFailureException::InitFailureException(const string& problem) :
	Exception("dataprovider::MultiThreadCommunication::InitFailureException", problem) {

}

MultiThreadCommunication::DestructionFailureException::DestructionFailureException(const string& problem) :
	Exception("dataprovider::MultiThreadCommunication::DestructionFailureException", problem) {
}

MultiThreadCommunication::MultiThreadCommunication() {
}

MultiThreadCommunication::~MultiThreadCommunication() {
}

void MultiThreadCommunication::initInstance(int argc, char* argv[], bool remote, const std::string& serverID) {

	if (_instance == 0) {
		_instance = new MultiThreadCommunication();
		_instance->init(argc, argv, remote, serverID);
	}
}

MultiThreadCommunication * MultiThreadCommunication::getInstance() {

	return _instance;
}

void MultiThreadCommunication::resetInstance() {

	if (_instance != 0) {
		_instance->reset();
		delete _instance;
	}
}

void MultiThreadCommunication::init(int argc, char* argv[], bool remote, const std::string& serverID) {

	_serverId = serverID;


	try {
		// Initialise direct objects
		// Initialise PropertyAccessor
		_propertyAccessor = new PropertyAccessorDirectImpl();

		// Initialise the ServantAccessor
		_servantAccessor = new ServantAccessorDirectImpl();

		// Initialise the CommandAccessor
		_commandAccessor = new CommandAccessorDirectImpl();

		// Initialise the LogAccessor
		_logAccessor = new LogAccessorDirectImpl();
		_serverLogAccessor = new DirectServerLogAccessor();


		string outputName = "main";
		if (serverID != "") {
			outputName = string("sim") + serverID;
		}

		common::out.setServerLogAccessor(_serverLogAccessor);
		common::out.setProcessName(outputName);

		common::slog.setServerLogAccessor(_serverLogAccessor);
		common::slog.setProcessName(outputName);

		common::err.setServerLogAccessor(_serverLogAccessor);
		common::err.setProcessName(outputName);

		// Initialise the SurveyAccessor
		_surveyAccessor = new SurveyAccessorDirectImpl();

		// Initialise the DataChangePublisher
		_dataChangePublisher = new DataChangePublisherDirectImpl();

		_conditionAccessor = new StateAccessorDirectImpl();

		out << "Data provider communications ready" << eol;

	} catch (Exception const & e) {

		throw link(InitFailureException("cannot initialise"), e);
	}
}

void MultiThreadCommunication::reset() {

	// Delete the objects
	delete _propertyAccessor;
	delete _servantAccessor;
	delete _commandAccessor;
	delete _logAccessor;
	delete _surveyAccessor;
	delete _conditionAccessor;
	delete _serverLogAccessor;
	delete _dataChangePublisher;
}

void MultiThreadCommunication::run() {

}

void MultiThreadCommunication::stop() {

}

void MultiThreadCommunication::shutdown(bool isRemote) {
}

PropertyAccessorDirectImpl* MultiThreadCommunication::getPropertyAccessor() {
	return _propertyAccessor;
}

ServantAccessorDirectImpl* MultiThreadCommunication::getServantAccessor() {
	return _servantAccessor;
}

CommandAccessorDirectImpl* MultiThreadCommunication::getCommandAccessor() {
	return _commandAccessor;
}

DataChangePublisherDirectImpl* MultiThreadCommunication::getDataChangePublisher() {
	return _dataChangePublisher;
}

LogAccessorDirectImpl* MultiThreadCommunication::getLogAccessor() {
	return _logAccessor;
}

SurveyAccessorDirectImpl* MultiThreadCommunication::getSurveyAccessor() {
	return _surveyAccessor;
}

StateAccessorDirectImpl* MultiThreadCommunication::getPropertyExpressionAccessor() {
	return _conditionAccessor;
}

}
