/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_MULTITHREADCOMMUNICATION_H_
#define DATAPROVIDER_MULTITHREADCOMMUNICATION_H_

#include <common/base/Exception.h>
#include <common/base/TimeCondition.h>

namespace common {

	class ServerLogAccessor;
}

namespace dataprovider {

class PropertyAccessorDirectImpl;
class ServantAccessorDirectImpl;
class CommandAccessorDirectImpl;
class LogAccessorDirectImpl;
class SurveyAccessorDirectImpl;
class DataChangePublisherDirectImpl;
class StateAccessorDirectImpl;

/**
 * Class defining a multi-process Communication.
 */
class MultiThreadCommunication {

public :
	/**
	 * Defined exceptions.
	 */
	class InitFailureException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		InitFailureException(const std::string& problem);
	};

	class DestructionFailureException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		DestructionFailureException(const std::string& problem);
	};

	/**
	 * Initialises the instance.
	 * throw InitFailureException
	 */
	static void initInstance(int argc, char* argv[], bool remote, const std::string& serverID);

	/**
	 * Gets the instance.
	 */
	static MultiThreadCommunication * getInstance();

	/**
	 * Resets the instance.
	 * throw DestructionFailureException
	 */
	static void resetInstance();

	/**
	 * Runs the communication.
	 */
	void run();

	/**
	 * Stops the communication.
	 */
	void stop();

	/**
	 * Stops the communication.
	 */
	void shutdown(bool isRemote);

	/**
	 * Gets the property accessor.
	 */
	PropertyAccessorDirectImpl* getPropertyAccessor();

	/**
	 * Gets the servant accessor.
	 */
	ServantAccessorDirectImpl* getServantAccessor();

	/**
	 * Gets the command accessor.
	 */
	CommandAccessorDirectImpl* getCommandAccessor();

	/**
	 * Gets the data change publisher.
	 */
	DataChangePublisherDirectImpl* getDataChangePublisher();

	/**
	 * Gets the log accessor.
	 */
	LogAccessorDirectImpl* getLogAccessor();

	/**
	 * Gets the survey accessor.
	 */
	SurveyAccessorDirectImpl* getSurveyAccessor();

	/**
	 * Gets the condition accessor.
	 */
	StateAccessorDirectImpl* getPropertyExpressionAccessor();

private:
	/**
	 * Constructor.
	 */
	MultiThreadCommunication();

	/**
	 * Virtual destructor.
	 */
	virtual ~MultiThreadCommunication();

	/**
	 * Initialises the instance.
	 * Context.
	 * throw InitFailureException
	 */
	void init(int argc, char* argv[], bool remote, const std::string& serverID);

	/**
	 * Resets the communication objects.
	 * throw DestructionFailureException
	 */
	void reset();

	PropertyAccessorDirectImpl* _propertyAccessor;
	ServantAccessorDirectImpl* _servantAccessor;
	CommandAccessorDirectImpl* _commandAccessor;

	DataChangePublisherDirectImpl* _dataChangePublisher;

	LogAccessorDirectImpl* _logAccessor;
	SurveyAccessorDirectImpl* _surveyAccessor;
	StateAccessorDirectImpl* _conditionAccessor;
	common::ServerLogAccessor* _serverLogAccessor;

	bool _running;
	common::TimeCondition _condition;
	std::string _serverId;

	static MultiThreadCommunication * _instance;
};

}

#endif /*DATAPROVIDER_MULTITHREADCOMMUNICATION_H_*/
