/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_COMMANDACCESSORDIRECTIMPL_H_
#define DATAPROVIDER_COMMANDACCESSORDIRECTIMPL_H_

#include <boost/thread/mutex.hpp>
#include <common/base/Exception.h>
#include <dataprovider/database/Database.h>
#include <dataprovider/dataelement/Command.h>
#include <vector>
#include <common/utils/CommandDescriptor.h>

namespace common {

class DataChangeSubscriberDirect;

}

namespace dataprovider {

/**
 * Direct server implementation of the class defining a command accessor to transfer command related data.
 */
class CommandAccessorDirectImpl {

public:
	enum ActiveResult {EXECUTE, RECURSIVE, WAIT, SKIP};

	class NoSuchServantException : public common::Exception {
	public:
		NoSuchServantException();
	};

	class NoSuchCommandException : public common::Exception {
	public:
		NoSuchCommandException();
	};

	/**
	 * Constructor.
	 */
	CommandAccessorDirectImpl();

	/**
	 * Virtual destructor.
	 */
	virtual ~CommandAccessorDirectImpl();

	/**
	 * Gets the controller commands iterator by calling the database to fill the iterator.
	 */
	virtual std::vector<common::CommandDescriptor> getControllerCommands(unsigned long databaseID, unsigned long controllerID);

	/**
	 * Gets the driver commands iterator by calling the database to fill the iterator.
	 */
	virtual std::vector<common::CommandDescriptor> getDriverCommands(unsigned long databaseID, unsigned long driverID);

	/**
	 * Gets the servant ID from the command ID.
	 * throw NoSuchCommandException
	 */
	virtual unsigned long getServantID(unsigned long databaseID, unsigned long commandID);

	/**
	 * Gets the command ID.
	 * throw NoSuchServantException
	 */
	virtual unsigned long getCommandID(unsigned long databaseID, unsigned long servantID, const std::string& commandName);

	/**
	 * Gets the command name.
	 * throw NoSuchServantException
	 */
	virtual std::string getCommandName(unsigned long databaseID, unsigned long commandID);

	/**
	 * Sets the state of the command ID.
	 * throw NoSuchCommandException
	 */
	virtual void setCommandState(unsigned long databaseID, unsigned long servantID, unsigned long commandID, common::CommandState state);

	/**
	 * Sets the command active.
	 * throw NoSuchCommandException
	 */
	virtual ActiveResult setCommandActive(unsigned long databaseID, unsigned long servantID, unsigned long commandID, const std::string& servantType, unsigned long processorID, unsigned long threadID, common::DataChangeSubscriberDirect* subscriber);

	/**
	 * Gets the state of the command ID.
	 * throw NoSuchCommandException
	 */
	virtual common::CommandState getCommandState(unsigned long databaseID, unsigned long commandID);

	/**
	 * Returns the elapsed time of a command.
	 * throw NoSuchCommandException
	 */
	virtual double getCommandElapsedTimeMs(unsigned long databaseID, unsigned long commandID);

	/**
	 * Adds a command commandName to the servant of ID servantID.
	 * throw NoSuchServantException
	 */
	virtual unsigned long addCommandToServant(unsigned long databaseID, unsigned long servantID, const std::string& commandName);

	/**
	 * Gets the servant ID from the command ID.
	 * throw NoSuchCommandException
	 */
	virtual unsigned long getServantIDFromCommandID(unsigned long databaseID, unsigned long commandID);

	/**
	 * Gets the sister command ID from the command ID and command name.
	 * throw NoSuchCommandException
	 */
	virtual unsigned long getSisterCommandID(unsigned long databaseID, unsigned long commandID, const std::string& commandName);

	/**
	 * Returns true if command belongs to a ics servant.
	 * throw NoSuchCommandException
	 */
	virtual bool isCommandFromIcsServant(unsigned long databaseID, unsigned long commandID);

	/**
	 * Sets the progression.
	 * throw NoSuchCommandException
	 */
	virtual void setProgression(unsigned long databaseID, unsigned long commandID, double progression);

	/**
	 * Sets the progression unknown.
	 * throw NoSuchCommandException
	 */
	virtual void setUnknownProgression(unsigned long databaseID, unsigned long commandID);

	/**
	 * Gets the progression.
	 * throw NoSuchCommandException
	 */
	virtual double getProgression(unsigned long databaseID, unsigned long commandID);

	/**
	 * Gets the remaining time of a command.
	 * throw NoSuchCommandException
	 */
	virtual double getCommandRemainingTimeMs(unsigned long databaseID, unsigned long commandID);

	/**
	 * Gets the duration of a command.
	 * throw NoSuchCommandException
	 */
	virtual double getCommandDurationMs(unsigned long databaseID, unsigned long commandID);

	/**
	 * Gets the start date of the command.
	 */
	virtual int64 getCommandStartDate(unsigned long databaseID, unsigned long commandID);

	/**
	 * Sets the pre-run calculated duration of a command.
	 * throw NoSuchCommandException
	 */
	virtual void setCommandPreRunCalculatedDurationMs(unsigned long databaseID, unsigned long commandID, double duration);

	/**
	 * Gets the pre-run calculated duration of a command.
	 * throw NoSuchCommandException
	 */
	virtual double getCommandPreRunCalculatedDurationMs(unsigned long databaseID, unsigned long commandID);

	/**
 	 * Sets the command log ID.
 	 * throw NoSuchCommandException
 	 */
	virtual void setCommandLogID(unsigned long databaseID, unsigned long commandID, const std::string& logID);

	/**
	 * Sets the command from the command zone.
	 * throw NoSuchCommandException
	 */
	virtual void setCommandFromCommandZone(unsigned long databaseID, unsigned long commandID, bool fromCommandZone);

	/**
	 * Returns true if is a command.
	 */
	virtual bool commandExists(unsigned long databaseID, unsigned long commandID);

	/**
	 * Gets the description of the element.
	 */
	virtual std::string what(unsigned long databaseID, unsigned long elementID);

	/**
	 * Reports a bug.
	 */
	virtual void reportBug(int64 date);

private:
	std::vector<common::CommandDescriptor> createCommandList(const Database::CommandMap& commandMap);
};

}

#endif /*DATAPROVIDER_COMMANDACCESSORDIRECTIMPL_H_*/
