/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "LogAccessorDirectImpl.h"
#include <dataprovider/database/Database.h>
#include <dataprovider/log/LogHandler.h>
#include <dataprovider/log/CommandLog.h>
#include <dataprovider/log/SimpleLog.h>
#include <dataprovider/log/VariableChangeLog.h>
#include <dataprovider/log/ForLoopStartLog.h>
#include <dataprovider/log/ForLoopEndLog.h>
#include <dataprovider/log/IfEvaluationLog.h>
#include <dataprovider/log/BreakEvaluationLog.h>
#include <dataprovider/log/UserScanStartLog.h>
#include <dataprovider/log/ControllerLog.h>
#include <dataprovider/log/CommandZoneContentLog.h>
#include <dataprovider/log/CommandZoneEventLog.h>
#include <dataprovider/transportlayer/command/DirectLogCompletionHandler.h>
#include <dataprovider/log/ServerLogHandler.h>
#include <dataprovider/notification/NotificationServer.h>
#include <iostream>

using namespace std;
using namespace common;

namespace dataprovider {

LogAccessorDirectImpl::LogAccessorDirectImpl() {
}

LogAccessorDirectImpl::~LogAccessorDirectImpl() {
}
	
void LogAccessorDirectImpl::writeNLog(common::LogType type, common::LogLevel level, const std::list<common::LogElement>& lineElements, const std::string & imageName, unsigned long commandID, unsigned long servantID, bool writeLogFile, unsigned long requestID, command::LogCompletionHandlerDirectImpl* completionHandler) {
	// Only logging non-sync messages and write log file
	Log* log = new ControllerLog(type, level, lineElements, imageName, commandID, servantID);
	// Copying the string because log object is destroyed in parallel in dataprovider::LogHandler
	string logMessage = log->getMessage();
	if (writeLogFile && level != common::LOG_SYNC) {
		LogHandler::getInstance()->writeLog(log, requestID, new DirectLogCompletionHandler(completionHandler));
	} else {
		DirectLogCompletionHandler* logCompletionHandler = new DirectLogCompletionHandler(completionHandler);
		logCompletionHandler->terminate(requestID, true, "");

		delete logCompletionHandler;
	}

	// notifying
	NotificationServer::getInstance()->log(type, level, logMessage, commandID, servantID);
}

void LogAccessorDirectImpl::imageDataReadyLog(const std::string& image) {
	//Only notify the NotificationServer, do not send it to the logHandler
	NotificationServer::getInstance()->imageDataReady(image);

}

void LogAccessorDirectImpl::writeCommandLog(common::LogType type, common::LogLevel level, const std::string& logMessage, const std::string& logID) {
	// Only logging non-sync messages
	if (level != common::LOG_SYNC) {
		Log* log = new CommandLog(type, level, logMessage, logID);
		LogHandler::getInstance()->writeLog(log);
	}

	// notifying
	NotificationServer::getInstance()->log(type, level, logMessage);
}

void LogAccessorDirectImpl::writeSimpleLog(common::LogLevel level, const std::string& logMessage, bool writeLogFile) {
	// Only logging non-sync messages
	if (writeLogFile && level != common::LOG_SYNC) {
		Log* log = new SimpleLog(level, logMessage);
		LogHandler::getInstance()->writeLog(log);
	}
	// notifying
	NotificationServer::getInstance()->log(common::SINGLE, level, logMessage);
}

void LogAccessorDirectImpl::writeCommandZoneContentLog(const std::string& content) {
	Log* log = new CommandZoneContentLog(content);
	LogHandler::getInstance()->writeLog(log);
}

void LogAccessorDirectImpl::writeCommandZoneEventLog(const std::string& content) {
	Log* log = new CommandZoneEventLog(content);
	LogHandler::getInstance()->writeLog(log);
}

void LogAccessorDirectImpl::writeVariableChangeLog(common::LogLevel level, const std::string& variableName, const std::string& variableType, const std::string& variableValue, const std::string& message, const unsigned long messageParentId) {
	Log* log = new VariableChangeLog(level, variableName, variableType, variableValue, message, messageParentId);
	LogHandler::getInstance()->writeLog(log);
}

void LogAccessorDirectImpl::writeForLoopStartLog(common::LogLevel level, const std::string& variablesField, const std::string& message, unsigned long messageId) {
	Log* log = new ForLoopStartLog(level, variablesField, message, messageId);
	LogHandler::getInstance()->writeLog(log);
}

void LogAccessorDirectImpl::writeForLoopEndLog(common::LogLevel level, const std::string& message, unsigned long messageId) {
	Log* log = new ForLoopEndLog(level, message, messageId);
	LogHandler::getInstance()->writeLog(log);
}

void LogAccessorDirectImpl::writeIfEvaluationLog(common::LogLevel level, const std::string& name, const std::string& type, const std::string& value, const std::string& message) {
	Log* log = new IfEvaluationLog(level, name, type, value, message);
	LogHandler::getInstance()->writeLog(log);
}

void LogAccessorDirectImpl::writeBreakEvaluationLog(common::LogLevel level, const std::string& name, const std::string& type, const std::string& value, const std::string& message, const unsigned long messageParentId) {
	Log* log = new BreakEvaluationLog(level, name, type, value, message, messageParentId);
	LogHandler::getInstance()->writeLog(log);
}

void LogAccessorDirectImpl::writeUserScanStartLog(common::LogLevel level, const std::string& name, const std::string& message) {
	Log* log = new UserScanStartLog(level, name, message);
	LogHandler::getInstance()->writeLog(log);
}

void LogAccessorDirectImpl::writeServerLog(const std::string& processName, unsigned long threadID, bool error, const std::string& line) {
	
	common::OutputType logType;

	if (error) {
		logType = common::ERROR;

	} else {
		logType = common::NORMAL;
	}

	ServerLogHandler::getInstance()->writeLog(processName, threadID, logType, line);
}

}
