/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_PROPERTYACCESSORDIRECTIMPL_H_
#define DATAPROVIDER_PROPERTYACCESSORDIRECTIMPL_H_

#include <common/base/Exception.h>
#include <dataprovider/database/Database.h>
#include <vector>
#include <common/utils/PropertyDescriptor.h>

namespace dataprovider {

/**
 * Direct server implementation of the class defining a property accessor to transfer property related data.
 */
class PropertyAccessorDirectImpl {

public:
	enum DataType {BINARY, ASCII};

	class PropertyListCreationException {};
	class NoSuchServantException {};
	class NoSuchPropertyException {};
	class BadPropertyTypeException {};
	class IndexOutOfBoundsException {};
	class NotConfiguredException {};
	class NotResizableException {};
	class FileNotFoundException {};
	class NullPropertyReferenceException {};

	/**
	 * Constructor.
	 */
	PropertyAccessorDirectImpl();

	/**
	 * destructor.
	 */
	~PropertyAccessorDirectImpl();

	/**
	 * Gets the controller properties iterator by calling the database to fill the iterator.
	 * throw NoSuchServantException
	 */
	std::vector<common::PropertyDescriptor> getControllerProperties(unsigned long databaseID, unsigned long controllerID);

	/**
	 * Gets the driver properties iterator by calling the database to fill the iterator.
	 * throw NoSuchServantException
	 */
	std::vector<common::PropertyDescriptor> getDriverProperties(unsigned long databaseID, unsigned long driverID);

	/**
	 * Adds a property to the Servant (if it doesn't already exist) and returns the ID of the added property
	 * throw NoSuchServantException
	 */
	unsigned long addPropertyToServant(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const std::string& propertyType, bool spy, bool saveBox, bool & added);

	/**
	 * Adds a dynamic property to the Servant (if it doesn't already exist) and returns the ID of the added property
	 * throw NoSuchServantException
	 */
	unsigned long addDynamicPropertyToServant(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const std::string& propertyType, bool spy);

	/**
	 * Removes a property from the servant.
	 *  throw NoSuchServantException, NoSuchPropertyException
	 */
	void removePropertyFromServant(unsigned long databaseID, unsigned long servantID, unsigned long propertyID);

	/**
	 * Sets a the settings format and save to a property.
	 * throw nsdp::PropertyAccessor::NoSuchPropertyException
	 */
	void setPropertySettings(unsigned long databaseID, unsigned long propertyID, const std::string& propertyFormat, bool save, bool spy);

	/**
	 * Resizes the dynamic property.
	 * throw NoSuchPropertyException
 	 */
	void resizeDynamicProperty(unsigned long databaseID, unsigned long propertyID, int32 size);

	/**
	 * Gets the dynamic property size.
	 * throw NoSuchPropertyException
 	 */
	long getDynamicPropertySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the dynamic property index.
	 * throw NoSuchPropertyException
 	 */
	unsigned long getDynamicPropertyIDAtIndex(unsigned long databaseID, unsigned long propertyID, int32 index, bool resize);

	/**
	 * Gets the index of given property ID. returns 0 if property does not belong here
	 * throw NoSuchPropertyException
 	 */
	int32 getIndexOfDynamicPropertyID(unsigned long databaseID, unsigned long parentPropertyID, unsigned long propertyID);

	/**
	 * Returns the property type.
	 * throw NoSuchPropertyException
	 */
	std::string getPropertyType(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Returns the property name.
	 * throw NoSuchPropertyException
	 */
	std::string getPropertyName(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Returns the property name.
	 * throw NoSuchPropertyException
	 */
	std::pair<std::string, long> getCompletePropertyName(unsigned long databaseID, unsigned long servantID, unsigned long propertyID);

	/**
	 * Returns true if is a property.
	 */
	bool propertyExists(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Returns the servant ID from the property ID.
	 * throw NoSuchPropertyException
	 */
	unsigned long getServantIDFromPropertyID(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Returns the property ID.
	 * throw NoSuchPropertyException
	 */
	unsigned long getPropertyID(unsigned long databaseID, unsigned long servantID, const std::string& propertyName);

	/**
	 * Gets the int8 value.
	 * throw NoSuchPropertyException
	 */
	int8 getInt8Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the int16 value.
	 * throw NoSuchPropertyException
	 */
	int16 getInt16Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the uint16 value.
	 * throw NoSuchPropertyException
	 */
	uint16 getUint16Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the int32 value.
	 * throw NoSuchPropertyException
	 */
	int32 getInt32Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the uint32 value.
	 * throw NoSuchPropertyException
	 */
	uint32 getUint32Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the int64 value.
	 * throw NoSuchPropertyException
	 */
	int64 getInt64Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the uint64 value.
	 * throw NoSuchPropertyException
	 */
	uint64 getUint64Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the float32 value.
	 * throw NoSuchPropertyException
	 */
	float32 getFloat32Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the float64 value.
	 * throw NoSuchPropertyException
	 */
	float64 getFloat64Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the string value.
	 * throw NoSuchPropertyException
	 */
	std::string getStringValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the binary string value.
	 * throw NoSuchPropertyException
	 */
	std::vector<int8> getBinaryStringValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the boolean value.
	 * throw NoSuchPropertyException
	 */
	bool getBooleanValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics boolean value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsBooleanValue(unsigned long databaseID, unsigned long propertyID, bool value);

	/**
	 * Sets the ics boolean formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsBooleanValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, bool value, const std::string& formattedValue);

	/**
	 * Gets the ics boolean value.
	 * throw NoSuchPropertyException
	 */
	bool getIcsBooleanValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int16 value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsInt16Value(unsigned long databaseID, unsigned long propertyID, int16 value);

	/**
	 * Sets the ics int16 formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsInt16ValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, int16 value, const std::string& formattedValue);

	/**
	 * Gets the ics uint16 value.
	 * throw NoSuchPropertyException
	 */
	int16 getIcsInt16Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics uint16 value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsUint16Value(unsigned long databaseID, unsigned long propertyID, uint16 value);

	/**
	 * Sets the ics uint16 formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsUint16ValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, uint16 value, const std::string& formattedValue);

	/**
	 * Gets the ics uint16 value.
	 * throw NoSuchPropertyException
	 */
	uint16 getIcsUint16Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int32 value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsInt32Value(unsigned long databaseID, unsigned long propertyID, int32 value);

	/**
	 * Sets the ics int32 formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsInt32ValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, int32 value, const std::string& formattedValue);

	/**
	 * Gets the ics uint16 value.
	 * throw NoSuchPropertyException
	 */
	int32 getIcsInt32Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics uint32 value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsUint32Value(unsigned long databaseID, unsigned long propertyID, uint32 value);

	/**
	 * Sets the ics uint32 formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsUint32ValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, uint32 value, const std::string& formattedValue);

	/**
	 * Gets the ics uint32 value.
	 * throw NoSuchPropertyException
	 */
	uint32 getIcsUint32Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int64 value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsInt64Value(unsigned long databaseID, unsigned long propertyID, int64 value);

	/**
	 * Sets the ics int64 formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsInt64ValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, int64 value, const std::string& formattedValue);

	/**
	 * Gets the ics uint16 value.
	 * throw NoSuchPropertyException
	 */
	int64 getIcsInt64Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics uint64 value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsUint64Value(unsigned long databaseID, unsigned long propertyID, uint64 value);

	/**
	 * Sets the ics uint64 formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsUint64ValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, uint64 value, const std::string& formattedValue);

	/**
	 * Gets the ics uint64 value.
	 * throw NoSuchPropertyException
	 */
	uint64 getIcsUint64Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics float32 value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsFloat32Value(unsigned long databaseID, unsigned long propertyID, float32 value);

	/**
	 * Sets the ics float32 formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsFloat32ValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, float32 value, const std::string& formattedValue);

	/**
	 * Gets the ics float32 value.
	 * throw NoSuchPropertyException
	 */
	float32 getIcsFloat32Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics float64 value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsFloat64Value(unsigned long databaseID, unsigned long propertyID, float64 value);

	/**
	 * Sets the ics float64 formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsFloat64ValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, float64 value, const std::string& formattedValue);

	/**
	 * Gets the ics float64 value.
	 * throw NoSuchPropertyException
	 */
	float64 getIcsFloat64Value(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics string value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsStringValue(unsigned long databaseID, unsigned long propertyID, const std::string& value);

	/**
	 * Sets the ics string formatted value.
	 * throw NoSuchPropertyException
	 */
	bool setIcsStringValueAndFormattedValue(unsigned long databaseID, unsigned long propertyID, const std::string& value, const std::string& formattedValue);

	/**
	 * Gets the ics string value.
	 * throw NoSuchPropertyException
	 */
	std::string getIcsStringValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int16 array value.
	 * throw NoSuchPropertyException
	 */
	void setIcsInt16ArrayValue(unsigned long databaseID, unsigned long propertyID, int16* value);

	/**
	 * Gets the ics int16 array value.
	 * throw NoSuchPropertyException
	 */
	int16* getIcsInt16ArrayValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics uint16 array value.
	 * throw NoSuchPropertyException
	 */
	void setIcsUint16ArrayValue(unsigned long databaseID, unsigned long propertyID, uint16* value);

	/**
	 * Gets the ics uint16 array value.
	 * throw NoSuchPropertyException
	 */
	uint16* getIcsUint16ArrayValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int32 array value.
	 * throw NoSuchPropertyException
	 */
	void setIcsInt32ArrayValue(unsigned long databaseID, unsigned long propertyID, int32* value);

	/**
	 * Gets the ics int32 array value.
	 * throw NoSuchPropertyException
	 */
	int32* getIcsInt32ArrayValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics uint32 array value.
	 * throw NoSuchPropertyException
	 */
	void setIcsUint32ArrayValue(unsigned long databaseID, unsigned long propertyID, uint32* value);

	/**
	 * Gets the ics uint32 array value.
	 * throw NoSuchPropertyException
	 */
	uint32* getIcsUint32ArrayValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int64 array value.
	 * throw NoSuchPropertyException
	 */
	void setIcsInt64ArrayValue(unsigned long databaseID, unsigned long propertyID, int64* value);

	/**
	 * Gets the ics int64 array value.
	 * throw NoSuchPropertyException
	 */
	int64* getIcsInt64ArrayValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the ics uint64 array value.
	 * throw NoSuchPropertyException
	 */
	uint64* getIcsUint64ArrayValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics uint64 array value.
	 * throw NoSuchPropertyException
	 */
	void setIcsUint64ArrayValue(unsigned long databaseID, unsigned long propertyID, uint64* value);

	/**
	 * Sets the ics float32 array value.
	 * throw NoSuchPropertyException
	 */
	void setIcsFloat32ArrayValue(unsigned long databaseID, unsigned long propertyID, float32* value);

	/**
	 * Gets the ics float32 array value.
	 * throw NoSuchPropertyException
	 */
	float32* getIcsFloat32ArrayValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics float64 array value.
	 * throw NoSuchPropertyException
	 */
	void setIcsFloat64ArrayValue(unsigned long databaseID, unsigned long propertyID, float64* value);

	/**
	 * Gets the ics float64 array value.
	 * throw NoSuchPropertyException
	 */
	float64* getIcsFloat64ArrayValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int32 array size.
	 * throw NoSuchPropertyException
	 */
	void setIcsInt16ArraySize(unsigned long databaseID, unsigned long propertyID, int32 size);

	/**
	 * Gets the ics int16 array size.
	 * throw NoSuchPropertyException
	 */
	int32 getIcsInt16ArraySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics uint16 array size.
	 * throw NoSuchPropertyException
	 */
	void setIcsUint16ArraySize(unsigned long databaseID, unsigned long propertyID, int32 size);

	/**
	 * Gets the ics uin162 array size.
	 * throw NoSuchPropertyException
	 */
	int32 getIcsUint16ArraySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int32 array size.
	 * throw NoSuchPropertyException
	 */
	void setIcsInt32ArraySize(unsigned long databaseID, unsigned long propertyID, int32 size);

	/**
	 * Gets the ics int32 array size.
	 * throw NoSuchPropertyException
	 */
	int32 getIcsInt32ArraySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics uint32 array size.
	 * throw NoSuchPropertyException
	 */
	void setIcsUint32ArraySize(unsigned long databaseID, unsigned long propertyID, int32 size);

	/**
	 * Gets the ics uint32 array size.
	 * throw NoSuchPropertyException
	 */
	int32 getIcsUint32ArraySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int64 array size.
	 * throw NoSuchPropertyException
	 */
	void setIcsInt64ArraySize(unsigned long databaseID, unsigned long propertyID, int32 size);

	/**
	 * Gets the ics int64 array size.
	 * throw NoSuchPropertyException
	 */
	int32 getIcsInt64ArraySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics uint64 array size.
	 * throw NoSuchPropertyException
	 */
	void setIcsUint64ArraySize(unsigned long databaseID, unsigned long propertyID, int32 size);

	/**
	 * Gets the ics uint64 array size.
	 * throw NoSuchPropertyException
	 */
	int32 getIcsUint64ArraySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics float32 array size.
	 * throw NoSuchPropertyException
	 */
	void setIcsFloat32ArraySize(unsigned long databaseID, unsigned long propertyID, int32 size);

	/**
	 * Gets the ics float32 array size.
	 * throw NoSuchPropertyException
	 */
	int32 getIcsFloat32ArraySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics float64 array size.
	 * throw NoSuchPropertyException
	 */
	void setIcsFloat64ArraySize(unsigned long databaseID, unsigned long propertyID, int32 size);

	/**
	 * Gets the ics float64 array size.
	 * throw NoSuchPropertyException
	 */
	int32 getIcsFloat64ArraySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Sets the ics int16 array value.
	 * throw NoSuchPropertyException
	 */
	void getIcsInt16ArrayValue(unsigned long databaseID, unsigned long propertyID, std::vector<int16> & array);

	/**
	 * Sets the ics uint16 array value.
	 * throw NoSuchPropertyException
	 */
	void getIcsUint16ArrayValue(unsigned long databaseID, unsigned long propertyID, std::vector<uint16> & array);

	/**
	 * Sets the ics int32 array value.
	 * throw NoSuchPropertyException
	 */
	void getIcsInt32ArrayValue(unsigned long databaseID, unsigned long propertyID, std::vector<int32> & array);

	/**
	 * Sets the ics uint32 array value.
	 * throw NoSuchPropertyException
	 */
	void getIcsUint32ArrayValue(unsigned long databaseID, unsigned long propertyID, std::vector<uint32> & array);

	/**
	 * Sets the ics int64 array value.
	 * throw NoSuchPropertyException
	 */
	void getIcsInt64ArrayValue(unsigned long databaseID, unsigned long propertyID, std::vector<int64> & array);

	/**
	 * Sets the ics uint64 array value.
	 * throw NoSuchPropertyException
	 */
	void getIcsUint64ArrayValue(unsigned long databaseID, unsigned long propertyID, std::vector<uint64> & array);

	/**
	 * Sets the ics float32 array value.
	 * throw NoSuchPropertyException
	 */
	void getIcsFloat32ArrayValue(unsigned long databaseID, unsigned long propertyID, std::vector<float32> & array);

	/**
	 * Sets the ics float64 array value.
	 * throw NoSuchPropertyException
	 */
	void getIcsFloat64ArrayValue(unsigned long databaseID, unsigned long propertyID, std::vector<float64> & array);

	/**
	 * Gets the array size.
	 * throw NoSuchPropertyException
	 */
	int32 getArraySize(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Gets the encoded int32 array value.
	 * throw NoSuchPropertyException
	 */
	void getEncodedInt16ArrayValue(unsigned long databaseID, unsigned long propertyID, std::string & encodedString);

	/**
	 * Gets the encoded uint16 array value.
	 * throw NoSuchPropertyException
	 */
	void getEncodedUint16ArrayValue(unsigned long databaseID, unsigned long propertyID, std::string & encodedString);

	/**
	 * Gets the encoded int32 array value.
	 * throw NoSuchPropertyException
	 */
	void getEncodedInt32ArrayValue(unsigned long databaseID, unsigned long propertyID, std::string & encodedString);

	/**
	 * Gets the encoded uint32 array value.
	 * throw NoSuchPropertyException
	 */
	void getEncodedUint32ArrayValue(unsigned long databaseID, unsigned long propertyID, std::string & encodedString);

	/**
	 * Gets the encoded int64 array value.
	 * throw NoSuchPropertyException
	 */
	void getEncodedInt64ArrayValue(unsigned long databaseID, unsigned long propertyID, std::string & encodedString);

	/**
	 * Gets the encoded uint64 array value.
	 * throw NoSuchPropertyException
	 */
	void getEncodedUint64ArrayValue(unsigned long databaseID, unsigned long propertyID, std::string & encodedString);

	/**
	 * Gets the encoded float32 array value.
	 * throw NoSuchPropertyException
	 */
	void getEncodedFloat32ArrayValue(unsigned long databaseID, unsigned long propertyID, std::string & encodedString);

	/**
	 * Gets the encoded float64 array value.
	 * throw NoSuchPropertyException
	 */
	void getEncodedFloat64ArrayValue(unsigned long databaseID, unsigned long propertyID, std::string & encodedString);

	/**
	 * Gets the formatted value.
	 * throw NoSuchPropertyException
	 */
	std::string getFormattedValue(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Updates the ics property.
	 * throw NoSuchPropertyException
	 */
	void updateIcsProperty(unsigned long databaseID, unsigned long propertyID);

	/**
	 * Returns true if is a scannable property.
	 * throw NoSuchPropertyException
	 */
	bool isScannableProperty(unsigned long databaseID, const std::string & controllerType, const std::string & wantedPropertyName) const;

	/**
	 * Gets the property info.
	 * throw NoSuchPropertyException
	 */
	Database::PropertyInfoItem getPropertyInfoItem(unsigned long databaseID, const std::string & controllerType, const std::string & wantedPropertyName) const;

	/**
	 * Gets the property infos.
	 * throw NoSuchPropertyException
	 */
	std::vector<Database::PropertyInfoItem> getPropertyInfoItems(unsigned long databaseID, const std::string & controllerType) const;

	/**
	 * Gets the property infos.
	 * throw NoSuchPropertyException
	 */
	std::vector<Database::PropertyInfoItem> getAllPropertyInfoItems(unsigned long databaseID, const std::string & controllerType) const;

	/**
	 * Adds the property info.
	 */
	void addPropertyInfoItem(unsigned long databaseID, const std::string & controllerType, const std::string & wantedPropertyName, const std::string & actualPropertyName, const std::string & offsetPropertyName, const std::string & shortPropertyName, bool isScan, bool isSpy, bool isDynamic);

	/**
	 * Gets the number of scannable properties.
	 */
	int32 getNumberOfScannableProperties(unsigned long databaseID, const std::string & controllerType) const;

	/**
	 * Sets the constant property unit.
	 */
	void setConstantPropertyUnit(unsigned long databaseID, unsigned long propertyID, const std::string & value) const;

	/**
	 * Sets the property unit.
	 */
	void setPropertyUnit(unsigned long databaseID, unsigned long propertyID, unsigned long unitPropertyID) const;

	/**
	 * Gets the property unit.
	 */
	unsigned long getUnitPropertyID(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Sets the enumerated values property for the property with propertyID.
	 */
	void setEnumeratedValues(unsigned long databaseID, unsigned long propertyID, unsigned long valuesPropertyID) const;

	/**
	 * Sets the enumerated labels property for the property with propertyID.
	 */
	void setEnumeratedLabels(unsigned long databaseID, unsigned long propertyID, unsigned long labelsPropertyID) const;

	/**
	 * Gets the enumerated values property for the property with propertyID.
	 */
	unsigned long getEnumeratedValues(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Gets the enumerated labels property for the property with propertyID.
	 */
	unsigned long getEnumeratedLabels(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Returns true if property is saved in box.
	 */
	bool isPropertySavedInCommandBox(unsigned long databaseID, unsigned long propertyID) const;

	/**
	 * Adds the servant type.
	 */
	void addServantType(unsigned long databaseID, unsigned long servantID, const std::string& type, bool atFront) const;

	/**
	 * Returns the whole scannable properties.
	 */
	std::string getWholeScannableProperties(unsigned long databaseID) const;

	/**
	 * Sets the servant in debug.
	 */
	void setServantInDebug(unsigned long databaseID, unsigned long servantID) const;


	std::vector<unsigned long> getPropertiesSavedInBox(unsigned long databaseID) const;

private:
	std::vector<common::PropertyDescriptor> createPropertyList(const Database::PropertyMap& propertyMap);
};

}

#endif /*DATAPROVIDER_PROPERTYACCESSORDIRECTIMPL_H_*/
