/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ServantAccessorDirectImpl.h"
#include <common/base/CommonTypes.h>
#include <common/output/OutputStream.h>
#include <dataprovider/notification/NotificationServer.h>
#include <iostream>
#include <string>

using namespace std;
using namespace boost;
using namespace common;

namespace dataprovider {

ServantAccessorDirectImpl::IndexOutOfBoundsException::IndexOutOfBoundsException(long size) :
	_size(size) {
}

long ServantAccessorDirectImpl::IndexOutOfBoundsException::getSize() const {
	return _size;
}

ServantAccessorDirectImpl::ServantAccessorDirectImpl() {
}

ServantAccessorDirectImpl::~ServantAccessorDirectImpl() {
}

void ServantAccessorDirectImpl::getWholeConfiguration(unsigned long databaseID, std::string& result) {

	configuration::ServantData data;

	createDriversForWholeConfiguration(databaseID, data, Database::getInstance(databaseID)->getDrivers());
	createControllersForWholeConfiguration(databaseID, data, Database::getInstance(databaseID)->getControllers());

	data.SerializeToString(&result);
}

void ServantAccessorDirectImpl::getWholeConfigurationUpdate(unsigned long databaseID, std::string& result) {

	configuration::ServantDataUpdate data;

	createDriversForWholeConfigurationUpdate(databaseID, data, Database::getInstance(databaseID)->getDrivers());
	createControllersForWholeConfigurationUpdate(databaseID, data, Database::getInstance(databaseID)->getControllers());

	data.SerializeToString(&result);
}

void ServantAccessorDirectImpl::getControllerConfiguration(unsigned long databaseID, unsigned long servantID, std::string& result) {

	try {
		unsigned long controllerID = servantID;

		Database::ServantItem item;

		if (Database::getInstance(databaseID)->isController(servantID)) {
			item = Database::getInstance(databaseID)->getController(controllerID);
		} else if (Database::getInstance(databaseID)->isDriver(servantID)) {
			item = Database::getInstance(databaseID)->getDriver(controllerID);
		}

		configuration::ServantDescriptor servantDescriptor;
		servantDescriptor.set_name(item.name);
		servantDescriptor.set_type(item.type);
		servantDescriptor.set_clienttype(item.clientType);
		servantDescriptor.set_id(controllerID);
		servantDescriptor.set_family(item.family);
		servantDescriptor.set_subfamily(item.subfamily);
		servantDescriptor.set_command(item.isCommand);
		servantDescriptor.set_visible(item.isVisible);
		servantDescriptor.set_enabled(item.isEnabled);
		servantDescriptor.set_root(item.isRoot);
		servantDescriptor.set_channel(0);

		if (Database::getInstance(databaseID)->isController(servantID)) {
			createPropertyListForWholeConfiguration(&servantDescriptor, Database::getInstance(databaseID)->getControllerProperties(controllerID));
			createCommandListForWholeConfiguration(&servantDescriptor, Database::getInstance(databaseID)->getControllerCommands(controllerID));
			createChildServantListForWholeConfiguration(&servantDescriptor, Database::getInstance(databaseID)->getControllerChildServants(controllerID));
		} else if (Database::getInstance(databaseID)->isDriver(servantID)) {
			createPropertyListForWholeConfiguration(&servantDescriptor, Database::getInstance(databaseID)->getDriverProperties(controllerID));
			createCommandListForWholeConfiguration(&servantDescriptor, Database::getInstance(databaseID)->getDriverCommands(controllerID));
			createChildServantListForWholeConfiguration(&servantDescriptor, Database::getInstance(databaseID)->getDriverChildServants(controllerID));
		}

		servantDescriptor.SerializeToString(&result);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::getServantDynamicProperties(unsigned long databaseID, unsigned long servantID, std::string& result) {

	try {
		configuration::ServantDynamicProperties properties;

		Database::PropertyMap propertyMap;

		if (Database::getInstance(databaseID)->isController(servantID)) {
			propertyMap = Database::getInstance(databaseID)->getControllerProperties(servantID);
		} else if (Database::getInstance(databaseID)->isDriver(servantID)) {
			propertyMap = Database::getInstance(databaseID)->getDriverProperties(servantID);
		}

		for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {

			bool isDynamic = (it->second.itemType == Database::PropertyItem::DYNAMIC);

			if (isDynamic) {
				configuration::PropertyDescriptor * propertyDescriptor = properties.add_properties();
				propertyDescriptor->set_name(it->first);
				propertyDescriptor->set_type(it->second.propertyType);
				propertyDescriptor->set_dynamic(true);

				const vector<unsigned long>& IDs = it->second.IDs;

				for (vector<unsigned long>::const_iterator i = IDs.begin(); i != IDs.end(); ++i) {
					propertyDescriptor->add_idarray(*i);
				}

				if (it->second.enumeratedValuesID != 0) {
					propertyDescriptor->set_enumeratedvaluesid(it->second.enumeratedValuesID);
				}

				if (it->second.enumeratedLabelsID != 0) {
					propertyDescriptor->set_enumeratedlabelsid(it->second.enumeratedLabelsID);
				}
			}
		}

		properties.SerializeToString(&result);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::getServantSpyProperties(unsigned long databaseID, unsigned long servantID, std::string& result) {

	// the algorithm could be improved by having less information about SPY
	// if only setpoint has SPY property (and not actual)
	// OR cache the informations

	try {
		configuration::ServantSpyProperties properties;

		Database::PropertyMap propertyMap;

		if (Database::getInstance(databaseID)->isController(servantID)) {
			propertyMap = Database::getInstance(databaseID)->getControllerProperties(servantID);
		} else if (Database::getInstance(databaseID)->isDriver(servantID)) {
			propertyMap = Database::getInstance(databaseID)->getDriverProperties(servantID);
		}

		set<unsigned long> processedPropertyIDs;

		// first traversal, only iterate setpoints
		for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {
			processServantSpySetpointProperty(databaseID, servantID, it->first, it->second, processedPropertyIDs, properties);
		}

		// second traversal, remaining properties that are actual by deduction
		for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {
			processServantSpyProperty(databaseID, servantID, it->first, it->second, processedPropertyIDs, properties);
		}

		properties.SerializeToString(&result);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::getServantSpyDynamicProperties(unsigned long databaseID, unsigned long servantID, std::string& result) {

	try {
		configuration::ServantSpyProperties properties;

		Database::PropertyMap propertyMap;

		if (Database::getInstance(databaseID)->isController(servantID)) {
			propertyMap = Database::getInstance(databaseID)->getControllerProperties(servantID);
		} else if (Database::getInstance(databaseID)->isDriver(servantID)) {
			propertyMap = Database::getInstance(databaseID)->getDriverProperties(servantID);
		}

		for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {
			processServantSpyDynamicProperty(databaseID, servantID, it->first, it->second, properties);
		}

		properties.SerializeToString(&result);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

std::vector<common::ServantDescriptor> ServantAccessorDirectImpl::getControllers(unsigned long databaseID) {
	return createServantList(Database::getInstance(databaseID)->getControllers());
}

std::vector<common::ServantDescriptor> ServantAccessorDirectImpl::getDrivers(unsigned long databaseID) {
	return createServantList(Database::getInstance(databaseID)->getDrivers());
}

std::vector<common::ChildServantDescriptor> ServantAccessorDirectImpl::getChildServants(unsigned long databaseID, unsigned long servantID) {

	if (Database::getInstance(databaseID)->isController(servantID)) {
		return createChildServantList(Database::getInstance(databaseID)->getControllerChildServants(servantID));
	}

	if (Database::getInstance(databaseID)->isDriver(servantID)) {
		return createChildServantList(Database::getInstance(databaseID)->getDriverChildServants(servantID));
	}

	throw NoSuchServantException();
}

std::vector<common::ServantDescriptor> ServantAccessorDirectImpl::createServantList(const std::vector<Database::ServantItem>& servants) {

	vector<common::ServantDescriptor> result;

	// fill up the list iterator with ServantDescriptors
	for (vector<Database::ServantItem>::const_iterator it = servants.begin(); it != servants.end(); it++) {
		result.push_back(common::ServantDescriptor(it->ID, it->name, it->type, it->isRoot));
	}

	return result;
}

std::vector<common::ChildServantDescriptor> ServantAccessorDirectImpl::createChildServantList(const Database::ChildServantMap& childServantMap) {

	vector<common::ChildServantDescriptor> result;

	// fill up the list iterator with ServantDescriptors
	for (Database::ChildServantMap::const_iterator it = childServantMap.begin(); it != childServantMap.end(); it++) {

		// it->first = n_name, it->second.first = isList, it->second.second = names
		string n_name = it->first;

		Database::ChildServant const & childServant = it->second;
		bool isList = childServant.isList;
		vector<pair<long, string> > const & indexesAndNames = childServant.children;

		// indexesAndNames[i].first = index, indexesAndNames[i].second = name

		vector<long> indexArray;
		vector<string> nameArray;

		for (vector<pair<long, string> >::const_iterator c = indexesAndNames.begin(); c != indexesAndNames.end(); c++) {
			indexArray.push_back(c->first);
			nameArray.push_back(c->second);
		}

		result.push_back(common::ChildServantDescriptor(n_name, isList, indexArray, nameArray, childServant.size));
	}

	return result;
}

bool ServantAccessorDirectImpl::servantExists(unsigned long databaseID, unsigned long servantID) {

	return Database::getInstance(databaseID)->isServant(servantID);
}

std::string ServantAccessorDirectImpl::getServantName(unsigned long databaseID, unsigned long servantID) {

	try {
		return Database::getInstance(databaseID)->getServantName(servantID);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::addServant(unsigned long databaseID, unsigned long servantID, const std::string& nServantName, bool isList) {

	try {
		Database::getInstance(databaseID)->addServantToParent(servantID, nServantName, isList);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

long ServantAccessorDirectImpl::getServantListSize(unsigned long databaseID, unsigned long servantID, const std::string& nServantName) {

	try {
		return Database::getInstance(databaseID)->getServantListSize(servantID, nServantName);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::resizeServantList(unsigned long databaseID, unsigned long servantID, const std::string& nServantName, long size) {

	try {
		Database::getInstance(databaseID)->resizeServantList(servantID, nServantName, size);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

ServantAccessorDirectImpl::ServantType ServantAccessorDirectImpl::getServantType(unsigned long databaseID, unsigned long servantID) {

	if (Database::getInstance(databaseID)->isController(servantID)) {
		return ServantAccessorDirectImpl::CONTROLLER;
	}

	if (Database::getInstance(databaseID)->isDriver(servantID)) {
		return ServantAccessorDirectImpl::DRIVER;
	}

	throw NoSuchServantException();
}

std::string ServantAccessorDirectImpl::getControllerType(unsigned long databaseID, unsigned long controllerID) {
	try {
		return Database::getInstance(databaseID)->getControllerType(controllerID);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

std::string ServantAccessorDirectImpl::getDriverType(unsigned long databaseID, unsigned long driverID) {
	try {
		return Database::getInstance(databaseID)->getDriverType(driverID);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::setMode(unsigned long databaseID, unsigned long driverID, common::DriverMode mode) {

	try {
		Database::getInstance(databaseID)->setMode(driverID, mode);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

common::DriverMode ServantAccessorDirectImpl::getMode(unsigned long databaseID, unsigned long driverID) {

	try {
		return Database::getInstance(databaseID)->getMode(driverID);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

unsigned long ServantAccessorDirectImpl::getDriverParentID(unsigned long databaseID, unsigned long driverID) {

	try {
		return Database::getInstance(databaseID)->getDriverParentID(driverID);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

unsigned long ServantAccessorDirectImpl::getProtocolIDForDriver(unsigned long databaseID, unsigned long driverID) {

	try {
		return 0;

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

unsigned long ServantAccessorDirectImpl::getEnabledOrUniqueServantID(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex) {

	try {
		return Database::getInstance(databaseID)->getEnabledOrUniqueServantID(parentID, nServantName, childIndex);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	} catch (Database::ServantDisabledException& e) {
		throw ServantDisabledException();
	} catch (Database::IndexOutOfBoundsException& e) {
		throw IndexOutOfBoundsException(e.getSize());
	}
}

std::vector<unsigned long> ServantAccessorDirectImpl::getServantIDs(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex) {

	try {
		return Database::getInstance(databaseID)->getServantIDs(parentID, nServantName, childIndex);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	} catch (Database::IndexOutOfBoundsException& e) {
		throw IndexOutOfBoundsException(e.getSize());
	}
}

unsigned long ServantAccessorDirectImpl::createController(unsigned long databaseID, const std::string& controllerName, const std::string& controllerType, const std::string& family, bool isCommand, common::HashMap<std::string,  std::pair<std::string , std::string>> roleMap) {

	try {
		return Database::getInstance(databaseID)->createController(controllerName,controllerType, family, isCommand, roleMap);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

unsigned long ServantAccessorDirectImpl::createDriver(unsigned long databaseID, const std::string& driverName, const std::string& driverType, const std::string& driverMode, const std::string& protocolType, unsigned long parentID, const std::string& nDriverName, long childIndex) {

	common::DriverMode mode;
	if (driverMode == "real") {
		mode = common::REAL_MODE;
	} else if (driverMode == "perfect") {
		mode = common::PERFECT_MODE;
	} else if (driverMode == "simulated") {
		mode = common::SIMULATED_MODE;
	} else {
		err << "Error : mode " << driverMode << " not managed" << eol;
	}

	try {
		return Database::getInstance(databaseID)->createDriver(driverName, driverType, protocolType, mode, parentID, nDriverName, childIndex);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

unsigned long ServantAccessorDirectImpl::cloneController(unsigned long databaseID, unsigned long controllerID) {

	try {
		return Database::getInstance(databaseID)->cloneController(controllerID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::assignServantFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex, unsigned long newParentID, const std::string& newNServantName, long newChildIndex) {

	try {
		Database::getInstance(databaseID)->assignServant(parentID, nServantName, childIndex, newParentID, newNServantName, newChildIndex);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	} catch (Database::ServantAssignmentException& e) {
		throw ServantAssignmentException();
	}
}

void ServantAccessorDirectImpl::assignServant(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex, unsigned long newServantID) {

	try {
		Database::getInstance(databaseID)->assignServant(parentID, nServantName, childIndex, newServantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	} catch (Database::ServantAssignmentException& e) {
		throw ServantAssignmentException();
	}
}

void ServantAccessorDirectImpl::assignServantWithID(unsigned long databaseID, unsigned long servantID, bool newServant, bool localServant) {

	try {
		Database::getInstance(databaseID)->assignServant(servantID, newServant, localServant);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

unsigned long ServantAccessorDirectImpl::removeServantFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nName, long index) {

	try {
		return Database::getInstance(databaseID)->removeServantFromParent(parentID, nName, index);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::removeServant(unsigned long databaseID, unsigned long servantID) {

	try {
		Database::getInstance(databaseID)->removeServant(servantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::setServantInitialised(unsigned long databaseID, unsigned long servantID, bool initialised) {

	try {
		Database::getInstance(databaseID)->setServantInitialised(servantID, initialised);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

bool ServantAccessorDirectImpl::isServantInitialised(unsigned long databaseID, unsigned long servantID) {

	try {
		return Database::getInstance(databaseID)->isServantInitialised(servantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

bool ServantAccessorDirectImpl::isServantInitialisedFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex) {

	try {
		return Database::getInstance(databaseID)->isServantInitialised(parentID, nServantName, childIndex);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

unsigned long ServantAccessorDirectImpl::clone() {
	return Database::getInstance()->clone();
}

void ServantAccessorDirectImpl::resetElapsedSimulationTime(unsigned long databaseID, unsigned long driverID) {

	try {
		Database::getInstance(databaseID)->resetElapsedSimulationTime(driverID);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

double ServantAccessorDirectImpl::getElapsedSimulationTimeMs(unsigned long databaseID, unsigned long driverID) {

	try {
		return Database::getInstance(databaseID)->getElapsedSimulationTimeMs(driverID);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::enableServant(unsigned long databaseID, unsigned long servantID, bool all) {

	try {
		if (all) {
			Database::getInstance(databaseID)->enableServant(servantID);
		} else {
			Database::getInstance(databaseID)->enableOnlyServant(servantID);
		}

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	} catch (Database::ServantEnableException& e) {
		throw ServantEnableException();
	}
}

void ServantAccessorDirectImpl::enableServantFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex) {

	try {
		Database::getInstance(databaseID)->enableServant(parentID, nServantName, childIndex);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	} catch (Database::ServantEnableException& e) {
		throw ServantEnableException();
	}
}

void ServantAccessorDirectImpl::disableServant(unsigned long databaseID, unsigned long servantID, bool all) {

	try {
		if (all) {
			Database::getInstance(databaseID)->disableServant(servantID);
		} else {
			Database::getInstance(databaseID)->disableOnlyServant(servantID);
		}

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::disableServantFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex) {

	try {
		Database::getInstance(databaseID)->disableServant(parentID, nServantName, childIndex);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

bool ServantAccessorDirectImpl::isServantEnabled(unsigned long databaseID, unsigned long servantID) {

	try {
		return Database::getInstance(databaseID)->isServantEnabled(servantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

bool ServantAccessorDirectImpl::isServantEnabledFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex) {

	try {
		return Database::getInstance(databaseID)->isServantEnabled(parentID, nServantName, childIndex);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

std::vector<unsigned long> ServantAccessorDirectImpl::configureServant(unsigned long databaseID, unsigned long servantID, const std::string& configurationName) {

	// To remove
	return vector<unsigned long>();
}

std::string ServantAccessorDirectImpl::what(unsigned long databaseID, unsigned long elementID) {

	return Database::getInstance(databaseID)->what(elementID, false);
}

void ServantAccessorDirectImpl::setClientType(unsigned long databaseID, unsigned long servantID, const std::string& clientType) {
	Database::getInstance(databaseID)->setClientType(servantID, clientType);
}

bool ServantAccessorDirectImpl::isServantPausable(unsigned long databaseID, unsigned long servantID) {
	try {
		return Database::getInstance(databaseID)->isServantPausable(servantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

long ServantAccessorDirectImpl::getDriverChannel(unsigned long databaseID, unsigned long servantID) {
	try {
		return Database::getInstance(databaseID)->getDriverChannel(servantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

unsigned long ServantAccessorDirectImpl::getServantID(unsigned long databaseID, const std::string& name) {
	try {
		return Database::getInstance(databaseID)->getServantID(name);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::setServantVisible(unsigned long databaseID, unsigned long servantID, bool value) {
	try {
		Database::getInstance(databaseID)->setServantVisible(servantID, value);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::setServantFavourite(unsigned long databaseID, unsigned long servantID, bool value) {
	try {
		Database::getInstance(databaseID)->setServantFavourite(servantID, value);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::setServantCommand(unsigned long databaseID, unsigned long servantID, bool value) {
	try {
		Database::getInstance(databaseID)->setServantCommand(servantID, value);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::enableConfigurationChanged(unsigned long databaseID) {
	NotificationServer::getInstance()->enableConfigurationChanged(databaseID);
}

void ServantAccessorDirectImpl::disableConfigurationChanged(unsigned long databaseID) {
	NotificationServer::getInstance()->disableConfigurationChanged(databaseID);
}

void ServantAccessorDirectImpl::configurationChanged(unsigned long databaseID) {
	NotificationServer::getInstance()->configurationChanged(databaseID, 0, common::TREE);
}

void ServantAccessorDirectImpl::setServantFamily(unsigned long databaseID, unsigned long servantID, const std::string& familyName, const std::string& subfamilyName) {
	try {
		Database::getInstance(databaseID)->setServantFamily(servantID, familyName, subfamilyName);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

std::string ServantAccessorDirectImpl::getServantFamily(unsigned long databaseID, unsigned long servantID) {
	try {
		return Database::getInstance(databaseID)->getServantFamily(servantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

std::string ServantAccessorDirectImpl::getServantSubfamily(unsigned long databaseID, unsigned long servantID) {
	try {
		return Database::getInstance(databaseID)->getServantSubfamily(servantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::getWebSpyProperties(unsigned long databaseID, configuration::WebSpyContainer& aContainer) {

	configuration::ServantData data;

	createControllersForWholeConfiguration(databaseID, data, Database::getInstance(databaseID)->getControllers());

	for(auto controller : data.controllers()) {
		if (controller.enabled()) {
			configuration::WebSpyServantContainer* webspyservant = aContainer.add_servants();
			webspyservant->set_type(controller.type());
			webspyservant->set_name(controller.name());
			webspyservant->set_visible(controller.visible());

			Database::PropertyMap propertyMap = Database::getInstance(databaseID)->getControllerProperties(controller.id());
			configuration::ServantSpyProperties properties;
			set<unsigned long> processedPropertyIDs;

			// first traversal, only iterate setpoints
			for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {
				processServantSpySetpointProperty(databaseID, controller.id(), it->first, it->second, processedPropertyIDs, properties);
			}

			// second traversal, remaining properties that are actual by deduction
			for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {
				processServantSpyProperty(databaseID, controller.id(), it->first, it->second, processedPropertyIDs, properties);
			}

			// Dynamic one
			for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {
				processServantSpyDynamicProperties(databaseID, controller.id(), it->first, it->second, properties);
			}

			for(auto property : properties.properties()) {
				configuration::WebSpyProperty* webspyproperties = webspyservant->add_properties();
				webspyproperties->set_id(property.actualid());
				string indice;
				if (property.has_dynamicindice()) {
					indice = "." + lexical_cast<string>(property.dynamicindice());
				}
				if (property.actualid() != 0) {
					webspyproperties->set_name(Database::getInstance(databaseID)->getPropertyName(property.actualid()) + indice);
					webspyproperties->set_value(Database::getInstance(databaseID)->getFormattedValue(property.actualid()));
				}
				if (property.setpointid() != 0) {
					webspyproperties->set_setpoint_name(Database::getInstance(databaseID)->getPropertyName(property.setpointid()) + indice);
					webspyproperties->set_setpoint_value(Database::getInstance(databaseID)->getFormattedValue(property.setpointid()));
				}
				if (property.unitid() != 0) {
					webspyproperties->set_unit(Database::getInstance(databaseID)->getFormattedValue(property.unitid()));
				}
				if (property.has_alias() != 0) {
					webspyproperties->set_alias(property.alias());
				}
			}

		}
	}
}

bool ServantAccessorDirectImpl::isServantFavourite(unsigned long databaseID, unsigned long servantID) {
	try {
		return Database::getInstance(databaseID)->isServantFavourite(servantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

bool ServantAccessorDirectImpl::isServantEnabledAndInitialised(unsigned long databaseID, unsigned long servantID) {

	try {
		return isServantEnabled(databaseID, servantID)
			&& isServantInitialised(databaseID, servantID);

	} catch (Database::NoSuchServantException& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::createControllersForWholeConfiguration(unsigned long databaseID, configuration::ServantData& data, const std::vector<Database::ServantItem>& servants) {

	// fill up the list with ServantDescriptors
	for (vector<Database::ServantItem>::const_iterator it = servants.begin(); it != servants.end(); it++) {

		unsigned long controllerID = it->ID;

		// it->first = id, it->second.first = name,  it->second.second = type
		configuration::ServantDescriptor * servantDescriptor = data.add_controllers();
		servantDescriptor->set_name(it->name);
		servantDescriptor->set_type(it->type);
		servantDescriptor->set_clienttype(it->clientType);
		servantDescriptor->set_id(controllerID);
		servantDescriptor->set_family(it->family);
		servantDescriptor->set_subfamily(it->subfamily);
		servantDescriptor->set_command(it->isCommand);
		servantDescriptor->set_visible(it->isVisible);
		servantDescriptor->set_enabled(it->isEnabled);
		servantDescriptor->set_root(it->isRoot);
		servantDescriptor->set_channel(0);

		createPropertyListForWholeConfiguration(servantDescriptor, Database::getInstance(databaseID)->getControllerProperties(controllerID));
		createCommandListForWholeConfiguration(servantDescriptor, Database::getInstance(databaseID)->getControllerCommands(controllerID));
		createChildServantListForWholeConfiguration(servantDescriptor, Database::getInstance(databaseID)->getControllerChildServants(controllerID));
	}
}

void ServantAccessorDirectImpl::createDriversForWholeConfiguration(unsigned long databaseID, configuration::ServantData& data, const std::vector<Database::ServantItem>& servants) {

	// fill up the list with ServantDescriptors
	for (vector<Database::ServantItem>::const_iterator it = servants.begin(); it != servants.end(); it++) {

		unsigned long driverID = it->ID;

		// it->first = id, it->second.first = name,  it->second.second = type
		configuration::ServantDescriptor * servantDescriptor = data.add_drivers();
		servantDescriptor->set_name(it->name);
		servantDescriptor->set_type(it->type);
		servantDescriptor->set_clienttype(it->clientType);
		servantDescriptor->set_id(driverID);
		servantDescriptor->set_family(it->family);
		servantDescriptor->set_subfamily(it->subfamily);
		servantDescriptor->set_command(it->isCommand);
		servantDescriptor->set_visible(it->isVisible);
		servantDescriptor->set_enabled(it->isEnabled);
		servantDescriptor->set_root(it->isRoot);
		// get the driver channel
		servantDescriptor->set_channel(Database::getInstance(databaseID)->getDriverChannel(driverID));

		common::DriverMode mode = Database::getInstance(databaseID)->getMode(driverID);

		if (mode == common::REAL_MODE) {
			servantDescriptor->set_mode(configuration::ServantDescriptor_Mode_REAL);

		} else if (mode == common::SIMULATED_MODE) {
			servantDescriptor->set_mode(configuration::ServantDescriptor_Mode_SIMULATED);

		} else if (mode == common::PERFECT_MODE) {
			servantDescriptor->set_mode(configuration::ServantDescriptor_Mode_PERFECT);

		} else {
			servantDescriptor->set_mode(configuration::ServantDescriptor_Mode_REAL);
		}

		createPropertyListForWholeConfiguration(servantDescriptor, Database::getInstance(databaseID)->getDriverProperties(driverID));
		createCommandListForWholeConfiguration(servantDescriptor, Database::getInstance(databaseID)->getDriverCommands(driverID));
		createChildServantListForWholeConfiguration(servantDescriptor, Database::getInstance(databaseID)->getDriverChildServants(driverID));
	}
}

void ServantAccessorDirectImpl::createPropertyListForWholeConfiguration(configuration::ServantDescriptor * servantDescriptor, const Database::PropertyMap& propertyMap) {

	for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {

		configuration::PropertyDescriptor * propertyDescriptor = servantDescriptor->add_properties();

		propertyDescriptor->set_name(it->first);
		propertyDescriptor->set_type(it->second.propertyType);
		propertyDescriptor->set_dynamic(it->second.itemType == Database::PropertyItem::DYNAMIC);

		const vector<unsigned long>& IDs = it->second.IDs;

		for (vector<unsigned long>::const_iterator i = IDs.begin(); i != IDs.end(); ++i) {
			propertyDescriptor->add_idarray(*i);
		}

		if (it->second.enumeratedValuesID != 0) {
			propertyDescriptor->set_enumeratedvaluesid(it->second.enumeratedValuesID);
		}

		if (it->second.enumeratedLabelsID != 0) {
			propertyDescriptor->set_enumeratedlabelsid(it->second.enumeratedLabelsID);
		}
	}
}

void ServantAccessorDirectImpl::createCommandListForWholeConfiguration(configuration::ServantDescriptor * servantDescriptor, const Database::CommandMap& commandMap) {

	for (Database::CommandMap::const_iterator it = commandMap.begin(); it != commandMap.end(); it++) {

		configuration::CommandDescriptor * commandDescriptor = servantDescriptor->add_commands();

		commandDescriptor->set_name(it->first);
		commandDescriptor->set_id(it->second);
	}
}

void ServantAccessorDirectImpl::createChildServantListForWholeConfiguration(configuration::ServantDescriptor * servantDescriptor, const Database::ChildServantMap& childServantMap) {

	for (Database::ChildServantMap::const_iterator it = childServantMap.begin(); it != childServantMap.end(); it++) {

		Database::ChildServant const & childServant = it->second;

		configuration::ChildDescriptor * childDescriptor = servantDescriptor->add_children();

		childDescriptor->set_name(it->first);
		childDescriptor->set_list(childServant.isList);

		vector<pair<long, string> > const & indexesAndNames = childServant.children;

		for (vector<pair<long, string> >::const_iterator c = indexesAndNames.begin(); c != indexesAndNames.end(); c++) {
			configuration::ChildDescriptor_Child * child = childDescriptor->add_childarray();
			child->set_index(c->first);
			child->set_name(c->second);
		}
	}

}


void ServantAccessorDirectImpl::createControllersForWholeConfigurationUpdate(unsigned long databaseID, configuration::ServantDataUpdate& data, const std::vector<Database::ServantItem>& servants) {

	// fill up the list with ServantInfos
	for (vector<Database::ServantItem>::const_iterator it = servants.begin(); it != servants.end(); it++) {

		unsigned long controllerID = it->ID;

		// it->first = id, it->second.first = name,  it->second.second = type
		configuration::ServantInfo * servantInfo = data.add_controllers();
		servantInfo->set_id(controllerID);
		servantInfo->set_family(it->family);
		servantInfo->set_subfamily(it->subfamily);
		servantInfo->set_command(it->isCommand);
		servantInfo->set_visible(it->isVisible);
		servantInfo->set_enabled(it->isEnabled);
	}
}

void ServantAccessorDirectImpl::createDriversForWholeConfigurationUpdate(unsigned long databaseID, configuration::ServantDataUpdate& data, const std::vector<Database::ServantItem>& servants) {

	// fill up the list with ServantInfos
	for (vector<Database::ServantItem>::const_iterator it = servants.begin(); it != servants.end(); it++) {

		unsigned long driverID = it->ID;

		// it->first = id, it->second.first = name,  it->second.second = type
		configuration::ServantInfo * servantInfo = data.add_drivers();
		servantInfo->set_id(driverID);
		servantInfo->set_family(it->family);
		servantInfo->set_subfamily(it->subfamily);
		servantInfo->set_visible(it->isVisible);
		servantInfo->set_enabled(it->isEnabled);
	}
}


void ServantAccessorDirectImpl::processServantSpySetpointProperty(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const Database::PropertyItem& item, std::set<unsigned long>& processedPropertyIDs, configuration::ServantSpyProperties& properties) {

	bool isSpy = item.isSpy;
	bool isDynamic = (item.itemType == Database::PropertyItem::DYNAMIC);

	if (!isSpy || isDynamic || item.IDs.empty()) {
		return;
	}

	unsigned long propertyID = item.IDs.front();

	// only process if not processed
	if (processedPropertyIDs.find(propertyID) != processedPropertyIDs.end()) {
		return;
	}

	bool setpoint = false;
	Database::PropertyInfoItem infoItem;

	string controllerType = Database::getInstance(databaseID)->getControllerType(servantID);

	try {
		infoItem = Database::getInstance(databaseID)->getSetpointPropertyInfoItem(controllerType, propertyName);
		setpoint = true;

	} catch (const Database::NoPropertyInfo& e) {
		// do nothing, trying next type
	}

	if (!setpoint) {
		return;
	}

	configuration::SpyPropertyDescriptor * spyPropertyDescriptor = properties.add_properties();

	unsigned long actualPropertyID = Database::getInstance(databaseID)->getPropertyID(servantID, infoItem.actualPropertyName);

	spyPropertyDescriptor->set_actualid(actualPropertyID);
	processedPropertyIDs.insert(actualPropertyID);

	spyPropertyDescriptor->set_setpointid(propertyID);
	processedPropertyIDs.insert(propertyID);

	if (infoItem.offsetPropertyName != "") {
		unsigned long offsetPropertyID = Database::getInstance(databaseID)->getPropertyID(servantID, infoItem.offsetPropertyName);
		spyPropertyDescriptor->set_offsetid(offsetPropertyID);
		processedPropertyIDs.insert(offsetPropertyID);
	} else {
		// setting 0 to offset id
		spyPropertyDescriptor->set_offsetid(0);
	}

	unsigned long unitID = Database::getInstance(databaseID)->getUnitPropertyID(propertyID);
	spyPropertyDescriptor->set_unitid(unitID);
	processedPropertyIDs.insert(unitID);

	spyPropertyDescriptor->set_alias(infoItem.aliasPropertyName);
}

void ServantAccessorDirectImpl::processServantSpyProperty(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const Database::PropertyItem& item, std::set<unsigned long>& processedPropertyIDs, configuration::ServantSpyProperties& properties) {

	bool isSpy = item.isSpy;
	bool isDynamic = (item.itemType == Database::PropertyItem::DYNAMIC);

	if (!isSpy || isDynamic || item.IDs.empty()) {
		return;
	}

	unsigned long propertyID = item.IDs.front();

	// only process if not processed
	if (processedPropertyIDs.find(propertyID) != processedPropertyIDs.end()) {
		return;
	}

	Database::PropertyInfoItem infoItem;

	string controllerType = Database::getInstance(databaseID)->getControllerType(servantID);

	try {
		infoItem = Database::getInstance(databaseID)->getNoSetpointPropertyInfoItem(controllerType, propertyName);

	} catch (const Database::NoPropertyInfo& e) {
		// do nothing
	}

	configuration::SpyPropertyDescriptor * spyPropertyDescriptor = properties.add_properties();

	spyPropertyDescriptor->set_actualid(propertyID);
	processedPropertyIDs.insert(propertyID);

	// setting 0 to setpoint and offset ids
	spyPropertyDescriptor->set_setpointid(0);
	spyPropertyDescriptor->set_offsetid(0);

	unsigned long unitID = Database::getInstance(databaseID)->getUnitPropertyID(propertyID);
	spyPropertyDescriptor->set_unitid(unitID);
	processedPropertyIDs.insert(unitID);

	spyPropertyDescriptor->set_alias(infoItem.aliasPropertyName);
}

void ServantAccessorDirectImpl::processServantSpyDynamicProperty(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const Database::PropertyItem& item, configuration::ServantSpyProperties& properties) {

	bool isSpy = item.isSpy;
	bool isDynamic = (item.itemType == Database::PropertyItem::DYNAMIC);

	if (!isSpy || !isDynamic || item.IDs.empty()) {
		return;
	}

	unsigned long propertyID = item.IDs.front();

	configuration::SpyPropertyDescriptor * spyPropertyDescriptor = properties.add_properties();

	spyPropertyDescriptor->set_actualid(propertyID);

	// setting 0 to setpoint and offset ids
	spyPropertyDescriptor->set_setpointid(0);
	spyPropertyDescriptor->set_offsetid(0);

	unsigned long unitID = Database::getInstance(databaseID)->getUnitPropertyID(propertyID);
	spyPropertyDescriptor->set_unitid(unitID);

	// we do not copy the alias because we do not have the information here
	spyPropertyDescriptor->set_alias("");
}

void ServantAccessorDirectImpl::processServantSpyDynamicProperties(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const Database::PropertyItem& item, configuration::ServantSpyProperties& properties) {

	bool isSpy = item.isSpy;
	bool isDynamic = (item.itemType == Database::PropertyItem::DYNAMIC);

	if (!isSpy || !isDynamic || item.IDs.empty()) {
		return;
	}

	std::vector<unsigned long>::const_iterator iter = item.IDs.begin();
	unsigned long mainid = *iter;
	unsigned long indice = 0;
	while(++iter != item.IDs.end()) {
		unsigned long propertyID = *iter;

		configuration::SpyPropertyDescriptor * spyPropertyDescriptor = properties.add_properties();

		spyPropertyDescriptor->set_actualid(*iter);

		// setting 0 to setpoint and offset ids
		spyPropertyDescriptor->set_setpointid(0);
		spyPropertyDescriptor->set_offsetid(0);

		unsigned long unitID = Database::getInstance(databaseID)->getUnitPropertyID(mainid);
		spyPropertyDescriptor->set_unitid(unitID);

		// we do not copy the alias because we do not have the information here
		spyPropertyDescriptor->set_alias("");

		spyPropertyDescriptor->set_dynamicindice(indice);
		++indice;
	}
}

void ServantAccessorDirectImpl::getServantPropertiesValue(unsigned long databaseID, unsigned long servantID, std::string& result) {

	// the algorithm could be improved by having less information about SPY
	// if only setpoint has SPY property (and not actual)
	// OR cache the informations

	try {
		configuration::ServantPropertiesValue properties;

		Database::PropertyMap propertyMap;

		if (Database::getInstance(databaseID)->isController(servantID)) {
			propertyMap = Database::getInstance(databaseID)->getControllerProperties(servantID);
		} else if (Database::getInstance(databaseID)->isDriver(servantID)) {
			propertyMap = Database::getInstance(databaseID)->getDriverProperties(servantID);
		}
		string propertyType;

		// first traversal, only iterate setpoints
		for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {
			Database::PropertyItem item = it->second;
			propertyType = Database::getInstance(databaseID)->getPropertyType(item.IDs[0]);
			string name = it->first;
			if (item.itemType == Database::PropertyItem::ELEMENTARY) {
				configuration::PropertyValue * propertyValue = properties.add_properties();
				propertyValue->set_name(name);
				if (propertyType == TypeToString<uint16>::convert()) {
					propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<uint16>(item.IDs[0])));
				} else if (propertyType == TypeToString<uint32>::convert()) {
					propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<uint32>(item.IDs[0])));
				} else if (propertyType == TypeToString<uint64>::convert()) {
					propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<uint64>(item.IDs[0])));
				} else if (propertyType == TypeToString<int16>::convert()) {
					propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<int16>(item.IDs[0])));
				} else if (propertyType == TypeToString<int32>::convert()) {
					propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<int32>(item.IDs[0])));
				} else if (propertyType == TypeToString<int64>::convert()) {
					propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<int64>(item.IDs[0])));
				} else if (propertyType == TypeToString<float32>::convert()) {
					propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<float32>(item.IDs[0])));
				} else if (propertyType == TypeToString<float64>::convert()) {
					propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<float64>(item.IDs[0])));
				} else if (propertyType == TypeToString<bool>::convert()) {
					if (Database::getInstance(databaseID)->getValue<bool>(item.IDs[0]) == true) {
						propertyValue->set_value("true");
					}
					else {
						propertyValue->set_value("false");
					}
				} else if (propertyType == TypeToString<string>::convert()) {
					propertyValue->set_value(Database::getInstance(databaseID)->getValue<string>(item.IDs[0]));
				} else {
					propertyValue->set_value("");
				}
			}
			else {
				for (uint32 i = 1; i < item.IDs.size(); ++i) {
					configuration::PropertyValue * propertyValue = properties.add_properties();
					ostringstream dynname;
					dynname << name << "." << i - 1;
					propertyValue->set_name(dynname.str());
					if (propertyType == TypeToString<uint16>::convert()) {
						propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<uint16>(item.IDs[i])));
					} else if (propertyType == TypeToString<uint32>::convert()) {
						propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<uint32>(item.IDs[i])));
					} else if (propertyType == TypeToString<uint64>::convert()) {
						propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<uint64>(item.IDs[i])));
					} else if (propertyType == TypeToString<int16>::convert()) {
						propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<int16>(item.IDs[i])));
					} else if (propertyType == TypeToString<int32>::convert()) {
						propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<int32>(item.IDs[i])));
					} else if (propertyType == TypeToString<int64>::convert()) {
						propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<int64>(item.IDs[i])));
					} else if (propertyType == TypeToString<float32>::convert()) {
						propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<float32>(item.IDs[i])));
					} else if (propertyType == TypeToString<float64>::convert()) {
						propertyValue->set_value(to_string(Database::getInstance(databaseID)->getValue<float64>(item.IDs[i])));
					} else if (propertyType == TypeToString<bool>::convert()) {
						if (Database::getInstance(databaseID)->getValue<bool>(item.IDs[i]) == true) {
							propertyValue->set_value("true");
						}
						else {
							propertyValue->set_value("false");
						}
					} else if (propertyType == TypeToString<string>::convert()) {
						propertyValue->set_value(Database::getInstance(databaseID)->getValue<string>(item.IDs[i]));
					} else {
						propertyValue->set_value("");
					}
				}
			}
		}

		properties.SerializeToString(&result);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

void ServantAccessorDirectImpl::getServantCommandsState(unsigned long databaseID, unsigned long servantID, std::string& result) {

	// the algorithm could be improved by having less information about SPY
	// if only setpoint has SPY property (and not actual)
	// OR cache the informations

	try {
		configuration::ServantCommandsState commands;

		Database::CommandMap commandMap;

		if (Database::getInstance(databaseID)->isController(servantID)) {
			commandMap = Database::getInstance(databaseID)->getControllerCommands(servantID);
		} else if (Database::getInstance(databaseID)->isDriver(servantID)) {
			commandMap = Database::getInstance(databaseID)->getDriverCommands(servantID);
		}

		// first traversal, only iterate setpoints
		for (Database::CommandMap::const_iterator it = commandMap.begin(); it != commandMap.end(); it++) {
			uint32 id = it->second;
			string name = it->first;
			configuration::CommandStateValue * commandState = commands.add_commands();
			commandState->set_name(name);
			commandState->set_value(Database::getInstance(databaseID)->getCommandState(id));
		}

		commands.SerializeToString(&result);

	} catch (Exception& e) {
		throw NoSuchServantException();
	}
}

bool ServantAccessorDirectImpl::isServantInDebug(unsigned long databaseID, unsigned long servantID) const {
	return Database::getInstance(databaseID)->isServantInDebug(servantID);
}

vector<unsigned long> ServantAccessorDirectImpl::getPropertiesSavedInBox(unsigned long databaseID) {
	return Database::getInstance(databaseID)->getPropertiesSavedInBox();
}




}
