/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DATAPROVIDER_SERVANTACCESSORDIRECTIMPL_H_
#define DATAPROVIDER_SERVANTACCESSORDIRECTIMPL_H_

#include <boost/thread/mutex.hpp>
#include <common/base/Exception.h>
#include <dataprovider/database/Database.h>
#include <vector>
#include <common/utils/ChildServantDescriptor.h>
#include <common/utils/ServantDescriptor.h>
#include <ServantConfiguration.pb.h>

namespace dataprovider {

/**
 * Direct server implementation of the class defining a servant accessor to transfer servant related data.
 */
class ServantAccessorDirectImpl {

public:
	enum ServantType {CONTROLLER, DRIVER};

	class NoSuchServantException {};
	class NoSuchChildException {};
	class NoSuchServantTypeException {};
	class NoSuchServantChildNameException {};

	class IndexOutOfBoundsException {

	public:
		IndexOutOfBoundsException(long size);
		long getSize() const;

	private:
		long _size;
	};

	class ServantDisabledException {};
	class ServantEnableException {};
	class ServantAssignmentException {};
	class CannotConfigureServantException {};

	/**
	 * Constructor.
	 */
	ServantAccessorDirectImpl();

	/**
	 * Virtual destructor.
	 */
	virtual ~ServantAccessorDirectImpl();

	/**
	 * Gets the whole configuration including controllers, drivers, properties, commands.
	 */
	virtual void getWholeConfiguration(unsigned long databaseID, std::string& result);

	/**
	 * Gets the whole configuration update including servant enabled/disabled.
	 */
	virtual void getWholeConfigurationUpdate(unsigned long databaseID, std::string& result);

	/**
	 * Gets the servant configuration.
	 * throw nsdp::ServantAccessor::NoSuchServantException
	 */
	virtual void getControllerConfiguration(unsigned long databaseID, unsigned long servantID, std::string& result);

	/**
	 * Gets the servant dynamic properties.
	 * throw nsdp::ServantAccessor::NoSuchServantException
	 */
	virtual void getServantDynamicProperties(unsigned long databaseID, unsigned long servantID, std::string& result);

	/**
	 * Gets the servant spy properties.
	 * throw nsdp::ServantAccessor::NoSuchServantException
	 */
	virtual void getServantSpyProperties(unsigned long databaseID, unsigned long servantID, std::string& result);

	/**
	 * Gets the servant spy dynamic properties.
	 * throw nsdp::ServantAccessor::NoSuchServantException
	 */
	virtual void getServantSpyDynamicProperties(unsigned long databaseID, unsigned long servantID, std::string& result);



	/**
	 * Gets the controllers iterator by calling the database to fill the iterator.
	 */
	virtual std::vector<common::ServantDescriptor> getControllers(unsigned long databaseID);

	/**
	 * Gets the drivers iterator by calling the database to fill the iterator.
	 */
	virtual std::vector<common::ServantDescriptor> getDrivers(unsigned long databaseID);

	/**
	 * Gets the child servants iterator by calling the database to fill the iterator.
	 * throw ServantAccessorDirectImpl::NoSuchServantException, ServantAccessorDirectImpl::IteratorFailureException
	 */
	virtual std::vector<common::ChildServantDescriptor> getChildServants(unsigned long databaseID, unsigned long servantID);

	/**
	 * Gets the servant name.
	 */
	virtual bool servantExists(unsigned long databaseID, unsigned long servantID);

	/**
	 * Gets the servant name.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual std::string getServantName(unsigned long databaseID, unsigned long servantID);

	/**
	 * Adds the servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual void addServant(unsigned long databaseID, unsigned long servantID, const std::string& nServantName, bool isList);

	/**
	 * Gets the servant list size.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual long getServantListSize(unsigned long databaseID, unsigned long servantID, const std::string& nServantName);

	/**
	 * Resizes the servant list.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual void resizeServantList(unsigned long databaseID, unsigned long servantID, const std::string& nServantName, long size);

	/**
	 * Gets the servant type.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual ServantAccessorDirectImpl::ServantType getServantType(unsigned long databaseID, unsigned long servantID);

	/**
	 * Gets the controller type.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual std::string getControllerType(unsigned long databaseID, unsigned long controllerID);

	/**
	 * Gets the driver type.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual std::string getDriverType(unsigned long databaseID, unsigned long driverID);

	/**
	 * Sets the mode of execution for a driver.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual void setMode(unsigned long databaseID, unsigned long driverID, common::DriverMode mode);

	/**
	 * Returns the mode of execution for a driver.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual common::DriverMode getMode(unsigned long databaseID, unsigned long driverID);

	/**
	 * Returns the ID for the parent of the driver.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual unsigned long getDriverParentID(unsigned long databaseID, unsigned long driverID);

	/**
	 * Returns the ID for the protocol associated with the driver.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual unsigned long getProtocolIDForDriver(unsigned long databaseID, unsigned long driverID);

	/**
	 * Returns the ID of the servant for which we have the parentID, the n name and the child index.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual unsigned long getEnabledOrUniqueServantID(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex);

	/**
	 * Returns the ID of the servants for which we have the parentID, the n name and the child index.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual std::vector<unsigned long> getServantIDs(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex);

	/**
	 * Creates a new controller of type controllerType with name controllerName
	 * (if it doesn't already exist). Returns the ID associated with the controller.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual unsigned long createController(unsigned long databaseID, const std::string& controllerName, const std::string& controllerType, const std::string& family, bool isCommand, common::HashMap<std::string,  std::pair<std::string , std::string>> roleMap);

	/**
	 * Creates a new driver of type driverType with name driverName
	 * (if it doesn't already exist). Returns the ID associated with the driver.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual unsigned long createDriver(unsigned long databaseID, const std::string& driverName, const std::string& driverType, const std::string& protocolType, const std::string& driverMode, unsigned long parentID, const std::string& nDriverName, long childIndex);

	/**
	 * Clones a controller.
	 */
	virtual unsigned long cloneController(unsigned long databaseID, unsigned long controllerID);

	/**
	 * Assigns a servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual void assignServantFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex, unsigned long newParentID, const std::string& newNServantName, long newChildIndex);

	/**
	 * Assigns a servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual void assignServant(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex, unsigned long newServantID);

	/**
	 * Assigns a servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual void assignServantWithID(unsigned long databaseID, unsigned long servantID, bool newServant, bool localServant);

	/**
	 * Removes a servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException, ServantAccessorDirectImpl::NoSuchServantChildException
	 */
	virtual unsigned long removeServantFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nName, long index);

	/**
	 * Removes a servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual void removeServant(unsigned long databaseID, unsigned long servantID);

	/**
	 * Set the servant initialised.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual void setServantInitialised(unsigned long databaseID, unsigned long servantID, bool initialised);

	/**
	 * Returns true if the servant has been initialised.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual bool isServantInitialised(unsigned long databaseID, unsigned long servantID);

	/**
	 * Returns true if the servant is initialised.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual bool isServantInitialisedFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex);

	/**
	 * Resets the elapsed simulation time.
	 * throw NoSuchServantException
	 */
	virtual void resetElapsedSimulationTime(unsigned long databaseID, unsigned long driverID);

	/**
	 * Gets the elapsed simulation time.
	 * throw NoSuchServantException
	 */
	virtual double getElapsedSimulationTimeMs(unsigned long databaseID, unsigned long driverID);

	/**
	 * Clones the dataprovider.
	 */
	virtual unsigned long clone();

	/**
	 * Enables the servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException, ServantAccessorDirectImpl::ServantEnableException
	 */
	virtual void enableServant(unsigned long databaseID, unsigned long servantID, bool all);

	/**
	 * Enables the servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException, ServantAccessorDirectImpl::ServantEnableException
	 */
	virtual void enableServantFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex);

	/**
	 * Disables the servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException, ServantAccessorDirectImpl::ServantDisableException
	 */
	virtual void disableServant(unsigned long databaseID, unsigned long servantID, bool all);

	/**
	 * Disables the servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException, ServantAccessorDirectImpl::ServantDisableException
	 */
	virtual void disableServantFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex);

	/**
	 * Returns true if the servant is enabled.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual bool isServantEnabled(unsigned long databaseID, unsigned long servantID);

	/**
	 * Returns true if the servant is enabled.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual bool isServantEnabledFromParent(unsigned long databaseID, unsigned long parentID, const std::string& nServantName, long childIndex);

	/**
	 * Configures the servant.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual std::vector<unsigned long> configureServant(unsigned long databaseID, unsigned long servantID, const std::string& configurationName);

	/**
	 * Gets the description of the element.
	 */
	virtual std::string what(unsigned long databaseID, unsigned long elementID);

	/**
	 * Sets the type. Allows developers of N to fake a type, useful for inherited
	 * classes with same view in client
	 */
	virtual void setClientType(unsigned long databaseID, unsigned long servantID, const std::string& clientType);

	/**
	 * Returns true if the servant is pausable.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual bool isServantPausable(unsigned long databaseID, unsigned long servantID);

	/**
	 * Returns the channel of the driver.
	 * throw ServantAccessorDirectImpl::NoSuchServantException
	 */
	virtual long getDriverChannel(unsigned long databaseID, unsigned long servantID);

	/**
	 * Gets the servant ID from the name.
	 */
	virtual unsigned long getServantID(unsigned long databaseID, const std::string& name);

	/**
	 * Sets the servant visible.
	 */
	virtual void setServantVisible(unsigned long databaseID, unsigned long servantID, bool value);

	/**
	 * Sets the servant favourite.
	 */
	virtual void setServantFavourite(unsigned long databaseID, unsigned long servantID, bool value);

	/**
	 * Sets the servant command.
	 */
	virtual void setServantCommand(unsigned long databaseID, unsigned long servantID, bool value);

	/**
	 * Enables the configuration changed.
	 */
	virtual void enableConfigurationChanged(unsigned long databaseID);

	/**
	 * Disables the configuration changed.
	 */
	virtual void disableConfigurationChanged(unsigned long databaseID);

	/**
	 * Notifies a configuration changed for all the servants.
	 */
	virtual void configurationChanged(unsigned long databaseID);

	/**
	 * Adds the family to the servant.
	 */
	virtual void setServantFamily(unsigned long databaseID, unsigned long servantID, const std::string& familyName, const std::string& subfamilyName);

	/**
	 * Gets the servant family.
	 */
	virtual std::string getServantFamily(unsigned long databaseID, unsigned long servantID);

	/**
	 * Gets the servant subfamily.
	 */
	virtual std::string getServantSubfamily(unsigned long databaseID, unsigned long servantID);

	/**
	 * Returns true if the servant is favourite.
	 */
	virtual bool isServantFavourite(unsigned long databaseID, unsigned long servantID);

	/**
	 * Returns true if the servant is enabled and initialised.
	 */
	bool isServantEnabledAndInitialised(unsigned long databaseID, unsigned long servantID);

	/**
	 * Gets the servant properties value.
	 * throw nsdp::ServantAccessor::NoSuchServantException
	 */
	virtual void getServantPropertiesValue(unsigned long databaseID, unsigned long servantID, std::string& result);

	/**
	 * Gets the servant properties value.
	 * throw nsdp::ServantAccessor::NoSuchServantException
	 */
	virtual void getServantCommandsState(unsigned long databaseID, unsigned long servantID, std::string& result);

	void getWebSpyProperties(unsigned long databaseID, configuration::WebSpyContainer& aContainer);

	/**
	 * Returns true if the the servant is in debug.
	 */
	bool isServantInDebug(unsigned long databaseID, unsigned long servantID) const;


	/**
	 * Gets ids of property to save in commandbox.
	 */
	std::vector<unsigned long> getPropertiesSavedInBox(unsigned long databaseID);


private:
	/**
	 * throw ServantListIteratorCreationException
	 */
	std::vector<common::ServantDescriptor> createServantList(const std::vector<Database::ServantItem>& servants);

	/**
	 * throw ServantListIteratorCreationException
	 */
	std::vector<common::ChildServantDescriptor> createChildServantList(const Database::ChildServantMap& childServantMap);

	/**
	 * Creates the controller data for the whole configuration.
	 */
	void createControllersForWholeConfiguration(unsigned long databaseID, configuration::ServantData& data, const std::vector<Database::ServantItem>& servants);

	/**
	 * Creates the driver data for the whole configuration.
	 */
	void createDriversForWholeConfiguration(unsigned long databaseID, configuration::ServantData& data, const std::vector<Database::ServantItem>& servants);

	/**
	 * Creates the servant properties for the whole configuration.
	 */
	void createPropertyListForWholeConfiguration(configuration::ServantDescriptor * servantDescriptor, const Database::PropertyMap& propertyMap);

	/**
	 * Creates the servant properties for the whole configuration.
	 */
	void createCommandListForWholeConfiguration(configuration::ServantDescriptor * servantDescriptor, const Database::CommandMap& commandMap);

	/**
	 * Creates the child servant list for the whole configuration.
	 */
	void createChildServantListForWholeConfiguration(configuration::ServantDescriptor * servantDescriptor, const Database::ChildServantMap& childServantMap);

	/**
	 * Creates the controller data for the whole configuration update.
	 */
	void createControllersForWholeConfigurationUpdate(unsigned long databaseID, configuration::ServantDataUpdate& data, const std::vector<Database::ServantItem>& servants);

	/**
	 * Creates the driver data for the whole configuration update.
	 */
	void createDriversForWholeConfigurationUpdate(unsigned long databaseID, configuration::ServantDataUpdate& data, const std::vector<Database::ServantItem>& servants);

	/**
	 * Processes the property item for a setpoint property.
	 */
	void processServantSpySetpointProperty(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const Database::PropertyItem& item, std::set<unsigned long>& processedPropertyIDs, configuration::ServantSpyProperties& properties);

	/**
	 * Processes the property item for a property.
	 */
	void processServantSpyProperty(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const Database::PropertyItem& item, std::set<unsigned long>& processedPropertyIDs, configuration::ServantSpyProperties& properties);

	/**
	 * Processes the property item for a dynamic property.
	 */
	void processServantSpyDynamicProperty(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const Database::PropertyItem& item, configuration::ServantSpyProperties& properties);

	/**
	 * Processes the property item for a dynamic property.
	 */
	void processServantSpyDynamicProperties(unsigned long databaseID, unsigned long servantID, const std::string& propertyName, const Database::PropertyItem& item, configuration::ServantSpyProperties& properties);

};

}

#endif /*DATAPROVIDER_SERVANTACCESSORDIRECTIMPL_H_*/
