/*
 * Copyright 2022 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef ATP_ADAPTIVETHREADPOOL_H_
#define ATP_ADAPTIVETHREADPOOL_H_

#include "ConcurrentQueue.h"
#include <thread>
#include <list>
#include <functional>
#include <memory>
#include <atomic>
#include <cstdint>
#include <future>
#include <string>
#include <iostream>

namespace atp {

/**
 * Class defining a thread pool adapting the number of worker threads so that a new submission is executed immediately.
 */
class AdaptiveThreadPool {

public:
	/**
	 * Constructor.
	 * \param maxNumberOfWorkers The maximum number of workers.
	 * \param workerTimeout The maximum time of inactivity for a worker before stopping.
	 */
	explicit AdaptiveThreadPool(int maxNumberOfWorkers = -1, int workerTimeout = -1) :
		m_maxNumberOfWorkers{maxNumberOfWorkers}, m_workerTimeout{workerTimeout} {
	}

	/**
	 * Destructor.
	 */
	~AdaptiveThreadPool() {
		terminate();
	}

	/**
	 * Terminates the pool by canceling the worker threads and joining them.
	 */
	void terminate() {

		// Terminate if not already done.
		if (!m_terminated) {

			// Set terminated.
			m_terminated = true;

			// Cancels the queue so that any pop() call is returning.
			m_taskQueue.cancel();

			std::unique_lock<std::mutex> lock{m_mutex};

			// Join the workers.
			for (auto& w : m_workers) {
				w.second->join();
				delete w.second;
			}
		}
	}

	/**
	 * Returns true if the pool is terminated.
	 * \return True if terminated.
	 */
	bool isTerminated() const {
		return m_terminated;
	}

	/**
	 * Gets the number of worker threads.
	 * \return The number of worker threads.
	 */
	int getNumberOfWorkers() {

		std::unique_lock<std::mutex> lock{m_mutex};

		return m_workers.size();
	}

	/**
	 * Executes the function. If there is no worker available then a new one is created.
	 * \return The future of the executed task
	 */
	template<typename Function, typename ... Args>
	auto execute(Function &&f, Args &&... args) -> std::future<typename std::result_of<Function(Args...)>::type> {

		// Create a task wrapper to "remove" the type of the function called.
		using returnType = typename std::result_of<Function(Args...)>::type;
		using taskType = std::packaged_task<returnType()>;

		auto task = std::make_shared<taskType>(std::bind(std::forward<Function>(f), std::forward<Args>(args)...));
		auto taskResult = task->get_future();
		auto taskWrapper = [task]() {
			(*task)();
		};

		try {
			// Add the wrapper to the task queue.
			m_taskQueue.push(taskWrapper);
		}
		catch (const Canceled&) {
			return taskResult;
		}

		// Mutex for the free workers counter and the workers list.
		{
			std::unique_lock<std::mutex> lock{m_mutex};

			// Check if there is an available worker.
			if (m_freeWorkers == 0) {

				// Check the number of workers.
				if (m_maxNumberOfWorkers == -1
				|| m_workers.size() < m_maxNumberOfWorkers) {
					addWorker();
				}
			}
			else {
				// Decrease the number of free workers because there is one that will be used to execute the task.
				m_freeWorkers--;
			}
		}

		// Return the future.
		return taskResult;
	}

private:
	void addWorker() {

		// Create a new worker if there is none available.
		//m_workers.emplace_back(
		std::thread* worker = new std::thread([this] {

			// Loop on the tasks.
			while (true) {

				try {
					// Get a task.
					std::function<void()> task = m_taskQueue.pop(m_workerTimeout);

					// Execute the task.
					task();
				}
				catch (const Canceled&) {
					// Return if the queue has been canceled.
					return;
				}
				catch (const Timeout&) {
					// Remove the worker if timeout occurred.
					removeWorker(std::this_thread::get_id());

					return;
				}

				// This worker is free.
				m_freeWorkers++;
			}

		});

		m_workers[worker->get_id()] = worker;
	}

	void removeWorker(const std::thread::id& id) {

		// The pool is already terminated, there is no reason to remove the worker.
		// Try to get the mutex before testing terminated can lead to a deadlock.
		if (m_terminated) {
			return;
		}

		std::unique_lock<std::mutex> lock{m_mutex};

		// This was a free worker.
		m_freeWorkers--;

		// Remove the thread from the map.
		auto it = m_workers.find(id);

		// We need to detach the thread before deleting the object.
		// Is there a possibility to have problem with the detached thread i.e. when terminating the program?
		it->second->detach();
		delete it->second;

		// Erase the entry.
		m_workers.erase(it);
	}

	int m_maxNumberOfWorkers{-1};
	int m_workerTimeout{-1};
	std::unordered_map<std::thread::id, std::thread*> m_workers;
	std::mutex m_mutex;
	std::atomic_int m_freeWorkers{0};
	ConcurrentQueue<std::function<void()>> m_taskQueue;
	std::atomic_bool m_terminated{false};
};

}

#endif
