/*
 * Copyright 2022 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef ATP_CONCURRENTQUEUE_H_
#define ATP_CONCURRENTQUEUE_H_

#include <queue>
#include <thread>
#include <mutex>
#include <condition_variable>

namespace atp {

/**
 * Class defining an exception thrown when the queue has been canceled.
 */
class Canceled : public std::exception {

public:
	/**
	 * Constructor.
	 */
	Canceled() {}

	/**
	 * What function.
	 */
	virtual const char* what() const noexcept {
		return "Queue has been canceled";
	}
};

/**
 * Class defining an exception thrown when the pop() call has timed out.
 */
class Timeout : public std::exception {

public:
	/**
	 * Constructor.
	 */
	Timeout() {}

	/**
	 * What function.
	 */
	virtual const char* what() const noexcept {
		return "Time out while popping";
	}
};

/**
 * Class implementing a cancelable concurrent queue.
 */
template<typename Type>
class ConcurrentQueue {

public:
	/**
	 * Destructor.
	 */
	~ConcurrentQueue() {
	}

	/**
	 * Cancels the queue becoming not usable.
	 */
	void cancel() {

		std::unique_lock<std::mutex> lock{m_mutex};

		m_canceled = true;

		// Notify all the wait calls.
		m_condition.notify_all();
	}

	/**
	 * Returns true if is canceled.
	 * \return True if is canceled.
	 */
	bool canceled() {

		std::unique_lock<std::mutex> lock{m_mutex};

		return m_canceled;
	}

	/**
	 * Gets the front item. Blocking call until there is an item.
	 * \return An item or default value if the queue has been canceled.
	 */
	Type pop(int timeout = -1) {

		std::unique_lock<std::mutex> lock{m_mutex};

		if (m_canceled) {
			throw Canceled();
		}

		while (m_queue.empty()) {

			if (timeout == -1) {
				m_condition.wait(lock);
			}
			else {
				std::cv_status status = m_condition.wait_for(lock, std::chrono::seconds(timeout));
				if (status == std::cv_status::timeout) {
					throw Timeout();
				}
			}

			if (m_canceled) {
				throw Canceled();
			}
		}

		auto item = m_queue.front();
		m_queue.pop();

		return item;
	}

	/**
	 * Pushes an item or does nothing if the queue has been canceled.
	 * \param item An item.
	 */
	void push(const Type & item) {

		std::unique_lock<std::mutex> lock{m_mutex};

		if (m_canceled) {
			throw Canceled();
		}

		m_queue.push(item);
		lock.unlock();
		m_condition.notify_one();
	}

	/**
	 * Returns the size of the queue.
	 * \return The size of the queue.
	 */
	typename std::queue<Type>::size_type size() {

		std::unique_lock<std::mutex> lock{m_mutex};
		return m_queue.size();
	}

private:
	std::queue<Type> m_queue;
	std::mutex m_mutex;
	std::condition_variable m_condition;
	bool m_canceled{false};
};

}

#endif
