/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <core/message/Shutdown.h>
#include <core/message/commandzone/NewEstimatedTimeMessage.h>
#include <core/commandzone/CommandZoneEventServer.h>
#include <common/base/Queue.h>
#include <common/output/OutputStream.h>
#include <common/base/ServerProperties.h>
#include <iostream>
#include "CLEstimatedTimeManager.h"
#include "CommandLineManager.h"

using namespace std;
using namespace common;

CLEstimatedTimeManager* CLEstimatedTimeManager::_instance = 0;
bool CLEstimatedTimeManager::_running= false;

const std::string CLEstimatedTimeManager::USE_ESTIMATED_TIME = "useEstimatedTime";

CLEstimatedTimeManager::CLEstimatedTimeManager() :
	_queue(new common::Queue<core::Message>(1024)),
	_thread(nullptr),
	_estimatedRunning(false),
	_activated(true) {
	try {
		_useEstimatedTime = common::ServerProperties::getInstance()->get<bool>(USE_ESTIMATED_TIME);
	} catch (const Exception& e) {
		out << "Setting default value true to property '" << USE_ESTIMATED_TIME << "'" << eol;
		_useEstimatedTime = true;
	}
	if ((Simulated::activated == true) || (Simulated::for3d == true)) {
		_useEstimatedTime = false;
	}
}

CLEstimatedTimeManager::~CLEstimatedTimeManager() {
	delete _queue;
}

void CLEstimatedTimeManager::initInstance() {
	if (_instance == nullptr) {
		_instance = new CLEstimatedTimeManager();
		_instance->init();
	}
}

void CLEstimatedTimeManager::resetInstance() {
	if (_instance != nullptr) {
		_instance->reset();
	}
}

CLEstimatedTimeManager* CLEstimatedTimeManager::getInstance() {
	return _instance;
}


void CLEstimatedTimeManager::init() {
	_running = true;
	_thread = new std::thread(&CLEstimatedTimeManager::run);
}

void CLEstimatedTimeManager::reset() {

	_queue->put(new core::Shutdown(false));

	_thread->join();
	delete _thread;

}

//void CLEstimatedTimeManager::setActivationState(bool state) {
//	if (_useEstimatedTime == false) {
//		_activated = false;
//	}
//	else {
//		_activated = state;
//	}
//	if (_activated) {
//		newEstimatedTime();
//	}
//}

void CLEstimatedTimeManager::newEstimatedTime() {
	if (_useEstimatedTime) {
		if (_estimatedRunning == false) {
			_queue->put(new core::NewEstimatedTimeMessage(0, 0, true));
		}
	}
}

void CLEstimatedTimeManager::getEstimatedTime() {
	if (_useEstimatedTime) {
		if (_estimatedRunning == false) {
			_queue->put(new core::NewEstimatedTimeMessage(0, 0, false));
		}
	}
}

/**
 * Main loop
 */
void CLEstimatedTimeManager::run() {

	_running = true;
	while (_running) {

		core::Message* message = 0;

		try {
			message = _instance->_queue->get();

			_instance->processMessage(message);

		} catch (common::Queue<core::Message>::UnavailableException& e) {
			_running = false;
		} catch (exception& e) {
			err << "STL exception in " << message->getTypeString() << " : " << e.what() << eol;
		} catch (...) {
			err << "Unexpected exception in " << message->getTypeString() << eol;
		}

		delete message;
	}

}

void CLEstimatedTimeManager::processMessage(core::Message* message) {
	if (_useEstimatedTime) {
		_estimatedRunning = true;
		if (message->getType() == core::Message::SHUTDOWN) {
			processMessage(static_cast<core::Shutdown*>(message));

		} else if (message->getType() == core::Message::COMMAND_ZONE_EVENT) {
			processMessage(static_cast<core::CommandZoneEvent*>(message));

		}
		_estimatedRunning = false;
	}
}

void CLEstimatedTimeManager::processMessage(core::Shutdown* message) {

	_running = false;
}

void CLEstimatedTimeManager::processMessage(core::CommandZoneEvent* commandZoneEvent) {
	if (_useEstimatedTime) {
		_estimatedRunning = true;
		if (commandZoneEvent->getEventType() == core::CommandZoneEvent::NEW_ESTIMATED_TIME) {
			if (static_cast<core::NewEstimatedTimeMessage*>(commandZoneEvent)->getDoCompute() == true) {
				CommandLineManager::getInstance()->computeEstimatedTime();
			}
			else {
				CommandLineManager::getInstance()->getLastEstimatedTime(0);
			}
		}
		_estimatedRunning = false;
	}
}
