/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief CommandLine Manager (Singleton)
 * \author J. Locatelli
 * \date 12-01-2011
 */

#include "CommandLineCommand.h"
#include "CommonCmdDef.h"
#include <ostream>
#include <dataprovider/database/Database.h>
#include "CommandSystem/CommandLine/CommandLineManager.h"
#include "Utilities/StringUtilities.h"
#include "CommandSystem/CommandLine/CommonCmdDef.h"


using namespace std;

CommandLineCommand::CommandLineCommand(const CommandLineCommand& command) {
	_controller = command._controller;
	_properties = command._properties;
	_baseproperties = command._baseproperties;
	_simpleproperties = command._simpleproperties;
	_stop = command._stop;
	_start = command._start;
}

void CommandLineCommand::applyProperties() const {
	for (auto prop : _properties) {
//		_controller->attach(CommandLineManager::getInstance());
		string propertyName = prop.first;
		string lvalue = prop.second;
//		cout << "applyProperties : " << propertyName << " -> " << lvalue << endl;
		try {

			if (propertyName.empty() == false) {
				try {
					uint32 servantID = dataprovider::Database::getInstance(0)->getServantID(_controller->getName());
					uint32 propertyID = dataprovider::Database::getInstance(0)->getPropertyID(servantID, propertyName);
					uint32 enumlabelsID = dataprovider::Database::getInstance(0)->getEnumeratedLabelsProperty(propertyID);
					if (enumlabelsID != 0) {
						uint32 enumvaluesID = dataprovider::Database::getInstance(0)->getEnumeratedValuesProperty(propertyID);
						// Get the dynamic property size.
						uint32 enumeratedLabelsSize = dataprovider::Database::getInstance(0)->getDynamicPropertySize(enumlabelsID);
						// Search for the label.
						for (int i = 0; i < enumeratedLabelsSize; ++i) {
							uint32 labelPropertyID = dataprovider::Database::getInstance(0)->getDynamicPropertyIDAtIndex(enumlabelsID, i, false);
							std::string label = dataprovider::Database::getInstance(0)->getValue<string>(labelPropertyID);

							// Check the label.
							if (label == lvalue) {
								uint32 valuePropertyID = dataprovider::Database::getInstance(0)->getDynamicPropertyIDAtIndex(enumvaluesID, i, false);
								lvalue = dataprovider::Database::getInstance(0)->getFormattedValue(valuePropertyID);
								break;
							}
						}
					}
				} catch (...) {
				}
				string ptype = _controller->getPropertyType(propertyName);
				if (ptype == UINT16_PROPERTY) {
					// Test conversion
					if (lvalue != "=") {
						uint16 val = StringUtilities::convert<uint16>(lvalue);
						if (val == numeric_limits<uint16>::max()) {
							throw CommonCmdDef::BadInput(prop.second);
						}
						_controller->setValue(propertyName, val);
					}
				}else if (ptype == UINT32_PROPERTY) {
					// Test conversion
					if (lvalue != "=") {
						uint32 val = StringUtilities::convert<uint32>(lvalue);
						if (val == numeric_limits<uint32>::max()) {
							throw CommonCmdDef::BadInput(prop.second);
						}
						_controller->setValue(propertyName, val);
					}
				}else if (ptype == UINT64_PROPERTY) {
					// Test conversion
					if (lvalue != "=") {
						uint64 val = StringUtilities::convert<uint64>(lvalue);
						if (val == numeric_limits<uint64>::max()) {
							throw CommonCmdDef::BadInput(prop.second);
						}
						_controller->setValue(propertyName, val);
					}
				} else if (ptype == INT16_PROPERTY) {
					// Test conversion
					if (lvalue != "=") {
						int16 val = StringUtilities::convert<int16>(lvalue);
						if (val == numeric_limits<int16>::max()) {
							throw CommonCmdDef::BadInput(prop.second);
						}
						_controller->setValue(propertyName, val);
					}
				}else if (ptype == INT32_PROPERTY) {
					// Test conversion
					if (lvalue != "=") {
						int32 val = StringUtilities::convert<int32>(lvalue);
						if (val == numeric_limits<int32>::max()) {
							throw CommonCmdDef::BadInput(prop.second);
						}
						_controller->setValue(propertyName, val);
					}
				}else if (ptype == INT64_PROPERTY) {
					// Test conversion
					if (lvalue != "=") {
						int64 val = StringUtilities::convert<int64>(lvalue);
						if (val == numeric_limits<int64>::max()) {
							throw CommonCmdDef::BadInput(prop.second);
						}
						_controller->setValue(propertyName, val);
					}
				} else if (ptype == FLOAT32_PROPERTY) {
					if (lvalue != "=") {
						float32 val = StringUtilities::convert<float32>(lvalue);
						if (val == numeric_limits<float32>::max()) {
							throw CommonCmdDef::BadInput(prop.second);
						}
						_controller->setValue(propertyName, val);
					}
				} else if (ptype == FLOAT64_PROPERTY) {
					if (lvalue != "=") {
						float64 val = StringUtilities::convert<float64>(lvalue);
						if (val == numeric_limits<float64>::max()) {
							throw CommonCmdDef::BadInput(prop.second);
						}
						_controller->setValue(propertyName, val);
					}
				} else if (ptype == BOOLEAN_PROPERTY) {
					if (lvalue != "=") {
						// Test conversion
						bool val = false;
						if ((lvalue == "0") || (lvalue == "false")) {
							val = false;
						}
						else if ((lvalue == "1") || (lvalue == "true")) {
							val = true;
						}
						else {
							throw CommonCmdDef::BadInput(prop.second);
						}
						_controller->setValue(propertyName, val);
					}
				} else if (ptype == STRING_PROPERTY) {
					_controller->setValue(propertyName, lvalue);
				} else {
					CommandLineManager::getInstance()->debug(
							_controller->getName() + " property " + propertyName + " " + ptype + " doesn''t implemented");
					// Print Error
					CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + prop.first);
				}
			}
		}
		catch (dataprovider::Database::NoSuchServantException& e) {
//			_controller->detach(CommandLineManager::getInstance());
			throw CommonCmdDef::BadInput("Cannot find controller : " + _controller->getName());
		}
		catch (dataprovider::Database::NoSuchPropertyException& e) {
//			_controller->detach(CommandLineManager::getInstance());
			throw CommonCmdDef::BadInput("Property not found: controller " + _controller->getName() + " , property " + propertyName);
		}
		catch (dataprovider::Database::BadPropertyTypeException& e) {
//			_controller->detach(CommandLineManager::getInstance());
			throw CommonCmdDef::BadInput("Bad property type: controller " + _controller->getName() + " , property " + propertyName);
		} catch (dataprovider::Database::IndexOutOfBoundsException & e) {
//			_controller->detach(CommandLineManager::getInstance());
			throw CommonCmdDef::BadInput("Bad index on dynamic property type: controller " + _controller->getName() + " , property " + propertyName);
		} catch (CommonCmdDef::BadInput& e) {
//			_controller->detach(CommandLineManager::getInstance());
			throw;
		} catch (CannotSetValue& e) {
//			_controller->detach(CommandLineManager::getInstance());
			if (StringUtilities::cmp(e.what(), ExperimentController::LIMIT_STR) == true) {
				throw CommonCmdDef::BadInput("Cannot set value on: controller " + _controller->getName() + " , property " + propertyName, true);
			}
			else {
				throw CommonCmdDef::BadInput("Cannot set value on: controller " + _controller->getName() + " , property " + propertyName);
			}
		} catch (...) {
			CommandLineManager::getInstance()->debug(_controller->getName() + " property " + propertyName + " doesn''t existed");
			_controller->detach(CommandLineManager::getInstance());
			throw CommonCmdDef::BadInput("Property not found: controller " + _controller->getName() + " , property " + propertyName);
		}
//		_controller->detach(CommandLineManager::getInstance());
	}
	for (auto prop : _baseproperties) {
		try {
			string ptype = prop.first->getType();
//			cout << _controller->getName() << " : " << prop.first->getName() << " : " << ptype << " : " << prop.second << endl;
			if (ptype == FLOAT32_PROPERTY) {
				*dynamic_cast<Property<float32>*>(prop.first) = boost::lexical_cast<float32>(prop.second);
			} else 	if (ptype == FLOAT64_PROPERTY) {
				*dynamic_cast<Property<float64>*>(prop.first) = boost::lexical_cast<float64>(prop.second);
			} else if (ptype == UINT16_PROPERTY) {
				*dynamic_cast<Property<uint16>*>(prop.first) = boost::lexical_cast<uint16>(prop.second);
			} else if (ptype == UINT32_PROPERTY) {
				*dynamic_cast<Property<uint32>*>(prop.first) = boost::lexical_cast<uint32>(prop.second);
			} else if (ptype == UINT32_PROPERTY) {
				*dynamic_cast<Property<uint64>*>(prop.first) = boost::lexical_cast<uint64>(prop.second);
			} else if (ptype == INT16_PROPERTY) {
				*dynamic_cast<Property<int16>*>(prop.first) = boost::lexical_cast<int16>(prop.second);
			} else if (ptype == INT32_PROPERTY) {
				*dynamic_cast<Property<int32>*>(prop.first) = boost::lexical_cast<int32>(prop.second);
			} else if (ptype == INT32_PROPERTY) {
				*dynamic_cast<Property<int64>*>(prop.first) = boost::lexical_cast<int64>(prop.second);
			} else if (ptype == BOOLEAN_PROPERTY) {
				bool val;
				if (prop.second == "false") {
					*dynamic_cast<Property<bool>*>(prop.first) = false;
				}
				else if (prop.second == "true") {
					*dynamic_cast<Property<bool>*>(prop.first) = true;
				}
				else {
					*dynamic_cast<Property<bool>*>(prop.first) = boost::lexical_cast<bool>(prop.second);
				}
			} else if (ptype == STRING_PROPERTY) {
				*dynamic_cast<Property<string>*>(prop.first) = prop.second;
			} else {
				common::err << _controller->getName() << " : " << prop.first->getName() << " : " << ptype
						<< " :  not managed in command line" << common::eol;
			}
		} catch (boost::bad_lexical_cast &e) {
			throw CommonCmdDef::BadInput("Bad cast value: " + prop.second);
		}
	}
	for (auto prop : _simpleproperties) {
		try {
			string ptype = prop.first->getType();
//			cout << _controller->getName() << " : " << prop.first->getName() << " : " << ptype  << " : " << prop.second  << endl;
			if (ptype == FLOAT32_PROPERTY) {
				*dynamic_cast<SimpleProperty<float32>*>(prop.first) = boost::lexical_cast<float32>(prop.second);
			} else 	if (ptype == FLOAT64_PROPERTY) {
				*dynamic_cast<SimpleProperty<float64>*>(prop.first) = boost::lexical_cast<float64>(prop.second);
			} else if (ptype == UINT16_PROPERTY) {
				*dynamic_cast<SimpleProperty<uint16>*>(prop.first) = boost::lexical_cast<uint16>(prop.second);
			} else if (ptype == UINT32_PROPERTY) {
				*dynamic_cast<SimpleProperty<uint32>*>(prop.first) = boost::lexical_cast<uint32>(prop.second);
			} else if (ptype == UINT64_PROPERTY) {
				*dynamic_cast<SimpleProperty<uint64>*>(prop.first) = boost::lexical_cast<uint64>(prop.second);
			} else if (ptype == INT16_PROPERTY) {
				*dynamic_cast<SimpleProperty<int16>*>(prop.first) = boost::lexical_cast<int16>(prop.second);
			} else if (ptype == INT32_PROPERTY) {
				*dynamic_cast<SimpleProperty<int32>*>(prop.first) = boost::lexical_cast<int32>(prop.second);
			} else if (ptype == INT64_PROPERTY) {
				*dynamic_cast<SimpleProperty<int64>*>(prop.first) = boost::lexical_cast<int64>(prop.second);
			} else if (ptype == BOOLEAN_PROPERTY) {
				if (prop.second == "false") {
					*dynamic_cast<SimpleProperty<bool>*>(prop.first) = false;
				}
				else if (prop.second == "true") {
					*dynamic_cast<SimpleProperty<bool>*>(prop.first) = true;
				}
				else {
					*dynamic_cast<SimpleProperty<bool>*>(prop.first) = boost::lexical_cast<bool>(prop.second);
				}
			} else if (ptype == STRING_PROPERTY) {
				*dynamic_cast<SimpleProperty<string>*>(prop.first) = prop.second;
			} else {
				common::err << _controller->getName() << " : " << prop.first->getName() << " : " << ptype
						<< " :  not managed in command line" << common::eol;
			}
		} catch (boost::bad_lexical_cast &e) {
			throw CommonCmdDef::BadInput("Bad cast value: " + prop.second);
		}
	}
}

std::map<std::string, std::string> CommandLineCommand::getPropertiesMap() {
	std::map<std::string, std::string> properties;
	for (auto prop : _properties) {
		properties[prop.first] = prop.second;
	}
	for (auto prop : _baseproperties) {
		properties[prop.first->getName()] = prop.second;
	}
	for (auto prop : _simpleproperties) {
		properties[prop.first->getName()] = prop.second;
	}
	return properties;
}
