/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief CommandLine Manager (Singleton)
 * \author J. Locatelli
 * \date 12-01-2011
 */

#include "CommandLineManager.h"
#include "CLEstimatedTimeManager.h"
#include "CommandLineParser.h"
#include "CommonCmdDef.h"
#include "LoggingServices/Log.h"
#include "LoggingServices/Level.h"
#include "ErrorsAlarmsManager/ErrorDef.h"
#include "InstrumentAbstraction/ExperimentController.h"
#include "InstrumentManager/InstrumentManager.h"
#include "DataProvider/ChangeAspect.h"
#include "CommandNames.h"
#include "InstrumentAbstraction/ClientCommand.h"
#include "Scheduler/Sender.h"
#include "Utilities/StringUtilities.h"
#include "common/base/ServerProperties.h"
#include <core/scheduler/Sender.h>
#include <core/dataaccess/DataAccess.h>
#include <ics/IcsServerManager.h>
#include <dataprovider/database/Database.h>
#include <boost/algorithm/string.hpp>
#include <iostream>
#include <fstream>
#include <InstrumentManager/InstrumentManager.h>

using namespace std;

std::string CommandLineManager::m_ParserType = "";
const std::string CommandLineManager::COUNT_REPEATER_TYPE = "count_repeater";

/**
 * Init static members
 */
std::map<std::string, CommandLineManager*> CommandLineManager::m_Instances;
const std::string CommandLineManager::DEFAULT = "default";

const string CommandLineManager::COMMANDLINE_STATE_PROPERTY = "commandline_state";
const string CommandLineManager::ACTUAL_COMMAND_PROPERTY = "actual_command";
const string CommandLineManager::LAST_COMMAND_PROPERTY = "last_command";
const string CommandLineManager::XBU_RUNNING_PROPERTY = "xbu_running";
const string CommandLineManager::XBU_FILENAME_PROPERTY = "xbu_filename";
const string CommandLineManager::XBU_FULL_FILENAME_PROPERTY = "xbu_full_filename";
const string CommandLineManager::XBU_NUMBER_LINES_PROPERTY = "xbu_number_lines";
const string CommandLineManager::XBU_CURRENT_LINE_PROPERTY = "xbu_current_line";
const string CommandLineManager::XBU_PREVIOUS_COMMAND_PROPERTY = "xbu_previous_command";
const string CommandLineManager::XBU_CURRENT_COMMAND_PROPERTY = "xbu_current_command";
const string CommandLineManager::XBU_NEXT_COMMAND_PROPERTY = "xbu_next_command";
const string CommandLineManager::FORLOOP_RUNNING_PROPERTY = "forloop_running";
const string CommandLineManager::FORLOOP_NUMBER_STEPS_PROPERTY = "forloop_number_steps";
const string CommandLineManager::FORLOOP_CURRENT_STEP_PROPERTY = "forloop_current_step";
const string CommandLineManager::PAL_CONTENTS_PROPERTY = "pal_contents";
const string CommandLineManager::PAL_FULL_FILENAME_PROPERTY = "pal_full_filename";

/*
 * Constructor
 */
CommandLineManager::CommandLineManager() :
	m_SpyController(NULL), m_CommandLineActivated(false), m_StartActivated(false), m_PauseActivated(false),
			m_StopActivated(false), m_SimulatedCommand(false) {

	// Get the default command line type
	string filename;
	filename.append(common::ServerProperties::getInstance()->getConfigPath());
	filename.append("/CommandLineType.txt");
	ifstream file;
	file.open(filename.c_str());
	if (file.is_open()) {
		string type;
		file >> type;
		try {
			setParserType(type);
		} catch (BadParser&) {
			// cannot happen
			m_ParserType = CommonCmdDef::GENERIC_PARSER;
		}
		file.close();
	}
	else {
		setParserType(CommonCmdDef::GENERIC_PARSER);
	}
}

/*
 * Destructor
 */
CommandLineManager::~CommandLineManager() {
}

/*
 * getInstance
 */
CommandLineManager* CommandLineManager::getInstance(const std::string& name) {

	map<string, CommandLineManager*>::iterator i = m_Instances.find(name);

	if (i == m_Instances.end()) {
		m_Instances[name] = new CommandLineManager();
		CLEstimatedTimeManager::initInstance();
	}
	return m_Instances[name];
}

/*
 * reset
 */
void CommandLineManager::reset() {

	CLEstimatedTimeManager::resetInstance();
	for(auto& i : m_Instances) {
		delete i.second;
	}
	m_Instances.clear();
}

/*
 * setParserType
 */
void CommandLineManager::setParserType(const std::string& aParser)  {
	m_ParserType = aParser;
	m_Parser.reset(new CommandLineParser());
	try {
		m_Parser->setParser(m_ParserType);
	}
	catch (BadParser&) {
		m_ParserType = CommonCmdDef::GENERIC_PARSER;
		throw;
	}
}

/*
 * getParserType
 */
string CommandLineManager::getParserType() {
	return m_ParserType;
}

/*
 * executeCommand
 */
void CommandLineManager::executeCommand(const string& text, bool watchdog, bool sync) {

	if (m_SpyController == NULL) {
		try {
			m_SpyController = (AbstractController*) InstrumentManager::getInstance()->getExperimentController(
					"CommandLineSpy");
			setCommandLineState(m_CommandLineActivated);
		} catch (...) {
			m_SpyController = NULL;
		}
	}

	m_PauseActivated = false;
	m_StopActivated = false;

	log("Command> " + text);

	// Create the Items vector from the command line text
	{
		std::scoped_lock lock1(m_ItsMutex);
		m_Its = text;
	}

	if (m_Its.end() == false) {

		if (watchdog == false) {
			CLEstimatedTimeManager::getInstance()->newEstimatedTime();
			_estimatedTimeRefreshCounter.clear();
		}
		// Execute command in other thread
		if (!sync) {
			std::thread td(std::bind(&CommandLineParser::executeCommand, m_Parser, m_Its));
			td.detach();
		} else {
			m_Parser->executeCommand(m_Its);
		}
	}
	else {
		sendSynchroEvent();
	}

	if (m_SpyController != NULL) {
		m_SpyController->setValue(LAST_COMMAND_PROPERTY, text);
		setSpyActualCommand(text);
	}
}

/*
 * simulateCommand
 */
void CommandLineManager::simulateCommand(const string& text, bool sync) {

	m_PauseActivated = false;
	m_StopActivated = false;

	log("Command> " + text);

	// Create the Items vector from the command line text
	Items its(text);

	m_SimulatedCommand = true;
	if (its.end() == false) {
		m_Parser->executeCommand(its);
		//sendSynchroEvent();
	}
	else {
		sendSynchroEvent();
	}
	m_SimulatedCommand = false;
}

/*
 * pauseCommands
 */
void CommandLineManager::pauseCommands() {
	std::scoped_lock lock1(m_StartedControllerMutex);
	m_PauseActivated = true;
	for (const auto& sc : m_StartedController) {
		ExperimentController* ec = sc.second.getController();
		Sender<ClientCommand>::post(ClientCommand(ec, COMMAND_PAUSE, true, false));
	}
}

/*
 * restartCommands
 */
void CommandLineManager::restartCommands() {
	std::scoped_lock lock1(m_StartedControllerMutex);
	m_PauseActivated = false;
	for (const auto& sc : m_StartedController) {
		ExperimentController* ec = sc.second.getController();
		Sender<ClientCommand>::post(ClientCommand(ec, COMMAND_CONTINUE, true, false));
	}
}

/*
 * stopCommands
 */
void CommandLineManager::stopCommands() {
	Log::getInstance()->log(Level::s_Debug, "Client command line stop requested");
	std::scoped_lock lock1(m_StartedControllerMutex);
	m_PauseActivated = false;
	m_StopActivated = true;

	// Stop
	for (const auto& sc : m_StartedController) {
		ExperimentController* ec = sc.second.getController();
		if (ec->getType() != COUNT_REPEATER_TYPE) {
			Sender<ClientCommand>::post(ClientCommand(ec, COMMAND_STOP, true, false));
		}
	}
}

/*
 * stopAtEndCommands
 */
void CommandLineManager::stopAtEndCommands() {
	std::scoped_lock lock1(m_StartedControllerMutex);

	m_PauseActivated = false;
	m_StopActivated = true;

	// Stop
	for (const auto& sc : m_StartedController) {
		ExperimentController* ec = sc.second.getController();
		if (ec->getType() != COUNT_REPEATER_TYPE) {
			if (ec->containsCommand(COMMAND_STOP_AT_END) == true) {
				Sender<ClientCommand>::post(ClientCommand(ec, COMMAND_CONTINUE, true, false));
				Sender<ClientCommand>::post(ClientCommand(ec, COMMAND_STOP_AT_END, true, false));
			} else {
				Sender<ClientCommand>::post(ClientCommand(ec, COMMAND_STOP, true, false));
			}
		}
	}
}

/*
 * output
 */
void CommandLineManager::output(const std::string& text) {
	Log::getInstance()->log(Level::s_Info, text);
}

/*
 * outputError
 */
void CommandLineManager::outputError(int32  id, const std::string& text) {
	string buf = "Parser error";

	// Parse error type for completing message
	switch (id) {
	case COMMAND_REFUSED_ERROR:
		buf = "Command refused";
		buf.append(text);
		break;
	case COMMAND_LINE_BAD_COMMAND:
		buf = "Parser error, Unrecognised command ";
		buf.append(text);
		break;
	case COMMAND_LINE_ILLEGAL_PARAM:
		buf = "Parser error ";
		buf.append(text);
		break;
	case COMMAND_LINE_DISACTIVED_PARAM:
		buf = text;
		buf.append(" isn't activated");
		break;
	default:
		if (text.empty() == false) {
			buf = text;
		}
	}

	// Log error
	Log::getInstance()->log(Level::s_Error, buf);
}

/*
 * debug
 */
void CommandLineManager::debug(const std::string& text) {
	Log::getInstance()->log(Level::s_Debug, text);
}

/*
 * log
 */
void CommandLineManager::log(const std::string& text) {
	Log::getInstance()->log(Level::s_Info, text);
}

/*
 * outputWarning
 */
void CommandLineManager::outputWarning(const std::string& text) {
	Log::getInstance()->log(Level::s_Warning, text);
}

/*
 * sendSynchroEvent
 */
void CommandLineManager::sendSynchroEvent() {
	Log::getInstance()->log(Level::s_CLSync, "");
}

/*
 * startControllers
 */
void CommandLineManager::startControllers(const std::string& command, const vector<CommandLineCommand>& controllers,
		bool sync) {

	if (m_SimulatedCommand == true) {
		return;
	}
	if (controllers.empty() == true) {
		return;
	}
	{
		std::scoped_lock lock1(m_StartedControllerMutex);
		for(const auto& clc : controllers) {
			ExperimentController* ec = clc.getController();
			if (clc.doStopBefore() == true) {
				if (ec->containsCommand("stop")) {
					ec->detach(this);
					if (ec->commandStatus.isRunning() == true) {
						Sender<ClientCommand>::post(ClientCommand(ec, "stop", false, false));
						while(ec->commandStatus.isRunning() == true) {
							std::this_thread::sleep_for(std::chrono::milliseconds(100));
						}
					}
				}
			}
		}
	}
	{
		std::scoped_lock lock1(m_StartedControllerMutex);
		for(const auto& clc : controllers) {
			clc.applyProperties();
		}
	}

	{
		std::scoped_lock lock1(m_StartedControllerMutex);
		for(const auto& clc : controllers) {
			ExperimentController* ec = clc.getController();
			// add listener to experiment controller
			ec->attach(this);
		}
	}

	m_StartActivated = true;

	if (sync == true) {

		// run the commands in threads in a group
		vector<thread> group;
		for(const auto& clc : controllers) {
			if (clc.toStart() == true) {
				ExperimentController* ec = clc.getController();
				m_StartedController[ec->getName()] = clc;
				group.push_back(std::thread(std::bind(&CommandLineManager::startControllerCommand, ec, command, true)));
			}
		}

		// wait for the termination of the threads
		for(auto& t : group) {
			t.join();
		}


	} else {
		// start the commands without waiting
		for(const auto& clc : controllers) {
			if (clc.toStart() == true) {
				ExperimentController* ec = clc.getController();
				startControllerCommand(ec, command, false);
			}
		}
		m_StartActivated = false;
		sendSynchroEvent();
	}
}

/*
 * stopControllers
 */
void CommandLineManager::stopControllers(const vector<ExperimentController*>& controllers) {

	// run the functions waiting in threads in a group
	vector<thread> group;

	for(const auto& clc : controllers) {
		group.push_back(std::thread(std::bind(&CommandLineManager::stopAndWaitForControllerCommand, clc)));
	}

	// wait for the termination of the threads
	for(auto& t : group) {
		t.join();
	}
}

/*
 * update
 */
void CommandLineManager::update(ChangeAspect* changeAspect) {

	switch (changeAspect->getType()) {
	case PROPERTY:
		break;

	case PROGRESS:
		if (m_PauseActivated == false) {
			string controllername = changeAspect->getControllerName();
			int32 progression = changeAspect->getLongParam();

			if (progression == 100) {
				std::scoped_lock lock1(m_StartedControllerMutex);
				map<string, CommandLineCommand>::iterator iter = m_StartedController.find(controllername);
				if (iter != m_StartedController.end()) {
					ExperimentController* ec = iter->second.getController();
					ec->detach(this);
					m_StartedController.erase(iter);
				}
				m_StartActivated = false;
			}
			else {
				if (_estimatedTimeRefreshCounter.getTime() > 10.) {
					CLEstimatedTimeManager::getInstance()->newEstimatedTime();
					_estimatedTimeRefreshCounter.clear();
				}
			}
		}
		break;

	case ERROR:
		if (Simulated::activated == true) {
			m_Parser->incNbError();
		}
		//outputError(changeAspect->getLongParam(), changeAspect->getStringParam());
		break;

	default:
		break;
	}
}

/*
 *  setSpyActualCommand
 */
void CommandLineManager::setSpyActualCommand(const std::string& text) {
	if (m_SpyController != NULL) {
		m_SpyController->updateValue(ACTUAL_COMMAND_PROPERTY, text);
	}
}

/*
 *  xbuSpy
 */
void CommandLineManager::xbuSpy(const std::string& filename, uint32 numberOfLines, uint32 currentLineNumber,
		const std::string& previousLineString, const std::string& currentLineString, const std::string& nextLineString) {
	if (m_SpyController != NULL) {
		if (filename.empty() == true) {
			m_SpyController->updateValue(XBU_RUNNING_PROPERTY, false);
		} else {
			m_SpyController->updateValue(XBU_FILENAME_PROPERTY, filename);
			m_SpyController->updateValue(XBU_NUMBER_LINES_PROPERTY, (int32 ) numberOfLines);
			m_SpyController->updateValue(XBU_CURRENT_LINE_PROPERTY, (int32 ) currentLineNumber);
			ostringstream sbuf1;
			if ((currentLineNumber - 1) != 0) {
				sbuf1 << currentLineNumber - 1 << " - " << previousLineString;
			}
			m_SpyController->updateValue(XBU_PREVIOUS_COMMAND_PROPERTY, sbuf1.str());
			ostringstream sbuf2;
			sbuf2 << currentLineNumber << " - " << currentLineString;
			m_SpyController->updateValue(XBU_CURRENT_COMMAND_PROPERTY, sbuf2.str());
			ostringstream sbuf3;
			if (currentLineNumber != numberOfLines) {
				sbuf3 << currentLineNumber + 1 << " - " << nextLineString;
			}
			m_SpyController->updateValue(XBU_NEXT_COMMAND_PROPERTY, sbuf3.str());
			m_SpyController->updateValue(XBU_RUNNING_PROPERTY, true);
		}
	}
}

/*
 *  forloopSpy
 */
void CommandLineManager::forloopSpy(uint32 numberOfSteps, uint32 currentStep) {
	if (m_SpyController != NULL) {
		if ((numberOfSteps == 0) && (currentStep == 0)) {
			m_SpyController->updateValue(FORLOOP_RUNNING_PROPERTY, false);
		} else {
			m_SpyController->updateValue(FORLOOP_NUMBER_STEPS_PROPERTY, (int32 ) numberOfSteps);
			m_SpyController->updateValue(FORLOOP_CURRENT_STEP_PROPERTY, (int32 ) currentStep);
			m_SpyController->updateValue(FORLOOP_RUNNING_PROPERTY, true);
		}
	}
}

/*
 * setCommandLineState
 */
void CommandLineManager::setCommandLineState(bool state) {
	m_CommandLineActivated = state;
	if (m_SpyController != NULL) {
		m_SpyController->setValue(COMMANDLINE_STATE_PROPERTY, m_CommandLineActivated);
	}
}

void CommandLineManager::startControllerCommand(ExperimentController * controller, const std::string& command, bool blocking) {

	try {

		int32 databaseID = 0;
		int32 servantID = (int32) ::core::DataAccess::getInstance()->getServantID(controller->getName());
		int32 commandID = (int32) ::core::DataAccess::getInstance()->getCommandID(databaseID, servantID, command);

		::core::Sender::newCommand(databaseID, commandID, false, blocking);
	} catch (const common::Exception & e) {
		common::err << "Commandline cannot launch " << controller->getName() << " " << command << common::eol;
	}
}

void CommandLineManager::waitForControllerCommand(ExperimentController * controller) {

	try {
		int32 databaseID = 0;
		int32 servantID = (int32) ::core::DataAccess::getInstance()->getServantID(controller->getName());

		try {
			int32 commandID = (int32) ::core::DataAccess::getInstance()->getCommandID(databaseID, servantID, COMMAND_START);

			::core::Sender::waitForTermination(databaseID, commandID);

		} catch (const common::Exception & e) {
			common::out << "Commandline cannot find start command for " << controller->getName() << common::eol;
		}

	} catch (const common::Exception & e) {
		common::err << "Commandline cannot wait for " << controller->getName() << common::eol;
	}
}

void CommandLineManager::stopAndWaitForControllerCommand(ExperimentController * controller) {

	try {
		int32 databaseID = 0;
		int32 servantID = (int32) ::core::DataAccess::getInstance()->getServantID(controller->getName());

		try {
			int32 commandID = (int32) ::core::DataAccess::getInstance()->getCommandID(databaseID, servantID, COMMAND_START);

			::core::Sender::stopCommandAndWaitForTermination(databaseID, commandID);

		} catch (const common::Exception & e) {
			common::out << "Commandline cannot find start command for " << controller->getName() << common::eol;
		}

	} catch (const common::Exception & e) {
		common::err << "Commandline cannot wait for " << controller->getName() << common::eol;
	}
}

/*
 * computeEstimatedTime
 */
void CommandLineManager::computeEstimatedTimeControllers(const std::string& command, const std::vector<CommandLineCommand>& controllers, bool doprogression) {
	vector<float64> parrallelTimes;
	for(auto cont : controllers) {
		float64 localestime = cont.getController()->computeEstimatedTime(cont.getPropertiesMap());
		if (m_StartActivated && doprogression) {
			if (cont.getController()->commandStatus.isRunning()) {
				localestime *= (100 - cont.getController()->commandProgression()) / 100.;
			}
		}
		parrallelTimes.push_back(localestime);
	}
	float64 maxtime = -numeric_limits<float64>::max();
	for(auto partime : parrallelTimes) {
		if (partime > maxtime) {
			maxtime = partime;
		}
	}
	_estimatedTime += maxtime;
}

void CommandLineManager::computeEstimatedTime() {

	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::getInstance()->onEstimatedTimeComputingStatus(1);
	}
	Counter c1;
	Items estIts;
	CommandLineParser* clp = nullptr;
	{
		std::scoped_lock lock1(m_ItsMutex);
		estIts = m_Its.clone();
		clp = m_Parser->clone();
	}
	try {
		map<string, ExperimentController*> experimentControllersMap =
				InstrumentManager::getInstance()->getAllInstalledExperimentControllersMap(false);
		for (auto it : experimentControllersMap) {
			it.second->resetEstimatedTime();
		}
		_estimatedTime = 0;
		// start the execution
		clp->computeEstimatedTime(estIts, true);
	}
	catch (...) {
		_estimatedTime = -1.0;
	}
	delete clp;
	if (ics::ServerManager::getInstance() != 0) {
		ics::ServerManager::getInstance()->onEstimatedTimeChanged(_estimatedTime);
		ics::ServerManager::getInstance()->onEstimatedTimeComputingStatus(0);
	}

}

pair<unsigned long, double> CommandLineManager::getLastEstimatedTime(unsigned long databaseID) {
	return make_pair(0, _estimatedTime);
}


/*
 * setXBURunning
 */
void CommandLineManager::setXBURunning(const std::string& filename, bool value) {
	m_XBUActivated = value;
	if (value == false) {
		if (m_XbuCurrentLine.find(filename) != m_XbuCurrentLine.end()) {
			m_XbuCurrentLine.erase(filename);
		}
	}
	getSpyController()->updateValue(CommandLineManager::XBU_FULL_FILENAME_PROPERTY, filename);
}

/*
 * setXBUCurrentLine
 */
void CommandLineManager::setXBUCurrentLine(const std::string& filename, uint32 value) {
	m_XbuCurrentLine[filename] = value;
}

/*
 * getXBUCurrentLine
 */
uint32 CommandLineManager::getXBUCurrentLine(const std::string& filename) {
	if (m_XbuCurrentLine.find(filename) != m_XbuCurrentLine.end()) {
		return (m_XbuCurrentLine[filename] - 1);
	}
	return 0;
}

/*
 * setForloopRunning
 */
void CommandLineManager::setForloopRunning(const std::string& loopname, bool value) {
	m_ForloopActivated = value;
	if (value == false) {
		if (m_ForloopCurrentLine.find(loopname) != m_ForloopCurrentLine.end()) {
			m_ForloopCurrentLine.erase(loopname);
		}
	}
}

/*
 * setForloopCurrentPoint
 */
void CommandLineManager::setForloopCurrentPoint(const std::string& loopname, uint32 value) {
	m_ForloopCurrentLine[loopname] = value;
}

/*
 * getForloopCurrentPoint
 */
uint32 CommandLineManager::getForloopCurrentPoint(const std::string& loopname) {
	if (m_ForloopCurrentLine.find(loopname) != m_ForloopCurrentLine.end()) {
		return (m_ForloopCurrentLine[loopname] - 1);
	}
	return 0;
}

/*
 * setIfRunning
 */
void CommandLineManager::setIfRunning(const std::string& ifname, bool value) {
	m_IfActivated = value;
	if (value == false) {
		if (m_IfCurrentLine.find(ifname) != m_IfCurrentLine.end()) {
			m_IfCurrentLine.erase(ifname);
		}
	}
}

/*
 * setIfCurrentPoint
 */
void CommandLineManager::setIfCurrentPoint(const std::string& ifname, uint32 value) {
	m_IfCurrentLine[ifname] = value;
}

/*
 * getIfCurrentPoint
 */
uint32 CommandLineManager::getIfCurrentPoint(const std::string& ifname) {
	if (m_IfCurrentLine.find(ifname) != m_IfCurrentLine.end()) {
		return (m_IfCurrentLine[ifname]);
	}
	return 0;
}

