/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief CommandLine Manager (Singleton)
 * \author J. Locatelli
 * \date 12-01-2011
 */

#ifndef COMMANDLINEMANAGER_H
#define COMMANDLINEMANAGER_H

#include <string>
#include <map>
#include <mutex>
#include <boost/thread.hpp>
#include <common/base/BaseTypes.h>
#include "boost/filesystem/path.hpp"
#include "InstrumentAbstraction/ExperimentController.h"
#include "CommonCmdDef.h"
#include "CommandLineCommand.h"
#include "CommandSystem/CommandLine/Items.h"
#include <Utilities/Counter.h>

#include "DataProvider/Property/TypedProperty/Property.h"
#include "DataProvider/Subscriber.h"

class ChangeAspect;
class CommandLineParser;
class AbstractController;

class BadParser{};

/*!
 * \class CommandLineManager
 * \brief CommandLine Manager (Singleton)
 *
 * \par
 * This class is the manager of all command line executions. It create a new thread which executes the decode method.
 * \par
 * This class also memorize the last parser type which
 */
class CommandLineManager: public Subscriber {

public:

	/**
	 * Default name.
	 */
	static const std::string DEFAULT;

	/*!
	 * \enum Define the type of parser used.
	 */

	/*!
	 * \brief Destructor
	 */
	virtual ~CommandLineManager();

	/*!
	 * \brief Singleton accessor
	 *
	 * \return the pointer on the singleton
	 */
	static CommandLineManager* getInstance(const std::string& name = DEFAULT);

	/*!
	 * \brief Singleton reset (delete class)
	 */
	static void reset();

	/*!
	 * \brief Method called for executing a command
	 *
	 * \param[in] text the command to treat
	 */
	void executeCommand(const std::string& text, bool watchdog, bool sync = false);

	/*!
	 * \brief Method called for executing a command
	 *
	 * \param[in] text the command to treat
	 */
	void simulateCommand(const std::string& text, bool sync = false);

	/*!
	 * \brief Method called for pausing commands
	 */
	void pauseCommands();

	/*!
	 * \brief Method called for restarting commands
	 */
	void restartCommands();

	/*!
	 * \brief Method called for stopping commands
	 */
	void stopCommands();

	/*!
	 * \brief Method called for stopping at end commands
	 */
	void stopAtEndCommands();

	/*!
	 * \brief Send a output text to client subscriber
	 *
	 * \param[in] text the text to send
	 */
	void output(const std::string& text);

	/*!
	 * \brief Send a error text to client subscriber
	 *
	 * \param[in] text the text to send
	 */
	void outputError(int32  id, const std::string& text);

	/*!
	 * \brief Send a warning text to client subscriber
	 *
	 * \param[in] text the text to send
	 */
	void outputWarning(const std::string& text);

	/*!
	 * \brief Log debug stuff
	 *
	 * \param[in] text the text to be log
	 */
	void debug(const std::string& text);

	/*!
	 * \brief Log stuff
	 *
	 * \param[in] text the text to be log
	 */
	void log(const std::string& text);

	/*!
	 * \brief Send a synchro event to client subscriber
	 */
	void sendSynchroEvent();

	/*!
	 * \brief Execute a parallel start on the controller list
	 *
	 * \param[in] command The command to send to controller list
	 * \param[in] controllers The controller list to start
	 * \param[in] sync The flag for telling if it synchronized, waiting for the end of actions
	 */
	void startControllers(const std::string& command, const std::vector<CommandLineCommand>& controllers, bool sync);

	/*!
	 * \brief Execute a stop on the controller list
	 *
	 * \param[in] controllers The controller list to start
	 */
	void stopControllers(const std::vector<ExperimentController*>& controllers);

	/*!
	 * \brief Method for changing the type of parser
	 *
	 * \param[in] aParser The new parser
	 * \throws BadParser the parser chosen isn't existed
	 */
	void setParserType(const std::string& aParser) ;

	/*!
	 * \brief Get the actual parser name
	 *
	 * \return The actual parser name
	 * \throws BadParser the parser chosen isn't existed
	 */
	std::string getParserType();

	/*!
	 * \brief Method called for testing if a command is running.
	 *
	 * \return The flag value
	 */
	bool isStarted() const;

	/*!
	 * \brief Method called for testing if the command is paused.
	 *
	 * \return The flag value
	 */
	bool isPaused() const;

	/*!
	 * \brief Get the stop flag value
	 *
	 * \return The flag value
	 */
	bool isStopped() const;

	/*!
	 * \brief Get the xbu running flag value
	 *
	 * \return The flag value
	 */
	bool isXBURunning() const;

	/*!
	 * \brief Set the xbu running flag value
	 * \param[in] value The new parser
	 *
	 * \return The flag value
	 */
	void setXBURunning(const std::string& filename, bool value);

	/**
	 * \brief Set actual command value to the spy command line controller
	 *
	 * \param[in] text The command text
	 */
	void setSpyActualCommand(const std::string& text);

	/**
	 * \brief Set actual xbu values needed by the spy command line controller
	 *
	 * \param[in] filename The XBU filename
	 * \param[in] numberOfLines The total number of lines of the current XBU
	 * \param[in] currentLineNumber The current line of the current XBU
	 * \param[in] previousLineString The previous line command text
	 * \param[in] currentLineString The current line command text
	 * \param[in] nextLineString The next line command text
	 */
	void xbuSpy(const std::string& filename, uint32 numberOfLines, uint32 currentLineNumber,
			const std::string& previousLineString, const std::string& currentLineString, const std::string& nextLineString);

	/**
	 * \brief Set actual ForLoop values needed by the spy command line controller
	 *
	 * \param[in] numberOfSteps The total number of steps
	 * \param[in] currentStep The current step
	 */
	void forloopSpy(uint32 numberOfSteps, uint32 currentStep);

	/**
	 * \briefget spy controller
	 *
	 * \return The controller link
	 */
	AbstractController* getSpyController();

	/*!
	 * \brief Method called for setting if the state of command line
	 * Server need to know client state (command box or command line). The WebSpy client wiil use
	 * this info for printing or not the commande line spy
	 *
	 * \param[in] state The flag value
	 */
	void setCommandLineState(bool state);

	/*!
	 * \brief Function waiting for the termination of the start command of the controller using the core::Sender class.
	 *
	 * \param[in] controller the controller
	 * \param[in] command the name of the command
	 */
	static void waitForControllerCommand(ExperimentController * controller);

	/*!
	 * \brief Function stopping the start command of the controller using the core::Sender class and waiting for its termination.
	 *
	 * \param[in] controller the controller
	 * \param[in] command the name of the command
	 */
	static void stopAndWaitForControllerCommand(ExperimentController * controller);

	void computeEstimatedTimeControllers(const std::string& command, const std::vector<CommandLineCommand>& controllers, bool doprogression);
	void computeEstimatedTime();
	std::pair<unsigned long, double> getLastEstimatedTime(unsigned long databaseID);

	void setXBUCurrentLine(const std::string& filename, uint32 value);
	uint32 getXBUCurrentLine(const std::string& filename);


	void setForloopRunning(const std::string& loopname, bool value);
	void setForloopCurrentPoint(const std::string& loopname, uint32 value);
	uint32 getForloopCurrentPoint(const std::string& loopname);

	void setIfRunning(const std::string& ifname, bool value);
	void setIfCurrentPoint(const std::string& ifname, uint32 value);
	uint32 getIfCurrentPoint(const std::string& ifname);

	boost::filesystem::path m_XBUPATH;
	boost::filesystem::path m_PALPATH;
	boost::filesystem::path m_SETTINGSPATH;

protected:

	static const std::string COUNT_REPEATER_TYPE;

	/**
	 * \brief Method used by the publishers to notify changes
	 *
	 * \param[in] changeAspect The change aspect to treat
	 */
	virtual void update(ChangeAspect* changeAspect);

private:

	/*!
	 * \brief Singleton constructor
	 *
	 * \param[in] The proxy for the command line
	 */
	CommandLineManager();

	/*!
	 * \brief Function starting the controller command using the core::Sender class.
	 *
	 * \param[in] controller the controller
	 * \param[in] command the name of the command
	 */
	static void startControllerCommand(ExperimentController * controller, const std::string& command, bool blocking);

	std::shared_ptr<CommandLineParser> m_Parser;
	AbstractController* m_SpyController;
	std::recursive_mutex m_ItsMutex;
	std::recursive_mutex m_StartedControllerMutex; //! The mutex used for protecting m_StartedController linst
	std::map<std::string, CommandLineCommand> m_StartedController; //! The started controller list (controller is running list)
	Items m_Its;
	bool m_CommandLineActivated; //! The command line is activated in the client
	bool m_StartActivated; //! The start is activated
	bool m_PauseActivated; //! The pause is activated
	bool m_StopActivated; //! The stop is activated
	bool m_XBUActivated; //! The Command line running a xbu file
	bool m_ForloopActivated; //! The Command line running a xbu file
	bool m_IfActivated; //! The Command line running a xbu file
	bool m_SimulatedCommand; //! The flag said this command is just for testing

	static std::map<std::string, CommandLineManager*> m_Instances; //! singleton instance
	static std::string m_ParserType; //! Last Command Parser Type

	float64 _estimatedTime;
	Counter _estimatedTimeRefreshCounter;

	std::map<std::string, uint32> m_XbuCurrentLine; //! The started controller list (controller is running list)
	std::map<std::string, uint32> m_ForloopCurrentLine; //! The started controller list (controller is running list)
	std::map<std::string, uint32> m_IfCurrentLine; //! The started controller list (controller is running list)

public:

	static const std::string COMMANDLINE_STATE_PROPERTY;
	static const std::string ACTUAL_COMMAND_PROPERTY;
	static const std::string LAST_COMMAND_PROPERTY;
	static const std::string XBU_RUNNING_PROPERTY;
	static const std::string XBU_FILENAME_PROPERTY;
	static const std::string XBU_FULL_FILENAME_PROPERTY;
	static const std::string XBU_NUMBER_LINES_PROPERTY;
	static const std::string XBU_CURRENT_LINE_PROPERTY;
	static const std::string XBU_PREVIOUS_COMMAND_PROPERTY;
	static const std::string XBU_CURRENT_COMMAND_PROPERTY;
	static const std::string XBU_NEXT_COMMAND_PROPERTY;
	static const std::string FORLOOP_RUNNING_PROPERTY;
	static const std::string FORLOOP_NUMBER_STEPS_PROPERTY;
	static const std::string FORLOOP_CURRENT_STEP_PROPERTY;
	static const std::string PAL_CONTENTS_PROPERTY;
	static const std::string PAL_FULL_FILENAME_PROPERTY;

};

/*
 * isStarted
 */
inline bool CommandLineManager::isStarted() const {
	return m_StartActivated;
}

/*
 * isPaused
 */
inline bool CommandLineManager::isPaused() const {
	return m_PauseActivated;
}

/*
 * isStopped
 */
inline bool CommandLineManager::isStopped() const {
	return m_StopActivated;
}

/*
 * isXBURunning
 */
inline bool CommandLineManager::isXBURunning() const {
	return m_XBUActivated;
}

/*
 * getSpyController
 */
inline AbstractController* CommandLineManager::getSpyController() {
	return m_SpyController;
}

#endif //COMMANDLINEMANAGER_H
