/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief CommandLine Parser
 * \author J. Locatelli
 * \date 12-01-2011
 */

#include "CommandSystem/CommandLine/CommandLineParser.h"
#include "CommandSystem/CommandLine/CommandLineManager.h"
#include "CommandSystem/CommandLine/CommonCmdDef.h"
#include "CommandSystem/CommandLine/Parsers/AbstractParser.h"
#include "Module/ModuleManager.h"
#include <ics/IcsServerManager.h>

#include <iostream>

using namespace std;

/*
 * Constructor
 */
CommandLineParser::CommandLineParser() : m_CurrentParser(nullptr) {
}

/*
 * Destructor
 */
CommandLineParser::~CommandLineParser() {
	delete m_CurrentParser;
}

CommandLineParser* CommandLineParser::clone() {
	initParser();
	CommandLineParser* clp = new CommandLineParser();
	clp->m_CurrentParser = m_CurrentParser->clone();
	clp->m_ParserType = m_ParserType;
	return clp;
}


void CommandLineParser::setParser(const std::string& parserType)  {
	m_ParserType = parserType;
}

void CommandLineParser::initParser() {

	if (m_CurrentParser == nullptr) {
		// Try first last parser
		try {
			m_CurrentParser = ModuleManager::getInstance()->makeParser(m_ParserType);
		} catch (ModuleManager::NoSuchParserClass const &) {
			m_CurrentParser = nullptr;
		}
		if (m_CurrentParser == nullptr) {
			m_CurrentParser = ModuleManager::getInstance()->makeParser(CommonCmdDef::GENERIC_PARSER);
			throw BadParser();
		}
	}
}

/*
 * executeCommand
 */
void CommandLineParser::executeCommand(const Items& items) {

	initParser();

	// Give back hand to client if synchronization isn't expected
	if (m_CurrentParser->isSynchronized() == false) {
		CommandLineManager::getInstance()->sendSynchroEvent();
	}
	try {
		parse(items);
		if (ics::ServerManager::getInstance() != 0) {
			ics::ServerManager::getInstance()->onEstimatedTimeChanged(0);
		}
	}
	catch(CommonCmdDef::BadInput& e) {
		// Error , bad input, no parser could parse this command
		CommandLineManager::getInstance()->outputError(COMMAND_LINE_BAD_COMMAND, ": " + e.m_Input);
	}

	// Give back hand to client if synchronization is expected
	if (m_CurrentParser->isSynchronized() == true) {
		CommandLineManager::getInstance()->sendSynchroEvent();
	}
}

/*
 * parse
 */
void CommandLineParser::parse(Items items)  {

	// init the parser, get current state of instrument configuration
	m_CurrentParser->init();

	// Try with the current used parser
	bool found = m_CurrentParser->decode(items, true);
	if (found == true) {
		goto fin;
	}

	fin: if (found == false) {
		// Nothing found so error
		throw CommonCmdDef::BadInput();
	}
}

/*
 * computeEstimatedTime
 */
void CommandLineParser::computeEstimatedTime(Items items, bool doprogression) {

	m_CurrentParser->init();

	try {
		m_CurrentParser->decode(items, true);
	}
	catch (exception& e) {
		common::err << "computeEstimatedTime exception : " << e.what() << common::eol;
	}

}


