/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief Items class
 * \author J. Locatelli
 * \date 02-02-2011
 */

#include "CommandSystem/CommandLine/Items.h"
#include "Utilities/StringUtilities.h"
#include <iostream>
#include <boost/tokenizer.hpp>
#include <boost/algorithm/string.hpp>

using namespace std;
using namespace boost;

const char Items::COMMA1[] = "'";
const char Items::COMMA2[] = "`";
const char Items::COMMAS[] = "\"";
const char Items::IN_PARENTHESE[] = "(";
const char Items::OUT_PARENTHESE[] = ")";
const char Items::IN_BRACE[] = "{";
const char Items::OUT_BRACE[] = "}";
const char Items::IN_BRAKET[] = "[";
const char Items::OUT_BRAKET[] = "]";

/*
 * constructor
 */
Items::Items() {
	// empty
}

/*
 * constructor
 */
Items::Items(const std::string& text) {

	// Add space between separator
	string newtext = text;
	replace_all(newtext,";"," ; ");

	// Slipt test with tokenizer
	typedef tokenizer<char_separator<char> > tokenizer;
	char_separator<char> sep(" \t");
	tokenizer tok1(newtext, sep);
	tokenizer::iterator iter = tok1.begin();

	string buf; // Storage of an expression, which contains space
	bool doconcat = false; // Flag concat tokenizer on or off

	while (iter != tok1.end()) {
		if (doconcat == false) {
			// concat is off
			// Test to find a container character
			if ((StringUtilities::cmp(*iter, COMMA1, 1) == true) || (StringUtilities::cmp(*iter, COMMA2, 1) == true)
					|| (StringUtilities::cmp(*iter, COMMAS, 1) == true) || (StringUtilities::cmp(*iter, IN_BRACE, 1) == true)
					|| (StringUtilities::cmp(*iter, IN_BRAKET, 1) == true) || (StringUtilities::cmp(*iter, IN_PARENTHESE, 1)
					== true)) {
				// Is there end of expression character
				if (((*iter).find(COMMA1) == ((*iter).size() - 1)) || ((*iter).find(COMMA2) == ((*iter).size() - 1))
						|| ((*iter).find(COMMAS) == ((*iter).size() - 1)) || ((*iter).find(OUT_BRACE) == ((*iter).size() - 1))
						|| ((*iter).find(OUT_BRAKET) == ((*iter).size() - 1)) || ((*iter).find(OUT_PARENTHESE) == ((*iter).size()
						- 1))) {
					// Yes, so save item with containers characters
					push_back((*iter).substr(0, (*iter).size()));
				}
				else {
					// Start to concat tokenizer, remove container character at start
					doconcat = true;
					buf = (*iter).substr(1);
				}
			} else {
				// Normal input, save item
				push_back(*iter);
			}
		} else {
			// Concat is on
			// Test if concat has to be finished
			if (((*iter).find(COMMA1) == ((*iter).size() - 1)) || ((*iter).find(COMMA2) == ((*iter).size() - 1))
					|| ((*iter).find(COMMAS) == ((*iter).size() - 1)) || ((*iter).find(OUT_BRACE) == ((*iter).size() - 1))
					|| ((*iter).find(OUT_BRAKET) == ((*iter).size() - 1)) || ((*iter).find(OUT_PARENTHESE) == ((*iter).size()
					- 1))) {
				// Yes, stop concat
				doconcat = false;
				// Save to item without last character which is the container character
				buf.append(" " + (*iter).substr(0, (*iter).size() - 1));
				push_back(buf);
			} else {
				// Concat tokenizer to temporary buffer
				buf.append(" " + *iter);
			}
		}
		++iter;
	}
}

/*
 * Copy constructor
 */
Items::Items(const Items& items) {
	// Copy items
	vector<string>::const_iterator iter1 = items.m_Items.begin();
	while (iter1 != items.m_Items.end()) {
		m_Items.push_back(*iter1);
		++iter1;
	}
	// Copy Types
	vector<ItemType>::const_iterator iter2 = items.m_Types.begin();
	while (iter2 != items.m_Types.end()) {
		m_Types.push_back(*iter2);
		++iter2;
	}
	// Reset iterator
	m_ItemsIterator = m_Items.begin();
	m_TypesIterator = m_Types.begin();
	// Copy command string
	m_String = items.m_String;
}

/*
 * Copy constructor
 */
Items Items::clone() {
	Items clone;
	// Copy items
	vector<string>::const_iterator iter1 = m_Items.begin();
	while (iter1 != m_Items.end()) {
		clone.m_Items.push_back(*iter1);
		++iter1;
	}
	// Copy Types
	vector<ItemType>::const_iterator iter2 = m_Types.begin();
	while (iter2 != m_Types.end()) {
		clone.m_Types.push_back(*iter2);
		++iter2;
	}
	// Reset iterator
	clone.m_ItemsIterator = m_Items.begin();
	clone.m_TypesIterator = m_Types.begin();
	// Copy command string
	clone.m_String = m_String;
	return clone;
}

/*
 * push_back
 */
void Items::push_back(const std::string& value) {
	// Add item to items vector
	m_Items.push_back(value);

	ItemType atype = STRING;

	// test if value is a numeric type
	int32 ival = StringUtilities::convert<int32>(value);
	if (ival != numeric_limits<int32>::max()) {
		atype = VALUE;
	} else {
		float64 fval = StringUtilities::convert<float64>(value);
		if (fval != numeric_limits<float64>::max()) {
			atype = VALUE;
		}
	}

	// Add item type to items type vector
	m_Types.push_back(atype);

	// Reset items iterator
	m_ItemsIterator = m_Items.begin();
	m_TypesIterator = m_Types.begin();

	// Append to command string
	if (m_String.empty() == true) {
		m_String = value;
	} else {
		m_String.append(" " + value);
	}

}

/*
 * operator <<
 */
std::ostream& operator <<(std::ostream& os, const Items& its) {
	Items litems(its);
	uint32 i = 0;
	while (litems.end() == false) {
		os << "Item " << i << " : " << *litems << " (";
		switch (litems.getType()) {
		case Items::VALUE:
			os << "Value";
			break;
		case Items::STRING:
			os << "String";
			break;
		case Items::OTHER:
			os << "Other";
			break;
		}
		os << ")" << endl;
		++litems;
		++i;
	}
	return (os);
}
