/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief Items class
 * \author J. Locatelli
 * \date 02-02-2011
 */

#ifndef ITEMS_H
#define ITEMS_H

#include <vector>
#include <string>
#include <common/base/BaseTypes.h>

/*!
 * \class Items
 * \brief Contains command line items information
 *
 * The command line string is split into a vector of string. Each string got is type String or value.
 */
class Items {
public:

	/*!
	 * \brief constructor
	 *
	 * \param[in] text The text to decode to items
	 */
	Items();

	/*!
	 * \brief constructor
	 *
	 * \param[in] text The text to decode to items
	 */
	Items(const std::string& text);

	/*!
	 * \brief copy constructor
	 *
	 * \param[in] items Object to copy
	 */
	Items(const Items& items);

	Items clone();

	/*!
	 * \enum Type of a item
	 */
	enum ItemType {
		VALUE, //! The item is a value
		STRING, //! The item is a string
		OTHER //! No type found
	};

	/*!
	 * \brief push back a string into the list of items
	 *
	 * \param[in] value The string to push back
	 */
	void push_back(const std::string& value);

	/*!
	 * \brief Test if the iterator of items list is at end position
	 *
	 * \return If iterator is at and or not
	 */
	bool end() const;

	/*!
	 * \brief Test if next item is a numeric value
	 *
	 * \return The result of the test
	 */
	bool nextIsNumeric() const;

	/*!
	 * \brief Test if the iterator of items list is at end position
	 *
	 * \return If iterator is at and or not
	 */
	uint32 getSize() const;

	/*!
	 * \brief Test if  item  is inside to int32 value
	 *
	 */
	template<class T> bool outside(T down , T up) const{
		if (m_TypesIterator == m_Types.end()) {
			return true;
		}
		else if (*m_TypesIterator != VALUE) {
			return true;
		}
		else {
			std::istringstream sbuf(*m_ItemsIterator);
			T value;
			sbuf >> value;
			if ((value < down) || (value > up)) {
				return true;
			} else {
				return false;
			}
		}
	}

	/*!
	 * \brief Get type of the current item
	 *
	 * \return The type
	 */
	ItemType getType() const;

	/*!
	 * \brief Get string of concat items. Rebuild command text
	 *
	 * \return The text
	 */
	const std::string& getString() const;

	/*!
	 * \brief Operator * overload
	 *
	 * \param[in] its The items object reference for getting actual item
	 * \return The item points by iterator
	 */
	friend std::string operator * (const Items& its);

	/*!
	 * \brief Operator ++ overload
	 *
	 * \param[in/out] its The items object reference for incrementing items iterators
	 */
	friend void operator ++ (Items& its);

	/*!
	 * \brief Operator -- overload
	 *
	 * \param[in/out] its The items object reference for decrementing items iterators
	 */
	friend void operator -- (Items& its);

	/*!
	 * \brief Operator <<+ overload
	 *
	 * \param[in/out] s The output stream where printing items
	 * \param[in] its The items to print
	 * \return The output stream where the items was print
	 */
	friend std::ostream& operator << (std::ostream& s,const Items& its);

private:

	std::string m_String; //! Local storage of command text, concat of items
	std::vector<std::string> m_Items; //! List of item
	std::vector<ItemType> m_Types; //! List of item's type
	std::vector<std::string>::iterator m_ItemsIterator; //! Iterator on item's list
	std::vector<ItemType>::iterator m_TypesIterator; //! Iterator on item's type list

	/*!
	 * List of expression container characters
	 */
	static const char COMMA1[];
	static const char COMMA2[];
	static const char COMMAS[];
	static const char IN_PARENTHESE[];
	static const char OUT_PARENTHESE[];
	static const char IN_BRACE[];
	static const char OUT_BRACE[];
	static const char IN_BRAKET[];
	static const char OUT_BRAKET[];

};

/*
 * operator *
 */
inline std::string operator * (const Items& its) {
	return *(its.m_ItemsIterator);
}

/*
 * operator ++
 */
inline void operator ++ (Items& its) {
	if (its.m_ItemsIterator != its.m_Items.end()) {
		++its.m_ItemsIterator;
	}
	if (its.m_TypesIterator != its.m_Types.end()) {
		++its.m_TypesIterator;
	}
}

/*
 * operator --
 */
inline void operator -- (Items& its) {
	if (its.m_ItemsIterator != its.m_Items.begin()) {
		--its.m_ItemsIterator;
	}
	if (its.m_TypesIterator != its.m_Types.begin()) {
		--its.m_TypesIterator;
	}
}

/*
 * end
 */
inline bool Items::end() const {
	return m_ItemsIterator == m_Items.end();
}

/*
 * getSize
 */
inline uint32 Items::getSize() const {
	return (uint32) m_Items.size();
}

/*
 * getType
 */
inline Items::ItemType Items::getType() const {
	if (m_TypesIterator != m_Types.end()) {
		return *m_TypesIterator;
	}
	return OTHER;
}


/*
 * getString
 */
inline const std::string& Items::getString() const {
	return m_String;
}

/*
 * nextIsNumeric
 */
inline bool Items::nextIsNumeric() const {
	bool ok = false;
	if (m_TypesIterator != m_Types.end()) {
		std::vector<ItemType>::iterator iter = m_TypesIterator;
		++iter;
		if ((iter) != m_Types.end()) {
			ok = (*iter == VALUE);
		}
	}
	return ok;
}

#endif
