/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief Abstract class describes the parser
 * \author J. Locatelli
 * \date 12-01-2010
 */

#include "CommandSystem/CommandLine/Parsers/AbstractParser.h"
#include "CommandSystem/CommandLine/Parsers/Common/CdCmd.h"
#include "CommandSystem/CommandLine/Parsers/Common/ParserCmd.h"
#include "CommandSystem/CommandLine/Parsers/Common/PwdCmd.h"
#include "CommandSystem/CommandLine/Parsers/Common/XBUCmd.h"
#include "CommandSystem/CommandLine/Parsers/Common/ShowCmd.h"
#include "CommandSystem/CommandLine/Parsers/Common/RefreshCmd.h"
#include "CommandSystem/CommandLine/Parsers/Common/WaitCmd.h"
#include "InstrumentManager/InstrumentManager.h"
#include <dataprovider/database/Database.h>
#include "DataProvider/Units.h"
#include <common/base/FileUtilities.h>
#include "CommandSystem/CommandLine/CommandLineManager.h"

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/format.hpp>
#include <algorithm>

using namespace std;
using namespace boost;
using namespace boost::posix_time;

/*
 * Constructor
 */
AbstractParser::AbstractParser() :
	m_output(new ostringstream()), m_xbuLevel(0), m_nbErrors(0), m_forEstimatedTime(false) {
	// Empty
}

/*
 * Destructor
 */
AbstractParser::~AbstractParser() {
	if (m_output != NULL) {
		delete m_output;
	}
}

void AbstractParser::copy(const AbstractParser* parser) {
	m_forEstimatedTime = true;
}


/*
 * decode
 */
bool AbstractParser::decode(Items items, bool doestimatedprogression)  {
	bool found = true;

	/*
	 * First item is the command
	 */
	if (StringUtilities::cmp(*items, CommonCmdDef::PARSER_CMD, 6) == true) {
		// Wait command
		ParserCmd cmd(this);
		cmd.parse(items, doestimatedprogression);
	} else if ((StringUtilities::cmp(*items, CommonCmdDef::DO_CMD, 2) == true) || (StringUtilities::cmp(*items,
			CommonCmdDef::XBU_CMD, 3) == true)) {
		if (Simulated::activated == true) {
			if (m_xbuLevel == 0) {
				m_nbErrors = 0;
			}
			++m_xbuLevel;
		}
		// Clear command
		XBUCmd cmd(this);
		cmd.parse(items, doestimatedprogression);
		if (Simulated::activated == true) {
			--m_xbuLevel;
			if (m_xbuLevel == 0) {
//				time_duration td = seconds((int64) Simulated::estimatedTime);
//				ostringstream sbuf1;
//				if (Simulated::estimatedTime < 0.0001) {
//					sbuf1 << "Could not Estimate Acquisition Time";
//					CommandLineManager::getInstance()->outputWarning(sbuf1.str());
//				} else {
//					sbuf1 << "Estimated Acquisition Time = " << to_simple_string(td);
//					CommandLineManager::getInstance()->outputWarning(sbuf1.str());
//				}
				ostringstream sbuf2;
				sbuf2 << "Number of Errors = " << m_nbErrors;
				CommandLineManager::getInstance()->outputWarning(sbuf2.str());
			}
		}
	} else if (StringUtilities::cmp(*items, CommonCmdDef::SHOW_CMD, 4) == true) {
		// Show command
		ShowCmd cmd(this);
		cmd.parse(items, doestimatedprogression);
	} else if (StringUtilities::cmp(*items, CommonCmdDef::PWD_CMD, 3) == true) {
		// Pwd command
		PwdCmd cmd(this);
		cmd.parse(items, doestimatedprogression);
	} else if (StringUtilities::cmp(*items, CommonCmdDef::CD_CMD, 2) == true) {
		// Cd command
		CdCmd cmd(this);
		cmd.parse(items, doestimatedprogression);
	} else if (StringUtilities::cmp(*items, CommonCmdDef::WAIT_CMD, 4) == true) {
		// Wait command
		WaitCmd cmd(this);
		cmd.parse(items, doestimatedprogression);
	} else if ((StringUtilities::cmp(*items, CommonCmdDef::REFRESH_CMD, 7) == true) && ((Simulated::activated == true) || (Simulated::visa == true))) {
		// Refresh command
		RefreshCmd cmd(this);
		cmd.parse(items, doestimatedprogression);
	} else if ((StringUtilities::cmp(*items, CommonCmdDef::COMMENT_CMD1, 1) || StringUtilities::cmp(*items,
			CommonCmdDef::COMMENT_CMD2, 1) == true)) {
		// Log text
		if ((*items).size() > 1) {
			*m_output << "Comment: " << (*items).substr(1) << " ";
		} else {
			*m_output << "Comment: ";
		}
		++items;
		while (items.end() == false) {
			*m_output << *items;
			++items;
			if (items.end() == false) {
				*m_output << " ";
			}
		}
		outputEndl();
	} else if (StringUtilities::cmp(*items, CommonCmdDef::STOP_CMD, 4) == true) {
		// Stop command
		CommandLineManager::getInstance()->stopCommands();
	} else if (StringUtilities::cmp(*items, CommonCmdDef::PAUSE_CMD, 4) == true) {
		// Pause command
		CommandLineManager::getInstance()->pauseCommands();
	} else if (StringUtilities::cmp(*items, CommonCmdDef::CONTINU_CMD, 4) == true) {
		// Continu command
		CommandLineManager::getInstance()->restartCommands();
	} else {
		found = false;
	}
	if (m_forEstimatedTime == false) {
		// TO DO remove possibility of many commands on the same line?
		++items;
		if (items.end() == false) {
			if (StringUtilities::cmp(*items, CommonCmdDef::SEPARATOR)) {
				string buf;
				++items;
				while (items.end() == false) {
					buf.append(*items);
					buf.append(" ");
					++items;
				}

				CommandLineManager::getInstance()->executeCommand(buf, false ,true);
			}
		}
	}
	return found;
}

/*
 * init
 */
void AbstractParser::init() {
	m_controllers = InstrumentManager::getInstance()->getInstalledExperimentControllersMap();
	m_nicknameControllers = InstrumentManager::getInstance()->getInstalledExperimentControllersNicknameMap();
	m_generics = InstrumentManager::getInstance()->getGenericPropertiesList();
}

/*
 * isSynchronized
 */
bool AbstractParser::isSynchronized() const {
	if (m_forEstimatedTime) {
		return true;
	}
	else {
		// Check XBU running
		return (CommandLineManager::getInstance()->isXBURunning() || isParserSynchronized());
	}
}

/*
 * isController
 */
bool AbstractParser::isController(const std::string& item, ExperimentController** anEC, uint32 size) const {

	map<string, ExperimentController*>::const_iterator controllerIter;

	int32 nbfound = 0;

	/*
	 * Check for nickname controllers
	 */
	controllerIter = m_nicknameControllers.begin();
	while (controllerIter != m_nicknameControllers.end()) {
		if (StringUtilities::cmp(item, controllerIter->first) == true) {
			// Found the controller
			if (controllerIter->second->getName() != controllerIter->first) {
				*anEC = controllerIter->second;
				++nbfound;
			}
		}
		++controllerIter;
	}

	if (nbfound == 0) {
		/*
		 * Check controllers
		 */
		controllerIter = m_controllers.begin();
		while (controllerIter != m_controllers.end()) {
			if (StringUtilities::cmp(item, controllerIter->first, size) == true) {
				// Found the controller
				*anEC = controllerIter->second;
				++nbfound;
			}
			++controllerIter;
		}
	}
	if (nbfound > 1) {
		return isController(item, anEC);
	}
	return (nbfound == 1);
}

/*
 * isControllerWithProperty
 */
bool AbstractParser::isControllerWithProperty(const std::string& item, ExperimentController** anEC, const std::string& property, uint32 size) const {

	map<string, ExperimentController*>::const_iterator controllerIter = m_controllers.begin();
	string tmp;

	int32 nbfound = 0;

	/*
	 * Check controllers
	 */
	while (controllerIter != m_controllers.end()) {
		if (StringUtilities::cmp(item, controllerIter->first, size) == true) {
			// Found the controller
			ExperimentController* cont = controllerIter->second;
			if (isProperty(property, cont, tmp) == true) {
				*anEC = cont;
				++nbfound;
			}
		}
		++controllerIter;
	}

	/*
	 * Check for nickname controllers
	 */
	controllerIter = m_nicknameControllers.begin();
	while (controllerIter != m_nicknameControllers.end()) {
		if (StringUtilities::cmp(item, controllerIter->first) == true) {
			// Found the controller
			ExperimentController* cont = controllerIter->second;
			if ((isProperty(property, cont, tmp) == true) && (cont->getName() != controllerIter->first)) {
				*anEC = cont;
				++nbfound;
			}
		}
		++controllerIter;
	}
	if (nbfound > 1) {
		return isControllerWithProperty(item, anEC, property);
	}
	return (nbfound == 1);
}

/*
 * isGenericProperty
 */
bool AbstractParser::isGenericProperty(const std::string& item, ExperimentController** anEC, string& propertyName) const {

	vector<GenericPropertyDescription>::const_iterator iter = m_generics.begin();
	int32 nbfound = 0;

	while (iter != m_generics.end()) {
		string alias = (*iter).getAlias();
		if (StringUtilities::cmp(item, alias) == true) {
			*anEC = (*iter).getController();
			propertyName = (*iter).getPropertyName();
			++nbfound;
			break;
		}
		++iter;
	}
	return (nbfound == 1);
}

/*
 * isProperty
 */
bool AbstractParser::isProperty(const std::string& item, ExperimentController* anEC, string& propertyName) const {

	vector<AnyProperty*> properties = anEC->getAllProperties();
	vector<AnyProperty*>::iterator iter = properties.begin();
	bool found = false;

	while (iter != properties.end()) {
		string name = (*iter)->getName();
		if (StringUtilities::cmp(item, name) == true) {
			propertyName = name;
			found = true;
			break;
		}
		++iter;
	}
	return found;
}

/*
 * getGenericPropertyName
 */
string AbstractParser::getGenericPropertyName(ExperimentController* controller, const string& inpname, const string& reference,
		const string& newref) const  {
	string pname;
	map<string, string> buf = controller->getAliasPropertyMap();
	if (buf.size() > 0) {
		map<string, string>::iterator ibuf = buf.begin();
		while (ibuf != buf.end()) {
			if ((StringUtilities::cmp(ibuf->first, inpname) == true) || (inpname.empty() == true)) {
				pname = ibuf->second;
				if (ibuf->second.find(reference) != string::npos) {
					replace_first(pname, reference, newref);
					break;
				}
			}
			++ibuf;
		}
		if (ibuf == buf.end()) {
			// Return the input name as it seems that it is not an alias.
			return pname;
		}
	}
	else {
		throw CommonCmdDef::BadInput("No generic property");
	}
	return pname;
}

/*
 * getControllerByType
 */
ExperimentController* AbstractParser::getControllerByType(const std::string& aType) {
	ExperimentController * ec = InstrumentManager::getInstance()->getFirstExperimentControllerByType(aType, false);
	if (ec == NULL) {
		CommandLineManager::getInstance()->debug(aType + " doesn''t existed");
	}
	return ec;
}

/*
 * getControllerByFunction
 */
ExperimentController* AbstractParser::getControllerByFunction(const std::string& aFunction)
		 {
	ExperimentController * ec = InstrumentManager::getInstance()->getFirstExperimentControllerByFunction(aFunction);
	if (ec == NULL) {
		CommandLineManager::getInstance()->debug(aFunction + " doesn''t existed");
		throw CommonCmdDef::BadInput("Controller not found with function: " + aFunction);
	}
	return ec;
}

/*
 * getControllerByName
 */
ExperimentController* AbstractParser::getControllerByName(const std::string& aName)  {
	// Convert all name to lower case
	string name = aName;
	to_lower(name);
	string buf;
	map<string, ExperimentController*>::const_iterator controllerIter = m_controllers.begin();
	/*
	 * Check controllers
	 */
	while (controllerIter != m_controllers.end()) {
		// Found the controller
		buf = controllerIter->first;
		to_lower(buf);
		if (buf == name) {
			return controllerIter->second;
		}
		++controllerIter;
	}
	CommandLineManager::getInstance()->debug(aName + " doesn''t existed");
	throw CommonCmdDef::BadInput("Controller not found: " + aName);
}

/*
 * getPropertyValue
 */
boost::any AbstractParser::getPropertyValue(ExperimentController* controller, const std::string& propertyName)
		 {
	try {
		return controller->getValue(propertyName);
	} catch (...) {
		CommandLineManager::getInstance()->debug(controller->getName() + " " + propertyName + " doesn''t existed");
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + propertyName);
	}

}

/*
 * getPropertyValueAsString
 */
string AbstractParser::getPropertyValueAsString(ExperimentController* controller, const std::string& propertyName)
		 {
	ostringstream buf;
	try {
		if (propertyName.empty() == false) {
			any value = controller->getValue(propertyName);
			string ptype = controller->getPropertyType(propertyName);
			if (ptype == UINT16_PROPERTY) {
				buf << any_cast<uint16> (value);
			} else if (ptype == UINT32_PROPERTY) {
				buf << any_cast<uint32> (value);
			} else if (ptype == UINT64_PROPERTY) {
				buf << any_cast<uint64> (value);
			} else if (ptype == INT16_PROPERTY) {
				buf << any_cast<int16> (value);
			} else if (ptype == INT32_PROPERTY) {
				buf << any_cast<int32> (value);
			} else if (ptype == INT64_PROPERTY) {
				buf << any_cast<int64> (value);
			} else if (ptype == FLOAT32_PROPERTY) {
				buf << any_cast<float32> (value);
			} else if (ptype == FLOAT64_PROPERTY) {
				buf << any_cast<float64> (value);
			} else if (ptype == BOOLEAN_PROPERTY) {
				if (any_cast<bool> (value) == true) {
					buf << "1";
				} else {
					buf << "0";
				}
			} else if (ptype == STRING_PROPERTY) {
				buf << any_cast<string> (value);
			} else {
				CommandLineManager::getInstance()->debug(
						controller->getName() + " property " + propertyName + " " + ptype + " doesn''t implemented");
				// Print Error
				CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + controller->getName());
			}
		}
	} catch (...) {
		CommandLineManager::getInstance()->debug(controller->getName() + " " + propertyName + " doesn''t existed");
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + propertyName);
	}
	return buf.str();
}

/*
 * setProperty
 */
void AbstractParser::setProperty(ExperimentController* controller, const string& name, const std::string& propertyName,
		const std::string& value, const std::string& prefix)  {
	if (m_forEstimatedTime) return;

	controller->attach(CommandLineManager::getInstance());
	string lvalue = value;
	try {

		if (propertyName.empty() == false) {
			try {
				uint32 servantID = dataprovider::Database::getInstance(0)->getServantID(controller->getName());
				uint32 propertyID = dataprovider::Database::getInstance(0)->getPropertyID(servantID, propertyName);
				uint32 enumlabelsID = dataprovider::Database::getInstance(0)->getEnumeratedLabelsProperty(propertyID);
				if (enumlabelsID != 0) {
					uint32 enumvaluesID = dataprovider::Database::getInstance(0)->getEnumeratedValuesProperty(propertyID);
					// Get the dynamic property size.
					uint32 enumeratedLabelsSize = dataprovider::Database::getInstance(0)->getDynamicPropertySize(enumlabelsID);
					// Search for the label.
					for (int i = 0; i < enumeratedLabelsSize; ++i) {
						uint32 labelPropertyID = dataprovider::Database::getInstance(0)->getDynamicPropertyIDAtIndex(enumlabelsID, i, false);
						std::string label = dataprovider::Database::getInstance(0)->getValue<string>(labelPropertyID);

						// Check the label.
						if (label == lvalue) {
							uint32 valuePropertyID = dataprovider::Database::getInstance(0)->getDynamicPropertyIDAtIndex(enumvaluesID, i, false);
							lvalue = dataprovider::Database::getInstance(0)->getFormattedValue(valuePropertyID);
							break;
						}
					}
				}
			} catch (...) {
			}
			string ptype = controller->getPropertyType(propertyName);
			if (ptype == UINT16_PROPERTY) {
				// Test conversion
				if (lvalue != "=") {
					uint16 val = StringUtilities::convert<uint16>(lvalue);
					if (val == numeric_limits<uint16>::max()) {
						throw CommonCmdDef::BadInput(value);
					}
					controller->setValue(propertyName, val);
				}
			} else if (ptype == UINT32_PROPERTY) {
				// Test conversion
				if (lvalue != "=") {
					uint32 val = StringUtilities::convert<uint32>(lvalue);
					if (val == numeric_limits<uint32>::max()) {
						throw CommonCmdDef::BadInput(value);
					}
					controller->setValue(propertyName, val);
				}
			} else if (ptype == UINT64_PROPERTY) {
				// Test conversion
				if (lvalue != "=") {
					uint64 val = StringUtilities::convert<uint64>(lvalue);
					if (val == numeric_limits<uint64>::max()) {
						throw CommonCmdDef::BadInput(value);
					}
					controller->setValue(propertyName, val);
				}
			} else if (ptype == INT16_PROPERTY) {
				// Test conversion
				if (lvalue != "=") {
					int16 val = StringUtilities::convert<int16>(lvalue);
					if (val == numeric_limits<int16>::max()) {
						throw CommonCmdDef::BadInput(value);
					}
					controller->setValue(propertyName, val);
				}
			} else if (ptype == INT32_PROPERTY) {
				// Test conversion
				if (lvalue != "=") {
					int32 val = StringUtilities::convert<int32>(lvalue);
					if (val == numeric_limits<int32>::max()) {
						throw CommonCmdDef::BadInput(value);
					}
					controller->setValue(propertyName, val);
				}
			} else if (ptype == INT64_PROPERTY) {
				// Test conversion
				if (lvalue != "=") {
					int64 val = StringUtilities::convert<int64>(lvalue);
					if (val == numeric_limits<int64>::max()) {
						throw CommonCmdDef::BadInput(value);
					}
					controller->setValue(propertyName, val);
				}
			} else if (ptype == FLOAT32ARRAY_PROPERTY) {
				if (lvalue != "=") {
					float32 val = StringUtilities::convert<float32>(lvalue);
					if (val == numeric_limits<float32>::max()) {
						throw CommonCmdDef::BadInput(value);
					}
					controller->setValue(propertyName, val);
				}
			} else if (ptype == FLOAT64ARRAY_PROPERTY) {
				if (lvalue != "=") {
					float64 val = StringUtilities::convert<float64>(lvalue);
					if (val == numeric_limits<float64>::max()) {
						throw CommonCmdDef::BadInput(value);
					}
					controller->setValue(propertyName, val);
				}
			} else if (ptype == BOOLEAN_PROPERTY) {
				if (lvalue != "=") {
					// Test conversion
					bool val = false;
					if ((lvalue == "0") || (lvalue == "false")) {
						val = false;
					}
					else if ((lvalue == "1") || (lvalue == "true")) {
						val = true;
					}
					else {
						throw CommonCmdDef::BadInput(value);
					}
					controller->setValue(propertyName, val);
				}
			} else if (ptype == STRING_PROPERTY) {
				controller->setValue(propertyName, lvalue);
			} else {
				CommandLineManager::getInstance()->debug(
						controller->getName() + " property " + propertyName + " " + ptype + " doesn''t implemented");
				// Print Error
				CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
			}
			if (name.empty() == false) {
				outputProperty(controller, name, propertyName, "+ " + prefix);
			}
		}
	}
	catch (dataprovider::Database::NoSuchServantException& e) {
		controller->detach(CommandLineManager::getInstance());
		throw CommonCmdDef::BadInput("Cannot find controller : " + controller->getName());
	}
	catch (dataprovider::Database::NoSuchPropertyException& e) {
		controller->detach(CommandLineManager::getInstance());
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + propertyName);
	}
	catch (dataprovider::Database::BadPropertyTypeException& e) {
		controller->detach(CommandLineManager::getInstance());
		throw CommonCmdDef::BadInput("Bad property type: controller " + controller->getName() + " , property " + propertyName);
	} catch (dataprovider::Database::IndexOutOfBoundsException & e) {
		controller->detach(CommandLineManager::getInstance());
		throw CommonCmdDef::BadInput("Bad index on dynamic property type: controller " + controller->getName() + " , property " + propertyName);
	} catch (CommonCmdDef::BadInput& e) {
		controller->detach(CommandLineManager::getInstance());
		throw;
	} catch (CannotSetValue& e) {
		controller->detach(CommandLineManager::getInstance());
		if (StringUtilities::cmp(e.what(), ExperimentController::LIMIT_STR) == true) {
			throw CommonCmdDef::BadInput("Cannot set value on: controller " + controller->getName() + " , property " + propertyName, true);
		}
		else {
			throw CommonCmdDef::BadInput("Cannot set value on: controller " + controller->getName() + " , property " + propertyName);
		}
	} catch (...) {
		CommandLineManager::getInstance()->debug(controller->getName() + " property " + propertyName + " doesn''t existed");
		controller->detach(CommandLineManager::getInstance());
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + propertyName);
	}
	controller->detach(CommandLineManager::getInstance());
}

/*
 * setTimeOutProperty
 */
void AbstractParser::setTimeOutProperty(ExperimentController* controller, const string& name, const std::string& value,
		const std::string& valueUnit, const std::string& prefix)  {
	if (m_forEstimatedTime) return;

	try {
		string ptype = controller->getPropertyType(CommonCmdDef::TO_NAME);
		if (ptype == INT32_PROPERTY) {
			// Test conversion
			int32 val = StringUtilities::convert<int32>(value);
			if (val == numeric_limits<int32>::max()) {
				throw CommonCmdDef::BadInput(value);
			}
			if (val == 0) {
				controller->setValue(CommonCmdDef::USE_TO_NAME, false);
			} else {
				string ptypeUnit = controller->getPropertyType(CommonCmdDef::TO_TYPE_NAME);
				if (ptypeUnit != STRING_PROPERTY) {
					CommandLineManager::getInstance()->debug(
							controller->getName() + " property " + CommonCmdDef::TO_TYPE_NAME + " " + ptypeUnit
									+ " doesn''t implemented");
					// Print Error
					CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
					throw CommonCmdDef::BadInput(valueUnit);
				}
				controller->setValue(CommonCmdDef::USE_TO_NAME, true);
				if (StringUtilities::cmp(valueUnit, unit::HOUR, 1) == true) {
					controller ->setValue(CommonCmdDef::TO_TYPE_NAME, string(unit::HOUR));
				} else if (StringUtilities::cmp(valueUnit, unit::MINUTE, 1) == true) {
					controller ->setValue(CommonCmdDef::TO_TYPE_NAME, string(unit::MINUTE));
				} else if (StringUtilities::cmp(valueUnit, unit::SECOND, 1) == true) {
					controller ->setValue(CommonCmdDef::TO_TYPE_NAME, string(unit::SECOND));
				}
				controller->setValue(CommonCmdDef::TO_NAME, val);
			}
			if (name.empty() == false) {
				outputTimeOutProperty(controller, name, "+ " + prefix);
			}
		} else {
			CommandLineManager::getInstance()->debug(
					controller->getName() + " property " + CommonCmdDef::TO_NAME + " " + ptype + " doesn''t implemented");
			// Print Error
			CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
			throw CommonCmdDef::BadInput(value);
		}
	} catch (CommonCmdDef::BadInput& e) {
		throw;
	} catch (CannotSetValue& e) {
		throw;
	} catch (...) {
		CommandLineManager::getInstance()->debug(
				controller->getName() + " property " + CommonCmdDef::TO_NAME + " doesn''t existed");
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + name);
	}
}

/*
 * setTimeOutProperty
 */
void AbstractParser::setTimeOutProperty(ExperimentController *controller, const std::string &value,	const std::string &valueUnit, CommandLineCommand& clc)  {
	if (m_forEstimatedTime) return;

	try {
		string ptype = controller->getPropertyType(CommonCmdDef::TO_NAME);
		if (ptype == INT32_PROPERTY) {
			// Test conversion
			int32 val = StringUtilities::convert<int32>(value);
			if (val == numeric_limits<int32>::max()) {
				throw CommonCmdDef::BadInput(value);
			}
			if (val == 0) {
				clc.push_property(CommonCmdDef::USE_TO_NAME, "false");
			} else {
				string ptypeUnit = controller->getPropertyType(CommonCmdDef::TO_TYPE_NAME);
				if (ptypeUnit != STRING_PROPERTY) {
					CommandLineManager::getInstance()->debug(
							controller->getName() + " property " + CommonCmdDef::TO_TYPE_NAME + " " + ptypeUnit
									+ " doesn''t implemented");
					// Print Error
					ostringstream sbuf;
					sbuf << " : " << CommonCmdDef::TO_NAME;
					CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, sbuf.str());
					throw CommonCmdDef::BadInput(valueUnit);
				}
				clc.push_property(CommonCmdDef::USE_TO_NAME, "true");
				if (StringUtilities::cmp(valueUnit, unit::HOUR, 1) == true) {
					clc.push_property(CommonCmdDef::TO_TYPE_NAME, string(unit::HOUR));
				} else if (StringUtilities::cmp(valueUnit, unit::MINUTE, 1) == true) {
					clc.push_property(CommonCmdDef::TO_TYPE_NAME, string(unit::MINUTE));
				} else if (StringUtilities::cmp(valueUnit, unit::SECOND, 1) == true) {
					clc.push_property(CommonCmdDef::TO_TYPE_NAME, string(unit::SECOND));
				}
				clc.push_property(CommonCmdDef::TO_NAME, to_string(val));
			}
		} else {
			CommandLineManager::getInstance()->debug(
					controller->getName() + " property " + CommonCmdDef::TO_NAME + " " + ptype + " doesn''t implemented");
			// Print Error
			ostringstream sbuf;
			sbuf << " : " << CommonCmdDef::TO_NAME;
			CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, sbuf.str());
			throw CommonCmdDef::BadInput(value);
		}
	} catch (CommonCmdDef::BadInput& e) {
		throw;
	} catch (CannotSetValue& e) {
		throw;
	} catch (...) {
		CommandLineManager::getInstance()->debug(
				controller->getName() + " property " + CommonCmdDef::TO_NAME + " doesn''t existed");
		ostringstream sbuf;
		sbuf << "Property not found: controller " << controller->getName() << " , property " << CommonCmdDef::TO_NAME;
		throw CommonCmdDef::BadInput(sbuf.str());
	}
}

/*
 * setStabilisationProperty
 */
void AbstractParser::setStabilisationProperty(ExperimentController* controller, const string& name, const std::string& value,
		const std::string& valueUnit, const std::string& prefix)  {
	if (m_forEstimatedTime) return;

	try {
		string ptype = controller->getPropertyType(CommonCmdDef::STABILISATION_NAME);
		if (ptype == INT32_PROPERTY) {
			// Test conversion
			int32 val = StringUtilities::convert<int32>(value);
			if (val == numeric_limits<int32>::max()) {
				throw CommonCmdDef::BadInput(value);
			}
			if (val == 0) {
				controller->setValue(CommonCmdDef::USE_STABILISATION_NAME, false);
			} else {
				string ptypeUnit = controller->getPropertyType(CommonCmdDef::STABILISATION_TYPE_NAME);
				if (ptypeUnit != STRING_PROPERTY) {
					CommandLineManager::getInstance()->debug(
							controller->getName() + " property " + CommonCmdDef::TO_TYPE_NAME + " " + ptypeUnit
									+ " doesn''t implemented");
					// Print Error
					CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
					throw CommonCmdDef::BadInput(valueUnit);
				}
				controller->setValue(CommonCmdDef::USE_STABILISATION_NAME, true);
				if (StringUtilities::cmp(valueUnit, unit::HOUR, 1) == true) {
					controller ->setValue(CommonCmdDef::STABILISATION_TYPE_NAME, string(unit::HOUR));
				} else if (StringUtilities::cmp(valueUnit, unit::MINUTE, 1) == true) {
					controller ->setValue(CommonCmdDef::STABILISATION_TYPE_NAME, string(unit::MINUTE));
				} else if (StringUtilities::cmp(valueUnit, unit::SECOND, 1) == true) {
					controller ->setValue(CommonCmdDef::STABILISATION_TYPE_NAME, string(unit::SECOND));
				}
				controller->setValue(CommonCmdDef::STABILISATION_NAME, val);
			}
			if (name.empty() == false) {
				outputStabilisationProperty(controller, name, "+ " + prefix);
			}
		} else {
			CommandLineManager::getInstance()->debug(
					controller->getName() + " property " + CommonCmdDef::STABILISATION_NAME + " " + ptype
							+ " doesn''t implemented");
			// Print Error
			CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
			throw CommonCmdDef::BadInput(value);
		}
	} catch (CommonCmdDef::BadInput& e) {
		throw;
	} catch (CannotSetValue& e) {
		throw;
	} catch (...) {
		CommandLineManager::getInstance()->debug(
				controller->getName() + " property " + CommonCmdDef::STABILISATION_NAME + " doesn''t existed");
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + name);
	}
}

/*
 * setStabilisationProperty
 */
void AbstractParser::setStabilisationProperty(ExperimentController *controller, const std::string &value,	const std::string &valueUnit, CommandLineCommand& clc)  {
	if (m_forEstimatedTime) return;

	try {
		string ptype = controller->getPropertyType(CommonCmdDef::STABILISATION_NAME);
		if (ptype == INT32_PROPERTY) {
			// Test conversion
			int32 val = StringUtilities::convert<int32>(value);
			if (val == numeric_limits<int32>::max()) {
				throw CommonCmdDef::BadInput(value);
			}
			if (val == 0) {
				clc.push_property(CommonCmdDef::USE_STABILISATION_NAME, "false");
			} else {
				string ptypeUnit = controller->getPropertyType(CommonCmdDef::STABILISATION_TYPE_NAME);
				if (ptypeUnit != STRING_PROPERTY) {
					CommandLineManager::getInstance()->debug(
							controller->getName() + " property " + CommonCmdDef::TO_TYPE_NAME + " " + ptypeUnit
									+ " doesn''t implemented");
					// Print Error
					ostringstream sbuf;
					sbuf << " : " << CommonCmdDef::STABILISATION_NAME;
					CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + sbuf.str());
					throw CommonCmdDef::BadInput(valueUnit);
				}
				clc.push_property(CommonCmdDef::USE_STABILISATION_NAME, "true");
				if (StringUtilities::cmp(valueUnit, unit::HOUR, 1) == true) {
					clc.push_property(CommonCmdDef::STABILISATION_TYPE_NAME, string(unit::HOUR));
				} else if (StringUtilities::cmp(valueUnit, unit::MINUTE, 1) == true) {
					clc.push_property(CommonCmdDef::STABILISATION_TYPE_NAME, string(unit::MINUTE));
				} else if (StringUtilities::cmp(valueUnit, unit::SECOND, 1) == true) {
					clc.push_property(CommonCmdDef::STABILISATION_TYPE_NAME, string(unit::SECOND));
				}
				controller->setValue(CommonCmdDef::STABILISATION_NAME, val);
			}
		} else {
			CommandLineManager::getInstance()->debug(
					controller->getName() + " property " + CommonCmdDef::STABILISATION_NAME + " " + ptype
							+ " doesn''t implemented");
			// Print Error
			ostringstream sbuf;
			sbuf << " : " << CommonCmdDef::STABILISATION_NAME;
			CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, sbuf.str());
			throw CommonCmdDef::BadInput(value);
		}
	} catch (CommonCmdDef::BadInput& e) {
		throw;
	} catch (CannotSetValue& e) {
		throw;
	} catch (...) {
		CommandLineManager::getInstance()->debug(
				controller->getName() + " property " + CommonCmdDef::STABILISATION_NAME + " doesn''t existed");
		ostringstream sbuf;
		sbuf << "Property not found: controller " << controller->getName() << " , property " << CommonCmdDef::TO_NAME;
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + CommonCmdDef::STABILISATION_NAME);
	}
}

/*
 * outputProperty
 */
void AbstractParser::outputProperty(ExperimentController* controller, const string& name, const string& propertyName,
		const std::string& prefix)  {
	if (m_forEstimatedTime) return;

	try {
		if (propertyName.empty() == false) {
			string ptype = controller->getPropertyType(propertyName);
			if (ptype == UINT16_PROPERTY) {
				*m_output << CommonCmdDef::TAB << prefix << name << " = "
						<< any_cast<uint16> (controller->getValue(propertyName));
			}
			else if (ptype == UINT32_PROPERTY) {
				*m_output << CommonCmdDef::TAB << prefix << name << " = "
						<< any_cast<uint32> (controller->getValue(propertyName));
			}
			else if (ptype == UINT64_PROPERTY) {
				*m_output << CommonCmdDef::TAB << prefix << name << " = "
						<< any_cast<uint64> (controller->getValue(propertyName));
			}
			else if (ptype == INT16_PROPERTY) {
				*m_output << CommonCmdDef::TAB << prefix << name << " = "
						<< any_cast<int16> (controller->getValue(propertyName));
			}
			else if (ptype == INT32_PROPERTY) {
				*m_output << CommonCmdDef::TAB << prefix << name << " = "
						<< any_cast<int32> (controller->getValue(propertyName));
			}
			else if (ptype == INT64_PROPERTY) {
				*m_output << CommonCmdDef::TAB << prefix << name << " = "
						<< any_cast<int64> (controller->getValue(propertyName));
			}
			else if (ptype == FLOAT32_PROPERTY) {
				if (controller->containsProperty("nb_decimal_places") == true) {
					stringstream fformat;
					fformat << "%." << any_cast<int32> (controller->getValue("nb_decimal_places")) << "f ";
					*m_output << CommonCmdDef::TAB << prefix << name << " = " << format(fformat.str()) % any_cast<float32> (
							controller->getValue(propertyName));
				} else {
					*m_output << CommonCmdDef::TAB << prefix << name << " = " << any_cast<float32> (
							controller->getValue(propertyName));
				}
			}
			else if (ptype == FLOAT64_PROPERTY) {
				if (controller->containsProperty("nb_decimal_places") == true) {
					stringstream fformat;
					fformat << "%." << any_cast<int32> (controller->getValue("nb_decimal_places")) << "f ";
					*m_output << CommonCmdDef::TAB << prefix << name << " = " << format(fformat.str()) % any_cast<float64> (
							controller->getValue(propertyName));
				} else {
					*m_output << CommonCmdDef::TAB << prefix << name << " = " << any_cast<float64> (
							controller->getValue(propertyName));
				}
			}
			else if (ptype == BOOLEAN_PROPERTY) {
				if (any_cast<bool> (controller->getValue(propertyName)) == true) {
					*m_output << CommonCmdDef::TAB << prefix << name << " = " << "1";
				} else {
					*m_output << CommonCmdDef::TAB << prefix << name << " = " << "0";
				}
			} else if (ptype == STRING_PROPERTY) {
				*m_output << CommonCmdDef::TAB << prefix << name << " = "
						<< any_cast<string> (controller->getValue(propertyName));
			} else {
				CommandLineManager::getInstance()->debug(
						controller->getName() + " property " + propertyName + " " + ptype + " doesn''t implemented");
				// Print Error
				CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
			}
		} else {
			*m_output << CommonCmdDef::TAB << prefix << name << " = 0";
		}
	} catch (...) {
		CommandLineManager::getInstance()->debug(controller->getName() + " property " + propertyName + " doesn''t existed");
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + name);
	}
}

/*
 * outputProperty
 */
void AbstractParser::outputTimeOutProperty(ExperimentController* controller, const string& name, const std::string& prefix)
		 {
	if (m_forEstimatedTime) return;

	try {
		if (any_cast<bool> (getPropertyValue(controller, CommonCmdDef::USE_TO_NAME)) == true) {
			string ptype = controller->getPropertyType(CommonCmdDef::TO_NAME);
			if (ptype == INT32_PROPERTY) {
				string ptypeUnit = controller->getPropertyType(CommonCmdDef::TO_TYPE_NAME);
				if (ptypeUnit != STRING_PROPERTY) {
					CommandLineManager::getInstance()->debug(
							controller->getName() + " property " + CommonCmdDef::TO_TYPE_NAME + " " + ptypeUnit
									+ " doesn''t implemented");
					// Print Error
					CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
					throw CommonCmdDef::BadInput(CommonCmdDef::TO_TYPE_NAME);
				}
				*m_output << CommonCmdDef::TAB << prefix << name << " = " << any_cast<int32> (
						controller->getValue(CommonCmdDef::TO_NAME)) << " " << any_cast<string> (
						controller->getValue(CommonCmdDef::TO_TYPE_NAME));
			} else {
				CommandLineManager::getInstance()->debug(
						controller->getName() + " property " + CommonCmdDef::TO_NAME + " " + ptype + " doesn''t implemented");
				// Print Error
				CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
				throw CommonCmdDef::BadInput(CommonCmdDef::TO_NAME);
			}
		} else {
			*m_output << CommonCmdDef::TAB << prefix << name << " = 0";
		}
	} catch (CommonCmdDef::BadInput& e) {
		throw;
	} catch (...) {
		CommandLineManager::getInstance()->debug(
				controller->getName() + " property " + CommonCmdDef::TO_TYPE_NAME + " doesn''t existed");
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + name);
	}
}

/*
 * outputStabilisationProperty
 */
void AbstractParser::outputStabilisationProperty(ExperimentController* controller, const string& name, const std::string& prefix)
		 {
	if (m_forEstimatedTime) return;

	try {
		if (any_cast<bool> (getPropertyValue(controller, CommonCmdDef::USE_STABILISATION_NAME)) == true) {
			string ptype = controller->getPropertyType(CommonCmdDef::STABILISATION_NAME);
			if (ptype == INT32_PROPERTY) {
				string ptypeUnit = controller->getPropertyType(CommonCmdDef::STABILISATION_TYPE_NAME);
				if (ptypeUnit != STRING_PROPERTY) {
					CommandLineManager::getInstance()->debug(
							controller->getName() + " property " + CommonCmdDef::STABILISATION_TYPE_NAME + " " + ptypeUnit
									+ " doesn''t implemented");
					// Print Error
					CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
					throw CommonCmdDef::BadInput(CommonCmdDef::STABILISATION_TYPE_NAME);
				}
				*m_output << CommonCmdDef::TAB << prefix << name << " = " << any_cast<int32> (
						controller->getValue(CommonCmdDef::STABILISATION_NAME)) << " " << any_cast<string> (
						controller->getValue(CommonCmdDef::STABILISATION_TYPE_NAME));
			} else {
				CommandLineManager::getInstance()->debug(
						controller->getName() + " property " + CommonCmdDef::STABILISATION_NAME + " " + ptype + " doesn''t implemented");
				// Print Error
				CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + name);
				throw CommonCmdDef::BadInput(CommonCmdDef::STABILISATION_NAME);
			}
		} else {
			*m_output << CommonCmdDef::TAB << prefix << name << " = 0";
		}
	} catch (CommonCmdDef::BadInput& e) {
		throw;
	} catch (...) {
		CommandLineManager::getInstance()->debug(
				controller->getName() + " property " + CommonCmdDef::STABILISATION_TYPE_NAME + " doesn''t existed");
		throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + name);
	}
}

/*
 * outputEndl
 */
void AbstractParser::outputEndl() {
	if (m_forEstimatedTime) return;

	if (m_output != NULL) {
		if (m_output->str().empty() == false) {
			CommandLineManager::getInstance()->output(m_output->str());
		}
		delete m_output;
		m_output = new std::ostringstream();
	}
}

/*
 * outputError
 */
void AbstractParser::outputError(int32 id, const std::string& text) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->outputError(id, text);
	if (Simulated::activated == true) {
		m_nbErrors++;
	}
}

/*
 * convertStringToTimeAndType
 */
bool AbstractParser::convertStringToTimeAndType(const string input, string& time, string& type) const {
	time = input.substr(0, input.size() - 1);
	type = input.substr(input.size() - 1, 1);
	if ((StringUtilities::cmp(type, unit::HOUR, 1) == true) || (StringUtilities::cmp(type, unit::MINUTE, 1)
			== true) || (StringUtilities::cmp(type, unit::SECOND, 1) == true)) {
		return true;
	} else {
		time = input;
		return false;
	}
}

/*
 * convertPropertyStringToAny
 */
boost::any AbstractParser::convertPropertyStringToAny(ExperimentController* controller, const string& propertyName, const string& value) const {
	boost::any anyvalue;
	if (propertyName.empty() == false) {
		string ptype = controller->getPropertyType(propertyName);
		if (ptype == UINT16_PROPERTY) {
			// Test conversion
			uint16 val = StringUtilities::convert<uint16>(value);
			if (val == numeric_limits<uint16>::max()) {
				throw CommonCmdDef::BadInput(value);
			}
			anyvalue = val;
		} else if (ptype == UINT32_PROPERTY) {
				// Test conversion
				uint32 val = StringUtilities::convert<uint32>(value);
				if (val == numeric_limits<uint32>::max()) {
					throw CommonCmdDef::BadInput(value);
				}
				anyvalue = val;
		} else if (ptype == UINT64_PROPERTY) {
				// Test conversion
				uint64 val = StringUtilities::convert<uint64>(value);
				if (val == numeric_limits<uint64>::max()) {
					throw CommonCmdDef::BadInput(value);
				}
				anyvalue = val;
		} else 	if (ptype == INT16_PROPERTY) {
			// Test conversion
			int16 val = StringUtilities::convert<int16>(value);
			if (val == numeric_limits<int16>::max()) {
				throw CommonCmdDef::BadInput(value);
			}
			anyvalue = val;
		} else if (ptype == INT32_PROPERTY) {
				// Test conversion
				int32 val = StringUtilities::convert<int32>(value);
				if (val == numeric_limits<int32>::max()) {
					throw CommonCmdDef::BadInput(value);
				}
				anyvalue = val;
		} else if (ptype == INT64_PROPERTY) {
				// Test conversion
				int64 val = StringUtilities::convert<int64>(value);
				if (val == numeric_limits<int64>::max()) {
					throw CommonCmdDef::BadInput(value);
				}
				anyvalue = val;
		} else if (ptype == FLOAT32_PROPERTY) {
			float32 val = StringUtilities::convert<float32>(value);
			if (val == numeric_limits<float32>::max()) {
				throw CommonCmdDef::BadInput(value);
			}
			anyvalue = val;
		} else if (ptype == FLOAT64_PROPERTY) {
			float64 val = StringUtilities::convert<float64>(value);
			if (val == numeric_limits<float64>::max()) {
				throw CommonCmdDef::BadInput(value);
			}
			anyvalue = val;
		} else if (ptype == BOOLEAN_PROPERTY) {
			// Test conversion
			bool val = StringUtilities::convert<bool>(value);
			anyvalue = val;
		} else if (ptype == STRING_PROPERTY) {
			anyvalue = value;
		} else {
			CommandLineManager::getInstance()->debug(
					controller->getName() + " property " + propertyName + " " + ptype + " doesn''t implemented");
			// Print Error
			CommandLineManager::getInstance()->outputError(COMMAND_LINE_ILLEGAL_PARAM, ": " + propertyName);
		}
	}
	return anyvalue;
}
/*
 * convertStringToTimeAndType
 */
string AbstractParser::getDataParamProperty(ExperimentController* controller, const std::string& name)
		 {
	string tmp;
	int32 n_group = any_cast<int32> (getPropertyValue(controller, CommonCmdDef::DATAPARAM_N_GROUPS_NAME));
	for (int32  n = 0; n < n_group; ++n) {
		ostringstream s1;
		s1 << CommonCmdDef::DATAPARAM_GROUP_SIZE_NAME << "." << n;
		int32 size = any_cast<int32> (getPropertyValue(controller, s1.str()));
		for (int32  i = 0; i < size; ++i) {
			ostringstream s2;
			s2 << CommonCmdDef::DATAPARAM_PROPERTY_NAME << "." << n << "." << i;
			string dataname = any_cast<string> (getPropertyValue(controller, s2.str()));
			if (StringUtilities::cmp(dataname, name) == true) {
				ostringstream s3;
				s3 << CommonCmdDef::DATAPARAM_VALUE_NAME << "." << n << "." << i;
				tmp = s3.str();
				return tmp;
			}
		}
	}
	throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + name);
}

/*
 * getFileName
 */
std::string AbstractParser::getFileName(const boost::filesystem::path& path1, const std::string& name, const std::string& defaultname,
		const std::string& extension) {
	boost::filesystem::path path2;
	string buf = name;
	to_lower(buf);
	if (buf.find("shared/") != string::npos) {
		path2 = path1.parent_path() / "Shared";
		buf = name.substr(7);
	}
	else {
		path2 = path1;
		buf = name;
	}
	string filename;
	try {
		filename = common::FileUtilities::getFileName(path2, buf, defaultname, extension);
	} catch (common::FileUtilities::NotRegularFile& e) {
		throw CommonCmdDef::BadInput("File " + name + " doesn't exist");
	}
	return filename;
}

/*
 * getXbuPath
 */
boost::filesystem::path& AbstractParser::getXbuPath() {
	return CommandLineManager::getInstance()->m_XBUPATH;
}

/*
 * getPalPath
 */
boost::filesystem::path& AbstractParser::getPalPath() {
	return CommandLineManager::getInstance()->m_PALPATH;
}

/*
 * getSettingsPath
 */
boost::filesystem::path& AbstractParser::getSettingsPath() {
	return CommandLineManager::getInstance()->m_SETTINGSPATH;
}

/*
 * forloopSpy
 */
void AbstractParser::forloopSpy(uint32 numberOfSteps, uint32 currentStep) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->forloopSpy(numberOfSteps, currentStep);
}

/*
 * setSpyActualCommand
 */
void AbstractParser::setSpyActualCommand(const std::string& text) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->setSpyActualCommand(text);
}

/*
 * isStopped
 */
bool AbstractParser::isStopped() const {
	return CommandLineManager::getInstance()->isStopped();
}

/*
 * startControllers
 */
void AbstractParser::startControllers(const std::string& command, const std::vector<CommandLineCommand>& controllers, bool sync, bool doprogression) {
	if (m_forEstimatedTime) {
		if (command == "start") {
//			cout << "AbstractParser::computeEstimatedTimeControllers -> " << command << " -> " << controllers.size() << endl;
//			for (auto c : controllers){
//				cout << "size : " << c.getPropertiesMap().size() << endl;
//				for (auto p : c.getPropertiesMap()){
//					cout << p.first << " - " << p.second << endl;
//				}
//			}
			CommandLineManager::getInstance()->computeEstimatedTimeControllers(command, controllers, doprogression);
		}
	}
	else {
		CommandLineManager::getInstance()->startControllers(command, controllers, sync);
	}
}

/*
 * setXBURunning
 */
void AbstractParser::setXBURunning(const std::string& filename, bool value) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->setXBURunning(filename, value);
}

/*
 * setXBUCurrentLine
 */
void AbstractParser::setXBUCurrentLine(const std::string& filename, uint32 value) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->setXBUCurrentLine(filename, value);
}

/*
 * getXBUCurrentLine
 */
uint32 AbstractParser::getXBUCurrentLine(const std::string& filename) {
	return CommandLineManager::getInstance()->getXBUCurrentLine(filename);
}

/*
 * outputWarning
 */
void AbstractParser::outputWarning(const std::string& text) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->outputWarning(text);
}

/*
 * xbuSpy
 */
void AbstractParser::xbuSpy(const std::string& filename, uint32 numberOfLines, uint32 currentLineNumber,
		const std::string& previousLineString, const std::string& currentLineString, const std::string& nextLineString) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->xbuSpy(filename, numberOfLines, currentLineNumber, previousLineString, currentLineString, nextLineString);
}

/*
 * xbuSpy
 */
void AbstractParser::stopCommands() {
	CommandLineManager::getInstance()->stopCommands();
}

/*
 * xbuSpy
 */
void AbstractParser::stopControllers(const std::vector<ExperimentController*>& controllers) {
	CommandLineManager::getInstance()->stopControllers(controllers);
}

/*
 * setSpyPalContents
 */
void AbstractParser::setSpyPalContents(const std::string& text) {
	CommandLineManager::getInstance()->getSpyController()->updateValue(CommandLineManager::PAL_CONTENTS_PROPERTY, text);
}

/*
 * setSpyPalContents
 */
void AbstractParser::setSpyPalFullFilename(const std::string& filename) {
	CommandLineManager::getInstance()->getSpyController()->updateValue(CommandLineManager::PAL_FULL_FILENAME_PROPERTY, filename);
}

/*
 * setForloopRunning
 */
void AbstractParser::setForloopRunning(const std::string& loopname, bool value) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->setForloopRunning(loopname, value);
}

/*
 * setForloopCurrentPoint
 */
void AbstractParser::setForloopCurrentPoint(const std::string& loopname, uint32 value) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->setForloopCurrentPoint(loopname, value);
}

/*
 * getForloopCurrentPoint
 */
uint32 AbstractParser::getForloopCurrentPoint(const std::string& loopname) {
	return CommandLineManager::getInstance()->getForloopCurrentPoint(loopname);
}


/*
 * setIfRunning
 */
void AbstractParser::setIfRunning(const std::string& ifname, bool value) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->setIfRunning(ifname, value);
}

/*
 * setIfCurrentPoint
 */
void AbstractParser::setIfCurrentPoint(const std::string& ifname, uint32 value) {
	if (m_forEstimatedTime) return;

	CommandLineManager::getInstance()->setIfCurrentPoint(ifname, value);
}

/*
 * getIfCurrentPoint
 */
uint32 AbstractParser::getIfCurrentPoint(const std::string& ifname) {
	return CommandLineManager::getInstance()->getIfCurrentPoint(ifname);
}

