/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief Abstract class describes the parser
 * \author J. Locatelli
 * \date 12-01-2010
 */

#ifndef ABSTRACTPARSER_H
#define ABSTRACTPARSER_H

#include <string>
#include <vector>
#include <map>
#include <boost/any.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem.hpp>

#include "CommandSystem/CommandLine/Items.h"
#include "CommandSystem/CommandLine/CommandLineCommand.h"
#include "InstrumentAbstraction/ExperimentController.h"
#include "ErrorsAlarmsManager/ErrorDef.h"
#include "InstrumentManager/GenericPropertyDescription.h"
#include "Utilities/StringUtilities.h"
#include "CommandSystem/CommandLine/CommonCmdDef.h"

/*!
 * \class AbstractParser
 * \brief Abstract class describes the parser
 *
 * \par
 * This class is the virtual class of all command line parsers.
 * It contains all common functionalities.
 */
class AbstractParser {

public:

	AbstractParser();
	virtual ~AbstractParser();

	virtual AbstractParser* clone() = 0;
	void copy(const AbstractParser *parser);

	/*!
	 * \brief initialise the parser
	 * Fill info of controllers map and generics vector
	 */
	void init();

	/*!
	 * \brief Decode the command line
	 *
	 * \param[in] items list of item contains on the command line
	 * \throws BadInput syntax error when paring line
	 */
	virtual bool
	decode(Items items, bool doestimatedprogression);

	/*!
	 * \brief Return the info that this parser is sequential or not and check if xbu
	 *
	 * \return the synchronize action
	 */
	bool isSynchronized() const;

	/*!
	 * \brief Return the info that this parser is sequential or not
	 *
	 * \return the synchronize action
	 */
	virtual bool isParserSynchronized() const = 0;

	/*!
	 * \brief Try to find if item is correspond to a controller name
	 *
	 * \param[in] item the input text
	 * \param[out] anEC link to the found ExperimentController
	 * \param[in] size number of character to be used for comparing strings
	 * \return found or not
	 */
	bool
	isController(const std::string &item, ExperimentController **anEC, uint32 size = 0) const;

	/*!
	 * \brief Try to find if item is correspond to a controller name
	 *
	 * \param[in] item the input text
	 * \param[out] anEC link to the found ExperimentController
	 * \param[in] property
	 * \param[in] size number of character to be used for comparing strings
	 * \return found or not
	 */
	bool
	isControllerWithProperty(const std::string &item, ExperimentController **anEC, const std::string &property,
			uint32 size = 0) const;

	/*!
	 * \brief Try to find if item is correspond to a generic (with alias) property
	 *
	 * \param[in] item the input text
	 * \param[out] anEC link to the ExperimentController which contains the found property
	 * \param[out] propertyName real property name of the found property alias
	 * \return found or not
	 */
	bool
	isGenericProperty(const std::string &item, ExperimentController **anEC, std::string &propertyName) const;

	/*!
	 * \brief Try to find if item is correspond to a property
	 *
	 * \param[in] item the input text
	 * \param[out] anEC link to the ExperimentController which contains the found property
	 * \param[out] propertyName real property name of the found property alias
	 * \return found or not
	 */
	bool
	isProperty(const std::string &item, ExperimentController *anEC, std::string &propertyName) const;

	/*!
	 * \brief Convert a generic property name from a reference to a new one
	 *
	 * \param[in] controller The controller link which contains the property to convert
	 * \param[in] inpname Original property name
	 * \param[in] reference Reference string to replace in the property name
	 * \param[in] newref New reference to set in the property name
	 * \return the new property name
	 * \throws ControllerNotFound Controller not found
	 */
	std::string getGenericPropertyName(ExperimentController *controller, const std::string &inpname,
			const std::string &reference, const std::string &newref) const;

	/*!
	 * \brief Get the controller link of the first found of the wanted type
	 *
	 * \param[in] aType The type of controller we need
	 * \return link to the found controller
	 */
	ExperimentController* getControllerByType(const std::string &aType);

	/*!
	 * \brief Get the controller link of the first found of the using function
	 *
	 * \param[in] aFunction The function of controller we need
	 * \return link to the found controller
	 * \throws ControllerNotFound Controller not found
	 */
	ExperimentController* getControllerByFunction(const std::string &aFunction);

	/*!
	 * \brief Get the controller link which have the wanted name
	 *
	 * \param[in] aType The type of controller we need
	 * \return link to the found controller
	 * \throws ControllerNotFound Controller not found
	 */
	ExperimentController* getControllerByName(const std::string &aName);

	/*!
	 * \brief Get property value of an ExperiementController
	 *
	 * \param[in] controller The controller link which contains the property
	 * \param[in] propertyName The property name
	 * \return The property value
	 * \throws PropertyNotFound Property not found
	 */
	boost::any getPropertyValue(ExperimentController *controller, const std::string &propertyName);

	/*!
	 * \brief Get property value (string form) of an ExperiementController
	 *
	 * \param[in] controller The controller link which contains the property
	 * \param[in] propertyName The property name
	 * \return The property value in a string form
	 * \throws PropertyNotFound Property not found
	 */
	std::string getPropertyValueAsString(ExperimentController *controller, const std::string &propertyName);

	/*!
	 * \brief Set the property value and output it
	 *
	 * \param[in] controller The controller link which contains the property to set
	 * \param[in] name The string to print for this property value
	 * \param[in] propertyName The property name
	 * \param[in] value The value to set
	 * \param[in] prefix Prefix to set before name
	 * \throws PropertyNotFound Property not found
	 * \throws BadInput syntax error when paring line
	 */
	void setProperty(ExperimentController *controller, const std::string &name, const std::string &propertyName,
			const std::string &value, const std::string &prefix = "");

	/*!
	 * \brief Set the property value and output it
	 *
	 * \param[in] controller The controller link which contains the property to set
	 * \param[in] name The string to print for this property value
	 * \param[in] value The value to set
	 * \param[in] valueUnit The unit value to set
	 * \param[in] prefix Prefix to set before name
	 * \throws PropertyNotFound Property not found
	 * \throws BadInput syntax error when paring line
	 */
	void setTimeOutProperty(ExperimentController *controller, const std::string &value, const std::string &valueUnit, CommandLineCommand& clc);

	/*!
	 * \brief Set the property value and output it
	 *
	 * \param[in] controller The controller link which contains the property to set
	 * \param[in] name The string to print for this property value
	 * \param[in] value The value to set
	 * \param[in] valueUnit The unit value to set
	 * \param[in] prefix Prefix to set before name
	 * \throws PropertyNotFound Property not found
	 * \throws BadInput syntax error when paring line
	 */
	void setTimeOutProperty(ExperimentController *controller, const std::string &name, const std::string &value,
			const std::string &valueUnit, const std::string &prefix = "");

	/*!
	 * \brief Set the property value and output it
	 *
	 * \param[in] controller The controller link which contains the property to set
	 * \param[in] name The string to print for this property value
	 * \param[in] value The value to set
	 * \param[in] valueUnit The unit value to set
	 * \param[in] prefix Prefix to set before name
	 * \throws PropertyNotFound Property not found
	 * \throws BadInput syntax error when paring line
	 */
	void setStabilisationProperty(ExperimentController *controller, const std::string &value,	const std::string &valueUnit, CommandLineCommand& clc);

	/*!
	 * \brief Set the property value and output it
	 *
	 * \param[in] controller The controller link which contains the property to set
	 * \param[in] name The string to print for this property value
	 * \param[in] value The value to set
	 * \param[in] valueUnit The unit value to set
	 * \param[in] prefix Prefix to set before name
	 * \throws PropertyNotFound Property not found
	 * \throws BadInput syntax error when paring line
	 */
	void setStabilisationProperty(ExperimentController *controller, const std::string &name, const std::string &value,
			const std::string &valueUnit, const std::string &prefix = "");

	/*!
	 * \brief Add an end of line on actual output stream
	 */
	void outputEndl();

	/*!
	 * \brief Add the text to the output stream
	 */
	void outputText(const std::string text);

	/*!
	 * \brief Send a error text to client subscriber
	 *
	 * \param[in] text the text to send
	 */
	void outputError(int32 id, const std::string &text);

	/*!
	 * \brief Output the property value
	 *
	 * \param[in] controller The controller link which contains the property to output
	 * \param[in] name The string to print for this property value
	 * \param[in] propertyName The property name to output
	 * \param[in] prefix Prefix to set before name
	 * \throws PropertyNotFound Property not found
	 */
	void outputProperty(ExperimentController *controller, const std::string &name, const std::string &propertyName,
			const std::string &prefix = "");

	/*!
	 * \brief Output the property value
	 *
	 * \param[in] controller The controller link which contains the property to output
	 * \param[in] name The string to print for this property value
	 * \param[in] prefix Prefix to set before name
	 * \throws PropertyNotFound Property not found
	 */
	void outputTimeOutProperty(ExperimentController *controller, const std::string &name, const std::string &prefix = "");

	/*!
	 * \brief Output the property value
	 *
	 * \param[in] controller The controller link which contains the property to output
	 * \param[in] name The string to print for this property value
	 * \param[in] prefix Prefix to set before name
	 * \throws PropertyNotFound Property not found
	 */
	void outputStabilisationProperty(ExperimentController *controller, const std::string &name, const std::string &prefix = "");

	/*!
	 * \brief Get the list of controllers
	 *
	 * \return The list of controllers
	 */
	const std::map<std::string, ExperimentController*>& getControllers() const;

	bool convertStringToTimeAndType(const std::string input, std::string &time, std::string &type) const;

	boost::any convertPropertyStringToAny(ExperimentController *controller, const std::string &propertyName,
			const std::string &value) const;

	/*!
	 * \brief Get property value (string form) of an ExperiementController
	 *
	 * \param[in] controller The controller link which contains the property
	 * \param[in] name The name on data param
	 * \return The property value in a string form
	 * \throws PropertyNotFound Property not found
	 */
	std::string getDataParamProperty(ExperimentController *controller, const std::string &name);

	static std::string getFileName(const boost::filesystem::path &path1, const std::string &name, const std::string &defaultname,
			const std::string &extension);

	template<typename Type> Type cast(const std::string &value);

	void incNbError() {
		++m_nbErrors;
	}

	boost::filesystem::path& getXbuPath();

	boost::filesystem::path& getPalPath();

	boost::filesystem::path& getSettingsPath();

	void forloopSpy(uint32 numberOfSteps, uint32 currentStep);

	void setSpyActualCommand(const std::string& text);

	bool isStopped() const;

	void startControllers(const std::string& command, const std::vector<CommandLineCommand>& controllers, bool sync, bool doprogression);

	void setXBURunning(const std::string& filename, bool value);
	void setXBUCurrentLine(const std::string& filename, uint32 value);
	uint32 getXBUCurrentLine(const std::string& filename);

	void outputWarning(const std::string& text);

	void xbuSpy(const std::string& filename, uint32 numberOfLines, uint32 currentLineNumber,
			const std::string& previousLineString, const std::string& currentLineString, const std::string& nextLineString);

	void stopCommands();

	void stopControllers(const std::vector<ExperimentController*>& controllers);

	void setSpyPalContents(const std::string& text);
	void setSpyPalFullFilename(const std::string& filename);

	bool isForEstimatedTime() const;

	void setForloopRunning(const std::string& loopname, bool value);
	void setForloopCurrentPoint(const std::string& loopname, uint32 value);
	uint32 getForloopCurrentPoint(const std::string& loopname);

	void setIfRunning(const std::string& ifname, bool value);
	void setIfCurrentPoint(const std::string& ifname, uint32 value);
	uint32 getIfCurrentPoint(const std::string& ifname);

protected:

	//! Map of get from InstrumentManager which contains controller name as key, and link to the controller as value
	std::map<std::string, ExperimentController*> m_controllers;
	//! Map of get from InstrumentManager which contains controller nickname as key, and link to the controller as value
	std::map<std::string, ExperimentController*> m_nicknameControllers;
	//! Contains the description of Generic property
	std::vector<GenericPropertyDescription> m_generics;
	//! Link on the actual string stream output
	std::ostringstream *m_output;

	uint32 m_xbuLevel;
	uint32 m_nbErrors;
	bool m_forEstimatedTime;
};

/*
 * outputText
 */
inline void AbstractParser::outputText(const std::string text) {
	*m_output << text;
}

/*
 * getControllers
 */
inline const std::map<std::string, ExperimentController*>& AbstractParser::getControllers() const {
	return m_controllers;
}

/*
 * isForEstimatedTime
 */
inline bool AbstractParser::isForEstimatedTime() const {
	return m_forEstimatedTime;
}

template<typename Type> Type AbstractParser::cast(const std::string &value) {
	try {
		return boost::lexical_cast<Type>(value);
	} catch (boost::bad_lexical_cast &e) {
		throw CommonCmdDef::BadInput("Bad cast value: " + value);
	}
}

#endif //ABSTRACTPARSER_H
