/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief For Loop command
 * \author J. Locatelli
 * \date 15-02-2011
 */

#include "CommandSystem/CommandLine/Parsers/Common/ForLoopCmd.h"
#include "CommandSystem/CommandLine/Parsers/Common/ForLoopLine.h"
#include "SpecificParser.h"

#include <boost/algorithm/string.hpp>
#include <fstream>
#include <iostream>
using namespace std;

/*
 * Constructor
 */
ForLoopCmd::ForLoopCmd(AbstractParser* aParser, XBUInfo& xbuinfo, ForLoopCmd* aParent) :
		AbstractCmd(aParser), m_XBUInfo(xbuinfo), m_Parent(aParent) {
	// moved $ into name chars
	m_MathParser.DefineNameChars("0123456789_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ$");
	m_MathParser.DefineOprtChars("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+-*^/?<>=#!%&|~'_");
	m_MathParser.DefineInfixOprtChars("/+-*^?<>=#!%&|~'_");
}

/*
 * Destructor
 */
ForLoopCmd::~ForLoopCmd() {
}

/*
 * parse
 */
void ForLoopCmd::parse(Items& items, bool doestimatedprogression)  {
	// Parse for loop line
	bool more;
	m_ForLoopOK = true;
	try {
		do {
			more = false;
			// Get variable name value
			++items;
			if (items.end() == true) {
				// Nothing found so error
				throw CommonCmdDef::BadInput("No variable name");
			}
			if (StringUtilities::cmp(*items, CommonCmdDef::VARIABLE, 1) == false) {
				// Nothing found so error
				throw CommonCmdDef::BadInput("Missed $ before variable name");
			}
			string variable = *items;

			// Get started value
			++items;
			if (items.end() == true) {
				// Nothing found so error
				throw CommonCmdDef::BadInput("Inputs are missing");
			}
			float64 start = StringUtilities::convert<float64>(*items);

			// check to
			++items;
			if (items.end() == true) {
				// Nothing found so error
				throw CommonCmdDef::BadInput("Inputs are missing");
			}
			if (StringUtilities::cmp(*items, CommonCmdDef::TO) == true) {
				// Get end value
				++items;
				if (items.end() == true) {
					// Nothing found so error
					throw CommonCmdDef::BadInput("Missed end value");
				}
				float64 end = StringUtilities::convert<float64>(*items);

				// Get type of loop
				++items;
				if (items.end() == true) {
					// Nothing found so error
					throw CommonCmdDef::BadInput("Missed type of loop (step or np)");
				}
				ForLoopLine::ForLoopType type;
				if (StringUtilities::cmp(*items, "step") == true) {
					type = ForLoopLine::STEP;
				} else if (StringUtilities::cmp(*items, "np") == true) {
					type = ForLoopLine::NBPOINT;
				} else {
					// Nothing found so error
					throw CommonCmdDef::BadInput("Bad for loop type " + *items);
				}

				// Get values
				++items;
				if (items.end() == true) {
					// Nothing found so error
					throw CommonCmdDef::BadInput("Inputs are missing");
				}
				float64 step = 0, nbpoint = 0;
				switch (type) {
				case ForLoopLine::STEP:
					step = StringUtilities::convert<float64>(*items);
					break;
				case ForLoopLine::NBPOINT:
					nbpoint = StringUtilities::convert<float64>(*items);
					break;
//				case ForLoopLine::RANGE:
//					// Not possible
//					break;
				}
				m_forLoopLines.push_back(ForLoopLine(variable, start, end, step, nbpoint, type));
				++items;
			} else {
				// come back to first value
				--items;
				--items;
				vector<float64> values;
				// List of values
				while (true) {
					++items;
					if (items.end() == true) {
						break;
					}
					if (StringUtilities::cmp(*items, CommonCmdDef::SEPARATOR) == true) {
						break;
					}
					// Get next value
					float64 value = StringUtilities::convert<float64>(*items);
					if (value == numeric_limits<float64>::max()) {
						throw CommonCmdDef::BadInput("Need numeric value");
					}
					values.push_back(value);
				}
				m_forLoopLines.push_back(ForLoopLine(variable, values));
			}

			// Is an other variable
			if (items.end() == false) {
				if (StringUtilities::cmp(*items, CommonCmdDef::SEPARATOR) == true) {
					more = true;
				}
			}
		} while (more == true);

		// Check loop are corrected
		checkForLoopLine();
	} catch (CommonCmdDef::BadInput& e) {
		m_ForLoopOK = false;
		throw;
	}

	// File Commands until the endfor
	uint32 i = m_XBUInfo.getCurrentLine();
	int32 loop = -1;
	vector<Items>::iterator iter = m_XBUInfo.getCommands().begin();
	while (iter != m_XBUInfo.getCommands().end()) {
		Items tmp = *iter;
		if (tmp.getSize() == 0) {
			addCommand(tmp, i);
			++i;
		} else if (StringUtilities::cmp(*tmp, CommonCmdDef::FORLOOP_CMD)) {
			++loop;
			if (loop > 0) {
				addCommand(tmp, i);
			}
			++i;
		} else if (StringUtilities::cmp(*tmp, CommonCmdDef::END_FORLOOP_CMD)) {
			if (loop == 0) {
				addCommand(tmp, i);
				// end of this for loop
				break;
			} else {
				addCommand(tmp, i);
				++i;
				--loop;
			}
		}
		else {
			addCommand(tmp, i);
			++i;
		}
		++iter;
	}

	run();
}

/*
 * run
 */
void ForLoopCmd::run()  {

	// Check if for loop parsing was OK
	if (m_ForLoopOK == false) {
		return;
	}
	m_DoBrake = false;

	ostringstream forloopname;
	/*
	 * Print log info of loops
	 */
	vector<ForLoopLine>::iterator iterForLoopLines = m_forLoopLines.begin();
	while (iterForLoopLines != m_forLoopLines.end()) {
		ostringstream buf1;
		buf1 << *iterForLoopLines;
		forloopname << *iterForLoopLines << " , ";
		m_Parser->outputText(buf1.str());
		m_Parser->outputEndl();
		++iterForLoopLines;
	}

	bool fisrt = true;

	uint32 startpoint = 0;
	if (m_Parser->isForEstimatedTime()) {
		startpoint = m_Parser->getForloopCurrentPoint(forloopname.str());
	}
	/*
	 *  Loop
	 */
	uint32 maxpoints = (*m_forLoopLines.begin()).getNbPoint();
	if (m_Parser->isForEstimatedTime()) {
		if (maxpoints > 200) {
			// no estimated
			return;
		}
	}

	m_Parser->setForloopRunning(forloopname.str(), true);
	for (uint32 i = startpoint; i < maxpoints; ++i) {

		// Set for loop spy
		m_Parser->forloopSpy(maxpoints, i + 1);
		m_Parser->setForloopCurrentPoint(forloopname.str(),  i + 1);

		// Write log info, current value of variable
		vector<Items> newCommands;
		vector<uint32> newLineNumber;
		iterForLoopLines = m_forLoopLines.begin();
		while (iterForLoopLines != m_forLoopLines.end()) {
			// Eval current value
			float64 currentValue = (*iterForLoopLines).getCurrentValue();
			ostringstream buf1;
			buf1 << "ForLoop: " << (*iterForLoopLines).getVariableName() << " = " << currentValue;
			m_Parser->outputText(buf1.str());
			m_Parser->outputEndl();
			++iterForLoopLines;
		}

		// Replacing variable by expression result
		vector<Items>::iterator iterCommands = m_Commands.begin();
		vector<uint32>::iterator iterNumber = m_LineNumber.begin();
		while (iterCommands != m_Commands.end()) {
			// Loop on each items of this commands, and create new one by replacing loop variable expression
			Items its(*iterCommands);
			if (its.getSize() == 0) {
				newCommands.push_back(its);
				newLineNumber.push_back(*iterNumber);
			}
			else if (StringUtilities::cmp(*its, CommonCmdDef::END_FORLOOP_CMD) == true) {
				newCommands.push_back(its);
				newLineNumber.push_back(*iterNumber);
			}
			else if (StringUtilities::cmp(*its, CommonCmdDef::FORLOOP_CMD) == false) {
				Items newits;
				while (its.end() == false) {
					bool found = false;
					// Check varaible is present for each loops
					iterForLoopLines = m_forLoopLines.begin();
					while (iterForLoopLines != m_forLoopLines.end()) {
						m_MathParser.ClearVar();
						float64 currentValue = (*iterForLoopLines).getCurrentValue();
						m_MathParser.DefineVar((*iterForLoopLines).getVariableName(), &currentValue);
						// Try to replace variable $x by calculating value
						if ((*its).find((*iterForLoopLines).getVariableName()) != string::npos) {
							ostringstream sbuf;
							m_MathParser.SetExpr(*its);
							try {
								float64 eval_value = m_MathParser.Eval();
								sbuf << eval_value;
								newits.push_back(sbuf.str());
							} catch (mu::Parser::exception_type &e) {
								// Expression isn't corrected, so replace just variable by current value
								sbuf << currentValue;
								string buf = *its;
								boost::replace_all(buf, (*iterForLoopLines).getVariableName(), sbuf.str());
								newits.push_back(buf);
							}
							found = true;
						}
						++iterForLoopLines;
					}
					// No variable found, so take directly original items
					if (found == false) {
						newits.push_back(*its);
					}
					++its;
				}
				// Add new command in the new command vector used after for execution
				newCommands.push_back(newits);
				newLineNumber.push_back(*iterNumber);
			} else {
				// If for loop command, replace nothing
				newCommands.push_back(its);
				newLineNumber.push_back(*iterNumber);
			}
			++iterCommands;
			++iterNumber;
		}

		/*
		 * Run commands
		 */
		// Execute the list of new commands
		vector<Items>::iterator iterItemsCommands = newCommands.begin();
		vector<uint32>::iterator iterLineNumber = newLineNumber.begin();
		while (iterItemsCommands != newCommands.end()) {
			uint32 currentline = *iterLineNumber;
			// Normal command (no for loop)
			vector<Items> buf;
			vector<Items>::iterator nextiterItemsCommands = iterItemsCommands;
			while (nextiterItemsCommands != newCommands.end()) {
				buf.push_back(*nextiterItemsCommands);
				++nextiterItemsCommands;
			}
			XBUInfo info(m_XBUInfo.getXBUFile(), currentline, buf);
			SpecificParser cmd(m_Parser->isForEstimatedTime());
			if (((*iterItemsCommands).getSize() != 0) && (StringUtilities::cmp(*(*iterItemsCommands), CommonCmdDef::END_FORLOOP_CMD) == false)) {
				ostringstream buf1;
				buf1 << "XBU (ForLoop) " << m_XBUInfo.getXBUFile().filename().string() << " line " << currentline << " > "
						<< (*iterItemsCommands).getString();
				m_Parser->outputText(buf1.str());
				m_Parser->outputEndl();
				m_Parser->setSpyActualCommand((*iterItemsCommands).getString());
				bool found = cmd.decode(*iterItemsCommands, m_Parser, info, fisrt, this);
				if (found == false) {
					m_Parser->outputError(COMMAND_LINE_BAD_COMMAND, "");
				}
				fisrt = false;
			}
			iterItemsCommands += info.getCurrentLine() - currentline + 1;
			iterLineNumber += info.getCurrentLine() - currentline + 1;
			// Stopped ?
			if (m_Parser->isStopped() == true) {
				break;
			}
			// Break ?
			if (m_DoBrake == true) {
				break;
			}
		}

		// Stopped ?
		if (m_Parser->isStopped() == true) {
			break;
		}
		// Break ?
		if (m_DoBrake == true) {
			break;
		}

		// Next points in Loops
		iterForLoopLines = m_forLoopLines.begin();
		while (iterForLoopLines != m_forLoopLines.end()) {
			(*iterForLoopLines).next();
			++iterForLoopLines;
		}
	}
	m_XBUInfo.incCurrentLine(m_Commands.size());

	// Finished, reset spy
	m_Parser->forloopSpy(0, 0);
}

/*
 * addCommand
 */
void ForLoopCmd::addCommand(const Items& items, uint32 n) {
	if (m_ForLoopOK == true) {
		m_Commands.push_back(items);
		m_LineNumber.push_back(n);
	}
}

/*
 * doBrake
 */
void ForLoopCmd::doBrake() {
	m_DoBrake = true;
}

/*
 * checkForLoopLine
 */
void ForLoopCmd::checkForLoopLine()  {
	/*
	 * Check first inside this for loop
	 */
	// Check variable names
	// Check that loops as the same number of total nb points
	vector<ForLoopLine>::iterator iterForLoopLines = m_forLoopLines.begin();
	while (iterForLoopLines != m_forLoopLines.end()) {
		vector<ForLoopLine>::iterator iterForLoopLines2 = iterForLoopLines;
		++iterForLoopLines2;
		while (iterForLoopLines2 != m_forLoopLines.end()) {
			if ((*iterForLoopLines).getVariableName() == (*iterForLoopLines2).getVariableName()) {
				// Same name is used
				ostringstream buf1;
				buf1 << "Common name of variables : " << (*iterForLoopLines).getVariableName() << " , "
						<< (*iterForLoopLines2).getVariableName();
				throw CommonCmdDef::BadInput(buf1.str());
			}
			if ((*iterForLoopLines).getNbPoint() != (*iterForLoopLines2).getNbPoint()) {
				ostringstream buf1;
				buf1 << "Same number of points is needed : " << (*iterForLoopLines).getNbPoint() << " , "
						<< (*iterForLoopLines2).getNbPoint();
				throw CommonCmdDef::BadInput(buf1.str());
			}
			++iterForLoopLines2;
		}
		++iterForLoopLines;
	}
	checkWithParentForLoopLine();
}

/*
 * checkWithParentForLoopLine
 */
void ForLoopCmd::checkWithParentForLoopLine()  {
	// First called of parent for checking variables name conflits
	if (m_Parent != NULL) {
		m_Parent->checkWithParentForLoopLine(m_forLoopLines);
	}
}

/*
 * checkWithParentForLoopLine
 */
void ForLoopCmd::checkWithParentForLoopLine(const vector<ForLoopLine>& forLoopLines)  {
	// Checking variables name conflits with loops parameters
	vector<ForLoopLine>::const_iterator iterForLoopLines = m_forLoopLines.begin();
	while (iterForLoopLines != m_forLoopLines.end()) {
		vector<ForLoopLine>::const_iterator iterForLoopLines2 = forLoopLines.begin();
		while (iterForLoopLines2 != forLoopLines.end()) {
			if ((*iterForLoopLines).getVariableName() == (*iterForLoopLines2).getVariableName()) {
				// Same name is used
				ostringstream buf1;
				buf1 << "Common name of variables : " << *iterForLoopLines << " , " << *iterForLoopLines2;
				throw CommonCmdDef::BadInput(buf1.str());
			}
			++iterForLoopLines2;
		}
		++iterForLoopLines;
	}
	// Continu check with other parent
	if (m_Parent != NULL) {
		m_Parent->checkWithParentForLoopLine(forLoopLines);
	}
}
