/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief For Loop command
 * \author J. Locatelli
 * \date 15-02-2011
 */

#ifndef FORLOOPCMD_H
#define FORLOOPCMD_H

#include "CommandSystem/CommandLine/Parsers/Common/AbstractCmd.h"
#include "XBUInfo.h"
#include <muParser.h>
#include <iostream>

class ForLoopLine;

/*!
 * \class ForLoopCmd
 * \brief For Loop
 *
 * \par
 * This class implement the for loop command.
 * There are 3 kinds of for loop:
 * - for $(variable) (start) to (end) np (number of points)
 * - for $(variable) (start) to (end) step (delta value)
 * - for $(variable) (value1) (value2) (value3) ...
 *
 * Each for loop has to be finished by the "endfor" command
 *
 * Example:
 *    for $i 0 to 3 step 1
 *    	set title "toto $i"
 *    endfor
 *
 *
 */
class ForLoopCmd: public AbstractCmd {

public:

	/*!
	 * \brief Constructor
	 */
	ForLoopCmd(AbstractParser* aParser, XBUInfo& xbuinfo, ForLoopCmd* aParent = NULL);

	/*!
	 * \brief Destructor
	 */
	virtual ~ForLoopCmd();

	/*!
	 * \brief Decode the command line
	 *
	 * \param[in] items list of item contains on the command line
	 * \throws bad_any_cast bad cast on boost any property
	 * \throws BadInput syntax error when paring line
	 * \throws PropertyNotFound property not found
	 * \throws ControllerNotFound controller not found
	 * \throws CannotSetValue could not set property value
	 */
	virtual void parse(Items& items, bool doestimatedprogression) ;

	/*!
	 * \brief Run the foor loop command list
	 */
	void run() ;

	void doBrake();

private:

	/*!
	 * \brief Add command to for loop execution list
	 *
	 * \param[in] items command items
	 * \param[in] n xbu line number
	 */
	void addCommand(const Items& items, uint32 n);

	/*!
	 * \brief check for loop conflits inside this command
	 *
	 * \throws BadInput syntax error when paring line, for loor for isn't ok
	 */
	void checkForLoopLine() ;

	/*!
	 * \brief check variable name conflit with parent for loop variables
	 *
	 * \throws BadInput syntax error when paring line, for loor for isn't ok
	 */
	void checkWithParentForLoopLine() ;

	/*!
	 * \brief check variable name conflit with parent for loop variables
	 *
	 * \throws BadInput syntax error when paring line, for loor for isn't ok
	 */
	void checkWithParentForLoopLine(const std::vector<ForLoopLine>& forLoopLines) ;

	XBUInfo& m_XBUInfo;
	ForLoopCmd* m_Parent;	//! For Loop parent if for loop is executed inside an other for loop

	bool m_ForLoopOK; //! For loop looks ok flag

	std::vector<ForLoopLine> m_forLoopLines; //! For loop lines (description of a loop) vector

	std::vector<Items> m_Commands; //! Command list to be executed by the for loop
	std::vector<uint32> m_LineNumber; //! Commands XBU line number. Use for print info.

	mu::Parser m_MathParser; //! Math parser used for calculating current value from a string expression

	bool m_DoBrake;
};

#endif // FORLOOPCMD_H
