/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief ForLoopLine info class
 * \author J. Locatelli
 * \date 15-02-2011
 */

#include "CommandSystem/CommandLine/Parsers/Common/ForLoopLine.h"

#include <cmath>
using namespace std;

/*
 * Constructor
 */
ForLoopLine::ForLoopLine(const std::string& name, float64 start, float64 end, float64 step, float64 np, ForLoopType type) :
	m_varName(name), m_varStart(start), m_varEnd(end), m_varStep(step), m_varNbPoint(np), m_type(type), m_step(0) {
	m_currentValue = m_varStart;
	switch (m_type) {
	case STEP:
		// Compute number of points
		if (m_varStep == 0) {
			m_varNbPoint = 1;
		} else {
			// Check error of step and correct it in case..
			if (((m_varEnd - m_varStart) > 0) && (m_varStep < 0)) {
				m_varStep = -m_varStep;
			} else if (((m_varEnd - m_varStart) < 0) && (m_varStep > 0)) {
				m_varStep = -m_varStep;
			}
			m_varNbPoint = ceil(fabs((m_varEnd - m_varStart) / m_varStep)) + 1;
		}
		// Fill values
		for (uint32 i = 0; i < m_varNbPoint - 1; ++i) {
			float64 val = m_varStart + i * m_varStep;
			m_values.push_back(val);
		}
		m_values.push_back(m_varEnd);
		break;
	case NBPOINT:
		// Fill values
		if (m_varNbPoint <= 0) {
			m_currentValue = m_varEnd;
			m_varNbPoint = 1;
		} else {
			for (uint32 i = 0; i < m_varNbPoint - 1; ++i) {
				float64 val = m_varStart + i * (m_varEnd - m_varStart) / (m_varNbPoint - 1);
				m_values.push_back(val);
			}
			m_values.push_back(m_varEnd);
		}
		break;
	case RANGE:
		break;
	}
}

/*
 * Constructor
 */
ForLoopLine::ForLoopLine(const std::string& name, const std::vector<float64>& vals) :
	m_varName(name), m_varStart(0), m_varEnd(0), m_varStep(0), m_currentValue(0), m_type(RANGE), m_values(vals), m_step(0) {
	if (m_values.empty() == false) {
		m_currentValue = m_values[0];
	}
	switch (m_type) {
//	case STEP:
//		break;
//	case NBPOINT:
//		break;
	case RANGE:
		m_varNbPoint = m_values.size();
		break;
	}
}

/*
 * copy Constructor
 */
ForLoopLine::ForLoopLine(const ForLoopLine& f) {
	m_varName = f.m_varName;
	m_varStart = f.m_varStart;
	m_varEnd = f.m_varEnd;
	m_varStep = f.m_varStep;
	m_varNbPoint = f.m_varNbPoint;
	m_currentValue = f.m_currentValue;
	m_type = f.m_type;
	m_values = f.m_values;
	m_step = f.m_step;
}

/*
 * operator <<
 */
ostream& operator <<(ostream& os, const ForLoopLine& loop) {
	os << "ForLoop " << loop.m_varName << " ";
	switch (loop.m_type) {
	case ForLoopLine::STEP:
		os << loop.m_varStart << " to " << loop.m_varEnd << " step " << loop.m_varStep;
		break;
	case ForLoopLine::NBPOINT:
		os << loop.m_varStart << " to " << loop.m_varEnd << " np " << loop.m_varNbPoint;
		break;
	case ForLoopLine::RANGE: {
		std::vector<float64>::const_iterator iter = loop.m_values.begin();
		while (iter != loop.m_values.end()) {
			os << *iter;
			++iter;
			if ((iter != loop.m_values.end())) {
				os << " ";
			}
		}
	}
	}
	return os;
}
