/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief ForLoopLine info class
 * \author J. Locatelli
 * \date 15-02-2011
 */

#ifndef FORLOOPLINE_H
#define FORLOOPLINE_H

#include <iostream>
#include <string>
#include <vector>
#include <common/base/BaseTypes.h>

/*!
 * \class ForLoopLine
 * \brief ForLoopLine info class
 *
 * \par
 * This class contains information of For Loop command.
 */
class ForLoopLine {

public:

	/*!
	 * \enum Type of for loop
	 */
	typedef enum {
		RANGE, //! List of values
		STEP, //! Input is the number of step (delta)
		NBPOINT
	//! Input is the number of points
	} ForLoopType;

	/*!
	 * \brief constructor for STEP and NBPOINT loop
	 *
	 * \param[in] name The name of the variable ($x)
	 * \param[in] start The started value
	 * \param[in] end The ended value
	 * \param[in] step The step (delta)
	 * \param[in] np The number of points
	 * \param[in] type The type of for loop
	 */
	ForLoopLine(const std::string& name, float64 start, float64 end, float64 step, float64 np, ForLoopType type);

	/*!
	 * \brief constructor for RANGE loop
	 *
	 * \param[in] name The name of the variable ($x)
	 * \param[in] vals The list of values
	 */
	ForLoopLine(const std::string& name, const std::vector<float64>& vals);

	/*!
	 * \brief copy constructor
	 *
	 * \param[in] f The reference object for copying
	 */
	ForLoopLine(const ForLoopLine& f);

	/*!
	 * \brief Getter of the variable name
	 *
	 * \return The name
	 */
	const std::string& getVariableName() const;

	/*!
	 * \brief Getter of the number of point
	 *
	 * \return The number
	 */
	uint32 getNbPoint() const;

	/*!
	 * \brief Set current value at the begin of for loop
	 */
	void begin();

	/*!
	 * \brief Set current value at the next value
	 */
	void next();

	/*!
	 * \brief Getter of the current value
	 *
	 * \return The value
	 */
	float64 getCurrentValue() const;

	/*!
	 * \brief Operator == overload
	 *
	 * \param[in] f The ForLoopLine object to compare with
	 */
	bool operator==(const ForLoopLine& f) const {
		return m_varName == f.m_varName;
	}

	/*!
	 * \brief Operator <<+ overload
	 *
	 * \param[in/out] s The output stream where printing items
	 * \param[in] its The ForLoopLine to print
	 * \return The output stream where the items was print
	 */
	friend std::ostream& operator <<(std::ostream& os, const ForLoopLine& loop);

private:

	std::string m_varName; //! The variable name
	float64 m_varStart; //! The started value
	float64 m_varEnd; //! The ended value
	float64 m_varStep; //! The step (delta) value
	float64 m_varNbPoint; //! The number of points
	float64 m_currentValue; //! The current value
	ForLoopType m_type; //! The for loop type
	std::vector<float64> m_values; //! The list of values
	uint32 m_step; //! the actual step number of the loop

};

/*
 * getVariableName
 */
inline const std::string& ForLoopLine::getVariableName() const {
	return m_varName;
}

/*
 * getNbPoint
 */
inline uint32 ForLoopLine::getNbPoint() const {
	return (uint32) m_varNbPoint;
}

/*
 * getCurrentValue
 */
inline float64 ForLoopLine::getCurrentValue() const {
	return m_currentValue;
}

/*
 * begin
 */
inline void ForLoopLine::begin() {
	m_step = 0;
	if (m_values.empty() == false) {
		m_currentValue = m_values[m_step];
	}
}

/*
 * next
 */
inline void ForLoopLine::next() {
	++m_step;
	if (m_step < m_values.size()) {
		m_currentValue = m_values[m_step];
	}
}

#endif // FORLOOPLINE_H
