/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief IfLine info class
 * \author J. Locatelli
 * \date 25-03-2015
 */

#include "CommandSystem/CommandLine/Parsers/Common/IfLine.h"

#include <boost/tokenizer.hpp>
using namespace std;
using namespace boost;

/*
 * Constructor
 */
IfLine::IfLine(AbstractParser* aParser, const std::string& command)  :
		m_Parser(aParser), m_Command(command) {
	m_MathParser.DefineNameChars("0123456789_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ$.");
	m_MathParser.DefineOprtChars("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+-*^/?<>=#!%&|~'_");
	m_MathParser.DefineInfixOprtChars("not+-*^/?<>=#!%&|~'_");
	m_MathParser.EnableBuiltInOprt(false);

	m_MathParser.DefineInfixOprt("not", notOperator);
	m_MathParser.DefineInfixOprt("!", notOperator);
	m_MathParser.DefineOprt("or", addOperator, 0);
	m_MathParser.DefineOprt("||", addOperator, 0);
	m_MathParser.DefineOprt("and", multiplyOperator, 1);
	m_MathParser.DefineOprt("&&", multiplyOperator, 1);
	m_MathParser.DefineOprt("<", lesserOperator, 2);
	m_MathParser.DefineOprt("<=", lesserOrEqualsOperator, 2);
	m_MathParser.DefineOprt(">", greaterOperator, 2);
	m_MathParser.DefineOprt(">=", greaterOrEqualsOperator, 2);
	m_MathParser.DefineOprt("==", equalsOperator, 2);
	m_MathParser.DefineOprt("=", equalsOperator, 2);
	m_MathParser.DefineOprt("equal", equalsOperator, 2);
	m_MathParser.DefineOprt("!=", notEqualsOperator, 2);
	m_MathParser.DefineOprt("+", addOperator, 3);
	m_MathParser.DefineOprt("-", lessOperator, 3);
	m_MathParser.DefineOprt("%", modOperator, 3);
	m_MathParser.DefineOprt("/", divideOperator, 4);
	m_MathParser.DefineOprt("*", multiplyOperator, 4);

	m_MathParser.DefineConst("true", 1.0);
	m_MathParser.DefineConst("false", 0.0);

	m_MathParser.DefineConst("unknown", 0);
	m_MathParser.DefineConst("idle", 1);
	m_MathParser.DefineConst("running", 2);
	m_MathParser.DefineConst("warning", 4);
	m_MathParser.DefineConst("error", 8);

	m_Data = new MuParserUserData(m_Parser, &m_Values);
	m_MathParser.SetVarFactory(addValue, m_Data);

	m_MathParser.SetExpr(m_Command);
}

/*
 * Destructor
 */
IfLine::~IfLine() {
	delete m_Data;
	vector<float64*>::iterator iter = m_Values.begin();
	while (iter != m_Values.end()) {
		delete (*iter);
		++iter;
	}
}

/*
 * evaluate
 */
bool IfLine::evaluate()  {
	try {
		return (m_MathParser.Eval() != 0);
	} catch (mu::Parser::exception_type &e) {
		return false;
	}
}

float64 * IfLine::addValue(const char * valueName, void * pUserData)  {
	MuParserUserData* data = (MuParserUserData*) pUserData;
	string buf(valueName);
	// searching name
	size_t dotPosition = buf.find_first_of('.');
	if (dotPosition != string::npos) {
		string controllerName = buf.substr(0, dotPosition);
		string propertyName = buf.substr(dotPosition + 1);
		// Get controller
		ExperimentController* controller = NULL;
		if (data->parser->isController(controllerName, &controller)) {
			ExperimentController* ec2 = NULL;
			string tmp;
			if (data->parser->isGenericProperty(propertyName, &ec2, tmp)) {
				propertyName = tmp;
			} else if (data->parser->isProperty(propertyName, controller, tmp)) {
				propertyName = tmp;
			} else {
				throw CommonCmdDef::BadInput("Property not found: controller " + controller->getName() + " , property " + propertyName);
			}
			string lbuf = data->parser->getPropertyValueAsString(controller, propertyName);
			float64* val = new float64();
			*val = data->parser->cast<float64>(lbuf);
			data->values->push_back(val);
			return val;
		} else {
			throw CommonCmdDef::BadInput("Controller not found: " + controllerName);
		}
	} else {
		throw CommonCmdDef::BadInput(buf);
	}
}

/*
 * operator <<
 */
ostream& operator <<(ostream& os, const IfLine& loop) {
	return os;
}

float64 IfLine::notOperator(float64 a) {
	if (a > 0.0) {
		return 0.0;
	}
	return 1.0;
}

float64 IfLine::multiplyOperator(float64 a, float64 b) {
	return a * b;
}

float64 IfLine::addOperator(float64 a, float64 b) {
	return a + b;
}

float64 IfLine::divideOperator(float64 a, float64 b) {
	return a / b;
}

float64 IfLine::lessOperator(float64 a, float64 b) {
	return a - b;
}

float64 IfLine::lesserOperator(float64 a, float64 b) {
	return a < b;
}

float64 IfLine::lesserOrEqualsOperator(float64 a, float64 b) {
	return a <= b;
}

float64 IfLine::greaterOperator(float64 a, float64 b) {
	return a > b;
}

float64 IfLine::greaterOrEqualsOperator(float64 a, float64 b) {
	return a >= b;
}

float64 IfLine::equalsOperator(float64 a, float64 b) {
	return a == b;
}

float64 IfLine::notEqualsOperator(float64 a, float64 b) {
	return a != b;
}

float64 IfLine::modOperator(float64 a, float64 b) {
	return fmod(a, b);
}
