/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief XBU or DO command
 * \author J. Locatelli
 * \date 10-02-2011
 */

#include "CommandSystem/CommandLine/Parsers/Common/XBUCmd.h"
#include "XBUInfo.h"
#include "SpecificParser.h"

#include <iostream>
#include <fstream>
using namespace std;
#include "boost/filesystem/operations.hpp"
using namespace boost::filesystem;

/*
 * Constructor
 */
XBUCmd::XBUCmd(AbstractParser* aParser) :
		AbstractCmd(aParser) {
}

/*
 * Destructor
 */
XBUCmd::~XBUCmd() {
}

void XBUCmd::parse(Items& items, bool doestimatedprogression)  {
	// Get file name
	++items;
	if (items.end() == true) {
		// Nothing found so error
		throw CommonCmdDef::BadInput("File name is missing");
	}

	string filename = *items;
	path xbufile = AbstractParser::getFileName(m_Parser->getXbuPath(), filename, "", ".xbu");

	uint32 repetitions = 1;
	++items;
	if (items.end() == false) {
		repetitions = m_Parser->cast<uint32>(*items);
	}

	bool first = true;

	for(uint32 rep = 0;rep < repetitions ; ++rep) {
		time_t last_modification_time = last_write_time(xbufile);
		uint32 currentLine = 0;
		if (m_Parser->isForEstimatedTime() == true) {
			currentLine = m_Parser->getXBUCurrentLine(xbufile.string());
		}
		// Loop
		m_Parser->setXBURunning(xbufile.string(), true);
		uint32 xbulines = getNumberLines(xbufile.string());
		ostringstream buf0;
		buf0 << "Contains " << xbulines << " lines";
		if (repetitions > 1) {
			buf0 << " repetition " << rep + 1 << " of " << repetitions;
		}
		m_Parser->outputText(buf0.str());
		m_Parser->outputEndl();

		string tmpprev, tmp, tmpnext;
		while (true) {
			// Check if file changed from last
			time_t new_time = last_write_time(xbufile);
			if (new_time != last_modification_time) {
				last_modification_time = new_time;
				uint32 newxbulines = getNumberLines(xbufile.string());
				ostringstream buf1;
				buf1 << "XBU file " << filename << " changed since the last command, " << xbulines << " -> " << newxbulines
						<< " lines";
				m_Parser->outputWarning(buf1.str());
				xbulines = newxbulines;
				// Force recopie to serdon
				m_Parser->setXBURunning("", true);
				m_Parser->setXBURunning(xbufile.string(), true);
			}

			std::ifstream XBUFile;
			// Open file
			XBUFile.open(xbufile.string().c_str());
			if (XBUFile.is_open() == false) {
				// Nothing found so error
				throw CommonCmdDef::BadInput("Could not open " + filename);
			}

			// Go to current line
			for (uint32 i = 0; i < currentLine; ++i) {
				if (XBUFile.eof() == false) {
					getline(XBUFile, tmpprev);
				}
			}
			if (XBUFile.eof() == false) {
				// Get current command
				getline(XBUFile, tmp);
				++currentLine;
				m_Parser->setXBUCurrentLine(xbufile.string(), currentLine);
				// Parse command
				Items its(tmp);
				if (its.getSize() != 0) {
					ostringstream buf1;
					buf1 << "XBU " << filename << " line " << currentLine << " > " << tmp;
					m_Parser->outputText(buf1.str());
					m_Parser->outputEndl();

					tmpnext.clear();
					// Get next line for counting number or lines and get next line for spy
					if (XBUFile.eof() == false) {
						getline(XBUFile, tmpnext);
					}
					m_Parser->xbuSpy(filename, xbulines, currentLine, tmpprev, tmp, tmpnext);
					m_Parser->setSpyActualCommand(its.getString());
					vector<Items> buf;
					buf.push_back(tmp);
					buf.push_back(tmpnext);
					while (XBUFile.eof() == false) {
						getline(XBUFile, tmpnext);
						buf.push_back(tmpnext);
					}
					XBUInfo info(xbufile, currentLine, buf);
					SpecificParser cmd(m_Parser->isForEstimatedTime());
					bool found = cmd.decode(its, m_Parser, info, first);
					if (found == false) {
						m_Parser->outputError(COMMAND_LINE_BAD_COMMAND, tmp);
					}
					first = false;
					if (m_Parser->isStopped() == true) {
						ostringstream buf1;
						buf1 << "XBU " << filename << " stopped at line " << currentLine;
						m_Parser->outputText(buf1.str());
						m_Parser->outputEndl();
						break;
					}
					currentLine = info.getCurrentLine();
					m_Parser->setXBUCurrentLine(xbufile.string(), currentLine);
				}
			} else {
				// End of file
				ostringstream buf1;
				buf1 << "XBU " << filename << " end of buffer ";
				m_Parser->outputText(buf1.str());
				m_Parser->outputEndl();
				break;
			}
		}
		m_Parser->setXBURunning(xbufile.string(), false);
		m_Parser->xbuSpy("", 0, 0, "", "", "");
		if (m_Parser->isStopped() == true) {
			break;
		}
	}
}

uint32 XBUCmd::getNumberLines(const std::string& filename) {
	std::ifstream XBUFile;
	uint32 xbulines = 0;
	string tmp;
	XBUFile.open(filename.c_str());
	while (XBUFile.eof() == false) {
		getline(XBUFile, tmp);
		++xbulines;
	}
	XBUFile.close();
	return xbulines;
}
