/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CommandSystem/SimulatedConfigurationManager.h"
#include <CommonResponses.pb.h>
#include <ServantManagerRequests.pb.h>
#include <common/base/ServerProperties.h>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/string.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <ics/InstrumentManager/HardwareManager.h>
#include <ics/InstrumentManager/InstrumentManager.h>
#include <ics/PersistenceServices/Configuration/TestControllerConfig.h>
#include <ics/PersistenceServices/Configuration/ExperimentControllerConfig.h>
#include <sstream>

SimulatedConfigurationManager* SimulatedConfigurationManager::m_instance = new SimulatedConfigurationManager();

using namespace std;
using namespace boost;
using namespace cameo;

SimulatedConfigurationManager::InitFailureException::InitFailureException(const std::string& problem) :
	Exception("SimulatedConfigurationManager::InitFailureException", problem) {
}

SimulatedConfigurationManager* SimulatedConfigurationManager::getInstance() {
	return m_instance;
}

void SimulatedConfigurationManager::resetInstance() {
	delete m_instance;
}

void SimulatedConfigurationManager::loadConfiguration() {

	cout << "Loading configuration from real server" << endl;

	HardwareManager::initInstance();
	InstrumentManager::initInstance();

	loadTestControllers();
	loadExperimentControllers();

	HardwareManager::getInstance()->initTestControllers();

	InstrumentManager::getInstance()->configExperimentControllers();
	InstrumentManager::getInstance()->initExperimentControllers();

	HardwareManager::getInstance()->loadReports();
	InstrumentManager::getInstance()->loadReports();

}

SimulatedConfigurationManager::SimulatedConfigurationManager() {
}

SimulatedConfigurationManager::~SimulatedConfigurationManager() {
}

void SimulatedConfigurationManager::init() {

	cout << "Connection to real server" << endl;

	// Connect to the running Nomad server.
	string nomadAppNames[] = {"ns", "nsd", "nsg", "nsv"};

	std::unique_ptr<App> nomad;

	for (int i = 0; i < 4; ++i) {
		// Connect to the server.
		nomad = This::getServer().connect(nomadAppNames[i]);
		if (nomad) {
			break;
		}
	}

	if (!nomad) {
		cerr << "Unable to connect to the Nomad server" << endl;
		return;
	}

	cout << "Connected to " << *nomad << endl;

	// create the requester
	m_servantManagerRequester = coms::Requester::create(*nomad, "servant_manager");
	m_servantManagerRequester->init();

	cout << "Created requester " << *m_servantManagerRequester << endl;
}

void SimulatedConfigurationManager::loadExperimentControllers() {

	// Create the message type and request.
	servantmanager::Message type;
	type.set_type(servantmanager::Message::GetIcsExperimentControllersConfiguration);
	servantmanager::EmptyRequest request;

	m_servantManagerRequester->sendTwoParts(type.SerializeAsString(), request.SerializeAsString());

	std::optional<string> responseMessage = m_servantManagerRequester->receive();

	common::StringResponse response;
	response.ParseFromString(responseMessage.value());

	// getting experiment controllers
	string configurationContent = response.value();
	//cout << configurationContent << endl;

	istringstream buf(configurationContent);
	vector<ExperimentControllerConfig> configuration;

	boost::archive::text_iarchive ia(buf);
	ia >> configuration;

	// load configuration
	InstrumentManager::getInstance()->loadExperimentControllers(configuration);
}

void SimulatedConfigurationManager::loadTestControllers() {

	// Create the message type and request.
	servantmanager::Message type;
	type.set_type(servantmanager::Message::GetIcsTestControllersConfiguration);
	servantmanager::EmptyRequest request;

	m_servantManagerRequester->sendTwoParts(type.SerializeAsString(), request.SerializeAsString());

	std::optional<string> responseMessage = m_servantManagerRequester->receive();

	common::StringResponse response;
	response.ParseFromString(responseMessage.value());

	// getting test controllers
	string configurationContent = response.value();
	//cout << configurationContent << endl;

	istringstream buf(configurationContent);
	vector<TestControllerConfigNode> configuration;

	boost::archive::text_iarchive ia(buf);
	ia >> configuration;

	// load configuration
	HardwareManager::getInstance()->loadTestControllers(configuration);
}


bool SimulatedConfigurationManager::refreshConfiguration() {
	bool ret1 = refreshTestControllers();
	bool ret2 = refreshExperimentControllers();
	return ret1 && ret2;
}

bool SimulatedConfigurationManager::refreshExperimentControllers() {

	// Create the message type and request.
	servantmanager::Message type;
	type.set_type(servantmanager::Message::GetIcsExperimentControllersConfiguration);
	servantmanager::EmptyRequest request;

	m_servantManagerRequester->sendTwoParts(type.SerializeAsString(), request.SerializeAsString());

	std::optional<std::string> responseMessage = m_servantManagerRequester->receive();

	if (responseMessage.has_value()) {
		return false;
	}

	common::StringResponse response;
	response.ParseFromString(responseMessage.value());

	if (response.has_error()) {
		return false;
	}

	// getting experiment controllers
	string configurationContent = response.value();
	//cout << configurationContent << endl;
	istringstream buf(configurationContent);
	vector<ExperimentControllerConfig> configuration;

	boost::archive::text_iarchive ia(buf);
	ia >> configuration;

	// load configuration
	try {
		InstrumentManager::getInstance()->refreshExperimentControllers(configuration);
	} catch (...) {
		return false;
	}
	return true;
}

bool SimulatedConfigurationManager::refreshTestControllers() {
	// Create the message type and request.
	servantmanager::Message type;
	type.set_type(servantmanager::Message::GetIcsTestControllersConfiguration);
	servantmanager::EmptyRequest request;

	m_servantManagerRequester->sendTwoParts(type.SerializeAsString(), request.SerializeAsString());

	std::optional<std::string> responseMessage = m_servantManagerRequester->receive();

	if (responseMessage.has_value()) {
		return false;
	}

	common::StringResponse response;
	response.ParseFromString(responseMessage.value());

	if (response.has_error()) {
		return false;
	}

	// getting test controllers
	string configurationContent = response.value();
	//cout << configurationContent << endl;

	istringstream buf(configurationContent);
	vector<TestControllerConfigNode> configuration;

	boost::archive::text_iarchive ia(buf);
	ia >> configuration;

	try {
		HardwareManager::getInstance()->refreshTestControllers(configuration);
	} catch (...) {
		return false;
	}
	return true;
}
