/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DataProvider/ChangeAspectFactory.h"
#include "DataProvider/ChangeAspect.h"
#include "LoggingServices/Level.h"

using namespace std;

ChangeAspectFactory::ChangeAspectFactory() :
	m_nbCreatedChangeAspects(0), m_nbDeletedChangeAspects(0) {

}

ChangeAspectFactory::~ChangeAspectFactory() {
	cleanFactory();
}

ChangeAspect*
ChangeAspectFactory::createPropertyChangeAspect(const string& controllerType,
		const string& controllerName, const string& propertyName) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(PROPERTY, controllerType,
			controllerName);
	ca->setStringParam(propertyName);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);
	return ca;
}

ChangeAspect*
ChangeAspectFactory::createProgressChangeAspect(const string& controllerType,
		const string& controllerName, int32 progress) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(PROGRESS, controllerType,
			controllerName);
	ca->setLongParam(progress);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);
	return ca;
}

ChangeAspect*
ChangeAspectFactory::createErrorChangeAspect(const string& controllerType,
		const string& controllerName, int32 error) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(ERROR, controllerType, controllerName);
	ca->setLongParam(error);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);

	//ChangeAspect caa = (*ca); //test

	return ca;
}

ChangeAspect*
ChangeAspectFactory::createDataFileChangeAspect(const string& controllerType,
		const string& controllerName, int32 level, const string& dataconfigName) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(DATAFILE, controllerType,
			controllerName);
	ca->setStringParam(dataconfigName);
	ca->setLongParam(level);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);

	return ca;
}

ChangeAspect*
ChangeAspectFactory::createSynchroChangeAspect(const string& controllerType,
		const string& controllerName, const string& commandName, int32 startStop) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca =
			new ChangeAspect(SYNCHRO, controllerType, controllerName);
	ca->setStringParam(commandName);
	ca->setLongParam(startStop);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);

	return ca;
}

ChangeAspect*
ChangeAspectFactory::createRefuseChangeAspect(const string& controllerType,
		const string& controllerName, const string& commandName) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(REFUSE, controllerType, controllerName);
	ca->setStringParam(commandName);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);

	return ca;
}

ChangeAspect*
ChangeAspectFactory::createPropertyChangeAspect(const string& controllerType,
		const string& propertyName) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(PROPERTY, controllerType, "");
	ca->setStringParam(propertyName);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);
	return ca;
}

ChangeAspect*
ChangeAspectFactory::createProgressChangeAspect(const string& controllerType,
		int32 progress) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(PROGRESS, controllerType, "");
	ca->setLongParam(progress);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);
	return ca;
}

ChangeAspect*
ChangeAspectFactory::createErrorChangeAspect(const string& controllerType,
		int32 error) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(ERROR, controllerType, "");
	ca->setLongParam(error);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);

	//ChangeAspect caa = (*ca); //test

	return ca;
}

ChangeAspect*
ChangeAspectFactory::createCommandResetChangeAspect(
		const string& controllerType, const string& controllerName, int32 type) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(RESET_COMMAND, controllerType,
			controllerName);
	ca->setLongParam(type);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);

	return ca;
}

ChangeAspect* ChangeAspectFactory::createLogChangeAspect(
		const string& controllerType, const string& controllerName) {
	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(LOG_CHANGE, controllerType, controllerName);
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);

	return ca;
}

ChangeAspect* ChangeAspectFactory::createLogChangeAspect(
		const string& controllerType) {
	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	ChangeAspect* ca = new ChangeAspect(LOG_CHANGE, controllerType, "");
	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);

	return ca;
}

ChangeAspect*
ChangeAspectFactory::cloneChangeAspect(ChangeAspect* changeAspect) {

	if (changeAspect == 0) {
		

		return 0;
	}

	ChangeAspect* ca = 0;
	ca = changeAspect->clone();

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	m_nbCreatedChangeAspects++;
	m_createdChangedAspects.push_back(ca);
	return ca;
}

ChangeAspect*
ChangeAspectFactory::cloneChangeAspectAndFollow(ChangeAspect* changeAspect) {

	if (changeAspect == 0) {
		

		return 0;
	}

	ChangeAspect* resultCa = 0;
	ChangeAspect* currentChangeAspect = 0;
	ChangeAspect* lastClonedChangeAspect = 0;
	ChangeAspect* previousClonedChangeAspect = 0;

	resultCa = cloneChangeAspect(changeAspect);
	currentChangeAspect = changeAspect;
	previousClonedChangeAspect = resultCa;

	while (currentChangeAspect->hasNext()) {
		currentChangeAspect = currentChangeAspect->getNext();
		lastClonedChangeAspect = cloneChangeAspect(currentChangeAspect);
		previousClonedChangeAspect->setNext(lastClonedChangeAspect);
		previousClonedChangeAspect = lastClonedChangeAspect;
	}

	return resultCa;
}

void ChangeAspectFactory::cleanFactory() {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	//Delete all created ChangeAspects
	vector<ChangeAspect*>::const_iterator caIterator;
	for (caIterator = m_createdChangedAspects.begin(); caIterator
			!= m_createdChangedAspects.end(); caIterator++) {
		ChangeAspect* changeAspect = *caIterator;
		if (changeAspect != 0) {
			delete changeAspect;
			m_nbDeletedChangeAspects++;
		}
	}
	m_createdChangedAspects.clear();
}

void ChangeAspectFactory::remove(ChangeAspect* deadAspect) {

	// Serialise access to createChangeAspects
	std::scoped_lock changeAspectsLock(_createdChangeAspectsMutex);

	/*
	 * Look for change aspect in list and destroy/remove it
	 */
	ChangeAspect* changeAspect = NULL;
	vector<ChangeAspect*>::iterator caIterator =
			m_createdChangedAspects.begin();
	while (caIterator != m_createdChangedAspects.end()) {
		changeAspect = *caIterator;
		if (changeAspect == deadAspect) {
			break;
		} else {
			++caIterator;
		}
	}
	if (caIterator != m_createdChangedAspects.end()) {
		delete changeAspect;

		m_nbDeletedChangeAspects++;
		m_createdChangedAspects.erase(caIterator);
	}
}
