/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CHANGEASPECTFACTORY_H
#define CHANGEASPECTFACTORY_H

#include <common/base/BaseTypes.h>
#include "ChangeAspectType.h"
#include <string>
#include <vector>
#include <mutex>

class ChangeAspect;

class ChangeAspectFactory {

public:
	/**
	 * Constructor
	 */
	ChangeAspectFactory();

	/**
	 * Destructor
	 */
	~ChangeAspectFactory();

	/**
	 * To create a PROPERTY ChangeAspect
	 */
	ChangeAspect* createPropertyChangeAspect(const std::string& controllerType, const std::string& controllerName,
			const std::string& propertyName);

	ChangeAspect* createPropertyChangeAspect(const std::string& controllerType, const std::string& propertyName);

	/**
	 * To create a PROGRESS ChangeAspect
	 */
	ChangeAspect* createProgressChangeAspect(const std::string& controllerType, const std::string& controllerName,
			int32 progress);

	ChangeAspect* createProgressChangeAspect(const std::string& controllerType, int32 progress);

	/**
	 * To create an ERROR ChangeAspect
	 */
	ChangeAspect* createErrorChangeAspect(const std::string& controllerType, const std::string& controllerName, int32 error);

	ChangeAspect* createErrorChangeAspect(const std::string& controllerType, int32 error);

	/**
	 * To create an DATAFILE ChangeAspect
	 */
	ChangeAspect* createDataFileChangeAspect(const std::string& controllerType, const std::string& controllerName, int32 level,
			const std::string& dataconfigName = "");

	ChangeAspect* createSynchroChangeAspect(const std::string& controllerType, const std::string& controllerName,
			const std::string& commandName, int32 startStop);

	ChangeAspect* createRefuseChangeAspect(const std::string& controllerType, const std::string& controllerName,
			const std::string& commandName);

	ChangeAspect* createCommandResetChangeAspect(const std::string& controllerType, const std::string& controllerName,
			int32 type);

	/**
	 * To create a LOG ChangeAspect. The address is converted into a int32 value.
	 */
	ChangeAspect* createLogChangeAspect(const std::string& controllerType, const std::string& controllerName);

	ChangeAspect* createLogChangeAspect(const std::string& controllerType);

	/**
	 * To clone a ChangeAspect
	 */
	ChangeAspect* cloneChangeAspect(ChangeAspect* changeAspect);

	/**
	 * To clone a ChangeAspect and linked ChangeAspects
	 */
	ChangeAspect* cloneChangeAspectAndFollow(ChangeAspect* changeAspect);

	/**
	 * To delete all ChangeAspects created by this factory
	 */
	void cleanFactory();

	void remove(ChangeAspect* changeAspect);

private:
	/**
	 * All created ChangeAspect instances are stored in a vector
	 * When ChangeAspectFactory destructor is called,
	 * all instances stored in the vector are deleted
	 */
	std::vector<ChangeAspect*> m_createdChangedAspects;

	/**
	 * Counts the number of ChangeAspects created by the
	 * factory (used in unit test)
	 */
	int32 m_nbCreatedChangeAspects;

	/**
	 * Counts the number of deleted ChangeAspects (used in unit test)
	 */
	int32 m_nbDeletedChangeAspects;

	std::mutex _createdChangeAspectsMutex;

};

#endif
