/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DataProvider/CommandManager.h"
#include <command/transportlayer/MultiThreadProcessorCommunication.h>
#include <command/transportlayer/dataprovider/ServantAccessor.h>
#include <command/transportlayer/dataprovider/CommandAccessor.h>
#include "CommandNames.h"

using namespace std;
using namespace command;

CommandManager::CommandManager(const std::string& name) :
	_name(name),
	_servantID(0),
	_commandAccessor(0) {

	if (MultiThreadProcessorCommunication::getInstance() != 0) {
		_servantID = (uint32) MultiThreadProcessorCommunication::getInstance()->getServantAccessor()->getServantID(name);
		_commandAccessor = MultiThreadProcessorCommunication::getInstance()->getCommandAccessor();
	}
}

CommandManager::~CommandManager() {
}

CommandManager* CommandManager::clone(const std::string& name) const {

	CommandManager* manager = new CommandManager(name);

	// copying the commands
	for (map<string, uint32>::const_iterator c = _commands.begin(); c != _commands.end(); ++c) {
		manager->addCommand(c->first);
	}

	return manager;
}

void CommandManager::renameServant(const std::string& name) {
	_name = name;
}

std::string CommandManager::getCommands() const {
	return "";
}

void CommandManager::addCommand(const std::string& commandName) {

	if (_commandAccessor != 0) {
		try {
			uint32 commandID = (uint32) _commandAccessor->addCommandToServant(_servantID, commandName);
			_commands[commandName] = commandID;
		} catch (...) {

		}
	}
}

void CommandManager::removeCommand(const std::string& commandName) {
}

bool CommandManager::commandExists(const std::string& commandName) const {

	map<string, uint32>::const_iterator c = _commands.find(commandName);

	if (c != _commands.end()) {
		return true;
	}

	return false;

}

void CommandManager::setCommandState(const std::string& commandName, common::CommandState state) const {
	if (_commandAccessor != 0) {
		map<string, uint32>::const_iterator c = _commands.find(commandName);
		if (c != _commands.end()) {
			_commandAccessor->setCommandState(_servantID, c->second, state);
		}
	}
}

common::CommandState CommandManager::getCommandState(const std::string& commandName) const {
	if (_commandAccessor != 0) {
		map<string, uint32>::const_iterator c = _commands.find(commandName);
		if (c != _commands.end()) {
			return _commandAccessor->getCommandState(c->second);
		}
	}

	return common::UNKNOWN;
}

void CommandManager::setRunningCommandState(const std::string& fromCommand, common::CommandState state) const {

	// INACTIVE, PENDING, STARTING, ACTIVE, PAUSED, PAUSING, STOPPING, RESUMING, UNKNOWN

	int counter = 0;
	for (map<string, uint32>::const_iterator c = _commands.begin(); c != _commands.end(); ++c) {
		// read and status are excluded from the list of commands
		if (c->first != "read" && c->first != "status" && c->first != fromCommand) {
			common::CommandState commandState = _commandAccessor->getCommandState(c->second);
			if (commandState != common::INACTIVE) {
				_commandAccessor->setCommandState(_servantID, c->second, state);
				counter++;
			}
		}
	}

	if (counter > 1) {
		common::warn << "More than one command of " << _name << " was not INACTIVE while changing state" << common::eol;
	}
}

bool CommandManager::hasRunningCommandState(common::CommandState state) const {

	for (map<string, uint32>::const_iterator c = _commands.begin(); c != _commands.end(); ++c) {
		// read and status are excluded from the list of commands
		if (c->first != "read" && c->first != "status") {
			common::CommandState commandState = _commandAccessor->getCommandState(c->second);
			if (commandState == state) {
				return true;
			}
		}
	}

	return false;
}

bool CommandManager::hasOnlyInactiveCommands() const {
	for (map<string, uint32>::const_iterator c = _commands.begin(); c != _commands.end(); ++c) {
		// read and status are excluded from the list of commands
		if (c->first != "read" && c->first != "status") {
			common::CommandState commandState = _commandAccessor->getCommandState(c->second);
			if (commandState != common::INACTIVE) {
				return false;
			}
		}
	}

	return true;
}

std::vector<std::string> CommandManager::getCommandList() const {

	vector<string> result;

	for (map<string, uint32>::const_iterator c = _commands.begin(); c != _commands.end(); ++c) {
		result.push_back(c->first);
	}

	return result;
}

void CommandManager::setProgression(float64 value) {
	if (_commandAccessor != 0) {
		map<string, uint32>::iterator c = _commands.find(COMMAND_START);
		if (c != _commands.end()) {
			_commandAccessor->setProgression(c->second, value);
		}
	}
}
