/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DataProvider/Property/AnyProperty.h"
#include <command/transportlayer/dataprovider/PropertyAccessor.h>
#include <core/dataaccess/DataAccess.h>
#include <boost/format.hpp>

using namespace boost;
using namespace std;

AnyProperty::CannotSetValue::CannotSetValue() :
	NotOk(0) {
}

AnyProperty::CannotSetValue::CannotSetValue(const std::string &reason) :
	NotOk(0, reason) {
}

template<class T>
void AnyProperty::convert(ostream& s, const T& value, const string& format) const {
	if (format.empty() == true) {
		s << any_cast<T> (value);
	} else {
		if (format.find("%x") != string::npos) {
			s << "0x" << boost::format(format) % any_cast<T> (value);
		} else {
			s << boost::format(format) % any_cast<T> (value);
		}
	}
}

/**
 * Constructor aimed at array types. The size can still be modified
 * using the setSize method.
 */
AnyProperty::AnyProperty(const std::string& propertyName,
		const std::string& propertyAlias, const std::string& propertyType,
		bool propertySaveFlag, const std::string& propertyFormat,
		bool propertySpyFlag, bool propertyScanFlag, bool propertySaveBoxFlag, int32 propertySize) :
		_ID(0), _propertyAccessor(0),
		_fromType("<legacy>"),
		_name(propertyName), _alias(propertyAlias), _type(propertyType), _save(propertySaveFlag), _format(propertyFormat),
		_spy(propertySpyFlag), _scannable(propertyScanFlag), _saveBox(propertySaveBoxFlag), _extraSize(0), _size(propertySize) {

	if (propertyType == BOOLEAN_PROPERTY) {
		_typeID = 1;
		setValue(false);
	} else if (propertyType == INT16_PROPERTY) {
		_typeID = 2;
		setValue(int16(0));
	} else if (propertyType == UINT16_PROPERTY) {
		_typeID = 3;
		setValue(uint16(0));
	} else if (propertyType == INT32_PROPERTY) {
		_typeID = 4;
		setValue(int32(0));
	} else if (propertyType == UINT32_PROPERTY) {
		_typeID = 5;
		setValue(uint32(0));
	} else if (propertyType == INT64_PROPERTY) {
		_typeID = 6;
		setValue(int64(0));
	} else if (propertyType == UINT64_PROPERTY) {
		_typeID = 7;
		setValue(uint64(0));
	} else if (propertyType == FLOAT32_PROPERTY) {
		_typeID = 8;
		setValue(float32(0.0));
	} else if (propertyType == FLOAT64_PROPERTY) {
		_typeID = 9;
		setValue(float64(0.0));
	} else if (propertyType == STRING_PROPERTY) {
		_typeID = 10;
		setValue(string(""));
	} else if (propertyType == INT16ARRAY_PROPERTY) {
		_typeID = 11;
		setValue((int16*) 0);
	} else if (propertyType == UINT16ARRAY_PROPERTY) {
		_typeID = 12;
		setValue((uint16*) 0);
	} else if (propertyType == INT32ARRAY_PROPERTY) {
		_typeID = 13;
		setValue((int32*) 0);
	} else if (propertyType == UINT32ARRAY_PROPERTY) {
		_typeID = 14;
		setValue((uint32*) 0);
	} else if (propertyType == INT64ARRAY_PROPERTY) {
		_typeID = 15;
		setValue((int64*) 0);
	} else if (propertyType == UINT64ARRAY_PROPERTY) {
		_typeID = 16;
		setValue((uint64*) 0);
	} else if (propertyType == FLOAT32ARRAY_PROPERTY) {
		_typeID = 17;
		setValue((float32*) 0);
	} else if (propertyType == FLOAT64ARRAY_PROPERTY) {
		_typeID = 18;
		setValue((float64*) 0);
	}
}

AnyProperty::~AnyProperty() {
}

void AnyProperty::setFromType(const std::string& fromType) {
	_fromType = fromType;
}

const std::string& AnyProperty::getFromType() const {
	return _fromType;
}

AnyProperty* AnyProperty::clone() const {

	AnyProperty* property = new AnyProperty(_name, _alias, _type, _save, _format, _spy, _scannable, _saveBox, _size);
	property->setFromType(_fromType);
	property->setExtraSize(_extraSize);
	property->setValue(getValue());

	return property;
}

unsigned long AnyProperty::getID() const {
	return _ID;
}

void AnyProperty::init(unsigned long ID, command::PropertyAccessor* propertyAccessor) {
	_ID = ID;
	_propertyAccessor = propertyAccessor;

	// initialize settings
	// do not set values of size and value
	// because in case of creation of a float64 AnyProperty, the default values would be sent to data provider
	_propertyAccessor->setPropertySettings(_ID, _format, _save, _spy);
}

void AnyProperty::setSave(bool value) {
	_save = value;
}

void AnyProperty::setFormat(const std::string& format) {
	_format = format;
}

void AnyProperty::setSpy(bool value) {
	_spy = value;
}

void AnyProperty::setScannable(bool value) {
	_scannable = value;
}

void AnyProperty::setSaveBox(bool value) {
	_saveBox = value;
}

void AnyProperty::updateAttributes() {
	// property settings
	_propertyAccessor->setPropertySettings(_ID, _format, _save, _spy);
}

const string& AnyProperty::getName() const {
	return _name;
}

const string& AnyProperty::getAlias() const {
	return _alias;
}

const string& AnyProperty::getType() const {
	return _type;
}

bool AnyProperty::isSave() const {
	return _save;
}

const string& AnyProperty::getFormat() const {
	return _format;
}

bool AnyProperty::isScannable() const {
	return _scannable;
}

bool AnyProperty::isSpy() const {
	return _spy;
}

bool AnyProperty::isSaveBox() const {
	return _saveBox;
}

int32 AnyProperty::getExtraSize() const {
	return _extraSize;
}

void AnyProperty::setExtraSize(int32 inputSize) {
	_extraSize = inputSize;
}

int32 AnyProperty::getSize() const {

	if (_ID == 0) {
		return _size;
	}

	switch (_typeID) {
	case 11:
		return _propertyAccessor->getIcsInt16ArraySize(_ID);
	case 12:
		return _propertyAccessor->getIcsUint16ArraySize(_ID);
	case 13:
		return _propertyAccessor->getIcsInt32ArraySize(_ID);
	case 14:
		return _propertyAccessor->getIcsUint32ArraySize(_ID);
	case 15:
		return _propertyAccessor->getIcsInt64ArraySize(_ID);
	case 16:
		return _propertyAccessor->getIcsUint64ArraySize(_ID);
	case 17:
		return _propertyAccessor->getIcsFloat32ArraySize(_ID);
	case 18:
		return _propertyAccessor->getIcsFloat64ArraySize(_ID);
	}

	return 0;
}

void AnyProperty::setSize(int32 value) {

	if (_ID == 0) {
		_size = value;
		return;
	}


	switch (_typeID) {
	case 11:
		_propertyAccessor->setIcsInt16ArraySize(_ID, value);
		break;
	case 12:
		_propertyAccessor->setIcsUint16ArraySize(_ID, value);
		break;
	case 13:
		_propertyAccessor->setIcsInt32ArraySize(_ID, value);
		break;
	case 14:
		_propertyAccessor->setIcsUint32ArraySize(_ID, value);
		break;
	case 15:
		_propertyAccessor->setIcsInt64ArraySize(_ID, value);
		break;
	case 16:
		_propertyAccessor->setIcsUint64ArraySize(_ID, value);
		break;
	case 17:
		_propertyAccessor->setIcsFloat32ArraySize(_ID, value);
		break;
	case 18:
		_propertyAccessor->setIcsFloat64ArraySize(_ID, value);
		break;
	}
}

boost::any AnyProperty::getValue() const {

	if (_ID == 0) {
		return _value;
	}

	switch (_typeID) {
	case 1:
		return _propertyAccessor->getIcsBooleanValue(_ID);
	case 2:
		return _propertyAccessor->getIcsInt16Value(_ID);
	case 3:
		return _propertyAccessor->getIcsUint16Value(_ID);
	case 4:
		return _propertyAccessor->getIcsInt32Value(_ID);
	case 5:
		return _propertyAccessor->getIcsUint32Value(_ID);
	case 6:
		return _propertyAccessor->getIcsInt64Value(_ID);
	case 7:
		return _propertyAccessor->getIcsUint64Value(_ID);
	case 8:
		return _propertyAccessor->getIcsFloat32Value(_ID);
	case 9:
		return _propertyAccessor->getIcsFloat64Value(_ID);
	case 10:
		return _propertyAccessor->getIcsStringValue(_ID);
	case 11:
		return _propertyAccessor->getIcsInt16ArrayValue(_ID);
	case 12:
		return _propertyAccessor->getIcsUint16ArrayValue(_ID);
	case 13:
		return _propertyAccessor->getIcsInt32ArrayValue(_ID);
	case 14:
		return _propertyAccessor->getIcsUint32ArrayValue(_ID);
	case 15:
		return _propertyAccessor->getIcsInt64ArrayValue(_ID);
	case 16:
		return _propertyAccessor->getIcsUint64ArrayValue(_ID);
	case 17:
		return _propertyAccessor->getIcsFloat32ArrayValue(_ID);
	case 18:
		return _propertyAccessor->getIcsFloat64ArrayValue(_ID);
	}

	return 1L;
}

void throwCannotSetValue(unsigned long ID, int32 typeID, const std::string& propertyName, const boost::any& value) {

	unsigned long servantID = ::core::DataAccess::getInstance()->getServantIDFromPropertyID(0, ID);
	std::string servantName = ::core::DataAccess::getInstance()->getServantName(0, servantID);

	std::ostringstream os;
	os << "Cannot set property " << propertyName << " of " << servantName;

	switch (typeID) {
	case 1:
		os << " with value " << any_cast<bool>(value) << endl;
		break;
	case 2:
		os << " with value " << any_cast<int16>(value) << endl;
		break;
	case 3:
		os << " with value " << any_cast<uint16>(value) << endl;
		break;
	case 4:
		os << " with value " << any_cast<int32>(value) << endl;
		break;
	case 5:
		os << " with value " << any_cast<uint32>(value) << endl;
		break;
	case 6:
		os << " with value " << any_cast<int64>(value) << endl;
		break;
	case 7:
		os << " with value " << any_cast<uint64>(value) << endl;
		break;
	case 8:
		os << " with value " << any_cast<float32>(value) << endl;
		break;
	case 9:
		os << " with value " << any_cast<float64>(value) << endl;
		break;
	case 10:
		os << " with value " << any_cast<string>(value) << endl;
		break;
	}

	throw CannotSetValue(os.str());
}

void AnyProperty::setValue(boost::any value) {

	if (_ID == 0) {
		_value = value;
		return;
	}

	bool ok = true;

	switch (_typeID) {
	case 1:
		ok = _propertyAccessor->setIcsBooleanValue(_ID, any_cast<bool>(value));
		break;
	case 2:
		ok = _propertyAccessor->setIcsInt16Value(_ID, any_cast<int16>(value));
		break;
	case 3:
		ok = _propertyAccessor->setIcsUint16Value(_ID, any_cast<uint16>(value));
		break;
	case 4:
		ok = _propertyAccessor->setIcsInt32Value(_ID, any_cast<int32>(value));
		break;
	case 5:
		ok = _propertyAccessor->setIcsUint32Value(_ID, any_cast<uint32>(value));
		break;
	case 6:
		ok = _propertyAccessor->setIcsInt64Value(_ID, any_cast<int64>(value));
		break;
	case 7:
		ok = _propertyAccessor->setIcsUint64Value(_ID, any_cast<uint64>(value));
		break;
	case 8:
		ok = _propertyAccessor->setIcsFloat32Value(_ID, any_cast<float32>(value));
		break;
	case 9:
		ok = _propertyAccessor->setIcsFloat64Value(_ID, any_cast<float64>(value));
		break;
	case 10:
		ok = _propertyAccessor->setIcsStringValue(_ID, any_cast<string>(value));
		break;
	case 11:
		_propertyAccessor->setIcsInt16ArrayValue(_ID, any_cast<int16*>(value));
		break;
	case 12:
		_propertyAccessor->setIcsUint16ArrayValue(_ID, any_cast<uint16*>(value));
		break;
	case 13:
		_propertyAccessor->setIcsInt32ArrayValue(_ID, any_cast<int32*>(value));
		break;
	case 14:
		_propertyAccessor->setIcsUint32ArrayValue(_ID, any_cast<uint32*>(value));
		break;
	case 15:
		_propertyAccessor->setIcsInt64ArrayValue(_ID, any_cast<int64*>(value));
		break;
	case 16:
		_propertyAccessor->setIcsUint64ArrayValue(_ID, any_cast<uint64*>(value));
		break;
	case 17:
		_propertyAccessor->setIcsFloat32ArrayValue(_ID, any_cast<float32*>(value));
		break;
	case 18:
		_propertyAccessor->setIcsFloat64ArrayValue(_ID, any_cast<float64*>(value));
		break;
	}

	if (!ok) {
		throwCannotSetValue(_ID, _typeID, _name, value);
	}
}

void AnyProperty::setValueAndFormattedValue(boost::any value, const std::string& formattedValue) {

	if (_ID == 0) {
		_value = value;
		return;
	}

	bool ok = true;

	switch (_typeID) {
	case 1:
		ok = _propertyAccessor->setIcsBooleanValueAndFormattedValue(_ID, any_cast<bool>(value), formattedValue);
		break;
	case 2:
		ok = _propertyAccessor->setIcsInt16ValueAndFormattedValue(_ID, any_cast<int16>(value), formattedValue);
		break;
	case 3:
		ok = _propertyAccessor->setIcsUint16ValueAndFormattedValue(_ID, any_cast<uint16>(value), formattedValue);
		break;
	case 4:
		ok = _propertyAccessor->setIcsInt32ValueAndFormattedValue(_ID, any_cast<int32>(value), formattedValue);
		break;
	case 5:
		ok = _propertyAccessor->setIcsUint32ValueAndFormattedValue(_ID, any_cast<uint32>(value), formattedValue);
		break;
	case 6:
		ok = _propertyAccessor->setIcsInt64ValueAndFormattedValue(_ID, any_cast<int64>(value), formattedValue);
		break;
	case 7:
		ok = _propertyAccessor->setIcsUint64ValueAndFormattedValue(_ID, any_cast<uint64>(value), formattedValue);
		break;
	case 8:
		ok = _propertyAccessor->setIcsFloat32ValueAndFormattedValue(_ID, any_cast<float32>(value), formattedValue);
		break;
	case 9:
		ok = _propertyAccessor->setIcsFloat64ValueAndFormattedValue(_ID, any_cast<float64>(value), formattedValue);
		break;
	case 10:
		ok = _propertyAccessor->setIcsStringValueAndFormattedValue(_ID, any_cast<string>(value), formattedValue);
		break;
	}

	if (!ok) {
		throwCannotSetValue(_ID, _typeID, _name, value);
	}
}

void AnyProperty::update() {

	if (_ID == 0) {
		return;
	}

	_propertyAccessor->updateIcsProperty(_ID);
}

void AnyProperty::setUnit(const std::string& value) {

	_propertyAccessor->setConstantPropertyUnit(_ID, value);
}

void AnyProperty::setUnit(unsigned long propertyID) {

	_propertyAccessor->setPropertyUnit(_ID, propertyID);
}

void AnyProperty::setEnumeratedValues(unsigned long propertyID) {

	_propertyAccessor->setEnumeratedValues(_ID, propertyID);
}

void AnyProperty::setEnumeratedLabels(unsigned long propertyID) {

	_propertyAccessor->setEnumeratedLabels(_ID, propertyID);
}

std::ostream& AnyProperty::toString(std::ostream& s, const boost::any& value, const std::string& format) const {

	string type = getType();

	if (type == INT16_PROPERTY) {
		convert<int16> (s, any_cast<int16> (value), format);
	} else if (type == UINT16_PROPERTY) {
		convert<uint16> (s, any_cast<uint16> (value), format);
	} else if (type == INT32_PROPERTY) {
		convert<int32> (s, any_cast<int32> (value), format);
	} else if (type == UINT32_PROPERTY) {
		convert<uint32> (s, any_cast<uint32> (value), format);
	} else if (type == INT64_PROPERTY) {
		convert<int64> (s, any_cast<int64> (value), format);
	} else if (type == UINT64_PROPERTY) {
		convert<uint64> (s, any_cast<uint64> (value), format);
	} else if (type == FLOAT32_PROPERTY) {
		convert<float32> (s, any_cast<float32> (value), format);
	} else if (type == FLOAT64_PROPERTY) {
		convert<float64> (s, any_cast<float64> (value), format);
	} else if (type == STRING_PROPERTY) {
		convert<string> (s, any_cast<string> (value), format);
	} else if (type == BOOLEAN_PROPERTY) {
		if (any_cast<bool> (value)) {
			s << "true";
		} else {
			s << "false";
		}
	} else if (type == INT16ARRAY_PROPERTY) {
		int16* data = any_cast<int16*> (value);
		for (uint32 i = 0; i < (uint32) getSize(); i++) {
			convert<int16> (s, data[i], format);
			if (i < ((uint32) getSize() - 1)) {
				s << "|";
			}
		}
	} else if (type == UINT16ARRAY_PROPERTY) {
		uint16* data = any_cast<uint16*> (value);
		for (uint32 i = 0; i < (uint32) getSize(); i++) {
			convert<uint16> (s, data[i], format);
			if (i < ((uint32) getSize() - 1)) {
				s << "|";
			}
		}
	} else if (type == INT32ARRAY_PROPERTY) {
		int32* data = any_cast<int32*> (value);
		for (uint32 i = 0; i < (uint32) getSize(); i++) {
			convert<int32> (s, data[i], format);
			if (i < ((uint32) getSize() - 1)) {
				s << "|";
			}
		}
	} else if (type == UINT32ARRAY_PROPERTY) {
		uint32* data = any_cast<uint32*> (value);
		for (uint32 i = 0; i < (uint32) getSize(); i++) {
			convert<uint32> (s, data[i], format);
			if (i < ((uint32) getSize() - 1)) {
				s << "|";
			}
		}
	} else if (type == INT64ARRAY_PROPERTY) {
		int64* data = any_cast<int64*> (value);
		for (uint32 i = 0; i < (uint32) getSize(); i++) {
			convert<int64> (s, data[i], format);
			if (i < ((uint32) getSize() - 1)) {
				s << "|";
			}
		}
	} else if (type == UINT64ARRAY_PROPERTY) {
		uint64* data = any_cast<uint64*> (value);
		for (uint32 i = 0; i < (uint32) getSize(); i++) {
			convert<uint64> (s, data[i], format);
			if (i < ((uint32) getSize() - 1)) {
				s << "|";
			}
		}
	} else if (type == FLOAT32ARRAY_PROPERTY) {
		float32* data = any_cast<float32*> (value);
		for (uint32 i = 0; i < (uint32) getSize(); i++) {
			convert<float32> (s, data[i], format);
			if (i < ((uint32) getSize() - 1)) {
				s << "|";
			}
		}
	} else if (type == FLOAT64ARRAY_PROPERTY) {
		float64* data = any_cast<float64*> (value);
		for (uint32 i = 0; i < (uint32) getSize(); i++) {
			convert<float64> (s, data[i], format);
			if (i < ((uint32) getSize() - 1)) {
				s << "|";
			}
		}
	}
	return s;
}

std::ostream& operator<< (std::ostream& s, const AnyProperty& property) {
	return property.toString(s, property.getValue(), property.getFormat());
}
