/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef ANYPROPERTY_H
#define ANYPROPERTY_H

#include "ErrorsAlarmsManager/Exceptions.h"
#include <boost/any.hpp>
#include <string>
#include <iostream>

namespace command {

class PropertyAccessor;

}

/**
 * Provides an interface to the underlying state variables of Elementary
 * Controllers and ScientificControllers. The idea is to provide the 
 * user with a uniform view of all controllers and provide an interface to
 * configure and control them. The property therefore allows for users
 * to modify values and determine current states.
 */
class AnyProperty {

	friend std::ostream& operator<< (std::ostream&, const AnyProperty&);
public:
	class CannotSetValue: public NotOk {
	public:
		CannotSetValue();
		CannotSetValue(const std::string &reason);
	};


	/**
	 * Constructor aimed at array types. The size can still be modified
	 * using the setSize method.
	 */
	AnyProperty(const std::string& propertyName, const std::string& propertyAlias,
			const std::string& propertyType, bool propertySaveFlag,
			const std::string& propertyFormat, bool propertySpyFlag,
			bool propertyScanFlag, bool propertySaveBoxFlag, int32 propertySize = 1);

	/**
	 * Destructor
	 */
	virtual ~AnyProperty();

	/**
	 * Sets the from type.
	 */
	void setFromType(const std::string& fromType);

	/**
	 * Sets the from type.
	 */
	const std::string& getFromType() const;

	/**
	 * Clones the property.
	 */
	AnyProperty* clone() const;

	/**
	 * Gets the ID.
	 */
	unsigned long getID() const;

	/**
	 * Sets the ID.
	 */
	void init(unsigned long ID, command::PropertyAccessor* propertyAccessor);

	/**
	 * Sets the save attribute.
	 */
	void setSave(bool value);

	/**
	 * Sets the format attribute.
	 */
	void setFormat(const std::string& format);

	/**
	 * Sets the spy attribute.
	 */
	void setSpy(bool value);

	/**
	 * Sets the property scannable.
	 */
	void setScannable(bool value);

	/**
	 * Sets the save box attribute.
	 */
	void setSaveBox(bool value);

	/**
	 * Updates the attributes.
	 */
	void updateAttributes();

	/**
	 * Accessor for the name attribute.
	 */
	const std::string& getName() const;

	/**
	 * Accessor for the name attribute.
	 */
	const std::string& getAlias() const;

	/**
	 * Accessor for the type attribute.
	 */
	const std::string& getType() const;

	/**
	 * Accessor for the save attribute.
	 */
	bool isSave() const;

	/**
	 * Accessor for the save attribute.
	 */
	const std::string& getFormat() const;

	/**
	 * Accessor for the scannable attribute.
	 */
	bool isScannable() const;

	/**
	 * Accessor for the spy attribute.
	 */
	bool isSpy() const;

	/**
	 * Accessor for the save box attribute.
	 */
	bool isSaveBox() const;

	/**
	 * Accessor for the size attribute.
	 */
	int32 getSize() const;

	/**
	 * Modifies the size attribute.
	 */
	void setSize(int32 inputSize);

	/**
	 * Accessor for the Extrasize attribute.
	 */
	int32 getExtraSize() const;

	/**
	 * Modifies the Extrasize attribute.
	 */
	void setExtraSize(int32 inputSize);

	/**
	 * Sets the value of the property.
	 */
	void setValue(boost::any value);

	/**
	 * Sets the value of the property and its formatted value.
	 */
	void setValueAndFormattedValue(boost::any value, const std::string& formattedValue);

	/**
	 * returns the ActualValue of the property
	 */
	boost::any getValue() const;

	/**
	 * Notifies the update of the property.
	 */
	void update();

	/**
	 * Sets the unit constant.
	 */
	void setUnit(const std::string& value);

	/**
	 * Sets the unit with property ID.
	 */
	void setUnit(unsigned long propertyID);

	/**
	 * Sets the enumerated values with property ID.
	 */
	void setEnumeratedValues(unsigned long propertyID);

	/**
	 * Sets the enumerated labels with property ID.
	 */
	void setEnumeratedLabels(unsigned long propertyID);

	/**
	 * Returns the formatted value.
	 */
	std::ostream& toString(std::ostream& s, const boost::any& value, const std::string& format) const;

private:
	template<class T>
	void convert(std::ostream& s, const T& value, const std::string& format) const;

	unsigned long _ID;
	command::PropertyAccessor* _propertyAccessor;
	int32 _typeID;

	std::string _fromType;
	std::string _name;
	std::string _alias;
	std::string _type;
	bool _save;
	std::string _format;
	bool _spy;
	bool _scannable;
	bool _saveBox;
	int32 _extraSize;
	int32 _size;

	boost::any _value;
};

/**
 * Operator to print the value.
 */
std::ostream& operator<< (std::ostream& s, const AnyProperty& property);

static const std::string FLOAT32_PROPERTY = "float32";
static const std::string FLOAT64_PROPERTY = "float64";
static const std::string INT8_PROPERTY = "int8";
static const std::string INT16_PROPERTY = "int16";
static const std::string INT32_PROPERTY = "int32";
static const std::string INT64_PROPERTY = "int64";
static const std::string UINT16_PROPERTY = "uint16";
static const std::string UINT32_PROPERTY = "uint32";
static const std::string UINT64_PROPERTY = "uint64";
static const std::string INT16ARRAY_PROPERTY = "int16array";
static const std::string INT32ARRAY_PROPERTY = "int32array";
static const std::string INT64ARRAY_PROPERTY = "int64array";
static const std::string UINT16ARRAY_PROPERTY = "uint16array";
static const std::string UINT32ARRAY_PROPERTY = "uint32array";
static const std::string UINT64ARRAY_PROPERTY = "uint64array";
static const std::string FLOAT32ARRAY_PROPERTY = "float32array";
static const std::string FLOAT64ARRAY_PROPERTY = "float64array";
static const std::string STRING_PROPERTY = "string";
static const std::string BOOLEAN_PROPERTY = "boolean";

static const std::string DEVICE_CALLER = "Device";
static const std::string PROTOCOL_CALLER = "Protocol";
static const std::string ABSTRACT_CALLER = "Abstract";


#endif //PROPERTY_H
