/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef PROPERTY_INFOS_H
#define PROPERTY_INFOS_H

#include <string>
#include <string.h>
#include <map>
#include <stdexcept>
#include <common/base/BaseTypes.h>

namespace command {

class PropertyAccessor;

}

/**
 * Class defining an exception.
 */
class NoPropertyInfo : public std::runtime_error {

public :
	/**
	 * Constructor.
	 */
	NoPropertyInfo(const std::string & controllerType, const std::string & wantedPropertyName);
};

/**
 * Singleton class to manage property information.
 * Provides information for setpoint and actual relation as well as scannable or spy information.
 */
class PropertyInfos {

public:
	/**
	 * Initialises the instance.
	 */
	static void initInstance();

	/**
	 * Gets the instance.
	 */
	static PropertyInfos * getInstance();

	/**
	 * Item of the map.
	 */
	struct Item {
		std::string actualPropertyName;
		std::string offsetPropertyName;
		std::string shortPropertyName;
		bool isScan;
		bool isSpy;
		bool isDynamic;

		Item() : isScan(false), isSpy(false), isDynamic(false) {}
	};

	/**
	 * Returns true if the property is scannable.
	 */
	bool isScannable(const std::string & controllerType, const std::string & wantedPropertyName) const;

	/**
	 * Gets the property item. Exists if there is a relation between the actual and the setpoint property.
	 */
	Item getItem(const std::string & controllerType, const std::string & wantedPropertyName) const;

	/**
	 * Adds the item.
	 */
	void addItem(const std::string & controllerType, const std::string & wantedPropertyName, const std::string & actualPropertyName, const std::string & offsetPropertyName, const std::string & shortPropertyName, bool isScan, bool isSpy, bool isDynamic);

	/**
	 * Returns the number of properties for a controller type.
	 */
	int32 getNumberOfScannableProperties(const std::string & controllerType) const;

private:
	PropertyInfos();
	void init();

	command::PropertyAccessor* _propertyAccessor;

	static PropertyInfos * _instance;
};

#endif
