/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DataProvider/Property/PropertyManager.h"
#include "DataProvider/Property/AnyProperty.h"
#include "DataProvider/Property/PropertyInfos.h"
#include <command/transportlayer/MultiThreadProcessorCommunication.h>
#include <command/transportlayer/dataprovider/ServantAccessor.h>
#include <command/transportlayer/dataprovider/PropertyAccessor.h>
#include <ics/TypeConverter.h>

using namespace std;
using namespace boost;
using namespace command;

PropertyManager::UnknownProperty::UnknownProperty(const std::string& servantName, const std::string& name) :
		_servantName(servantName), _name(name) {
		_message = "Unknown property ";
		_message += name;
		_message += " in " + _servantName;
}

const char* PropertyManager::UnknownProperty::what() const noexcept {
	return _message.c_str();
}

PropertyManager::CannotSetValue::CannotSetValue(const std::string& name) {
	//_message = "Cannot set value of property ";
	_message += name;
}

const char* PropertyManager::CannotSetValue::what() const noexcept {
	return _message.c_str();
}

PropertyManager::PropertyManager(const std::string& name) :
	_name(name),
	_servantID(0),
	_propertyAccessor(0) {

	if (MultiThreadProcessorCommunication::getInstance() != 0) {
		_servantID = (uint32) MultiThreadProcessorCommunication::getInstance()->getServantAccessor()->getServantID(name);
		_propertyAccessor = MultiThreadProcessorCommunication::getInstance()->getPropertyAccessor();
	}
}

PropertyManager::~PropertyManager() {

	lock_guard<recursive_mutex> lock(_propertiesMutex);

	// delete the properties
	for (vector<AnyProperty*>::iterator p = _properties.begin(); p != _properties.end(); p++) {
		delete *p;
	}

}

PropertyManager* PropertyManager::clone(const std::string& name) const {

	PropertyManager* manager = new PropertyManager(name);

	// copying the properties
	for (vector<AnyProperty*>::const_iterator op = _properties.begin(); op != _properties.end(); op++) {
		// copying the property
		AnyProperty * cloneProperty = (*op)->clone();
		manager->addProperty(cloneProperty);
		cloneProperty->setValue((*op)->getValue());
	}

	return manager;
}

void PropertyManager::renameServant(const std::string& name) {
	_name = name;
}

bool PropertyManager::addProperty(AnyProperty* property) {

	try {
		if (_propertyAccessor != 0) {

			// check type
			string type = convertTypeFromIcs(property->getType());

			DynamicPropertyInfo info = getDynamicPropertyInfo(property->getName());

			uint32 propertyID = 0;
			bool added = false;

			// create the property if it does not exist
			if (info.isDynamic) {
				uint32 dynamicPropertyID = (uint32) _propertyAccessor->addDynamicPropertyToServant(_servantID, info.name, type, property->isSpy());
				propertyID = (uint32) _propertyAccessor->getDynamicPropertyIDAtIndex(dynamicPropertyID, info.index, true);

			} else {
				propertyID = (uint32) _propertyAccessor->addPropertyToServant(_servantID, property->getName(), type, property->isSpy(), property->isSaveBox(), added);
			}

			property->init(propertyID, _propertyAccessor);

			lock_guard<recursive_mutex> lock(_propertiesMutex);

			// insert for the order
			_properties.push_back(property);
			_propertiesMap[property->getName()] = property;

			return added;

		} else {
			// should only occur with tests
			_properties.push_back(property);
			_propertiesMap[property->getName()] = property;
		}

	} catch (...) {
		_properties.push_back(property);
		_propertiesMap[property->getName()] = property;
	}

	return true;
}

unsigned long PropertyManager::addDynamicProperty(const std::string& name, const std::string& type) {

	return _propertyAccessor->addDynamicPropertyToServant(_servantID, name, type, false);
}

AnyProperty * PropertyManager::addProperty(const std::string& fromType, const std::string& name, const std::string& alias, const std::string& type, const std::string& format, bool save, bool spy, bool scan, bool saveBox) {

	//cout << "PropertyManager " << _name << " addProperty " << name << endl;

	if (containsProperty(name)) {

		lock_guard<recursive_mutex> lock(_propertiesMutex);

		AnyProperty * property = _propertiesMap[name];

		if (fromType != property->getFromType()) {
			common::warn << _name << " defines property " << name << " in " << property->getFromType() << " and " << fromType << common::eol;
		}

		return property;
	}

	AnyProperty * property = new AnyProperty(name, alias, type, save, format, spy, scan, saveBox);
	property->setFromType(fromType);
	addProperty(property);

	return property;
}

void PropertyManager::removeProperty(const string& propertyName)  {

	lock_guard<recursive_mutex> lock(_propertiesMutex);

	map<string, AnyProperty*>::iterator p = _propertiesMap.find(propertyName);

	if (p != _propertiesMap.end()) {

		DynamicPropertyInfo info = getDynamicPropertyInfo(propertyName);

		AnyProperty* property = p->second;
		string type = convertTypeFromIcs(property->getType());
		_propertiesMap.erase(p);
		vector<AnyProperty*>::iterator op;
		for (op = _properties.begin(); op != _properties.end(); op++) {
			if ((*op)->getName() == propertyName) {
				_properties.erase(op);
				break;
			}
		}

		if (_propertyAccessor != 0) {

			if (info.isDynamic) {

				// calculating actual size
				int32 newSize = 0;
				for (map<string, AnyProperty*>::const_iterator cp = _propertiesMap.begin(); cp != _propertiesMap.end(); cp++) {

					DynamicPropertyInfo propertyInfo = getDynamicPropertyInfo(cp->first);

					if (propertyInfo.isDynamic && propertyInfo.name == info.name && propertyInfo.index >= newSize) {
						newSize = propertyInfo.index + 1;
					}
				}

				uint32 dynamicPropertyID = (uint32) _propertyAccessor->addDynamicPropertyToServant(_servantID, info.name, type, property->isSpy());
				_propertyAccessor->resizeDynamicProperty(dynamicPropertyID, newSize);

			} else {
				_propertyAccessor->removePropertyFromServant(_servantID, property->getID());
			}
		}

		delete property;
	}
}

AnyProperty* PropertyManager::getProperty(const string& propertyName) const  {

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	map<string, AnyProperty*>::const_iterator p = _propertiesMap.find(propertyName);
	if (p == _propertiesMap.end()) {
		throw UnknownProperty(_name, propertyName);
	}

	return p->second;
}

const string& PropertyManager::getType(const string& propertyName) const  {

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	map<string, AnyProperty*>::const_iterator p = _propertiesMap.find(propertyName);
	if (p == _propertiesMap.end()) {
		throw UnknownProperty(_name, propertyName);
	}

	return p->second->getType();
}

const string& PropertyManager::getAlias(const string& propertyName)	 {

	return getProperty(propertyName)->getAlias();
}

boost::any PropertyManager::getValue(const std::string& propertyName)  {

	lock_guard<recursive_mutex> lock(_propertiesMutex);

	map<string, AnyProperty*>::const_iterator p = _propertiesMap.find(propertyName);
	if (p == _propertiesMap.end()) {
		throw UnknownProperty(_name, propertyName);
	}

	return p->second->getValue();
}

void PropertyManager::setValue(const std::string& propertyName, const boost::any& value)  {

	lock_guard<recursive_mutex> lock(_propertiesMutex);

	map<string, AnyProperty*>::iterator p = _propertiesMap.find(propertyName);
	if (p == _propertiesMap.end()) {
		throw UnknownProperty(_name, propertyName);
	}

	try {
		p->second->setValue(value);
	} catch (AnyProperty::CannotSetValue& e) {
		throw CannotSetValue(e.what());
	}
}

void PropertyManager::setValueAndFormattedValue(const std::string& propertyName, const boost::any& value, const std::string& formattedValue)  {

	lock_guard<recursive_mutex> lock(_propertiesMutex);

	map<string, AnyProperty*>::iterator p = _propertiesMap.find(propertyName);
	if (p == _propertiesMap.end()) {
		throw UnknownProperty(_name, propertyName);
	}

	try {
		p->second->setValueAndFormattedValue(value, formattedValue);
	} catch (AnyProperty::CannotSetValue& e) {
		throw CannotSetValue(e.what());
	}
}

int32 PropertyManager::getSize(const std::string& propertyName) const  {

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	map<string, AnyProperty*>::const_iterator p = _propertiesMap.find(propertyName);
	if (p == _propertiesMap.end()) {
		throw UnknownProperty(_name, propertyName);
	}

	return p->second->getSize();
}

void PropertyManager::setSize(const std::string& propertyName, int32 value)  {

	lock_guard<recursive_mutex> lock(_propertiesMutex);

	map<string, AnyProperty*>::iterator p = _propertiesMap.find(propertyName);
	if (p == _propertiesMap.end()) {
		throw UnknownProperty(_name, propertyName);
	}

	p->second->setSize(value);
}

int32 PropertyManager::getExtraSize(const std::string& propertyName) const  {

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	map<string, AnyProperty*>::const_iterator p = _propertiesMap.find(propertyName);
	if (p == _propertiesMap.end()) {
		throw UnknownProperty(_name, propertyName);
	}

	return p->second->getExtraSize();
}

void PropertyManager::setExtraSize(const std::string& propertyName, int32 value)	 {

	lock_guard<recursive_mutex> lock(_propertiesMutex);

	map<string, AnyProperty*>::iterator p = _propertiesMap.find(propertyName);
	if (p == _propertiesMap.end()) {
		throw UnknownProperty(_name, propertyName);
	}

	p->second->setExtraSize(value);
}

std::string PropertyManager::getPropertyList() const {

	string result = "";

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	for (vector<AnyProperty*>::const_iterator p = _properties.begin(); p != _properties.end(); p++) {
		AnyProperty* property = *p;
		result.append(property->getType() + " " + property->getName() + "\n");
	}

	return result;
}

std::string PropertyManager::getSavePropertyList() const {

	string result = "";

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	for (vector<AnyProperty*>::const_iterator p = _properties.begin(); p != _properties.end(); p++) {
		AnyProperty* property = *p;

		if (property->isSave()) {
			result.append(property->getType() + " " + property->getName() + "\n");
		}
	}

	return result;
}

std::string PropertyManager::getAliasPropertyList() const {

	string result = "";

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	vector<AnyProperty*>::const_iterator p = _properties.begin();
	while (p != _properties.end()) {
		AnyProperty* property = *p;
		if (!property->getAlias().empty()) {
			result.append(property->getName());
			result.append(",");
			result.append(property->getAlias());
			if (++p != _properties.end()) {
				result.append(",");
			}
		}
		else {
			p++;
		}
	}

	return result;
}

std::string PropertyManager::getGenericAliasProperties() const {

	string result = "";

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	vector<AnyProperty*>::const_iterator p = _properties.begin();
	while (p != _properties.end()) {
		AnyProperty* property = *p;
		if (!property->getAlias().empty() && property->isSave()) {
			result.append(property->getName());
			result.append(",");
			result.append(property->getAlias());
			if (++p != _properties.end()) {
				result.append(",");
			}
		}
		else {
			p++;
		}
	}

	return result;
}

std::vector<AnyProperty*> PropertyManager::getSpyPropertyList() const {

	vector<AnyProperty*> result;

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	for (vector<AnyProperty*>::const_iterator p = _properties.begin(); p != _properties.end(); p++) {
		AnyProperty* property = *p;

		if ((property->getType() != UINT16ARRAY_PROPERTY) && (property->getType() != UINT32ARRAY_PROPERTY)
				&& (property->getType() != UINT64ARRAY_PROPERTY) && (property->getType() != INT16ARRAY_PROPERTY)
				&& (property->getType() != INT32ARRAY_PROPERTY) && (property->getType() != INT64ARRAY_PROPERTY)
				&& (property->getType() != FLOAT32ARRAY_PROPERTY) && (property->getType() != FLOAT64ARRAY_PROPERTY)) {
			result.push_back(*p);
		}
	}

	return result;
}

std::vector<AnyProperty*> PropertyManager::getPropertiesToSave() const {
	vector<AnyProperty*> result;

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	for (vector<AnyProperty*>::const_iterator p = _properties.begin(); p != _properties.end(); p++) {

		if ((*p)->isSave()) {
			result.push_back(*p);
		}
	}

	return result;
}

std::vector<AnyProperty*> PropertyManager::getProperties() const {
	vector<AnyProperty*> result;

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	for (vector<AnyProperty*>::const_iterator p = _properties.begin(); p != _properties.end(); p++) {

		if (((*p)->getType() == UINT16ARRAY_PROPERTY) && ((*p)->isSave() != true)) {
			// do not push
		} else if (((*p)->getType() == UINT32ARRAY_PROPERTY) && ((*p)->isSave() != true)) {
			// do not push
		} else if (((*p)->getType() == UINT64ARRAY_PROPERTY) && ((*p)->isSave() != true)) {
			// do not push
		} else if (((*p)->getType() == INT16ARRAY_PROPERTY) && ((*p)->isSave() != true)) {
			// do not push
		} else if (((*p)->getType() == INT32ARRAY_PROPERTY) && ((*p)->isSave() != true)) {
			// do not push
		} else if (((*p)->getType() == INT64ARRAY_PROPERTY) && ((*p)->isSave() != true)) {
			// do not push
		} else if (((*p)->getType() == FLOAT32ARRAY_PROPERTY) && ((*p)->isSave() != true)) {
			// do not push
		} else if (((*p)->getType() == FLOAT64ARRAY_PROPERTY) && ((*p)->isSave() != true)) {
			// do not push
		} else {
			result.push_back(*p);
		}
	}

	return result;
}

std::vector<AnyProperty*> PropertyManager::getAllProperties() const {
	vector<AnyProperty*> result;

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	for (vector<AnyProperty*>::const_iterator p = _properties.begin(); p != _properties.end(); p++) {
		result.push_back(*p);
	}

	return result;
}

std::vector<std::string> PropertyManager::getLocalProperties() const {
	vector<string> result;

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	for (vector<AnyProperty*>::const_iterator p = _properties.begin(); p != _properties.end(); p++) {
		result.push_back((*p)->getName());
	}

	return result;
}

bool PropertyManager::containsProperty(const std::string& propertyName) {

	lock_guard<recursive_mutex> lock(_propertiesMutex);

	if (_propertiesMap.find(propertyName) == _propertiesMap.end()) {
		return false;
	}

	return true;
}

bool PropertyManager::isManyScanProperties() const {
	return false;
}

std::string PropertyManager::getPropertyName(uint32 propertyID) const {

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	for (map<string, AnyProperty*>::const_iterator p = _propertiesMap.begin(); p != _propertiesMap.end(); p++) {
		if (p->second->getID() == propertyID) {
			return p->first;
		}
	}

	throw UnknownProperty(_name, "?");
}

uint32 PropertyManager::getPropertyID(const std::string& propertyName) const {
	if (_propertyAccessor != 0) {
		return _propertyAccessor->getPropertyID(_servantID, propertyName);
	}
	return 0;
}

uint32 PropertyManager::getServantID() const {
	return _servantID;
}

void PropertyManager::addServantType(const std::string& type, bool atFront) {

	if (_propertyAccessor != 0) {
		_propertyAccessor->addServantType(_servantID, type, atFront);
	}
}

PropertyManager::DynamicPropertyInfo PropertyManager::getDynamicPropertyInfo(const std::string& propertyName) {

	DynamicPropertyInfo result;

	size_t dotIndex = propertyName.find_last_of(".");

	if (dotIndex != string::npos) {
		result.isDynamic = true;
		result.name = propertyName.substr(0, dotIndex);

		istringstream is(propertyName.substr(dotIndex + 1));
		is >> result.index;

	} else {
		result.isDynamic = false;
	}

	return result;
}

std::map<std::string, std::string> PropertyManager::getAliasPropertyMap() const {

	lock_guard<recursive_mutex> lock(const_cast<std::recursive_mutex&>(_propertiesMutex));

	map<string, string> buf;

	// iterate through the list of properties returning the data type
	// and name of property - all of these going into one string
	vector<AnyProperty*>::const_iterator it = _properties.begin();
	while (it != _properties.end()) {
		AnyProperty* property = *it;
		if (property->getAlias().empty() == false) {
			if (buf.find(property->getAlias()) == buf.end()) {
				buf[property->getAlias()] = property->getName();
			}
		}
		it++;
	}
	return buf;
}

void PropertyManager::sendArrayEvent(const std::string& propertyName) {
	getProperty(propertyName)->update();
}

void PropertyManager::setEnumeratedValues(unsigned long propertyID, unsigned long valuesPropertyID) const {
	_propertyAccessor->setEnumeratedValues(propertyID, valuesPropertyID);
}

void PropertyManager::setEnumeratedLabels(unsigned long propertyID, unsigned long labelsPropertyID) const {
	_propertyAccessor->setEnumeratedLabels(propertyID, labelsPropertyID);
}

void PropertyManager::setServantInDebug() {
	_propertyAccessor->setServantInDebug(_servantID);
}
