/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef PROPERTYMANAGER_H
#define PROPERTYMANAGER_H

#include "ErrorsAlarmsManager/Exceptions.h"
#include <boost/any.hpp>
#include <mutex>
#include <vector>
#include <map>

namespace command {

class PropertyAccessor;

}

class AnyProperty;
class PropertyDescription;

/**
 * This class is a manager of the properties for a device driver ao a controller.
 */
class PropertyManager {

	friend class PropertyRefresher;
	friend class PropertyUpdater;

public:
	class UnknownProperty : public std::exception {
	public:
		UnknownProperty(const std::string& servantName, const std::string& name);

		std::string getServantName() const;
		std::string getName() const;

		virtual const char* what() const noexcept;

	private:
		std::string _servantName;
		std::string _name;
		std::string _message;
	};

	class BadTypeProperty {};

	class CannotSetValue : public std::exception {
	public:
		CannotSetValue(const std::string& name);

		virtual const char* what() const noexcept;

	private:
		std::string _message;
	};

	PropertyManager(const std::string& name);
	virtual ~PropertyManager();

	PropertyManager* clone(const std::string& name) const;

	void renameServant(const std::string& name);

	bool addProperty(AnyProperty* property);
	AnyProperty * addProperty(const std::string& fromType, const std::string& name, const std::string& alias, const std::string& type, const std::string& format, bool save, bool spy, bool scan, bool saveBox);
	unsigned long addDynamicProperty(const std::string& name, const std::string& type);
	void removeProperty(const std::string& propertyName) ;

	AnyProperty* getProperty(const std::string& propertyName) const ;
	const std::string& getType(const std::string& propertyName) const ;
	const std::string& getAlias(const std::string& propertyName) ;

	boost::any getValue(const std::string& propertyName) ;
	void setValue(const std::string& propertyName, const boost::any& aValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, const boost::any& aValue, const std::string& formattedValue) ;

	int32 getSize(const std::string& propertyName) const ;
	void setSize(const std::string& propertyName, int32 value) ;
	int32 getExtraSize(const std::string& propertyName) const ;
	void setExtraSize(const std::string& propertyName, int32 value)	;

	std::string getPropertyList() const;
	std::string getSavePropertyList() const;
	std::string getAliasPropertyList() const;
	std::string getGenericAliasProperties() const;
	std::vector<AnyProperty*> getSpyPropertyList() const;
	std::vector<AnyProperty*> getPropertiesToSave() const;
	std::vector<AnyProperty*> getProperties() const;
	std::vector<AnyProperty*> getAllProperties() const;
	std::vector<std::string> getLocalProperties() const;
	std::map<std::string, std::string> getAliasPropertyMap() const;

	bool containsProperty(const std::string& propertyName);

	bool isManyScanProperties() const;

	std::string getPropertyName(uint32 propertyID) const;

	uint32 getPropertyID(const std::string& propertyName) const;

	/**
	 * Returns the servant ID. Added method.
	 */
	uint32 getServantID() const;

	/**
	 * Adds the servant type.
	 */
	void addServantType(const std::string& type, bool atFront);

	/**
	 * Sends an array event.
	 */
	void sendArrayEvent(const std::string& propertyName);

	/**
	 * Structure for dynamic properties
	 */
	struct DynamicPropertyInfo {
		bool isDynamic;
		std::string name;
		int32 index;
	};

	/**
	 * Gets the dynamic properties info.
	 */
	static DynamicPropertyInfo getDynamicPropertyInfo(const std::string& propertyName);

	/**
	 * Sets the enumerated values property.
	 */
	void setEnumeratedValues(unsigned long propertyID, unsigned long valuesPropertyID) const;

	/**
	 * Sets the enumerated labels property.
	 */
	void setEnumeratedLabels(unsigned long propertyID, unsigned long labelsPropertyID) const;

	/**
	 * Sets the servant in debug.
	 */
	void setServantInDebug();

private:
	std::string _name;
	uint32 _servantID;

	command::PropertyAccessor* _propertyAccessor;

	std::recursive_mutex _propertiesMutex;

	std::vector<AnyProperty*> _properties;
	std::map<std::string, AnyProperty*> _propertiesMap;
};

#endif
