/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "BaseDynamicProperty.h"
#include "DataProvider/Property/PropertyInfos.h"
#include <sstream>
#include <boost/lexical_cast.hpp>

using namespace common;

std::string propertyNameAtIndex(const std::string& name, int32 index) {
	std::ostringstream os;
	os << name << "." << index;

	return os.str();
}

std::string propertyNameAtIndex(const std::string& name, int32 index1, int32 index2) {
	std::ostringstream os;
	os << name << "." << index1 << "." << index2;

	return os.str();
}

BaseDynamicProperty::BadIndexException::BadIndexException(const std::string& name, int32 index)
: Exception("BaseDynamicProperty::BadIndexException", std::string("bad index <") + boost::lexical_cast<std::string>(index) + "> for " + name) {
}

BaseDynamicProperty::BaseDynamicProperty() :
	m_propertyManager(0),
	m_spy(false),
	m_save(false),
	m_scan(false),
	m_hex(false),
	m_logLevel(-1),
	m_copied(false) {
}

BaseDynamicProperty::BaseDynamicProperty(const BaseDynamicProperty & property) :
	m_name(property.m_name),
	m_alias(property.m_alias),
	m_propertyManager(property.m_propertyManager),
	m_spy(property.m_spy),
	m_save(property.m_save),
	m_scan(property.m_scan),
	m_hex(property.m_hex),
	m_logLevel(property.m_logLevel),
	m_copied(property.m_copied) {
}

void BaseDynamicProperty::init(PropertyManager * propertyManager, int32 flags, const std::string& name, const std::string& alias) {
	m_name = name;
	m_alias = alias;
	m_propertyManager = propertyManager;

	m_spy = ((SPY & flags) == 1);
	m_save = (((SAVE & flags) >> 1) == 1);
	m_scan = (((SCAN & flags) >> 2) == 1);
	m_hex = (((HEX & flags) >> 4) == 1);
	bool logInfo = (((::LOG_INFO & flags) >> 4) == 1);
	bool logDebug = (((::LOG_DEBUG & flags) >> 5) == 1);

	if (logInfo) {
		m_logLevel = Level::s_Info;
	}
	else if (logDebug) {
		m_logLevel = Level::s_Debug;
	}
}

void BaseDynamicProperty::addPropertyInfo(const std::string & controllerType, const std::string & wantedPropertyName, const std::string & actualPropertyName, const std::string & shortPropertyName) {
	PropertyInfos::getInstance()->addItem(controllerType, wantedPropertyName, actualPropertyName, "", shortPropertyName, false, false, true);
}

void BaseDynamicProperty::copy(PropertyManager * propertyManager, const BaseDynamicProperty & property) {
	m_name = property.m_name;
	m_alias = property.m_alias;
	m_propertyManager = propertyManager;

	m_spy = property.m_spy;
	m_save = property.m_save;
	m_scan = property.m_scan;
	m_hex = property.m_hex;

	m_logLevel = property.m_logLevel;
	m_copied = true;
}

const std::string& BaseDynamicProperty::getName() const {
	return m_name;
}

const std::string& BaseDynamicProperty::getAlias() const {
	return m_alias;
}

bool BaseDynamicProperty::isLogging() const {
	return m_logLevel.getLevel() != -1 && !m_copied;
}
