/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef BASEDYNAMICPROPERTY_T_H
#define BASEDYNAMICPROPERTY_T_H

#include "ValueProperty.h"
#include "DynamicRefresher.h"
#include "common/base/Exception.h"

std::string propertyNameAtIndex(const std::string& name, int32 index);
std::string propertyNameAtIndex(const std::string& name, int32 index1, int32 index2);

template<typename Type>
struct BaseServantDynamicProperty {

	virtual ~BaseServantDynamicProperty() {}

	virtual BaseServantDynamicProperty<Type> * clone() const = 0;
	virtual std::string getServantType() const = 0;
	virtual std::string getServantName() const = 0;
	virtual std::string getName() const = 0;
	virtual Level getLogLevel(Level const & level) const = 0;

	virtual void setValue(int32 index, const Type& value) = 0;
	virtual void setValue(int32 index, const Type& value, const std::string& formattedValue) = 0;
	virtual void setValue(int32 index1, int32 index2, const Type& value) = 0;
	virtual void setValue(int32 index1, int32 index2, const Type& value, const std::string& formattedValue) = 0;

	virtual void updateValue(int32 index, const Type& value) = 0;
	virtual void updateValue(int32 index, const Type& value, const std::string& formattedValue) = 0;
	virtual void updateValue(int32 index1, int32 index2, const Type& value) = 0;
	virtual void updateValue(int32 index1, int32 index2, const Type& value, const std::string& formattedValue) = 0;

	virtual Type getValue(int32 index) = 0;
	virtual Type getValue(int32 index1, int32 index2) = 0;

	virtual void setSize(int32 index, int32 size) = 0;
	virtual void setSize(int32 index1, int32 index2, int32 size) = 0;

	virtual int32 getSize(int32 index) = 0;
	virtual int32 getSize(int32 index1, int32 index2) = 0;

	virtual void sendEvent(int32 index) = 0;
	virtual void sendEvent(int32 index1, int32 index2) = 0;

};

template<typename ServantType, typename Type>
struct ServantDynamicProperty : public BaseServantDynamicProperty<Type> {

	ServantDynamicProperty(ServantType * servant, const std::string& name) :
		m_servant(servant), m_name(name) {}

	virtual ~ServantDynamicProperty() {}

	virtual BaseServantDynamicProperty<Type> * clone() const {
		return new ServantDynamicProperty<ServantType, Type>(*this);
	}

	virtual std::string getServantType() const {
		return m_servant->getType();
	}

	virtual std::string getServantName() const {
		return m_servant->getName();
	}

	virtual std::string getName() const {
		return m_name;
	}

	virtual Level getLogLevel(Level const & level) const {
		return m_servant->getLogLevel(level);
	}

	virtual void setValue(int32 index, const Type& value) {
		m_servant->setValue(propertyNameAtIndex(m_name, index), value);
	}

	virtual void setValue(int32 index, const Type& value, const std::string& formattedValue) {
		m_servant->setValueAndFormattedValue(propertyNameAtIndex(m_name, index), value, formattedValue);
	}

	virtual void setValue(int32 index1, int32 index2, const Type& value) {
		m_servant->setValue(propertyNameAtIndex(m_name, index1, index2), value);
	}

	virtual void setValue(int32 index1, int32 index2, const Type& value, const std::string& formattedValue) {
		m_servant->setValueAndFormattedValue(propertyNameAtIndex(m_name, index1, index2), value, formattedValue);
	}

	virtual void updateValue(int32 index, const Type& value) {
		m_servant->updateValue(propertyNameAtIndex(m_name, index), value);
	}

	virtual void updateValue(int32 index, const Type& value, const std::string& formattedValue) {
		m_servant->updateValueAndFormattedValue(propertyNameAtIndex(m_name, index), value, formattedValue);
	}

	virtual void updateValue(int32 index1, int32 index2, const Type& value) {
		m_servant->updateValue(propertyNameAtIndex(m_name, index1, index2), value);
	}

	virtual void updateValue(int32 index1, int32 index2, const Type& value, const std::string& formattedValue) {
		m_servant->updateValueAndFormattedValue(propertyNameAtIndex(m_name, index1, index2), value, formattedValue);
	}

	virtual Type getValue(int32 index) {
		return boost::any_cast<Type>(m_servant->getValue(propertyNameAtIndex(m_name, index)));
	}

	virtual Type getValue(int32 index1, int32 index2) {
		return boost::any_cast<Type>(m_servant->getValue(propertyNameAtIndex(m_name, index1, index2)));
	}
	
	virtual void setSize(int32 index, int32 size) {
		m_servant->setSize(propertyNameAtIndex(m_name, index), size);
	}

	virtual void setSize(int32 index1, int32 index2, int32 size) {
		m_servant->setSize(propertyNameAtIndex(m_name, index1, index2), size);
	}

	virtual int32 getSize(int32 index) {
		return m_servant->getSize(propertyNameAtIndex(m_name, index));
	}

	virtual int32 getSize(int32 index1, int32 index2) {
		return m_servant->getSize(propertyNameAtIndex(m_name, index1, index2));
	}

	virtual void sendEvent(int32 index) {
		m_servant->sendArrayEvent(propertyNameAtIndex(m_name, index));
	}

	virtual void sendEvent(int32 index1, int32 index2) {
		m_servant->sendArrayEvent(propertyNameAtIndex(m_name, index1, index2));
	}

	ServantType * m_servant;
	std::string m_name;
};

class BaseDynamicProperty {

public:
	/**
	 * Exception for servant enable.
	 */
	class BadIndexException : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		BadIndexException(const std::string& name, int32 index);
	};

	/**
	 * Constructor.
	 */
	BaseDynamicProperty();

	/**
	 * Copy constructor is copying each member and clones the servant property member.
	 */
	BaseDynamicProperty(const BaseDynamicProperty & property);
	virtual ~BaseDynamicProperty() {}

	void addPropertyInfo(const std::string & controllerType, const std::string & wantedPropertyName, const std::string & actualPropertyName, const std::string & shortPropertyName);

	void copy(PropertyManager * propertyManager, const BaseDynamicProperty & property);
	const std::string & getName() const;
	const std::string & getAlias() const;

	virtual void formatValues() = 0;

protected:
	void init(PropertyManager * propertyManager, int32 flags, const std::string& name, const std::string& alias);
	bool isLogging() const;

	std::string m_name;
	std::string m_alias;
	PropertyManager * m_propertyManager;
	bool m_spy;
	bool m_save;
	bool m_scan;
	bool m_hex;
	Level m_logLevel;
	bool m_copied;
};

#endif
