/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "BaseProperty.h"
#include "DataProvider/Property/PropertyInfos.h"

using namespace std;

const std::string PropertyTypeToString<float32>::value = FLOAT32_PROPERTY;
const std::string PropertyTypeToString<float64>::value = FLOAT64_PROPERTY;
const std::string PropertyTypeToString<int16>::value = INT16_PROPERTY;
const std::string PropertyTypeToString<int32>::value = INT32_PROPERTY;
const std::string PropertyTypeToString<int64>::value = INT64_PROPERTY;
const std::string PropertyTypeToString<uint16>::value = UINT16_PROPERTY;
const std::string PropertyTypeToString<uint32>::value = UINT32_PROPERTY;
const std::string PropertyTypeToString<uint64>::value = UINT64_PROPERTY;
const std::string PropertyTypeToString<int16 *>::value = INT16ARRAY_PROPERTY;
const std::string PropertyTypeToString<int32 *>::value = INT32ARRAY_PROPERTY;
const std::string PropertyTypeToString<int64 *>::value = INT64ARRAY_PROPERTY;
const std::string PropertyTypeToString<uint16 *>::value = UINT16ARRAY_PROPERTY;
const std::string PropertyTypeToString<uint32 *>::value = UINT32ARRAY_PROPERTY;
const std::string PropertyTypeToString<uint64 *>::value = UINT64ARRAY_PROPERTY;
const std::string PropertyTypeToString<float32 *>::value = FLOAT32ARRAY_PROPERTY;
const std::string PropertyTypeToString<float64 *>::value = FLOAT64ARRAY_PROPERTY;
const std::string PropertyTypeToString<string>::value = STRING_PROPERTY;
const std::string PropertyTypeToString<bool>::value = BOOLEAN_PROPERTY;

BaseProperty::BaseProperty() :
	m_property(0),
	m_qualifier(ACTUAL),
	m_logLevel(-1),
	m_copied(false) {
}

BaseProperty::BaseProperty(const BaseProperty & property) :
	m_property(property.m_property),
	m_servantType(property.m_servantType),
	m_qualifier(property.m_qualifier),
	m_logLevel(property.m_logLevel),
	m_copied(property.m_copied) {
}

void BaseProperty::init(const std::string& fromType, PropertyManager * propertyManager, const std::string& name, const std::string& alias, const std::string& type, int32 flags) {
	// analyse Flags
	bool spy = ((SPY & flags) == 1);
	bool save = (((SAVE & flags) >> 1) == 1);
	bool scan = (((SCAN & flags) >> 2) == 1);
	bool hex = (((HEX & flags) >> 3) == 1);
	bool logInfo = (((LOG_INFO & flags) >> 4) == 1);
	bool logDebug = (((LOG_DEBUG & flags) >> 5) == 1);
	bool saveBox = (((SAVE_BOX & flags) >> 6) == 1);

	if (logInfo) {
		m_logLevel = Level::s_Info;
	}
	else if (logDebug) {
		m_logLevel = Level::s_Debug;
	}

	string format;
	if (hex) {
		format = "%x";
	}

	// add a new property
	m_property = propertyManager->addProperty(fromType, name, alias, type, format, save, spy, scan, saveBox);
	m_qualifier = ACTUAL;
}

void BaseProperty::copy(PropertyManager * propertyManager, const BaseProperty & property) {
	m_property = propertyManager->getProperty(property.getName());
	m_servantType = property.m_servantType;
	m_qualifier = property.m_qualifier;
	m_logLevel = property.m_logLevel;
	m_copied = true;
}

void BaseProperty::setFlags(int32 flags) {

	if (m_property == 0) {
		return;
	}

	bool spy = ((SPY & flags) == 1);
	bool save = (((SAVE & flags) >> 1) == 1);
	bool scan = (((SCAN & flags) >> 2) == 1);
	bool hex = (((HEX & flags) >> 3) == 1);
	bool logInfo = (((LOG_INFO & flags) >> 4) == 1);
	bool logDebug = (((LOG_DEBUG & flags) >> 5) == 1);
	bool saveBox = (((SAVE_BOX & flags) >> 6) == 1);

	if (logInfo) {
		m_logLevel = Level::s_Info;
	}
	else if (logDebug) {
		m_logLevel = Level::s_Debug;
	}

	m_property->setSave(save);
	m_property->setSpy(spy);
	m_property->setSaveBox(saveBox);

	string format;
	if (hex) {
		format = "%x";
		m_property->setFormat(format);
	}

	// re-add the property
	m_property->updateAttributes();
}

void BaseProperty::addPropertyInfo(const std::string & controllerType, const std::string & wantedPropertyName, const std::string & actualPropertyName, const std::string & offsetPropertyName, const std::string & shortPropertyName, bool isScan, bool isSpy) {
	PropertyInfos::getInstance()->addItem(controllerType, wantedPropertyName, actualPropertyName, offsetPropertyName, shortPropertyName, isScan, isSpy, false);
}

void BaseProperty::setQualifier(BaseProperty::QualifierType qualifier) {
	m_qualifier = qualifier;
}

BaseProperty::QualifierType BaseProperty::getQualifier() const {
	return m_qualifier;
}

unsigned long BaseProperty::getID() const {
	return m_property->getID();
}

const std::string& BaseProperty::getName() const {
	return m_property->getName();
}

const std::string& BaseProperty::getAlias() const {
	return m_property->getAlias();
}

const std::string& BaseProperty::getType() const {
	return m_property->getType();
}

boost::any BaseProperty::getValue() const {
	return m_property->getValue();
}

std::string BaseProperty::getUnit() const {
	return "";
}

bool BaseProperty::isLogging() const {
	return m_logLevel.getLevel() != -1 && !m_copied;
}
