/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef BASEPROPERTY_T_H
#define BASEPROPERTY_T_H

#include <DataProvider/Property/AnyProperty.h>
#include <DataProvider/Property/PropertyManager.h>
#include <LoggingServices/Level.h>
#include <common/base/LogElement.h>
#include "AliasString.h"

// Flags
static const int32 NOSAVE = 0;
static const int32 SPY  = 1;       // 1u << 0
static const int32 SAVE = 2;       // 1u << 1
static const int32 SCAN = 4;       // 1u << 2
static const int32 HEX = 8;        // 1u << 3
static const int32 LOG_INFO = 16;  // 1u << 4
static const int32 LOG_DEBUG = 32; // 1u << 5
static const int32 SAVE_BOX = 64; // 1u << 6

static const int32 SETPOINT = 1;
static const int32 OFFSET = 2;

class BaseProperty {

	friend class Unit;

public:
	enum QualifierType {ACTUAL, SETPOINT, OFFSET};

	BaseProperty();
	BaseProperty(const BaseProperty & property);
	void copy(PropertyManager * propertyManager, const BaseProperty & property);

	virtual ~BaseProperty() {}

	void addPropertyInfo(const std::string & controllerType, const std::string & wantedPropertyName, const std::string & actualPropertyName, const std::string & offsetPropertyName, const std::string & shortPropertyName, bool isScan, bool isSpy);
	void setQualifier(QualifierType qualifier);
	void setFlags(int32 flags);

	unsigned long getID() const;
	const std::string & getName() const;
	const std::string & getAlias() const;
	const std::string & getType() const;
	QualifierType getQualifier() const;
	boost::any getValue() const;

	virtual std::string getUnit() const;
	virtual void formatValue() = 0;
	virtual common::LogElement toPropertyLog(bool hidden = false) const = 0;

protected:
	void init(const std::string& fromType, PropertyManager * propertyManager, const std::string& name, const std::string& alias, const std::string& type, int32 flags);
	bool isLogging() const;

	AnyProperty * m_property;
	std::string m_servantType;
	QualifierType m_qualifier;
	Level m_logLevel;
	bool m_copied;
};

template<typename Type>
struct PropertyTypeToString;

template<>
struct PropertyTypeToString<float32> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<float64> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<int16> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<int32> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<int64> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<uint16> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<uint32> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<uint64> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<uint16 *> {
	static const std::string value;
};
template<>
struct PropertyTypeToString<uint32 *> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<uint64 *> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<int16 *> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<int32 *> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<int64 *> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<float32 *> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<float64 *> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<std::string> {
	static const std::string value;
};

template<>
struct PropertyTypeToString<bool> {
	static const std::string value;
};

template<typename Type>
struct BaseServantProperty {

	virtual ~BaseServantProperty() {}

	virtual BaseServantProperty<Type> * clone() const = 0;
	virtual std::string getServantType() const = 0;
	virtual std::string getServantName() const = 0;
	virtual Level getLogLevel(Level const & level) const = 0;
	virtual void setValue(const Type& value) = 0;
	virtual void setValue(const Type& value, const std::string& formattedValue) = 0;
	virtual void updateValue(const Type& value) = 0;
	virtual void updateValue(const Type& value, const std::string& formattedValue) = 0;
	virtual Type getValue() = 0;
	virtual void setSize(int32  size) = 0;
	virtual int32 getSize() = 0;
	virtual void sendArrayEvent() = 0;
	virtual void setColumnSize(int32 size) = 0;
	virtual int32 getColumnSize() = 0;
};

template<typename ServantType, typename Type>
struct ServantProperty : public BaseServantProperty<Type> {

	ServantProperty(ServantType * servant, const std::string& name) :
		m_servant(servant), m_name(name) {

		// add the type
		m_servant->addType(ServantType::TYPE);
	}

	virtual ~ServantProperty() {}

	virtual BaseServantProperty<Type> * clone() const {
		return new ServantProperty<ServantType, Type>(*this);
	}

	virtual std::string getServantType() const {
		return m_servant->getType();
	}

	virtual std::string getServantName() const {
		return m_servant->getName();
	}

	virtual Level getLogLevel(Level const & level) const {
		return m_servant->getLogLevel(level);
	}

	virtual void setValue(const Type& value) {
		m_servant->setValue(m_name, value);
	}

	virtual void setValue(const Type& value, const std::string& formattedValue) {
		m_servant->setValueAndFormattedValue(m_name, value, formattedValue);
	}

	virtual void updateValue(const Type& value) {
		m_servant->updateValue(m_name, value);
	}

	virtual void updateValue(const Type& value, const std::string& formattedValue) {
		m_servant->updateValueAndFormattedValue(m_name, value, formattedValue);
	}

	virtual Type getValue() {
		return boost::any_cast<Type>(m_servant->getValue(m_name));
	}

	virtual void setSize(int32 size) {
		m_servant->setSize(m_name, size);
	}

	virtual int32 getSize() {
		return m_servant->getSize(m_name);
	}

	virtual void sendArrayEvent() {
		m_servant->sendArrayEvent(m_name);
	}

	virtual void setColumnSize(int32 size) {
		m_servant->setExtraSize(m_name, size);
	}

	virtual int32 getColumnSize() {
		return m_servant->getExtraSize(m_name);
	}

	ServantType * m_servant;
	std::string m_name;
};

#endif
