/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DYNAMICARRAYPROPERTY_T_H
#define DYNAMICARRAYPROPERTY_T_H

#include "ValueDynamicProperty.h"

template<typename Type>
class DynamicArrayProperty : public ValueDynamicProperty<Type *> {

public:
	typedef Type * TypePtr;

	using ValueDynamicProperty<Type *>::m_size;
	using ValueDynamicProperty<Type *>::m_fromType;
	using ValueDynamicProperty<Type *>::m_servantProperty;

	// Do not use using because of CDT bug from version 11.5 causing a systematic failure of the indexing
	//using BaseDynamicProperty::m_name;
	//using BaseDynamicProperty::m_propertyManager;
	//using BaseDynamicProperty::m_spy;
	//using BaseDynamicProperty::m_save;
	//using BaseDynamicProperty::m_scan;
	//using BaseDynamicProperty::m_hex;

	DynamicArrayProperty() : ValueDynamicProperty<Type *>(), m_columnSize(1) {}

	template<typename ServantType, int aliasLength>
	void init(ServantType * servant, int32 flags, const std::string& name, const char(&aliasName)[aliasLength]) {

		std::string alias = aliasName;

		// Do not check for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			alias = checkString<EXECUTION_WARNING, 0, 10>(name, aliasName);
			if (formatAlias(alias)) {
				common::warn << "Alias for '" << name << "' renamed into '" << alias << "'" << common::eol;
			}
		}

		ValueDynamicProperty<Type *>::init(servant, flags, name, alias);
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name) {
		init(servant, flags, name, "");
	}

	void resize(int32 value);

	void setSize(int32 index, int32 size);
	int32 getSize(int32 index);
	void sendEvent(int32 index);

	void setColumnSize(int32 size);
	int32 getColumnSize();

	void set(int32 index, const DynamicArrayProperty<Type>::TypePtr& value);
	void update(int32 index, const DynamicArrayProperty<Type>::TypePtr& value);
	virtual void formatValues();

protected:
	int32 m_columnSize;

private:
	void operator=(const DynamicArrayProperty<Type>&);
};

template<typename Type>
void DynamicArrayProperty<Type>::resize(int32 newSize) {

	for (int32 i = m_size; i < newSize; ++i) {

		std::string format;
		if (BaseDynamicProperty::m_hex) {
			format = "%x";
		}

		std::string propertyName = propertyNameAtIndex(BaseDynamicProperty::m_name, i);
		if (!BaseDynamicProperty::m_propertyManager->containsProperty(propertyName)) {
			AnyProperty * property = BaseDynamicProperty::m_propertyManager->addProperty(m_fromType, propertyName, "", PropertyTypeToString<Type *>::value, format, BaseDynamicProperty::m_save, BaseDynamicProperty::m_spy, BaseDynamicProperty::m_scan, false);
			property->setExtraSize(m_columnSize);
		}
	}

	for (int32 i = newSize; i < m_size; ++i) {
		BaseDynamicProperty::m_propertyManager->removeProperty(propertyNameAtIndex(BaseDynamicProperty::m_name, i));
	}

	m_size = newSize;
}

template<typename Type>
void DynamicArrayProperty<Type>::setSize(int32 index, int32 size) {
	this->m_servantProperty->setSize(index, size);
}

template<typename Type>
int32 DynamicArrayProperty<Type>::getSize(int32 index) {
	return this->m_servantProperty->getSize(index);
}

template<typename Type>
void DynamicArrayProperty<Type>::sendEvent(int32 index) {
	this->m_servantProperty->sendEvent(index);
}

template<typename Type>
void DynamicArrayProperty<Type>::setColumnSize(int32 size) {
	m_columnSize = size;
}

template<typename Type>
int32 DynamicArrayProperty<Type>::getColumnSize() {
	return m_columnSize;
}

template<typename Type>
void DynamicArrayProperty<Type>::set(int32 index, const DynamicArrayProperty<Type>::TypePtr& value) {
	m_servantProperty->setValue(index, value);
}

template<typename Type>
void DynamicArrayProperty<Type>::update(int32 index, const DynamicArrayProperty<Type>::TypePtr& value) {
	m_servantProperty->updateValue(index, value);
}

template<typename Type>
void DynamicArrayProperty<Type>::formatValues() {
	// do nothing in case of array
}

#endif
