/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DYNAMICARRAYPROPERTY2_T_H
#define DYNAMICARRAYPROPERTY2_T_H

#include "ValueDynamicProperty2.h"

template<typename Type>
class DynamicArrayProperty2 : public ValueDynamicProperty2<Type *> {

public:
	typedef Type * TypePtr;

	using ValueDynamicProperty2<Type *>::m_size;
	using ValueDynamicProperty2<Type *>::m_fromType;
	using ValueDynamicProperty2<Type *>::m_servantProperty;
	using ValueDynamicProperty2<Type *>::resize;
	using ValueDynamicProperty2<Type *>::getSize;

	// Do not use using because of CDT bug from version 11.5 causing a systematic failure of the indexing
	//using BaseDynamicProperty::m_name;
	//using BaseDynamicProperty::m_propertyManager;
	//using BaseDynamicProperty::m_spy;
	//using BaseDynamicProperty::m_save;
	//using BaseDynamicProperty::m_scan;
	//using BaseDynamicProperty::m_hex;


	DynamicArrayProperty2() :
		ValueDynamicProperty2<Type *>(), m_columnSize(1) {
	}

	template<typename ServantType, int aliasLength>
	void init(ServantType * servant, int32 flags, const std::string& name, const char(&aliasName)[aliasLength]) {

		std::string alias = aliasName;

		// Do not check for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			alias = checkString<EXECUTION_WARNING, 0, 10>(name, aliasName);
			if (formatAlias(alias)) {
				common::warn << "Alias for '" << name << "' renamed into '" << alias << "'" << common::eol;
			}
		}

		ValueDynamicProperty2<Type *>::init(servant, flags, name, alias);
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name) {
		init(servant, flags, name, "");
	}

	void resize(int32 index1, int32 value);

	void setSize(int32 index1, int32 index2, int32 size);
	int32 getSize(int32 index1, int32 index2);
	void sendEvent(int32 index1, int32 index2);

	void setColumnSize(int32 size);
	int32 getColumnSize();

	void set(int32 index1, int32 index2, const DynamicArrayProperty2<Type>::TypePtr& value);
	void update(int32 index1, int32 index2, const DynamicArrayProperty2<Type>::TypePtr& value);
	virtual void formatValues();

protected:
	int32 m_columnSize;

private:
	void operator=(const DynamicArrayProperty2<Type>&);
};

template<typename Type>
void DynamicArrayProperty2<Type>::resize(int32 index1, int32 newSize) {

	// auto resize the size
	if (index1 >= static_cast<int32>(m_size.size())) {
		ValueDynamicProperty2<Type *>::resize(index1 + 1);
	}

	int32 actualSize = getSize(index1);

	for (int32 j = actualSize; j < newSize; ++j) {

		std::string format;
		if (BaseDynamicProperty::m_hex) {
			format = "%x";
		}

		std::string propertyName = propertyNameAtIndex(BaseDynamicProperty::m_name, index1, j);
		if (!BaseDynamicProperty::m_propertyManager->containsProperty(propertyName)) {
			AnyProperty * property = BaseDynamicProperty::m_propertyManager->addProperty(m_fromType, propertyName, "", PropertyTypeToString<Type *>::value, format, BaseDynamicProperty::m_save, BaseDynamicProperty::m_spy, BaseDynamicProperty::m_scan, false);
			property->setExtraSize(m_columnSize);
		}
	}

	for (int32 j = newSize; j < actualSize; ++j) {
		BaseDynamicProperty::m_propertyManager->removeProperty(propertyNameAtIndex(BaseDynamicProperty::m_name, index1, j));
	}

	m_size[index1] = newSize;
}

template<typename Type>
void DynamicArrayProperty2<Type>::setSize(int32 index1, int32 index2, int32 size) {
	this->m_servantProperty->setSize(index1, index2, size);
}

template<typename Type>
int32 DynamicArrayProperty2<Type>::getSize(int32 index1, int32 index2) {
	return this->m_servantProperty->getSize(index1, index2);
}

template<typename Type>
void DynamicArrayProperty2<Type>::sendEvent(int32 index1, int32 index2) {
	this->m_servantProperty->sendEvent(index1, index2);
}

template<typename Type>
void DynamicArrayProperty2<Type>::setColumnSize(int32 size) {
	m_columnSize = size;
}

template<typename Type>
int32 DynamicArrayProperty2<Type>::getColumnSize() {
	return m_columnSize;
}

template<typename Type>
void DynamicArrayProperty2<Type>::set(int32 index1, int32 index2, const DynamicArrayProperty2<Type>::TypePtr& value) {
	m_servantProperty->setValue(index1, index2, value);
}

template<typename Type>
void DynamicArrayProperty2<Type>::update(int32 index1, int32 index2, const DynamicArrayProperty2<Type>::TypePtr& value) {
	m_servantProperty->updateValue(index1, index2, value);
}

template<typename Type>
void DynamicArrayProperty2<Type>::formatValues() {
	// do nothing in case of array
}

#endif
