/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DYNAMICPROPERTY_T_H
#define DYNAMICPROPERTY_T_H

#include "ValueDynamicProperty.h"
#include "LoggingServices/LogStream.h"

template<typename Type, int32 LinkedProperties = 0>
class DynamicProperty;

template<typename Type>
class DynamicProperty<Type, 0> : public ValueDynamicProperty<Type> {

public:
	using ValueDynamicProperty<Type>::m_servantProperty;
	using ValueDynamicProperty<Type>::copy;
	using ValueDynamicProperty<Type>::toString;
	using ValueDynamicProperty<Type>::getSize;
	using ValueDynamicProperty<Type>::resize;
	using ValueDynamicProperty<Type>::get;
	using ValueDynamicProperty<Type>::m_enumeratedValuesID;
	using ValueDynamicProperty<Type>::m_enumeratedLabelsID;
	using ValueDynamicProperty<Type>::m_size;
	using ValueDynamicProperty<Type>::m_ID;
	using BaseDynamicProperty::m_propertyManager;
	using BaseDynamicProperty::m_name;
	using BaseDynamicProperty::m_logLevel;
	using BaseDynamicProperty::getAlias;
	using BaseDynamicProperty::isLogging;
	using Unit::getUnit;

	DynamicProperty() :
		ValueDynamicProperty<Type>() {}

	/**
	 * Copy constructor is copying each member and clones the servant property member.
	 */
	DynamicProperty(const DynamicProperty<Type> & property) :
		ValueDynamicProperty<Type>(property) {
	}

	template<typename ServantType, int aliasLength>
	void init(ServantType * servant, int32 flags, const std::string& name, const char(&aliasName)[aliasLength]) {

		std::string alias = aliasName;

		// Do not check for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			alias = checkString<EXECUTION_WARNING, 0, 10>(name, aliasName);
			if (formatAlias(alias)) {
				common::warn << "Alias for '" << name << "' renamed into '" << alias << "'" << common::eol;
			}
		}

		ValueDynamicProperty<Type>::init(servant, flags, name, alias);

		// Do not set property info for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			this->addPropertyInfo(ServantType::TYPE, "", name, alias);
		}
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name) {
		init(servant, flags, name, "");
	}

	void unsetFormatter() {
		this->unsetFormatFunction();
	}

	template<typename Function>
	void setFormatter(Function f) {
		this->setFormatFunction(f);
	}

	template<typename Function, typename Object>
	void setFormatter(Function f, Object obj) {
		this->setFormatFunction(f, obj);
	}

	void set(int32 index, const Type& value);
	void update(int32 index, const Type& value);
	virtual void formatValues();
	void setEnumeratedValues(const DynamicProperty<Type>& valuesProperty);
	void setEnumeratedLabels(const DynamicProperty<std::string>& labelsProperty);
	void copyValues(const DynamicProperty<Type> & property);

protected:
	void log(int32 index, const std::string& formattedValue, const std::string& previousFormattedValue) const;

private:
	void operator=(const DynamicProperty<Type>&);
};

template<typename Type>
void DynamicProperty<Type, 0>::log(int32 index, const std::string& newFormattedValue, const std::string& previousFormattedValue) const {

	if (getAlias() == "") {
		common::warn << "The property '" << m_servantProperty->getServantName() << "." << m_name << "' has no alias and cannot be logged" << common::eol;
		return;
	}

	LogStream ls(m_servantProperty->getServantType(), m_servantProperty->getServantName(), m_servantProperty->getLogLevel(m_logLevel));
	ls << getAlias() << "." << index << " " << previousFormattedValue << " --> "  << newFormattedValue << getUnit() << endlog;
}

template<typename Type>
void DynamicProperty<Type, 0>::set(int32 index, const Type& value) {
	if (Simulated::activated || Simulated::visa || Simulated::for3d) {
		m_servantProperty->setValue(index, value, toString(index, value));
	}
	else {
		// Get the previous value.
		Type previousValue = get(index);

		// Logging if active and the previous value is different from the new one.
		if (isLogging() && value != previousValue) {
			// Get the previous and new value.
			std::string previousStringValue(toString(index, previousValue));
			std::string newStringValue(toString(index, value));

			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->setValue(index, value, toString(index, value));

			// Log the change of value.
			log(index, newStringValue, previousStringValue);
		}
		else {
			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->setValue(index, value, toString(index, value));
		}
	}
}

template<typename Type>
void DynamicProperty<Type, 0>::update(int32 index, const Type& value) {
	if (Simulated::activated || Simulated::visa || Simulated::for3d) {
		m_servantProperty->updateValue(index, value, toString(index, value));
	}
	else  {
		// Get the previous value.
		Type previousValue = get(index);

		// Logging if active and the previous value is different from the new one.
		if (isLogging() && value != previousValue) {
			// Get the previous and new value.
			std::string previousStringValue(toString(index, previousValue));
			std::string newStringValue(toString(index, value));

			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->updateValue(index, value, toString(index, value));

			// Log the change of value.
			log(index, newStringValue, previousStringValue);
		}
		else {
			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->updateValue(index, value, toString(index, value));
		}
	}
}

template<typename Type>
void DynamicProperty<Type, 0>::formatValues() {

	int32 size = getSize();
	for (int32 i = 0; i < size; ++i) {
		update(i, get(i));
	}
}

template<typename Type>
void DynamicProperty<Type, 0>::setEnumeratedValues(const DynamicProperty<Type>& valuesProperty) {

	m_enumeratedValuesID = valuesProperty.getID();

	// Set the dynamic property.
	m_propertyManager->setEnumeratedValues(m_ID, m_enumeratedValuesID);

	// Set all the properties.
	for (int32 i = 0; i < m_size; ++i) {
		AnyProperty * property = m_propertyManager->getProperty(propertyNameAtIndex(m_name, i));
		property->setEnumeratedValues(m_enumeratedValuesID);
	}
}

template<typename Type>
void DynamicProperty<Type, 0>::setEnumeratedLabels(const DynamicProperty<std::string>& labelsProperty) {

	m_enumeratedLabelsID = labelsProperty.getID();

	// Set the dynamic property.
	m_propertyManager->setEnumeratedLabels(m_ID, m_enumeratedLabelsID);

	// Set all the properties.
	for (int32 i = 0; i < m_size; ++i) {
		AnyProperty * property = m_propertyManager->getProperty(propertyNameAtIndex(m_name, i));
		property->setEnumeratedLabels(m_enumeratedLabelsID);
	}
}

template<typename Type>
void DynamicProperty<Type, 0>::copyValues(const DynamicProperty<Type> & property) {
	resize(property.m_size);

	// set all the properties
	for (int32 i = 0; i < m_size; ++i) {
		update(i, property.get(i));
	}
}

// LinkedProperties = SETPOINT
// DynamicProperty members for actual and setpoint

template<typename Type>
class DynamicProperty<Type, SETPOINT> {

public:
	DynamicProperty<Type> actual;
	DynamicProperty<Type> setpoint;

	DynamicProperty() {}

	/**
	 * Copy constructor is copying each member.
	 */
	DynamicProperty(const DynamicProperty<Type, SETPOINT> & property) :
		actual(property.actual),
		setpoint(property.setpoint) {
	}

	template<typename ServantType, int aliasLength>
	void init(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName, const char(&aliasName)[aliasLength]) {

		std::string alias = aliasName;

		// Do not check for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			alias = checkString<EXECUTION_WARNING, 0, 10>(name, aliasName);
			if (formatAlias(alias)) {
				common::warn << "Alias for '" << name << "' renamed into '" << alias << "'" << common::eol;
			}
		}

		// No flag for the actual value.
		actual.ValueDynamicProperty<Type>::init(servant, 0, name, alias);
		setpoint.ValueDynamicProperty<Type>::init(servant, flags, setpointName, alias);

		//actual.init(servant, 0, name, aliasName);
		//setpoint.init(servant, flags, setpointName, aliasName);

		// Set qualifier?

		// Do not set property info for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			actual.addPropertyInfo(ServantType::TYPE, setpointName, name, alias);
		}
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName) {
		init(servant, flags, name, setpointName, "");
	}

	template<typename ServantType>
	void copy(ServantType * servant, const DynamicProperty<Type, SETPOINT> & property) {
		actual.copy(servant, property.actual);
		setpoint.copy(servant, property.setpoint);
	}

	void setUnit(const std::string & value) {
		actual.setUnit(value);
		setpoint.setUnit(value);
	}

	void setUnit(ValueProperty<std::string> & property) {
		actual.setUnit(property);
		setpoint.setUnit(property);
	}

	void unsetFormatter() {
		actual.unsetFormatFunction();
		setpoint.unsetFormatFunction();
	}

	template<typename Function>
	void setFormatter(Function f) {
		actual.setFormatFunction(f);
		setpoint.setFormatFunction(f);
	}

	template<typename Function, typename Object>
	void setFormatter(Function f, Object obj) {
		actual.setFormatFunction(f, obj);
		setpoint.setFormatFunction(f, obj);
	}

	void resize(int32 value) {
		actual.resize(value);
		setpoint.resize(value);
	}

	int32 getSize() const {
		return actual.getSize();
	}

	void formatValues() {
		actual.formatValues();
		setpoint.formatValues();
	}

	void setEnumeratedValues(const DynamicProperty<Type>& valuesProperty) {
		actual.setEnumeratedValues(valuesProperty);
		setpoint.setEnumeratedValues(valuesProperty);
	}

	void setEnumeratedLabels(const DynamicProperty<std::string>& labelsProperty) {
		actual.setEnumeratedLabels(labelsProperty);
		setpoint.setEnumeratedLabels(labelsProperty);
	}

};

#endif
