/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DYNAMICPROPERTY2_T_H
#define DYNAMICPROPERTY2_T_H

#include "ValueDynamicProperty2.h"
#include "LoggingServices/LogStream.h"

template<typename Type>
class DynamicProperty2 : public ValueDynamicProperty2<Type> {

public:
	using ValueDynamicProperty2<Type>::m_servantProperty;
	using ValueDynamicProperty2<Type>::toString;
	using ValueDynamicProperty2<Type>::getSize;
	using ValueDynamicProperty2<Type>::resize;
	using ValueDynamicProperty2<Type>::get;
	using ValueDynamicProperty2<Type>::m_enumeratedValuesID;
	using ValueDynamicProperty2<Type>::m_enumeratedLabelsID;
	using ValueDynamicProperty2<Type>::m_size;
	using ValueDynamicProperty2<Type>::m_ID;
	using BaseDynamicProperty::m_propertyManager;
	using BaseDynamicProperty::m_name;
	using BaseDynamicProperty::m_logLevel;
	using BaseDynamicProperty::getAlias;
	using BaseDynamicProperty::isLogging;
	using Unit::getUnit;

	DynamicProperty2() :
		ValueDynamicProperty2<Type>() {}

	/**
	 * Copy constructor is copying each member and clones the servant property member.
	 */
	DynamicProperty2(const DynamicProperty2<Type> & property) :
		ValueDynamicProperty2<Type>(property) {
	}

	template<typename ServantType, int aliasLength>
	void init(ServantType * servant, int32 flags, const std::string& name, const char(&aliasName)[aliasLength]) {

		std::string alias = aliasName;

		// Do not check for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			alias = checkString<EXECUTION_WARNING, 0, 10>(name, aliasName);
			if (formatAlias(alias)) {
				common::warn << "Alias for '" << name << "' renamed into '" << alias << "'" << common::eol;
			}
		}

		ValueDynamicProperty2<Type>::init(servant, flags, name, alias);

		// Do not set property info for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			this->addPropertyInfo(ServantType::TYPE, "", name, alias);
		}
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name) {
		init(servant, flags, name, "");
	}

	void unsetFormatter() {
		this->unsetFormatFunction();
	}

	template<typename Function>
	void setFormatter(Function f) {
		this->setFormatFunction(f);
	}

	template<typename Function, typename Object>
	void setFormatter(Function f, Object obj) {
		this->setFormatFunction(f, obj);
	}

	void set(int32 index1, int32 index2, const Type& value);
	void update(int32 index1, int32 index2, const Type& value);
	virtual void formatValues();
	void setEnumeratedValues(const DynamicProperty<Type>& valuesProperty);
	void setEnumeratedLabels(const DynamicProperty<std::string>& labelsProperty);
	void copyValues(const DynamicProperty2<Type> & property);

protected:
	void log(int32 index1, int32 index2, const std::string& newFormattedValue, const std::string& previousFormattedValue) const;

private:
	void operator=(const DynamicProperty2<Type>&);
};

template<typename Type>
void DynamicProperty2<Type>::log(int32 index1, int32 index2, const std::string& newFormattedValue, const std::string& previousFormattedValue) const {

	if (getAlias() == "") {
		common::warn << "The property '" << m_servantProperty->getServantName() << "." << m_name << "' has no alias and cannot be logged" << common::eol;
		return;
	}

	LogStream ls(m_servantProperty->getServantType(), m_servantProperty->getServantName(), m_servantProperty->getLogLevel(m_logLevel));
	ls << getAlias() << "." << index1 << "." << index2 << " " << previousFormattedValue << " --> " << newFormattedValue << getUnit() << endlog;
}

template<typename Type>
void DynamicProperty2<Type>::set(int32 index1, int32 index2, const Type& value) {
	if (Simulated::activated || Simulated::visa || Simulated::for3d) {
		m_servantProperty->setValue(index1, index2, value, toString(index1, index2, value));
	}
	else {
		// Get the previous value.
		Type previousValue = get(index1, index2);

		// Logging if active and the previous value is different from the new one.
		if (isLogging() && value != previousValue) {
			// Get the previous and new value.
			std::string previousStringValue(toString(index1, index2, previousValue));
			std::string newStringValue(toString(index1, index2, value));

			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->setValue(index1, index2, value, toString(index1, index2, value));

			// Log the change of value.
			log(index1, index2, newStringValue, previousStringValue);
		}
		else {
			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->setValue(index1, index2, value, toString(index1, index2, value));
		}
	}
}

template<typename Type>
void DynamicProperty2<Type>::update(int32 index1, int32 index2, const Type& value) {
	if (Simulated::activated || Simulated::visa || Simulated::for3d) {
		m_servantProperty->updateValue(index1, index2, value, toString(index1, index2, value));
	}
	else {
		// Get the previous value.
		Type previousValue = get(index1, index2);

		// Logging if active and the previous value is different from the new one.
		if (isLogging() && value != previousValue) {
			// Get the previous and new value.
			std::string previousStringValue(toString(index1, index2, previousValue));
			std::string newStringValue(toString(index1, index2, value));

			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->updateValue(index1, index2, value, toString(index1, index2, value));

			// Log the change of value.
			log(index1, index2, newStringValue, previousStringValue);
		}
		else {
			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->updateValue(index1, index2, value, toString(index1, index2, value));
		}
	}
}

template<typename Type>
void DynamicProperty2<Type>::formatValues() {

	int32 size1 = getSize();
	for (int32 i = 0; i < size1; ++i) {
		int32 size2 = getSize(i);
		for (int32 j = 0; j < size2; ++j) {
			update(i, j, get(i, j));
		}
	}
}

template<typename Type>
void DynamicProperty2<Type>::setEnumeratedValues(const DynamicProperty<Type>& valuesProperty) {

	m_enumeratedValuesID = valuesProperty.getID();

	// Set the dynamic property.
	m_propertyManager->setEnumeratedValues(m_ID, m_enumeratedValuesID);

	// Set all the properties.
	int32 size1 = getSize();
	for (int32 i = 0; i < size1; ++i) {
		int32 size2 = getSize(i);
		for (int32 j = 0; j < size2; ++j) {
			AnyProperty * property = m_propertyManager->getProperty(propertyNameAtIndex(m_name, i, j));
			property->setEnumeratedValues(m_enumeratedValuesID);
		}
	}
}

template<typename Type>
void DynamicProperty2<Type>::setEnumeratedLabels(const DynamicProperty<std::string>& labelsProperty) {

	m_enumeratedLabelsID = labelsProperty.getID();

	// Set the dynamic property.
	m_propertyManager->setEnumeratedLabels(m_ID, m_enumeratedLabelsID);

	// Set all the properties.
	int32 size1 = getSize();
	for (int32 i = 0; i < size1; ++i) {
		int32 size2 = getSize(i);
		for (int32 j = 0; j < size2; ++j) {
			AnyProperty * property = m_propertyManager->getProperty(propertyNameAtIndex(m_name, i, j));
			property->setEnumeratedLabels(m_enumeratedLabelsID);
		}
	}
}

template<typename Type>
void DynamicProperty2<Type>::copyValues(const DynamicProperty2<Type> & property) {

	// set all the properties
	int32 size1 = property.getSize();
	resize(size1);

	for (int32 i = 0; i < size1; ++i) {
		int32 size2 = property.getSize(i);
		resize(i, size2);

		for (int32 j = 0; j < size2; ++j) {
			update(i, j, property.get(i, j));
		}
	}
}

#endif
