/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DYNAMICREFRESHER_H
#define DYNAMICREFRESHER_H

#include <boost/any.hpp>
#include <common/base/Function.h>
#include "ArgumentType.h"

class BaseDynamicRefresher {

public:
	virtual ~BaseDynamicRefresher() {}
	virtual bool hasRefresher() const = 0;
	virtual void refresh(int32 index, const boost::any & value) = 0;
	virtual void copy(Servant * servant, const BaseDynamicRefresher * refresher) = 0;
};

template<typename Type>
class DynamicRefresher : public BaseDynamicRefresher {

	friend class PropertyRefresher;

public:
	DynamicRefresher() : m_refresher(nullptr) {}

	virtual ~DynamicRefresher() {
		delete m_refresher;
	}

	virtual bool hasRefresher() const {
		return m_refresher != nullptr;
	}

	virtual void refresh(int32 index, const boost::any & value);

private:
	typedef typename ArgumentType<Type>::value ValueType;

	template<typename Class, typename Object>
	void setRefresher(void (Class::* f)(int32, ValueType), Object * obj) {
		m_refresher = new common::Function2<void, int32, ValueType, Object>(std::bind(f, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3), obj);
	}

	template<typename Class, typename Param1, typename Object, typename Arg1>
	void setRefresher(void (Class::* f)(Param1, int32, ValueType), Object * obj, Arg1 arg1) {
		m_refresher = new common::Function2<void, int32, ValueType, Object>(std::bind(f, std::placeholders::_1, arg1, std::placeholders::_2, std::placeholders::_3), obj);
	}

	template<typename Class, typename Param1, typename Param2, typename Object, typename Arg1, typename Arg2>
	void setRefresher(void (Class::* f)(Param1, Param2, int32, ValueType), Object * obj, Arg1 arg1, Arg2 arg2) {
		m_refresher = new common::Function2<void, int32, ValueType, Object>(std::bind(f, std::placeholders::_1, arg1, arg2, std::placeholders::_2, std::placeholders::_3), obj);
	}

	template<typename Class, typename Param1, typename Param2, typename Param3, typename Object, typename Arg1, typename Arg2, typename Arg3>
	void setRefresher(void (Class::* f)(Param1, Param2, Param3, int32, ValueType), Object * obj, Arg1 arg1, Arg2 arg2, Arg3 arg3) {
		m_refresher = new common::Function2<void, int32, ValueType, Object>(std::bind(f, std::placeholders::_1, arg1, arg2, arg3, std::placeholders::_2, std::placeholders::_3), obj);
	}

	template<typename Class, typename Param1, typename Param2, typename Param3, typename Param4, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
	void setRefresher(void (Class::* f)(Param1, Param2, Param3, Param4, int32, ValueType), Object * obj, Arg1 arg1, Arg2 arg2, Arg3 arg3, Arg4 arg4) {
		m_refresher = new common::Function2<void, int32, ValueType, Object>(std::bind(f, std::placeholders::_1, arg1, arg2, arg3, arg4, std::placeholders::_2, std::placeholders::_3), obj);
	}

	virtual void copy(Servant * servant, const BaseDynamicRefresher * refresher);

	common::BaseFunction2<void, int32, ValueType> * m_refresher;
};

template<typename Type>
void DynamicRefresher<Type>::refresh(int32 index, const boost::any & value) {
	m_refresher->execute(index, boost::any_cast<Type>(value));
}

template<typename Type>
void DynamicRefresher<Type>::copy(Servant * servant, const BaseDynamicRefresher * refresher) {

	const common::BaseFunction2<void, int32, ValueType> * refresherToCopy = dynamic_cast<const DynamicRefresher<Type> *>(refresher)->m_refresher;
	if (refresherToCopy != nullptr) {
		m_refresher = refresherToCopy->clone();
		m_refresher->setObject(servant);
	}
}

#endif
