/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef PROPERTY_T_H
#define PROPERTY_T_H

#include "ValueProperty.h"
#include "DynamicProperty.h"
#include "LoggingServices/LogStream.h"
#include <bitset>

template<typename Type>
class SimpleProperty : public ValueProperty<Type> {

public:
	using ValueProperty<Type>::m_servantProperty;
	using ValueProperty<Type>::toString;
	using ValueProperty<Type>::get;
	using BaseProperty::m_property;
	//using BaseProperty::m_logLevel;  // Do not use using because of CDT bug from version 11.5 causing a systematic failure of the indexing
	//using BaseProperty::m_qualifier; // Do not use using because of CDT bug from version 11.5 causing a systematic failure of the indexing
	using BaseProperty::getAlias;
	using BaseProperty::getName;
	using BaseProperty::isLogging;

	SimpleProperty() {}
	virtual ~SimpleProperty() {}

	void operator=(const Type& value);
	void set(const Type& value);
	void update(const Type& value);
	virtual void formatValue();
	void setEnumeratedValues(const DynamicProperty<Type>& valuesProperty);
	void setEnumeratedLabels(const DynamicProperty<std::string>& labelsProperty);

protected:
	void log(const std::string& previousFormattedValue) const;
};

template<typename Type>
void SimpleProperty<Type>::log(const std::string& previousFormattedValue) const {

	if (getAlias() == "") {
		common::warn << "The property '" << m_servantProperty->getServantName() << "." << getName() << "' has no alias and cannot be logged" << common::eol;
		return;
	}

	LogStream ls(m_servantProperty->getServantType(), m_servantProperty->getServantName(), m_servantProperty->getLogLevel(BaseProperty::m_logLevel));
	ls << getAlias() << " ";

	if (BaseProperty::m_qualifier == OFFSET) {
		ls << "offset ";
	}
	ls << previousFormattedValue << " --> " << *this << endlog;
}

template<typename Type>
void SimpleProperty<Type>::operator=(const Type& value) {
	if (Simulated::activated || Simulated::visa || Simulated::for3d) {
		m_servantProperty->setValue(value, toString(value));
	}
	else {
		// Get the previous value.
		Type previousValue = get();

		// Logging if active and the previous value is different from the new one.
		if (isLogging() && value != previousValue) {
			// Get the previous string value.
			std::string previousStringValue(toString(previousValue));

			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->setValue(value, toString(value));

			// Log the change of value.
			log(previousStringValue);
		}
		else {
			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->setValue(value, toString(value));
		}
	}
}

template<typename Type>
void SimpleProperty<Type>::set(const Type& value) {

	if (Simulated::activated || Simulated::visa || Simulated::for3d) {
		m_servantProperty->setValue(value, toString(value));
	}
	else {
		// Get the previous value.
		Type previousValue = get();

		// Logging if active and the previous value is different from the new one.
		if (isLogging() && value != previousValue) {

			// Get the previous value.
			std::string previousStringValue(toString(previousValue));

			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->setValue(value, toString(value));

			// Log the change of value.
			log(previousStringValue);
		}
		else {
			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->setValue(value, toString(value));
		}
	}
}

template<typename Type>
void SimpleProperty<Type>::update(const Type& value) {

	if (Simulated::activated || Simulated::visa || Simulated::for3d) {
		m_servantProperty->updateValue(value, toString(value));
	}
	else {
		// Get the previous value.
		Type previousValue = get();

		// Logging if active and the previous value is different from the new one.
		if (isLogging() && value != previousValue) {
			// Get the previous value.
			std::string previousStringValue(toString(previousValue));

			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->updateValue(value, toString(value));

			// Log the change of value.
			log(previousStringValue);
		}
		else {
			// Format the value, otherwise asynchronous problems may occur.
			m_servantProperty->updateValue(value, toString(value));
		}
	}
}

template<typename Type>
void SimpleProperty<Type>::formatValue() {
	update(get());
}

template<typename Type>
void SimpleProperty<Type>::setEnumeratedValues(const DynamicProperty<Type>& valuesProperty) {
	m_property->setEnumeratedValues(valuesProperty.getID());
}

template<typename Type>
void SimpleProperty<Type>::setEnumeratedLabels(const DynamicProperty<std::string>& labelsProperty) {
	m_property->setEnumeratedLabels(labelsProperty.getID());
}

template<typename Type, int32 LinkedProperties = 0>
class Property;

// LinkedProperties = 0
template<typename Type>
class Property<Type, 0> : public SimpleProperty<Type> {

public:
	using SimpleProperty<Type>::operator=;

	Property() {}

	template<typename ServantType, int aliasLength>
	void init(ServantType * servant, int32 flags, const std::string& name, const char(&aliasName)[aliasLength]) {

		std::string alias = aliasName;

		// Do not check for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			alias = checkString<EXECUTION_WARNING, 0, 10>(name, aliasName);
			if (formatAlias(alias)) {
				common::warn << "Alias for '" << name << "' renamed into '" << alias << "'" << common::eol;
			}
		}

		ValueProperty<Type>::init(servant, name, alias, flags);

		// Do not set property info for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			bool spy = ((SPY & flags) == 1);
			this->addPropertyInfo(ServantType::TYPE, "", name, "", alias, false, spy);
		}
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name) {
		init(servant, flags, name, "");
	}

	template<typename ServantType, int aliasLength>
	Property(ServantType * servant, int32 flags, const std::string& name, const char(&aliasName)[aliasLength]) {
		init(servant, flags, name, aliasName);
	}

	template<typename ServantType>
	Property(ServantType * servant, int32 flags, const std::string& name) {
		init(servant, flags, name, "");
	}

	template<typename ServantType>
	void copy(ServantType * servant, const Property<Type, 0> & property) {
		ValueProperty<Type>::copy(servant, property);
	}

	void unsetFormatter() {
		this->unsetFormatFunction();
	}

	template<typename Function>
	void setFormatter(Function f) {
		this->setFormatFunction(f);
	}

	template<typename Function, typename Object>
	void setFormatter(Function f, Object obj) {
		this->setFormatFunction(f, obj);
	}

	template<typename ServantType>
	void clear(ServantType * servant, const std::string& name) {
		clear(servant, name);
	}

private:
	void operator=(const Property<Type, 0>&);
};

// LinkedProperties = SETPOINT
// only SPY is used for actual and offset properties
template<typename Type>
class Property<Type, SETPOINT> : public SimpleProperty<Type> {

public:
	using SimpleProperty<Type>::operator=;
	using ValueProperty<Type>::setUnit;
	using SimpleProperty<Type>::setEnumeratedValues;
	using SimpleProperty<Type>::setEnumeratedLabels;

	Property() {}

	template<typename ServantType, int aliasLength>
	void init(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName, const char(&aliasName)[aliasLength]) {

		std::string alias = aliasName;

		// Do not check for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			alias = checkString<EXECUTION_WARNING, 1, 10>(name, aliasName);
			if (formatAlias(alias)) {
				common::warn << "Alias for '" << name << "' renamed into '" << alias << "'" << common::eol;
			}
		}

		// Remove LOG flags because SETPOINT properties cannot be logged.
		bool logInfo = (((::LOG_INFO & flags) >> 4) == 1);
		bool logDebug = (((::LOG_DEBUG & flags) >> 5) == 1);

		if (logInfo || logDebug) {
			common::warn << "Setpoint property '" << name << "' cannot log" << common::eol;
		}

		flags &= ~(1u << 4);
		flags &= ~(1u << 5);

		// Call to init of the base class. Only keep the SPY flag for the actual property.
		ValueProperty<Type>::init(servant, name, alias, flags & 1);

		setpoint.init(servant, setpointName, alias, flags);
		setpoint.setQualifier(BaseProperty::SETPOINT);

		// Do not set property info for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			bool scan = (((SCAN & flags) >> 2) == 1);
			bool spy = ((SPY & flags) == 1);
			this->addPropertyInfo(ServantType::TYPE, setpointName, name, "", alias, scan, spy);
		}
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName) {
		init(servant, flags, name, setpointName, "");
	}

	template<typename ServantType, int aliasLength>
	Property(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName, const char(&aliasName)[aliasLength]) {
		init(servant, flags, name, setpointName, aliasName);
	}

	template<typename ServantType>
	Property(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName) {
		init(servant, flags, name, setpointName, "");
	}

	template<typename ServantType>
	void copy(ServantType * servant, const Property<Type, SETPOINT> & property) {
		ValueProperty<Type>::copy(servant, property);

		setpoint.copy(servant, property.setpoint);
	}

	void setUnit(const std::string & value) {
		ValueProperty<Type>::setUnit(value);
		setpoint.setUnit(value);
	}

	void setUnit(ValueProperty<std::string> & property) {
		ValueProperty<Type>::setUnit(property);
		setpoint.setUnit(property);
	}

	void setFormat(const std::string & value) {
		Format::setFormat(value);
		setpoint.setFormat(value);
	}

	void setFormat(ValueProperty<std::string> & property) {
		Format::setFormat(property);
		setpoint.setFormat(property);
	}

	void unsetFormatter() {
		this->unsetFormatFunction();
		setpoint.unsetFormatFunction();
	}

	template<typename Function>
	void setFormatter(Function f) {
		this->setFormatFunction(f);
		setpoint.setFormatFunction(f);
	}

	template<typename Function, typename Object>
	void setFormatter(Function f, Object obj) {
		this->setFormatFunction(f, obj);
		setpoint.setFormatFunction(f, obj);
	}

	void setEnumeratedValues(const DynamicProperty<Type>& valuesProperty) {
		SimpleProperty<Type>::setEnumeratedValues(valuesProperty);
		setpoint.setEnumeratedValues(valuesProperty);
	}

	void setEnumeratedLabels(const DynamicProperty<std::string>& labelsProperty) {
		SimpleProperty<Type>::setEnumeratedLabels(labelsProperty);
		setpoint.setEnumeratedLabels(labelsProperty);
	}

	SimpleProperty<Type> setpoint;

private:
	void operator=(const Property<Type, SETPOINT>&);
};

// LinkedProperties = SETPOINT | OFFSET
// only SPY is used for actual and offset properties
template<typename Type>
class Property<Type, SETPOINT | OFFSET> : public SimpleProperty<Type> {

public:
	using SimpleProperty<Type>::operator=;

	Property() {}

	template<typename ServantType, int aliasLength>
	void init(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName, const std::string& offsetName, const char(&aliasName)[aliasLength]) {

		std::string alias = aliasName;

		// Do not check for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			alias = checkString<EXECUTION_WARNING, 1, 10>(name, aliasName);
			if (formatAlias(alias)) {
				common::warn << "Alias for '" << name << "' renamed into '" << alias << "'" << common::eol;
			}
		}

		// OFFSET property is logged whereas the SETPOINT is not.
		// Keep the SPY, LOG_INFO and LOG_DEBUG flags for the offset property.
		int32 offsetFlags = flags & ((1u << 1) | (1u << 4) | (1u << 5));

		// Remove LOG flags because SETPOINT properties and the actual associated property cannot be logged.
		flags &= ~(1u << 4);
		flags &= ~(1u << 5);

		// Call to init of the base class.
		ValueProperty<Type>::init(servant, name, alias, flags & 1);
		setpoint.init(servant, setpointName, alias, flags);
		setpoint.setQualifier(BaseProperty::SETPOINT);

		// Init the offset property at last.
		offset.init(servant, offsetName, alias, offsetFlags);
		offset.setQualifier(BaseProperty::OFFSET);

		// Do not set property info for drivers.
		if (servant->getServantClass() != DRIVER_CLASS) {
			bool scan = (((SCAN & flags) >> 2) == 1);
			bool spy = ((SPY & flags) == 1);
			this->addPropertyInfo(ServantType::TYPE, setpointName, name, offsetName, alias, scan, spy);
		}
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName, const std::string& offsetName) {
		init(servant, flags, name, setpointName, offsetName, "");
	}

	template<typename ServantType, int aliasLength>
	Property(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName, const std::string& offsetName, const char(&aliasName)[aliasLength]) {
		init(servant, flags, name, setpointName, offsetName, aliasName);
	}

	template<typename ServantType>
	Property(ServantType * servant, int32 flags, const std::string& name, const std::string& setpointName, const std::string& offsetName) {
		init(servant, flags, name, setpointName, offsetName, "");
	}

	template<typename ServantType>
	void copy(ServantType * servant, const Property<Type, SETPOINT | OFFSET> & property) {
		ValueProperty<Type>::copy(servant, property);

		setpoint.copy(servant, property.setpoint);
		offset.copy(servant, property.offset);
	}

	void setUnit(const std::string & value) {
		ValueProperty<Type>::setUnit(value);
		setpoint.setUnit(value);
		offset.setUnit(value);
	}

	void setUnit(ValueProperty<std::string> & property) {
		ValueProperty<Type>::setUnit(property);
		setpoint.setUnit(property);
		offset.setUnit(property);
	}

	void setFormat(const std::string & value) {
		Format::setFormat(value);
		setpoint.setFormat(value);
		offset.setFormat(value);
	}

	void setFormat(ValueProperty<std::string> & property) {
		Format::setFormat(property);
		setpoint.setFormat(property);
		offset.Format(property);
	}

	void unsetFormatter() {
		this->unsetFormatFunction();
		setpoint.unsetFormatFunction();
		offset.unsetFormatFunction();
	}

	template<typename Function>
	void setFormatter(Function f) {
		this->setFormatFunction(f);
		setpoint.setFormatFunction(f);
		offset.setFormatFunction(f);
	}

	template<typename Function, typename Object>
	void setFormatter(Function f, Object obj) {
		this->setFormatFunction(f, obj);
		setpoint.setFormatFunction(f, obj);
		offset.setFormatFunction(f, obj);
	}

	void setEnumeratedValues(const DynamicProperty<Type>& valuesProperty) {
		SimpleProperty<Type>::setEnumeratedValues(valuesProperty);
		setpoint.setEnumeratedValues(valuesProperty);
		offset.setEnumeratedValues(valuesProperty);
	}

	void setEnumeratedLabels(const DynamicProperty<std::string>& labelsProperty) {
		SimpleProperty<Type>::setEnumeratedLabels(labelsProperty);
		setpoint.setEnumeratedLabels(labelsProperty);
		offset.setEnumeratedLabels(labelsProperty);
	}

	SimpleProperty<Type> setpoint;
	SimpleProperty<Type> offset;

private:
	void operator=(const Property<Type, SETPOINT | OFFSET>&);
};

#endif
