/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef STATUSPROPERTY_T_H
#define STATUSPROPERTY_T_H

#include "BaseProperty.h"
#include <sstream>

class StatusProperty: public BaseProperty {

	friend std::ostream& operator<<(std::ostream&, const StatusProperty&);

public:
	StatusProperty() :
			m_servantProperty(0) {
	}

	/**
	 * Copy constructor is copying each member and clones the servant property member.
	 */
	StatusProperty(const StatusProperty& property) :
			BaseProperty(property), m_servantProperty(property.m_servantProperty->clone()) {
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name, const std::string& alias = "") {
		BaseProperty::init(ServantType::TYPE, servant->getPropertyManager(), name, alias, PropertyTypeToString<int32>::value,
				flags);
		m_servantProperty = new ServantProperty<ServantType, int32>(servant, name);
		m_servantProperty->updateValue(UNKNOWN, UNKNOWN_STR);
	}

	template<typename ServantType>
	void copy(ServantType * servant, const StatusProperty & property) {
		BaseProperty::copy(servant->getPropertyManager(), property);
		m_servantProperty = new ServantProperty<ServantType, int32>(servant, property.getName());
		m_servantProperty->updateValue(property());
	}

	template<typename ServantType>
	StatusProperty(ServantType * servant, int32 flags, const std::string& name, const std::string& alias = "") {
		init(servant, flags, name, alias);
	}

	virtual ~StatusProperty() {
		delete m_servantProperty;
	}

	std::string getServantName();

	int32 get() const;
	int32 operator()() const;

	void set(const StatusProperty& property);

	void setIdle();
	void setRunning();
	void setWarning();
	void setError();
	void setFixed();
	void setUnknown();

	bool isIdle() const;
	bool isRunning() const;
	bool isWarning() const;
	bool isError() const;
	bool isFixed() const;
	bool isUnknown() const;

	static const std::string IDLE_STR;
	static const std::string RUNNING_STR;
	static const std::string WARNING_STR;
	static const std::string ERROR_STR;
	static const std::string FIXED_STR;
	static const std::string UNKNOWN_STR;

	std::ostream& toString(std::ostream& s) const {
		switch (boost::any_cast<int32>(m_property->getValue())) {
		case IDLE:
			s << IDLE_STR;
			break;
		case RUNNING:
			s << RUNNING_STR;
			break;
		case WARNING:
			s << WARNING_STR;
			break;
		case ERROR:
			s << ERROR_STR;
			break;
		case FIXED:
			s << FIXED_STR;
			break;
		default:
			s << UNKNOWN_STR;
			break;
		}
		return s;
	}

	virtual void formatValue();
	virtual common::LogElement toPropertyLog(bool hidden = false) const;

protected:
	BaseServantProperty<int32> * m_servantProperty;

private:
	enum {
		UNKNOWN = 0, IDLE = 1, RUNNING = 2, WARNING = 4, ERROR = 8, FIXED = 16
	};
};

#endif
