/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef VALUEDYNAMICPROPERTY_T_H
#define VALUEDYNAMICPROPERTY_T_H

#include "BaseDynamicProperty.h"
#include "BaseProperty.h"
#include "Unit.h"
#include "Format.h"
#include "Formatter.h"
#include "DynamicRefresher.h"
#include "Utilities/Simulated.h"

template<typename Type>
class ValueDynamicProperty : public BaseDynamicProperty, public Unit, public Format, public Formatter<Type>, public DynamicRefresher<Type> {

public:
	ValueDynamicProperty() :
		BaseDynamicProperty(),
		m_size(0),
		m_servantProperty(0),
		m_ID(0),
		m_unitPropertyID(0),
		m_enumeratedValuesID(0),
		m_enumeratedLabelsID(0) {}

	/**
	 * Copy constructor is copying each member and clones the servant property member.
	 */
	ValueDynamicProperty(const ValueDynamicProperty<Type> & property) :
		BaseDynamicProperty(property),
		Unit(property),
		Format(property),
		Formatter<Type>(property),
		m_size(property.m_size),
		m_fromType(property.m_fromType),
		m_servantProperty(property.m_servantProperty->clone()),
		m_ID(0),
		m_unitValue(property.m_unitValue),
		m_unitPropertyID(property.m_unitPropertyID),
		m_enumeratedValuesID(property.m_enumeratedValuesID),
		m_enumeratedLabelsID(property.m_enumeratedLabelsID) {
	}

	template<typename ServantType>
	void copy(ServantType * servant, const ValueDynamicProperty<Type> & property) {

		BaseDynamicProperty::copy(servant->getPropertyManager(), property);
		Unit::copy(property);
		Format::copy(property);
		Formatter<Type>::copy(servant, property);

		m_size = property.m_size;
		m_fromType = property.m_fromType;
		m_servantProperty = new ServantDynamicProperty<ServantType, Type>(servant, property.m_name);
		m_ID = servant->getPropertyManager()->addDynamicProperty(m_name, PropertyTypeToString<Type>::value);
		m_unitValue = property.m_unitValue;
		m_unitPropertyID = property.m_unitPropertyID;
		m_enumeratedValuesID = property.m_enumeratedValuesID;
		m_enumeratedLabelsID = property.m_enumeratedLabelsID;

		// no copy of refresher (too difficult to do it with reference copies)
		servant->registerDynamicProperty(m_name, this);
	}

	template<typename ServantType>
	ValueDynamicProperty(ServantType * servant, int32 flags, const std::string& name, const std::string& alias) :
		BaseDynamicProperty(),
		m_size(0),
		m_servantProperty(0),
		m_ID(0),
		m_unitPropertyID(0),
		m_enumeratedValuesID(0),
		m_enumeratedLabelsID(0) {
		init(servant, flags, name, alias);
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name, const std::string& alias) {

		BaseDynamicProperty::init(servant->getPropertyManager(), flags, name, alias);

		m_size = 0;
		m_fromType = ServantType::TYPE;
		m_servantProperty = new ServantDynamicProperty<ServantType, Type>(servant, name);
		m_ID = servant->getPropertyManager()->addDynamicProperty(name, PropertyTypeToString<Type>::value);
		m_unitValue = "";
		m_unitPropertyID = 0;
		m_enumeratedValuesID = 0;
		m_enumeratedLabelsID = 0;

		servant->registerDynamicProperty(name, this);
	}

	template<typename ServantType>
	void clear(ServantType * servant, const std::string& name) {
		servant->unregisterDynamicProperty(name);
	}

	~ValueDynamicProperty() {
		delete m_servantProperty;
	}

	std::string getServantName() const;
	unsigned long getID() const;

	void resize(int32 value);
	int32 getSize() const;

	Type get(int32 index) const;
	Type operator()(int32 index) const;

	void setUnit(const std::string & value);
	void setUnit(const ValueProperty<std::string> & property);
	std::string getUnit() const;

	std::ostream& toString(int32 index, std::ostream& s, const Type& value) const;
	std::string toString(int32 index, const Type& value) const;
	std::string toString(int32 index) const;

protected:
	int32 m_size;
	std::string m_fromType;
	BaseServantDynamicProperty<Type> * m_servantProperty;
	unsigned long m_ID;
	std::string m_unitValue;
	unsigned long m_unitPropertyID;
	unsigned long m_enumeratedValuesID;
	unsigned long m_enumeratedLabelsID;
};

template<typename Type>
std::string ValueDynamicProperty<Type>::getServantName() const {
	return m_servantProperty->getServantName();
}

template<typename Type>
unsigned long ValueDynamicProperty<Type>::getID() const {
	return m_ID;
}

template<typename Type>
void ValueDynamicProperty<Type>::resize(int32 newSize) {

	for (int32 i = m_size; i < newSize; ++i) {

		std::string format;
		if (m_hex) {
			format = "%x";
		}

		std::string propertyName = propertyNameAtIndex(m_name, i);
		if (!m_propertyManager->containsProperty(propertyName)) {
			AnyProperty * property = m_propertyManager->addProperty(m_fromType, propertyName, m_alias, PropertyTypeToString<Type>::value, format, m_save, m_spy, m_scan, false);

			if (m_unitPropertyID != 0) {
				property->setUnit(m_unitPropertyID);
			} else if (m_unitValue != "") {
				property->setUnit(m_unitValue);
			}

			if (m_enumeratedValuesID != 0) {
				property->setEnumeratedValues(m_enumeratedValuesID);
			}

			if (m_enumeratedLabelsID != 0) {
				property->setEnumeratedLabels(m_enumeratedLabelsID);
			}
		}
	}

	for (int32 i = newSize; i < m_size; ++i) {
		m_propertyManager->removeProperty(propertyNameAtIndex(m_name, i));
	}

	m_size = newSize;
}

template<typename Type>
int32 ValueDynamicProperty<Type>::getSize() const {
	return m_size;
}

template<typename Type>
Type ValueDynamicProperty<Type>::get(int32 index) const {
	return boost::any_cast<Type>(m_propertyManager->getProperty(propertyNameAtIndex(m_name, index))->getValue());
}

template<typename Type>
Type ValueDynamicProperty<Type>::operator()(int32 index) const {
	return get(index);
}

template<typename Type>
void ValueDynamicProperty<Type>::setUnit(const std::string & value) {
	Unit::setUnit(value);

	m_unitValue = value;
	m_unitPropertyID = 0;

	// set all the properties
	for (int32 i = 0; i < m_size; ++i) {
		AnyProperty * property = m_propertyManager->getProperty(propertyNameAtIndex(m_name, i));
		property->setUnit(m_unitValue);
	}
}

template<typename Type>
void ValueDynamicProperty<Type>::setUnit(const ValueProperty<std::string> & property) {
	Unit::setUnit(property);

	m_unitValue = "";
	m_unitPropertyID = property.getID();

	// set all the properties
	for (int32 i = 0; i < m_size; ++i) {
		AnyProperty * property = m_propertyManager->getProperty(propertyNameAtIndex(m_name, i));
		property->setUnit(m_unitPropertyID);
	}
}

template<typename Type>
std::string ValueDynamicProperty<Type>::getUnit() const {
	return Unit::getUnit();
}

template<typename Type>
std::ostream& ValueDynamicProperty<Type>::toString(int32 index, std::ostream& s, const Type& value) const {

	try {
		if (Formatter<Type>::isFormatterDefined()) {
			s << Formatter<Type>::format(value);
		} else {
			// use default format from AnyProperty
			m_propertyManager->getProperty(propertyNameAtIndex(m_name, index))->toString(s, value, getFormat());
		}
	} catch (...) {
		return s << "?";
	}

	return s;
}

template<typename Type>
std::string ValueDynamicProperty<Type>::toString(int32 index, const Type& value) const {

	std::ostringstream s;
	toString(index, s, value);

	return s.str();

}

template<typename Type>
std::string ValueDynamicProperty<Type>::toString(int32 index) const {
	return toString(index, get(index));

}

#endif
