/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef VALUEDYNAMICPROPERTY2_T_H
#define VALUEDYNAMICPROPERTY2_T_H

#include "BaseDynamicProperty.h"
#include "BaseProperty.h"
#include "Unit.h"
#include "Format.h"
#include "Formatter.h"
#include "DynamicRefresher2.h"
#include "Utilities/Simulated.h"

template<typename Type>
class ValueDynamicProperty2 : public BaseDynamicProperty, public Unit, public Format, public Formatter<Type>, public DynamicRefresher2<Type> {
public:
	ValueDynamicProperty2() :
		BaseDynamicProperty(),
		m_servantProperty(0),
		m_ID(0),
		m_unitPropertyID(0),
		m_enumeratedValuesID(0),
		m_enumeratedLabelsID(0) {}

	/**
	 * Copy constructor is copying each member and clones the servant property member.
	 */
	ValueDynamicProperty2(const ValueDynamicProperty2<Type> & property) :
		BaseDynamicProperty(property),
		Unit(property),
		Format(property),
		Formatter<Type>(property),
		m_size(property.m_size),
		m_fromType(property.m_fromType),
		m_servantProperty(property.m_servantProperty->clone()),
		m_ID(0),
		m_unitValue(property.m_unitValue),
		m_unitPropertyID(property.m_unitPropertyID),
		m_enumeratedValuesID(property.m_enumeratedValuesID),
		m_enumeratedLabelsID(property.m_enumeratedLabelsID) {
	}

	template<typename ServantType>
	void init(ServantType * servant, int32 flags, const std::string& name, const std::string& alias) {

		BaseDynamicProperty::init(servant->getPropertyManager(), flags, name, alias);

		m_fromType = ServantType::TYPE;
		m_servantProperty = new ServantDynamicProperty<ServantType, Type>(servant, name);
		m_unitValue = "";
		m_unitPropertyID = 0;
		m_enumeratedValuesID = 0;
		m_enumeratedLabelsID = 0;

		servant->registerDynamicProperty2(name, this);
	}

	template<typename ServantType>
	void copy(ServantType * servant, const ValueDynamicProperty2<Type> & property) {

		BaseDynamicProperty::copy(servant->getPropertyManager(), property);
		Unit::copy(property);
		Format::copy(property);
		Formatter<Type>::copy(servant, property);

		m_size = property.m_size;
		m_fromType = property.m_fromType;
		m_servantProperty = new ServantDynamicProperty<ServantType, Type>(servant, property.m_name);
		m_unitValue = property.m_unitValue;
		m_unitPropertyID = property.m_unitPropertyID;
		m_enumeratedValuesID = property.m_enumeratedValuesID;
		m_enumeratedLabelsID = property.m_enumeratedLabelsID;

		// no copy of refresher (too difficult to do it with reference copies)
		servant->registerDynamicProperty2(m_name, this);
	}

	template<typename ServantType>
	ValueDynamicProperty2(ServantType * servant, int32 flags, const std::string& name) :
		BaseDynamicProperty(),
		m_servantProperty(0),
		m_ID(0),
		m_unitPropertyID(0),
		m_enumeratedValuesID(0),
		m_enumeratedLabelsID(0) {
		init(servant, flags, name);
	}

	template<typename ServantType>
	void clear(ServantType * servant, const std::string& name) {
		servant->unregisterDynamicProperty2(name);
	}

	~ValueDynamicProperty2() {
		delete m_servantProperty;
	}

	std::string getServantName() const;
	unsigned long getID() const;

	void resize(int32 value);
	void resize(int32 index1, int32 value);

	int32 getSize() const;
	int32 getSize(int32 index1) const;

	Type get(int32 index1, int32 index2) const;
	Type operator()(int32 index1, int32 index2) const;

	void setUnit(const std::string & value);
	void setUnit(const ValueProperty<std::string> & property);
	std::string getUnit() const;

	std::ostream& toString(int32 index1, int32 index2, std::ostream& s, const Type& value) const;
	std::string toString(int32 index1, int32 index2, const Type& value) const;
	std::string toString(int32 index1, int32 index2) const;

protected:
	std::vector<int32> m_size;
	std::string m_fromType;
	BaseServantDynamicProperty<Type> * m_servantProperty;
	unsigned long m_ID;
	std::string m_unitValue;
	unsigned long m_unitPropertyID;
	unsigned long m_enumeratedValuesID;
	unsigned long m_enumeratedLabelsID;
};

template<typename Type>
std::string ValueDynamicProperty2<Type>::getServantName() const {
	return m_servantProperty->getServantName();
}

template<typename Type>
unsigned long ValueDynamicProperty2<Type>::getID() const {
	return m_ID;
}

template<typename Type>
void ValueDynamicProperty2<Type>::resize(int32 value) {

	if (value < static_cast<int32>(m_size.size())) {
		// remove properties
		for (int32 i = value; i < static_cast<int32>(m_size.size()); ++i) {
			// iterate the dimension to remove each property
			for (int32 j = 0; j < m_size[i]; ++j) {
				m_propertyManager->removeProperty(propertyNameAtIndex(m_name, i, j));
			}
		}
	}

	// in case value is > size of vector, do not add any property, resize size vector by adding 0

	m_size.resize(value, 0);
}

template<typename Type>
void ValueDynamicProperty2<Type>::resize(int32 index1, int32 newSize) {

	// auto resize the size
	if (index1 >= static_cast<int32>(m_size.size())) {
		resize(index1 + 1);
	}

	int32 actualSize = getSize(index1);

	for (int32 j = actualSize; j < newSize; ++j) {

		std::string format;
		if (m_hex) {
			format = "%x";
		}

		std::string propertyName = propertyNameAtIndex(m_name, index1, j);
		if (!m_propertyManager->containsProperty(propertyName)) {
			AnyProperty * property = m_propertyManager->addProperty(m_fromType, propertyName, m_alias, PropertyTypeToString<Type>::value, format, m_save, m_spy, m_scan, false);

			if (m_unitPropertyID != 0) {
				property->setUnit(m_unitPropertyID);
			} else if (m_unitValue != "") {
				property->setUnit(m_unitValue);
			}

			if (m_enumeratedValuesID != 0) {
				property->setEnumeratedValues(m_enumeratedValuesID);
			}

			if (m_enumeratedLabelsID != 0) {
				property->setEnumeratedLabels(m_enumeratedLabelsID);
			}
		}
	}

	for (int32 j = newSize; j < actualSize; ++j) {
		m_propertyManager->removeProperty(propertyNameAtIndex(m_name, index1, j));
	}

	m_size[index1] = newSize;
}

template<typename Type>
int32 ValueDynamicProperty2<Type>::getSize() const {
	return static_cast<int32>(m_size.size());
}

template<typename Type>
int32 ValueDynamicProperty2<Type>::getSize(int32 index1) const {
	try {
		return m_size.at(index1);
	} catch (...) {
		throw BadIndexException(m_name, index1);
	}
}

template<typename Type>
Type ValueDynamicProperty2<Type>::get(int32 index1, int32 index2) const {
	return boost::any_cast<Type>(m_propertyManager->getProperty(propertyNameAtIndex(m_name, index1, index2))->getValue());
}

template<typename Type>
Type ValueDynamicProperty2<Type>::operator()(int32 index1, int32 index2) const {
	return get(index1, index2);
}

template<typename Type>
void ValueDynamicProperty2<Type>::setUnit(const std::string & value) {
	Unit::setUnit(value);

	m_unitValue = value;
	m_unitPropertyID = 0;

	// set all the properties
	for (int32 i = 0; i < static_cast<int32>(m_size.size()); ++i) {
		for (int32 j = 0; j < m_size[i]; ++j) {
			AnyProperty * property = m_propertyManager->getProperty(propertyNameAtIndex(m_name, i));
			property->setUnit(m_unitValue);
		}
	}
}

template<typename Type>
void ValueDynamicProperty2<Type>::setUnit(const ValueProperty<std::string> & property) {
	Unit::setUnit(property);

	m_unitValue = "";
	m_unitPropertyID = property.getID();

	// set all the properties
	for (int32 i = 0; i < static_cast<int32>(m_size.size()); ++i) {
		for (int32 j = 0; j < m_size[i]; ++j) {
			AnyProperty * property = m_propertyManager->getProperty(propertyNameAtIndex(m_name, i));
			property->setUnit(m_unitPropertyID);
		}
	}
}

template<typename Type>
std::string ValueDynamicProperty2<Type>::getUnit() const {
	return Unit::getUnit();
}

template<typename Type>
std::ostream& ValueDynamicProperty2<Type>::toString(int32 index1, int32 index2, std::ostream& s, const Type& value) const {

	try {
		if (Formatter<Type>::isFormatterDefined()) {
			s << Formatter<Type>::format(value);
		} else {
			// use default format from AnyProperty
			m_propertyManager->getProperty(propertyNameAtIndex(m_name, index1, index2))->toString(s, value, getFormat());
		}
	} catch (...) {
		return s << "?";
	}

	return s;
}

template<typename Type>
std::string ValueDynamicProperty2<Type>::toString(int32 index1, int32 index2, const Type& value) const {

	std::ostringstream s;
	toString(index1, index2, s, value);

	return s.str();

}

template<typename Type>
std::string ValueDynamicProperty2<Type>::toString(int32 index1, int32 index2) const {
	return toString(index1, index2, get(index1, index2));
}

#endif
