/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef VALUEPROPERTY_T_H
#define VALUEPROPERTY_T_H

#include "BaseProperty.h"
#include "Unit.h"
#include "Format.h"
#include "Formatter.h"
#include "Refresher.h"
#include "Utilities/Simulated.h"
#include <sstream>

template<typename Type>
class ValueProperty : public BaseProperty, public Unit, public Format, public Formatter<Type>, public Refresher<Type> {

	template<class T>
	friend std::ostream& operator<< (std::ostream&, const ValueProperty<T>&);

public:
	ValueProperty() :
		m_servantProperty(0) {
	}

	/**
	 * Copy constructor is copying each member and clones the servant property member.
	 */
	ValueProperty(const ValueProperty<Type> & property) :
		BaseProperty(property),
		Unit(property),
		Format(property),
		Formatter<Type>(property),
		m_servantProperty(property.m_servantProperty->clone()) {
	}

	template<typename ServantType>
	void init(ServantType * servant, const std::string& name, const std::string& alias, int32 flags) {
		BaseProperty::init(ServantType::TYPE, servant->getPropertyManager(), name, alias, PropertyTypeToString<Type>::value, flags);
		m_servantProperty = new ServantProperty<ServantType, Type>(servant, name);
		servant->registerProperty(name, this);
	}

	template<typename ServantType>
	void copy(ServantType * servant, const ValueProperty<Type> & property) {
		BaseProperty::copy(servant->getPropertyManager(), property);
		Unit::copy(property);
		Format::copy(property);
		Formatter<Type>::copy(servant, property);
		m_servantProperty = new ServantProperty<ServantType, Type>(servant, property.getName());

		// no copy of refresher (too difficult to do it with reference copies)
		servant->registerProperty(property.getName(), this);
	}

	template<typename ServantType>
	ValueProperty(ServantType * servant, const std::string& name, const std::string& alias, int32 flags) : m_servantProperty(0) {
		init(servant, name, alias, flags);
	}

	template<typename ServantType>
	void clear(ServantType * servant, const std::string& name) {
		servant->unregisterProperty(name);
	}

	virtual ~ValueProperty() {
		delete m_servantProperty;
	}

	std::string getServantName() const;

	Type get() const;
	Type operator()() const;

	void setUnit(const std::string & value);
	void setUnit(const ValueProperty<std::string> & property);
	std::string getUnit() const;

	std::ostream& toString(std::ostream& s, const Type& value) const;
	std::string toString(const Type& value) const;
	std::string toString() const;

	virtual common::LogElement toPropertyLog(bool hidden = false) const;

protected:
	BaseServantProperty<Type> * m_servantProperty;
};

template<typename Type>
std::string ValueProperty<Type>::getServantName() const {
	return m_servantProperty->getServantName();
}

template<typename Type>
Type ValueProperty<Type>::get() const {
	return boost::any_cast<Type>(m_property->getValue());
}

template<typename Type>
Type ValueProperty<Type>::operator()() const {
	return boost::any_cast<Type>(m_property->getValue());
}

template<typename Type>
void ValueProperty<Type>::setUnit(const std::string & value) {
	Unit::setUnit(value);
	m_property->setUnit(value);
}

template<typename Type>
void ValueProperty<Type>::setUnit(const ValueProperty<std::string> & property) {
	Unit::setUnit(property);
	m_property->setUnit(property.getID());
}

template<typename Type>
std::string ValueProperty<Type>::getUnit() const {
	return Unit::getUnit();
}

template<class Type>
std::ostream& ValueProperty<Type>::toString(std::ostream& s, const Type& value) const {
	try {
		if (Formatter<Type>::isFormatterDefined()) {
			s << Formatter<Type>::format(value);
		} else {
			// use default format from AnyProperty
			m_property->toString(s, value, getFormat());
		}
	} catch (...) {
		return s << "?";
	}

	return s;
}

template<class Type>
std::string ValueProperty<Type>::toString(const Type& value) const {
	std::ostringstream s;
	toString(s, value);

	return s.str();
}

template<class Type>
std::string ValueProperty<Type>::toString() const {
	return toString(get());
}

template<class Type>
common::LogElement ValueProperty<Type>::toPropertyLog(bool hidden) const {

	std::ostringstream s;
	toString(s, get());

	std::string name;
	if (getAlias() != "") {
		name = getAlias();
	} else {
		name = getName();
	}

	std::string qualifier = common::LogElement::ACTUAL;
	if (getQualifier() == BaseProperty::SETPOINT) {
		qualifier = common::LogElement::SETPOINT;
	} else if (getQualifier() == BaseProperty::OFFSET) {
		qualifier = common::LogElement::OFFSET;
	}

	return common::LogElement(true, hidden, name, PropertyTypeToString<Type>::value, qualifier, getUnit(), s.str());
}

template<class Type>
std::ostream& operator<< (std::ostream& s, const ValueProperty<Type>& property) {
	return property.toString(s, property.get());
}

#endif
