/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include<iostream>
using namespace std;

#include "DataProvider/Property/XMLFileCollector.h"
#include "DataProvider/Property/PropertyDescription.h"
#include "DataProvider/Property/PropertyDescriptionList.h"
#include "PersistenceServices/XMLErrors.h"
#include "ErrorsAlarmsManager/ErrorDef.h"
#include "ErrorsAlarmsManager/Exceptions.h"
#include "common/base/ServerProperties.h"
#include <pugixml.hpp>

/**
 * Init static members
 */
const string XMLFileCollector::ROOT_TAG="AllProperties";
const string XMLFileCollector::PROPERTIES_CONTAINER_TAG="PropertiesContainer";
const string XMLFileCollector::PROPERTIES_CONTAINER_TYPE_ATTRIBUTE="type";
const string XMLFileCollector::PROPERTY_TAG="Property";
const string XMLFileCollector::NAME_ATTRIBUTE="name";
const string XMLFileCollector::TYPE_ATTRIBUTE="type";
const string XMLFileCollector::SAVE_ATTRIBUTE="save";
const string XMLFileCollector::FORMAT_ATTRIBUTE="format";
const string XMLFileCollector::SPY_ATTRIBUTE="spy";
const string XMLFileCollector::SCAN_ATTRIBUTE="scan";
const string XMLFileCollector::ALIAS_ATTRIBUTE="alias";
const string XMLFileCollector::VALUE_ATTRIBUTE="value";

const string XMLFileCollector::DEVICE="Device";
const string XMLFileCollector::ABSTRACT="Abstract";

XMLFileCollector* XMLFileCollector::m_instance = 0;

void XMLFileCollector::initInstance() {
	if (m_instance == 0) {
		m_instance = new XMLFileCollector();
	}
}

XMLFileCollector* XMLFileCollector::getInstance() {
	return m_instance;
}

/*
 * Constructor
 */
XMLFileCollector::XMLFileCollector() {
}

/*
 * Destructor 
 */
XMLFileCollector::~XMLFileCollector() {
	// Delete all description objects
	map<string,PropertyDescriptionList*>::iterator iter;
	iter = m_DriverPropertyDescriptionMap.begin();
	while (iter != m_DriverPropertyDescriptionMap.end()) {
		// Delete all description objects
		delete iter->second;
		++iter;
	}
	iter = m_ControllerPropertyDescriptionMap.begin();
	while (iter != m_ControllerPropertyDescriptionMap.end()) {
		// Delete all description objects
		delete iter->second;
		++iter;
	}
	iter = m_PropertyDescriptionMap.begin();
	while (iter != m_PropertyDescriptionMap.end()) {
		// Delete all description objects
		delete iter->second;
		++iter;
	}
}

/*
 * getPartOfList 
 */
void XMLFileCollector::getPartOfList(const string& aCaller, const string &aTag,	vector<PropertyDescription*>& desc) {

	bool found = false;

	if (aCaller == DEVICE) {
		if (m_DriverPropertyDescriptionMap.find(aTag)!= m_DriverPropertyDescriptionMap.end()) {
			found = true;
			m_DriverPropertyDescriptionMap[aTag]->copyPropertyDescriptionList(desc);
		}
	}
	else if (aCaller == ABSTRACT) {
		if (m_ControllerPropertyDescriptionMap.find(aTag)!= m_ControllerPropertyDescriptionMap.end()) {
			found = true;
			m_ControllerPropertyDescriptionMap[aTag]->copyPropertyDescriptionList(desc);
		}
	}

	// if not in abstract and device list, then try in other list
	if (!found) {
		if (m_PropertyDescriptionMap.find(aTag)!= m_PropertyDescriptionMap.end()) {
			m_PropertyDescriptionMap[aTag]->copyPropertyDescriptionList(desc);
		}
	}
}

void XMLFileCollector::loadPropertyFile(const std::string& file, const std::string& aCaller) {

	pugi::xml_document pDoc;
	pugi::xml_parse_result result = pDoc.load_file(file.c_str());
	if (result) {
		for (auto noderoot : pDoc.children()) {
			if (noderoot.name() == ROOT_TAG) {
				for (auto node : noderoot.children()) {
					if (node.name() == PROPERTIES_CONTAINER_TAG) {
						pugi::xml_attribute attr = node.attribute(PROPERTIES_CONTAINER_TYPE_ATTRIBUTE.c_str());
						string id;
						if (attr) {
							id = attr.value();
						}
						else {
							missingAttributeError(file, (uint32) node.offset_debug(), PROPERTIES_CONTAINER_TYPE_ATTRIBUTE, PROPERTIES_CONTAINER_TAG);
							throw NotOk(XML_PARSING_PROPERTIES_FILE_ERROR);
						}
						PropertyDescriptionList* description = new PropertyDescriptionList();
						for (auto element : node.children()) {
							if (element.name() == PROPERTY_TAG) {
								string name, alias, type, value, save, format, spy, scan;
								for(auto attr : element.attributes()) {
									if (attr.name() == NAME_ATTRIBUTE) {
										name = attr.value();
									}
									else if (attr.name() == TYPE_ATTRIBUTE) {
										type = attr.value();
									}
									else if (attr.name() == VALUE_ATTRIBUTE) {
										value = attr.value();
									}
									else if (attr.name() == SAVE_ATTRIBUTE) {
										save = attr.value();
									}
									else if (attr.name() == FORMAT_ATTRIBUTE) {
										format = attr.value();
									}
									else if (attr.name() == ALIAS_ATTRIBUTE) {
										alias = attr.value();
									}
									else if (attr.name() == SPY_ATTRIBUTE) {
										spy = attr.value();
									}
									else if (attr.name() == SCAN_ATTRIBUTE) {
										scan = attr.value();
									}
									else {
										unknownAttributeError(file, (uint32) element.offset_debug(), attr.name());
										throw NotOk(XML_PARSING_PROPERTIES_FILE_ERROR);
									}
								}
								// Create description object of the property
								PropertyDescription *tmp = new PropertyDescription(name,alias,type,save,format,spy,scan,value);
								// Add to the local list for the destruction of the object
								description->addPropertyDescription(tmp);
							}

						}
						if (aCaller == DEVICE) {
							m_DriverPropertyDescriptionMap[id] = description;
						}
						else if (aCaller == ABSTRACT) {
							m_ControllerPropertyDescriptionMap[id] = description;
						} else {
							m_PropertyDescriptionMap[id] = description;
						}
					}
				}
			}
			else {
				unknownTagError(file, (uint32) noderoot.offset_debug(), noderoot.name());
				throw NotOk(XML_PARSING_PROPERTIES_FILE_ERROR);
			}
		}
	}
	else {
		XMLErrors::printError(file, (uint32) result.offset, result.description());
		throw NotOk(XML_PARSING_PROPERTIES_FILE_ERROR);
	}

}

/*
 * loadFile
 */
void XMLFileCollector::loadFile(const string& aCaller) {

	// Init the property file name
	string fileName;
	fileName.append(common::ServerProperties::getInstance()->getConfigPath());
	fileName.append("/.");
	fileName.append(aCaller);
	fileName.append("Property");
	fileName.append(".xml");

	loadPropertyFile(fileName, aCaller);
}

