/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DataProvider/Publisher.h"
#include "DataProvider/Subscriber.h"
#include "DataProvider/ChangeAspect.h"
#include "LoggingServices/Level.h"
#include <sstream>
#include <algorithm>

using namespace std;

/**
 * Initialize static members
 */
string Publisher::m_idPrefix = string("pub-");
int32 Publisher::m_idIndex = 0;
recursive_mutex Publisher::m_publisherMutex;

Publisher::Publisher() :
	m_attached(false) {
	m_idIndex++;
	stringstream ss;
	ss << m_idPrefix << m_idIndex;
	m_publisherId = ss.str();
}

Publisher::Publisher(const Publisher&) :
	m_attached(false) {

}

Publisher::~Publisher() {

}

void Publisher::attach(Subscriber* subscriber) {

	//cout << this << " (" << typeid(*this).name() << ") attaches " << subscriber << "(" << typeid(*subscriber).name() << ")" << endl;

	/*
	 * Lock the publisher during an attach
	 */

	//cout << pthread_self() << " attach" << endl;

	lock_guard<recursive_mutex> pubLock(m_publisherMutex);
	//LOCK(pubLock, "attach", m_publisherMutex);

	//cout << pthread_self() << " attach ok" << endl;

	/*
	 * Verify that we've not had a null pointer sent to us
	 */
	if (subscriber == 0) {

		return;

	} else {

		if (find(m_subscribers.begin(), m_subscribers.end(), subscriber)
				== m_subscribers.end()) {
			m_subscribers.push_back(subscriber);
			m_attached = true;
		}
	}
}

void Publisher::detach(Subscriber* subscriber) {

	//usleep(100);

	//cout << pthread_self() << " - " << this << " (" << typeid(*this).name() << ") detaches " << subscriber << "(" << typeid(*subscriber).name() << ")" << endl;


	//cout << pthread_self() << " detach" << endl;

	/*
	 * Lock the publisher during a detach
	 */
	lock_guard<recursive_mutex> pubLock(m_publisherMutex);
	//LOCK(pubLock, "detach", m_publisherMutex);

	//cout << pthread_self() << " detach ok" << endl;

	/*
	 * Verify that we've not had a null pointer sent to us
	 */
	if (subscriber == 0) {
		return;
	}

	list<Subscriber*>::iterator subscriberIterator = find(
			m_subscribers.begin(), m_subscribers.end(), subscriber);
	if (subscriberIterator != m_subscribers.end()) {
		*subscriberIterator = 0;
	}
	if (m_subscribers.empty() == true) {
		m_attached = false;
	}
	//cout << pthread_self() << " - " << this << " (" << typeid(*this).name() << ") detached" << endl;
}

void Publisher::notify(ChangeAspect* changeAspect) {

	//cout << pthread_self() << " - " << this << " (" << typeid(*this).name() << ") notifies" << endl;

	/*
	 * Verify that we've not had a null pointer sent to us
	 */
	if (changeAspect == 0) {
		return;
	}

	//cout << std::hex << pthread_self() << " - notify" << endl << std::dec;

	/*
	 * Copy the array so that any detaches that occur do not corrupt the
	 * vector.
	 */
	lock_guard<recursive_mutex> pubLock(m_publisherMutex);

//	cout << std::hex << pthread_self() << " - publisher mutex ok" << endl << std::dec;

	/*
	 * Perform all the updates
	 */
	for (list<Subscriber*>::iterator subscriberIterator = m_subscribers.begin(); subscriberIterator
			!= m_subscribers.end(); subscriberIterator++) {
		if (subscriberIterator != m_subscribers.end()) {
			//LOCK(pubLock, "notify", m_publisherMutex);

			/*
			 * Verify that the subscriber still exists
			 */
			if (*subscriberIterator != 0) {
				//cout << pthread_self() << " - " << this << " updates " << *subscriberIterator << " (" << typeid(**subscriberIterator).name() << ")" << endl;
				(*subscriberIterator)->update(changeAspect);
				//cout << pthread_self() << " - " << this << " updated" << endl;
			}
		}
	}

	/*
	 * Remove dead subscribers.
	 */

	//LOCK(pubLock, "notify", m_publisherMutex);

	for (list<Subscriber*>::iterator subscriberIterator = m_subscribers.begin(); subscriberIterator
			!= m_subscribers.end(); subscriberIterator++) {
		if (*subscriberIterator == 0) {
			subscriberIterator = m_subscribers.erase(subscriberIterator);
		}
	}

//	cout << std::hex << pthread_self() << " - " << this << " (" << typeid(*this).name() << ") notifies end" << endl << std::dec;
}

