/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DeviceContainer/DeviceDriver.h"
#include "DeviceContainer/DeviceDriverState.h"
#include "DeviceContainer/Observer.h"
#include "DeviceContainer/Spy.h"
#include "DataProvider/Property/AnyProperty.h"
#include "DataProvider/ChangeAspect.h"
#include "DataProvider/ChangeAspectFactory.h"
#include "common/base/ServerProperties.h"

// Trace include
#include "LoggingServices/Level.h"

#include <cstdio>
using namespace std;
using namespace boost;

/*
 * Static variable init
 */
const char DeviceDriver::PROGRESSION_PROPERTY[] = "progression";

const string DeviceDriver::ROOT_DEVICE_TYPE_DEVICE_CONTAINER = "root";
const string DeviceDriver::BRANCH_DEVICE_TYPE_DEVICE_CONTAINER = "branch";
const string DeviceDriver::LEAF_DEVICE_TYPE_DEVICE_CONTAINER = "leaf";

const string DeviceDriver::TYPE = "DeviceDriver";

DeviceDriver::BadValue::BadValue(const std::string& type, const std::string& name, Publisher* publisher,
		const std::string& info) :
		NotOk(type, name, publisher, BAD_VALUE_ERROR, info) {
}

DeviceDriver::TimeOut::TimeOut(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info) :
		NotOk(type, name, publisher, TIME_OUT_ERROR, info) {
}

DeviceDriver::NoDeviceDriver::NoDeviceDriver() :
		NotOk(0) {
}

DeviceDriver::NoDeviceDriver::NoDeviceDriver(const std::string& type, const std::string& name, Publisher* publisher,
		const std::string& info) :
		NotOk(type, name, publisher, NO_DEVICE_DRIVER_ERROR, info) {
}

DeviceDriver::NotEnoughMemory2::NotEnoughMemory2(const std::string& type, const std::string& name, Publisher* publisher,
		const std::string& info) :
		NotOk(type, name, publisher, NOTENOUGH_MEMORY_ERROR, info) {
}

DeviceDriver::Disconnected::Disconnected() :
		NotOk(0) {
}

DeviceDriver::Disconnected::Disconnected(const std::string& type, const std::string& name, Publisher* publisher,
		const std::string& info) :
		NotOk(type, name, publisher, DISCONNECTED_ERROR, info) {
}

DeviceDriver::BusError::BusError(const std::string& type, const std::string& name, Publisher* publisher, int32 ExceptionsID,
		const std::string& info) :
		NotOk(type, name, publisher, ExceptionsID, info) {
}

DeviceDriver::ChildBadType::ChildBadType() :
		NotOk(0) {
}

DeviceDriver::ChildBadChannel::ChildBadChannel() :
		NotOk(0) {
}

DeviceDriver::ChildBadConf::ChildBadConf() :
		NotOk(0) {
}

DeviceDriver::ChildIsUsed::ChildIsUsed() :
		NotOk(0) {
}

DeviceDriver::ParentBadType::ParentBadType() :
		NotOk(0) {
}

DeviceDriver::ParentBadConf::ParentBadConf() :
		NotOk(0) {
}

DeviceDriver::ParentIsUsed::ParentIsUsed() :
		NotOk(0) {
}

DeviceDriver::DeviceDriver() :
		m_runningMode(NORMAL_MODE_DEVICE_CONTAINER), m_realState(0), m_perfectState(
				0), m_simulatedState(0), m_currentState(0), m_parent(NULL), m_CommandManager(0), m_Enable(true), m_PauseRequested(
				false), m_PauseActivated(false), m_module(0), m_propertyManager(0) {
}

DeviceDriver::DeviceDriver(const string& aType, const string& aName, const string& aParentFunction,
		DeviceDriverState* aRealState, DeviceDriverState* aPerfectState, DeviceDriverState* aSimulatedState) :
		m_runningMode(NORMAL_MODE_DEVICE_CONTAINER), m_realState(aRealState), m_perfectState(
				aPerfectState), m_simulatedState(aSimulatedState), m_currentState(m_perfectState), m_parent(NULL), m_CommandManager(
				new CommandManager(aName)), m_Enable(true), m_PauseRequested(false), m_PauseActivated(false), m_module(0), m_propertyManager(
				new PropertyManager(aName)) {
	// Init Parent Function
	registerParentFunction(aParentFunction);
	constructor(aType, aName);
}

DeviceDriver::DeviceDriver(const std::string& aType, const std::string& aName, const std::string& aParentFunction) :
		m_runningMode(NORMAL_MODE_DEVICE_CONTAINER), m_realState(0), m_perfectState(
				0), m_simulatedState(0), m_currentState(0), m_parent(NULL), m_CommandManager(new CommandManager(aName)), m_Enable(
				true), m_PauseRequested(false), m_PauseActivated(false), m_module(0), m_propertyManager(
				new PropertyManager(aName)) {

	// Init Parent Function
	registerParentFunction(aParentFunction);
	constructor(aType, aName);
}

DeviceDriver::DeviceDriver(const std::string& aName) :
		m_runningMode(NORMAL_MODE_DEVICE_CONTAINER), m_realState(0), m_perfectState(
				0), m_simulatedState(0), m_currentState(0), m_parent(NULL), m_CommandManager(new CommandManager(aName)), m_Enable(
				true), m_PauseRequested(false), m_PauseActivated(false), m_module(0), m_propertyManager(
				new PropertyManager(aName)) {
	constructor("", aName);
}

void DeviceDriver::constructor(const std::string& aType, const std::string& aName) {
	status.init(this, NOSAVE, "status");
	commandProgression.init(this, NOSAVE, "progression");
	version.init(this, SAVE, "version");
	date.init(this, NOSAVE, "date");
	deviceType.init(this, NOSAVE, "device_type");
	leafChannel.init(this, NOSAVE, "leaf_channel");
	numberLeaf.init(this, SAVE, "number_leaf");
	deviceName.init(this, NOSAVE, "name");
	type.init(this, NOSAVE, "type");
	state.init(this, SAVE, "state");
	startActivated.init(this, NOSAVE, "started");
	disconnected.init(this, NOSAVE, "disconnected");
	commandStatus.init(this, NOSAVE, "commandStatus");

	commandProgression.update(PROGRESSION_END_DEVICE_CONTAINER);
	version.update("none");

	// Init property
	type.update(aType);
	deviceName.update(aName);

	registerRefresher(state, &DeviceDriver::refreshStateProperty, this);
	registerRefresher(numberLeaf, &DeviceDriver::refreshNumberLeafProperty, this);
	registerRefresher(commandProgression, &DeviceDriver::refreshProgression, this);
}

void DeviceDriver::init(const std::string& aName) {

	// initialise only once
	if (m_CommandManager == 0) {
		m_CommandManager = new CommandManager(aName);
		m_propertyManager = new PropertyManager(aName);

		constructor("", aName);
	}
}

/*
 * Destructor
 */
DeviceDriver::~DeviceDriver() {

	string tmp = deviceType();

	Spy::getInstance()->unregisterDeviceDriver(this);
	Observer::getInstance()->unregisterDeviceDriver(this);

	if ((tmp == ROOT_DEVICE_TYPE_DEVICE_CONTAINER) || (tmp == BRANCH_DEVICE_TYPE_DEVICE_CONTAINER)) {
		/*
		 * Set child parent link to zero
		 */
		for (int32 i = 1; i <= numberLeaf(); ++i) {
			if (m_child[i - 1] != 0) {
				try {
					m_child[i - 1]->removeParent();
					removeChild(i);
				} catch (ParentBadConf& e) {
//					cerr << "Parent not configured" << endl;
				}
			}
		}
	}

	// Test also if driver is enable, other
	if ((tmp == LEAF_DEVICE_TYPE_DEVICE_CONTAINER) || (tmp == BRANCH_DEVICE_TYPE_DEVICE_CONTAINER)) {

		if ((m_Enable == true) || (getRunningMode() != NORMAL_MODE_DEVICE_CONTAINER)) {
			/*
			 * Set parent child link to zero
			 */
			try {
				DeviceDriver* base = getParent();

				// Searching for this in the children list of the parent
				for (int32 i = 1; i <= base->numberLeaf(); ++i) {
					if (base->getChild(i) == this) {
						base->removeChild(i);
					}
				}
			} catch (ParentBadConf& e) {
//				cerr << "Parent not configured for " << getName() << endl;
			}
		}
	}

	delete m_realState;
	delete m_perfectState;
	delete m_simulatedState;

	delete m_propertyManager;
	delete m_CommandManager;
}

ServantClass DeviceDriver::getServantClass() const {
	return DRIVER_CLASS;
}

std::string DeviceDriver::getServantInfo() const {
	return "Driver " + getName() + " (" + getType() + ")";
}

std::string DeviceDriver::getHomePath() const {
	return common::ServerProperties::getInstance()->getHomePath();
}

std::string DeviceDriver::getConfigPath() const {
	return common::ServerProperties::getInstance()->getConfigPath();
}

void DeviceDriver::setRunningMode(uint32 mode) {
	m_runningMode = mode;
}

void DeviceDriver::registerStates(DeviceDriverState* aRealState, DeviceDriverState* aPerfectState,
		DeviceDriverState* aSimulatedState) {
	m_realState = aRealState;
	m_perfectState = aPerfectState;
	m_simulatedState = aSimulatedState;
	m_currentState = m_perfectState;
}

void DeviceDriver::registerRealState(DeviceDriverState* aRealState) {
	m_realState = aRealState;
}

bool DeviceDriver::hasRealState() const {
	return (m_realState != 0);
}

void DeviceDriver::setModule(Module * module) {
	m_module = module;
}

Module * DeviceDriver::getModule() const {
	return m_module;
}

std::vector<std::string> DeviceDriver::getCommandList() const {
	return m_CommandManager->getCommandList();
}

/*
 * setValue
 */
void DeviceDriver::setValue(const std::string& propertyName, boost::any aValue)
		 {

	try {
		refreshValue(propertyName, aValue);
		boost::any oldValue = getValue(propertyName);
		m_propertyManager->setValue(propertyName, aValue);
		hasChanged(propertyName, oldValue, aValue);
	} catch (PropertyManager::CannotSetValue&) {
		throw CannotSetValue();
	}
}

/*
 * setValue
 */
void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, boost::any aValue, const std::string& formattedValue)
		 {

	try {
		refreshValue(propertyName, aValue);
		boost::any oldValue = getValue(propertyName);
		m_propertyManager->setValue(propertyName, aValue);
		hasChanged(propertyName, oldValue, aValue);
	} catch (PropertyManager::CannotSetValue&) {
		throw CannotSetValue();
	}
}

/*
 * updateValue
 */
void DeviceDriver::updateValue(const std::string& propertyName, boost::any aValue)
		 {

	try {
		boost::any oldValue = getValue(propertyName);
		m_propertyManager->setValue(propertyName, aValue);
		hasChanged(propertyName, oldValue, aValue);
	} catch (PropertyManager::CannotSetValue&) {
		throw CannotSetValue();
	}
}

/*
 * updateValue
 */
void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, boost::any aValue, const std::string& formattedValue)
		 {

	try {
		boost::any oldValue = getValue(propertyName);
		m_propertyManager->setValueAndFormattedValue(propertyName, aValue, formattedValue);
		hasChanged(propertyName, oldValue, aValue);
	} catch (PropertyManager::CannotSetValue&) {
		throw CannotSetValue();
	}
}

void DeviceDriver::refreshValue(const std::string& propertyName, const boost::any& value)  {

	PropertyRefresher::refresh(propertyName, value);
	refreshSetValue(propertyName, value);
}

/*
 * hasChanged
 */
void DeviceDriver::hasChanged(const string& propertyName, boost::any oldValue,
		boost::any newValue) {

	string propertyType = m_propertyManager->getType(propertyName);
	bool hasChanged = false;
	try {
		if (propertyType == UINT16_PROPERTY) {
			hasChanged = (any_cast<uint16>(newValue) != any_cast<uint16>(oldValue));
		} else if (propertyType == UINT32_PROPERTY) {
			hasChanged = (any_cast<uint32>(newValue) != any_cast<uint32>(oldValue));
		} else if (propertyType == UINT64_PROPERTY) {
			hasChanged = (any_cast<uint64>(newValue) != any_cast<uint64>(oldValue));
		} else if (propertyType == INT16_PROPERTY) {
			hasChanged = (any_cast<int16>(newValue) != any_cast<int16>(oldValue));
		} else if (propertyType == INT32_PROPERTY) {
			hasChanged = (any_cast<int32>(newValue) != any_cast<int32>(oldValue));
		} else if (propertyType == INT64_PROPERTY) {
			hasChanged = (any_cast<int64>(newValue) != any_cast<int64>(oldValue));
		} else if (propertyType == FLOAT32_PROPERTY) {
			hasChanged = (any_cast<float32>(newValue) != any_cast<float32>(oldValue));
		} else if (propertyType == FLOAT64_PROPERTY) {
			hasChanged = (any_cast<float64>(newValue) != any_cast<float64>(oldValue));
		} else if (propertyType == BOOLEAN_PROPERTY) {
			hasChanged = (any_cast<bool>(newValue) != any_cast<bool>(oldValue));
		} else if (propertyType == STRING_PROPERTY) {
			hasChanged = (any_cast<string>(newValue) != any_cast<string>(oldValue));
		}
	} catch (...) {
		hasChanged = true;
	}
	if (hasChanged == true) {
		ChangeAspectFactory aCAFactory;

		ChangeAspect* aCA = aCAFactory.createPropertyChangeAspect(getType(), getName(), propertyName);
		notify(aCA);
	}
}

/*
 * refreshProgression
 */
void DeviceDriver::refreshProgression(int32 value) {
	// only send the progress
	sendProgressEvent(value);
}

/*
 * sendProgressEvent
 */
void DeviceDriver::sendProgressEvent(int32 progress) {

	ChangeAspectFactory aCAFactory;

	ChangeAspect* progressChangeAspect = aCAFactory.createProgressChangeAspect(getType(), getName(), progress);

	if (getPropertyManager()->containsProperty(PROGRESSION_PROPERTY)) {
		bool hasChanged = (progress != commandProgression());
		if (hasChanged == true) {
			commandProgression.update(progress);
			notify(progressChangeAspect);

			// Sending for nomad2
			m_CommandManager->setProgression(progress);
		}
	} else {
		notify(progressChangeAspect);

		// Sending for nomad2
		m_CommandManager->setProgression(progress);
	}
}

/*
 * sendArrayEvent
 */
void DeviceDriver::sendArrayEvent(const string& propertyName) {

	string propertyType = m_propertyManager->getType(propertyName);

	if ((propertyType == UINT16ARRAY_PROPERTY) || (propertyType == UINT32ARRAY_PROPERTY)
			|| (propertyType == UINT64ARRAY_PROPERTY) || (propertyType == INT16ARRAY_PROPERTY)
			|| (propertyType == INT32ARRAY_PROPERTY) || (propertyType == INT64ARRAY_PROPERTY)
			|| (propertyType == FLOAT32ARRAY_PROPERTY)|| (propertyType == FLOAT64ARRAY_PROPERTY)) {
		ChangeAspectFactory aCAFactory;

		ChangeAspect* aCA = aCAFactory.createPropertyChangeAspect(getType(), getName(), propertyName);
		notify(aCA);
	}

	// sending notification
	getPropertyManager()->sendArrayEvent(propertyName);
}

/*
 * isChildEmpty
 */
bool DeviceDriver::isChildEmpty() const

{
	bool empty = true;

	if (deviceType() != LEAF_DEVICE_TYPE_DEVICE_CONTAINER) {
		for (uint32 i = 0; i < (uint32) numberLeaf(); ++i) {
			if (m_child[i] != NULL) {
				empty = false;
				break;
			}
		}
	}
	return empty;
}

/*
 * getPortID
 */
string DeviceDriver::getPortID() const {
	if (deviceType() == ROOT_DEVICE_TYPE_DEVICE_CONTAINER)
		return getName();
	else if (m_parent != 0)
		return m_parent->getPortID();

	return "";
}

/*
 * getChild
 */
DeviceDriver* DeviceDriver::getChild(int32 channel) const 

{
	DeviceDriver* device = 0;

	if (deviceType() == LEAF_DEVICE_TYPE_DEVICE_CONTAINER) {
		// Error bad type
		//cerr << "Function not allowed for this device type" << endl;
		throw ChildBadType();
	} else {
		if ((channel <= 0) || (channel > numberLeaf())) {
			// Error channel doesn't exist
			cerr << "Error channel getChild" << channel << " , " << numberLeaf() << endl;
			throw ChildBadChannel();
		} else {
			device = m_child[channel - 1];
		}
	}
	return device;
}

/*
 * getParent
 */
DeviceDriver* DeviceDriver::getParent() const  {
	DeviceDriver* device = 0;

	if (deviceType() == ROOT_DEVICE_TYPE_DEVICE_CONTAINER) {
		// Error bad type
		cerr << "Function not allowed for this device type" << endl;
		throw ParentBadType();
	} else {
		if (m_parent == 0) {
			// Error not configurated
			//cerr << "Parent not configured" << endl;
			throw ParentBadConf();
		} else {
			device = m_parent;
		}
	}
	return device;
}
;

/*
 * setParent
 */
void DeviceDriver::setParent(DeviceDriver* aDevice)  {

	if (deviceType() == ROOT_DEVICE_TYPE_DEVICE_CONTAINER) {
		// Function not allowed
//		cerr << "Function not allowed for this device type 2" << endl;
		throw ParentBadType();
	} else if (getParentFunction() != aDevice->getFunction()) {
		cerr << "Devices not compatible: " << deviceName() << " has parent function " << getParentFunction() << " but parent " << aDevice->deviceName() << " has function " << aDevice->getFunction() << endl;
		throw ChildBadConf();
	} else if (m_parent != 0) {
		// Error , already set , remove first
		//cerr << "Error parent already used" << endl;
		throw ParentIsUsed();
	} else {
		m_parent = aDevice;
	}
}

/*
 * removeParent
 */
void DeviceDriver::removeParent()  {

	if (deviceType() == ROOT_DEVICE_TYPE_DEVICE_CONTAINER) {
		// Function not allowed
		cerr << "Function not allowed for this device type 3" << endl;
		throw ParentBadType();
	} else {
		m_parent = 0;
	}
}

/*
 * setChild
 */
void DeviceDriver::setChild(DeviceDriver* aDevice, const int32 aChannel, const int32 aSize)  {
	if (deviceType() == LEAF_DEVICE_TYPE_DEVICE_CONTAINER) {
		// Function not allowed
		cerr << "Function not allowed for this device type" << endl;
		throw ChildBadType();
	} else if ((aChannel <= 0) || (aChannel > numberLeaf())) {
		// Error channel doesn't exist
		cerr << "Error channel  , setChild" << aChannel << endl;
		throw ChildBadChannel();
	} else if (m_child[aChannel - 1] != 0) {
		// Error , already set , remove first
		//cerr << "Error channel already used , " << aChannel << endl;
		throw ChildIsUsed();
	} else {
		m_child[aChannel - 1] = aDevice;
		aDevice->leafChannel = aChannel;
	}
}

/*
 * removeChild
 */
void DeviceDriver::removeChild(const int32 aChannel)  {

	if (deviceType() == LEAF_DEVICE_TYPE_DEVICE_CONTAINER) {
		// Function not allowed
		//cerr << "Function not allowed for this device type" << endl;
		throw ChildBadType();
	} else if ((aChannel <= 0) || (aChannel > numberLeaf())) {
		// Error channel doesn't exist
		//cerr << "Error channel removeDeviceLink  , " << aChannel  << endl;
		throw ChildBadChannel();
	} else {
		m_child[aChannel - 1] = 0;
	}
}

/*
 * isConfigured
 */
bool DeviceDriver::isConfigured() const {

	

	bool conf = true;
	string tmp = deviceType();

	if ((tmp == LEAF_DEVICE_TYPE_DEVICE_CONTAINER) || (tmp == BRANCH_DEVICE_TYPE_DEVICE_CONTAINER)) {
		try {
			getParent();
		} catch (ParentBadConf& e) {
			conf = false;
		}
	}
	return conf;
}

/*
 * getRootDeviceType
 */
std::string DeviceDriver::getRootDeviceType() {
	string type;

	// Check during arborecence, if there if a rs232 or gpib function,
	// To put the command on a other thread can slow down vme)
	if (deviceType() == ROOT_DEVICE_TYPE_DEVICE_CONTAINER) {
		type = getType();
	} else {
		if (isConfigured()) {
			type = getParent()->getRootDeviceType();
		}
	}
	return type;
}

/*
 * Register a Spy Command
 */
void DeviceDriver::registerSpyCommand(const string& commandName, int32 numberOfCycles) {

	bool exists = false;

	// Check if the command already exists
	for (size_t i = 0; i < m_SpyCommands.size(); i++) {
		if (m_SpyCommands[i] == commandName) {
			exists = true;
		}
	}

	if (!exists) {

		// Compute the new period according to the spy cycle duration.
		int32 realNumberOfCycles = numberOfCycles * 5000 / Spy::getInstance()->getCycleDurationMs();
		if (realNumberOfCycles == 0) {
			realNumberOfCycles = 1;
		}

		int32 realPeriodMs = realNumberOfCycles * Spy::getInstance()->getCycleDurationMs();

		if (realPeriodMs != numberOfCycles * 5000) {
			common::warn << "Real spy period of '" << commandName << "' is " << realPeriodMs << "ms" << common::eol;
		}

		m_SpyCommands.push_back(commandName);
		m_SpyNumberOfCycles[commandName] = realNumberOfCycles;
		m_SpyCounter[commandName] = 0;
	}

}

void DeviceDriver::registerSpyCommandMs(const std::string& commandName, int32 periodMs) {

	bool exists = false;

	// Check if the command already exists
	for (size_t i = 0; i < m_SpyCommands.size(); i++) {
		if (m_SpyCommands[i] == commandName) {
			exists = true;
		}
	}

	if (!exists) {

		// Compute the new period according to the spy cycle duration.
		int32 realNumberOfCycles = periodMs / Spy::getInstance()->getCycleDurationMs();
		if (realNumberOfCycles == 0) {
			realNumberOfCycles = 1;
		}

		int32 realPeriodMs = realNumberOfCycles * Spy::getInstance()->getCycleDurationMs();

		if (realPeriodMs != periodMs) {
			common::warn << "Real spy period of '" << commandName << "' is " << realPeriodMs << "ms" << common::eol;
		}

		m_SpyCommands.push_back(commandName);
		m_SpyNumberOfCycles[commandName] = realNumberOfCycles;
		m_SpyCounter[commandName] = 0;
	}
}

/*
 * Register a Observer Command
 */
void DeviceDriver::registerObserverCommand(const string& commandName, int32 numberOfCycles) {

	bool exists = false;
	
	// Check if the command already exists
	for (size_t i = 0; i < m_ObserverCommands.size(); i++) {
		if (m_ObserverCommands[i] == commandName) {
			exists = true;
		}
	}

	if (!exists) {

		// Compute the new period according to the observer cycle duration.
		int32 realNumberOfCycles = numberOfCycles * 20000 / Observer::getInstance()->getCycleDurationUs();
		if (realNumberOfCycles == 0) {
			realNumberOfCycles = 1;
		}

		int32 realPeriodUs = realNumberOfCycles * Observer::getInstance()->getCycleDurationUs();

		if (realPeriodUs != numberOfCycles * 20000) {
			common::warn << "Real observer period of '" << commandName << "' is " << realPeriodUs / 1000.0 << "ms" << common::eol;
		}

		m_ObserverCommands.push_back(commandName);
		m_ObserverNumberOfCycles[commandName] = realNumberOfCycles;
		m_ObserverCounter[commandName] = 0;
	}
}

void DeviceDriver::registerObserverCommandMs(const std::string& commandName, int32 periodMs) {
	
	bool exists = false;

	// Check if the command already exists
	for (size_t i = 0; i < m_ObserverCommands.size(); i++) {
		if (m_ObserverCommands[i] == commandName) {
			exists = true;
		}
	}

	if (!exists) {

		// Compute the new period according to the observer cycle duration.
		int32 realNumberOfCycles = periodMs * 1000 / Observer::getInstance()->getCycleDurationUs();
		if (realNumberOfCycles == 0) {
			realNumberOfCycles = 1;
		}

		int32 realPeriodUs = realNumberOfCycles * Observer::getInstance()->getCycleDurationUs();

		if (realPeriodUs != periodMs * 1000) {
			common::warn << "Real observer period of '" << commandName << "' is " << realPeriodUs / 1000.0 << "ms" << common::eol;
		}

		m_ObserverCommands.push_back(commandName);
		m_ObserverNumberOfCycles[commandName] = realNumberOfCycles;
		m_ObserverCounter[commandName] = 0;
	}
}

/*
 * getSpyCommands
 */
vector<string> DeviceDriver::getSpyCommands() {

	vector<string> commandList;

	/*
	 * First of all verify that this object is subscribed too: there is not
	 * point sending commands if there is nothing to update. The isAttached()
	 * method is implemented by the Publisher class.
	 */

	if (isAttached()) {
		/*
		 * loop through list of registered spy commands: increment the counter
		 * of each command and check to see if the command needs to be sent to
		 * the Spy.
		 */
		for (size_t i = 0; i < m_SpyCommands.size(); i++) {
			string commandName = m_SpyCommands[i];

			m_SpyCounter[commandName] = m_SpyCounter[commandName] + 1;

			if (m_SpyCounter[commandName] >= m_SpyNumberOfCycles[commandName]) {
				commandList.push_back(commandName);
				m_SpyCounter[commandName] = 0;
			}

		}
	}
	return commandList;

}

/*
 * getObserverCommands
 */
vector<string> DeviceDriver::getObserverCommands() {

	vector<string> commandList;

	/*
	 * First of all verify that this object is subscribed too: there is not
	 * point sending commands if there is nothing to update. The isAttached()
	 * method is implemented by the Publisher class.
	 */

	if (isAttached()) {
		/*
		 * loop through list of registered observer commands: increment the counter
		 * of each command and check to see if the command needs to be sent to
		 * the Observer.
		 */
		for (size_t i = 0; i < m_ObserverCommands.size(); i++) {
			string commandName = m_ObserverCommands[i];

			// Same code than for spy.
			m_ObserverCounter[commandName] = m_ObserverCounter[commandName] + 1;

			if (m_ObserverCounter[commandName] >= m_ObserverNumberOfCycles[commandName]) {
				commandList.push_back(commandName);
				m_ObserverCounter[commandName] = 0;
			}
		}
	}

	return commandList;

}

/*
 * resetObserverCounters
 */
void DeviceDriver::resetObserverCounters() {
	for (size_t i = 0; i < m_ObserverCommands.size(); i++) {
		string commandName = m_ObserverCommands[i];
		m_ObserverCounter[commandName] = 0;
	}
}
/*end mbus */

/*
 * sendErrorEvent
 */
void DeviceDriver::sendErrorEvent(int32 errorId, const string& errorMessage) {

	ChangeAspectFactory changeAspectFactory;
	ChangeAspect* errorChangeAspect = changeAspectFactory.createErrorChangeAspect(getType(), getName(), errorId);

	errorChangeAspect->setStringParam(errorMessage);

	notify(errorChangeAspect);

}

/*
 * setEnabled
 */
void DeviceDriver::setEnabled(bool aValue) {
	m_Enable = aValue;
	if (aValue == false) {
		updateValue("disconnected", false);
		m_DisconnectedCounter.clear();
	}
}

/*
 * setDisconnected
 */
void DeviceDriver::setDisconnected() {
	if (disconnected() == false) {
		updateValue("disconnected", true);
		string buf;
		buf.append("Device ");
		buf.append(getName());
		buf.append(" is disconnected, Nomad unuses it now.");
		sendErrorEvent(DISCONNECTED_ERROR, buf);
		Disconnected(getType(), getName(), this, buf);
	}
}

/*
 * clearDisconnected
 */
void DeviceDriver::clearDisconnected() {
	map<string, uint32>::const_iterator it = m_DisconnectedCounter.begin();
	while (it != m_DisconnectedCounter.end()) {
		clearDisconnectedCounter(it->first);
		++it;
	}
	setValue("disconnected", false);
}

/*
 * getDisconnectedCounter
 */
uint32 DeviceDriver::getDisconnectedCounter(const string& command) const {
	map<string, uint32>::const_iterator it = m_DisconnectedCounter.find(command);
	if (it != m_DisconnectedCounter.end()) {
		return it->second;
	}
	return 0;
}

/*
 * containsCategory
 */
bool DeviceDriver::containsCategory(const std::string& category) {

	std::vector<std::string>::iterator i = m_categoryList.begin();
	bool found = false;
	while (i != m_categoryList.end() && !found) {
		if (*i == category) {
			found = true;
		} else {
			i++;
		}
	}

	return found;
}

/*
 * getPropertyManager
 */
PropertyManager* DeviceDriver::getPropertyManager() const {
	return m_propertyManager;
}

/*
 * getValue
 */
boost::any DeviceDriver::getValue(const std::string& propertyName) const
		 {
	return m_propertyManager->getValue(propertyName);
}

/*
 * getSize
 */
int32 DeviceDriver::getSize(const std::string& propertyName) const
		 {
	return m_propertyManager->getSize(propertyName);
}

/*
 * setSize
 */
void DeviceDriver::setSize(const std::string& propertyName, int32 size) const
		 {
	m_propertyManager->setSize(propertyName, size);
}

/*
 * getExtraSize
 */
int32 DeviceDriver::getExtraSize(const std::string& propertyName) const
		 {
	return m_propertyManager->getExtraSize(propertyName);
}

/*
 * setExtraSize
 */
void DeviceDriver::setExtraSize(const std::string& propertyName, int32 size) const
		 {
	m_propertyManager->setExtraSize(propertyName, size);
}

/*
 * getProperties
 */
std::string DeviceDriver::getProperties() {
	return m_propertyManager->getPropertyList();
}

/*
 * isTerminated
 */
bool DeviceDriver::isTerminated() {
	return (commandProgression() == PROGRESSION_END_DEVICE_CONTAINER);
}

/*
 * initCommand
 */
void DeviceDriver::initCommand(const std::string& commandName) const {
	m_CommandManager->addCommand(commandName);
}

/*
 * commandExists
 */
bool DeviceDriver::commandExists(const std::string& commandName) const {
	return m_CommandManager->commandExists(commandName);
}

/*
 * getCommands
 */
std::string DeviceDriver::getCommands() const {
	return m_CommandManager->getCommands();
}

/*
 * write
 */
void DeviceDriver::write(const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, const void * aData) {
	getParent()->write(this, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * read
 */
void DeviceDriver::read(const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, void * aData) {
	getParent()->read(this, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * readMBLT
 */
void DeviceDriver::readMBLT(const uint32 anOffset, const uint32 numberOfBytes,	void * aData) {
	getParent()->readMBLT(this, anOffset, numberOfBytes, aData);
}

/*
 * write
 */
void DeviceDriver::write(const DeviceDriver* aDeviceDriver, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, const void * aData) {
	getParent()->write(anOffset, sizeOfData, numberOfData, aData);
}

/*
 * read
 */
void DeviceDriver::read(const DeviceDriver* aDeviceDriver, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, void * aData) {
	getParent()->read(anOffset, sizeOfData, numberOfData, aData);
}

/*
 * readMBLT
 */
void DeviceDriver::readMBLT(const DeviceDriver* aDeviceDriver, const uint32 anOffset, const uint32 numberOfBytes,	void * aData) {
	getParent()->readMBLT(anOffset, numberOfBytes, aData);
}

/*
 * write
 */
void DeviceDriver::write(const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData,
		const void * aData) {
	getParent()->write(this, anAddressType, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * read
 */
void DeviceDriver::read(const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData,
		void * aData) {
	getParent()->read(this, anAddressType, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * write
 */
void DeviceDriver::write(const DeviceDriver *aDeviceDriver, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData,
		const uint32 numberOfData, const void * aData) {
	getParent()->write(anAddressType, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * read
 */
void DeviceDriver::read(const DeviceDriver *aDeviceDriver, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData,
		const uint32 numberOfData, void * aData) {
	getParent()->read(anAddressType, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * write
 */
void DeviceDriver::write(const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData,
		const void * aData) {
	getParent()->write(this, anAddress, anAddressType, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * read
 */
void DeviceDriver::read(const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData,
		void * aData) {
	getParent()->read(this, anAddress, anAddressType, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * read
 */
void DeviceDriver::readMBLT(const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset,
		const uint32 numberOfBytes, void *aData) {
	getParent()->readMBLT(this, anAddress, anAddressType, anOffset, numberOfBytes, aData);
}

/*
 * write
 */
void DeviceDriver::write(const DeviceDriver *aDeviceDriver, const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData,
		const uint32 numberOfData, const void * aData) {
	getParent()->write(anAddress, anAddressType, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * read
 */
void DeviceDriver::read(const DeviceDriver *aDeviceDriver, const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData,
		const uint32 numberOfData, void * aData) {
	getParent()->read(anAddress, anAddressType, anOffset, sizeOfData, numberOfData, aData);
}

/*
 * read
 */
void DeviceDriver::readMBLT(const DeviceDriver *aDeviceDriver, const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset,
		const uint32 numberOfBytes, void *aData) {
	getParent()->readMBLT(anAddress, anAddressType, anOffset, numberOfBytes, aData);
}

/*
 * write
 */
void DeviceDriver::write(const string& aData, bool doClear) {
	getParent()->write(this, aData, doClear);
}

/*
 * read
 */
void DeviceDriver::read(string& aData, const string aEnd) {
	getParent()->read(this, aData, aEnd);
}

/*
 * write
 */
void DeviceDriver::write(const DeviceDriver* aDeviceDriver, const std::string& aData, bool doClear) {
	getParent()->write(aData, doClear);
}

/*
 * read
 */
void DeviceDriver::read(const DeviceDriver* aDeviceDriver, std::string& aData, const std::string aEnd) {
	getParent()->read(aData, aEnd);
}

/*
 * write
 */
void DeviceDriver::write(const uint8 *aData, int32 aDataLen) {
	getParent()->write(this, aData, aDataLen);
}

/*
 * read
 */
void DeviceDriver::read(uint8 *aData, int32 aDataLen) {
	getParent()->read(this, aData, aDataLen);
}

/*
 * write
 */
void DeviceDriver::write(const DeviceDriver* aDeviceDriver, const uint8 *aData, int32 aDataLen) {
	getParent()->write(aData, aDataLen);
}

/*
 * read
 */
void DeviceDriver::read(const DeviceDriver* aDeviceDriver, char unsigned *aData, int32 aDataLen) {
	getParent()->read(aData, aDataLen);
}

/*
 * read
 */
void DeviceDriver::readsp(string& aData) {
	getParent()->readsp(this, aData);
}

/*
 * readsp
 */
void DeviceDriver::readsp(const DeviceDriver* aDeviceDriver, std::string& aData) {
	getParent()->readsp(aData);
}

/*
 * setName
 */
void DeviceDriver::setName(const std::string& newName) {
	deviceName = newName;
}

/*
 * getName
 */
std::string DeviceDriver::getName() const {
	return deviceName();
}

/*
 * getType
 */
std::string DeviceDriver::getType() const {
	return type();
}

/*
 * setType
 */
void DeviceDriver::setType(const std::string& aType) {
	type = aType;
	m_propertyManager->addServantType(aType, true);
}

void DeviceDriver::addType(const std::string& aType) {
	m_propertyManager->addServantType(aType, false);
}

/*
 * getCurrentState
 */
DeviceDriverState* DeviceDriver::getCurrentState() {
	return m_currentState;
}

/*
 * getRunningMode
 */
uint32 DeviceDriver::getRunningMode() {
	return m_runningMode;
}

/*
 * getChannel
 */
int32 DeviceDriver::getChannel() const {
	return leafChannel();
}

/*
 * getNumberOfChannels
 */
uint32 DeviceDriver::getNumberOfChannels() const {
	return (uint32) numberLeaf();
}

/*
 * getFunction
 */
std::string DeviceDriver::getFunction() const {
	return m_Function;
}
/*
 * getParentFunction
 */
std::string DeviceDriver::getParentFunction() const {
	return m_ParentFunction;
}

/*
 * registerFunction
 */
void DeviceDriver::registerFunction(const std::string& function) {
	m_Function = function;
}

/*
 * registerParentFunction
 */
void DeviceDriver::registerParentFunction(const std::string& function) {
	m_ParentFunction = function;
}

/*
 * incrementDisconnectedCounter
 */
void DeviceDriver::incrementDisconnectedCounter(const std::string& command) {
	++m_DisconnectedCounter[command];
}

/*
 * clearDisconnectedCounter
 */
void DeviceDriver::clearDisconnectedCounter(const std::string& command) {
	m_DisconnectedCounter[command] = 0;
}

/*
 * isDisconnected
 */
bool DeviceDriver::isDisconnected() const {
	return disconnected();
}

/*
 * getProgression
 */
int32 DeviceDriver::getProgression() const {
	return commandProgression();
}

/*
 * setPauseRequested
 */
void DeviceDriver::setPauseRequested(bool avalue) {
	m_PauseRequested = avalue;
}

/*
 * pauseRequested
 */
bool DeviceDriver::pauseRequested() const {
	return m_PauseRequested;
}

/*
 * refreshStateProperty
 */
void DeviceDriver::refreshStateProperty(const std::string& value) {
	// Change the state
	if (value == REAL_DEVICE_CONTAINER) {
		m_currentState = m_realState;

		if (m_currentState == nullptr) {
			common::warn << "Real state of " << type() << " is null" << common::eol;
		}

	} else if (value == PERFECT_DEVICE_CONTAINER) {
		m_currentState = m_perfectState;
	} else if (value == SIMULATED_DEVICE_CONTAINER) {
		m_currentState = m_simulatedState;
	}
}

/*
 * refreshNumberLeafProperty
 */
void DeviceDriver::refreshNumberLeafProperty(int32 value)  {
	if (m_child.size() != static_cast<uint32>(value)) {
		m_child.clear();
		for (int32 i = 0; i < value; ++i) {
			m_child.push_back(0);
		}
	}
}

Level DeviceDriver::getLogLevel(Level const & level) const {
	return Level::s_Debug;
}
