/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DEVICEDRIVER_H
#define DEVICEDRIVER_H

#include <common/base/BaseTypes.h>
#include "DeviceContainer/Def/CommonDef.h"
#include "DataProvider/Property/PropertyManager.h"
#include "DataProvider/CommandManager.h"
#include "DataProvider/Publisher.h"
#include "DataProvider/Property/TypedProperty/PropertyRefresher.h"
#include "ErrorsAlarmsManager/Exceptions.h"
#include "DataProvider/Property/TypedProperty/Property.h"
#include "DataProvider/Property/TypedProperty/ArrayProperty.h"
#include "DataProvider/Property/TypedProperty/DynamicProperty.h"
#include "DataProvider/Property/TypedProperty/DynamicProperty2.h"
#include "DataProvider/Property/TypedProperty/DynamicArrayProperty.h"
#include "DataProvider/Property/TypedProperty/StatusProperty.h"
#include "LoggingServices/Level.h"
#include <string>
#include <vector>
#include <map>
#include <iterator>

class DeviceDriverState;
class AnyProperty;

namespace command {

	class Servant;
}

class Module;

/**
 * The device driver class is the base class.
 * It makes the link between hardware device and the protocol.
 */
class DeviceDriver: public Publisher, public PropertyRefresher {

public:
	/**
	 * TYPE member used for adding properties with from type.
	 */
	static const std::string TYPE;

	/*
	 * Exceptions used in an execute command
	 */
	class BadValue: public NotOk {
	public:
		BadValue(const std::string& type, const std::string& name,
				Publisher* publisher, const std::string& info);
	};

	/*
	 * Hardware Time out
	 */
	class TimeOut: public NotOk {
	public:
		TimeOut(const std::string& type, const std::string& name,
				Publisher* publisher, const std::string& info);
	};

	/*
	 * No device Driver
	 */
	class NoDeviceDriver: public NotOk {
	public:
		NoDeviceDriver();
		NoDeviceDriver(const std::string& type, const std::string& name,
				Publisher* publisher, const std::string& info);
	};

	/*
	 * Memory allocation error
	 */
	class NotEnoughMemory2: public NotOk {

	public:
		NotEnoughMemory2(const std::string& type, const std::string& name,
				Publisher* publisher, const std::string& info);
	};

	/*
	 * Device is disconnected
	 */
	class Disconnected: public NotOk {

	public:
		Disconnected();
		Disconnected(const std::string& type, const std::string& name,
				Publisher* publisher, const std::string& info);
	};

	/*
	 * Hardware Bus error
	 */
	class BusError: public NotOk {

	public:
		BusError(const std::string& type, const std::string& name,
				Publisher* publisher, int32 ExceptionsID,
				const std::string& info);
	};

	/*
	 * Exceptions used in local
	 */
	class ChildBadType: public NotOk {

	public:
		ChildBadType();
	};

	class ChildBadChannel: public NotOk {

	public:
		ChildBadChannel();
	};

	class ChildBadConf: public NotOk {

	public:
		ChildBadConf();
	};

	class ChildIsUsed: public NotOk {

	public:
		ChildIsUsed();
	};

	class ParentBadType: public NotOk {

	public:
		ParentBadType();
	};

	class ParentBadConf: public NotOk {

	public:
		ParentBadConf();
	};

	class ParentIsUsed: public NotOk {

	public:
		ParentIsUsed();
	};

	/**
	 * Default constructor.
	 */
	DeviceDriver();

	/**
	 * The constructor of the class DeviceDriver.
	 * It sets the two privates properties.
	 *
	 * @param the type of the device driver
	 * @param the name of the device driver
	 * @param the running mode of the device driver
	 */
	DeviceDriver(const std::string& aType, const std::string& aName, const std::string& aParentFunction, DeviceDriverState* aRealState, DeviceDriverState* aPerfectState, DeviceDriverState* aSimulatedState);

	/**
	 * The constructor of the class DeviceDriver.
	 * It sets the two privates properties.
	 *
	 * @param the type of the device driver
	 * @param the name of the device driver
	 * @param the running mode of the device driver
	 */
	DeviceDriver(const std::string& aType, const std::string& aName, const std::string& aParentFunction);

	/**
	 * The constructor of the class DeviceDriver.
	 * It sets the two privates properties.
	 *
	 * @param the name of the device driver
	 */
	DeviceDriver(const std::string& aName);

	/**
	 * The destructor of the class DeviceDriver.
	 */
	virtual ~DeviceDriver();

	/**
	 * Returns the driver class.
	 */
	virtual ServantClass getServantClass() const;

	/**
	 * Gets the servant info.
	 */
	virtual std::string getServantInfo() const;

	/**
	 * Returns the home path.
	 */
	std::string getHomePath() const;

	/**
	 * Returns the config path.
	 */
	std::string getConfigPath() const;

	/**
	 * Sets the running mode.
	 */
	void setRunningMode(uint32 mode);

	/**
	 * Sets the states.
	 */
	void registerStates(DeviceDriverState* aRealState, DeviceDriverState* aPerfectState, DeviceDriverState* aSimulatedState);

	/**
	 * Sets the real state.
	 */
	void registerRealState(DeviceDriverState* aRealState);

	/**
	 * Returns true if the real state has been set.
	 */
	bool hasRealState() const;

	/**
	 * Sets the parent module.
	 */
	void setModule(Module * module);

	/**
	 * Gets the parent module.
	 */
	Module * getModule() const;

	/**
	 * Returns the list of commands.
	 */
	std::vector<std::string> getCommandList() const;

	/**
	 * Gets the property manager.
	 */
	PropertyManager* getPropertyManager() const;

	/**
	 * Returns the actual value of the property.
	 *
	 * @param the property name
	 * @return the property actual value
	 */
	boost::any getValue(const std::string& propertyName) const	;

	/**
	 * Set the actual value of the property.
	 *
	 * @param the property name
	 * @param the actual value
	 */
	void setValue(const std::string& propertyName, boost::any aValue) ;
	void setValue(const std::string& propertyName, bool aValue) ;
	void setValue(const std::string& propertyName, uint16 aValue) ;
	void setValue(const std::string& propertyName, uint32 aValue) ;
	void setValue(const std::string& propertyName, uint64 aValue) ;
	void setValue(const std::string& propertyName, int16 aValue) ;
	void setValue(const std::string& propertyName, int32 aValue) ;
	void setValue(const std::string& propertyName, int64 aValue) ;
	void setValue(const std::string& propertyName, float32 aValue) ;
	void setValue(const std::string& propertyName, float64 aValue) ;

	void setValue(const std::string& propertyName, uint16* aValue) ;
	void setValue(const std::string& propertyName, uint32* aValue) ;
	void setValue(const std::string& propertyName, uint64* aValue) ;
	void setValue(const std::string& propertyName, int16* aValue) ;
	void setValue(const std::string& propertyName, int32* aValue) ;
	void setValue(const std::string& propertyName, int64* aValue) ;
	void setValue(const std::string& propertyName, float32* aValue) ;
	void setValue(const std::string& propertyName, float64* aValue) ;

	/**
	 * Set the actual value of the property.
	 * Adds the formatted value.
	 *
	 * @param the property name
	 * @param the actual value
	 */
	void setValueAndFormattedValue(const std::string& propertyName, boost::any aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, bool aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, uint16 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, uint32 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, uint64 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, int16 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, int32 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, int64 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, float32 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, float64 aValue, const std::string& formattedValue) ;

	/**
	 * Returns the size of the property.
	 *
	 * @param the property name
	 * @return the set point property
	 */
	int32 getSize(const std::string& propertyName) const ;

	/**
	 * Set the size of the property.
	 *
	 * @param the property name
	 * @param the size value
	 */
	void setSize(const std::string& propertyName, int32 size) const ;

	/**
	 * Returns the extra size of the property.
	 *
	 * @param the property name
	 * @return the set point property
	 */
	int32 getExtraSize(const std::string& propertyName) const ;

	/**
	 * Set the extra size of the property.
	 *
	 * @param the property name
	 * @param the size value
	 */
	void setExtraSize(const std::string& propertyName, int32 size) const ;

	/**
	 * getProperties returns a std::stringified list of all properties available
	 * to a particular device. The property type (float64, long...) is
	 * returned also. The format is type:name\ntype:name\ntype... The name is
	 * the reference name used in all the get/set functions and the type
	 * allows clients to specify which get/set method to use.
	 */
	std::string getProperties();

	/**
	 * Set the actual value of the property without calling refresh method.
	 * Internal used
	 * @param the property name
	 * @param the actual value
	 */
	void updateValue(const std::string& propertyName, boost::any aValue) ;
	void updateValue(const std::string& propertyName, bool aValue) ;
	void updateValue(const std::string& propertyName, uint16 aValue) ;
	void updateValue(const std::string& propertyName, uint32 aValue) ;
	void updateValue(const std::string& propertyName, uint64 aValue) ;
	void updateValue(const std::string& propertyName, int16 aValue) ;
	void updateValue(const std::string& propertyName, int32 aValue) ;
	void updateValue(const std::string& propertyName, int64 aValue) ;
	void updateValue(const std::string& propertyName, float32 aValue) ;
	void updateValue(const std::string& propertyName, float64 aValue) ;

	void updateValue(const std::string& propertyName, uint16* aValue) ;
	void updateValue(const std::string& propertyName, uint32* aValue) ;
	void updateValue(const std::string& propertyName, uint64* aValue) ;
	void updateValue(const std::string& propertyName, int16* aValue) ;
	void updateValue(const std::string& propertyName, int32* aValue) ;
	void updateValue(const std::string& propertyName, int64* aValue) ;
	void updateValue(const std::string& propertyName, float32* aValue) ;
	void updateValue(const std::string& propertyName, float64* aValue) ;

	/**
	 * Set the actual value of the property without calling refresh method.
	 * Internal used
	 * @param the property name
	 * @param the actual value
	 */
	void updateValueAndFormattedValue(const std::string& propertyName, boost::any aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, bool aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, uint16 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, uint32 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, uint64 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, int16 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, int32 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, int64 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, float32 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, float64 aValue, const std::string& formattedValue) ;

	/**
	 * Returns a std::stringified list of available commands.
	 * Each command name is terminated by a \\n to separate the different names.
	 */
	std::string getCommands() const;

	/**
	 * Method to verify if a particular command exists.
	 */
	bool commandExists(const std::string& commandName) const;

	/*!
	 * \brief Method call for executing a command
	 *
	 * \param[in] aCommand: the command to apply on the device
	 */
	virtual void execute(const std::string& aCommand) = 0;

	/**
	 * Events handle methods
	 */
	void sendErrorEvent(int32 errorId, const std::string& errorMessage);
	void sendArrayEvent(const std::string& propertyName);
	void sendProgressEvent(int32 progress);

	/**
	 * Sets the name.
	 */
	void setName(const std::string& newName);

	/**
	 * Gets the name.
	 */
	std::string getName() const;

	/**
	 * Gets the type.
	 */
	std::string getType() const;

	/**
	 * Sets the type.
	 */
	void setType(const std::string& aType);

	/**
	 * Adds the type.
	 */
	void addType(const std::string& aType);

	/**
	 * Gets the current state.
	 */
	DeviceDriverState* getCurrentState();

	/**
	 * Gets the running mode.
	 */
	uint32 getRunningMode();

	/**
	 * Return device function.
	 */
	virtual std::string getFunction() const;

	/**
	 * Return parent device function.
	 */
	virtual std::string getParentFunction() const;

	/**
	 * Gets the progression.
	 */
	int32 getProgression() const;

	/**
	 * Returns the log level.
	 */
	Level getLogLevel(Level const & level) const;

public:
	/*
	 *---------------------------------------------------------------------
	 * Updating utilities
	 *---------------------------------------------------------------------
	 */

	/**
	 * This method returns the property m_terminated.
	 * The flag returns the action status (terminated or not).
	 *
	 * @return the flag terminated
	 */
	bool isTerminated();

	/*!
	 * \brief Method called before changing the property value
	 *
	 * This method is called for doing pre-treatment before changing the property value
	 *
	 * \param[in] propertyName: the property name
	 * \param[in] aValue: the new value of the property
	 * \throws CannotSetValue: the value isn't corrected, or property couldn't be changed
	 */
	virtual void refreshSetValue(std::string propertyName, const boost::any aValue)  {
		// TODO , to remove after module intgration, or remove all inherite method DeviceDriver::refreshSetValue
	}

	/**
	 * Calls refresher's followed by refreshSetValue.
	 */
	void refreshValue(const std::string& propertyName, const boost::any& value) ;

	/*
	 *---------------------------------------------------------------------
	 * Driver utilities
	 *---------------------------------------------------------------------
	 */

	/**
	 * Verifies if the Controller is configured and can be used.
	 */
	virtual bool isConfigured() const;

	/**
	 * Give the root protcol of this driver (vme,rs233,gpib...)
	 */
	std::string getRootDeviceType();

	virtual void write(const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, const void * aData);
	virtual void read(const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, void * aData);
	virtual void readMBLT(const uint32 anOffset, const uint32 numberOfBytes, void * aData);

	virtual void write(const DeviceDriver *aDeviceDriver, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, const void * aData);
	virtual void read(const DeviceDriver *aDeviceDriver, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, void * aData);
	virtual void readMBLT(const DeviceDriver* aDeviceDriver, const uint32 anOffset, const uint32 numberOfBytes, void * aData);

	virtual void write(const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, const void * aData);
	virtual void read(const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, void * aData);

	virtual void write(const DeviceDriver *aDeviceDriver, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, const void * aData);
	virtual void read(const DeviceDriver *aDeviceDriver, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, void * aData);

	virtual void write(const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, const void * aData);
	virtual void read(const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, void * aData);
	virtual void readMBLT(const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 numberOfBytes, void * aData);

	virtual void write(const DeviceDriver *aDeviceDriver, const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, const void * aData);
	virtual void read(const DeviceDriver *aDeviceDriver, const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 sizeOfData, const uint32 numberOfData, void * aData);
	virtual void readMBLT(const DeviceDriver* aDeviceDriver, const uint32 anAddress, const uint32 anAddressType, const uint32 anOffset, const uint32 numberOfBytes, void * aData);

	virtual void write(const std::string& aData, bool doClear = false);
	virtual void read(std::string& aData, const std::string aEnd);

	virtual void write(const DeviceDriver* aDeviceDriver, const std::string& aData, bool doClear = false);
	virtual void read(const DeviceDriver* aDeviceDriver, std::string& aData, const std::string aEnd);

	virtual void write(const uint8 *aData, int32 dataLen);
	virtual void read(uint8 *aData, int32 dataLen);

	virtual void write(const DeviceDriver* aDeviceDriver, const uint8 *aData, int32 dataLen);
	virtual void read(const DeviceDriver* aDeviceDriver, uint8 *aData, int32 dataLen);

	virtual void readsp(std::string& aData);
	virtual void readsp(const DeviceDriver* aDeviceDriver, std::string& aData);

	/*
	 *---------------------------------------------------------------------
	 * Link utilities
	 *---------------------------------------------------------------------
	 */

	/**
	 * Set the device link on parent
	 *
	 * @param the device parent or child
	 */
	void setParent(DeviceDriver* aDevice) ;

	/**
	 * Remove the device link on parent.
	 */
	void removeParent() ;

	/**
	 * Get the device link list
	 */
	DeviceDriver* getParent() const ;

	/**
	 * Set the device link on child
	 *
	 * @param the device parent or child
	 * @param the channel of the device
	 */
	void setChild(DeviceDriver* aDevice, const int32 aChannel, const int32 aSize = 1) ;

	/**
	 * Remove the device link on child.
	 *
	 * @param the channel of the device
	 */
	void removeChild(const int32 aChannel) ;

	/**
	 * Returns the device link of the parent or child. If there is no child for the
	 specified channel, a null reference is returned.
	 *
	 * @return the device link
	 */
	DeviceDriver* getChild(int32 channel) const ;

	/**
	 * Return the channel position of a device in the list of device link.
	 */
	int32 getChannel() const;

	/**
	 * Return the number channels (list of device link).
	 */
	uint32 getNumberOfChannels() const;

	/**
	 * Return true if the list of device link is empty.
	 * Parent got no child
	 */
	bool isChildEmpty() const;

	/**
	 * Gets the ID of the port. Searches for the root ancestor and getting its name.
	 */
	std::string getPortID() const;

	/**
	 * Get the Spy command list for update notification
	 */
	std::vector<std::string> getSpyCommands();

	/**
	 * Get the Observer command list for update notification
	 */
	std::vector<std::string> getObserverCommands();

	/**
	 * Returns true if category is present.
	 */
	bool containsCategory(const std::string& category);

	/**
	 * Disable flag, temporary, need for destructor to know enable state
	 * will be disapire when integrates data base
	 */
	virtual void setEnabled(bool aValue);

	/*
	 * reset the observer counter
	 */
	void resetObserverCounters();

	/*
	 *---------------------------------------------------------------------
	 * Disconnect method
	 *---------------------------------------------------------------------
	 */
	void incrementDisconnectedCounter(const std::string& command);
	void clearDisconnectedCounter(const std::string& command);
	uint32 getDisconnectedCounter(const std::string& command) const;
	void setDisconnected();
	void clearDisconnected();
	bool isDisconnected() const;

	void setPauseRequested(bool avalue);
	bool pauseRequested() const;

protected:
	void constructor(const std::string& aType, const std::string& aName);
	void init(const std::string& aName);

	/*
	 * initCommand allows subclasses to add new properties to the
	 * CommandManager list.
	 */
	void initCommand(const std::string& commandName) const;

	/**
	 * Registers a Spy command for the update notification. The real period is numberOfCycles multiplied by 5 seconds.
	 */
	void registerSpyCommand(const std::string& commandName, int32 numberOfCycles);

	/**
	 * Registers a Spy command for the update notification with a period in milliseconds.
	 */
	void registerSpyCommandMs(const std::string& commandName, int32 periodMs);

	/**
	 * Registers an Observer command for the update notification. The real period is numberOfCycles multiplied by 20 milliseconds.
	 */
	void registerObserverCommand(const std::string& commandName, int32 numberOfCycles);

	/**
	 * Registers an Observer command for the update notification with a period in milliseconds.
	 */
	void registerObserverCommandMs(const std::string& commandName, int32 periodMs);

	/**
	 * Register the device function
	 */
	void registerFunction(const std::string& function);

	/**
	 * Register the parent device function
	 */
	void registerParentFunction(const std::string& function);

private:
	/**
	 * Test if a property has changed. Compare new and old values.
	 */
	void hasChanged(const std::string& propertyName, boost::any oldValue, boost::any newValue);
	void refreshProgression(int32 value);

	uint32 m_runningMode;
	std::vector<std::string> m_categoryList;
	DeviceDriverState* m_realState;
	DeviceDriverState* m_perfectState;
	DeviceDriverState* m_simulatedState;
	DeviceDriverState* m_currentState;

	DeviceDriver* m_parent;
	std::vector<DeviceDriver*> m_child;

	CommandManager* m_CommandManager;

	std::string m_Function;
	std::string m_ParentFunction;

	std::vector<std::string> m_SpyCommands;
	std::map<std::string, int32> m_SpyNumberOfCycles;
	std::map<std::string, int32> m_SpyCounter;
	std::vector<std::string> m_ObserverCommands;
	std::map<std::string, int32> m_ObserverNumberOfCycles;
	std::map<std::string, int32> m_ObserverCounter;

	std::map<std::string, boost::any> m_ChangeValues;

	bool m_Enable;
	bool m_PauseRequested;
	bool m_PauseActivated;

	std::map<std::string, uint32> m_DisconnectedCounter;

	Module * m_module;

	static const char PROGRESSION_PROPERTY[];

	PropertyManager* m_propertyManager;

public:
	Property<int32> status;
	Property<int32> commandProgression;
	Property<std::string> version;
	Property<std::string> date;
	Property<std::string> deviceType;
	Property<int32> leafChannel;
	Property<int32> numberLeaf;
	Property<std::string> deviceName;
	Property<std::string> type;
	Property<std::string> state;
	Property<bool> startActivated;
	Property<bool> disconnected;
	StatusProperty commandStatus;

	/*
	 * type of device
	 */
	static const std::string ROOT_DEVICE_TYPE_DEVICE_CONTAINER;
	static const std::string BRANCH_DEVICE_TYPE_DEVICE_CONTAINER;
	static const std::string LEAF_DEVICE_TYPE_DEVICE_CONTAINER;

protected:

	void refreshStateProperty(const std::string& value);
	virtual void refreshNumberLeafProperty(int32  value) ;

};

static const char NONE_FUNCTION[] = "none";

inline void DeviceDriver::setValue(const std::string& propertyName, bool aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, uint16 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, uint32 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, uint64 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, int16 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, int32 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, int64 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, float32 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, float64 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, uint16* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, uint32* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, uint64* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, int16* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, int32* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, int64* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, float32* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValue(const std::string& propertyName, float64* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, bool aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, uint16 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, uint32 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, uint64 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, int16 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, int32 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, int64 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, float32 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::setValueAndFormattedValue(const std::string& propertyName, float64 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::updateValue(const std::string& propertyName, bool aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, uint16 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, uint32 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, uint64 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, int16 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, int32 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, int64 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, float32 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, float64 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, uint16* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, uint32* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, uint64* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, int16* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, int32* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, int64* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, float32* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValue(const std::string& propertyName, float64* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, bool aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, uint16 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, uint32 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, uint64 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, int16 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, int32 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, int64 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, float32 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void DeviceDriver::updateValueAndFormattedValue(const std::string& propertyName, float64 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

#endif //DEVICEDRIVER_H
