/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DEVICEDRIVERCREATOR_H
#define DEVICEDRIVERCREATOR_H

#include <string>
#include <map>
#include "DeviceContainer/DeviceDriver.h"

/**
 * The DeviceDriverCreator class is a singleton which delivers the generics
 * functions for creating DeviceDriver objects. This class is the interface 
 * for other modules. 
 */
class DeviceDriverCreator {

protected:

	/**
	 * The constructor of the class DeviceDriverCreator.
	 */
	DeviceDriverCreator();

public:
	/**
	 * Get the instance.
	 *
	 * @return the pointer of the singleton class (itself)
	 */
	static DeviceDriverCreator* getInstance();

	/**
	 * Instance deletion.
	 */
	static void reset();

	/**
	 * Initialises the list of controllers loaded in modules.
	 */
	void initFromModules();

	/**
	 * Give the list of all device ID.
	 *
	 * @return the list
	 */
	std::string getKnownIDs() const;

	/**
	 * Check if a device ID exist.
	 *
	 * @param the device ID
	 * @return the result found or not
	 */
	bool verifyType(const std::string& aDriverID) const;

	/**
	 * It creates the write device driver in accordance
	 * with the category and type of device that you wanted
	 *
	 * @param the type of device
	 * @param the name of device
	 * @param the running mode (real, sim ...)
	 * @return the pointer on the created device driver class
	 */
	DeviceDriver* create(const std::string& aType, const std::string& aName, const uint32 aRunningMode = NORMAL_MODE_DEVICE_CONTAINER, bool create = true);

	/**
	 * Gets the list of device driver types.
	 */
	const std::vector<std::string>& getDeviceDriverList() const;

	/**
	 * Structure for storing meta information of drivers.
	 */
	struct TypeInfo {
		std::string cppClass;
		std::string cppClassName;
		std::string categories;
	};

	/**
	 * Creates all the type to get the full list of (type, class)
	 */
	std::map<std::string, TypeInfo> retrieveAllTypeClasses();

private:

	std::vector<std::string> m_DriverTypes;

	// type, class, class name, categories
	std::map<std::string, TypeInfo> m_typeClassMap;

	/**
	 * @link
	 * @shapeType PatternLink
	 * @pattern <{DeviceDriverCreator}>
	 * @supplierRole Singleton factory
	 */
	/*# DeviceDriverCreator _singleton; */
	static DeviceDriverCreator* m_instance;

};
#endif //DEVICEDRIVERCREATOR_H
