/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "HardwareCommand.h"
#include "DeviceContainer/DeviceDriver.h"
#include "DeviceContainer/Spy.h"
#include "DeviceContainer/Observer.h"

#include <iostream>
#include <thread>
#include "CommandNames.h"

using namespace std;
using namespace boost;

HardwareCommand::HardwareCommand(DeviceDriver* aDeviceDriver,
		const std::string& name) :
	m_deviceDriver(aDeviceDriver), m_name(name) {
}

HardwareCommand::~HardwareCommand() {
}

void HardwareCommand::execute() {

	bool exec = true;

	try {
		if (!m_deviceDriver->isConfigured()) {
			exec = false;
		}

		if (m_deviceDriver->isDisconnected()) {
			exec = false;
		}

		if (exec) {

			if (m_name == COMMAND_PAUSE) {
				m_deviceDriver->setPauseRequested(true);
			}
			// Call device execute
			m_deviceDriver->execute(m_name);
			// Clear disconnected counter
			m_deviceDriver->clearDisconnectedCounter((m_name));

			if ((m_deviceDriver->pauseRequested() == true) && (m_name
					== COMMAND_START)) {
				// Pause was requested, so do it now
				m_deviceDriver->execute(COMMAND_PAUSE);
				m_deviceDriver->execute(COMMAND_STATUS);
			} else if ((m_deviceDriver->pauseRequested() == true) && ((m_name
						== COMMAND_CONTINUE) || (m_name
								== COMMAND_STOP))) {
				// Clear pause requested
				m_deviceDriver->setPauseRequested(false);
			} else if ((m_name != COMMAND_STATUS) && (m_name
					!= COMMAND_READ)) {
				m_deviceDriver->execute(COMMAND_STATUS);
			}

		} else {
			// Force terminated to free waiting of the abstract controller
			m_deviceDriver->sendProgressEvent(0);
			// Force terminated to free waiting of the abstract controller
			m_deviceDriver->sendProgressEvent(PROGRESSION_END_DEVICE_CONTAINER);
		}
	} catch (const DeviceDriver::BusError& e) {
		checkDisconnected();
	} catch (const DeviceDriver::TimeOut& e) {
		checkDisconnected();
	} catch (const Exceptions& e) {
		// Continu throw for unitary test
		if (m_deviceDriver->getRunningMode() == NORMAL_MODE_DEVICE_CONTAINER) {
		} else {
			throw ;
		}
		// Force terminated to free waiting of the abstract controller
		m_deviceDriver->sendProgressEvent(0);
		// Force terminated to free waiting of the abstract controller
		m_deviceDriver->sendProgressEvent(PROGRESSION_END_DEVICE_CONTAINER);
	} catch (const boost::bad_any_cast& e) {
		cout << "Bad cast " << m_deviceDriver->getName()
		<< ", command = " << m_name << endl;
		if (m_deviceDriver->getRunningMode() != NORMAL_MODE_DEVICE_CONTAINER) {
			throw;
		}
		// Force terminated to free waiting of the abstract controller
		m_deviceDriver->sendProgressEvent(0);
		// Force terminated to free waiting of the abstract controller
		m_deviceDriver->sendProgressEvent(PROGRESSION_END_DEVICE_CONTAINER);
	} catch (const std::exception& e) {
		cout << "Error in command " << m_name << " of " << m_deviceDriver->getName() << " : " << e.what() << endl;
	} catch (...) {
		cout << "Unknown error : " << m_deviceDriver->getName()
		<< ", command = " << m_name << endl;
		// Continu throw for unitary test
		if (m_deviceDriver->getRunningMode() != NORMAL_MODE_DEVICE_CONTAINER) {
			throw;
		}
		// Force terminated to free waiting of the abstract controller
		m_deviceDriver->sendProgressEvent(0);
		// Force terminated to free waiting of the abstract controller
		m_deviceDriver->sendProgressEvent(PROGRESSION_END_DEVICE_CONTAINER);
	}
}

void HardwareCommand::stop() {
	// The command is valid since it has been launched by the scheduler. That is why no test is required.

	m_deviceDriver->execute(COMMAND_STOP);
}

bool HardwareCommand::isBlocking(Command& aCommand) {
	HardwareCommand& c = dynamic_cast<HardwareCommand&> (aCommand);

	return m_deviceDriver->getPortID() == c.getDeviceDriver()->getPortID();
}

bool HardwareCommand::reject(bool blocked) {
	// The commands are never rejected.
	return false;
}

void HardwareCommand::checkDisconnected() {
	// Continu throw for unitary test
//	if (m_deviceDriver->getRunningMode() == NORMAL_MODE_DEVICE_CONTAINER) {
//	} else {
//		throw ;
//	}
	// Test if disconnected
	if (m_deviceDriver->getDisconnectedCounter(m_name) >= 10) {
		m_deviceDriver->setDisconnected();
	}
	else {
		m_deviceDriver->incrementDisconnectedCounter(m_name);
	}
}

DeviceDriver* HardwareCommand::getDeviceDriver() const {
	return m_deviceDriver;
}

const std::string& HardwareCommand::getName() const {
	return m_name;
}

ostream& operator<<(ostream& os, const HardwareCommand &c) {
	os << c.toString();
	return os;
}

bool HardwareCommand::equal(Command* c1) {
	return ((static_cast<HardwareCommand*> (c1)->getDeviceDriver()
			== m_deviceDriver)
			&& (static_cast<HardwareCommand*> (c1)->getName() == m_name));
}

std::string HardwareCommand::toString() const {
	return getDeviceDriver()->getName() + " " + getName();
}
