/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DeviceContainer/Observer.h"
#include "DeviceContainer/Spy.h"
#include "Scheduler/Sender.h"
#include "DeviceContainer/HardwareCommand.h"
#include "DeviceContainer/DeviceDriver.h"
#include "DeviceContainer/DeviceDriverCounter.h"
#include "LoggingServices/Level.h"
#include "Utilities/Simulated.h"
#include <common/base/ServerProperties.h>

using namespace std;
using namespace common;

Observer* Observer::_instance = new Observer();

const std::string Observer::OBSERVER_CYCLE_DURATION_US = "observerCycleDuration-us";

Observer::Observer() :
	_sleepTimeUs(20000), _active(0), _running(false), _thread(nullptr) {
}

Observer::~Observer() {

	if (_thread != nullptr) {
		delete _thread;
	}
}

Observer* Observer::getInstance() {
	return _instance;
}

void Observer::init() {

	try {
		_sleepTimeUs = ServerProperties::getInstance()->get<int32>(OBSERVER_CYCLE_DURATION_US);
	}
	catch (const Exception& e) {
		out << "Setting default value 20000us to property 'observerCycleDuration-us'" << eol;
	}
}

void Observer::reset() {
	{
		std::scoped_lock lkObserver(_observerMutex);

		_deviceDriverCounters.clear();
		_running = false;
	}

	_thread->join();
}

void Observer::resetInstance() {
	_instance->reset();
	delete _instance;
}

int32 Observer::getCycleDurationUs() const {
	return _sleepTimeUs;
}

void Observer::execute() {
	_thread = new std::thread(&Observer::loop);
}

void Observer::loop() {
	Observer::getInstance()->loopPhases();
}

void Observer::loopPhases() {

	_running = true;

	while (_running) {
		_phase = "";

		{
			std::scoped_lock lkObserver(_cmdObserverMutex);
			_phase = _phase + " treatment ";
		}

		treatmentPhase();

		{
			std::scoped_lock lkObserver(_cmdObserverMutex);
			_phase = _phase + " read ";
		}

		readPhase();

		{
			std::scoped_lock lkObserver(_cmdObserverMutex);
			_phase = _phase + " write ";
		}

		writePhase();

		{
			std::scoped_lock lkObserver(_cmdObserverMutex);
			_phase = _phase + " control ";
		}

		controlCyclePhase();

		{
			std::scoped_lock lkObserver(_cmdObserverMutex);
			_phase = _phase + " sleep ";
		}

		if (Simulated::activated == true) {
			usleep(2000);
		}
		else {
			usleep(_sleepTimeUs);
		}
	}

}

void Observer::readPhase() {

	std::scoped_lock lkObserver(_observerMutex);

	_startDate = common::Date();

	// Loop through all the registered DeviceDrivers
	vector<DeviceDriverCounter*>::iterator DeviceDriverCounterIter;

	for (DeviceDriverCounterIter = _deviceDriverCounters.begin(); DeviceDriverCounterIter != _deviceDriverCounters.end(); DeviceDriverCounterIter++) {
		// Check if the frequency and counter are the same: if so this
		// one is to be refreshed
		DeviceDriverCounter* theDeviceDriverCounter = *DeviceDriverCounterIter;

		// Increment the counter
		theDeviceDriverCounter->incrementCounter();

		if ((theDeviceDriverCounter->getPeriod() == theDeviceDriverCounter->getCounter()) || (Simulated::activated == true)) {

			// Reset the counter
			theDeviceDriverCounter->resetCounter();

			DeviceDriver* theDeviceDriver = theDeviceDriverCounter->getDeviceDriver();

			// Get the list of commands from the DeviceDriver needed for the
			// refresh (can vary from call to call - handled by the DeviceDriver)
			vector<string> commandList = theDeviceDriver->getObserverCommands();

			// Loop through the commands and execute each on in turn
			for (size_t listIndex = 0; listIndex < commandList.size(); listIndex++) {

				string aCommand = commandList[listIndex];
				Sender<HardwareCommand>::post(HardwareCommand(theDeviceDriver, aCommand));
			}
		}
	}

}

void Observer::writePhase() {
	// do nothing
}

void Observer::treatmentPhase() {
	// do nothing
}

void Observer::controlCyclePhase() {
	// do nothing
}

void Observer::registerDeviceDriver(DeviceDriver* aDeviceDriver, int32 aFrequency) {

	// Lock the Map during operation
	std::scoped_lock lkObserver(_observerMutex);

	vector<DeviceDriverCounter*>::iterator DeviceDriverCounterIter;

	for (DeviceDriverCounterIter = _deviceDriverCounters.begin(); DeviceDriverCounterIter != _deviceDriverCounters.end(); DeviceDriverCounterIter++) {
		if ((*DeviceDriverCounterIter)->getDeviceDriver() == aDeviceDriver) {
			return;
		}
	}

	//Create a new container for the DeviceDriver and frequency
	DeviceDriverCounter* aDeviceDriverCounter = new DeviceDriverCounter(aDeviceDriver, aFrequency);

	// insert the new DeviceDriver in the list
	_deviceDriverCounters.push_back(aDeviceDriverCounter);

	// Deactivate the DD from the Spy thread if it is there
	Spy::getInstance()->deactivateDeviceDriver(aDeviceDriver);

	// Keep a note on how many DDs are registered: we take charge of the
	// Change Manager if it is us.
	_active++;
}

void Observer::unregisterDeviceDriver(DeviceDriver* aDeviceDriver) {

	// Lock the Map during operation
	std::scoped_lock lkObserver(_observerMutex);
	vector<DeviceDriverCounter*>::iterator theIterator;
	vector<DeviceDriverCounter*>::iterator theOneToDie;
	bool thereIsOne = false;

	// Loop through the vector of DeviceDriverCounters and delete the correct one
	for (theIterator = _deviceDriverCounters.begin(); theIterator != _deviceDriverCounters.end(); theIterator++) {

		DeviceDriverCounter* theDeviceDriverCounter = *theIterator;

		if (theDeviceDriverCounter->getDeviceDriver() == aDeviceDriver) {
			theOneToDie = theIterator;
			thereIsOne = true;
		}
	}

	if (thereIsOne) {
		DeviceDriverCounter* theDeviceDriverCounter = *theOneToDie;
		aDeviceDriver->resetObserverCounters();
		delete theDeviceDriverCounter;
		_deviceDriverCounters.erase(theOneToDie);
	}

	// Reactivate the DD on the Spy thread if it is there
	Spy::getInstance()->activateDeviceDriver(aDeviceDriver);

	// Keep a note on how many DDs are registered: we take charge of the
	// Change Manager if it is us.
	_active--;
	if (_active == 0) {
		//Spy::instance()->inCharge();
	}

}

string Observer::getPhase() {
	std::scoped_lock lkObserver(_cmdObserverMutex);
	return _phase;
}

vector<DeviceDriverCounter*> Observer::getDeviceDriverCounters() {
	std::scoped_lock lkObserver(_cmdObserverMutex);
	return _deviceDriverCounters;
}
/*
int32 Observer::getListDeviceDriverActivate() {
	std::scoped_lock lkObserver(_cmdObserverMutex);
	return _active;
}*/

void Observer::printDurationSinceStart() {

	int duration;
	{
		std::scoped_lock lkSpy(_observerMutex);

		common::Date date = common::Date();
		duration = (date - _startDate).getMilliseconds();
	}

	cout << "Duration since observer start " << duration << "ms" << endl;

}
