/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DeviceContainer/Spy.h"
#include "Scheduler/Sender.h"
#include "DeviceContainer/HardwareCommand.h"
#include "DeviceContainer/DeviceDriver.h"
#include "DeviceContainer/DeviceDriverCounter.h"
#include "LoggingServices/Level.h"
#include "Utilities/Simulated.h"
#include <common/base/ServerProperties.h>

using namespace std;
using namespace common;

Spy* Spy::_instance = new Spy();

const std::string Spy::SPY_CYCLE_DURATION_MS = "spyCycleDuration-ms";

Spy::Spy() :
	_sleepTimeMs(5000), _inCharge(true), _running(false), _thread(nullptr) {
}

Spy::~Spy() {

	if (_thread != nullptr) {
		delete _thread;
	}
}

Spy* Spy::getInstance() {
	return _instance;
}

void Spy::init() {

	try {
		_sleepTimeMs = ServerProperties::getInstance()->get<int32>(SPY_CYCLE_DURATION_MS);
	}
	catch (const Exception& e) {
		out << "Setting default value 5000ms to property 'spyCycleDuration-ms'" << eol;
	}
}

void Spy::stop() {
	{
		std::scoped_lock lkSpy(_spyMutex);

		_deviceDriverCounters.clear();
		_deactivationQueue.clear();
		_activationQueue.clear();
		_running = false;
	}

	_thread->join();
}

void Spy::resetInstance() {
	delete _instance;
}

int32 Spy::getCycleDurationMs() const {
	return _sleepTimeMs;
}

void Spy::execute() {
	_thread = new std::thread(&Spy::loop);
}

void Spy::loop() {
	Spy::getInstance()->loopPhases();
}

void Spy::loopPhases() {

	_running = true;

	while (_running) {
		_phase = "";

		if (_inCharge && !Simulated::activated) {

			{
				std::scoped_lock lkSpy(_cmdSpyMutex);
				_phase = _phase + " treatment ";
			}

			treatmentPhase();

			{
				std::scoped_lock lkSpy(_cmdSpyMutex);
				_phase = _phase + " read ";
			}

			readPhase();

			{
				std::scoped_lock lkSpy(_cmdSpyMutex);
				_phase = _phase + " write ";
			}

			writePhase();

			{
				std::scoped_lock lkSpy(_cmdSpyMutex);
				_phase = _phase + " control ";
			}

			controlCyclePhase();

			{
				std::scoped_lock lkSpy(_cmdSpyMutex);
				_phase = _phase + " sleep ";
			}

		}

		usleep(_sleepTimeMs * 1000);
	}
}

void Spy::readPhase() {

	std::scoped_lock lkSpy(_spyMutex);

	// Loop through all the registered DeviceDrivers
	vector<DeviceDriverCounter*>::iterator DeviceDriverCounterIter;

	for (DeviceDriverCounterIter = _deviceDriverCounters.begin(); DeviceDriverCounterIter != _deviceDriverCounters.end(); DeviceDriverCounterIter++) {

		// Check if the frequency and counter are the same: if so this
		// one is to be refreshed
		DeviceDriverCounter* theDeviceDriverCounter = *DeviceDriverCounterIter;

		// Increment the counter
		theDeviceDriverCounter->incrementCounter();

		if (theDeviceDriverCounter->getPeriod() == theDeviceDriverCounter->getCounter()) {

			// Reset the counter
			theDeviceDriverCounter->resetCounter();

			DeviceDriver* theDeviceDriver = theDeviceDriverCounter->getDeviceDriver();

			// Get the list of commands from the DeviceDriver needed for the
			// refresh (can vary from call to call - handled by the DeviceDriver)
			vector<string> commandList = theDeviceDriver->getSpyCommands();

			// Loop through the commands and execute each on in turn
			for (size_t listIndex = 0; listIndex < commandList.size(); listIndex++) {

				string aCommand = commandList[listIndex];
				Sender<HardwareCommand>::post(HardwareCommand(theDeviceDriver, aCommand));
			}

		}
	}

}

void Spy::writePhase() {
	// do nothing
}

void Spy::treatmentPhase() {
	std::scoped_lock lkSpy(_spyMutex);

	vector<DeviceDriverCounter*>::iterator theIterator;

	// Treat the activation and deactivation queues
	for (size_t i = 0; i < _activationQueue.size(); i++) {
		DeviceDriver* aDeviceDriver = _activationQueue[i];

		for (theIterator = _deviceDriverCounters.begin(); theIterator != _deviceDriverCounters.end(); theIterator++) {

			DeviceDriverCounter* theDeviceDriverCounter = *theIterator;

			if (theDeviceDriverCounter->getDeviceDriver() == aDeviceDriver) {
				theDeviceDriverCounter->activate();
			}
		}
	}
	_activationQueue.clear();

	for (size_t i = 0; i < _deactivationQueue.size(); i++) {

		DeviceDriver* aDeviceDriver = _deactivationQueue[i];

		// Loop through the vector of DeviceDriverCounters
		for (theIterator = _deviceDriverCounters.begin(); theIterator != _deviceDriverCounters.end(); theIterator++) {

			DeviceDriverCounter* theDeviceDriverCounter = *theIterator;
			if (theDeviceDriverCounter->getDeviceDriver() == aDeviceDriver) {
				theDeviceDriverCounter->deactivate();

			}
		}
	}
	_deactivationQueue.clear();

}

void Spy::controlCyclePhase() {
	// do nothing
}

void Spy::registerDeviceDriver(DeviceDriver* aDeviceDriver, int32 aFrequency) {

	// Lock the Map during operation
	std::scoped_lock lkSpy(_spyMutex);

	// Create a new container for the DeviceDriver and frequency
	DeviceDriverCounter* aDeviceDriverCounter = new DeviceDriverCounter(aDeviceDriver, aFrequency);

	// insert the new DeviceDriver in the list
	_deviceDriverCounters.push_back(aDeviceDriverCounter);

}

void Spy::unregisterDeviceDriver(DeviceDriver* aDeviceDriver) {

	// Lock the Map during operation
	std::scoped_lock lkSpy(_spyMutex);

	vector<DeviceDriverCounter*>::iterator theIterator;
	vector<DeviceDriverCounter*>::iterator theOneToDie;
	bool thereIsOne = false;

	// Loop through the vector of DeviceDriverCounters and delete the correct one
	for (theIterator = _deviceDriverCounters.begin(); theIterator != _deviceDriverCounters.end(); theIterator++) {

		DeviceDriverCounter* theDeviceDriverCounter = *theIterator;

		if (theDeviceDriverCounter->getDeviceDriver() == aDeviceDriver) {
			theOneToDie = theIterator;
			thereIsOne = true;
		}
	}

	if (thereIsOne) {
		DeviceDriverCounter* theDeviceDriverCounter = *theOneToDie;
		delete theDeviceDriverCounter;
		_deviceDriverCounters.erase(theOneToDie);
	}

}

void Spy::activateDeviceDriver(DeviceDriver* aDeviceDriver) {

	// Lock the Map during operation
	std::scoped_lock lkSpy(_spyMutex);

	// Iterate through de-activation queue to see if the DD exists
	// there too. This action cancels the deactivation.
	vector<DeviceDriver*>::iterator i = _deactivationQueue.begin();
	bool found = false;
	while (i != _deactivationQueue.end()) {
		if (*i == aDeviceDriver) {
			found = true;
			_deactivationQueue.erase(i);
			break;
		} else {
			i++;
		}
	}

	if (!found) {
		_activationQueue.push_back(aDeviceDriver);
	}

}

void Spy::deactivateDeviceDriver(DeviceDriver* aDeviceDriver) {

	// Lock the Map during operation
	std::scoped_lock lkSpy(_spyMutex);

	// Iterate through activation queue to see if the DD exists
	// there too. This action cancels the activation.
	vector<DeviceDriver*>::iterator i = _activationQueue.begin();
	bool found = false;
	while (i != _activationQueue.end()) {
		if (*i == aDeviceDriver) {
			found = true;
			_activationQueue.erase(i);
			break;
		} else {
			i++;
		}
	}

	if (!found) {
		_deactivationQueue.push_back(aDeviceDriver);
	}
}

string Spy::getPhase() {
	std::scoped_lock lkSpy(_cmdSpyMutex);
	return _phase;
}

vector<DeviceDriver*> Spy::getListDeviceDriverActivate() {
	std::scoped_lock lkSpy(_cmdSpyMutex);
	return _activationQueue;
}

vector<DeviceDriverCounter*> Spy::getDeviceDriverCounter() {
	std::scoped_lock lkSpy(_cmdSpyMutex);
	return _deviceDriverCounters;
}

bool Spy::isInCharge() const {
	return _inCharge;
}

void Spy::inCharge() {
	_inCharge = true;
}

void Spy::notInCharge() {
	_inCharge = false;
}
