/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "InstrumentManager/NameManager.h"
#include "InstrumentAbstraction/AbstractController.h"
#include "InstrumentAbstraction/ExperimentController.h"
#include "DataProvider/ChangeAspect.h"
#include "DataProvider/Property/AnyProperty.h"
#include "HistogrammingServices/HistogramConfigLoader.h"
#include "HistogrammingServices/HistogramManager.h"
#include "common/base/ServerProperties.h"
#include <common/base/Convert.h>
#include <fstream>

using namespace std;

const std::string HistogramConfigLoader::HISTOGRAM_CONFIG_NAME = "HistoFileConfig.txt";
const std::string HistogramConfigLoader::DEFAULT_TIME = "default";

HistogramConfigLoader::HistogramConfigLoader() {
}

HistogramConfigLoader::~HistogramConfigLoader() {
}

void HistogramConfigLoader::loadConfig() {

	string home;
	home.append(common::ServerProperties::getInstance()->getConfigPath());
	home.append("/");

	string filename;
	filename.append(home);
	filename.append(HISTOGRAM_CONFIG_NAME);

	cout << "Loading config " << filename << endl;

	ifstream configFile(filename.c_str());

	if (!configFile) {
		cout << "\tNo histogram file configuration: " << endl;
	} else {
		bool error = false;
		bool done = false;

		cout << "\tReading histogram file configuration: " << endl;
		while (!error && !configFile.eof() && !done) {
			string line;
			getline(configFile, line);
			//cout << line << endl;
			if (line == "<EOF>") {
				done = true;

			} else if (!line.empty()) {
				decodeDataController(line);
			}
		}

	}

	configFile.close();
}

void HistogramConfigLoader::decodeDataController(const string& line) {

	// Get the abstract controller name and property name.
	size_t index[2];
	index[0] = line.find_first_of(',');
	string controllerName = line.substr(0, index[0]);
	index[1] = line.find(',', index[0] + 1);

	string propertyName;

	if (index[1] != string::npos) {
		propertyName = line.substr(index[0] + 1, index[1] - index[0] - 1);
	}
	else {
		propertyName = line.substr(index[0] + 1);
	}

	// Active time value.
	optional<int> value;
	if (index[1] != string::npos) {
		string valueString = line.substr(index[1] + 1);
		if (valueString == DEFAULT_TIME) {
			value = -1;
		}
		else {
			value = std::stoi(valueString);
		}
	}
	else {
		// Set default value.
		value = -1;
	}

	AbstractController* controller = getAbstractController(controllerName);

	if (controller != nullptr) {
		AnyProperty* property = getProperty(controller, propertyName);

		if (property != nullptr) {

			// Test dynamic property.
			PropertyManager::DynamicPropertyInfo info = PropertyManager::getDynamicPropertyInfo(propertyName);

			if (!info.isDynamic) {
				HistogramManager::getInstance()->addHistogram(controller->getName(), property->getName(), value);
			} else {
				HistogramManager::getInstance()->addHistogram(controller->getName(), info.name, info.index, value);
			}
		}
	}



}

AbstractController* HistogramConfigLoader::getAbstractController(const string& controllerName) {

	AbstractController* controller = nullptr;

	try {
		controller = NameManager::getInstance()->getController(controllerName);
	}
	catch (NameManager::ControllerNotFound& cnf) {
		common::err << "Cannot find " << controllerName << " in histogram config file" << common::eol;
	}
	return controller;

}

AnyProperty* HistogramConfigLoader::getProperty(AbstractController* ac, const string& propertyName) {

	AnyProperty* property = nullptr;

	if (ac) {
		try {
			property = ac->getPropertyManager()->getProperty(propertyName);
		}
		catch (PropertyManager::UnknownProperty& up) {
			common::err << "Cannot find Property " << propertyName << " for " << ac->getName() << " in histogram config file" << common::eol;
		}
	}

	return property;

}
