/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "IcsServerManager.h"
#include "DataProvider/Property/PropertyInfos.h"
#include "DataProvider/ChangeAspect.h"
#include "CommandSystem/ConfigurationManager.h"
#include "CommandSystem/SimulatedConfigurationManager.h"
#include "Scheduler/Scheduler.h"
#include "InstrumentAbstraction/ClientCommand.h"
#include "HistogrammingServices/HistogramManager.h"
#include "InstrumentManager/InstrumentManager.h"
#include "InstrumentManager/HardwareManager.h"
#include "InstrumentAbstraction/ExperimentController.h"
#include "Module/ModuleManager.h"
#include "DataProvider/ChangeAspectFactory.h"
#include "ErrorsAlarmsManager/ErrorCatalogManager.h"
#include <Utilities/Assert.h>
#include <common/base/ServerProperties.h>
#include <dataprovider/iconandfamily/ServantIconAndFamilyMaps.h>
#include <fstream>
#include "DeviceContainer/HardwareCommand.h"
#include "DeviceContainer/Observer.h"
#include "DeviceContainer/Spy.h"
#include <core/commandzone/AtomicCommandBox.h>
#include <core/commandzone/CommandBoxFactory.h>
#include <core/commandzone/CommandZone.h>

using namespace std;
using namespace common;

namespace ics {

ServerManager* ServerManager::_instance = 0;

const std::string ServerManager::TYPE = "ServerManager";
const std::string ServerManager::NAME = "ServerManager";
const std::string ServerManager::STATE_PROPERTY = "state";
const std::string ServerManager::ESTIMATED_TIME_PROPERTY = "estimated_time";
const std::string ServerManager::ESTIMATED_TIME_STATUS_PROPERTY = "estimated_time_status";
const std::string ServerManager::SCAN_FILENAME_PROPERTY = "scan_filename";

const std::string ServerManager::MODULES_PATH = "modulesPath";

void ServerManager::initInstance() {

	if (_instance == 0) {
		_instance = new ServerManager();
	}
}

ServerManager* ServerManager::getInstance() {
	return _instance;
}

void ServerManager::resetInstance() {

	if (_instance != 0) {
		_instance->reset();
		delete _instance;
	}
}

void ServerManager::init() {
	ErrorCatalogManager::initInstance();
	PropertyInfos::initInstance();
	ModuleManager::initInstance();

	try {
		ModuleManager::getInstance()->loadModulesFromSequence(ServerProperties::getInstance()->getNomadHomePath() + "/mod");
	} catch (const Exception & e) {
		cout << "Unable to load modules : " << e.getMessage() << endl;
	}

	// load here the families and icons because the module must be loaded
	// read plugin files (both controller and family folders) to retrieve image keys necessary for log messages
	dataprovider::ServantIconAndFamilyMaps::getInstance()->readClientPluginFiles();

	ExperimentControllerCreator::getInstance()->initFromModules();
	DeviceDriverCreator::getInstance()->initFromModules();

	ThreadPool::init();

	Scheduler<HardwareCommand>::getInstance()->start();
	Scheduler<ClientCommand>::getInstance()->start();

	Observer::getInstance()->init();
	Spy::getInstance()->init();

	Observer::getInstance()->execute();

	if (Simulated::visa == true) {
		ConfigurationManager::getInstance()->loadConfiguration();
		ConfigurationManager::getInstance()->backupConfiguration();

		ConfigurationManager::getInstance()->start();
	}
	else if ((!Simulated::activated) && (!Simulated::for3d)) {

		ConfigurationManager::getInstance()->loadConfiguration();
		ConfigurationManager::getInstance()->backupConfiguration();

		HistogramManager::getInstance()->init();

		ConfigurationManager::getInstance()->start();

	} else {
		SimulatedConfigurationManager::getInstance()->init();
		SimulatedConfigurationManager::getInstance()->loadConfiguration();
	}

	Spy::getInstance()->execute();
}

void ServerManager::reset() {

	Spy::getInstance()->stop();

	Scheduler<HardwareCommand>::getInstance()->stop();
	Scheduler<ClientCommand>::getInstance()->stop();

	ThreadPool::reset();

	// waiting for scheduler
	while (!Scheduler<HardwareCommand>::getInstance()->isReady() || !Scheduler<ClientCommand>::getInstance()->isReady()) {
		int32 commands = Scheduler<HardwareCommand>::getInstance()->getNumberOfCommands() + Scheduler<ClientCommand>::getInstance()->getNumberOfCommands();
		if (commands == 1) {
			cout << commands << " command to wait for..." << endl;
		} else {
			cout << commands << " commands to wait for..." << endl;
		}

		cout << "Hardware commands:" << endl;
		cout << Scheduler<HardwareCommand>::getInstance()->getCurrentCommandsInfo().toString() << endl;
		cout << "Client commands:" << endl;
		cout << Scheduler<ClientCommand>::getInstance()->getCurrentCommandsInfo().toString() << endl;

		sleep(1);
	}

	//TODO It is necessary to check that all commands are terminated to ensure that reset() will execute safely.
	Scheduler<HardwareCommand>::getInstance()->reset();
	Scheduler<ClientCommand>::getInstance()->reset();

	if ((!Simulated::activated && !Simulated::for3d) || (Simulated::visa == true)) {
		ConfigurationManager::resetInstance();

	} else {
		SimulatedConfigurationManager::resetInstance();
	}

	Spy::resetInstance();
	Observer::resetInstance();

	ModuleManager::resetInstance();
}

void ServerManager::forceReset() {

	// only reset configuration manager
	if ((!Simulated::activated && !Simulated::for3d) || (Simulated::visa == true)) {
		ConfigurationManager::resetInstance(false);
	} else {
		SimulatedConfigurationManager::resetInstance();
	}
}

void ServerManager::onCommandZoneStarted() {
	if (Simulated::activated == false || (Simulated::visa == true)) {
		// notify subscribers
		ChangeAspectFactory factory;

		// the server manager is not a controller but we can use the publisher/subscriber mechanism
		ChangeAspect* changeAspect = factory.createPropertyChangeAspect(TYPE, NAME, STATE_PROPERTY);
		changeAspect->setLongParam((int32) 1);
		notify(changeAspect);
	}
}

void ServerManager::onCommandZoneStopped() {
	if (Simulated::activated == false || (Simulated::visa == true)) {
		// notify subscribers
		ChangeAspectFactory factory;

		// the server manager is not a controller but we can use the publisher/subscriber mechanism
		ChangeAspect* changeAspect = factory.createPropertyChangeAspect(TYPE, NAME, STATE_PROPERTY);
		changeAspect->setLongParam((int32) 0);
		notify(changeAspect);
	}
}

void ServerManager::onEstimatedTimeChanged(double estimatedTime) {
	if (Simulated::activated == false || (Simulated::visa == true)) {
		// notify subscribers
		ChangeAspectFactory factory;

		// the server manager is not a controller but we can use the publisher/subscriber mechanism
		ChangeAspect* changeAspect = factory.createPropertyChangeAspect(TYPE, NAME, ESTIMATED_TIME_PROPERTY);
		changeAspect->setLongParam((int32) estimatedTime);
		notify(changeAspect);
	}
}

void ServerManager::onEstimatedTimeComputingStatus(long status) {
	if (Simulated::activated == false || (Simulated::visa == true)) {
		// notify subscribers
		ChangeAspectFactory factory;

		// the server manager is not a controller but we can use the publisher/subscriber mechanism
		ChangeAspect* changeAspect = factory.createPropertyChangeAspect(TYPE, NAME, ESTIMATED_TIME_STATUS_PROPERTY);
		changeAspect->setLongParam(status);
		notify(changeAspect);
	}
}

void ServerManager::onScanCommandBoxLaunch(const std::string filename) {
	if (Simulated::activated == false || (Simulated::visa == true)) {
		_ScanCommandBoxFileName = filename;

		// notify subscribers
		ChangeAspectFactory factory;

		// the server manager is not a controller but we can use the publisher/subscriber mechanism
		ChangeAspect* changeAspect = factory.createPropertyChangeAspect(TYPE, NAME, SCAN_FILENAME_PROPERTY);
		notify(changeAspect);
	}
}

void ServerManager::unloadUnusedModules() {
	ModuleManager::getInstance()->unloadUnusedModules();

	vector<string> modules = ModuleManager::getInstance()->getLoadedModules();

	cout << "Modules:" << endl;
	cout << "<modules>" << endl;
	for (vector<string>::const_iterator m = modules.begin(); m != modules.end(); ++m) {
		cout << "  <include name=\"" << *m << "\"/>" << endl;
	}
	cout << "</modules>" << endl;
}

bool ServerManager::test() {

	// Iterating the controller map
	const map<string, ExperimentController*>& controllerMap = InstrumentManager::getInstance()->getAllExperimentControllers();

	for (map<string, ExperimentController*>::const_iterator c = controllerMap.begin(); c != controllerMap.end(); c++) {

		ExperimentController* controller = c->second;
		if (controller->isEnabled() == true) {
			controller->test();
		}
	}

	return !Assert::onError();
}

bool ServerManager::testEstimatedTime() {

//	// Iterating the controller map
//	const map<string, ExperimentController*>& controllerMap = InstrumentManager::getInstance()->getAllExperimentControllers();
//
//	for (map<string, ExperimentController*>::const_iterator c = controllerMap.begin(); c != controllerMap.end(); c++) {
//
//		ExperimentController* controller = c->second;
//		out << "try add " << controller->getName() << " " << controller->isCommand() << eol;
//		if (controller->isCommand() == true) {
//			try {
//				out << "try" << eol;
// 				bool issettings = (dynamic_cast<SettingsController*>(controller) != nullptr);
//				out << "add " << controller->getName() << " (" << controller->getType() <<")" << " " << issettings << eol;
//				core::CommandBox* commandBox = core::CommandBoxFactory::getInstance()->createAtomicCommandBox(controller->getName(), issettings);
//				out << "add to end" << eol;
//				core::CommandZone::getInstance()->addCommandBoxAtEnd(commandBox);
//			} catch (core::CommandBoxFactory::NoSuchCommandException& e) {
//				err << "Add command box for controller " << controller->getName() << " (" << controller->getType() <<")" << eol;
//			}
//		}
//	}
//	out <<" OK finished" << eol;
//	core::CommandZone::getInstance()->computeEstimatedTime("", 0);
//	core::CommandZone::getInstance()->resetCommandZone(0);
	return !Assert::onError();
}

}
