/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "InstrumentAbstraction/AbstractController.h"
#include "LoggingServices/Level.h"
#include "DeviceContainer/DeviceDriver.h"
#include "DataProvider/Property/PropertyManager.h"
#include "DataProvider/CommandManager.h"
#include "DataProvider/ChangeAspect.h"
#include "DataProvider/ChangeAspectFactory.h"
#include "DataProvider/Property/PropertyInfos.h"
#include "DataProvider/Property/TypedProperty/Property.h"
#include "PersistenceServices/Configuration/PropertyConfig.h"

#include <boost/algorithm/string.hpp>
#include <boost/format.hpp>

using namespace std;
using namespace boost;

namespace family {

const std::string NONE = "none";
const std::string HIDDEN = "hidden";

}

std::recursive_mutex AbstractController::m_longArrayMutex;
std::recursive_mutex AbstractController::m_float64ArrayMutex;

const std::string AbstractController::FAVORITES = "Favorites";

AbstractController::UnknownPropertyType::UnknownPropertyType(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info) :
		m_type(type), m_name(name), m_publisher(publisher), m_info(info) {
}

const std::string & AbstractController::UnknownPropertyType::getType() const {
	return m_type;
}

const std::string & AbstractController::UnknownPropertyType::getName() const {
	return m_name;
}

Publisher* AbstractController::UnknownPropertyType::getPublisher() const {
	return m_publisher;
}

const std::string & AbstractController::UnknownPropertyType::getInfo() const {
	return m_info;
}

AbstractController::UnknownProperty::UnknownProperty(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info) :
		NotOk(type, name, publisher, UNKNOWN_PROPERTY_ERROR, info) {
}

AbstractController::UnknownProperty::UnknownProperty(UnknownPropertyType const & exception) :
		NotOk(exception.getType(), exception.getName(), exception.getPublisher(), UNKNOWN_PROPERTY_ERROR, exception.getInfo()) {
}

AbstractController::TimeOutOccured::TimeOutOccured() :
		NotOk(0) {
}

AbstractController::ControllerNotConfigured::ControllerNotConfigured() :
		NotOk(0) {
}

AbstractController::ControllerNotConfigured::ControllerNotConfigured(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info) :
		NotOk(type, name, publisher, NOT_CONFIGURED_ERROR) {
}

AbstractController::WrongPropertyType::WrongPropertyType(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info) :
		NotOk(type, name, publisher, BAD_TYPE_PROPERTY_ERROR, info) {
}

AbstractController::UnknownCommand::UnknownCommand(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info) :
		NotOk(type, name, publisher, BAD_TYPE_PROPERTY_ERROR, info) {
}

AbstractController::ExecutionError::ExecutionError() :
		NotOk(0) {
}

AbstractController::ExecutionError::ExecutionError(const std::string& message) :
		NotOk(0, message) {
}

/*
 * Constructor
 */
AbstractController::AbstractController(const string& type, const string& name, const string& controllerType) :
		m_PropertyManager(NULL), m_CommandManager(NULL), m_ServantManager(NULL) {
	constructor(type, name, controllerType);
	m_clone = false;
	
}

/*
 * Constructor
 */
AbstractController::AbstractController() :
		m_PropertyManager(NULL), m_CommandManager(NULL), m_ServantManager(NULL) {
	m_clone = false;
}

void AbstractController::constructor(const std::string& type, const std::string& name, const string& controllerType) {
	m_type = type;
	m_name = name;
	m_controllerType = controllerType;
	m_Enabled = false;
	if (m_PropertyManager == NULL) {
		m_PropertyManager = new PropertyManager(name);
	}
	if (m_CommandManager == NULL) {
		m_CommandManager = new CommandManager(name);
	}
	if (m_ServantManager == NULL) {
		m_ServantManager = new ServantManager(name);
	}
	m_Histogramming = true;
}

AbstractController::AbstractController(const AbstractController& controller) {

	m_clone = true;
	m_type = controller.m_type;
	m_name = controller.m_name + ".clone";
	m_controllerType = controller.m_controllerType;
	m_Enabled = controller.m_Enabled;

	m_PropertyManager = controller.m_PropertyManager->clone(m_name);
	m_CommandManager = controller.m_CommandManager->clone(m_name);
	m_ServantManager = controller.m_ServantManager->clone(m_name);

	m_Functions = controller.m_Functions;

	m_defaultFamily = controller.m_defaultFamily;
	m_defaultSubfamily = controller.m_defaultSubfamily;

	m_Histogramming = controller.m_Histogramming;
}


void AbstractController::copy(const AbstractController& controller) {
	m_clone = true;
	m_type = controller.m_type;
	m_name = controller.m_name + ".clone";
	m_controllerType = controller.m_controllerType;
	m_Enabled = controller.m_Enabled;

	m_PropertyManager = controller.m_PropertyManager->clone(m_name);
	m_CommandManager = controller.m_CommandManager->clone(m_name);
	m_ServantManager = controller.m_ServantManager->clone(m_name);

	m_Functions = controller.m_Functions;

	m_defaultFamily = controller.m_defaultFamily;
	m_defaultSubfamily = controller.m_defaultSubfamily;

	m_Histogramming = controller.m_Histogramming;
}

/*
 * Destructor
 */
AbstractController::~AbstractController() {

	

	/*
	 * Delete Property and Command Controllers and set to NULL
	 */
	delete m_PropertyManager;
	delete m_CommandManager;
	delete m_ServantManager;

}

ServantClass AbstractController::getServantClass() const {
	return CONTROLLER_CLASS;
}

std::vector<std::string> AbstractController::getCommandList() const {
	return m_CommandManager->getCommandList();
}

/*
 * registerFunction
 */
void AbstractController::registerFunction(const string& function)  {
	m_Functions.push_back(function);
}

/*
 * initCommand
 */
void AbstractController::initCommand(const string& commandName) const {
	m_CommandManager->addCommand(commandName);
}

/*
 * removeCommand
 */
void AbstractController::removeCommand(const string& commandName) {
	m_CommandManager->removeCommand(commandName);
}

/*
 * containsFunction
 */
bool AbstractController::containsFunction(const string& function) {

	vector<string>::iterator i = m_Functions.begin();
	bool found = false;
	while (i != m_Functions.end() && !found) {
		if (*i == function) {
			found = true;
		} else {
			i++;
		}
	}
	return found;
}

bool AbstractController::containsCommand(const std::string& commandName) const {
	return m_CommandManager->commandExists(commandName);
}

/*
 * hasChanged
 */
void AbstractController::hasChanged(const string& propertyName, boost::any oldValue, boost::any newValue) {

	string propertyType = getPropertyType(propertyName);
	bool hasChanged = false;
	try {
		if (propertyType == UINT16_PROPERTY) {
			hasChanged = (any_cast<uint16>(newValue) != any_cast<uint16>(oldValue));
		} else if (propertyType == UINT32_PROPERTY) {
			hasChanged = (any_cast<uint32>(newValue) != any_cast<uint32>(oldValue));
		} else if (propertyType == UINT64_PROPERTY) {
			hasChanged = (any_cast<uint64>(newValue) != any_cast<uint64>(oldValue));
		} else if (propertyType == INT16_PROPERTY) {
			hasChanged = (any_cast<int16>(newValue) != any_cast<int16>(oldValue));
		} else if (propertyType == INT32_PROPERTY) {
			hasChanged = (any_cast<int32>(newValue) != any_cast<int32>(oldValue));
		} else if (propertyType == INT64_PROPERTY) {
			hasChanged = (any_cast<int64>(newValue) != any_cast<int64>(oldValue));
		} else if (propertyType == FLOAT32_PROPERTY) {
			hasChanged = (any_cast<float32>(newValue) != any_cast<float32>(oldValue));
		} else if (propertyType == FLOAT64_PROPERTY) {
			hasChanged = (any_cast<float64>(newValue) != any_cast<float64>(oldValue));
		} else if (propertyType == BOOLEAN_PROPERTY) {
			hasChanged = (any_cast<bool>(newValue) != any_cast<bool>(oldValue));
		} else if (propertyType == STRING_PROPERTY) {
			hasChanged = (any_cast<string>(newValue) != any_cast<string>(oldValue));
		}
	} catch (...) {
		cout << "Error  has changed " << endl;
		hasChanged = true;
	}
	if (hasChanged == true) {
		ChangeAspectFactory aCAFactory;

		ChangeAspect* aCA = aCAFactory.createPropertyChangeAspect(m_type, m_name, propertyName);
		notify(aCA);
	}
}

void AbstractController::refreshValue(const std::string& propertyName, const boost::any& value)  {

	PropertyRefresher::refresh(propertyName, value);
	refreshSetValue(propertyName, value);
}

/*
 * setValue
 */
void AbstractController::setValue(const std::string& propertyName, boost::any aValue)  {

	try {
		refreshValue(propertyName, aValue);
		boost::any oldValue = getValue(propertyName);
		m_PropertyManager->setValue(propertyName, aValue);
		formatValue(propertyName);
		hasChanged(propertyName, oldValue, aValue);
	} catch (const PropertyManager::UnknownProperty& e) {
		throw UnknownProperty(m_type, m_name, this, std::string("set ") + propertyName + ": " + e.what());
	} catch (PropertyManager::CannotSetValue& e) {
		throw CannotSetValue(e.what());
	}
}

/*
 * setValue
 */
void AbstractController::setValueAndFormattedValue(const std::string& propertyName, boost::any aValue, const std::string& formattedValue)  {

	try {
		refreshValue(propertyName, aValue);
		boost::any oldValue = getValue(propertyName);
		m_PropertyManager->setValueAndFormattedValue(propertyName, aValue, formattedValue);
		hasChanged(propertyName, oldValue, aValue);
	} catch (const PropertyManager::UnknownProperty& e) {
		throw UnknownProperty(m_type, m_name, this, std::string("set ") + propertyName + ": " + e.what());
	} catch (const PropertyManager::CannotSetValue& e) {
		throw CannotSetValue(e.what());
	}
}

void AbstractController::setTypedValue(const std::string& propertyName, const boost::any& value)  {

	AnyProperty* anyProperty = m_PropertyManager->getProperty(propertyName);

	// Checking if the property is dynamic or not.
	PropertyManager::DynamicPropertyInfo info = PropertyManager::getDynamicPropertyInfo(propertyName);

	string ptype = anyProperty->getType();

	if (!info.isDynamic) {

		// The property is not dynamic.
		if (anyProperty->getType() == UINT16_PROPERTY) {
			SimpleProperty<uint16> * property = dynamic_cast<SimpleProperty<uint16> *>(PropertyContainer::m_properties[propertyName]);
			if (property != 0) {
				property->set(boost::any_cast<uint16>(value));
			} else {
				cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<uint16>" << endl;
			}

		} else if (anyProperty->getType() == UINT32_PROPERTY) {
				SimpleProperty<uint32> * property = dynamic_cast<SimpleProperty<uint32> *>(PropertyContainer::m_properties[propertyName]);
				if (property != 0) {
					property->set(boost::any_cast<uint32>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<uint32>" << endl;
				}

		} else if (anyProperty->getType() == UINT64_PROPERTY) {
				SimpleProperty<uint64> * property = dynamic_cast<SimpleProperty<uint64> *>(PropertyContainer::m_properties[propertyName]);
				if (property != 0) {
					property->set(boost::any_cast<uint64>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<uint64>" << endl;
				}
		} else if (anyProperty->getType() == INT16_PROPERTY) {
			SimpleProperty<int16> * property = dynamic_cast<SimpleProperty<int16> *>(PropertyContainer::m_properties[propertyName]);
			if (property != 0) {
				property->set(boost::any_cast<int16>(value));
			} else {
				cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<int16>" << endl;
			}

		} else if (anyProperty->getType() == INT32_PROPERTY) {
				SimpleProperty<int32> * property = dynamic_cast<SimpleProperty<int32> *>(PropertyContainer::m_properties[propertyName]);
				if (property != 0) {
					property->set(boost::any_cast<int32>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<int32>" << endl;
				}

		} else if (anyProperty->getType() == INT64_PROPERTY) {
				SimpleProperty<int64> * property = dynamic_cast<SimpleProperty<int64> *>(PropertyContainer::m_properties[propertyName]);
				if (property != 0) {
					property->set(boost::any_cast<int64>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<int64>" << endl;
				}
		} else if (anyProperty->getType() == FLOAT32_PROPERTY) {
			SimpleProperty<float32> * property = dynamic_cast<SimpleProperty<float32> *>(PropertyContainer::m_properties[propertyName]);
			if (property != 0) {
				property->set(boost::any_cast<float32>(value));
			} else {
				cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<float32>" << endl;
			}

		} else if (anyProperty->getType() == FLOAT64_PROPERTY) {
			SimpleProperty<float64> * property = dynamic_cast<SimpleProperty<float64> *>(PropertyContainer::m_properties[propertyName]);
			if (property != 0) {
				property->set(boost::any_cast<float64>(value));
			} else {
				cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<float64>" << endl;
			}

		} else if (anyProperty->getType() == STRING_PROPERTY) {
			SimpleProperty<string> * property = dynamic_cast<SimpleProperty<string> *>(PropertyContainer::m_properties[propertyName]);
			if (property != 0) {
				property->set(boost::any_cast<string>(value));
			} else {
				cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<string>" << endl;
			}

		} else if (anyProperty->getType() == BOOLEAN_PROPERTY) {
			SimpleProperty<bool> * property = dynamic_cast<SimpleProperty<bool> *>(PropertyContainer::m_properties[propertyName]);
			if (property != 0) {
				property->set(boost::any_cast<bool>(value));
			} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type SimpleProperty<bool>" << endl;
				}
		} else {
			cerr << "Cannot call setTypedValue on an array type" << endl;
		}
	} else {
		// test double dynamic property
		PropertyManager::DynamicPropertyInfo info2 = PropertyManager::getDynamicPropertyInfo(info.name);

		if (!info2.isDynamic) {

			// The property is dynamic.
			if (anyProperty->getType() == UINT16_PROPERTY) {
				DynamicProperty<uint16> * property = dynamic_cast<DynamicProperty<uint16> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<uint16>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<uint16>" << endl;
				}

			} else if (anyProperty->getType() == UINT32_PROPERTY) {
				DynamicProperty<uint32> * property = dynamic_cast<DynamicProperty<uint32> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<uint32>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<uint32>" << endl;
				}

			} else if (anyProperty->getType() == UINT64_PROPERTY) {
				DynamicProperty<uint64> * property = dynamic_cast<DynamicProperty<uint64> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<uint64>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<uint64>" << endl;
				}

			} else if (anyProperty->getType() == INT16_PROPERTY) {
				DynamicProperty<int16> * property = dynamic_cast<DynamicProperty<int16> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<int16>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<int16>" << endl;
				}

			} else if (anyProperty->getType() == INT32_PROPERTY) {
				DynamicProperty<int32> * property = dynamic_cast<DynamicProperty<int32> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<int32>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<int32>" << endl;
				}

			} else if (anyProperty->getType() == INT64_PROPERTY) {
				DynamicProperty<int64> * property = dynamic_cast<DynamicProperty<int64> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<int64>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<int64>" << endl;
				}

			} else if (anyProperty->getType() == FLOAT64_PROPERTY) {
				DynamicProperty<float64> * property = dynamic_cast<DynamicProperty<float64> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<float64>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<float64>" << endl;
				}

			} else if (anyProperty->getType() == FLOAT32_PROPERTY) {
				DynamicProperty<float32> * property = dynamic_cast<DynamicProperty<float32> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<float32>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<float32>" << endl;
				}

			} else if (anyProperty->getType() == STRING_PROPERTY) {
				DynamicProperty<string> * property = dynamic_cast<DynamicProperty<string> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<string>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<string>" << endl;
				}

			} else if (anyProperty->getType() == BOOLEAN_PROPERTY) {
				DynamicProperty<bool> * property = dynamic_cast<DynamicProperty<bool> *>(PropertyContainer::m_dynamicProperties[info.name]);
				if (property != 0) {
					property->set(info.index, boost::any_cast<bool>(value));
				} else {
						cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty<bool>" << endl;
					}
			} else {
				cerr << "Cannot call setTypedValue on an array type" << endl;
			}

		} else {

			// The property is double dynamic.
			if (anyProperty->getType() == UINT16_PROPERTY) {
				DynamicProperty2<uint16> * property = dynamic_cast<DynamicProperty2<uint16> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<uint16>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<uint16>" << endl;
				}

			} else if (anyProperty->getType() == UINT32_PROPERTY) {
				DynamicProperty2<uint32> * property = dynamic_cast<DynamicProperty2<uint32> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<uint32>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<uint32>" << endl;
				}

			} else if (anyProperty->getType() == UINT64_PROPERTY) {
				DynamicProperty2<uint64> * property = dynamic_cast<DynamicProperty2<uint64> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<uint64>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<uint64>" << endl;
				}

			} else if (anyProperty->getType() == INT16_PROPERTY) {
				DynamicProperty2<int16> * property = dynamic_cast<DynamicProperty2<int16> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<int16>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<int16>" << endl;
				}

			} else if (anyProperty->getType() == INT32_PROPERTY) {
				DynamicProperty2<int32> * property = dynamic_cast<DynamicProperty2<int32> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<int32>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<int32>" << endl;
				}

			} else if (anyProperty->getType() == INT64_PROPERTY) {
				DynamicProperty2<int64> * property = dynamic_cast<DynamicProperty2<int64> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<int64>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<int64>" << endl;
				}

			} else if (anyProperty->getType() == FLOAT32_PROPERTY) {
				DynamicProperty2<float32> * property = dynamic_cast<DynamicProperty2<float32> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<float32>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<float32>" << endl;
				}

			} else if (anyProperty->getType() == FLOAT64_PROPERTY) {
				DynamicProperty2<float64> * property = dynamic_cast<DynamicProperty2<float64> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<float64>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<float64>" << endl;
				}

			} else if (anyProperty->getType() == STRING_PROPERTY) {
				DynamicProperty2<string> * property = dynamic_cast<DynamicProperty2<string> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<string>(value));
				} else {
					cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<string>" << endl;
				}

			} else if (anyProperty->getType() == BOOLEAN_PROPERTY) {
				DynamicProperty2<bool> * property = dynamic_cast<DynamicProperty2<bool> *>(PropertyContainer::m_dynamicProperties2[info2.name]);
				if (property != 0) {
					property->set(info2.index, info.index, boost::any_cast<bool>(value));
				} else {
						cerr << "Cannot call setTypedValue on " << propertyName << " of " << getName() << " that is not of type DynamicProperty2<bool>" << endl;
					}
			} else {
				cerr << "Cannot call setTypedValue on an array type" << endl;
			}
		}
	}
}

void AbstractController::setValueSynchronized(const std::string& propertyName, boost::any aValue)  {

	std::scoped_lock pubLock(Publisher::m_publisherMutex);

	setTypedValue(propertyName, aValue);
}

/*
 * getValue
 */
boost::any AbstractController::getValue(const string& propertyName)  {

	try {
		return m_PropertyManager->getValue(propertyName);
	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	}
}

/*
 * getPropertyType
 */
const string& AbstractController::getPropertyType(const string& propertyName)  {

	try {
		return m_PropertyManager->getType(propertyName);
	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownPropertyType(m_type, m_name, this, propertyName);

	}
}

/*
 * getPropertyAlias
 */
const string& AbstractController::getPropertyAlias(const string& propertyName)  {

	try {
		return m_PropertyManager->getAlias(propertyName);
	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	}
}

std::string AbstractController::getPropertyUnit(const std::string& propertyName) {

	auto it = PropertyContainer::m_properties.find(propertyName);

	if (it != PropertyContainer::m_properties.end()) {
		return it->second->getUnit();
	}

	return "";
}

/*
 * getSize
 */
int32 AbstractController::getSize(const string& propertyName)  {

	try {
		return m_PropertyManager->getSize(propertyName);
	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	}
}

/*
 * getExtraSize
 */
int32 AbstractController::getExtraSize(const string& propertyName)  {

	try {
		return m_PropertyManager->getExtraSize(propertyName);
	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	}
}

/*
 * setSize
 */
void AbstractController::setSize(const string& propertyName, int32 aSize)  {

	try {
		m_PropertyManager->setSize(propertyName, aSize);
	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	}
}

/*
 * setExtraSize
 */
void AbstractController::setExtraSize(const string& propertyName, int32 size)  {

	try {
		m_PropertyManager->setExtraSize(propertyName, size);
	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	}
}

/*
 * updateValue
 */
void AbstractController::updateValue(const std::string& propertyName, boost::any propertyValue)  {

	try {
		boost::any oldValue = getValue(propertyName);
		m_PropertyManager->setValue(propertyName, propertyValue);
		formatValue(propertyName);
		hasChanged(propertyName, oldValue, propertyValue);
	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	} catch (PropertyManager::CannotSetValue& e) {
		throw CannotSetValue(e.what());
	}
}

/*
 * updateValue
 */
void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, boost::any propertyValue, const std::string& formattedValue)  {

	try {
		boost::any oldValue = getValue(propertyName);
		m_PropertyManager->setValueAndFormattedValue(propertyName, propertyValue, formattedValue);
		hasChanged(propertyName, oldValue, propertyValue);
	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	} catch (PropertyManager::CannotSetValue& e) {
		throw CannotSetValue(e.what());
	}
}

/*
 * updateValue
 */
void AbstractController::updateValue(const std::string& propertyName, boost::any propertyValue, int32 size)  {

	try {
		m_PropertyManager->setValue(propertyName, propertyValue);
		m_PropertyManager->setSize(propertyName, size);
		formatValue(propertyName);

	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	} catch (PropertyManager::CannotSetValue& e) {
		throw CannotSetValue(e.what());
	}
}

void AbstractController::refreshSetValue(std::string propertyName, const boost::any aValue)  {
	// do nothing
}

/*
 * sendSynchroEvent
 */
void AbstractController::sendSynchroEvent(const string& commandName, int32 state) {

	ChangeAspectFactory aCAFactory;
	ChangeAspect* aCA = aCAFactory.createSynchroChangeAspect(m_type, m_name, commandName, state);
	notify(aCA);

}

/*
 * sendRefuseEvent
 */
void AbstractController::sendRefuseEvent(const string& commandName, const string& cName, const string& cCommandName) {

	stringstream message;
	message << cName << " command " << cCommandName << " refuses " << getName() << "::" << commandName;

	ChangeAspectFactory aCAFactory;

	ChangeAspect* errorChangeAspect = aCAFactory.createErrorChangeAspect(m_type, m_name, COMMAND_REFUSED_ERROR);

	errorChangeAspect->setStringParam(message.str());
	log(Level::s_Error) << message.str() << endlog;
	notify(errorChangeAspect);

	ChangeAspect* aCA = aCAFactory.createRefuseChangeAspect(m_type, m_name, commandName);
	notify(aCA);

}

/*
 * sendRefuseEvent
 */
void AbstractController::sendRefuseEvent(const string& commandName) {

	stringstream message;
	message << commandName << " command refused, delay it";

	ChangeAspectFactory aCAFactory;

	ChangeAspect* errorChangeAspect = aCAFactory.createErrorChangeAspect(m_type, m_name, COMMAND_REFUSED_ERROR);

	errorChangeAspect->setStringParam(message.str());
	log(Level::s_Error) << message.str() << endlog;
	notify(errorChangeAspect);

	ChangeAspect* aCA = aCAFactory.createRefuseChangeAspect(m_type, m_name, commandName);
	notify(aCA);

}

/*
 * initProperties
 */
void AbstractController::initProperties() {
	vector<AnyProperty *> properties;
	if (m_clone) {
		properties = m_PropertyManager->getAllProperties();
	}
	else {
		properties = m_PropertyManager->getPropertiesToSave();
	}
	vector<AnyProperty *>::iterator iter;
	for (iter = properties.begin(); iter != properties.end(); ++iter) {
		string name = (*iter)->getName();
		try {
			refreshValue(name, (*iter)->getValue());
			int32 size = (*iter)->getSize();
			updateValue(name, (*iter)->getValue());
			setSize(name, size);
		} catch (CannotSetValue&) {
			ostringstream sbuf;
			sbuf << "Cannot set property " << name;
			m_Warnings.push_back(sbuf.str());
		}
	}
}

/*
 * initProperties
 */
void AbstractController::initProperties(const vector<PropertyConfig>& aProperties, int32 pass) {
	/*
	 * Init properties in pass = 1 , just update value to avoid affected value to
	 * when controller is disable. Some controller are linked to many controllers.
	 * For example , sample temperature can be use by different controller like orange
	 * or cryoloop. When orange is select, sample temperature is affect by orange and loop
	 * can't affect this controller.
	 *
	 */

	/*
	 * Init the controller properties (static only)
	 */
	vector<PropertyConfig>::const_iterator propertiesIterator;
	for (propertiesIterator = aProperties.begin(); propertiesIterator != aProperties.end(); propertiesIterator++) {

		// Get the property from property manager
		string name = propertiesIterator->getName();

		bool found = true;
		string type;

		try {
			type = getPropertyType(name);
		} catch (UnknownPropertyType & e) {
			if (pass == 3) {
				found = false;
			} else {
				if (Simulated::activated == false) {
					ostringstream sbuf;
					sbuf << "Property " << name << " does not exist";
					m_Errors.push_back(sbuf.str());
				}
				goto fin;
				//            	throw UnknownProperty(e);
			}
		}

		//    cout << "name = " << name << " , type = " << type << endl;
		if (found) {
			try {
				// Cast the string to the write value type
				if (type == UINT16_PROPERTY) {
					size_t loc = propertiesIterator->getValue().find("0x");
					if (loc != string::npos) {
						string val = propertiesIterator->getValue().substr(loc + 2);
						if (val.size() == 4) {
							// Do this in two time due to the 0x80000000 value, because strtol return int,
							// so 0x80000000 it is the sign, strtol return 0x7FFFFFFF
							uint16 high;
							high = (uint16) strtoul(val.c_str(), NULL, 16);
							try {
								//if (pass != 1) {
									refreshValue(name, high);
								//}
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
							updateValue(name, high);
						} else {
							//cout << name << " " << strtol(propertiesIterator->getValue().c_str(),NULL,16) << endl;

							try {
								//if (pass != 1) {
									refreshValue(name, (uint16) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
								//}
								updateValue(name, (uint16) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
						}
					} else {
						// 	    cout << strtol(propertiesIterator->getValue().c_str(),NULL,10) << endl;
						try {
							//if (pass != 1) {
								refreshValue(name, (uint16) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
							//}
						} catch (CannotSetValue&) {
							if (pass != 3) {
								throw;
							}
						}
						updateValue(name, (uint16) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
					}
				}else if (type == UINT32_PROPERTY) {
					size_t loc = propertiesIterator->getValue().find("0x");
					if (loc != string::npos) {
						string val = propertiesIterator->getValue().substr(loc + 2);
						if (val.size() == 8) {
							// Do this in two time due to the 0x80000000 value, because strtol return int,
							// so 0x80000000 it is the sign, strtol return 0x7FFFFFFF
							uint32 high;
							high = (uint32) strtoul(val.c_str(), NULL, 16);
							try {
								//if (pass != 1) {
									refreshValue(name, high);
								//}
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
							updateValue(name,(uint32)  high);
						} else {
							//cout << name << " " << strtol(propertiesIterator->getValue().c_str(),NULL,16) << endl;

							try {
								//if (pass != 1) {
									refreshValue(name, (uint32) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
								//}
								updateValue(name, (uint32) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
						}
					} else {
						// 	    cout << strtol(propertiesIterator->getValue().c_str(),NULL,10) << endl;
						try {
							//if (pass != 1) {
								refreshValue(name, (uint32) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
							//}
						} catch (CannotSetValue&) {
							if (pass != 3) {
								throw;
							}
						}
						updateValue(name, (uint32) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
					}
				} else if (type == UINT64_PROPERTY) {
					size_t loc = propertiesIterator->getValue().find("0x");
					if (loc != string::npos) {
						string val = propertiesIterator->getValue().substr(loc + 2);
						if (val.size() == 16) {
							// Do this in two time due to the 0x80000000 value, because strtol return int,
							// so 0x80000000 it is the sign, strtol return 0x7FFFFFFF
							uint64 high;
							high = (int64) strtol(val.c_str(), NULL, 16);
							try {
								//if (pass != 1) {
									refreshValue(name, high);
								//}
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
							updateValue(name, high);
						} else {
							//cout << name << " " << strtol(propertiesIterator->getValue().c_str(),NULL,16) << endl;

							try {
								//if (pass != 1) {
									refreshValue(name, (uint64) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
								//}
								updateValue(name, (uint64) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
						}
					} else {
						// 	    cout << strtol(propertiesIterator->getValue().c_str(),NULL,10) << endl;
						try {
							//if (pass != 1) {
								refreshValue(name, (uint64) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
							//}
						} catch (CannotSetValue&) {
							if (pass != 3) {
								throw;
							}
						}
						updateValue(name, (uint64) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
					}
				} else if (type == INT16_PROPERTY) {
					size_t loc = propertiesIterator->getValue().find("0x");
					if (loc != string::npos) {
						string val = propertiesIterator->getValue().substr(loc + 2);
						if (val.size() == 4) {
							// Do this in two time due to the 0x80000000 value, because strtol return int,
							// so 0x80000000 it is the sign, strtol return 0x7FFFFFFF
							int16 high;
							high = (int16) strtol(val.c_str(), NULL, 16);
							try {
								//if (pass != 1) {
									refreshValue(name, high);
								//}
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
							updateValue(name, high);
						} else {
							//cout << name << " " << strtol(propertiesIterator->getValue().c_str(),NULL,16) << endl;

							try {
								//if (pass != 1) {
									refreshValue(name, (int16) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
								//}
								updateValue(name, (int16) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
						}
					} else {
						// 	    cout << strtol(propertiesIterator->getValue().c_str(),NULL,10) << endl;
						try {
							//if (pass != 1) {
								refreshValue(name, (int16) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
							//}
						} catch (CannotSetValue&) {
							if (pass != 3) {
								throw;
							}
						}
						updateValue(name, (int16) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
					}
				}else if (type == INT32_PROPERTY) {
					size_t loc = propertiesIterator->getValue().find("0x");
					if (loc != string::npos) {
						string val = propertiesIterator->getValue().substr(loc + 2);
						if (val.size() == 8) {
							// Do this in two time due to the 0x80000000 value, because strtol return int,
							// so 0x80000000 it is the sign, strtol return 0x7FFFFFFF
							string val2 = val.substr(0, 4);
							string val3 = val.substr(4, 4);
							int32 high, low;
							high = (int32) strtol(val2.c_str(), NULL, 16);
							low = (int32) strtol(val3.c_str(), NULL, 16);
							try {
								//if (pass != 1) {
									refreshValue(name, (high << 16) + low);
								//}
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
							updateValue(name, (high << 16) + low);
						} else {
							//cout << name << " " << strtol(propertiesIterator->getValue().c_str(),NULL,16) << endl;

							try {
								//if (pass != 1) {
									refreshValue(name, (int32) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
								//}
								updateValue(name, (int32) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
						}
					} else {
						// 	    cout << strtol(propertiesIterator->getValue().c_str(),NULL,10) << endl;
						try {
							//if (pass != 1) {
								refreshValue(name, (int32) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
							//}
						} catch (CannotSetValue&) {
							if (pass != 3) {
								throw;
							}
						}
						updateValue(name, (int32) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
					}
				} else if (type == INT64_PROPERTY) {
					size_t loc = propertiesIterator->getValue().find("0x");
					if (loc != string::npos) {
						string val = propertiesIterator->getValue().substr(loc + 2);
						if (val.size() == 16) {
							// Do this in two time due to the 0x80000000 value, because strtol return int,
							// so 0x80000000 it is the sign, strtol return 0x7FFFFFFF
							string val2 = val.substr(0, 8);
							string val3 = val.substr(8, 16);
							int64 high, low;
							high = (int64) strtol(val2.c_str(), NULL, 16);
							low = (int64) strtol(val3.c_str(), NULL, 16);
							try {
								//if (pass != 1) {
									refreshValue(name, (high << 16) + low);
								//}
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
							updateValue(name, (high << 16) + low);
						} else {
							//cout << name << " " << strtol(propertiesIterator->getValue().c_str(),NULL,16) << endl;

							try {
								//if (pass != 1) {
									refreshValue(name, (int64) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
								//}
								updateValue(name, (int64) strtol(propertiesIterator->getValue().c_str(), NULL, 16));
							} catch (CannotSetValue&) {
								if (pass != 3) {
									throw;
								}
							}
						}
					} else {
						// 	    cout << strtol(propertiesIterator->getValue().c_str(),NULL,10) << endl;
						try {
							//if (pass != 1) {
								refreshValue(name, (int64) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
							//}
						} catch (CannotSetValue&) {
							if (pass != 3) {
								throw;
							}
						}
						updateValue(name, (int64) strtol(propertiesIterator->getValue().c_str(), NULL, 10));
					}
				} else if (type == FLOAT32_PROPERTY) {
					// 	  cout << strtod(propertiesIterator->getValue().c_str(),NULL) << endl;
					try {
						//if (pass != 1) {
							refreshValue(name, (float32) strtof(propertiesIterator->getValue().c_str(), NULL));
						//}
					} catch (CannotSetValue&) {
						if (pass != 3) {
							throw;
						}
					}
					updateValue(name, (float32) strtod(propertiesIterator->getValue().c_str(), NULL));
				} else if (type == FLOAT64_PROPERTY) {
					// 	  cout << strtod(propertiesIterator->getValue().c_str(),NULL) << endl;
					try {
						//if (pass != 1) {
							refreshValue(name, (float64) strtod(propertiesIterator->getValue().c_str(), NULL));
						//}
					} catch (CannotSetValue&) {
						if (pass != 3) {
							throw;
						}
					}
					updateValue(name, (float64) strtod(propertiesIterator->getValue().c_str(), NULL));
				} else if (type == STRING_PROPERTY) {
					// 	  cout << propertiesIterator->getValue() << endl;
					try {
						//if (pass != 1) {
							refreshValue(name, propertiesIterator->getValue());
						//}
					} catch (CannotSetValue&) {
						if (pass != 3) {
							throw;
						}
					}
					updateValue(name, propertiesIterator->getValue());
				} else if (type == BOOLEAN_PROPERTY) {
					if (propertiesIterator->getValue() == "true") {
						try {
							//if (pass != 1) {
								// 	      cout << true << endl;
								refreshValue(name, true);
							//}
						} catch (CannotSetValue&) {
							if (pass != 3) {
								throw;
							}
						}
						updateValue(name, true);
					} else {
						try {
							//if (pass != 1) {
								refreshValue(name, false);
							//}
						} catch (CannotSetValue&) {
							if (pass != 3) {
								throw;
							}
						}
						updateValue(name, false);
					}
				} else if (type == INT16ARRAY_PROPERTY) {
					size_t size = propertiesIterator->getSize();
					// check if property size correspond to config size
					if (getSize(name) != (int32) size) {
						if (pass == 2) {
							ostringstream sbuf;
							sbuf << "Error size on property" << name << " : " << getSize(name) << " , " << size;
							m_Errors.push_back(sbuf.str());
							throw CannotSetValue();
						}
					} else {
						getPropertyManager()->setExtraSize(name, propertiesIterator->getColsNb());
						vector<string> array = propertiesIterator->getArray();
						int16* data = any_cast<int16*>(getValue(name));

						for (unsigned i = 0; i < (uint32) size; ++i) {
							size_t loc;
							string value = array[i];
							if ((loc = value.find("0x")) != string::npos) {
								string val = value.substr(loc);
								data[i] = (int16) strtol(val.c_str(), NULL, 16);
							} else {
								data[i] = (int16) strtol(value.c_str(), NULL, 10);
							}
						}

						// when setValue is redefined
					//	if (pass == 2) {
					//		setValue(name, data);
					//	}
					}
				} else if (type == INT32ARRAY_PROPERTY) {
					size_t size = propertiesIterator->getSize();
					// check if property size correspond to config size
					if (getSize(name) != (int32) size) {
						if (pass == 2) {
							ostringstream sbuf;
							sbuf << "Error size on property" << name << " : " << getSize(name) << " , " << size;
							m_Errors.push_back(sbuf.str());
							throw CannotSetValue();
						}
					} else {
						getPropertyManager()->setExtraSize(name, propertiesIterator->getColsNb());
						vector<string> array = propertiesIterator->getArray();
						int32* data = any_cast<int32*>(getValue(name));

						for (unsigned i = 0; i < (uint32) size; ++i) {
							size_t loc;
							string value = array[i];
							if ((loc = value.find("0x")) != string::npos) {
								string val = value.substr(loc);
								data[i] = (int32) strtol(val.c_str(), NULL, 16);
							} else {
								data[i] = (int32) strtol(value.c_str(), NULL, 10);
							}
						}

						// when setValue is redefined
					//	if (pass == 2) {
					//		setValue(name, data);
					//	}
					}
				} else if (type == INT64ARRAY_PROPERTY) {
					size_t size = propertiesIterator->getSize();
					// check if property size correspond to config size
					if (getSize(name) != (int32) size) {
						if (pass == 2) {
							ostringstream sbuf;
							sbuf << "Error size on property" << name << " : " << getSize(name) << " , " << size;
							m_Errors.push_back(sbuf.str());
							throw CannotSetValue();
						}
					} else {
						getPropertyManager()->setExtraSize(name, propertiesIterator->getColsNb());
						vector<string> array = propertiesIterator->getArray();
						int64* data = any_cast<int64*>(getValue(name));

						for (unsigned i = 0; i < (uint32) size; ++i) {
							size_t loc;
							string value = array[i];
							if ((loc = value.find("0x")) != string::npos) {
								string val = value.substr(loc);
								data[i] = (int64) strtol(val.c_str(), NULL, 16);
							} else {
								data[i] = (int64) strtol(value.c_str(), NULL, 10);
							}
						}

						// when setValue is redefined
					//	if (pass == 2) {
					//		setValue(name, data);
					//	}
					}
				} else if (type == UINT16ARRAY_PROPERTY) {
					size_t size = propertiesIterator->getSize();
					// check if property size correspond to config size
					if (getSize(name) != (int32) size) {
						if (pass == 2) {
							ostringstream sbuf;
							sbuf << "Error size on property" << name << " : " << getSize(name) << " , " << size;
							m_Errors.push_back(sbuf.str());
							throw CannotSetValue();
						}
					} else {
						getPropertyManager()->setExtraSize(name, propertiesIterator->getColsNb());
						vector<string> array = propertiesIterator->getArray();
						uint16* data = any_cast<uint16*>(getValue(name));

						for (unsigned i = 0; i < (uint32) size; ++i) {
							size_t loc;
							string value = array[i];
							if ((loc = value.find("0x")) != string::npos) {
								string val = value.substr(loc);
								data[i] = (uint16) strtoul(val.c_str(), NULL, 16);
							} else {
								data[i] = (uint16) strtoul(value.c_str(), NULL, 10);
							}
						}

						// when setValue is redefined
					//	if (pass == 2) {
					//		setValue(name, data);
					//	}
					}
				} else if (type == UINT32ARRAY_PROPERTY) {
					size_t size = propertiesIterator->getSize();
					// check if property size correspond to config size
					if (getSize(name) != (int32) size) {
						if (pass == 2) {
							ostringstream sbuf;
							sbuf << "Error size on property" << name << " : " << getSize(name) << " , " << size;
							m_Errors.push_back(sbuf.str());
							throw CannotSetValue();
						}
					} else {
						getPropertyManager()->setExtraSize(name, propertiesIterator->getColsNb());
						vector<string> array = propertiesIterator->getArray();
						uint32* data = any_cast<uint32*>(getValue(name));

						for (unsigned i = 0; i < (uint32) size; ++i) {
							size_t loc;
							string value = array[i];
							if ((loc = value.find("0x")) != string::npos) {
								string val = value.substr(loc);
								data[i] = (uint32) strtoul(val.c_str(), NULL, 16);
							} else {
								data[i] = (uint32) strtoul(value.c_str(), NULL, 10);
							}
						}

						// when setValue is redefined
					//	if (pass == 2) {
					//		setValue(name, data);
					//	}
					}
				} else if (type == UINT64ARRAY_PROPERTY) {
					size_t size = propertiesIterator->getSize();
					// check if property size correspond to config size
					if (getSize(name) != (int32) size) {
						if (pass == 2) {
							ostringstream sbuf;
							sbuf << "Error size on property" << name << " : " << getSize(name) << " , " << size;
							m_Errors.push_back(sbuf.str());
							throw CannotSetValue();
						}
					} else {
						getPropertyManager()->setExtraSize(name, propertiesIterator->getColsNb());
						vector<string> array = propertiesIterator->getArray();
						uint64* data = any_cast<uint64*>(getValue(name));

						for (unsigned i = 0; i < (uint32) size; ++i) {
							size_t loc;
							string value = array[i];
							if ((loc = value.find("0x")) != string::npos) {
								string val = value.substr(loc);
								data[i] = (uint64) strtol(val.c_str(), NULL, 16);
							} else {
								data[i] = (uint64) strtol(value.c_str(), NULL, 10);
							}
						}

						// when setValue is redefined
					//	if (pass == 2) {
					//		setValue(name, data);
					//	}
					}
				} else if (type == FLOAT32ARRAY_PROPERTY) {
					size_t size = propertiesIterator->getSize();
					// check if property size correspond to config size
					if (getSize(name) != (int32) size) {
						if (pass == 2) {
							ostringstream sbuf;
							sbuf << "Error size on property" << name << " : " << getSize(name) << " , " << size;
							m_Errors.push_back(sbuf.str());
							throw CannotSetValue();
						}
					} else {
						getPropertyManager()->setExtraSize(name, propertiesIterator->getColsNb());
						vector<string> array = propertiesIterator->getArray();
						float32* data = any_cast<float32*>(getValue(name));

						for (unsigned i = 0; i < (uint32) size; ++i) {
							data[i] = (float32) strtod(array[i].c_str(), NULL);
						}

						// when setValue is redefined
						//if (pass == 2) {
						//	setValue(name, data);
						//}
					}
				} else if (type == FLOAT64ARRAY_PROPERTY) {
					size_t size = propertiesIterator->getSize();
					// check if property size correspond to config size
					if (getSize(name) != (int32) size) {
						if (pass == 2) {
							ostringstream sbuf;
							sbuf << "Error size on property" << name << " : " << getSize(name) << " , " << size;
							m_Errors.push_back(sbuf.str());
							throw CannotSetValue();
						}
					} else {
						getPropertyManager()->setExtraSize(name, propertiesIterator->getColsNb());
						vector<string> array = propertiesIterator->getArray();
						float64* data = any_cast<float64*>(getValue(name));

						for (unsigned i = 0; i < (uint32) size; ++i) {
							data[i] = (float64) strtod(array[i].c_str(), NULL);
						}

						// when setValue is redefined
						//if (pass == 2) {
						//	setValue(name, data);
						//}
					}
				} else {
					// Error , bad type
				}
			} catch (PropertyManager::UnknownProperty&) {
				ostringstream sbuf;
				sbuf << "Property " << name << " unknown for " << m_type;
				m_Errors.push_back(sbuf.str());
			} catch (CannotSetValue&) {
				if (pass == 2) {
					ostringstream sbuf;
					sbuf << "Cannot set property " << name;
					m_Warnings.push_back(sbuf.str());
				}
			} catch (AbstractController::ControllerNotConfigured&) {
				ostringstream sbuf;
				sbuf << "Cannot set property " << name << " because " << m_name << " is not configured";
				m_Errors.push_back(sbuf.str());
			}
		}
		fin: ;
	}
}

///*
// * initProperty
// */
//void AbstractController::initProperty(AnyProperty* property) const {
//	if (property->getType() == UINT16_PROPERTY) {
//		property->setValue(uint16(1));
//	} else if (property->getType() == UINT32_PROPERTY) {
//		property->setValue(uint32(1));
//	} else if (property->getType() == UINT64_PROPERTY) {
//		property->setValue(uint64(1));
//	} else if (property->getType() == INT16_PROPERTY) {
//		property->setValue(int16(1));
//	} else if (property->getType() == INT32_PROPERTY) {
//		property->setValue(int32(1));
//	} else if (property->getType() == INT64_PROPERTY) {
//		property->setValue(int64(1));
//	} else if (property->getType() == FLOAT32_PROPERTY) {
//		property->setValue((float32) 1.);
//	} else if (property->getType() == FLOAT64_PROPERTY) {
//		property->setValue((float64) 1.);
//	} else if (property->getType() == BOOLEAN_PROPERTY) {
//		property->setValue(false);
//	} else if (property->getType() == STRING_PROPERTY) {
//		property->setValue(string(""));
//	} else if (property->getType() == INT16ARRAY_PROPERTY) {
//		property->setValue((int16*) 0);
//		property->setSize(0);
//	} else if (property->getType() == INT32ARRAY_PROPERTY) {
//		property->setValue((int32*) 0);
//		property->setSize(0);
//	} else if (property->getType() == INT64ARRAY_PROPERTY) {
//		property->setValue((int64*) 0);
//		property->setSize(0);
//	} else if (property->getType() == UINT16ARRAY_PROPERTY) {
//		property->setValue((uint16*) 0);
//		property->setSize(0);
//	} else if (property->getType() == UINT32ARRAY_PROPERTY) {
//		property->setValue((uint32*) 0);
//		property->setSize(0);
//	} else if (property->getType() == UINT64ARRAY_PROPERTY) {
//		property->setValue((uint64*) 0);
//		property->setSize(0);
//	} else if (property->getType() == FLOAT32ARRAY_PROPERTY) {
//		property->setValue((float32*) 0);
//		property->setSize(0);
//	} else if (property->getType() == FLOAT64ARRAY_PROPERTY) {
//		property->setValue((float64*) 0);
//		property->setSize(0);
//	}
//	m_PropertyManager->addProperty(property);
//}
//
///*
// * initProperty
// */
//void AbstractController::initProperty(AnyProperty* property, boost::any value) const {
//	property->setValue(value);
//	m_PropertyManager->addProperty(property);
//}

/*
 * removeProperty
 */
void AbstractController::removeProperty(const std::string& propertyName) {

	try {
		m_PropertyManager->removeProperty(propertyName);

	} catch (PropertyManager::UnknownProperty&) {
		throw UnknownProperty(m_type, m_name, this, propertyName);
	}
}

/*
 * containsProperty
 */
bool AbstractController::containsProperty(const std::string& propertyName) {
	return m_PropertyManager->containsProperty(propertyName);
}

/*
 * getType
 */
const std::string& AbstractController::getType() const {
	return m_type;
}

void AbstractController::setType(const std::string& type) {
	m_type = type;
	m_PropertyManager->addServantType(type, true);
}

void AbstractController::addType(const std::string& type) {
	m_PropertyManager->addServantType(type, false);
}

/*
 * getControllerType
 */
const std::string& AbstractController::getControllerType() const {
	return m_controllerType;
}

/*
 * isManyScanProperties
 */
bool AbstractController::isManyScanProperties() const {
	vector<string> properties = m_PropertyManager->getLocalProperties();
	vector<string>::iterator it = properties.begin();
	uint32 i = 0;
	while (it != properties.end()) {
		if (PropertyInfos::getInstance()->isScannable(m_type, *it)) {
			++i;
//        	if ((m_Type == CURRENT_EXPERIMENT_CONTROLLER) | (m_Type == COMPOSITECURRENT_EXPERIMENT_CONTROLLER)) {
//        		break;
//        	}
			if (i > 1)
				break;
		}
		++it;
	}
	return (i > 1);
}

/*
 * getAllProperties
 */
std::vector<AnyProperty*> AbstractController::getAllProperties() const {
	return m_PropertyManager->getProperties();
}

/*
 * getAllAllProperties
 */
std::vector<AnyProperty*> AbstractController::getAllAllProperties() const {
	return m_PropertyManager->getAllProperties();
}

/*
 * getPropertyManager
 */
PropertyManager* AbstractController::getPropertyManager() const {
	return m_PropertyManager;
}

/*
 * setPropertyManager
 */
void AbstractController::setPropertyManager(PropertyManager* aPropertyManager) {
	m_PropertyManager = aPropertyManager;
}

/*
 * histogramming
 */
bool AbstractController::histogramming() const {
	return (m_Histogramming & m_Enabled);
}

/*
 * setHistogramming
 */
void AbstractController::setHistogramming(bool aValue) {
	m_Histogramming = aValue;
}

/*
 * isEnabled
 */
bool AbstractController::isEnabled() const {
	return m_Enabled;
}

/*
 * getName
 */
const std::string& AbstractController::getName() const {
	return m_name;
}

/*
 * setName
 */
void AbstractController::setName(const std::string& inputName) {
	m_name = inputName;
	m_PropertyManager->renameServant(inputName);
	m_CommandManager->renameServant(inputName);
	m_ServantManager->renameServant(inputName);
}

/*
 * getProperties
 */
std::string AbstractController::getProperties() const {
	return m_PropertyManager->getPropertyList();
}

/*
 * getSaveProperties
 */
std::string AbstractController::getSaveProperties() const {
	return m_PropertyManager->getSavePropertyList();
}

/*
 * getSpyProperties
 */
std::vector<AnyProperty *> AbstractController::getSpyProperties() const {
	return m_PropertyManager->getSpyPropertyList();
}

/*
 * getAliasProperties
 */
std::string AbstractController::getAliasProperties() const {
	return m_PropertyManager->getAliasPropertyList();
}

/*
 * getAliasProperties
 */
std::string AbstractController::getGenericAliasProperties() const {
	return m_PropertyManager->getGenericAliasProperties();
}

/*
 * getPropertiesToSave
 */
std::vector<AnyProperty *> AbstractController::getPropertiesToSave() const {
	return m_PropertyManager->getPropertiesToSave();
}

/*
 * getCommands
 */
std::string AbstractController::getCommands() const {
	return m_CommandManager->getCommands();
}

/*
 * commandExists
 */
bool AbstractController::commandExists(const std::string& commandName) const {
	return m_CommandManager->commandExists(commandName);
}

bool AbstractController::isLogging() const {
	return true;
}

/*
 * log
 */
LogStream AbstractController::log(Level const & level) {

	Level logLevel = level;
	if (!isLogging() && logLevel.getLevel() == 0) {
		logLevel = Level::s_Debug;
	}

	return LogStream(m_type, m_name, logLevel);
}

/*
 * Property log for database messages
 */
PLog AbstractController::plog(std::string const & propertyName, bool hidden) {

	// first search for the BaseProperty in PropertyRefresher
	map<string, BaseProperty *>::iterator r = PropertyContainer::m_properties.find(propertyName);

	if (r != PropertyContainer::m_properties.end()) {
		return PLog(r->second->toPropertyLog(), hidden);
	}

	if (getPropertyManager()->containsProperty(propertyName)) {
		AnyProperty* anyProperty = getPropertyManager()->getProperty(propertyName);
		std::string controllerType = getType();
		return PLog(anyProperty, controllerType, hidden);
	}
	return PLog(0, "", false);
}

Level AbstractController::getLogLevel(Level const & level) const {

	Level logLevel = level;
	if (!isLogging() && logLevel.getLevel() == 0) {
		logLevel = Level::s_Debug;
	}

	return logLevel;
}

/*
 * sleep
 */
void AbstractController::sleep(uint32 time) {
	if (Simulated::activated == false)
		::sleep(time);
}

/*
 * usleep
 */
void AbstractController::usleep(uint32 time) {
	if (Simulated::activated == false)
		::usleep(time);
}

/*
 * addFamily
 */
void AbstractController::setFamily(const std::string& familyName, const std::string& subfamilyName) {

	m_defaultFamily = familyName;
	m_defaultSubfamily = subfamilyName;

	// set the data provider
	getServantManager()->setFamily(familyName, subfamilyName);
}

void AbstractController::setConfigFamily(const std::string& configFamilyName, const std::string& configSubfamilyName) {

	string familyName;
	string subfamilyName;

	if (configFamilyName == "" || configFamilyName == m_defaultFamily) {
		familyName = m_defaultFamily;

		if (configSubfamilyName == "") {
			subfamilyName = m_defaultSubfamily;
		} else {
			subfamilyName = configSubfamilyName;
		}

	} else {
		familyName = configFamilyName;

		if (configSubfamilyName == "") {
			subfamilyName = family::NONE;
		} else {
			subfamilyName = configSubfamilyName;
		}
	}

	// set the data provider
	getServantManager()->setFamily(familyName, subfamilyName);
}

std::string AbstractController::getFamily() const {
	return getServantManager()->getFamily();
}

std::string AbstractController::getSubfamily() const {
	return getServantManager()->getSubfamily();
}

void AbstractController::setFavourite(bool value) {
	getServantManager()->setFavourite(value);
}

bool AbstractController::isFavourite() const {
	return getServantManager()->isFavourite();
}

std::string AbstractController::getConfigFamily() const {
	string family = getFamily();
	if (family != m_defaultFamily) {
		return family;
	}

	return "";
}

std::string AbstractController::getConfigSubfamily() const {
	string subfamily = getSubfamily();
	if (subfamily != m_defaultSubfamily) {
		return subfamily;
	}

	return "";
}

/*
 * updateError
 */
void AbstractController::updateError(ChangeAspect* aChangeAspect) {

	// no need to forward the error, we never redefine updateError except for GenericScan.
}

CommandManager* AbstractController::getCommandManager() const {
	return m_CommandManager;
}

ServantManager* AbstractController::getServantManager() const {
	return m_ServantManager;
}

void AbstractController::enable(bool value) {

	m_Enabled = value;

	if (value) {
		m_ServantManager->enable();
	} else {
		m_ServantManager->disable();
	}
}

/*
 * getAliasPropertyMap
 */
std::map<std::string, std::string> AbstractController::getAliasPropertyMap() const {
	return m_PropertyManager->getAliasPropertyMap();
}

void AbstractController::test() {
	// Do nothing
}

string AbstractController::formatDoubleValue(const string& format, float64 value) {
	ostringstream buf;
	if (format.empty() == true) {
		buf << value;
	} else {
		buf << boost::format(format) % value;
	}
	return buf.str();
}

