/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef ABSTRACTCONTROLLER_H
#define ABSTRACTCONTROLLER_H

#include <string>
#include <vector>

#include <common/base/BaseTypes.h>
#include "DataProvider/Property/AnyProperty.h"
#include "DataProvider/Property/TypedProperty/PropertyRefresher.h"
#include "DataProvider/Repeater.h"
#include "DataProvider/ServantManager.h"
#include "ErrorsAlarmsManager/Exceptions.h"
#include "ErrorsAlarmsManager/ErrorDef.h"
#include "LoggingServices/LogStream.h"
#include "LoggingServices/PLog.h"
#include "Utilities/Simulated.h"
#include "Utilities/Assert.h"
#include <boost/thread/mutex.hpp>

class AnyProperty;
class PropertyConfig;
class PropertyManager;
class CommandManager;
class ServantManager;

namespace controller {
	class Command;
}

namespace family {
	extern const std::string NONE;
	extern const std::string HIDDEN;
}

/**
 * The AbstractController acts as a superclass for the ElementaryController
 * and the ScientificController. It encapsulates the joint functionality of
 * the two classes. Both sub-classes have the ability the modify properties
 * to provide a user interface to the object and are also capable of executing
 * particular commands.
 */
class AbstractController: public Repeater, public PropertyRefresher {

	friend class LogStream;
	friend class controller::Command;
	friend class ClientCommand;
	friend class InstrumentManager;
	friend class ExperimentController;

public:

	/*
	 * temporary exceptions: should use Error manager
	 */
	class UnknownPropertyType {
	public:
		UnknownPropertyType(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info);

		const std::string & getType() const;
		const std::string & getName() const;
		Publisher* getPublisher() const;
		const std::string & getInfo() const;

	private:
		std::string m_type;
		std::string m_name;
		Publisher* m_publisher;
		std::string m_info;
	};

	class UnknownProperty: public NotOk {
	public:
		UnknownProperty(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info);

		UnknownProperty(UnknownPropertyType const & exception);
	};

	class TimeOutOccured: public NotOk {
	public:
		TimeOutOccured();
	};

	class ControllerNotConfigured: public NotOk {
	public:
		ControllerNotConfigured();
		ControllerNotConfigured(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info);
	};

	class WrongPropertyType: public NotOk {
	public:
		WrongPropertyType(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info);
	};

	class UnknownCommand: public NotOk {
	public:
		UnknownCommand(const std::string& type, const std::string& name, Publisher* publisher, const std::string& info);
	};

	class ExecutionError: public NotOk {
	public:
		ExecutionError();
		ExecutionError(const std::string& message);
	};

	AbstractController(const std::string& inputType, const std::string& inputName, const std::string& controllerType);

	/**
	 * Copy constructor.
	 */
	AbstractController(const AbstractController& controller);
	AbstractController();

	/**
	 * Destructor.
	 */
	virtual ~AbstractController();

	void copy(const AbstractController& controller);

	/**
	 * Returns the controller class.
	 */
	virtual ServantClass getServantClass() const;

	/**
	 * Gets the list of commands.
	 */
	std::vector<std::string> getCommandList() const;

	/**
	 * Returns a stringified list of all properties. Call is forwarded to
	 * the PropertyManager which handles formatting issues.
	 */
	virtual std::string getProperties() const;

	/**
	 * Returns a stringified list of all properties. Call is forwarded to
	 * the PropertyManager which handles formatting issues.
	 */
	virtual std::string getSaveProperties() const;

	/**
	 * Returns a stringified list of all properties. Call is forwarded to
	 * the PropertyManager which handles formatting issues.
	 */
	virtual std::string getAliasProperties() const;

	/**
	 * Returns a stringified list of all properties. Call is forwarded to
	 * the PropertyManager which handles formatting issues.
	 */
	virtual std::string getGenericAliasProperties() const;

	/**
	 * Returns a stringified list of all properties. Call is forwarded to
	 * the PropertyManager which handles formatting issues.
	 */
	virtual std::map<std::string, std::string> getAliasPropertyMap() const;

	/**
	 * Returns a std::stringified list of all properties. Call is forwarded to
	 * the PropertyManager which handles formatting issues.
	 */
	virtual std::vector<AnyProperty *> getSpyProperties() const;

	/**
	 * Returns a list of all properties to save. Call is forwarded to
	 * the PropertyManager.
	 */
	virtual std::vector<AnyProperty *> getPropertiesToSave() const;

	/**
	 * Returns a list of all properties to set on simulated server. Call is forwarded to
	 * the PropertyManager.
	 */
	std::vector<AnyProperty *> getAllProperties() const;

	/**
	 * Returns a list of all properties to set on simulated server. Call is forwarded to
	 * the PropertyManager.
	 */
	std::vector<AnyProperty *> getAllAllProperties() const;

	/**
	 * Returns the PropertyManager pointer.
	 */
	PropertyManager* getPropertyManager() const;

	/**
	 * Sets the PropertyManager pointer.
	 */
	void setPropertyManager(PropertyManager* aPropertyManager);

	/**
	 * Returns the actual value of particular a property.
	 * Return type is of type boost::any which needs to be recast to
	 * obtain the true value.
	 */
	virtual boost::any getValue(const std::string& propertyName) ;
//	template <class T> T getTypedValue(const std::string& propertyName) {
//		boost::any pvalue = getValue(propertyName);
//		std::string ptype = getPropertyType(propertyName);
//		T rvalue;
//		if (ptype == UINT16_PROPERTY) {
//			rvalue = (T) boost::any_cast<uint16>(pvalue);
//		}
//		else if (ptype == UINT32_PROPERTY) {
//			rvalue = (T) boost::any_cast<uint32>(pvalue);
//		}
//		else if (ptype == UINT64_PROPERTY) {
//			rvalue = (T) boost::any_cast<uint64>(pvalue);
//		}
//		else if (ptype == INT16_PROPERTY) {
//			rvalue = (T) boost::any_cast<int16>(pvalue);
//		}
//		else if (ptype == INT32_PROPERTY) {
//			rvalue = (T) boost::any_cast<int32>(pvalue);
//		}
//		else if (ptype == INT64_PROPERTY) {
//			rvalue = (T) boost::any_cast<int64>(pvalue);
//		}
//		else if (ptype == FLOAT32_PROPERTY) {
//			rvalue = (T) boost::any_cast<float32>(pvalue);
//		}
//		else if (ptype == FLOAT64_PROPERTY) {
//			rvalue = (T) boost::any_cast<float64>(pvalue);
//		}
//		else if (ptype == BOOLEAN_PROPERTY) {
//			rvalue = (T) boost::any_cast<bool>(pvalue);
//		}
////		else if (ptype == STRING_PROPERTY) {
////			rvalue = (T) boost::any_cast<std::string>(pvalue);
////		}
////		else if (ptype == UINT16ARRAY_PROPERTY) {
////			rvalue = (T) boost::any_cast<uint16*>(pvalue);
////		}
////		else if (ptype == UINT32ARRAY_PROPERTY) {
////			rvalue = (T) boost::any_cast<uint32*>(pvalue);
////		}
////		else if (ptype == UINT64ARRAY_PROPERTY) {
////			rvalue = (T) boost::any_cast<uint64*>(pvalue);
////		}
////		else if (ptype == INT16ARRAY_PROPERTY) {
////			rvalue = (T) boost::any_cast<int16*>(pvalue);
////		}
////		else if (ptype == INT32ARRAY_PROPERTY) {
////			rvalue = (T) boost::any_cast<int32*>(pvalue);
////		}
////		else if (ptype == INT64ARRAY_PROPERTY) {
////			rvalue = (T) boost::any_cast<int64*>(pvalue);
////		}
//		return rvalue;
//	}

	/**
	 * Sets the actual value of a particular Property.
	 * Parameter is of type boost::any which allows for all types to be used (including pointers).
	 */
	void setValue(const std::string& propertyName, boost::any aValue) ;
//	template <class T> void setValue(const std::string& propertyName, T aValue) {
//		setValue(propertyName, boost::any(aValue));
//	}
	void setValue(const std::string& propertyName, bool aValue) ;
	void setValue(const std::string& propertyName, int16 aValue) ;
	void setValue(const std::string& propertyName, int32 aValue) ;
	void setValue(const std::string& propertyName, int64 aValue) ;
	void setValue(const std::string& propertyName, uint16 aValue) ;
	void setValue(const std::string& propertyName, uint32 aValue) ;
	void setValue(const std::string& propertyName, uint64 aValue) ;
	void setValue(const std::string& propertyName, float32 aValue) ;
	void setValue(const std::string& propertyName, float64 aValue) ;
	void setValue(const std::string& propertyName, int16* aValue) ;
	void setValue(const std::string& propertyName, int32* aValue) ;
	void setValue(const std::string& propertyName, int64* aValue) ;
	void setValue(const std::string& propertyName, uint16* aValue) ;
	void setValue(const std::string& propertyName, uint32* aValue) ;
	void setValue(const std::string& propertyName, uint64* aValue) ;
	void setValue(const std::string& propertyName, float32* aValue) ;
	void setValue(const std::string& propertyName, float64* aValue) ;

	/**
	 * Sets the actual value of a particular property.
	 * Parameter is of type boost::any which allows for all types to be used (including pointers).
	 * Adds the formatted value.
	 */
	void setValueAndFormattedValue(const std::string& propertyName, boost::any aValue, const std::string& formattedValue) ;
//	template <class T> void setValueAndFormattedValue(const std::string& propertyName, T aValue, const std::string& formattedValue) {
//		setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
//	}
	void setValueAndFormattedValue(const std::string& propertyName, bool aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, int16 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, int32 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, int64 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, uint16 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, uint32 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, uint64 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, float32 aValue, const std::string& formattedValue) ;
	void setValueAndFormattedValue(const std::string& propertyName, float64 aValue, const std::string& formattedValue) ;

	/**
	 * Sets the typed value by calling the typed property so that the log is possible.
	 */
	void setTypedValue(const std::string& propertyName, const boost::any& value) ;

	/**
	 * Sets the value of the property by calling setValue after having acquired the publisher mutex.
	 * The method is called by the IcsController class and can be overridden in order to not acquire
	 * the mutex.
	 */
	virtual void setValueSynchronized(const std::string& propertyName, boost::any aValue) ;

	/**
	 * Returns the type of a particular property. This is necessary
	 * the property contains a pointer to an array of data.
	 */
	virtual const std::string& getPropertyType(const std::string& propertyName) ;

	/**
	 * Returns the alias of a particular property. This is necessary
	 * the property contains a pointer to an array of data.
	 */
	virtual const std::string& getPropertyAlias(const std::string& propertyName) ;

	/**
	 * Returns the property unit of a particular property.
	 */
	std::string getPropertyUnit(const std::string& propertyName);

	/**
	 * Returns the size of a particular property. This is necessary
	 * the property contains a pointer to an array of data.
	 */
	int32 getSize(const std::string& propertyName) ;

	/**
	 * Set the size of a particular property. This is necessary
	 * the property contains a pointer to an array of data.
	 */
	void setSize(const std::string& propertyName, int32 aSize) ;

	/**
	 * Returns the extra size of a particular property. This is necessary
	 * the property contains a pointer to an array of data.
	 */
	void setExtraSize(const std::string& propertyName, int32 size) ;

	/**
	 * Returns the extra size of a particular property. This is necessary
	 * the property contains a pointer to an array of data.
	 */
	int32 getExtraSize(const std::string& propertyName) ;

	/**
	 * Init the properties from a vector of PropertyConfig
	 */
	void initProperties(const std::vector<PropertyConfig>& aProperties, int32 pass);

	/**
	 * Init the properties.
	 */
	void initProperties();

	/**
	 * Returns true if has more than one scan properties.
	 */
	bool isManyScanProperties() const;

	/**
	 * Returns true if contains a property.
	 */
	bool containsProperty(const std::string& propertyName);

	/**
	 * Returns a stringified list of available commands.
	 * Each command name is terminated by a \\n to separate the different names.
	 */
	std::string getCommands() const;
	
	/**
	 * Sets the default family and the subfamily.
	 */
	void setFamily(const std::string& familyName, const std::string& subfamilyName = family::NONE);

	/**
	 * Returns the family of the controller.
	 */
	std::string getFamily() const;

	/**
	 * Returns the subfamily of the controller.
	 */
	std::string getSubfamily() const;

	/**
	 * Sets the controller favourite.
	 */
	void setFavourite(bool value);

	/**
	 * Returns true if the controller is favourite.
	 */
	bool isFavourite() const;

	/**
	 * Returns true if the command exists.
	 */
	bool commandExists(const std::string& commandName) const;

	/*!
	 *\brief execute
	 *  Execute a particular command if it has been implemented.
	 *\param command: The name of the command
	 *\param blocking: the execution must be blocking true or false
	 *\param logging:The action is able to add some logs
	 */
	virtual void execute(const std::string& command, bool blocking, bool logging = false) =0;

	/**
	 * Returns true if histogramming can be saved.
	 */
	bool histogramming() const;

	/**
	 * Gets the name.
	 */
	virtual const std::string& getName() const;

	/**
	 * Set the name.
	 */
	void setName(const std::string& inputName);

	/**
	 * Gets the type attribute.
	 */
	const std::string& getType() const;

	/**
	 * Sets the type attribute.
	 */
	void setType(const std::string& type);

	/**
	 * Adds the type attribute.
	 */
	void addType(const std::string& type);

	/**
	 * Verifies if the Controller is configured and can be used.
	 */
	virtual bool isConfigured() = 0;

	/*!
	 *\brief refreshSetValue
	 * Each time the value of a static property is modified by clients, the
	 * refresh method is called to update the state of the Controller.
	 *
	 *\param propertyName: The name of the property
	 *\param aValue: the value
	 *\exception  CannotSetValue
	 */
	virtual void refreshSetValue(std::string propertyName, const boost::any aValue) ;

	/**
	 * Calls refresher's followed by refreshSetValue.
	 */
	void refreshValue(const std::string& propertyName, const boost::any& value) ;

	/*
	 * Returns true if contains a function.
	 */
	virtual bool containsFunction(const std::string& function);

	/*
	 * Returns true if contains a command.
	 */
	bool containsCommand(const std::string& commandName) const;

	/**
	 * Get controller type.
	 */
	const std::string& getControllerType() const;

	/**
	 * Sends a synchro event.
	 */
	void sendSynchroEvent(const std::string& commandName, int32 state);

	/**
	 * Sends a refuse event.
	 */
	void sendRefuseEvent(const std::string& commandName, const std::string& cName, const std::string& cCommandName);

	/**
	 * Sends a refuse event.
	 */
	void sendRefuseEvent(const std::string& commandName);

	/**
	 * Sets the controller enabled.
	 */
	virtual void setEnabled(bool all, bool postconf = true) = 0;

	/**
	 * Sets the controller enabled for loading.
	 */
	virtual void setEnabledForLoading(bool all, bool postconf = true) = 0;

	/**
	 * Sets the controller disabled.
	 */
	virtual void setDisabled(bool all = false) = 0;

	/**
	 * Returns true if is enabled.
	 */
	bool isEnabled() const;

	/**
	 * Updates the value. No refresh is called.
	 */
	void updateValue(const std::string& propertyName, boost::any propertyValue) ;
//	template <class T> void updateValue(const std::string& propertyName, T aValue) {
//		updateValue(propertyName, boost::any(aValue));
//	}
	void updateValue(const std::string& propertyName, bool aValue) ;
	void updateValue(const std::string& propertyName, int16 aValue) ;
	void updateValue(const std::string& propertyName, int32 aValue) ;
	void updateValue(const std::string& propertyName, int64 aValue) ;
	void updateValue(const std::string& propertyName, uint16 aValue) ;
	void updateValue(const std::string& propertyName, uint32 aValue) ;
	void updateValue(const std::string& propertyName, uint64 aValue) ;
	void updateValue(const std::string& propertyName, float32 aValue) ;
	void updateValue(const std::string& propertyName, float64 aValue) ;
	void updateValue(const std::string& propertyName, int16* aValue) ;
	void updateValue(const std::string& propertyName, int32* aValue) ;
	void updateValue(const std::string& propertyName, int64* aValue) ;
	void updateValue(const std::string& propertyName, uint16* aValue) ;
	void updateValue(const std::string& propertyName, uint32* aValue) ;
	void updateValue(const std::string& propertyName, uint64* aValue) ;
	void updateValue(const std::string& propertyName, float32* aValue) ;
	void updateValue(const std::string& propertyName, float64* aValue) ;

	/**
	 * Updates the value. No refresh is called.
	 * Adds the formatted value.
	 */
	void updateValueAndFormattedValue(const std::string& propertyName, boost::any propertyValue, const std::string& formattedValue) ;
//	template <class T> void updateValueAndFormattedValue(const std::string& propertyName, T aValue, const std::string& formattedValue) {
//		updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
//	}
	void updateValueAndFormattedValue(const std::string& propertyName, bool aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, int16 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, int32 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, int64 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, uint16 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, uint32 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, uint64 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, float32 aValue, const std::string& formattedValue) ;
	void updateValueAndFormattedValue(const std::string& propertyName, float64 aValue, const std::string& formattedValue) ;

	/*
	 * Mutex use for big datas
	 * TODO REMOVE
	 */
	static std::recursive_mutex m_longArrayMutex;
	static std::recursive_mutex m_float64ArrayMutex;

	/**
	 * Sets the histogramming.
	 */
	void setHistogramming(bool aValue);

	/**
	 * Overriding of updateError functions.
	 */
	void updateError(ChangeAspect* aChangeAspect);

	/*!
	 *\brief computeEstimatedTime
	 * Use to estimate the time in simulation mode
	 *
	 */
	virtual void resetEstimatedTime() {
	}
	virtual float64 computeEstimatedTime(const std::map<std::string, std::string>& amap) {
		return 0.;
	}

	/**
	 * Sets the external sample.
	 */
	virtual void SetExtSample(bool value) {
	}

	/**
	 * Function to be reimplemented in the sub-classes for simple unitary testing.
	 */
	virtual void test();

	/**
	 * Formats a double value.
	 */
	std::string formatDoubleValue(const std::string& format, float64 value);

	/*
	 * Registers a command.
	 */
	void initCommand(const std::string& commandName) const;

	/**
	 * Returns a log stream.
	 */
	LogStream log(Level const & level);

	/**
	 * Returns a property log.
	 */
	PLog plog(std::string const & propertyName, bool hidden = false);

	/**
	 * Returns the log level.
	 */
	Level getLogLevel(Level const & level) const;

	virtual void fillScanList(std::list<std::pair<std::string, std::string> > & list) = 0;

	const std::vector<std::string>& getErrors() const;
	const std::vector<std::string>& getWarnings() const;
	void pushWarning(const std::string& warning);
	void pushError(const std::string& error);
	void clearErrors();
	void clearWarnings();

	/**
	 * Returns true if the controller is favourite.
	 */
	bool isClone() const;

protected:

	void constructor(const std::string& type, const std::string& name, const std::string& controllerType);

//	/**
//	 * Registers the property.
//	 */
//	void initProperty(AnyProperty* property) const;
//
//	/**
//	 * Registers the property.
//	 */
//	void initProperty(AnyProperty* property, boost::any value) const;

	/**
	 * Removes the property.
	 */
	void removeProperty(const std::string& propertyName);

	/**
	 * Removes the command.
	 */
	void removeCommand(const std::string& commandName);

	/**
	 * Updates the value with size attribute.
	 */
	void updateValue(const std::string& propertyName, boost::any propertyValue, int32 size) ;
//	template <class T> void updateValue(const std::string& propertyName, T aValue, int32 size) {
//		updateValue(propertyName, boost::any(aValue), size);
//	}
	void updateValue(const std::string& propertyName, bool aValue, int32 size) ;
	void updateValue(const std::string& propertyName, uint16 aValue, int32 size) ;
	void updateValue(const std::string& propertyName, uint32 aValue, int32 size) ;
	void updateValue(const std::string& propertyName, uint64 aValue, int32 size) ;
	void updateValue(const std::string& propertyName, int16 aValue, int32 size) ;
	void updateValue(const std::string& propertyName, int32 aValue, int32 size) ;
	void updateValue(const std::string& propertyName, int64 aValue, int32 size) ;
	void updateValue(const std::string& propertyName, float32 aValue, int32 size) ;
	void updateValue(const std::string& propertyName, float64 aValue, int32 size) ;
	void updateValue(const std::string& propertyName, uint16* aValue, int32 size) ;
	void updateValue(const std::string& propertyName, uint32* aValue, int32 size) ;
	void updateValue(const std::string& propertyName, uint64* aValue, int32 size) ;
	void updateValue(const std::string& propertyName, int16* aValue, int32 size) ;
	void updateValue(const std::string& propertyName, int32* aValue, int32 size) ;
	void updateValue(const std::string& propertyName, int64* aValue, int32 size) ;
	void updateValue(const std::string& propertyName, float32* aValue, int32 size) ;
	void updateValue(const std::string& propertyName, float64* aValue, int32 size) ;

	/**
	 * Registers the function.
	 */
	void registerFunction(const std::string& function) ;

	/**
	 * Returns true if is logging. Used to avoid test in sub-commands.
	 */
	virtual bool isLogging() const;

	/**
	 * Local sleep, no sleep when simulated mode is activated.
	 */
	void sleep(uint32 time);

	/**
	 * Local usleep, no sleep when simulated mode is activated.
	 */
	void usleep(uint32 time);

	/**
	 * For inherited classes.
	 */
	CommandManager* getCommandManager() const;

	/**
	 * For inherited classes.
	 */
	ServantManager* getServantManager() const;

	/**
	 * Local variables
	 */
	std::string m_type;
	std::string m_name;
	std::string m_controllerType;

	/**
	 * Enables the controller.
	 */
	void enable(bool value);

	bool m_clone;

private:
	/**
	 * Tests if a property has changed. Compares old and new value.
	 */
	void hasChanged(const std::string& propertyName, boost::any oldValue, boost::any newValue);

	/**
	 * Sets the configuration family and the subfamily.
	 */
	void setConfigFamily(const std::string& configFamilyName, const std::string& configSubfamilyName);

	/**
	 * Gets the configuration family.
	 */
	std::string getConfigFamily() const;

	/**
	 * Gets the configuration subfamily.
	 */
	std::string getConfigSubfamily() const;

	bool m_Enabled;

	PropertyManager* m_PropertyManager;
	CommandManager* m_CommandManager;
	ServantManager* m_ServantManager;

	std::vector<std::string> m_Functions;
	std::string m_defaultFamily;
	std::string m_defaultSubfamily;
	bool m_Histogramming;
	std::vector<std::string> m_Errors;
	std::vector<std::string> m_Warnings;

	static const std::string FAVORITES;
};

inline void AbstractController::setValue(const std::string& propertyName, bool aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, uint16 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, uint32 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, uint64 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, int16 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, int32 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, int64 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, float32 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, float64 aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, uint16* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, uint32* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, uint64* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, int16* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, int32* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, int64* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, float32* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValue(const std::string& propertyName, float64* aValue)  {
	setValue(propertyName, boost::any(aValue));
}

inline void AbstractController::setValueAndFormattedValue(const std::string& propertyName, bool aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::setValueAndFormattedValue(const std::string& propertyName, uint16 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::setValueAndFormattedValue(const std::string& propertyName, uint32 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::setValueAndFormattedValue(const std::string& propertyName, uint64 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::setValueAndFormattedValue(const std::string& propertyName, int16 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::setValueAndFormattedValue(const std::string& propertyName, int32 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::setValueAndFormattedValue(const std::string& propertyName, int64 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::setValueAndFormattedValue(const std::string& propertyName, float32 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::setValueAndFormattedValue(const std::string& propertyName, float64 aValue, const std::string& formattedValue)  {
	setValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValue(const std::string& propertyName, bool aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, int16 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, int32 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, int64 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, uint16 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, uint32 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, uint64 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, float32 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, float64 aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, int16* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, int32* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, int64* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, uint16* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, uint32* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, uint64* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, float32* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValue(const std::string& propertyName, float64* aValue)  {
	updateValue(propertyName, boost::any(aValue));
}

inline void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, bool aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, uint16 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, uint32 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, uint64 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, int16 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, int32 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, int64 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, float32 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValueAndFormattedValue(const std::string& propertyName, float64 aValue, const std::string& formattedValue)  {
	updateValueAndFormattedValue(propertyName, boost::any(aValue), formattedValue);
}

inline void AbstractController::updateValue(const std::string& propertyName, bool aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, uint16 aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, uint32 aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, uint64 aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, int16 aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, int32 aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, int64 aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, float32 aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, float64 aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, uint16* aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, uint32* aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, uint64* aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, int16* aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, int32* aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, int64* aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, float32* aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline void AbstractController::updateValue(const std::string& propertyName, float64* aValue, int32 size)  {
	updateValue(propertyName, boost::any(aValue), size);
}

inline const std::vector<std::string>& AbstractController::getErrors() const {
	return m_Errors;
}

inline const std::vector<std::string>& AbstractController::getWarnings() const {
	return m_Warnings;
}

inline void AbstractController::pushWarning(const std::string& warning) {
	m_Warnings.push_back(warning);
}

inline void AbstractController::pushError(const std::string& error) {
	m_Errors.push_back(error);
}

inline void AbstractController::clearErrors() {
	m_Errors.clear();
}

inline void AbstractController::clearWarnings() {
	m_Warnings.clear();
}

inline 	bool AbstractController::isClone() const {
	return m_clone;
}

static const std::string TEST_CONTROLLER_TYPE = "TestController";
static const std::string EXPERIMENT_CONTROLLER_TYPE = "ExperimentController";

#endif
