/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ClientCommand.h"
#include "CommandNames.h"
#include "InstrumentAbstraction/AbstractController.h"
#include "CommandNames.h"
#include "DataProvider/CommandManager.h"
#include "Scheduler/AgentPolicy/AgentPolicy.h"
#include <LoggingServices/LogStream.h>
#include <LoggingServices/Log.h>

#include <iostream>

using namespace std;

ClientCommand::ClientCommand(AbstractController* aController, const std::string& name, bool logging, bool blocking) :
	m_controller(aController), m_name(name), m_logging(logging), m_blocking(blocking) {
}

ClientCommand::~ClientCommand() {
}

void ClientCommand::execute() {

	setCommandStateBeforeExecution();

	/*
	 * Send a synchro event to client saying that a command is starting
	 */
	if (m_name == COMMAND_PAUSE) {
		m_controller->sendSynchroEvent(m_name, 2);
	} else {
		m_controller->sendSynchroEvent(m_name, 0);
	}

	//AgentPolicy::getInstance()->activateEC(m_controller);

	m_controller->setActive();
	m_controller->execute(m_name, true, m_logging);
	m_controller->unsetActive();

	AgentPolicy::getInstance()->deactivateEC(m_controller);

	/*
	 * Send a synchro event to client saying that a command has finished
	 */
	if (m_name != COMMAND_PAUSE && m_name != COMMAND_CONTINUE) {
		m_controller->sendSynchroEvent(m_name, 1);
	}

	setCommandStateAfterExecution();
}

void ClientCommand::stop() {
	// Executing the STOP command. Called by the scheduler when exiting.

	m_controller->execute(COMMAND_STOP, false);
}

bool ClientCommand::isBlocking(Command& aCommand) {
	ClientCommand& c = dynamic_cast<ClientCommand&> (aCommand);

	//	cout << m_controller->getName() << "::" << m_name << " ! " << c.m_controller->getName()<< "::" << c.m_name << endl;
	//	cout << m_blocking << " ! " << c.m_blocking << endl;
	if (c.m_blocking == false)
		return false;

	// We can reimplement AgentPolicy to the new framework...

	bool isOk = AgentPolicy::getInstance()->canI(c.m_controller);

	bool notok = (!isOk
			&& c.m_name != COMMAND_STOP
			&& c.m_name != COMMAND_READ
			&& c.m_name != COMMAND_STATUS
			&& c.m_name != COMMAND_PAUSE
			&& c.m_name != COMMAND_CONTINUE
			&& c.m_name != COMMAND_STOP_AT_END);
	//	if (notok) {
	//		cout << m_controller->getName() << "::" << m_name << " refuses " << c.m_controller->getName()<< "::" << c.m_name << endl;
	//
	//		c.m_controller->sendRefuseEvent(c.m_name,m_controller->getName(),m_name);
	//	}
	return notok;
}

bool ClientCommand::reject(bool blocked) {
	return blocked;
}

AbstractController* ClientCommand::getController() const {
	return m_controller;
}

const std::string& ClientCommand::getName() const {
	return m_name;
}

ostream& operator<<(ostream& os, const ClientCommand &c) {
	os << c.toString();
	return os;
}

bool ClientCommand::equal(Command* c1) {
	return ((static_cast<ClientCommand*> (c1)->getController() == m_controller)
			&& (static_cast<ClientCommand*> (c1)->getName() == m_name));
}

void ClientCommand::setCommandStateBeforeExecution() {

	m_controller->getCommandManager()->setCommandState(m_name, common::ACTIVE);

	if (m_name == COMMAND_PAUSE) {
		m_controller->getCommandManager()->setRunningCommandState(m_name, common::PAUSING);
	} else if (m_name == COMMAND_CONTINUE) {
		m_controller->getCommandManager()->setRunningCommandState(m_name, common::RESUMING);
	} else if (m_name == COMMAND_STOP) {
		m_controller->getCommandManager()->setRunningCommandState(m_name, common::STOPPING);
	}
}

void ClientCommand::setCommandStateAfterExecution() {

	if (m_name == COMMAND_PAUSE) {
		m_controller->getCommandManager()->setRunningCommandState(m_name, common::PAUSED);
	} else if (m_name == COMMAND_CONTINUE) {
		m_controller->getCommandManager()->setRunningCommandState(m_name, common::ACTIVE);
	} else if (m_name == COMMAND_STOP) {
		m_controller->getCommandManager()->setRunningCommandState(m_name, common::STOPPED);
	}

	m_controller->getCommandManager()->setCommandState(m_name, common::INACTIVE);
}

std::string ClientCommand::toString() const {
	return getController()->getName() + " " + getName();
}
