/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */
#ifndef CONTROLLER_PTR_H
#define CONTROLLER_PTR_H

#include "BasePtr.h"

class ExperimentController;

template<typename ControllerType>
class ControllerPtr : public BasePtr {

	template<typename>
	friend class ControllerPtr;

public:
	ControllerPtr();
	ControllerPtr(ExperimentController * pointerOwner, const std::string & role);
	ControllerPtr(const ControllerPtr<ControllerType> & ptr);

	void init(ExperimentController * pointerOwner, const std::string & role);
	bool isInited();

	template<typename ControllerFromType>
	void copy(const ControllerPtr<ControllerFromType> & ptr) {
		 m_controller = dynamic_cast<ControllerType *>(ptr.m_controller);

		if (m_controller == 0) {
			throw BadConversion(ControllerFromType::TYPE, ControllerType::TYPE);
		}

		m_abstractController = ptr.m_abstractController;
		m_role = ptr.m_role;
	}

	void reset(ExperimentController * pointerOwner);

	AbstractController ** getControllerPointer();

	virtual std::string getType() const;
	virtual bool assign(AbstractController * controller);
	virtual bool isAssigned() const;

	ControllerType * operator->() const;
	ControllerType * get() const;
	const std::string& getRole() const;
	ControllerPtr<ControllerType>& operator=(const ControllerPtr<ControllerType>& contptr) {
	    if (this != &contptr) { // self-assignment check expected
	    	m_controller = contptr.m_controller;
	    	m_abstractController = contptr.m_abstractController;
	    	m_role = contptr.m_role;
	    }
	    return *this;
	}

	template<typename TargetControllerType>
	TargetControllerType * cast() const {
		TargetControllerType * target = dynamic_cast<TargetControllerType *>(m_controller);
		if (target == 0) {
			throw BadConversion(ControllerType::TYPE, TargetControllerType::TYPE);
		}
	}

private:
	ControllerType * m_controller;
	AbstractController * m_abstractController;
	std::string m_role;
};

template<typename ControllerType>
ControllerPtr<ControllerType>::ControllerPtr() :
	m_controller(0),
	m_abstractController(0) {
}

template<typename ControllerType>
ControllerPtr<ControllerType>::ControllerPtr(ExperimentController * pointerOwner, const std::string & role) :
	m_controller(0),
	m_abstractController(0) {
	init(pointerOwner, role);
}

template<typename ControllerType>
ControllerPtr<ControllerType>::ControllerPtr(const ControllerPtr<ControllerType> & ptr) :
	m_controller(ptr.m_controller),
	m_abstractController(ptr.m_abstractController),
	m_role(ptr.m_role) {
}

template<typename ControllerType>
std::string ControllerPtr<ControllerType>::getType() const {
	return ControllerType::TYPE;
}

template<typename ControllerType>
AbstractController ** ControllerPtr<ControllerType>::getControllerPointer() {
	return &m_abstractController;
}

template<typename ControllerType>
bool ControllerPtr<ControllerType>::assign(AbstractController * controller) {

	// test null
	if (controller == 0) {
		m_controller = 0;
		m_abstractController = 0;
		return true;
	}

	// test compatibility of controller
	ControllerType * typedController = dynamic_cast<ControllerType *>(controller);

	if (typedController != 0) {
		m_controller = typedController;
		m_abstractController = controller;
		return true;
	} else {
		m_controller = 0;
		m_abstractController = 0;
	}

	return false;
}

template<typename ControllerType>
bool ControllerPtr<ControllerType>::isAssigned() const {
	return (m_controller != 0);
}

template<typename ControllerType>
ControllerType * ControllerPtr<ControllerType>::operator->() const {
	return m_controller;
}

template<typename ControllerType>
ControllerType * ControllerPtr<ControllerType>::get() const {
	return m_controller;
}

template<typename ControllerType>
const std::string& ControllerPtr<ControllerType>::getRole() const{
	return m_role;
}

template<typename ControllerType>
bool operator!=(const ControllerPtr<ControllerType>& controller1, const ControllerPtr<ControllerType>& controller2) {
	return (controller1.get()->getName() != controller2.get()->getName());
}

template<typename ControllerType>
bool operator==(const ControllerPtr<ControllerType>& controller1, const ControllerPtr<ControllerType>& controller2) {
	return (controller1.get()->getName() == controller2.get()->getName());
}

#endif
