/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */
#ifndef DRIVER_PTR_H
#define DRIVER_PTR_H

#include "BasePtr.h"
#include "TestController.h"
#include "Scheduler/Sender.h"
#include "ClientCommand.h"

class ExperimentController;

template<typename DriverType>
class DriverAdapter : public DriverType {

public:
	DriverAdapter(const std::string& name, const std::string& type, DeviceDriver * driver);

	virtual void execute(const std::string& command);
	virtual void refreshSetValue(std::string propertyName, const boost::any aValue) ;

private:
	DeviceDriver * m_driver;
};

template<typename DriverType>
DriverAdapter<DriverType>::DriverAdapter(const std::string& name, const std::string& type, DeviceDriver * driver) :
	DriverType(name), m_driver(driver) {

	this->DeviceDriver::type = type;
}

template<typename DriverType>
void DriverAdapter<DriverType>::execute(const std::string& command) {
	// forward to real driver
	if (m_driver != 0) {
		m_driver->execute(command);
	}
}

template<typename DriverType>
void DriverAdapter<DriverType>::refreshSetValue(std::string propertyName, const boost::any aValue)  {
	// forward to real driver
	if (m_driver != 0) {
		m_driver->refreshSetValue(propertyName, aValue);
	}
}

template<typename DriverType>
class DriverPtr : public BasePtr {

	template<typename>
	friend class DriverPtr;

public:
	DriverPtr();
	DriverPtr(ExperimentController * pointerOwner, const std::string & role);
	DriverPtr(const DriverPtr<DriverType> & ptr);
	~DriverPtr();

	void init(ExperimentController * pointerOwner, const std::string & role);
	bool isInited();

	template<typename DriverFromType>
	void copy(const DriverPtr<DriverFromType> & ptr) {
		m_driver = dynamic_cast<DriverType *>(ptr.m_driver);

		if (m_driver == 0) {
			throw BadConversion(DriverFromType::TYPE, DriverType::TYPE);
		}

		m_abstractController = ptr.m_abstractController;
		m_ownerName = ptr.m_ownerName;
		m_role = ptr.m_role;
		m_driverAdapter = ptr.m_driverAdapter;
	}

	AbstractController ** getControllerPointer();

	virtual std::string getType() const;
	virtual bool assign(AbstractController * controller);
	virtual bool isAssigned() const;

	DriverType * operator->() const;
	DriverType * get() const;
	const std::string& getRole() const;

	template<typename TargetDriverType>
	TargetDriverType * cast() const {
		TargetDriverType * target = dynamic_cast<TargetDriverType *>(m_driver);
		if (target == 0) {
			throw BadConversion(DriverType::TYPE, TargetDriverType::TYPE);
		}

		return target;
	}

	void execute(const std::string& command, bool blocking, bool observerIfNotBlocking = false);
	void executeParallel(const std::string& command);
	void setTimeout(const std::string & command, int32 sec);
	void setEnabled(bool enabled, bool all, bool postConfiguration = true);

private:
	DriverType * m_driver;
	AbstractController * m_abstractController;
	std::string m_ownerName;
	std::string m_role;
	bool m_driverAdapter;
};

template<typename DriverType>
DriverPtr<DriverType>::DriverPtr() :
	m_driver(0),
	m_abstractController(0),
	m_driverAdapter(false) {
}

template<typename DriverType>
DriverPtr<DriverType>::DriverPtr(ExperimentController * pointerOwner, const std::string & role) :
	m_driver(0),
	m_abstractController(0),
	m_driverAdapter(false) {
	init(pointerOwner, role);
}

template<typename DriverType>
DriverPtr<DriverType>::DriverPtr(const DriverPtr<DriverType> & ptr) :
	m_driver(ptr.m_driver),
	m_abstractController(ptr.m_abstractController),
	m_ownerName(ptr.m_ownerName),
	m_role(ptr.m_role),
	m_driverAdapter(ptr.m_driverAdapter) {
}

template<typename DriverType>
DriverPtr<DriverType>::~DriverPtr() {
//	// delete the driver in case it was generated
//  Remove it, do seg fault if copy a driverptr
//	if (m_driverAdapter) {
//		delete m_driver;
//	}
}

template<typename DriverType>
std::string DriverPtr<DriverType>::getType() const {
	return DriverType::TYPE;
}

template<typename DriverType>
AbstractController ** DriverPtr<DriverType>::getControllerPointer() {
	return &m_abstractController;
}

template<typename DriverType>
bool DriverPtr<DriverType>::assign(AbstractController * controller) {

	TestController * deviceController = dynamic_cast<TestController *>(controller);

	if (deviceController == 0) {
		return false;
	}

	DeviceDriver * deviceDriver = deviceController->getDeviceDriver();

	// test compatibility of driver
	DriverType * typedDriver = dynamic_cast<DriverType *>(deviceDriver);

	if (typedDriver != 0) {
		m_driver = typedDriver;
		m_abstractController = controller;
		return true;

	} else {
		// create the bridge, first get the type value otherwise it will be initialised by the adapter
		std::string type = deviceDriver->type();
		m_driver = new DriverAdapter<DriverType>(deviceDriver->getName(), type, deviceDriver);
		m_driverAdapter = true;
		m_abstractController = controller;

		common::warn << "Created a driver adapter for " << m_role << " of " << m_ownerName << common::eol;

		return true;
	}

	return false;
}

template<typename DriverType>
bool DriverPtr<DriverType>::isAssigned() const {
	return (m_driver != 0);
}

template<typename DriverType>
DriverType * DriverPtr<DriverType>::operator->() const {
	return m_driver;
}

template<typename DriverType>
DriverType * DriverPtr<DriverType>::get() const {
	return m_driver;
}

template<typename DriverType>
const std::string& DriverPtr<DriverType>::getRole() const{
	return m_role;
}

template<typename DriverType>
void DriverPtr<DriverType>::execute(const std::string& command, bool blocking, bool observerIfNotBlocking) {
	dynamic_cast<TestController *>(m_abstractController)->execute(command, blocking, false, observerIfNotBlocking);
}

template<typename DriverType>
void DriverPtr<DriverType>::executeParallel(const std::string& command) {
	Sender<ClientCommand>::post(ClientCommand(dynamic_cast<TestController *>(m_abstractController), command));
}

template<typename DriverType>
void DriverPtr<DriverType>::setTimeout(const std::string & command, int32 sec) {
	dynamic_cast<TestController *>(m_abstractController)->setTimeOut(command, sec);
}

template<typename DriverType>
void DriverPtr<DriverType>::setEnabled(bool enabled, bool all, bool postConfiguration) {
	if (enabled) {
		m_abstractController->setEnabled(all, postConfiguration);
	} else {
		m_abstractController->setDisabled(all);
	}
}

template<typename DriverType>
bool operator!=(const DriverPtr<DriverType>& driver1, const DriverPtr<DriverType>& driver2) {
	return (driver1.get()->getName() != driver2.get()->getName());
}

template<typename DriverType>
bool operator==(const DriverPtr<DriverType>& driver1, const DriverPtr<DriverType>& driver2) {
	return (driver1.get()->getName() == driver2.get()->getName());
}


#endif
