/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "InstrumentAbstraction/ExperimentController.h"
#include "DataProvider/Property/AnyProperty.h"
#include "DataProvider/Property/PropertyManager.h"
#include "DataProvider/CommandManager.h"
#include "DataProvider/ServantManager.h"
#include "DataProvider/ChangeAspect.h"
#include "DataProvider/ChangeAspectFactory.h"
#include "ErrorsAlarmsManager/ErrorCatalogManager.h"
#include "ErrorsAlarmsManager/Error.h"
#include "InstrumentManager/InstrumentManager.h"
#include "Module/Module.h"
#include "common/base/ServerProperties.h"
#include <sstream>
#include "dataprovider/database/Database.h"
#include <core/dataaccess/DataAccess.h>
#include "CommandNames.h"

using namespace std;
using namespace boost;

const string ExperimentController::TYPE = "ExperimentController";
const string ExperimentController::LIMIT_STR = "limit";

ExperimentController::FunctionMismatch::FunctionMismatch(const string &role) :
	runtime_error(role) {
}

ExperimentController::UnknownController::UnknownController(const std::string& name) :
	runtime_error(string("unknown controller ") + name) {
}

ExperimentController::ExperimentController(const string& type, const string& name) :
	AbstractController(type, name, EXPERIMENT_CONTROLLER_TYPE) {
	constructor(type, name);
}

ExperimentController::ExperimentController(const string& name) :
	AbstractController("", name, EXPERIMENT_CONTROLLER_TYPE) {
	constructor("", name);
}

ExperimentController::ExperimentController() :
	AbstractController(),
	m_NRequiredControllers(0),
	m_Configured(false),
	m_saved(false),
	m_inited(false),
	m_module(nullptr),
	m_CloneController(true) {
}

void ExperimentController::constructor(const string& type, const string& name) {

	// Init the controller only once.
	// This is necessary in case of virtual inheritance.
	if (m_inited) {
		return;
	}

	m_NRequiredControllers = 0;
	m_Configured = false;
	m_saved = true;
	m_module = 0;
	isCommand.init(this, SAVE, "isCommand");
	nickname.init(this, SAVE, "nickname");
	commandProgression.init(this, NOSAVE | SPY, "progression");
	commandStatus.init(this, NOSAVE | SPY, "commandStatus");
	driverDisconnected.init(this, NOSAVE | SPY, "driver_disconnected");
	isCommand.update(false);
	m_CloneController = false;
	m_inited = true;

	registerFunction(EXPERIMENT_CONTROLLER);

	registerRefresher(commandProgression, &ExperimentController::refreshProgression, this);
}

void ExperimentController::init(const string& name) {
	AbstractController::constructor("", name, "");
	constructor("", name);
}

ExperimentController::ExperimentController(const ExperimentController & controller) :
	AbstractController(controller) {
	m_RequiredControllers = controller.m_RequiredControllers;
	m_RequiredType = controller.m_RequiredType;
	m_RequiredFunctions = controller.m_RequiredFunctions;
	m_NRequiredControllers = controller.m_NRequiredControllers;
	m_ChangeValues = controller.m_ChangeValues;
	m_Configured = controller.m_Configured;
	m_saved = controller.m_saved;
	m_module = controller.m_module;
	m_CloneController = true;

	isCommand.copy(this, controller.isCommand);
	nickname.copy(this, controller.nickname);
	commandStatus.copy(this, controller.commandStatus);
	commandProgression.copy(this, controller.commandProgression);
	driverDisconnected.copy(this, controller.driverDisconnected);

	// do not copy
	// - m_servantPointers
	// - m_servantFunctions
}

void ExperimentController::copy(const ExperimentController & controller) {
	AbstractController::copy(controller);
	m_RequiredControllers = controller.m_RequiredControllers;
	m_RequiredType = controller.m_RequiredType;
	m_RequiredFunctions = controller.m_RequiredFunctions;
	m_NRequiredControllers = controller.m_NRequiredControllers;
	m_ChangeValues = controller.m_ChangeValues;
	m_Configured = controller.m_Configured;
	m_saved = controller.m_saved;
	m_module = controller.m_module;
	m_CloneController = true;

	isCommand.copy(this, controller.isCommand);
	nickname.copy(this, controller.nickname);
	commandStatus.copy(this, controller.commandStatus);
	commandProgression.copy(this, controller.commandProgression);
	driverDisconnected.copy(this, controller.driverDisconnected);

	// do not copy
	// - m_servantPointers
	// - m_servantFunctions
}

ExperimentController::~ExperimentController() {
}

ExperimentController* ExperimentController::clone(bool all) const {
	if (m_module != 0) {
		ExperimentController* clone = m_module->copyController(this);
		clone->postConfiguration();

		return clone;
	}
	return 0;
}

void ExperimentController::setModule(Module * module) {
	m_module = module;
}

Module * ExperimentController::getModule() const {
	return m_module;
}

std::string ExperimentController::getHomePath() const {
	return common::ServerProperties::getInstance()->getHomePath();
}

std::string ExperimentController::getConfigPath() const {
	return common::ServerProperties::getInstance()->getConfigPath();
}

std::string ExperimentController::getServantInfo() const {
	return "Controller " + getName() + " (" + getType() + ")";
}

string ExperimentController::getConfiguration() {

	string configString = "";

	map<string, AbstractController**>::const_iterator controllerIterator;

	/*
	 * Iterate through the map of required Controllers.
	 * Return the role, function and name if cconfigured
	 */
	if (m_NRequiredControllers > 0) {
		controllerIterator = m_RequiredControllers.begin();
		while (controllerIterator != m_RequiredControllers.end()) {

			string role = controllerIterator->first;
			string function = m_RequiredFunctions[role];
			string name = "";
			/*
			 * Get the configured AC (if it exists)
			 */
			AbstractController* currentAC = *(controllerIterator->second);

			/*
			 * Get the name of the controller if configured
			 */
			if (currentAC == NULL) {
				name = "not configured";
			} else {
				//	configString.append(currentAC->getName());
				name = currentAC->getName();
			}

			/*
			 * Check to see if the family is 'hidden', if so, don't display it
			 */
			bool addToList = true;
			if (currentAC != NULL) {
				if (currentAC->getFamily() == family::HIDDEN) {
					addToList = false;
				}
			}
			if (addToList) {
				configString.append(role + "\t" + function + "\t" + name + "\n");
			}

			controllerIterator++;
		}
	}
	return configString;

}

string ExperimentController::getConfigurationWithHidden() {

	string configStrgwithHidden = "";

	map<string, AbstractController**>::const_iterator controllerIterator;

	/*
	 * Iterate through the map of required Controllers.
	 * Return the role, function and name if cconfigured
	 */
	if (m_NRequiredControllers > 0) {
		controllerIterator = m_RequiredControllers.begin();
		while (controllerIterator != m_RequiredControllers.end()) {

			string role = controllerIterator->first;
			string function = m_RequiredFunctions[role];
			string name = "";
			/*
			 * Get the configured AC (if it exists)
			 */
			AbstractController* currentAC = *(controllerIterator->second);

			/*
			 * Get the name of the controller if configured
			 */
			if (currentAC == NULL) {
				name = "not configured";
			} else {
				//	configString.append(currentAC->getName());
				name = currentAC->getName();
			}

			configStrgwithHidden.append(role + "\t" + function + "\t" + name + "\n");

			controllerIterator++;
		}
	}
	// cout << configStrgwithHidden << endl;
	return configStrgwithHidden;
}

map<string, pair<string, string>> ExperimentController::getConfigurationRoles() {

	map<string, AbstractController**>::const_iterator controllerIterator;
	map<string, pair<string, string>> roles;

	/*
	 * Iterate through the map of required Controllers.
	 * Return the role, function and name if cconfigured
	 */
	if (m_NRequiredControllers > 0) {
		controllerIterator = m_RequiredControllers.begin();
		while (controllerIterator != m_RequiredControllers.end()) {

			string role = controllerIterator->first;
			string name = "";
			/*
			 * Get the configured AC (if it exists)
			 */
			AbstractController* currentAC = *(controllerIterator->second);

			/*
			 * Get the name of the controller if configured
			 */
			if (currentAC == NULL) {
				name = "not configured";
			} else {
				name = currentAC->getName();
			}

			roles[role] = make_pair(name, m_RequiredType[role]);
			controllerIterator++;
		}
	}

	return roles;
}

void ExperimentController::configure(const string& role, AbstractController* ac)  {

	// Check to ensure that the AbstractController is valid
	if (ac != 0) {

		if (m_servantPointers.find(role) != m_servantPointers.end()) {
			// Check to see if a controller already exists: remove it if there is one
			if (*m_RequiredControllers[role] != 0) {
				AbstractController* deadController = *m_RequiredControllers[role];
				deadController->detach(this);
			}

			bool assigned = m_servantPointers[role]->assign(ac);

			if (!assigned) {
				throw FunctionMismatch(m_servantPointers[role]->getType());
			}

			try {
				ExperimentController* controller = dynamic_cast<ExperimentController*>(ac);
				if (controller != NULL) {
					registerUpdater(controller->driverDisconnected, &ExperimentController::updateDisconnected, this);
				}
			}
			catch(...) {

			}
			try {
				TestController* controller = dynamic_cast<TestController*>(ac);
				if (controller != NULL) {
					registerUpdater(controller->getDeviceDriver()->disconnected, &ExperimentController::updateDisconnected, this);
				}
			}
			catch(...) {

			}
			return;
		}

		// Check to ensure that the AbstractController is correct
		if (ac->containsFunction(m_RequiredFunctions[role])) {
			// Check to see if a controller already exists: remove it if there is one
			if (*m_RequiredControllers[role] != 0) {
				AbstractController* deadController = *m_RequiredControllers[role];
				deadController->detach(this);
			}

			*m_RequiredControllers[role] = ac;

			try {
				ExperimentController* controller = dynamic_cast<ExperimentController*>(ac);
				if (controller != NULL) {
					registerUpdater(controller->driverDisconnected, &ExperimentController::updateDisconnected, this);
				}
			}
			catch(...) {

			}
			try {
				TestController* controller = dynamic_cast<TestController*>(ac);
				if (controller != NULL) {
					registerUpdater(controller->getDeviceDriver()->disconnected, &ExperimentController::updateDisconnected, this);
				}
			}
			catch(...) {

			}
		} else {
			throw FunctionMismatch(m_RequiredFunctions[role]);
		}

	} else
		cerr << "ExperimentController:: Got a NULL AbstractController\n";

}

void ExperimentController::unConfigure(const AbstractController* const deadController) {

	map<string, AbstractController**>::iterator acIterator;
	AbstractController* currentController;
	string role;
	string groupName;

	/*
	 * Iterate through the installed AbstractControllerss checking for this one.
	 * If its there, remove it and destroy links to its propertyController
	 * also.
	 */
	for (acIterator = m_RequiredControllers.begin(); acIterator != m_RequiredControllers.end(); acIterator++) {
		/*
		 * get pointer to AbstractController which is in the second pos of map
		 * and check to see if it matches the dead one
		 */
		role = acIterator->first;
		currentController = *(acIterator->second);
		if (currentController == deadController) {

			*m_RequiredControllers[role] = NULL;

			/*
			 * Unsubscribe from controller
			 */
			currentController->detach(this);

		}
	}

}

//void ExperimentController::registerConfiguration(const string& role, const string& function, AbstractController** ac) {
//
//	/*
//	 * Ensure that the registered AC is initialised to NULL
//	 */
//	*ac = NULL;
//
//	m_RequiredFunctions[role] = function;
//	m_RequiredControllers[role] = ac;
//	m_NRequiredControllers++;
//
//}
//
//void ExperimentController::unregisterConfiguration(const string& role) {
//
//	if (m_RequiredFunctions.find(role) != m_RequiredFunctions.end()) {
//		m_RequiredFunctions.erase(role);
//	}
//
//	if (m_RequiredControllers.find(role) != m_RequiredControllers.end()) {
//		m_RequiredControllers.erase(role);
//	}
//
//	m_NRequiredControllers--;
//}

bool ExperimentController::isConfigured() {

	map<string, AbstractController**>::const_iterator acIterator;
	bool isOK = true;

	/*
	 * Iterate through the map of required controllers.
	 * Check to see if all DMs are configured (ie, non-NULL pointers)
	 */
	if (m_NRequiredControllers > 0) {
		for (acIterator = m_RequiredControllers.begin(); acIterator != m_RequiredControllers.end(); acIterator++) {

			string role = acIterator->first;
			/*
			 * Check to see if a AC has been fitted to role
			 */
			AbstractController* currentAC = *(acIterator->second);
			if (currentAC == NULL) {
				isOK &= false;
			} else {
				/*
				 * Then check to see if the AC is actually configured itself
				 */
				isOK &= currentAC->isConfigured();
			}
		}
	}
	return isOK;

}

void ExperimentController::activate() {

	map<string, AbstractController**>::const_iterator acIterator;

	/*
	 * Forward the activation call to all required Abstract Controllers
	 */
	for (acIterator = m_RequiredControllers.begin(); acIterator != m_RequiredControllers.end(); acIterator++) {

		AbstractController* currentAC = *(acIterator->second);

		/*
		 * Verify that the AC is not null (it shouldn't be because checks
		 * should have been prior to arriving here that all ACs are
		 * configured... but just in case.
		 */
		if (currentAC != NULL) {
			currentAC->setActive();
		}

	}
}

void ExperimentController::deactivate() {

	map<string, AbstractController**>::const_iterator acIterator;

	/*
	 * Forward the deactivation call to all required Abstract Controllers
	 */
	for (acIterator = m_RequiredControllers.begin(); acIterator != m_RequiredControllers.end(); acIterator++) {

		AbstractController* currentAC = *(acIterator->second);

		/*
		 * Verify that the AC is not null (it shouldn't be because checks
		 * should have been prior to arriving here that all ACs are
		 * configured... but just in case.
		 */
		if (currentAC != NULL) {
			currentAC->unsetActive();
		}

	}
}

AbstractController*
ExperimentController::findControllerWithName(const string& name) {

	map<string, AbstractController**>::const_iterator controllerIterator;
	AbstractController* theController = NULL;
	bool found = false;

	if (m_NRequiredControllers > 0) {
		controllerIterator = m_RequiredControllers.begin();
		while (controllerIterator != m_RequiredControllers.end() && !found) {

			AbstractController* currentAC = *(controllerIterator->second);
			if (currentAC != NULL) {
				if (currentAC->getName() == name) {
					found = true;
					theController = currentAC;
				}
			}
			controllerIterator++;
		}
	}

	return theController;

}

string ExperimentController::getControllerNameWithRole(const string& role)  {

	/*
	 * Find the controller with given role... if role exists
	 */
	map<string, AbstractController**>::const_iterator controllerIterator;
	bool found = false;
	string name = "";

	if (m_NRequiredControllers > 0) {
		controllerIterator = m_RequiredControllers.begin();
		while (controllerIterator != m_RequiredControllers.end() && !found) {

			/*
			 * Compare the registered roles with the passed paramter
			 */
			if (controllerIterator->first == role) {
				found = true;
				AbstractController* controller = *(controllerIterator->second);

				/*
				 * Verify that the controller is configured
				 */
				if (controller) {
					name = controller->getName();
				} else {
					throw ControllerNotConfigured(m_type, m_name, this, role);
				}
			} else {
				controllerIterator++;
			}

		}
	}

	if (found) {
		return name;
	} else {
		throw RoleDoesNotExist();
	}

}

void ExperimentController::sendArrayEvent(const string& propertyName) {

	string propertyType = getPropertyType(propertyName);

	if ((propertyType == UINT16ARRAY_PROPERTY) || (propertyType == UINT32ARRAY_PROPERTY)
			|| (propertyType == UINT64ARRAY_PROPERTY) || (propertyType == INT16ARRAY_PROPERTY)
			|| (propertyType == INT32ARRAY_PROPERTY) || (propertyType == INT64ARRAY_PROPERTY)
			|| (propertyType == FLOAT32ARRAY_PROPERTY)|| (propertyType == FLOAT64ARRAY_PROPERTY)) {
		ChangeAspectFactory aCAFactory;

		ChangeAspect* aCA = aCAFactory. createPropertyChangeAspect(m_type, m_name, propertyName);
		notify(aCA);
	}

	// sending notification
	getPropertyManager()->sendArrayEvent(propertyName);
}

void ExperimentController::sendProgressEvent(int32  progress) {

	ChangeAspectFactory aCAFactory;

	ChangeAspect* progressChangeAspect = aCAFactory.createProgressChangeAspect(m_type, m_name, progress);

	if (getPropertyManager()->containsProperty(PROGRESSION_PROPERTY)) {
		bool hasChanged = (progress != any_cast<int32> (getValue(PROGRESSION_PROPERTY)));
		if (hasChanged == true) {
			updateValue(PROGRESSION_PROPERTY, progress);
			notify(progressChangeAspect);

			// Sending for nomad2
			getCommandManager()->setProgression(progress);
		}
	} else {
		notify(progressChangeAspect);

		// Sending for nomad2
		getCommandManager()->setProgression(progress);
	}
}

void ExperimentController::sendErrorEvent(int32  errorId, const string& errorMessage) {

	ChangeAspectFactory changeAspectFactory;
	ChangeAspect * errorChangeAspect = changeAspectFactory.createErrorChangeAspect(m_type, m_name, errorId);

	//  string message = ErrorCatalogManager::instance()->
	//    getAddError(errorChangeAspect->getLongParam())->getMasterMessage();

	//  message.append("-> ");
	//  message.append(errorMessage);

	errorChangeAspect->setStringParam(errorMessage);
	log(Level::s_Error) << errorMessage << endlog;
	notify(errorChangeAspect);

}

/*
 * setEnable
 */
//void ExperimentController::setEnabled(bool all,bool postconf) {
//	map<string, AbstractController**>::const_iterator ecIterator;
//	if (all) {
//		for (ecIterator = m_RequiredControllers.begin(); ecIterator != m_RequiredControllers.end(); ecIterator++) {
//			AbstractController* controller = *(ecIterator->second);
//			if (controller != NULL) {
//				if (find(m_ExcludeSetEnableController.begin(), m_ExcludeSetEnableController.end(), ecIterator->first)
//						== m_ExcludeSetEnableController.end()) {
//					controller->setEnabled(true, false);
//				}
//			}
//		}
//	}
//	m_ExcludeSetEnableController.clear();
//	for (ecIterator = m_RequiredControllers.begin(); ecIterator != m_RequiredControllers.end(); ecIterator++) {
//		AbstractController* controller = *(ecIterator->second);
//		if (controller != NULL) {
//			controller->attach(this);
//		}
//	}
//
//	// Init properties (set value) because now controller is in used
//	initProperties();
//
//	setConfigured(true);
//	if (postconf) postConfiguration();
//
//	//	// Init properties (set value) because now controller is in used
//	//	initProperties();
//
//	execute(INIT_COMMAND, false, true);
//	enable(true);
//
//	updateDisconnected();
//}

void ExperimentController::setEnabledForLoading(bool all, bool postconf) {

	map<string, AbstractController**>::const_iterator ecIterator;
	if (all) {
		for (ecIterator = m_RequiredControllers.begin(); ecIterator != m_RequiredControllers.end(); ecIterator++) {
			AbstractController* controller = *(ecIterator->second);
			if (controller != NULL) {
				if (find(m_ExcludeSetEnableController.begin(), m_ExcludeSetEnableController.end(), ecIterator->first) == m_ExcludeSetEnableController.end()) {
					controller->setEnabledForLoading(true, true);
				}
			}
		}
	}
	m_ExcludeSetEnableController.clear();
	for (ecIterator = m_RequiredControllers.begin(); ecIterator != m_RequiredControllers.end(); ecIterator++) {
		AbstractController* controller = *(ecIterator->second);
		if (controller != NULL) {
			controller->attach(this);
		}
	}

	// Init properties (set value) because now controller is in used
	initProperties();

	setConfigured(true);
	if (postconf) {
		postConfiguration();
	}

	// Set the controller enabled.
	enable(true);

	updateDisconnected();
}

/*
 * setEnable
 */
void ExperimentController::setEnabled(bool all, bool postconf) {

	setEnabledForLoading(all, postconf);

	// The init command is called when the controller is not enabled from the configuration file.
	execute(COMMAND_INIT, false, true);
}


/*
 * setDisabled
 */
void ExperimentController::setDisabled(bool all) {
	if (all) {
		map<string, AbstractController**>::const_iterator ecIterator;
		for (ecIterator = m_RequiredControllers.begin(); ecIterator != m_RequiredControllers.end(); ecIterator++) {
			AbstractController* controller = *(ecIterator->second);
			if (controller != NULL) {
				if (find(m_ExcludeSetEnableController.begin(), m_ExcludeSetEnableController.end(), ecIterator->first) == m_ExcludeSetEnableController.end()) {
					controller->setDisabled(true);
				}
			}
		}
	}
	m_ExcludeSetEnableController.clear();
	map<string, AbstractController**>::const_iterator ecIterator;
	for (ecIterator = m_RequiredControllers.begin(); ecIterator != m_RequiredControllers.end(); ecIterator++) {
		AbstractController* controller = *(ecIterator->second);
		if (controller != NULL) {
			controller->detach(this);
		}
	}
	setConfigured(false);
	enable(false);
}

void ExperimentController::setVisible(bool value) {

	// Set visible.
	getServantManager()->setVisible(value);

	// Set command to appear in the launch pad.
	getServantManager()->setCommand(value);
	setValue(IS_COMMAND_PROPERTY, value);
}

void ExperimentController::setSaved(bool value) {
	m_saved = value;
}

bool ExperimentController::isSaved() const {
	return m_saved;
}

ExperimentController * ExperimentController::getController(std::string const & name)  {
	// direct call to InstrumentManager
	if (InstrumentManager::getInstance()->controllerExists(name)) {
		return InstrumentManager::getInstance()->getExperimentController(name);
	}

	throw UnknownController(name);
}

std::string ExperimentController::concat(const std::string & name, int32 i) {
	ostringstream os;
	os << name + "." << i;

	return os.str();
}

std::string ExperimentController::concat(const std::string & name, int32 i, int32 j) {
	ostringstream os;
	os << name << "." << i << "." << j;

	return os.str();
}

const map<string, AbstractController**> & ExperimentController::getRequiredControllers() const {
	return m_RequiredControllers;
}

void ExperimentController::setConfigured(bool aValue) {
	m_Configured = aValue;
}

const map<string, string> & ExperimentController::getRequiredFunctions() const {
	return m_RequiredFunctions;
}

AbstractController* ExperimentController::getControllerWithRole(const string& role) const  {
	AbstractController* controller = NULL;
	map<string, AbstractController**>::const_iterator controllerIterator;
	controllerIterator = m_RequiredControllers.find(role);
	if (controllerIterator != m_RequiredControllers.end()) {
		controller = *(controllerIterator->second);
		if (controller == NULL) {
			throw ControllerNotConfigured();
		}
	}
	else {
		throw RoleDoesNotExist();
	}
	return controller;
}

void ExperimentController::unregisterController(const string& role) {
	
	if (m_servantPointers.find(role) != m_servantPointers.end()) {
		m_servantPointers.erase(role);
	}

	if (m_RequiredFunctions.find(role) != m_RequiredFunctions.end()) {
		m_RequiredFunctions.erase(role);
	}

	if (m_RequiredControllers.find(role) != m_RequiredControllers.end()) {
		m_RequiredControllers.erase(role);
	}

	m_NRequiredControllers--;
}

void ExperimentController::addCommandFunction(const string& name, controller::Command * function) {
	m_servantFunctions[name] = function;
}

bool ExperimentController::isLogging() const {

	if (m_logging.get() == 0) {
		return true;
	}

	return *m_logging.get();
}

void ExperimentController::refreshProgression(int32 value) {
	// only send the progress
	sendProgressEvent(value);
}

void ExperimentController::execute(const std::string& command, bool blocking, bool logging) {

	m_logging.reset(new bool(logging));

	map<string, controller::Command *>::iterator f = m_servantFunctions.find(command);
	if (f != m_servantFunctions.end()) {
		f->second->execute();
	}

	// can generate memory leak if command throws
	delete m_logging.release();
}

void ExperimentController::addExcludeSetEnableController(const std::string& name) {
	m_ExcludeSetEnableController.push_back(name);
}

void ExperimentController::updateDisconnected() {
	bool disconnected = false;
	map<string, AbstractController**>::const_iterator ecIterator;
	for (ecIterator = m_RequiredControllers.begin(); ecIterator != m_RequiredControllers.end(); ecIterator++) {
		try {
			ExperimentController* controller = dynamic_cast<ExperimentController*>(*(ecIterator->second));
			if (controller != NULL) {
				if (controller->driverDisconnected() == true) {
					disconnected = true;
					break;
				}
			}
		}
		catch (...) {

		}
		try {
			TestController* controller = dynamic_cast<TestController*>(*(ecIterator->second));
			if (controller != NULL) {
				if (controller->getDeviceDriver()->disconnected() == true) {
					disconnected = true;
					break;
				}
			}
		}
		catch (...) {

		}
	}
	driverDisconnected.update(disconnected);
}

void ExperimentController::fillScanList(std::list<std::pair<std::string, std::string> > & list) {
	using namespace dataprovider;
	vector<string> properties = m_PropertyManager->getLocalProperties();
	vector<string>::iterator it = properties.begin();
	uint32 i = 0;
	while (it != properties.end()) {
		if (Database::getInstance(0)->isScannableProperty(m_type, *it) == true) {
			if (find(list.begin(), list.end(), make_pair(m_name, *it)) ==list.end()) {
				list.push_back(make_pair(m_name, *it));
			}
		}
		++it;
	}
//	// File this scan property
//	uint32 id = Database::getInstance(0)->getServantID(m_Name);
//	Database::PropertyMap propertyMap = Database::getInstance(0)->getControllerProperties(id);
//	for (Database::PropertyMap::const_iterator it = propertyMap.begin(); it != propertyMap.end(); it++) {
//		if (Database::getInstance(0)->isScannableProperty(m_Type, it->first) == true) {
//			if (find(list.begin(), list.end(), make_pair(m_Name, it->first)) ==list.end()) {
//				list.push_back(make_pair(m_Name, it->first));
//			}
//		}
//	}
	// Fill role property
	map<string, AbstractController**>::const_iterator ecIterator;
	for (ecIterator = m_RequiredControllers.begin(); ecIterator != m_RequiredControllers.end(); ecIterator++) {
		AbstractController* controller = *(ecIterator->second);
		if (controller != NULL) {
			if (controller->isEnabled() == true) {
				controller->fillScanList(list);
			}
		}
	}
}

void ExperimentController::setDebug() {
	m_PropertyManager->setServantInDebug();
}

ExperimentController* ExperimentController::cloneController(const std::string& name) {
	unsigned long servantCloneId = ::core::ServantCloner::getInstance()->cloneOrGetController(::core::DataAccess::getInstance()->getServantID(name));
	string clonename = ::core::DataAccess::getInstance()->getServantName(0, servantCloneId);
	return InstrumentManager::getInstance()->getExperimentController(clonename);
}

void ExperimentController::removeCloneController(ExperimentController* ec) {
	if (m_CloneController) {
		if (::core::ServantCloner::getInstance()->removeCloneController(::core::DataAccess::getInstance()->getServantID(ec->getName()))) {
			InstrumentManager::getInstance()->removeController(ec->getName());
		}
	}
}

