/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*! File ExperimentController.cpp
 *\brief  main class to control all the Experiment Controller
 *\version 1.0
 *\Date Jun 2010
 *
 */
#ifndef EXPERIMENTCONTROLLER_H
#define EXPERIMENTCONTROLLER_H

#include <map>
#include <vector>
#include <stdexcept>
//#include <boost/function.hpp>
//#include <boost/bind.hpp>
//#include <boost/thread/tss.hpp>

#include "InstrumentAbstraction/AbstractController.h"
#include "InstrumentAbstraction/ControllerPtr.h"
#include "InstrumentAbstraction/DriverPtr.h"
#include "InstrumentAbstraction/Commands/Command.h"
#include "DataProvider/Property/TypedProperty/ValueProperty.h"
#include "DataProvider/Property/TypedProperty/ValueDynamicProperty.h"
#include "DataProvider/Property/TypedProperty/ValueDynamicProperty2.h"
#include "DataProvider/Property/TypedProperty/StatusProperty.h"
#include <core/cloner/ServantCloner.h>

class Module;

class ExperimentController: public AbstractController {

	template<typename >
	friend class ControllerPtr;

	template<typename >
	friend class DriverPtr;

	template<typename, typename >
	friend struct ServantProperty;

	template<typename, typename >
	friend struct ServantDynamicProperty;

	template<typename >
	friend class ControllerFactory;

	friend class InstrumentManager;

public:
	/**
	 * TYPE member used for adding properties with from type.
	 */
	static const std::string TYPE;

	/*
	 * temporary exceptions: should use Error manager
	 */
	class FunctionMismatch: public std::runtime_error {
	public:
		FunctionMismatch(const std::string &role);
	};
	class RoleDoesNotExist {
	};

	/**
	 * Exception for the getController method.
	 */
	class UnknownController: std::runtime_error {

	public:
		/**
		 * Constructor.
		 */
		UnknownController(const std::string& name);
	};

	ExperimentController(const ExperimentController & controller);
	ExperimentController();
	virtual ~ExperimentController();

	void init(const std::string& name);
	void copy(const ExperimentController & controller);

	void setModule(Module * module);
	Module * getModule() const;

	virtual ExperimentController* clone(bool all = false) const;

	virtual std::string getConfiguration();
	virtual std::string getConfigurationWithHidden();
	virtual std::map<std::string, std::pair<std::string, std::string>> getConfigurationRoles();

	std::string getHomePath() const;
	std::string getConfigPath() const;

	/**
	 * Gets the servant info.
	 */
	virtual std::string getServantInfo() const;

	virtual void configure(const std::string& role, AbstractController* ac) ;

	virtual void unConfigure(const AbstractController* const ac);

	/*
	 * Method from ExperimentDirector (mediator pattern). Implement a
	 * default 'do nothing' method rather than having virtual pure (for
	 * those experiment Controllers who aren't directors).
	 */
	virtual void experimentComponentChanged(ExperimentController* aComponent) {
	}

	virtual std::string getControllerNameWithRole(const std::string& role) ;

	/**
	 * Return the list of required controllers.
	 */
	const std::map<std::string, AbstractController**> & getRequiredControllers() const;

	/**
	 * Return the list of required functions (role and type)
	 */
	const std::map<std::string, std::string> & getRequiredFunctions() const;

	/**
	 * Set enable the controller
	 */
	virtual void setEnabled(bool all, bool postconf = true);

	/**
	 * Set disable the controller
	 */
	virtual void setDisabled(bool all = false);

	/**
	 * Set enable the controller
	 */
	void addExcludeSetEnableController(const std::string& name);

	/**
	 * Sets the controller visible or not.
	 */
	void setVisible(bool value);

	/**
	 * Sets the controller saved or not in the configuration file.
	 */
	void setSaved(bool value);

	/**
	 * Returns true if is saved in the configuration file.
	 */
	bool isSaved() const;

	/**
	 * Returns the required controller whose name is name
	 */
	ExperimentController * getController(std::string const & name) ;

	/*
	 * Returns the string concatenation of "name.i"
	 */
	static std::string concat(const std::string & name, int32 i);

	/*
	 * Returns the string concatenation of "name.i.j"
	 */
	static std::string concat(const std::string & name, int32 i, int32 j);

	/*!
	 *\brief postConfiguration
	 * Call after server restart when all controller are initialized
	 *
	 */
	virtual void postConfiguration() {
	}

	virtual bool isConfigured();

	AbstractController* getControllerWithRole(const std::string& role) const ;

	void addCommandFunction(const std::string& name, controller::Command * function);

	virtual void execute(const std::string& command, bool blocking, bool logging = false);

	virtual void fillScanList(std::list<std::pair<std::string, std::string> > & list);

	template<typename Type> void refreshPropertyisStriclyPositive(const std::string& name, Type value);
	template<typename Type> void refreshPropertyisPositive(const std::string& name, Type value) ;

	ExperimentController* cloneController(const std::string& name);
	void removeCloneController(ExperimentController* ec);

protected:

	ExperimentController(const std::string& controllerType, const std::string& controllerName);
	ExperimentController(const std::string& controllerName);

	virtual void setEnabledForLoading(bool all, bool postconf);

	virtual void activate();
	virtual void deactivate();

//	void registerConfiguration(const std::string& role, const std::string& function, AbstractController** ac);
//	void unregisterConfiguration(const std::string& role);

	template<typename ControllerType>
	void registerController(const std::string& role, ControllerPtr<ControllerType>& controllerPointer);
	void unregisterController(const std::string& role);

	template<typename DriverType>
	void registerDriver(const std::string& role, DriverPtr<DriverType>& driverPointer);

	void unregisterUpdaters(const std::string& servantName);

	template<typename ControllerType>
	void unregisterUpdaters(ControllerPtr<ControllerType>& controllerPointer);

	template<typename DriverType>
	void unregisterUpdaters(DriverPtr<DriverType>& driverPointer);

	template<typename Type>
	void unregisterUpdater(ValueProperty<Type>& property);

	template<typename Type, typename Function, typename Object>
	void registerUpdater(ValueProperty<Type>& property, Function f, Object obj);

	template<typename Type, typename Function, typename Object, typename Arg1>
	void registerUpdater(ValueProperty<Type>& property, Function f, Object obj, Arg1 arg1);

	template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2>
	void registerUpdater(ValueProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2);

	template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
	void registerUpdater(ValueProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3);

	template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
	void registerUpdater(ValueProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3, Arg4 arg4);

	template<typename Type>
	void unregisterUpdater(ValueDynamicProperty<Type>& property);

	template<typename Type, typename Function, typename Object>
	void registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj);

	template<typename Type, typename Function, typename Object, typename Arg1>
	void registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj, Arg1 arg1);

	template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2>
	void registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2);

	template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
	void registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3);

	template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
	void registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3,
			Arg4 arg4);

	template<typename Type>
	void unregisterUpdater(ValueDynamicProperty2<Type>& property);

	template<typename Type, typename Function, typename Object>
	void registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj);

	template<typename Type, typename Function, typename Object, typename Arg1>
	void registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj, Arg1 arg1);

	template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2>
	void registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2);

	template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
	void registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3);

	template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
	void registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3,
			Arg4 arg4);

	template<typename ControllerType>
	void unregisterProgression(ControllerPtr<ControllerType>& controller);

	template<typename ControllerType, typename Function, typename Object>
	void registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj);

	template<typename ControllerType, typename Function, typename Object, typename Arg1>
	void registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1);

	template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2>
	void registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1, Arg2 arg2);

	template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
	void registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1, Arg2 arg2,
			Arg3 arg3);

	template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3,
			typename Arg4>
	void registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3,
			Arg4 arg4);

	template<typename DriverType>
	void unregisterProgression(DriverPtr<DriverType>& driver);

	template<typename DriverType, typename Function, typename Object>
	void registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj);

	template<typename DriverType, typename Function, typename Object, typename Arg1>
	void registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1);

	template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2>
	void registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1, Arg2 arg2);

	template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
	void registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3);

	template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
	void registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3, Arg4 arg4);

	template<typename ControllerType>
	void unregisterStatus(ControllerPtr<ControllerType>& controller);

	template<typename ControllerType, typename Function, typename Object>
	void registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj);

	template<typename ControllerType, typename Function, typename Object, typename Arg1>
	void registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1);

	template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2>
	void registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1, Arg2 arg2);

	template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
	void registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3);

	template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3,
			typename Arg4>
	void registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3,
			Arg4 arg4);

	template<typename DriverType>
	void unregisterStatus(DriverPtr<DriverType>& controller);

	template<typename DriverType, typename Function, typename Object>
	void registerStatus(DriverPtr<DriverType>& controller, Function f, Object obj);

	template<typename DriverType, typename Function, typename Object, typename Arg1>
	void registerStatus(DriverPtr<DriverType>& controller, Function f, Object obj, Arg1 arg1);

	template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2>
	void registerStatus(DriverPtr<DriverType>& controller, Function f, Object obj, Arg1 arg1, Arg2 arg2);

	template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
	void registerStatus(DriverPtr<DriverType>& controller, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3);

	template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
	void registerStatus(DriverPtr<DriverType>& controller, Function f, Object obj, Arg1 arg1, Arg2 arg2, Arg3 arg3, Arg4 arg4);

	template<typename Type>
	void copyValue(SimpleProperty<Type>& property, typename ArgumentType<Type>::value value);

	template<typename Type>
	void copyValueByUpdate(SimpleProperty<Type>& property, typename ArgumentType<Type>::value value);

	template<typename Type>
	void copyPropertyValue(SimpleProperty<Type>& property, SimpleProperty<Type>& sourceProperty);

	template<typename Type>
	void copyPropertyValueByUpdate(SimpleProperty<Type>& property, SimpleProperty<Type>& sourceProperty);

	template<typename Type>
	void registerValueCopier(SimpleProperty<Type>& property, SimpleProperty<Type>& targetProperty);

	template<typename Type>
	void registerValueCopierByUpdate(SimpleProperty<Type>& property, SimpleProperty<Type>& targetProperty);

	template<typename Type>
	void registerPropertyCopier(SimpleProperty<Type>& property, SimpleProperty<Type>& targetProperty);

	template<typename Type>
	void registerPropertyCopierByUpdate(SimpleProperty<Type>& property, SimpleProperty<Type>& targetProperty);

	void sendProgressEvent(int32 progress);
	void sendArrayEvent(const std::string& propertyName);
	void sendErrorEvent(int32 errorId, const std::string& errorMessage);

	void setConfigured(bool aValue);

	virtual bool isLogging() const;

	void refreshProgression(int32 value);

	void setDebug();

private:

	void constructor(const std::string& type, const std::string& name);

	void updateDisconnected();
	AbstractController* findControllerWithName(const std::string& name);

	std::map<std::string, AbstractController**> m_RequiredControllers;
	std::map<std::string, std::string> m_RequiredType;
	std::map<std::string, std::string> m_RequiredFunctions;
	int32 m_NRequiredControllers;
	std::map<std::string, boost::any> m_ChangeValues;
	bool m_Configured = false;
	bool m_saved = false;
	bool m_inited = false;

	Module * m_module;
	std::map<std::string, BasePtr *> m_servantPointers;
	std::map<std::string, controller::Command *> m_servantFunctions;
	boost::thread_specific_ptr<bool> m_logging;

	std::vector<std::string> m_ExcludeSetEnableController;

	bool m_CloneController;

public:
	Property<bool> isCommand;
	Property<std::string> nickname;
	Property<int32> commandProgression;
	StatusProperty commandStatus;
	Property<bool> driverDisconnected;

	static const std::string LIMIT_STR;

};

static const char EXPERIMENT_CONTROLLER[] = "experiment";

template<typename ControllerType>
void ExperimentController::registerController(const std::string& role, ControllerPtr<ControllerType>& controllerPointer) {

	m_servantPointers[role] = &controllerPointer;
	m_RequiredControllers[role] = controllerPointer.getControllerPointer();
	m_RequiredType[role] = EXPERIMENT_CONTROLLER_TYPE;
	m_NRequiredControllers++;
}

template<typename ControllerType>
void ControllerPtr<ControllerType>::init(ExperimentController * pointerOwner, const std::string & role) {
	m_role = role;
	pointerOwner->registerController(role, *this);
}

template<typename ControllerType>
bool ControllerPtr<ControllerType>::isInited() {
	return (m_role != "");
}

template<typename ControllerType>
void ControllerPtr<ControllerType>::reset(ExperimentController * pointerOwner) {
	assign(0);
	pointerOwner->unregisterController(m_role);
}

template<typename DriverType>
void ExperimentController::registerDriver(const std::string& role, DriverPtr<DriverType>& driverPointer) {

	m_servantPointers[role] = &driverPointer;
	m_RequiredControllers[role] = driverPointer.getControllerPointer();
	m_RequiredType[role] = TEST_CONTROLLER_TYPE;
	m_NRequiredControllers++;
}

template<typename DriverType>
void DriverPtr<DriverType>::init(ExperimentController * pointerOwner, const std::string & role) {

	m_ownerName = pointerOwner->getName();
	m_role = role;

	pointerOwner->registerDriver(role, *this);
}

template<typename DriverType>
bool DriverPtr<DriverType>::isInited() {
	return (m_role != "");
}

inline void ExperimentController::unregisterUpdaters(const std::string& servantName) {
	PropertyUpdater::unregisterPropertyUpdaters(servantName);
}

template<typename ControllerType>
void ExperimentController::unregisterUpdaters(ControllerPtr<ControllerType>& controllerPointer) {
	if (controllerPointer.isAssigned()) {
		PropertyUpdater::unregisterPropertyUpdaters(controllerPointer.get()->getName());
	}
}

template<typename DriverType>
void ExperimentController::unregisterUpdaters(DriverPtr<DriverType>& driverPointer) {
	if (driverPointer.isAssigned()) {
		PropertyUpdater::unregisterPropertyUpdaters(driverPointer.get()->getName());
	}
}

template<typename Type>
void ExperimentController::unregisterUpdater(ValueProperty<Type>& property) {
	PropertyUpdater::unregisterPropertyUpdater(property.getServantName(), property.getName());
}

template<typename Type, typename Function, typename Object>
void ExperimentController::registerUpdater(ValueProperty<Type>& property, Function f, Object obj) {
	PropertyUpdater::registerPropertyUpdater(property.getServantName(), property.getName(), f, obj);
}

template<typename Type, typename Function, typename Object, typename Arg1>
void ExperimentController::registerUpdater(ValueProperty<Type>& property, Function f, Object obj, Arg1 arg1) {
	PropertyUpdater::registerPropertyUpdater(property.getServantName(), property.getName(), f, obj, arg1);
}

template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2>
void ExperimentController::registerUpdater(ValueProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2) {
	PropertyUpdater::registerPropertyUpdater(property.getServantName(), property.getName(), f, obj, arg1, arg2);
}

template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
void ExperimentController::registerUpdater(ValueProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3) {
	PropertyUpdater::registerPropertyUpdater(property.getServantName(), property.getName(), f, obj, arg1, arg2, arg3);
}

template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
void ExperimentController::registerUpdater(ValueProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3, Arg4 arg4) {
	PropertyUpdater::registerPropertyUpdater(property.getServantName(), property.getName(), f, obj, arg1, arg2, arg3, arg4);
}

template<typename Type>
void ExperimentController::unregisterUpdater(ValueDynamicProperty<Type>& property) {
	PropertyUpdater::unregisterPropertyUpdater(property.getServantName(), property.getName());
}

template<typename Type, typename Function, typename Object>
void ExperimentController::registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj) {
	PropertyUpdater::registerDynamicPropertyUpdater(property.getServantName(), property.getName(), f, obj);
}

template<typename Type, typename Function, typename Object, typename Arg1>
void ExperimentController::registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj, Arg1 arg1) {
	PropertyUpdater::registerDynamicPropertyUpdater(property.getServantName(), property.getName(), f, obj, arg1);
}

template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2>
void ExperimentController::registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2) {
	PropertyUpdater::registerDynamicPropertyUpdater(property.getServantName(), property.getName(), f, obj, arg1, arg2);
}

template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
void ExperimentController::registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3) {
	PropertyUpdater::registerDynamicPropertyUpdater(property.getServantName(), property.getName(), f, obj, arg1, arg2, arg3);
}

template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
void ExperimentController::registerUpdater(ValueDynamicProperty<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3, Arg4 arg4) {
	PropertyUpdater::registerDynamicPropertyUpdater(property.getServantName(), property.getName(), f, obj, arg1, arg2, arg3,
			arg4);
}

template<typename Type>
void ExperimentController::unregisterUpdater(ValueDynamicProperty2<Type>& property) {
	PropertyUpdater::unregisterPropertyUpdater(property.getServantName(), property.getName());
}

template<typename Type, typename Function, typename Object>
void ExperimentController::registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj) {
	PropertyUpdater::registerDynamicPropertyUpdater2(property.getServantName(), property.getName(), f, obj);
}

template<typename Type, typename Function, typename Object, typename Arg1>
void ExperimentController::registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj, Arg1 arg1) {
	PropertyUpdater::registerDynamicPropertyUpdater2(property.getServantName(), property.getName(), f, obj, arg1);
}

template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2>
void ExperimentController::registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj, Arg1 arg1,
		Arg2 arg2) {
	PropertyUpdater::registerDynamicPropertyUpdater2(property.getServantName(), property.getName(), f, obj, arg1, arg2);
}

template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
void ExperimentController::registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3) {
	PropertyUpdater::registerDynamicPropertyUpdater2(property.getServantName(), property.getName(), f, obj, arg1, arg2, arg3);
}

template<typename Type, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
void ExperimentController::registerUpdater(ValueDynamicProperty2<Type>& property, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3, Arg4 arg4) {
	PropertyUpdater::registerDynamicPropertyUpdater2(property.getServantName(), property.getName(), f, obj, arg1, arg2, arg3,
			arg4);
}

template<typename ControllerType>
void ExperimentController::unregisterProgression(ControllerPtr<ControllerType>& controller) {
	unregisterUpdater(controller->commandProgression);
}

template<typename ControllerType, typename Function, typename Object>
void ExperimentController::registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj) {
	registerUpdater(controller->commandProgression, f, obj);
}

template<typename ControllerType, typename Function, typename Object, typename Arg1>
void ExperimentController::registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1) {
	registerUpdater(controller->commandProgression, f, obj, arg1);
}

template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2>
void ExperimentController::registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1,
		Arg2 arg2) {
	registerUpdater(controller->commandProgression, f, obj, arg1, arg2);
}

template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
void ExperimentController::registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1,
		Arg2 arg2, Arg3 arg3) {
	registerUpdater(controller->commandProgression, f, obj, arg1, arg2, arg3);
}

template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
void ExperimentController::registerProgression(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1,
		Arg2 arg2, Arg3 arg3, Arg4 arg4) {
	registerUpdater(controller->commandProgression, f, obj, arg1, arg2, arg3, arg4);
}

template<typename DriverType>
void ExperimentController::unregisterProgression(DriverPtr<DriverType>& driver) {
	unregisterUpdater(driver->commandProgression);
}

template<typename DriverType, typename Function, typename Object>
void ExperimentController::registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj) {
	registerUpdater(driver->commandProgression, f, obj);
}

template<typename DriverType, typename Function, typename Object, typename Arg1>
void ExperimentController::registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1) {
	registerUpdater(driver->commandProgression, f, obj, arg1);
}

template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2>
void ExperimentController::registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1, Arg2 arg2) {
	registerUpdater(driver->commandProgression, f, obj, arg1, arg2);
}

template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
void ExperimentController::registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3) {
	registerUpdater(driver->commandProgression, f, obj, arg1, arg2, arg3);
}

template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
void ExperimentController::registerProgression(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3, Arg4 arg4) {
	registerUpdater(driver->commandProgression, f, obj, arg1, arg2, arg3, arg4);
}

template<typename ControllerType>
void ExperimentController::unregisterStatus(ControllerPtr<ControllerType>& controller) {
	PropertyUpdater::unregisterPropertyUpdater(controller->commandStatus.getServantName(), controller->commandStatus.getName());
}

template<typename ControllerType, typename Function, typename Object>
void ExperimentController::registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj) {
	PropertyUpdater::registerPropertyUpdater(controller->commandStatus.getServantName(), controller->commandStatus.getName(), f,
			obj);
}

template<typename ControllerType, typename Function, typename Object, typename Arg1>
void ExperimentController::registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1) {
	PropertyUpdater::registerPropertyUpdater(controller->commandStatus.getServantName(), controller->commandStatus.getName(), f,
			obj, arg1);
}

template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2>
void ExperimentController::registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1,
		Arg2 arg2) {
	PropertyUpdater::registerPropertyUpdater(controller->commandStatus.getServantName(), controller->commandStatus.getName(), f,
			obj, arg1, arg2);
}

template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
void ExperimentController::registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1,
		Arg2 arg2, Arg3 arg3) {
	PropertyUpdater::registerPropertyUpdater(controller->commandStatus.getServantName(), controller->commandStatus.getName(), f,
			obj, arg1, arg2, arg3);
}

template<typename ControllerType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
void ExperimentController::registerStatus(ControllerPtr<ControllerType>& controller, Function f, Object obj, Arg1 arg1,
		Arg2 arg2, Arg3 arg3, Arg4 arg4) {
	PropertyUpdater::registerPropertyUpdater(controller->commandStatus.getServantName(), controller->commandStatus.getName(), f,
			obj, arg1, arg2, arg3, arg4);
}

template<typename DriverType>
void ExperimentController::unregisterStatus(DriverPtr<DriverType>& driver) {
	PropertyUpdater::unregisterPropertyUpdater(driver->commandStatus.getServantName(), driver->commandStatus.getName());
}

template<typename DriverType, typename Function, typename Object>
void ExperimentController::registerStatus(DriverPtr<DriverType>& driver, Function f, Object obj) {
	PropertyUpdater::registerPropertyUpdater(driver->commandStatus.getServantName(), driver->commandStatus.getName(), f, obj);
}

template<typename DriverType, typename Function, typename Object, typename Arg1>
void ExperimentController::registerStatus(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1) {
	PropertyUpdater::registerPropertyUpdater(driver->commandStatus.getServantName(), driver->commandStatus.getName(), f, obj,
			arg1);
}

template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2>
void ExperimentController::registerStatus(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1, Arg2 arg2) {
	PropertyUpdater::registerPropertyUpdater(driver->commandStatus.getServantName(), driver->commandStatus.getName(), f, obj,
			arg1, arg2);
}

template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3>
void ExperimentController::registerStatus(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3) {
	PropertyUpdater::registerPropertyUpdater(driver->commandStatus.getServantName(), driver->commandStatus.getName(), f, obj,
			arg1, arg2, arg3);
}

template<typename DriverType, typename Function, typename Object, typename Arg1, typename Arg2, typename Arg3, typename Arg4>
void ExperimentController::registerStatus(DriverPtr<DriverType>& driver, Function f, Object obj, Arg1 arg1, Arg2 arg2,
		Arg3 arg3, Arg4 arg4) {
	PropertyUpdater::registerPropertyUpdater(driver->commandStatus.getServantName(), driver->commandStatus.getName(), f, obj,
			arg1, arg2, arg3, arg4);
}

template<typename Type>
void ExperimentController::copyValue(SimpleProperty<Type>& property, typename ArgumentType<Type>::value value) {
	property = value;
}

template<typename Type>
void ExperimentController::copyValueByUpdate(SimpleProperty<Type>& property, typename ArgumentType<Type>::value value) {
	property.update(value);
}

template<typename Type>
void ExperimentController::copyPropertyValue(SimpleProperty<Type>& property, SimpleProperty<Type>& sourceProperty) {
	property = sourceProperty();
}

template<typename Type>
void ExperimentController::copyPropertyValueByUpdate(SimpleProperty<Type>& property, SimpleProperty<Type>& sourceProperty) {
	property.update(sourceProperty());
}

template<typename Type>
void ExperimentController::registerPropertyCopier(SimpleProperty<Type>& property, SimpleProperty<Type>& targetProperty) {
	registerUpdater(property, &ExperimentController::copyPropertyValue<Type>, this, targetProperty, property);
}

template<typename Type>
void ExperimentController::registerPropertyCopierByUpdate(SimpleProperty<Type>& property,
		SimpleProperty<Type>& targetProperty) {
	registerUpdater(property, &ExperimentController::copyPropertyValueByUpdate<Type>, this, targetProperty, property);
}

template<typename Type>
void ExperimentController::registerValueCopier(SimpleProperty<Type>& property, SimpleProperty<Type>& targetProperty) {
	registerRefresher(property, &ExperimentController::copyValue<Type>, this, targetProperty);
}

template<typename Type>
void ExperimentController::registerValueCopierByUpdate(SimpleProperty<Type>& property, SimpleProperty<Type>& targetProperty) {
	registerRefresher(property, &ExperimentController::copyValueByUpdate<Type>, this, targetProperty);
}

/*
 * refreshPropertyisStriclyPositive
 */
template<typename Type> void ExperimentController::refreshPropertyisStriclyPositive(const std::string& name, Type value)  {
	if ((value > 0) == false) {
		throw CannotSetValue("  cannot set property " + name + " on controller " + m_name + ", value must be stricly positive");
	}
}

/*
 * refreshPropertyisPositive
 */
template<typename Type> void ExperimentController::refreshPropertyisPositive(const std::string& name, Type value)  {
	if (value < 0) {
		throw CannotSetValue("  cannot set property " + name + " on controller " + m_name + ", value must be positive");
	}
}

/*
 * Property name
 */

/*
 * Generic
 */
static const std::string STATUS_PROPERTY = "status";
static const std::string PROGRESSION_PROPERTY = "progression";
static const std::string IS_COMMAND_PROPERTY = "isCommand";

#endif
