/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "InstrumentAbstraction/TestController.h"
#include "CommandNames.h"
#include "DataProvider/ChangeAspectFactory.h"
#include "DataProvider/ChangeAspect.h"
#include "DeviceContainer/HardwareCommand.h"
#include "CommandNames.h"
#include "LoggingServices/Log.h"
#include "Scheduler/Sender.h"
#include "Scheduler/Terminated.h"
#include "DeviceContainer/Spy.h"
#include "DeviceContainer/Observer.h"
//#include "Utilities/Simulated.h"

using namespace std;
using namespace boost;

/*
 * Constructor
 */
TestController::TestController(const string& type, const string& name) :
	AbstractController(type, name, TEST_CONTROLLER_TYPE), m_DeviceDriver(NULL), m_Parent(NULL), m_TimeOut(-1) {
}

/*
 * Destructor
 */
TestController::~TestController() {

	setPropertyManager(NULL);

	/*
	 * Detach from publisher
	 */
	if (m_DeviceDriver != NULL) {
		m_DeviceDriver->detach(this);
	}

}

std::string TestController::getServantInfo() const {
	return "Driver " + getName() + " (" + getType() + ")";
}

/*
 * refreshSetValue
 */
void TestController::refreshSetValue(string propertyName, const boost::any aValue)  {
	try {
		getDeviceDriver()->refreshValue(propertyName, aValue);
	} catch (CannotSetValue&) {
		throw CannotSetValue();
	}
}

/*
 * updateProperty
 */
void TestController::updateProperty(ChangeAspect* aChangeAspect) {

	/*
	 * Just forward properties but change the name to reflect how it
	 * appears in the PropertyManager.
	 */

	ChangeAspectFactory aCAFactory;

	string propertyName = aChangeAspect->getStringParam();

	ChangeAspect* aCA = aCAFactory.createPropertyChangeAspect(getType(), getName(), propertyName);
	notify(aCA);

}

/*
 * updateProgress
 */
void TestController::updateProgress(ChangeAspect* aChangeAspect) {

	/*
	 * Just forward properties but change the name to reflect how it
	 * appears in the PropertyManager.
	 */

	ChangeAspectFactory aCAFactory;

	int32 progression = aChangeAspect->getLongParam();

	if (progression == 100) {
		Observer::getInstance()->unregisterDeviceDriver(m_DeviceDriver);
	}
	ChangeAspect* aCA = aCAFactory.createProgressChangeAspect(getType(), getName(), progression);
	notify(aCA);

}

void TestController::updateError(ChangeAspect* ca) {
	/*
	 * Just forward properties but change the name to reflect how it
	 * appears in the PropertyManager.
	 */
	notify(ca);
}
;

/*
 * isConfigured
 */
bool TestController::isConfigured() {

	if (m_DeviceDriver == NULL) {
		return false;
	} else {
		return m_DeviceDriver->isConfigured();
	}
}

/*
 * setDeviceDriver
 */
void TestController::setDeviceDriver(DeviceDriver* aDriver) {

	m_DeviceDriver = aDriver;

	/*
	 * Now that we've fixed the DeviceDriver we can recuperate all
	 * of its properties.
	 */
	if (m_DeviceDriver != NULL) {
		setPropertyManager(m_DeviceDriver->getPropertyManager());

		/*
		 * Now that we've fixed the DeviceDriver we can recuperate all
		 * of its commands.
		 */
		string commandlist = m_DeviceDriver->getCommands();
		size_t position = 0;
		// generated command from the list
		do {
			position = commandlist.find('\n', 0);
			if (position != string::npos) {
				string command = commandlist.substr(0, position);
				initCommand(command);
				commandlist.erase(0, position + 1);
			}
		} while (!commandlist.empty());
	}
}

/*
 * setParent
 */
void TestController::setParent(TestController* aDevice, int32 aChannel) {
	m_DeviceDriver->setParent(aDevice->getDeviceDriver());
	m_Parent = aDevice;
	m_DeviceDriver->leafChannel = aChannel;
}

/*
 * getParent
 */
TestController* TestController::getParent(bool localThrow)  {
	if (m_Parent) {
		return m_Parent;
	} else {
		if (localThrow) {
			throw AbstractController::ControllerNotConfigured();
		} else {
			throw AbstractController::ControllerNotConfigured(m_type, m_name, this, string("Parent"));
		}
	}
}

/*
 * removeParentLink
 */
void TestController::removeParent()  {
	if (m_Parent) {
		m_DeviceDriver->removeParent();
		m_Parent = NULL;
	} else {
		throw AbstractController::ControllerNotConfigured(m_type, m_name, this, string("Parent"));
	}
}

/*
 * setChild
 */
void TestController::setChild(TestController* aDevice, const int32 aChannel) {
	m_DeviceDriver->setChild(aDevice->getDeviceDriver(), aChannel);
}

/*
 * setTCChild
 */
void TestController::setTCChild(TestController* aDevice, const int32 aChannel) {
	m_child.insert(pair<int32, TestController*>(aChannel - 1, aDevice));
}

/*
 * removeChild
 */
void TestController::removeChild(const int32 aChannel) {
	//m_child.erase(aChannel);
	m_DeviceDriver->removeChild(aChannel);
}

void TestController::removeTCChild(const int32 aChannel) {
	m_child.erase(aChannel - 1);
}

/*
 * getNumberOfChilds
 */
bool TestController::getChannelsDisable() const {
	bool disable = true;
	for (multimap<int32, TestController*>::const_iterator iter = m_child.begin(); iter != m_child.end(); ++iter) {
		TestController* tc = (*iter).second;
		if (tc->isEnabled() == true) {
			disable = false;
		}
	}
	return disable;
}

/*
 * activate
 */
void TestController::activate() {
}

/*
 * deactivate
 */
void TestController::deactivate() {
}

void TestController::execute(const std::string& command, bool blocking, bool logging)  {
	execute(command, blocking, logging, false);
}

/*
 * execute
 */
void TestController::execute(const string& command, bool blocking, bool logging, bool observerIfNotBlocking)  {

	if (isEnabled()) {

#ifndef TESTS

		if (command == COMMAND_PAUSE) {
			Sender<HardwareCommand>::pause(true);
		} else if ((command != COMMAND_READ) && (command != COMMAND_STATUS)) {
			Sender<HardwareCommand>::pause(false);
		}

		// change blocking in case of read and status
		if (blocking && (command == COMMAND_READ || command == COMMAND_STATUS)) {
			blocking = false;
		}

		if ((command == COMMAND_PAUSE) || (command == COMMAND_CONTINUE)) {

			Sender<HardwareCommand>::begin();
			Sender<HardwareCommand>::post(HardwareCommand(m_DeviceDriver, command));
			Sender<HardwareCommand>::endIf(End<HardwareCommand>(), -1);

		} else if (blocking) {

			// Synchronized command
			Sender<HardwareCommand>::begin();

			Terminated terminated(m_DeviceDriver);

			Sender<HardwareCommand>::post(HardwareCommand(m_DeviceDriver, command));

			if ((command != COMMAND_INIT) && (command != COMMAND_READ) && (command != COMMAND_STATUS) && (command != COMMAND_CONTINUE) && (command != COMMAND_PAUSE)) {
				Observer::getInstance()->registerDeviceDriver(m_DeviceDriver, 1);
			}

			Sender<HardwareCommand>::endIf(End<HardwareCommand>() && terminated, m_TimeOut);

		} else {

			Sender<HardwareCommand>::begin();
			Sender<HardwareCommand>::post(HardwareCommand(m_DeviceDriver, command));
			Sender<HardwareCommand>::endIf(End<HardwareCommand>(), m_TimeOut);

			if (observerIfNotBlocking) {
				Observer::getInstance()->registerDeviceDriver(m_DeviceDriver, 1);
			}

		}

		if ((m_TimeOut > -1) && (Sender<HardwareCommand>::isTimedOut())) {

			log(Level::s_Warning) << command << " timed out after " << m_TimeOut << "s" << endlog;
			m_TimeOut = -1;

			throw TimeOutOccured();
		}

		m_TimeOut = -1;

#else
		HardwareCommand c(m_DeviceDriver,command);
		c.execute();
		HardwareCommand status(m_DeviceDriver,COMMAND_STATUS);
		for (;;) {
			if (m_DeviceDriver != 0)
			status.execute();
			if (m_DeviceDriver->isTerminated()) break;
			usleep(50000);
		}
#endif
	}
}

/*
 * setEnabled
 */
void TestController::setEnabledForLoading(bool all, bool postconf) {
	if (!isEnabled()) {
		try {
			forceEnabled(false);
		} catch (...) {
			// do nothing
		}
	}
}

/*
 * setEnabled
 */
void TestController::setEnabled(bool all, bool postconf) {
	setEnabledForLoading(all, postconf);
}

void TestController::forceEnabled(bool reconnect) {
	// Init properties (set value) because now controller is in used
	initProperties();
	// Set Parent link if necessary
	try {
		// Check if parent is enable, if not enable it before
		if (getParent(true)->isEnabled() == false) {
			getParent(true)->setEnabled(false);
		}
		if (reconnect == false) {
			getParent(true)->setChild(this, getChannel());
		}
	} catch (AbstractController::ControllerNotConfigured&) {
		// No Parent, catch for root device
	}
//	} catch (DeviceDriver::ChildIsUsed&) {
//		// Already set
//	} catch (DeviceDriver::ParentBadType&) {
//		// Root device doesn't need to set parent
//	} catch (DeviceDriver::ParentBadConf&) {
//		// not configured
//	}
	// Set spy
	Spy::getInstance()->registerDeviceDriver(m_DeviceDriver, 1);
	m_DeviceDriver->attach(this);
	m_DeviceDriver->setEnabled(true);
	m_DeviceDriver->clearDisconnected();
	enable(true);

	// Initialise driver
	try {
		m_DeviceDriver->execute(COMMAND_INIT);

	} catch (...) {
		cerr << "Failed to initialise " << getName() << endl;
		throw;
	}
}

/*
 * setDisabled
 */
void TestController::setDisabled(bool all) {
	// set disabled
	// Detach spy
	Spy::getInstance()->unregisterDeviceDriver(m_DeviceDriver);
	m_DeviceDriver->detach(this);
	m_DeviceDriver->setEnabled(false);
	if (isEnabled()) {
		// Remove child link to parent
		try {
			// try to get channel number to remove child in the parent, if it existed
			uint32 channel = getChannel();
			getParent(true)->removeChild(channel);
		} catch (AbstractController::ControllerNotConfigured&) {
			// Not configured
		} catch (DeviceDriver::ParentBadType&) {
			// Root device doesn't need to set parent
		} catch (DeviceDriver::NoDeviceDriver&) {
			// It is not necessary to remove because it doesn't exist in children list
			// It is true at the beginning of the server
		}
	}

	enable(false);
	// Check if parent is enable, if not enable it before
	try {
		if (getParent(true)->getChannelsDisable() == true) {
			if (getParent(true)->isEnabled() == true) {
				getParent(true)->setDisabled(false);
			}
		}
	} catch (AbstractController::ControllerNotConfigured&) {
		// Not configured
	}
}

/*
 * getDeviceDriver
 */
DeviceDriver* TestController::getDeviceDriver() const {
	return m_DeviceDriver;
}

/*
 * getChannel
 */
int32 TestController::getChannel() const {
	return m_DeviceDriver->getChannel();
}

/*
 * getRunningMode
 */
bool TestController::isReal() const {
	bool real = any_cast<std::string>(m_DeviceDriver->getValue(STATE_DEVICE_CONTAINER)).find(REAL_DEVICE_CONTAINER) != std::string::npos;
	return real;
}

/*
 * getParentName
 */
std::string TestController::getParentName() const {
	std::string parent = "root";
	if (m_Parent != NULL) {
		parent = m_Parent->getName();
	}
	return parent;
}

/*
 * containsFunction
 */
bool TestController::containsFunction(const std::string& function) {
	return m_DeviceDriver->containsCategory(function);
}

/*
 * clearDisconnected
 */
void TestController::clearDisconnected() {
	m_DeviceDriver->clearDisconnected();
}

/*
 * setTimeOut
 */
void TestController::setTimeOut(std::string const & command, int32 sec) {
	m_TimeOut = sec;
}

/*
 * getChilds
 */
const std::multimap<int32, TestController*>& TestController::getChilds() {
	return m_child;
}

void TestController::setValueAndFormat(const std::string& propertyName, boost::any aValue) {

	AbstractController::setValue(propertyName, aValue);
}
