/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef TESTCONTROLLER_H
#define TESTCONTROLLER_H

#include <string>
#include <map>
#include <iostream>

#include "InstrumentAbstraction/AbstractController.h"
#include "DeviceContainer/DeviceDriver.h"

class ChangeAspect;

/**
 * The ElementaryController is a subclass of the AbstractController. It
 * provides the interface for a specific element of the instrument, typically
 * a device, or a particular functionality of a complex device. It again
 * is subclassed to provide concrete instances of devices.
 */
class TestController: public AbstractController {

public:
	/**
	 * Execute a particular command if it has been implemented. Forwards its call to execute with observerIfNotBlocking parameter.
	 */
	virtual void execute(const std::string& command, bool blocking, bool logging) ;

	/**
	 * Execute implementation with observerIfNotBlocking parameter.
	 */
	void execute(const std::string& command, bool blocking, bool logging, bool observerIfNotBlocking) ;

	/**
	 * Constructor. Must be supplied a name and an ID such as motor, gate, etc
	 */
	TestController(const std::string& inputType, const std::string& inputName);

	/**
	 * Destructor
	 */
	virtual ~TestController();

	/**
	 * Gets the servant info.
	 */
	virtual std::string getServantInfo() const;

	/**
	 * This function is called before each set of a property in order
	 * to modify our table or Device Container properties
	 */
	virtual void refreshSetValue(std::string propertyName, const boost::any aValue);

	/**
	 * Method to set the DeviceDriver for the controller.
	 */
	void setDeviceDriver(DeviceDriver* aDriver);

	/**
	 * Return the device driver in use. Used mainly by subclasses - maybe
	 * this should then be a protected data.
	 */
	DeviceDriver* getDeviceDriver() const;

	/**
	 * Set the device link on parent
	 *
	 * @param the device parent or child
	 */
	void setParent(TestController* aDevice, int32 aChannel);

	/**
	 * Remove the device link on parent.
	 */
	void removeParent() ;

	/**
	 * Set the device link on child
	 *
	 * @param the device parent or child
	 * @param the channel of the device
	 */
	void setChild(TestController* aDevice, const int32 aChannel);
	void setTCChild(TestController* aDevice, const int32 aChannel);

	/**
	 * Gets the channel of the TestController.
	 */
	int32 getChannel() const;
	bool getChannelsDisable() const;

	/**
	 * Gets the channel of the TestController.
	 */
	std::string getParentName() const;

	/**
	 * Remove the device link on child.
	 *
	 * @param the channel of the device
	 */
	void removeChild(const int32 aChannel);
	//    throw(AbstractController::ControllerNotConfigured);
	void removeTCChild(const int32 aChannel);

	/**
	 * Verifies if the Controller is configured and can be used.
	 */
	virtual bool isConfigured();

	/**
	 * Get the parent link.
	 */
	TestController* getParent(bool localThrow = false)
			;
	const std::multimap<int32, TestController*>& getChilds();

	virtual bool containsFunction(const std::string& function);

	/**
	 * Set the timeout.
	 * This is set the time out for the next command
	 * After command timeout is reset
	 * This is no global timeout, each controler have to set timeout before an command
	 */
	void setTimeOut(std::string const & command, int32 sec);

	/**
	 * Set enable the controller
	 */
	virtual void setEnabled(bool all ,bool postconf = true);

	void forceEnabled(bool reconnect);


	/**
	 * Set disable the controller
	 */
	virtual void setDisabled(bool all = false);

	void clearDisconnected();

	bool isReal() const;

	void setValueAndFormat(const std::string& propertyName, boost::any aValue);

	virtual void fillScanList(std::list<std::pair<std::string, std::string> > & list) {
	}

protected:
	virtual void setEnabledForLoading(bool all, bool postconf);

	/*
	 * Perform necessary operations when a client has made a setActive/
	 * unsetActive method call on the Repeater class. We need to put the
	 * Device Driver onto the Observer thread for increased update
	 * frequency.
	 */
	virtual void activate();

	virtual void deactivate();

	/*!
	 *\brief updateProperty
	 * Implementation of updateProperty() for the publisher subscriber pattern.
	 */
	virtual void updateProperty(ChangeAspect* aChangeAspect);

	/*!
	 *\brief updateProgress
	 * Implementation of updateProgress() for the publisher subscriber pattern.
	 */
	virtual void updateProgress(ChangeAspect* aChangeAspect);

	/*!
	 *\brief updateError
	 * Implementation of updateError() for the publisher subscriber pattern.
	 */
	virtual void updateError(ChangeAspect* ca);

private:

	/**
	 * @supplierCardinality 1
	 * @clientCardinality 1*/
	DeviceDriver* m_DeviceDriver;

	/**
	 * Parent link for need client hardware manager
	 */
	TestController* m_Parent;
	int32 m_TimeOut;

	std::multimap<int32, TestController*> m_child;

};

static const std::string STATUS_TEST_CONTROLLER = "status";

#endif
