/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <cstdlib>
#include <sstream>
#include <boost/tokenizer.hpp>
#include <boost/format.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/archive/text_oarchive.hpp>
#include "InstrumentManager/HardwareManager.h"
#include "InstrumentManager/InstrumentManager.h"
#include "InstrumentManager/NameManager.h"
#include "InstrumentAbstraction/TestControllerCreator.h"
#include "DeviceContainer/DeviceDriverCreator.h"
#include "DeviceContainer/Spy.h"
#include "CommandNames.h"
#include "PersistenceServices/Configuration/TestControllerConfig.h"
#include <dataprovider/notification/NotificationServer.h>

#include <ics/ServantConfiguration.h>
#include <common/output/OutputStream.h>
#include <common/base/Convert.h>
#include <set>

using namespace std;
using namespace ics;
using namespace common;
using namespace boost;

HardwareManager* HardwareManager::_instance = 0;

HardwareManager::HardwareManager() :
	_functionMap(0) {
}

HardwareManager::~HardwareManager() {

	// Delete all TestControllers and DeviceDrivers
	// Take care to delete all the children before the parent
	while (_installedTestControllers.size() != 0) {

		set<string> controllersToRemove;

		for (map<string, TestController*>::iterator t = _installedTestControllers.begin(); t != _installedTestControllers.end(); t++) {

			TestController* currentTC = t->second;
			if (currentTC) {
				DeviceDriver* currentDD = currentTC->getDeviceDriver();

				if (currentDD->isChildEmpty()) {
					delete currentTC;
					delete currentDD;
					controllersToRemove.insert(t->first);
				}
			}
		}

		for (set<string>::iterator c = controllersToRemove.begin(); c != controllersToRemove.end(); c++) {
			_installedTestControllers.erase(*c);
		}
	}

	delete _functionMap;
}

void HardwareManager::initInstance() {
	if (_instance == 0) {
		_instance = new HardwareManager();
	}
}

HardwareManager* HardwareManager::getInstance() {
	return _instance;
}


void HardwareManager::reset() {
	delete _instance;
}

HardwareManager::HardwareManagerError::HardwareManagerError(const std::string& message) :
	Exception("HardwareManager::HardwareManagerError", message) {
}

void HardwareManager::loadTestControllers() {

	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	vector<TestControllerConfig> testControllerConfigs = ServantConfiguration::getInstance()->getTestControllerList();

	for (vector<TestControllerConfig>::const_reverse_iterator c = testControllerConfigs.rbegin(); c != testControllerConfigs.rend(); ++c) {
		loadTestController(*c);
	}

	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
}

void HardwareManager::loadTestControllers(const vector<TestControllerConfigNode>& configuration, bool checkParentChannel) {
	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	vector<TestControllerConfigNode>::const_iterator end = configuration.end();
	for (vector<TestControllerConfigNode>::const_iterator c = configuration.begin(); c != end; ++c) {
		if ((checkParentChannel && c->config.getParentChannel() == -1) || !checkParentChannel) {
			loadTestController(c->config);
			loadTestControllers(c->children, false);
		}
	}

	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
}

TestController* HardwareManager::getTestController(const std::string& testControllerName)  {

	// Check that the Controller exists
	if (!controllerExists(testControllerName)) {
		throw HardwareManagerError(testControllerName);
	}

	return _installedTestControllers[testControllerName];
}

DeviceDriver* HardwareManager::getDriver(const std::string& testControllerName)  {
	return getTestController(testControllerName)->getDeviceDriver();
}

bool HardwareManager::controllerExists(const std::string& controllerName) const {

	for (map<string, TestController*>::const_iterator c = _installedTestControllers.begin(); c != _installedTestControllers.end(); c++) {
		if (c->first == controllerName) {
			return true;
		}
	}

	return false;
}

bool HardwareManager::controllerIDExists(const std::string& ID) const {
	return false;
}

TestController* HardwareManager::loadTestController(const TestControllerConfig& conf)  {

	TestController* newTestController = 0;

	out << "Loading " << conf.getName() << " (" << conf.getType() << ")" << eol;

	// Check that an ElementaryController with the same name
	// hasn't already been installed.
	if (NameManager::getInstance()->controllerExists(conf.getName())) {
		ostringstream sbuf;
		sbuf << "controller name already exists : " << conf.getName();
		throw HardwareManagerError(sbuf.str());
	}

	// Then check if controller type is a known type
	if (!TestControllerCreator::getInstance()->verifyType(conf.getType())) {
		ostringstream sbuf;
		sbuf << "unknown Type : " << conf.getType();
		throw HardwareManagerError(sbuf.str());
	}

	// Create new ElementaryController
	newTestController = TestControllerCreator::getInstance()->create(conf.getType(), conf.getName());

	// Create DeviceDriver
	DeviceDriver* device = DeviceDriverCreator::getInstance()->create(conf.getType(), conf.getName());

	// Add DeviceDriver to TestController
	newTestController->setDeviceDriver(device);

	// Init parent if there is one
	if (!conf.getParentName().empty()) {
		TestController* parent = getTestController(conf.getParentName());
		newTestController->setParent(parent, conf.getParentChannel());
		parent->setTCChild(newTestController, conf.getParentChannel());
	}

	try {
		// first attempt to set properties (for wanted)
		newTestController->initProperties(conf.getProperties(), 1);
	} catch (...) {
		ostringstream sbuf;
		sbuf << "error first init properties attempt : " << conf.getName() << " (" << conf.getType() << ")";
		throw HardwareManagerError(sbuf.str());
	}
	if (Simulated::for3d == true) {
		newTestController->setValue(STATE_DEVICE_CONTAINER, string(SIMULATED_DEVICE_CONTAINER));
	}
	else if ((Simulated::activated == true) || (Simulated::visa == true)) {
		newTestController->setValue(STATE_DEVICE_CONTAINER, string(PERFECT_DEVICE_CONTAINER));
	}
	out << "\tState : " << any_cast<string>(newTestController->getValue(STATE_DEVICE_CONTAINER)) << eol;

	// Add ElementaryController to the map
	_installedTestControllers[conf.getName()] = newTestController;
	_orderedInstalledTestControllers.push_back(newTestController);
	_installedTestControllersEnable[conf.getName()] = conf.isEnabled();

	return newTestController;

}

void HardwareManager::initTestControllers() {

	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	for(auto controller : _orderedInstalledTestControllers) {
		common::out << "Initialising " << controller->getName() << " (" << controller->getType() << ")" << common::eol;

		try {
			controller->getDeviceDriver()->formatAllValues();
		}
		catch (...) {
			cerr << "Bad formated property value" << endl;
		}

		// set enable or disable
		if (_installedTestControllersEnable[controller->getName()] == true) {
			try {
				controller->setEnabled(false);
			} catch (DeviceDriver::TimeOut& e) {
			}
			catch (...) {
				ostringstream sbuf;
				sbuf << "can't enable : " << controller->getName() << " (" << controller->getType() << ")";
				throw HardwareManagerError(sbuf.str());
			}
		} else {
			try {
				// Initialise properties
				controller->initProperties();
				controller->setDisabled();
			}
			catch (...) {
				ostringstream sbuf;
				sbuf << "can't disable : " << controller->getName() << " (" << controller->getType() << ")";
				throw HardwareManagerError(sbuf.str());;
			}
		}
	}

	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);

}

void HardwareManager::loadReports() {
	for(auto controller : _orderedInstalledTestControllers) {
		const vector<string>& warnings = controller->getWarnings();
		const vector<string>& errors = controller->getErrors();
		if ((warnings.empty() == false) || (errors.empty() == false)) {
			common::title << "Report for driver: " << controller->getName() << " (" << controller->getType() << ")" << common::eol;
			for(auto w : warnings) {
				common::warn << w << common::eol;
			}
			for(auto e : errors) {
				common::err << e << common::eol;
			}
		}
	}
}

void HardwareManager::saveTestControllers(bool isBackup) {
	vector<TestControllerConfigNode> configs;
	// Iterating the TestControllers via the map
	for (map<string, TestController*>::const_iterator t = _installedTestControllers.begin(); t != _installedTestControllers.end(); ++t) {
		if (t->second->getChannel() == 0) {
			saveATestControllers(t->second, configs);
		}
	}

	ServantConfiguration::getInstance()->saveTestControllerList(configs, isBackup);
}

TestControllerConfig HardwareManager::getTestControllerConfig(TestController* testController, bool all) {
	// Filling the TestControllerConfig from the attributes of the TestController
	TestControllerConfig result;
	try {
		TestController *parent = testController->getParent(true);
		result.setParentChannel(testController->getChannel());
		result.setParentName(parent->getName());
	} catch (AbstractController::ControllerNotConfigured&) {
		result.setParentChannel(-1);
		result.setParentName("");
	}

	result.setName(testController->getName());
	result.setType(testController->getType());
	result.setEnabled(testController->isEnabled());

	// Setting properties
	vector<AnyProperty *> properties;
	if (all == true) {
		properties = testController->getAllProperties();
	} else {
		properties = testController->getPropertiesToSave();
	}
	for (vector<AnyProperty *>::const_iterator p = properties.begin(); p
			!= properties.end(); ++p) {
		ostringstream value;
		string type = (*p)->getType();
		value << static_cast<AnyProperty> (*(*p));
		if ((type == UINT16_PROPERTY) || (type == INT16_PROPERTY) || (type == UINT32_PROPERTY) || (type == INT32_PROPERTY)
				|| (type == UINT64_PROPERTY) || (type == INT64_PROPERTY) || (type == FLOAT32_PROPERTY)
				|| (type == FLOAT64_PROPERTY) || (type == STRING_PROPERTY) || (type == BOOLEAN_PROPERTY)) {
			result.addProperty(PropertyConfig((*p)->getName(), value.str()));
		} else if ((type == INT16ARRAY_PROPERTY) || (type == INT32ARRAY_PROPERTY) || (type == INT64ARRAY_PROPERTY)
				|| (type == UINT16ARRAY_PROPERTY) || (type == UINT32ARRAY_PROPERTY) || (type == UINT64ARRAY_PROPERTY)
				|| (type == FLOAT64ARRAY_PROPERTY) || (type == FLOAT32ARRAY_PROPERTY)) {
			typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
			boost::char_separator<char> sep("|");
			string buf = value.str();
			tokenizer tokens(buf, sep);
			vector<string> array;
			for (tokenizer::iterator tok_iter = tokens.begin(); tok_iter
					!= tokens.end(); ++tok_iter) {
				array.push_back(*tok_iter);
			}
			int32 extrasize =
					testController->getPropertyManager()->getExtraSize(
							(*p)->getName());
			result.addProperty(
					PropertyConfig((*p)->getName(), array, extrasize));
		}
	}

	return result;
}

void HardwareManager::saveATestControllers(TestController *TC, std::vector<TestControllerConfigNode>& configs, bool all) {
	if (TC != 0) {
		TestControllerConfigNode aConfig;
		aConfig.config = getTestControllerConfig(TC, all);
		multimap<int32, TestController*> childs = TC->getChilds();
		multimap<int32, TestController*>::iterator c = childs.begin();
		while (c != childs.end()) {
			saveATestControllers(c->second, aConfig.children, all);
			++c;
		}
		configs.push_back(aConfig);
	}
}

void HardwareManager::setPaused(bool value) {
	if (value) {
		pause();
	} else {
		restart();
	}
}

void HardwareManager::pause() {
	Spy::getInstance()->notInCharge();

	out << "Hardware paused" << eol;
}

void HardwareManager::restart() {

	out << "Hardware restarted" << eol;

	init();
	Spy::getInstance()->inCharge();
}

void HardwareManager::init() {

	bool finished;

	set<string> initialisedTestControllers;

	do {
		finished = true;
		map<string, TestController*>::const_iterator iter;
		for (iter = _installedTestControllers.begin(); iter != _installedTestControllers.end(); ++iter) {
			string name = iter->first;
			if (initialisedTestControllers.find(name) == initialisedTestControllers.end()) {
				// Not initialised
				TestController* tc = iter->second;
				TestController* ptc = 0;
				bool init = false;
				try {
					ptc = tc->getParent(true);

					if (initialisedTestControllers.find(ptc->getName()) != initialisedTestControllers.end()) {
						init = true;
					} else {
						finished = false;
					}
				} catch (...) {
					// no parent, meaning that it is a root controller
					init = true;
				}
				if (init == true) {
					out << "Initialising " << name << "..." << eol;
					tc->clearDisconnected();
					tc->execute(COMMAND_INIT, false, true);
					initialisedTestControllers.insert(name);
				}
			}
		}

	} while (finished == false);
}

std::string HardwareManager::getTestControllersConfig() {

	vector<TestControllerConfigNode> configs;
	// Iterating the TestControllers via the map
	for (map<string, TestController*>::const_iterator t = _installedTestControllers.begin(); t != _installedTestControllers.end(); ++t) {
		if (t->second->getChannel() == 0) {
			saveATestControllers(t->second, configs, true);
		}
	}

	ostringstream buf;
	boost::archive::text_oarchive oa(buf);
	oa << configs;

	return buf.str();
}

void HardwareManager::createFunctionDriverMap() {

	int32 i = 0;
	_functionMap = new map<string, FunctionParent>();
	for (vector<string>::const_iterator d = DeviceDriverCreator::getInstance()->getDeviceDriverList().begin(); d != DeviceDriverCreator::getInstance()->getDeviceDriverList().end(); d++) {

		ostringstream os;
		os << "d" << i;

		DeviceDriver* driver = DeviceDriverCreator::getInstance()->create(*d, os.str());
		i++;


		FunctionParent functionParent;
		functionParent.function = driver->getFunction();
		functionParent.parentFunction = driver->getParentFunction();
		(*_functionMap) [string(*d)] = functionParent;
		//delete driver;
	}

}

std::vector<std::string> HardwareManager::getDriverTypesWithParentFunction(const std::string& parentFunction) {

	if (_functionMap == 0) {
		createFunctionDriverMap();
	}

	vector<string> result;

	for (map<string, FunctionParent>::iterator f = _functionMap->begin(); f != _functionMap->end(); f++) {

		string driverType = f->first;
		FunctionParent functionParent = f->second;
		if (functionParent.parentFunction == parentFunction) {
			result.push_back(driverType);
		}
	}

	return result;
}

void HardwareManager::createDriver(const std::string& parentName, const std::string& type, const std::string& name, const std::string& mode, int32 channel) {

	// Create a configuration structure
	TestControllerConfig conf(name, type, parentName);
	conf.setParentChannel(channel);

	//conf.setEnabled(false);
	conf.setEnabled(true);

	try {
		TestController* controller = loadTestController(conf);
		controller->setValue(STATE_DEVICE_CONTAINER, mode);

	} catch (...) {
		throw HardwareManagerError(name);
	}

}

void HardwareManager::removeDriver(const std::string& driverName) {

	try {
		map<string, TestController*>::iterator t = _installedTestControllers.find(driverName);
		if (t != _installedTestControllers.end()) {

			TestController* currentTC = t->second;
			if (currentTC != 0) {
				DeviceDriver* currentDD = currentTC->getDeviceDriver();

				if (currentDD->isChildEmpty()) {

					try {
						TestController* parentTC = currentTC->getParent(true);
						if (parentTC != 0) {
							parentTC->removeTCChild(currentTC->getChannel());
						}
					} catch(AbstractController::ControllerNotConfigured& e) {

					}
					delete currentTC;
					delete currentDD;
					_installedTestControllers.erase(t);

				} else {
					ostringstream sbuf;
					sbuf << "cannot remove driver " << driverName << " : has still children";
					throw HardwareManagerError(sbuf.str());
				}
			}
		} else {
			cerr << "Cannot remove driver " << driverName << endl;
		}

	} catch (...) {
		ostringstream sbuf;
		sbuf << "cannot remove driver " << driverName;
		throw HardwareManagerError(sbuf.str());
	}
}

vector<string> HardwareManager::getDriversWithType(const std::string& type) {

	vector<string> result;
	for (map<string, TestController*>::iterator f = _installedTestControllers.begin(); f != _installedTestControllers.end(); f++) {

		TestController* currentDriver = f->second;
		if (currentDriver->containsFunction(type)) {
			result.push_back(currentDriver->getName());
		}
	}
	return result;
}

map<string,TestController*> HardwareManager::getDriversPtrWithType(const std::string& type) {

	map<string,TestController*> result;
	for (map<string, TestController*>::iterator f = _installedTestControllers.begin(); f != _installedTestControllers.end(); f++) {

		TestController* currentDriver = f->second;
		if (currentDriver->getType() == type) {
			result.insert(*f);
		}
	}
	return result;
}

map<string, TestController*> HardwareManager::getAllInstalledDrivers() const {
	return _installedTestControllers;
}


void HardwareManager::refreshTestControllers(const std::vector<TestControllerConfigNode>& configuration, bool checkParentChannel) {
	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);
	vector<TestControllerConfigNode>::const_iterator end = configuration.end();
	for (vector<TestControllerConfigNode>::const_iterator c = configuration.begin(); c != end; ++c) {
		if ((checkParentChannel && c->config.getParentChannel() == -1) || !checkParentChannel) {
			refreshTestController(c->config);
			refreshTestControllers(c->children, false);
		}
	}
	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
	dataprovider::NotificationServer::getInstance()->configurationChanged(0, 0, common::TREE);
}

TestController* HardwareManager::refreshTestController(const TestControllerConfig& conf) {

	TestController* newTestController = _installedTestControllers[conf.getName()];

	out << "Refreshing " << conf.getName() << " (" << conf.getType() << ") , initialising..." << eol;

	try {
		// first attempt to set properties (for wanted)
		newTestController->initProperties(conf.getProperties(), 1);

	} catch (...) {
		cerr << "Error first init properties attempt : " << conf.getName() << " (" << conf.getType() << ")" << endl;
		throw;
	}

	newTestController->getDeviceDriver()->formatAllValues();

	if (Simulated::for3d == true) {
		newTestController->setValue(STATE_DEVICE_CONTAINER, string(SIMULATED_DEVICE_CONTAINER));
	}
	else if ((Simulated::activated == true) || (Simulated::visa == true)) {
		newTestController->setValue(STATE_DEVICE_CONTAINER, string(PERFECT_DEVICE_CONTAINER));
	}

	// set enable or disable
	if (conf.isEnabled()) {
		try {
			newTestController->setEnabled(false);
		}
		catch (...) {
			cerr << "Cannot enable : " << conf.getName() << " (" << conf.getType() << ")" << endl;
			throw;
		}
	} else {
		try {
			// Initialise properties
			newTestController->initProperties();
			newTestController->setDisabled();
		}
		catch (...) {
			cerr << "Cannot disable : " << conf.getName() << " (" << conf.getType() << ")" << endl;
			throw;
		}
	}

	return newTestController;

}
