/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "InstrumentManager/InstrumentManager.h"
#include "InstrumentManager/NameManager.h"
#include "InstrumentManager/HardwareManager.h"
#include "InstrumentAbstraction/ExperimentControllerCreator.h"
#include "InstrumentAbstraction/ExperimentController.h"
#include "InstrumentAbstraction/ControllerCommandNames.h"
#include "InstrumentManager/GenericPropertyDescription.h"
#include "Utilities/WebSpyDescription.h"
#include <ics/ServantConfiguration.h>
#include <dataprovider/ics/ServantLoader.h>
#include <dataprovider/database/DatabaseLoader.h>

#undef eol

#include <boost/tokenizer.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/algorithm/string.hpp>
#include <common/base/Convert.h>
#include <core/dataaccess/DataAccess.h>
#include <core/cloner/ServantCloner.h>
#include <dataprovider/notification/NotificationServer.h>
#include "Utilities/Counter.h"

#include <cstdlib>

using namespace ics;
using namespace std;
using namespace common;
using namespace boost;

InstrumentManager* InstrumentManager::_instance = 0;

InstrumentManager::InstrumentManagerError::InstrumentManagerError(const std::string &message) :
		Exception("InstrumentManager::InstrumentManagerError", message) {
}

InstrumentManager::InstrumentManager() :
		_createdControllerCounterID(1) {
}

InstrumentManager::~InstrumentManager() {

	for (map<string, ExperimentController*>::iterator c = _installedExperimentControllers.begin(); c != _installedExperimentControllers.end(); c++) {
		delete c->second;
	}
}

void InstrumentManager::initInstance() {
	if (_instance == 0) {
		_instance = new InstrumentManager();
	}
}

InstrumentManager* InstrumentManager::getInstance() {
	return _instance;
}

void InstrumentManager::reset() {
	delete _instance;
}

void InstrumentManager::loadExperimentControllers() {

	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	vector<ExperimentControllerConfig> experimentControllerConfigs = ServantConfiguration::getInstance()->getExperimentControllerList();

	for (vector<ExperimentControllerConfig>::const_iterator c = experimentControllerConfigs.begin(); c != experimentControllerConfigs.end(); ++c) {
		loadExperimentController(*c);
	}

	initControllerCounterID();

	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
}

void InstrumentManager::loadExperimentControllers(const std::vector<ExperimentControllerConfig>& configuration) {

	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	vector<ExperimentControllerConfig>::const_iterator end = configuration.end();
	for (vector<ExperimentControllerConfig>::const_iterator c = configuration.begin(); c != end; ++c) {
		loadExperimentController(*c);
	}

	initControllerCounterID();

	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
}

ExperimentController* InstrumentManager::loadExperimentController(const ExperimentControllerConfig& config, bool dopostconfig) {

	ExperimentController* newExperimentController = 0;

	string name = config.getName();
	string type = config.getType();

	out << "Loading " << name << " (" << type << ")" << endOfLine;

	// Check that the ExperimentController hasn't already been installed.
	if (NameManager::getInstance()->controllerExists(name)) {
		ostringstream sbuf;
		sbuf << "the " << name << " controller already exists!";
		throw InstrumentManagerError(sbuf.str());
	}

	// Then check if controller Type is a known Type
	if (!ExperimentControllerCreator::getInstance()->verifyType(type)) {
		ostringstream sbuf;
		sbuf << "the " << name << " controller has unknown type!";
		throw InstrumentManagerError(sbuf.str());
	}

	// Create new ExperimentController
	newExperimentController = ExperimentControllerCreator::getInstance()->create(type, name);

	// Init the families
	newExperimentController->setConfigFamily(config.getFamily(), config.getSubfamily());
	newExperimentController->setFavourite(config.isFavourite());

	try {
		// first attempt to set properties, Pass is set to 3 to be more permissive.
		newExperimentController->initProperties(config.getProperties(), 3);

	} catch (...) {
		cerr << "Error first init properties attempt : " << name << " (" << type << ")\n";
		throw;
	}

	// Add the new ExperimentController to the controller list
	_installedExperimentControllers[name] = newExperimentController;
	_orderedInstalledExperimentControllers.push_back(newExperimentController);
	if (dopostconfig == true) {
		if (config.isEnabled() == true) {
			_installedExperimentControllersEnable.push_back(newExperimentController);
		}
		else {
			_installedExperimentControllersDisable.push_back(newExperimentController);
		}
	}
	_installedExperimentControllersRoles[name] = config.getRoles();

	return newExperimentController;
}

void InstrumentManager::configExperimentControllers() {
	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);
	// Do Disable controller First
	for(auto controller : _installedExperimentControllersDisable) {
		configExperimentController(controller, false);
	}
	// Do Enable controller at last
	for(auto controller : _installedExperimentControllersEnable) {
		configExperimentController(controller, true);
	}
	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
}

void InstrumentManager::configExperimentController(ExperimentController* controller, bool enable) {
	common::out << "Configure " << controller->getName() << " (" << controller->getType() << ")" << common::endOfLine;

	string role;

	try {
		// Init the controller roles
		const vector<RoleConfig> roles = _installedExperimentControllersRoles[controller->getName()];
		vector<RoleConfig>::const_iterator rolesIterator;
		for (rolesIterator = roles.begin(); rolesIterator != roles.end(); rolesIterator++) {
			AbstractController* rolecontroller = 0;

			role = rolesIterator->getName();
			string name = rolesIterator->getWho();
			string roletype = rolesIterator->getRoleType();

			if ((name != "not configured") && (name != "<not configured>")) {
				if (roletype.empty() == true) {
					// TO BE removed when xerces version is removed
					roletype = NameManager::getInstance()->getControllerType(name);
				}

				if (roletype == EXPERIMENT_CONTROLLER_TYPE) {
					rolecontroller = getExperimentController(name);
				} else if (roletype == TEST_CONTROLLER_TYPE) {
					rolecontroller = HardwareManager::getInstance()->getTestController(name);
				}

				if (controller) {
					controller->configure(role, rolecontroller);
				} else {
					// Error
				}
			}
			else {
				ostringstream sbuf;
				sbuf << "Role " << role << " isn't configured";
				controller->pushWarning(sbuf.str());
			}
		}

	} catch (...) {
		ostringstream sbuf;
		sbuf << "error init roles attempt : " << controller->getName() << " (" << controller->getType() << ") cannot configure " << role << "\n";
		throw InstrumentManagerError(sbuf.str());
	}

	controller->postConfiguration();
	try {
		controller->formatAllValues();
	} catch (const std::exception& e) {
		err << "Format all values exception: " << e.what() << endOfLine;
	} catch (...) {
		err << "Format all values unknown exception" << endOfLine;
	}

	if (enable == true) {

		try {
			// Do not init the controller here but after having loaded all the controllers.
			controller->setEnabledForLoading(false, false);
		} catch (const std::exception& e) {
			ostringstream sbuf;
			sbuf << "cannot enable : " << controller->getName() << " (" << controller->getType() << ") : " << e.what();
			throw InstrumentManagerError(sbuf.str());
		} catch (...) {
			ostringstream sbuf;
			sbuf << "cannot enable : " << controller->getName() << " (" << controller->getType() << ")";
			throw InstrumentManagerError(sbuf.str());
		}
	} else {
		try {
			controller->setDisabled();
		} catch (...) {
			ostringstream sbuf;
			sbuf << "cannot disable : " << controller->getName() << " (" << controller->getType() << ")";
			throw InstrumentManagerError(sbuf.str());
		}
	}
}

void InstrumentManager::initExperimentControllers() {
	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);
	for(auto controller : _installedExperimentControllersEnable) {
		if (controller->containsCommand(INIT_COMMAND)) {
			out << "Initialising " << controller->getName() << " (" << controller->getType() << ")" << common::endOfLine;
			controller->execute(INIT_COMMAND, false, true);
		}
	}
	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
}

void InstrumentManager::loadReports() {
	for(auto controller : _orderedInstalledExperimentControllers) {
		const vector<string>& warnings = controller->getWarnings();
		const vector<string>& errors = controller->getErrors();
		if ((warnings.empty() == false) || (errors.empty() == false)) {
			ostringstream sbuf;
			common::title << "Report for controller: " << controller->getName() << " (" << controller->getType() << ")" << common::endOfLine;
			for(auto w : warnings) {
				common::warn << w << common::endOfLine;
			}
			for(auto e : errors) {
				common::err << e << common::endOfLine;
			}
		}
	}
}

ExperimentController* InstrumentManager::getExperimentController(const std::string& controllerName)  {

	if (!controllerExists(controllerName)) {
		throw InstrumentManagerError(controllerName);
	}

	return _installedExperimentControllers[controllerName];
}

bool InstrumentManager::controllerExists(const std::string& controllerName) const {

	for (map<string, ExperimentController*>::const_iterator c = _installedExperimentControllers.begin(); c != _installedExperimentControllers.end(); c++) {
		if (c->first == controllerName) {
			return true;
		}
	}

	return false;
}

bool InstrumentManager::controllerIDExists(const std::string& ID) const {
	return false;
}

std::string InstrumentManager::getPropertiesForGeneric() const {

	string result = "";

	for(auto ecIterator : _installedExperimentControllers) {
		ExperimentController* currentController = ecIterator.second;
		if (currentController->getName().find(".clone") == string::npos) {
			if (currentController->isEnabled()) {
				string properties = currentController->getGenericAliasProperties();
				if (!properties.empty()) {
					result.append(currentController->getName() + "," + properties + "\n");
				}
			}
		}
	}

	return result;
}

std::string InstrumentManager::getWebSpyProperties() const {

	WebSpyDescription theDescription;

	// Order in type of controller
	for(auto iter1 : _installedExperimentControllers) {
		ExperimentController* currentController = iter1.second;
		if (currentController->isEnabled() == true) {
			string controllertype = currentController->getType();
			string controllername = currentController->getName();
			if (controllername.find(".clone") == string::npos) {
				vector<AnyProperty*> properties = currentController->getSpyProperties();
				if (properties.empty() == false) {
					vector<AnyProperty*>::iterator iter2 = properties.begin();
					while (iter2 != properties.end()) {
						AnyProperty* p = *iter2;
						string propertytype = p->getType();
						string propertyname = p->getName();
						ostringstream svalue;
						bool ok = true;
						try {
							if (propertytype == UINT16_PROPERTY) {
								svalue << any_cast<uint16>(p->getValue());
							} else if (propertytype == UINT32_PROPERTY) {
								svalue << any_cast<uint32>(p->getValue());
							} else if (propertytype == UINT64_PROPERTY) {
								svalue << any_cast<uint64>(p->getValue());
							} else if (propertytype == INT16_PROPERTY) {
								svalue << any_cast<int16>(p->getValue());
							} else if (propertytype == INT32_PROPERTY) {
								svalue << any_cast<int32>(p->getValue());
							} else if (propertytype == INT64_PROPERTY) {
								svalue << any_cast<int64>(p->getValue());
							} else if (propertytype == FLOAT32_PROPERTY) {
								svalue << any_cast<float32>(p->getValue());
							} else if (propertytype == FLOAT64_PROPERTY) {
								svalue << any_cast<float64>(p->getValue());
							} else if (propertytype == BOOLEAN_PROPERTY) {
								svalue << any_cast<bool>(p->getValue());
							} else if (propertytype == STRING_PROPERTY) {
								string buf = any_cast<string>(p->getValue());
								if (buf.empty() == true) {
									svalue << " ";
								} else {
									svalue << any_cast<string>(p->getValue());
								}
							} else {
								ok = false;
							}
						}
						catch(...) {
							ok = false;
						}
						if (ok == true) {
							theDescription.add(controllertype, controllername, propertyname, svalue.str());
						}
						++iter2;
					}
				}
			}
		}
	}

	ostringstream sbuf;
	boost::archive::text_oarchive oa(sbuf);
	oa << theDescription;
	return sbuf.str();
}

vector<GenericPropertyDescription> InstrumentManager::getGenericPropertiesList() const {
	vector<GenericPropertyDescription> buf;
	for (auto ecIterator : _installedExperimentControllers) {
		ExperimentController* currentEC = ecIterator.second;
		if (currentEC->getName().find(".clone") == string::npos) {
			if (currentEC->isEnabled()) {
				string tmp = currentEC->getAliasProperties();
				vector<string> items;
				split(items, tmp, is_any_of(","), token_compress_on);
				if (items.empty() == false) {
					vector<string>::iterator iter = items.begin();
					while (iter != items.end()) {
						string pname = *iter;
						++iter;
						if (iter != items.end()) {
							string palias = *iter;
							GenericPropertyDescription gp(palias, currentEC, pname);
							buf.push_back(gp);
							++iter;
						}
					}
				}
			}
		}
	}
	return buf;
}

ExperimentControllerConfig InstrumentManager::getExperimentControllerConfig(ExperimentController* experimentController, int32 all) {
	// Filling the ExperimentControllerConfig from the attributes of the ExperimentController
	ExperimentControllerConfig result;

	result.setName(experimentController->getName());
	result.setType(experimentController->getType());
	result.setEnabled(experimentController->isEnabled());

	// Add families
	result.setFamily(experimentController->getConfigFamily());
	result.setSubfamily(experimentController->getConfigSubfamily());
	result.setFavourite(experimentController->isFavourite());

	// Add roles
	for (auto p : experimentController->getConfigurationRoles()) {
		RoleConfig role(p.second.second, p.first, p.second.first);
		result.addRole(role);
	}

	// Setting properties
	vector<AnyProperty *> properties;
	if (all == 2) {
		properties = experimentController->getAllAllProperties();
	} else if (all == 1) {
		properties = experimentController->getAllProperties();
	} else if (all == 0) {
		properties = experimentController->getPropertiesToSave();
	}
	for (vector<AnyProperty *>::const_iterator p = properties.begin(); p != properties.end(); ++p) {
		ostringstream value;
		string type = (*p)->getType();
		value << static_cast<AnyProperty>(*(*p));
		if ((type == UINT16_PROPERTY) || (type == INT16_PROPERTY) || (type == UINT32_PROPERTY) || (type == INT32_PROPERTY)
				|| (type == UINT64_PROPERTY) || (type == INT64_PROPERTY) || (type == FLOAT32_PROPERTY)
				|| (type == FLOAT64_PROPERTY) || (type == STRING_PROPERTY) || (type == BOOLEAN_PROPERTY)) {
			result.addProperty(PropertyConfig((*p)->getName(), value.str()));
		} else if ((type == INT16ARRAY_PROPERTY) || (type == INT32ARRAY_PROPERTY) || (type == INT64ARRAY_PROPERTY)
				|| (type == UINT16ARRAY_PROPERTY) || (type == UINT32ARRAY_PROPERTY) || (type == UINT64ARRAY_PROPERTY)
				|| (type == FLOAT64ARRAY_PROPERTY) || (type == FLOAT32ARRAY_PROPERTY)) {
			typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
			boost::char_separator<char> sep("|");
			string buf = value.str();
			tokenizer tokens(buf, sep);
			vector<string> array;
			for (tokenizer::iterator tok_iter = tokens.begin(); tok_iter != tokens.end(); ++tok_iter) {
				array.push_back(*tok_iter);
			}
			int32 extrasize = experimentController->getPropertyManager()->getExtraSize((*p)->getName());
			result.addProperty(PropertyConfig((*p)->getName(), array, extrasize));
		}
	}

	return result;
}

void InstrumentManager::saveExperimentControllers(bool isBackup) {

	// Create a vector of the configs
	vector<ExperimentControllerConfig> configs;
	for (auto i : _orderedInstalledExperimentControllers) {
		if (i->isSaved()) {
			configs.push_back(getExperimentControllerConfig(i, 0));
		}
	}

	ServantConfiguration::getInstance()->saveExperimentControllerList(configs, isBackup);
}

map<string, ExperimentController*> InstrumentManager::getAllInstalledExperimentControllersMap(bool useclone) const {
	map<string, ExperimentController*> themap;
	for(auto ecIterator : _installedExperimentControllers) {
		ExperimentController* currentEC = ecIterator.second;
		if (useclone == true) {
			if (currentEC->getName().find(".clone") != string::npos) {
				themap[currentEC->getName()] = currentEC;
			}
		}
		else {
			if (currentEC->getName().find(".clone") == string::npos) {
				themap[currentEC->getName()] = currentEC;
			}
		}
	}
	return themap;
}

map<string, ExperimentController*> InstrumentManager::getInstalledExperimentControllersMap() const {
	map<string, ExperimentController*> themap;
	for(auto ecIterator : _installedExperimentControllers) {
		ExperimentController* currentEC = ecIterator.second;
		if (currentEC->getName().find(".clone") == string::npos) {
			if (currentEC->isEnabled()) {
				themap[currentEC->getName()] = currentEC;
			}
		}
	}
	return themap;
}

map<string, ExperimentController*> InstrumentManager::getInstalledExperimentControllersNicknameMap() const {
	map<string, ExperimentController*> themap;
	for(auto ecIterator : _installedExperimentControllers) {
		ExperimentController* currentEC = ecIterator.second;
		if (currentEC->getName().find(".clone") == string::npos) {
			if (currentEC->isEnabled()) {
				string name = any_cast<string>(currentEC->getValue("nickname"));
				if (name.empty() == false) {
					themap[name] = currentEC;
				}
			}
		}
	}
	return themap;
}

ExperimentController*InstrumentManager::getFirstExperimentControllerByType(const string& controllerType, bool useclone)  {

	ExperimentController* wantedExperimentController = NULL;
	for(auto ecIterator : _installedExperimentControllers) {
		ExperimentController* currentEC = ecIterator.second;
		if (useclone) {
			if (currentEC->getName().find(".clone") != string::npos) {
				if ((currentEC->getType() == controllerType) && (currentEC->isEnabled() == true)) {
					wantedExperimentController = currentEC;
					break;
				}
			}
		}
		else {
			if (currentEC->getName().find(".clone") == string::npos) {
				if ((currentEC->getType() == controllerType) && (currentEC->isEnabled() == true)) {
					wantedExperimentController = currentEC;
					break;
				}
			}
		}
	}
	return wantedExperimentController;
}

ExperimentController*InstrumentManager::getFirstExperimentControllerByFunction(const string& aFunction)  {

	ExperimentController* wantedExperimentController = NULL;
	for(auto ecIterator : _installedExperimentControllers) {
		ExperimentController* currentEC = ecIterator.second;
		if (currentEC->getName().find(".clone") == string::npos) {
			if ((currentEC->containsFunction(aFunction) == true) && (currentEC->isEnabled() == true)) {
				wantedExperimentController = currentEC;
				break;
			}
		}
	}
	return wantedExperimentController;
}

std::string InstrumentManager::getExperimentControllersConfig() {
	// Create a vector of the configs
	vector<ExperimentControllerConfig> configs;
	for (auto i : _orderedInstalledExperimentControllers) {
		configs.push_back(getExperimentControllerConfig(i, 1));
	}

	ostringstream buf;
	boost::archive::text_oarchive oa(buf);
	oa << configs;
	return buf.str();
}

vector<string> InstrumentManager::getControllerTypes() {

	return ExperimentControllerCreator::getInstance()->getKnownTypes();
}

vector<string> InstrumentManager::getControllerRoles(const std::string& type) {

	vector<string> v;
	ExperimentController* experimentController = ExperimentControllerCreator::getInstance()->create(type, "tmp", "tc-0");

	if (experimentController != NULL) {
		map<string, string> roles = experimentController->getRequiredFunctions();

		for (map<string, string>::iterator it = roles.begin(); it != roles.end(); it++) {
			v.push_back(it->first);
			v.push_back(it->second);
		}

		delete experimentController;
	}
	return v;
}

vector<string> InstrumentManager::getControllersWithType(const std::string& type) {

	vector<string> result;
	for (map<string, ExperimentController*>::iterator f = _installedExperimentControllers.begin(); f != _installedExperimentControllers.end(); f++) {

		ExperimentController* currentEC = f->second;
		if (currentEC->containsFunction(type)) {
			result.push_back(currentEC->getName());
		}
	}
	return result;
}

ExperimentController* InstrumentManager::createController(const std::string& type, const std::string& name, const std::string& family, const std::string& subfamily, bool isCommand, const std::map<std::string, std::pair<std::string, std::string>>& roleMap) {
	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	ExperimentController * controller = nullptr;
	ExperimentControllerConfig conf(name, type);
	conf.setFamily(family);
	conf.setSubfamily(subfamily);
	conf.setEnabled(true);

	// iterating the role map
	for (auto r : roleMap) {
		conf.addRole(RoleConfig(r.second.second, r.first, r.second.first));
	}

	ics::ServantLoader servantLoader;
	dataprovider::DatabaseLoader loader;
	dataprovider::ControllerElement* element = servantLoader.createController(conf);
	unsigned long controllerID = loader.createController(element);
	loader.connectControllerServant(element);
	controller = loadExperimentController(conf, false);

	// setting isCommand
	controller->setValue(IS_COMMAND_PROPERTY, isCommand);

	configExperimentController(controller, true);

	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
	dataprovider::NotificationServer::getInstance()->configurationChanged(0, controllerID, common::CONTROLLER_ADDED);

	// The controller is created
	_createdControllerCounterID++;
	return controller;
}

void InstrumentManager::destroyController(const std::string& name) {
	//dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	dataprovider::DatabaseLoader loader;
	loader.destroyController(name);

	auto c = _installedExperimentControllers.find(name);
	if (c != _installedExperimentControllers.end()) {
		auto iter = _orderedInstalledExperimentControllers.begin();
		while(iter != _orderedInstalledExperimentControllers.end()) {
			if (*iter == c->second) {
				break;
			}
			++iter;
		}
		if (iter != _orderedInstalledExperimentControllers.end()) {
			_orderedInstalledExperimentControllers.erase(iter);
		}
		delete c->second;
		_installedExperimentControllers.erase(c);
		_installedExperimentControllersRoles.erase(name);
	}

	//dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
	//dataprovider::NotificationServer::getInstance()->configurationChanged(0, 0, common::TREE);

	// Let the Database send the configuration changed notification for a controller removed.
}

void InstrumentManager::renameController(const std::string& oldname, const std::string& newname) {
	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	dataprovider::DatabaseLoader loader;
	unsigned long controllerID = loader.renameController(oldname, newname);

	auto iec = _installedExperimentControllers.extract(oldname);
	iec.key() = newname;
	ExperimentController* ec = iec.mapped();
	ec->setName(newname);
	_installedExperimentControllers.insert(std::move(iec));
	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
	dataprovider::NotificationServer::getInstance()->configurationChanged(0, controllerID, common::CONTROLLER_ADDED);
}

ExperimentController* InstrumentManager::cloneController(const std::string& controllerName) {

	map<string, ExperimentController*>::iterator c = _installedExperimentControllers.find(controllerName);
	if (c != _installedExperimentControllers.end()) {
		ExperimentController* clone = c->second->clone();
		clone->setSaved(false);
		clone->setFamily(family::HIDDEN);
		_installedExperimentControllers[clone->getName()] = clone;

		return clone;
	}

	return 0;
}

void InstrumentManager::removeController(const std::string& controllerName) {

	map<string, ExperimentController*>::iterator c = _installedExperimentControllers.find(controllerName);
	if (c != _installedExperimentControllers.end()) {
		delete c->second;
		_installedExperimentControllers.erase(c);
	}
}

void InstrumentManager::initControllers(const std::vector<ExperimentControllerConfig>& configuration) {

	vector<ExperimentControllerConfig>::const_iterator end = configuration.end();
	for (vector<ExperimentControllerConfig>::const_iterator c = configuration.begin(); c != end; ++c) {

		string controllerName = c->getName();
		if (_installedExperimentControllers.find(controllerName) != _installedExperimentControllers.end()) {
			ExperimentController* controller = _installedExperimentControllers[controllerName];
			if (controller->isEnabled()) {
				if (controller->containsCommand(INIT_COMMAND)) {
					out << "initialising " << controllerName << " (" << controller->getType() << ")" << endOfLine;
					controller->execute(INIT_COMMAND, false, true);
				}
			}
		}
	}
}

void InstrumentManager::initControllerCounterID() {
	_createdControllerCounterID += 1;
}


void InstrumentManager::refreshExperimentControllers(const std::vector<ExperimentControllerConfig>& configuration) {

	dataprovider::NotificationServer::getInstance()->disableConfigurationChanged(0);

	vector<ExperimentControllerConfig>::const_iterator end = configuration.end();
	for (vector<ExperimentControllerConfig>::const_iterator c = configuration.begin(); c != end; ++c) {
		refreshExperimentController(*c);
	}

	initControllers(configuration);

	dataprovider::NotificationServer::getInstance()->enableConfigurationChanged(0);
	dataprovider::NotificationServer::getInstance()->configurationChanged(0, 0, common::TREE);
}

ExperimentController* InstrumentManager::refreshExperimentController(const ExperimentControllerConfig& config, bool dopostconfig) {

	string name = config.getName();
	string type = config.getType();

//	cout << "refreshing " << name << " (" << type << ")\n";

	ExperimentController* newExperimentController = _installedExperimentControllers[name];

	try {
		// first attempt to set properties, Pass is set to 3 to be more permissive.
		newExperimentController->initProperties(config.getProperties(), 3);
	} catch (...) {
		err << "Error first init properties attempt : " << name << " (" << type << ")" << endOfLine;
		throw;
	}
	string role;

	try {
		if (dopostconfig) {
			newExperimentController->postConfiguration();
		}
		try {
			newExperimentController->formatAllValues();
		} catch (...) {
			warn << "Format all values exception" << endOfLine;
		}

	} catch (...) {
		err << "Error init roles attempt : " << name << " (" << type << ") cannot configure " << role << endOfLine;
		throw;
	}

	if (config.isEnabled()) {

		try {
			// Do not init the controller here but after having loaded all the controllers.
			newExperimentController->setEnabledForLoading(false, false);

		} catch (...) {
			err << "Cannot enable : " << name << " (" << type << ")" << endOfLine;
			throw;
		}

	} else {
		try {
			newExperimentController->setDisabled();
		} catch (...) {
			err << "Cannot disable : " << name << " (" << type << ")" << endOfLine;
			throw;
		}
	}

	return newExperimentController;
}

