/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef EXPERIMENTCONTROLMANAGER_H
#define EXPERIMENTCONTROLMANAGER_H

#include <sys/types.h>
#include <signal.h>
#include <string>
#include <map>
#include <vector>

#include "InstrumentAbstraction/ExperimentControllerCreator.h"
#include "InstrumentAbstraction/ExperimentController.h"

class GenericPropertyDescription;
class ExperimentController;
class ExperimentControllerConfig;
class RoleConfig;

class InstrumentManager {

public:
	class InstrumentManagerError : public common::Exception {
	public:
		/**
		 * Constructor.
		 */
		InstrumentManagerError(const std::string& message);
	};

	static void initInstance();
	static InstrumentManager* getInstance();
	static void reset();

	void loadExperimentControllers();
	void configExperimentControllers();
	void initExperimentControllers();
	void loadReports();
	void loadExperimentControllers(const std::vector<ExperimentControllerConfig>& configuration);
	void saveExperimentControllers(bool isBackup);
	ExperimentController* loadExperimentController(const ExperimentControllerConfig& config, bool dopostconfig = true);
	void refreshExperimentControllers(const std::vector<ExperimentControllerConfig>& configuration);

	ExperimentController* getExperimentController(const std::string& controllerName) ;
	ExperimentController* getFirstExperimentControllerByType(const std::string& controllerType, bool useclone) ;
	ExperimentController* getFirstExperimentControllerByFunction(const std::string& aFunction) ;
	bool controllerExists(const std::string& controllerName) const;
	bool controllerIDExists(const std::string& ID) const;
	std::string getPropertiesForGeneric() const;
	std::string getWebSpyProperties() const;
	std::vector<GenericPropertyDescription> getGenericPropertiesList() const;
	std::map<std::string, ExperimentController*> getInstalledExperimentControllersMap() const;
	std::map<std::string, ExperimentController*> getAllInstalledExperimentControllersMap(bool useclone) const;
	std::map<std::string, ExperimentController*> getInstalledExperimentControllersNicknameMap() const;
	const std::map<std::string, ExperimentController*>& getAllExperimentControllers() const;

	std::string getExperimentControllersConfig();
	std::vector<std::string> getControllerTypes();
	std::vector<std::string> getControllerRoles(const std::string& type);
	std::vector<std::string> getControllersWithType(const std::string& type);

	void initControllers(const std::vector<ExperimentControllerConfig>& configuration);
	void initControllerCounterID();
	ExperimentController * createController(const std::string& type, const std::string& name, const std::string& family, const std::string& subfamily, bool isCommand, const std::map<std::string, std::pair<std::string, std::string>>& roleMap);
	void destroyController(const std::string& name);
	void renameController(const std::string& oldname, const std::string& newname);
	ExperimentController* cloneController(const std::string& controllerName);
	void removeController(const std::string& controllerName);

	ExperimentController* refreshExperimentController(const ExperimentControllerConfig& config, bool dopostconfig = true);
	ExperimentControllerConfig getExperimentControllerConfig(ExperimentController* experimentController, int32 all);

private:
	InstrumentManager();
	~InstrumentManager();

	void configExperimentController(ExperimentController* controller, bool enable);

	static InstrumentManager* _instance;

	std::vector<ExperimentController*> _orderedInstalledExperimentControllers;
	std::map<std::string, ExperimentController*> _installedExperimentControllers;
	std::vector<ExperimentController*> _installedExperimentControllersEnable;
	std::vector<ExperimentController*> _installedExperimentControllersDisable;
	std::map<std::string, std::vector<RoleConfig>> _installedExperimentControllersRoles;
	int32 _createdControllerCounterID;
};

inline const std::map<std::string, ExperimentController*>& InstrumentManager::getAllExperimentControllers() const {
	return _installedExperimentControllers;
}

#endif
