/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Log.h"
#include "InstrumentManager/NameManager.h"
#include "InstrumentAbstraction/AbstractController.h"
#include "DataProvider/Property/PropertyManager.h"
#include "LoggingServices/Level.h"
#include <command/log/LogHandler.h>
#include <command/log/BaseLog.h>
#include <command/log/DatabaseLog.h>
#include <command/log/SimpleLog.h>
#include <common/output/OutputStream.h>

using namespace common;

Log * Log::_instance = 0;

Log::Log() {

}

Log::~Log() {

}

Log * Log::getInstance() {

	if (_instance == 0) {
		_instance = new Log();
	}

	return _instance;
}

void Log::reset() {

}

common::LogLevel convertLevel(Level const & level) {

	if (level.getLevel() == 0) {
		return common::LOG_INFO;
	} else if (level.getLevel() == 1) {
		return common::LOG_WARNING;
	} else if (level.getLevel() == 2) {
		return common::LOG_ERROR;
	} else if (level.getLevel() == 3) {
		return common::LOG_DEBUG;
	} else if (level.getLevel() == 4) {
		return common::LOG_SYNC;
	}

	return common::LOG_INFO;
}
//Log method used for "NotOk" messages (from Exceptions messages class)
void Log::log(std::string const & controllerType, std::string const & controllerName, Level const & level, std::string const & message) {

	std::list<LogElement> lineElements;
	lineElements.push_back(LogElement(false, false, "", "", "", "", message));
	std::string imageName = "null";
	log(controllerType, controllerName, level, lineElements, imageName);

}

//Log method used in new controller log messages
void Log::log(std::string const & controllerType, std::string const & controllerName, Level const & level, std::list<LogElement> const & lineElements, std::string const & imageName) {

	command::DatabaseLog dblog(common::SINGLE, convertLevel(level), lineElements, imageName);

	try {
		dblog.setServantID(NameManager::getInstance()->getController(controllerName)->getPropertyManager()->getServantID());
		command::LogHandler::getInstance()->write(dblog, !Simulated::activated && !Simulated::visa && !Simulated::for3d);
		// Sending error
		if (level.getLevel() == 2) {
			err << getErrorMessage(lineElements) << eol;
		}
	} catch (const NameManager::ControllerNotFound & e) {
		// do nothing
	}
}


//Text messages from Command line and both ascii and nexus writer
void Log::log(Level const & level, std::string const & message) {

	command::SimpleLog simpleLog(common::SINGLE, convertLevel(level), message);

	// do not set commandID
	// do not use address
	command::LogHandler::getInstance()->write(simpleLog, !Simulated::activated && !Simulated::visa && !Simulated::for3d);

	// Sending error
	if (level.getLevel() == 2) {
		err << message << eol;
	}

}

void Log::log(Level const & level, std::string const & propertyName, std::string const & propertyType, int32 numor, std::string const & imageName) {
	std::list<LogElement> lineElements;
	std::stringstream numorValue;
	numorValue << numor;

	lineElements.push_back(LogElement(false, false, propertyName, propertyType, "", "", numorValue.str()));
	command::DatabaseLog dblog(common::SINGLE, convertLevel(level), lineElements, imageName);
	command::LogHandler::getInstance()->write(dblog, !Simulated::activated && !Simulated::visa && !Simulated::for3d);
}

/**
 * Notify to NomadGUI (offscreen generator) that an image is ready to be rendered
 */
void Log::imageDataReadyLog(std::string const & image) {
	command::LogHandler::getInstance()->imageDataReady(image);
}

std::string Log::getErrorMessage(std::list<LogElement> const & lineElements) {
	//The error message is the output message (value + unit)
	std::string errorMessage;
	for (std::list<LogElement>::const_iterator it = lineElements.begin(); it != lineElements.end(); it++) {
			LogElement logElement = *it;
			errorMessage += logElement.getValue() + logElement.getUnit();
	}
	return errorMessage;
}
