/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Module.h"
#include "ModuleManager.h"
#include <dataprovider/iconandfamily/ServantIconAndFamilyMaps.h>

using namespace std;


const std::string Module::DIRECT = "direct";
const std::string Module::INDIRECT = "indirect";
const std::string Module::SKIP = "skip";

Module::Module(const boost::filesystem::path & path, const std::string & name, void * handle) :
	m_path(path), m_name(name), m_handle(handle), m_testPtr(0) {
}

Module::~Module() {
	reset();
}

void Module::reset() {
	for (map<string, BaseControllerFactory *>::iterator f = m_controllerFactories.begin(); f != m_controllerFactories.end(); ++f) {
		delete f->second;
	}
	for (map<string, BaseDriverFactory *>::iterator f = m_driverFactories.begin(); f != m_driverFactories.end(); ++f) {
		delete f->second;
	}
	for (map<string, BaseParserFactory *>::iterator f = m_parserFactories.begin(); f != m_parserFactories.end(); ++f) {
		delete f->second;
	}
}

void Module::setDependency(const std::string& value) {
	m_dependency = value;
}

const std::string & Module::getDependency() const {
	return m_dependency;
}

const boost::filesystem::path & Module::getPath() const {
	return m_path;
}

const std::string & Module::getName() const {
	return m_name;
}

void * Module::getHandle() const {
	return m_handle;
}

std::vector<std::string> Module::getControllerTypes() const {

	vector<string> result;

	for (map<string, BaseControllerFactory *>::const_iterator f = m_controllerFactories.begin(); f != m_controllerFactories.end(); ++f) {
		result.push_back(f->first);
	}

	return result;
}

bool Module::findControllerClass(const std::string & type) const {
	return m_controllerFactories.find(type) != m_controllerFactories.end();
}

ExperimentController * Module::makeController(const std::string & type, const std::string & name) {
	map<string, BaseControllerFactory *>::iterator f = m_controllerFactories.find(type);

	if (f != m_controllerFactories.end()) {
		return f->second->make(name);
	}

	return 0;
}

ExperimentController * Module::copyController(const ExperimentController * controller) {
	map<string, BaseControllerFactory *>::const_iterator f = m_controllerFactories.find(controller->getType());

	if (f != m_controllerFactories.end()) {
		return f->second->copy(controller);
	}

	return 0;
}

std::vector<std::string> Module::getDriverTypes() const {

	vector<string> result;

	for (map<string, BaseDriverFactory *>::const_iterator f = m_driverFactories.begin(); f != m_driverFactories.end(); ++f) {
		result.push_back(f->first);
	}

	return result;
}

bool Module::findDriverClass(const std::string & type) const {
	return m_driverFactories.find(type) != m_driverFactories.end();
}

bool Module::findRealDriverClass(const std::string & type) const {
	return m_realDriverFactories.find(type) != m_realDriverFactories.end();
}

DeviceDriver * Module::makeDriver(const std::string & type, const std::string & name) {
	map<string, BaseDriverFactory *>::iterator f = m_driverFactories.find(type);

	if (f != m_driverFactories.end()) {
		return f->second->make(name);
	}

	return 0;
}

DeviceDriverState * Module::makeRealDriverState(const std::string & type, DeviceDriver * owner) {

	map<string, BaseRealDriverFactory *>::iterator f = m_realDriverFactories.find(type);

	if (f != m_realDriverFactories.end()) {
		return f->second->make(owner);
	}

	return 0;
}

bool Module::findParserClass(const std::string & type) const {
	return m_parserFactories.find(type) != m_parserFactories.end();
}

AbstractParser * Module::makeParser(const std::string & type) {
	map<string, BaseParserFactory *>::iterator f = m_parserFactories.find(type);

	if (f != m_parserFactories.end()) {
		return f->second->make();
	}

	return 0;
}

void Module::setTestFunction(bool (*testPtr)()) {
	m_testPtr = testPtr;
}

bool Module::hasTestFunction() const {
	return (m_testPtr != 0);
}

bool Module::test() {

	if (m_testPtr != 0) {
		return (*m_testPtr)();
	}

	return true;
}

std::string Module::getResourcePath(const std::string& fileName) {
	return ModuleManager::getInstance()->getResourceFilePath(fileName);
}

void Module::loadFamilyImageKeys(const std::string& fileName) {
	dataprovider::ServantIconAndFamilyMaps::getInstance()->readFamilyImageKeys(ModuleManager::getInstance()->getResourceFilePath(fileName));
}
