/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef MODULE_H_
#define MODULE_H_

#include <string>
#include <iostream>
#include <map>
#include <vector>
#include "ControllerFactory.h"
#include "DriverFactory.h"
#include "RealDriverFactory.h"
#include "ParserFactory.h"

template<typename T>
struct ControllerClass {};

template<typename T>
struct DriverClass {};

template<typename T, typename RT>
struct RealDriverClass {};

template<typename T>
struct ParserClass {};

class Module {

public:
	Module(const boost::filesystem::path & path, const std::string & name, void * handle);
	~Module();

	static const std::string DIRECT;
	static const std::string INDIRECT;
	static const std::string SKIP;

	void reset();
	void setDependency(const std::string& value);
	const std::string & getDependency() const;

	const boost::filesystem::path & getPath() const;
	const std::string & getName() const;
	void * getHandle() const;

	template<typename Type>
	Module & operator<< (const ControllerClass<Type> & controller) {

		if (m_controllerFactories.find(Type::TYPE) != m_controllerFactories.end()) {
			std::cerr << "error, controller " << Type::TYPE << " already registered" << std::endl;
		}

		m_controllerFactories[Type::TYPE] = new ControllerFactory<Type>();

		return *this;
	}

	template<typename Type>
	Module & operator<< (const DriverClass<Type> & driver) {

		if (m_driverFactories.find(Type::TYPE) != m_driverFactories.end()) {
			std::cerr << "error, driver " << Type::TYPE << " already registered" << std::endl;
		}

		m_driverFactories[Type::TYPE] = new DriverFactory<Type>();

		return *this;
	}

	template<typename Type, typename RealType>
	Module & operator<< (const RealDriverClass<Type, RealType> & driver) {

		if (m_realDriverFactories.find(Type::TYPE) != m_realDriverFactories.end()) {
			std::cerr << "error, real driver " << Type::TYPE << " already registered" << std::endl;
		}

		m_realDriverFactories[Type::TYPE] = new RealDriverFactory<Type, RealType>();

		return *this;
	}

	template<typename Type>
	Module & operator<< (const ParserClass<Type> & parser) {

		if (m_parserFactories.find(Type::TYPE) != m_parserFactories.end()) {
			std::cerr << "error, parser " << Type::TYPE << " already registered" << std::endl;
		}

		m_parserFactories[Type::TYPE] = new ParserFactory<Type>();

		return *this;
	}

	std::vector<std::string> getControllerTypes() const;
	bool findControllerClass(const std::string & type) const;
	ExperimentController * makeController(const std::string & type, const std::string & name);
	ExperimentController * copyController(const ExperimentController * controller);

	std::vector<std::string> getDriverTypes() const;
	bool findDriverClass(const std::string & type) const;
	bool findRealDriverClass(const std::string & type) const;

	DeviceDriver * makeDriver(const std::string & type, const std::string & name);
	DeviceDriverState * makeRealDriverState(const std::string & type, DeviceDriver * owner);

	bool findParserClass(const std::string & type) const;
	AbstractParser * makeParser(const std::string & type);

	void setTestFunction(bool (*testPtr)());
	bool hasTestFunction() const;
	bool test();

	/**
	 * Returns the absolute path of the file filename.
	 * The function may throw an exception if there are multiple or zero files with the same name.
	 * In that case, provide the relative module path.
	 */
	static std::string getResourcePath(const std::string& fileName);

	/**
	 * Loads the family image keys.
	 * The function may throw an exception if there are multiple or zero files with the same name.
	 */
	static void loadFamilyImageKeys(const std::string& fileName);

private:
	boost::filesystem::path m_path;
	std::string m_name;
	void * m_handle;
	std::map<std::string, BaseControllerFactory *> m_controllerFactories;
	std::map<std::string, BaseDriverFactory *> m_driverFactories;
	std::map<std::string, BaseRealDriverFactory *> m_realDriverFactories;
	std::map<std::string, BaseParserFactory *> m_parserFactories;
	bool (*m_testPtr)();
	std::string m_dependency;
};

#define EXPORT_MODULE extern "C"

#endif
