/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "ModuleManager.h"
#include "LoadSequenceParser.h"
#include "DataProvider/Property/PropertyInfos.h"
#include <common/base/Date.h>
#include <common/base/FileUtilities.h>
#include <common/base/ServerProperties.h>
#include <common/output/Color.h>
#include <boost/filesystem/operations.hpp>
#include <dlfcn.h>
#include <fstream>
#include <algorithm>
#include <time.h>
#include "Utilities/Counter.h"

using namespace common;
using namespace std;
namespace fs = boost::filesystem;

ModuleManager * ModuleManager::m_instance = 0;
const std::string ModuleManager::LIB_PREFIX = "lib";
const std::string ModuleManager::INIT_FUNCTION_NAME = "init";
const std::string ModuleManager::TEST_FUNCTION_NAME = "test";
const std::string ModuleManager::INIT_MODULE_FUNCTION_NAME = "initModule";
const std::string ModuleManager::PROPERTY_FILE_NAME = "Properties.xml";
const std::string ModuleManager::DEVICE_PROPERTY_FILE_NAME = "DeviceProperties.xml";
const std::string ModuleManager::CONTROLLER_PROPERTY_FILE_NAME = "ControllerProperties.xml";
const std::string ModuleManager::SCANNABLE_PROPERTY_FILE_NAME = "ScannableProperties.xml";
const std::string ModuleManager::MODULE_LOAD_SEQUENCE_FILE_NAME = "LoadSequence.xml";
const std::string ModuleManager::MODULE_LOG_FILE_NAME = "minconfig.log";

AmbiguousFile::AmbiguousFile(std::string const & fileName)
: Exception("ModuleManager::AmbiguousFile", string("ambiguous file ") + fileName) {
}

NoSuchFile::NoSuchFile(std::string const & fileName)
: Exception("NoSuchFile", string("no file ") + fileName) {
}

ModuleManager::NoSuchControllerClass::NoSuchControllerClass(std::string const & type)
: Exception("ModuleManager::NoSuchControllerClass", string("no controller class for ") + type) {
}

ModuleManager::NoSuchDriverClass::NoSuchDriverClass(std::string const & type)
: Exception("ModuleManager::NoSuchDriverClass", string("no driver class for ") + type) {
}

ModuleManager::NoSuchParserClass::NoSuchParserClass(std::string const & type)
: Exception("ModuleManager::NoSuchParserClass", string("no parser class for ") + type) {
}

ModuleManager::NoSuchModule::NoSuchModule(std::string const & name)
: Exception("ModuleManager::NoSuchModule", string("no module ") + name) {
}

ModuleManager::NoTestFunction::NoTestFunction(std::string const & name)
: Exception("ModuleManager::NoTestFunction", string("no test function for ") + name) {
}

ModuleManager::NoSequenceFile::NoSequenceFile(std::string const & name)
: Exception("ModuleManager::NoSequenceFile", string("no sequence file ") + name) {
}

ModuleManager * ModuleManager::getInstance() {
	return m_instance;
}

void ModuleManager::initInstance() {

	if (m_instance == 0) {
		m_instance = new ModuleManager();
	}
}

void ModuleManager::resetInstance() {
	delete m_instance;
	m_instance = 0;
}

ModuleManager::ModuleManager() : m_sequenceId(0) {
}

ModuleManager::~ModuleManager() {

	for (list<Module *>::iterator m = m_modules.begin(); m != m_modules.end(); ++m) {

		Module * module = *m;
		
		// unload the library and delete the module object
		void * handle = module->getHandle();

		delete module;
		dlclose(handle);

	}
}

bool ModuleManager::loadModulesForOrdering(const std::string& directoryPath) {

	// first load the file
	loadModuleLibrariesFromSequence(directoryPath, false, false);
	vector<fs::path> loadedModuleLibs = getLibraryPathList();

	vector<fs::path> moduleLibs;

	try {
		moduleLibs = FileUtilities::getFilePathsWithExtension(directoryPath, "so");

	} catch (const FileUtilities::NoSuchDirectoryException & e) {
		cerr << "Cannot load modules from " << directoryPath << endl;
		return true;
	}

	// loading the modules iteratively
	vector<fs::path> remainingModuleLibs;
	map<fs::path, string> moduleErrors;

	size_t loadedModulesCounter = 0;

	for (vector<fs::path>::const_iterator f = moduleLibs.begin(); f != moduleLibs.end(); ++f) {

		// try to load if not already loaded
		if (find(loadedModuleLibs.begin(), loadedModuleLibs.end(), *f) == loadedModuleLibs.end()) {

			string modulePath = f->parent_path().string();
			string moduleName = getModuleName(FileUtilities::nameToString(f->filename()));

			// the shell command lsof -p <pid> shows the list of loaded libraries
			void * handle = dlopen(f->string().c_str(), RTLD_GLOBAL | RTLD_NOW);

			// initialise the module
			bool error = false;
			std::string errorMessage;

			if (handle != 0) {
				void (*initPtr)(Module &);
				void * initSymbol = dlsym(handle, INIT_FUNCTION_NAME.c_str());

				if (initSymbol != 0) {
					loadedModuleLibs.push_back(*f);
					++loadedModulesCounter;

				} else {
					error = true;
					errorMessage = "no init function";
				}

			} else {
				error = true;
				errorMessage = string(dlerror());
			}

			if (error) {
				remainingModuleLibs.push_back(*f);
				moduleErrors[*f] = errorMessage;
			}
		}
	}

	// copying remaining module libs and continue the iteration
	moduleLibs = remainingModuleLibs;

	// if no modules were loaded, the iteration is finished
	if (loadedModulesCounter == 0) {

		cout << "done" << endl;

		// print errors
		if (remainingModuleLibs.size() > 0) {
			ostringstream ss;

			ss << "The following modules were not loaded:\n";
			for (vector<fs::path>::const_iterator f = moduleLibs.begin(); f != moduleLibs.end(); ++f) {
				ss << "\t" << getModuleName(FileUtilities::nameToString(f->filename())) << ": " << moduleErrors[*f] << "\n";
			}

			cout << color::rize(ss.str(), "Red", "Default", "Default") << endl;
		}
		return true;
	}

	writeModuleLoadSequence(directoryPath, loadedModuleLibs);

	return false;
}

void ModuleManager::writeModuleLoadSequence(const std::string& directoryPath, const vector<fs::path>& loadedModuleLibs) {

	string filename = directoryPath + "/" + MODULE_LOAD_SEQUENCE_FILE_NAME;
	ofstream sequenceFile(filename.c_str(), ios::out);

	// Write the number of seconds as id.
	 time_t seconds = time(NULL);

	sequenceFile << "<sequence id=\"" << seconds << "\">\n";

	for (vector<fs::path>::const_iterator f = loadedModuleLibs.begin(); f != loadedModuleLibs.end(); ++f) {

		string moduleLib = f->string();
		string relativeModuleLib = moduleLib;

		// verifying moduleLib starts with directoryPath
		if (moduleLib.substr(0, directoryPath.size()) == directoryPath) {
			// cutting beginning of string
			relativeModuleLib = moduleLib.substr(directoryPath.size() + 1);
		}

		sequenceFile << "  <module library=\"" + relativeModuleLib + "\"/>\n";
	}

	sequenceFile << "</sequence>\n";
	sequenceFile.close();
}

void ModuleManager::writeModuleLoadSequence(const std::string& directoryPath) {

	string filename = directoryPath + "/" + MODULE_LOAD_SEQUENCE_FILE_NAME;
	ofstream sequenceFile(filename.c_str(), ios::out);

	// Write the number of seconds as id.
	 time_t seconds = time(NULL);

	 sequenceFile << "<sequence id=\"" << seconds << "\">\n";

	for (list<Module *>::const_iterator m = m_modules.begin(); m != m_modules.end(); ++m) {

		Module * module = *m;
		string moduleLib = module->getPath().string();
		string relativeModuleLib = moduleLib;

		// verifying moduleLib starts with directoryPath
		if (moduleLib.substr(0, directoryPath.size()) == directoryPath) {
			// cutting beginning of string
			relativeModuleLib = moduleLib.substr(directoryPath.size() + 1);
		}

		// only write modules that are not skipped
		if (module->getDependency() != Module::SKIP) {
			string dependency;
			if (module->getDependency() != "") {
				dependency = " dependency=\"" + module->getDependency() + "\"";
			}

			sequenceFile << "  <module library=\"" + relativeModuleLib + "\"" + dependency + "/>\n";
		}
	}

	sequenceFile << "</sequence>\n";
	sequenceFile.close();
}

void ModuleManager::writeModuleConfig(const std::string& filename) {

	ofstream moduleFile(filename.c_str(), ios::out);
	moduleFile << "<modules>\n";

	for (list<Module *>::const_iterator m = m_modules.begin(); m != m_modules.end(); ++m) {

		Module * module = *m;
		moduleFile << "  <include name=\"" + module->getName() + "\"/>\n";

	}

	moduleFile << "</modules>\n";
	moduleFile.close();
}

void ModuleManager::loadModulesFromSequence(const std::string& directoryPath) {

	out << "Loading modules from sequence " << directoryPath << eol;

	loadModuleResourceFiles(directoryPath);
	loadModuleLibrariesFromSequence(directoryPath, false, true);
}

int ModuleManager::getSequenceId() const {
	return m_sequenceId;
}

bool ModuleManager::loadModuleLibrariesFromSequence(const std::string& directoryPath, bool minimize, bool initModuleFunction) {

	LoadSequenceParser parser;
	parser.init();

	string sequenceFileName = directoryPath + "/" + MODULE_LOAD_SEQUENCE_FILE_NAME;

	fs::path sequenceFilePath(sequenceFileName);
	if (!fs::exists(sequenceFilePath)) {
		throw NoSequenceFile(sequenceFileName);
	}

	parser.parseXMLFile(sequenceFileName);

	m_sequenceId = parser.getId();
	const vector<pair<string, string> > & moduleList = parser.getSequence();

	// in case of minimize, look for the last with no dependency
	string libraryFilenameToSkip;
	Module * moduleToSkip = 0;

	if (minimize) {
		for (vector<pair<string, string> >::const_reverse_iterator f = moduleList.rbegin(); f != moduleList.rend(); ++f) {
			if (f->second == "") {
				libraryFilenameToSkip = f->first;
				break;
			}
		}

		if (libraryFilenameToSkip == "") {
			// end of iteration
			return true;
		}
	}

	// log file
	ofstream logFile;

	if (minimize) {
		string logFileName = directoryPath + "/" + MODULE_LOG_FILE_NAME;
		logFile.open(logFileName.c_str(), ios::app);
	}

	vector<fs::path> loadedLibraries;
	bool error = false;

	for (vector<pair<string, string> >::const_iterator f = moduleList.begin(); f != moduleList.end(); ++f) {

		string relativeLibraryFilename = f->first;
		string libraryFilename = directoryPath + "/" + relativeLibraryFilename;
		fs::path libraryPath(libraryFilename);
		string modulePath = libraryPath.parent_path().string();
		string moduleName = getModuleName(FileUtilities::nameToString(libraryPath.filename()));
		string errorMessage;

		// create the module even if it is skipped in case of minimizing
		if (minimize && libraryFilenameToSkip == relativeLibraryFilename) {
			moduleToSkip = new Module(libraryPath, moduleName, 0);
			addModule(moduleToSkip);

		} else {

			out << "Loading module " << libraryFilename << eol;

			// the shell command lsof -p <pid> shows the list of loaded libraries
			void * handle = dlopen(libraryFilename.c_str(), RTLD_GLOBAL | RTLD_NOW);

			Module * module = new Module(libraryPath, moduleName, handle);
			module->setDependency(f->second);

			addModule(module);

			if (handle != 0) {
				void (*initPtr)(Module &);
				void * initSymbol = dlsym(handle, INIT_FUNCTION_NAME.c_str());

				if (initSymbol != 0) {
					// how to check the type?
					*reinterpret_cast<void **>(&initPtr) = initSymbol;
					(*initPtr)(*module);

					// execute initModule function if present and not while minimizing
					if (initModuleFunction) {
						bool (*initModulePtr)();
						void * initModuleSymbol = dlsym(handle, INIT_MODULE_FUNCTION_NAME.c_str());
						if (initModuleSymbol != 0) {
							*reinterpret_cast<void **>(&initModulePtr) = initModuleSymbol;

							try {
								(*initModulePtr)();

							} catch (...) {
								err << "Exception while initialising a module" << eol;
								throw;
							}
						}
					}

					// searching for test function
					bool (*testPtr)();
					void * testSymbol = dlsym(handle, TEST_FUNCTION_NAME.c_str());
					if (testSymbol != 0) {
						*reinterpret_cast<void **>(&testPtr) = testSymbol;
						module->setTestFunction(testPtr);
					}

					loadedLibraries.push_back(libraryPath);

				} else {
					error = true;
					errorMessage = "no init symbol";
				}

			} else {
				error = true;
				errorMessage = string(dlerror());
			}

			if (error) {
				if (minimize) {
					logFile << "Cannot load module " << moduleName << " (" << relativeLibraryFilename << "): " << errorMessage << endl;
				} else {
					err << "Cannot load module " << moduleName << " (" << relativeLibraryFilename << "): " << errorMessage << eol;
				}
			}
		}
	}

	// final set of dependency
	if (moduleToSkip != 0) {
		if (error) {
			moduleToSkip->setDependency(Module::INDIRECT);
		} else {
			moduleToSkip->setDependency(Module::SKIP);
		}
	}

	if (minimize) {
		logFile.close();
	}

	return false;
}

std::vector<boost::filesystem::path> ModuleManager::getLibraryPathList() {

	std::vector<boost::filesystem::path> loadedLibraries;

	for (list<Module *>::iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;
		loadedLibraries.push_back(module->getPath());
	}
	return loadedLibraries;
}

void ModuleManager::setDirectDependenciesInSequence(const std::string& directoryPath, const std::string& configPath) {

	// first load the file
	loadModuleLibrariesFromSequence(directoryPath, false, false);
	vector<fs::path> loadedModuleLibs = getLibraryPathList();

	map<fs::path, std::string> moduleDependencies;

	// drivers
	string hardwareFilename = configPath + "/HardwareConfig.xml";

	pugi::xml_document hardwareDoc;
	hardwareDoc.load_file(hardwareFilename.c_str());
	set<string> driverTypes;
	for (auto children : hardwareDoc.children()) {
		if (children.name() == string("DriverConfig")) {
			driverTypes = readDriverTypes(children);
		}
	}

	for (set<string>::const_iterator t = driverTypes.begin(); t != driverTypes.end(); ++t) {
		const string & type = *t;

		// search for type in the modules
		for (list<Module *>::iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
			Module * module = *m;
			if (module->findDriverClass(type)) {
				module->setDependency(Module::DIRECT);
			}
		}
	}

	// controllers
	string instrumentFilename = configPath + "/InstrumentConfig.xml";

	pugi::xml_document instrumentDoc;
	instrumentDoc.load_file(instrumentFilename.c_str());

	set<string> controllerTypes;
	for (auto children : instrumentDoc.children()) {
		if (children.name() == string("ControllerConfig")) {
			controllerTypes = readControllerTypes(children);
		}
	}

	for (set<string>::const_iterator t = controllerTypes.begin(); t != controllerTypes.end(); ++t) {
		const string & type = *t;

		// search for type in the modules
		for (list<Module *>::iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
			Module * module = *m;
			if (module->findControllerClass(type)) {
				module->setDependency(Module::DIRECT);
			}
		}
	}

	writeModuleLoadSequence(directoryPath);
}

bool ModuleManager::loadModulesForMinimizing(const std::string& directoryPath, const std::string& moduleConfigFilename) {

	bool finished = loadModuleLibrariesFromSequence(directoryPath, true, false);

	// write only if not finished
	if (!finished) {
		writeModuleLoadSequence(directoryPath);
		writeModuleConfig(moduleConfigFilename);
	} else {
		out << "done" << eol;
	}

	return finished;
}

void ModuleManager::unloadUnusedModules() {

	// this function is now unused, because the unloading is not effective. Some libraries cannot be unloaded because of static variables.
}

void ModuleManager::loadModuleResourceFiles(const std::string& directoryPath) {

	Counter c1;
	out << "Loading resource files from modules : " << directoryPath << eol;

	vector<fs::path> propertyFiles;

	try {
		propertyFiles = FileUtilities::getFilePathsWithExtension(directoryPath, "xml");

	} catch (const FileUtilities::NoSuchDirectoryException & e) {
		cout << "Cannot load property files from " << directoryPath << endl;
		return;
	}

	out << "Time 1 : " << c1.getTime() << eol;

	// loading the XML files
	for (vector<fs::path>::const_iterator f = propertyFiles.begin(); f != propertyFiles.end(); ++f) {
		// add in the resources map
		string filename = f->filename().c_str();
		m_moduleResources.insert(make_pair(filename, f->c_str()));
	}

	out << "Time 2 : " << c1.getTime() << eol;

	// loading the properties files
	propertyFiles = FileUtilities::getFilePathsWithExtension(directoryPath, "properties");

	for (vector<fs::path>::const_iterator f = propertyFiles.begin(); f != propertyFiles.end(); ++f) {
		// add in the resources map
		string filename = f->filename().c_str();
		m_moduleResources.insert(make_pair(filename, f->c_str()));
	}

	out << "Time 3 : " << c1.getTime() << eol;
	// loading the jar files
	propertyFiles = FileUtilities::getFilePathsWithExtension(directoryPath, "jar");

	for (vector<fs::path>::const_iterator f = propertyFiles.begin(); f != propertyFiles.end(); ++f) {
		// add in the resources map
		string filename = f->filename().c_str();
		m_moduleResources.insert(make_pair(filename, f->c_str()));
	}
	out << "Time 4 : " << c1.getTime() << eol;

	out << "Loaded resource files" << eol;
}

void ModuleManager::addModule(Module * module) {
	m_modules.push_back(module);

	vector<string> controllerTypes = module->getControllerTypes();

	for (vector<string>::const_iterator t = controllerTypes.begin(); t != controllerTypes.end(); ++t) {
		if (m_moduleControllers.find(*t) != m_moduleControllers.end()) {
			warn << *t << " is already created by module " << m_moduleControllers[*t] << eol;
			warn << *t << " will not be created by module " << module->getName() << eol;
		} else {
			m_moduleControllers[*t] = module->getName();
		}
	}

	vector<string> driverTypes = module->getDriverTypes();

	for (vector<string>::const_iterator t = driverTypes.begin(); t != driverTypes.end(); ++t) {
		if (m_moduleDrivers.find(*t) != m_moduleDrivers.end()) {
			warn << *t << " is already created by module " << m_moduleDrivers[*t] << eol;
			warn << *t << " will not be created by module " << module->getName() << eol;
		} else {
			m_moduleDrivers[*t] = module->getName();
		}
	}
}

std::vector<std::string> ModuleManager::getLoadedModules() const {

	vector<string> result;
	for (list<Module *>::const_iterator m = m_modules.begin(); m != m_modules.end(); ++m) {

		Module * module = *m;
		result.push_back(module->getName());
	}

	return result;
}

std::vector<std::string> ModuleManager::getAllControllerTypes() const {

	vector<string> result;

	for (list<Module *>::const_iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;

		vector<string> controllerList = module->getControllerTypes();
		for (vector<string>::const_iterator c = controllerList.begin(); c != controllerList.end(); ++c) {
			result.push_back(*c);
		}
	}

	return result;
}

ExperimentController * ModuleManager::makeController(const std::string& type, const std::string& name) {

	// search for type in the modules
	for (list<Module *>::iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;
		if (module->findControllerClass(type)) {
			module->setDependency(Module::DIRECT);
			ExperimentController * controller = module->makeController(type, name);
			controller->setModule(module);

			return controller;
		}
	}

	throw NoSuchControllerClass(type);
}

std::vector<std::string> ModuleManager::getAllDriverTypes() const {

	vector<string> result;

	for (list<Module *>::const_iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;

		vector<string> driverList = module->getDriverTypes();
		for (vector<string>::const_iterator c = driverList.begin(); c != driverList.end(); ++c) {
			result.push_back(*c);
		}
	}

	return result;
}

Module * ModuleManager::findModuleForDriver(const std::string& type) const {

	// search for type in the modules
	for (list<Module *>::const_iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;
		if (module->findDriverClass(type)) {
			return module;
		}
	}

	return nullptr;
}

Module * ModuleManager::findModuleForRealDriver(const std::string& type) const {

	// search for type in the modules
	for (list<Module *>::const_iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;
		if (module->findRealDriverClass(type)) {
			return module;
		}
	}

	return nullptr;
}

DeviceDriver * ModuleManager::makeDriver(const std::string& type, const std::string& name) {

	Module * module = findModuleForDriver(type);

	if (module) {
		module->setDependency(Module::DIRECT);
		DeviceDriver * driver = module->makeDriver(type, name);

		// Try to create the real driver state if it is not set in the driver.
		if (!driver->hasRealState()) {

			Module * moduleForReal = findModuleForRealDriver(type);

			if (moduleForReal) {
				DeviceDriverState * realState = moduleForReal->makeRealDriverState(type, driver);
				driver->registerRealState(realState);

				out << "Registered real driver for " << type << eol;
			}
			else {
				out << "Cannot register real driver for " << type << eol;
			}
		}

		driver->setModule(module);

		return driver;
	}

	throw NoSuchDriverClass(type);
}

AbstractParser * ModuleManager::makeParser(const std::string& type) {

	// search for type in the modules
	for (list<Module *>::iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;
		if (module->findParserClass(type)) {
			AbstractParser * parser = module->makeParser(type);

			return parser;
		}
	}

	throw NoSuchParserClass(type);
}

std::string ModuleManager::getModuleName(const std::string& fileName) {
	int32 pos = 3;
	if (fileName.substr(0, 3) != LIB_PREFIX) {
		pos = 0;
	}

	return fileName.substr(pos, fileName.size() - 3 - pos);
}

bool ModuleManager::testModule(const std::string& name) {
	for (list<Module *>::iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;
		if (module->getName() == name) {

			if (module->hasTestFunction()) {
				return module->test();
			} else {
				throw NoTestFunction(name);
			}
		}
	}

	throw NoSuchModule(name);
}

std::map<std::string, std::string> ModuleManager::getAllTypeModules() const {

	map<string, string> result;

	for (list<Module *>::const_iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;

		vector<string> driverList = module->getDriverTypes();
		for (vector<string>::const_iterator c = driverList.begin(); c != driverList.end(); ++c) {
			result[*c] = module->getName();
		}

		vector<string> controllerList = module->getControllerTypes();
		for (vector<string>::const_iterator c = controllerList.begin(); c != controllerList.end(); ++c) {
			if (result.find(*c) != result.end()) {
				warn << "Controller " << *c << " already added" << eol;
			}
			result[*c] = module->getName();
		}
	}

	return result;
}

std::map<std::string, std::vector<std::string> > ModuleManager::getControllerTypesByModule() const {

	map<string, vector<string> > result;

	for (list<Module *>::const_iterator m = m_modules.begin(); m != m_modules.end(); ++m) {
		Module * module = *m;
		result[module->getName()] = module->getControllerTypes();
	}

	return result;
}

std::set<std::string> ModuleManager::readDriverTypes(const pugi::xml_node& items) {

	set<string> result;
	for (auto driverconfig : items.children()) {
		for (auto driver : driverconfig.children()) {
			if (driver.name() == string("Type")) {
				string type = driver.child_value();
				result.insert(type);
			}
			else if (driver.name() == string("Children")) {
				set<string> childrenTypes = readDriverTypes(driver);
				result.insert(childrenTypes.begin(), childrenTypes.end());
			}
		}
	}

	return result;
}

std::set<std::string> ModuleManager::readControllerTypes(const pugi::xml_node& items) {

	set<string> result;

	for (auto controllerconfig : items.children()) {
		for (auto controller : controllerconfig.children()) {
			if (controller.name() == string("Type")) {
				string type = controller.child_value();
				result.insert(type);
			}
		}
	}

	return result;
}

const std::multimap<std::string, std::string>& ModuleManager::getResourceFiles() const {
	return m_moduleResources;
}

std::vector<std::string> ModuleManager::getResourceFilesByExtension(const std::string& extension) const {

	vector<string> result;

	// Iterate the resources.
	for (map<string, string>::const_iterator r = m_moduleResources.begin(); r != m_moduleResources.end(); ++r) {
		// Get the path and test the extension
		fs::path filePath(r->first);
		if (filePath.extension() == ".jar") {
			result.push_back(r->first);
		}
	}

	return result;
}

std::string ModuleManager::readResourceFile(const std::string& type, const std::string& fileName) {

	// type is not used

	// find the file
	map<string, string>::iterator r = m_moduleResources.find(fileName);
	if (r == m_moduleResources.end()) {
		throw NoSuchFile(fileName);
	}

	ifstream inFile;
	inFile.open(r->second.c_str());

	stringstream stream;
	stream << inFile.rdbuf();

	return stream.str();
}

void ModuleManager::writeResourceFile(const std::string& type, const std::string& fileName, const std::string& content) {

	Module * module = 0;

	// search for type in the modules
	for (list<Module *>::iterator m = m_modules.begin(); m != m_modules.end(); ++m) {

		if ((*m)->findDriverClass(type) || (*m)->findControllerClass(type)) {
			module = *m;
			break;
		}
	}

	if (module == 0) {
		err << "Cannot find module for type " << type << eol;
		return;
	}

	// the files must be moved after the creation in the correct location
	fs::path moduleTmpPath = ServerProperties::getInstance()->getNomadHomePath() + "/mod/tmp";

	if (!fs::exists(moduleTmpPath)) {
		fs::create_directory(moduleTmpPath);
	}

	moduleTmpPath /= fileName;

	ofstream outFile(moduleTmpPath.c_str(), ios::out);
	outFile << content << endl;
	outFile.close();

	string filename = fileName.c_str();
	m_moduleResources.insert(make_pair(filename, moduleTmpPath.c_str()));

	out << "Written file " << moduleTmpPath.string() << eol;
}

std::string ModuleManager::getResourceFilePath(const std::string& fileName) {

	// get the filename without relative path
	fs::path filePath(fileName);
	string shortName = FileUtilities::nameToString(filePath.filename());

	int counter = 0;
	string fullPath;

	pair<multimap<string, string>::iterator, multimap<string, string>::iterator> ret = m_moduleResources.equal_range(shortName);
	for (std::multimap<string, string>::iterator it = ret.first; it != ret.second; ++it) {
		string longFileName = it->second;

		// check if fileName matches longFileName
		string subFileName = longFileName.substr(longFileName.length() - fileName.length());

		if (subFileName == fileName) {
			fullPath = longFileName;
			counter++;
		}
	}

	if (counter == 0) {
		throw NoSuchFile(fileName);
	}

	if (counter > 1) {
		throw AmbiguousFile(fileName);
	}

	return fullPath;
}

time_t ModuleManager::getResourceFileLastTime(const std::string& type, const std::string& fileName) {
	// find the file
	map<string, string>::iterator r = m_moduleResources.find(fileName);
	if (r == m_moduleResources.end()) {
		throw NoSuchFile(fileName);
	}

	time_t date;
	try {
		date = FileUtilities::getFileLastTime(r->second);
	}
	catch (const FileUtilities::NotRegularFile& e) {
		throw NoSuchFile(fileName);
	}

	return date;
}
