/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef MODULEMANAGER_H_
#define MODULEMANAGER_H_

#include "Module.h"
#include <common/base/Exception.h>
#include <pugixml.hpp>
#include <vector>
#include <set>

/**
 * Exception class for resource file path.
 */
class AmbiguousFile : public common::Exception {

public:
	/**
	 * Constructor.
	 */
	AmbiguousFile(std::string const & fileName);
};

/**
 * Exception class for resource file.
 */
class NoSuchFile : public common::Exception {

public:
	/**
	 * Constructor.
	 */
	NoSuchFile(std::string const & fileName);
};


class ModuleManager {

public:
	/**
	 * Exception class for unloaded controller.
	 */
	class NoSuchControllerClass : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchControllerClass(std::string const & type);
	};

	/**
	 * Exception class for unloaded driver.
	 */
	class NoSuchDriverClass : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchDriverClass(std::string const & type);
	};

	/**
	 * Exception class for unloaded parser.
	 */
	class NoSuchParserClass : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchParserClass(std::string const & type);
	};

	/**
	 * Exception class for module.
	 */
	class NoSuchModule : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSuchModule(std::string const & name);
	};

	/**
	 * Exception class for test function.
	 */
	class NoTestFunction : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoTestFunction(std::string const & name);
	};

	/**
	 * Exception class for sequence file.
	 */
	class NoSequenceFile : public common::Exception {

	public:
		/**
		 * Constructor.
		 */
		NoSequenceFile(std::string const & name);
	};

	/**
	 * Returns the instance.
	 */
	static ModuleManager * getInstance();

	/**
	 * Initialises the instance.
	 */
	static void initInstance();

	/**
	 * Resets the instance.
	 */
	static void resetInstance();

	/**
	 * Loads the modules located in directoryPath for ordering the load sequence.
	 * Returns true if no more modules were loaded.
	 */
	bool loadModulesForOrdering(const std::string& directoryPath);

	/**
	 * Loads the modules located in directoryPath.
	 */
	void loadModulesFromSequence(const std::string& directoryPath);

	/**
	 * Gets the sequence id.
	 */
	int getSequenceId() const;

	/**
	 * Generates the direct dependencies in the load sequence.
	 */
	void setDirectDependenciesInSequence(const std::string& directoryPath, const std::string& configPath);

	/**
	 * Loads the modules located in directoryPath for minimizing the load sequence.
	 * Returns true if no more modules were loaded.
	 */
	bool loadModulesForMinimizing(const std::string& directoryPath, const std::string& moduleConfigFilename);

	/**
	 * Unloads the modules that are not used.
	 */
	void unloadUnusedModules();

	/**
	 * Gets the loaded modules.
	 */
	std::vector<std::string> getLoadedModules() const;

	/**
	 * Gets all the controller types.
	 */
	std::vector<std::string> getAllControllerTypes() const;

	/**
	 * Makes the controller with type.
	 * throws: NoSuchControllerClass
	 */
	ExperimentController * makeController(const std::string& type, const std::string& name);

	/**
	 * Gets all the driver types.
	 */
	std::vector<std::string> getAllDriverTypes() const;

	/**
	 * Makes the driver with type.
	 * throws: NoSuchDriverClass
	 */
	DeviceDriver * makeDriver(const std::string& type, const std::string& name);

	/**
	 * Makes the parser with type.
	 * throws: NoSuchParserClass
	 */
	AbstractParser * makeParser(const std::string& type);

	/**
	 * Gets the map of type module.
	 */
	std::map<std::string, std::string> getAllTypeModules() const;

	/**
	 * Gets the map of controller types by module.
	 */
	std::map<std::string, std::vector<std::string> > getControllerTypesByModule() const;

	/**
	 * Tests the module name.
	 * throws: NoSuchModule
	 */
	bool testModule(const std::string& name);

	/**
	 * Returns the map of resources. The first item is the file name and the second item the absolute path of the file.
	 */
	const std::multimap<std::string, std::string>& getResourceFiles() const;

	/**
	 * Returns the list of resources with extension.
	 */
	std::vector<std::string> getResourceFilesByExtension(const std::string& extension) const;

	/**
	 * Reads the resource file.
	 * throws: NoSuchFile
	 */
	std::string readResourceFile(const std::string& type, const std::string& fileName);

	/**
	 * Writes the resource file.
	 * throws: NoSuchFile
	 */
	void writeResourceFile(const std::string& type, const std::string& fileName, const std::string& content);

	/**
	 * Gets the resource file path. The parameter can be a relative (partial) file name to make it non ambiguous.
	 * throws: AmbiguousFile, NoSuchFile
	 */
	std::string getResourceFilePath(const std::string& fileName);

	/**
	 * Reads the resource file last time.
	 * throws: NoSuchFile
	 */
	time_t getResourceFileLastTime(const std::string& type, const std::string& fileName);

private:
	ModuleManager();
	~ModuleManager();

	bool loadModuleLibrariesFromSequence(const std::string& directoryPath, bool minimize, bool initModuleFunction);
	std::vector<boost::filesystem::path> getLibraryPathList();
	void loadModuleResourceFiles(const std::string& directoryPath);
	static std::string getModuleName(const std::string& fileName);
	void addModule(Module * module);
	void writeModuleLoadSequence(const std::string& directoryPath, const std::vector<boost::filesystem::path>& loadedModuleLibs);
	void writeModuleLoadSequence(const std::string& directoryPath);
	void writeModuleConfig(const std::string& filename);

	std::set<std::string> readDriverTypes(const pugi::xml_node& items);
	std::set<std::string> readControllerTypes(const pugi::xml_node& items);

	Module * findModuleForDriver(const std::string& type) const;
	Module * findModuleForRealDriver(const std::string& type) const;

	int m_sequenceId;
	std::list<Module *> m_modules;
	std::map<std::string, std::string> m_moduleControllers;
	std::map<std::string, std::string> m_moduleDrivers;

	std::multimap<std::string, std::string> m_moduleResources;

	static ModuleManager * m_instance;

	static const std::string LIB_PREFIX;
	static const std::string INIT_FUNCTION_NAME;
	static const std::string TEST_FUNCTION_NAME;
	static const std::string INIT_MODULE_FUNCTION_NAME;
	static const std::string PROPERTY_FILE_NAME;
	static const std::string DEVICE_PROPERTY_FILE_NAME;
	static const std::string CONTROLLER_PROPERTY_FILE_NAME;
	static const std::string SCANNABLE_PROPERTY_FILE_NAME;
	static const std::string MODULE_LOAD_SEQUENCE_FILE_NAME;
	static const std::string MODULE_LOG_FILE_NAME;
};

#endif
