/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "FileManager.h"
#include <common/base/FileUtilities.h>
#include <iostream>
#include <boost/property_tree/xml_parser.hpp>
#include "DataProvider/Property/PropertyInfos.h"
#include <common/base/FileUtilities.h>
#include <common/base/ServerProperties.h>
#include <boost/filesystem/operations.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/property_tree/exceptions.hpp>
#include <dlfcn.h>
#include <fstream>
#include <algorithm>
#include <stdio.h>
#include <dirent.h>
#include <boost/filesystem.hpp>


namespace pt = boost::property_tree;
namespace fs = boost::filesystem;

using namespace boost::filesystem;
using namespace common;
using namespace std;
using namespace rapidjson;


const std::string FileManager::NAME = "name";
const std::string FileManager::ISDIRECTORY = "isDirectory";
const std::string FileManager::CHILDREN = "children";


FileManager * FileManager::m_instance = 0;

FileManager * FileManager::getInstance() {
	if (m_instance == 0) {
		initInstance();
	}
	return m_instance;
}

void FileManager::initInstance() {
	if (m_instance == 0) {
		m_instance = new FileManager();
	}
}

void FileManager::resetInstance() {
	delete m_instance;
	m_instance = 0;
}

FileManager::FileManager() {
	const char* home = getenv("HOME");
	if (home != 0) {
		_homePath = string(home);
	} else {
		common::warn << "$HOME is not set" << common::eol;
	}
}

FileManager::~FileManager() {
}


std::string FileManager::getFileContent(const std::string& fileName) {

	string filePath = _homePath + "/" + fileName;
	fs::ifstream inFile;
	inFile.open(filePath);
	stringstream stream;
	stream << inFile.rdbuf();
	inFile.close();

	return stream.str();
}


std::string FileManager::getBinaryFileContent(const std::string& fileName) {

	string filePath = _homePath + "/" + fileName;

	streampos size;
	string fileContent;
	std::ifstream file(filePath.c_str(), ios::in | ios::binary | ios::ate);

	if (file.is_open()) {

		size = file.tellg();

		// Reserve the string.
		fileContent.resize(size);

		// Get the allocated array.
		char * array = (char *) fileContent.data();

		file.seekg(0, ios::beg);
		file.read(array, size);
		file.close();

	} else {
		common::warn << "File " << filePath << " cannot be read." << common::eol;
	}
	return fileContent;
}


std::string FileManager::getFileContentAbsolutePath(const std::string& fileName) {

	fs::ifstream inFile;
	inFile.open(fileName);
	stringstream stream;
	stream << inFile.rdbuf();
	inFile.close();

	return stream.str();
}


void FileManager::saveFile(const std::string& fileName, const std::string& content, const bool addServerHome) {

	fs::path filePath;

	if (addServerHome) {
		filePath = _homePath / fileName;
	} else {
		filePath = fileName;
	}

	if (!fs::exists(filePath.parent_path())) {
		fs::create_directory(filePath.parent_path());
	}

	fs::ofstream outFile(filePath.string(), ios::out);

	if (outFile.is_open()) {
		outFile << content << endl;
		outFile.close();
	} else {
		common::warn << "Cannot write to file " << filePath.string() << common::eol;
	}
}

void FileManager::saveBinaryFile(const std::string& fileName, const std::string& content, const bool addServerHome) {

	fs::path filePath;

	if (addServerHome) {
		filePath = _homePath / fileName;
	} else {
		filePath = fileName;
	}

	if (!fs::exists(filePath.parent_path())) {
		fs::create_directory(filePath.parent_path());
	}

	std::ofstream file;
	file.open(filePath.string(), ios::binary);

	if (file.is_open()) {
		file << content;
		file.close();
	} else {
		common::warn << "Cannot write to file " << filePath.string() << common::eol;
	}
}

std::string FileManager::getTree(const std::string& rootDirectory, const std::string& extension, bool recurse) {

	rapidjson::StringBuffer s;
	PrettyWriter<StringBuffer> writer(s);

	// Clean writer and stringbuffer
	s.Clear();

	// Start new message
	writer.StartObject();

	getFileTree(_homePath + "/" + rootDirectory, extension, writer, recurse);

	writer.EndObject();

	return s.GetString();
}


std::string FileManager::getFileTree(const path currentPath, const std::string& extension, rapidjson::PrettyWriter<rapidjson::StringBuffer> &writer, bool recurse) {
	try {
		if (fs::exists(currentPath)) {
			if (fs::is_directory(currentPath)) {

//				bool goOn = true;
				//	if (extension.length() > 0) {
				//		goOn = atLeastOneFile(currentPath.c_str(), extension);
				//	}

//				if (goOn) {
					path directoryName = currentPath.filename();
					if (directoryName == ".trash" || directoryName == "oldScripts") {
						// ignore hidden directory
					} else {
						writer.Key(NAME.c_str());
						writer.String(currentPath.c_str());

						writer.Key(ISDIRECTORY.c_str());
						writer.String("true");

						if (recurse) {
							writer.Key(CHILDREN.c_str());
							writer.StartArray();
							// Walk trough children
							for (directory_entry& x : directory_iterator(currentPath)) {
								writer.StartObject();
								getFileTree(x.path(), extension, writer, recurse);
								writer.EndObject();
							}
							writer.EndArray();
						}
					}
//				}
			} else {
				if ((currentPath.extension() == extension) || (extension.length() == 0)) {
					writer.Key(NAME.c_str());
					writer.String(currentPath.c_str());
					writer.Key(ISDIRECTORY.c_str());
					writer.String("false");
				}
			}
		} else {
			common::warn << currentPath.string() << " does not exist" << common::eol;
		}
	} catch (const filesystem_error& ex) {
		common::warn << "Exception while getting file tree " << ex.what() << common::eol;
	}
	return "";
}


bool FileManager::atLeastOneFile(const std::string& path, const std::string& extension) {
	std::vector<std::string> files = getFilesInDir(path, extension, true);
	return (files.size() > 0);
}


/**
 * File is not really deleted but moved in an hidden directory.
 */
void FileManager::deleteFile(const std::string& fileName) {
	try {
		path src = _homePath / fs::path(fileName);
		path parent(src.parent_path());
		path atomicFileName(src.filename());

		path dest = parent / path(".trash");
		if (!fs::exists(dest)) {
			fs::create_directories(dest);
		}
		dest = dest / atomicFileName;
		fs::rename(src, dest);
	} catch (const filesystem_error& ex) {
		common::warn << "Exception while deleting file " << fileName << ":" << ex.what() << common::eol;
	}
}




std::vector<std::string> FileManager::getFilesInDir(const std::string& path, const std::string& extension, const bool recursive) {

	std::vector<std::string> files;
	boost::filesystem::path dir(path);
	if (boost::filesystem::exists(path) && boost::filesystem::is_directory(path)) {

		if (recursive) {
			boost::filesystem::recursive_directory_iterator it(path);
			boost::filesystem::recursive_directory_iterator endit;
			while (it != endit) {
				if (boost::filesystem::is_regular_file(*it) && (extension=="")?true:it->path().extension() == extension) {
					files.push_back(it->path().string());
				}
				++it;
			}
		} else {
			boost::filesystem::directory_iterator it(path);
			boost::filesystem::directory_iterator endit;
			while (it != endit) {
				if(boost::filesystem::is_regular_file(*it) && (extension=="")?true:it->path().extension() == extension) {
					files.push_back(it->path().string());
				}
				++it;
			}
		}
	}
	return files;
}


void FileManager::createDirectory(const std::string& parent, const std::string& directoryName) {
	try {
		path dest = _homePath / fs::path(parent) / fs::path(directoryName);
		if (!fs::exists(dest)) {
			fs::create_directories(dest);
		}
	} catch (const filesystem_error& ex) {
		common::warn << "Exception while creating directory " << directoryName << " in " << parent << ":" << ex.what() << common::eol;
	}
}


void FileManager::appendToFile(const std::string& fullFileName, const std::string& stringToAppend) {

	// Getting the path
	size_t pos = fullFileName.find_last_of('/');

	if ((pos >= 0) && (pos < fullFileName.size())) {
		string filePath = fullFileName.substr(0, pos + 1);
		path dest = _homePath / fs::path(filePath);

		if (!fs::exists(dest)) {
			// Create directory if it does not exist
			createDirectory(filePath, "");
		}
	}

	fs::ofstream outFile(_homePath + "/" + fullFileName.c_str(), ios::out | ios::app);
	if (outFile.is_open()) {
		outFile << stringToAppend << endl;
	}
	else {
		common::warn << "Error in FileManager::appendToFile(), unable to open file: " << fullFileName << common::eol;
	}
}

bool FileManager::fileExists(const std::string& fileName) {
	string filePath;
	filePath = _homePath + "/" + fileName;
	return fs::exists(filePath);
}
