/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "PropertyConfig.h"
#include "RoleConfig.h"
#include "XMLFileConfiguration2.h"
#include "common/base/ServerProperties.h"
#include "InstrumentAbstraction/AbstractController.h"
#include "PersistenceServices/XMLErrors.h"
#include "Module/ModuleManager.h"

#include <cstdlib>
#include <fstream>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/filesystem/operations.hpp>

using namespace boost::posix_time;
using namespace boost::filesystem;
using namespace std;

/**
 * Init static members
 */
const string XMLFileConfiguration2::HARDWARE_CONF_FILE_NAME = "HardwareConfig.xml";
const string XMLFileConfiguration2::INSTRUMENT_CONF_FILE_NAME = "InstrumentConfig.xml";
const string XMLFileConfiguration2::VERSION_TAG = "version";
const string XMLFileConfiguration2::MODULES_ID_TAG = "modules_id";
const string XMLFileConfiguration2::DRIVERCONFIG_TAG = "DriverConfig";
const string XMLFileConfiguration2::DRIVER_TAG = "Driver";
const string XMLFileConfiguration2::CONTROLLERCONFIG_TAG = "ControllerConfig";
const string XMLFileConfiguration2::CONTROLLER_TAG = "Controller";
const string XMLFileConfiguration2::NAME_TAG = "Name";
const string XMLFileConfiguration2::TYPE_TAG = "Type";
const string XMLFileConfiguration2::ENABLED_TAG = "Enabled";
const string XMLFileConfiguration2::PROPERTY_TAG = "Property";
const string XMLFileConfiguration2::NAME_ATTRIBUTE = "name";
const string XMLFileConfiguration2::CHILDREN_TAG = "Children";
const string XMLFileConfiguration2::CHANNEL_ATTRIBUTE = "channel";
const string XMLFileConfiguration2::FAMILY_TAG = "MainFamily";
const string XMLFileConfiguration2::SUBFAMILY_TAG = "SubFamily";
const string XMLFileConfiguration2::CONTROLLERPTR_TAG = "ControllerPtr";
const string XMLFileConfiguration2::DRIVERPTR_TAG = "DriverPtr";
const string XMLFileConfiguration2::PROPERTYARRAY_TAG = "PropertyArray";
const string XMLFileConfiguration2::PROPERTYARRAY_CSV_TAG = "|";
const string XMLFileConfiguration2::PROPERTYARRAY_COLS_TAG = "cols";
const string XMLFileConfiguration2::ROW_TAG = "Row";

/*
 * Constructor
 */
XMLFileConfiguration2::XMLFileConfiguration2() {
	// Create config files name
	string home;
	home.append(common::ServerProperties::getInstance()->getConfigPath());
	home.append("/");

	m_hardware_filename.append(home);
	m_hardware_filename.append(HARDWARE_CONF_FILE_NAME);
	m_instrument_filename.append(home);
	m_instrument_filename.append(INSTRUMENT_CONF_FILE_NAME);

	home = common::ServerProperties::getInstance()->getConfigPath();
	home.append("_save");
	home.append("/");
	m_instrument_backup_filename.append(home);
	m_instrument_backup_filename.append("/");
	m_instrument_backup_filename.append(INSTRUMENT_CONF_FILE_NAME);
	m_hardware_backup_filename.append(home);
	m_hardware_backup_filename.append("/");
	m_hardware_backup_filename.append(HARDWARE_CONF_FILE_NAME);

	// Create if it isn't existed directory save for backup config
	string buf = "mkdir -p " + home;
	::system(buf.c_str());

}

/*
 * Destructor
 */
XMLFileConfiguration2::~XMLFileConfiguration2() {
}

/************************************************************************
 *
 * TEST CONTROLLER
 *
 ************************************************************************/

/*
 * loadTestControllers
 */
int32 XMLFileConfiguration2::loadTestControllers(vector<TestControllerConfig> &configs) {
	pugi::xml_document doc;
	pugi::xml_parse_result result = doc.load_file(m_hardware_filename.c_str());
	if (result) {
		for (auto noderoot : doc.children()) {
			if (noderoot.name() == DRIVERCONFIG_TAG) {
				if (noderoot.attribute(VERSION_TAG.c_str()).as_int() > 1) {
					for (auto node : noderoot.children()) {
						if (node.name() == DRIVER_TAG) {
							loadTestController(node, "", configs);
						} else {
							unknownTagError(m_hardware_filename, (uint32) node.offset_debug(), node.name());
						}
					}
				} else {
					return -1;
				}
			} else {
				return -1;
			}
		}
	} else {
		printError(m_hardware_filename, (uint32) result.offset, result.description());
	}
	return 0;
}

/*
 * saveTestControllers
 */
void XMLFileConfiguration2::saveTestControllers(const vector<TestControllerConfigNode> &configs, bool isBackup) {
	pugi::xml_document doc;
	pugi::xml_node rootElem = doc.append_child(DRIVERCONFIG_TAG.c_str());
	rootElem.append_attribute(VERSION_TAG.c_str()) = "2";

	// Save the module sequence id.
	int moduleSequenceId = ModuleManager::getInstance()->getSequenceId();
	std::string moduleSequenceIdString = std::to_string(moduleSequenceId);
	rootElem.append_attribute(MODULES_ID_TAG.c_str()) = moduleSequenceIdString.c_str();

	for (auto c : configs) {
		saveTestControllerNode(rootElem, c);
	}

	string buf;
	if (isBackup) {
		buf = m_hardware_backup_filename;
		ptime t(second_clock::local_time());
		buf.append(".");
		buf.append(to_simple_string(t));
	} else {
		buf = m_hardware_filename;
	}
	doc.save_file(buf.c_str(), PUGIXML_TEXT("  "));
}

/*
 * createTestController
 */
void XMLFileConfiguration2::loadTestController(const pugi::xml_node &node, const std::string &parentName,
		std::vector<TestControllerConfig> &configs) {

	TestControllerConfig config;

	// Getting the channel attribute
	pugi::xml_attribute attr = node.attribute(CHANNEL_ATTRIBUTE.c_str());
	if (!attr) {
		missingAttributeError(m_hardware_filename, (uint32) node.offset_debug(), CHANNEL_ATTRIBUTE, DRIVER_TAG);
	}
	string parentChannel = attr.value();
	config.setParentChannel(stoi(parentChannel));
	config.setParentName(parentName);

	bool isnametag = false;
	bool istypetag = false;

	// Iterating the children
	for (auto children : node.children()) {
		if (children.name() == NAME_TAG) {
			config.setName(children.child_value());
			isnametag = true;
		} else if (children.name() == TYPE_TAG) {
			config.setType(children.child_value());
			istypetag = true;
		} else if (children.name() == ENABLED_TAG) {
			if (children.child_value() == string("true")) {
				config.setEnabled(true);
			} else {
				config.setEnabled(false);
			}
		} else if (children.name() == PROPERTY_TAG) {
			attr = children.attribute(NAME_ATTRIBUTE.c_str());
			string propertyName;
			if (attr) {
				propertyName = attr.value();
			} else {
				missingAttributeError(m_hardware_filename, (uint32) children.offset_debug(), NAME_ATTRIBUTE, PROPERTY_TAG);
			}
			string propertyValue = children.child_value();
			config.addProperty(PropertyConfig(propertyName, propertyValue));
		} else if (children.name() == PROPERTYARRAY_TAG) {
			pugi::xml_attribute attr = children.attribute(NAME_ATTRIBUTE.c_str());
			string propertyName;
			if (attr) {
				propertyName = attr.value();
			} else {
				missingAttributeError(m_hardware_filename, (uint32) children.offset_debug(), NAME_ATTRIBUTE, PROPERTY_TAG);
			}
			attr = children.attribute(PROPERTYARRAY_COLS_TAG.c_str());
			string propertyArrayCols;
			if (attr) {
				propertyArrayCols = attr.value();
			} else {
				missingAttributeError(m_hardware_filename, (uint32) children.offset_debug(), PROPERTYARRAY_COLS_TAG,
						PROPERTYARRAY_TAG);
			}
			int32 ncol = stoi(propertyArrayCols);
			vector<string> myString;
			for (auto row : children.children()) {
				if (row.name() == ROW_TAG) {
					string arrayValue = row.child_value();
					uint32 lastindex = 0;
					for (uint32 colId = 0; colId < (uint32) ncol; colId++) {
						uint32 myindex = arrayValue.find(PROPERTYARRAY_CSV_TAG, lastindex);
						myString.push_back(arrayValue.substr(lastindex, myindex - lastindex));
						lastindex = myindex + 1;
					}
				} else {
					unknownTagError(m_hardware_filename, (uint32) row.offset_debug(), row.name());
				}
			}
			config.addProperty(PropertyConfig(propertyName, myString, ncol));
		} else if (children.name() == CHILDREN_TAG) {
			for (auto childrennode : children.children()) {
				if (childrennode.name() == DRIVER_TAG) {
					loadTestController(childrennode, config.getName(), configs);
				} else {
					unknownTagError(m_hardware_filename, (uint32) childrennode.offset_debug(), childrennode.name());
				}
			}
		} else {
			unknownTagError(m_hardware_filename, (uint32) children.offset_debug(), children.name());
		}
	}
	if (isnametag == false) {
		missingTagNameError(m_hardware_filename, node.offset_debug(), node.name(), NAME_TAG);
	}
	if (istypetag == false) {
		missingTagNameError(m_hardware_filename, node.offset_debug(), node.name(), TYPE_TAG);
	}

	// Adding the TestController configuration to the list
	// Copying the map into a vector
	configs.push_back(config);
}

/*
 * Creates the DOMNode from a TestControllerConfig. Creates recursively the children.
 *
 */
pugi::xml_node XMLFileConfiguration2::saveTestControllerNode(pugi::xml_node &parent,
		const TestControllerConfigNode &configNode) {

	pugi::xml_node node = parent.append_child(DRIVER_TAG.c_str());

	ostringstream channel;
	channel << configNode.config.getParentChannel();
	node.append_attribute(CHANNEL_ATTRIBUTE.c_str()) = channel.str().c_str();

	node.append_child(NAME_TAG.c_str()).append_child(pugi::node_pcdata).set_value(configNode.config.getName().c_str());
	node.append_child(TYPE_TAG.c_str()).append_child(pugi::node_pcdata).set_value(configNode.config.getType().c_str());

	string enabled = "false";
	if (configNode.config.isEnabled()) {
		enabled = "true";
	}
	node.append_child(ENABLED_TAG.c_str()).append_child(pugi::node_pcdata).set_value(enabled.c_str());

	vector<PropertyConfig> properties = configNode.config.getProperties();
	for (auto p : properties) {
		if (p.isArray() == false) {
			pugi::xml_node propnode = node.append_child(PROPERTY_TAG.c_str());
			propnode.append_attribute(NAME_ATTRIBUTE.c_str()) = p.getName().c_str();
			propnode.append_child(pugi::node_pcdata).set_value(p.getValue().c_str());
		}
	}

	for (auto p : properties) {
		if (p.isArray() & (p.getSize() != 0)) {
			vector<string> myarrayVector;
			string myseparator;
			ostringstream omycolsnbs;
			omycolsnbs << p.getColsNb();
			string mycolsnbs;
			int32 rows;
			pugi::xml_node array = node.append_child(PROPERTYARRAY_TAG.c_str());
			array.append_attribute(NAME_ATTRIBUTE.c_str()) = p.getName().c_str();
			array.append_attribute(PROPERTYARRAY_COLS_TAG.c_str()) = omycolsnbs.str().c_str();
			myarrayVector = p.getArray();
			if (p.getColsNb() != 0) {
				rows = int32(p.getSize() / p.getColsNb());
			} else {
				rows = 0;
			}

			int32 ind = 0;
			for (int32 myRowsInd = 0; myRowsInd < rows; ++myRowsInd) {
				string myarrayString;
				for (int32 myColsInd = 0; myColsInd < p.getColsNb(); myColsInd++) {
					myarrayString.append(myarrayVector.at(ind));
					if (myColsInd != (p.getColsNb() - 1)) {
						myarrayString.append(PROPERTYARRAY_CSV_TAG);
					}
					ind++;
				}
				array.append_child(ROW_TAG.c_str()).append_child(pugi::node_pcdata).set_value(myarrayString.c_str());
			}
		}
	}

	// Children

	vector<TestControllerConfigNode> children = configNode.children;
	if (children.size() != 0) {
		pugi::xml_node childrenNode = node.append_child(CHILDREN_TAG.c_str());
		for (auto c : children) {
			saveTestControllerNode(childrenNode, c);
		}
	}
	return node;
}

/************************************************************************
 *
 * EXPERIMENT CONTROLLER
 *
 ************************************************************************/

/*
 * loadExperimentControllers
 */
int32 XMLFileConfiguration2::loadExperimentControllers(vector<ExperimentControllerConfig> &configs) {
	pugi::xml_document doc;
	pugi::xml_parse_result result = doc.load_file(m_instrument_filename.c_str());
	if (result) {
		for (auto node : doc.children()) {
			if (node.name() == CONTROLLERCONFIG_TAG) {
				if (node.attribute(VERSION_TAG.c_str()).as_int() > 1) {
					for (auto node2 : node.children()) {
						if (node2.name() == CONTROLLER_TAG) {
							loadExperimentController(node2, configs);
						} else {
							unknownTagError(m_instrument_filename, (uint32) node2.offset_debug(), node2.name());
						}
					}
				} else {
					return -1;
				}
			} else {
				return -1;
			}
		}
	} else {
		printError(m_instrument_filename, (uint32) result.offset, result.description());
	}
	return 0;
}

/*
 * saveExperimentControllers
 */
void XMLFileConfiguration2::saveExperimentControllers(const vector<ExperimentControllerConfig> &configs, bool isBackup) {

	pugi::xml_document doc;
	pugi::xml_node rootElem = doc.append_child(CONTROLLERCONFIG_TAG.c_str());
	rootElem.append_attribute(VERSION_TAG.c_str()) = "2";

	// Save the module sequence id.
	int moduleSequenceId = ModuleManager::getInstance()->getSequenceId();
	std::string moduleSequenceIdString = std::to_string(moduleSequenceId);
	rootElem.append_attribute(MODULES_ID_TAG.c_str()) = moduleSequenceIdString.c_str();

	for (auto c : configs) {
		saveExperimentController(rootElem, c);
	}

	string buf;
	if (isBackup) {
		buf = m_instrument_backup_filename;
		ptime t(second_clock::local_time());
		buf.append(".");
		buf.append(to_simple_string(t));
	} else {
		buf = m_instrument_filename;
	}
	doc.save_file(buf.c_str(), PUGIXML_TEXT("  "));
}

/*
 * createExperimentControllerDOMNode
 */
void XMLFileConfiguration2::saveExperimentController(pugi::xml_node &parent, const ExperimentControllerConfig &configNode) {
	pugi::xml_node node = parent.append_child(CONTROLLER_TAG.c_str());

	node.append_child(NAME_TAG.c_str()).append_child(pugi::node_pcdata).set_value(configNode.getName().c_str());
	node.append_child(TYPE_TAG.c_str()).append_child(pugi::node_pcdata).set_value(configNode.getType().c_str());

	string enabled = "false";
	if (configNode.isEnabled()) {
		enabled = "true";
	}
	node.append_child(ENABLED_TAG.c_str()).append_child(pugi::node_pcdata).set_value(enabled.c_str());

	// Local properties

	vector<PropertyConfig> properties = configNode.getProperties();
	for (auto p : properties) {
		if (p.isArray() == false) {
			pugi::xml_node propnode = node.append_child(PROPERTY_TAG.c_str());
			propnode.append_attribute(NAME_ATTRIBUTE.c_str()) = p.getName().c_str();
			propnode.append_child(pugi::node_pcdata).set_value(p.getValue().c_str());
		}
	}

	for (auto p : properties) {
		if (p.isArray() & (p.getSize() != 0)) {
			vector<string> myarrayVector;
			string myseparator;
			ostringstream omycolsnbs;
			omycolsnbs << p.getColsNb();
			string mycolsnbs;
			int32 rows;

			pugi::xml_node array = node.append_child(PROPERTYARRAY_TAG.c_str());
			array.append_attribute(NAME_ATTRIBUTE.c_str()) = p.getName().c_str();
			array.append_attribute(PROPERTYARRAY_COLS_TAG.c_str()) = omycolsnbs.str().c_str();
			myarrayVector = p.getArray();
			if (p.getColsNb() != 0) {
				rows = int32(p.getSize() / p.getColsNb());
			} else {
				rows = 0;
			}

			int32 ind = 0;
			for (int32 myRowsInd = 0; myRowsInd < rows; ++myRowsInd) {
				string myarrayString;
				for (int32 myColsInd = 0; myColsInd < p.getColsNb(); myColsInd++) {
					myarrayString.append(myarrayVector.at(ind));
					if (myColsInd != (p.getColsNb() - 1)) {
						myarrayString.append(PROPERTYARRAY_CSV_TAG);
					}
					ind++;
				}
				array.append_child(ROW_TAG.c_str()).append_child(pugi::node_pcdata).set_value(myarrayString.c_str());
			}
		}
	}

	string family = configNode.getFamily();
	if (family.empty() == false) {
		node.append_child(FAMILY_TAG.c_str()).append_child(pugi::node_pcdata).set_value(family.c_str());
	}

	string subfamily = configNode.getSubfamily();
	if (subfamily.empty() == false) {
		node.append_child(SUBFAMILY_TAG.c_str()).append_child(pugi::node_pcdata).set_value(subfamily.c_str());
	}

	vector<RoleConfig> configs = configNode.getRoles();
	for (auto p : configs) {
		if (p.getRoleType() == EXPERIMENT_CONTROLLER_TYPE) {
			pugi::xml_node cnode = node.append_child(CONTROLLERPTR_TAG.c_str());
			cnode.append_attribute(NAME_ATTRIBUTE.c_str()) = p.getName().c_str();
			cnode.append_child(pugi::node_pcdata).set_value(p.getWho().c_str());
		} else if (p.getRoleType() == TEST_CONTROLLER_TYPE) {
			pugi::xml_node cnode = node.append_child(DRIVERPTR_TAG.c_str());
			cnode.append_attribute(NAME_ATTRIBUTE.c_str()) = p.getName().c_str();
			cnode.append_child(pugi::node_pcdata).set_value(p.getWho().c_str());
		}
	}
}

/*
 * createExperimentController
 */
void XMLFileConfiguration2::loadExperimentController(const pugi::xml_node &node,
		std::vector<ExperimentControllerConfig> &configs) {
	ExperimentControllerConfig config;

	bool isnametag = false;
	bool istypetag = false;

	for (auto children : node.children()) {
		if (children.name() == NAME_TAG) {
			config.setName(children.child_value());
			isnametag = true;
		} else if (children.name() == TYPE_TAG) {
			config.setType(children.child_value());
			istypetag = true;
		} else if (children.name() == ENABLED_TAG) {
			if (children.child_value() == string("true")) {
				config.setEnabled(true);
			} else {
				config.setEnabled(false);
			}
		} else if (children.name() == PROPERTY_TAG) {
			pugi::xml_attribute attr = children.attribute(NAME_ATTRIBUTE.c_str());
			string propertyName;
			if (attr) {
				propertyName = attr.value();
			} else {
				missingAttributeError(m_instrument_filename, (uint32) children.offset_debug(), NAME_ATTRIBUTE, PROPERTY_TAG);
			}
			string propertyValue = children.child_value();
			config.addProperty(PropertyConfig(propertyName, propertyValue));
		} else if (children.name() == PROPERTYARRAY_TAG) {
			pugi::xml_attribute attr = children.attribute(NAME_ATTRIBUTE.c_str());
			string propertyName;
			if (attr) {
				propertyName = attr.value();
			} else {
				missingAttributeError(m_instrument_filename, (uint32) children.offset_debug(), NAME_ATTRIBUTE, PROPERTY_TAG);
			}
			attr = children.attribute(PROPERTYARRAY_COLS_TAG.c_str());
			string propertyArrayCols;
			if (attr) {
				propertyArrayCols = attr.value();
			} else {
				missingAttributeError(m_instrument_filename, (uint32) children.offset_debug(), PROPERTYARRAY_COLS_TAG,
						PROPERTYARRAY_TAG);
			}
			int32 ncol = stoi(propertyArrayCols);
			vector<string> myString;
			for (auto row : children.children()) {
				if (row.name() == ROW_TAG) {
					string arrayValue = row.child_value();
					uint32 lastindex = 0;
					for (uint32 colId = 0; colId < (uint32) ncol; colId++) {
						uint32 myindex = arrayValue.find(PROPERTYARRAY_CSV_TAG, lastindex);
						myString.push_back(arrayValue.substr(lastindex, myindex - lastindex));
						lastindex = myindex + 1;
					}
				} else {
					unknownTagError(m_instrument_filename, (uint32) row.offset_debug(), row.name());
				}
			}
			config.addProperty(PropertyConfig(propertyName, myString, ncol));
		} else if (children.name() == FAMILY_TAG) {
			config.setFamily(children.child_value());
		} else if (children.name() == SUBFAMILY_TAG) {
			config.setSubfamily(children.child_value());
		} else if (children.name() == TYPE_TAG) {
			config.setType(children.child_value());
		} else if (children.name() == CONTROLLERPTR_TAG) {
			pugi::xml_attribute attr = children.attribute(NAME_ATTRIBUTE.c_str());
			string roleName;
			if (attr) {
				roleName = attr.value();
			} else {
				missingAttributeError(m_instrument_filename, (uint32) children.offset_debug(), NAME_ATTRIBUTE,
						CONTROLLERPTR_TAG);
			}
			config.addRole(RoleConfig(EXPERIMENT_CONTROLLER_TYPE, roleName, children.child_value()));
		} else if (children.name() == DRIVERPTR_TAG) {
			pugi::xml_attribute attr = children.attribute(NAME_ATTRIBUTE.c_str());
			string roleName;
			if (attr) {
				roleName = attr.value();
			} else {
				missingAttributeError(m_instrument_filename, (uint32) children.offset_debug(), NAME_ATTRIBUTE,
						DRIVERPTR_TAG);
			}
			config.addRole(RoleConfig(TEST_CONTROLLER_TYPE, roleName, children.child_value()));
		} else {
			unknownTagError(m_instrument_filename, (uint32) children.offset_debug(), children.name());
		}
	}
	if (isnametag == false) {
		missingTagNameError(m_instrument_filename, node.offset_debug(), node.name(), NAME_TAG);
	}
	if (istypetag == false) {
		missingTagNameError(m_instrument_filename, node.offset_debug(), node.name(), TYPE_TAG);
	}
	configs.push_back(config);
}

/*
 * printError
 */
std::string XMLFileConfiguration2::printError(const std::string &file, uint32 offset, const std::string &message) {
	throw ConfigurationError(XMLErrors::printError(file, offset, message));
}

