/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef XMLFILECONFIGURATION2_H
#define XMLFILECONFIGURATION2_H

#include <string>
#include <vector>

#include "PersistenceServices/Configuration/AbstractConfiguration.h"
#include <pugixml.hpp>
#include "PersistenceServices/XMLErrors.h"

/*!
 * \class XMLFileConfiguration2
 * \brief Save and load xml instrument and hardware configuration
 */
class XMLFileConfiguration2: public AbstractConfiguration, public XMLErrors {

public:

	/*!
	 * \brief Constructor
	 */
	XMLFileConfiguration2();

	/*!
	 * \brief Destructor
	 */
	virtual ~XMLFileConfiguration2();

	/*!
	 * \brief Load test controllers from xml file
	 * \param[out] configs TestController configuration vector
	 * \return error code
	 * \throws XmlConfigurationError Xml File Configuration Error
	 */
	virtual int32 loadTestControllers(std::vector<TestControllerConfig> &configs);

	/*!
	 * \brief Save test controllers in the xml file
	 * \param[in] configs TestController configuration vector
	 * \return error code
	 * \throws XmlConfigurationError Xml File Configuration Error
	 */
	virtual void saveTestControllers(const std::vector<TestControllerConfigNode> &configs, bool isBackup);

	/*!
	 * \brief Load experiment controllers from xml file
	 * \param[out] configs ExperimentController configuration vector
	 * \return error code
	 * \throws XmlConfigurationError Xml File Configuration Error
	 */
	virtual int32 loadExperimentControllers(std::vector<ExperimentControllerConfig> &configs);

	/*!
	 * \brief Save experiment controllers in the xml file
	 * \param[in] configs ExperimentController configuration vector
	 * \return error code
	 * \throws XmlConfigurationError Xml File Configuration Error
	 */
	virtual void saveExperimentControllers(const std::vector<ExperimentControllerConfig> &configs, bool isBackup);

private:

	/*!
	 * \brief Load test controller node of the xml file
	 * \param[in] node Xml current Node
	 * \param[in] parentName Parent name of the current node
	 * \param[out] configs TestController configuration vector
	 * \throws XmlConfigurationError Xml File Configuration Error
	 */
	void loadTestController(const pugi::xml_node &node, const std::string &parentName,
			std::vector<TestControllerConfig> &configs);

	/*!
	 * \brief Save test controller node in the xml file
	 * \param[in] parent Xml parent Node
	 * \param[in] configs TestController configuration vector
	 * \return xml node
	 */
	pugi::xml_node saveTestControllerNode(pugi::xml_node &parent, const TestControllerConfigNode &configNode);

	/*!
	 * \brief Load experiment controller node of the xml file
	 * \param[in] node Xml current Node
	 * \param[out] configs ExperimentController configuration vector
	 * \throws XmlConfigurationError Xml File Configuration Error
	 */
	void loadExperimentController(const pugi::xml_node &node, std::vector<ExperimentControllerConfig> &configs);

	/*!
	 * \brief Save experiment controller node in the xml file
	 * \param[in] parent Xml parent Node
	 * \param[in] configs ExperimentController configuration vector
	 * \return xml node
	 */
	void saveExperimentController(pugi::xml_node &parent, const ExperimentControllerConfig &configNode);

	/*!
	 * \brief Print error in terminal using red bold characters
	 * \param[in] file XML file name
	 * \param[in] offset Character offset in the file
	 * \param[in] message Message to print
	 * \return the string error
	 */
	virtual std::string printError(const std::string &file, uint32 offset, const std::string &message);

	//! Name of configuration files
	static const std::string HARDWARE_CONF_FILE_NAME;
	static const std::string INSTRUMENT_CONF_FILE_NAME;

	//! XML tags
	static const std::string VERSION_TAG;
	static const std::string MODULES_ID_TAG;
	static const std::string DRIVERCONFIG_TAG;
	static const std::string DRIVER_TAG;
	static const std::string CONTROLLERCONFIG_TAG;
	static const std::string CONTROLLER_TAG;
	static const std::string NAME_TAG;
	static const std::string TYPE_TAG;
	static const std::string ENABLED_TAG;
	static const std::string PROPERTY_TAG;
	static const std::string NAME_ATTRIBUTE;
	static const std::string CHILDREN_TAG;
	static const std::string CHANNEL_ATTRIBUTE;
	static const std::string FAMILY_TAG;
	static const std::string SUBFAMILY_TAG;
	static const std::string CONTROLLERPTR_TAG;
	static const std::string DRIVERPTR_TAG;
	static const std::string PROPERTYARRAY_TAG;
	static const std::string PROPERTYARRAY_CSV_TAG;
	static const std::string PROPERTYARRAY_COLS_TAG;
	static const std::string ROW_TAG;

	//! Complete name with path of configuration files
	std::string m_hardware_filename;
	std::string m_instrument_filename;
	std::string m_hardware_backup_filename;
	std::string m_instrument_backup_filename;

};

#endif
