/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "LoggingServices/Level.h"

#include "AgentCollectorProxy.h"
#include "AgentPolicy.h"
#include "InstrumentManager/NameManager.h"

using namespace std;

/**
 * Init singleton instance
 */
AgentPolicy* AgentPolicy::m_instance = 0;

/*
 * Destructor
 */
AgentPolicy::~AgentPolicy() {
	m_activeECs.clear();
	if (m_agentCollector != NULL) {
		delete m_agentCollector;
		m_agentCollector = 0;
	}
	m_instance = 0;
}

/*
 * getInstance
 */
AgentPolicy* AgentPolicy::getInstance() {
	if (m_instance == 0) {
		m_instance = new AgentPolicy();
	}
	return m_instance;
}

/*
 * activateEC
 */
void AgentPolicy::activateEC(AbstractController* anEC) {
	if (anEC == 0) {
		return;
	}

	lock_guard<recursive_mutex> lock(m_activeECsMutex);

	bool found = false;
	vector<string>::iterator ecIDsIterator;
	for (ecIDsIterator = m_activeECs.begin(); ecIDsIterator
			!= m_activeECs.end() && (found == false); ecIDsIterator++) {
		if (*ecIDsIterator == anEC->getName()) {
			found = true;
		}
	}
	if (found == false) {
		m_activeECs.push_back(anEC->getName());
	}
}

/*
 * deactivateEC
 */
void AgentPolicy::deactivateEC(AbstractController* anEC) {
	if (anEC == 0) {
		return;
	}

	lock_guard<recursive_mutex> lock(m_activeECsMutex);

	vector<string>::iterator ecIDsIterator;
	for (ecIDsIterator = m_activeECs.begin(); ecIDsIterator
			!= m_activeECs.end(); ecIDsIterator++) {
		if (*ecIDsIterator == anEC->getName()) {
			m_activeECs.erase(ecIDsIterator);
			break;
		}
	}
}

/*
 * canI
 */
bool AgentPolicy::canI(AbstractController* newEC) {
	if (newEC == 0) {
		return false;
	}

	lock_guard<recursive_mutex> lock1(m_activeECsMutex);

	if (m_activeECs.size() == 0) {
		return true;
	}

	vector<string>::iterator activatedIDsIterator;

	lock_guard<recursive_mutex> lock2(m_compatibilityTableMutex);

	string ecToActivate = newEC->getName();

	// Loop on activated IDs: for each activatedID, search if it is
	// in the authorized IDs
	for (activatedIDsIterator = m_activeECs.begin(); activatedIDsIterator
			!= m_activeECs.end(); activatedIDsIterator++) {

		if (*activatedIDsIterator != ecToActivate) {
			multimap<string,string>::iterator iter = m_compatibilityTable.begin();
			while (iter != m_compatibilityTable.end()) {
				if ((iter->first == *activatedIDsIterator) && (iter->second == ecToActivate)) {
					return false;
				}
				if ((iter->second == *activatedIDsIterator) && (iter->first == ecToActivate)) {
					return false;
				}
				++iter;
			}
		} else {
			// Could not move the same if already activate.
			return false;
		}
	}
	return true;
}

/*
 * Constructor
 */
AgentPolicy::AgentPolicy() {
	m_agentCollector = new AgentCollectorProxy();
	m_agentCollector->loadTable(m_compatibilityTable);
}

/*
 * checkTable
 */
void AgentPolicy::checkTable(const string& controllerList) {
	multimap<string,string>::iterator iter = m_compatibilityTable.begin();
	while (iter != m_compatibilityTable.end()) {
		cout << iter->first << " - " << iter->second << endl;
		if (NameManager::getInstance()->controllerExists(iter->first) == false) {
			cerr << "Agent Policy Table Warning !!! Controller " <<  iter->first << " doesn't exist" << endl;
		}
		if (NameManager::getInstance()->controllerExists(iter->second) == false) {
			cerr << "Agent Policy Table Warning !!! Controller " <<  iter->second << " doesn't exist" << endl;
		}
		iter++;
	}
}
