/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CONDITION_H
#define CONDITION_H

#include <mutex>
#include <condition_variable>

/**
 * Condition interface to define conditions to the end of synchronization for Sender posts.
 */
class Condition
{
public:
  /**
   * Constructor
   */
  Condition();
  /**
   * Constructor
   */
  virtual ~Condition() {};
  /**
   * Notify the current condition of the current thread
   */
  void notify();
   
  /**
   * Return the condition value.
   */
  virtual bool operator()() = 0;
  
  /**
   * Called at the end of an endIf function.
   */
  virtual void finalise();
  
  /**
   * Sets the condition finished
   */
  virtual void setFinished();
  
  /**
   * Returns true if the condition is finished
   */
  bool isFinished() const;

private:
  std::mutex* m_mutex;
  std::condition_variable* m_condition;
  bool m_finished;
};

/**
 * Class defining a new Condition from the union of two Condition's
 */
class OrCondition : public Condition
{
public:
  OrCondition(const Condition& A, const Condition& B)
  : m_A(const_cast<Condition&>(A)), m_B(const_cast<Condition&>(B)) {}

  virtual bool operator()();
  virtual void finalise();
  virtual void setFinished();

private:
  Condition& m_A;
  Condition& m_B;
};

/**
 * Overloaded operator+ to provide an intuitive syntax
 */
inline OrCondition operator|| (const Condition& A, const Condition& B)
{
  return OrCondition(A, B);
}

/**
 * Class defining a new Condition from the intersection of two Condition's
 */
class AndCondition : public Condition
{
public:
  AndCondition(const Condition& A, const Condition& B)
  : m_A(const_cast<Condition&>(A)), m_B(const_cast<Condition&>(B)) {}

  virtual bool operator()();
  virtual void finalise();
  virtual void setFinished();

private:
  Condition& m_A;
  Condition& m_B;
};

/**
 * Overloaded operator* to provide an intuitive syntax
 */
inline AndCondition operator&& (const Condition& A, const Condition& B)
{
  return AndCondition(A, B);
}

/**
 * Class defining a new Condition from the negation of a Condition
 */
class NotCondition : public Condition
{
public:
  NotCondition(const Condition& A)
  : m_A(const_cast<Condition&>(A)) {}
  
  virtual bool operator()();
  virtual void finalise();
  virtual void setFinished();

private:
  Condition& m_A;
};

/**
 * Overloaded operator! to provide an intuitive syntax
 */
inline NotCondition operator! (const Condition& A)
{
  return NotCondition(A);
}

/**
 * Class defining a new Condition from the negation of a Condition
 */
class Eval : public Condition
{
public:
  Eval(bool value)
  : m_value(value) {}
  
  virtual bool operator()();

private:
  bool m_value;
};

#endif
