/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "Scheduler/Scheduler.h"
#include "InstrumentAbstraction/ClientCommand.h"
#include "DeviceContainer/HardwareCommand.h"
#include "InstrumentAbstraction/AbstractController.h"
#include <common/output/OutputStream.h>

using namespace std;

std::string CommandsInfo::toString() const {
	std::string result;

	result += "current commands:\n";
	for (std::list<std::string>::const_iterator c = currentCommandNames.begin(); c != currentCommandNames.end(); c++) {
		result += "\t" + *c + "\n";
	}
	result += "waiting commands:\n";
	for (std::list<std::string>::const_iterator c = waitingCommandNames.begin(); c != waitingCommandNames.end(); c++) {
		result += "\t" + *c + "\n";
	}

	return result;
}

template<>
void Scheduler<ClientCommand>::s_execute(SenderCommand<ClientCommand>* aCommand) {
	//std::cout << "executing " << aCommand->getCommand() << " (" << aCommand->getIndex() << ") at address " << &(aCommand->getCommand()) << std::endl;

	// Starting command
	aCommand->getCommand().execute();

	//std::cout << "end executing " << aCommand->getCommand() << " (" << aCommand->getIndex() << ") at address " << &(aCommand->getCommand()) << std::endl;

	// Notifying for synchronization
	if (aCommand->getMutex() != 0) {

		// Synchronization
		std::scoped_lock lock(*aCommand->getMutex());

		// Erasing command from active commands
		std::list<SenderCommand<ClientCommand>*> &activeCommands = *aCommand->getActiveCommands();
		std::list<SenderCommand<ClientCommand>*>::iterator c = std::find(activeCommands.begin(), activeCommands.end(), aCommand);
		if (c != activeCommands.end()) {
			activeCommands.erase(c);
		}

		// Notifying Sender
		aCommand->getCondition()->notify_one();
	}

	// Notifying Scheduler
	Scheduler<ClientCommand>::getInstance()->m_notifyEndOfCommand(aCommand);

	// Deleting command
	delete aCommand;

	//cout << "d " << aCommand << endl;

	//  std::cout << "end execute" << std::endl;
}

template<>
void Scheduler<HardwareCommand>::s_execute(SenderCommand<HardwareCommand>* aCommand) {
	//std::cout << "executing " << aCommand->getCommand() << " (" << aCommand->getIndex() << ") at address " << &(aCommand->getCommand()) << std::endl;
	//std::cout << "executing hardware command " << aCommand->getCommand() << std::endl;

	// Starting command
	aCommand->getCommand().execute();

	//std::cout << "end executing " << aCommand->getCommand() << " (" << aCommand->getIndex() << ") at address " << &(aCommand->getCommand()) << std::endl;

	// Notifying for synchronization
	if (aCommand->getMutex() != 0) {

		// Synchronization
		std::scoped_lock lock(*aCommand->getMutex());

		// Erasing command from active commands
		std::list<SenderCommand<HardwareCommand>*> &activeCommands = *aCommand->getActiveCommands();
		std::list<SenderCommand<HardwareCommand>*>::iterator c = std::find(activeCommands.begin(), activeCommands.end(), aCommand);
		if (c != activeCommands.end()) {
			activeCommands.erase(c);
		}

		// Notifying Sender
		aCommand->getCondition()->notify_one();
	}

	// Notifying Scheduler
	Scheduler<HardwareCommand>::getInstance()->m_notifyEndOfCommand(aCommand);

	// Deleting command
	delete aCommand;

	//cout << "deleted " << aCommand << endl;

	//  std::cout << "end execute" << std::endl;
}



template<>
void Scheduler<ClientCommand>::m_processNewCommand(SenderCommand<ClientCommand>* aCommand) {
	std::scoped_lock lock(m_commandsMutex);

	aCommand->setIndex(m_index++);

	bool blocked = m_isBlocked(aCommand);
//	bool rejected = false;//aCommand->getCommand().reject(blocked);

	//std::cout << "process command " << aCommand->getCommand() << " (" << aCommand->getIndex() << ") " << blocked << ", " << rejected << std::endl;

//	if (!rejected) {
		if (blocked) {
			// aCommands becomes a waiting command
			m_waitingCommands.push_back(aCommand);

			if (m_waitingCommands.size() % 100 == 0) {
				cout << "Client waiting commands = " << m_waitingCommands.size() << endl;
			}

			if (aCommand->getCommand().getController() != NULL) {
				common::warn << "Scheduler refuses command " << aCommand->getCommand().toString() << " and delays it" << common::eol;
				aCommand->getCommand().getController()->sendRefuseEvent(aCommand->getCommand().getName());
			}

		} else {

			//if (aCommand->getCommand().getName() == "start") {
			//	std::cout << "process " << aCommand->getCommand() << std::endl;
			//}

			//cout << "p " << aCommand->getCommand().getController() << endl;
			//cout << aCommand->getCommand().getController()->getName() << endl;

			// Place her the activation of ec in agent policy instead of ClientCommand classs, because there is a leak
			// synchronism. The test of agent couldn't working id the register of agent policy not done before test of blocking on the parallel commad.
			if (aCommand->getCommand().getController() != NULL) {
				AgentPolicy::getInstance()->activateEC(aCommand->getCommand().getController());
			}

			// Submit command.
			ThreadPool::get().execute(std::bind(&Scheduler<ClientCommand>::s_execute, aCommand));

			// Adding command to history
			Scheduler<ClientCommand>::getInstance()->m_addCommandHistory(aCommand->getCommand());

			// aCommands becomes a current command
			m_currentCommands.push_back(aCommand);
		}
//	} else {
//		common::warn << "Scheduler rejects the command " << aCommand->getCommand().toString() << common::eol;
//
//		if (aCommand->getMutex() != 0) {
//			// Synchronization
//			std::scoped_lock lock2(*aCommand->getMutex());
//
//			// Erasing command from active commands
//			std::list<SenderCommand<ClientCommand>*> &activeCommands = *aCommand->getActiveCommands();
//			std::list<SenderCommand<ClientCommand>*>::iterator c = std::find(activeCommands.begin(), activeCommands.end(), aCommand);
//			if (c != activeCommands.end())
//				activeCommands.erase(c);
//
//			// Notifying Sender
//			aCommand->getCondition()->notify_one();
//		}
//	}
}

template<>
void Scheduler<HardwareCommand>::m_processNewCommand(SenderCommand<HardwareCommand>* aCommand) {
	std::scoped_lock lock(m_commandsMutex);

	aCommand->setIndex(m_index++);

	bool blocked = m_isBlocked(aCommand);
//	bool rejected = false;//aCommand->getCommand().reject(blocked);

	//std::cout << "process command " << aCommand->getCommand() << " (" << aCommand->getIndex() << ") " << blocked << ", " << rejected << std::endl;

//	if (!rejected) {
		if (blocked) {
			// check if command must wait
			if (!m_isWaitingCommand(aCommand)) {
				m_waitingCommands.push_back(aCommand);
			} else {
				//cout << "rejects already waiting command " << aCommand->getCommand() << endl;
				if (aCommand->getMutex() != 0) {
					// Synchronization
					std::scoped_lock lock2(*aCommand->getMutex());

					// Erasing command from active commands
					std::list<SenderCommand<HardwareCommand>*> &activeCommands = *aCommand->getActiveCommands();
					std::list<SenderCommand<HardwareCommand>*>::iterator c = std::find(activeCommands.begin(), activeCommands.end(), aCommand);
					if (c != activeCommands.end())
						activeCommands.erase(c);

					// Notifying sender
					aCommand->getCondition()->notify_one();
				}

				delete aCommand;

				return;
			}

			if (m_waitingCommands.size() % 200 == 0) {
				cout << "Hardware waiting commands = " << m_waitingCommands.size() << endl;
			}

			if (aCommand->getCommand().getController() != NULL) {
				common::warn << "Scheduler refuses command " << aCommand->getCommand().toString() << " and delays it" << common::eol;
				aCommand->getCommand().getController()->sendRefuseEvent(aCommand->getCommand().getName());
			}
		} else {

			//if (aCommand->getCommand().getName() == "start") {
			//	std::cout << "process " << aCommand->getCommand() << std::endl;
			//}

			// Place her the activation of ec in agent policy instead of ClientCommand classs, because there is a leak
			// synchronism. The test of agent couldn't working id the register of agent policy not done before test of blocking on the parallel commad.
			if (aCommand->getCommand().getController() != NULL) {
				AgentPolicy::getInstance()->activateEC(aCommand->getCommand().getController());
			}

			// Submit command.
			ThreadPool::get().execute(std::bind(&Scheduler<HardwareCommand>::s_execute, aCommand));

			// Adding command to history
			Scheduler<HardwareCommand>::getInstance()->m_addCommandHistory(aCommand->getCommand());

			// aCommands becomes a current command
			m_currentCommands.push_back(aCommand);
		}
//	} else {
//		common::warn << "Scheduler rejects the command " << aCommand->getCommand().toString() << common::eol;
//
//		if (aCommand->getMutex() != 0) {
//			// Synchronization
//			std::scoped_lock lock2(*aCommand->getMutex());
//
//			// Erasing command from active commands
//			std::list<SenderCommand<HardwareCommand>*> &activeCommands = *aCommand->getActiveCommands();
//			std::list<SenderCommand<HardwareCommand>*>::iterator c = std::find(activeCommands.begin(), activeCommands.end(), aCommand);
//			if (c != activeCommands.end())
//				activeCommands.erase(c);
//
//			// Notifying Sender
//			aCommand->getCondition()->notify_one();
//		}
//	}
}



