/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef SCHEDULER_H
#define SCHEDULER_H

#include "Scheduler/SenderCommand.h"
#include "Scheduler/TodoList.h"
#include "Scheduler/AgentPolicy/AgentPolicy.h"
#include "Scheduler/ThreadPool.h"

#include <boost/bind.hpp>
#include <iostream>
#include <vector>
#include <utility>
#include <memory>
#include <thread>
#include <atomic>

class CommandsInfo {

public:
	CommandsInfo() :
		nbthreads(0) {
	}

	std::list<std::string> currentCommandNames;
	std::list<std::string> waitingCommandNames;
	int32 nbthreads;

	std::string toString() const;
};

/**
 * Class dispatching Commands.
 */
template<class T>
class Scheduler {

public:
	/**
	 * Singleton method
	 */
	static Scheduler<T>* getInstance();

	/**
	 * Start the thread.
	 */
	void start();

	/**
	 * Stops the scheduler. Empties the waiting command list.
	 */
	void stop();

	/**
	 * Delete the singleton. Use it only when the server is being shutdowned. Otherwise,
	 * if start is called afterwards, a new thread will be launched and there will be two
	 * threads for s_loop.
	 */
	void reset();

	/**
	 * Return true if there is no command to process or active
	 */
	bool isReady();

	/**
	 * Gets the number of commands.
	 */
	int32 getNumberOfCommands();

	/**
	 * Get the command history. Return the list of commands executed.
	 */
	std::pair<std::vector<T>, int32> getCommandHistory();

	/**
	 * historySize
	 */
	static const int32 historySize = 32;

	/**
	 * Scheduler view information. Used by SchedulerView.
	 */
	CommandsInfo getCurrentCommandsInfo();

	/**
	 * Get the waiting commands info. Used by SchedulerView.
	 */
	CommandsInfo getWaitingCommandsInfo();

private:
	/**
	 * Constructor
	 */
	Scheduler();

	/**
	 * Destructor
	 */
	~Scheduler();

	/**
	 * Adds a new command into the history.
	 */
	void m_addCommandHistory(const T& aCommand);

	/**
	 * Process a new command
	 */
	void m_processNewCommand(SenderCommand<T>* aCommand);

	/**
	 * Loop for the thread
	 */
	static void s_loop();

	/**
	 * Execute the sender command in a new thread.
	 */
	static void s_execute(SenderCommand<T>* aCommand);

	/**
	 * Notify the Scheduler the end of the command. Called in the launched
	 * threads.
	 */
	void m_notifyEndOfCommand(SenderCommand<T>* aCommand);

	/**
	 * Return true if the command can be launched.
	 */
	bool m_isBlocked(SenderCommand<T>* aCommand);

	/**
	 * Returns true if the command is already waiting to be executed.
	 */
	bool m_isWaitingCommand(SenderCommand<T>* aCommand);

	/**
	 * Singleton instance
	 */
	static Scheduler<T>* s_instance;
	static std::mutex m_instanceMutex;

	std::unique_ptr<std::thread> m_thread;

	/**
	 * Commands
	 */
	std::list<SenderCommand<T>*> m_currentCommands;
	std::list<SenderCommand<T>*> m_waitingCommands;
	std::mutex m_commandsMutex;

	/**
	 * Command history
	 */
	std::vector<T> m_commandHistory;
	int32 m_historyCursor;
	std::mutex m_historyMutex;

	int32 m_index;

};

/*******************************************************************************
 * Template implementation. This must be done in the .h file. Does not
 * compile in the .cpp file (template class constraint).
 */

template<class T>
Scheduler<T>* Scheduler<T>::s_instance = 0;

template<class T>
std::mutex Scheduler<T>::m_instanceMutex;

template<class T>
Scheduler<T>* Scheduler<T>::getInstance() {
	if (s_instance == 0) {
		// Double-checked optimization pattern
		std::scoped_lock lock(m_instanceMutex);
		if (s_instance == 0)
			s_instance = new Scheduler<T> ();
	}

	return s_instance;
}

template<class T>
void Scheduler<T>::start() {

	// Launching thread
	m_thread.reset(new std::thread(&Scheduler<T>::s_loop));
}

template<class T>
void Scheduler<T>::stop() {

	std::scoped_lock lock(m_commandsMutex);

	m_waitingCommands.clear();
	TodoList<T>::getInstance()->stop();

	// Join the thread.
	m_thread->join();
}

template<class T>
void Scheduler<T>::reset() {
	if (s_instance != nullptr) {
		delete s_instance;
	}
}

template<class T>
bool Scheduler<T>::isReady() {
	std::scoped_lock lock(m_commandsMutex);

	return m_currentCommands.empty() && m_waitingCommands.empty();
}

template<class T>
int32 Scheduler<T>::getNumberOfCommands() {
	std::scoped_lock lock(m_commandsMutex);

	return (int32) (m_currentCommands.size() + m_waitingCommands.size());
}

template<class T>
std::pair<std::vector<T>, int32> Scheduler<T>::getCommandHistory() {
	std::scoped_lock lock(m_historyMutex);

	return std::pair<std::vector<T>, int32>(m_commandHistory, m_historyCursor - 1);
}

template<class T>
Scheduler<T>::Scheduler() :
	m_commandHistory(historySize, T()),
	m_historyCursor(0),
	m_index(0) {
}

template<class T>
Scheduler<T>::~Scheduler() {
}

template<class T>
void Scheduler<T>::m_addCommandHistory(const T& aCommand) {
	// Saving command in history
	std::scoped_lock lock(m_historyMutex);

	m_commandHistory[m_historyCursor] = aCommand;

	// updating cursor
	++m_historyCursor;
	m_historyCursor %= historySize;
}

template<class T>
void Scheduler<T>::s_loop() {
	// infinite loop for the thread
	while (true) {
		// Get new command and launch a thread
		SenderCommand<T>* aCommand = TodoList<T>::getInstance()->consume();

		// A command null means that stop has been called.
		if (aCommand == nullptr) {
			break;
		}

		// Processing command
		Scheduler<T>::getInstance()->m_processNewCommand(aCommand);
	}
}

template<class T>
void Scheduler<T>::m_notifyEndOfCommand(SenderCommand<T>* aCommand) {
	std::scoped_lock lock(m_commandsMutex);

	// Erasing aCommand from current commands
	typename std::list<SenderCommand<T>*>::iterator c = std::find(m_currentCommands.begin(), m_currentCommands.end(), aCommand);
	if (c != m_currentCommands.end()) {
		m_currentCommands.erase(c);

		// Trying to launch a new command
		// Iterating the list
		typename std::list<SenderCommand<T>*>::iterator w = m_waitingCommands.begin();
		while (w != m_waitingCommands.end()) {
			bool blocked = m_isBlocked(*w);
//			bool rejected = false;//(*w)->getCommand().reject(blocked);

//			if (!rejected) {
				if (blocked) {
					++w;
				} else {
					// Place her the activation of ec in agent policy instead of ClientCommand classs, because there is a leak
					// synchronism. The test of agent couldn't working id the register of agent policy not done before test of blocking on the parallel commad.
					if ((*w)->getCommand().getController() != NULL) {
						AgentPolicy::getInstance()->activateEC((*w)->getCommand().getController());
					}

					// Submit command.
					ThreadPool::get().execute(std::bind(&Scheduler<T>::s_execute, *w));

					// Adding command to history
					Scheduler<T>::getInstance()->m_addCommandHistory((*w)->getCommand());

					// adding to current commands.
					m_currentCommands.push_back(*w);

					// erasing from waiting commands and getting next command.
					w = m_waitingCommands.erase(w);
				}
//			} else {
//				// what to do if rejected. first erase from list
//				w = m_waitingCommands.erase(w);
//			}
		}
	}
}

template<class T>
bool Scheduler<T>::m_isBlocked(SenderCommand<T>* aCommand) {
	// No mutex! m_commandsMutex is supposed to be already locked.
	// Iterating the current commands
	for (typename std::list<SenderCommand<T>*>::const_iterator c = m_currentCommands.begin(); c != m_currentCommands.end(); ++c) {
		if ((*c)->getCommand().isBlocking(aCommand->getCommand())) {
			//std::cout << "command " << (*c)->getCommand() << "(" << (*c)->getIndex() << ") blocks " << aCommand->getCommand() << " (" << aCommand->getIndex() << ")" << std::endl;
			return true;
		}
	}

	return false;
}

template<class T>
bool Scheduler<T>::m_isWaitingCommand(SenderCommand<T>* aCommand) {

	// Test if command already post , just for read or status
	if ((aCommand->getPCommand()->getName() == "read") || (aCommand->getPCommand()->getName() == "status")) {
		for (typename std::list<SenderCommand<T>*>::const_iterator c = m_waitingCommands.begin(); c != m_waitingCommands.end(); c++) {
			if (aCommand->getPCommand()->equal((*c)->getPCommand())) {
				// Already in the list so don't add it.
				return true;
			}
		}
	}
	return false;
}

template<class T>
CommandsInfo Scheduler<T>::getCurrentCommandsInfo() {
	CommandsInfo info;
	std::scoped_lock lock(m_commandsMutex);

	for (typename std::list<SenderCommand<T>*>::const_iterator c = m_currentCommands.begin(); c != m_currentCommands.end(); ++c) {

		std::ostringstream os;
		os << (*c)->getCommand();

		info.currentCommandNames.push_back(os.str());
	}

	return info;
}

template<class T>
CommandsInfo Scheduler<T>::getWaitingCommandsInfo() {
	CommandsInfo info;
	std::scoped_lock lock(m_commandsMutex);

	for (typename std::list<SenderCommand<T>*>::const_iterator c = m_waitingCommands.begin(); c != m_waitingCommands.end(); ++c) {
		info.waitingCommandNames.push_back((*c)->getCommand().getController()->getName() + " " + (*c)->getCommand().getName());
	}

	return info;
}

#endif
