/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef SENDER_H
#define SENDER_H

#include "Scheduler/TodoList.h"
#include "Scheduler/Condition.h"
#include "Scheduler/SenderCommand.h"
#include <common/base/Date.h>
#include <boost/thread/tss.hpp>
#include <memory>
#include <list>
#include <mutex>
#include <condition_variable>
#include <iostream>

template<class T> struct End;
template<class T> class Sender;

class SenderShared {

	// friend classes
	friend class Condition;
	template<class T> friend class Sender;

	thread_local static std::unique_ptr<std::mutex> m_mutex;
	thread_local static std::unique_ptr<std::condition_variable> m_ready;
};

/**
 * Template class parameterized by the Command type. Manages the synchronization
 * termination of the commands.
 */
template<class T>
class Sender {

	friend struct End<T> ;

public:
	/**
	 * Begin a group of commands which are synchronized.
	 */
	static void begin();

	/**
	 * Post a new command.
	 */
	static bool post(const T &aCommand);

	/**
	 * End of a group of commands. The method is blocking as int32 as the commands
	 * posted after the last begin are not terminated.
	 */
	static void endIf(const Condition& condition = End<T> (), int32 sec = -1);

	/**
	 * Pauses the execution.
	 */
	static void pause(bool state);

	/**
	 * To use after endif to check whether a time out has occurred
	 */
	static bool isTimedOut();

private:
	thread_local static std::unique_ptr<std::unique_lock<std::mutex>> m_lock;
	thread_local static std::unique_ptr<std::list<SenderCommand<T>*> > m_commands;
	thread_local static std::unique_ptr<bool> m_timedOut;

	static std::mutex m_postMutex;
	static bool m_pauseState;
};

/*******************************************************************************
 * Template implementation. This must be done in the .h file. Does not
 * compile in the .cpp file (template class constraint).
 */

template<class T>
struct End: public Condition {
	virtual bool operator()() {
		return ((*Sender<T>::m_commands.get()).size() == 0);
	}

};

template<class T>
thread_local std::unique_ptr<std::unique_lock<std::mutex>> Sender<T>::m_lock;

template<class T>
thread_local std::unique_ptr<std::list<SenderCommand<T>*> > Sender<T>::m_commands;

template<class T>
thread_local std::unique_ptr<bool> Sender<T>::m_timedOut;

template<class T>
std::mutex Sender<T>::m_postMutex;

template<class T>
bool Sender<T>::m_pauseState = false;

template<class T>
void Sender<T>::begin() {
	// Initializing if not done
	if (SenderShared::m_mutex.get() == nullptr) {

		std::mutex *p1 = new std::mutex();
		SenderShared::m_mutex.reset(p1);
		std::condition_variable *p2 = new std::condition_variable();
		SenderShared::m_ready.reset(p2);
		std::unique_lock<std::mutex> *p3 = new std::unique_lock<std::mutex>(*SenderShared::m_mutex.get());
		m_lock.reset(p3);

		std::list<SenderCommand<T>*> *p4 = new std::list<SenderCommand<T>*>();
		m_commands.reset(p4);
		bool* atimedOut = m_timedOut.release();
		if (atimedOut != nullptr) {
			delete atimedOut;
		}
		m_timedOut.reset(new bool(false));
	}
}

template<class T>
bool Sender<T>::post(const T &aCommand) {
	std::scoped_lock lock(m_postMutex);

	SenderCommand<T>* command = new SenderCommand<T> (aCommand, SenderShared::m_mutex.get(), SenderShared::m_ready.get(), m_commands.get());

	if (TodoList<T>::getInstance()->isPosted(command) == false) {
		// Incrementing N if synchronization is needed
		if (SenderShared::m_mutex.get() != nullptr) {
			(*m_commands.get()).push_back(command);
		}

		// Posting to TodoList
		TodoList<T>::getInstance()->post(command);

		return true;

	}

	delete command;

	return false;
}

template<class T>
void Sender<T>::endIf(const Condition& condition, int32 sec) {
	if (sec > 0) {

		while (!const_cast<Condition&> (condition)()) {
			do {
				if (SenderShared::m_ready.get()->wait_for(*m_lock.get(), std::chrono::seconds(sec)) == std::cv_status::timeout) {
					if (m_pauseState == false) {
						// timeout
						*m_timedOut.get() = true;
						break;
					}
				}
			} while (m_pauseState == true);

			if (*m_timedOut.get() == true) {
				break;
			}
		}
	} else {
		while (!const_cast<Condition&> (condition)()) {
			SenderShared::m_ready.get()->wait(*m_lock.get());
		}
	}

	const_cast<Condition&> (condition).setFinished();

	m_lock.get()->unlock();
	std::unique_lock<std::mutex>* alock = m_lock.release();
	delete alock;

	const_cast<Condition&> (condition).finalise();

	// advising remaining active commands that there is no synchronization any more
	for (typename std::list<SenderCommand<T>*>::iterator c = (*m_commands.get()).begin(); c != (*m_commands.get()).end(); ++c) {
		(*c)->reset();
	}

	// Releasing condition and counter.
	std::list<SenderCommand<T>*>* alist = m_commands.release();
	delete alist;

	if (SenderShared::m_mutex.get() != nullptr) {
		SenderShared::m_mutex.reset();
		SenderShared::m_ready.reset();
	}
}

template<class T>
void Sender<T>::pause(bool state) {
	m_pauseState = state;
}

template<class T>
bool Sender<T>::isTimedOut() {
	return *m_timedOut.get();
}

#endif
