/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef SENDERCOMMAND_H
#define SENDERCOMMAND_H

#include <common/base/BaseTypes.h>
#include <list>
#include <mutex>
#include <condition_variable>

/**
 * Template class parameterized by the Command type. Encapsulates the command
 * and information relative to its synchronization. It is implemented with pointers
 * to Sender mutex, condition and counter. In case of timeout, meaning that Sender
 * mutex is released, it is advised implying a lock system.
 */

template<class T>
class SenderCommand {

public:
	/**
	 * Constructor.
	 */
	SenderCommand(const T& aCommand, std::mutex* mutex, std::condition_variable* condition, std::list<SenderCommand<T>*>* activeCommands);

	/**
	 * Gets the associated command (const object).
	 */
	const T &getCommand() const;

	/**
	 * Gets the associated command (non const object).
	 */
	T &getCommand();

	/**
	 * Gets the associated command pointer (non const object).
	 */
	T *getPCommand();

	/**
	 * Gets the associated mutex.
	 */
	std::mutex* getMutex();

	/**
	 * Gets the list of active commands.
	 */
	std::list<SenderCommand<T>*>* getActiveCommands();

	/**
	 * Gets the associated condition.
	 */
	std::condition_variable* getCondition();

	/**
	 * Resets the sender command.
	 */
	void reset();

	/**
	 * Sets the index.
	 */
	void setIndex(int32 index);

	/**
	 * Gets the index.
	 */
	int32 getIndex() const;

private:
	T m_command;
	std::mutex* m_mutex;
	std::condition_variable* m_condition;
	std::list<SenderCommand<T>*>* m_activeCommands;

	int32 m_index;
};

/*******************************************************************************
 * Template implementation. This must be done in the .h file. Does not
 * compile in the .cpp file (template class constraint).
 */
template<class T>
SenderCommand<T>::SenderCommand(const T& aCommand, std::mutex* mutex, std::condition_variable* condition, std::list<SenderCommand<T>*>* activeCommands) :
	m_command(aCommand),
	m_mutex(mutex),
	m_condition(condition),
	m_activeCommands(activeCommands),
	m_index(0) {
}

template<class T>
inline const T &SenderCommand<T>::getCommand() const {
	return m_command;
}

template<class T>
inline T &SenderCommand<T>::getCommand() {
	return m_command;
}

template<class T>
inline T *SenderCommand<T>::getPCommand() {
	return &m_command;
}

template<class T>
inline std::mutex* SenderCommand<T>::getMutex() {
	return m_mutex;
}

template<class T>
inline std::list<SenderCommand<T>*>* SenderCommand<T>::getActiveCommands() {
	return m_activeCommands;
}

template<class T>
inline std::condition_variable* SenderCommand<T>::getCondition() {
	return m_condition;
}

template<class T>
void SenderCommand<T>::reset() {
	m_mutex = 0;
	m_condition = 0;
	m_activeCommands = 0;
}

template<class T>
void SenderCommand<T>::setIndex(int32 index) {
	m_index = index;
}

template<class T>
int32 SenderCommand<T>::getIndex() const {
	return m_index;
}

#endif
