/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef TODOLIST_H
#define TODOLIST_H

#include "Scheduler/SenderCommand.h"
#include <deque>
#include <map>
#include <mutex>
#include <condition_variable>
#include <typeinfo>

class AbstractTodoList {

protected:
	static std::map<std::string, AbstractTodoList*> s_instances;
};

class TodoMutex {

public:
	static std::mutex m_instanceMutex;
};


/**
 * Class managing concurrent access to the list of actions posted by
 * AbstractControllerProxyImpl(producer) and processed by Scheduler(consumer).
 */
template<class T>
class TodoList: public AbstractTodoList {

public:
	/**
	 * Destructor.
	 */
	~TodoList();

	/**
	 * Singleton method.
	 */
	static TodoList<T>* getInstance();

	/**
	 * Stops the list.
	 */
	void stop();

	/**
	 * Check if command on the pipe
	 */
	bool isPosted(SenderCommand<T>* aCommand);

	/**
	 * Add a new command.
	 */
	void post(SenderCommand<T>* aCommand);

	/**
	 * Add a new command at the top of the list.
	 */
	void postAtHead(SenderCommand<T>* aCommand);

	/**
	 * Consume the first element in the list.
	 */
	SenderCommand<T>* consume();

	/**
	 * Returns true if the list is empty.
	 */
	bool isEmpty();

private:
	/**
	 * Constructor.
	 */
	TodoList();

	/**
	 * list of commands waiting for process.
	 */
	std::deque<SenderCommand<T>*> m_todo;
	std::mutex m_todoMutex;
	std::condition_variable m_todoNotEmpty;
};

/*******************************************************************************
 * Template implementation. This must be done in the .h file. Does not
 * compile in the .cpp file (template class constraint).
 */

template<class T>
TodoList<T>* TodoList<T>::getInstance() {
	// Finding system name string
	std::scoped_lock lock(TodoMutex::m_instanceMutex);
	{
		std::map<std::string, AbstractTodoList*>::iterator l = AbstractTodoList::s_instances.find(typeid(T).name());

		if (l != AbstractTodoList::s_instances.end()) {
			return static_cast<TodoList<T>*> ((*l).second);

		} else {
			TodoList<T>* instance = new TodoList<T> ();
			AbstractTodoList::s_instances[typeid(T).name()] = instance;
			return instance;
		}
	}

}

template<class T>
void TodoList<T>::stop() {
	post(nullptr);
}

template<class T>
bool TodoList<T>::isPosted(SenderCommand<T>* aCommand) {

	std::unique_lock<std::mutex> lock(m_todoMutex);

	// Test if command already post , just for read or status
	if ((aCommand->getPCommand()->getName() == "read") || (aCommand->getPCommand()->getName() == "status")) {
		typename std::deque<SenderCommand<T>*>::iterator iter = m_todo.begin();
		while (iter != m_todo.end()) {
			if (aCommand->getPCommand()->equal((*iter)->getPCommand()) == true) {
				// Already in the list so don't add it.
				return true;
			}
			++iter;
		}
	}
	return false;
}

template<class T>
void TodoList<T>::post(SenderCommand<T>* aCommand) {
	std::unique_lock<std::mutex> lock(m_todoMutex);
	m_todo.push_back(aCommand);
	m_todoNotEmpty.notify_one();
}

template<class T>
void TodoList<T>::postAtHead(SenderCommand<T>* aCommand) {
	std::unique_lock<std::mutex> lock(m_todoMutex);

	// Add at the end of the queue the new DM and the command
	m_todo.push_back(aCommand);
	m_todoNotEmpty.notify_one();
}

template<class T>
SenderCommand<T>* TodoList<T>::consume() {
	std::unique_lock<std::mutex> lock(m_todoMutex);

	// waiting until the todoNotEmpty condition is fulfilled
	if (m_todo.empty()) {
		while (m_todo.empty()) {
			m_todoNotEmpty.wait(lock);
		}
	}
	// getting the first element
	SenderCommand<T>* todo = m_todo.front();
	m_todo.pop_front();

	return todo;
}

template<class T>
bool TodoList<T>::isEmpty() {
	std::unique_lock<std::mutex> lock(m_todoMutex);

	return m_todo.empty();
}

template<class T>
TodoList<T>::TodoList() {
}

template<class T>
TodoList<T>::~TodoList() {
	TodoList<T>* instance = AbstractTodoList::s_instances[typeid(T).name()];
	if (instance != NULL) {
		delete instance;
		AbstractTodoList::s_instances[typeid(T).name()] = NULL;
	}
}

#endif
