/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "MadPort.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <common/base/ServerProperties.h>

using namespace std;

PID2Parameters::PID2Parameters() :
	_Kp(0.0),
	_Ki(0.0),
	_Kd(0.0),
	_amortizement(0.0),
	_I(0.0) {

}

void PID2Parameters::setKp(float64 value) {
	_Kp = value;
}

void PID2Parameters::setKi(float64 value) {
	_Ki = value;
}

void PID2Parameters::setKd(float64 value) {
	_Kd = value;
}

void PID2Parameters::setAmortizement(float64 value) {
	_amortizement = value;
}

void PID2Parameters::setI(float64 value) {
	_I = value;
}

float64 PID2Parameters::getKp() const {
	return _Kp;
}

float64 PID2Parameters::getKi() const {
	return _Ki;
}

float64 PID2Parameters::getKd() const {
	return _Kd;
}

float64 PID2Parameters::getAmortizement() const {
	return _amortizement;
}

float64 PID2Parameters::getI() const {
	return _I;
}

DTI::DTI() :
	_wantedTemperature(0.0),
	_loopingTime(0),
	_useFixedPID2Parameters(true) {
}

void DTI::setWantedTemperature(float64 value) {
	_wantedTemperature = value;
}

void DTI::setLoopingTime(int32 value) {
	_loopingTime = value;
}

void DTI::setUseFixedPID2Parameters(bool value) {
	_useFixedPID2Parameters = value;
}

void DTI::setPID2File(const std::string & value) {

	if (value != _PID2FileName) {
		_PID2FileName = value;
		loadFile();
	}
}

float64 DTI::getWantedTemperature() const {
	return _wantedTemperature;
}

int32 DTI::getLoopingTime() const {
	return _loopingTime;
}

bool DTI::useFixedPID2Parameters() const {
	return _useFixedPID2Parameters;
}

const PID2Parameters & DTI::getPID2Parameters() const {
	return _PID2Parameters;
}

PID2Parameters & DTI::getPID2Parameters() {
	return _PID2Parameters;
}

const std::string & DTI::getPID2FileName() const {
	return _PID2FileName;
}

void DTI::getParametersFromPID2File() {

	// iterating the PID2 parameter entries
	for (vector<pair<float64, PID2Parameters> >::const_iterator p = _predefinedParameters.begin(); p != _predefinedParameters.end(); ++p) {
		if (_wantedTemperature <= p->first) {
			_PID2Parameters = p->second;

			cout << "Using PID2 parameters " << _PID2Parameters.getKp() << " " << _PID2Parameters.getKi() << " " << _PID2Parameters.getKd() << " " << _PID2Parameters.getAmortizement() << " " << _PID2Parameters.getI() << endl;

			return;
		}
	}

	cerr << "Cannot find PID2 parameters for temperature " << _wantedTemperature << endl;
}

void DTI::loadFile() {

	cout << "Loading PID2 file " << _PID2FileName << endl;

	_predefinedParameters.clear();

	if (_PID2FileName == "") {
		return;
	}

	string buffer;

	string fileName = common::ServerProperties::getInstance()->getConfigPath() + "/" + _PID2FileName;
	ifstream file(fileName.c_str());

	if (file.is_open() && !file.eof()) {

		while (getline(file, buffer)) {
			istringstream ls(buffer);

			float64 temperature;
			float64 Kp;
			float64 Ki;
			float64 Kd;
			float64 amortizement;
			float64 I;

			ls >> temperature >> Kp >> Ki >> Kd >> amortizement >> I;

			PID2Parameters parameters;

			parameters.setKp(Kp);
			parameters.setKi(Ki);
			parameters.setKd(Kd);
			parameters.setAmortizement(amortizement);
			parameters.setI(I);

			_predefinedParameters.push_back(pair<float64, PID2Parameters>(temperature, parameters));

			cout << "Added parameters " << temperature << " : " << Kp << " " << Ki << " " << Kd << " " << amortizement << " " << I << endl;
		}

		file.close();

	} else {
		cerr << "Cannot open file " << _PID2FileName << " for PID2 parameters" << endl;
	}

}
