/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef DTI_H
#define DTI_H

#include <map>
#include <string>
#include <vector>
#include <common/base/BaseTypes.h>

/**
 * Class owning the PID2 parameters.
 */
class PID2Parameters {

public:
	/**
	 * Constructor.
	 */
	PID2Parameters();

	/**
	 * Sets the Kp parameter.
	 */
	void setKp(float64 value);

	/**
	 * Sets the Ki parameter.
	 */
	void setKi(float64 value);

	/**
	 * Sets the Kd parameter.
	 */
	void setKd(float64 value);

	/**
	 * Sets the amortizement parameter.
	 */
	void setAmortizement(float64 value);

	/**
	 * Sets the I parameter.
	 */
	void setI(float64 value);

	/**
	 * Gets the Kp parameter.
	 */
	float64 getKp() const;

	/**
	 * Gets the Ki parameter.
	 */
	float64 getKi() const;

	/**
	 * Gets the Kd parameter.
	 */
	float64 getKd() const;

	/**
	 * Gets the amortizement parameter.
	 */
	float64 getAmortizement() const;

	/**
	 * Gets the I parameter.
	 */
	float64 getI() const;

private:
	float64 _Kp;
	float64 _Ki;
	float64 _Kd;
	float64 _amortizement;
	float64 _I;
};

/**
 * Structure defining the used members of the DTI Mad port.
 */
class DTI {

public:
	/**
	 * Constructor.
	 */
	DTI();

	/**
	 * Sets the wanted temperature.
	 */
	void setWantedTemperature(float64 value);

	/**
	 * Sets the looping time.
	 */
	void setLoopingTime(int32 value);

	/**
	 * Sets the value true if the fixed PID2 parameters are used, false otherwise.
	 */
	void setUseFixedPID2Parameters(bool value);

	/**
	 * Sets the PID2 file.
	 */
	void setPID2File(const std::string & value);

	/**
	 * Gets the wanted temperature.
	 */
	float64 getWantedTemperature() const;

	/**
	 * Gets the looping time.
	 */
	int32 getLoopingTime() const;

	/**
	 * Gets the value true if the fixed PID2 parameters are used, false otherwise.
	 */
	bool useFixedPID2Parameters() const;

	/**
	 * Gets the PID2 parameters.
	 */
	const PID2Parameters & getPID2Parameters() const;

	/**
	 * Gets the PID2 parameters.
	 */
	PID2Parameters & getPID2Parameters();

	/**
	 * Sets the PID2 filenNme.
	 */
	const std::string & getPID2FileName() const;

	/**
	 * Loads the PID2 parameters from the PID2 file.
	 */
	void getParametersFromPID2File();

private:
	void loadFile();

	float64 _wantedTemperature;
	int32 _loopingTime;
	bool _useFixedPID2Parameters;
	PID2Parameters _PID2Parameters;
	std::string _PID2FileName;
	// Pair of (temperature, parameters)
	std::vector<std::pair<float64, PID2Parameters> > _predefinedParameters;

	static const int32 LINE_SIZE = 2048;
};

#endif
