/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "PID2Controller.h"
#include <iostream>
#include <sstream>
#include <cmath>
#include <boost/format.hpp>

using namespace std;
using namespace boost;

float64 sign(float64 x, float64 y) {
	if (y >= 0) {
		return abs(x);
	}

	return -abs(x);
}

const float64 PID2Controller::TAU2I = 20.0;
const float64 PID2Controller::ERROR_MAX = 0.1;

PID2Controller::PID2Controller() :
	_sampleTemperature(0.0),
	_regulationTemperature(0.0),
	_setpointTemperature(0.0),
	_loopingTime(0),
	_usePID2(true),
	_P(0.0),
	_I(0.0),
	_Imax(0.0),
	_D(0.0),
	_Dmax(0.0),
	_Dfiltered1(0.0),
	_Dfiltered2(0.0),
	_calculatedSetpointTemperature(0.0),
	_setNewTemperatureSetpoint(false),
	_initialisation(true),
	_iterationCounter(0),
	_initialError(0.0),
	_tau1(0.0),
	_tau2(0.0),
	_tolerance(0.0) {
}

const DTI & PID2Controller::getDTI() const {
	return _dti;
}

void PID2Controller::setSampleTemperature(float64 value) {
	_sampleTemperature = value;

	// update temperature values
	if ((int32) _sampleTemperatureValues.size() == STABILITY_WINDOW) {
		_sampleTemperatureValues.pop_front();
	}

	_sampleTemperatureValues.push_back(value);
}

void PID2Controller::setRegulationTemperature(float64 value) {
	_regulationTemperature = value;
}

void PID2Controller::setSetpointTemperature(float64 value) {
	_setpointTemperature = value;
}

void PID2Controller::setI(float64 value) {
	_I = value;
	//cout << "I sets " << value << endl;
}

void PID2Controller::setTolerance(float64 value) {
	_tolerance = value;
}

DTI & PID2Controller::getDTI() {
	return _dti;
}

void PID2Controller::init(float64 sampleTemperature) {

	_sampleTemperature = sampleTemperature;
	_usePID2 = false;

	// check sample temperature readable
	if (_sampleTemperature > 9000.0) {
		cout << "Sample thermometer not readable" << endl;
		cout << "Cannot initialise external pid algorithm" << endl;
	}

	string text;

	// manual - fixed parameters -
	if (_dti.useFixedPID2Parameters()) {
		_usePID2 = true;
		text="MANUAL";

	} else {
		// get from a file
		_usePID2 = true;
		text = "From FILE";
		_dti.getParametersFromPID2File();
	}

	_PID2Parameters.setKp(_dti.getPID2Parameters().getKp());
	_PID2Parameters.setKi(_dti.getPID2Parameters().getKi());
	_PID2Parameters.setKd(_dti.getPID2Parameters().getKd());
	_PID2Parameters.setAmortizement(_dti.getPID2Parameters().getAmortizement());
	_I = _dti.getPID2Parameters().getI();

	if (_usePID2) {
		cout << "External PID2 algorithm invoked : ";
		cout << "PID parameters (" << text << ") " << _PID2Parameters.getKp() << " " << _PID2Parameters.getKi() << " " << _PID2Parameters.getKd();
		cout << "  Amortizement : " << _PID2Parameters.getAmortizement() << endl;
		cout << "  Initial I : " << _I << endl;

		if (_dti.getLoopingTime() == 0) {
			_dti.setLoopingTime(10);
		}

		cout << "Sampling time: " << _dti.getLoopingTime() << " secs" << endl;
		_loopingTime = _dti.getLoopingTime();

		// set up pid parameters in common pid
		if (_PID2Parameters.getKp() != 0.0) {
			_tau1 = _PID2Parameters.getKd() / _PID2Parameters.getKp();
		}

		_tau2 = TAU2I;

		if (_tau2 < _tau1) {
			_tau2 = _tau1;
		}

		// initialise the derives
		_P = 0.0;
		// don't touch the integral term in changes (_I)
		_D = 0.0;

		_Dfiltered1 = _sampleTemperature;
		_Dfiltered2 = _sampleTemperature;
		_initialError = _dti.getWantedTemperature() - _sampleTemperature;

		// absolute max permitted correction is 15% with min of 10 deg
		// changed from 15% to 20%
		_Imax = _dti.getWantedTemperature() * 0.20;

		if (_Imax < 10.0) {
			_Imax = 10.0;
		}


		//cout << "init PID2" << endl;
		//cout << "st = " << _sampleTemperature << endl;
		//cout << "rt = " << _regulationTemperature << endl;
		//cout << "sp = " << _setpointTemperature << endl;
		//cout << "lt = " << _loopingTime << endl;
		//cout << "P  = " << _P << endl;
		//cout << "I  = " << _I << endl;
		//cout << "Im = " << _Imax << endl;
		//cout << "D  = " << _D << endl;
		//cout << "Dm = " << _Dmax << endl;
		//cout << "D1 = " << _Dfiltered1 << endl;
		//cout << "D2 = " << _Dfiltered2 << endl;
		//cout << "cs = " << _calculatedSetpointTemperature << endl;
		//cout << "ie = " << _initialError << endl;
		//cout << "t1 = " << _tau1 << endl;
		//cout << "t2 = " << _tau2 << endl << endl;


		// Do not take a maximum Imax?
		// take out 20 degree ceiling
		/*		if (_Imax > 20.0) {
			_Imax = 20.0;
		}
		*/
		//cout << "Imax = " << _Imax << endl;

		// initialises counter used to ensure a certain time to initialise
		_initialisation = true;
		_iterationCounter = 0;
	}

}

void PID2Controller::readParametersFromFile(const std::string & filename) {


}

void PID2Controller::calculate() {

	// check sample temperature still readable
	_setNewTemperatureSetpoint = false;
	if (_sampleTemperature > 9000.0) {
		cout << "Error reading sample temperature - inhibit pid2";
		cout << "  Reset set point to " << _dti.getWantedTemperature() << endl;

		_usePID2 = false;
		// re-choose sampling time as a function of monitoring state
		// simplified looping time compared to Mad
		_loopingTime = 10; // TEMP?

		_calculatedSetpointTemperature = _dti.getWantedTemperature();
		_setNewTemperatureSetpoint = true;

		_log = "error reading sample temperature";

	} else {
		// calculate filtered derivative
		float64 derivative1 = (_sampleTemperature - _Dfiltered1) / (_dti.getLoopingTime() + _tau1);
		_Dfiltered1 = _Dfiltered1 + derivative1 * _dti.getLoopingTime();
		float64 derivative2 = (_sampleTemperature - _Dfiltered2) / (_dti.getLoopingTime() + _tau2);
		_Dfiltered2 = _Dfiltered2 + derivative2 * _dti.getLoopingTime();


		//cout << "tau1, loopingTime, Dfiltered1, derivative1 " << _tau1 << " " << _dti.getLoopingTime() << " " << _Dfiltered1 << " " << derivative1 << endl;

		//	calculate derivative term (dfd)
		_D = derivative1 * _PID2Parameters.getKd() * -1.0;

		//cout << "D " << _D << endl;

		// calculate dmax = max(abs(der2))
		_Dmax = max(abs(derivative2), _Dmax);

		// calculate relative and absolute deviations
		float64 error = _dti.getWantedTemperature() - _sampleTemperature;

		float64 absoluteError = error / _initialError; // no verification on initialError ?
		float64 relativeError = abs(_setpointTemperature - _regulationTemperature);

		// calculate proportional deviation (dfp)
		_P = _PID2Parameters.getKp() * error;

		// calculate change in integral term (ddfi)
		float64 deltaI = 0.0;

		// check we've initialised far enough
		if (_initialisation) {
			_iterationCounter++;
			if (_iterationCounter > 10) {
				_initialisation = false;
			}

			//cout << "initialisation" << endl;
		}

		float64 derivativeMax = 0.0;
		if (!_initialisation) {
			if (_PID2Parameters.getAmortizement() == 0.0) {
				derivativeMax = 0.0;
			} else {
				derivativeMax = _Dmax / _PID2Parameters.getAmortizement();
			}

			//cout << "first I test absolute error, derivative2, derivativeMax : " << absoluteError << " " << derivative2 << " " << derivativeMax << endl;

			if (absoluteError < 0.0 || abs(derivative2) < derivativeMax) {
				// play with I if regulation and set point are within 1
				// warning : West controller returns integer values
				if (relativeError < 2.0) {
					if (abs(error) > ERROR_MAX) {
						error = sign(ERROR_MAX, error);
					}
					deltaI = _PID2Parameters.getKi() * 0.01 * error * _dti.getLoopingTime();
					_I = _I + deltaI;
					//cout << "new I term : " << deltaI << " " << _I << endl;
				//} else {
				//	cout << "relative error too large, no modification to I" << endl;
				}
			}
		}

		if (abs(_I) > _Imax) {
			_I = sign(_Imax, _I);
			//cout << "integral plafonded, I recalculated " << _I << endl;
		}

		// calculate correction term (df)
		float64 MV = _P + _I + _D;

		if (abs(MV) > _Imax) {
			MV = sign(_Imax, MV);
			//cout << "MV plafonded" << endl;
		}

		//cout << "P, I, D : " << _P << " " << _I << " " << _D << " correction : " << MV << endl;

		_calculatedSetpointTemperature = _dti.getWantedTemperature() + MV;


		//cout << "calculated setpoint = " << _calculatedSetpointTemperature << endl;
		//cout << "actual setpoint = " << _setpointTemperature << endl;
		//cout << "setpoint diff = " << (_calculatedSetpointTemperature - _setpointTemperature) << endl;

		// really correct?
		if (abs(_calculatedSetpointTemperature - _setpointTemperature) > 0.001) {
			//cout << "correct by : " << MV << endl << endl;
			_setNewTemperatureSetpoint = true;
			//cout << "new setpoint" << endl;
		}


		_loopingTime = _dti.getLoopingTime();

		if (_setNewTemperatureSetpoint) {
			// iret = dtctrl_sendt (ddy.set) ? consequence : setpoint + ddy.const == ddy.set?
		}

		ostringstream os;

		os << "st/rt/sp/lt/P/I/Im/D/Dm/D1/D2/cs/ie/t1/t2/d1/d2/e/ae/re";

		os << format("%10.3f ") % _sampleTemperature
				<< format("%10.3f ") % _regulationTemperature				// const
				<< format("%10.3f ") % _setpointTemperature					// consp
				<< format("%10.3f ") % _loopingTime							// loop
				<< format("%10.3f ") % _P									// dfp
				<< format("%10.3f ") % _I									// dfi
				<< format("%10.3f ") % _Imax								// dfmax
				<< format("%10.3f ") % _D									// dfd
				<< format("%10.3f ") % _Dmax								// dmax
				<< format("%10.3f ") % _Dfiltered1							// tf1
				<< format("%10.3f ") % _Dfiltered2							// tf2
				<< format("%10.3f ") % _calculatedSetpointTemperature		// set
				<< format("%10.3f ") % _initialError						// saut
				<< format("%10.3f ") % _tau1								// tau1
				<< format("%10.3f ") % _tau2								// tau2
				<< format("%10.3f ") % derivative1							// der1
				<< format("%10.3f ") % derivative2							// der2
				<< format("%10.3f ") % error								// ec
				<< format("%10.3f ") % absoluteError						// eca
				<< format("%10.3f ") % relativeError;						// ecr




//		os << "sp/rt/st/dp/di/dd/csp";
//		os << format("%10.3f ") % _setpointTemperature
//			<< format("%10.3f ") % _regulationTemperature
//			<< format("%10.3f ") % _sampleTemperature
//			<< format("%10.3f ") % _P
//			<< format("%10.3f ") % _I
//			<< format("%10.3f ") % _D;

		_log = os.str();
	}
}

int32 PID2Controller::getLoopingTime() const {
	return _loopingTime;
}

bool PID2Controller::isNewTemperatureSetpoint() const {
	return _setNewTemperatureSetpoint;
}

float64 PID2Controller::getCalculatedSetpointTemperature() const {
	return _calculatedSetpointTemperature;
}

float64 PID2Controller::getP() const {
	return _P;
}

float64 PID2Controller::getI() const {
	return _I;
}

float64 PID2Controller::getD() const {
	return _D;
}

std::string PID2Controller::getLog() const {
	return _log;
}

bool PID2Controller::isStable() const {

	if ((int32) _sampleTemperatureValues.size() == STABILITY_WINDOW) {

		for (int32 i = 0; i < STABILITY_WINDOW; i++) {
			if (abs(_dti.getWantedTemperature() - _sampleTemperatureValues[i]) > _tolerance) {
				return false;
			}
		}

		return true;
	}

	return false;
}
