/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef PID2CONTROLLER_H
#define PID2CONTROLLER_H

#include "MadPort.h"
#include <deque>

/**
 */
class PID2Controller {

public:
	/**
	 * Constructor.
	 */
	PID2Controller();

	/**
	 * Sets the sample temperature.
	 */
	void setSampleTemperature(float64 value);

	/**
	 * Sets the regulation temperature.
	 */
	void setRegulationTemperature(float64 value);

	/**
	 * Sets the setpoint temperature.
	 */
	void setSetpointTemperature(float64 value);

	/**
	 * Sets the integral term.
	 */
	void setI(float64 value);

	/**
	 * Sets the tolerance.
	 */
	void setTolerance(float64 value);

	/**
	 * Gets the DTI structure.
	 */
	const DTI & getDTI() const;

	/**
	 * Gets the DTI structure.
	 */
	DTI & getDTI();

	/**
	 * Initialises the PID2 calculator.
	 */
	void init(float64 sampleTemperature);

	/**
	 * Calculates a new PID2 iteration.
	 */
	void calculate();

	/**
	 * Gets the looping time.
	 */
	int32 getLoopingTime() const;

	/**
	 * Returns true if a new temperature set point has to be set.
	 */
	bool isNewTemperatureSetpoint() const;

	/**
	 * Gets the temperature setpoint.
	 */
	float64 getCalculatedSetpointTemperature() const;

	/**
	 * Gets the P value.
	 */
	float64 getP() const;

	/**
	 * Gets the I value.
	 */
	float64 getI() const;

	/**
	 * Gets the I value.
	 */
	float64 getD() const;

	/**
	 * Returns true if temperature is stable.
	 */
	bool isStable() const;

	/**
	 * Gets the log information.
	 */
	std::string getLog() const;

private:
	void readParametersFromFile(const std::string & filename);

	DTI _dti;

	float64 _sampleTemperature;
	float64 _regulationTemperature;
	float64 _setpointTemperature;
	int32 _loopingTime;
	bool _usePID2;
	PID2Parameters _PID2Parameters;

	// calculated PID2 components
	float64 _P;
	float64 _I;
	float64 _Imax;
	float64 _D;
	float64 _Dmax;
	float64 _Dfiltered1;
	float64 _Dfiltered2;

	float64 _calculatedSetpointTemperature;
	bool _setNewTemperatureSetpoint;
	bool _initialisation;
	int32 _iterationCounter;
	float64 _initialError;
	float64 _tau1;
	float64 _tau2;

	std::string _log;

	float64 _tolerance;

	static const float64 TAU2I;
	static const float64 ERROR_MAX;
	static const int32 STABILITY_WINDOW = 10;

	std::deque<float64> _sampleTemperatureValues;
};

#endif