/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief This class propose a unit tests facilities
 * \author J. Locatelli
 * \date 08-02-2017
 */

#ifndef UNITTEST_H
#define UNITTEST_H

#include <common/base/BaseTypes.h>
#include <fstream>
#include <cstdlib>
#include <execinfo.h>
#include "InstrumentAbstraction/ExperimentController.h"

namespace utilities {

/*!
 * \class LangmuirDriver
 * \brief Main class for the Langmuir device driver
 *
 * \par
 * Class for Langmuir driver. Implements DeviceDriver methods.
 * \par
 * The Langmuir device is a windows computer which running the Nima software made with LabView.
 * It is connected to the nomad server using RS232 protocol.
 */
class UnitTest {

public:

	/*!
	 * \Constructor
	 */
	UnitTest(ExperimentController* cont);

	/*!
	 * \Destructor
	 */
	~UnitTest();

	bool checkTrueImpl(const std::string& caller, const std::string& file, int32 line, bool val);
	bool checkNotTrueImpl(std::string caller, std::string file, int32 line, bool val);
	bool checkInt32EqualImpl(std::string caller, std::string file, int32 line, int32 val1, int32 val2);
	bool checkUint32EqualImpl(std::string caller, std::string file, int32 line, int32 val1, int32 val2);
	bool checkFloat64EqualImpl(std::string caller, std::string file, int32 line, float64 val1, float64 val2, float64 eps);
	bool checkInt32NotEqualImpl(std::string caller, std::string file, int32 line, int32 val1, int32 val2);
	bool checkUint32NotEqualImpl(std::string caller, std::string file, int32 line, int32 val1, int32 val2);
	bool checkFloat64NotEqualImpl(std::string caller, std::string file, int32 line, float64 val1, float64 val2, float64 eps);
	bool checkStringEqualImpl(std::string caller, std::string file, int32 line, const std::string& val1,
			const std::string& val2);
	bool checkStringNotEqualImpl(std::string caller, std::string file, int32 line, const std::string& val1,
			const std::string& val2);

	template<typename Type> bool checkPropertyCannotSetValueImpl(std::string caller, std::string file, int32 line,
			SimpleProperty<Type>& prop, Type val, bool doexception);

	template<typename Object, typename Function, typename Param1,
			typename Arg1> bool checkThrowImpl1(std::string caller,
			std::string file, int32 line, bool doexception, void (Function::*f)(Param1) 
			, Function* obj, Arg1& arg1);
	template<typename Object, typename Function, typename Param1, typename Param2,
			typename Arg1, typename Arg2> bool checkThrowImpl2(std::string caller,
			std::string file, int32 line, bool doexception, void (Function::*f)(Param1, Param2) 
			, Function* obj, Arg1& arg1, Arg2& arg2);
	template<typename Object, typename Function, typename Param1, typename Param2, typename Param3,
			typename Arg1, typename Arg2, typename Arg3> bool checkThrowImpl3(std::string caller,
			std::string file, int32 line, bool doexception, void (Function::*f)(Param1, Param2, Param3) 
			, Function* obj, Arg1& arg1, Arg2& arg2, Arg3& arg3);
	template<typename Object, typename Function, typename Param1, typename Param2, typename Param3, typename Param4,
			typename Arg1, typename Arg2, typename Arg3, typename Arg4> bool checkThrowImpl4(std::string caller,
			std::string file, int32 line, bool doexception, void (Function::*f)(Param1, Param2, Param3, Param4) 
			, Function* obj, Arg1& arg1, Arg2& arg2, Arg3& arg3, Arg4& arg4);

	static const float64 EPS;
private:

	void logFaillure(std::string caller, std::string file, int32 line, const std::string& message);

	std::ofstream m_file;
	ExperimentController* m_controller;
};

}

#include "UnitTest.inl"

#endif
