/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

/*!
 * \brief This class propose a unit tests facilities
 * \author J. Locatelli
 * \date 08-02-2017
 */

#include <common/base/BaseTypes.h>
#include <fstream>
#include <cstdlib>
#include <execinfo.h>

namespace utilities {

/*
 * checkPropertyCannotSetValue
 */
template<typename Type> bool UnitTest::checkPropertyCannotSetValueImpl(std::string caller, std::string file, int32 line,
		SimpleProperty<Type>& prop, Type val, bool doexception) {
	bool exception = false;
	try {
		prop = val;
	} catch (CannotSetValue& e) {
		exception = true;
	}
	if (doexception == true) {
		if (exception == false) {
			std::ostringstream buf;
			buf << "property " << prop.getName() << " not doing cannot set value " << val << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return exception;
	} else {
		if (exception == true) {
			std::ostringstream buf;
			buf << "property " << prop.getName() << " doing cannot set value " << val << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return !exception;
	}
}

template<typename Object, typename Function, typename Param1, typename Arg1>
bool UnitTest::checkThrowImpl1(std::string caller, std::string file, int32 line, bool doexception,
		void (Function::*f)(Param1) , Function* obj, Arg1& arg1) {

	bool exception = false;
	try {
		(obj->*f)(arg1);
	} catch (Object&) {
		exception = true;
	}
	if (doexception == true) {
		if (exception == false) {
			std::ostringstream buf;
			buf << typeid(f).name() << " not doing " << typeid(Object).name() << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return exception;
	} else {
		if (exception == true) {
			std::ostringstream buf;
			buf << typeid(f).name() << " doing " << typeid(Object).name() << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return !exception;
	}
}

template<typename Object, typename Function, typename Param1, typename Param2, typename Arg1,
		typename Arg2>
bool UnitTest::checkThrowImpl2(std::string caller, std::string file, int32 line, bool doexception,
		void (Function::*f)(Param1, Param2) , Function* obj, Arg1& arg1, Arg2& arg2) {

	bool exception = false;
	try {
		(obj->*f)(arg1, arg2);
	} catch (Object&) {
		exception = true;
	}
	if (doexception == true) {
		if (exception == false) {
			std::ostringstream buf;
			buf << typeid(f).name() << " not doing " << typeid(Object).name() << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return exception;
	} else {
		if (exception == true) {
			std::ostringstream buf;
			buf << typeid(f).name() << " doing " << typeid(Object).name() << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return !exception;
	}
}

template<typename Object, typename Function, typename Param1, typename Param2, typename Param3, typename Arg1,
		typename Arg2, typename Arg3>
bool UnitTest::checkThrowImpl3(std::string caller, std::string file, int32 line, bool doexception,
		void (Function::*f)(Param1, Param2, Param3) , Function* obj, Arg1& arg1, Arg2& arg2,
		Arg3& arg3) {

	bool exception = false;
	try {
		(obj->*f)(arg1, arg2, arg3);
	} catch (Object&) {
		exception = true;
	}
	if (doexception == true) {
		if (exception == false) {
			std::ostringstream buf;
			buf << typeid(f).name() << " not doing " << typeid(Object).name() << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return exception;
	} else {
		if (exception == true) {
			std::ostringstream buf;
			buf << typeid(f).name() << " doing " << typeid(Object).name() << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return !exception;
	}
}

template<typename Object, typename Function, typename Param1, typename Param2, typename Param3, typename Param4, typename Arg1,
		typename Arg2, typename Arg3, typename Arg4>
bool UnitTest::checkThrowImpl4(std::string caller, std::string file, int32 line, bool doexception,
		void (Function::*f)(Param1, Param2, Param3, Param4) , Function* obj, Arg1& arg1, Arg2& arg2,
		Arg3& arg3, Arg4& arg4) {

	bool exception = false;
	try {
		(obj->*f)(arg1, arg2, arg3, arg4);
	} catch (Object&) {
		exception = true;
	}
	if (doexception == true) {
		if (exception == false) {
			std::ostringstream buf;
			buf << typeid(f).name() << " not doing " << typeid(Object).name() << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return exception;
	} else {
		if (exception == true) {
			std::ostringstream buf;
			buf << typeid(f).name() << " doing " << typeid(Object).name() << std::endl;
			logFaillure(caller, file, line, buf.str());
		}
		return !exception;
	}
}

#define checkTrue(x) checkTrueImpl(__FUNCTION__,__FILE__,__LINE__,x)
#define checkNotTrue(x) checkNotTrueImpl(__FUNCTION__,__FILE__,__LINE__,x)
#define checkInt32Equal(x,y) checkInt32EqualImpl(__FUNCTION__,__FILE__,__LINE__,x,y)
#define checkUint32Equal(x,y) checkUint32EqualImpl(__FUNCTION__,__FILE__,__LINE__,x,y)
#define checkFloat64Equal(x,y,z) checkFloat64EqualImpl(__FUNCTION__,__FILE__,__LINE__,x,y,z)
#define checkInt32NotEqual(x,y) checkInt32NotEqualImpl(__FUNCTION__,__FILE__,__LINE__,x,y)
#define checkUint32NotEqual(x,y) checkUint32NotEqualImpl(__FUNCTION__,__FILE__,__LINE__,x,y)
#define checkFloat64NotEqual(x,y,z) checkFloat64NotEqualImpl(__FUNCTION__,__FILE__,__LINE__,x,y,z)
#define checkStringEqual(x,y) checkStringEqualImpl(__FUNCTION__,__FILE__,__LINE__,x,y)
#define checkStringNotEqual(x,y) checkStringNotEqualImpl(__FUNCTION__,__FILE__,__LINE__,x,y)
#define checkPropertyCannotSetValue(x,y,z) checkPropertyCannotSetValueImpl(__FUNCTION__,__FILE__,__LINE__,x,y,z)
#define checkThrow1(ErrorType,x,y,z,a1) checkThrowImpl1<ErrorType>(__FUNCTION__,__FILE__,__LINE__,x,y,z,a1)
#define checkThrow2(ErrorType,x,y,z,a1,a2) checkThrowImpl2<ErrorType>(__FUNCTION__,__FILE__,__LINE__,x,y,z,a1,a2)
#define checkThrow3(ErrorType,x,y,z,a1,a2,a3) checkThrowImpl3<ErrorType>(__FUNCTION__,__FILE__,__LINE__,x,y,z,a1,a2,a3)
#define checkThrow4(ErrorType,x,y,z,a1,a2,a3,a4) checkThrowImpl4<ErrorType>(__FUNCTION__,__FILE__,__LINE__,x,y,z,a1,a2,a3,a4)

}
