/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#define BOOST_AUTO_TEST_MAIN
#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <common/base/Convert.h>
#include <common/base/Date.h>
#include <common/base/Equals.h>
#include <common/base/Compare.h>
#include <common/base/StringUtilities.h>
#include <common/base/TypeToString.h>
#include <common/base/RealArray.h>
#include <limits>
#include <iostream>
#include <fstream>
#include <boost/format.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <algorithm>
#include <boost/locale.hpp>

using namespace common;
using namespace boost;
using namespace std;

/**
 * Class template to control at compile-time the size of strings.
 */
template<int Min, int Max, int Current>
struct StringLength {
	static const std::string& value(const std::string& s) {
		return StringLength<Min, Max, Current - 1>::value(s);
	}
};

template<int Min, int Max>
struct StringLength<Min, Max, Max>;

template<int Min, int Max>
struct StringLength<Min, Max, -1>;

template<int Min, int Max>
struct StringLength<Min, Max, Min> {
	static const std::string& value(const std::string& s) {
		return s;
	}
};

template<int Min, int Max, int Length>
std::string checkString(const char(&p)[Length]) {
	return StringLength<Min, Max + 1, Length - 1>::value(p);
}

template<typename T>
string convertFloat(T value) {

	if (value > 1E6 || value < 1E-3) {
		return (boost::format("%E") % value).str();
	}

	return (boost::format("%10.3f") % value).str();
}



BOOST_AUTO_TEST_SUITE(suite);

void compareStrings(std::string const & a, std::string const & b) {

	if (a.size() != b.size()) {
		cout << "sizes are " << a.size() << ", " << b.size() << endl;
	}

	for (int i = 0; i < a.length(); ++i) {
		if (a[i] != b[i]) {
			cout << "difference with char " << i << endl;
			return;
		}
	}

	cout << "comparison terminated" << endl;
}

void printString(std::string const & s) {

	for (int i = 0; i < s.length(); ++i) {
		cout << "char " << static_cast<int>(s[i]) << endl;
	}
}

BOOST_AUTO_TEST_CASE(testConvert1)
{
	cout << "testing checkString" << endl;
	cout << checkString<1, 8>("12345678") << endl;

	cout << "testing string utf-8" << endl;

	char data[256 - 32];
	for (int i = 32; i < 256; ++i) {
		data[i - 32] = static_cast<char>(i);
	}

	std::string sas(data);


	cout << "testing utf8s" << endl;

	std::string utf8s = boost::locale::conv::to_utf<char>(sas, "Latin1");

	string tests = "hello, this is a test \r\n\b possible?";

	utf8s = boost::locale::conv::to_utf<char>(tests, "Latin1");
	cout << "utf8s = " << utf8s << endl;

	char cc[1];
	cc[0] = 255;
	string ccs(cc);

	utf8s = boost::locale::conv::to_utf<char>(ccs, "Latin1");

	compareStrings(tests, utf8s);

	printString(utf8s);

	double dd = 12323.5678;

	stringstream ddValue1;
	ddValue1 << boost::format("%10.2d") % dd;
	cout << "dd %10.2d  = |" << ddValue1.str() << "|" << endl;

	stringstream ddValue2;
	ddValue2 << boost::format("%10.2f") % dd;
	cout << "dd %10.2f  = |" << ddValue2.str() << "|" << endl;

	stringstream ddsValue1;
	string dds = "12323.56788989";
	ddsValue1 << boost::format("%10.2d") % dds;
	cout << "dds %10.2d = |" << ddsValue1.str() << "|" << endl;

	stringstream ddsValue2;
	dds = "12323.567888890";
	ddsValue2 << boost::format("%10s") % dds;
	cout << "dds %10s   = |" << ddsValue2.str() << "|" << endl;

	stringstream ddsValue3;
	dds = "12323";
	ddsValue3 << boost::format("%10s") % dds;
	cout << "dds %10s   = |" << ddsValue3.str() << "|" << endl;


	string testString = "Content-LEngth    :    232389  ";
	string::size_type colonPosition = testString.find_first_of(':');
	string propertyField = testString.substr(0, colonPosition);
	string valueField = testString.substr(colonPosition + 1, testString.size() - colonPosition - 1);
	boost::trim(propertyField);
	boost::trim(valueField);

	boost::to_lower(propertyField);

	cout << "property:" << propertyField << endl;
	cout << "value:" << valueField << endl;


	int a1 = 4;
	int a2 = 2;
	int a3 = 19;
	int a4 = -1;

	int all[3];
	all[0] = a1;
	all[1] = a2;
	all[2] = a3;

	int alb[3] = {a1, a2, a3};

	sort(alb, alb + 3);

	for (int i = 0; i < 3; i++) {
		cout << "alb " << alb[i] << endl;
	}


	float64 fv = 12.9;

	cout << "converting " << fv << " to " << static_cast<int32>(fv + 0.5) << endl;

	fv = 13.001;

	cout << "converting " << fv << " to " << static_cast<int32>(fv + 0.5) << endl;

	StringHelper::setMaxNumberOfPrintedArrayElements(5);

	cout << "A1 = " << to<RealArray<int32> >(string("[1, 2, -5, -6]")) << endl;
	cout << "A2 = " << to<RealArray<int32> >(string("[1]")) << endl;
	cout << "A3 = " << to<RealArray<int32> >(string("[]")) << endl;
	cout << "A4 = " << to<RealArray<int32> >(string("[  ]")) << endl;

	cout << to<float>(2) << endl;
	cout << to<int>(2.1f) << endl;
	cout << to<string>(12) << endl;
	cout << to<string>(-123.35E12) << endl;

	cout << to<double>(string("-123.35E12")) << endl;
	cout << to<int>(string("-12312")) << endl;

	string values = "0, 13, 32136";
	vector<long> floatValues = to<vector<long> >(StringUtilities::split(values, ",; "));

	for (int i = 0; i < floatValues.size(); i++) {
		cout << floatValues[i] << endl;
	}

	cout << to<string>("toto") << endl;

	string s = "tata";
	cout << to<string>(s) << endl;

	double dv = 1.;

	std::ostringstream os;
	os << dv;

	cout << "1 = " << std::to_string(dv) << " " << os.str() << endl;

	dv = 12.3429;

	os.clear();
	os << dv;

	cout << "12.3429 = " << std::to_string(dv) << " " << os.str() << endl;



/*
	int8 a = to<int8>(string("0"));
	long al = to<int8>(string("0"));

	cout << "a = " << (int)a << endl;
	cout << "a = " << to<string>(a) << endl;
	cout << "al = " << al << endl;

	cout << "infinity" << endl;

	cout << numeric_limits<double>::infinity() << endl;

	double dinf = numeric_limits<double>::infinity();
	cout << dinf << endl;

	int iinf = numeric_limits<int>::infinity();
	cout << iinf << endl;

	int i = dinf;
	cout << i << endl;

	double d = 1 + (dinf + 2)*3 + 4;
	cout << d << endl;

	if (!finite(d)) {
		cout << "inf number" << endl;
	}

	int j = 0;

	cout << "compare  = " << Compare::greaterThan(j, (-1)) << endl;

	string value = "titato";

	int8 c = value.operator[](2);

	cout << "c = " << c << endl;

	value.replace(3, 1, string(1, 'u'));

	cout << "value = " << value << endl;

	cout << "0xA = " << to<int32>(string("0xA")) << endl;

	cout << "array = " << to<string>(to<RealArray<int32> >(string("[1, 2, 3, 4, 5, 6, 7, 8]"))) << endl;

	int32 v = 132;

	cout << ">>" << to<string>(v, "%-8i") << "<<" << endl;
	cout << ">>" << to<string>(v, "%x") << "<<" << endl;

	cout << "array = " << to<string>(to<RealArray<int32> >(string("[1, 2, 3, 4, 5, 6, 7, 8]")), "%-4i") << endl;

	cout << to<int32>(string("fc"), string("%x")) << endl;
	cout << to<int32>(string("0xfc")) << endl;
	cout << to<int32>(string("0xfc"), string("%x")) << endl;

	cout << to<int32>(string("2147483647")) << endl;

	cout << "array = " << to<string>(to<RealArray<int32> >(string("[0xff , 0xaa , a8]"), "%x"), "%x") << endl;

	cout << "a8 = " << to<int32>(string("a8"), "%x") << endl;

	RealArray<string> sa = to<RealArray<string> >(string("[toto,, sdw,]"));

	cout << "sa[0] = " << sa[0] << endl;

	cout << "array = " << to<string>(sa) << endl;

	int32 v0 = -1717986918;
	int32 v1 = 1069128089;

	int64 v0a = 0x00000000ffffffff & v0;
	int64 v1a = 0x00000000ffffffff & v1;

	cout << "v0a = " << v0a << endl;
	cout << hex << "v0a = " << v0a << endl;
	cout << hex << "v1a = " << v1a << endl;

	cout << hex << "v0 = " << v0 << endl;
	cout << hex << "v1 = " << v1 << endl;

	int64 v2 = v1;
	cout << hex << "v2 = " << v2 << endl;

	int64 v0b = v0;
	cout << hex << "v0b = " << v0b << endl;

	cout << hex << "v2 = " << v2 << endl;
	v2 = v2 << 32;
	cout << hex << "v2 = " << v2 << endl;
	v2 = v2 | (0x00000000ffffffff & v0);
	cout << hex << "v2 = " << v2 << endl;

	char* ss = "totoeoe";

	cout << "len = " << strlen(ss) << endl;

	string cs;
	int length = cs.copy(ss, strlen(ss), 0);
	cs[length] = '\0';

	cout << "cs = " << cs << endl;

	double dp = 8000.9135875454321;

	cout << to<string>(dp) << endl;

	double d1 = 1235.15654;
	cout << "1235.15654 = " << to<string>(d1) << ", " << convertFloat(d1) << endl;

	double d2 = 123515.654;
	cout << "123515.654 = " << to<string>(d2) << ", " << convertFloat(d2)  << endl;

	double d3 = 12351565.4;
	cout << "12351565.4 = " << to<string>(d3) << ", " << convertFloat(d3)  << endl;

	double d4 = 0.00012351565;
	cout << "0.00012351565 = " << to<string>(d4) << ", " << convertFloat(d4)  << endl;

	double d5 = 0.0012351565;
	cout << "0.0012351565 = " << to<string>(d5) << ", " << convertFloat(d5)  << endl;

	double d6 = 0.00092351565;
	cout << "0.00092351565 = " << to<string>(d6) << ", " << convertFloat(d6)  << endl;

	cout << endl;

	float fd1 = 1235.15654;
	cout << "1235.15654 = " << to<string>(fd1) << ", " << convertFloat(fd1) << endl;

	float fd2 = 123515.654;
	cout << "123515.654 = " << to<string>(fd2) << ", " << convertFloat(fd2)  << endl;

	float fd3 = 12351565.4;
	cout << "12351565.4 = " << to<string>(fd3) << ", " << convertFloat(fd3)  << endl;

	float fd4 = 0.00012351565;
	cout << "0.00012351565 = " << to<string>(fd4) << ", " << convertFloat(fd4)  << endl;

	float fd5 = 0.0012351565;
	cout << "0.0012351565 = " << to<string>(fd5) << ", " << convertFloat(fd5)  << endl;

	float fd6 = 0.00092351565;
	cout << "0.00092351565 = " << to<string>(fd6) << ", " << convertFloat(fd6)  << endl;

	s = "123";
	cout << s << " = " << to<int>(s) << endl;

	s = "123.265";
	cout << s << " = " << to<float>(s) << endl;

	s = "1a23";
	cout << s << " = " << to<int>(s) << endl;
*/
	dv = 3234.6767823;
	cout << "formatted " << (boost::format("%E") % dv) << endl;
}

BOOST_AUTO_TEST_SUITE_END();
